Package: RoBMA
Title: Robust Bayesian Meta-Analyses
Version: 4.0.0
Maintainer: František Bartoš <f.bartos96@gmail.com>
Authors@R: c( 
    person("František", "Bartoš",     role = c("aut", "cre"),
    email   = "f.bartos96@gmail.com", comment = c(ORCID = "0000-0002-0018-5573")),
    person("Maximilian", "Maier",     role = "aut",
    email   = "maximilianmaier0401@gmail.com", comment = c(ORCID = "0000-0002-9873-6096")),
    person("Eric-Jan", "Wagenmakers", role = "ths",
    comment = c(ORCID = "0000-0003-1596-1034")),
    person("Joris", "Goosen",         role = "ctb"),
    person("Matthew", "Denwood", role="cph", 
           comment="Original copyright holder of some modified code where indicated."),
    person("Martyn", "Plummer", role="cph", 
           comment="Original copyright holder of some modified code where indicated.")
    )
Description: A framework for Bayesian meta-analysis, including model estimation,
      prior specification, model comparison, prediction, summaries, visualizations,
      and diagnostics. The package fits single and model-averaged meta-analytic,
      meta-regression, multilevel, publication bias adjusted, and generalized linear 
      mixed models The model-averaged meta-analytic models combine competing models 
      based on their predictive performance, weight inference by posterior model probabilities, 
      and test model components using Bayes factors (e.g., effect vs. no effect; 
      Bartoš et al., 2022, <doi:10.1002/jrsm.1594>; Maier, Bartoš & Wagenmakers, 2022,
      <doi:10.1037/met0000405>; Bartoš et al., 2025, <doi:10.1037/met0000737>).
      Users can specify flexible prior distributions for effect sizes, heterogeneity,
      publication bias (including selection models and PET-PEESE), and moderators.
URL: https://fbartos.github.io/RoBMA/
BugReports: https://github.com/FBartos/RoBMA/issues
License: GPL-3
Encoding: UTF-8
LazyData: true
RoxygenNote: 7.3.3
SystemRequirements: JAGS >= 4.3.1 (https://mcmc-jags.sourceforge.io/)
NeedsCompilation: yes
Depends: R (>= 4.0.0)
Imports: BayesTools (>= 0.3.0), bridgesampling, loo, MASS, parallel,
        runjags, rjags, stats, graphics, mvtnorm, scales, Rdpack,
        rlang, coda, ggplot2
Suggests: metafor, posterior, weightr, lme4, fixest, metaBMA, emmeans,
        metadat, testthat, vdiffr, knitr, rmarkdown, covr
LinkingTo: mvtnorm
RdMacros: Rdpack
VignetteBuilder: knitr
Packaged: 2026-05-07 09:33:12 UTC; fbart
Author: František Bartoš [aut, cre] (ORCID:
    <https://orcid.org/0000-0002-0018-5573>),
  Maximilian Maier [aut] (ORCID: <https://orcid.org/0000-0002-9873-6096>),
  Eric-Jan Wagenmakers [ths] (ORCID:
    <https://orcid.org/0000-0003-1596-1034>),
  Joris Goosen [ctb],
  Matthew Denwood [cph] (Original copyright holder of some modified code
    where indicated.),
  Martyn Plummer [cph] (Original copyright holder of some modified code
    where indicated.)
Repository: CRAN
Date/Publication: 2026-05-07 13:31:05 UTC
Built: R 4.6.0; x86_64-apple-darwin20; 2026-05-07 14:15:46 UTC; unix
Archs: RoBMA.so.dSYM

# Generated by roxygen2: do not edit by hand

S3method(add_loo,brma)
S3method(add_marglik,brma)
S3method(add_waic,brma)
S3method(as.matrix,brma_samples)
S3method(as_draws,brma)
S3method(as_draws,brma_samples)
S3method(as_draws,default)
S3method(as_draws_array,brma)
S3method(as_draws_array,brma_samples)
S3method(as_draws_array,default)
S3method(as_draws_df,brma)
S3method(as_draws_df,brma_samples)
S3method(as_draws_df,default)
S3method(as_draws_list,brma)
S3method(as_draws_list,brma_samples)
S3method(as_draws_list,default)
S3method(as_draws_matrix,brma)
S3method(as_draws_matrix,brma_samples)
S3method(as_draws_matrix,default)
S3method(as_draws_rvars,brma)
S3method(as_draws_rvars,brma_samples)
S3method(as_draws_rvars,default)
S3method(as_zplot,brma)
S3method(bayes_factor,brma)
S3method(bf,brma)
S3method(blup,brma)
S3method(bridge_sampler,brma)
S3method(bridgesampling::bayes_factor,brma)
S3method(bridgesampling::bf,brma)
S3method(bridgesampling::bridge_sampler,brma)
S3method(bridgesampling::logml,brma)
S3method(bridgesampling::post_prob,brma)
S3method(check_loo,brma)
S3method(coef,brma)
S3method(cooks.distance,brma)
S3method(covratio,brma)
S3method(dfbetas,brma)
S3method(dffits,brma)
S3method(fitted,brma)
S3method(funnel,brma)
S3method(galbraith,brma)
S3method(hatvalues,brma)
S3method(hist,zplot_brma)
S3method(influence,brma)
S3method(interpret,brma)
S3method(interpret,default)
S3method(lines,zplot_brma)
S3method(logLik,brma)
S3method(logml,brma)
S3method(loo,brma)
S3method(loo::waic,brma)
S3method(loo_compare,brma)
S3method(loo_compare,loo)
S3method(loo_weights,brma)
S3method(marginal_means,brma)
S3method(metafor::funnel,brma)
S3method(metafor::regplot,brma)
S3method(nobs,brma)
S3method(plot,brma)
S3method(plot,marginal_means.brma)
S3method(plot,only_priors.brma)
S3method(plot,zplot_brma)
S3method(plot_diagnostic,brma)
S3method(plot_diagnostic_autocorrelation,brma)
S3method(plot_diagnostic_density,brma)
S3method(plot_diagnostic_trace,brma)
S3method(plot_pet_peese,brma)
S3method(plot_prior,brma)
S3method(plot_prior,prior)
S3method(plot_weightfunction,brma)
S3method(pooled_effect,brma)
S3method(pooled_heterogeneity,brma)
S3method(post_prob,brma)
S3method(predict,brma)
S3method(print,RoBMA_data)
S3method(print,brma)
S3method(print,brma_samples)
S3method(print,covratio.brma)
S3method(print,dfbetas.brma)
S3method(print,infl.brma)
S3method(print,interpret.brma)
S3method(print,logLik.brma)
S3method(print,marginal_means.brma)
S3method(print,only_priors.brma)
S3method(print,summary.brma)
S3method(print,summary.brma_samples)
S3method(print,summary.marginal_means.brma)
S3method(print,summary.zplot_brma)
S3method(print,summary_heterogeneity.brma)
S3method(print,summary_models.RoBMA)
S3method(print,vif.brma)
S3method(print,zplot_brma)
S3method(print_prior,brma)
S3method(print_prior,prior)
S3method(qqnorm,brma)
S3method(radial,brma)
S3method(ranef,brma)
S3method(regplot,brma)
S3method(residuals,brma)
S3method(rstandard,brma)
S3method(rstudent,brma)
S3method(summary,brma)
S3method(summary,brma_samples)
S3method(summary,marginal_means.brma)
S3method(summary,zplot_brma)
S3method(summary_heterogeneity,brma)
S3method(summary_models,RoBMA)
S3method(summary_models,default)
S3method(true_effects,brma)
S3method(update,brma)
S3method(vif,brma)
S3method(waic,brma)
S3method(zplot,brma)
S3method(zplot,zplot_brma)
export(BMA)
export(BMA.glmm)
export(BMA.norm)
export(RoBMA)
export(RoBMA.get_option)
export(RoBMA.options)
export(add_loo)
export(add_marglik)
export(add_waic)
export(as_draws)
export(as_draws_array)
export(as_draws_df)
export(as_draws_list)
export(as_draws_matrix)
export(as_draws_rvars)
export(as_zplot)
export(bPEESE)
export(bPET)
export(bayes_factor)
export(bf)
export(blup)
export(bridge_sampler)
export(brma)
export(brma.glmm)
export(brma.norm)
export(bselmodel)
export(check_loo)
export(contr.independent)
export(contr.meandif)
export(contr.orthonormal)
export(covratio)
export(dffits)
export(estimate_unit_information_sd)
export(funnel)
export(galbraith)
export(interpret)
export(logml)
export(loo)
export(loo_compare)
export(loo_weights)
export(marginal_means)
export(plot_diagnostic)
export(plot_diagnostic_autocorrelation)
export(plot_diagnostic_density)
export(plot_diagnostic_trace)
export(plot_pet_peese)
export(plot_prior)
export(plot_weightfunction)
export(pooled_effect)
export(pooled_heterogeneity)
export(post_prob)
export(print_prior)
export(prior)
export(prior_PEESE)
export(prior_PET)
export(prior_factor)
export(prior_informed)
export(prior_none)
export(prior_weightfunction)
export(radial)
export(ranef)
export(regplot)
export(set_autofit_control)
export(set_convergence_checks)
export(summary_heterogeneity)
export(summary_models)
export(true_effects)
export(vif)
export(waic)
export(wf_cumulative)
export(wf_fixed)
export(wf_independent)
export(zplot)
importFrom(BayesTools,is.prior)
importFrom(BayesTools,is.prior.PEESE)
importFrom(BayesTools,is.prior.PET)
importFrom(BayesTools,is.prior.factor)
importFrom(BayesTools,is.prior.independent)
importFrom(BayesTools,is.prior.mixture)
importFrom(BayesTools,is.prior.none)
importFrom(BayesTools,is.prior.point)
importFrom(BayesTools,is.prior.simple)
importFrom(BayesTools,is.prior.spike_and_slab)
importFrom(BayesTools,is.prior.weightfunction)
importFrom(Rdpack,reprompt)
importFrom(bridgesampling,bayes_factor)
importFrom(bridgesampling,bf)
importFrom(bridgesampling,bridge_sampler)
importFrom(bridgesampling,logml)
importFrom(bridgesampling,post_prob)
importFrom(graphics,hist)
importFrom(graphics,lines)
importFrom(rlang,.data)
importFrom(stats,coef)
importFrom(stats,cooks.distance)
importFrom(stats,dfbetas)
importFrom(stats,fitted)
importFrom(stats,hatvalues)
importFrom(stats,influence)
importFrom(stats,logLik)
importFrom(stats,model.matrix)
importFrom(stats,nobs)
importFrom(stats,plogis)
importFrom(stats,predict)
importFrom(stats,qlogis)
importFrom(stats,qqnorm)
importFrom(stats,residuals)
importFrom(stats,rstandard)
importFrom(stats,rstudent)
importFrom(stats,terms)
importFrom(stats,update)
importFrom(stats,vcov)
importFrom(utils,capture.output)
importFrom(utils,getS3method)

README
================

<!-- README.md is generated from README.Rmd. Please edit that file -->

<!-- badges: start -->

[![R-CMD-check](https://github.com/FBartos/RoBMA/actions/workflows/R-CMD-check.yaml/badge.svg)](https://github.com/FBartos/RoBMA/actions/workflows/R-CMD-check.yaml)
[![R-unit-tests](https://github.com/FBartos/RoBMA/actions/workflows/R-CMD-tests.yaml/badge.svg)](https://github.com/FBartos/RoBMA/actions/workflows/R-CMD-tests.yaml)
[![test-coverage](https://github.com/FBartos/RoBMA/actions/workflows/test-coverage.yaml/badge.svg)](https://github.com/FBartos/RoBMA/actions/workflows/test-coverage.yaml)
[![Codecov](https://codecov.io/gh/FBartos/RoBMA/graph/badge.svg)](https://app.codecov.io/gh/FBartos/RoBMA)
[![CRAN
status](https://www.r-pkg.org/badges/version/RoBMA)](https://CRAN.R-project.org/package=RoBMA)
<!-- badges: end -->

# Robust Bayesian Meta-Analysis (RoBMA)

The `RoBMA` R package fits a comprehensive collection of Bayesian
meta-analytic models (Bartoš & Maier, 2020). It provides single-model
fits, meta-regression, location-scale models, multilevel models, fully
Bayesian model-averaged ensembles, and built-in publication-bias
adjustment using selection models, PET-PEESE, and Robust Bayesian
Meta-Analysis (RoBMA). Posterior sampling is performed by JAGS through a
custom C++ module that implements the weighted distributions used in
selection models. Prior handling, plotting, and Bayesian diagnostics are
delegated to the companion `BayesTools` package.

This introduction maps the package at a high level and points to the
more detailed vignettes that follow.

## System Requirements

The `RoBMA` R package requires JAGS 4.3.1 or newer (Plummer, 2003).
Install JAGS from <https://mcmc-jags.sourceforge.io/> before installing
the R package.

``` r
install.packages("RoBMA")
```

**Note on backwards compatibility.** The 4.0 release is not backwards
compatible with earlier versions of the package. The *Manuscript
Companions* section accompanies the methodological papers in which the
`RoBMA` R package was originally developed and illustrated; the
vignettes there have been updated to the 4.0 version of the package, and
the numerical results differ from the published values because the
underlying algorithms have changed. To reproduce the published analyses
exactly, install `RoBMA` version 3.6.1 together with `BayesTools`
version 0.2.23:

``` r
remotes::install_version("BayesTools", version = "0.2.23")
remotes::install_version("RoBMA",      version = "3.6.1")
```

## Main Functions

The package exposes a small set of high-level fitting functions.

| Function | What it fits |
|:---|:---|
| [`brma()`](man/brma.Rd), [`brma.norm()`](man/brma.Rd) | Bayesian random-effects meta-analysis |
| [`brma.glmm()`](man/brma.glmm.Rd) | Bayesian GLMM meta-analysis (binomial / log OR or Poisson / log IRR) |
| [`bselmodel()`](man/bselmodel.Rd) | Bayesian weight-function selection model |
| [`bPET()`](man/bPET.Rd), [`bPEESE()`](man/bPEESE.Rd) | Bayesian PET / PEESE publication-bias adjustment |
| [`BMA()`](man/BMA.Rd), [`BMA.norm()`](man/BMA.Rd) | Bayesian model-averaging across presence / absence of effect and heterogeneity (no bias adjustment) |
| [`BMA.glmm()`](man/BMA.glmm.Rd) | Bayesian model-averaging for GLMM meta-analysis |
| [`RoBMA()`](man/RoBMA.Rd) | Robust Bayesian model-averaging including publication-bias models |

`brma()` and `RoBMA()` are the main workhorses. `brma()` fits a single
model in the same spirit as `metafor::rma()`; `RoBMA()` averages across
an ensemble that includes publication-bias adjustment. `BMA()` is
`RoBMA()` without the bias-adjustment models.

All fitting functions

- share an effect-size interface (`yi`, `sei` / `vi`, `measure`
  arguments; apart from glmm),
- can be extended into a meta-regression (`mods` argument),
- can be extended into a location-scale model (`scale` argument),
- can be extended into a multilevel model (`cluster` argument),
- and all produce objects that work with the same set of inference
  helpers (`summary()`, `plot()`, `predict()`, `loo()`, `funnel()`,
  `regplot()`, residuals and influence diagnostics, …).

## Vignette Map

The vignettes are organized into four sections.

### Foundations

| Vignette | Topic |
|:---|:---|
| [*Introduction to RoBMA*](vignettes/v00-introduction.Rmd) | Package overview, fitting functions, and vignette map |
| [*Prior Distributions*](vignettes/v01-prior-distributions.Rmd) | Default, informed, and custom prior distributions; rescaling guidance |
| [*Bayesian Meta-Analysis*](vignettes/v02-bayesian-meta-analysis.Rmd) | Random-effects meta-analysis with `brma()`, compared with `metafor::rma()` |
| [*Feature Coverage*](vignettes/v03-feature-coverage.Rmd) | Overview of available functionality across model families |

The foundations section gives the package overview, prior specification,
a baseline `brma()` workflow, and current feature coverage. The
[*Bayesian Meta-Analysis*](vignettes/v02-bayesian-meta-analysis.Rmd)
vignette walks through the BCG-vaccine example from the `metafor`
package in `brma()`, covering summaries, meta-regression, marginal
means, residuals, influence, LOO, and the standard meta-analytic plots.

### Correspondence with `metafor`

| Vignette | Topic |
|:---|:---|
| [*Multilevel Meta-Analysis*](vignettes/v10-metafor-parity-multilevel.Rmd) | `brma()` with `cluster` and `metafor::rma.mv()` for 3-level data |
| [*Publication-Bias Adjustment*](vignettes/v11-metafor-parity-publication-bias.Rmd) | `bselmodel()` / `bPET()` / `bPEESE()` and `metafor::selmodel()` |
| [*Location-Scale Models*](vignettes/v12-metafor-parity-location-scale.Rmd) | `brma()` with `scale` and `metafor::rma.ls()` |
| [*Generalized Linear Mixed-Effects Meta-Analysis*](vignettes/v13-metafor-parity-glmm.Rmd) | `brma.glmm()` and `metafor::rma.glmm()` for binomial and Poisson outcomes |

Each vignette starts from a `metafor` package analysis and shows the
matching `RoBMA` R package syntax, output, and diagnostics.

### Bayesian Model Averaging

| Vignette | Topic |
|:---|:---|
| [*Bayesian Model Averaging*](vignettes/v20-bayesian-model-averaging.Rmd) | Accounting for model uncertainty across presence and absence of effect and heterogeneity, with posterior model probabilities and inclusion Bayes factors |
| [*Robust Bayesian Meta-Analysis*](vignettes/v21-robust-bayesian-meta-analysis.Rmd) | Extending the ensemble to publication-bias models used in `RoBMA()`, with `PSMA`, `PP`, and bespoke ensemble specifications |

The first vignette introduces model averaging on an ensemble of models
with and without effect and heterogeneity. The second extends the
ensemble to publication-bias models used in `RoBMA()`.

### Manuscript Companions

| Vignette | Companion paper / dataset |
|:---|:---|
| [*Adjusting for Publication Bias Tutorial*](vignettes/v30-tutorial.Rmd) | Bartoš et al. (2022), JASP / R tutorial on `Lui2015` |
| [*Robust Bayesian Meta-Regression*](vignettes/v31-robma-metaregression.Rmd) | Bartoš et al. (2025), RoBMA-reg on `Andrews2021` |
| [*Multilevel Robust Bayesian Meta-Analysis*](vignettes/v32-robma-multilevel.Rmd) | Bartoš et al. (2026), Multilevel RoBMA on `Johnides2025` |
| [*Multilevel Robust Bayesian Meta-Regression*](vignettes/v33-robma-multilevel-metaregression.Rmd) | Multilevel RoBMA-reg on `Kroupova2021` |
| [*Informed Bayesian Meta-Analysis in Medicine*](vignettes/v34-bma-norm-medicine.Rmd) | Bartoš et al. (2021), informed prior distributions for medical meta-analysis (continuous outcomes) |
| [*Informed Bayesian Meta-Analysis with Binary Outcomes*](vignettes/v35-bma-glmm-medicine.Rmd) | Bartoš et al. (2023), informed prior distributions for medical meta-analysis (binary and time-to-event outcomes) |
| [*Zplot Publication-Bias Diagnostics*](vignettes/v36-zplot.Rmd) | Bartoš & Schimmack (2025), zplot diagnostics on `Hoppen2025` |

These vignettes reproduce or update analyses from published papers and
serve as references when citing the corresponding methodological work.

## Where to Start

- For readers familiar with the `metafor` package: read [*Bayesian
  Meta-Analysis*](vignettes/v02-bayesian-meta-analysis.Rmd), then the
  *Correspondence with `metafor`* vignettes for the model family they
  use most.
- For readers new to Bayesian meta-analysis: read [*Prior
  Distributions*](vignettes/v01-prior-distributions.Rmd) and then
  [*Bayesian Meta-Analysis*](vignettes/v02-bayesian-meta-analysis.Rmd).
- For publication-bias adjustment: read [*Robust Bayesian
  Meta-Analysis*](vignettes/v21-robust-bayesian-meta-analysis.Rmd), with
  [*Adjusting for Publication Bias
  Tutorial*](vignettes/v30-tutorial.Rmd) as a worked example.

## References

<div id="refs" class="references csl-bib-body hanging-indent"
entry-spacing="0" line-spacing="2">

<div id="ref-bartos2021bayesian" class="csl-entry">

Bartoš, F., Gronau, Q. F., Timmers, B., Otte, W. M., Ly, A., &
Wagenmakers, E.-J. (2021). Bayesian model-averaged meta-analysis in
medicine. *Statistics in Medicine*, *40*(30), 6743–6761.
<https://doi.org/10.1002/sim.9170>

</div>

<div id="ref-RoBMA" class="csl-entry">

Bartoš, F., & Maier, M. (2020). *RoBMA: An R package for robust Bayesian
meta-analyses*. <https://CRAN.R-project.org/package=RoBMA>

</div>

<div id="ref-bartos2020adjusting" class="csl-entry">

Bartoš, F., Maier, M., Quintana, D. S., & Wagenmakers, E.-J. (2022).
Adjusting for publication bias in JASP and R — Selection models,
PET-PEESE, and robust Bayesian meta-analysis. *Advances in Methods and
Practices in Psychological Science*, *5*(3), 1–19.
<https://doi.org/10.1177/25152459221109259>

</div>

<div id="ref-bartos2023robust" class="csl-entry">

Bartoš, F., Maier, M., Stanley, T., & Wagenmakers, E.-J. (2025). Robust
Bayesian meta-regression: Model-averaged moderation analysis in the
presence of publication bias. *Psychological Methods*.
<https://doi.org/10.1037/met0000737>

</div>

<div id="ref-bartos2025robust" class="csl-entry">

Bartoš, F., Maier, M., & Wagenmakers, E.-J. (2026). Robust Bayesian
multilevel meta-analysis: Adjusting for publication bias in the presence
of dependent effect sizes. *Behavior Research Methods*.
<https://doi.org/10.31234/osf.io/9tgp2_v1>

</div>

<div id="ref-bartos2023empirical" class="csl-entry">

Bartoš, F., Otte, W. M., Gronau, Q. F., Timmers, B., Ly, A., &
Wagenmakers, E.-J. (2023). *Empirical prior distributions for Bayesian
meta-analyses of binary and time-to-event outcomes*.
<https://doi.org/10.48550/arXiv.2306.11468>

</div>

<div id="ref-bartos2025zcurve" class="csl-entry">

Bartoš, F., & Schimmack, U. (2025). Zplot: A visual diagnostic for
publication bias in meta-analysis. In *arXiv*.
<https://doi.org/10.48550/arXiv.2509.07171>

</div>

<div id="ref-plummer2003jags" class="csl-entry">

Plummer, M. (2003). JAGS: A program for analysis of Bayesian graphical
models using Gibbs sampling. *Proceedings of the 3rd International
Workshop on Distributed Statistical Computing (DSC 2003)*, 1–10.

</div>

</div>

## ----setup, include = FALSE---------------------------------------------------
is_check <- ("CheckExEnv" %in% search()) ||
              any(c("_R_CHECK_TIMINGS_", "_R_CHECK_LICENSE_") %in% names(Sys.getenv())) ||
              !file.exists("../models/MedicineBMA/fit_BMA.RDS")
knitr::opts_chunk$set(
  collapse  = TRUE,
  comment   = "#>",
  eval      = !is_check,
  dev      = "png")
if(.Platform$OS.type == "windows"){
  knitr::opts_chunk$set(dev.args = list(type = "cairo"))
}

## ----include = FALSE----------------------------------------------------------
library(RoBMA)
# we pre-load the RoBMA models, the fitting time is around 2-5 minutes
fit_BMA   <- readRDS(file = "../models/MedicineBMA/fit_BMA.RDS")
fit_BMAb  <- readRDS(file = "../models/MedicineBMA/fit_BMAb.RDS")
fit_RoBMA <- readRDS(file = "../models/MedicineBMA/fit_RoBMA.RDS")

## ----include = FALSE, eval = FALSE--------------------------------------------
# # R package version updating
# library(RoBMA)
# 
# data("Poulsen2006", package = "RoBMA")
# 
# fit_BMA <- RoBMA(d = Poulsen2006$d, se = Poulsen2006$se, study_names = Poulsen2006$study,
#                  priors_effect        = prior_informed(name = "oral health", parameter = "effect", type = "smd"),
#                  priors_heterogeneity = prior_informed(name = "oral health", parameter = "heterogeneity", type = "smd"),
#                  priors_bias          = NULL,
#                  transformation = "cohens_d", seed = 1, parallel = TRUE)
# 
# fit_BMAb <- RoBMA(d = Poulsen2006$d, se = Poulsen2006$se, study_names = Poulsen2006$study,
#                   priors_effect        = prior_informed(name = "oral health", parameter = "effect", type = "smd"),
#                   priors_heterogeneity = prior_informed(name = "oral health", parameter = "heterogeneity", type = "smd"),
#                   priors_bias          = NULL,
#                   seed = 1, parallel = TRUE)
# 
# fit_RoBMA <- RoBMA(d = Poulsen2006$d, se = Poulsen2006$se, study_names = Poulsen2006$study,
#                    priors_effect        = prior_informed(name = "oral health", parameter = "effect", type = "smd"),
#                    priors_heterogeneity = prior_informed(name = "oral health", parameter = "heterogeneity", type = "smd"),
#                    seed = 1, parallel = TRUE)
# 
# saveRDS(fit_BMA, file = "../models/MedicineBMA/fit_BMA.RDS")
# saveRDS(fit_BMAb, file = "../models/MedicineBMA/fit_BMAb.RDS")
# saveRDS(fit_RoBMA, file = "../models/MedicineBMA/fit_RoBMA.RDS")

## -----------------------------------------------------------------------------
library(RoBMA)

data("Poulsen2006", package = "RoBMA")
Poulsen2006

## -----------------------------------------------------------------------------
summary(fit_BMA, conditional = TRUE)

## ----fig_mu_BMA, dpi = 300, fig.width = 6, fig.height = 4, out.width = "75%", fig.align = "center"----
plot(fit_BMA, parameter = "mu", prior = TRUE)

## ----fig_mu_BMA_cond, dpi = 300, fig.width = 6, fig.height = 4, out.width = "75%", fig.align = "center"----
plot(fit_BMA, parameter = "mu", prior = TRUE, conditional = TRUE)

## ----fig_models, dpi = 300, fig.height = 4.5, fig.width = 7, out.width = '75%', fig.align = "center"----
plot_models(fit_BMA)

## ----fig_forest, dpi = 300, fig.height = 4.5, fig.width = 7, out.width = '75%', fig.align = "center"----
forest(fit_BMA, conditional = TRUE)

## -----------------------------------------------------------------------------
summary(fit_BMAb, conditional = TRUE)

## -----------------------------------------------------------------------------
summary(fit_RoBMA, conditional = TRUE)

## ----fig_mu_RoBMA_cond, dpi = 300, fig.width = 6, fig.height = 4, out.width = "75%", fig.align = "center"----
plot(fit_RoBMA, parameter = "mu", prior = TRUE, conditional = TRUE)

## ----setup, include = FALSE---------------------------------------------------
is_check <- ("CheckExEnv" %in% search()) ||
             any(c("_R_CHECK_TIMINGS_", "_R_CHECK_LICENSE_") %in% names(Sys.getenv())) ||
             !file.exists("../models/CustomEnsembles/Bem_update1.RDS") 
knitr::opts_chunk$set(
  collapse = TRUE,
  comment  = "#>",
  eval     = !is_check,
  dev      = "png"
)
if(.Platform$OS.type == "windows"){
  knitr::opts_chunk$set(dev.args = list(type = "cairo"))
}

## ----include = FALSE, eval = FALSE--------------------------------------------
# # R package version updating
# library(RoBMA)
# 
# data("Bem2011", package = "RoBMA")
# fit <- RoBMA(d = Bem2011$d, se = Bem2011$se, study_names = Bem2011$study,
#              priors_effect = NULL, priors_heterogeneity = NULL, priors_bias = NULL,
#              priors_effect_null        = prior("spike", parameters = list(location = 0)),
#              priors_heterogeneity_null = prior("spike", parameters = list(location = 0)),
#              priors_bias_null          = prior_none(),
#              seed = 1)
# 
# fit <- update(fit,
#               prior_effect             = prior("normal", parameters = list(mean = .15, sd = .10), truncation = list(lower = 0)),
#               prior_heterogeneity_null = prior("spike",  parameters = list(location = 0)),
#               prior_bias_null          = prior_none())
# saveRDS(fit, file = "../models/CustomEnsembles/Bem_update1.RDS")
# 
# fit <- update(fit,
#               prior_effect        = prior("normal", parameters = list(mean = .15, sd = .10), truncation = list(lower = 0)),
#               prior_heterogeneity = prior("invgamma", parameters = list(shape = 1, scale = .15)),
#               prior_bias_null     = prior_none())
# fit <- update(fit,
#               prior_effect_null        = prior("spike",     parameters = list(location = 0)),
#               prior_heterogeneity_null = prior("spike",     parameters = list(location = 0)),
#               prior_bias               = prior_weightfunction("one.sided", parameters = list(alpha = c(1, 1), steps = c(0.05))))
# fit <- update(fit,
#               prior_effect_null        = prior("spike",     parameters = list(location = 0)),
#               prior_heterogeneity_null = prior("spike",     parameters = list(location = 0)),
#               prior_bias               = prior_weightfunction("one.sided", parameters = list(alpha = c(1, 1, 1), steps = c(0.05, 0.10))))
# fit <- update(fit,
#               prior_effect_null        = prior("spike",     parameters = list(location = 0)),
#               prior_heterogeneity_null = prior("spike",     parameters = list(location = 0)),
#               prior_bias               = prior_PET("Cauchy", parameters = list(0, 1),  truncation = list(lower = 0)))
# fit <- update(fit,
#               prior_effect_null        = prior("spike",     parameters = list(location = 0)),
#               prior_heterogeneity_null = prior("spike",     parameters = list(location = 0)),
#               prior_bias               = prior_PEESE("Cauchy", parameters = list(0, 5),  truncation = list(lower = 0)))
# saveRDS(fit, file = "../models/CustomEnsembles/Bem_update2.RDS")

## -----------------------------------------------------------------------------
library(RoBMA)

data("Bem2011", package = "RoBMA")
Bem2011

## -----------------------------------------------------------------------------
fit <- RoBMA(d = Bem2011$d, se = Bem2011$se, study_names = Bem2011$study,
             priors_effect = NULL, priors_heterogeneity = NULL, priors_bias = NULL,
             priors_effect_null        = prior("spike", parameters = list(location = 0)),
             priors_heterogeneity_null = prior("spike", parameters = list(location = 0)),
             priors_bias_null          = prior_none(),
             seed = 1)

## -----------------------------------------------------------------------------
summary(fit, type = "models")

## ----fig_mu_prior, dpi = 300, fig.width = 4, fig.height = 3, out.width = "50%", fig.align = "center"----
plot(prior("normal", parameters = list(mean = .15, sd = .10), truncation = list(lower = 0)))

## ----include = FALSE----------------------------------------------------------
# these fits are relatively fast, but we reduce the knitting time considerably
fit <- readRDS(file = "../models/CustomEnsembles/Bem_update1.RDS")

## -----------------------------------------------------------------------------
summary(fit, type = "models")

## ----include = FALSE----------------------------------------------------------
fit <- readRDS(file = "../models/CustomEnsembles/Bem_update2.RDS")

## -----------------------------------------------------------------------------
summary(fit, type = "models")

## -----------------------------------------------------------------------------
summary(fit)

## ----fig_mu_posterior, dpi = 300, fig.width = 4, fig.height = 3.5, out.width = "50%", fig.align = "center"----
plot(fit, parameter = "mu", prior = TRUE)

## ----fig_weightfunction_posterior, dpi = 300, fig.width = 5, fig.height = 4, out.width = "75%", fig.align = "center"----
plot(fit, parameter = "weightfunction", prior = TRUE)

## ----fig_PETPEESE_posterior, dpi = 300, fig.width = 5, fig.height = 4, out.width = "75%", fig.align = "center"----
plot(fit, parameter = "PET-PEESE", prior = TRUE)

## ----setup, include = FALSE---------------------------------------------------
is_check <- ("CheckExEnv" %in% search()) ||
              any(c("_R_CHECK_TIMINGS_", "_R_CHECK_LICENSE_") %in% names(Sys.getenv())) ||
              !file.exists("../models/ReproducingBMA/BMA_PowerPoseTest.RDS")
knitr::opts_chunk$set(
  collapse  = TRUE,
  comment   = "#>",
  eval      = !is_check,
  dev      = "png")
if(.Platform$OS.type == "windows"){
  knitr::opts_chunk$set(dev.args = list(type = "cairo"))
}

## ----include = FALSE----------------------------------------------------------
library(RoBMA)
# we pre-load the RoBMA models, the fitting time is around 2-5 minutes
fit_BMA_test   <- readRDS(file = "../models/ReproducingBMA/BMA_PowerPoseTest.RDS")
fit_BMA_est    <- readRDS(file = "../models/ReproducingBMA/BMA_PowerPoseEst.RDS")
fit_RoBMA_test <- readRDS(file = "../models/ReproducingBMA/PowerPoseTest.RDS")
fit_RoBMA_est  <- readRDS(file = "../models/ReproducingBMA/PowerPoseEst.RDS")

## ----include = FALSE, eval = FALSE--------------------------------------------
# # R package version updating
# library(RoBMA)
# 
# data("power_pose", package = "metaBMA")
# 
# fit_RoBMA_test <- RoBMA(d = power_pose$effectSize, se = power_pose$SE, study_names = power_pose$study,
#                         priors_effect  = prior(
#                           distribution = "cauchy",
#                           parameters = list(location = 0, scale = 1/sqrt(2)),
#                           truncation = list(0, Inf)),
#                         priors_heterogeneity = prior(
#                           distribution = "invgamma",
#                           parameters = list(shape = 1, scale = 0.15)),
#                         priors_bias = NULL,
#                         transformation = "cohens_d", seed = 1, parallel = TRUE)
# 
# fit_RoBMA_est  <- RoBMA(d = power_pose$effectSize, se = power_pose$SE, study_names = power_pose$study,
#                         priors_effect  = prior(
#                           distribution = "cauchy",
#                           parameters = list(location = 0, scale = 1/sqrt(2))),
#                         priors_heterogeneity = prior(
#                           distribution = "invgamma",
#                           parameters = list(shape = 1, scale = 0.15)),
#                         priors_bias = NULL,
#                         priors_effect_null = NULL,
#                         transformation = "cohens_d", seed = 2, parallel = TRUE)
# 
# saveRDS(fit_RoBMA_test, file = "../models/ReproducingBMA/PowerPoseTest.RDS")
# saveRDS(fit_RoBMA_est, file = "../models/ReproducingBMA/PowerPoseEst.RDS")

## -----------------------------------------------------------------------------
data("power_pose", package = "metaBMA")
power_pose[,c("study", "effectSize", "SE")]

## -----------------------------------------------------------------------------
fit_BMA_test$inclusion

round(fit_BMA_est$estimates,2)

## -----------------------------------------------------------------------------
summary(fit_RoBMA_test)

summary(fit_RoBMA_est)

## ----fig_mu_est, dpi = 300, fig.width = 4, fig.height = 4, out.width = "50%", fig.align = "center"----
plot(fit_RoBMA_est, parameter = "mu", prior = TRUE, xlim = c(-1, 1))

## ----fig_mu_test, dpi = 300, fig.width = 4, fig.height = 4, out.width = "50%", fig.align = "center"----
plot(fit_RoBMA_test, parameter = "mu", prior = TRUE, xlim = c(-.5, 1))

## ----fig_models, dpi = 300, fig.height = 4.5, fig.width = 7, out.width = '75%', fig.align = "center"----
plot_models(fit_RoBMA_est)

## ----fig_forest, dpi = 300, fig.height = 4.5, fig.width = 7, out.width = '75%', fig.align = "center"----
forest(fit_RoBMA_est)

## ----setup, include = FALSE---------------------------------------------------
is_check <- ("CheckExEnv" %in% search()) ||
              any(c("_R_CHECK_TIMINGS_", "_R_CHECK_LICENSE_") %in% names(Sys.getenv())) ||
              !file.exists("../models/FastRoBMA/fit_RoBMA.bridge.RDS")
knitr::opts_chunk$set(
  collapse  = TRUE,
  comment   = "#>",
  eval      = !is_check,
  dev      = "png")
if(.Platform$OS.type == "windows"){
  knitr::opts_chunk$set(dev.args = list(type = "cairo"))
}

## ----include = FALSE----------------------------------------------------------
library(RoBMA)
# we pre-load the RoBMA models, the fitting time is around 2-5 minutes
fit_RoBMA.bridge  <- readRDS(file = "../models/FastRoBMA/fit_RoBMA.bridge.RDS")
fit_RoBMA.ss      <- readRDS(file = "../models/FastRoBMA/fit_RoBMA.ss.RDS")

## ----include = FALSE, eval = FALSE--------------------------------------------
# library(RoBMA)
# 
# t1_bridge <- Sys.time()
# fit_RoBMA.bridge    <- RoBMA.reg(
#   ~ measure + age, data = Andrews2021,
#   algorithm = "bridge",
#   chains = 3, sample = 5000, burnin = 2000, adapt = 500, parallel = TRUE
# )
# t2_bridge <- Sys.time()
# 
# t1_ss <- Sys.time()
# fit_RoBMA.ss  <- RoBMA.reg(
#   ~ measure + age, data = Andrews2021,
#   algorithm = "ss",
#   chains = 6, sample = 10000, burnin = 2500, adapt = 2500, parallel = TRUE
# )
# t2_ss <- Sys.time()
# 
# # save memory space
# fit_RoBMA.bridge <- RoBMA:::.remove_model_posteriors(fit_RoBMA.bridge)
# fit_RoBMA.bridge <- RoBMA:::.remove_model_margliks(fit_RoBMA.bridge)
# fit_RoBMA.ss$model$fit <- NULL
# 
# saveRDS(fit_RoBMA.bridge, file = "../models/FastRoBMA/fit_RoBMA.bridge.RDS",    compress = "xz")
# saveRDS(fit_RoBMA.ss,     file = "../models/FastRoBMA/fit_RoBMA.ss.RDS",        compress = "xz")

## -----------------------------------------------------------------------------
library(RoBMA)
data("Andrews2021", package = "RoBMA")
head(Andrews2021)

## -----------------------------------------------------------------------------
summary(fit_RoBMA.bridge)
summary(fit_RoBMA.ss)

## ----fig_effect, dpi = 300, fig.width = 12, fig.height = 6, out.width = "100%", fig.align = "center"----
par(mfrow = c(1, 2), mar = c(4, 4, 2, 5))
plot(fit_RoBMA.bridge, parameter = "mu", prior = TRUE, main = "Bridge Sampling")
plot(fit_RoBMA.ss,     parameter = "mu", prior = TRUE, main = "Spike-and-Slab")

## -----------------------------------------------------------------------------
marginal_summary(fit_RoBMA.bridge)
marginal_summary(fit_RoBMA.ss)

## ----fig_marginal, dpi = 300, fig.width = 12, fig.height = 6, out.width = "100%", fig.align = "center"----
par(mfrow = c(1, 2))
marginal_plot(fit_RoBMA.bridge, parameter = "measure", conditional = TRUE, prior = TRUE, xlim = c(-1, 1), ylim = c(0, 7), main = "Bridge Sampling")
marginal_plot(fit_RoBMA.ss,     parameter = "measure", conditional = TRUE, prior = TRUE, xlim = c(-1, 1), ylim = c(0, 7), main = "Spike-and-Slab")

## ----setup, include = FALSE---------------------------------------------------
is_check <- ("CheckExEnv" %in% search()) ||
              any(c("_R_CHECK_TIMINGS_", "_R_CHECK_LICENSE_") %in% names(Sys.getenv())) ||
              !file.exists("../models/MultilevelRoBMARegression/zfit_Havrankova2025.RDS")
knitr::opts_chunk$set(
  collapse  = TRUE,
  comment   = "#>",
  eval      = !is_check,
  dev       = "png")
if(.Platform$OS.type == "windows"){
  knitr::opts_chunk$set(dev.args = list(type = "cairo"))
}

## ----include = FALSE----------------------------------------------------------
library(RoBMA)
zfit_reg <- readRDS(file = "../models/MultilevelRoBMARegression/zfit_Havrankova2025.RDS")
fit_reg  <- zfit_reg
class(fit_reg) <- class(fit_reg)[!class(fit_reg) %in% "zcurve_RoBMA"]

## ----include = FALSE, eval = FALSE--------------------------------------------
# # R package version updating
# library(RoBMA)
# data("Havrankova2025", package = "RoBMA")
# 
# # Prior scaling
# fit_fe        <- metafor::rma(yi = y, sei = se, data = Havrankova2025, method = "FE")
# unti_scale    <- fit_fe$se * sqrt(sum(Havrankova2025$N))
# prior_scale   <- unti_scale * 0.5
# 
# df_reg  <- data.frame(
#   y               = Havrankova2025$y,
#   se              = Havrankova2025$se,
#   facing_customer = Havrankova2025$facing_customer,
#   study_id        = Havrankova2025$study_id
# )
# 
# fit_reg <- RoBMA.reg(
#   ~ facing_customer,
#   study_ids      = df_reg$study_id,
#   data           = df_reg,
#   rescale_priors = prior_scale,
#   prior_scale = "none", transformation = "none",
#   algorithm = "ss", sample = 20000, burnin = 10000, adapt = 10000,
#   thin = 5, parallel = TRUE, autofit = FALSE, seed = 1)
# )
# 
# zfit_reg    <- as_zcurve(fit_reg)
# saveRDS(zfit_reg, file = "../models/MultilevelRoBMARegression/zfit_Havrankova2025.RDS", compress = "xz")

## -----------------------------------------------------------------------------
library(RoBMA)
data("Havrankova2025", package = "RoBMA")

fit_fe        <- metafor::rma(yi = y, sei = se, data = Havrankova2025, method = "FE")
unti_scale    <- fit_fe$se * sqrt(sum(Havrankova2025$N))
prior_scale   <- unti_scale * 0.5

## ----eval = FALSE-------------------------------------------------------------
# df_reg  <- data.frame(
#   y               = Havrankova2025$y,
#   se              = Havrankova2025$se,
#   facing_customer = Havrankova2025$facing_customer,
#   study_id        = Havrankova2025$study_id
# )
# 
# fit_reg <- RoBMA.reg(
#   ~ facing_customer,
#   study_ids      = df_reg$study_id,
#   data           = df_reg,
#   rescale_priors = prior_scale,
#   prior_scale    = "none", transformation = "none",
#   algorithm = "ss", sample = 20000, burnin = 10000, adapt = 10000,
#   thin = 5, parallel = TRUE, autofit = FALSE, seed = 1)
# )

## -----------------------------------------------------------------------------
summary(fit_reg)

## -----------------------------------------------------------------------------
marginal_summary(fit_reg)

## ----eval = FALSE-------------------------------------------------------------
# zfit_reg <- as_zcurve(fit_reg)

## ----fig.width = 6, fig.height = 4--------------------------------------------
par(mar = c(4,4,0,0))
plot(zfit_reg, by.hist = 0.25, plot_extrapolation = FALSE, from = -4, to = 8)

## ----setup, include = FALSE---------------------------------------------------
is_check <- ("CheckExEnv" %in% search()) ||
              any(c("_R_CHECK_TIMINGS_", "_R_CHECK_LICENSE_") %in% names(Sys.getenv())) ||
              !file.exists("../models/zcurve/zcurve_RE_Hoppen2025.RDS")
knitr::opts_chunk$set(
  collapse  = TRUE,
  comment   = "#>",
  eval      = !is_check,
  dev      = "png",
  fig.width = 7,
  fig.height = 5)
if(.Platform$OS.type == "windows"){
  knitr::opts_chunk$set(dev.args = list(type = "cairo"))
}

## ----include = FALSE, eval = FALSE--------------------------------------------
# # R package version updating
# library(RoBMA)
# 
# # Ease of retrieval - Hoppen2025
# data("Weingarten2018", package = "RoBMA")
# Weingarten2018 <- Weingarten2018[Weingarten2018$standard_paradigm & Weingarten2018$proximal_dataset, ]
# fit_RE_Weingarten2018 <- NoBMA(r = Weingarten2018$r_xy, n = round(Weingarten2018$N), study_ids = Weingarten2018$paper_id,
#                            priors_effect_null = NULL, priors_heterogeneity_null = NULL,
#                            algorithm = "ss", sample = 10000, burnin = 10000, adapt = 10000,
#                            chains = 5, parallel = TRUE, seed = 1, save = "min")
# fit_RoBMA_Weingarten2018 <- RoBMA(r = Weingarten2018$r_xy, n = round(Weingarten2018$N), study_ids = Weingarten2018$paper_id,
#                               algorithm = "ss", sample = 10000, burnin = 10000, adapt = 10000,
#                               chains = 5, parallel = TRUE, seed = 1, save = "min")
# 
# zcurve_RE_Weingarten2018    <- as_zcurve(fit_RE_Weingarten2018)
# zcurve_RoBMA_Weingarten2018 <- as_zcurve(fit_RoBMA_Weingarten2018)
# 
# saveRDS(zcurve_RE_Weingarten2018,    file = "../models/zcurve/zcurve_RE_Weingarten2018.RDS", compress = "xz")
# saveRDS(zcurve_RoBMA_Weingarten2018, file = "../models/zcurve/zcurve_RoBMA_Weingarten2018.RDS", compress = "xz")
# 
# 
# # Social comparison example - Hoppen2025
# data("Hoppen2025", package = "RoBMA")
# fit_RE_Hoppen2025 <- NoBMA(d = Hoppen2025$d, se = sqrt(Hoppen2025$v),
#                            priors_effect_null = NULL, priors_heterogeneity_null = NULL,
#                            algorithm = "ss", sample = 10000, burnin = 5000, adapt = 5000,
#                            chains = 5, parallel = TRUE, seed = 1, save = "min")
# fit_RoBMA_Hoppen2025 <- RoBMA(d = Hoppen2025$d, se = sqrt(Hoppen2025$v),
#                               algorithm = "ss", sample = 10000, burnin = 5000, adapt = 5000,
#                               chains = 5, parallel = TRUE, seed = 1, save = "min")
# 
# zcurve_RE_Hoppen2025    <- as_zcurve(fit_RE_Hoppen2025)
# zcurve_RoBMA_Hoppen2025 <- as_zcurve(fit_RoBMA_Hoppen2025)
# 
# saveRDS(zcurve_RE_Hoppen2025,    file = "../models/zcurve/zcurve_RE_Hoppen2025.RDS", compress = "xz")
# saveRDS(zcurve_RoBMA_Hoppen2025, file = "../models/zcurve/zcurve_RoBMA_Hoppen2025.RDS", compress = "xz")
# 
# # ChatGPT example - Wang2025
# data("Wang2025", package = "RoBMA")
# Wang2025 <- Wang2025[Wang2025$Learning_effect == "Learning performance", ]
# fit_RE_Wang2025 <- NoBMA(d = Wang2025$g, se = Wang2025$se,
#                          priors_effect_null = NULL, priors_heterogeneity_null = NULL,
#                          algorithm = "ss", sample = 10000, burnin = 5000, adapt = 5000,
#                          chains = 5, parallel = TRUE, seed = 1, save = "min")
# fit_RoBMA_Wang2025 <- RoBMA(d = Wang2025$g, se = Wang2025$se,
#                             algorithm = "ss", sample = 10000, burnin = 5000, adapt = 5000,
#                             chains = 5, parallel = TRUE, seed = 1, save = "min")
# 
# zcurve_RE_Wang2025    <- as_zcurve(fit_RE_Wang2025)
# zcurve_RoBMA_Wang2025 <- as_zcurve(fit_RoBMA_Wang2025)
# 
# saveRDS(zcurve_RE_Wang2025,    file = "../models/zcurve/zcurve_RE_Wang2025.RDS", compress = "xz")
# saveRDS(zcurve_RoBMA_Wang2025, file = "../models/zcurve/zcurve_RoBMA_Wang2025.RDS", compress = "xz")
# 
# # Many Labs 2 example - ManyLabs16
# data("ManyLabs16", package = "RoBMA")
# fit_RE_ManyLabs16 <- NoBMA(d = ManyLabs16$y, se = ManyLabs16$se,
#                            priors_effect_null = NULL, priors_heterogeneity_null = NULL,
#                            algorithm = "ss", sample = 10000, burnin = 5000, adapt = 5000,
#                            chains = 5, parallel = TRUE, seed = 1, save = "min")
# fit_RoBMA_ManyLabs16 <- RoBMA(d = ManyLabs16$y, se = ManyLabs16$se,
#                               algorithm = "ss", sample = 10000, burnin = 5000, adapt = 5000,
#                               chains = 5, parallel = TRUE, seed = 1, save = "min")
# 
# zcurve_RE_ManyLabs16    <- as_zcurve(fit_RE_ManyLabs16)
# zcurve_RoBMA_ManyLabs16 <- as_zcurve(fit_RoBMA_ManyLabs16)
# 
# saveRDS(zcurve_RE_ManyLabs16,    file = "../models/zcurve/zcurve_RE_ManyLabs16.RDS", compress = "xz")
# saveRDS(zcurve_RoBMA_ManyLabs16, file = "../models/zcurve/zcurve_RoBMA_ManyLabs16.RDS", compress = "xz")

## ----include = FALSE----------------------------------------------------------
knitr::opts_chunk$set(echo = TRUE)

# preload the fitted models
zcurve_RE_Weingarten2018    <- readRDS(file = "../models/zcurve/zcurve_RE_Weingarten2018.RDS")
zcurve_RoBMA_Weingarten2018 <- readRDS(file = "../models/zcurve/zcurve_RoBMA_Weingarten2018.RDS")
zcurve_RE_Hoppen2025    <- readRDS(file = "../models/zcurve/zcurve_RE_Hoppen2025.RDS")
zcurve_RoBMA_Hoppen2025 <- readRDS(file = "../models/zcurve/zcurve_RoBMA_Hoppen2025.RDS")
zcurve_RE_Wang2025      <- readRDS(file = "../models/zcurve/zcurve_RE_Wang2025.RDS")
zcurve_RoBMA_Wang2025   <- readRDS(file = "../models/zcurve/zcurve_RoBMA_Wang2025.RDS")
zcurve_RE_ManyLabs16    <- readRDS(file = "../models/zcurve/zcurve_RE_ManyLabs16.RDS")
zcurve_RoBMA_ManyLabs16 <- readRDS(file = "../models/zcurve/zcurve_RoBMA_ManyLabs16.RDS")

fit_RE_Weingarten2018    <- zcurve_RE_Weingarten2018
fit_RoBMA_Weingarten2018 <- zcurve_RoBMA_Weingarten2018
fit_RE_Hoppen2025    <- zcurve_RE_Hoppen2025
fit_RoBMA_Hoppen2025 <- zcurve_RoBMA_Hoppen2025
fit_RE_Wang2025      <- zcurve_RE_Wang2025
fit_RoBMA_Wang2025   <- zcurve_RoBMA_Wang2025
fit_RE_ManyLabs16    <- zcurve_RE_ManyLabs16
fit_RoBMA_ManyLabs16 <- zcurve_RoBMA_ManyLabs16

class(fit_RE_Weingarten2018)    <- class(fit_RE_Weingarten2018)[!class(fit_RE_Weingarten2018)       %in% "zcurve_RoBMA"]
class(fit_RoBMA_Weingarten2018) <- class(fit_RoBMA_Weingarten2018)[!class(fit_RoBMA_Weingarten2018) %in% "zcurve_RoBMA"]
class(fit_RE_Hoppen2025)    <- class(fit_RE_Hoppen2025)[!class(fit_RE_Hoppen2025)       %in% "zcurve_RoBMA"]
class(fit_RoBMA_Hoppen2025) <- class(fit_RoBMA_Hoppen2025)[!class(fit_RoBMA_Hoppen2025) %in% "zcurve_RoBMA"]
class(fit_RE_Wang2025)      <- class(fit_RE_Wang2025)[!class(fit_RE_Wang2025)           %in% "zcurve_RoBMA"]
class(fit_RoBMA_Wang2025)   <- class(fit_RoBMA_Wang2025)[!class(fit_RoBMA_Wang2025)     %in% "zcurve_RoBMA"]
class(fit_RE_ManyLabs16)    <- class(fit_RE_ManyLabs16)[!class(fit_RE_ManyLabs16)       %in% "zcurve_RoBMA"]
class(fit_RoBMA_ManyLabs16) <- class(fit_RoBMA_ManyLabs16)[!class(fit_RoBMA_ManyLabs16) %in% "zcurve_RoBMA"]

## ----message = FALSE----------------------------------------------------------
library("RoBMA")

## -----------------------------------------------------------------------------
# Load the ease-of-retrieval dataset
data("Weingarten2018", package = "RoBMA")
# Filter to standard paradigm and proximal dataset as in the original analysis
Weingarten2018 <- Weingarten2018[Weingarten2018$standard_paradigm & Weingarten2018$proximal_dataset, ]
head(Weingarten2018)

## ----eval = FALSE-------------------------------------------------------------
# # Fit random-effects model (unadjusted for publication bias)
# fit_RE_Weingarten2018 <- NoBMA(r = Weingarten2018$r_xy, n = round(Weingarten2018$N),
#                                study_ids = Weingarten2018$paper_id,
#                                priors_effect_null = NULL, priors_heterogeneity_null = NULL,
#                                algorithm = "ss", sample = 10000, burnin = 10000, adapt = 10000,
#                                chains = 5, parallel = TRUE, seed = 1)
# 
# # Fit RoBMA model (adjusted for publication bias)
# fit_RoBMA_Weingarten2018 <- RoBMA(r = Weingarten2018$r_xy, n = round(Weingarten2018$N),
#                                   study_ids = Weingarten2018$paper_id,
#                                   algorithm = "ss", sample = 10000, burnin = 10000, adapt = 10000,
#                                   chains = 5, parallel = TRUE, seed = 1)

## -----------------------------------------------------------------------------
# Random-effects model results
summary(fit_RE_Weingarten2018, output_scale = "r")

# RoBMA model results  
summary(fit_RoBMA_Weingarten2018, output_scale = "r")

## ----eval = FALSE-------------------------------------------------------------
# # Create z-curve objects
# zcurve_RE_Weingarten2018    <- as_zcurve(fit_RE_Weingarten2018)
# zcurve_RoBMA_Weingarten2018 <- as_zcurve(fit_RoBMA_Weingarten2018)

## ----fig.cap="Ease-of-Retrieval Effect: Model Fit Assessment"-----------------
# Create histogram of observed z-statistics
hist(zcurve_RoBMA_Weingarten2018, from = -3, to = 6, by = 0.25)

# Add model-implied distributions
lines(zcurve_RE_Weingarten2018, from = -3, to = 6, col = "black", lty = 2, lwd = 2)
lines(zcurve_RoBMA_Weingarten2018, from = -3, to = 6, col = "blue", lty = 2, lwd = 2)

# Add legend
legend("topright", 
       legend = c("Random-Effects", "RoBMA"), 
       col = c("black", "blue"), 
       lty = 2, lwd = 2)

## ----fig.cap="Ease-of-Retrieval Effect: Extrapolation Analysis"---------------
# Plot extrapolation to pre-publication bias
plot(zcurve_RoBMA_Weingarten2018, from = -3, to = 6, by.hist = 0.25)

## -----------------------------------------------------------------------------
# Extract z-curve summary metrics
summary(zcurve_RoBMA_Weingarten2018)

## -----------------------------------------------------------------------------
# Load the social comparison dataset
data("Hoppen2025", package = "RoBMA")
head(Hoppen2025)

## ----eval = FALSE-------------------------------------------------------------
# # Fit random-effects model (unadjusted for publication bias)
# fit_RE_Hoppen2025 <- NoBMA(d = Hoppen2025$d, se = sqrt(Hoppen2025$v),
#                            priors_effect_null = NULL, priors_heterogeneity_null = NULL,
#                            algorithm = "ss", sample = 10000, burnin = 5000, adapt = 5000,
#                            chains = 5, parallel = TRUE, seed = 1)
# 
# # Fit RoBMA model (adjusted for publication bias)
# fit_RoBMA_Hoppen2025 <- RoBMA(d = Hoppen2025$d, se = sqrt(Hoppen2025$v),
#                               algorithm = "ss", sample = 10000, burnin = 5000, adapt = 5000,
#                               chains = 5, parallel = TRUE, seed = 1)

## -----------------------------------------------------------------------------
# Random-effects model results
summary(fit_RE_Hoppen2025)

## -----------------------------------------------------------------------------
# RoBMA model results  
summary(fit_RoBMA_Hoppen2025)

## ----eval = FALSE-------------------------------------------------------------
# # Create z-curve objects
# zcurve_RE_Hoppen2025    <- as_zcurve(fit_RE_Hoppen2025)
# zcurve_RoBMA_Hoppen2025 <- as_zcurve(fit_RoBMA_Hoppen2025)

## ----fig.cap="Social Comparison: Model Fit Assessment"------------------------
# Create histogram of observed z-statistics
hist(zcurve_RoBMA_Hoppen2025)

# Add model-implied distributions
lines(zcurve_RE_Hoppen2025, col = "black", lty = 2, lwd = 2)
lines(zcurve_RoBMA_Hoppen2025, col = "blue", lty = 2, lwd = 2)

# Add legend
legend("topright", 
       legend = c("Random-Effects", "RoBMA"), 
       col = c("black", "blue"), 
       lty = 2, lwd = 2)

## ----fig.cap="Social Comparison: Extrapolation Analysis"----------------------
# Plot extrapolation to pre-publication bias
plot(zcurve_RoBMA_Hoppen2025)

## -----------------------------------------------------------------------------
# Extract z-curve summary metrics
summary(zcurve_RoBMA_Hoppen2025)

## -----------------------------------------------------------------------------
# Load the ChatGPT dataset
data("Wang2025", package = "RoBMA")
# Select learning performance studies
Wang2025 <- Wang2025[Wang2025$Learning_effect == "Learning performance", ]
head(Wang2025)

## ----eval = FALSE-------------------------------------------------------------
# # Fit models
# fit_RE_Wang2025 <- NoBMA(d = Wang2025$g, se = Wang2025$se,
#                          priors_effect_null = NULL, priors_heterogeneity_null = NULL,
#                          algorithm = "ss", sample = 10000, burnin = 5000, adapt = 5000,
#                          chains = 5, parallel = TRUE, seed = 1)
# fit_RoBMA_Wang2025 <- RoBMA(d = Wang2025$g, se = Wang2025$se,
#                             algorithm = "ss", sample = 10000, burnin = 5000, adapt = 5000,
#                             chains = 5, parallel = TRUE, seed = 1)
# 
# # Create z-curve objects
# zcurve_RE_Wang2025    <- as_zcurve(fit_RE_Wang2025)
# zcurve_RoBMA_Wang2025 <- as_zcurve(fit_RoBMA_Wang2025)

## ----fig.cap="ChatGPT: Model Fit Assessment"----------------------------------
# Assess model fit
hist(zcurve_RoBMA_Wang2025, from = -2, to = 8)
lines(zcurve_RE_Wang2025, col = "black", lty = 2, lwd = 2, from = -2, to = 8)
lines(zcurve_RoBMA_Wang2025, col = "blue", lty = 2, lwd = 2, from = -2, to = 8)
legend("topright", 
       legend = c("Random-Effects", "RoBMA"), 
       col = c("black", "blue"), 
       lty = 2, lwd = 2)

## ----fig.cap="ChatGPT: Extrapolation Analysis"--------------------------------
# Examine extrapolation
plot(zcurve_RoBMA_Wang2025, from = -2, to = 8)

## -----------------------------------------------------------------------------
summary(fit_RE_Wang2025)
summary(fit_RoBMA_Wang2025)

## -----------------------------------------------------------------------------
# Extract z-curve summary metrics for ChatGPT data
summary(zcurve_RoBMA_Wang2025)

## -----------------------------------------------------------------------------
# Load the Many Labs 2 framing effect data
data("ManyLabs16", package = "RoBMA")
head(ManyLabs16)

## ----eval = FALSE-------------------------------------------------------------
# # Fit models and create z-curve objects
# fit_RE_ManyLabs16 <- NoBMA(d = ManyLabs16$y, se = ManyLabs16$se,
#                            priors_effect_null = NULL, priors_heterogeneity_null = NULL,
#                            algorithm = "ss", sample = 10000, burnin = 5000, adapt = 5000,
#                            chains = 5, parallel = TRUE, seed = 1)
# fit_RoBMA_ManyLabs16 <- RoBMA(d = ManyLabs16$y, se = ManyLabs16$se,
#                               algorithm = "ss", sample = 10000, burnin = 5000, adapt = 5000,
#                               chains = 5, parallel = TRUE, seed = 1)
# 
# zcurve_RE_ManyLabs16    <- as_zcurve(fit_RE_ManyLabs16)
# zcurve_RoBMA_ManyLabs16 <- as_zcurve(fit_RoBMA_ManyLabs16)

## ----fig.cap="Framing Effects: Model Fit Assessment"--------------------------
# Assess model fit - should show good agreement
hist(zcurve_RoBMA_ManyLabs16)
lines(zcurve_RE_ManyLabs16, col = "black", lty = 2, lwd = 2)
lines(zcurve_RoBMA_ManyLabs16, col = "blue", lty = 2, lwd = 2)
legend("topleft", 
       legend = c("Random-Effects", "RoBMA"), 
       col = c("black", "blue"), 
       lty = 2, lwd = 2)

## ----fig.cap="Framing Effects: Extrapolation Analysis"------------------------
# Extrapolation should show minimal change
plot(zcurve_RoBMA_ManyLabs16)

## ----eval = FALSE-------------------------------------------------------------
# summary(fit_RE_ManyLabs16)
# summary(fit_RoBMA_ManyLabs16)

## -----------------------------------------------------------------------------
# Extract z-curve summary metrics for Many Labs 2 data
summary(zcurve_RoBMA_ManyLabs16)

## ----setup, include = FALSE---------------------------------------------------
is_check <- ("CheckExEnv" %in% search()) ||
              any(c("_R_CHECK_TIMINGS_", "_R_CHECK_LICENSE_") %in% names(Sys.getenv())) ||
              !file.exists("../models/MedicineBiBMA/fit.RDS")
knitr::opts_chunk$set(
  collapse  = TRUE,
  comment   = "#>",
  eval      = !is_check,
  dev      = "png")
if(.Platform$OS.type == "windows"){
  knitr::opts_chunk$set(dev.args = list(type = "cairo"))
}

## ----include = FALSE----------------------------------------------------------
library(RoBMA)
# we pre-load the RoBMA models, the fitting time is around 2-5 minutes
fit <- readRDS(file = "../models/MedicineBiBMA/fit.RDS")

## ----include = FALSE, eval = FALSE--------------------------------------------
# # R package version updating
# library(RoBMA)
# 
# data("Poulsen2006", package = "RoBMA")
# 
# # p. 73 in https://www.cochranelibrary.com/cdsr/doi/10.1002/14651858.CD007094.pub5/epdf/full
# events_experimental        <- c(5, 2)
# events_control             <- c(0, 0)
# observations_experimental  <- c(35, 40)
# observations_control       <- c(39, 40)
# study_names <- c("Paul 2007", "Shadkam 2010")
# 
# # domain specific prior distributions: Acute Respiratory Infections
# fit <- BiBMA(
#   x1          = events_experimental,
#   x2          = events_control,
#   n1          = observations_experimental,
#   n2          = observations_control,
#   study_names = study_names,
#   priors_effect        = prior_informed("Acute Respiratory Infections", type = "logOR", parameter = "effect"),
#   priors_heterogeneity = prior_informed("Acute Respiratory Infections", type = "logOR", parameter = "heterogeneity"),
#   seed = 1
# )
# 
# saveRDS(fit,  file = "../models/MedicineBiBMA/fit.RDS")

## -----------------------------------------------------------------------------
library(RoBMA)

events_experimental        <- c(5, 2)
events_control             <- c(0, 0)
observations_experimental  <- c(35, 40)
observations_control       <- c(39, 40)
study_names <- c("Paul 2007", "Shadkam 2010")

## -----------------------------------------------------------------------------
summary(fit, conditional = TRUE, output_scale = "OR")

## ----fig_mu_BMA, dpi = 300, fig.width = 6, fig.height = 4, out.width = "75%", fig.align = "center"----
plot(fit, parameter = "mu", prior = TRUE, conditional = TRUE)

## ----setup, include = FALSE---------------------------------------------------
is_check <- ("CheckExEnv" %in% search()) ||
              any(c("_R_CHECK_TIMINGS_", "_R_CHECK_LICENSE_") %in% names(Sys.getenv())) ||
              !file.exists("../models/HierarchicalBMA/fit.RDS")
knitr::opts_chunk$set(
  collapse  = TRUE,
  comment   = "#>",
  eval      = !is_check,
  dev      = "png")
if(.Platform$OS.type == "windows"){
  knitr::opts_chunk$set(dev.args = list(type = "cairo"))
}

## ----include = FALSE----------------------------------------------------------
library(RoBMA)
# we pre-load the RoBMA models, the fitting time is around 2-5 minutes
fit.0             <- readRDS(file = "../models/HierarchicalBMA/fit.0.RDS")
fit               <- readRDS(file = "../models/HierarchicalBMA/fit.RDS")
fit_BMA           <- readRDS(file = "../models/HierarchicalBMA/fit_BMA.RDS")
hierarchical_test <- readRDS(file = "../models/HierarchicalBMA/hierarchical_test.RDS")

## ----include = FALSE, eval = FALSE--------------------------------------------
# # R package version updating
# library(RoBMA)
# 
# data("dat.konstantopoulos2011", package = "metadat")
# dat <- dat.konstantopoulos2011
# 
# fit.0 <- RoBMA(d = dat$yi, v = dat$vi,
#              priors_effect_null        = NULL,
#              priors_heterogeneity_null = NULL,
#              priors_bias               = NULL,
#              parallel = TRUE, seed = 1)
# 
# fit <- RoBMA(d = dat$yi, v = dat$vi, study_ids = dat$district,
#              priors_effect_null        = NULL,
#              priors_heterogeneity_null = NULL,
#              priors_bias               = NULL,
#              parallel = TRUE, seed = 1)
# 
# fit_BMA <- RoBMA(d = dat$yi, v = dat$vi, study_ids = dat$district,
#                  priors_bias = NULL,
#                  parallel = TRUE, seed = 1)
# 
# hierarchical_test <- RoBMA(d = dat$yi, v = dat$vi, study_ids = dat$district,
#                            priors_heterogeneity_null = NULL,
#                            priors_hierarchical_null = prior(distribution = "spike", parameters = list("location" = 0)),
#                            priors_bias = NULL,
#                            parallel = TRUE, seed = 1)
# 
# saveRDS(fit.0,             file = "../models/HierarchicalBMA/fit.0.RDS",             compress = "xz")
# saveRDS(fit,               file = "../models/HierarchicalBMA/fit.RDS",               compress = "xz")
# saveRDS(fit_BMA,           file = "../models/HierarchicalBMA/fit_BMA.RDS",           compress = "xz")
# saveRDS(hierarchical_test, file = "../models/HierarchicalBMA/hierarchical_test.RDS", compress = "xz")

## -----------------------------------------------------------------------------
data("dat.konstantopoulos2011", package = "metadat")
dat <- dat.konstantopoulos2011

head(dat)

## -----------------------------------------------------------------------------
fit_metafor.0 <- metafor::rma(yi = yi, vi = vi, data = dat)
fit_metafor.0

## -----------------------------------------------------------------------------
fit_metafor <- metafor::rma.mv(yi, vi, random = ~ school | district, data = dat)
fit_metafor

## -----------------------------------------------------------------------------
summary(fit.0, type = "individual")

## -----------------------------------------------------------------------------
summary(fit, type = "individual")

## ----fig_rho, dpi = 300, fig.width = 4, fig.height = 3, out.width = "50%", fig.align = "center"----
par(mar = c(2, 4, 0, 0))
plot(fit, parameter = "rho", prior = TRUE)

## -----------------------------------------------------------------------------
summary(fit_BMA)

## -----------------------------------------------------------------------------
summary(hierarchical_test)

## ----vignette-nowrap-options, include = FALSE, eval = [encoded data, 9944 characters]
options(width = 10000)

## ----setup, include = [encoded data, 9976 characters]
source("vignette-cache.R", local = knitr::knit_global())

cached_fits <- vignette_cache(
  name     = "v10-metafor-parity-multilevel",
  objects  = c("fit1_brma"),
  packages = c("metafor", "metadat")
)

knitr::opts_chunk$set(
  collapse    = TRUE,
  comment     = "#>",
  eval        = vignette_cache_eval(cached_fits),
  message     = FALSE,
  warning     = FALSE,
  fig.width   = 7,
  fig.height  = 3.5,
  dev         = "png",
  fig.retina  = 3
)
if (.Platform$OS.type == "windows") {
  knitr::opts_chunk$set(dev.args = list(type = "cairo"))
}

## ----load-models, include = [encoded data, 9970 characters]
library("metafor")
library("RoBMA")

data("dat.konstantopoulos2011", package = "metadat")
dat <- dat.konstantopoulos2011

vignette_cache_load(cached_fits)

## ----precompute-models, include = FALSE, eval = [encoded data, 9950 characters]
# library("metafor")
# library("RoBMA")
# 
# data("dat.konstantopoulos2011", package = "metadat")
# dat <- dat.konstantopoulos2011
# 
# fit1_brma <- brma(
#   yi = yi, vi = vi, measure = "SMD",
#   cluster = district,
#   data = dat, seed = 1
# )
# 
# vignette_cache_save(cached_fits)

## [encoded data, 9997 characters]
data("dat.konstantopoulos2011", package = "metadat")
dat <- dat.konstantopoulos2011
head(dat)

## [encoded data, 9997 characters]
fit1_metafor <- metafor::rma.mv(yi, vi, random = ~ school | district, data = dat)
fit1_metafor

## ----fit1-brma, eval = [encoded data, 9975 characters]
# fit1_brma <- brma(
#   yi = yi, vi = vi, measure = "SMD",
#   cluster = district,
#   data = dat, seed = 1
# )

## [encoded data, 9997 characters]
summary(fit1_brma, include_mcmc_diagnostics = FALSE)

## [encoded data, 9997 characters]
confint(fit1_metafor)

## [encoded data, 9997 characters]
summary_heterogeneity(fit1_brma)

## ----tau-rho-posterior, fig.width = 7, fig.height = 3.[encoded data, 9944 characters]
par(mfrow = c(1, 2), mar = c(4, 4, 2, 1), cex.axis = 0.8, cex.lab = 0.8, cex.main = 0.75)
plot(fit1_brma, parameter = "tau", prior = TRUE, main = "tau", xlim = c(0, 1))
plot(fit1_brma, parameter = "rho", prior = TRUE, main = "rho", xlim = c(0, 1))

## ----vignette-nowrap-options, include = FALSE, eval = [encoded data, 9944 characters]
options(width = 10000)

## ----setup, include = [encoded data, 9976 characters]
source("vignette-cache.R", local = knitr::knit_global())

cached_fits <- vignette_cache(
  name    = "v32-robma-multilevel",
  objects = c("fit", "fit_simple")
)

knitr::opts_chunk$set(
  collapse    = TRUE,
  comment     = "#>",
  eval        = vignette_cache_eval(cached_fits),
  message     = FALSE,
  warning     = FALSE,
  fig.width   = 7,
  fig.height  = 3.5,
  dev         = "png",
  fig.retina  = 3
)
if (.Platform$OS.type == "windows") {
  knitr::opts_chunk$set(dev.args = list(type = "cairo"))
}

## ----load-models, include = [encoded data, 9970 characters]
library("RoBMA")

data("Johnides2025", package = "RoBMA")

vignette_cache_load(cached_fits)

## ----precompute-models, include = FALSE, eval = [encoded data, 9950 characters]
# library("RoBMA")
# 
# data("Johnides2025", package = "RoBMA")
# 
# prior_effect_361        <- prior("normal",   list(mean = 0, sd = 1))
# prior_heterogeneity_361 <- prior("invgamma", list(shape = 1, scale = 0.15))
# 
# fit <- RoBMA(
#   yi = d, sei = se, measure = "SMD", cluster = study,
#   prior_effect        = prior_effect_361,
#   prior_heterogeneity = prior_heterogeneity_361,
#   adapt = 5000, burnin = 5000, sample = 10000,
#   parallel  = TRUE, seed = 1, data = Johnides2025
# )
# 
# fit_simple <- RoBMA(
#   yi = d, sei = se, measure = "SMD",
#   prior_effect        = prior_effect_361,
#   prior_heterogeneity = prior_heterogeneity_361,
#   adapt = 5000, burnin = 5000, sample = 10000,
#   parallel  = TRUE, seed = 1, data = Johnides2025
# )
# 
# vignette_cache_save(cached_fits)

## [encoded data, 9997 characters]
library(RoBMA)
data("Johnides2025", package = "RoBMA")

## [encoded data, 9997 characters]
head(Johnides2025)

## [encoded data, 9997 characters]
prior_effect_361        <- prior("normal",   list(mean = 0, sd = 1))
prior_heterogeneity_361 <- prior("invgamma", list(shape = 1, scale = 0.15))

## ----eval = [encoded data, 9986 characters]
# fit <- RoBMA(
#   yi = d, sei = se, measure = "SMD", cluster = study,
#   prior_effect        = prior_effect_361,
#   prior_heterogeneity = prior_heterogeneity_361,
#   adapt = 5000, burnin = 5000, sample = 10000,
#   parallel  = TRUE, seed = 1, data = Johnides2025
# )

## [encoded data, 9997 characters]
summary(fit)

## [encoded data, 9997 characters]
summary_heterogeneity(fit)

## [encoded data, 9997 characters]
summary_models(fit)

## ----fig.width = 6, fig.height = [encoded data, 9965 characters]
plot_weightfunction(fit)

## ----eval = [encoded data, 9986 characters]
# fit_simple <- RoBMA(
#   yi = d, sei = se, measure = "SMD",
#   prior_effect        = prior_effect_361,
#   prior_heterogeneity = prior_heterogeneity_361,
#   adapt = 5000, burnin = 5000, sample = 10000,
#   parallel  = TRUE, seed = 1, data = Johnides2025
# )

## [encoded data, 9997 characters]
summary(fit_simple)

## ----vignette-nowrap-options, include = FALSE, eval = [encoded data, 9944 characters]
options(width = 10000)

## ----setup, include = [encoded data, 9976 characters]
source("vignette-cache.R", local = knitr::knit_global())

cached_fits <- vignette_cache(
  name     = "v20-bayesian-model-averaging",
  objects  = c("fit_RE", "fit_FE", "fit_BMA_he", "fit_BMA", "fit_BMA_mod"),
  packages = c("metafor", "metadat")
)

knitr::opts_chunk$set(
  collapse    = TRUE,
  comment     = "#>",
  eval        = vignette_cache_eval(cached_fits),
  message     = FALSE,
  warning     = FALSE,
  fig.width   = 7,
  fig.height  = 3.5,
  dev         = "png",
  fig.retina  = 3
)
if (.Platform$OS.type == "windows") {
  knitr::opts_chunk$set(dev.args = list(type = "cairo"))
}

## ----load-models, include = [encoded data, 9970 characters]
library("RoBMA")

data("dat.baskerville2012", package = "metadat")
dat <- dat.baskerville2012

vignette_cache_load(cached_fits)

## ----precompute-models, include = FALSE, eval = [encoded data, 9950 characters]
# library("RoBMA")
# 
# data("dat.baskerville2012", package = "metadat")
# dat <- dat.baskerville2012
# dat$tailor <- as.factor(dat$tailor)
# 
# fit_RE <- brma(
#   yi = smd, sei = se, measure = "SMD",
#   data = dat, seed = 1
# )
# fit_RE <- add_marglik(fit_RE)
# 
# fit_FE <- brma(
#   yi = smd, sei = se, measure = "SMD",
#   prior_heterogeneity = prior("spike", list(location = 0)),
#   data = dat, seed = 1
# )
# fit_FE <- add_marglik(fit_FE)
# 
# fit_BMA_he <- BMA(
#   yi = smd, sei = se, measure = "SMD",
#   prior_effect_null = NULL,
#   data = dat, seed = 1
# )
# 
# fit_BMA <- BMA(
#   yi = smd, sei = se, measure = "SMD",
#   data = dat, seed = 1
# )
# 
# fit_BMA_mod <- BMA(
#   yi = smd, sei = se, measure = "SMD",
#   mods = ~ tailor,
#   data = dat, seed = 1
# )
# 
# vignette_cache_save(cached_fits)

## ----set-up, include = [encoded data, 9975 characters]
library("RoBMA")

data("dat.baskerville2012", package = "metadat")
dat <- dat.baskerville2012
dat$tailor <- as.factor(dat$tailor)
head(dat[, c("year", "smd", "se", "tailor")])

## [encoded data, 9997 characters]
res <- metafor::rma.uni(
  yi = smd, sei = se,
  data = dat, method = "REML"
)
res

## ----fit-RE, eval = [encoded data, 9978 characters]
# fit_RE <- brma(
#   yi = smd, sei = se, measure = "SMD",
#   data = dat, seed = 1
# )
# fit_RE <- add_marglik(fit_RE)

## [encoded data, 9997 characters]
summary(fit_RE, include_mcmc_diagnostics = FALSE)

## ----fit-FE, eval = [encoded data, 9978 characters]
# fit_FE <- brma(
#   yi = smd, sei = se, measure = "SMD",
#   prior_heterogeneity = prior("spike", list(location = 0)),
#   data = dat, seed = 1
# )
# fit_FE <- add_marglik(fit_FE)

## [encoded data, 9997 characters]
summary(fit_FE, include_mcmc_diagnostics = FALSE)

## [encoded data, 9997 characters]
bf(fit_RE, fit_FE)

## ----fit-BMA-he, eval = [encoded data, 9974 characters]
# fit_BMA_he <- BMA(
#   yi = smd, sei = se, measure = "SMD",
#   prior_effect_null = NULL,
#   data = dat, seed = 1
# )

## [encoded data, 9997 characters]
summary(fit_BMA_he)

## ----fit-BMA, eval = [encoded data, 9977 characters]
# fit_BMA <- BMA(
#   yi = smd, sei = se, measure = "SMD",
#   data = dat, seed = 1
# )

## [encoded data, 9997 characters]
summary(fit_BMA)

## ----plot-BMA-mu, fig.width = 6, fig.height = 4, out.width = "75%", fig.align = "center"[encoded data, 9910 characters]
par(mar = c(4, 4, 1, 4))
plot(fit_BMA, parameter = "mu", prior = TRUE, xlim = c(-1, 1.5), lwd = 2)

## ----plot-BMA-tau, fig.width = 6, fig.height = 4, out.width = "75%", fig.align = "center"[encoded data, 9909 characters]
par(mar = c(4, 4, 1, 4))
plot(fit_BMA, parameter = "tau", prior = TRUE, xlim = c(0, 1), lwd = 2)

## [encoded data, 9997 characters]
summary(fit_BMA, conditional = TRUE)

## [encoded data, 9997 characters]
res_mod <- metafor::rma.uni(
  yi = smd, sei = se, mods = ~ tailor,
  data = dat, method = "REML"
)
res_mod

## ----fit-BMA-mod, eval = [encoded data, 9973 characters]
# fit_BMA_mod <- BMA(
#   yi = smd, sei = se, measure = "SMD",
#   mods = ~ tailor,
#   data = dat, seed = 1
# )

## [encoded data, 9997 characters]
summary(fit_BMA_mod)

## ----regplot-BMA-mod, fig.width = 5, fig.height = 3.5, fig.align = "center"[encoded data, 9923 characters]
par(mar = c(4, 4, 1, 1))
regplot(fit_BMA_mod, mod = "tailor")

## ----regplot-BMA-emm, fig.width = 5, fig.height = 3.5, fig.align = "center"[encoded data, 9923 characters]
par(mar = c(4, 4, 1, 1))
emm <- marginal_means(fit_BMA_mod)
emm

plot(emm, parameter = "tailor", prior = TRUE)

## [encoded data, 9997 characters]
fit_priors <- BMA(
  yi = smd, sei = se, measure = "SMD",
  rescale_priors = 0.5,
  data = dat, only_priors = TRUE
)
print_prior(fit_priors, parameter = "mu")

## [encoded data, 9997 characters]
fit_priors <- BMA(
  yi = smd, sei = se, measure = "SMD",
  prior_effect_null = prior("normal", list(mean = 0, sd = 0.05)),
  data = dat, only_priors = TRUE
)
print_prior(fit_priors, parameter = "mu")

## [encoded data, 9997 characters]
fit_priors <- BMA(
  yi = smd, sei = se, measure = "SMD",
  mods = ~ tailor,
  prior_mods_null = list(tailor = NULL),
  data = dat, only_priors = TRUE
)
print_prior(fit_priors, parameter_mods = "tailor")

## [encoded data, 9997 characters]
fit_priors <- BMA(
  yi = smd, sei = se, measure = "SMD",
  prior_heterogeneity = prior(
    distribution = "normal",
    parameters   = list(mean = 0, sd = 0.25),
    truncation   = list(lower = 0)
  ),
  data = dat, only_priors = TRUE
)
print_prior(fit_priors, parameter = "tau")

## ----vignette-nowrap-options, include = FALSE, eval = [encoded data, 9944 characters]
options(width = 10000)

## ----setup, include = [encoded data, 9976 characters]
source("vignette-cache.R", local = knitr::knit_global())

cached_fits <- vignette_cache(
  name     = "v21-robust-bayesian-meta-analysis",
  objects  = c("fit_RoBMA", "fit_RoBMA_custom"),
  packages = c("metadat")
)

knitr::opts_chunk$set(
  collapse    = TRUE,
  comment     = "#>",
  eval        = vignette_cache_eval(cached_fits),
  message     = FALSE,
  warning     = FALSE,
  fig.width   = 7,
  fig.height  = 3.5,
  dev         = "png",
  fig.retina  = 3
)
if (.Platform$OS.type == "windows") {
  knitr::opts_chunk$set(dev.args = list(type = "cairo"))
}

## ----load-models, include = [encoded data, 9970 characters]
library("RoBMA")

data("dat.lehmann2018", package = "metadat")
dat <- dat.lehmann2018

vignette_cache_load(cached_fits)

## ----precompute-models, include = FALSE, eval = [encoded data, 9950 characters]
# library("RoBMA")
# 
# data("dat.lehmann2018", package = "metadat")
# dat <- dat.lehmann2018
# 
# fit_RoBMA <- RoBMA(
#   yi = yi, vi = vi, measure = "SMD",
#   data = dat, seed = 1
# )
# 
# fit_RoBMA_custom <- RoBMA(
#   yi = yi, vi = vi, measure = "SMD",
#   prior_bias = list(
#     prior_PET(
#       distribution = "Cauchy",
#       parameters   = list(location = 0, scale = 1),
#       truncation   = list(lower = 0, upper = Inf)
#     ),
#     prior_PEESE(
#       distribution = "Cauchy",
#       parameters   = list(location = 0, scale = 5),
#       truncation   = list(lower = 0, upper = Inf)
#     ),
#     prior_weightfunction(
#       "one-sided",
#       steps   = c(0.025),
#       weights = wf_cumulative(c(1, 1))
#     )
#   ),
#   data = dat, seed = 1
# )
# 
# vignette_cache_save(cached_fits)

## ----data, include = [encoded data, 9977 characters]
library("RoBMA")

data("dat.lehmann2018", package = "metadat")
dat <- dat.lehmann2018
head(dat[, c("Short_Title", "Year", "yi", "vi")])

## ----fit-RoBMA, eval = [encoded data, 9975 characters]
# fit_RoBMA <- RoBMA(
#   yi = yi, vi = vi, measure = "SMD",
#   data = dat, seed = 1
# )

## [encoded data, 9997 characters]
summary(fit_RoBMA)

## [encoded data, 9997 characters]
summary_models(fit_RoBMA)

## ----plot-RoBMA-mu, fig.width = 6, fig.height = 4, out.width = "75%", fig.align = "center"[encoded data, 9908 characters]
par(mar = c(4, 4, 1, 4))
plot(fit_RoBMA, parameter = "mu", prior = TRUE, xlim = c(-1, 1), lwd = 2)

## [encoded data, 9997 characters]
interpret(fit_RoBMA, scope = "all")

## [encoded data, 9997 characters]
fit_priors <- RoBMA(
  yi = yi, vi = vi, measure = "SMD",
  model_type = "PP",
  data = dat, only_priors = TRUE
)
print_prior(fit_priors, parameter = "bias")

## ----fit-RoBMA-custom, eval = [encoded data, 9968 characters]
# fit_RoBMA_custom <- RoBMA(
#   yi = yi, vi = vi, measure = "SMD",
#   prior_bias = list(
#     prior_PET(
#       distribution = "Cauchy",
#       parameters   = list(location = 0, scale = 1),
#       truncation   = list(lower = 0, upper = Inf)
#     ),
#     prior_PEESE(
#       distribution = "Cauchy",
#       parameters   = list(location = 0, scale = 5),
#       truncation   = list(lower = 0, upper = Inf)
#     ),
#     prior_weightfunction(
#       "one-sided",
#       steps   = c(0.025),
#       weights = wf_cumulative(c(1, 1))
#     )
#   ),
#   data = dat, seed = 1
# )

## [encoded data, 9997 characters]
interpret(fit_RoBMA_custom, scope = c("components", "estimates"))

## ----vignette-nowrap-options, include = FALSE, eval = [encoded data, 9944 characters]
options(width = 10000)

## ----setup, include = [encoded data, 9976 characters]
source("vignette-cache.R", local = knitr::knit_global())

cached_fits <- vignette_cache(
  name     = "v30-tutorial",
  objects  = c(
    fit_RoBMA  = "fit_RoBMA_Lui2015.RDS",
    fit_RoBMA2 = "fit_RoBMA_perinull_Lui2015.RDS"
  ),
  packages = c("metafor", "weightr")
)

knitr::opts_chunk$set(
  collapse    = TRUE,
  comment     = "#>",
  eval        = vignette_cache_eval(cached_fits),
  message     = FALSE,
  warning     = FALSE,
  fig.width   = 7,
  fig.height  = 3.5,
  dev         = "png",
  fig.retina  = 3
)
if (.Platform$OS.type == "windows") {
  knitr::opts_chunk$set(dev.args = list(type = "cairo"))
}

prepare_30_tutorial <- function(envir = parent.frame()) {

  library("metafor")
  library("weightr")
  library("RoBMA")

  data("Lui2015", package = "RoBMA")
  df <- Lui2015

  dfz <- metafor::escalc(
    ri = r, ni = n, measure = "ZCOR",
    data = df
  )

  prior_effect_361        <- prior("normal", list(mean = 0, sd = 0.5))
  prior_heterogeneity_361 <- prior("invgamma", list(shape = 1, scale = 0.075))
  prior_bias_361          <- list(
    prior_weightfunction("two-sided", c(0.05),              wf_cumulative(c(1, 1)),       prior_weights = 1/12),
    prior_weightfunction("two-sided", c(0.05, 0.10),        wf_cumulative(c(1, 1, 1)),    prior_weights = 1/12),
    prior_weightfunction("one-sided", c(0.05),              wf_cumulative(c(1, 1)),       prior_weights = 1/12),
    prior_weightfunction("one-sided", c(0.025, 0.05),       wf_cumulative(c(1, 1, 1)),    prior_weights = 1/12),
    prior_weightfunction("one-sided", c(0.05, 0.50),        wf_cumulative(c(1, 1, 1)),    prior_weights = 1/12),
    prior_weightfunction("one-sided", c(0.025, 0.05, 0.50), wf_cumulative(c(1, 1, 1, 1)), prior_weights = 1/12),
    prior_PET("Cauchy",   list(location = 0, scale = 1),  truncation = list(0, Inf), prior_weights = 1/4),
    prior_PEESE("Cauchy", list(location = 0, scale = 10), truncation = list(0, Inf), prior_weights = 1/4)
  )

  prior_effect_perinull_361      <- prior("normal", list(mean = 0.30, sd = 0.10), truncation = list(0, Inf))
  prior_effect_null_perinull_361 <- prior("normal", list(mean = 0, sd = 0.05))

  list2env(
    list(
      df                                = df,
      dfz                               = dfz,
      prior_effect_361                  = prior_effect_361,
      prior_heterogeneity_361           = prior_heterogeneity_361,
      prior_bias_361                    = prior_bias_361,
      prior_effect_perinull_361         = prior_effect_perinull_361,
      prior_effect_null_perinull_361    = prior_effect_null_perinull_361
    ),
    envir = envir
  )

  invisible(NULL)
}

## ----load-models, include = [encoded data, 9970 characters]
prepare_30_tutorial()
vignette_cache_load(cached_fits)

## ----precompute-models, include = FALSE, eval = [encoded data, 9950 characters]
# prepare_30_tutorial()
# fit_RoBMA <- RoBMA(
#   yi = yi, vi = vi, measure = "ZCOR",
#   prior_effect                 = prior_effect_361,
#   prior_heterogeneity          = prior_heterogeneity_361,
#   prior_bias                   = prior_bias_361,
#   data = dfz, seed = 1, parallel = TRUE
# )
# 
# fit_RoBMA2 <- RoBMA(
#   yi = yi, vi = vi, measure = "ZCOR",
#   prior_effect                 = prior_effect_perinull_361,
#   prior_effect_null            = prior_effect_null_perinull_361,
#   prior_heterogeneity          = prior_heterogeneity_361,
#   prior_bias                   = prior_bias_361,
#   data = dfz, seed = 2, parallel = TRUE
# )
# 
# vignette_cache_save(cached_fits)

## [encoded data, 9997 characters]
library("metafor")
library("weightr")
library("RoBMA")

## [encoded data, 9997 characters]
head(df)

## [encoded data, 9997 characters]
df$r

## [encoded data, 9997 characters]
dfz <- metafor::escalc(
  ri = r, ni = n, measure = "ZCOR",
  data = df
)
head(dfz)

## [encoded data, 9997 characters]
fit_rma <- metafor::rma(yi = yi, vi = vi, data = dfz)
fit_rma

## [encoded data, 9997 characters]
metafor::transf.ztor(fit_rma$b)

## [encoded data, 9997 characters]
fit_PET <- lm(yi ~ I(sqrt(vi)), weights = 1/vi, data = dfz)
summary(fit_PET)

## [encoded data, 9997 characters]
metafor::transf.ztor(summary(fit_PET)$coefficients["(Intercept)", "Estimate"])

## [encoded data, 9997 characters]
fit_PEESE <- lm(yi ~ vi, weights = 1/vi, data = dfz)
summary(fit_PEESE)

## [encoded data, 9997 characters]
fit_4PSM <- weightr::weightfunct(
  effect = dfz$yi, v = dfz$vi,
  steps = c(0.025, 0.05), table = TRUE
)
fit_4PSM

## [encoded data, 9997 characters]
fit_3PSM <- weightr::weightfunct(
  effect = dfz$yi, v = dfz$vi,
  steps = c(0.025), table = TRUE
)
fit_3PSM

## [encoded data, 9997 characters]
metafor::transf.ztor(fit_3PSM$adj_est[2])

## [encoded data, 9997 characters]
fit_sel_z <- metafor::selmodel(fit_rma, type = "stepfun", steps = c(0.025))
fit_sel_z

## [encoded data, 9997 characters]
prior_effect_361        <- prior("normal", list(mean = 0, sd = 0.5))
prior_heterogeneity_361 <- prior("invgamma", list(shape = 1, scale = 0.075))
prior_bias_361          <- list(
  prior_weightfunction("two-sided", c(0.05),             wf_cumulative(c(1, 1)),       prior_weights = 1/12),
  prior_weightfunction("two-sided", c(0.05, 0.10),       wf_cumulative(c(1, 1, 1)),    prior_weights = 1/12),
  prior_weightfunction("one-sided", c(0.05),             wf_cumulative(c(1, 1)),       prior_weights = 1/12),
  prior_weightfunction("one-sided", c(0.025, 0.05),      wf_cumulative(c(1, 1, 1)),    prior_weights = 1/12),
  prior_weightfunction("one-sided", c(0.05, 0.50),       wf_cumulative(c(1, 1, 1)),    prior_weights = 1/12),
  prior_weightfunction("one-sided", c(0.025, 0.05, 0.50), wf_cumulative(c(1, 1, 1, 1)), prior_weights = 1/12),
  prior_PET("Cauchy",   list(location = 0, scale = 1),  truncation = list(0, Inf), prior_weights = 1/4),
  prior_PEESE("Cauchy", list(location = 0, scale = 10), truncation = list(0, Inf), prior_weights = 1/4)
)

## ----fit-RoBMA, eval = [encoded data, 9975 characters]
# fit_RoBMA <- RoBMA(
#   yi = yi, vi = vi, measure = "ZCOR",
#   prior_effect                 = prior_effect_361,
#   prior_heterogeneity          = prior_heterogeneity_361,
#   prior_bias                   = prior_bias_361,
#   data = dfz, seed = 1, parallel = TRUE
# )

## [encoded data, 9997 characters]
summary(fit_RoBMA)

## [encoded data, 9997 characters]
pooled_effect(fit_RoBMA, output_measure = "COR")

## [encoded data, 9997 characters]
summary(fit_RoBMA, conditional = TRUE)

## [encoded data, 9997 characters]
summary_models(fit_RoBMA, type = "individual")

## [encoded data, 9997 characters]
plot_diagnostic_trace(fit_RoBMA, parameter = "mu")

## ----fig.width = 6, fig.height = 4, out.width = "75%", fig.align = "center"[encoded data, 9923 characters]
par(mar = c(4, 4, 1, 4))
plot(fit_RoBMA, parameter = "mu", prior = TRUE, output_measure = "COR", xlim = c(-1, 1))

## [encoded data, 9997 characters]
prior_effect_perinull_361      <- prior("normal", list(mean = 0.30, sd = 0.10), truncation = list(0, Inf))
prior_effect_null_perinull_361 <- prior("normal", list(mean = 0, sd = 0.05))

## ----fit-RoBMA2, eval = [encoded data, 9974 characters]
# fit_RoBMA2 <- RoBMA(
#   yi = yi, vi = vi, measure = "ZCOR",
#   prior_effect                 = prior_effect_perinull_361,
#   prior_effect_null            = prior_effect_null_perinull_361,
#   prior_heterogeneity          = prior_heterogeneity_361,
#   prior_bias                   = prior_bias_361,
#   data = dfz, seed = 2, parallel = TRUE
# )

## [encoded data, 9997 characters]
summary_models(fit_RoBMA2, type = "individual")

## ----vignette-nowrap-options, include = FALSE, eval = [encoded data, 9944 characters]
options(width = 10000)

## ----setup, include = [encoded data, 9976 characters]
source("vignette-cache.R", local = knitr::knit_global())

cached_fits <- vignette_cache(
  name    = "v35-bma-glmm-medicine",
  objects = c("fit_BMA")
)

knitr::opts_chunk$set(
  collapse   = TRUE,
  comment    = "#>",
  eval       = vignette_cache_eval(cached_fits),
  message    = FALSE,
  warning    = FALSE,
  dev        = "png",
  fig.retina = 3
)
if (.Platform$OS.type == "windows") {
  knitr::opts_chunk$set(dev.args = list(type = "cairo"))
}

## ----include = [encoded data, 9983 characters]
library(RoBMA)

events_experimental        <- c(5, 2)
events_control             <- c(0, 0)
observations_experimental  <- c(35, 40)
observations_control       <- c(39, 40)
study_names                <- c("Paul 2007", "Shadkam 2010")

vignette_cache_load(cached_fits)

## ----include = FALSE, eval = [encoded data, 9969 characters]
# # R package version updating
# library(RoBMA)
# 
# events_experimental        <- c(5, 2)
# events_control             <- c(0, 0)
# observations_experimental  <- c(35, 40)
# observations_control       <- c(39, 40)
# study_names                <- c("Paul 2007", "Shadkam 2010")
# 
# fit_BMA <- BMA.glmm(
#   ai = events_experimental, n1i = observations_experimental,
#   ci = events_control, n2i = observations_control,
#   slab = study_names, measure = "OR",
#   prior_informed_field    = "medicine",
#   prior_informed_subfield = "Acute Respiratory Infections",
#   seed = 1, parallel = TRUE
# )
# 
# vignette_cache_save(cached_fits)

## [encoded data, 9997 characters]
library(RoBMA)

events_experimental        <- c(5, 2)
events_control             <- c(0, 0)
observations_experimental  <- c(35, 40)
observations_control       <- c(39, 40)
study_names                <- c("Paul 2007", "Shadkam 2010")

## [encoded data, 9997 characters]
summary(fit_BMA, conditional = TRUE, include_mcmc_diagnostics = FALSE)

## [encoded data, 9997 characters]
pooled_effect(fit_BMA, conditional = TRUE, transform = "EXP")
pooled_heterogeneity(fit_BMA, conditional = TRUE)

## ----fig_mu_BMA, fig.width = 6, fig.height = 4, out.width = "75%", fig.align = "center"[encoded data, 9911 characters]
par(mar = c(4, 4, 0, 4))
plot(fit_BMA, parameter = "mu", prior = TRUE, conditional = TRUE, lwd = 2)

## ----vignette-nowrap-options, include = FALSE, eval = [encoded data, 9944 characters]
options(width = 10000)

## ----setup, include = [encoded data, 9976 characters]
knitr::opts_chunk$set(
  collapse = TRUE,
  comment  = "#>",
  eval     = FALSE
)

article_url <- function(page) {
  paste0(page, ".html")
}

reference_url <- function(page) {
  paste0("../reference/", page, ".html")
}

## ----vignette-nowrap-options, include = FALSE, eval = [encoded data, 9944 characters]
options(width = 10000)

## ----setup, include = [encoded data, 9976 characters]
source("vignette-cache.R", local = knitr::knit_global())

cached_fits <- vignette_cache(
  name     = "v12-metafor-parity-location-scale",
  objects  = c("fit1_brma", "fit2_brma", "fit3_brma"),
  packages = c("metafor", "metadat")
)

knitr::opts_chunk$set(
  collapse    = TRUE,
  comment     = "#>",
  eval        = vignette_cache_eval(cached_fits),
  message     = FALSE,
  warning     = FALSE,
  fig.width   = 7,
  fig.height  = 3.5,
  dev         = "png",
  fig.retina  = 3
)
if (.Platform$OS.type == "windows") {
  knitr::opts_chunk$set(dev.args = list(type = "cairo"))
}

## ----load-models, include = [encoded data, 9970 characters]
library("metafor")
library("RoBMA")

data("dat.bangertdrowns2004", package = "metadat")
dat <- dat.bangertdrowns2004
dat$ni100 <- dat$ni / 100
dat$meta  <- as.factor(dat$meta)
dat$imag  <- as.factor(dat$imag)

vignette_cache_load(cached_fits)

## ----precompute-models, include = FALSE, eval = [encoded data, 9950 characters]
# library("metafor")
# library("RoBMA")
# 
# data("dat.bangertdrowns2004", package = "metadat")
# dat <- dat.bangertdrowns2004
# dat$ni100 <- dat$ni / 100
# dat$meta  <- as.factor(dat$meta)
# dat$imag  <- as.factor(dat$imag)
# 
# fit1_brma <- brma(
#   yi = yi, vi = vi, measure = "SMD",
#   data = dat, seed = 1
# )
# 
# fit2_brma <- brma(
#   yi = yi, vi = vi, measure = "SMD",
#   mods  = ~ ni100,
#   scale = ~ ni100,
#   data = dat, seed = 1
# )
# 
# fit3_brma <- brma(
#   yi = yi, vi = vi, measure = "SMD",
#   mods  = ~ ni100 + meta,
#   scale = ~ ni100 + imag,
#   data = dat, seed = 1
# )
# 
# vignette_cache_save(cached_fits)

## [encoded data, 9997 characters]
data("dat.bangertdrowns2004", package = "metadat")
dat <- dat.bangertdrowns2004
dat$ni100 <- dat$ni / 100
dat$meta  <- as.factor(dat$meta)
dat$imag  <- as.factor(dat$imag)

## [encoded data, 9997 characters]
fit1_metafor <- metafor::rma(yi, vi, data = dat)
fit1_metafor

## ----fit1-brma, eval = [encoded data, 9975 characters]
# fit1_brma <- brma(
#   yi = yi, vi = vi, measure = "SMD",
#   data = dat, seed = 1
# )

## [encoded data, 9997 characters]
summary(fit1_brma, include_mcmc_diagnostics = FALSE)

## [encoded data, 9997 characters]
fit2_metafor <- suppressWarnings(metafor::rma(
  yi, vi,
  mods  = ~ ni100,
  scale = ~ ni100,
  data = dat
))
fit2_metafor

## ----fit2-brma, eval = [encoded data, 9975 characters]
# fit2_brma <- brma(
#   yi = yi, vi = vi, measure = "SMD",
#   mods  = ~ ni100,
#   scale = ~ ni100,
#   data = dat, seed = 1
# )

## [encoded data, 9997 characters]
summary(fit2_brma, include_mcmc_diagnostics = FALSE)

## ----regplot-pi, fig.width = 7, fig.height = 3.[encoded data, 9951 characters]
par(mfrow = c(1, 2), mar = c(4, 4, 2, 1), cex.axis = 0.8, cex.lab = 0.8, cex.main = 0.75)
metafor::regplot(fit2_metafor, mod = "ni100", pi = TRUE,
                 main = "metafor", xlim = c(0, 6), ylim = c(-1, 2))
regplot(fit2_brma, mod = "ni100", pi = TRUE,
        main = "RoBMA",   xlim = c(0, 6), ylim = c(-1, 2))

## [encoded data, 9997 characters]
fit3_metafor <- suppressWarnings(metafor::rma(
  yi, vi,
  mods  = ~ ni100 + meta,
  scale = ~ ni100 + imag,
  data = dat
))
fit3_metafor

## ----fit3-brma, eval = [encoded data, 9975 characters]
# fit3_brma <- brma(
#   yi = yi, vi = vi, measure = "SMD",
#   mods  = ~ ni100 + meta,
#   scale = ~ ni100 + imag,
#   data = dat, seed = 1
# )

## [encoded data, 9997 characters]
summary(fit3_brma, include_mcmc_diagnostics = FALSE)

## ----scale-posterior, fig.width = 5, fig.height = 3.[encoded data, 9946 characters]
par(mar = c(4, 4, 1, 1))
plot(fit2_brma, parameter_scale = "ni100", prior = TRUE, xlim = c(-2, 2))

## ----vignette-nowrap-options, include = FALSE, eval = [encoded data, 9944 characters]
options(width = 10000)

## ----setup, include = [encoded data, 9976 characters]
source("vignette-cache.R", local = knitr::knit_global())

cached_fits <- vignette_cache(
  name     = "v36-zplot",
  objects  = c(
    "fit_RE_Weingarten2018",
    "fit_RoBMA_Weingarten2018",
    "fit_RE_Hoppen2025",
    "fit_RoBMA_Hoppen2025",
    "fit_RE_Wang2025",
    "fit_RoBMA_Wang2025",
    "fit_RE_ManyLabs16",
    "fit_RoBMA_ManyLabs16"
  ),
  packages = c("metafor")
)

knitr::opts_chunk$set(
  collapse   = TRUE,
  comment    = "#>",
  eval       = vignette_cache_eval(cached_fits),
  message    = FALSE,
  warning    = FALSE,
  fig.width  = 7,
  fig.height = 5,
  dev        = "png",
  fig.retina = 3
)
if (.Platform$OS.type == "windows") {
  knitr::opts_chunk$set(dev.args = list(type = "cairo"))
}

## ----load-models, include = [encoded data, 9970 characters]
library("RoBMA")
vignette_cache_load(cached_fits)

## ----precompute-models, include = FALSE, eval = [encoded data, 9950 characters]
# library("RoBMA")
# 
# # Ease of retrieval - Weingarten2018
# data("Weingarten2018", package = "RoBMA")
# Weingarten2018 <- Weingarten2018[Weingarten2018$standard_paradigm & Weingarten2018$proximal_dataset, ]
# Weingarten2018_z <- metafor::escalc(
#   ri = r_xy, ni = round(N), measure = "ZCOR", data = Weingarten2018
# )
# fit_RE_Weingarten2018 <- brma(
#   yi = yi, vi = vi, measure = "ZCOR",
#   cluster = paper_id, data = Weingarten2018_z,
#   sample = 10000, burnin = 10000, adapt = 10000,
#   chains = 5, parallel = TRUE, seed = 1
# )
# fit_RoBMA_Weingarten2018 <- RoBMA(
#   yi = yi, vi = vi, measure = "ZCOR",
#   cluster = paper_id, data = Weingarten2018_z,
#   sample = 10000, burnin = 10000, adapt = 10000,
#   chains = 5, parallel = TRUE, seed = 1
# )
# 
# # Social comparison - Hoppen2025
# data("Hoppen2025", package = "RoBMA")
# fit_RE_Hoppen2025 <- brma(
#   yi = d, sei = sqrt(v), measure = "SMD",
#   data = Hoppen2025,
#   sample = 10000, burnin = 5000, adapt = 5000,
#   chains = 5, parallel = TRUE, seed = 1
# )
# fit_RoBMA_Hoppen2025 <- RoBMA(
#   yi = d, sei = sqrt(v), measure = "SMD",
#   data = Hoppen2025,
#   sample = 10000, burnin = 5000, adapt = 5000,
#   chains = 5, parallel = TRUE, seed = 1
# )
# 
# # ChatGPT - Wang2025
# data("Wang2025", package = "RoBMA")
# Wang2025 <- Wang2025[Wang2025$Learning_effect == "Learning performance", ]
# fit_RE_Wang2025 <- brma(
#   yi = g, sei = se, measure = "SMD",
#   data = Wang2025,
#   sample = 10000, burnin = 5000, adapt = 5000,
#   chains = 5, parallel = TRUE, seed = 1
# )
# fit_RoBMA_Wang2025 <- RoBMA(
#   yi = g, sei = se, measure = "SMD",
#   data = Wang2025,
#   sample = 10000, burnin = 5000, adapt = 5000,
#   chains = 5, parallel = TRUE, seed = 1
# )
# 
# # Many Labs 2 - ManyLabs16
# data("ManyLabs16", package = "RoBMA")
# fit_RE_ManyLabs16 <- brma(
#   yi = y, sei = se, measure = "SMD",
#   data = ManyLabs16,
#   sample = 10000, burnin = 5000, adapt = 5000,
#   chains = 5, parallel = TRUE, seed = 1
# )
# fit_RoBMA_ManyLabs16 <- RoBMA(
#   yi = y, sei = se, measure = "SMD",
#   data = ManyLabs16,
#   sample = 10000, burnin = 5000, adapt = 5000,
#   chains = 5, parallel = TRUE, seed = 1
# )
# 
# vignette_cache_save(cached_fits)

## ----message = [encoded data, 9983 characters]
library("RoBMA")

## [encoded data, 9997 characters]
data("Weingarten2018", package = "RoBMA")
Weingarten2018 <- Weingarten2018[Weingarten2018$standard_paradigm & Weingarten2018$proximal_dataset, ]
head(Weingarten2018)

## ----eval = [encoded data, 9986 characters]
# Weingarten2018_z <- metafor::escalc(
#   ri = r_xy, ni = round(N), measure = "ZCOR", data = Weingarten2018
# )
# 
# fit_RE_Weingarten2018 <- brma(
#   yi = yi, vi = vi, measure = "ZCOR",
#   cluster = paper_id, data = Weingarten2018_z,
#   sample = 10000, burnin = 10000, adapt = 10000,
#   chains = 5, parallel = TRUE, seed = 1)
# 
# fit_RoBMA_Weingarten2018 <- RoBMA(
#   yi = yi, vi = vi, measure = "ZCOR",
#   cluster = paper_id, data = Weingarten2018_z,
#   sample = 10000, burnin = 10000, adapt = 10000,
#   chains = 5, parallel = TRUE, seed = 1)

## [encoded data, 9997 characters]
summary(fit_RE_Weingarten2018)

summary(fit_RoBMA_Weingarten2018)

## ----fig.cap="Ease-of-Retrieval Effect: Model Fit Assessment"[encoded data, 9937 characters]
hist(as_zplot(fit_RoBMA_Weingarten2018), from = -3, to = 6, by = 0.25)

lines(as_zplot(fit_RE_Weingarten2018),    from = -3, to = 6, col = "black", lty = 2, lwd = 2)
lines(as_zplot(fit_RoBMA_Weingarten2018), from = -3, to = 6, col = "blue",  lty = 2, lwd = 2)

legend(
  "topright",
  legend = c("Random-Effects", "RoBMA"),
  col = c("black", "blue"),
  lty = 2, lwd = 2)

## ----fig.cap="Ease-of-Retrieval Effect: Extrapolation Analysis"[encoded data, 9935 characters]
zplot(fit_RoBMA_Weingarten2018, from = -3, to = 6, by.hist = 0.25)

## [encoded data, 9997 characters]
summary(as_zplot(fit_RoBMA_Weingarten2018))

## [encoded data, 9997 characters]
data("Hoppen2025", package = "RoBMA")
head(Hoppen2025)

## ----eval = [encoded data, 9986 characters]
# fit_RE_Hoppen2025 <- brma(
#   yi = d, sei = sqrt(v), measure = "SMD",
#   data = Hoppen2025,
#   sample = 10000, burnin = 5000, adapt = 5000,
#   chains = 5, parallel = TRUE, seed = 1)
# 
# fit_RoBMA_Hoppen2025 <- RoBMA(
#   yi = d, sei = sqrt(v), measure = "SMD",
#   data = Hoppen2025,
#   sample = 10000, burnin = 5000, adapt = 5000,
#   chains = 5, parallel = TRUE, seed = 1)

## [encoded data, 9997 characters]
summary(fit_RE_Hoppen2025)

## [encoded data, 9997 characters]
summary(fit_RoBMA_Hoppen2025)

## ----fig.cap="Social Comparison: Model Fit Assessment"[encoded data, 9944 characters]
hist(as_zplot(fit_RoBMA_Hoppen2025))

lines(as_zplot(fit_RE_Hoppen2025),    col = "black", lty = 2, lwd = 2)
lines(as_zplot(fit_RoBMA_Hoppen2025), col = "blue",  lty = 2, lwd = 2)

legend(
  "topright",
  legend = c("Random-Effects", "RoBMA"),
  col = c("black", "blue"),
  lty = 2, lwd = 2)

## ----fig.cap="Social Comparison: Extrapolation Analysis"[encoded data, 9942 characters]
zplot(fit_RoBMA_Hoppen2025)

## [encoded data, 9997 characters]
summary(as_zplot(fit_RoBMA_Hoppen2025))

## [encoded data, 9997 characters]
data("Wang2025", package = "RoBMA")
Wang2025 <- Wang2025[Wang2025$Learning_effect == "Learning performance", ]
head(Wang2025)

## ----eval = [encoded data, 9986 characters]
# fit_RE_Wang2025 <- brma(
#   yi = g, sei = se, measure = "SMD",
#   data = Wang2025,
#   sample = 10000, burnin = 5000, adapt = 5000,
#   chains = 5, parallel = TRUE, seed = 1)
# fit_RoBMA_Wang2025 <- RoBMA(
#   yi = g, sei = se, measure = "SMD",
#   data = Wang2025,
#   sample = 10000, burnin = 5000, adapt = 5000,
#   chains = 5, parallel = TRUE, seed = 1)

## ----fig.cap="ChatGPT: Model Fit Assessment"[encoded data, 9954 characters]
hist(as_zplot(fit_RoBMA_Wang2025), from = -2, to = 8)
lines(as_zplot(fit_RE_Wang2025),    col = "black", lty = 2, lwd = 2, from = -2, to = 8)
lines(as_zplot(fit_RoBMA_Wang2025), col = "blue",  lty = 2, lwd = 2, from = -2, to = 8)
legend(
  "topright",
  legend = c("Random-Effects", "RoBMA"),
  col = c("black", "blue"),
  lty = 2, lwd = 2)

## ----fig.cap="ChatGPT: Extrapolation Analysis"[encoded data, 9952 characters]
zplot(fit_RoBMA_Wang2025, from = -2, to = 8)

## [encoded data, 9997 characters]
summary(fit_RE_Wang2025)
summary(fit_RoBMA_Wang2025)

## [encoded data, 9997 characters]
summary(as_zplot(fit_RoBMA_Wang2025))

## [encoded data, 9997 characters]
data("ManyLabs16", package = "RoBMA")
head(ManyLabs16)

## ----eval = [encoded data, 9986 characters]
# fit_RE_ManyLabs16 <- brma(
#   yi = y, sei = se, measure = "SMD",
#   data = ManyLabs16,
#   sample = 10000, burnin = 5000, adapt = 5000,
#   chains = 5, parallel = TRUE, seed = 1)
# fit_RoBMA_ManyLabs16 <- RoBMA(
#   yi = y, sei = se, measure = "SMD",
#   data = ManyLabs16,
#   sample = 10000, burnin = 5000, adapt = 5000,
#   chains = 5, parallel = TRUE, seed = 1)

## ----fig.cap="Framing Effects: Model Fit Assessment"[encoded data, 9946 characters]
hist(as_zplot(fit_RoBMA_ManyLabs16))
lines(as_zplot(fit_RE_ManyLabs16),    col = "black", lty = 2, lwd = 2)
lines(as_zplot(fit_RoBMA_ManyLabs16), col = "blue",  lty = 2, lwd = 2)
legend(
  "topleft",
  legend = c("Random-Effects", "RoBMA"),
  col = c("black", "blue"),
  lty = 2, lwd = 2)

## ----fig.cap="Framing Effects: Extrapolation Analysis"[encoded data, 9944 characters]
zplot(fit_RoBMA_ManyLabs16)

## ----eval = [encoded data, 9986 characters]
# summary(fit_RE_ManyLabs16)
# summary(fit_RoBMA_ManyLabs16)

## [encoded data, 9997 characters]
summary(as_zplot(fit_RoBMA_ManyLabs16))

## ----vignette-nowrap-options, include = FALSE, eval = [encoded data, 9944 characters]
options(width = 10000)

## ----setup, include = [encoded data, 9976 characters]
source("vignette-cache.R", local = knitr::knit_global())

cached_fits <- vignette_cache(
  name     = "v11-metafor-parity-publication-bias",
  objects  = c(
    "fit_PET_brma", "fit_PEESE_brma", "fit_selmodel_brma",
    "fit_PET_brma_flex", "fit_PEESE_brma_flex", "fit_selmodel_brma_flex"
  ),
  packages = c("metafor", "metadat")
)

knitr::opts_chunk$set(
  collapse    = TRUE,
  comment     = "#>",
  eval        = vignette_cache_eval(cached_fits),
  message     = FALSE,
  warning     = FALSE,
  fig.width   = 7,
  fig.height  = 3.5,
  dev         = "png",
  fig.retina  = 3
)
if (.Platform$OS.type == "windows") {
  knitr::opts_chunk$set(dev.args = list(type = "cairo"))
}

## ----load-models, include = [encoded data, 9970 characters]
library("metafor")
library("RoBMA")

data("dat.lehmann2018", package = "metadat")
dat <- dat.lehmann2018

vignette_cache_load(cached_fits)

## ----precompute-models, include = FALSE, eval = [encoded data, 9950 characters]
# library("metafor")
# library("RoBMA")
# 
# data("dat.lehmann2018", package = "metadat")
# dat <- dat.lehmann2018
# 
# fit_PET_brma <- bPET(
#   yi = yi, vi = vi, measure = "SMD",
#   data = dat, seed = 1
# )
# 
# fit_PEESE_brma <- bPEESE(
#   yi = yi, vi = vi, measure = "SMD",
#   data = dat, seed = 1
# )
# 
# fit_selmodel_brma <- bselmodel(
#   yi = yi, vi = vi, measure = "SMD",
#   data = dat, seed = 1
# )
# 
# fit_PET_brma_flex <- bPET(
#   yi = yi, vi = vi, measure = "SMD",
#   prior_bias = prior_PET(
#     distribution = "cauchy",
#     parameters = list(location = 0, scale = 1),
#     truncation = list(lower = -Inf, upper = Inf)
#   ),
#   data = dat, seed = 1
# )
# 
# fit_PEESE_brma_flex <- bPEESE(
#   yi = yi, vi = vi, measure = "SMD",
#   prior_bias = prior_PEESE(
#     distribution = "cauchy",
#     parameters = list(location = 0, scale = 5),
#     truncation = list(lower = -Inf, upper = Inf)
#   ),
#   data = dat, seed = 1
# )
# 
# fit_selmodel_brma_flex <- bselmodel(
#   yi = yi, vi = vi, measure = "SMD",
#   prior_bias = prior_weightfunction(
#     "one-sided",
#     steps   = c(0.025),
#     weights = wf_independent(
#       prior("normal", parameters = list(mean = 0, sd = 0.5)),
#       scale = "log_omega"
#     )
#   ),
#   data = dat, seed = 1
# )
# 
# vignette_cache_save(cached_fits)

## [encoded data, 9997 characters]
data("dat.lehmann2018", package = "metadat")
dat <- dat.lehmann2018

## [encoded data, 9997 characters]
fit_PET_metafor <- metafor::rma(yi, vi, mods = ~ sqrt(vi), data = dat)
fit_PET_metafor

## ----fit-PET-brma, eval = [encoded data, 9972 characters]
# fit_PET_brma <- bPET(
#   yi = yi, vi = vi, measure = "SMD",
#   data = dat, seed = 1
# )

## [encoded data, 9997 characters]
summary(fit_PET_brma, include_mcmc_diagnostics = FALSE)

## ----PET-posterior, fig.width = 7, fig.height = 3.[encoded data, 9948 characters]
par(mfrow = c(1, 2), mar = c(4, 4, 2, 1), cex.axis = 0.8, cex.lab = 0.8, cex.main = 0.75)
plot(fit_PET_brma,  parameter = "PET", prior = TRUE, main = "PET coefficient", xlim = c(0, 4))
plot_pet_peese(fit_PET_brma, prior = TRUE, main = "PET regression")

## [encoded data, 9997 characters]
fit_PEESE_metafor <- metafor::rma(yi, vi, mods = ~ vi, data = dat)
fit_PEESE_metafor

## ----fit-PEESE-brma, eval = [encoded data, 9970 characters]
# fit_PEESE_brma <- bPEESE(
#   yi = yi, vi = vi, measure = "SMD",
#   data = dat, seed = 1
# )

## [encoded data, 9997 characters]
summary(fit_PEESE_brma, include_mcmc_diagnostics = FALSE)

## ----PEESE-posterior, fig.width = 7, fig.height = 3.[encoded data, 9946 characters]
par(mfrow = c(1, 2), mar = c(4, 4, 2, 1), cex.axis = 0.8, cex.lab = 0.8, cex.main = 0.75)
plot(fit_PEESE_brma, parameter = "PEESE", prior = TRUE, main = "PEESE coefficient", xlim = c(0, 8))
plot_pet_peese(fit_PEESE_brma, prior = TRUE, main = "PEESE regression")

## [encoded data, 9997 characters]
fit_rma_metafor      <- metafor::rma(yi, vi, data = dat, method = "ML")
fit_selmodel_metafor <- metafor::selmodel(fit_rma_metafor, type = "stepfun", steps = .025, decreasing = TRUE)
fit_selmodel_metafor

## ----fit-selmodel-brma, eval = [encoded data, 9967 characters]
# fit_selmodel_brma <- bselmodel(
#   yi = yi, vi = vi, measure = "SMD",
#   data = dat, seed = 1
# )

## [encoded data, 9997 characters]
summary(fit_selmodel_brma, include_mcmc_diagnostics = FALSE)

## ----selmodel-omega, fig.width = 7, fig.height = 3.[encoded data, 9947 characters]
par(mfrow = c(1, 2), mar = c(4, 4, 2, 1), cex.axis = 0.8, cex.lab = 0.8, cex.main = 0.75)
plot(fit_selmodel_brma, parameter = "omega", prior = TRUE, xlim = c(0, 1))

## ----selmodel-weightfunction, fig.width = 4.5, fig.height = 3.[encoded data, 9936 characters]
par(mar = c(4, 4, 2, 1), cex.axis = 0.8, cex.lab = 0.8, cex.main = 0.75)
plot_weightfunction(fit_selmodel_brma, prior = TRUE, main = "Weight-function")

## ----funnel-plots, fig.width = 7, fig.height = 3.[encoded data, 9949 characters]
par(mfrow = c(1, 3), mar = c(4, 4, 2, 1), cex.axis = 0.8, cex.lab = 0.8, cex.main = 0.75)
xlim <- c(-2, 2.5); ylim <- c(0, 0.8)
funnel(fit_PET_brma,      main = "PET",       xlim = xlim, ylim = ylim)
funnel(fit_PEESE_brma,    main = "PEESE",     xlim = xlim, ylim = ylim)
funnel(fit_selmodel_brma, main = "Sel. Mod.", xlim = xlim, ylim = ylim)

## ----fit-PET-brma-flex, eval = [encoded data, 9967 characters]
# fit_PET_brma_flex <- bPET(
#   yi = yi, vi = vi, measure = "SMD",
#   prior_bias = prior_PET(
#     distribution = "cauchy",
#     parameters = list(location = 0, scale = 1),
#     truncation = list(lower = -Inf, upper = Inf)
#   ),
#   data = dat, seed = 1
# )

## [encoded data, 9997 characters]
summary(fit_PET_brma_flex, include_mcmc_diagnostics = FALSE)

## ----PET-flex-comparison, fig.width = 7, fig.height = 3.[encoded data, 9942 characters]
par(mfrow = c(1, 2), mar = c(4, 4, 2, 1), cex.axis = 0.8, cex.lab = 0.8, cex.main = 0.85)
plot(fit_PET_brma,      parameter = "PET", prior = TRUE,
  main = "PET (default)",     xlim = c(-2, 4))
abline(v = 0, lty = 2, col = "grey50")
plot(fit_PET_brma_flex, parameter = "PET", prior = TRUE,
  main = "PET (untruncated)", xlim = c(-2, 4))
abline(v = 0, lty = 2, col = "grey50")

## ----fit-PEESE-brma-flex, eval = [encoded data, 9965 characters]
# fit_PEESE_brma_flex <- bPEESE(
#   yi = yi, vi = vi, measure = "SMD",
#   prior_bias = prior_PEESE(
#     distribution = "cauchy",
#     parameters = list(location = 0, scale = 5),
#     truncation = list(lower = -Inf, upper = Inf)
#   ),
#   data = dat, seed = 1
# )

## [encoded data, 9997 characters]
summary(fit_PEESE_brma_flex, include_mcmc_diagnostics = FALSE)

## ----PEESE-flex-comparison, fig.width = 7, fig.height = 3.[encoded data, 9940 characters]
par(mfrow = c(1, 2), mar = c(4, 4, 2, 1), cex.axis = 0.8, cex.lab = 0.8, cex.main = 0.85)
plot(fit_PEESE_brma, parameter = "PEESE", prior = TRUE, main = "PEESE (default)", xlim = c(-5, 10))
abline(v = 0, lty = 2, col = "grey50")
plot(fit_PEESE_brma_flex, parameter = "PEESE", prior = TRUE, main = "PEESE (untruncated)", xlim = c(-5, 10))
abline(v = 0, lty = 2, col = "grey50")

## ----fit-selmodel-brma-flex, eval = [encoded data, 9962 characters]
# fit_selmodel_brma_flex <- bselmodel(
#   yi = yi, vi = vi, measure = "SMD",
#   prior_bias = prior_weightfunction(
#     "one-sided",
#     steps   = c(0.025),
#     weights = wf_independent(
#       prior("normal", parameters = list(mean = 0, sd = 0.5)),
#       scale = "log_omega"
#     )
#   ),
#   data = dat, seed = 1
# )

## [encoded data, 9997 characters]
summary(fit_selmodel_brma_flex, include_mcmc_diagnostics = FALSE)

## ----selmodel-flex-comparison, fig.width = 7, fig.height = 3.[encoded data, 9937 characters]
par(mfrow = c(1, 2), mar = c(4, 4, 2, 1), cex.axis = 0.8, cex.lab = 0.8, cex.main = 0.85)
plot(fit_selmodel_brma, parameter = "omega", prior = TRUE, main = "omega (default)", xlim = c(0, 3))
abline(v = 1, lty = 2, col = "grey50")
plot(fit_selmodel_brma_flex, parameter = "omega", prior = TRUE, main = "omega (log_omega)", xlim = c(0, 3))
abline(v = 1, lty = 2, col = "grey50")

## ----vignette-nowrap-options, include = FALSE, eval = [encoded data, 9944 characters]
options(width = 10000)

## ----setup, include = [encoded data, 9976 characters]
source("vignette-cache.R", local = knitr::knit_global())

cached_fits <- vignette_cache(
  name     = "v02-bayesian-meta-analysis",
  objects  = c("fit1_brma", "fit2_brma", "fit3_brma"),
  packages = c("metafor", "metadat")
)

knitr::opts_chunk$set(
  collapse    = TRUE,
  comment     = "#>",
  eval        = vignette_cache_eval(cached_fits),
  message     = FALSE,
  warning     = FALSE,
  fig.width   = 7,
  fig.height  = 3.5,
  dev         = "png",
  fig.retina  = 3
)
if (.Platform$OS.type == "windows") {
  knitr::opts_chunk$set(dev.args = list(type = "cairo"))
}

## ----load-models, include = [encoded data, 9970 characters]
library("metafor")
library("RoBMA")

data("dat.bcg", package = "metadat")

dat <- metafor::escalc(
  ai = tpos, bi = tneg, ci = cpos, di = cneg,
  measure = "RR", data = dat.bcg
)

vignette_cache_load(cached_fits)

## ----precompute-models, include = FALSE, eval = [encoded data, 9950 characters]
# library("metafor")
# library("RoBMA")
# 
# data("dat.bcg", package = "metadat")
# 
# dat <- metafor::escalc(
#   ai = tpos, bi = tneg, ci = cpos, di = cneg,
#   measure = "RR", data = dat.bcg
# )
# 
# fit1_brma <- brma(
#   yi = yi, vi = vi, measure = "RR",
#   data = dat, seed = 1
# )
# fit1_brma <- add_loo(fit1_brma)
# fit1_brma <- add_marglik(fit1_brma)
# 
# fit2_brma <- brma(
#   yi = yi, vi = vi, measure = "RR",
#   mods = ~ ablat,
#   data = dat, seed = 1
# )
# fit2_brma <- add_loo(fit2_brma)
# fit2_brma <- add_marglik(fit2_brma)
# 
# fit3_brma <- brma(
#   yi = yi, vi = vi, measure = "RR",
#   mods = ~ ablat + alloc,
#   data = dat, seed = 1
# )
# fit3_brma <- add_loo(fit3_brma)
# fit3_brma <- add_marglik(fit3_brma)
# 
# vignette_cache_save(cached_fits)

## [encoded data, 9997 characters]
data("dat.bcg", package = "metadat")

dat <- metafor::escalc(
  ai = tpos, bi = tneg, ci = cpos, di = cneg,
  measure = "RR", data = dat.bcg
)

## [encoded data, 9997 characters]
fit1_metafor <- metafor::rma(yi, vi, data = dat, method = "REML")
fit1_metafor

## ----fit1-brma, eval = [encoded data, 9975 characters]
# fit1_brma <- brma(
#   yi = yi, vi = vi, measure = "RR",
#   data = dat, seed = 1
# )

## [encoded data, 9997 characters]
summary(fit1_brma, include_mcmc_diagnostics = FALSE)

## [encoded data, 9997 characters]
pooled_effect(fit1_brma, transform = "EXP")

## ----fit1-brma-plot, fig.width = 5, fig.height = 3.[encoded data, 9947 characters]
par(mar = c(4, 4, 1, 1))
plot(fit1_brma, parameter = "mu", prior = TRUE, xlim = c(-3, 3))

## [encoded data, 9997 characters]
confint(fit1_metafor)

## [encoded data, 9997 characters]
summary_heterogeneity(fit1_brma)

## ----funnel-plot, fig.width = 7, fig.height = 3.[encoded data, 9950 characters]
par(mfrow = c(1, 2), mar = c(4, 4, 2, 1), cex.axis = 0.8, cex.lab = 0.8, cex.main = 0.75)
metafor::funnel(fit1_metafor, main = "metafor", xlim = c(-2, 1), ylim = c(0, 0.8))
funnel(fit1_brma, main = "RoBMA", xlim = c(-2, 1), ylim = c(0, 0.8), sampling_heterogeneity = FALSE)

## [encoded data, 9997 characters]
fit2_metafor <- metafor::rma(yi, vi, mods = ~ ablat, data = dat)
fit2_metafor

## ----fit2-brma, eval = [encoded data, 9975 characters]
# fit2_brma <- brma(
#   yi = yi, vi = vi, measure = "RR",
#   mods = ~ ablat,
#   data = dat, seed = 1
# )

## [encoded data, 9997 characters]
summary(fit2_brma, include_mcmc_diagnostics = FALSE)

## ----regplot-continuous, fig.width = 7, fig.height = 3.[encoded data, 9943 characters]
par(mfrow = c(1, 2), mar = c(4, 4, 2, 1), cex.axis = 0.8, cex.lab = 0.8, cex.main = 0.75)
metafor::regplot(fit2_metafor, main = "metafor", xlim = c(10, 60), ylim = c(-2, 0.5))
regplot(fit2_brma, main = "RoBMA", xlim = c(10, 60), ylim = c(-2, 0.5))

## [encoded data, 9997 characters]
new_data <- data.frame(ablat = c(15, 35, 55))
predict(fit2_brma, newdata = new_data, type = "terms")

## ----add-marglik, eval = [encoded data, 9973 characters]
# fit1_brma <- add_marglik(fit1_brma)
# fit2_brma <- add_marglik(fit2_brma)

## [encoded data, 9997 characters]
bf(fit2_brma, fit1_brma)

## [encoded data, 9997 characters]
fit3_metafor <- metafor::rma(yi, vi, mods = ~ ablat + alloc, data = dat)

## ----fit3-brma, eval = [encoded data, 9975 characters]
# fit3_brma <- brma(
#   yi = yi, vi = vi, measure = "RR",
#   mods = ~ ablat + alloc,
#   data = dat, seed = 1
# )

## [encoded data, 9997 characters]
summary(fit3_brma, include_mcmc_diagnostics = FALSE)

## ----regplot-categorical, fig.width = 7, fig.height = 3.[encoded data, 9942 characters]
par(mfrow = c(1, 2), mar = c(4, 4, 2, 1), cex.axis = 0.8, cex.lab = 0.8, cex.main = 0.75)
metafor::regplot(fit3_metafor, mod = "allocrandom", main = "metafor: random dummy", ylim = c(-2, 0.5))
regplot(fit3_brma, mod = "alloc", main = "RoBMA: alloc factor", ylim = c(-2, 0.5))

## [encoded data, 9997 characters]
emm <- marginal_means(fit3_brma)
summary(emm)

## ----marginal-means-plot, fig.width = 4.5, fig.height = 3.[encoded data, 9940 characters]
par(mar = c(4, 4, 1, 1), cex.axis = 0.8, cex.lab = 0.8, cex.main = 0.75)
plot(emm, parameter = "alloc", xlim = c(-2, 0.5), lwd = 2)

## ----add-loo, eval = [encoded data, 9977 characters]
# fit3_brma <- add_loo(fit3_brma)

## [encoded data, 9997 characters]
head(rstudent(fit3_metafor)$z)
head(rstudent(fit3_brma)$z)

## ----qqplot, fig.width = 7, fig.height = 3.[encoded data, 9955 characters]
par(mfrow = c(1, 2), mar = c(4, 4, 2, 1), cex.axis = 0.8, cex.lab = 0.8, cex.main = 0.75)
qqnorm(fit3_metafor, main = "metafor: studentized", xlim = c(-1.5, 1.5), ylim = c(-3, 3))
qqnorm(fit3_brma, main = "RoBMA: LOO-PIT", xlim = c(-1.5, 1.5), ylim = c(-3, 3))

## [encoded data, 9997 characters]
inf_brma <- influence(fit3_brma)
head(inf_brma$inf)

## ----vignette-nowrap-options, include = FALSE, eval = [encoded data, 9944 characters]
options(width = 10000)

## ----setup, include = [encoded data, 9976 characters]
knitr::opts_chunk$set(
  collapse   = TRUE,
  comment    = "#>",
  message    = FALSE,
  warning    = FALSE,
  fig.width  = 6,
  fig.height = 4
)

library(RoBMA)

has_metafor <- requireNamespace("metafor", quietly = TRUE)
has_bcg     <- has_metafor && requireNamespace("metadat", quietly = TRUE)

## ----bcg-data, eval = [encoded data, 9976 characters]
data("dat.bcg", package = "metadat")

bcg <- metafor::escalc(
  ai = tpos, bi = tneg, ci = cpos, di = cneg,
  measure = "RR", data = dat.bcg
)

## ----bcg-default-priors, eval = [encoded data, 9966 characters]
bcg_priors <- brma(
  yi = yi, vi = vi, measure = "RR",
  data = bcg, only_priors = TRUE
)

plot_prior(bcg_priors,  parameter = "mu")
plot_prior(bcg_priors,  parameter = "tau")
print_prior(bcg_priors)

## ----bcg-rescale, eval = [encoded data, 9973 characters]
bcg_wider_priors <- brma(
  yi = yi, vi = vi, measure = "RR",
  rescale_priors = 2,
  data = bcg, only_priors = TRUE
)

print_prior(bcg_wider_priors)

## ----Kroupova2021-data, eval = [encoded data, 9967 characters]
data("Kroupova2021", package = "RoBMA")

Kroupova2021 <- metafor::escalc( 
  ri = r, ni = sample_size, measure = "ZCOR",
  data = Kroupova2021
)

## ----Kroupova2021-priors, eval = [encoded data, 9965 characters]
Kroupova2021_priors <- brma(
  yi = yi, vi = vi, measure = "ZCOR",
  data = Kroupova2021, only_priors = TRUE
)

print_prior(Kroupova2021_priors)

## [encoded data, 9997 characters]
data("Havrankova2025", package = "RoBMA")
head(Havrankova2025)

## [encoded data, 9997 characters]
Havrankova_priors <- brma(
  yi = y, sei = se, ni = N, measure = "GEN",
  data = Havrankova2025, only_priors = TRUE
)

print_prior(Havrankova_priors)

## [encoded data, 9997 characters]
Havrankova_UISD <- estimate_unit_information_sd(
  sei = Havrankova2025$se,
  ni  = Havrankova2025$N
)
Havrankova_UISD

## [encoded data, 9997 characters]
Havrankova_manual_priors <- brma(
  yi = y, sei = se, measure = "GEN",
  prior_unit_information_sd = Havrankova_UISD,
  data = Havrankova2025, only_priors = TRUE
)

print_prior(Havrankova_manual_priors)

## [encoded data, 9997 characters]
Havrankova_treatment <- brma(
  yi = y, sei = se, ni = N, measure = "GEN",
  mods = ~ facing_customer + N,
  set_contrast_factor_predictors = "treatment",
  data = Havrankova2025, only_priors = TRUE
)

print_prior(Havrankova_treatment, parameter_mods = "facing_customer")

## [encoded data, 9997 characters]
Havrankova_meandif <- brma(
  yi = y, sei = se, ni = N, measure = "GEN",
  mods = ~ facing_customer + N,
  set_contrast_factor_predictors = "meandif",
  data = Havrankova2025, only_priors = TRUE
)

print_prior(Havrankova_meandif, parameter_mods = "facing_customer")

## [encoded data, 9997 characters]
plot_prior(Havrankova_meandif, parameter_mods = "N")
plot_prior(Havrankova_meandif, parameter_mods = "N", 
  standardized_coefficients = FALSE)

## [encoded data, 9997 characters]
Havrankova_per_mod <- brma(
  yi = y, sei = se, ni = N, measure = "GEN",
  mods = ~ facing_customer + N,
  prior_mods = list(
    facing_customer = prior_factor(
      distribution = "mnormal",
      parameters = list(mean = 0, sd = 5),
      contrast   = "meandif"
    )
  ),
  data = Havrankova2025, only_priors = TRUE
)

print_prior(Havrankova_per_mod)

## ----bcg-informed, eval = [encoded data, 9972 characters]
bcg_informed_priors <- brma(
  yi = yi, vi = vi, measure = "RR",
  prior_informed_field    = "medicine",
  prior_informed_subfield = "Cochrane",
  data = bcg, only_priors = TRUE
)

print_prior(bcg_informed_priors)

## ----psychology-informed, eval = [encoded data, 9965 characters]
psychology_priors <- brma(
  yi = yi, vi = vi, measure = "ZCOR",
  prior_effect        = prior_informed("Oosterwijk"),
  prior_heterogeneity = prior_informed("van Erp", parameter = "heterogeneity"),
  data = Kroupova2021, only_priors = TRUE
)

print_prior(psychology_priors)

## ----custom-effect-priors, eval = [encoded data, 9964 characters]
bcg_custom_priors <- brma(
  yi = yi, vi = vi, measure = "RR",
  prior_effect        = prior(
    distribution = "normal",
    parameters = list(mean = 0, sd = 0.5)
  ),
  prior_heterogeneity = prior(
    distribution = "normal",
    parameters = list(mean = 0, sd = 0.25),
    truncation = list(lower = 0, upper = Inf)
  ),
  data = bcg, only_priors = TRUE
)

print_prior(bcg_custom_priors)

## ----bma-priors, eval = [encoded data, 9974 characters]
bcg_bma_priors <- BMA(
  yi = yi, vi = vi, measure = "RR",
  data = bcg, only_priors = TRUE
)

print_prior(bcg_bma_priors)

## ----bma-no-null, eval = [encoded data, 9973 characters]
bcg_estimation_priors <- BMA(
  yi = yi, vi = vi, measure = "RR",
  prior_effect_null = NULL,
  data = bcg, only_priors = TRUE
)

print_prior(bcg_estimation_priors, parameter = "mu")

## ----vignette-nowrap-options, include = FALSE, eval = [encoded data, 9944 characters]
options(width = 10000)

## ----setup, include = [encoded data, 9976 characters]
source("vignette-cache.R", local = knitr::knit_global())

cached_fits <- vignette_cache(
  name     = "v31-robma-metaregression",
  objects  = c("fit_BMA", "fit_RoBMA"),
  packages = c("metafor", "emmeans")
)

knitr::opts_chunk$set(
  collapse    = TRUE,
  comment     = "#>",
  eval        = vignette_cache_eval(cached_fits),
  message     = FALSE,
  warning     = FALSE,
  fig.width   = 7,
  fig.height  = 3.5,
  dev         = "png",
  fig.retina  = 3
)
if (.Platform$OS.type == "windows") {
  knitr::opts_chunk$set(dev.args = list(type = "cairo"))
}

## ----load-models, include = [encoded data, 9970 characters]
library("RoBMA")

data("Andrews2021", package = "RoBMA")

Andrews2021_z <- metafor::escalc(
  ri = ri, ni = ni,
  measure = "ZCOR", data = Andrews2021
)

vignette_cache_load(cached_fits)

## ----precompute-models, include = FALSE, eval = [encoded data, 9950 characters]
# library("RoBMA")
# 
# data("Andrews2021", package = "RoBMA")
# 
# Andrews2021_z <- metafor::escalc(
#   ri = ri, ni = ni,
#   measure = "ZCOR", data = Andrews2021
# )
# 
# prior_effect_361 <- prior("normal", list(mean = 0, sd = 0.5))
# prior_heterogeneity_361 <- prior("invgamma", list(shape = 1, scale = 0.075))
# 
# fit_BMA <- BMA(
#   yi = yi, vi = vi, measure = "ZCOR",
#   mods = ~ measure + age,
#   prior_effect        = prior_effect_361,
#   prior_heterogeneity = prior_heterogeneity_361,
#   data = Andrews2021_z, parallel = TRUE, seed = 1
# )
# 
# fit_RoBMA <- RoBMA(
#   yi = yi, vi = vi, measure = "ZCOR",
#   mods = ~ measure + age,
#   prior_effect        = prior_effect_361,
#   prior_heterogeneity = prior_heterogeneity_361,
#   data = Andrews2021_z, parallel = TRUE, seed = 1
# )
# 
# vignette_cache_save(cached_fits)

## [encoded data, 9997 characters]
library(RoBMA)
data("Andrews2021", package = "RoBMA")
head(Andrews2021)[,c("slab", "ri", "ni", "measure", "age")]

## [encoded data, 9997 characters]
Andrews2021_z <- metafor::escalc(
  ri = ri, ni = ni,
  measure = "ZCOR", data = Andrews2021
)

## [encoded data, 9997 characters]
fit_rma <- metafor::rma(yi, vi, mods = ~ measure + age, data = Andrews2021_z)
fit_rma

## [encoded data, 9997 characters]
emm_rma <- emmeans::emmeans(metafor::emmprep(fit_rma), specs = "measure")
emm_rma
metafor::transf.ztor(as.data.frame(emm_rma)[, c("emmean", "asymp.LCL", "asymp.UCL")])

## [encoded data, 9997 characters]
prior_effect_361 <- prior("normal", list(mean = 0, sd = 0.5))
prior_heterogeneity_361 <- prior("invgamma", list(shape = 1, scale = 0.075))

## ----fit-bma, eval = [encoded data, 9977 characters]
# fit_BMA <- BMA(
#   yi = yi, vi = vi, measure = "ZCOR",
#   mods = ~ measure + age,
#   prior_effect        = prior_effect_361,
#   prior_heterogeneity = prior_heterogeneity_361,
#   data = Andrews2021_z, parallel = TRUE, seed = 1
# )

## [encoded data, 9997 characters]
summary(fit_BMA, include_mcmc_diagnostics = FALSE)
pooled_effect(fit_BMA, output_measure = "COR")

## [encoded data, 9997 characters]
BMA_marginal <- marginal_means(fit_BMA, output_measure = "COR")
summary(BMA_marginal)

## ----fig_BMA, fig.width = 6, fig.height = 4, out.width = "75%", fig.align = "center"[encoded data, 9914 characters]
par(mar = c(4, 4, 1, 4))
plot(BMA_marginal, parameter = "measure", lwd = 2)

## ----fit-robma, eval = [encoded data, 9975 characters]
# fit_RoBMA <- RoBMA(
#   yi = yi, vi = vi, measure = "ZCOR",
#   mods = ~ measure + age,
#   prior_effect        = prior_effect_361,
#   prior_heterogeneity = prior_heterogeneity_361,
#   data = Andrews2021_z, parallel = TRUE, seed = 1
# )

## [encoded data, 9997 characters]
summary(fit_RoBMA, include_mcmc_diagnostics = FALSE)
pooled_effect(fit_RoBMA, output_measure = "COR")

## [encoded data, 9997 characters]
RoBMA_marginal <- marginal_means(fit_RoBMA, output_measure = "COR")
summary(RoBMA_marginal)

## ----fig_RoBMA, fig.width = 6, fig.height = 4, out.width = "75%", fig.align = "center"[encoded data, 9912 characters]
par(mar = c(4, 4, 1, 4))
plot(RoBMA_marginal, parameter = "measure", lwd = 2)

## ----vignette-nowrap-options, include = FALSE, eval = [encoded data, 9944 characters]
options(width = 10000)

## ----setup, include = [encoded data, 9976 characters]
source("vignette-cache.R", local = knitr::knit_global())

cached_fits <- vignette_cache(
  name     = "v33-robma-multilevel-metaregression",
  objects  = "fit_reg"
)

knitr::opts_chunk$set(
  collapse    = TRUE,
  comment     = "#>",
  eval        = vignette_cache_eval(cached_fits),
  message     = FALSE,
  warning     = FALSE,
  fig.width   = 7,
  fig.height  = 3.5,
  dev         = "png",
  fig.retina  = 3
)
if (.Platform$OS.type == "windows") {
  knitr::opts_chunk$set(dev.args = list(type = "cairo"))
}

## ----load-models, include = [encoded data, 9970 characters]
library("RoBMA")

data("Havrankova2025", package = "RoBMA")

unit_information_sd_361 <- estimate_unit_information_sd(
  sei = Havrankova2025$se,
  ni  = Havrankova2025$N
)
prior_scale_361 <- unit_information_sd_361 * 0.5

vignette_cache_load(cached_fits)

## ----precompute-models, include = FALSE, eval = [encoded data, 9950 characters]
# library("RoBMA")
# 
# data("Havrankova2025", package = "RoBMA")
# 
# unit_information_sd_361 <- estimate_unit_information_sd(
#   sei = Havrankova2025$se,
#   ni  = Havrankova2025$N
# )
# prior_scale_361 <- unit_information_sd_361 * 0.5
# 
# prior_effect_361        <- prior("normal",   list(mean = 0, sd = prior_scale_361))
# prior_heterogeneity_361 <- prior("invgamma", list(shape = 1, scale = 0.15 * prior_scale_361))
# prior_mods_361          <- list(
#   facing_customer = prior_factor("mnormal", list(mean = 0, sd = 0.25 * prior_scale_361), contrast = "meandif")
# )
# prior_bias_361          <- list(
#   prior_weightfunction("two-sided", c(0.05),             wf_cumulative(c(1, 1)),       prior_weights = 1/12),
#   prior_weightfunction("two-sided", c(0.05, 0.10),       wf_cumulative(c(1, 1, 1)),    prior_weights = 1/12),
#   prior_weightfunction("one-sided", c(0.05),             wf_cumulative(c(1, 1)),       prior_weights = 1/12),
#   prior_weightfunction("one-sided", c(0.025, 0.05),      wf_cumulative(c(1, 1, 1)),    prior_weights = 1/12),
#   prior_weightfunction("one-sided", c(0.05, 0.5),        wf_cumulative(c(1, 1, 1)),    prior_weights = 1/12),
#   prior_weightfunction("one-sided", c(0.025, 0.05, 0.5), wf_cumulative(c(1, 1, 1, 1)), prior_weights = 1/12),
#   prior_PET("Cauchy", list(0, 1), list(0, Inf), 1/4),
#   prior_PEESE("Cauchy", list(0, 5 / prior_scale_361), list(0, Inf), 1/4)
# )
# 
# fit_reg <- RoBMA(
#   yi = y, sei = se, ni = N, measure = "GEN",
#   mods = ~ facing_customer,
#   cluster = study_id,
#   prior_unit_information_sd = unit_information_sd_361,
#   prior_effect              = prior_effect_361,
#   prior_heterogeneity       = prior_heterogeneity_361,
#   prior_mods                = prior_mods_361,
#   prior_bias                = prior_bias_361,
#   sample = 20000, burnin = 10000, adapt = 10000,
#   thin = 5, parallel = TRUE, seed = 1,
#   data = Havrankova2025
# )
# 
# vignette_cache_save(cached_fits)

## [encoded data, 9997 characters]
prior_effect_361        <- prior("normal",   list(mean = 0, sd = 25.8))
prior_heterogeneity_361 <- prior("invgamma", list(shape = 1, scale = 0.15 * 25.8))
prior_mods_361          <- list(
  facing_customer = prior_factor("mnormal", list(mean = 0, sd = 0.25 * 25.8), contrast = "meandif")
)
prior_bias_361          <- list(
  prior_weightfunction("two-sided", c(0.05),             wf_cumulative(c(1, 1)),       prior_weights = 1/12),
  prior_weightfunction("two-sided", c(0.05, 0.10),       wf_cumulative(c(1, 1, 1)),    prior_weights = 1/12),
  prior_weightfunction("one-sided", c(0.05),             wf_cumulative(c(1, 1)),       prior_weights = 1/12),
  prior_weightfunction("one-sided", c(0.025, 0.05),      wf_cumulative(c(1, 1, 1)),    prior_weights = 1/12),
  prior_weightfunction("one-sided", c(0.05, 0.5),        wf_cumulative(c(1, 1, 1)),    prior_weights = 1/12),
  prior_weightfunction("one-sided", c(0.025, 0.05, 0.5), wf_cumulative(c(1, 1, 1, 1)), prior_weights = 1/12),
  prior_PET("Cauchy", list(0, 1), list(0, Inf), 1/4),
  prior_PEESE("Cauchy", list(0, 5 / 25.8), list(0, Inf), 1/4)
)

## ----fit-reg, eval = [encoded data, 9977 characters]
# fit_reg <- RoBMA(
#   yi = y, sei = se, measure = "GEN",
#   mods = ~ facing_customer,
#   cluster = study_id,
#   prior_effect              = prior_effect_361,
#   prior_heterogeneity       = prior_heterogeneity_361,
#   prior_mods                = prior_mods_361,
#   prior_bias                = prior_bias_361,
#   sample = 20000, burnin = 10000, adapt = 10000,
#   thin = 5, parallel = TRUE, seed = 1,
#   data = Havrankova2025
# )

## [encoded data, 9997 characters]
summary(fit_reg, include_mcmc_diagnostics = FALSE)

## [encoded data, 9997 characters]
pooled_effect(fit_reg)

## [encoded data, 9997 characters]
mm_reg <- marginal_means(fit_reg)
summary(mm_reg)

## ----zplot, fig.width = 6, fig.height = [encoded data, 9958 characters]
par(mar = c(4, 4, 0, 0))
zplot(fit_reg, by.hist = 0.25, plot_extrapolation = FALSE, from = -4, to = 8)

## ----vignette-nowrap-options, include = FALSE, eval = [encoded data, 9944 characters]
options(width = 10000)

## ----setup, include = [encoded data, 9976 characters]
source("vignette-cache.R", local = knitr::knit_global())

cached_fits <- vignette_cache(
  name    = "v34-bma-norm-medicine",
  objects = c("fit_BMA", "fit_RoBMA")
)

knitr::opts_chunk$set(
  collapse   = TRUE,
  comment    = "#>",
  eval       = vignette_cache_eval(cached_fits),
  message    = FALSE,
  warning    = FALSE,
  dev        = "png",
  fig.retina = 3
)
if (.Platform$OS.type == "windows") {
  knitr::opts_chunk$set(dev.args = list(type = "cairo"))
}

## ----include = [encoded data, 9983 characters]
library(RoBMA)

data("Poulsen2006", package = "RoBMA")

vignette_cache_load(cached_fits)

## ----include = FALSE, eval = [encoded data, 9969 characters]
# # R package version updating
# library(RoBMA)
# 
# data("Poulsen2006", package = "RoBMA")
# 
# fit_BMA <- BMA(
#   yi = d, sei = se, measure = "SMD",
#   prior_informed_field    = "medicine",
#   prior_informed_subfield = "Oral Health",
#   data = Poulsen2006, slab = study,
#   seed = 1, parallel = TRUE
# )
# 
# fit_RoBMA <- RoBMA(
#   yi = d, sei = se, measure = "SMD",
#   prior_informed_field    = "medicine",
#   prior_informed_subfield = "Oral Health",
#   data = Poulsen2006, slab = study,
#   seed = 1, parallel = TRUE
# )
# 
# vignette_cache_save(cached_fits)

## [encoded data, 9997 characters]
library(RoBMA)

data("Poulsen2006", package = "RoBMA")
Poulsen2006

## [encoded data, 9997 characters]
summary(fit_BMA, conditional = TRUE, include_mcmc_diagnostics = FALSE)

## [encoded data, 9997 characters]
pooled_effect(fit_BMA, conditional = TRUE)
pooled_heterogeneity(fit_BMA, conditional = TRUE)

## ----fig_mu_BMA, fig.width = 6, fig.height = 4, out.width = "75%", fig.align = "center"[encoded data, 9911 characters]
par(mar = c(4, 4, 0, 4))
plot(fit_BMA, parameter = "mu", prior = TRUE, xlim = c(-2, 2), lwd = 2)

## ----fig_mu_BMA_cond, fig.width = 6, fig.height = 4, out.width = "75%", fig.align = "center"[encoded data, 9906 characters]
par(mar = c(4, 4, 0, 4))
plot(fit_BMA, parameter = "mu", prior = TRUE, conditional = TRUE, xlim = c(-2, 2), lwd = 2)

## [encoded data, 9997 characters]
summary(fit_RoBMA, conditional = TRUE, include_mcmc_diagnostics = FALSE)

## [encoded data, 9997 characters]
pooled_effect(fit_BMA, conditional = TRUE)
pooled_effect(fit_RoBMA, conditional = TRUE)

## ----fig_mu_RoBMA_cond, fig.width = 6, fig.height = 4, out.width = "75%", fig.align = "center"[encoded data, 9904 characters]
par(mar = c(4, 4, 0, 4))
plot(fit_RoBMA, parameter = "mu", prior = TRUE, conditional = TRUE, xlim = c(-2, 2), lwd = 2)

## ----vignette-nowrap-options, include = FALSE, eval = [encoded data, 9944 characters]
options(width = 10000)

## ----setup, include = [encoded data, 9976 characters]
knitr::opts_chunk$set(
  collapse = TRUE,
  comment  = "#>",
  message  = FALSE,
  warning  = FALSE
)

tick    <- '<span class="yes">&#10003;</span>'
limited <- '<span class="limited">limited</span>'
no      <- '<span class="no">No</span>'
na_cell <- '<span class="na-cell">n/a</span>'

model_columns <- c(
  "brma.norm", "brma.glmm", "bselmodel", "bPET/bPEESE",
  "BMA.norm", "BMA.glmm", "RoBMA"
)

reference_links <- c(
  "add_loo()"                            = "../reference/add_loo.brma.html",
  "add_marglik()"                        = "../reference/add_marglik.brma.html",
  "add_waic()"                           = "../reference/add_waic.brma.html",
  "as_draws()"                           = "../reference/as_draws.brma.html",
  "as_draws_array()"                     = "../reference/as_draws.brma.html",
  "as_draws_df()"                        = "../reference/as_draws.brma.html",
  "as_draws_list()"                      = "../reference/as_draws.brma.html",
  "as_draws_matrix()"                    = "../reference/as_draws.brma.html",
  "as_draws_rvars()"                     = "../reference/as_draws.brma.html",
  "as_zplot()"                           = "../reference/as_zplot.brma.html",
  "bayes_factor()"                       = "../reference/bf.brma.html",
  "bf()"                                 = "../reference/bf.brma.html",
  "blup()"                               = "../reference/blup.brma.html",
  "bPET()"                               = "../reference/bPET.html",
  "bPEESE()"                             = "../reference/bPEESE.html",
  "brma()"                               = "../reference/brma.html",
  "brma.glmm()"                          = "../reference/brma.glmm.html",
  "BMA()"                                = "../reference/BMA.html",
  "BMA.glmm()"                           = "../reference/BMA.glmm.html",
  "bridge_sampler()"                     = "../reference/bridge_sampler.brma.html",
  "bselmodel()"                          = "../reference/bselmodel.html",
  "check_loo()"                          = "../reference/check_loo.brma.html",
  "coef()"                               = "../reference/coef.brma.html",
  "contr.meandif()"                      = "../reference/contr.BayesTools.html",
  "contr.orthonormal()"                  = "../reference/contr.BayesTools.html",
  "contr.treatment()"                    = "https://stat.ethz.ch/R-manual/R-devel/library/stats/html/contrast.html",
  "cooks.distance()"                     = "../reference/cooks.distance.brma.html",
  "covratio()"                           = "../reference/covratio.brma.html",
  "dfbetas()"                            = "../reference/dfbetas.brma.html",
  "dffits()"                             = "../reference/dffits.brma.html",
  "estimate_unit_information_sd()"       = "../reference/estimate_unit_information_sd.html",
  "fitted()"                             = "../reference/fitted.brma.html",
  "funnel()"                             = "../reference/funnel.html",
  "galbraith()"                          = "../reference/radial.html",
  "hatvalues()"                          = "../reference/hatvalues.brma.html",
  "influence()"                          = "../reference/influence.brma.html",
  "interpret()"                          = "../reference/interpret.html",
  "logLik()"                             = "../reference/logLik.brma.html",
  "logml()"                              = "../reference/logml.brma.html",
  "loo()"                                = "../reference/loo.brma.html",
  "loo::pareto_k_ids()"                  = "https://mc-stan.org/loo/reference/pareto-k-diagnostic.html",
  "loo::pareto_k_table()"                = "https://mc-stan.org/loo/reference/pareto-k-diagnostic.html",
  "loo_compare()"                        = "../reference/loo_compare.brma.html",
  "loo_weights()"                        = "../reference/loo_weights.brma.html",
  "marginal_means()"                     = "../reference/marginal_means.brma.html",
  "nobs()"                               = "../reference/nobs.brma.html",
  "plot()"                               = "../reference/plot.brma.html",
  "plot_diagnostic()"                    = "../reference/plot_diagnostic.html",
  "plot_diagnostic_autocorrelation()"    = "../reference/plot_diagnostic.html",
  "plot_diagnostic_density()"            = "../reference/plot_diagnostic.html",
  "plot_diagnostic_trace()"              = "../reference/plot_diagnostic.html",
  "plot_pet_peese()"                     = "../reference/plot_pet_peese.html",
  "plot_prior()"                         = "../reference/plot_prior.html",
  "plot_weightfunction()"                = "../reference/plot_weightfunction.html",
  "pooled_effect()"                      = "../reference/pooled_effect.brma.html",
  "pooled_heterogeneity()"               = "../reference/pooled_heterogeneity.brma.html",
  "post_prob()"                          = "../reference/post_prob.brma.html",
  "predict()"                            = "../reference/predict.brma.html",
  "print_prior()"                        = "../reference/print_prior.html",
  "prior()"                              = "../reference/prior.html",
  "prior_factor()"                       = "../reference/prior_factor.html",
  "prior_informed()"                     = "../reference/prior_informed.html",
  "prior_none()"                         = "../reference/prior_none.html",
  "prior_PEESE()"                        = "../reference/prior_PEESE.html",
  "prior_PET()"                          = "../reference/prior_PET.html",
  "prior_weightfunction()"               = "../reference/prior_weightfunction.html",
  "qqnorm()"                             = "../reference/qqnorm.brma.html",
  "radial()"                             = "../reference/radial.html",
  "ranef()"                              = "../reference/ranef.brma.html",
  "regplot()"                            = "../reference/regplot.html",
  "residuals()"                          = "../reference/residuals.brma.html",
  "RoBMA()"                              = "../reference/RoBMA.html",
  "rstandard()"                          = "../reference/rstandard.brma.html",
  "rstudent()"                           = "../reference/rstudent.brma.html",
  "simulate()"                           = "https://stat.ethz.ch/R-manual/R-devel/library/stats/html/simulate.html",
  "summary()"                            = "../reference/summary.brma.html",
  "summary_heterogeneity()"              = "../reference/summary_heterogeneity.brma.html",
  "summary_models()"                     = "../reference/summary_models.html",
  "true_effects()"                       = "../reference/true_effects.brma.html",
  "update()"                             = "../reference/update.brma.html",
  "vcov()"                               = "https://stat.ethz.ch/R-manual/R-devel/library/stats/html/vcov.html",
  "vif()"                                = "../reference/vif.brma.html",
  "waic()"                               = "../reference/waic.brma.html",
  "weights()"                            = "https://stat.ethz.ch/R-manual/R-devel/library/stats/html/weights.html",
  "wf_cumulative()"                      = "../reference/prior_weightfunction.html",
  "wf_fixed()"                           = "../reference/prior_weightfunction.html",
  "wf_independent()"                     = "../reference/prior_weightfunction.html",
  "forest()"                             = "https://wviechtb.github.io/metafor/reference/forest.rma.html",
  "baujat()"                             = "https://wviechtb.github.io/metafor/reference/baujat.html",
  "gosh()"                               = "https://wviechtb.github.io/metafor/reference/gosh.html",
  "labbe()"                              = "https://wviechtb.github.io/metafor/reference/labbe.html"
)

model_column_links <- list(
  "brma.norm"   = "../reference/brma.html",
  "brma.glmm"   = "../reference/brma.glmm.html",
  "bselmodel"   = "../reference/bselmodel.html",
  "bPET/bPEESE" = c("../reference/bPET.html", "../reference/bPEESE.html"),
  "BMA.norm"    = "../reference/BMA.html",
  "BMA.glmm"    = "../reference/BMA.glmm.html",
  "RoBMA"       = "../reference/RoBMA.html"
)

status_map <- c(
  Y = tick,
  L = limited,
  N = no,
  A = na_cell
)

feature_row <- function(topic, feature, pattern) {

  values <- strsplit(pattern, "", fixed = TRUE)[[1]]
  if (length(values) != length(model_columns)) {
    stop("Status pattern has wrong length: ", pattern, call. = FALSE)
  }

  out <- c(
    list(Topic = topic, Feature = feature),
    stats::setNames(as.list(status_map[values]), model_columns)
  )

  return(out)
}

feature_table <- function(rows) {

  rows <- as.data.frame(do.call(rbind, rows), stringsAsFactors = FALSE, check.names = FALSE)
  rows[] <- lapply(rows, function(x) {

    if (is.list(x)) {
      return(vapply(x, function(y) paste0(as.character(y), collapse = ""), character(1)))
    }
    return(as.character(x))
  })

  cat('
<style>
.feature-matrix {
  border-collapse: collapse;
  width: 100%;
  font-size: 0.84em;
}
.feature-matrix th,
.feature-matrix td {
  border: 1px solid #d0d7de;
  padding: 4px 6px;
  vertical-align: top;
}
.feature-matrix th {
  background: #f6f8fa;
}
.feature-matrix .topic-head {
  min-width: 60px;
  width: 60px;
}
.feature-matrix .feature-head {
  min-width: 25em;
}
.feature-matrix .model-head {
  height: 118px;
  min-width: 34px;
  max-width: 34px;
  width: 34px;
  padding: 0;
  position: relative;
  vertical-align: bottom;
}
.feature-matrix .rotated-header {
  bottom: 6px;
  display: block;
  left: 50%;
  line-height: 1;
  position: absolute;
  transform: rotate(-90deg);
  transform-origin: left center;
  white-space: nowrap;
}
.feature-matrix .topic-cell {
  background: #f6f8fa;
  min-width: 60px;
  width: 60px;
  padding: 0;
  text-align: center;
  vertical-align: middle;
}
.feature-matrix .topic-cell span {
  display: inline-block;
  transform: rotate(-90deg);
  white-space: normal;
  width: 8em;
  text-align: center;
  font-weight: 700;
  line-height: 1.1;
}
.feature-matrix .feature-cell {
  min-width: 25em;
}
.feature-matrix .feature-cell a {
  color: #0969da;
  text-decoration: none;
}
.feature-matrix .feature-cell a:hover {
  text-decoration: underline;
}
.feature-matrix .feature-cell a code {
  color: inherit;
}
.feature-matrix .status-cell {
  text-align: center;
  vertical-align: middle;
  white-space: nowrap;
  width: 34px;
}
.feature-matrix .yes {
  color: #1a7f37;
  font-weight: 700;
}
.feature-matrix .limited {
  color: #9a6700;
}
.feature-matrix .no,
.feature-matrix .na-cell {
  color: #6e7781;
}
</style>
')

  escape_html_text <- function(x) {

    x <- as.character(x)
    x <- gsub("&", "&amp;", x, fixed = TRUE)
    x <- gsub("<", "&lt;", x, fixed = TRUE)
    x <- gsub(">", "&gt;", x, fixed = TRUE)
    x <- gsub('"', "&quot;", x, fixed = TRUE)
    return(x)
  }

  format_code_span <- function(code) {

    code_html <- paste0("<code>", escape_html_text(code), "</code>")
    link      <- if (code %in% names(reference_links)) reference_links[[code]] else NA_character_

    if (!is.na(link)) {
      return(paste0('<a href="', escape_html_text(link), '">', code_html, "</a>"))
    }

    return(code_html)
  }

  escape_html <- function(x) {

    x       <- as.character(x)
    matches <- gregexpr("`[^`]+`", x, perl = TRUE)[[1]]

    if (matches[1] == -1L) {
      return(escape_html_text(x))
    }

    match_lengths <- attr(matches, "match.length")
    pieces        <- character(0)
    position      <- 1L

    for (j in seq_along(matches)) {
      match_start <- matches[j]
      match_end   <- match_start + match_lengths[j] - 1L

      if (match_start > position) {
        pieces <- c(pieces, escape_html_text(substr(x, position, match_start - 1L)))
      }

      code <- substr(x, match_start + 1L, match_end - 1L)
      pieces <- c(pieces, format_code_span(code))
      position <- match_end + 1L
    }

    if (position <= nchar(x)) {
      pieces <- c(pieces, escape_html_text(substr(x, position, nchar(x))))
    }

    return(paste0(pieces, collapse = ""))
  }

  format_model_column <- function(model_column) {

    link <- model_column_links[[model_column]]

    if (length(link) == 2L) {
      return(paste0(
        '<a href="', escape_html_text(link[[1]]), '">bPET</a>/',
        '<a href="', escape_html_text(link[[2]]), '">bPEESE</a>'
      ))
    }

    return(paste0(
      '<a href="', escape_html_text(link), '">',
      escape_html_text(model_column),
      "</a>"
    ))
  }

  topic_runs <- rle(rows$Topic)
  topic_start <- cumsum(c(1L, head(topic_runs$lengths, -1L)))
  topic_span <- rep(0L, nrow(rows))
  topic_span[topic_start] <- topic_runs$lengths

  cat('<div style="overflow-x: auto;">\n')
  cat('<table class="feature-matrix">\n')
  cat("<thead><tr>")
  cat('<th class="topic-head">Topic</th>')
  cat('<th class="feature-head">Feature</th>')
  for (model_column in model_columns) {
    cat('<th class="model-head"><span class="rotated-header">', format_model_column(model_column), "</span></th>", sep = "")
  }
  cat("</tr></thead>\n")
  cat("<tbody>\n")
  for (i in seq_len(nrow(rows))) {
    cat("<tr>")
    if (topic_span[i] > 0L) {
      cat(
        '<td class="topic-cell" rowspan="', topic_span[i], '"><span>',
        escape_html(rows$Topic[i]),
        "</span></td>",
        sep = ""
      )
    }
    cat('<td class="feature-cell">', escape_html(rows$Feature[i]), "</td>", sep = "")
    for (model_column in model_columns) {
      cat('<td class="status-cell">', rows[[model_column]][i], "</td>", sep = "")
    }
    cat("</tr>\n")
  }
  cat("</tbody>\n")
  cat("</table>\n")
  cat("</div>\n")
}

## ----feature-matrix, echo = FALSE, results = "asis"[encoded data, 9947 characters]
rows <- list(
  feature_row("Model / Structure", "Fixed- and random-effects models", "YYYYYYY"),
  feature_row("Model / Structure", "Moderation (`mods`)", "YYYYYYY"),
  feature_row("Model / Structure", "Location-scale models (`scale`)", "YYYYYYY"),
  feature_row("Model / Structure", "Multilevel models (`cluster`)", "YYYYYYY"),
  feature_row("Model / Structure", "Model averaging", "NNNNYYY"),
  feature_row("Model / Structure", "Inclusion Bayes factors", "NNNNYYY"),
  feature_row("Model / Structure", "General multivariate / covariance structures", "NNNNNNN"),
  feature_row("Model / Structure", "Robust / sandwich inference", "NNNNNNN"),
  feature_row("Model / Structure", "Refit / update (`update()`)", "YYYYYYY"),

  feature_row("Priors", "Default prior distributions", "YYYYYYY"),
  feature_row("Priors", "Empirical/informed prior distributions (`prior_informed()`)", "YYYYYYY"),
  feature_row("Priors", "Custom prior distributions (`prior()`, `prior_factor()`, `prior_weightfunction()`, `prior_none()`)", "YYYYYYY"),
  feature_row("Priors", "PET / PEESE prior distributions (`prior_PET()`, `prior_PEESE()`)", "NNNYNNY"),
  feature_row("Priors", "Prior-only inspection (`only_priors = TRUE`)", "YYYYYYY"),
  feature_row("Priors", "Weight-function shape (`wf_cumulative()`, `wf_fixed()`, `wf_independent()`)", "NNYNNNY"),
  feature_row("Priors", "Factor contrasts (`contr.treatment()`, `contr.meandif()`, `contr.orthonormal()`)", "YYYYYYY"),
  feature_row("Priors", "UISD estimation (`estimate_unit_information_sd()`)", "YYYYYYY"),

  feature_row("Publication Bias", "Selection models", "NNYNNNY"),
  feature_row("Publication Bias", "PET / PEESE models", "NNNYNNY"),
  feature_row("Publication Bias", "Model averaging over bias models", "NNNNNNY"),
  feature_row("Publication Bias", "Bias-adjusted summaries / predictions", "NNYYNNY"),
  feature_row("Publication Bias", "Trim-fill / fail-safe N", "NNNNNNN"),

  feature_row("Prediction", "Pooled effect / heterogeneity summaries (`pooled_effect()`, `pooled_heterogeneity()`)", "YYYYYYY"),
  feature_row("Prediction", "Heterogeneity decomposition (`summary_heterogeneity()`, tau², I², H²)", "YYYYYYY"),
  feature_row("Prediction", "Fitted-value extraction (`fitted()`)", "YYYYYYY"),
  feature_row("Prediction", "Prediction for new covariate values (`predict()`)", "YYYYYYY"),
  feature_row("Prediction", "Posterior predictive response summaries (`predict()`)", "YYYYYYY"),
  feature_row("Prediction", "True effects / BLUPs / random effects (`true_effects()`, `blup()`, `ranef()`)", "YYYYYYY"),
  feature_row("Prediction", "Estimated marginal means (`marginal_means()`)", "YYYYYYY"),

  feature_row("Plots", "Posterior plots (`plot()`)", "YYYYYYY"),
  feature_row("Plots", "Prior plots (`plot_prior()`)", "YYYYYYY"),
  feature_row("Plots", "Marginal means plots (`marginal_means()` + `plot()`)", "YYYYYYY"),
  feature_row("Plots", "Regression plots (`regplot()`)", "YYYYYYY"),
  feature_row("Plots", "Funnel plots (`funnel()`)", "YYYYYYY"),
  feature_row("Plots", "zplot diagnostics (`as_zplot()`/`zplot()`)", "YYYYYYY"),
  feature_row("Plots", "Weight-function plots (`plot_weightfunction()`)", "NNYNNNY"),
  feature_row("Plots", "PET-PEESE plots (`plot_pet_peese()`)", "NNNYNNY"),
  feature_row("Plots", "Radial / Galbraith plots (`radial()`, `galbraith()`)", "YLYYYLY"),
  feature_row("Plots", "Forest plots (`forest()`)", "NNNNNNN"),
  feature_row("Plots", "Baujat plots (`baujat()`)", "NNNNNNN"),
  feature_row("Plots", "GOSH plots (`gosh()`)", "NNNNNNN"),
  feature_row("Plots", "L'Abbe plots (`labbe()`)", "NNNNNNN"),

  feature_row("Residuals / Diagnostics", "Raw residuals (`residuals()`)", "YYYYYYY"),
  feature_row("Residuals / Diagnostics", "Pearson / standardized residuals (`rstandard()`)", "YNNYYNL"),
  feature_row("Residuals / Diagnostics", "Studentized residuals (LOO-PIT) (`rstudent()`)", "YYYYYYY"),
  feature_row("Residuals / Diagnostics", "Q-Q plots (`qqnorm()`)", "YYYYYYY"),
  feature_row("Residuals / Diagnostics", "Cook's distances (`cooks.distance()`)", "YNNYYNN"),
  feature_row("Residuals / Diagnostics", "DFBETAS (`dfbetas()`)", "YYYYYYY"),
  feature_row("Residuals / Diagnostics", "DFFITS (`dffits()`)", "YNNYYNN"),
  feature_row("Residuals / Diagnostics", "Covariance ratios (`covratio()`)", "YYYYYYY"),
  feature_row("Residuals / Diagnostics", "Hat values (`hatvalues()`)", "YNNYYNN"),
  feature_row("Residuals / Diagnostics", "Combined influence summary (`influence()`)", "YLLYYLL"),
  feature_row("Residuals / Diagnostics", "LOO / WAIC diagnostics (`check_loo()`, `loo::pareto_k_ids()`, `loo::pareto_k_table()`)", "YYYYYYY"),
  feature_row("Residuals / Diagnostics", "Moderator collinearity diagnostics (`vif()`)", "YYYYYYY"),
  feature_row("Residuals / Diagnostics", "Refit leave-one-out / permutation tests", "NNNNNNN"),

  feature_row("MCMC Diagnostics", "Posterior summaries (Rhat, ESS, MCMC error) (`summary()`)", "YYYYYYY"),
  feature_row("MCMC Diagnostics", "Trace plots (`plot_diagnostic_trace()`)", "YYYYYYY"),
  feature_row("MCMC Diagnostics", "Density plots (`plot_diagnostic_density()`)", "YYYYYYY"),
  feature_row("MCMC Diagnostics", "Autocorrelation plots (`plot_diagnostic_autocorrelation()`)", "YYYYYYY"),
  feature_row("MCMC Diagnostics", "Diagnostic-plot wrapper (`plot_diagnostic()`)", "YYYYYYY"),

  feature_row("Model Comparison", "Marginal likelihood (`add_marglik()`, `logml()`)", "YYYYNNN"),
  feature_row("Model Comparison", "WAIC/LOO (`add_waic()`, `add_loo()`, `waic()`, `loo()`)", "YYYYYYY"),
  feature_row("Model Comparison", "Bayes factors (`bf()`, `bayes_factor()`, `post_prob()`)", "YYYYYYY"),
  feature_row("Model Comparison", "WAIC comparison (`loo_compare()`, `loo_weights()`)", "YYYYYYY"),
  feature_row("Model Comparison", "LOO comparison (`loo_compare()`, `loo_weights()`)", "YYYYYYY"),
  feature_row("Model Comparison", "AIC / BIC", "NNNNNNN"),

  feature_row("Reporting", "Plain-text interpretation (`interpret()`)", "YYYYYYY"),
  feature_row("Reporting", "Sub-model summary (`summary_models()`)", "NNNNYYY"),
  feature_row("Reporting", "Prior inspection (`print_prior()`)", "YYYYYYY"),

  feature_row("Extraction", "Posterior draw extraction (`as_draws()`, `as_draws_array()`, `as_draws_df()`, `as_draws_list()`, `as_draws_matrix()`, `as_draws_rvars()`)", "YYYYYYY"),
  feature_row("Extraction", "Coefficients (`coef()`)", "YYYYYYY"),
  feature_row("Extraction", "Point-wise log-likelihood (`logLik()`)", "YYYYYYY"),
  feature_row("Extraction", "Sample size (`nobs()`)", "YYYYYYY"),
  feature_row("Extraction", "Variance-covariance matrix (`vcov()`)", "NNNNNNN"),
  feature_row("Extraction", "Credible intervals (`summary()`, `summary_heterogeneity()`, `pooled_effect()`, `pooled_heterogeneity()`)", "YYYYYYY"),
  feature_row("Extraction", "Model weights (`weights()`)", "NNNNNNN"),
  feature_row("Extraction", "Simulated responses (`simulate()`)", "NNNNNNN")
)

# Intentionally skipped exported functions in this feature table:
# bridge_sampler(), contr.independent(), RoBMA.options(), RoBMA.get_option(),
# set_autofit_control(), set_convergence_checks().

feature_table(rows)

##' RoBMA: Robust Bayesian Meta-Analysis
##'
##' RoBMA provides Bayesian meta-analysis, meta-regression, multilevel
##' meta-analysis, model averaging, and publication-bias adjustment. The main
##' user-facing fitters are \code{\link{RoBMA}}, \code{\link{BMA}},
##' \code{\link{brma}}, \code{\link{brma.glmm}}, \code{\link{bselmodel}},
##' \code{\link{bPET}}, and \code{\link{bPEESE}}.
##'
##' @name RoBMA-package
##' @author Frantisek Bartos \email{f.bartos96@@gmail.com}
##' @keywords package
##' @aliases RoBMA-package RoBMA_package RoBMA.package
##' @docType package
##' @section User guide:
##' See \insertCite{bartos2021no;textual}{RoBMA},
##' \insertCite{maier2020robust;textual}{RoBMA},
##' \insertCite{bartos2020adjusting;textual}{RoBMA}, and
##' \insertCite{bartos2025robust;textual}{RoBMA} for the RoBMA methodology.
##' Use \code{vignette(package = "RoBMA")} to list installed vignettes.
##'
##' @references \insertAllCited{}
##' @importFrom BayesTools is.prior is.prior.none is.prior.point is.prior.simple is.prior.factor is.prior.PET is.prior.PEESE is.prior.weightfunction
##' @importFrom BayesTools is.prior.independent is.prior.spike_and_slab is.prior.mixture
##' @importFrom Rdpack reprompt
##' @importFrom rlang .data
"_PACKAGE"

# ============================================================================ #
# brma.unit_level.R
# ============================================================================ #
#
# Shared helpers for diagnostics that need to distinguish the output unit from
# the conditioning depth.
#
# ============================================================================ #


# ---------------------------------------------------------------------------- #
# .normalize_unit
# ---------------------------------------------------------------------------- #
#
# @param unit character; output/deletion unit.
#
# @return normalized unit value.
#
# ---------------------------------------------------------------------------- #
.normalize_unit <- function(unit) {

  return(match.arg(unit, c("estimate", "cluster")))
}


# ---------------------------------------------------------------------------- #
# .normalize_conditioning_depth
# ---------------------------------------------------------------------------- #
#
# @param conditioning_depth character; conditioning depth.
#
# @return normalized conditioning depth value.
#
# ---------------------------------------------------------------------------- #
.normalize_conditioning_depth <- function(conditioning_depth) {

  return(match.arg(conditioning_depth, c("marginal", "cluster", "estimate")))
}


# ---------------------------------------------------------------------------- #
# .check_unit_conditioning_depth
# ---------------------------------------------------------------------------- #
#
# Validate combinations of unit and conditioning depth for diagnostics.
#
# @param object             brma object.
# @param unit               character; output/deletion unit.
# @param conditioning_depth character; conditioning depth.
# @param caller             character; caller name for error messages.
#
# @return invisible NULL.
#
# ---------------------------------------------------------------------------- #
.check_unit_conditioning_depth <- function(object, unit, conditioning_depth, caller) {

  is_multilevel <- .is_multilevel(object)

  if (unit == "cluster" && !is_multilevel) {
    stop(caller, " with unit = 'cluster' is only available for multilevel models.",
         call. = FALSE)
  }

  if (conditioning_depth == "cluster" && !is_multilevel) {
    stop(caller, " with conditioning_depth = 'cluster' is only available for multilevel models.",
         call. = FALSE)
  }

  if (unit == "cluster" && conditioning_depth == "estimate") {
    stop(caller, " does not support unit = 'cluster' with conditioning_depth = 'estimate'.",
         call. = FALSE)
  }

  return(invisible(NULL))
}


# ---------------------------------------------------------------------------- #
# .loo_conditioning_depth_from_unit
# ---------------------------------------------------------------------------- #
#
# LOO exposes only the deletion unit. The implied conditioning depth is stored
# as metadata.
#
# @param unit character; output/deletion unit.
#
# @return character; implied conditioning depth.
#
# ---------------------------------------------------------------------------- #
.loo_conditioning_depth_from_unit <- function(unit) {

  unit <- .normalize_unit(unit)

  if (unit == "estimate") {
    return("estimate")
  } else {
    return("cluster")
  }
}


# ---------------------------------------------------------------------------- #
# .get_cluster_indices
# ---------------------------------------------------------------------------- #
#
# @param object brma object.
#
# @return named list of integer vectors, one per cluster.
#
# ---------------------------------------------------------------------------- #
.get_cluster_indices <- function(object) {

  if (!.is_multilevel(object)) {
    return(NULL)
  }

  outcome_data <- object[["data"]][["outcome"]]
  cluster      <- outcome_data[["cluster"]]
  indices      <- split(seq_along(cluster), cluster)
  labels       <- .get_cluster_labels(object)

  names(indices) <- labels[names(indices)]

  return(indices)
}


# ---------------------------------------------------------------------------- #
# .get_cluster_labels
# ---------------------------------------------------------------------------- #
#
# @param object brma object.
#
# @return named character vector mapping numeric cluster indices to labels.
#
# ---------------------------------------------------------------------------- #
.get_cluster_labels <- function(object) {

  if (!.is_multilevel(object)) {
    return(NULL)
  }

  outcome_data <- object[["data"]][["outcome"]]
  cluster      <- outcome_data[["cluster"]]

  if ("cluster_label" %in% names(outcome_data)) {
    labels <- tapply(outcome_data[["cluster_label"]], cluster, function(x) x[1])
  } else {
    labels <- tapply(cluster, cluster, function(x) as.character(x[1]))
  }
  label_names <- names(labels)
  labels      <- as.character(labels)
  names(labels) <- label_names

  return(labels)
}


# ---------------------------------------------------------------------------- #
# .get_estimate_labels
# ---------------------------------------------------------------------------- #
#
# @param object brma object.
#
# @return character vector of estimate labels.
#
# ---------------------------------------------------------------------------- #
.get_estimate_labels <- function(object) {

  outcome_data <- object[["data"]][["outcome"]]

  if ("slab" %in% names(outcome_data)) {
    labels <- as.character(outcome_data[["slab"]])
  } else {
    labels <- as.character(seq_len(nrow(outcome_data)))
  }

  return(labels)
}


# ---------------------------------------------------------------------------- #
# .get_outcome_hash
# ---------------------------------------------------------------------------- #
#
# @param object brma object.
#
# @return deterministic hash of the outcome target.
#
# ---------------------------------------------------------------------------- #
.get_outcome_hash <- function(object) {

  outcome_type <- .outcome_type(object)
  outcome_data <- object[["data"]][["outcome"]]

  payload <- list(outcome_type = outcome_type)

  if (outcome_type == "norm") {
    payload[["outcome"]] <- list(
      yi  = unname(as.numeric(outcome_data[["yi"]])),
      sei = unname(as.numeric(outcome_data[["sei"]]))
    )
  } else if (outcome_type == "bin") {
    payload[["outcome"]] <- list(
      ai  = unname(as.integer(outcome_data[["ai"]])),
      ci  = unname(as.integer(outcome_data[["ci"]])),
      n1i = unname(as.integer(outcome_data[["n1i"]])),
      n2i = unname(as.integer(outcome_data[["n2i"]]))
    )
  } else if (outcome_type == "pois") {
    payload[["outcome"]] <- list(
      x1i = unname(as.integer(outcome_data[["x1i"]])),
      x2i = unname(as.integer(outcome_data[["x2i"]])),
      t1i = unname(as.numeric(outcome_data[["t1i"]])),
      t2i = unname(as.numeric(outcome_data[["t2i"]]))
    )
  }

  if ("weights" %in% names(outcome_data)) {
    payload[["weights"]] <- unname(as.numeric(outcome_data[["weights"]]))
  }

  if ("cluster" %in% names(outcome_data)) {
    payload[["cluster"]] <- unname(as.integer(outcome_data[["cluster"]]))
  }

  if ("cluster_label" %in% names(outcome_data)) {
    payload[["cluster_label"]] <- unname(as.character(outcome_data[["cluster_label"]]))
  }

  bytes <- as.integer(serialize(payload, NULL, version = 3))
  hash1 <- 5381
  hash2 <- 0

  for (byte in bytes) {
    hash1 <- (hash1 * 33 + byte) %% 2147483647
    hash2 <- (hash2 * 65599 + byte) %% 2147483629
  }

  return(paste0(
    sprintf("%08x", as.integer(hash1)),
    sprintf("%08x", as.integer(hash2))
  ))
}


# ---------------------------------------------------------------------------- #
# .check_brma_compare_targets
# ---------------------------------------------------------------------------- #
#
# Reject Bayes factor/model-probability comparisons across different data.
#
# @param objects list of brma objects.
# @param caller  character; caller name for error messages.
#
# @return invisible NULL.
#
# ---------------------------------------------------------------------------- #
.check_brma_compare_targets <- function(objects, caller) {

  data_hashes <- vapply(objects, .get_outcome_hash, character(1))

  if (length(unique(data_hashes)) > 1) {
    stop(caller, " requires models fitted to the same outcome data.",
         call. = FALSE)
  }

  return(invisible(NULL))
}


# ---------------------------------------------------------------------------- #
# .check_cluster_unit_deferred
# ---------------------------------------------------------------------------- #
#
# Error used by diagnostics whose cluster-unit implementation is intentionally
# deferred.
#
# @param caller character; caller name for error messages.
#
# @return stops.
#
# ---------------------------------------------------------------------------- #
.check_cluster_unit_deferred <- function(caller) {
  # Cluster residual diagnostics need a separate Mahalanobis/chi-square design
  stop(caller, " with unit = 'cluster' is not implemented currently.", call. = FALSE)
}


# ---------------------------------------------------------------------------- #
# .check_legacy_level_arg
# ---------------------------------------------------------------------------- #
#
# Reject the old residual-conditioning argument name when it arrives through
# `...`.
#
# @param dots   list; captured dots.
# @param caller character; caller name for error messages.
#
# @return invisible NULL.
#
# ---------------------------------------------------------------------------- #
.check_legacy_level_arg <- function(dots, caller) {

  if ("level" %in% names(dots)) {
    stop(caller, " uses 'conditioning_depth' for residual conditioning; ",
         "do not set 'level'.", call. = FALSE)
  }

  return(invisible(NULL))
}


# ---------------------------------------------------------------------------- #
# .add_loo_target_metadata
# ---------------------------------------------------------------------------- #
#
# Store the LOO/WAIC target on the returned loo object.
#
# @param object             loo or waic object.
# @param unit               character; output/deletion unit.
# @param conditioning_depth character; implied conditioning depth.
# @param targets            character; target labels.
# @param data_hash          character; hash of the outcome target.
#
# @return object with RoBMA target metadata.
#
# ---------------------------------------------------------------------------- #
.add_loo_target_metadata <- function(object, unit, conditioning_depth, targets,
                                     data_hash) {

  attr(object, "RoBMA_target") <- list(
    unit               = unit,
    conditioning_depth = conditioning_depth,
    n                  = length(targets),
    targets            = targets,
    data_hash          = data_hash
  )

  return(object)
}


# ---------------------------------------------------------------------------- #
# .get_loo_target_metadata
# ---------------------------------------------------------------------------- #
#
# @param object loo or waic object.
#
# @return target metadata list or NULL.
#
# ---------------------------------------------------------------------------- #
.get_loo_target_metadata <- function(object) {

  return(attr(object, "RoBMA_target", exact = TRUE))
}


# ---------------------------------------------------------------------------- #
# .check_loo_target
# ---------------------------------------------------------------------------- #
#
# Check that a stored LOO object matches the requested unit.
#
# @param object brma object.
# @param unit   character; requested unit.
#
# @return stored loo object.
#
# ---------------------------------------------------------------------------- #
.check_loo_target <- function(object, unit) {

  unit       <- .normalize_unit(unit)
  loo_store  <- object[["loo"]]

  if (is.null(loo_store)) {
    stop("LOO has not been computed. Call 'object <- add_loo(object)' first.",
         call. = FALSE)
  }

  loo_result <- loo_store[[unit]]

  if (is.null(loo_result)) {
    stop(
      "LOO with unit = '", unit, "' has not been computed. ",
      "Call 'object <- add_loo(object, unit = \"", unit, "\")' first.",
      call. = FALSE
    )
  }

  metadata <- .get_loo_target_metadata(loo_result)

  if (!is.null(metadata) && metadata[["unit"]] != unit) {
    stop(
      "Stored LOO was computed with unit = '", metadata[["unit"]],
      "'. Recompute with add_loo(object, unit = '", unit, "').",
      call. = FALSE
    )
  }

  return(loo_result)
}


# ---------------------------------------------------------------------------- #
# .check_waic_target
# ---------------------------------------------------------------------------- #
#
# Check that a stored WAIC object matches the requested unit.
#
# @param object brma object.
# @param unit   character; requested unit.
#
# @return stored waic object.
#
# ---------------------------------------------------------------------------- #
.check_waic_target <- function(object, unit) {

  unit       <- .normalize_unit(unit)
  waic_store <- object[["waic"]]

  if (is.null(waic_store)) {
    stop("WAIC has not been computed. Call 'object <- add_waic(object)' first.",
         call. = FALSE)
  }

  waic_result <- waic_store[[unit]]

  if (is.null(waic_result)) {
    stop(
      "WAIC with unit = '", unit, "' has not been computed. ",
      "Call 'object <- add_waic(object, unit = \"", unit, "\")' first.",
      call. = FALSE
    )
  }

  metadata <- .get_loo_target_metadata(waic_result)

  if (!is.null(metadata) && metadata[["unit"]] != unit) {
    stop(
      "Stored WAIC was computed with unit = '", metadata[["unit"]],
      "'. Recompute with add_waic(object, unit = '", unit, "').",
      call. = FALSE
    )
  }

  return(waic_result)
}


# ---------------------------------------------------------------------------- #
# .get_target_conditioning_depth
# ---------------------------------------------------------------------------- #
#
# Extract the conditioning-depth metadata, accepting older cached objects that
# used `level`.
#
# @param metadata list; RoBMA target metadata.
#
# @return character scalar.
#
# ---------------------------------------------------------------------------- #
.get_target_conditioning_depth <- function(metadata) {

  if (!is.null(metadata[["conditioning_depth"]])) {
    return(metadata[["conditioning_depth"]])
  }
  if (!is.null(metadata[["level"]])) {
    return(metadata[["level"]])
  }

  return(NA_character_)
}


# ---------------------------------------------------------------------------- #
# .check_loo_compare_targets
# ---------------------------------------------------------------------------- #
#
# Reject comparisons across different LOO targets.
#
# @param loo_objects list of loo objects.
#
# @return invisible NULL.
#
# ---------------------------------------------------------------------------- #
.check_loo_compare_targets <- function(loo_objects) {

  metadata <- lapply(loo_objects, .get_loo_target_metadata)
  missing  <- vapply(metadata, is.null, logical(1))

  if (length(metadata) <= 1) {
    return(invisible(NULL))
  }

  if (any(missing)) {
    stop("LOO/WAIC objects without RoBMA target metadata cannot be compared.",
         call. = FALSE)
  }

  missing_hash <- vapply(metadata, function(x) is.null(x[["data_hash"]]), logical(1))
  if (any(missing_hash)) {
    stop("LOO/WAIC objects without RoBMA data hashes cannot be compared.",
         call. = FALSE)
  }

  units                <- vapply(metadata, `[[`, character(1), "unit")
  conditioning_depths  <- vapply(metadata, .get_target_conditioning_depth, character(1))
  data_hashes          <- vapply(metadata, `[[`, character(1), "data_hash")

  if (length(unique(units)) > 1 ||
      length(unique(conditioning_depths)) > 1 ||
      length(unique(data_hashes)) > 1) {
    stop("LOO/WAIC objects with different data, unit, or conditioning-depth targets cannot be compared.",
         call. = FALSE)
  }

  return(invisible(NULL))
}

#' @title Summarize Model-Averaged Component Weights
#'
#' @description Creates marginal or individual model-weight summaries for
#' RoBMA-class product-space objects.
#'
#' @param object a fitted RoBMA-class product-space object, including
#' \code{RoBMA}, \code{BMA}/\code{BMA.norm}, and \code{BMA.glmm}.
#' @param type whether to summarize marginal component prior distributions
#'   (\code{"marginal"}) or individual model combinations (\code{"individual"}).
#' @param include_mcmc_diagnostics whether to include Bayes factor MCMC
#'   diagnostics in the output. Defaults to \code{TRUE}.
#' @param ... additional arguments
#'
#' @details Only mixture-prior components are summarized; non-mixture
#' components are omitted.
#'
#' @return A list of class \code{summary_models.RoBMA} with elements
#' \code{name} and \code{type}. For \code{type = "marginal"}, element
#' \code{marginal} contains component tables with columns such as
#' \code{prior_prob}, \code{post_prob}, and \code{inclusion_BF}. For
#' \code{type = "individual"}, element \code{individual} contains individual
#' model combinations and posterior probabilities.
#'
#' @examples \dontrun{
#' if (requireNamespace("metadat", quietly = TRUE)) {
#'   data(dat.lehmann2018, package = "metadat")
#'   fit <- RoBMA(yi = yi, vi = vi, data = dat.lehmann2018, measure = "SMD")
#'
#'   summary_models(fit)
#'   summary_models(fit, type = "individual")
#' }
#' }
#'
#' @export
summary_models <- function(object, ...) {

  UseMethod("summary_models")
}

#' @export
summary_models.default <- function(object, ...) {

  stop("'summary_models' is available only for RoBMA objects.", call. = FALSE)
}

#' @rdname summary_models
#' @export
summary_models.RoBMA <- function(object, type = "marginal",
                                 include_mcmc_diagnostics = TRUE, ...) {

  BayesTools::check_char(type, "type")
  BayesTools::check_bool(include_mcmc_diagnostics, "include_mcmc_diagnostics")
  type <- match.arg(type, c("marginal", "individual"))

  if (is.null(object[["fit"]]) || length(object[["fit"]]) == 0) {
    stop("'summary_models' requires a fitted RoBMA object.", call. = FALSE)
  }

  components             <- .summary_models_components(object)
  posterior_samples      <- .get_posterior_samples(object[["fit"]])
  posterior_samples_list <- if (include_mcmc_diagnostics) {
    .summary_models_posterior_samples_list(object[["fit"]])
  } else {
    NULL
  }

  if (length(components) == 0L) {
    stop("No mixture-prior model components found.", call. = FALSE)
  }

  output <- list(
    name = .summary.brma_model_names(object),
    type = type
  )

  if (type == "marginal") {
    output[["marginal"]] <- .summary_models_marginal_tables(
      components             = components,
      posterior_samples      = posterior_samples,
      posterior_samples_list = posterior_samples_list
    )
  } else {
    output[["individual"]] <- .summary_models_individual_table(
      components             = components,
      posterior_samples      = posterior_samples,
      posterior_samples_list = posterior_samples_list
    )
  }

  class(output) <- "summary_models.RoBMA"

  return(output)
}

#' @rdname summary_models
#' @param x a `summary_models.RoBMA` object.
#' @export
print.summary_models.RoBMA <- function(x, ...) {

  cat("\n")

  if (x[["type"]] == "marginal") {
    for (i in seq_along(x[["marginal"]])) {
      if (i > 1L) {
        cat("\n")
      }
      table <- x[["marginal"]][[i]]
      print(table)
    }
  } else {
    print(x[["individual"]])
  }

  cat("\n")

  return(invisible(x))
}

# Extract top-level model components used by summary_models().
.summary_models_components <- function(object) {

  prior_list <- attr(object[["fit"]], "prior_list")
  components <- list()

  effect_parameter <- if ("mu" %in% names(prior_list)) "mu" else "mu_intercept"
  components <- .summary_models_add_component(
    components = components,
    prior_list = prior_list,
    component  = "Effect",
    parameter  = effect_parameter
  )

  location_parameters <- grep("^mu_", names(prior_list), value = TRUE)
  location_parameters <- location_parameters[location_parameters != "mu_intercept"]
  for (parameter in location_parameters) {
    component <- paste0(
      "Location: ",
      .summary_parameter_label(sub("^mu_", "", parameter))
    )
    components <- .summary_models_add_component(
      components = components,
      prior_list = prior_list,
      component  = component,
      parameter  = parameter
    )
  }

  heterogeneity_parameter <- if ("tau" %in% names(prior_list)) "tau" else "log_tau_intercept"
  components <- .summary_models_add_component(
    components = components,
    prior_list = prior_list,
    component  = "Heterogeneity",
    parameter  = heterogeneity_parameter
  )

  scale_parameters <- grep("^log_tau_", names(prior_list), value = TRUE)
  scale_parameters <- scale_parameters[scale_parameters != "log_tau_intercept"]
  for (parameter in scale_parameters) {
    component <- paste0(
      "Scale: ",
      .summary_parameter_label(sub("^log_tau_", "", parameter))
    )
    components <- .summary_models_add_component(
      components = components,
      prior_list = prior_list,
      component  = component,
      parameter  = parameter
    )
  }

  components <- .summary_models_add_component(
    components = components,
    prior_list = prior_list,
    component  = "Publication Bias",
    parameter  = "bias"
  )

  return(components)
}

.summary_models_add_component <- function(components, prior_list, component, parameter) {

  if (!parameter %in% names(prior_list)) {
    return(components)
  }
  if (!BayesTools::is.prior.mixture(prior_list[[parameter]])) {
    return(components)
  }

  prior <- prior_list[[parameter]]
  components[[component]] <- list(
    component   = component,
    parameter   = parameter,
    prior       = prior,
    names       = .summary_models_prior_names(prior),
    hypothesis  = .summary_models_hypothesis(prior),
    prior_probs = .summary_models_prior_probs(prior)
  )

  return(components)
}

.summary_models_marginal_tables <- function(components, posterior_samples,
                                            posterior_samples_list = NULL) {

  tables <- list()

  for (component in names(components)) {

    info       <- components[[component]]
    indicators <- .summary_models_indicators(
      posterior_samples = posterior_samples,
      parameter         = info[["parameter"]],
      prior             = info[["prior"]]
    )
    post_probs <- vapply(seq_along(info[["names"]]), function(i) {

      mean(indicators == i)
    }, numeric(1))
    inclusion_BF <- .summary_models_inclusion_BF(
      prior_probs = info[["prior_probs"]],
      post_probs  = post_probs
    )
    BF_error_percent <- .summary_models_BF_error_percent(
      indicator_matrices = .summary_models_marginal_indicator_matrices(
        posterior_samples_list = posterior_samples_list,
        parameter              = info[["parameter"]],
        prior                  = info[["prior"]]
      ),
      prior_probs        = info[["prior_probs"]],
      post_probs         = post_probs,
      inclusion_BF       = inclusion_BF
    )

    table <- data.frame(
      Hypothesis   = info[["hypothesis"]],
      prior_prob   = info[["prior_probs"]],
      post_prob    = post_probs,
      inclusion_BF = inclusion_BF,
      row.names    = info[["names"]],
      check.names  = FALSE
    )

    table[["inclusion_BF"]] <- BayesTools::format_BF(
      table[["inclusion_BF"]],
      inclusion = TRUE
    )
    table <- .summary_models_add_BF_diagnostics(
      table            = table,
      BF_error_percent = BF_error_percent
    )

    class(table)             <- c("BayesTools_table", "RoBMA_summary_models_marginal", class(table))
    attr(table, "type")      <- c(
      "string_left", "prior_prob", "post_prob", "inclusion_BF",
      .summary_models_BF_diagnostic_types(BF_error_percent)
    )
    attr(table, "rownames")  <- TRUE
    attr(table, "title")     <- component
    attr(table, "footnotes") <- NULL
    attr(table, "warnings")  <- NULL

    tables[[component]] <- table
  }

  return(tables)
}

.summary_models_individual_table <- function(components, posterior_samples,
                                             posterior_samples_list = NULL) {

  component_names <- names(components)
  component_grid  <- expand.grid(
    lapply(components, function(info) seq_along(info[["names"]])),
    KEEP.OUT.ATTRS   = FALSE,
    stringsAsFactors = FALSE
  )

  prior_columns <- data.frame(
    lapply(seq_along(components), function(i) {

      info <- components[[i]]
      info[["names"]][component_grid[[i]]]
    }),
    check.names = FALSE
  )
  names(prior_columns) <- component_names

  prior_probs <- vapply(seq_len(nrow(component_grid)), function(i) {

    prod(vapply(seq_along(components), function(j) {

      components[[j]][["prior_probs"]][component_grid[i, j]]
    }, numeric(1)))
  }, numeric(1))

  indicators <- lapply(components, function(info) {

    .summary_models_indicators(
      posterior_samples = posterior_samples,
      parameter         = info[["parameter"]],
      prior             = info[["prior"]]
    )
  })

  post_probs <- vapply(seq_len(nrow(component_grid)), function(i) {

    selected <- rep(TRUE, nrow(posterior_samples))
    for (j in seq_along(components)) {
      selected <- selected & indicators[[j]] == component_grid[i, j]
    }
    mean(selected)
  }, numeric(1))

  inclusion_BF <- .summary_models_inclusion_BF(
    prior_probs = prior_probs,
    post_probs  = post_probs
  )
  BF_error_percent <- .summary_models_BF_error_percent(
    indicator_matrices = .summary_models_individual_indicator_matrices(
      components             = components,
      component_grid         = component_grid,
      posterior_samples_list = posterior_samples_list
    ),
    prior_probs        = prior_probs,
    post_probs         = post_probs,
    inclusion_BF       = inclusion_BF
  )

  table <- cbind(
    prior_columns,
    data.frame(
      prior_prob   = prior_probs,
      post_prob    = post_probs,
      inclusion_BF = BayesTools::format_BF(inclusion_BF, inclusion = TRUE),
      check.names  = FALSE
    )
  )
  table <- .summary_models_add_BF_diagnostics(
    table            = table,
    BF_error_percent = BF_error_percent
  )

  class(table)             <- c("BayesTools_table", "RoBMA_summary_models_individual", class(table))
  attr(table, "type")      <- c(
    rep("string_left", length(components)),
    "prior_prob", "post_prob", "inclusion_BF",
    .summary_models_BF_diagnostic_types(BF_error_percent)
  )
  attr(table, "rownames")  <- FALSE
  attr(table, "title")     <- "Individual Models"
  attr(table, "footnotes") <- NULL
  attr(table, "warnings")  <- NULL

  return(table)
}

.summary_models_prior_names <- function(prior) {

  prior_names <- names(prior)
  if (is.null(prior_names)) {
    prior_names <- rep("", length(prior))
  }

  missing_names <- !nzchar(prior_names)
  if (any(missing_names)) {
    prior_names[missing_names] <- vapply(prior[missing_names], function(p) {

      prior_name <- paste(capture.output(print(p)), collapse = " ")
      prior_name <- sub(" ~.*$", "", prior_name)
      trimws(prior_name)
    }, character(1))
  }

  return(make.unique(prior_names))
}

.summary_models_hypothesis <- function(prior) {

  components <- attr(prior, "components")
  hypothesis <- ifelse(components == "null", "Null", "Alternative")

  return(hypothesis)
}

.summary_models_prior_probs <- function(prior) {

  prior_probs <- attr(prior, "prior_weights")
  prior_probs <- prior_probs / sum(prior_probs)

  return(prior_probs)
}

.summary_models_indicators <- function(posterior_samples, parameter, prior) {

  return(.extract_posterior_indicator(
    posterior_samples = posterior_samples,
    parameter         = parameter,
    prior             = prior
  ))
}

.summary_models_posterior_samples_list <- function(fit) {

  posterior_samples_list <- suppressWarnings(coda::as.mcmc.list(fit))
  posterior_samples_list <- lapply(posterior_samples_list, as.matrix)

  return(posterior_samples_list)
}

.summary_models_indicators_list <- function(posterior_samples_list, parameter,
                                            prior) {

  if (is.null(posterior_samples_list)) {
    return(NULL)
  }

  indicators_list <- lapply(posterior_samples_list, function(samples) {

    .summary_models_indicators(
      posterior_samples = samples,
      parameter         = parameter,
      prior             = prior
    )
  })

  return(indicators_list)
}

.summary_models_marginal_indicator_matrices <- function(posterior_samples_list,
                                                        parameter, prior) {

  indicators_list <- .summary_models_indicators_list(
    posterior_samples_list = posterior_samples_list,
    parameter              = parameter,
    prior                  = prior
  )
  if (is.null(indicators_list)) {
    return(NULL)
  }

  indicator_matrices <- lapply(indicators_list, function(indicators) {

    out <- matrix(
      as.numeric(outer(indicators, seq_len(length(prior)), "==")),
      nrow = length(indicators),
      ncol = length(prior)
    )
    colnames(out) <- paste0("component_", seq_len(length(prior)))

    return(out)
  })

  return(indicator_matrices)
}

.summary_models_individual_indicator_matrices <- function(components,
                                                          component_grid,
                                                          posterior_samples_list) {

  if (is.null(posterior_samples_list)) {
    return(NULL)
  }

  indicators_list <- lapply(components, function(info) {

    .summary_models_indicators_list(
      posterior_samples_list = posterior_samples_list,
      parameter              = info[["parameter"]],
      prior                  = info[["prior"]]
    )
  })

  indicator_matrices <- lapply(seq_along(posterior_samples_list), function(i) {

    out <- matrix(
      0,
      nrow = nrow(posterior_samples_list[[i]]),
      ncol = nrow(component_grid)
    )
    for (j in seq_len(nrow(component_grid))) {
      selected <- rep(TRUE, nrow(posterior_samples_list[[i]]))
      for (k in seq_along(components)) {
        selected <- selected & indicators_list[[k]][[i]] == component_grid[j, k]
      }
      out[, j] <- as.numeric(selected)
    }
    colnames(out) <- paste0("model_", seq_len(nrow(component_grid)))

    return(out)
  })

  return(indicator_matrices)
}

.summary_models_BF_error_percent <- function(indicator_matrices, prior_probs,
                                             post_probs, inclusion_BF) {

  if (is.null(indicator_matrices)) {
    return(NULL)
  }

  indicator_mcmc <- coda::mcmc.list(lapply(indicator_matrices, coda::mcmc))
  mcmc_summary   <- summary(indicator_mcmc, quantiles = NULL)[["statistics"]]
  if (is.null(dim(mcmc_summary))) {
    mcmc_summary <- t(mcmc_summary)
  }

  MCMC_error        <- mcmc_summary[, "Time-series SE"]
  BF_error_percent <- 100 * MCMC_error / (post_probs * (1 - post_probs))
  invalid          <- post_probs <= 0 | post_probs >= 1 |
    prior_probs <= 0 | prior_probs >= 1 |
    !is.finite(inclusion_BF) | !is.finite(MCMC_error)
  BF_error_percent[invalid] <- NA_real_

  return(BF_error_percent)
}

.summary_models_add_BF_diagnostics <- function(table, BF_error_percent) {

  if (is.null(BF_error_percent)) {
    return(table)
  }

  table[["BF_error_percent"]] <- BF_error_percent
  attr(table[["BF_error_percent"]], "name") <- "error%(Inclusion BF)"

  return(table)
}

.summary_models_BF_diagnostic_types <- function(BF_error_percent) {

  if (is.null(BF_error_percent)) {
    return(character())
  }

  return("BF_error")
}

.summary_models_inclusion_BF <- function(prior_probs, post_probs) {

  inclusion_BF <- vapply(seq_along(prior_probs), function(i) {

    BayesTools::inclusion_BF(
      prior_probs = prior_probs,
      post_probs  = post_probs,
      is_null     = seq_along(prior_probs) != i
    )
  }, numeric(1))

  return(inclusion_BF)
}

context("Summary models")

source(testthat::test_path("common-functions.R"))

skip_if_no_fits()
fit_names <- list_fits()
fits      <- lazy_fits(fit_names, validate = FALSE)

summary_model_names <- intersect(
  names(fits),
  catalog_fits(class = c("BMA.norm", "BMA.glmm", "RoBMA"))
)

expect_summary_models_marginal_table <- function(table, name, component) {

  info <- paste0("summary_models marginal '", component, "' for '", name, "'")

  expect_s3_class(table, "RoBMA_summary_models_marginal")
  expect_true(is.data.frame(table), info = info)
  expect_true(nrow(table) > 0L, info = info)
  expect_true(all(c("Hypothesis", "prior_prob", "post_prob", "inclusion_BF") %in%
                    colnames(table)), info = info)
  expect_true("BF_error_percent" %in% colnames(table), info = info)
  expect_true(all(table[["Hypothesis"]] %in% c("Null", "Alternative")),
              info = info)
  expect_true(all(is.finite(table[["prior_prob"]])), info = info)
  expect_true(all(is.finite(table[["post_prob"]])), info = info)
  expect_true(all(is.finite(table[["BF_error_percent"]]) |
                    is.na(table[["BF_error_percent"]])), info = info)
  expect_equal(attr(table[["BF_error_percent"]], "name"),
               "error%(Inclusion BF)", info = info)
  expect_equal(sum(table[["prior_prob"]]), 1, tolerance = sqrt(.Machine$double.eps),
               info = info)
  expect_equal(sum(table[["post_prob"]]), 1, tolerance = sqrt(.Machine$double.eps),
               info = info)
  expect_equal(attr(table, "title"), component, info = info)
}

expect_summary_models_marginal <- function(out, name) {

  expect_s3_class(out, "summary_models.RoBMA")
  expect_equal(out[["type"]], "marginal")
  expect_type(out[["name"]], "character")
  expect_true(length(out[["name"]]) == 1L && nzchar(out[["name"]]))
  expect_type(out[["marginal"]], "list")
  expect_true(length(out[["marginal"]]) > 0L,
              info = paste0("marginal components for '", name, "'"))

  for (component in names(out[["marginal"]])) {
    expect_summary_models_marginal_table(
      out[["marginal"]][[component]],
      name      = name,
      component = component
    )
  }
}

expect_summary_models_individual <- function(out, marginal, name) {

  table <- out[["individual"]]
  info  <- paste0("summary_models individual for '", name, "'")

  expect_s3_class(out, "summary_models.RoBMA")
  expect_equal(out[["type"]], "individual")
  expect_s3_class(table, "RoBMA_summary_models_individual")
  expect_true(is.data.frame(table), info = info)
  expect_true(nrow(table) > 0L, info = info)
  expect_true(all(c("prior_prob", "post_prob", "inclusion_BF") %in%
                    colnames(table)), info = info)
  expect_true("BF_error_percent" %in% colnames(table), info = info)
  expect_true(all(names(marginal[["marginal"]]) %in% colnames(table)),
              info = info)
  expect_equal(
    nrow(table),
    prod(vapply(marginal[["marginal"]], nrow, integer(1))),
    info = info
  )
  expect_true(all(is.finite(table[["prior_prob"]])), info = info)
  expect_true(all(is.finite(table[["post_prob"]])), info = info)
  expect_true(all(is.finite(table[["BF_error_percent"]]) |
                    is.na(table[["BF_error_percent"]])), info = info)
  expect_equal(attr(table[["BF_error_percent"]], "name"),
               "error%(Inclusion BF)", info = info)
  expect_equal(sum(table[["prior_prob"]]), 1, tolerance = sqrt(.Machine$double.eps),
               info = info)
  expect_equal(sum(table[["post_prob"]]), 1, tolerance = sqrt(.Machine$double.eps),
               info = info)
  expect_equal(attr(table, "title"), "Individual Models", info = info)
}

expect_printed_summary_models <- function(out, name) {

  output <- capture.output(print(out))
  expect_true(any(nzchar(output)),
              info = paste0("printed summary_models for '", name, "'"))
  expect_false(any(grepl("__xXx__", output, fixed = TRUE)),
               info = paste0("printed summary_models labels for '", name, "'"))
}


test_that("summary_models marginal summaries have stable structure", {

  for (name in summary_model_names) {
    out <- summary_models(fits[[name]], type = "marginal")

    expect_summary_models_marginal(out, name)
    expect_printed_summary_models(out, name)
  }
})

test_that("summary_models individual summaries have stable structure", {

  for (name in summary_model_names) {
    marginal   <- summary_models(fits[[name]], type = "marginal")
    individual <- summary_models(fits[[name]], type = "individual")

    expect_summary_models_individual(individual, marginal, name)
    expect_printed_summary_models(individual, name)
  }
})

test_that("summary_models individual weights marginalize to component weights", {

  for (name in summary_model_names) {
    marginal   <- summary_models(fits[[name]], type = "marginal")
    individual <- summary_models(fits[[name]], type = "individual")[["individual"]]

    for (component in names(marginal[["marginal"]])) {
      component_table <- marginal[["marginal"]][[component]]

      for (level in rownames(component_table)) {
        selected <- individual[[component]] == level

        expect_equal(
          sum(individual[["prior_prob"]][selected]),
          component_table[level, "prior_prob"],
          tolerance = sqrt(.Machine$double.eps),
          info      = paste(name, component, level, "prior marginalization")
        )
        expect_equal(
          sum(individual[["post_prob"]][selected]),
          component_table[level, "post_prob"],
          tolerance = sqrt(.Machine$double.eps),
          info      = paste(name, component, level, "posterior marginalization")
        )
      }
    }
  }
})

test_that("summary_models can omit BF diagnostics", {

  skip_if_missing_fits("dat.lehmann2018_RoBMA")

  marginal <- summary_models(
    fits[["dat.lehmann2018_RoBMA"]],
    type                     = "marginal",
    include_mcmc_diagnostics = FALSE
  )
  individual <- summary_models(
    fits[["dat.lehmann2018_RoBMA"]],
    type                     = "individual",
    include_mcmc_diagnostics = FALSE
  )

  for (component in names(marginal[["marginal"]])) {
    table <- marginal[["marginal"]][[component]]
    expect_false("BF_error_percent" %in% colnames(table))
    expect_false("BF_error" %in% attr(table, "type"))
  }
  expect_false("BF_error_percent" %in% colnames(individual[["individual"]]))
  expect_false("BF_error" %in% attr(individual[["individual"]], "type"))
})

test_that("summary_models is RoBMA-only", {

  skip_if_missing_fits("bcg_meta-analysis")

  expect_error(
    summary_models(fits[["bcg_meta-analysis"]]),
    "RoBMA objects"
  )
})

test_that("summary_models decodes interaction labels", {

  skip_if_missing_fits("dat.lehmann2018_RoBMA_mods2")

  marginal_output <- capture.output(print(summary_models(
    fits[["dat.lehmann2018_RoBMA_mods2"]],
    type = "marginal"
  )))
  individual_output <- capture.output(print(summary_models(
    fits[["dat.lehmann2018_RoBMA_mods2"]],
    type = "individual"
  )))

  expect_false(any(grepl("__xXx__", marginal_output, fixed = TRUE)))
  expect_false(any(grepl("__xXx__", individual_output, fixed = TRUE)))
  expect_true(any(grepl("Preregistered:Gender", marginal_output, fixed = TRUE)))
  expect_true(any(grepl("Preregistered:Gender", individual_output, fixed = TRUE)))
})

# ============================================================================ #
# formula-design.R
# ============================================================================ #
#
# Shared accessors for BayesTools formula-design metadata.
#
# ============================================================================ #


# Return the BayesTools formula design for a brma formula parameter.
.fitted_formula_design <- function(object, parameter, required = TRUE) {

  design <- NULL

  if (!is.null(object[["fit"]])) {
    design <- BayesTools::JAGS_formula_design(object[["fit"]], parameter)
  }

  if (is.null(design)) {
    design <- .object_formula_design(object, parameter)
  }

  if (is.null(design) && required) {
    stop(
      "Fitted formula design metadata for parameter '", parameter,
      "' is missing. Refit the model with the current RoBMA/BayesTools build.",
      call. = FALSE
    )
  }

  return(design)
}


# Return formula design metadata stored on or reconstructed from a brma object.
.object_formula_design <- function(object, parameter) {

  if (!is.null(object[["formula_design"]][[parameter]])) {
    return(object[["formula_design"]][[parameter]])
  }

  source <- .fitted_formula_source(parameter)
  if (is.null(source) ||
      is.null(object[["data"]]) ||
      is.null(object[["data"]][[source]])) {
    return(NULL)
  }

  if (!is.null(object[["priors"]]) &&
      !is.null(object[["priors"]][[source]])) {
    return(.object_bayestools_formula_design(
      object    = object,
      parameter = parameter,
      source    = source
    ))
  }

  return(.object_data_formula_design(
    object    = object,
    parameter = parameter,
    source    = source
  ))
}


# Reconstruct BayesTools formula design for objects that store data and priors
# but do not have a fitted JAGS object, e.g. only_priors objects.
.object_bayestools_formula_design <- function(object, parameter, source) {

  formula_design <- BayesTools::JAGS_formula(
    formula       = .create_fit_formula_list(data = object[["data"]], parameter = source),
    parameter     = parameter,
    data          = .create_fit_formula_data_list(data = object[["data"]], parameter = source),
    prior_list    = .create_fit_formula_prior_list(priors = object[["priors"]], parameter = source),
    formula_scale = .data_standardize_continuous_predictors(object[["data"]])
  )[["formula_design"]]

  return(formula_design)
}


# Build minimal formula design metadata for data-only objects. These objects do
# not have priors, so BayesTools::JAGS_formula() cannot be used.
.object_data_formula_design <- function(object, parameter, source) {

  formula <- .create_fit_formula_list(data = object[["data"]], parameter = source)
  data    <- object[["data"]][[source]]
  terms   <- attr(data, "terms")

  if (is.null(terms)) {
    terms <- stats::terms(formula, data = data)
  }

  model_matrix <- stats::model.matrix(terms, data = data)
  column_names <- colnames(model_matrix)
  assign       <- attr(model_matrix, "assign")
  term_labels  <- attr(terms, "term.labels")
  term_labels  <- gsub(":", "__xXx__", term_labels, fixed = TRUE)

  if (is.null(assign)) {
    assign <- seq_len(ncol(model_matrix)) - 1L
  }

  qr_x    <- qr(model_matrix)
  aliased <- rep(FALSE, ncol(model_matrix))
  if (qr_x[["rank"]] < ncol(model_matrix)) {
    aliased[qr_x[["pivot"]][seq(qr_x[["rank"]] + 1L, ncol(model_matrix))]] <- TRUE
  }
  names(aliased) <- column_names

  predictors <- all.vars(formula)
  predictors <- predictors[predictors %in% names(data)]
  predictor_types <- stats::setNames(
    vapply(data[predictors], .formula_design_predictor_type, character(1)),
    predictors
  )

  model_terms <- term_labels
  if (isTRUE(attr(terms, "intercept") == 1L)) {
    model_terms <- c("intercept", model_terms)
  }
  model_terms_type <- stats::setNames(
    vapply(model_terms, .formula_design_term_type, character(1), predictor_types = predictor_types),
    model_terms
  )

  xlevels <- lapply(data[vapply(data, is.factor, logical(1))], levels)

  design <- list(
    parameter        = parameter,
    formula          = formula,
    model_frame      = data,
    model_matrix     = model_matrix,
    column_names     = column_names,
    raw_column_names = column_names,
    assign           = assign,
    terms            = terms,
    contrasts        = attr(model_matrix, "contrasts"),
    xlevels          = xlevels,
    predictors       = predictors,
    predictor_types  = predictor_types,
    model_terms      = model_terms,
    model_terms_type = model_terms_type,
    prior_list       = NULL,
    formula_scale    = NULL,
    rank             = qr_x[["rank"]],
    qr_pivot         = qr_x[["pivot"]],
    aliased          = aliased,
    transformed_terms = list(),
    random_effects    = list(),
    jags_data_names   = list()
  )
  class(design) <- c("BayesTools_formula_design", "list")

  return(design)
}


# Classify a formula predictor for minimal data-only designs.
.formula_design_predictor_type <- function(x) {

  if (is.factor(x) || is.character(x)) {
    return("factor")
  }

  return("continuous")
}


# Classify a formula term from the predictor types.
.formula_design_term_type <- function(term, predictor_types) {

  if (identical(term, "intercept")) {
    return("continuous")
  }

  components <- unlist(strsplit(term, "__xXx__", fixed = TRUE), use.names = FALSE)
  if (any(predictor_types[components] == "factor", na.rm = TRUE)) {
    return("factor")
  }

  return("continuous")
}


# Map a BayesTools formula parameter to the RoBMA data/prior source.
.fitted_formula_source <- function(parameter) {

  switch(
    parameter,
    "mu"      = "mods",
    "log_tau" = "scale",
    NULL
  )
}


# Map a RoBMA formula source name to the BayesTools formula parameter.
.fitted_formula_parameter <- function(source) {

  switch(
    source,
    "mods"  = "mu",
    "scale" = "log_tau",
    source
  )
}


# Return fitted model terms, optionally converted to user-facing interaction names.
.fitted_formula_terms <- function(object, parameter, include_intercept = TRUE,
                                  display = TRUE, required = TRUE) {

  design <- .fitted_formula_design(
    object    = object,
    parameter = parameter,
    required  = required
  )

  if (is.null(design)) {
    return(NULL)
  }

  terms <- design[["model_terms"]]
  if (!include_intercept) {
    terms <- terms[terms != "intercept"]
  }
  if (display) {
    terms <- .formula_design_display_names(terms)
  }

  return(terms)
}


# Return a fitted formula with an evaluation environment suitable for model.frame().
.fitted_formula_evaluable <- function(object, design, source) {

  formula <- design[["formula"]]
  if (is.null(formula)) {
    return(NULL)
  }

  stored_formula <- attr(object[["data"]][[source]], "formula")
  if (!is.null(stored_formula) && !is.null(environment(stored_formula))) {
    environment(formula) <- environment(stored_formula)
  } else {
    environment(formula) <- baseenv()
  }

  return(formula)
}


# Return fitted formula column display names keyed by internal matrix names.
.fitted_formula_column_display_map <- function(design) {

  if (is.null(design[["column_names"]]) || is.null(design[["raw_column_names"]])) {
    return(NULL)
  }

  if (length(design[["column_names"]]) != length(design[["raw_column_names"]])) {
    return(NULL)
  }

  return(stats::setNames(
    .formula_design_display_names(design[["raw_column_names"]]),
    design[["column_names"]]
  ))
}


# Convert BayesTools' JAGS-safe interaction separator to formula syntax.
.formula_design_display_names <- function(x) {

  return(gsub("__xXx__", ":", x, fixed = TRUE))
}


# ---------------------------------------------------------------------------- #
# .get_model_matrix
# ---------------------------------------------------------------------------- #
#
# Extract the fitted location model matrix X from a brma object.
#
# ---------------------------------------------------------------------------- #
.get_model_matrix <- function(object) {

  is_mods  <- .is_mods(object)
  is_PET   <- .is_PET(object)
  is_PEESE <- .is_PEESE(object)

  if (!is_mods) {
    K <- nrow(object[["data"]][["outcome"]])
    X <- matrix(1, nrow = K, ncol = 1)
    colnames(X) <- "(Intercept)"
    attr(X, "assign") <- 0L
  } else {
    design <- .fitted_formula_design(object, "mu", required = TRUE)
    X      <- design[["model_matrix"]]

    attr(X, "assign")       <- design[["assign"]]
    attr(X, "rank")         <- design[["rank"]]
    attr(X, "aliased")      <- design[["aliased"]]
    attr(X, "term_labels")  <- .fitted_formula_terms(
      object            = object,
      parameter         = "mu",
      include_intercept = FALSE,
      display           = TRUE,
      required          = TRUE
    )

    raw_colnames <- .fitted_formula_column_display_map(design)
    if (!is.null(raw_colnames)) {
      attr(X, "raw_colnames") <- raw_colnames
    }
  }

  if (is_PET || is_PEESE) {
    outcome_data <- object[["data"]][["outcome"]]
    direction    <- if (.effect_direction(object) == "negative") -1 else 1
    assign       <- attr(X, "assign")
    raw_colnames <- attr(X, "raw_colnames")
    aliased      <- attr(X, "aliased")
    term_labels  <- attr(X, "term_labels")
    if (is.null(assign)) {
      assign <- seq_len(ncol(X)) - 1L
    }
    next_assign  <- max(assign) + 1L

    if (is_PET) {
      X            <- cbind(X, PET = direction * outcome_data[["sei"]])
      assign       <- c(assign, next_assign)
      raw_colnames <- c(raw_colnames, PET = "PET")
      aliased      <- c(aliased, PET = FALSE)
      term_labels  <- c(term_labels, "PET")
      next_assign  <- next_assign + 1L
    }
    if (is_PEESE) {
      X            <- cbind(X, PEESE = direction * outcome_data[["sei"]]^2)
      assign       <- c(assign, next_assign)
      raw_colnames <- c(raw_colnames, PEESE = "PEESE")
      aliased      <- c(aliased, PEESE = FALSE)
      term_labels  <- c(term_labels, "PEESE")
      next_assign  <- next_assign + 1L
    }

    attr(X, "assign") <- assign
    if (!is.null(raw_colnames) && length(raw_colnames) == ncol(X)) {
      attr(X, "raw_colnames") <- raw_colnames
    }
    if (!is.null(aliased) && length(aliased) == ncol(X)) {
      attr(X, "aliased") <- aliased
    }
    if (!is.null(term_labels)) {
      attr(X, "term_labels") <- term_labels
    }
  }

  attr(X, "rank") <- qr(X)[["rank"]]

  return(X)
}

context("BayesTools forward API integration")

if (!exists("fit_cache_paths", mode = "function")) {
  source(testthat::test_path("common-functions.R"))
}

.read_cached_fit_with_formula_design <- function(name, parameter = "mu") {

  paths <- unique(c(
    fit_cache_paths(name)[["fit"]],
    testthat::test_path("test_files", "fits", paste0(name, ".RDS"))
  ))

  for (path in paths[file.exists(paths)]) {
    object <- suppressWarnings(try(readRDS(path), silent = TRUE))
    if (inherits(object, "try-error") || is.null(object[["fit"]])) {
      next
    }

    design <- BayesTools::JAGS_formula_design(object[["fit"]], parameter)
    if (!is.null(design)) {
      return(list(object = object, design = design))
    }
  }

  skip(paste0("Cached fit '", name, "' with formula design metadata is unavailable."))
}

test_that("BayesTools forward API guard passes for the active namespace", {

  expect_true(isTRUE(.check_bayestools_forward_api()))
  expect_true(all(
    c("fit", "parameter") %in% names(formals(BayesTools::JAGS_formula_design))
  ))
  expect_true("interpret_records" %in% getNamespaceExports("BayesTools"))
  expect_true(all(
    c("legend", "legend_title", "legend_labels", "legend_position") %in%
      names(formals(BayesTools::plot_marginal))
  ))
})

test_that(".get_model_matrix uses fitted BayesTools design metadata", {

  outcome <- data.frame(
    yi  = c(0.10, 0.20, 0.30),
    sei = c(0.20, 0.30, 0.40)
  )
  data <- list(
    outcome = outcome,
    mods    = data.frame(x = c(1, 2, 3))
  )
  attr(data, "mods")             <- TRUE
  attr(data, "effect_direction") <- "negative"

  X_design <- cbind(
    mu_intercept    = 1,
    x__xXx__factorB = c(0, 1, 0)
  )
  attr(X_design, "assign") <- c(0L, 1L)

  design <- list(
    model_matrix     = X_design,
    column_names     = colnames(X_design),
    raw_column_names = c("(Intercept)", "x:factorB"),
    assign           = c(0L, 1L),
    rank             = 2L,
    aliased          = stats::setNames(c(FALSE, FALSE), colnames(X_design))
  )
  fit <- structure(list(), class = "BayesTools_fit")
  attr(fit, "formula_design") <- list(mu = design)

  object <- list(
    data   = data,
    fit    = fit,
    priors = list(
      outcome = list(
        bias = BayesTools::prior_PET("normal", parameters = list(mean = 0, sd = 1))
      )
    )
  )

  X <- .get_model_matrix(object)

  expect_equal(as.numeric(X[, colnames(X_design)]), as.numeric(X_design))
  expect_equal(X[, "PET"], -outcome[["sei"]])
  expect_equal(attr(X, "assign"), c(0L, 1L, 2L))
  expect_equal(attr(X, "rank"), 3L)
  expect_equal(
    attr(X, "raw_colnames"),
    c(mu_intercept = "(Intercept)", x__xXx__factorB = "x:factorB", PET = "PET")
  )
})

test_that(".get_model_matrix preserves BayesTools design metadata from real fits", {

  cached <- .read_cached_fit_with_formula_design("bcg_meta-regression3")
  object <- cached[["object"]]
  design <- cached[["design"]]
  X      <- .get_model_matrix(object)

  expect_s3_class(design, "BayesTools_formula_design")
  expect_equal(dim(X), dim(design[["model_matrix"]]))
  expect_equal(as.numeric(X), as.numeric(design[["model_matrix"]]))
  expect_equal(colnames(X), design[["column_names"]])
  expect_equal(attr(X, "assign"), design[["assign"]])
  expect_equal(attr(X, "rank"), design[["rank"]])
  expect_equal(
    attr(X, "raw_colnames"),
    stats::setNames(design[["raw_column_names"]], design[["column_names"]])
  )
  expect_true(any(grepl("__xXx__", colnames(X), fixed = TRUE)))
  expect_true(any(grepl(":", attr(X, "raw_colnames"), fixed = TRUE)))
})

test_that("formula design accessors support current non-fitted objects", {

  dat <- data.frame(
    yi  = c(.10, .20, .15, .30),
    sei = c(.10, .12, .11, .15),
    x   = c(-1, 0, 1, 2),
    z   = factor(c("a", "b", "a", "b"))
  )

  prior_object <- BMA(
    yi           = yi,
    sei          = sei,
    mods         = ~ x + z,
    data         = dat,
    measure      = "SMD",
    only_priors  = TRUE
  )

  prior_design <- .fitted_formula_design(prior_object, "mu", required = TRUE)

  expect_s3_class(prior_design, "BayesTools_formula_design")
  expect_true(all(c("intercept", "x", "z") %in% prior_design[["model_terms"]]))
  expect_true("mu_x" %in% names(prior_design[["prior_list"]]))

  data_object <- brma.norm(
    yi        = yi,
    sei       = sei,
    mods      = ~ x + z,
    data      = dat,
    only_data = TRUE
  )

  data_design <- .fitted_formula_design(data_object, "mu", required = TRUE)

  expect_s3_class(data_design, "BayesTools_formula_design")
  expect_equal(nrow(data_design[["model_matrix"]]), nrow(dat))
  expect_equal(data_design[["assign"]], attr(data_design[["model_matrix"]], "assign"))
  expect_true(all(c("intercept", "x", "z") %in% data_design[["model_terms"]]))
})

test_that("Weightfunction observed p-values use selection mapping sign", {

  outcome <- data.frame(
    yi  = c(0.10, -0.20, 0.30),
    sei = c(0.20, 0.25, 0.50)
  )
  data <- list(outcome = outcome)
  attr(data, "effect_direction") <- "negative"

  object <- list(
    data   = data,
    priors = list(
      outcome = list(
        bias = BayesTools::prior_weightfunction(
          "one-sided",
          c(0.05),
          weights = BayesTools::wf_fixed(c(1, 0.5))
        )
      )
    )
  )

  expect_equal(
    .weightfunction_observed_p_values(object),
    stats::pnorm(-outcome[["yi"]] / outcome[["sei"]], lower.tail = FALSE)
  )
})

#' @title Summarize brma Object
#'
#' @description \code{summary.brma} creates summary tables for a
#' brma object. For RoBMA objects, inclusion summaries are printed before
#' parameter estimates.
#'
#' @param object a fitted brma object
#' @param probs quantiles of the posterior samples to be displayed.
#' Defaults to \code{c(.025, .50, .975)}
#' @param include_mcmc_diagnostics whether to include MCMC diagnostics in the output.
#' Defaults to \code{TRUE}.
#' @param standardized_coefficients whether to show standardized meta-regression coefficients.
#' Defaults to \code{FALSE}. When set to \code{TRUE}, standardized meta-regression
#' coefficients are returned for the intercept and continuous predictors. These coefficients
#' correspond to the standardized scale on which prior distributions are specified by default
#' (i.e., `standardize_continuous_predictors = TRUE`).
#' @param conditional whether to include conditional estimates for RoBMA
#'   product-space objects. Defaults to \code{FALSE}.
#' @param logBF whether to show inclusion Bayes factors on the log scale.
#' Defaults to \code{FALSE}.
#' @param BF01 whether to show inverse inclusion Bayes factors. Defaults to
#' \code{FALSE}.
#' @param ... additional arguments
#'
#' @return A list of class `summary.brma` with model name, optional RoBMA
#' inclusion tables, common estimates, moderator estimates, scale estimates,
#' publication-bias estimates, and optional conditional estimates. The printed
#' form displays the non-empty tables.
#'
#' @examples \dontrun{
#' if (requireNamespace("metadat", quietly = TRUE)) {
#'   data(dat.lehmann2018, package = "metadat")
#'
#'   fit <- bPET(
#'     yi      = yi,
#'     vi      = vi,
#'     data    = dat.lehmann2018,
#'     measure = "SMD",
#'     seed    = 1,
#'     silent  = TRUE
#'   )
#'
#'   summary(fit)
#' }
#' }
#'
#'
#' @seealso [brma()], [brma.glmm()]
#' @export
summary.brma       <- function(
    object, probs = c(.025, .50, .975),
    include_mcmc_diagnostics  = TRUE,
    standardized_coefficients = FALSE,
    conditional               = FALSE,
    logBF                     = FALSE,
    BF01                      = FALSE, ...) {

  ### model information
  is_mods       <- .is_mods(object)
  is_scale      <- .is_scale(object)
  is_multilevel <- .is_multilevel(object)
  is_bias       <- .is_bias(object)
  is_robma      <- .is_RoBMA(object)
  outcome_type  <- .outcome_type(object)

  BayesTools::check_bool(include_mcmc_diagnostics, "include_mcmc_diagnostics")
  BayesTools::check_bool(conditional, "conditional")
  BayesTools::check_bool(logBF, "logBF")
  BayesTools::check_bool(BF01, "BF01")
  if (conditional && !is_robma) {
    stop("'conditional' summaries are available only for RoBMA objects.",
         call. = FALSE)
  }

  ### special cases handling
  # deal with `only_data` fit
  if (is.null(object[["priors"]]) && is.null(object[["fit"]])) {
    return(object[["data"]])
  }

  ### provide common estimates
  common_parameters <- intersect(
    c("mu", "tau", "rho"),
    names(attr(object[["fit"]], "prior_list"))
  )
  estimates_common_pair <- .summary_estimates_pair(
    enabled                  = length(common_parameters) > 0L,
    object                   = object,
    probs                    = probs,
    include_mcmc_diagnostics = include_mcmc_diagnostics,
    is_robma                 = is_robma,
    conditional              = conditional,
    main_args                = list(
      transform_factors = TRUE,
      transform_scaled  = !standardized_coefficients,
      keep_parameters   = common_parameters,
      title             = if (is_mods || is_scale) "Common Estimates" else "Estimates"
    ),
    conditional_args         = list(
      transform_factors = TRUE,
      transform_scaled  = !standardized_coefficients,
      keep_parameters   = common_parameters,
      title             = if (is_mods || is_scale) {
        "Conditional Common Estimates"
      } else {
        "Conditional Estimates"
      }
    )
  )
  estimates_common             <- estimates_common_pair[["estimates"]]
  estimates_common_conditional <- estimates_common_pair[["conditional"]]

  ### provide regression estimates for the effect size meta-regression
  estimates_mods_pair <- .summary_estimates_pair(
    enabled                  = is_mods,
    object                   = object,
    probs                    = probs,
    include_mcmc_diagnostics = include_mcmc_diagnostics,
    is_robma                 = is_robma,
    conditional              = conditional,
    main_args                = list(
      transform_factors = TRUE,
      transform_scaled  = !standardized_coefficients,
      keep_formulas     = "mu",
      formula_prefix    = FALSE,
      title             = if (is_scale) "Location" else "Meta-Regression"
    ),
    conditional_args         = list(
      transform_factors = TRUE,
      transform_scaled  = !standardized_coefficients,
      keep_formulas     = "mu",
      formula_prefix    = FALSE,
      title             = if (is_scale) {
        "Conditional Location"
      } else {
        "Conditional Meta-Regression"
      }
    )
  )
  estimates_mods             <- estimates_mods_pair[["estimates"]]
  estimates_mods_conditional <- estimates_mods_pair[["conditional"]]

  ### provide regression estimates for the scale meta-regression
  scale_footnotes <- "exp(Intercept) corresponds to the between-study heterogeneity tau; the meta-regression coefficients correspond to the multiplicative effects on log-scale."
  estimates_scale_pair <- .summary_estimates_pair(
    enabled                  = is_scale,
    object                   = object,
    probs                    = probs,
    include_mcmc_diagnostics = include_mcmc_diagnostics,
    is_robma                 = is_robma,
    conditional              = conditional,
    main_args                = list(
      transform_factors = TRUE,
      transform_scaled  = !standardized_coefficients,
      keep_formulas     = "log_tau",
      formula_prefix    = FALSE,
      title             = "Scale",
      footnotes         = scale_footnotes
    ),
    conditional_args         = list(
      transform_factors = TRUE,
      transform_scaled  = !standardized_coefficients,
      keep_formulas     = "log_tau",
      formula_prefix    = FALSE,
      title             = "Conditional Scale",
      footnotes         = scale_footnotes
    )
  )
  estimates_scale             <- estimates_scale_pair[["estimates"]]
  estimates_scale_conditional <- estimates_scale_pair[["conditional"]]

  ### provide publication bias estimates
  bias_footnotes <- if (.is_weightfunction(object)) {
    "P-value intervals for publication bias weights omega correspond to one-sided p-values."
  }
  estimates_bias_pair <- .summary_estimates_pair(
    enabled                  = is_bias,
    object                   = object,
    probs                    = probs,
    include_mcmc_diagnostics = include_mcmc_diagnostics,
    is_robma                 = is_robma,
    conditional              = conditional,
    main_args                = list(
      keep_parameters = c("bias", "omega", "PET", "PEESE"),
      title           = "Publication Bias",
      footnotes       = bias_footnotes
    ),
    conditional_args         = list(
      keep_parameters = c("bias", "omega", "PET", "PEESE"),
      title           = "Conditional Publication Bias",
      footnotes       = bias_footnotes
    )
  )
  estimates_bias             <- estimates_bias_pair[["estimates"]]
  estimates_bias_conditional <- estimates_bias_pair[["conditional"]]

  ### provide RoBMA inclusion summaries
  if (is_robma) {
    inclusion <- .summary.RoBMA_inclusion_tables(
      object                   = object,
      include_mcmc_diagnostics = include_mcmc_diagnostics,
      logBF                    = logBF,
      BF01                     = BF01
    )
  } else {
    inclusion <- list(
      inclusion_components = list(),
      inclusion_mods       = list(),
      inclusion_scale      = list()
    )
  }

  out <- list(
    name                        = .summary.brma_model_names(object),
    inclusion_components        = inclusion[["inclusion_components"]],
    inclusion_mods              = inclusion[["inclusion_mods"]],
    inclusion_scale             = inclusion[["inclusion_scale"]],
    estimates                   = estimates_common,
    estimates_conditional       = estimates_common_conditional,
    estimates_mods              = estimates_mods,
    estimates_mods_conditional  = estimates_mods_conditional,
    estimates_scale             = estimates_scale,
    estimates_scale_conditional = estimates_scale_conditional,
    estimates_bias              = estimates_bias,
    estimates_bias_conditional  = estimates_bias_conditional
  )

  class(out) <- "summary.brma"
  attr(out, "mods")         <- is_mods
  attr(out, "scale")        <- is_scale
  attr(out, "multilevel")   <- is_multilevel
  attr(out, "bias")         <- is_bias
  attr(out, "RoBMA")        <- is_robma
  attr(out, "outcome_type") <- outcome_type

  return(out)
}

#' @rdname summary.brma
#' @param x a `summary.brma` or fitted `brma` object.
#' @export
print.summary.brma <- function(x, ...) {

  cat("\n")
  cat(x[["name"]])
  cat("\n")

  for (type in c(
    "inclusion_components", "inclusion_mods", "inclusion_scale",
    "estimates", "estimates_conditional",
    "estimates_mods", "estimates_mods_conditional",
    "estimates_scale", "estimates_scale_conditional",
    "estimates_bias", "estimates_bias_conditional"
  )) {
    if (length(x[[type]]) > 0) {
      cat("\n")
      print(x[[type]])
    }
  }


  cat("\n")

  return(invisible(x))
}

#' @rdname summary.brma
#' @export
print.brma <- function(x, ...) {
  print(summary(x, ...))
}

.summary_estimates_table <- function(object, probs, include_mcmc_diagnostics,
                                     is_robma, conditional = FALSE, ...) {

  args <- list(
    fit                = object[["fit"]],
    conditional        = conditional,
    remove_diagnostics = !include_mcmc_diagnostics,
    remove_inclusion   = if (conditional) TRUE else is_robma,
    probs              = probs
  )
  args <- c(args, list(...))
  if (.summary_function_has_argument(
    BayesTools::JAGS_estimates_table,
    "diagnostic_columns"
  )) {
    args[["diagnostic_columns"]] <- .summary_estimates_diagnostic_columns(
      include_mcmc_diagnostics
    )
  }

  return(do.call(BayesTools::JAGS_estimates_table, args))
}

.summary_estimates_pair <- function(enabled, object, probs,
                                    include_mcmc_diagnostics, is_robma,
                                    conditional, main_args,
                                    conditional_args = main_args) {

  if (!enabled) {
    return(list(estimates = list(), conditional = list()))
  }

  shared_args <- list(
    object                   = object,
    probs                    = probs,
    include_mcmc_diagnostics = include_mcmc_diagnostics,
    is_robma                 = is_robma
  )

  estimates <- do.call(
    .summary_estimates_table,
    c(shared_args, main_args)
  )

  if (conditional) {
    estimates_conditional <- do.call(
      .summary_estimates_table,
      c(shared_args, list(conditional = TRUE), conditional_args)
    )
  } else {
    estimates_conditional <- list()
  }

  return(list(
    estimates   = estimates,
    conditional = estimates_conditional
  ))
}

.summary.brma_model_names <- function(object) {

  is_mods       <- .is_mods(object)
  is_scale      <- .is_scale(object)
  is_multilevel <- .is_multilevel(object)

  if (.is_robust_RoBMA(object)) {
    model_name <- "Robust Bayesian Model-Averaged"
  } else if (.is_BMA(object)) {
    model_name <- "Bayesian Model-Averaged"
  } else {
    model_name <- "Bayesian"
  }

  if (is_multilevel) {
    model_name <- paste(model_name, "Multilevel")
  }

  if (is_scale) {
    model_name <- paste(model_name, "Location-Scale")
  } else if (is_mods) {
    model_name <- paste(model_name, "Mixed-Effect")
  } else if (!is_mods && !is_scale) {
    model_name <- paste(model_name, "Random-Effects")
  }

  if ("bselmodel" %in% class(object)) {
    model_name <- paste(model_name, "Selection")
  } else if ("bPET" %in% class(object)) {
    model_name <- paste(model_name, "PET")
  } else if ("bPEESE" %in% class(object)) {
    model_name <- paste(model_name, "PEESE")
  }

  n_effects <- nrow(object[["data"]][["outcome"]])

  model_name <- paste(model_name, "Model")
  if (is_multilevel) {
    n_clusters <- length(unique(object[["data"]][["outcome"]][["cluster"]]))
    model_name <- paste(model_name, sprintf("(k = %1$i, clusters = %2$i)", n_effects, n_clusters))
  } else {
    model_name <- paste(model_name, sprintf("(k = %1$i)", n_effects))
  }

  return(model_name)
}

# Split product-space inclusion inference into summary sections.
.summary.RoBMA_inclusion_tables <- function(object, include_mcmc_diagnostics,
                                            logBF, BF01) {

  args <- list(
    fit            = object[["fit"]],
    formula_prefix = TRUE
  )
  if (.summary_function_has_argument(
    BayesTools::JAGS_inference_table,
    "logBF"
  )) {
    args[["logBF"]] <- logBF
  } else if (isTRUE(logBF)) {
    stop("Installed BayesTools does not support 'logBF' in inclusion summaries.",
         call. = FALSE)
  }
  if (.summary_function_has_argument(
    BayesTools::JAGS_inference_table,
    "BF01"
  )) {
    args[["BF01"]] <- BF01
  } else if (isTRUE(BF01)) {
    stop("Installed BayesTools does not support 'BF01' in inclusion summaries.",
         call. = FALSE)
  }
  if (.summary_function_has_argument(
    BayesTools::JAGS_inference_table,
    "BF_diagnostic_columns"
  )) {
    args[["BF_diagnostic_columns"]] <- .summary_BF_diagnostic_columns(
      include_mcmc_diagnostics
    )
  } else {
    args[["BF_diagnostics"]] <- include_mcmc_diagnostics
  }

  inclusion <- do.call(BayesTools::JAGS_inference_table, args)
  if (!.summary_function_has_argument(
    BayesTools::JAGS_inference_table,
    "BF_diagnostic_columns"
  )) {
    inclusion <- .summary_filter_legacy_BF_diagnostics(
      inclusion                 = inclusion,
      include_mcmc_diagnostics = include_mcmc_diagnostics
    )
  }

  parameters <- attr(inclusion, "parameters")
  row_labels <- rownames(inclusion)

  core_map <- c(
    mu                = "Effect",
    mu_intercept      = "Effect",
    tau               = "Heterogeneity",
    log_tau_intercept = "Heterogeneity",
    bias              = "Publication Bias"
  )
  core_parameters <- names(core_map)[names(core_map) %in% parameters]
  core_indices    <- match(core_parameters, parameters)

  mods_indices <- grep("^mu_", parameters)
  mods_indices <- mods_indices[parameters[mods_indices] != "mu_intercept"]

  scale_indices <- grep("^log_tau_", parameters)
  scale_indices <- scale_indices[parameters[scale_indices] != "log_tau_intercept"]

  output <- list(
    inclusion_components = .summary.inclusion_subtable(
      table      = inclusion,
      indices    = core_indices,
      row_labels = unname(core_map[core_parameters]),
      title      = "Component Inclusion"
    ),
    inclusion_mods = .summary.inclusion_subtable(
      table      = inclusion,
      indices    = mods_indices,
      row_labels = .summary_parameter_label(
        sub("^\\(mu\\) ", "", row_labels[mods_indices])
      ),
      title      = if (.is_scale(object)) {
        "Location Inclusion"
      } else {
        "Meta-Regression Inclusion"
      }
    ),
    inclusion_scale = .summary.inclusion_subtable(
      table      = inclusion,
      indices    = scale_indices,
      row_labels = .summary_parameter_label(
        sub("^\\(log_tau\\) ", "", row_labels[scale_indices])
      ),
      title      = "Scale Inclusion"
    )
  )

  return(output)
}

.summary_estimates_diagnostic_columns <- function(include_mcmc_diagnostics) {

  if (isTRUE(include_mcmc_diagnostics)) {
    return(c("MCMC_error", "MCMC_SD_error", "ESS", "R_hat"))
  }

  return("none")
}

.summary_BF_diagnostic_columns <- function(include_mcmc_diagnostics) {

  if (isTRUE(include_mcmc_diagnostics)) {
    return("BF_error_percent")
  }

  return("none")
}

.summary_function_has_argument <- function(fun, argument) {

  return(argument %in% names(formals(fun)))
}

.summary_filter_legacy_BF_diagnostics <- function(inclusion,
                                                  include_mcmc_diagnostics) {

  keep_columns <- c("prior_prob", "post_prob", "inclusion_BF")
  if (isTRUE(include_mcmc_diagnostics)) {
    keep_columns <- c(keep_columns, "BF_error_percent")
  }
  keep_columns <- intersect(keep_columns, colnames(inclusion))

  custom_attributes <- attributes(inclusion)
  custom_attributes <- custom_attributes[!names(custom_attributes) %in%
    c("names", "row.names", "class")]

  inclusion <- inclusion[, keep_columns, drop = FALSE]
  for (attribute in names(custom_attributes)) {
    attr(inclusion, attribute) <- custom_attributes[[attribute]]
  }
  attr(inclusion, "type") <- unname(c(
    prior_prob       = "prior_prob",
    post_prob        = "post_prob",
    inclusion_BF     = "inclusion_BF",
    BF_error_percent = "BF_error"
  )[colnames(inclusion)])

  return(inclusion)
}

# Convert internal interaction separators back to formula syntax.
.summary_parameter_label <- function(label) {

  return(gsub("__xXx__", ":", label, fixed = TRUE))
}

# Create a labelled BayesTools inclusion subtable.
.summary.inclusion_subtable <- function(table, indices, row_labels, title) {

  if (length(indices) == 0) {
    return(list())
  }

  output <- table[indices, , drop = FALSE]
  output <- .summary.inclusion_subtable_restore_BF_attributes(
    output  = output,
    table   = table,
    indices = indices
  )
  rownames(output) <- row_labels

  attr(output, "parameters") <- row_labels
  attr(output, "title")      <- title

  return(output)
}

# Restore row-level BayesTools_BF attributes after table subsetting.
.summary.inclusion_subtable_restore_BF_attributes <- function(output, table,
                                                              indices) {

  for (column in intersect(colnames(output), colnames(table))) {
    if (inherits(table[[column]], "BayesTools_BF")) {
      output[[column]] <- table[[column]][indices]
    }
  }

  return(output)
}

#' @importFrom graphics hist lines
#' @importFrom stats coef cooks.distance dfbetas fitted hatvalues influence logLik model.matrix nobs plogis predict qlogis qqnorm residuals rstandard rstudent terms update vcov
#' @importFrom utils capture.output getS3method
NULL

.onLoad <- function(libname, pkgname) {

  requireNamespace("BayesTools")
  requireNamespace("runjags")
  requireNamespace("mvtnorm")

  .check_bayestools_forward_api()

  RoBMA.private$RoBMA_version   <- utils::packageDescription(pkgname, fields = "Version")
  RoBMA.private$module_location <- .RoBMA_module_location(libname, pkgname)
  RoBMA.private$lib_name        <- libname

  .load_RoBMA_module(pkgname = pkgname, warn = TRUE)
  .load_RoBMA_native_routines(pkgname = pkgname, libname = libname, warn = TRUE)
  .check_RoBMA_native_routines(pkgname = pkgname)

  setopts <- mget(".RoBMA.options", envir = .GlobalEnv, ifnotfound = list(.RoBMA.options = NULL))[[1]]
  if (!is.null(setopts)) {
    if (!is.list(setopts)) {
      warning("Ignoring invalid non-list .RoBMA.options on loading RoBMA.",
              call. = FALSE)
    } else {
      do.call("RoBMA.options", args = setopts)
    }
  }

  .check_max_cores()
  .register_posterior_methods()
  .register_loo_methods()
}

.onAttach <- function(libname, pkgname) {

  packageStartupMessage(paste0(
    "Welcome to RoBMA ", utils::packageVersion(pkgname), ".\n",
    "See `vignette('v00-introduction', package = 'RoBMA')` for introduction to the package."
  ))
}

.check_bayestools_forward_api <- function() {

  required <- c(
    "formula_add_intercept",
    "plot_transformed_prior",
    "JAGS_formula_design"
  )
  missing <- required[!vapply(
    required,
    function(x) exists(x, envir = asNamespace("BayesTools"), inherits = FALSE),
    logical(1)
  )]

  posterior_plot_args         <- names(formals(BayesTools::plot_posterior))
  marginal_plot_args          <- names(formals(BayesTools::plot_marginal))
  missing_posterior_plot_args <- setdiff(
    c("data", "show_data", "dots_data"),
    posterior_plot_args
  )
  missing_marginal_plot_args  <- setdiff(
    c("legend", "legend_title", "legend_labels", "legend_position"),
    marginal_plot_args
  )

  if (length(missing) > 0 ||
      length(missing_posterior_plot_args) > 0 ||
      length(missing_marginal_plot_args) > 0) {
    details <- character(0)
    if (length(missing) > 0) {
      details <- c(details, paste0("missing functions: ", paste(missing, collapse = ", ")))
    }
    if (length(missing_posterior_plot_args) > 0) {
      details <- c(
        details,
        paste0(
          "BayesTools::plot_posterior() missing arguments: ",
          paste(missing_posterior_plot_args, collapse = ", ")
        )
      )
    }
    if (length(missing_marginal_plot_args) > 0) {
      details <- c(
        details,
        paste0(
          "BayesTools::plot_marginal() missing arguments: ",
          paste(missing_marginal_plot_args, collapse = ", ")
        )
      )
    }

    stop(
      "RoBMA requires a BayesTools build with the forward APIs (",
      paste(details, collapse = "; "),
      ").",
      call. = FALSE
    )
  }

  invisible(TRUE)
}

.onUnload <- function(libpath) {

  tryCatch(
    {
      if ("RoBMA" %in% rjags::list.modules()) {
        rjags::unload.module("RoBMA")
      }
    },
    error = function(e) NULL
  )
}

.RoBMA_module_location <- function(libname, pkgname) {

  arch             <- if (.Platform$r_arch != "") .Platform$r_arch else ""
  module_location  <- file.path(libname, pkgname, "libs", arch)
  module_file      <- file.path(module_location, paste0("RoBMA", .Platform$dynlib.ext))

  if (file.exists(module_file)) {
    return(normalizePath(module_location, winslash = "/", mustWork = TRUE))
  }

  source_location <- .RoBMA_source_module_location(pkgname)
  if (!is.null(source_location)) {
    return(source_location)
  }

  dll_path <- .RoBMA_loaded_dll_path(pkgname)
  if (is.null(dll_path)) {
    return(NULL)
  }

  return(dirname(dll_path))
}

.RoBMA_source_module_location <- function(pkgname) {

  source_path <- tryCatch(getNamespaceInfo(pkgname, "path"), error = function(e) NULL)
  if (is.null(source_path)) {
    return(NULL)
  }

  arch <- if (.Platform$r_arch != "") .Platform$r_arch else ""
  paths <- c(
    file.path(source_path, "src", arch),
    file.path(source_path, "src")
  )
  paths <- unique(paths[nzchar(paths)])

  for (path in paths) {
    module_file <- file.path(path, paste0("RoBMA", .Platform$dynlib.ext))
    if (file.exists(module_file)) {
      return(normalizePath(path, winslash = "/", mustWork = TRUE))
    }
  }

  return(NULL)
}

.RoBMA_loaded_dll_path <- function(pkgname) {

  dlls <- getLoadedDLLs()
  if (!pkgname %in% names(dlls)) {
    return(NULL)
  }

  dll_path <- dlls[[pkgname]][["path"]]
  if (is.null(dll_path) || !file.exists(dll_path)) {
    return(NULL)
  }

  return(normalizePath(dll_path, winslash = "/", mustWork = TRUE))
}

.load_RoBMA_module <- function(pkgname = "RoBMA", path = RoBMA.private$module_location,
                               quiet = TRUE, warn = FALSE) {

  loaded <- tryCatch("RoBMA" %in% rjags::list.modules(), error = function(e) FALSE)
  if (loaded) {
    return(TRUE)
  }

  if (is.null(path) || !dir.exists(path)) {
    if (warn) {
      warning(
        "RoBMA JAGS module was not found in the installed package library. ",
        "Model fitting requires this module; reinstall RoBMA after installing JAGS >= 4.3.1.",
        call. = FALSE
      )
    }
    return(FALSE)
  }

  load_error <- NULL
  tryCatch(
    rjags::load.module("RoBMA", path = path, quiet = quiet),
    error = function(e) load_error <<- conditionMessage(e)
  )

  loaded <- tryCatch("RoBMA" %in% rjags::list.modules(), error = function(e) FALSE)
  if (!loaded && warn) {
    message <- paste0(
      "RoBMA JAGS module failed to load from '", path, "'. ",
      "Model fitting requires this module; reinstall RoBMA after installing JAGS >= 4.3.1."
    )
    if (!is.null(load_error)) {
      message <- paste0(message, " rjags error: ", load_error)
    }
    warning(message, call. = FALSE)
  }

  return(loaded)
}

.load_RoBMA_native_routines <- function(pkgname = "RoBMA", libname = RoBMA.private$lib_name,
                                        warn = FALSE) {

  if (isTRUE(.check_RoBMA_native_routines(pkgname = pkgname, warn = FALSE))) {
    return(TRUE)
  }

  load_error <- NULL
  tryCatch(
    library.dynam("RoBMA", pkgname, libname),
    error = function(e) load_error <<- conditionMessage(e)
  )

  if (!isTRUE(.check_RoBMA_native_routines(pkgname = pkgname, warn = FALSE)) &&
      !is.null(RoBMA.private$module_location)) {
    module_file <- file.path(RoBMA.private$module_location, paste0("RoBMA", .Platform$dynlib.ext))
    if (file.exists(module_file)) {
      tryCatch(
        dyn.load(module_file),
        error = function(e) load_error <<- conditionMessage(e)
      )
    }
  }

  loaded <- .check_RoBMA_native_routines(pkgname = pkgname, warn = FALSE)
  if (!isTRUE(loaded) && warn) {
    message <- paste0(
      "RoBMA native routines failed to load from the package DLL. ",
      "Compiled likelihood helpers will be unavailable; reinstall RoBMA after installing JAGS >= 4.3.1."
    )
    if (!is.null(load_error)) {
      message <- paste0(message, " R loader error: ", load_error)
    }
    warning(message, call. = FALSE)
  }

  return(isTRUE(loaded))
}

.check_RoBMA_native_routines <- function(pkgname = "RoBMA", warn = TRUE) {

  required_symbols <- c(
    "RoBMA_selnorm_kernel_loglik_matrix",
    "RoBMA_glmm_binom_marginal_loglik",
    "RoBMA_glmm_pois_marginal_loglik"
  )

  loaded <- vapply(
    required_symbols,
    is.loaded,
    FUN.VALUE = logical(1),
    PACKAGE   = pkgname
  )

  if (!all(loaded) && warn) {
    warning(
      "RoBMA native routines are not loaded from the package DLL. ",
      "Compiled likelihood helpers will be unavailable.",
      call. = FALSE
    )
  }

  return(invisible(all(loaded)))
}


# ---------------------------------------------------------------------------- #
# S3 method registration for posterior package
# ---------------------------------------------------------------------------- #

# Register as_draws methods with the posterior package when it is available.
# This allows posterior::as_draws(brma_object) to work correctly.
.register_posterior_methods <- function() {

  if (!requireNamespace("posterior", quietly = TRUE)) {
    return(invisible(NULL))
  }

  # Register methods for brma class
  .s3_register("posterior::as_draws",        "brma")
  .s3_register("posterior::as_draws_array",  "brma")
  .s3_register("posterior::as_draws_df",     "brma")
  .s3_register("posterior::as_draws_list",   "brma")
  .s3_register("posterior::as_draws_matrix", "brma")
  .s3_register("posterior::as_draws_rvars",  "brma")

  # Register methods for brma_samples class
  .s3_register("posterior::as_draws",        "brma_samples")
  .s3_register("posterior::as_draws_array",  "brma_samples")
  .s3_register("posterior::as_draws_df",     "brma_samples")
  .s3_register("posterior::as_draws_list",   "brma_samples")
  .s3_register("posterior::as_draws_matrix", "brma_samples")
  .s3_register("posterior::as_draws_rvars",  "brma_samples")

  invisible(NULL)
}

# Register loo and loo_compare methods with the loo package when it is available.
# This allows loo::loo(brma_object) and loo::loo_compare(brma_object) to work correctly.
.register_loo_methods <- function() {

  if (!requireNamespace("loo", quietly = TRUE)) {
    return(invisible(NULL))
  }

  # Register methods for brma class
  .s3_register("loo::loo",         "brma")
  .s3_register("loo::loo_compare", "brma")

  invisible(NULL)
}

# S3 method registration helper (adapted from vctrs package)
# This is a standard pattern used by tidyverse and other packages to
# conditionally register S3 methods with another package's generics.
.s3_register <- function(generic, class, method = NULL) {

  stopifnot(is.character(generic), length(generic) == 1)
  stopifnot(is.character(class), length(class) == 1)

  pieces <- strsplit(generic, "::")[[1]]
  stopifnot(length(pieces) == 2)
  package <- pieces[[1]]
  generic <- pieces[[2]]

  caller <- parent.frame()

  get_method_env <- function() {
    top <- topenv(caller)
    if (isNamespace(top)) {
      top
    } else {
      caller
    }
  }
  get_method <- function(method, env) {
    if (is.null(method)) {
      get(paste0(generic, ".", class), envir = env)
    } else {
      method
    }
  }

  register <- function(...) {
    envir <- asNamespace(package)
    method_fn <- get_method(method, get_method_env())
    registerS3method(generic, class, method_fn, envir = envir)
  }

  setHook(packageEvent(package, "onLoad"), register)

  if (isNamespaceLoaded(package)) {
    register()
  }

  invisible()
}

context("Input handling for brma.norm")

source(testthat::test_path("helper-contracts.R"))

# Test data for input specification tests
test_data <- data.frame(
  effect    = c(0.10, 0.25, 0.15, 0.30, 0.05),
  variance  = c(0.04, 0.06, 0.05, 0.08, 0.03),
  std_err   = sqrt(c(0.04, 0.06, 0.05, 0.08, 0.03)),
  n         = c(50L, 75L, 60L, 40L, 90L),
  wgt       = c(1.0, 1.5, 1.2, 0.8, 1.3),
  study     = c("A", "B", "C", "D", "E"),
  cluster   = c("g1", "g1", "g2", "g2", "g3"),
  stringsAsFactors = FALSE
)

expect_norm_outcome <- function(result, n, yi, sei = NULL, ni = NULL,
                                weights = NULL, slab = NULL, cluster = NULL) {

  expect_type(result, "list")
  expect_true("outcome" %in% names(result))
  expect_equal(nrow(result$outcome), n)
  expect_equal(result$outcome$yi, yi)

  if (!is.null(sei)) {
    expect_equal(result$outcome$sei, sei, tolerance = 1e-10)
  }
  if (!is.null(ni)) {
    expect_equal(result$outcome$ni, ni)
  }
  if (!is.null(weights)) {
    expect_equal(result$outcome$weights, weights)
  }
  if (!is.null(slab)) {
    expect_equal(result$outcome$slab, slab)
  }
  if (!is.null(cluster)) {
    expect_equal(result$outcome$cluster, cluster)
  }
}

formula_text <- function(x) {

  return(paste0(as.character(attr(x, "formula")), collapse = " "))
}


test_that("Input accepts supported outcome specifications", {

  skip_on_cran()

  external_weights <- c(2, 2, 2, 2, 2)

  input_cases <- list(
    list(
      expr = quote(brma.norm(
        yi        = c(0.10, 0.25, 0.15),
        sei       = c(0.20, 0.24, 0.22),
        only_data = TRUE
      )[["data"]]),
      n    = 3,
      yi   = c(0.10, 0.25, 0.15),
      sei  = c(0.20, 0.24, 0.22)
    ),
    list(
      expr = quote(brma.norm(
        yi        = effect,
        vi        = variance,
        data      = test_data,
        only_data = TRUE
      )[["data"]]),
      n    = 5,
      yi   = test_data$effect,
      sei  = sqrt(test_data$variance)
    ),
    list(
      expr = quote(brma.norm(
        yi        = "effect",
        sei       = "std_err",
        data      = test_data,
        only_data = TRUE
      )[["data"]]),
      n    = 5,
      yi   = test_data$effect,
      sei  = test_data$std_err
    ),
    list(
      expr = quote(brma.norm(
        yi        = effect,
        vi        = "variance",
        ni        = n,
        weights   = external_weights,
        data      = test_data,
        only_data = TRUE
      )[["data"]]),
      n       = 5,
      yi      = test_data$effect,
      sei     = sqrt(test_data$variance),
      ni      = test_data$n,
      weights = external_weights
    ),
    list(
      expr = quote(brma.norm(
        yi        = effect,
        sei       = std_err,
        ni        = n,
        slab      = study,
        cluster   = cluster,
        data      = test_data,
        only_data = TRUE
      )[["data"]]),
      n       = 5,
      yi      = test_data$effect,
      sei     = test_data$std_err,
      ni      = test_data$n,
      slab    = test_data$study,
      cluster = as.numeric(as.factor(test_data$cluster))
    )
  )

  for (case in input_cases) {
    expect_norm_outcome(
      eval(case$expr),
      n       = case$n,
      yi      = case$yi,
      sei     = case$sei,
      ni      = case$ni,
      weights = case$weights,
      slab    = case$slab,
      cluster = case$cluster
    )
  }
})


test_that("Input generates default slab when not provided", {

  skip_on_cran()

  result <- brma.norm(
    yi  = c(0.1, 0.2, 0.3),
    sei = c(0.1, 0.1, 0.1),
    only_data = TRUE
  )[["data"]]

  expect_equal(result$outcome$slab, c("Study 1", "Study 2", "Study 3"))
})


test_that("Input handles subset argument", {

  skip_on_cran()

  # Logical subset
  result_logical <- brma.norm(
    yi     = effect,
    sei    = std_err,
    subset = c(TRUE, FALSE, TRUE, FALSE, TRUE),
    data   = test_data,
    only_data = TRUE
  )[["data"]]

  expect_equal(nrow(result_logical$outcome), 3)
  expect_equal(result_logical$outcome$yi, test_data$effect[c(1, 3, 5)])

  # Numeric subset
  result_numeric <- brma.norm(
    yi     = effect,
    sei    = std_err,
    subset = c(1, 3, 5),
    data   = test_data,
    only_data = TRUE
  )[["data"]]

  expect_equal(nrow(result_numeric$outcome), 3)
  expect_equal(result_numeric$outcome$yi, test_data$effect[c(1, 3, 5)])
})


test_that("Effect direction is passed and detected after final row selection", {

  skip_on_cran()

  result_explicit <- RoBMA(
    yi               = c(-0.10, -0.20, -0.30),
    sei              = c(0.10, 0.10, 0.10),
    effect_direction = "negative",
    only_data        = TRUE
  )[["data"]]

  expect_equal(attr(result_explicit, "effect_direction"), "negative")

  result_robma_detect <- RoBMA(
    yi               = c(-0.10, -0.20, -0.30),
    sei              = c(0.10, 0.10, 0.10),
    effect_direction = "detect",
    only_data        = TRUE
  )[["data"]]

  expect_equal(attr(result_robma_detect, "effect_direction"), "negative")

  result_subset <- bselmodel(
    yi               = c(100, 90, -1, -2),
    sei              = c(0.10, 0.10, 0.10, 0.10),
    subset           = 3:4,
    effect_direction = "detect",
    only_data        = TRUE
  )[["data"]]

  expect_equal(attr(result_subset, "effect_direction"), "negative")

  data_na_direction <- data.frame(
    yi  = c(100, 90, 80, -1, -2),
    sei = c(0.10, 0.10, 0.10, 0.10, 0.10),
    mod = c(NA, NA, NA, 1, 2)
  )

  expect_warning(
    result_na <- bselmodel(
      yi               = yi,
      sei              = sei,
      mods             = ~ mod,
      data             = data_na_direction,
      effect_direction = "detect",
      only_data        = TRUE
    )[["data"]],
    regexp = "3 observation.*removed due to missing"
  )

  expect_equal(attr(result_na, "effect_direction"), "negative")
})


test_that("Input rejects invalid outcome specifications", {

  skip_on_cran()

  expect_error_cases(list(
    list(
      label  = "missing yi",
      expr   = quote(brma.norm(sei = c(0.1, 0.2), only_data = TRUE)),
      regexp = "yi"
    ),
    list(
      label  = "missing variance",
      expr   = quote(brma.norm(yi = c(0.1, 0.2), only_data = TRUE)),
      regexp = "vi|sei|variance|standard error"
    ),
    list(
      label  = "length mismatch",
      expr   = quote(brma.norm(
        yi        = c(0.1, 0.2, 0.3),
        sei       = c(0.1, 0.2),
        only_data = TRUE
      )),
      regexp = "length"
    ),
    list(
      label  = "invalid column reference",
      expr   = quote(brma.norm(
        yi        = nonexistent_column,
        sei       = std_err,
        data      = test_data,
        only_data = TRUE
      )),
      regexp = "nonexistent_column|Cannot find"
    )
  ))
})


test_that("Input handles variables from calling environment", {

  skip_on_cran()

  # Create variables in the current environment
  my_effects   <- c(0.15, 0.25, 0.35)
  my_std_errs  <- c(0.10, 0.12, 0.11)
  my_labels    <- c("Study X", "Study Y", "Study Z")

  result <- brma.norm(
    yi   = my_effects,
    sei  = my_std_errs,
    slab = my_labels,
    only_data = TRUE
  )[["data"]]

  expect_equal(result$outcome$yi, my_effects)
  expect_equal(result$outcome$sei, my_std_errs)
  expect_equal(result$outcome$slab, my_labels)
})


test_that("Input resolves direct column references inside functions", {

  skip_on_cran()

  # Wrapper function that uses NSE to pass column names
  # Note: When brma.norm is called from another function, the column names

  # must be available in the data frame or the calling environment of brma.norm
  wrapper_function <- function(data) {
    # Direct column references work because they are evaluated in data
    brma.norm(
      yi   = effect,
      sei  = std_err,
      data = data,
      only_data = TRUE
    )[["data"]]
  }

  # This tests that NSE works correctly when brma.norm is called from another function
  result <- wrapper_function(data = test_data)

  expect_type(result, "list")
  expect_equal(nrow(result$outcome), 5)
  expect_equal(result$outcome$yi, test_data$effect)
})


# ============================================================================
# Tests for moderator (mods) and scale input
# ============================================================================

# Extended test data with moderator variables
test_data_mods <- data.frame(
  effect     = c(0.10, 0.25, 0.15, 0.30, 0.05),
  variance   = c(0.04, 0.06, 0.05, 0.08, 0.03),
  std_err    = sqrt(c(0.04, 0.06, 0.05, 0.08, 0.03)),
  mod_cont   = c(1.5, 2.3, 1.8, 3.1, 0.9),
  mod_factor = factor(c("A", "B", "A", "B", "A")),
  mod_char   = c("low", "high", "low", "high", "low"),
  scale_var  = c(0.5, 1.0, 0.8, 1.2, 0.6),
  stringsAsFactors = FALSE
)


test_that("Moderator formulas preserve data columns, types, and formula metadata", {

  skip_on_cran()

  result <- brma.norm(
    yi   = effect,
    sei  = std_err,
    mods = ~ mod_cont + mod_factor,
    data = test_data_mods,
    only_data = TRUE
  )[["data"]]

  expect_type(result, "list")
  expect_true(!is.null(result$mods))
  expect_true(is.data.frame(result$mods))
  expect_equal(nrow(result$mods), 5)
  expect_true(all(c("mod_cont", "mod_factor") %in% names(result$mods)))
  expect_equal(result$mods$mod_cont, test_data_mods$mod_cont)
  expect_type(result$mods$mod_cont, "double")
  expect_s3_class(result$mods$mod_factor, "factor")
  expect_equal(levels(result$mods$mod_factor), levels(test_data_mods$mod_factor))
  expect_s3_class(attr(result$mods, "formula"), "formula")
  expect_equal(formula_text(result$mods), "~ mod_cont + mod_factor")
})


test_that("Mods accepts formula using vectors from environment", {

  skip_on_cran()

  env_mod1 <- c(10, 20, 30, 40, 50)
  env_mod2 <- factor(c("X", "Y", "X", "Y", "X"))

  result <- brma.norm(
    yi   = effect,
    sei  = std_err,
    mods = ~ env_mod1 + env_mod2,
    data = test_data_mods,
    only_data = TRUE
  )[["data"]]

  expect_true(!is.null(result$mods))
  expect_equal(nrow(result$mods), 5)
  expect_true("env_mod1" %in% names(result$mods))
  expect_true("env_mod2" %in% names(result$mods))
  expect_equal(result$mods$env_mod1, env_mod1)
  expect_equal(formula_text(result$mods), "~ env_mod1 + env_mod2")

  mf_check <- stats::model.frame(attr(result$mods, "formula"), data = result$mods)
  expect_equal(nrow(mf_check), 5)
})


test_that("Mods and scale apply logical and numeric subsets", {

  skip_on_cran()

  result_logical <- brma.norm(
    yi     = effect,
    sei    = std_err,
    mods   = ~ mod_cont + mod_factor,
    subset = c(TRUE, TRUE, FALSE, TRUE, FALSE),
    data   = test_data_mods,
    only_data = TRUE
  )[["data"]]

  result_numeric <- brma.norm(
    yi     = effect,
    sei    = std_err,
    mods   = ~ mod_cont,
    subset = c(2, 4),
    data   = test_data_mods,
    only_data = TRUE
  )[["data"]]

  result_scale <- brma.norm(
    yi        = effect,
    sei       = std_err,
    scale     = ~ scale_var,
    subset    = c(1, 3, 5),
    data      = test_data_mods,
    only_data = TRUE
  )[["data"]]

  expect_equal(nrow(result_logical$mods), 3)
  expect_equal(result_logical$mods$mod_cont, test_data_mods$mod_cont[c(1, 2, 4)])
  expect_equal(levels(result_logical$mods$mod_factor), c("A", "B"))
  expect_equal(formula_text(result_logical$mods), "~ mod_cont + mod_factor")

  expect_equal(nrow(result_numeric$mods), 2)
  expect_equal(result_numeric$mods$mod_cont, test_data_mods$mod_cont[c(2, 4)])
  expect_equal(formula_text(result_numeric$mods), "~ mod_cont")

  expect_equal(nrow(result_scale$scale), 3)
  expect_equal(result_scale$scale$scale_var, test_data_mods$scale_var[c(1, 3, 5)])
})


test_that("Scale formulas are accepted alone and with moderators", {

  skip_on_cran()

  scale_result <- brma.norm(
    yi        = effect,
    sei       = std_err,
    scale     = ~ scale_var,
    data      = test_data_mods,
    only_data = TRUE
  )[["data"]]

  joint_result <- brma.norm(
    yi    = effect,
    sei   = std_err,
    mods  = ~ mod_cont + mod_factor,
    scale = ~ scale_var,
    data  = test_data_mods,
    only_data = TRUE
  )[["data"]]

  expect_true(is.data.frame(scale_result$scale))
  expect_equal(nrow(scale_result$scale), 5)
  expect_true("scale_var" %in% names(scale_result$scale))
  expect_equal(scale_result$scale$scale_var, test_data_mods$scale_var)
  expect_equal(formula_text(scale_result$scale), "~ scale_var")

  expect_true(!is.null(joint_result$mods))
  expect_true(!is.null(joint_result$scale))
  expect_equal(nrow(joint_result$mods), 5)
  expect_equal(nrow(joint_result$scale), 5)
  expect_equal(formula_text(joint_result$mods), "~ mod_cont + mod_factor")
  expect_equal(formula_text(joint_result$scale), "~ scale_var")
})


test_that("Mods rejects length mismatch", {

  skip_on_cran()

  short_mod <- c(1, 2, 3)  # Only 3 values, but yi has 5

  expect_error(
    brma.norm(
      yi   = effect,
      sei  = std_err,
      mods = ~ short_mod,
      data = test_data_mods,
      only_data = TRUE
    ),
    regexp = "number of rows|must equal"
  )
})


test_that("Mods accepts matrix input", {

  skip_on_cran()

  mod_matrix <- matrix(1:10, nrow = 5, ncol = 2)
  colnames(mod_matrix) <- c("X1", "X2")

  result <- brma.norm(
    yi   = effect,
    sei  = std_err,
    mods = mod_matrix,
    data = test_data_mods,
    only_data = TRUE
  )[["data"]]

  expect_true(!is.null(result$mods))
  expect_true(is.data.frame(result$mods))
  expect_equal(nrow(result$mods), 5)
  expect_true(all(c("X1", "X2") %in% names(result$mods)))

  # Check formula attribute
  expect_equal(
    paste0(as.character(attr(result$mods, "formula")), collapse = " "),
    "~ X1 + X2")
})


test_that("Mods accepts data.frame input", {

  skip_on_cran()

  mod_df <- data.frame(
    predictor1 = c(1.1, 2.2, 3.3, 4.4, 5.5),
    predictor2 = factor(c("a", "b", "a", "b", "a"))
  )

  result <- brma.norm(
    yi   = effect,
    sei  = std_err,
    mods = mod_df,
    data = test_data_mods,
    only_data = TRUE
  )[["data"]]

  expect_true(!is.null(result$mods))
  expect_true(is.data.frame(result$mods))
  expect_equal(nrow(result$mods), 5)

  # Check types preserved
  expect_type(result$mods$predictor1, "double")
  expect_s3_class(result$mods$predictor2, "factor")

  # Check formula attribute
  expect_equal(
    paste0(as.character(attr(result$mods, "formula")), collapse = " "),
    "~ predictor1 + predictor2")
})


test_that("Mods formula attribute can be evaluated in clean environment", {

  skip_on_cran()

  # Use environment vectors that won't be available later
  local({
    temp_var <- c(100, 200, 300, 400, 500)

    result <- brma.norm(
      yi   = effect,
      sei  = std_err,
      mods = ~ temp_var,
      data = test_data_mods,
      only_data = TRUE
    )[["data"]]

    # Check formula attribute
    expect_equal(
      paste0(as.character(attr(result$mods, "formula")), collapse = " "),
      "~ temp_var")

    # Verify formula can be evaluated with the returned data.frame
    mf <- stats::model.frame(attr(result$mods, "formula"), data = result$mods)
    expect_equal(nrow(mf), 5)
    expect_equal(mf$temp_var, temp_var)
  })
})


test_that("Mods handles formula with LHS (with warning)", {

  skip_on_cran()

  expect_warning(
    result <- brma.norm(
      yi   = effect,
      sei  = std_err,
      mods = effect ~ mod_cont,  # LHS should be ignored
      data = test_data_mods,
      only_data = TRUE
    )[["data"]],
    regexp = "left-hand side|LHS"
  )

  expect_true(!is.null(result$mods))
  expect_true("mod_cont" %in% names(result$mods))
  expect_false("effect" %in% names(result$mods))  # LHS was ignored

  # Check formula attribute (should be ~ mod_cont, not effect ~ mod_cont)
  expect_equal(
    paste0(as.character(attr(result$mods, "formula")), collapse = " "),
    "~ mod_cont")
})


test_that("Mods accepts inline transformations in formula", {

  skip_on_cran()

  result <- brma.norm(
    yi   = effect,
    sei  = std_err,
    mods = ~ I(mod_cont - 2) + I(mod_cont^2),
    data = test_data_mods,
    only_data = TRUE
  )[["data"]]

  expect_true(!is.null(result$mods))
  expect_equal(nrow(result$mods), 5)

  # Check that transformations are correctly applied (as.numeric to strip AsIs class)
  expect_equal(as.numeric(result$mods$`I(mod_cont - 2)`), test_data_mods$mod_cont - 2)
  expect_equal(as.numeric(result$mods$`I(mod_cont^2)`), test_data_mods$mod_cont^2)

  # Check formula attribute preserves transformations
  expect_equal(
    paste0(as.character(attr(result$mods, "formula")), collapse = " "),
    "~ I(mod_cont - 2) + I(mod_cont^2)")
})


test_that("Mods accepts interaction terms in formula", {

  skip_on_cran()

  result <- brma.norm(
    yi   = effect,
    sei  = std_err,
    mods = ~ mod_cont * mod_factor,
    data = test_data_mods,
    only_data = TRUE
  )[["data"]]

  expect_true(!is.null(result$mods))
  expect_equal(nrow(result$mods), 5)

  # Should have both main effects and the original variables
  expect_true("mod_cont" %in% names(result$mods))
  expect_true("mod_factor" %in% names(result$mods))

  # Check that the formula attribute preserves the interaction structure
  expect_equal(
    paste0(as.character(attr(result$mods, "formula")), collapse = " "),
    "~ mod_cont * mod_factor")
})


test_that("Mods auto-converts character to factor", {

  skip_on_cran()

  result <- brma.norm(
    yi   = effect,
    sei  = std_err,
    mods = ~ mod_char,  # mod_char is character in test_data_mods
    data = test_data_mods,
    only_data = TRUE
  )[["data"]]

  expect_true(!is.null(result$mods))
  expect_true(is.factor(result$mods$mod_char))
  expect_equal(as.character(result$mods$mod_char), test_data_mods$mod_char)
})


test_that("yi formula auto-converts character moderators to factor", {

  skip_on_cran()

  result <- brma.norm(
    yi   = effect ~ mod_char,
    sei  = std_err,
    data = test_data_mods,
    only_data = TRUE
  )[["data"]]

  expect_true(!is.null(result$mods))
  expect_true(is.factor(result$mods$mod_char))
  expect_equal(as.character(result$mods$mod_char), test_data_mods$mod_char)
  expect_equal(paste0(as.character(attr(result$mods, "formula")), collapse = " "), "~ mod_char")
})


test_that("Mods accepts data$column syntax", {

  skip_on_cran()

  # Using $ syntax directly
  result <- brma.norm(
    yi   = test_data_mods$effect,
    sei  = test_data_mods$std_err,
    mods = ~ test_data_mods$mod_cont,
    only_data = TRUE
  )[["data"]]

  expect_true(!is.null(result$mods))
  expect_equal(nrow(result$mods), 5)

  # Check formula attribute
  expect_equal(
    paste0(as.character(attr(result$mods, "formula")), collapse = " "),
    "~ test_data_mods$mod_cont")
})


test_that("NULL mods and scale return NULL", {

  skip_on_cran()

  result <- brma.norm(
    yi   = effect,
    sei  = std_err,
    data = test_data_mods,
    only_data = TRUE
  )[["data"]]

  expect_null(result$mods)
  expect_null(result$scale)
})


# ============================================================================
# Tests for predictor validation (constant variables and single-level factors)
# ============================================================================

test_that("Degenerate moderator and scale designs are rejected", {

  skip_on_cran()

  test_data_const <- data.frame(
    effect    = c(0.10, 0.25, 0.15),
    std_err   = c(0.2, 0.25, 0.22),
    const_var = c(5, 5, 5)
  )

  test_data_single <- data.frame(
    effect     = c(0.10, 0.25, 0.15),
    std_err    = c(0.2, 0.25, 0.22),
    single_fac = factor(c("A", "A", "A"))
  )

  test_data_const_scale <- data.frame(
    effect     = c(0.10, 0.25, 0.15),
    std_err    = c(0.2, 0.25, 0.22),
    mod_cont   = c(1.5, 2.3, 1.8),
    const_scale = c(1, 1, 1)
  )

  expect_error_cases(list(
    list(
      label  = "constant numeric moderator",
      expr   = quote(brma.norm(
        yi        = effect,
        sei       = std_err,
        mods      = ~ const_var,
        data      = test_data_const,
        only_data = TRUE
      )),
      regexp = "zero variance"
    ),
    list(
      label  = "single-level factor moderator",
      expr   = quote(brma.norm(
        yi        = effect,
        sei       = std_err,
        mods      = ~ single_fac,
        data      = test_data_single,
        only_data = TRUE
      )),
      regexp = "only one level"
    ),
    list(
      label  = "subset drops all but one factor level",
      expr   = quote(brma.norm(
        yi        = effect,
        sei       = std_err,
        mods      = ~ mod_factor,
        subset    = c(TRUE, FALSE, TRUE, FALSE, TRUE),
        data      = test_data_mods,
        only_data = TRUE
      )),
      regexp = "only one level"
    ),
    list(
      label  = "constant scale predictor",
      expr   = quote(brma.norm(
        yi        = effect,
        sei       = std_err,
        mods      = ~ mod_cont,
        scale     = ~ const_scale,
        data      = test_data_const_scale,
        only_data = TRUE
      )),
      regexp = "zero variance"
    )
  ))
})


test_that("Valid mods with variation passes validation", {

  skip_on_cran()

  test_data_valid <- data.frame(
    effect   = c(0.10, 0.25, 0.15, 0.20),
    std_err  = c(0.2, 0.25, 0.22, 0.18),
    mod_cont = c(1.5, 2.3, 1.8, 3.0),
    mod_fac  = factor(c("A", "B", "A", "B"))
  )

  result <- brma.norm(
    yi   = effect,
    sei  = std_err,
    mods = ~ mod_cont + mod_fac,
    data = test_data_valid,
    only_data = TRUE
  )[["data"]]

  expect_true(!is.null(result$mods))
  expect_equal(nrow(result$mods), 4)
})


# ============================================================================
# Tests for yi ~ mods formula syntax
# ============================================================================

test_that("yi ~ mods formula syntax preserves variables and formula metadata", {

  skip_on_cran()

  formula_cases <- list(
    list(
      expr      = quote(brma.norm(
        yi        = effect ~ mod_cont,
        sei       = std_err,
        data      = test_data_mods,
        only_data = TRUE
      )[["data"]]),
      rows      = seq_len(5),
      variables = "mod_cont",
      formula   = "~ mod_cont"
    ),
    list(
      expr      = quote(brma.norm(
        effect ~ mod_cont + mod_factor,
        sei       = std_err,
        data      = test_data_mods,
        only_data = TRUE
      )[["data"]]),
      rows      = seq_len(5),
      variables = c("mod_cont", "mod_factor"),
      formula   = "~ mod_cont + mod_factor"
    ),
    list(
      expr      = quote(brma.norm(
        yi        = effect ~ 0 + mod_factor,
        sei       = std_err,
        data      = test_data_mods,
        only_data = TRUE
      )[["data"]]),
      rows      = seq_len(5),
      variables = "mod_factor",
      absent    = c("mod_factorA", "mod_factorB"),
      formula   = "~ 0 + mod_factor"
    ),
    list(
      expr      = quote(brma.norm(
        yi        = effect ~ mod_cont * mod_factor,
        sei       = std_err,
        data      = test_data_mods,
        only_data = TRUE
      )[["data"]]),
      rows      = seq_len(5),
      variables = c("mod_cont", "mod_factor"),
      formula   = "~ mod_cont * mod_factor"
    ),
    list(
      expr      = quote(brma.norm(
        yi        = effect ~ mod_cont,
        sei       = std_err,
        subset    = c(1, 3, 5),
        data      = test_data_mods,
        only_data = TRUE
      )[["data"]]),
      rows      = c(1, 3, 5),
      variables = "mod_cont",
      formula   = "~ mod_cont"
    )
  )

  for (case in formula_cases) {
    result <- eval(case$expr)

    expect_true(!is.null(result$mods))
    expect_equal(nrow(result$mods), length(case$rows))
    expect_equal(nrow(result$outcome), length(case$rows))
    expect_equal(result$outcome$yi, test_data_mods$effect[case$rows])
    expect_true(all(case$variables %in% names(result$mods)))
    expect_equal(formula_text(result$mods), case$formula)

    if ("mod_cont" %in% case$variables) {
      expect_equal(result$mods$mod_cont, test_data_mods$mod_cont[case$rows])
    }
    if ("mod_factor" %in% case$variables) {
      expect_s3_class(result$mods$mod_factor, "factor")
    }
    if (!is.null(case$absent)) {
      expect_false(any(case$absent %in% names(result$mods)))
    }
  }
})


test_that("yi ~ mods formula syntax rejects duplicate mods argument", {

  skip_on_cran()

  expect_error(
    brma.norm(
      yi   = effect ~ mod_cont,
      sei  = std_err,
      mods = ~ mod_factor,  # This should cause an error
      data = test_data_mods,
      only_data = TRUE
    ),
    regexp = "Cannot specify 'mods' when 'yi' is a formula"
  )
})


test_that("yi ~ mods formula returns equivalent results to mods = ~", {

  skip_on_cran()

  result1 <- brma.norm(
    yi   = effect ~ mod_cont + mod_factor,
    sei  = std_err,
    data = test_data_mods,
    only_data = TRUE
  )[["data"]]

  result2 <- brma.norm(
    yi   = effect,
    sei  = std_err,
    mods = ~ mod_cont + mod_factor,
    data = test_data_mods,
    only_data = TRUE
  )[["data"]]

  # Both should have the same yi values
  expect_equal(result1$outcome$yi, result2$outcome$yi)

  # yi ~ formula gives model matrix (expanded with intercept and dummies)
  # mods = ~ formula gives data.frame with original variables
  # So the structures differ, but both should have mod_cont values
  expect_true("mod_cont" %in% names(result1$mods))
  expect_true("mod_cont" %in% names(result2$mods))

  # Both should have same number of rows
  expect_equal(nrow(result1$mods), nrow(result2$mods))

  # Both should have same formula attribute
  expect_equal(paste0(as.character(attr(result1$mods, "formula")), collapse = " "), "~ mod_cont + mod_factor")
  expect_equal(paste0(as.character(attr(result2$mods, "formula")), collapse = " "), "~ mod_cont + mod_factor")
})


test_that("Input accepts metafor::escalc output", {

  skip_on_cran()
  skip_if_not_installed("metadat")
  skip_if_not_installed("metafor")

  # Load BCG vaccine data and compute log risk ratios
  data(dat.bcg, package = "metadat")
  dat <- metafor::escalc(measure = "RR", ai = tpos, bi = tneg, ci = cpos, di = cneg, data = dat.bcg)

  # Test that brma.norm can handle escalc output (yi and vi have special attributes)
  result <- brma.norm(
    yi      = yi,
    vi      = vi,
    data    = dat,
    measure = "RR",
    only_data = TRUE
  )[["data"]]

  expect_type(result, "list")
  expect_true("outcome" %in% names(result))
  expect_equal(nrow(result$outcome), nrow(dat))

  # Check that yi values are correctly extracted (as plain numeric, no attributes)

  expect_equal(result$outcome$yi, as.vector(dat$yi))
  expect_null(attributes(result$outcome$yi))

  # Check that sei is computed from vi
  expect_equal(result$outcome$sei, sqrt(as.vector(dat$vi)), tolerance = 1e-10)
  expect_null(attributes(result$outcome$sei))
})


# ----- NA Handling Tests -----

# Test data with NAs for NA handling tests
test_data_na <- data.frame(
  effect     = c(0.10, NA, 0.15, 0.30, 0.05, 0.20),
  variance   = c(0.04, 0.06, NA, 0.08, 0.03, 0.05),
  std_err    = sqrt(c(0.04, 0.06, 0.05, 0.08, 0.03, 0.05)),
  n          = c(50L, 75L, 60L, 40L, 90L, 55L),
  study      = c("A", "B", "C", "D", "E", "F"),
  cluster    = c("g1", "g1", "g2", "g2", "g3", "g3"),
  mod_cont   = c(1.5, 2.0, NA, 3.0, 1.0, 2.5),
  mod_factor = factor(c("low", "high", "low", NA, "medium", "high")),
  scale_var  = c(0.5, NA, 0.7, 0.8, 0.6, 0.9),
  stringsAsFactors = FALSE
)


test_that("Outcome NAs drop rows with warning", {

  skip_on_cran()

  na_cases <- list(
    list(
      yi          = c(0.10, NA, 0.15, 0.30, 0.05),
      sei         = c(0.20, 0.24, 0.22, 0.28, 0.17),
      regexp      = "1 observation.*removed due to missing",
      expected_yi = c(0.10, 0.15, 0.30, 0.05),
      expected_se = c(0.20, 0.22, 0.28, 0.17)
    ),
    list(
      yi          = c(0.10, 0.25, 0.15, 0.30, 0.05),
      sei         = c(0.20, 0.24, NA, 0.28, 0.17),
      regexp      = "1 observation.*removed due to missing",
      expected_yi = c(0.10, 0.25, 0.30, 0.05),
      expected_se = c(0.20, 0.24, 0.28, 0.17)
    ),
    list(
      yi          = c(0.10, NA, 0.15, 0.30, 0.05),
      sei         = c(0.20, 0.24, 0.22, NA, 0.17),
      regexp      = "2 observation.*removed due to missing",
      expected_yi = c(0.10, 0.15, 0.05),
      expected_se = c(0.20, 0.22, 0.17)
    )
  )

  for (case in na_cases) {
    expect_warning(
      result <- brma.norm(
        yi        = case$yi,
        sei       = case$sei,
        only_data = TRUE
      )[["data"]],
      regexp = case$regexp
    )

    expect_norm_outcome(
      result,
      n   = length(case$expected_yi),
      yi  = case$expected_yi,
      sei = case$expected_se
    )
  }
})


test_that("NA in moderators drops rows with warning", {

  skip_on_cran()

  # Data with NA in moderator (row 3)
  test_mods <- data.frame(
    effect   = c(0.10, 0.25, 0.15, 0.30, 0.05),
    std_err  = c(0.20, 0.24, 0.22, 0.28, 0.17),
    mod_cont = c(1.5, 2.0, NA, 3.0, 1.0)
  )

  expect_warning(
    result <- brma.norm(
      yi   = effect,
      sei  = std_err,
      mods = ~ mod_cont,
      data = test_mods,
      only_data = TRUE
    )[["data"]],
    regexp = "1 observation.*removed due to missing"
  )

  expect_equal(nrow(result$outcome), 4)
  expect_equal(nrow(result$mods), 4)
  expect_equal(result$outcome$yi, c(0.10, 0.25, 0.30, 0.05))
  expect_equal(result$mods$mod_cont, c(1.5, 2.0, 3.0, 1.0))
})


test_that("NA in factor moderator drops rows with warning", {

  skip_on_cran()

  # Data with NA in factor moderator (row 4)
  test_mods <- data.frame(
    effect     = c(0.10, 0.25, 0.15, 0.30, 0.05),
    std_err    = c(0.20, 0.24, 0.22, 0.28, 0.17),
    mod_factor = factor(c("low", "high", "low", NA, "medium"))
  )

  expect_warning(
    result <- brma.norm(
      yi   = effect,
      sei  = std_err,
      mods = ~ mod_factor,
      data = test_mods,
      only_data = TRUE
    )[["data"]],
    regexp = "1 observation.*removed due to missing"
  )

  expect_equal(nrow(result$outcome), 4)
  expect_equal(nrow(result$mods), 4)

  # Check factor levels are properly dropped
  expect_equal(levels(result$mods$mod_factor), c("high", "low", "medium"))
})


test_that("NA in scale moderator drops rows with warning", {

  skip_on_cran()

  # Data with NA in scale variable (row 2)
  test_scale <- data.frame(
    effect    = c(0.10, 0.25, 0.15, 0.30, 0.05),
    std_err   = c(0.20, 0.24, 0.22, 0.28, 0.17),
    scale_var = c(0.5, NA, 0.7, 0.8, 0.6)
  )

  expect_warning(
    result <- brma.norm(
      yi    = effect,
      sei   = std_err,
      scale = ~ scale_var,
      data  = test_scale,
      only_data = TRUE
    )[["data"]],
    regexp = "1 observation.*removed due to missing"
  )

  expect_equal(nrow(result$outcome), 4)
  expect_equal(nrow(result$scale), 4)
  expect_equal(result$outcome$yi, c(0.10, 0.15, 0.30, 0.05))
  expect_equal(result$scale$scale_var, c(0.5, 0.7, 0.8, 0.6))
})


test_that("NAs across outcome, mods, and scale are all handled", {

  skip_on_cran()

  # Data with NAs in multiple places
  # yi NA in row 1, mod NA in row 3, scale NA in row 5
  test_combined <- data.frame(
    effect    = c(NA, 0.25, 0.15, 0.30, 0.05, 0.20),
    std_err   = c(0.20, 0.24, 0.22, 0.28, 0.17, 0.19),
    mod_cont  = c(1.5, 2.0, NA, 3.0, 1.0, 2.5),
    scale_var = c(0.5, 0.6, 0.7, 0.8, NA, 0.9)
  )

  expect_warning(
    result <- brma.norm(
      yi    = effect,
      sei   = std_err,
      mods  = ~ mod_cont,
      scale = ~ scale_var,
      data  = test_combined,
      only_data = TRUE
    )[["data"]],
    regexp = "3 observation.*removed due to missing"
  )

  # Should have rows 2, 4, 6 remaining
  expect_equal(nrow(result$outcome), 3)
  expect_equal(nrow(result$mods), 3)
  expect_equal(nrow(result$scale), 3)
  expect_equal(result$outcome$yi, c(0.25, 0.30, 0.20))
  expect_equal(result$mods$mod_cont, c(2.0, 3.0, 2.5))
  expect_equal(result$scale$scale_var, c(0.6, 0.8, 0.9))
})


test_that("NA in weights or cluster is rejected instead of dropped", {

  skip_on_cran()

  # Data with NA in weights
  expect_error(
    result <- brma.norm(
      yi        = c(0.10, 0.25, 0.15, 0.30, 0.05),
      sei       = c(0.20, 0.24, 0.22, 0.28, 0.17),
      weights   = c(1.0, 1.5, NA, 0.8, 1.3),
      cluster   = c("g1", "g1", "g2", "g2", "g3"),
      only_data = TRUE
    )[["data"]], regexp = "argument must not contain missing values"
  )

  # Data with NA in cluster
  expect_error(
    result <- brma.norm(
      yi        = c(0.10, 0.25, 0.15, 0.30, 0.05),
      sei       = c(0.20, 0.24, 0.22, 0.28, 0.17),
      weights   = c(1.0, 1.5, 1, 0.8, 1.3),
      cluster   = c("g1", NA, "g2", "g2", "g3"),
      only_data = TRUE
    )[["data"]], regexp = "argument must not contain missing values"
  )
})


test_that("Slab and cluster metadata are refreshed after NA dropping", {

  skip_on_cran()

  expect_warning(
    default_slab <- brma.norm(
      yi        = c(0.10, NA, 0.15, 0.30, 0.05),
      sei       = c(0.20, 0.24, 0.22, 0.28, 0.17),
      only_data = TRUE
    )[["data"]],
    regexp = "1 observation.*removed due to missing"
  )

  expect_warning(
    cluster_result <- brma.norm(
      yi        = c(NA, NA, 0.15, 0.30, 0.05),
      sei       = c(0.20, 0.24, 0.22, 0.28, 0.17),
      cluster   = c("g1", "g1", "g2", "g2", "g3"),
      only_data = TRUE
    )[["data"]],
    regexp = "2 observation.*removed due to missing"
  )

  expect_warning(
    user_slab <- brma.norm(
      yi        = c(0.10, NA, 0.15, 0.30, 0.05),
      sei       = c(0.20, 0.24, 0.22, 0.28, 0.17),
      slab      = c("A", "B", "C", "D", "E"),
      only_data = TRUE
    )[["data"]],
    regexp = "1 observation.*removed due to missing"
  )

  expect_equal(default_slab$outcome$slab, c("Study 1", "Study 2", "Study 3", "Study 4"))
  expect_equal(cluster_result$outcome$cluster, c(1, 1, 2))
  expect_equal(user_slab$outcome$slab, c("A", "C", "D", "E"))
})


test_that("All-NA yi is rejected", {

  skip_on_cran()

  # All NAs in yi should error (caught during validation by BayesTools::check_real)
  expect_error(
    brma.norm(
      yi   = c(NA, NA, NA),
      sei  = c(0.20, 0.24, 0.22),
      only_data = TRUE
    ),
    regexp = "numeric vector"
  )
})


test_that("NA handling rejects empty post-drop data", {

  skip_on_cran()

  # Each row has at least one NA
  expect_error(
    suppressWarnings(brma.norm(
      yi   = c(NA, 0.25, 0.15),
      sei  = c(0.20, NA, 0.22),
      mods = ~ mod_cont,
      data = data.frame(
        mod_cont = c(1.5, 2.0, NA)
      ),
      only_data = TRUE
    )),
    regexp = "No observations remaining"
  )
})


test_that("NA handling supports yi ~ mods formula syntax", {

  skip_on_cran()

  # Data with NA in formula moderator
  test_formula <- data.frame(
    effect   = c(0.10, 0.25, 0.15, 0.30, 0.05),
    std_err  = c(0.20, 0.24, 0.22, 0.28, 0.17),
    mod_cont = c(1.5, NA, 2.5, 3.0, 1.0)
  )

  expect_warning(
    result <- brma.norm(
      yi   = effect ~ mod_cont,
      sei  = std_err,
      data = test_formula,
      only_data = TRUE
    )[["data"]],
    regexp = "1 observation.*removed due to missing"
  )

  expect_equal(nrow(result$outcome), 4)
  expect_equal(nrow(result$mods), 4)
  expect_equal(result$outcome$yi, c(0.10, 0.15, 0.30, 0.05))
  expect_equal(result$mods$mod_cont, c(1.5, 2.5, 3.0, 1.0))
})


test_that("NA handling supports subset argument", {

  skip_on_cran()

  # Apply subset first, then NA dropping
  # Original data has 5 rows, subset keeps rows 1-4, NA in row 2 gets dropped
  expect_warning(
    result <- brma.norm(
      yi     = c(0.10, NA, 0.15, 0.30, 0.05),
      sei    = c(0.20, 0.24, 0.22, 0.28, 0.17),
      subset = 1:4,
      only_data = TRUE
    )[["data"]],
    regexp = "1 observation.*removed due to missing"
  )

  # After subset (4 rows) then NA drop (1 row), should have 3 rows
  expect_equal(nrow(result$outcome), 3)
  expect_equal(result$outcome$yi, c(0.10, 0.15, 0.30))
})

test_that("Numeric subset must use integer indices", {

  skip_on_cran()

  expect_error(
    brma.norm(
      yi        = c(0.10, 0.25, 0.15),
      sei       = c(0.20, 0.24, 0.22),
      subset    = c(1, 2.5),
      only_data = TRUE
    ),
    regexp = "integer numeric indices"
  )
})


test_that("Formula attributes are preserved after NA dropping", {

  skip_on_cran()

  test_mods <- data.frame(
    effect   = c(0.10, 0.25, 0.15, NA, 0.05),
    std_err  = c(0.20, 0.24, 0.22, 0.28, 0.17),
    mod_cont = c(1.5, 2.0, 2.5, 3.0, 1.0)
  )

  expect_warning(
    result <- brma.norm(
      yi   = effect,
      sei  = std_err,
      mods = ~ mod_cont,
      data = test_mods,
      only_data = TRUE
    )[["data"]],
    regexp = "1 observation.*removed due to missing"
  )

  # Formula attribute should be preserved
  expect_equal(
    paste0(as.character(attr(result$mods, "formula")), collapse = " "),
    "~ mod_cont"
  )
})


test_that("NA in subset argument is rejected", {

  skip_on_cran()

  # Logical subset with NA
  expect_error(
    brma.norm(
      yi     = c(0.10, 0.25, 0.15, 0.30, 0.05),
      sei    = c(0.20, 0.24, 0.22, 0.28, 0.17),
      subset = c(TRUE, NA, TRUE, TRUE, FALSE),
      only_data = TRUE
    ),
    regexp = "must not contain NA"
  )

  # Numeric subset with NA
  expect_error(
    brma.norm(
      yi     = c(0.10, 0.25, 0.15, 0.30, 0.05),
      sei    = c(0.20, 0.24, 0.22, 0.28, 0.17),
      subset = c(1, NA, 3),
      only_data = TRUE
    ),
    regexp = "must not contain NA"
  )
})

# ============================================================================ #
# evaluate.R
# ============================================================================ #
#
# This file contains modular helper functions for evaluating posterior samples
# from brma model fits. These functions extract and transform MCMC samples for:
# - heterogeneity (tau) parameters
# - location (mu) parameters
# - true effects (theta)
# - GLMM-specific parameters (baserate, lograte)
#
# The functions are designed for:
# - Prediction (predict.brma)
# - Simulation from the posterior predictive distribution
# - Likelihood evaluation
# - Other downstream tasks requiring posterior sample manipulation
#
# Design principles:
# - Decomposed parameters: minimize memory by passing only required components
# - Vectorized operations: use outer(), sweep(), matrix ops instead of loops
# - Consistent return structures: always return lists with predictable components
#
# ============================================================================ #


# ---------------------------------------------------------------------------- #
# Posterior Extraction Helpers
# ---------------------------------------------------------------------------- #
#
# Keep posterior extraction centralized so helper functions can share an already
# materialized matrix in hot paths.
#
# ---------------------------------------------------------------------------- #
.get_posterior_samples <- function(fit, posterior_samples = NULL) {

  if (!is.null(posterior_samples)) {
    return(posterior_samples)
  }

  return(suppressWarnings(coda::as.mcmc(fit)))
}


# ---------------------------------------------------------------------------- #
# .extract_posterior_matrix
# ---------------------------------------------------------------------------- #
#
# Extract vectorized JAGS parameters such as theta[1], ..., theta[K] as an
# S x K matrix, preserving the requested column order.
#
# ---------------------------------------------------------------------------- #
.extract_posterior_matrix <- function(posterior_samples, parameter, K) {

  column_names <- paste0(parameter, "[", seq_len(K), "]")
  missing_cols <- setdiff(column_names, colnames(posterior_samples))

  if (length(missing_cols) > 0) {
    stop("Missing posterior column(s): ", paste(missing_cols, collapse = ", "),
         call. = FALSE)
  }

  return(as.matrix(posterior_samples[, column_names, drop = FALSE]))
}


# ---------------------------------------------------------------------------- #
# .extract_indexed_parameter_samples
# ---------------------------------------------------------------------------- #
#
# Extract scalar/vectorized posterior parameter columns, sorting fully indexed
# names such as `omega[10]` by numeric index instead of column order.
#
# ---------------------------------------------------------------------------- #
.extract_indexed_parameter_samples <- function(posterior_samples, parameter,
                                               n_expected = NULL,
                                               required = TRUE) {

  parameter_pattern <- gsub("([][{}()+*^$|\\\\?.])", "\\\\\\1", parameter)
  parameter_cols    <- grep(
    paste0("^", parameter_pattern, "(\\[|$)"),
    colnames(posterior_samples),
    value = TRUE
  )

  if (length(parameter_cols) == 0L) {
    if (required) {
      stop("Missing posterior ", parameter, " columns.", call. = FALSE)
    }
    return(NULL)
  }

  parameter_index <- suppressWarnings(as.integer(sub(
    paste0("^", parameter_pattern, "\\[([0-9]+)\\]$"),
    "\\1",
    parameter_cols
  )))
  if (all(!is.na(parameter_index))) {
    parameter_cols <- parameter_cols[order(parameter_index)]
  }

  if (!is.null(n_expected) && length(parameter_cols) != n_expected) {
    stop(
      "Expected ", n_expected, " posterior ", parameter, " column(s), found ",
      length(parameter_cols), ".",
      call. = FALSE
    )
  }

  return(as.matrix(posterior_samples[, parameter_cols, drop = FALSE]))
}


# ---------------------------------------------------------------------------- #
# .extract_omega_samples
# ---------------------------------------------------------------------------- #
#
# Extract selection-model omega columns from posterior samples in numeric order.
#
# @param posterior_samples posterior sample matrix.
#
# @return S x W matrix of omega samples.
#
# ---------------------------------------------------------------------------- #
.extract_omega_samples <- function(posterior_samples) {

  return(.extract_indexed_parameter_samples(posterior_samples, "omega"))
}


# ---------------------------------------------------------------------------- #
# .posterior_formula_fit
# ---------------------------------------------------------------------------- #
#
# Build a BayesTools::JAGS_evaluate_formula() input from already selected
# posterior rows while preserving formula scaling metadata from the JAGS fit.
#
# ---------------------------------------------------------------------------- #
.posterior_formula_fit <- function(fit, posterior_samples) {

  formula_fit <- if (inherits(posterior_samples, "mcmc")) {
    posterior_samples
  } else {
    coda::as.mcmc(as.matrix(posterior_samples))
  }

  attr(formula_fit, "formula_scale") <- attr(fit, "formula_scale")

  return(formula_fit)
}


# ---------------------------------------------------------------------------- #
# .evaluate.brma.tau
# ---------------------------------------------------------------------------- #
#
# Extract and compute heterogeneity (tau) posterior samples from a brma fit.
#
# This function handles:
# - Scale regression models: evaluates log_tau formula, then exponentiates
# - Simple models: extracts tau column and replicates to K columns
# - Multilevel models: splits tau into within/between components via rho
#
# @param fit         runjags fit object containing posterior samples
# @param scale_data  data.frame of scale predictors for formula evaluation
#                    (NULL if not a scale regression model)
# @param scale_formula formula object for scale regression (from attr(data$scale, "formula"))
# @param scale_priors list of priors for scale parameters
# @param is_scale    logical; whether model uses scale regression
# @param is_multilevel logical; whether model is 3-level (clustered)
# @param K           integer; number of observations (determines output columns)
#
# @param posterior_samples optional posterior sample matrix; avoids repeated
#                          coda extraction when supplied by callers.
#
# @return A list with three components (all S x K matrices):
#   - tau_total: total heterogeneity
#   - tau_within: estimate-level heterogeneity component
#   - tau_between: cluster-level heterogeneity component
#   For non-multilevel models: tau_within = tau (total), tau_between = 0 matrix
#   Total tau can be reconstructed as: sqrt(tau_within^2 + tau_between^2)
#
# ---------------------------------------------------------------------------- #
.evaluate.brma.tau <- function(fit, scale_data, scale_formula, scale_priors,
                               is_scale, is_multilevel, K,
                               posterior_samples = NULL) {

  posterior_samples <- .get_posterior_samples(fit, posterior_samples)
  S <- nrow(posterior_samples)  # number of posterior samples

  ### compute tau samples based on model type
  if (is_scale) {

    scale_priors <- .repair_formula_prior_list(
      prior_list = scale_priors,
      parameter  = "log_tau"
    )

    # scale regression: evaluate log_tau formula then exponentiate
    # BayesTools::JAGS_evaluate_formula returns K x S matrix, we need S x K
    log_tau_samples <- t(BayesTools::JAGS_evaluate_formula(
      fit        = .posterior_formula_fit(fit, posterior_samples),
      formula    = scale_formula,
      parameter  = "log_tau",
      data       = scale_data,
      prior_list = scale_priors
    ))
    tau_samples <- exp(log_tau_samples)

  } else {

    # simple model: extract tau column and replicate to K columns
    # matrix(vec, nrow = S, ncol = K) replicates vec across columns
    # equivalent to: for (k in 1:K) result[, k] <- posterior_samples[, "tau"]
    tau_samples <- matrix(posterior_samples[, "tau"], nrow = S, ncol = K)

  }

  ### split tau into within/between components for multilevel models
  if (is_multilevel) {

    # extract rho (proportion of variance at cluster-level)
    rho_samples <- posterior_samples[, "rho"]

    # clamp rho to [0, 1] to handle JAGS numerical precision issues
    # pmin/pmax are vectorized min/max: faster than rho[rho > 1] <- 1
    rho_samples <- pmin(pmax(rho_samples, 0), 1)

    # tau_within = tau * sqrt(1 - rho)  (estimate-level heterogeneity)
    # tau_between = tau * sqrt(rho)     (cluster-level heterogeneity)
    # multiplication by vector rho_samples broadcasts across columns
    tau_within_samples  <- tau_samples * sqrt(1 - rho_samples)
    tau_between_samples <- tau_samples * sqrt(rho_samples)

  } else {

    # non-multilevel: all heterogeneity is at estimate-level
    # tau_between is zero matrix for consistent interface
    tau_within_samples  <- tau_samples
    tau_between_samples <- matrix(0, nrow = S, ncol = K)

  }

  return(list(
    tau_total   = tau_samples,
    tau_within  = tau_within_samples,
    tau_between = tau_between_samples
  ))
}


# ---------------------------------------------------------------------------- #
# .evaluate.brma.mu
# ---------------------------------------------------------------------------- #
#
# Extract and compute location (mu) posterior samples from a brma fit.
#
# This function handles:
# - Meta-regression models: evaluates mu formula with moderators
# - Simple models: extracts mu column and replicates to K columns
# - PET/PEESE bias shifts (added when bias_adjusted = FALSE)
#
# The returned mu is on the original effect-size scale. For
# effect_direction = "negative", the JAGS likelihood already uses -mu against
# internally flipped outcomes, so this helper must not flip mu itself.
#
# @param fit              runjags fit object containing posterior samples
# @param outcome_data     data.frame with outcome info (must contain 'sei')
# @param mods_data        data.frame of moderators for formula evaluation
#                         (NULL if not a meta-regression model)
# @param mods_formula     formula object for meta-regression
# @param mods_priors      list of priors for moderator parameters
# @param is_mods          logical; whether model is meta-regression
# @param is_PET           logical; whether model includes PET adjustment
# @param is_PEESE         logical; whether model includes PEESE adjustment
# @param effect_direction character; "positive" or "negative" - direction of true effect
# @param bias_adjusted    logical; if TRUE, PET/PEESE adjustments are skipped
#                         (returns bias-adjusted mu); if FALSE, adds bias terms
# @param K                integer; number of observations
# @param posterior_samples optional posterior sample matrix; avoids repeated
#                          coda extraction when supplied by callers.
#
# @return S x K matrix of mu (location) posterior samples
#
# ---------------------------------------------------------------------------- #
.evaluate.brma.mu <- function(fit, outcome_data, mods_data, mods_formula, mods_priors,
                              is_mods, is_PET, is_PEESE, effect_direction,
                              bias_adjusted, K, posterior_samples = NULL) {

  posterior_samples <- .get_posterior_samples(fit, posterior_samples)
  S <- nrow(posterior_samples)

  ### compute base mu samples
  if (is_mods) {

    mods_priors <- .repair_formula_prior_list(
      prior_list = mods_priors,
      parameter  = "mu"
    )

    # meta-regression: evaluate mu formula with moderators
    # returns K x S, transpose to S x K
    mu_samples <- t(BayesTools::JAGS_evaluate_formula(
      fit        = .posterior_formula_fit(fit, posterior_samples),
      formula    = mods_formula,
      parameter  = "mu",
      data       = mods_data,
      prior_list = mods_priors
    ))

  } else {

    # simple model: replicate mu column to K columns
    mu_samples <- matrix(posterior_samples[, "mu"], nrow = S, ncol = K)

  }

  # NOTE: No effect direction flipping needed here.
  # The JAGS model uses `-mu` in the likelihood for negative effects, so `mu`
  # remains on the original effect-size scale. The data flip in
  # .create_fit_data() is matched by the likelihood sign.

  ### add PET adjustment when NOT incorporating publication bias adjustment
  # (i.e., when we want to show the biased predictions)
  # PET model in JAGS: yi_flipped ~ N(-mu + PET * sei, tau) for negative effects
  #                    yi ~ N(mu + PET * sei, tau) for positive effects
  # To get biased effect in original scale:
  #   - positive: E[yi] = mu + PET * sei
  #   - negative: E[yi_original] = -E[yi_flipped] = -(-mu + PET * sei) = mu - PET * sei
  # So the sign of PET adjustment depends on effect_direction
  if (is_PET && !bias_adjusted) {

    PET_samples <- posterior_samples[, "PET"]
    sei_vec     <- outcome_data[["sei"]]

    # vectorized: outer(PET_samples, sei_vec) creates S x K matrix
    # outer(a, b) computes a[i] * b[j] for all i,j pairs
    # direction multiplier: +1 for positive, -1 for negative
    direction <- ifelse(effect_direction == "negative", -1, 1)
    mu_samples <- mu_samples + direction * outer(PET_samples, sei_vec)

  }

  ### add PEESE adjustment when NOT incorporating publication bias adjustment
  # PEESE model: Same logic as PET but with sei^2
  if (is_PEESE && !bias_adjusted) {

    PEESE_samples <- posterior_samples[, "PEESE"]
    sei_sq_vec    <- outcome_data[["sei"]]^2

    # direction multiplier: +1 for positive, -1 for negative
    direction <- ifelse(effect_direction == "negative", -1, 1)
    mu_samples <- mu_samples + direction * outer(PEESE_samples, sei_sq_vec)

  }

  return(mu_samples)
}


# ---------------------------------------------------------------------------- #
# .evaluate.brma.bias_offset
# ---------------------------------------------------------------------------- #
#
# Compute the PET/PEESE bias offset on the original effect-size scale.
#
# For observed-data BLUPs with bias_adjusted = TRUE, the residual must be formed
# from the bias-corrected estimate:
#   yi - bias_offset - mu
# rather than:
#   yi - mu
#
# @param fit              runjags fit object containing posterior samples
# @param outcome_data     data.frame with outcome info (must contain 'sei')
# @param is_PET           logical; whether model includes PET adjustment
# @param is_PEESE         logical; whether model includes PEESE adjustment
# @param effect_direction character; "positive" or "negative" - direction of true effect
# @param K                integer; number of observations
# @param posterior_samples optional posterior sample matrix; avoids repeated
#                          coda extraction when supplied by callers.
#
# @return S x K matrix of additive bias offsets.
#
# ---------------------------------------------------------------------------- #
.evaluate.brma.bias_offset <- function(fit, outcome_data, is_PET, is_PEESE,
                                       effect_direction, K,
                                       posterior_samples = NULL) {

  posterior_samples <- .get_posterior_samples(fit, posterior_samples)
  S                 <- nrow(posterior_samples)
  bias_offset       <- matrix(0, nrow = S, ncol = K)
  direction         <- if (effect_direction == "negative") -1 else 1

  if (is_PET) {
    bias_offset <- bias_offset +
      direction * outer(posterior_samples[, "PET"], outcome_data[["sei"]])
  }

  if (is_PEESE) {
    bias_offset <- bias_offset +
      direction * outer(posterior_samples[, "PEESE"], outcome_data[["sei"]]^2)
  }

  return(bias_offset)
}


# ---------------------------------------------------------------------------- #
# .get_multilevel_block_indices
# ---------------------------------------------------------------------------- #
#
# Create a reusable block index list for multilevel covariance calculations.
#
# @param cluster integer vector of cluster identifiers
#
# @return A named list of integer index vectors, one per cluster.
#
# ---------------------------------------------------------------------------- #
.get_multilevel_block_indices <- function(cluster) {

  split(seq_along(cluster), cluster)
}


# ---------------------------------------------------------------------------- #
# .build_multilevel_marginal_covariance
# ---------------------------------------------------------------------------- #
#
# Construct the marginal covariance matrix for a 3-level normal model.
#
# The covariance decomposes into:
# - sampling variance: diag(vi)
# - estimate-level heterogeneity: diag(tau_within^2)
# - cluster-level heterogeneity: block-wise tcrossprod(tau_between)
#
# @param tau_within    numeric vector of length K with estimate-level SDs
# @param tau_between   numeric vector of length K with cluster-level SDs
# @param vi            numeric vector of length K with sampling variances
# @param block_indices list of observation indices for each cluster
#
# @return A K x K marginal covariance matrix.
#
# ---------------------------------------------------------------------------- #
.build_multilevel_marginal_covariance <- function(tau_within, tau_between, vi,
                                                  block_indices) {

  K <- length(vi)

  marginal_covariance <- diag(vi + tau_within^2, nrow = K, ncol = K)

  for (idx in block_indices) {
    marginal_covariance[idx, idx] <- marginal_covariance[idx, idx] +
      tcrossprod(tau_between[idx])
  }

  return(marginal_covariance)
}


# ---------------------------------------------------------------------------- #
# .solve_diagonal_rank_one_block
# ---------------------------------------------------------------------------- #
#
# Solve (diag(diagonal) + rank_one %*% t(rank_one))^-1 residual. The analytic
# Sherman-Morrison path is O(K) per cluster and falls back to Cholesky/generalized
# inverse when a diagonal element is non-positive.
#
# ---------------------------------------------------------------------------- #
.solve_diagonal_rank_one_block <- function(diagonal, rank_one, residual) {

  if (all(is.finite(diagonal)) && all(diagonal > 0)) {
    inv_diag_residual <- residual / diagonal
    inv_diag_rank_one <- rank_one / diagonal
    denom             <- 1 + sum(rank_one * inv_diag_rank_one)

    if (is.finite(denom) && denom > .Machine$double.eps) {
      correction <- inv_diag_rank_one * sum(rank_one * inv_diag_residual) / denom
      return(inv_diag_residual - correction)
    }
  }

  covariance <- diag(diagonal, nrow = length(diagonal), ncol = length(diagonal)) +
    tcrossprod(rank_one)
  chol_m <- try(chol(covariance), silent = TRUE)

  if (inherits(chol_m, "try-error")) {
    weights <- tryCatch(solve(covariance), error = function(e) MASS::ginv(covariance))
  } else {
    weights <- chol2inv(chol_m)
  }

  return(as.vector(weights %*% residual))
}


# ---------------------------------------------------------------------------- #
# .evaluate.brma.multilevel_blup.norm
# ---------------------------------------------------------------------------- #
#
# Compute exact same-data BLUP contributions for 3-level normal models.
#
# The conditional predictor is obtained from the marginal mixed-model identity:
#   b_hat = G M^{-1} (y - X beta)
# where G is decomposed into cluster-level and estimate-level components.
#
# @param mu_samples   S x K matrix of fixed-effect predictions
# @param tau_within   S x K matrix of estimate-level SDs
# @param tau_between  S x K matrix of cluster-level SDs
# @param yi           numeric vector of observed outcomes
# @param vi           numeric vector of sampling variances
# @param cluster      integer vector of cluster identifiers
#
# @return A list with S x K matrices `cluster` and `estimate`.
#
# ---------------------------------------------------------------------------- #
.evaluate.brma.multilevel_blup.norm <- function(mu_samples, tau_within, tau_between,
                                                yi, vi, cluster,
                                                bias_offset = NULL) {

  S <- nrow(mu_samples)
  K <- ncol(mu_samples)

  if (is.null(bias_offset)) {
    bias_offset <- matrix(0, nrow = S, ncol = K)
  } else if (!identical(dim(bias_offset), c(S, K))) {
    stop("'bias_offset' must have the same dimensions as 'mu_samples'.",
         call. = FALSE)
  }

  block_indices <- .get_multilevel_block_indices(cluster)

  cluster_contribution  <- matrix(0, nrow = S, ncol = K)
  estimate_contribution <- matrix(0, nrow = S, ncol = K)

  for (s in seq_len(S)) {
    for (idx in block_indices) {
      weighted_residual <- .solve_diagonal_rank_one_block(
        diagonal = vi[idx] + tau_within[s, idx]^2,
        rank_one = tau_between[s, idx],
        residual = yi[idx] - bias_offset[s, idx] - mu_samples[s, idx]
      )

      estimate_contribution[s, idx] <- tau_within[s, idx]^2 * weighted_residual

      cluster_scale <- sum(tau_between[s, idx] * weighted_residual)
      cluster_contribution[s, idx] <- tau_between[s, idx] * cluster_scale
    }
  }

  return(list(
    cluster  = cluster_contribution,
    estimate = estimate_contribution
  ))
}


# ---------------------------------------------------------------------------- #
# .evaluate.brma.cluster_effects
# ---------------------------------------------------------------------------- #
#
# Extract or sample cluster-level (gamma) random effects for multilevel models.
#
# For multilevel models, gamma[cluster] represents the standardized cluster-level
# random effect (i.e., gamma ~ N(0, 1)). The actual contribution to mu is
# gamma * tau_between.
#
# This function handles:
# - Same data: extracts fitted gamma samples from posterior
# - New data: samples new gamma from N(0, 1) (marginalizes over cluster effects)
#
# @param fit              runjags fit object (needed to extract gamma if same_data)
# @param tau_between      S x K matrix of cluster-level heterogeneity samples
# @param cluster          integer vector of length K; cluster ID for each observation
# @param same_data        logical; TRUE if predicting on original data (use fitted gamma)
# @param effect_direction character; "positive" or "negative" (currently unused but kept for interface)
# @param posterior_samples optional posterior sample matrix; avoids repeated
#                          coda extraction when supplied by callers.
#
# @return S x K matrix: contribution from cluster-level random effects
#         (gamma[cluster[k]] * tau_between[,k])
#         Can be added directly to mu samples.
#
# ---------------------------------------------------------------------------- #
.evaluate.brma.cluster_effects <- function(fit, tau_between, cluster,
                                           same_data, effect_direction,
                                           posterior_samples = NULL) {

  S <- nrow(tau_between)
  K <- ncol(tau_between)

  # NOTE: No direction flipping needed for cluster effects!
  # The JAGS model uses: "-gamma*tau_between" for negative effects, "+gamma*tau_between" for positive
  # But when converting to original scale:
  # E[yi_original] = -E[yi_flipped] = -(-mu - gamma*tau_between) = mu + gamma*tau_between
  # So the contribution to the original-scale effect is always +gamma*tau_between

  if (same_data) {

    # extract fitted gamma samples from posterior
    posterior_samples <- .get_posterior_samples(fit, posterior_samples)
    n_clusters        <- max(cluster)

    # extract all gamma columns at once: S x n_clusters matrix
    gamma_samples <- .extract_posterior_matrix(
      posterior_samples = posterior_samples,
      parameter         = "gamma",
      K                 = n_clusters
    )

    # gamma_samples[, cluster] reorders columns to match observations
    # this is S x K after reordering, element-wise multiply with tau_between
    cluster_contribution <- gamma_samples[, cluster, drop = FALSE] * tau_between

  } else {

    # new data: sample fresh gamma ~ N(0, 1) for each supplied cluster
    cluster_levels <- sort(unique(cluster))
    n_clusters     <- length(cluster_levels)
    gamma_new      <- matrix(stats::rnorm(S * n_clusters), nrow = S, ncol = n_clusters)
    cluster_map    <- match(cluster, cluster_levels)

    cluster_contribution <- gamma_new[, cluster_map, drop = FALSE] * tau_between

  }

  return(cluster_contribution)
}


# ---------------------------------------------------------------------------- #
# .evaluate.brma.true_effects.norm
# ---------------------------------------------------------------------------- #
#
# Compute posterior samples of true-effect summaries for normal models.
# For same-data selection models, the selection weight is constant after
# conditioning on yi, so the estimate-level shrinkage remains Gaussian
# conditional on the fitted location and heterogeneity draw.
#
# For same_data = TRUE: Uses empirical Bayes shrinkage to return the
# conditional BLUP mean, E(theta_i | y_i, mu_i, tau_i), for each posterior row:
#   theta_i = lambda_i * y_i + (1 - lambda_i) * mu_i
# where:
#   lambda_i = tau_within^2 / (tau_within^2 + se_i^2)
#
# If bias_offset is supplied, the observed estimate is first corrected by the
# posterior-row PET/PEESE offset:
#   theta_i = mu_i + lambda_i * (y_i - bias_offset_i - mu_i)
#
# For same_data = FALSE: Samples from the marginal distribution of true effects:
#   theta_i = mu_i + epsilon_i * tau_within_i, where epsilon_i ~ N(0, 1)
#
# IMPORTANT: For multilevel models, mu_samples must already include the
# cluster-level contribution (gamma * tau_between) before calling this function.
#
# @param mu_samples       S x K matrix of location samples (must include
#                         gamma * tau_between contribution for multilevel models)
# @param tau_within       S x K matrix of estimate-level heterogeneity
# @param yi               numeric vector of length K; observed effect sizes (used only if same_data = TRUE)
# @param sei              numeric vector of length K; standard errors (used only if same_data = TRUE)
# @param same_data        logical; TRUE for BLUP estimates, FALSE for marginal sampling
# @param bias_offset      optional S x K matrix of PET/PEESE offsets to subtract
#                         from yi before BLUP shrinkage.
#
# @return S x K matrix of true-effect BLUP means or new-effect posterior samples
#
# ---------------------------------------------------------------------------- #
.evaluate.brma.true_effects.norm <- function(mu_samples, tau_within, yi, sei,
                                             same_data, bias_offset = NULL) {

  S <- nrow(mu_samples)
  K <- ncol(mu_samples)

  if (same_data) {

    # BLUP: empirical Bayes conditional means for existing observations
    # lambda = tau^2 / (tau^2 + se^2) ranges from 0 (strong shrinkage)
    # to 1 (weak shrinkage).
    lambda <- tau_within^2
    lambda <- lambda / sweep(lambda, 2, sei^2, "+")

    if (is.null(bias_offset)) {
      bias_offset <- matrix(0, nrow = S, ncol = K)
    } else if (!identical(dim(bias_offset), c(S, K))) {
      stop("'bias_offset' must have the same dimensions as 'mu_samples'.",
           call. = FALSE)
    }

    # BLUP: weighted average of observed effect and model prediction
    # high tau -> lambda -> 1 -> trust data more
    # low tau -> lambda -> 0 -> trust model more
    true_effects_samples <- mu_samples +
      lambda * (sweep(mu_samples, 2, yi, function(mu, yi_i) yi_i - mu) - bias_offset)

  } else {

    # Marginal sampling: sample new theta from N(mu, tau_within)
    # epsilon ~ N(0, 1), then theta = mu + epsilon * tau_within
    epsilon <- matrix(stats::rnorm(S * K), nrow = S, ncol = K)
    true_effects_samples <- mu_samples + epsilon * tau_within

  }

  return(true_effects_samples)
}


# ---------------------------------------------------------------------------- #
# .evaluate.brma.true_effects.glmm
# ---------------------------------------------------------------------------- #
#
# Compute posterior samples of true effects (theta) for GLMM models.
#
# For GLMM models (binomial or Poisson), the estimate-level random effects
# (theta) are directly sampled in JAGS (not marginalized as in normal models).
# The true effect is:
#   true_effect_i = mu_i + theta_i * tau_within_i
#
# For same_data: extracts theta[k] from posterior samples
# For new_data: samples new theta ~ N(0, 1) to marginalize over random effects
#
# IMPORTANT: For multilevel models, mu_samples must already include the
# cluster-level contribution (gamma * tau_between) before calling this function.
# This is handled by .evaluate.brma.cluster_effects() in predict.R.
#
# @param fit              runjags fit object (needed to extract theta if same_data = TRUE)
# @param mu_samples       S x K matrix of location samples (must include
#                         gamma * tau_between contribution for multilevel models)
# @param tau_within       S x K matrix of estimate-level heterogeneity
# @param same_data        logical; TRUE if predicting on original data
# @param K                integer; number of observations
# @param posterior_samples optional posterior sample matrix; avoids repeated
#                          coda extraction when supplied by callers.
#
# @return S x K matrix of true effect (theta) posterior samples
#
# ---------------------------------------------------------------------------- #
.evaluate.brma.true_effects.glmm <- function(fit, mu_samples, tau_within, same_data, K,
                                             posterior_samples = NULL) {

  # add the estimate-level random effects (theta * tau_within) to mu
  theta_contribution <- .evaluate.brma.theta.glmm(
    fit               = fit,
    tau_within        = tau_within,
    same_data         = same_data,
    K                 = K,
    posterior_samples = posterior_samples
  )
  true_effects_samples <- mu_samples + theta_contribution

  return(true_effects_samples)
}


# ---------------------------------------------------------------------------- #
# .evaluate.brma.baserate
# ---------------------------------------------------------------------------- #
#
# Extract base-rate (pi) samples for binomial GLMM models.
#
# For binomial outcomes, logit(pi[i]) is the midpoint of the two arm logits.
# The log-odds ratio effect size is applied symmetrically around this midpoint.
#
# @param fit              runjags fit object containing posterior samples
# @param K                integer; number of observations
# @param posterior_samples optional posterior sample matrix; avoids repeated
#                          coda extraction when supplied by callers.
#
# @return A matrix of logit(pi) samples for each observation
#
# ---------------------------------------------------------------------------- #
.evaluate.brma.baserate <- function(fit, K, posterior_samples = NULL) {

  posterior_samples <- .get_posterior_samples(fit, posterior_samples)
  pi_samples        <- .extract_posterior_matrix(
    posterior_samples = posterior_samples,
    parameter         = "pi",
    K                 = K
  )

  return(.logit(pi_samples))
}


# ---------------------------------------------------------------------------- #
# .evaluate.brma.baserate_newdata
# ---------------------------------------------------------------------------- #
#
# Sample new-study base-rate values for binomial GLMM response prediction.
#
# @param prior_pi BayesTools prior object on pi.
# @param S        integer; number of posterior rows.
# @param K        integer; number of new observations.
#
# @return S x K matrix of logit(pi) samples.
#
# ---------------------------------------------------------------------------- #
.evaluate.brma.baserate_newdata <- function(prior_pi, S, K) {

  pi_samples <- .draw_prior_samples_matrix(prior = prior_pi, S = S, K = K)
  pi_samples <- pmin(pmax(pi_samples, .Machine$double.eps), 1 - .Machine$double.eps)

  return(.logit(pi_samples))
}


# ---------------------------------------------------------------------------- #
# .evaluate.brma.lograte
# ---------------------------------------------------------------------------- #
#
# Extract log-rate (phi) samples for Poisson GLMM models.
#
# For Poisson outcomes, phi[i] is the midpoint of the two arm log incidence
# rates. The incidence rate ratio effect size is applied symmetrically around
# this midpoint.
#
# @param fit              runjags fit object containing posterior samples
# @param K                integer; number of observations
# @param posterior_samples optional posterior sample matrix; avoids repeated
#                          coda extraction when supplied by callers.
#
# @return A matrix of log-rate (phi) samples for each observation
#
# ---------------------------------------------------------------------------- #
.evaluate.brma.lograte <- function(fit, K, posterior_samples = NULL) {

  posterior_samples <- .get_posterior_samples(fit, posterior_samples)

  return(.extract_posterior_matrix(
    posterior_samples = posterior_samples,
    parameter         = "phi",
    K                 = K
  ))
}


# ---------------------------------------------------------------------------- #
# .evaluate.brma.lograte_newdata
# ---------------------------------------------------------------------------- #
#
# Sample new-study log-rate values for Poisson GLMM response prediction.
#
# @param prior_phi BayesTools prior object on log(phi).
# @param S         integer; number of posterior rows.
# @param K         integer; number of new observations.
#
# @return S x K matrix of log-rate samples.
#
# ---------------------------------------------------------------------------- #
.evaluate.brma.lograte_newdata <- function(prior_phi, S, K) {

  return(.draw_prior_samples_matrix(prior = prior_phi, S = S, K = K))
}


# ---------------------------------------------------------------------------- #
# .draw_prior_samples_matrix
# ---------------------------------------------------------------------------- #
#
# Draw independent prior samples for GLMM newdata nuisance parameters.
#
# @param prior BayesTools prior object.
# @param S     integer; number of posterior rows.
# @param K     integer; number of new observations.
#
# @return S x K matrix of samples.
#
# ---------------------------------------------------------------------------- #
.draw_prior_samples_matrix <- function(prior, S, K) {

  values <- .draw_prior_samples(prior = prior, n = S * K)

  return(matrix(values, nrow = S, ncol = K))
}


# ---------------------------------------------------------------------------- #
# .draw_prior_samples
# ---------------------------------------------------------------------------- #
#
# Thin wrapper around BayesTools prior RNG.
#
# @param prior BayesTools prior object.
# @param n     integer; number of draws.
#
# @return numeric vector.
#
# ---------------------------------------------------------------------------- #
.draw_prior_samples <- function(prior, n) {

  return(as.numeric(BayesTools::rng(prior, n = n)))
}


# ---------------------------------------------------------------------------- #
# .evaluate.brma.theta.glmm
# ---------------------------------------------------------------------------- #
#
# Extract or sample estimate-level random effects (theta) for GLMM models.
#
# For GLMMs, theta[i] represents the standardized estimate-level random effect
# (i.e., theta ~ N(0, 1)). The actual random effect is theta * tau_within.
#
# @param fit              runjags fit object (needed to extract theta if same_data)
# @param tau_within       S x K matrix of estimate-level heterogeneity samples
# @param same_data        logical; TRUE if predicting on original data
# @param K                integer; number of observations
# @param posterior_samples optional posterior sample matrix; avoids repeated
#                          coda extraction when supplied by callers.
#
# @return S x K matrix: mu contribution from estimate-level random effects
#         (theta[k] * tau_within[,k])
#
# ---------------------------------------------------------------------------- #
.evaluate.brma.theta.glmm <- function(fit, tau_within, same_data, K,
                                      posterior_samples = NULL) {

  S <- nrow(tau_within)

  if (same_data) {

    posterior_samples  <- .get_posterior_samples(fit, posterior_samples)
    theta_contribution <- .extract_posterior_matrix(
      posterior_samples = posterior_samples,
      parameter         = "theta",
      K                 = K
    ) * tau_within

  } else {

    # new data: sample fresh theta ~ N(0, 1) for each observation
    theta_new          <- matrix(stats::rnorm(S * K), nrow = S, ncol = K)
    theta_contribution <- theta_new * tau_within

  }

  return(theta_contribution)
}


.logit <- function(p) {
  return(stats::qlogis(p))
}
.inv_logit <- function(x) {
  return(stats::plogis(x))
}


# ---------------------------------------------------------------------------- #
# .extract_use_normal
# ---------------------------------------------------------------------------- #
#
# Extract bias indicator and compute use_normal logical vector.
#
# For RoBMA objects with mixture bias priors, `bias_indicator` tracks which
# bias model generated each posterior sample. For brma objects with a single
# bias prior, no indicator column exists (all samples use the same model).
#
# This function is used for performance optimization in PDF/CDF/RNG functions.
# When `use_normal[i] = TRUE`, the sample uses a non-weightfunction bias model
# (PET, PEESE, or no bias), so the fast normal path can be used instead of
# the selected-normal computation.
#
# @param object brma or RoBMA object.
# @param posterior_samples optional posterior sample matrix; avoids repeated
#                          coda extraction when supplied by callers.
#
# @return A logical vector of length S (number of posterior samples):
#   - TRUE if sample uses non-weightfunction bias model (PET, PEESE, or no bias)
#   - FALSE if sample uses weightfunction bias model
#
# ---------------------------------------------------------------------------- #
.extract_use_normal <- function(object, posterior_samples = NULL) {

  priors <- object[["priors"]]
  fit    <- object[["fit"]]

  posterior_samples <- .get_posterior_samples(fit, posterior_samples)
  S <- nrow(posterior_samples)

  # check if bias_indicator column exists (RoBMA with mixture priors)
  has_bias_indicator <- "bias_indicator" %in% colnames(posterior_samples)

  if (has_bias_indicator) {

    # RoBMA case: multiple bias priors in mixture
    bias_indicator <- as.integer(posterior_samples[, "bias_indicator"])

    # extract bias priors and ensure list format
    priors_bias <- priors[["outcome"]][["bias"]]
    if (!BayesTools::is.prior.mixture(priors_bias)) {
      priors_bias <- list(priors_bias)
    }

    # identify which bias priors use selection kernels
    weightfunction_indices <- which(sapply(priors_bias, .prior_is_selection_kernel))

    # use_normal = TRUE for samples NOT from weightfunction priors
    use_normal <- !(bias_indicator %in% weightfunction_indices)

  } else {

    # brma case: single bias prior (or no bias)
    # check if the single prior is a weightfunction
    is_weightfunction <- .is_priors_weightfunction(priors)

    # if weightfunction, all samples use weighted path; otherwise all use normal
    use_normal <- rep(!is_weightfunction, S)

  }

  return(use_normal)
}


# ---------------------------------------------------------------------------- #
# .extract_bias_indicator
# ---------------------------------------------------------------------------- #
#
# Extract posterior bias branch indicators. Single-prior models do not monitor
# `bias_indicator`, but their branch is always 1.
#
# @param object brma object.
# @param posterior_samples optional posterior sample matrix; avoids repeated
#                          coda extraction when supplied by callers.
#
# @return integer vector of length S.
#
# ---------------------------------------------------------------------------- #
.extract_bias_indicator <- function(object, posterior_samples = NULL) {

  posterior_samples <- .get_posterior_samples(object[["fit"]], posterior_samples)
  S <- nrow(posterior_samples)

  if ("bias_indicator" %in% colnames(posterior_samples)) {
    return(as.integer(posterior_samples[, "bias_indicator"]))
  }

  return(rep(1L, S))
}

context("Evaluate posterior samples")
# ============================================================================ #
# test-02-evaluate.R
# ============================================================================ #
#
# Unit and integration tests for evaluate.R helper functions
# defined in R/evaluate.R
#
# Structure:
# 1. Unit tests: Mock posterior matrices (no JAGS fit needed)
# 2. Integration tests: Pre-fitted models from list_fits()
#
# ============================================================================ #

# Load common test helpers
source(testthat::test_path("common-functions.R"))
source(testthat::test_path("helper-contracts.R"))
source(testthat::test_path("helper-test-matrix.R"))
REFERENCE_DIR <<- testthat::test_path("..", "results", "evaluate")

# ============================================================================ #
# SECTION 1: Unit Tests with Mock Data
# ============================================================================ #
# These tests verify the computational logic of helper functions
# using mock posterior sample matrices (no JAGS fit required)
# ============================================================================ #

test_that(".evaluate.brma.true_effects.norm computes correct BLUPs", {

  # create mock data
  S <- 100  # samples
  K <- 4    # observations

  set.seed(1234)
  mu_samples <- matrix(rnorm(S * K, mean = 0.5, sd = 0.1), nrow = S, ncol = K)
  tau_within <- matrix(abs(rnorm(S * K, mean = 0.2, sd = 0.01)), nrow = S, ncol = K)
  yi  <- c(0.3, 0.5, 0.4, 0.6)
  sei <- c(0.1, 0.15, 0.25, 0.30)

  # compute true effects with same_data = TRUE (BLUPs)
  theta <- .evaluate.brma.true_effects.norm(
    mu_samples = mu_samples,
    tau_within = tau_within,
    yi         = yi,
    sei        = sei,
    same_data  = TRUE
  )

  # verify dimensions
  expect_equal(dim(theta), c(S, K))

  # verify shrinkage behavior: theta should be between yi and mu
  # (on average across samples)
  mean_theta <- colMeans(theta)
  mean_mu    <- colMeans(mu_samples)

  for (k in seq_len(K)) {
    # theta should be closer to data when tau is large relative to se
    # and closer to model when tau is small relative to se
    if (mean(tau_within[, k]) > sei[k]) {
      # high tau: theta closer to yi
      expect_true(abs(mean_theta[k] - yi[k]) < abs(mean_theta[k] - mean_mu[k]),
                  info = paste("Observation", k, "shrinks toward data with high tau"))
    } else {
      expect_true(abs(mean_theta[k] - yi[k]) > abs(mean_theta[k] - mean_mu[k]),
                  info = paste("Observation", k, "shrinks toward mean with high sei"))
    }
  }

  # verify lambda formula against a closed-form calculation
  # lambda = tau^2 / (tau^2 + se^2)
  # theta = lambda * yi + (1 - lambda) * mu
  sei_mat <- matrix(sei, nrow = S, ncol = K, byrow = TRUE)
  lambda  <- tau_within^2 / (tau_within^2 + sei_mat^2)
  yi_mat  <- matrix(yi, nrow = S, ncol = K, byrow = TRUE)
  expected_theta <- lambda * yi_mat + (1 - lambda) * mu_samples

  expect_equal(theta, expected_theta, tolerance = 1e-10)
})

test_that(".evaluate.brma.true_effects.norm returns BLUP means for same data", {

  S <- 25
  K <- 2

  mu_samples <- matrix(c(0.2, -0.1), nrow = S, ncol = K, byrow = TRUE)
  tau_within <- matrix(c(0.5, 0.25), nrow = S, ncol = K, byrow = TRUE)
  yi         <- c(1.0, -0.7)
  sei        <- c(0.2, 0.4)

  theta <- .evaluate.brma.true_effects.norm(
    mu_samples = mu_samples,
    tau_within = tau_within,
    yi         = yi,
    sei        = sei,
    same_data  = TRUE
  )

  lambda   <- tau_within[1, ]^2 / (tau_within[1, ]^2 + sei^2)
  expected <- mu_samples[1, ] + lambda * (yi - mu_samples[1, ])

  expect_equal(theta[1, ], expected, tolerance = 1e-12)
  expect_equal(apply(theta, 2, stats::var), c(0, 0), tolerance = 1e-14)
})

test_that(".evaluate.brma.true_effects.norm subtracts posterior-row bias offsets", {

  mu_samples <- matrix(
    c(
      0.0,  0.0,
      0.1, -0.1
    ),
    nrow = 2,
    byrow = TRUE
  )
  tau_within <- matrix(1, nrow = 2, ncol = 2)
  yi          <- c(0.4, -0.2)
  sei         <- c(1, 2)
  bias_offset <- matrix(
    c(
       0.2, -0.1,
      -0.3,  0.4
    ),
    nrow = 2,
    byrow = TRUE
  )

  theta <- .evaluate.brma.true_effects.norm(
    mu_samples  = mu_samples,
    tau_within  = tau_within,
    yi          = yi,
    sei         = sei,
    same_data   = TRUE,
    bias_offset = bias_offset
  )

  yi_samples     <- matrix(yi, nrow = 2, ncol = 2, byrow = TRUE)
  lambda         <- tau_within^2 / sweep(tau_within^2, 2, sei^2, "+")
  expected_theta <- mu_samples + lambda * (yi_samples - bias_offset - mu_samples)

  expect_equal(theta, expected_theta, tolerance = 1e-12)
})

test_that(".evaluate.brma.true_effects.norm samples from marginal with same_data = FALSE", {

  # test that same_data = FALSE samples from N(mu, tau) marginal distribution

  S <- 10000  # many samples for stable variance estimation
  K <- 3

  set.seed(9999)
  mu_samples <- matrix(0.4, nrow = S, ncol = K)  # constant mu
  tau_within <- matrix(0.25, nrow = S, ncol = K)  # constant tau

  # compute true effects with same_data = FALSE (marginal sampling)
  theta <- .evaluate.brma.true_effects.norm(
    mu_samples = mu_samples,
    tau_within = tau_within,
    yi         = NULL,  # not used when same_data = FALSE
    sei        = NULL,  # not used when same_data = FALSE
    same_data  = FALSE
  )

  # verify dimensions
  expect_equal(dim(theta), c(S, K))

  # verify mean is approximately mu
  for (k in seq_len(K)) {
    expect_equal(mean(theta[, k]), 0.4, tolerance = 0.02,
                 info = paste("Mean for observation", k))
  }

  # verify variance is approximately tau^2
  for (k in seq_len(K)) {
    expect_equal(var(theta[, k]), 0.25^2, tolerance = 0.005,
                 info = paste("Variance for observation", k))
  }
})

test_that(".outcome_rng.norm has correct sampling variance", {

  # create mock data
  S <- 10000  # many samples for stable variance estimation
  K <- 3

  set.seed(5678)
  mu_samples <- matrix(0.5, nrow = S, ncol = K)  # constant mu for variance test
  tau_within <- matrix(0.2, nrow = S, ncol = K)  # constant tau
  sei <- c(0.1, 0.2, 0.3)

  # compute response samples
  response <- .outcome_rng.norm(mu_samples, tau_within, sei)

  # verify dimensions
  expect_equal(dim(response), c(S, K))

  # verify sampling variance matches theoretical: Var = tau^2 + se^2
  for (k in seq_len(K)) {
    expected_var <- 0.2^2 + sei[k]^2
    observed_var <- var(response[, k])
    # use tolerance appropriate for 10k samples
    expect_equal(observed_var, expected_var, tolerance = 0.01,
                 info = paste("Variance for observation", k))
  }

  # verify means are close to mu
  for (k in seq_len(K)) {
    expect_equal(mean(response[, k]), 0.5, tolerance = 0.05,
                 info = paste("Mean for observation", k))
  }
})

test_that(".evaluate.brma.cluster_effects returns contribution matrix for new data", {

  S <- 10000  # many samples
  K <- 3

  set.seed(222)
  tau_between <- matrix(0.3, nrow = S, ncol = K)  # constant tau_between

  # for new data, gamma ~ N(0,1) is sampled per cluster, so contribution = gamma * tau_between
  # expected variance = tau_between^2 = 0.09
  contribution <- .evaluate.brma.cluster_effects(
    fit              = NULL,
    tau_between      = tau_between,
    cluster          = c(1, 1, 2),
    same_data        = FALSE,
    effect_direction = "positive"
  )

  expect_equal(dim(contribution), c(S, K))

  # variance should be approximately tau_between^2
  for (k in seq_len(K)) {
    expect_equal(var(contribution[, k]), 0.3^2, tolerance = 0.01,
                 info = paste("Variance for observation", k))
  }

  expect_equal(contribution[, 1], contribution[, 2])
})

test_that(".evaluate.brma.multilevel_blup.norm matches full covariance solve", {

  S       <- 8
  K       <- 5
  cluster <- c(1, 1, 2, 2, 2)

  set.seed(333)
  mu_samples  <- matrix(rnorm(S * K, mean = 0.2, sd = 0.1), nrow = S, ncol = K)
  tau_within  <- matrix(runif(S * K, min = 0.05, max = 0.25), nrow = S, ncol = K)
  tau_between <- matrix(runif(S * K, min = 0.05, max = 0.20), nrow = S, ncol = K)
  yi          <- c(0.10, 0.25, -0.10, 0.05, 0.30)
  vi          <- c(0.02, 0.03, 0.01, 0.04, 0.02)

  result <- .evaluate.brma.multilevel_blup.norm(
    mu_samples  = mu_samples,
    tau_within  = tau_within,
    tau_between = tau_between,
    yi          = yi,
    vi          = vi,
    cluster     = cluster
  )

  block_indices     <- .get_multilevel_block_indices(cluster)
  expected_cluster  <- matrix(0, nrow = S, ncol = K)
  expected_estimate <- matrix(0, nrow = S, ncol = K)

  for (s in seq_len(S)) {
    covariance <- .build_multilevel_marginal_covariance(
      tau_within    = tau_within[s, ],
      tau_between   = tau_between[s, ],
      vi            = vi,
      block_indices = block_indices
    )
    weighted_residual <- as.vector(chol2inv(chol(covariance)) %*%
      (yi - mu_samples[s, ]))

    expected_estimate[s, ] <- tau_within[s, ]^2 * weighted_residual

    for (idx in block_indices) {
      cluster_scale <- sum(tau_between[s, idx] * weighted_residual[idx])
      expected_cluster[s, idx] <- tau_between[s, idx] * cluster_scale
    }
  }

  expect_equal(result[["cluster"]], expected_cluster, tolerance = 1e-12)
  expect_equal(result[["estimate"]], expected_estimate, tolerance = 1e-12)
})

test_that(".evaluate.brma.multilevel_blup.norm subtracts posterior-row bias offsets", {

  S       <- 3
  K       <- 4
  cluster <- c(1, 1, 2, 2)

  mu_samples <- matrix(
    c(
      0.00, 0.10, 0.20, 0.30,
      0.05, 0.15, 0.25, 0.35,
      0.10, 0.20, 0.30, 0.40
    ),
    nrow = S,
    byrow = TRUE
  )
  tau_within  <- matrix(0.20, nrow = S, ncol = K)
  tau_between <- matrix(0.15, nrow = S, ncol = K)
  yi          <- c(0.30, 0.10, 0.50, 0.20)
  vi          <- c(0.02, 0.03, 0.02, 0.04)
  bias_offset <- matrix(
    c(
       0.05,  0.00, 0.10, 0.00,
      -0.05,  0.03, 0.00, 0.08,
       0.02, -0.04, 0.05, 0.01
    ),
    nrow = S,
    byrow = TRUE
  )

  result <- .evaluate.brma.multilevel_blup.norm(
    mu_samples  = mu_samples,
    tau_within  = tau_within,
    tau_between = tau_between,
    yi          = yi,
    vi          = vi,
    cluster     = cluster,
    bias_offset = bias_offset
  )

  block_indices     <- .get_multilevel_block_indices(cluster)
  expected_cluster  <- matrix(0, nrow = S, ncol = K)
  expected_estimate <- matrix(0, nrow = S, ncol = K)

  for (s in seq_len(S)) {
    covariance <- .build_multilevel_marginal_covariance(
      tau_within    = tau_within[s, ],
      tau_between   = tau_between[s, ],
      vi            = vi,
      block_indices = block_indices
    )
    residual <- yi - bias_offset[s, ] - mu_samples[s, ]
    weighted_residual <- as.vector(chol2inv(chol(covariance)) %*% residual)

    expected_estimate[s, ] <- tau_within[s, ]^2 * weighted_residual

    for (idx in block_indices) {
      cluster_scale <- sum(tau_between[s, idx] * weighted_residual[idx])
      expected_cluster[s, idx] <- tau_between[s, idx] * cluster_scale
    }
  }

  expect_equal(result[["cluster"]], expected_cluster, tolerance = 1e-12)
  expect_equal(result[["estimate"]], expected_estimate, tolerance = 1e-12)
})

test_that("variance ordering: mu < theta < response", {

  # this tests the theoretical relationship:
  # mu (fixed only) has least variance
  # theta (includes tau) has more variance
  # response (includes tau + se) has most variance

  S <- 5000
  K <- 4

  set.seed(444)
  # create realistic mock data
  mu_samples <- matrix(rnorm(S * K, mean = 0.33, sd = 0.05), nrow = S, ncol = K)
  tau_within <- matrix(abs(rnorm(S * K, mean = 0.25, sd = 0.05)), nrow = S, ncol = K)
  yi  <- c(0.30, 0.2, 0.5, 0.45)
  sei <- c(0.15, 0.1, 0.2, 0.12)

  # compute true effects and response
  theta <- .evaluate.brma.true_effects.norm(
    mu_samples = mu_samples,
    tau_within = tau_within,
    yi         = yi,
    sei        = sei,
    same_data  = TRUE
  )

  response <- .outcome_rng.norm(mu_samples, tau_within, sei)

  # compute variances across the distributions
  sd_mu       <- median(apply(mu_samples, 1, sd))
  sd_theta    <- median(apply(theta, 1, sd))
  sd_response <- median(apply(response, 1, sd))

  # theta variance should be less than or equal to mu variance
  # (shrinkage reduces variance toward the mean)
  # response variance should be larger than theta (adds sampling error)
  expect_true(sd_response > sd_theta,
              info = "response has more variance than theta")
  expect_true(sd_theta > sd_mu,
              info = "theta has more variance than terms")
})


# ============================================================================ #
# SECTION 2: Integration Tests with Pre-fitted Models
# ============================================================================ #
# These tests verify helper functions work correctly with actual brma fits
# ============================================================================ #

skip_if_no_fits()
fit_names     <- list_fits(tier = test_tier())
all_fit_names <- list_fits()
fits          <- lazy_fits(fit_names, validate = FALSE)
all_fits      <- lazy_fits(all_fit_names, validate = FALSE)

test_that(".evaluate.brma.tau returns correct structure", {

  for (name in names(fits)) {

    object <- fits[[name]]

    if (!inherits(object, "brma")) next

    priors       <- object[["priors"]]
    is_scale     <- .is_scale(object)
    is_multilevel <- .is_multilevel(object)
    K            <- nrow(object[["data"]][["outcome"]])

    # Direct tau extraction for scale formulas is exercised through predict().
    if (is_scale) next

    result <- .evaluate.brma.tau(
      fit           = object[["fit"]],
      scale_data    = object[["data"]][["scale"]],
      scale_formula = if (is_scale) attr(object[["data"]][["scale"]], "formula") else NULL,
      scale_priors  = priors[["scale"]],
      is_scale      = is_scale,
      is_multilevel = is_multilevel,
      K             = K
    )

    # verify structure
    expect_true(is.list(result), info = paste(name, ": returns list"))
    expect_true(all(c("tau_within", "tau_between") %in% names(result)),
                info = paste(name, ": has tau_within and tau_between components"))

    # verify dimensions
    S <- nrow(result[["tau_within"]])
    expect_equal(dim(result[["tau_within"]]), c(S, K), info = paste(name, ": tau_within dimensions"))
    expect_equal(dim(result[["tau_between"]]), c(S, K), info = paste(name, ": tau_between dimensions"))

    # verify positivity
    expect_true(all(result[["tau_within"]] >= 0), info = paste(name, ": tau_within is non-negative"))
    expect_true(all(result[["tau_between"]] >= 0), info = paste(name, ": tau_between is non-negative"))

    # verify relationship for non-multilevel: tau_between = 0
    if (!is_multilevel) {
      expect_true(all(result[["tau_between"]] == 0),
                  info = paste(name, ": non-multilevel has tau_between = 0"))
    }

    # verify total tau can be reconstructed: tau = sqrt(tau_within^2 + tau_between^2)
    tau_reconstructed <- sqrt(result[["tau_within"]]^2 + result[["tau_between"]]^2)
    expect_true(all(tau_reconstructed >= 0), info = paste(name, ": reconstructed tau is non-negative"))
  }
})

test_that(".evaluate.brma.mu returns correct dimensions", {

  for (name in names(fits)) {

    object <- fits[[name]]

    if (!inherits(object, "brma")) next

    priors           <- object[["priors"]]
    is_mods          <- .is_mods(object)
    is_PET           <- .is_PET(object)
    is_PEESE         <- .is_PEESE(object)
    effect_direction <- .effect_direction(object)
    outcome_data     <- object[["data"]][["outcome"]]
    K                <- nrow(outcome_data)

    # Direct moderator design handling is exercised through predict().
    if (is_mods) next

    mu_samples <- .evaluate.brma.mu(
      fit               = object[["fit"]],
      outcome_data      = outcome_data,
      mods_data         = object[["data"]][["mods"]],
      mods_formula      = if (is_mods) attr(object[["data"]][["mods"]], "formula") else NULL,
      mods_priors       = priors[["mods"]],
      is_mods           = is_mods,
      is_PET            = is_PET,
      is_PEESE          = is_PEESE,
      effect_direction  = effect_direction,
      bias_adjusted     = TRUE,
      K                 = K
    )

    # verify dimensions
    posterior_samples <- suppressWarnings(coda::as.mcmc(object[["fit"]]))
    S <- nrow(posterior_samples)
    expect_equal(dim(mu_samples), c(S, K),
                 info = paste(name, ": mu dimensions are S x K"))
  }
})

test_that("PET/PEESE bias offsets match column-wise algebra", {

  S   <- 100
  K   <- 5
  sei <- c(0.1, 0.2, 0.15, 0.25, 0.12)

  set.seed(666)
  mu_samples  <- matrix(rnorm(S * K), nrow = S, ncol = K)
  PET_samples <- rnorm(S, mean = 0.5, sd = 0.1)

  mu_loop <- mu_samples
  for (i in seq_len(K)) {
    mu_loop[, i] <- mu_loop[, i] + PET_samples * sei[i]
  }

  mu_vec <- mu_samples + outer(PET_samples, sei)

  expect_equal(mu_loop, mu_vec, tolerance = 1e-14,
               info = "Vectorized outer() matches loop for PET")

  # same test for PEESE (se^2)
  set.seed(777)
  mu_samples    <- matrix(rnorm(S * K), nrow = S, ncol = K)
  PEESE_samples <- rnorm(S, mean = 0.3, sd = 0.05)
  sei_sq        <- sei^2

  mu_loop <- mu_samples
  for (i in seq_len(K)) {
    mu_loop[, i] <- mu_loop[, i] + PEESE_samples * sei_sq[i]
  }

  mu_vec <- mu_samples + outer(PEESE_samples, sei_sq)

  expect_equal(mu_loop, mu_vec, tolerance = 1e-14,
               info = "Vectorized outer() matches loop for PEESE")
})

test_that(".evaluate.brma.bias_offset handles PET/PEESE and effect direction", {

  posterior_samples <- matrix(
    c(
      0.50, 1.00,
      0.25, 2.00
    ),
    nrow = 2,
    byrow = TRUE
  )
  colnames(posterior_samples) <- c("PET", "PEESE")

  outcome_data <- data.frame(sei = c(0.10, 0.20, 0.30))
  expected_positive <- outer(posterior_samples[, "PET"], outcome_data[["sei"]]) +
    outer(posterior_samples[, "PEESE"], outcome_data[["sei"]]^2)

  offset_positive <- .evaluate.brma.bias_offset(
    fit               = NULL,
    outcome_data      = outcome_data,
    is_PET            = TRUE,
    is_PEESE          = TRUE,
    effect_direction  = "positive",
    K                 = 3,
    posterior_samples = posterior_samples
  )
  offset_negative <- .evaluate.brma.bias_offset(
    fit               = NULL,
    outcome_data      = outcome_data,
    is_PET            = TRUE,
    is_PEESE          = TRUE,
    effect_direction  = "negative",
    K                 = 3,
    posterior_samples = posterior_samples
  )

  expect_equal(offset_positive, expected_positive, tolerance = 1e-12)
  expect_equal(offset_negative, -expected_positive, tolerance = 1e-12)
})

test_that("rho clamping handles boundary values", {

  # test that rho values outside [0, 1] are properly clamped

  # create mock rho with edge cases
  rho <- c(-0.01, 0, 0.5, 1, 1.001)
  rho_clamped <- pmin(pmax(rho, 0), 1)

  expect_equal(rho_clamped, c(0, 0, 0.5, 1, 1))

  # verify tau decomposition is valid after clamping
  tau <- 0.3
  tau_within  <- tau * sqrt(1 - rho_clamped)
  tau_between <- tau * sqrt(rho_clamped)

  # all values should be non-negative
  expect_true(all(tau_within >= 0))
  expect_true(all(tau_between >= 0))

  # verify Pythagorean relationship: tau^2 = tau_within^2 + tau_between^2
  tau_reconstructed <- sqrt(tau_within^2 + tau_between^2)
  expect_equal(tau_reconstructed, rep(tau, length(rho)), tolerance = 1e-10)
})

test_that("GLMM posterior extraction helpers are vectorized", {

  posterior_samples <- matrix(
    c(
      0.25, 0.75, -1.0, 1.0, 0.10, -0.20,
      0.40, 0.60, -0.5, 0.5, 0.30,  0.40
    ),
    nrow = 2,
    byrow = TRUE
  )
  colnames(posterior_samples) <- c(
    "pi[1]", "pi[2]", "phi[1]", "phi[2]", "theta[1]", "theta[2]"
  )

  tau_within <- matrix(c(0.2, 0.3, 0.4, 0.5), nrow = 2, byrow = TRUE)

  expect_equal(
    .evaluate.brma.baserate(fit = NULL, K = 2, posterior_samples = posterior_samples),
    stats::qlogis(posterior_samples[, c("pi[1]", "pi[2]")])
  )
  expect_equal(
    .evaluate.brma.lograte(fit = NULL, K = 2, posterior_samples = posterior_samples),
    posterior_samples[, c("phi[1]", "phi[2]")]
  )
  expect_equal(
    .evaluate.brma.theta.glmm(
      fit               = NULL,
      tau_within        = tau_within,
      same_data         = TRUE,
      K                 = 2,
      posterior_samples = posterior_samples
    ),
    posterior_samples[, c("theta[1]", "theta[2]")] * tau_within
  )
})

test_that("matrix replication patterns preserve dimensions", {

  # verify that matrix(vec, S, K, byrow = TRUE) works as expected
  # this pattern is used throughout the helpers

  S   <- 3
  K   <- 4
  vec <- c(0.1, 0.2, 0.3, 0.4)

  # byrow = TRUE: each row is the vector
  mat <- matrix(vec, nrow = S, ncol = K, byrow = TRUE)

  expect_equal(nrow(mat), S)
  expect_equal(ncol(mat), K)

  # each row should equal vec
  for (s in seq_len(S)) {
    expect_equal(mat[s, ], vec)
  }

  # byrow = FALSE: each column is the vector (recycled)
  vec2 <- c(1, 2, 3)
  mat2 <- matrix(vec2, nrow = S, ncol = K, byrow = FALSE)

  # each column should equal vec2
  for (k in seq_len(K)) {
    expect_equal(mat2[, k], vec2)
  }
})

# ============================================================================ #
# SECTION 3: Unit Tests for Aggregated Predictions (newdata = TRUE)
# ============================================================================ #
# These tests verify the aggregation logic for predict.brma with newdata = TRUE
# ============================================================================ #

test_that("aggregation matches rowMeans", {

  # test that rowMeans aggregation works as expected
  S <- 100
  K <- 5

  set.seed(888)
  mu_samples <- matrix(rnorm(S * K, mean = 0.3, sd = 0.1), nrow = S, ncol = K)

  # aggregation: rowMeans across K observations

  mu_aggregated <- matrix(rowMeans(mu_samples), ncol = 1)

  # verify dimensions
  expect_equal(dim(mu_aggregated), c(S, 1))

  # verify mean is preserved (on average)
  expect_equal(mean(mu_aggregated), mean(mu_samples), tolerance = 0.01)

  # verify aggregation against a direct matrix calculation
  expected_aggregated <- matrix(apply(mu_samples, 1, mean), ncol = 1)
  expect_equal(mu_aggregated, expected_aggregated, tolerance = 1e-14)
})

test_that("aggregation is no-op for identical columns (non-mods/scale models)", {

  # for models without moderators/scale, all columns are identical
  # aggregation should return the same value
  S <- 100
  K <- 5

  # create mock data where all columns are identical
  mu_values <- rnorm(S, mean = 0.5, sd = 0.1)
  mu_samples <- matrix(mu_values, nrow = S, ncol = K)

  # aggregation
  mu_aggregated <- matrix(rowMeans(mu_samples), ncol = 1)

  # should be identical to original column
  expect_equal(mu_aggregated[, 1], mu_values, tolerance = 1e-14)
})

test_that("aggregated true effects have expected variance", {

  # for aggregated effect predictions:
  # theta ~ N(mu, tau) where mu and tau are aggregated
  # variance of theta should be approximately var(mu) + mean(tau)^2

  S <- 10000  # many samples for stable variance estimation

  set.seed(999)
  mu_aggregated  <- matrix(rnorm(S, mean = 0.4, sd = 0.05), ncol = 1)
  tau_aggregated <- matrix(abs(rnorm(S, mean = 0.25, sd = 0.02)), ncol = 1)

  # sample true effects: theta = mu + rnorm(S) * tau
  theta_samples <- mu_aggregated + rnorm(S) * tau_aggregated

  # expected variance: Var(mu) + E[tau^2] = Var(mu) + Var(tau) + E[tau]^2
  expected_var <- var(mu_aggregated) + var(tau_aggregated) + mean(tau_aggregated)^2

  # observed variance should be close
  observed_var <- var(theta_samples)
  expect_equal(observed_var, expected_var, tolerance = 0.02,
               info = "Aggregated theta variance matches theory")
})


# ============================================================================ #
# SECTION 4: Integration Tests for predict.brma with newdata = TRUE
# ============================================================================ #
# These tests verify predict.brma with aggregated predictions using cached fits
# ============================================================================ #

skip_if_no_fits()

test_that("predict.brma with newdata = TRUE returns single aggregated prediction", {

  for (name in names(fits)) {

    object <- fits[[name]]

    # skip non-brma objects
    if (!inherits(object, "brma")) next

    # test type = "terms"
    result_terms <- predict(object, newdata = TRUE, type = "terms")

    expect_s3_class(result_terms, "brma_samples")
    expect_null(attr(result_terms, "data"), info = paste(name, ": data is NULL for aggregate"))
    expect_equal(nrow(summary(result_terms)), 1,
                 info = paste(name, ": has single row for aggregate terms"))

    # test type = "terms.scale"
    result_scale <- predict(object, newdata = TRUE, type = "terms.scale")

    expect_s3_class(result_scale, "brma_samples")
    expect_null(attr(result_scale, "data"), info = paste(name, ": data is NULL for aggregate scale"))
    expect_equal(nrow(summary(result_scale)), 1,
                 info = paste(name, ": has single row for aggregate scale"))

    # test type = "effect"
    result_effect <- predict(object, newdata = TRUE, type = "effect")

    expect_s3_class(result_effect, "brma_samples")
    expect_null(attr(result_effect, "data"), info = paste(name, ": data is NULL for aggregate effect"))
    expect_equal(nrow(summary(result_effect)), 1,
                 info = paste(name, ": has single row for aggregate effect"))
  }
})

test_that("predict.brma with newdata = TRUE returns S x 1 matrix", {

  for (name in names(fits)) {

    object <- fits[[name]]

    # skip non-brma objects
    if (!inherits(object, "brma")) next

    # get expected sample count
    posterior_samples <- suppressWarnings(coda::as.mcmc(object[["fit"]]))
    S <- nrow(posterior_samples)

    # test type = "terms"
    samples_terms <- predict(object, newdata = TRUE, type = "terms")
    expect_equal(dim(samples_terms), c(S, 1),
                 info = paste(name, ": terms samples are S x 1"))
    expect_equal(colnames(samples_terms), "mu",
                 info = paste(name, ": terms samples have 'mu' column name"))

    # test type = "terms.scale"
    samples_scale <- predict(object, newdata = TRUE, type = "terms.scale")
    expect_equal(dim(samples_scale), c(S, 1),
                 info = paste(name, ": scale samples are S x 1"))
    expect_equal(colnames(samples_scale), "tau",
                 info = paste(name, ": scale samples have 'tau' column name"))

    # test type = "effect"
    samples_effect <- predict(object, newdata = TRUE, type = "effect")
    expect_equal(dim(samples_effect), c(S, 1),
                 info = paste(name, ": effect samples are S x 1"))
    expect_equal(colnames(samples_effect), "theta",
                 info = paste(name, ": effect samples have 'theta' column name"))
  }
})

test_that("predict.brma rejects response-scale newdata predictions", {

  for (name in names(fits)) {

    object <- fits[[name]]

    # skip non-brma objects
    if (!inherits(object, "brma")) next

    # should throw error for type = "response"
    expect_error(
      predict(object, newdata = TRUE, type = "response"),
      "Aggregated predictions.*not available for type = 'response'",
      info = paste(name, ": aggregate + response is rejected")
    )
  }
})

test_that("aggregated mu equals rowMeans of non-aggregated mu", {

  for (name in names(fits)) {

    object <- fits[[name]]

    # skip non-brma objects
    if (!inherits(object, "brma")) next

    # get non-aggregated samples (as plain matrix)
    samples_full <- as.matrix(predict(object, newdata = NULL, type = "terms"))

    # get aggregated samples (as plain matrix)
    samples_agg <- as.matrix(predict(object, newdata = TRUE, type = "terms"))

    # aggregated should equal rowMeans of full
    expected_agg <- matrix(rowMeans(samples_full), ncol = 1)
    colnames(expected_agg) <- "mu"

    expect_equal(samples_agg, expected_agg, tolerance = 1e-10,
                 info = paste(name, ": aggregated mu equals rowMeans of full mu"))
  }
})

test_that("aggregated tau equals rowMeans of non-aggregated tau", {

  for (name in names(fits)) {

    object <- fits[[name]]

    # skip non-brma objects
    if (!inherits(object, "brma")) next

    # get non-aggregated samples (as plain matrix)
    samples_full <- as.matrix(predict(object, newdata = NULL, type = "terms.scale"))

    # get aggregated samples (as plain matrix)
    samples_agg <- as.matrix(predict(object, newdata = TRUE, type = "terms.scale"))

    # aggregated should equal rowMeans of full
    expected_agg <- matrix(rowMeans(samples_full), ncol = 1)
    colnames(expected_agg) <- "tau"

    expect_equal(samples_agg, expected_agg, tolerance = 1e-10,
                 info = paste(name, ": aggregated tau equals rowMeans of full tau"))
  }
})


# ============================================================================ #
# SECTION 3: Tests for .extract_use_normal()
# ============================================================================ #
# These tests verify the bias indicator extraction helper function
# that identifies which posterior samples use weightfunction vs normal path
# ============================================================================ #

test_that(".extract_omega_samples orders indexed omega columns", {

  posterior_samples <- matrix(
    seq_len(12),
    nrow = 3,
    ncol = 4
  )
  colnames(posterior_samples) <- c("omega[10]", "mu", "omega[2]", "omega[1]")

  omega <- .extract_omega_samples(posterior_samples)

  expect_equal(colnames(omega), c("omega[1]", "omega[2]", "omega[10]"))
  expect_equal(omega[, 1], posterior_samples[, "omega[1]"])
  expect_equal(omega[, 2], posterior_samples[, "omega[2]"])
  expect_equal(omega[, 3], posterior_samples[, "omega[10]"])
})

test_that(".extract_use_normal returns correct structure for brma without bias", {

  for (name in names(fits)) {

    object <- fits[[name]]

    # skip non-brma objects or objects with bias priors
    if (!inherits(object, "brma")) next
    if (.is_bias(object)) next

    # test the function
    use_normal <- .extract_use_normal(object)

    # get expected sample count
    posterior_samples <- suppressWarnings(coda::as.mcmc(object[["fit"]]))
    S <- nrow(posterior_samples)

    # verify structure
    expect_type(use_normal, "logical")
    expect_length(use_normal, S)

    # all samples should use normal path (no weightfunction)
    expect_true(all(use_normal),
                info = paste(name, ": brma without bias has all use_normal = TRUE"))
  }
})

test_that(".extract_use_normal returns correct structure for brma with weightfunction", {

  for (name in names(fits)) {

    object <- fits[[name]]

    # skip non-brma objects or objects without weightfunction priors
    if (!inherits(object, "brma")) next
    if (!.is_weightfunction(object)) next
    if (inherits(object, "RoBMA")) next  # RoBMA has mixture priors, test separately

    # test the function
    use_normal <- .extract_use_normal(object)

    # get expected sample count
    posterior_samples <- suppressWarnings(coda::as.mcmc(object[["fit"]]))
    S <- nrow(posterior_samples)

    # verify structure
    expect_type(use_normal, "logical")
    expect_length(use_normal, S)

    # all samples should use weighted path (all from weightfunction)
    expect_true(all(!use_normal),
                info = paste(name, ": brma with single weightfunction has all use_normal = FALSE"))
  }
})

test_that(".extract_use_normal returns correct structure for brma with PET", {

  for (name in names(fits)) {

    object <- fits[[name]]

    # skip non-brma objects or objects without PET priors
    if (!inherits(object, "brma")) next
    if (!.is_PET(object)) next
    if (.is_weightfunction(object)) next  # skip if also has weightfunction
    if (inherits(object, "RoBMA")) next

    # test the function
    use_normal <- .extract_use_normal(object)

    # get expected sample count
    posterior_samples <- suppressWarnings(coda::as.mcmc(object[["fit"]]))
    S <- nrow(posterior_samples)

    # verify structure
    expect_type(use_normal, "logical")
    expect_length(use_normal, S)

    # all samples should use normal path (PET is not a weightfunction)
    expect_true(all(use_normal),
                info = paste(name, ": brma with PET has all use_normal = TRUE"))
  }
})

test_that(".extract_use_normal returns correct structure for RoBMA", {

  robma_names <- intersect(names(fits), catalog_fits(class = "RoBMA", tier = test_tier()))
  if (length(robma_names) == 0L) {
    skip("No cached RoBMA fits available for use_normal branch checks.")
  }

  for (name in robma_names) {

    object <- fits[[name]]

    # test the function
    use_normal <- .extract_use_normal(object)

    # get expected sample count
    posterior_samples <- suppressWarnings(coda::as.mcmc(object[["fit"]]))
    S <- nrow(posterior_samples)

    # verify structure
    expect_type(use_normal, "logical")
    expect_length(use_normal, S)

    # RoBMA should have a mix of TRUE and FALSE (different bias models)
    # get bias_indicator to verify
    if ("bias_indicator" %in% colnames(posterior_samples)) {
      bias_indicator <- as.integer(posterior_samples[, "bias_indicator"])

      # extract bias priors
      priors_bias <- object[["priors"]][["outcome"]][["bias"]]
      if (!BayesTools::is.prior.mixture(priors_bias)) {
        priors_bias <- list(priors_bias)
      }

      # identify selected-normal kernel indices
      wf_indices <- which(sapply(priors_bias, .prior_is_selection_kernel))

      # verify use_normal is correctly computed
      expected_use_normal <- !(bias_indicator %in% wf_indices)
      expect_equal(use_normal, expected_use_normal,
                   info = paste(name, ": use_normal matches bias_indicator logic"))

      # verify we have both types of samples (unless all priors are one type)
      if (length(wf_indices) > 0 && length(wf_indices) < length(priors_bias)) {
        expect_true(any(use_normal) && any(!use_normal),
                    info = paste(name, ": RoBMA has mix of normal and weighted samples"))
      }
    }
  }
})

# ============================================================================ #
# SECTION 4: Integration Tests for .pdf.brma() and .cdf.brma() with use_normal
# ============================================================================ #
# These tests verify that PDF and CDF functions work correctly with the
# use_normal fast-path optimization
# ============================================================================ #

test_that(".pdf.brma returns finite log-likelihoods for weightfunction models", {

  for (name in names(fits)) {

    object <- fits[[name]]

    # skip non-brma objects or objects without weightfunction priors
    if (!inherits(object, "brma")) next
    if (!.is_weightfunction(object)) next

    # compute PDF (this internally uses use_normal optimization)
    log_lik <- .pdf.brma(object)

    # verify structure
    expect_true(is.matrix(log_lik),
                info = paste(name, ": log_lik is a matrix"))
    expect_true(all(is.finite(log_lik)),
                info = paste(name, ": all log-likelihoods are finite"))

    # verify dimensions match
    K <- length(.outcome_data_yi(object))
    expect_equal(ncol(log_lik), K,
                 info = paste(name, ": ncol matches number of observations"))
  }
})

test_that(".cdf.brma returns valid CDF values for weightfunction models", {

  for (name in names(fits)) {

    object <- fits[[name]]

    # skip non-brma objects or objects without weightfunction priors
    if (!inherits(object, "brma")) next
    if (!.is_weightfunction(object)) next

    # compute CDF (this internally uses use_normal optimization)
    cdf_vals <- .cdf.brma(object)

    # verify structure
    expect_true(is.matrix(cdf_vals),
                info = paste(name, ": cdf_vals is a matrix"))
    expect_true(all(is.finite(cdf_vals)),
                info = paste(name, ": all CDF values are finite"))
    expect_true(all(cdf_vals > 0 & cdf_vals < 1),
                info = paste(name, ": all CDF values are in (0, 1)"))

    # verify dimensions match
    K <- length(.outcome_data_yi(object))
    expect_equal(ncol(cdf_vals), K,
                 info = paste(name, ": ncol matches number of observations"))
  }
})

# ============================================================================ #
# brma.residuals.R
# ============================================================================ #
#
# This file contains the residuals method for brma model fits. Residuals are
# computed as the difference between observed effect sizes and fitted values.
#
# Design principles:
# - Consistent with predict.brma: use same underlying machinery
# - Simple interface: minimal required arguments
# - Supports both summary statistics and posterior samples
#
# ============================================================================ #


#' @title Residuals for brma Objects
#'
#' @description Computes residuals (observed minus fitted values) from a
#' fitted brma object, with options for standardization.
#'
#' @param object a fitted brma object
#' @param type the type of residuals to compute. Options are:
#' \itemize{
#'   \item \code{"outcome"} (default): Raw residuals (observed - fitted).
#'     Available for all model types.
#'   \item \code{"pearson"}: Pearson (semi-standardized) residuals, dividing
#'     raw residuals by the marginal standard error \eqn{\sqrt{v_i + \tau^2}}.
#'     Only available for normal outcome models without selection (weightfunction)
#'     bias adjustment.
#'   \item \code{"rstandard"}: Internally standardized residuals, dividing
#'     raw residuals by their standard errors computed using the hat matrix.
#'     Only available for normal outcome models without selection (weightfunction)
#'     bias adjustment.
#'   \item \code{"LOO-PIT"} (alias: \code{"rstudent"}): Leave-one-out probability
#'     integral transform (PIT) residuals computed via Pareto smoothed importance
#'     sampling. The LOO-CDF value for each observation is computed and transformed
#'     to standard normal quantiles via \eqn{\Phi^{-1}(u_i)}. Under a correctly
#'     specified model, these residuals should follow a standard normal distribution.
#'     This is the recommended standardized residual for Bayesian models as it properly
#'     accounts for estimation uncertainty and leverage. Available for all model
#'     types. Note: This requires that the loo has been computed previously (see
#'     \code{add_loo()} function).
#' }
#' @param unit output unit. Only \code{"estimate"} is implemented currently..
#' @param conditioning_depth conditioning depth for non-LOO residuals. \code{"marginal"}
#' uses fixed effects only, \code{"cluster"} conditions on cluster-level random
#' effects, and \code{"estimate"} conditions on the full estimate-level fitted
#' value. LOO-PIT residuals always use the estimate-unit LOO target.
#' @param bias_adjusted whether residuals should be computed from bias-adjusted
#' fitted values. Defaults to \code{FALSE}, which means residuals are computed
#' as the difference between observed values and raw (biased) predictions
#' including PET/PEESE terms. Set to \code{TRUE} to compute residuals from
#' bias-corrected fitted values. Applies to outcome and Pearson residuals. Note
#' that bias-adjusted residuals are not residuals in the traditional sense.
#' @param ... additional arguments.
#'
#' @details
#' Raw residuals (\code{type = "outcome"}) are computed as:
#' \deqn{e_i = y_i - \hat{\mu}_i}
#' where \eqn{y_i} is the observed effect size and \eqn{\hat{\mu}_i} is the
#' fitted value (prediction from the fixed effects).
#'
#' Pearson residuals (\code{type = "pearson"}) divide raw residuals by the
#' marginal standard error:
#' \deqn{r_i^{Pearson} = \frac{e_i}{\sqrt{v_i + \tau^2}}}
#' where \eqn{v_i} is the sampling variance and \eqn{\tau^2} is the
#' relevant heterogeneity component. Only available for normal outcome models
#' without selection (weightfunction) bias adjustment.
#'
#' Standardized residuals (\code{type = "rstandard"}) use the hat matrix to
#' compute residual standard errors that account for the uncertainty in
#' estimated coefficients:
#' \deqn{z_i = \frac{e_i}{\sqrt{[(I-H)M(I-H)']_{ii}}}}
#' where \eqn{H} is the hat matrix and \eqn{M} is the marginal variance-covariance
#' matrix. For models without moderators, this simplifies to the Pearson formula.
#' Only available for normal outcome models without selection (weightfunction)
#' bias adjustment.
#'
#' LOO-PIT residuals (\code{type = "LOO-PIT"}) are the Bayesian equivalent of
#' studentized deleted residuals \insertCite{vehtari2017practical}{RoBMA}. They
#' are computed via leave-one-out probability integral transformation:
#' \deqn{r_i = \Phi^{-1}(u_i)}
#' where \eqn{u_i = \sum_s w_{is} F(y_i | \theta^{(s)})} is the LOO-weighted CDF
#' value, \eqn{w_{is}} are the normalized PSIS weights, and \eqn{F} is the
#' cumulative distribution function of the estimate-unit predictive
#' distribution used by LOO. Under a correctly specified model, LOO-PIT
#' residuals should follow a standard normal distribution. Unlike traditional
#' standardized residuals, LOO-PIT residuals properly account for estimation
#' uncertainty and leverage without requiring a hat matrix. This is the
#' recommended method for standardized residuals in Bayesian meta-analysis.
#'
#' For meta-regression models, fitted values incorporate moderator effects.
#' For models without moderators, all fitted values equal the pooled effect.
#'
#' For GLMM models (binomial or Poisson), observed effect sizes and their
#' sampling variances are computed from the raw frequency data using the
#' same formulas as \code{metafor::escalc} with the default zero-cell
#' adjustment (adding 0.5 to all cells when any cell is zero). GLMM residuals
#' and LOO-PIT values are therefore approximate effect-size-scale diagnostics,
#' not exact PIT diagnostics for the raw count likelihood.
#'
#' The residuals are computed separately for each posterior sample,
#' naturally propagating uncertainty in model parameters to the residuals.
#'
#' @return A numeric vector of residual means, one per estimate.
#'
#' @examples \dontrun{
#' if (requireNamespace("metadat", quietly = TRUE)) {
#'   data(dat.lehmann2018, package = "metadat")
#'
#'   fit <- bPET(
#'     yi      = yi,
#'     vi      = vi,
#'     data    = dat.lehmann2018,
#'     measure = "SMD",
#'     seed    = 1,
#'     silent  = TRUE
#'   )
#'
#'   # raw residuals (default)
#'   residuals(fit)
#'
#'   # Pearson and internally standardized residuals
#'   residuals(fit, type = "pearson")
#'   residuals(fit, type = "rstandard")
#'
#'   # LOO-PIT residuals require stored LOO
#'   fit <- add_loo(fit)
#'   residuals(fit, type = "LOO-PIT")
#'
#'   # residuals from bias-adjusted predictions
#'   residuals(fit, bias_adjusted = TRUE)
#'
#'   # check Pareto k diagnostics
#'   plot(loo(fit))
#' }
#' }
#'
#' @references
#' \insertAllCited{}
#'
#' @seealso [predict.brma()], [blup.brma()], [pooled_effect()], [rstandard.brma()], [rstudent.brma()]
#' @exportS3Method
residuals.brma <- function(object, type = "outcome", unit = "estimate",
                           conditioning_depth = "marginal",
                           bias_adjusted = FALSE, ...) {

  # input validation
  dots                         <- list(...)
  conditioning_depth_specified <- !missing(conditioning_depth)
  .check_legacy_level_arg(dots, "residuals()")

  BayesTools::check_char(type, "type", allow_values = c("outcome", "pearson", "rstandard", "rstudent", "LOO-PIT"))
  unit                         <- .normalize_unit(unit)
  conditioning_depth           <- .normalize_conditioning_depth(conditioning_depth)
  BayesTools::check_bool(bias_adjusted, "bias_adjusted")

  if (is.element(type, c("LOO-PIT", "rstudent")) && conditioning_depth_specified) {
    stop(
      "LOO-PIT residuals use the estimate-unit LOO target; ",
      "do not set 'conditioning_depth'.",
      call. = FALSE
    )
  }

  .check_unit_conditioning_depth(
    object             = object,
    unit               = unit,
    conditioning_depth = conditioning_depth,
    caller             = "residuals()"
  )

  if (unit == "cluster") {
    .check_cluster_unit_deferred("residuals()")
  }

  # extract model characteristics for error checking
  outcome_type      <- .outcome_type(object)
  is_weightfunction <- .is_weightfunction(object)

  # check for unsupported type/model combinations
  .check_residual_type_availability(type, outcome_type, is_weightfunction)

  out <- .residuals_estimate.brma(
    object             = object,
    type               = type,
    conditioning_depth = conditioning_depth,
    bias_adjusted      = bias_adjusted
  )

  # clean names
  if (!is.data.frame(out)) {
    names(out) <- NULL
  }
  return(out)
}


# ---------------------------------------------------------------------------- #
# .residuals_estimate.brma
# ---------------------------------------------------------------------------- #
#
# Estimate-unit residuals, one value per effect-size estimate.
#
# @param object        brma object.
# @param type          character; residual type.
# @param conditioning_depth character; conditioning depth.
# @param bias_adjusted      logical; whether PET/PEESE bias adjustments are removed.
#
# @return numeric vector of residuals or standardized residuals.
#
# ---------------------------------------------------------------------------- #
.residuals_estimate.brma <- function(object, type, conditioning_depth,
                                     bias_adjusted) {

  if (is.element(type, c("LOO-PIT", "rstudent"))) {
    return(.standardized_residuals_loopit(object))
  }

  if (type == "rstandard") {
    return(rstandard.brma(
      model              = object,
      unit               = "estimate",
      conditioning_depth = conditioning_depth
    )[["z"]])
  }

  pred_type <- switch(conditioning_depth,
    "marginal" = "terms",
    "cluster"  = "cluster",
    "estimate" = "estimate"
  )

  fitted_samples <- predict.brma(
    object        = object,
    newdata       = NULL,
    type          = pred_type,
    bias_adjusted = bias_adjusted,
    quiet         = TRUE
  )

  yi <- .outcome_data_yi(object)
  K  <- length(yi)
  S  <- nrow(fitted_samples)

  yi_mat        <- matrix(yi, nrow = S, ncol = K, byrow = TRUE)
  resid_samples <- yi_mat - fitted_samples

  if (type == "pearson") {
    se_samples    <- .pearson_residual_se_samples(
      object             = object,
      conditioning_depth = conditioning_depth
    )
    resid_samples <- resid_samples / se_samples
  }

  return(colMeans(resid_samples))
}


# ---------------------------------------------------------------------------- #
# .pearson_residual_se_samples
# ---------------------------------------------------------------------------- #
#
# Compute posterior samples of Pearson residual standard errors.
#
# @param object brma object.
# @param conditioning_depth character; conditioning depth.
#
# @return S x K matrix of standard errors.
#
# ---------------------------------------------------------------------------- #
.pearson_residual_se_samples <- function(object, conditioning_depth) {

  is_multilevel <- .is_multilevel(object)
  is_scale      <- .is_scale(object)
  priors        <- object[["priors"]]
  vi            <- .outcome_data_vi(object)
  weights       <- .outcome_data_weights(object)
  K             <- length(vi)

  tau_result <- .evaluate.brma.tau(
    fit           = object[["fit"]],
    scale_data    = object[["data"]][["scale"]],
    scale_formula = if (is_scale) .create_fit_formula_list(data = object[["data"]], "scale") else NULL,
    scale_priors  = priors[["scale"]],
    is_scale      = is_scale,
    is_multilevel = is_multilevel,
    K             = K
  )

  tau_within  <- tau_result[["tau_within"]]
  tau_between <- tau_result[["tau_between"]]
  if (is.null(tau_between)) {
    tau_between <- matrix(0, nrow = nrow(tau_within), ncol = K)
  }

  vi_mat      <- matrix(vi, nrow = nrow(tau_within), ncol = K, byrow = TRUE)
  weights_mat <- matrix(weights, nrow = nrow(tau_within), ncol = K, byrow = TRUE)

  se2 <- switch(conditioning_depth,
    "marginal" = (vi_mat + tau_within^2) / weights_mat + tau_between^2,
    "cluster"  = (vi_mat + tau_within^2) / weights_mat,
    "estimate" = vi_mat / weights_mat
  )

  return(sqrt(se2))
}


#' @title Internally Standardized Residuals for brma Objects
#'
#' @description Computes internally standardized residuals from a fitted brma
#' object using the hat matrix. Returns a data frame with raw residuals,
#' standard errors, and standardized residuals (z-values). Available for normal
#' outcome models only.
#'
#' @param model a fitted brma object.
#' @param unit output unit. Only \code{"estimate"} is implemented currently.
#' @param conditioning_depth conditioning depth. Options are:
#' \itemize{
#'   \item \code{"marginal"} (default): Residuals from fixed effects predictions
#'     (\eqn{observed - X\beta}).
#'   \item \code{"cluster"}: Residuals from cluster-level predictions (observed -
#'     (\eqn{X\beta + gamma})). Only available for multilevel (3-level) models.
#'   \item \code{"estimate"}: Residuals from BLUPs,
#'     i.e., deviations of the observed effect sizes from the best linear
#'     unbiased predictions of the estimate-specific true effects (observed - theta).
#' }
#' @param ... additional arguments (currently ignored)
#'
#' @details
#' This function returns a data frame with three columns matching the output
#' of \code{metafor::rstandard}:
#' \itemize{
#'   \item \code{resid}: Raw residuals (observed - fitted values)
#'   \item \code{se}: Standard errors of the residuals
#'   \item \code{z}: Standardized residuals (resid / se)
#' }
#'
#' Internally standardized residuals divide the observed residuals by their
#' corresponding standard errors computed using the hat matrix. For a correctly
#' specified model, these residuals should approximately follow a standard
#' normal distribution.
#'
#' This function is only available for normal outcome models without selection
#' (weightfunction) bias adjustment. For other model types, use
#' \code{\link{rstudent.brma}} which uses LOO-PIT.
#'
#' @return A data frame with columns:
#' \itemize{
#'   \item \code{resid}: Raw residuals
#'   \item \code{se}: Standard errors of the residuals
#'   \item \code{z}: Standardized residuals
#' }
#' @examples \dontrun{
#' if (requireNamespace("metadat", quietly = TRUE)) {
#'   data(dat.lehmann2018, package = "metadat")
#'   fit <- brma(yi = yi, vi = vi, data = dat.lehmann2018, measure = "SMD")
#'
#'   # marginal internally standardized residuals (default)
#'   rstandard(fit)
#'
#'   # estimate-level (BLUP-based) residuals
#'   rstandard(fit, conditioning_depth = "estimate")
#' }
#' }
#'
#' @seealso [rstudent.brma()], [residuals.brma()], [blup.brma()], [predict.brma()]
#' @exportS3Method
rstandard.brma <- function(model, unit = "estimate",
                           conditioning_depth = "marginal", ...) {

  dots <- list(...)
  .check_legacy_level_arg(dots, "rstandard()")

  unit               <- .normalize_unit(unit)
  conditioning_depth <- .normalize_conditioning_depth(conditioning_depth)
  .check_unit_conditioning_depth(
    object             = model,
    unit               = unit,
    conditioning_depth = conditioning_depth,
    caller             = "rstandard()"
  )

  if (unit == "cluster") {
    .check_cluster_unit_deferred("rstandard()")
  }

  # check model type availability
  outcome_type      <- .outcome_type(model)
  is_weightfunction <- .is_weightfunction(model)

  if (outcome_type != "norm") {
    stop(
      "rstandard is only available for normal outcome models. ",
      "Use rstudent() for GLMM models.",
      call. = FALSE
    )
  }
  if (is_weightfunction) {
    stop(
      "rstandard is not available for selection models (weightfunction). ",
      "Use rstudent() for selection models.",
      call. = FALSE
    )
  }
  # compute residuals and standard errors using the same GLS projection
  hat_res <- .compute_hat_matrix_samples(
    object             = model,
    conditioning_depth = conditioning_depth,
    return_se          = TRUE,
    return_resid       = TRUE,
    summarize          = TRUE
  )

  # construct output data frame matching metafor::rstandard format
  out <- data.frame(
    resid = hat_res[["resid"]],
    se    = hat_res[["se"]],
    z     = hat_res[["z"]]
  )
  out <- .diagnostic_set_rownames(out, model)

  return(out)
}


#' @title Externally Standardized (Studentized) Residuals for brma Objects
#'
#' @description Computes externally standardized residuals (also called
#' studentized residuals or standardized deleted residuals) from a fitted brma
#' object using LOO-PIT (Leave-One-Out Probability Integral Transform).
#' Returns a data frame with raw residuals, standard errors, and standardized
#' residuals (z-values).
#'
#' @param model a fitted brma object.
#' @param unit output unit. Only \code{"estimate"} is available for LOO-PIT
#' residuals.
#' @param conditioning_depth unused for LOO-PIT residuals. LOO-PIT residuals
#' always use the estimate-unit LOO target.
#' @param ... additional arguments (currently ignored)
#'
#' @details
#' This function returns a data frame with three columns matching the output
#' of \code{metafor::rstudent}:
#' \itemize{
#'   \item \code{resid}: LOO predictive residuals (observed - fitted values)
#'   \item \code{se}: LOO predictive standard errors when available
#'   \item \code{z}: Externally standardized residuals (LOO-PIT transformed)
#' }
#'
#' LOO-PIT residuals are the Bayesian equivalent of studentized deleted
#' residuals. They are computed via leave-one-out probability integral
#' transformation using Pareto smoothed importance sampling. For each
#' observation, the LOO-weighted CDF is computed and transformed to a
#' standard normal quantile.
#'
#' Under a correctly specified model, LOO-PIT residuals should follow a
#' standard normal distribution. Large absolute values may indicate outliers
#' or model misspecification.
#'
#' The \code{z} column is the primary standardized diagnostic. The \code{resid}
#' and \code{se} columns are raw-scale companions computed from LOO predictive
#' moments using the normalized PSIS weights. For selection models, these moments
#' are computed from the fitted selected-normal predictive distribution. For
#' GLMMs, they are computed on the approximate effect-size scale used by the
#' LOO-PIT diagnostic; they are not exact PIT diagnostics for the raw count
#' likelihood.
#'
#' Unlike \code{\link{rstandard.brma}} (which uses the hat matrix), LOO-PIT
#' residuals properly account for estimation uncertainty and leverage without
#' requiring explicit hat matrix computation. This makes \code{rstudent.brma}
#' suitable for all model types including selection models and GLMMs.
#'
#' @return A data frame with columns:
#' \itemize{
#'   \item \code{resid}: Raw residuals
#'   \item \code{se}: Standard errors of the residuals
#'   \item \code{z}: Externally standardized residuals (LOO-PIT)
#' }
#'
#' @examples \dontrun{
#' if (requireNamespace("metadat", quietly = TRUE)) {
#'   data(dat.lehmann2018, package = "metadat")
#'   fit <- brma(yi = yi, vi = vi, data = dat.lehmann2018, measure = "SMD")
#'   fit <- add_loo(fit)
#'
#'   # externally standardized residuals
#'   rstudent(fit)
#'
#'   # check Pareto k values
#'   plot(loo(fit))
#' }
#' }
#'
#' @seealso [rstandard.brma()], [residuals.brma()], [loo.brma()], [blup.brma()]
#' @exportS3Method
rstudent.brma <- function(model, unit = "estimate",
                          conditioning_depth = "marginal", ...) {

  dots                         <- list(...)
  .psis_context                <- dots[[".psis_context"]]
  dots[[".psis_context"]]      <- NULL
  conditioning_depth_specified <- !missing(conditioning_depth)
  .check_legacy_level_arg(dots, "rstudent()")

  unit <- .normalize_unit(unit)

  if (unit == "cluster") {
    stop(
      "Cluster-unit rstudent residuals are not available because multivariate ",
      "LOO-PIT residuals are not unique. Use unit = 'estimate'.",
      call. = FALSE
    )
  }

  if (conditioning_depth_specified) {
    stop(
      "LOO-PIT residuals use the estimate-unit LOO target; ",
      "do not set 'conditioning_depth'.",
      call. = FALSE
    )
  }

  .check_unit_conditioning_depth(
    object             = model,
    unit               = unit,
    conditioning_depth = "estimate",
    caller             = "rstudent()"
  )

  setup <- .estimate_likelihood_setup.brma(model)

  # extract PSIS object once and reuse it for PIT and LOO expectations
  psis_context <- .diagnostic_psis_context(model, .psis_context)
  psis_object  <- psis_context[["psis_object"]]
  psis_weights <- psis_context[["psis_weights"]]

  # get LOO-PIT z values using the estimate-unit LOO target
  .diagnostic_check_loo(model, psis_context, unit = "estimate")

  if (setup[["outcome_type"]] == "norm" && setup[["is_weightfunction"]]) {
    summary <- .loo_predictive_selnorm_summary_estimate(
      object       = model,
      setup        = setup,
      psis_weights = psis_weights
    )

    out <- data.frame(
      resid = summary[["resid"]],
      se    = summary[["se"]],
      z     = summary[["z"]]
    )
    out <- .diagnostic_set_rownames(out, model)

    return(out)
  }

  z <- .standardized_residuals_loopit(
    object       = model,
    psis_weights = psis_weights,
    check        = FALSE
  )

  moments <- .loo_predictive_moments_estimate(
    object       = model,
    setup        = setup,
    psis_object  = psis_object,
    psis_weights = psis_weights
  )

  resid <- moments[["resid"]]
  se    <- moments[["se"]]

  # construct output data frame matching metafor::rstudent format
  out <- data.frame(
    resid = resid,
    se    = se,
    z     = z
  )
  out <- .diagnostic_set_rownames(out, model)

  return(out)
}


# ---------------------------------------------------------------------------- #
# .loo_predictive_moments_estimate
# ---------------------------------------------------------------------------- #
#
# Estimate-unit LOO predictive residual mean and standard deviation.
#
# The LOO-PIT z-value is the primary standardized residual. These moments provide
# raw-scale companions for the output table and residual funnel. The aggregation
# uses normalized PSIS weights on analytic predictive first and second moments,
# matching loo::E_loo(..., type = "mean") values without recomputing Pareto k.
#
# @param object       brma object.
# @param setup        output from .estimate_likelihood_setup.brma().
# @param psis_object  PSIS object from loo.
# @param psis_weights optional normalized PSIS weights.
#
# @return list with numeric vectors resid and se.
#
# ---------------------------------------------------------------------------- #
.loo_predictive_moments_estimate <- function(object, setup, psis_object,
                                             psis_weights = NULL) {

  yi                <- setup[["yi"]]
  sei               <- setup[["sei"]]
  K                 <- setup[["K"]]
  mu_samples        <- setup[["mu"]]
  tau_within        <- setup[["tau_within"]]
  outcome_type      <- setup[["outcome_type"]]
  is_weightfunction <- setup[["is_weightfunction"]]

  sei_mat <- matrix(sei, nrow = nrow(mu_samples), ncol = K, byrow = TRUE)

  if (is.null(psis_weights)) {
    psis_weights <- loo::weights.importance_sampling(
      psis_object,
      log       = FALSE,
      normalize = TRUE
    )
  }

  if (outcome_type == "norm" && is_weightfunction) {
    summary <- .loo_predictive_selnorm_summary_estimate(
      object       = object,
      setup        = setup,
      psis_weights = psis_weights
    )

    return(list(
      resid = summary[["resid"]],
      se    = summary[["se"]]
    ))
  } else {
    pred_var       <- tau_within^2 + sei_mat^2
    mean_samples   <- mu_samples
    second_samples <- pred_var + mu_samples^2
  }

  mean_samples   <- as.matrix(mean_samples)
  second_samples <- as.matrix(second_samples)

  pred_mean   <- colSums(psis_weights * mean_samples)
  pred_second <- colSums(psis_weights * second_samples)

  resid <- yi - pred_mean
  se    <- sqrt(pmax(pred_second - pred_mean^2, 0))

  return(list(
    resid = resid,
    se    = se
  ))
}


# ---------------------------------------------------------------------------- #
# .loo_predictive_selnorm_summary_estimate
# ---------------------------------------------------------------------------- #
#
# Fused selected-normal LOO-PIT and predictive moment reducer.
#
# ---------------------------------------------------------------------------- #
.loo_predictive_selnorm_summary_estimate <- function(object, setup,
                                                     psis_weights) {

  yi                <- setup[["yi"]]
  sei               <- setup[["sei"]]
  K                 <- setup[["K"]]
  mu_samples        <- setup[["mu"]]
  tau_within        <- setup[["tau_within"]]
  posterior_samples <- setup[["posterior_samples"]]
  S                 <- nrow(mu_samples)
  sei_mat           <- matrix(sei, nrow = S, ncol = K, byrow = TRUE)
  total_sd          <- sqrt(tau_within^2 + sei_mat^2)

  selection_context <- .selection_context(
    object            = object,
    posterior_samples = posterior_samples
  )
  summary <- .selection_step_weighted_summary(
    yi                = yi,
    mean              = mu_samples,
    sd                = total_sd,
    sei               = sei,
    psis_weights      = psis_weights,
    selection_context = selection_context
  )

  u_values <- pmax(pmin(summary[["cdf"]], 1 - 1e-10), 1e-10)
  resid    <- yi - summary[["mean"]]
  se       <- sqrt(pmax(summary[["second"]] - summary[["mean"]]^2, 0))

  return(list(
    resid = resid,
    se    = se,
    z     = stats::qnorm(u_values),
    u     = u_values
  ))
}


# ---------------------------------------------------------------------------- #
# .outcome_moments.selnorm
# ---------------------------------------------------------------------------- #
#
# Per-posterior-sample first and second moments for selected-normal kernels.
#
# ---------------------------------------------------------------------------- #
.outcome_moments.selnorm <- function(mu_samples, tau_within, sei,
                                     selection_context) {

  S        <- nrow(mu_samples)
  K        <- ncol(mu_samples)
  sei_mat  <- matrix(sei, nrow = S, ncol = K, byrow = TRUE)
  total_sd <- sqrt(tau_within^2 + sei_mat^2)

  return(.selection_step_moments_matrix(
    mean              = mu_samples,
    sd                = total_sd,
    sei               = sei,
    selection_context = selection_context
  ))
}

# ---------------------------------------------------------------------------- #
# .residuals_cluster.brma
# ---------------------------------------------------------------------------- #
#
# Deferred cluster-unit residual diagnostics.
#
# @param object brma object.
# @param type   character; residual type requested by the public caller.
# @param conditioning_depth character; conditioning depth.
#
# @return stops with a deferred-design error.
#
# ---------------------------------------------------------------------------- #
.residuals_cluster.brma <- function(object, type, conditioning_depth) {

  .check_cluster_unit_deferred("residuals()")
}


# ---------------------------------------------------------------------------- #
# .standardized_residuals_loopit
# ---------------------------------------------------------------------------- #
#
# Compute LOO-PIT residuals.
#
# This is the Bayesian equivalent of studentized deleted residuals. It computes
# leave-one-out probability integral transform (PIT) residuals using PSIS weights
# to reweight the posterior samples.
#
# The algorithm:
# 1. Compute LOO via PSIS (reusing loo.brma)
# 2. Extract normalized PSIS weights (S x K matrix)
# 3. Compute CDF matrix F(yi | parameters^(s)) (S x K matrix) for the LOO target
# 4. For each observation i, compute LOO-weighted CDF:
#    u_i = sum_s w_{is} * F(yi | theta^(s))
# 5. Transform to standard normal: r_i = qnorm(u_i)
#
# Unlike traditional standardized residuals, LOO-PIT residuals:
# - Account for estimation uncertainty (integrate over posterior)
# - Account for leverage (PSIS effectively removes yi from posterior)
# - Work for all model types (selection models, GLMMs)
#
# @param object brma object.
#
# @return If loo_only = TRUE, returns the psis_loo object.
#         Otherwise, returns numeric vector of LOO-PIT residuals.
#
# ---------------------------------------------------------------------------- #
.standardized_residuals_loopit <- function(object, psis_weights = NULL, check = TRUE) {

  # extract PSIS object and get normalized weights
  if (is.null(psis_weights)) {
    psis_weights <- loo_weights(object, unit = "estimate")
  }

  # check Pareto k diagnostics and warn if unreliable
  if (check) {
    check_loo(object, unit = "estimate")
  }

  setup <- .estimate_likelihood_setup.brma(object)
  if (setup[["outcome_type"]] == "norm" && setup[["is_weightfunction"]]) {
    summary <- .loo_predictive_selnorm_summary_estimate(
      object       = object,
      setup        = setup,
      psis_weights = psis_weights
    )
    return(summary[["z"]])
  }

  # compute CDF matrix (S x K) for the estimate-unit LOO target
  cdf_matrix <- .cdf_lik_estimate.brma(object)

  # compute LOO-weighted CDF for each observation
# u_i = sum_s w_{is} * F(yi | parameters^(s))
  # this is a weighted average across posterior samples
  u_values <- colSums(psis_weights * cdf_matrix)

  # clamp to avoid infinite quantiles
  u_values <- pmax(pmin(u_values, 1 - 1e-10), 1e-10)

  # transform to standard normal quantiles: r_i = qnorm(u_i)
  resid_point <- stats::qnorm(u_values)

  return(resid_point)
}


# ---------------------------------------------------------------------------- #
# .check_residual_type_availability
# ---------------------------------------------------------------------------- #
#
# Helper function to validate that the requested residual type is compatible
# with the model's outcome type and presence of selection (weightfunction).
#
# @param type Residual type requested
# @param outcome_type Model outcome type
# @param is_weightfunction Whether model includes weightfunction
#
# @return NULL (throws error if invalid combination)
# @keywords internal
#
# ---------------------------------------------------------------------------- #
.check_residual_type_availability <- function(type, outcome_type, is_weightfunction) {
  if (type == "pearson") {
    if (outcome_type != "norm") {
      stop(
        "Pearson residuals are only available for normal outcome models. ",
        "Use type = 'LOO-PIT' for GLMM models.",
        call. = FALSE
      )
    }
    if (is_weightfunction) {
      stop(
        "Pearson residuals are not available for selection models (weightfunction). ",
        "Use type = 'LOO-PIT' for selection models.",
        call. = FALSE
      )
    }
  }

  if (type == "rstandard") {
    if (outcome_type != "norm") {
      stop(
        "Standardized residuals (rstandard) are only available for normal outcome models. ",
        "Use type = 'LOO-PIT' for GLMM models.",
        call. = FALSE
      )
    }
    if (is_weightfunction) {
      stop(
        "Standardized residuals (rstandard) are not available for selection models (weightfunction). ",
        "Use type = 'LOO-PIT' for selection models.",
        call. = FALSE
      )
    }
  }

  return(invisible(NULL))
}

# ============================================================================ #
# brma.dfbetas.R
# ============================================================================ #
#
# This file contains the dfbetas method for brma model fits. DFBETAS values
# measure the influence of each observation on the estimated model coefficients.
#
# Instead of computationally expensive refitting (leave-one-out), we use
# Pareto Smoothed Importance Sampling (PSIS) weights from the LOO-CV
# approximation to estimate the leave-one-out coefficients.
#
# ============================================================================ #


#' @title DFBETAS for brma Objects
#'
#' @description Computes DFBETAS (Difference in BETAS, standardized) for a
#' fitted brma object. DFBETAS measures the influence of each observation on
#' the estimated model coefficients. Positive values indicate that deleting the
#' observation yields a smaller estimate, negative values indicate that deleting
#' the observation yields a larger estimate.
#'
#' @param model a fitted brma object.
#' @param type type of parameters to be summarized. Defaults to \code{"mods"}
#' (for the effect size and meta-regression coefficients). The other options are
#' \code{"scale"} (for the heterogeneity and scale-regression coefficients) and
#' \code{"bias"} (for omega, PET, and PEESE publication-bias parameters).
#' @param standardized_coefficients whether to show standardized meta-regression coefficients.
#' Defaults to \code{FALSE}. When set to \code{TRUE}, standardized meta-regression
#' coefficients are returned for the intercept and continuous predictors. These coefficients
#' correspond to the standardized scale on which prior distributions are specified by default
#' (i.e., `standardize_continuous_predictors = TRUE`).
#' @param transform_factors whether to transform factors to their original names.
#' Defaults to \code{TRUE}.
#' @param return_loo_estimates whether to return the leave-one-out coefficient
#' estimates used to compute DFBETAS instead of standardized DFBETAS values.
#' Defaults to \code{FALSE}.
#' @param ... additional arguments (currently ignored).
#'
#' @details
#' This function computes DFBETAS values using the Leave-One-Out (LOO)
#' approximation based on Pareto Smoothed Importance Sampling (PSIS) weights.
#' Ideally, DFBETAS is defined as:
#' \deqn{DFBETAS_{ij} = \frac{\hat{\beta}_j - \hat{\beta}_{j(-i)}}{SE(\hat{\beta}_{j(-i)})}}
#' where \eqn{\hat{\beta}_j} is the estimate of the \eqn{j}-th coefficient using
#' the full data, \eqn{\hat{\beta}_{j(-i)}} is the estimate when observation \eqn{i}
#' is omitted, and \eqn{SE(\hat{\beta}_{j(-i)})} is the standard error of the
#' coefficient when observation \eqn{i} is omitted.
#'
#' In the Bayesian context using LOO approximation:
#' \itemize{
#'   \item \eqn{\hat{\beta}_{j(-i)}} is estimated as the importance sampling
#'     weighted mean of the posterior samples, using PSIS weights \eqn{w_{is}}.
#'   \item \eqn{SE(\hat{\beta}_{j(-i)})} is estimated as the importance sampling
#'     weighted standard deviation of the posterior samples.
#' }
#'
#' This approximation allows computing influence statistics without refitting
#' the model \eqn{K} times, making it computationally efficient.
#' For \code{type = "bias"}, fixed identification parameters (e.g., the reference
#' \eqn{\omega = 1} interval) can have zero LOO posterior standard deviation.
#' These parameters are retained in the output, but their DFBETAS values are
#' reported as \code{NaN} because the standardized diagnostic is undefined.
#'
#' Note: This function requires that LOO-CV has been computed for the model
#' using \code{\link{add_loo}}.
#'
#' @return If `return_loo_estimates = FALSE`, a data frame with \eqn{K} rows
#' (observations) and \eqn{P} columns (parameters), containing DFBETAS values.
#' If `return_loo_estimates = TRUE`, returns the corresponding leave-one-out
#' coefficient estimates. Row names correspond to study labels (if available)
#' or indices.
#'
#' @examples \dontrun{
#' if (requireNamespace("metadat", quietly = TRUE)) {
#'   data(dat.lehmann2018, package = "metadat")
#'   fit <- bPET(yi = yi, vi = vi, data = dat.lehmann2018, measure = "SMD")
#'   fit <- add_loo(fit)
#'
#'   inf <- dfbetas(fit)
#'   plot(inf[, 1], type = "h")
#' }
#' }
#'
#' @seealso \code{\link{add_loo}}, \code{\link{loo_weights.brma}}
#' @importFrom stats dfbetas
#' @exportS3Method
dfbetas.brma <- function(model, type = "mods", standardized_coefficients = FALSE,
                         transform_factors = TRUE,
                         return_loo_estimates = FALSE, ...) {

  dots <- list(...)
  .weights <- dots[[".weights"]]
  BayesTools::check_char(type, "type", allow_values = c("mods", "scale", "bias"))

  # get PSIS weights (S x K matrix)
  weights <- .diagnostic_psis_weights(model, .weights)

  # determine whether to extract formula (for meta-regression) or parameter (for intercept-only)
  is_mods  <- .is_mods(model)
  is_scale <- .is_scale(model)
  is_bias  <- .is_bias(model)

  if (type == "mods") {
    if (is_mods) {
      # meta-regression: means mu is a formula
      samples_table <- BayesTools::JAGS_estimates_table(
        fit                = model[["fit"]],
        keep_formulas      = "mu",
        remove_diagnostics = TRUE,
        transform_factors  = transform_factors,
        transform_scaled   = !standardized_coefficients,
        return_samples     = TRUE
      )
    } else {
      # random/fixed effects: mu is a parameter (intercept)
      samples_table <- BayesTools::JAGS_estimates_table(
        fit                = model[["fit"]],
        keep_parameters    = "mu",
        remove_diagnostics = TRUE,
        transform_factors  = transform_factors,
        transform_scaled   = !standardized_coefficients,
        return_samples     = TRUE
      )
    }
  } else if (type == "scale") {
    if (is_scale) {
      # scale-regression: tau is modeled via log_tau formula
      samples_table <- BayesTools::JAGS_estimates_table(
        fit                = model[["fit"]],
        keep_formulas      = "log_tau",
        remove_diagnostics = TRUE,
        transform_factors  = transform_factors,
        return_samples     = TRUE
      )
    } else {
      # random/fixed effects: tau is a parameter (intercept)
      samples_table <- BayesTools::JAGS_estimates_table(
        fit                = model[["fit"]],
        keep_parameters    = "tau",
        remove_diagnostics = TRUE,
        transform_factors  = transform_factors,
        return_samples     = TRUE
      )
    }
  } else if (type == "bias") {
    if (!is_bias) {
      stop("type = 'bias' is only available for models with publication bias adjustment.", call. = FALSE)
    }

    samples_table <- .dfbetas_bias_samples(model)
  }


  # samples_table is a flattened data frame (S rows x P columns)
  # ensure it's a matrix for computation
  samples_mat <- as.matrix(samples_table)

  if (ncol(samples_mat) == 0) {
    stop("No parameters available for DFBETAS with the requested type.", call. = FALSE)
  }

  # dimensions
  S <- nrow(samples_mat) # number of samples
  P <- ncol(samples_mat) # number of coefficients
  K <- ncol(weights)     # number of observations

  # 1. Compute LOO-weighted means for each observation i and parameter j
  # beta_loo[i, j] = sum_s (w_{is} * beta_{js})
  # weights is S x K -> t(weights) is K x S
  # samples_mat is S x P
  # K x S %*% S x P -> K x P
  beta_loo <- crossprod(weights, samples_mat)

  # 2. Compute Robust Weighted SD (SE LOO)
  # Uses centered moment calculation: sum(w * (x - mu)^2)
  # This avoids catastrophic cancellation issues with E[x^2] - E[x]^2
  se_loo <- matrix(NA, nrow = K, ncol = P)

  for (j in seq_len(P)) {
    # Get samples for parameter j (vector length S)
    beta_s <- samples_mat[, j]

    # Get LOO means for parameter j (vector length K)
    mu_k <- beta_loo[, j]

    # Efficient Centering:
    # Use outer() to create an (S x K) matrix of differences
    # element [s, k] = beta_s[s] - mu_k[k]
    # This vectorizes the subtraction of every sample from every LOO mean
    diff_mat <- outer(beta_s, mu_k, "-")

    # Square differences
    diff_sq <- diff_mat^2

    # Weighted Sum of Squares (Variance)
    # colSums of (S x K) * (S x K) -> Vector length K
    # This is the "Population Variance" of the LOO posterior distribution
    var_k <- colSums(weights * diff_sq)

    # Store SE
    se_loo[, j] <- sqrt(var_k)
  }

  # 3. Compute full posterior means (unweighted / equal weights)
  # beta_full[j] = mean(beta_{js})
  beta_full <- colMeans(samples_mat)

  # expand beta_full to K x P matrix for vectorized subtraction
  beta_full_mat <- matrix(beta_full, nrow = K, ncol = P, byrow = TRUE)

  # 4. Return LOO estimates if requested
  if (return_loo_estimates) {
    beta_loo_df <- as.data.frame(beta_loo)
    colnames(beta_loo_df) <- colnames(samples_mat)
    beta_loo_df <- .diagnostic_set_rownames(beta_loo_df, model)
    return(beta_loo_df)
  }

  # 4. Compute DFBETAS
  # (beta_full - beta_loo) / se_loo
  dfbetas_val <- (beta_full_mat - beta_loo) / se_loo
  undefined   <- apply(se_loo <= sqrt(.Machine$double.eps), 2, any)
  if (any(undefined)) {
    dfbetas_val[, undefined] <- NaN
  }

  # convert to data frame
  dfbetas_df <- as.data.frame(dfbetas_val)
  colnames(dfbetas_df) <- colnames(samples_mat)
  dfbetas_df <- .diagnostic_set_rownames(dfbetas_df, model)

  if (any(undefined)) {
    dfbetas_df <- .diagnostic_with_note(
      dfbetas_df,
      class = "dfbetas.brma",
      note  = .diagnostic_zero_variance_note(
        diagnostic = "DFBETAS",
        parameters = colnames(samples_mat)[undefined]
      )
    )
  }

  return(dfbetas_df)
}


#' @exportS3Method
print.dfbetas.brma <- function(x, ...) {

  note <- attr(x, "note")
  class(x) <- "data.frame"
  print(x, ...)
  .print_diagnostic_note(note)

  return(invisible(x))
}


# ---------------------------------------------------------------------------- #
# .dfbetas_bias_samples
# ---------------------------------------------------------------------------- #
#
# Extract publication-bias posterior samples without using JAGS_estimates_table.
# BayesTools summary tables always run formula/factor post-processing, which can
# fail for RoBMA mixture-prior objects when we only need raw omega/PET/PEESE draws.
#
# ---------------------------------------------------------------------------- #
.dfbetas_bias_samples <- function(model) {

  posterior_samples <- .get_posterior_samples(model[["fit"]])
  bias_samples      <- list()
  omega_cols        <- grep("^omega(\\[|$)", colnames(posterior_samples), value = TRUE)

  if (length(omega_cols) > 0L) {
    bias_samples[["omega"]] <- .extract_omega_samples(posterior_samples)
  }

  for (par in c("PET", "PEESE")) {
    if (par %in% colnames(posterior_samples)) {
      bias_samples[[par]] <- as.matrix(posterior_samples[, par, drop = FALSE])
    }
  }

  if (length(bias_samples) == 0L) {
    return(matrix(nrow = nrow(posterior_samples), ncol = 0))
  }

  return(do.call(cbind, bias_samples))
}

# Native interface helpers ----------------------------------------------------

.native_numeric_matrix <- function(x) {

  if (!is.matrix(x)) {
    x <- as.matrix(x)
  }
  if (!is.double(x)) {
    storage.mode(x) <- "double"
  }
  return(x)
}

.native_numeric_vector <- function(x) {

  if (!is.double(x)) {
    x <- as.double(x)
  }
  return(x)
}

.native_integer_vector <- function(x) {

  if (!is.integer(x)) {
    x <- as.integer(x)
  }
  return(x)
}

# ============================================================================ #
# brma.marginal_means.R
# ============================================================================ #
#
# Estimated marginal means for brma objects with moderators.
#
# The expensive computation is delegated to BayesTools::as_marginal_inference()
# and stored in a small RoBMA-side result object. Summary and plot methods only
# format or render the stored BayesTools marginal-inference object.
#
# ============================================================================ #


#' @title Estimated Marginal Means
#'
#' @description S3 generic for estimated marginal means. The \code{brma}
#' method works with fitted moderator models and stores the BayesTools
#' marginal-inference object for \code{summary()} and \code{plot()}.
#'
#' @param object a fitted model object.
#' @param ... additional arguments passed to methods.
#'
#' @return A method-specific estimated marginal means object.
#'
#' @seealso [summary()], [plot()], [summary.brma()], [regplot()]
#' @export
marginal_means <- function(object, ...) {

  UseMethod("marginal_means")
}


#' @title Estimated Marginal Means for brma Objects
#'
#' @description Computes estimated marginal means for a fitted \code{brma}
#' object with moderators using \code{BayesTools::as_marginal_inference()}.
#'
#' @param object a fitted \code{brma} object with moderators.
#' @param null_hypothesis point null hypothesis used for inclusion Bayes
#' factors. Defaults to \code{0}.
#' @param normal_approximation whether prior and posterior density at the null
#' should be approximated with a normal distribution. Defaults to \code{FALSE}.
#' @param n_samples number of samples/grid points used by BayesTools for
#' marginal prior densities. Defaults to \code{10000}.
#' @param bf whether inclusion Bayes factors should be shown by default in
#' summaries. Defaults to \code{TRUE} for RoBMA/BMA objects and \code{FALSE}
#' for single-model \code{brma} objects.
#' @inheritParams predict.brma
#' @param ... additional arguments (currently ignored).
#'
#' @return A list of class \code{marginal_means.brma} containing the
#' BayesTools \code{marginal_inference} object and parameter metadata.
#'
#' @examples \dontrun{
#' if (requireNamespace("metadat", quietly = TRUE) &&
#'     requireNamespace("metafor", quietly = TRUE)) {
#'   data(dat.bcg, package = "metadat")
#'   dat <- metafor::escalc(
#'     measure = "RR",
#'     ai      = tpos,
#'     bi      = tneg,
#'     ci      = cpos,
#'     di      = cneg,
#'     data    = dat.bcg
#'   )
#'
#'   fit <- brma(yi = yi, vi = vi, mods = ~ alloc, data = dat, measure = "RR")
#'   mm  <- marginal_means(fit)
#'   summary(mm)
#'   plot(mm, parameter = "alloc")
#' }
#' }
#'
#' @seealso [summary()], [plot()], [summary.brma()], [regplot()]
#' @export
marginal_means.brma <- function(object, null_hypothesis = 0,
                                normal_approximation = FALSE,
                                n_samples = 10000,
                                output_measure = NULL, transform = NULL,
                                bf = NULL, ...) {

  if (!.is_mods(object)) {
    stop("'marginal_means' requires a model with moderators.", call. = FALSE)
  }
  if (is.null(object[["fit"]]) || length(object[["fit"]]) == 0L) {
    stop("'marginal_means' requires a fitted brma object.", call. = FALSE)
  }

  BayesTools::check_real(null_hypothesis, "null_hypothesis", check_length = 1)
  BayesTools::check_bool(normal_approximation, "normal_approximation")
  BayesTools::check_int(n_samples, "n_samples", lower = 2)
  model_averaged <- .is_RoBMA(object)
  bf             <- .marginal_means_resolve_bf(model_averaged, bf)
  effect_transform <- .effect_output_setup(
    object         = object,
    output_measure = output_measure,
    transform      = transform
  )

  design           <- .fitted_formula_design(object, "mu", required = TRUE)
  formula          <- .fitted_formula_evaluable(object, design, "mods")
  terms            <- .marginal_means_terms(design)
  parameters       <- BayesTools::JAGS_parameter_names(
    parameters        = terms,
    formula_parameter = "mu"
  )
  conditional_list <- .marginal_means_conditional_list(
    terms      = terms,
    parameters = parameters
  )

  inference <- suppressWarnings(BayesTools::as_marginal_inference(
    model                = object[["fit"]],
    marginal_parameters  = parameters,
    parameters           = parameters,
    conditional_list     = conditional_list,
    conditional_rule     = "OR",
    formula              = formula,
    null_hypothesis      = null_hypothesis,
    normal_approximation = normal_approximation,
    n_samples            = n_samples,
    silent               = TRUE,
    force_plots          = TRUE
  ))

  available_parameters <- Reduce(
    intersect,
    list(
      parameters,
      names(inference[["averaged"]]),
      names(inference[["conditional"]]),
      names(inference[["inference"]])
    )
  )

  if (length(available_parameters) == 0L) {
    stop("No marginal means are available for this model.", call. = FALSE)
  }

  term_map <- data.frame(
    term             = terms,
    parameter        = parameters,
    label            = terms,
    check.names      = FALSE,
    stringsAsFactors = FALSE
  )
  term_map <- term_map[term_map[["parameter"]] %in% available_parameters, , drop = FALSE]

  output <- list(
    inference              = inference,
    parameters             = available_parameters,
    term_map               = term_map,
    formula                = formula,
    null_hypothesis        = null_hypothesis,
    normal_approximation   = normal_approximation,
    n_samples              = n_samples,
    input_measure          = .measure(object),
    effect_transform       = effect_transform,
    model_averaged         = model_averaged,
    bf                     = bf,
    name                   = .summary.brma_model_names(object)
  )

  class(output) <- c("marginal_means.brma", "marginal_means")

  return(output)
}


#' @title Summarize Estimated Marginal Means
#'
#' @description Summarizes estimated marginal means stored in a
#' \code{marginal_means.brma} object using
#' \code{BayesTools::marginal_estimates_table()}.
#'
#' @param object a \code{marginal_means.brma} object.
#' @param type for RoBMA product-space objects, whether to summarize
#' model-averaged (\code{"averaged"}) or conditional (\code{"conditional"})
#' marginal means. Defaults to \code{"averaged"} and is available only for
#' RoBMA marginal means.
#' @param probs quantiles of the posterior distribution to be displayed.
#' Defaults to \code{c(.025, .50, .975)}.
#' @param logBF whether to show inclusion Bayes factors on the log scale.
#' Defaults to \code{FALSE}.
#' @param BF01 whether to show inverse inclusion Bayes factors. Defaults to
#' \code{FALSE}.
#' @param bf whether to show inclusion Bayes factors. Defaults to the setting
#' stored by \code{marginal_means()}.
#' @inheritParams predict.brma
#' @param ... additional arguments (currently ignored).
#'
#' @return A \code{BayesTools_table} of class
#' \code{summary.marginal_means.brma}.
#'
#' @export
summary.marginal_means.brma <- function(object, type = NULL,
                                        probs = c(.025, .50, .975),
                                        logBF = FALSE, BF01 = FALSE,
                                        bf = NULL,
                                        output_measure = NULL, transform = NULL, ...) {

  type <- .marginal_means_type(object = object, type = type)
  BayesTools::check_real(probs, "probs", allow_NULL = TRUE, check_length = 0)
  BayesTools::check_bool(logBF, "logBF")
  BayesTools::check_bool(BF01, "BF01")
  bf <- .marginal_means_resolve_bf(object[["model_averaged"]], bf, object[["bf"]])

  if (missing(output_measure) && missing(transform)) {
    effect_transform <- object[["effect_transform"]]
  } else {
    effect_transform <- .effect_output_setup_measure(
      input_measure  = object[["input_measure"]],
      output_measure = if (missing(output_measure)) NULL else output_measure,
      transform      = if (missing(transform)) NULL else transform
    )
  }

  samples    <- .transform_marginal_samples_effect(
    samples          = object[["inference"]][[type]],
    effect_transform = effect_transform
  )
  inference  <- object[["inference"]][["inference"]]
  parameters <- object[["parameters"]]
  parameters <- parameters[
    parameters %in% names(samples) &
      parameters %in% names(inference)
  ]

  if (length(parameters) == 0L) {
    stop("No marginal means are available for type = '", type, "'.",
         call. = FALSE)
  }

  estimates <- BayesTools::marginal_estimates_table(
    samples        = samples,
    inference      = inference,
    parameters     = parameters,
    probs          = probs,
    logBF          = logBF,
    BF01           = BF01,
    formula_prefix = FALSE,
    title          = .effect_output_title(
      title = if (isTRUE(object[["model_averaged"]])) {
        switch(
          type,
          "averaged"    = "Model-Averaged Marginal Means:",
          "conditional" = "Conditional Marginal Means:"
        )
      } else {
        "Marginal Means:"
      },
      effect_transform = effect_transform
    ),
    footnotes      = effect_transform[["note"]]
  )

  if (!bf) {
    estimates <- .marginal_means_drop_bf(estimates)
  }

  class(estimates) <- c("summary.marginal_means.brma", class(estimates))
  attr(estimates, "marginal_type") <- type

  return(estimates)
}


#' @title Print Estimated Marginal Means
#'
#' @description Prints the estimated marginal means summary.
#'
#' @param x a \code{marginal_means.brma} object.
#' @param ... additional arguments passed to \code{summary()}.
#'
#' @return Returns \code{x} invisibly.
#'
#' @export
print.marginal_means.brma <- function(x, ...) {

  print(summary(x, ...))

  return(invisible(x))
}


#' @title Print Summary of Estimated Marginal Means
#'
#' @description Prints a summary table of estimated marginal means.
#'
#' @param x a \code{summary.marginal_means.brma} object.
#' @param ... additional arguments (currently ignored).
#'
#' @return Returns \code{x} invisibly.
#'
#' @export
print.summary.marginal_means.brma <- function(x, ...) {

  class(x) <- setdiff(class(x), "summary.marginal_means.brma")

  cat("\n")
  print(x)
  cat("\n")

  return(invisible(x))
}


#' @title Plot Estimated Marginal Means
#'
#' @description Plots estimated marginal means stored in a
#' \code{marginal_means.brma} object using \code{BayesTools::plot_marginal()}.
#'
#' @param x a \code{marginal_means.brma} object.
#' @param parameter moderator term to plot. Use the original term name, for
#' example \code{"measure"}, \code{"intercept"} for the intercept when
#' available, \code{"mu"} as an intercept alias, or the internal parameter name,
#' for example \code{"mu_measure"}.
#' @param type for RoBMA product-space objects, whether to plot model-averaged
#' (\code{"averaged"}) or conditional (\code{"conditional"}) marginal means.
#' Defaults to \code{"averaged"} and is available only for RoBMA marginal
#' means.
#' @param prior whether the marginal prior distribution should be added to the
#' plot. Defaults to \code{FALSE}.
#' @param plot_type whether to use base R graphics (\code{"base"}) or ggplot2
#' (\code{"ggplot"}). Defaults to \code{"base"}.
#' @param dots_prior list of additional graphical arguments passed to the prior
#' plotting function.
#' @inheritParams predict.brma
#' @param ... additional graphical arguments passed to
#' \code{BayesTools::plot_marginal()}.
#'
#' @return \code{NULL} invisibly if \code{plot_type = "base"} or a ggplot object
#' if \code{plot_type = "ggplot"}.
#'
#' @export
plot.marginal_means.brma <- function(x, parameter, type = NULL,
                                     prior = FALSE, plot_type = "base",
                                     dots_prior = NULL,
                                     output_measure = NULL, transform = NULL, ...) {

  type <- .marginal_means_type(object = x, type = type)
  BayesTools::check_bool(prior, "prior")
  BayesTools::check_char(plot_type, "plot_type", allow_values = c("base", "ggplot"))

  if (missing(output_measure) && missing(transform)) {
    effect_transform <- x[["effect_transform"]]
  } else {
    effect_transform <- .effect_output_setup_measure(
      input_measure  = x[["input_measure"]],
      output_measure = if (missing(output_measure)) NULL else output_measure,
      transform      = if (missing(transform)) NULL else transform
    )
  }

  if (missing(parameter) || is.null(parameter)) {
    stop("The 'parameter' argument must be specified. Available terms are: ",
         .marginal_means_available_terms(x), ".", call. = FALSE)
  }

  selected <- .marginal_means_select_parameter(x, parameter)
  samples  <- x[["inference"]][[type]]

  if (!selected[["parameter"]] %in% names(samples)) {
    stop("No marginal means are available for parameter '",
         selected[["term"]], "' and type = '", type, "'.", call. = FALSE)
  }

  n_levels <- length(samples[[selected[["parameter"]]]])
  dots     <- .set_dots_plot(..., n_levels = n_levels)
  if (is.null(dots[["xlab"]])) {
    dots[["xlab"]] <- .plot_parameter_label("mu", effect_transform)
  }
  if (is.null(dots[["legend_title"]])) {
    dots[["legend_title"]] <- selected[["label"]]
  }
  prior_component_counts <- .marginal_means_prior_component_counts(
    samples[[selected[["parameter"]]]]
  )
  dots_prior <- .set_dots_prior_marginal_means(
    dots_prior             = dots_prior,
    n_levels               = n_levels,
    prior_component_counts = prior_component_counts
  )

  args                          <- dots
  args$samples                  <- samples
  args$parameter                <- selected[["parameter"]]
  args$plot_type                <- plot_type
  args$prior                    <- prior
  args$n_points                 <- 1000
  args$transformation           <- .effect_plot_transformation(effect_transform)
  args$transformation_arguments <- NULL
  args$transformation_settings  <- FALSE
  args$par_name                 <- dots[["xlab"]]
  args$dots_prior               <- dots_prior

  plot <- suppressMessages(do.call(BayesTools::plot_marginal, args))

  if (plot_type == "base") {
    return(invisible(plot))
  } else if (plot_type == "ggplot") {
    return(plot)
  }
}


# Extract model terms used by the moderator formula.
.marginal_means_terms <- function(formula) {

  if (inherits(formula, "BayesTools_formula_design")) {
    terms <- .formula_design_display_names(formula[["model_terms"]])
  } else {
    formula_terms <- stats::terms(formula)
    terms         <- c(
      if (attr(formula_terms, "intercept") == 1L) "intercept",
      attr(formula_terms, "term.labels")
    )
  }

  if (length(terms) == 0L) {
    stop("No moderator terms found.", call. = FALSE)
  }

  return(terms)
}


# Build BayesTools conditional-list specification for marginal means.
.marginal_means_conditional_list <- function(terms, parameters) {

  intercept_parameter <- parameters[terms == "intercept"]

  conditional_list <- lapply(seq_along(parameters), function(i) {

    c(
      if (length(intercept_parameter) > 0L && terms[i] != "intercept") {
        intercept_parameter
      },
      parameters[i]
    )
  })
  names(conditional_list) <- parameters

  return(conditional_list)
}


# Resolve whether marginal-means summaries should include inclusion BFs.
.marginal_means_resolve_bf <- function(model_averaged, bf = NULL, default = NULL) {

  if (is.null(default)) {
    default <- isTRUE(model_averaged)
  }

  if (is.null(bf)) {
    return(default)
  }

  BayesTools::check_bool(bf, "bf")

  return(bf)
}


# Remove BF columns and BF-only warnings from marginal-means summaries.
.marginal_means_drop_bf <- function(table) {

  table_type <- attr(table, "type")
  if (is.null(table_type) || !any(table_type == "inclusion_BF")) {
    return(table)
  }

  keep        <- table_type != "inclusion_BF"
  table_attrs <- attributes(table)
  table       <- table[, keep, drop = FALSE]

  class(table)        <- table_attrs[["class"]]
  attr(table, "type") <- table_type[keep]

  copy_attrs <- setdiff(
    names(table_attrs),
    c("names", "row.names", "class", "type")
  )
  for (attr_name in copy_attrs) {
    attr(table, attr_name) <- table_attrs[[attr_name]]
  }

  warnings <- attr(table, "warnings")
  if (!is.null(warnings)) {
    warnings <- warnings[!grepl("Savage-Dickey", warnings, fixed = TRUE)]
    if (length(warnings) == 0L) {
      warnings <- NULL
    }
    attr(table, "warnings") <- warnings
  }

  return(table)
}


# Resolve marginal-means summary/plot type.
.marginal_means_type <- function(object, type) {

  if (is.null(type)) {
    return("averaged")
  }

  BayesTools::check_char(type, "type")
  type <- match.arg(type, c("averaged", "conditional"))

  if (!isTRUE(object[["model_averaged"]])) {
    stop("The 'type' argument is available only for RoBMA marginal means.",
         call. = FALSE)
  }

  return(type)
}


# Select a marginal-means parameter from raw or internal names.
.marginal_means_select_parameter <- function(x, parameter) {

  BayesTools::check_char(parameter, "parameter", check_length = 1)

  term_map <- x[["term_map"]]
  lookup   <- parameter

  if (lookup == "mu") {
    lookup <- "intercept"
  }

  index <- match(lookup, term_map[["parameter"]])

  if (is.na(index)) {
    index <- match(lookup, term_map[["term"]])
  }

  if (is.na(index)) {
    jags_lookup <- BayesTools::JAGS_parameter_names(
      parameters        = lookup,
      formula_parameter = "mu"
    )
    index <- match(jags_lookup, term_map[["parameter"]])
  }

  if (is.na(index)) {
    stop("Unknown marginal means parameter '", parameter,
         "'. Available terms are: ", .marginal_means_available_terms(x),
         ".", call. = FALSE)
  }

  return(as.list(term_map[index, , drop = FALSE]))
}


# Format available marginal-means terms for error messages.
.marginal_means_available_terms <- function(x) {

  term_map <- x[["term_map"]]
  terms    <- unique(term_map[["term"]])

  return(paste0("'", terms, "'", collapse = ", "))
}


# Count prior components plotted for each marginal level.
.marginal_means_prior_component_counts <- function(samples_parameter) {

  if (!is.list(samples_parameter)) {
    samples_parameter <- list(samples_parameter)
  }

  counts <- vapply(samples_parameter, function(sample) {

    prior_density <- attr(sample, "prior_density")
    if (is.null(prior_density)) {
      return(1L)
    }

    count <- 0L
    if (!is.null(prior_density[["density"]])) {
      count <- count + 1L
    }
    if (!is.null(prior_density[["points"]]) &&
        NROW(prior_density[["points"]]) > 0L) {
      count <- count + 1L
    }

    return(max(count, 1L))
  }, integer(1))

  return(counts)
}


# Configure prior-line defaults for marginal means plots.
.set_dots_prior_marginal_means <- function(dots_prior, n_levels,
                                           prior_component_counts) {

  if (is.null(dots_prior)) {
    dots_prior <- list()
  }

  n_prior_levels <- sum(prior_component_counts)
  if (is.null(dots_prior[["col"]]) && n_levels == 1L) {
    dots_prior[["col"]] <- "black"
  } else if (is.null(dots_prior[["col"]]) && n_levels > 1L) {
    level_col <- .plot_level_palette(n_levels)
    dots_prior[["col"]] <- rep(level_col, prior_component_counts)
  } else if (length(dots_prior[["col"]]) == 1L) {
    dots_prior[["col"]] <- rep(dots_prior[["col"]], n_prior_levels)
  } else if (length(dots_prior[["col"]]) == n_levels &&
             n_prior_levels != n_levels) {
    dots_prior[["col"]] <- rep(dots_prior[["col"]], prior_component_counts)
  }
  if (is.null(dots_prior[["lty"]])) {
    dots_prior[["lty"]] <- rep(2, n_prior_levels)
  } else if (length(dots_prior[["lty"]]) == 1L) {
    dots_prior[["lty"]] <- rep(dots_prior[["lty"]], n_prior_levels)
  } else if (length(dots_prior[["lty"]]) == n_levels &&
             n_prior_levels != n_levels) {
    dots_prior[["lty"]] <- rep(dots_prior[["lty"]], prior_component_counts)
  }
  if (!is.null(dots_prior[["linetype"]]) &&
      length(dots_prior[["linetype"]]) == 1L) {
    dots_prior[["linetype"]] <- rep(dots_prior[["linetype"]], n_prior_levels)
  } else if (!is.null(dots_prior[["linetype"]]) &&
             length(dots_prior[["linetype"]]) == n_levels &&
             n_prior_levels != n_levels) {
    dots_prior[["linetype"]] <- rep(
      dots_prior[["linetype"]],
      prior_component_counts
    )
  }

  return(dots_prior)
}

# ============================================================================ #
# brma.dffits.R
# ============================================================================ #
#
# This file contains the dffits method for brma model fits. DFFITS measures
# the number of standard deviations that the fitted value changes if observation
# i is removed.
#
# ============================================================================ #


#' @export
dffits <- function(model, ...) UseMethod("dffits")

#' @title DFFITS for brma Objects
#'
#' @description Computes DFFITS (Difference in FITS, standardized) for a
#' fitted brma object. DFFITS measures how much the fitted value for observation
#' \eqn{i} changes if observation \eqn{i} is removed, standardized by the
#' estimated standard error of the fit.
#'
#' @param model a fitted normal-outcome \code{brma} object without a
#' weightfunction component.
#' @param ... additional arguments (currently ignored).
#'
#' @details
#' DFFITS values are computed as a PSIS leave-one-out deletion diagnostic. For
#' each observation \eqn{i}, the leave-one-out posterior mean fitted value at
#' that observation is estimated with normalized PSIS weights and compared to
#' the full-posterior fitted value:
#' \deqn{DFFITS_i =
#'   \frac{\hat{\mu}_i - \hat{\mu}_{i(-i)}}{SD_{(-i)}(\mu_i)}}
#'
#' This targets deletion influence on fitted values directly. It does not use
#' LOO-PIT residuals, which are predictive outlier diagnostics rather than
#' fitted-value deletion diagnostics.
#'
#' Estimate-unit LOO must first be computed with
#' \code{model <- add_loo(model, unit = "estimate")}. If the leave-one-out
#' posterior SD of a fitted value is near zero, the corresponding DFFITS value
#' is returned as \code{NA}.
#'
#' @return A named numeric vector of DFFITS values, one for each observation.
#'
#' @examples \dontrun{
#' if (requireNamespace("metadat", quietly = TRUE)) {
#'   data(dat.lehmann2018, package = "metadat")
#'   fit <- bPET(yi = yi, vi = vi, data = dat.lehmann2018, measure = "SMD")
#'   fit <- add_loo(fit)
#'
#'   dffits(fit)
#' }
#' }
#'
#' @seealso \code{\link{influence.brma}}, \code{\link{cooks.distance.brma}}, \code{\link{hatvalues.brma}}
#' @aliases dffits
#' @exportS3Method
dffits.brma <- function(model, ...) {

    # the function relies on the normal-normal hat matrix
    outcome_type      <- .outcome_type(model)
    is_weightfunction <- .is_weightfunction(model)

    if (outcome_type != "norm") {
      stop("dffits is only available for normal outcome models.", call. = FALSE)
    }
    if (is_weightfunction) {
      stop("dffits is not available for selection models (weightfunction).", call. = FALSE)
    }

    fit_samples <- .influence_fit_samples(model)
    weights     <- .diagnostic_psis_weights(model)
    dffits_vec  <- .dffits_internal(fit_samples, weights)
    dffits_vec  <- .diagnostic_set_names(dffits_vec, model)

    return(dffits_vec)
}

.dffits_internal <- function(fit_samples, weights) {

  summary <- .psis_fit_influence_summary(fit_samples, weights)

  delta <- summary[["full_fit"]] - diag(summary[["loo_fit"]])
  se    <- sqrt(diag(summary[["loo_var"]]))
  out   <- delta / se
  out[se <= sqrt(.Machine$double.eps)] <- NA_real_

  names(out) <- colnames(fit_samples)

  return(out)
}


# ---------------------------------------------------------------------------- #
# .influence_fit_samples
# ---------------------------------------------------------------------------- #
#
# Posterior fitted-value samples for fixed location terms used by PSIS deletion
# diagnostics.
#
# ---------------------------------------------------------------------------- #
.influence_fit_samples <- function(model) {

  data              <- model[["data"]]
  priors            <- model[["priors"]]
  posterior_samples <- .get_posterior_samples(model[["fit"]])
  K                 <- nrow(data[["outcome"]])

  fit_samples <- .evaluate.brma.mu(
    fit               = model[["fit"]],
    outcome_data      = data[["outcome"]],
    mods_data         = data[["mods"]],
    mods_formula      = if (.is_mods(model)) .create_fit_formula_list(data = data, "mods") else NULL,
    mods_priors       = priors[["mods"]],
    is_mods           = .is_mods(model),
    is_PET            = .is_PET(model),
    is_PEESE          = .is_PEESE(model),
    effect_direction  = .effect_direction(model),
    bias_adjusted     = FALSE,
    K                 = K,
    posterior_samples = posterior_samples
  )

  colnames(fit_samples) <- .get_estimate_labels(model)

  return(fit_samples)
}


# ---------------------------------------------------------------------------- #
# .psis_fit_influence_summary
# ---------------------------------------------------------------------------- #
#
# Full and PSIS leave-one-out fitted-value moments.
#
# ---------------------------------------------------------------------------- #
.psis_fit_influence_summary <- function(fit_samples, weights) {

  if (!is.matrix(fit_samples)) {
    fit_samples <- as.matrix(fit_samples)
  }
  if (!is.matrix(weights)) {
    weights <- as.matrix(weights)
  }
  if (nrow(fit_samples) != nrow(weights) ||
      ncol(fit_samples) != ncol(weights)) {
    stop("'fit_samples' and 'weights' must have the same dimensions.",
         call. = FALSE)
  }

  full_fit <- colMeans(fit_samples)
  loo_fit  <- crossprod(weights, fit_samples)
  loo_m2   <- crossprod(weights, fit_samples^2)
  loo_var  <- pmax(loo_m2 - loo_fit^2, 0)

  return(list(
    full_fit = full_fit,
    loo_fit  = loo_fit,
    loo_var  = loo_var
  ))
}

context("Influence diagnostics")

source(testthat::test_path("common-functions.R"))
source(testthat::test_path("helper-contracts.R"))
source(testthat::test_path("helper-test-matrix.R"))
source(testthat::test_path("helper-metafor.R"))

test_that("PSIS fitted-value influence helpers match direct calculations", {

  fit_samples <- matrix(
    c(
      0, 1, 2,
      1, 2, 3,
      2, 3, 4,
      3, 4, 5
    ),
    nrow  = 4,
    byrow = TRUE
  )
  colnames(fit_samples) <- paste0("study", 1:3)

  weights <- matrix(
    c(
      .4, .3, .2, .1,
      .1, .2, .3, .4,
      .25, .25, .25, .25
    ),
    nrow  = 4,
    byrow = FALSE
  )

  full_fit <- colMeans(fit_samples)
  loo_fit  <- crossprod(weights, fit_samples)
  loo_m2   <- crossprod(weights, fit_samples^2)
  loo_var  <- loo_m2 - loo_fit^2

  expected_dffits <- (full_fit - diag(loo_fit)) / sqrt(diag(loo_var))
  expected_cook_delta <- sweep(loo_fit, 2, full_fit, "-")
  expected_cook_delta <- -expected_cook_delta
  expected_cook <- rowSums(
    (expected_cook_delta %*% .symmetric_ginv(stats::cov(fit_samples))) *
      expected_cook_delta
  ) / 2

  summary <- .psis_fit_influence_summary(fit_samples, weights)
  expect_equal(summary[["full_fit"]], full_fit)
  expect_equal(summary[["loo_fit"]], loo_fit)
  expect_equal(summary[["loo_var"]], pmax(loo_var, 0))
  expect_equal(.dffits_internal(fit_samples, weights), expected_dffits)
  expect_equal(unname(.cooks.distance_internal(fit_samples, weights, P = 2)),
               expected_cook)
})

test_that("PSIS tau deletion helper aggregates remaining scale rows", {

  tau_samples <- matrix(
    c(
      1, 3, 5,
      2, 4, 6,
      3, 5, 7,
      4, 6, 8
    ),
    nrow  = 4,
    byrow = TRUE
  )
  weights <- matrix(
    c(
      .4, .3, .2, .1,
      .1, .2, .3, .4,
      .25, .25, .25, .25
    ),
    nrow  = 4,
    byrow = FALSE
  )

  deleted_tau <- cbind(
    rowMeans(tau_samples[, -1, drop = FALSE]),
    rowMeans(tau_samples[, -2, drop = FALSE]),
    rowMeans(tau_samples[, -3, drop = FALSE])
  )
  expected <- colSums(weights * deleted_tau)

  expect_equal(.influence_tau_del_from_samples(tau_samples, weights), expected)
})

skip_if_no_fits()
skip_if_not_installed("metafor")

fit_names <- list_fits()
fits      <- lazy_fits(fit_names, validate = FALSE)
info      <- lazy_infos(fit_names, validate = FALSE)

for_each_case(influence_metafor_cases(), function(case) {
  test_that_case("Influence diagnostics match metafor", case, {
    expect_influence_matches_metafor(case)
  })
})

test_that("Influence tau.del for location-scale models uses aggregate deleted tau", {

  name <- "bangertdrowns2004_location-scale"
  skip_if_missing_fits(name)

  fit_brma <- fits[[name]]

  tau_result <- .evaluate.brma.tau(
    fit           = fit_brma[["fit"]],
    scale_data    = fit_brma[["data"]][["scale"]],
    scale_formula = .create_fit_formula_list(data = fit_brma[["data"]], "scale"),
    scale_priors  = fit_brma[["priors"]][["scale"]],
    is_scale      = TRUE,
    is_multilevel = FALSE,
    K             = nrow(fit_brma[["data"]][["outcome"]])
  )
  tau_samples <- tau_result[["tau_total"]]
  weights     <- loo_weights(fit_brma)
  deleted_tau <- (rowSums(tau_samples) - tau_samples) / (ncol(tau_samples) - 1L)
  expected    <- colSums(weights * deleted_tau)

  inf_brma <- suppressWarnings(influence(fit_brma))
  scale_loo_coef_1 <- suppressWarnings(
    dfbetas(fit_brma, type = "scale", return_loo_estimates = TRUE)[[1]]
  )

  expect_equal(unname(inf_brma[["inf"]][["tau.del"]]), unname(expected),
               tolerance = 1e-12)
  expect_false(isTRUE(all.equal(unname(inf_brma[["inf"]][["tau.del"]]),
                                unname(scale_loo_coef_1))))
})

test_that("Influence stats for selection model are available", {

  name <- "dat.lehmann2018-3PSM"
  skip_if_missing_fits(name)

  inf_brma <- suppressWarnings(influence(fits[[name]]))
  expect_true(!is.null(inf_brma))
})

test_that("Influence diagnostics and scalar outputs use study labels", {

  name <- "bcg_meta-analysis"
  skip_if_missing_fits(name)

  fit_brma <- fits[[name]]
  labels   <- RoBMA:::.diagnostic_study_labels(fit_brma)
  inf_brma <- suppressWarnings(influence(fit_brma))

  expect_equal(rownames(inf_brma[["inf"]]), labels)
  expect_equal(rownames(inf_brma[["dfbs"]]), labels)
  expect_equal(names(covratio(fit_brma)), labels)
  expect_equal(names(dffits(fit_brma)), labels)
  expect_equal(names(cooks.distance(fit_brma)), labels)
})

test_that("DFFITS rejects unsupported model families", {

  model_names <- c(
    "bcg_glmm",
    "bcg_BMA.glmm",
    "dat.lehmann2018-3PSM",
    "dat.lehmann2018_RoBMA"
  )
  skip_if_missing_fits(model_names)

  expect_error(dffits(fits[["bcg_glmm"]]), "only available for normal outcome models")
  expect_error(dffits(fits[["bcg_BMA.glmm"]]), "only available for normal outcome models")
  expect_error(dffits(fits[["dat.lehmann2018-3PSM"]]), "not available for selection models")
  expect_error(dffits(fits[["dat.lehmann2018_RoBMA"]]), "not available for selection models")
})

test_that("Influence stats for model-averaging fits are internally consistent", {

  cases <- data.frame(
    name = c("dat.lehmann2018_BMA.norm", "bcg_BMA.glmm", "dat.lehmann2018_RoBMA"),
    unsupported_cook = c(NA, "normal outcome models", "selection models"),
    stringsAsFactors = FALSE
  )
  cases[["inf_cols"]] <- I(list(
    c("rstudent", "dffits", "cook.d", "cov.r", "tau.del", "hat"),
    c("rstudent", "cov.r", "tau.del"),
    c("rstudent", "cov.r", "tau.del")
  ))
  skip_if_missing_fits(cases[["name"]])

  for (i in seq_len(nrow(cases))) {
    name     <- cases[["name"]][[i]]
    fit_brma <- fits[[name]]
    inf_brma <- suppressWarnings(influence(fit_brma))

    expect_influence_object(inf_brma, nobs(fit_brma), cases[["inf_cols"]][[i]],
                            info = name)

    unsupported_cook <- cases[["unsupported_cook"]][[i]]
    if (!is.na(unsupported_cook)) {
      expect_false(any(c("dffits", "cook.d", "hat") %in% names(inf_brma[["inf"]])))
      expect_error(cooks.distance(fit_brma), unsupported_cook)
    }
  }
})

library(testthat)
library(RoBMA)

on_cran <- get("on_cran", envir = asNamespace("testthat"), inherits = FALSE)

if (on_cran()) {
  test_check("RoBMA", filter = "cran-smoke")
} else {
  test_check("RoBMA")
}

#' @title Predict From brma Object
#'
#' @description \code{predict.brma} predicts values
#'
#' @inheritParams summary.brma
#' @param probs quantiles of the posterior samples to be displayed when the
#' returned `brma_samples` object is printed. Defaults to `c(.025, .975)`.
#' @param newdata specification for prediction data. Defaults to \code{NULL} which
#' corresponds to prediction for the observed data. Alternatives are:
#' \itemize{
#'   \item{\code{TRUE} returns a single aggregated prediction (either the scalar
#'   parameter for models without moderators/scale, or the average across the
#'   model matrix for regression models). Not available for \code{type = "response"}
#'   since observation-level sampling variance is required.}
#'   \item{A data.frame (for meta-regression) or a named list with effect size
#'   measure and variability metrics (for meta-analysis) for new studies. The input
#'   must contain the variables used by the requested prediction type. Outcome
#'   columns are optional for \code{type = "terms"} unless PET/PEESE bias terms
#'   are included via \code{bias_adjusted = FALSE}.}
#' }
#' @param type type of prediction to be performed. Options are:
#' \itemize{
#'   \item{\code{"terms"} (alias: \code{"marginal"}): Fixed-effect parameters only (mu).
#'   Produces the mean effect size estimate at the given predictor levels,
#'   not accounting for random effects.}
#'   \item{\code{"cluster"}: Fixed effects plus cluster-level random effects (mu + gamma).
#'   Only available for multilevel (3-level) models.}
#'   \item{\code{"estimate"} (aliases: \code{"effect"}, \code{"blup"}): True
#'   effects (mu + gamma + theta). For existing normal data, returns posterior
#'   draws of conditional BLUP means, not simulated latent-effect draws. For
#'   new data, samples estimate-level random effects.}
#'   \item{\code{"response"} (alias: \code{"outcome"}): Predicted observed values (yi).
#'   Incorporates both heterogeneity and sampling variability.}
#'   \item{\code{"terms.scale"}: Scale parameter (tau), incorporating scale
#'   regression if present.}
#' }
#' @param as_measure logical; whether to convert GLMM response predictions from
#' simulated counts to continuity-corrected effect-size estimators (logOR for
#' binomial, logIRR for Poisson). Defaults to \code{TRUE}. Only relevant for
#' GLMM models with \code{type = "response"}. When \code{FALSE}, returns raw
#' frequency data (counts). Use \code{type = "estimate"} for latent logOR/logIRR
#' predictions.
#' @param output_measure effect-size measure for location/effect predictions.
#' Defaults to the fitted measure. Supported conversions are among \code{"SMD"},
#' \code{"COR"}, \code{"ZCOR"}, and \code{"OR"}; \code{"RR"}, \code{"HR"},
#' \code{"IRR"}, \code{"RD"}, and \code{"GEN"} can only be returned on their
#' fitted measure. Use \code{transform = "EXP"} for ratio-scale output from
#' log-scale measures.
#' @param transform optional display transformation. Currently \code{"EXP"}
#' exponentiates log-scale measures \code{"OR"}, \code{"RR"}, \code{"HR"},
#' and \code{"IRR"}.
#' @param bias_adjusted whether predictions should adjust for publication bias.
#' Defaults to \code{FALSE}. When \code{TRUE}:
#' \itemize{
#'   \item{PET/PEESE terms are NOT added to the mean parameter (mu), returning
#'   the bias-corrected effect estimate.}
#'   \item{For \code{type = "response"} with selection models, samples from
#'   the ordinary normal predictive distribution instead of the selected-normal
#'   distribution, simulating what would be observed without publication bias.}
#' }
#' When \code{FALSE}:
#' \itemize{
#'   \item{PET/PEESE terms ARE added to mu, returning predictions that include
#'   the expected bias (i.e., what we expect to observe given publication bias).}
#'   \item{For \code{type = "response"} with selection models, samples from
#'   the selected-normal distribution reflecting the selective publishing
#'   process.}
#' }
#' @param conditional whether to return conditional posterior predictions for
#' RoBMA product-space objects. For location predictions, samples are conditioned
#' on the effect component; for \code{type = "terms.scale"}, samples are
#' conditioned on the heterogeneity component. Conditional samples are flattened
#' to one chain after subsetting posterior rows.
#' @param quiet logical; whether to suppress informational messages about
#' prediction scale and bias adjustment.
#'
#' @details
#' \strong{Type hierarchy:}
#' \itemize{
#'   \item{\code{"terms"}: mu (fixed effects only)}
#'   \item{\code{"cluster"}: mu + gamma (adds cluster-level random effect)}
#'   \item{\code{"estimate"}: mu + gamma + theta (adds estimate-level random effect)}
#'   \item{\code{"response"}: mu + gamma + theta + epsilon (adds sampling error)}
#' }
#' For existing normal observations, \code{type = "estimate"} reports the
#' conditional BLUP mean \eqn{E(\theta_i \mid y_i, \mu_i, \tau_i)} for each
#' posterior draw. It is therefore an empirical-Bayes summary, not a draw from
#' the full latent-effect posterior \eqn{\theta_i \mid y_i}. For new data,
#' estimate-level random effects are sampled from their model distribution.
#'
#' For RoBMA product-space objects, conditional posterior predictions subset
#' posterior rows according to model indicators. This removes the original
#' chain structure, so returned \code{brma_samples} objects are intentionally
#' stored as one flattened chain.
#'
#' Note that in contrast to \link[metafor]{predict}, the \code{type = "response"} produces
#' predictions for the new effect size estimates. To obtain results corresponding to
#' metafor's predict function, use \code{type = "terms"} for the mean effect size
#' and \code{type = "estimate"} for true effects (prediction interval).
#'
#' \strong{Likelihood weights:} If the model was fitted with \code{weights},
#' the weights affect the posterior fit, log-likelihood, LOO/WAIC, and
#' existing-data conditional diagnostics such as BLUP shrinkage and leverage.
#' They do not change the observation-level sampling error used by
#' \code{type = "response"} for normal models: response predictions simulate
#' raw future effect-size estimates with the supplied \code{sei}, not
#' \code{sei / sqrt(weight)}.
#'
#' @examples \dontrun{
#' if (requireNamespace("metadat", quietly = TRUE) &&
#'     requireNamespace("metafor", quietly = TRUE)) {
#'   data(dat.bcg, package = "metadat")
#'   dat <- metafor::escalc(
#'     measure = "RR",
#'     ai      = tpos,
#'     bi      = tneg,
#'     ci      = cpos,
#'     di      = cneg,
#'     data    = dat.bcg
#'   )
#'
#'   fit <- brma(
#'     yi      = yi,
#'     vi      = vi,
#'     mods    = ~ ablat + year,
#'     data    = dat,
#'     measure = "RR",
#'     seed    = 1,
#'     silent  = TRUE
#'   )
#'
#'   predict(fit, type = "terms")
#'   predict(fit, newdata = TRUE, type = "terms")
#'   predict(fit, type = "estimate")
#' }
#' }
#'
#' @return A \code{brma_samples} object containing posterior samples. When printed,
#' displays a summary table via \code{BayesTools::ensemble_estimates_table}. The
#' underlying samples matrix can be accessed directly (the object inherits from matrix)
#' or via \code{summary()} to obtain the summary table. The samples can also be converted
#' to \pkg{posterior} draws formats using \code{as_draws()} and related functions.
#' @seealso [pooled_effect()], [pooled_heterogeneity()], [blup()]
#' @export
predict.brma <- function(object, newdata = NULL,
                         type = "terms",
                         as_measure = TRUE,
                         output_measure = NULL,
                         transform = NULL,
                         probs = c(.025, .975),
                         bias_adjusted = FALSE,
                         quiet = FALSE,
                         conditional = FALSE,
                         ...){

  dots <- list(...)

  # normalize type aliases
  type <- match.arg(type, c("terms", "marginal", "cluster", "estimate", "effect", "blup",
                            "response", "outcome", "terms.scale"))
  type <- switch(type,
    "marginal" = "terms",
    "effect"   = "estimate",
    "blup"     = "estimate",
    "outcome"  = "response",
    type  # default: keep as is
  )

  # input validation
  BayesTools::check_bool(as_measure, "as_measure")
  BayesTools::check_bool(bias_adjusted, "bias_adjusted")
  BayesTools::check_bool(conditional, "conditional")
  BayesTools::check_bool(quiet, "quiet")
  if (conditional && !.is_RoBMA(object)) {
    stop("'conditional' predictions are available only for RoBMA objects.",
         call. = FALSE)
  }

  # check newdata: NULL, TRUE, or data.frame/list
  if (!is.null(newdata) && !isTRUE(newdata)) {
    if (!is.data.frame(newdata) && !is.list(newdata)) {
      stop("'newdata' must be NULL, TRUE, a data.frame, or a named list.", call. = FALSE)
    }
    if (is.list(newdata) && !is.data.frame(newdata) && is.null(names(newdata))) {
      stop("'newdata' list must be named.", call. = FALSE)
    }
  }

  # check incompatible options: aggregate predictions not available for response
  if (isTRUE(newdata) && type == "response") {
    stop("Aggregated predictions (newdata = TRUE) are not available for type = 'response' ",
         "because observation-level sampling variance is required.", call. = FALSE)
  }

  ### types of predictions
  # terms:       fixed effects terms for the overall effect (mu) / incorporating mods if present
  # cluster:     terms + cluster-level random effects (mu + gamma) - multilevel only
  # terms.scale: fixed effects terms for the overall heterogeneity (tau) / incorporating scale if present
  # estimate:    incorporating heterogeneity into terms to obtain the true effects
  #              (via empirical Bayes for existing data, new random effect sampled for new data)
  # response:    incorporating heterogeneity and sampling variability

  ### dispatch between prediction on the current data vs. new data
  if (is.null(newdata)) {

    # existing data are used
    same_data <- TRUE
    aggregate <- FALSE
    new_data  <- object[["data"]]

  } else if (isTRUE(newdata)) {

    # aggregated prediction: use original data but marginalize random effects
    # and average across observations (for mods/scale models)
    same_data <- FALSE
    aggregate <- TRUE
    new_data  <- object[["data"]]

  } else {

    # prepare newdata using the same settings as the original fit
    same_data <- FALSE
    aggregate <- FALSE
    new_data  <- .prepare_newdata(
      object        = object,
      newdata       = newdata,
      type          = type,
      bias_adjusted = bias_adjusted
    )

  }

  ### extract priors and structural information about the model
  priors            <- object[["priors"]]
  is_mods           <- .is_mods(object)
  is_multilevel     <- .is_multilevel(object)
  is_scale          <- .is_scale(object)
  is_PET            <- .is_PET(object)
  is_PEESE          <- .is_PEESE(object)
  is_weightfunction <- .is_weightfunction(object)
  is_weights        <- .is_weights(object)
  outcome_type      <- .outcome_type(object)
  effect_direction  <- .effect_direction(object)
  posterior_samples <- .get_posterior_samples(object[["fit"]], dots[[".posterior_samples"]])
  effect_transform  <- .effect_output_setup(
    object         = object,
    output_measure = output_measure,
    transform      = transform
  )

  if (type == "terms.scale" && .effect_output_requested(effect_transform)) {
    stop(
      "'output_measure' and 'transform' are only available for effect-size ",
      "predictions, not for type = 'terms.scale'.",
      call. = FALSE
    )
  }
  if (type == "response" && !isTRUE(as_measure) &&
      .effect_output_requested(effect_transform)) {
    stop(
      "'output_measure' and 'transform' require type = 'response' predictions ",
      "to be returned as effect-size measures (as_measure = TRUE).",
      call. = FALSE
    )
  }

  # check: cluster type requires multilevel model
  if (type == "cluster" && !is_multilevel) {
    stop("type = 'cluster' is only available for multilevel (3-level) models. ",
         "Use type = 'terms' for non-multilevel models.", call. = FALSE)
  }

  ### extract outcome data and fit data for convenience
  outcome_data      <- new_data[["outcome"]]
  fit_data          <- .create_fit_data(data = new_data, priors = priors)

  # outcome dimensions
  K_original <- nrow(outcome_data)
  K          <- K_original

  ### obtain tau samples using helper function
  # returns list(tau_within, tau_between) - all S x K matrices
  # see .evaluate.brma.tau() in evaluate.R for details
  tau_result          <- .evaluate.brma.tau(
    fit               = object[["fit"]],
    scale_data        = new_data[["scale"]],
    scale_formula     = if (is_scale) .create_fit_formula_list(data = new_data, "scale") else NULL,
    scale_priors      = priors[["scale"]],
    is_scale          = is_scale,
    is_multilevel     = is_multilevel,
    K                 = K_original,
    posterior_samples = posterior_samples
  )
  tau_within_samples  <- tau_result[["tau_within"]]
  tau_between_samples <- tau_result[["tau_between"]]

  ### aggregate tau samples if requested (for terms.scale only at this point)
  ### mu aggregation happens after mu computation
  if (aggregate && type == "terms.scale") {
    # average tau across observations (rows are samples, columns are observations)
    # for non-scale models, all columns are identical so this is a no-op
    # for scale models, this averages across the model matrix
    if (is_scale && K_original > 1 && !quiet) {
      message("Aggregated prediction averages tau across the scale model matrix (K = ", K_original, " observations).")
    }
    tau_within_samples  <- matrix(rowMeans(tau_within_samples), ncol = 1)
    tau_between_samples <- matrix(rowMeans(tau_between_samples), ncol = 1)
    K <- 1L
  }

  ### extract MCMC chain info for brma_samples construction
  chain_info <- .brma_samples_chain_info(
    fit       = object[["fit"]],
    n_samples = nrow(posterior_samples)
  )
  n_chains <- chain_info[["n_chains"]]
  n_iter   <- chain_info[["n_iter"]]

  ### return only tau samples if type = "terms.scale" is selected
  if (type == "terms.scale") {
    # reconstruct total tau from components: tau = sqrt(tau_within^2 + tau_between^2)
    tau_samples <- sqrt(tau_within_samples^2 + tau_between_samples^2)

    # rename samples
    colnames(tau_samples) <- if (aggregate) "tau" else paste0("tau[", seq_len(K), "]")

    out <- .new_brma_samples(
      samples  = tau_samples,
      n_chains = n_chains,
      n_iter   = n_iter,
      title    = if (aggregate) "Aggregated Scale Term Posterior Prediction:" else "Scale Term Posterior Prediction:",
      probs    = probs,
      data     = if (aggregate) NULL else new_data
    )
    return(.condition_prediction_samples(
      object            = object,
      samples           = out,
      conditional       = conditional,
      parameters        = .conditional_heterogeneity_parameters(object),
      posterior_samples = posterior_samples,
      quiet             = quiet
    ))
  }

  ### get the base mu samples using helper function
  # returns S x K matrix of location samples
  # see .evaluate.brma.mu() in evaluate.R for details
  mu_samples <- .evaluate.brma.mu(
    fit               = object[["fit"]],
    outcome_data      = outcome_data,
    mods_data         = new_data[["mods"]],
    mods_formula      = if (is_mods) .create_fit_formula_list(data = new_data, "mods") else NULL,
    mods_priors       = priors[["mods"]],
    is_mods           = is_mods,
    is_PET            = is_PET,
    is_PEESE          = is_PEESE,
    effect_direction  = effect_direction,
    bias_adjusted     = bias_adjusted,
    K                 = K_original,
    posterior_samples = posterior_samples
  )

  ### aggregate mu and tau samples if requested (for terms/cluster/estimate types)
  if (aggregate && type %in% c("terms", "cluster", "estimate")) {
    # average mu across observations (rows are samples, columns are observations)
    # for non-mods models, all columns are identical so this is a no-op
    # for mods models, this averages across the model matrix
    if (is_mods && K_original > 1 && !quiet) {
      message("Aggregated prediction averages mu across the moderator model matrix (K = ", K_original, " observations).")
    }
    mu_samples          <- matrix(rowMeans(mu_samples), ncol = 1)
    tau_within_samples  <- matrix(rowMeans(tau_within_samples), ncol = 1)
    tau_between_samples <- matrix(rowMeans(tau_between_samples), ncol = 1)
    K <- 1L
  }

  ### return only mu samples if type = "terms" is selected
  # terms incorporate fixed effects only (i.e., random effects are not incorporated)
  if (type == "terms") {
    # rename samples
    colnames(mu_samples) <- if (aggregate) "mu" else paste0("mu[", seq_len(K), "]")

    out <- .new_effect_brma_samples(
      samples          = mu_samples,
      n_chains         = n_chains,
      n_iter           = n_iter,
      title            = if (aggregate) "Aggregated Location Term Posterior Prediction:" else "Location Term Posterior Prediction:",
      probs            = probs,
      data             = if (aggregate) NULL else new_data,
      effect_transform = effect_transform
    )
    return(.condition_prediction_samples(
      object            = object,
      samples           = out,
      conditional       = conditional,
      parameters        = .conditional_effect_parameters(object),
      posterior_samples = posterior_samples,
      quiet             = quiet
    ))
  }

  ### include 3-level (cluster-level) random-effects using helper function
  # returns contribution matrix gamma[cluster] * tau_between for multilevel models
  # see .evaluate.brma.cluster_effects() in evaluate.R for details
  # for aggregated predictions (same_data = FALSE, K = 1), function samples new gamma ~ N(0,1)
  blup_vi <- NULL
  if (outcome_type == "norm") {
    blup_vi <- outcome_data[["sei"]]^2
    if (is_weights) {
      blup_vi <- blup_vi / outcome_data[["weights"]]
    }
  }

  blup_bias_offset <- NULL
  if (outcome_type == "norm" && same_data && bias_adjusted && (is_PET || is_PEESE)) {
    blup_bias_offset <- .evaluate.brma.bias_offset(
      fit               = object[["fit"]],
      outcome_data      = outcome_data,
      is_PET            = is_PET,
      is_PEESE          = is_PEESE,
      effect_direction  = effect_direction,
      K                 = K,
      posterior_samples = posterior_samples
    )
  }

  # Exact block BLUP applies when the Gaussian residual model matches the target.
  # For weightfunction models, gamma is sampled under the selected-data
  # likelihood; keep those draws and apply estimate-level shrinkage conditional
  # on them instead of replacing them by the ordinary Gaussian block shortcut.
  multilevel_blup     <- NULL
  use_multilevel_blup <- is_multilevel && outcome_type == "norm" && same_data &&
    !is_weightfunction

  if (use_multilevel_blup) {
    multilevel_blup <- .evaluate.brma.multilevel_blup.norm(
      mu_samples  = mu_samples,
      tau_within  = tau_within_samples,
      tau_between = tau_between_samples,
      yi          = outcome_data[["yi"]],
      vi          = blup_vi,
      cluster     = fit_data[["cluster"]],
      bias_offset = blup_bias_offset
    )
  }

  if (is_multilevel) {
    if (!is.null(multilevel_blup)) {
      cluster_contribution <- multilevel_blup[["cluster"]]
    } else {
      cluster <- if (aggregate) seq_len(K) else fit_data[["cluster"]]
      cluster_contribution <- .evaluate.brma.cluster_effects(
        fit               = object[["fit"]],
        tau_between       = tau_between_samples,
        cluster           = cluster,
        same_data         = same_data,
        effect_direction  = effect_direction,
        posterior_samples = posterior_samples
      )
    }
    mu_samples <- mu_samples + cluster_contribution
  }

  ### return cluster-level predictions if type = "cluster" is selected
  # cluster incorporates fixed effects + cluster-level random effects (mu + gamma)
  if (type == "cluster") {
    # rename samples
    colnames(mu_samples) <- if (aggregate) "mu_cluster" else paste0("mu_cluster[", seq_len(K), "]")

    out <- .new_effect_brma_samples(
      samples          = mu_samples,
      n_chains         = n_chains,
      n_iter           = n_iter,
      title            = if (aggregate) "Aggregated Cluster-Level Posterior Prediction:" else "Cluster-Level Posterior Prediction:",
      probs            = probs,
      data             = if (aggregate) NULL else new_data,
      effect_transform = effect_transform
    )
    return(.condition_prediction_samples(
      object            = object,
      samples           = out,
      conditional       = conditional,
      parameters        = .conditional_effect_parameters(object),
      posterior_samples = posterior_samples,
      quiet             = quiet
    ))
  }

  ### create true-effect prediction
  # dispatches between normal and GLMM approaches
  # both helpers handle same_data dispatch internally:
  # - same_data = TRUE: use fitted values (BLUPs for normal, extracted theta for GLMM)
  # - same_data = FALSE: sample from marginal distribution N(mu, tau_within)
  if (type == "estimate") {

    if (!is.null(multilevel_blup)) {
      true_effects_samples <- mu_samples + multilevel_blup[["estimate"]]
    } else if (outcome_type == "norm") {
      true_effects_samples <- .evaluate.brma.true_effects.norm(
        mu_samples  = mu_samples,
        tau_within  = tau_within_samples,
        yi          = outcome_data[["yi"]],
        sei         = sqrt(blup_vi),
        same_data   = same_data,
        bias_offset = blup_bias_offset
      )
    } else {
      true_effects_samples <- .evaluate.brma.true_effects.glmm(
        fit               = object[["fit"]],
        mu_samples        = mu_samples,
        tau_within        = tau_within_samples,
        same_data         = same_data,
        K                 = K,
        posterior_samples = posterior_samples
      )
    }

    # rename samples
    colnames(true_effects_samples) <- if (aggregate) "theta" else paste0("theta[", seq_len(K), "]")

    out <- .new_effect_brma_samples(
      samples          = true_effects_samples,
      n_chains         = n_chains,
      n_iter           = n_iter,
      title            = if (same_data && outcome_type == "norm") {
        "True Effects (BLUP Means):"
      } else if (aggregate) {
        "Aggregated True Effect Posterior Prediction:"
      } else {
        "True Effect Posterior Prediction:"
      },
      probs            = probs,
      data             = if (aggregate) NULL else new_data,
      effect_transform = effect_transform
    )
    return(.condition_prediction_samples(
      object            = object,
      samples           = out,
      conditional       = conditional,
      parameters        = .conditional_effect_parameters(object),
      posterior_samples = posterior_samples,
      quiet             = quiet
    ))

  }

  ### create observed effects prediction
  if (type == "response") {

    # different model types have different output structures
    if (is.element(outcome_type, c("bin", "pois"))) {

      if (as_measure) {
        .check_glmm_response_as_measure(outcome_type, outcome_data)
      }

      # the estimate-level effects are not marginalized for GLMMs
      # include the sampled random effects or marginalize over them for new data
      # see .evaluate.brma.theta.glmm() in evaluate.R for details
      theta_contribution <- .evaluate.brma.theta.glmm(
        fit               = object[["fit"]],
        tau_within        = tau_within_samples,
        same_data         = same_data,
        K                 = K,
        posterior_samples = posterior_samples
      )
      mu_samples <- mu_samples + theta_contribution

      if (outcome_type == "bin") {

        ### incorporate with base-rate using helper function
        # Existing data use fitted pi[k]. Newdata samples new pi from the
        # estimate-specific prior; reusing fitted pi[k] would index old studies.
        if (same_data) {
          if (K != nrow(object[["data"]][["outcome"]])) {
            stop("Existing-data GLMM response prediction has inconsistent K.",
                 call. = FALSE)
          }
          logit_baserate <- .evaluate.brma.baserate(
            fit               = object[["fit"]],
            K                 = K,
            posterior_samples = posterior_samples
          )
        } else {
          logit_baserate <- .evaluate.brma.baserate_newdata(
            prior_pi = priors[["outcome"]][["pi"]],
            S        = nrow(mu_samples),
            K        = K
          )
        }

        ### sample outcome using RNG helper
        outcome_samples <- .outcome_rng.binom(
          mu_samples     = mu_samples,
          logit_baserate = logit_baserate,
          n1i            = outcome_data[["n1i"]],
          n2i            = outcome_data[["n2i"]]
        )

      } else if (outcome_type == "pois") {

        ### incorporate with log-rate using helper function
        # Existing data use fitted phi[k]. Newdata samples new phi from the
        # estimate-specific prior; reusing fitted phi[k] would index old studies.
        if (same_data) {
          if (K != nrow(object[["data"]][["outcome"]])) {
            stop("Existing-data GLMM response prediction has inconsistent K.",
                 call. = FALSE)
          }
          log_phi <- .evaluate.brma.lograte(
            fit               = object[["fit"]],
            K                 = K,
            posterior_samples = posterior_samples
          )
        } else {
          log_phi <- .evaluate.brma.lograte_newdata(
            prior_phi = priors[["outcome"]][["phi"]],
            S         = nrow(mu_samples),
            K         = K
          )
        }

        ### sample outcome using RNG helper
        outcome_samples <- .outcome_rng.pois(
          mu_samples = mu_samples,
          log_phi    = log_phi,
          t1i        = outcome_data[["t1i"]],
          t2i        = outcome_data[["t2i"]]
        )

      }

      # convert to effect size measure if requested (default)
      if (as_measure) {
        outcome_samples <- .glmm_response_counts_to_measure(
          outcome_samples = outcome_samples,
          outcome_data    = outcome_data,
          outcome_type    = outcome_type
        )

      }


    } else if (outcome_type == "norm") {

      # normal outcome models dispatch between:
      # - bias_adjusted = TRUE: sample from unweighted normal (as if no bias)
      # - bias_adjusted = FALSE with weightfunction: sample from selected-normal
      # - bias_adjusted = FALSE without weightfunction: sample from unweighted normal

      if (bias_adjusted || !is_weightfunction) {

        # sample from unweighted normal distribution
        # y[s,k] ~ N(mu[s,k], sqrt(tau_within[s,k]^2 + sei[k]^2))
        outcome_samples <- .outcome_rng.norm(
          mu_samples = mu_samples,
          tau_within = tau_within_samples,
          sei        = outcome_data[["sei"]]
        )

      } else {

        selection_context <- .selection_context(
          object            = object,
          posterior_samples = posterior_samples,
          newdata           = new_data
        )

        outcome_samples <- .outcome_rng.selnorm(
          mu_samples        = mu_samples,
          tau_within        = tau_within_samples,
          sei               = outcome_data[["sei"]],
          selection_context = selection_context
        )

      }

      # rename samples
      colnames(outcome_samples) <- paste0("yi[", seq_len(K), "]")
    }

    out <- .new_effect_brma_samples(
      samples          = outcome_samples,
      n_chains         = n_chains,
      n_iter           = n_iter,
      title            = .response_prediction_title(outcome_type, as_measure),
      probs            = probs,
      data             = new_data,
      effect_transform = effect_transform
    )
    return(.condition_prediction_samples(
      object            = object,
      samples           = out,
      conditional       = conditional,
      parameters        = .conditional_effect_parameters(object),
      posterior_samples = posterior_samples,
      quiet             = quiet
    ))
  }
}


# ---------------------------------------------------------------------------- #
# .check_glmm_response_as_measure
# ---------------------------------------------------------------------------- #
#
# Check whether simulated GLMM counts can be converted to effect-size estimator
# draws. Raw count predictions can allow zero exposure/person-time, but the
# derived logOR/logIRR estimator cannot.
#
# ---------------------------------------------------------------------------- #
.check_glmm_response_as_measure <- function(outcome_type, outcome_data) {

  if (outcome_type == "bin" &&
      (any(outcome_data[["n1i"]] <= 0, na.rm = TRUE) ||
       any(outcome_data[["n2i"]] <= 0, na.rm = TRUE))) {
    stop(
      "GLMM response predictions with as_measure = TRUE require positive ",
      "'n1i' and 'n2i'. Use as_measure = FALSE for raw count predictions ",
      "or type = 'estimate' for latent logOR predictions.",
      call. = FALSE
    )
  }

  if (outcome_type == "pois" &&
      (any(outcome_data[["t1i"]] <= 0, na.rm = TRUE) ||
       any(outcome_data[["t2i"]] <= 0, na.rm = TRUE))) {
    stop(
      "Poisson response predictions with as_measure = TRUE require positive ",
      "'t1i' and 't2i'. Use as_measure = FALSE for raw count predictions ",
      "or type = 'estimate' for latent logIRR predictions.",
      call. = FALSE
    )
  }

  return(invisible(TRUE))
}


# ---------------------------------------------------------------------------- #
# .glmm_response_counts_to_measure
# ---------------------------------------------------------------------------- #
#
# Convert simulated GLMM response counts to escalc-style effect-size estimator
# draws. This is not the latent effect posterior; it is the sampling distribution
# of the continuity-corrected estimator applied to posterior predictive counts.
#
# ---------------------------------------------------------------------------- #
.glmm_response_counts_to_measure <- function(outcome_samples, outcome_data,
                                             outcome_type) {

  K <- nrow(outcome_data)

  if (outcome_type == "bin") {
    ai <- outcome_samples[, seq(1, 2 * K, by = 2), drop = FALSE]
    ci <- outcome_samples[, seq(2, 2 * K, by = 2), drop = FALSE]

    n1i_mat <- matrix(outcome_data[["n1i"]], nrow = nrow(ai), ncol = K, byrow = TRUE)
    n2i_mat <- matrix(outcome_data[["n2i"]], nrow = nrow(ai), ncol = K, byrow = TRUE)
    bi      <- n1i_mat - ai
    di      <- n2i_mat - ci
    zero    <- ai == 0 | bi == 0 | ci == 0 | di == 0

    ai_adj <- ai + 0.5 * zero
    bi_adj <- bi + 0.5 * zero
    ci_adj <- ci + 0.5 * zero
    di_adj <- di + 0.5 * zero

    out <- log((ai_adj * di_adj) / (bi_adj * ci_adj))

  } else if (outcome_type == "pois") {
    x1i <- outcome_samples[, seq(1, 2 * K, by = 2), drop = FALSE]
    x2i <- outcome_samples[, seq(2, 2 * K, by = 2), drop = FALSE]

    t1i_mat <- matrix(outcome_data[["t1i"]], nrow = nrow(x1i), ncol = K, byrow = TRUE)
    t2i_mat <- matrix(outcome_data[["t2i"]], nrow = nrow(x1i), ncol = K, byrow = TRUE)
    zero    <- x1i == 0 | x2i == 0

    x1i_adj <- x1i + 0.5 * zero
    x2i_adj <- x2i + 0.5 * zero

    out <- log((x1i_adj / t1i_mat) / (x2i_adj / t2i_mat))

  } else {
    stop("Unsupported GLMM response outcome type.", call. = FALSE)
  }

  colnames(out) <- paste0("yi[", seq_len(K), "]")

  return(out)
}


# ---------------------------------------------------------------------------- #
# .response_prediction_title
# ---------------------------------------------------------------------------- #
#
# Human-readable prediction title matching the response estimand.
#
# ---------------------------------------------------------------------------- #
.response_prediction_title <- function(outcome_type, as_measure) {

  if (is.element(outcome_type, c("bin", "pois"))) {
    if (as_measure) {
      return("Observed Effect-Size Estimator Posterior Prediction:")
    }
    return("Observed Counts Posterior Prediction:")
  }

  return("Observations Posterior Prediction:")
}

.condition_prediction_samples <- function(object, samples, conditional,
                                          parameters, posterior_samples = NULL,
                                          quiet = FALSE) {

  if (!conditional) {
    return(samples)
  }

  keep <- .conditional_parameter_rows(
    object            = object,
    parameters        = parameters,
    posterior_samples = posterior_samples,
    rule              = "OR"
  )
  sample_matrix <- as.matrix(samples)[keep, , drop = FALSE]

  if (nrow(sample_matrix) == 0L) {
    stop("No posterior samples remain after conditioning.", call. = FALSE)
  }

  if (!quiet) {
    message("Conditional posterior samples flattened to one chain.")
  }

  return(.new_brma_samples(
    samples          = sample_matrix,
    n_chains         = 1L,
    n_iter           = nrow(sample_matrix),
    title            = paste("Conditional", attr(samples, "title")),
    probs            = attr(samples, "probs"),
    data             = attr(samples, "data"),
    effect_transform = attr(samples, "effect_transform")
  ))
}

.conditional_effect_parameters <- function(object) {

  prior_names <- names(attr(object[["fit"]], "prior_list"))

  if ("mu" %in% prior_names) {
    return("mu")
  }
  if ("mu_intercept" %in% prior_names) {
    return("mu_intercept")
  }

  parameters <- grep("^mu_", prior_names, value = TRUE)
  if (length(parameters) > 0L) {
    return(parameters)
  }

  stop("No location parameter is available for conditional prediction.",
       call. = FALSE)
}

.conditional_heterogeneity_parameters <- function(object) {

  prior_names <- names(attr(object[["fit"]], "prior_list"))

  if ("tau" %in% prior_names) {
    return("tau")
  }
  if ("log_tau_intercept" %in% prior_names) {
    return("log_tau_intercept")
  }

  parameters <- grep("^log_tau_", prior_names, value = TRUE)
  if (length(parameters) > 0L) {
    return(parameters)
  }

  stop("No heterogeneity parameter is available for conditional prediction.",
       call. = FALSE)
}

.conditional_parameter_rows <- function(object, parameters,
                                        posterior_samples = NULL,
                                        rule = "OR") {

  BayesTools::check_char(parameters, "parameters", check_length = 0)
  BayesTools::check_char(rule, "rule", allow_values = c("AND", "OR"))

  posterior_samples <- .get_posterior_samples(object[["fit"]], posterior_samples)
  conditions <- vapply(
    parameters,
    .conditional_parameter_rows_single,
    logical(nrow(posterior_samples)),
    object            = object,
    posterior_samples = posterior_samples
  )

  if (rule == "AND") {
    keep <- apply(conditions, 1, all)
  } else {
    keep <- apply(conditions, 1, any)
  }

  if (!any(keep)) {
    warning("No samples left after conditioning.", call. = FALSE,
            immediate. = TRUE)
  }

  return(keep)
}

.conditional_parameter_rows_single <- function(parameter, object,
                                               posterior_samples) {

  prior_list <- attr(object[["fit"]], "prior_list")
  if (!parameter %in% names(prior_list)) {
    stop("Missing prior for conditional parameter '", parameter, "'.",
         call. = FALSE)
  }

  prior <- prior_list[[parameter]]
  S     <- nrow(posterior_samples)

  if (BayesTools::is.prior.spike_and_slab(prior)) {
    indicator <- .extract_posterior_indicator(
      posterior_samples = posterior_samples,
      parameter         = parameter,
      prior             = prior
    )
    return(indicator == 1)
  }

  if (BayesTools::is.prior.mixture(prior)) {
    components <- attr(prior, "components")
    if (!all(components %in% c("null", "alternative"))) {
      stop(
        "Conditional mixture posterior distributions are available only ",
        "for 'null' and 'alternative' components.",
        call. = FALSE
      )
    }
    indicator <- .extract_posterior_indicator(
      posterior_samples = posterior_samples,
      parameter         = parameter,
      prior             = prior
    )
    return(indicator %in% which(components == "alternative"))
  }

  warning(
    "The parameter '", parameter, "' is not a conditional parameter. ",
    "All samples are assumed to come from the conditional posterior distribution.",
    call. = FALSE,
    immediate. = TRUE
  )
  return(rep(TRUE, S))
}

.conditional_indicator_column <- function(posterior_samples, parameter) {

  return(.extract_posterior_indicator(
    posterior_samples = posterior_samples,
    parameter         = parameter
  ))
}

context("GLMM response measure prediction")


test_that("GLMM response count conversion matches continuity-corrected estimators", {

  bin_data <- data.frame(
    ai  = c(0, 0),
    ci  = c(0, 0),
    n1i = c(10, 20),
    n2i = c(12, 25)
  )
  bin_counts <- matrix(
    c(
      0, 3, 10, 25,
      4, 0, 15, 20
    ),
    nrow = 2,
    byrow = TRUE
  )

  bin_out <- .glmm_response_counts_to_measure(
    outcome_samples = bin_counts,
    outcome_data    = bin_data,
    outcome_type    = "bin"
  )

  bin_expected <- matrix(
    c(
      log(((0 + .5) * (12 - 3 + .5)) / ((10 - 0 + .5) * (3 + .5))),
      log(((10 + .5) * (25 - 25 + .5)) / ((20 - 10 + .5) * (25 + .5))),
      log(((4 + .5) * (12 - 0 + .5)) / ((10 - 4 + .5) * (0 + .5))),
      log((15 * (25 - 20)) / ((20 - 15) * 20))
    ),
    nrow = 2,
    byrow = TRUE
  )

  expect_equal(unname(bin_out), bin_expected, tolerance = 1e-12)
  expect_equal(colnames(bin_out), c("yi[1]", "yi[2]"))

  pois_data <- data.frame(
    x1i = c(0, 0),
    x2i = c(0, 0),
    t1i = c(10, 20),
    t2i = c(5, 40)
  )
  pois_counts <- matrix(
    c(
      0, 2, 4, 0,
      3, 6, 0, 8
    ),
    nrow = 2,
    byrow = TRUE
  )

  pois_out <- .glmm_response_counts_to_measure(
    outcome_samples = pois_counts,
    outcome_data    = pois_data,
    outcome_type    = "pois"
  )

  pois_expected <- matrix(
    c(
      log(((0 + .5) / 10) / ((2 + .5) / 5)),
      log(((4 + .5) / 20) / ((0 + .5) / 40)),
      log((3 / 10) / (6 / 5)),
      log(((0 + .5) / 20) / ((8 + .5) / 40))
    ),
    nrow = 2,
    byrow = TRUE
  )

  expect_equal(unname(pois_out), pois_expected, tolerance = 1e-12)
  expect_equal(colnames(pois_out), c("yi[1]", "yi[2]"))
})


test_that("GLMM response as_measure rejects undefined denominators", {

  expect_error(
    .check_glmm_response_as_measure(
      outcome_type = "bin",
      outcome_data = data.frame(n1i = c(10, 0), n2i = c(12, 12))
    ),
    "positive 'n1i' and 'n2i'"
  )

  expect_error(
    .check_glmm_response_as_measure(
      outcome_type = "pois",
      outcome_data = data.frame(t1i = c(10, 0), t2i = c(12, 12))
    ),
    "positive 't1i' and 't2i'"
  )

  expect_true(isTRUE(.check_glmm_response_as_measure(
    outcome_type = "pois",
    outcome_data = data.frame(t1i = c(10, 20), t2i = c(12, 15))
  )))
})


test_that("Poisson raw response RNG allows zero exposure as zero counts", {

  set.seed(1)

  out <- .outcome_rng.pois(
    mu_samples = matrix(c(0.2, -0.1, 0.3, 0.5), nrow = 2),
    log_phi    = matrix(0, nrow = 2, ncol = 2),
    t1i        = c(0, 3),
    t2i        = c(4, 0)
  )

  expect_equal(out[, "x1i[1]"], c(0, 0))
  expect_equal(out[, "x2i[2]"], c(0, 0))
  expect_true(all(out >= 0))
  expect_true(all(out == floor(out)))
})

# ============================================================================ #
# brma.funnel.R
# ============================================================================ #
#
# This file contains the funnel plot functions for brma objects.
#
# Two modes of dispatch:
# 1) No mods AND no scale: Shows observed effect sizes against the fitted
#    sampling distribution
# 2) All other cases: Shows residuals against a standard-error funnel
#
# ============================================================================ #


### funnel plot functions ----
#' @export
funnel <- function(x, ...) UseMethod("funnel")


#' @title Funnel Plot for brma Object
#'
#' @description \code{funnel.brma} creates a funnel plot for a fitted brma object.
#' For intercept-only models without scale regression, the default outcome mode
#' displays observed effect sizes against the fitted sampling distribution. For
#' models with location or scale moderators, the default residual mode displays
#' residuals against a standard-error funnel.
#'
#' @param x a fitted brma object
#' @param residual whether to use residual mode. Defaults to not specified,
#' which means the function automatically determines the mode:
#' \itemize{
#'   \item not specified: For intercept-only models without scale
#'     regression, displays observed effect sizes against the fitted sampling
#'     distribution funnel; for models with moderators or scale regression,
#'     automatically uses residual mode.
#'   \item \code{FALSE}: explicitly requests outcome mode.
#'   \item \code{TRUE}: explicitly requests residual mode, displaying residuals
#'     on the x-axis and using \code{type} to determine how those residuals are
#'     computed.
#' }
#' @param type the type of residuals to use when in residual mode.
#' Options are:
#' \itemize{
#'   \item \code{"LOO-PIT"} (alias: \code{"rstudent"}; default): Leave-one-out
#'     probability integral transform residuals returned by
#'     \code{\link{rstudent.brma}}.
#'   \item \code{"rstandard"}: Internally standardized residuals using
#'     \code{\link{rstandard.brma}}. Only available for normal outcome models.
#'   \item \code{"outcome"}: Raw outcome residuals from
#'     \code{\link{residuals.brma}} with \code{type = "outcome"}.
#' }
#' Only used when funnel is in residual mode.
#' @param unit output unit for residual mode. Only \code{"estimate"} is
#' implemented in this pass.
#' @param conditioning_depth residual conditioning depth for residual mode.
#' Options are \code{"marginal"}, \code{"cluster"}, and \code{"estimate"}.
#' The default LOO-PIT residual path is available only with marginal
#' conditioning depth.
#' @param sampling_heterogeneity whether heterogeneity should be incorporated
#' into the sampling distribution funnel. Defaults to \code{TRUE}. Only used
#' in outcome mode and ignored in residual mode.
#' @param sampling_bias whether publication bias should be incorporated into the
#' sampling distribution funnel. Defaults to \code{TRUE}. Only used when
#' \code{residual = FALSE} or when automatic mode selects outcome mode. Ignored
#' in residual mode. When \code{TRUE} and the model
#' includes selection models (weightfunction), uses selected-normal quantiles.
#' When \code{TRUE} and the model includes PET/PEESE, incorporates the expected
#' skew from these regression adjustments.
#' @param max_samples maximum number of posterior samples used for
#' model-averaged publication-bias funnel contours. Defaults to \code{10000}.
#' Use \code{Inf} to use all posterior samples.
#' @param plot_type whether to use a base plot \code{"base"} or ggplot2
#' \code{"ggplot"} for plotting. Defaults to \code{"base"}.
#' @param ... additional graphical arguments to customize the plot appearance:
#' \describe{
#'   \item{xlim, ylim}{numeric vectors of length 2 specifying axis limits}
#'   \item{xlab, ylab}{character strings for axis labels}
#'   \item{main}{character string for plot title (default: no title)}
#'   \item{pch}{point symbol (default: 21, filled circle). Use standard R pch values.}
#'   \item{col}{point border color (default: "black")}
#'   \item{bg}{point fill/background color (default: "#A6A6A6")}
#'   \item{cex}{point size multiplier for base graphics (default: 1)}
#'   \item{size}{point size for ggplot2 (default: 2)}
#'   \item{las}{axis-label style for base graphics (default: 1)}
#'   \item{back}{background region color (default: "grey"). Set to \code{NA} to suppress.}
#'   \item{shade}{funnel region fill color (default: "white"). Set to \code{NA} to suppress.}
#'   \item{lty}{line type for funnel edges and center line (default: "dotted")}
#'   \item{col.line}{color for funnel edge lines (default: "black")}
#'   \item{refline}{numeric override for the reference line. By default,
#'   residual mode uses 0, while outcome mode uses the center of the fitted
#'   sampling distribution, which may be curved when PET/PEESE bias adjustment
#'   is incorporated.}
#'   \item{col.refline}{color of vertical reference line (default: "black")}
#'   \item{as_data}{if \code{TRUE}, returns plot data instead of creating a plot}
#' }
#'
#' @details
#' The funnel plot has two modes. If \code{residual} is not specified, the mode
#' is chosen automatically from the fitted model: intercept-only models without
#' scale regression use outcome mode, whereas models with location or scale
#' moderators use residual mode.
#'
#' \strong{Outcome mode} (intercept-only models without scale regression):
#' Displays observed effect sizes on the x-axis and standard errors on the
#' y-axis. The reference line follows the center of the fitted sampling
#' distribution. When \code{sampling_bias = FALSE}, this center is the pooled
#' effect; when PET/PEESE bias adjustment is incorporated, the center line can
#' vary with the standard error. The funnel region represents the central 95\%
#' region of the sampling distribution, optionally incorporating heterogeneity
#' and publication bias.
#'
#'
#' \strong{Residual mode} (models with moderators or scale regression):
#' Displays residuals on the x-axis and standard errors on the
#' y-axis. The funnel region represents the central 95\% region of
#' \eqn{N(0, \mathrm{SE}^2)}. With \code{type = "LOO-PIT"}, the plotted
#' residuals and standard errors are the raw-scale LOO predictive companions
#' returned by \code{\link{rstudent.brma}}; the PIT-normalized \code{z} values
#' are used by \code{\link{qqnorm.brma}} and influence diagnostics. With
#' \code{type = "rstandard"}, the plotted values are internally standardized
#' residual companions from \code{\link{rstandard.brma}}. With
#' \code{type = "outcome"}, these are raw outcome residuals. Under a correctly
#' specified model, most points should fall within this region.
#'
#' The \code{type} argument controls how residuals are computed in residual
#' mode. See \code{\link{residuals.brma}} for details on each type.
#' The \code{sampling_heterogeneity} and \code{sampling_bias} arguments are
#' ignored in residual mode.
#'
#' For GLMM models, observed effect sizes are computed from the raw frequency
#' data using formulas equivalent to \code{metafor::escalc}. Residual-mode
#' GLMM funnels use approximate effect-size-scale residual/PIT companions, not
#' exact PIT diagnostics for the raw count likelihood.
#'
#' @return If \code{as_data = TRUE}, \code{funnel.brma} returns a list with the
#' data used for plotting, including the plotted points, funnel polygons,
#' plotting limits, labels, and reference line. Otherwise, it returns
#' \code{NULL} invisibly if \code{plot_type = "base"} or a ggplot object if
#' \code{plot_type = "ggplot"}.
#'
#' @examples \dontrun{
#' if (requireNamespace("metadat", quietly = TRUE) &&
#'     requireNamespace("metafor", quietly = TRUE)) {
#'   data(dat.bcg, package = "metadat")
#'   dat <- metafor::escalc(
#'     measure = "RR",
#'     ai      = tpos,
#'     bi      = tneg,
#'     ci      = cpos,
#'     di      = cneg,
#'     data    = dat.bcg
#'   )
#'
#'   fit <- brma(yi = yi, vi = vi, data = dat, measure = "RR")
#'   funnel(fit)
#'   funnel(fit, pch = 19, col = "blue", bg = "lightblue")
#'
#'   fit_reg <- brma(
#'     yi      = yi,
#'     vi      = vi,
#'     mods    = ~ ablat + year,
#'     data    = dat,
#'     measure = "RR"
#'   )
#'   fit_reg <- add_loo(fit_reg)
#'   funnel(fit_reg)
#'   funnel(fit_reg, type = "outcome")
#'
#'   funnel_data <- funnel(fit, as_data = TRUE)
#'   funnel(fit, plot_type = "ggplot")
#' }
#' }
#'
#' @seealso [residuals.brma()], [rstandard.brma()], [rstudent.brma()],
#'   [predict.brma()]
#' @aliases funnel
#' @export
#' @exportS3Method metafor::funnel
#' @rdname funnel
funnel.brma <- function(x, residual, type = "LOO-PIT",
                        unit = "estimate", conditioning_depth = "marginal",
                        sampling_heterogeneity = TRUE, sampling_bias = TRUE,
                        max_samples = 10000,
                        plot_type = "base", ...) {
  # input validation
  conditioning_depth_specified <- !missing(conditioning_depth)
  dots                         <- list(...)
  .check_legacy_level_arg(dots, "funnel()")

  BayesTools::check_bool(sampling_heterogeneity, "sampling_heterogeneity")
  BayesTools::check_bool(sampling_bias, "sampling_bias")
  BayesTools::check_char(plot_type, "plot_type", allow_values = c("base", "ggplot"))
  max_samples <- .normalize_funnel_max_samples(max_samples)

  # set up graphical arguments with defaults
  dots <- .set_dots_funnel(dots)

  # get model characteristics
  is_mods  <- .is_mods(x)
  is_scale <- .is_scale(x)

  # determine mode: residual mode if mods/scale present, or if explicitly requested

  if (missing(residual)) {
    is_residual <- is_mods || is_scale
  } else {
    BayesTools::check_bool(residual, "residual")
    is_residual <- residual
  }

  # generate funnel data based on mode
  if (is_residual) {
    BayesTools::check_char(type, "type", allow_values = c("outcome", "rstandard", "LOO-PIT", "rstudent"))
    unit               <- .normalize_unit(unit)
    conditioning_depth <- .normalize_conditioning_depth(conditioning_depth)
    .check_unit_conditioning_depth(
      object             = x,
      unit               = unit,
      conditioning_depth = conditioning_depth,
      caller             = "funnel()"
    )

    if (unit == "cluster") {
      .check_cluster_unit_deferred("funnel()")
    }

    if ((type == "LOO-PIT" || type == "rstudent") && conditioning_depth_specified) {
      stop(
        "LOO-PIT residuals use the estimate-unit LOO target; ",
        "do not set 'conditioning_depth'.",
        call. = FALSE
      )
    }

    # sampling heterogeneity/bias ignored in residual mode - explicitly set to FALSE
    sampling_heterogeneity <- FALSE
    sampling_bias          <- FALSE
    funnel_data <- .funnel_data_residual(
      x                  = x,
      type               = type,
      unit               = unit,
      conditioning_depth = conditioning_depth,
      dots               = dots
    )
  } else {
    funnel_data <- .funnel_data_outcome(
      x                      = x,
      sampling_heterogeneity = sampling_heterogeneity,
      sampling_bias          = sampling_bias,
      max_samples            = max_samples,
      dots                   = dots
    )
  }

  # allow data return for programmatic access
  if (isTRUE(dots[["as_data"]])) {
    return(funnel_data)
  }

  # create plot
  if (plot_type == "ggplot") {
    return(.funnel_plot_ggplot(funnel_data, dots))
  } else {
    .funnel_plot_base(funnel_data, dots)
    return(invisible(NULL))
  }
}


# ---------------------------------------------------------------------------- #
# .funnel_data_outcome
# ---------------------------------------------------------------------------- #
#
# Generate funnel plot data for outcome mode (no mods AND no scale).
#
# Shows observed effect sizes against the fitted sampling distribution funnel.
#
# @param x                      brma object
# @param sampling_heterogeneity logical; incorporate tau into funnel
# @param sampling_bias          logical; incorporate bias into funnel
# @param max_samples            maximum posterior samples for biased contours
# @param dots                   list of graphical parameters
#
# @return list with funnel plot data components
#
# ---------------------------------------------------------------------------- #
.funnel_data_outcome <- function(x, sampling_heterogeneity, sampling_bias,
                                 max_samples, dots) {
  # get model characteristics
  is_weightfunction <- .is_weightfunction(x)
  is_PET            <- .is_PET(x)
  is_PEESE          <- .is_PEESE(x)
  effect_direction  <- .effect_direction(x)

  # get observed effect sizes (yi) - these go on the x-axis
  yi <- .outcome_data_yi(x)

  # get pooled effect estimate - this is the funnel center
  mu      <- pooled_effect(x)
  mu_mean <- summary(mu)["mu", "Mean"]

  # get standard errors
  se <- .outcome_data_sei(x)
  K  <- length(se)

  # get tau for incorporating heterogeneity into sampling distribution
  if (sampling_heterogeneity) {
    tau <- .get_funnel_tau(x)
  } else {
    tau <- 0
  }

  # compute standard-error plotting range
  se_range <- pretty(c(0, max(se)))

  # apply custom y-axis limits if provided
  if (!is.null(dots[["ylim"]])) {
    se_range <- pretty(dots[["ylim"]])
  }

  # set axis labels - using "Observed Effect Size" for outcome mode
  xlab <- if (!is.null(dots[["xlab"]])) dots[["xlab"]] else "Observed Effect Size"
  ylab <- if (!is.null(dots[["ylab"]])) dots[["ylab"]] else "Standard Error"

  # set data for reference line
  if (!is.null(dots[["refline"]])) {
    df_refline <- data.frame(
      x = rep(dots[["refline"]], 2),
      y = range(se_range)
    )
  } else {
    # use model center
    df_refline <- NULL
  }

  # determine y-axis plot limit before computing clipped contours
  if (!is.null(dots[["ylim"]])) ylim <- dots[["ylim"]] else ylim <- range(se_range)

  # compute funnel region STRICTLY within ylim to avoid artifacts
  # generate sequence within ylim
  se_sequence_clipped <- seq(min(ylim), max(ylim), length.out = 100)

  # compute contours for clipped sequence
  ci_offsets_clipped <- .get_funnel_quantiles(
    x                      = x,
    se_sequence            = se_sequence_clipped,
    mu                     = mu_mean,
    tau                    = tau,
    sampling_heterogeneity = sampling_heterogeneity,
    sampling_bias          = sampling_bias,
    is_weightfunction      = is_weightfunction,
    is_PET                 = is_PET,
    is_PEESE               = is_PEESE,
    effect_direction       = effect_direction,
    max_samples            = max_samples
  )
  ci_left  <- ci_offsets_clipped$lower
  ci_right <- ci_offsets_clipped$upper

  # create data frames for plotting
  x_range <- pretty(c(ci_left, ci_right, yi))

  # apply custom x-axis limits after funnel computation
  if (!is.null(dots[["xlim"]])) {
    x_range <- pretty(dots[["xlim"]])
  }

  if (!is.null(dots[["xlim"]])) xlim <- dots[["xlim"]] else xlim <- range(x_range)

  # Clamp CIs to xlim to prevent drawing outside plot area (for POLYGON)
  ci_left_clipped  <- pmax(ci_left, min(xlim))
  ci_right_clipped <- pmin(ci_right, max(xlim))

  # Compute reference line for plotting (if not user-specified)
  if (is.null(df_refline)) {
      df_refline <- data.frame(
        x = ci_offsets_clipped$mid,
        y = se_sequence_clipped
      )
      # Clip refline to xlim just in case
      df_refline <- .clip_line_x(df_refline$x, df_refline$y, xlim)
  }

  # create output data structures
  # Points: observed data
  df_points <- data.frame(
    x = yi,
    y = se
  )
  # Funnel: use CLIPPED data for filled polygon to respect limits
  df_funnel <- data.frame(
    x = c(rev(ci_left_clipped), ci_right_clipped),
    y = c(rev(se_sequence_clipped), se_sequence_clipped)
  )
  # Edges: use CLIPPED lines using exact intersection
  edge1 <- .clip_line_x(ci_left, se_sequence_clipped, xlim)
  edge2 <- .clip_line_x(ci_right, se_sequence_clipped, xlim)

  df_funnel_edge1 <- data.frame(
    x = edge1$x,
    y = edge1$y
  )
  df_funnel_edge2 <- data.frame(
    x = edge2$x,
    y = edge2$y
  )
  df_background <- data.frame(
    x = c(min(xlim), max(xlim), max(xlim), min(xlim)),
    y = c(min(ylim), min(ylim), max(ylim), max(ylim))
  )

  return(list(
    points       = df_points,
    funnel       = df_funnel,
    funnel_edge1 = df_funnel_edge1,
    funnel_edge2 = df_funnel_edge2,
    background   = df_background,
    x_range      = x_range, # Keep original ticks for axis drawing if needed
    y_range      = se_range,
    xlim         = xlim,
    ylim         = ylim,
    xlab         = xlab,
    ylab         = ylab,
    refline      = df_refline
  ))
}


# ---------------------------------------------------------------------------- #
# .funnel_data_residual
# ---------------------------------------------------------------------------- #
#
# Generate funnel plot data for residual mode (mods OR scale).
#
# @param x                  brma object
# @param type               character; type of residuals
# @param unit               character; output unit
# @param conditioning_depth character; conditioning depth
# @param dots               list of graphical parameters
#
# @return list with funnel plot data components
#
# ---------------------------------------------------------------------------- #
.funnel_data_residual <- function(x, type, unit, conditioning_depth, dots) {

  # get residuals based on type
  # rstandard and rstudent return data.frame with resid, se, z columns
  if (type == "rstandard") {
    res_obj <- rstandard.brma(
      model              = x,
      unit               = unit,
      conditioning_depth = conditioning_depth
    )
    res <- res_obj$resid
    se  <- res_obj$se
  } else if (type == "LOO-PIT" || type == "rstudent") {
    res_obj <- rstudent.brma(x, unit = unit)
    res <- res_obj$resid
    se  <- res_obj$se
  } else if (type == "outcome") {
    # raw outcome residuals
    res_obj <- residuals.brma(
      object             = x,
      type               = "outcome",
      unit               = unit,
      conditioning_depth = conditioning_depth
    )
    if (is.data.frame(res_obj)) {
      res <- res_obj$resid
      se  <- res_obj$se
    } else {
      res <- res_obj
      se  <- .outcome_data_sei(x)
    }
  }
  K <- length(se)

  # compute funnel region - NO tau incorporation for residual funnel
  # for residuals, the expected distribution is N(0, se^2)
  # so bounds are +/- 1.96 * se
  # se_range determines the axis ticks
  se_range    <- pretty(c(0, max(se)))

  # determine plot limits
  # For xlim: pretty range of residuals + quantiles
  # For ylim: defaults to se_range range
  x_init_range <- pretty(c(stats::qnorm(0.025) * max(se_range), stats::qnorm(0.975) * max(se_range), res))

  if (!is.null(dots[["xlim"]])) xlim <- dots[["xlim"]] else xlim <- range(x_init_range)
  if (!is.null(dots[["ylim"]])) ylim <- dots[["ylim"]] else ylim <- range(se_range)

  # generate sequence strictly within ylim for clean polygons
  se_sequence_clipped <- seq(min(ylim), max(ylim), length.out = 100)

  # funnel bounds are quantiles of N(0, se^2) -> qnorm(p) * se
  ci_left  <- stats::qnorm(0.025) * se_sequence_clipped
  ci_right <- stats::qnorm(0.975) * se_sequence_clipped

  # Clamp CIs to xlim to prevent drawing outside plot area (for POLYGON)
  ci_left_clipped  <- pmax(ci_left, min(xlim))
  ci_right_clipped <- pmin(ci_right, max(xlim))

  # set axis labels
  xlab <- if (!is.null(dots[["xlab"]])) dots[["xlab"]] else "Residual"
  ylab <- if (!is.null(dots[["ylab"]])) dots[["ylab"]] else "Standard Error"

  # set data for reference line (residuals centered at 0 or user specified)
  if (!is.null(dots[["refline"]])) {
    refline_x <- dots[["refline"]]
  } else {
    refline_x <- 0
  }
  df_refline <- data.frame(
    x = rep(refline_x, 2),
    y = range(se_range)
  )

  # create output data structures
  df_points <- data.frame(
    x                  = res,
    y                  = se,
    unit               = rep(unit, K),
    conditioning_depth = rep(conditioning_depth, K)
  )
  if (exists("res_obj") && is.data.frame(res_obj) && "cluster" %in% names(res_obj)) {
    df_points[["cluster"]]     <- res_obj[["cluster"]]
    df_points[["n_estimates"]] <- res_obj[["n_estimates"]]
  }
  # Funnel: use CLIPPED data for filled polygon
  df_funnel <- data.frame(
    x = c(rev(ci_left_clipped), ci_right_clipped),
    y = c(rev(se_sequence_clipped), se_sequence_clipped)
  )
  # Edges: use CLIPPED lines using exact intersection
  edge1 <- .clip_line_x(ci_left, se_sequence_clipped, xlim)
  edge2 <- .clip_line_x(ci_right, se_sequence_clipped, xlim)

  df_funnel_edge1 <- data.frame(
    x = edge1$x,
    y = edge1$y
  )
  df_funnel_edge2 <- data.frame(
    x = edge2$x,
    y = edge2$y
  )
  df_background <- data.frame(
    x = c(min(xlim), max(xlim), max(xlim), min(xlim)),
    y = c(min(ylim), min(ylim), max(ylim), max(ylim))
  )

  return(list(
    points       = df_points,
    funnel       = df_funnel,
    funnel_edge1 = df_funnel_edge1,
    funnel_edge2 = df_funnel_edge2,
    background   = df_background,
    x_range      = x_init_range, # ticks
    y_range      = se_range,     # ticks
    xlim         = xlim,
    ylim         = ylim,
    xlab         = xlab,
    ylab         = ylab,
    refline      = df_refline
  ))
}


# ---------------------------------------------------------------------------- #
# .funnel_plot_base
# ---------------------------------------------------------------------------- #
#
# Create funnel plot using base R graphics.
#
# @param data list of funnel plot data from .funnel_data_* functions
# @param dots list of graphical parameters
#
# @return NULL invisibly
#
# ---------------------------------------------------------------------------- #
.funnel_plot_base <- function(data, dots) {
  # extract graphical parameters
  pch   <- dots[["pch"]]
  col   <- dots[["col"]]
  bg    <- dots[["bg"]]
  cex   <- dots[["cex"]]
  back  <- dots[["back"]]
  shade <- dots[["shade"]]
  lty   <- dots[["lty"]]
  col_line <- dots[["col.line"]]
  col_refline <- dots[["col.refline"]]
  main <- dots[["main"]]
  las  <- dots[["las"]]

  # extract data components
  df_points <- data$points
  df_funnel <- data$funnel
  df_funnel_edge1 <- data$funnel_edge1
  df_funnel_edge2 <- data$funnel_edge2
  df_background <- data$background
  x_range  <- data$x_range
  se_range <- data$y_range
  xlab <- data$xlab
  ylab <- data$ylab
  # extract data components
  df_points <- data$points
  df_funnel <- data$funnel
  df_funnel_edge1 <- data$funnel_edge1
  df_funnel_edge2 <- data$funnel_edge2
  df_background <- data$background

  # use limits from data object or dots for PLOTTING LIMITS
  if (!is.null(dots[["xlim"]])) xlim_plot <- dots[["xlim"]] else xlim_plot <- data$xlim
  if (!is.null(dots[["ylim"]])) ylim_plot <- dots[["ylim"]] else ylim_plot <- data$ylim

  # Generate pretty ticks based on the limits
  x_ticks <- pretty(xlim_plot)
  y_ticks <- pretty(ylim_plot)

  xlab <- data$xlab
  ylab <- data$ylab
  df_refline <- data$refline

  # set up the plot area with reversed y-axis
  graphics::plot(
    NA, NA,
    xlim = range(xlim_plot),
    ylim = rev(range(ylim_plot)),
    xlab = xlab,
    ylab = ylab,
    main = main,
    type = "n",
    axes = FALSE,
    las  = las
  )
  graphics::axis(1, at = x_ticks, las = las)
  graphics::axis(2, at = rev(y_ticks), las = las)

  # draw background polygon (if not suppressed)
  if (!is.na(back)) {
    graphics::polygon(df_background$x, df_background$y, col = back, border = NA)
  }

  # draw funnel polygon (if not suppressed)
  if (!is.na(shade)) {
    graphics::polygon(df_funnel$x, df_funnel$y, col = shade, border = NA)
  }

  # vertical reference line (now potentially curved)
  graphics::lines(df_refline$x, df_refline$y, lty = lty, col = col_refline)

  # funnel edge lines
  graphics::lines(df_funnel_edge1$x, df_funnel_edge1$y, lty = lty, col = col_line)
  graphics::lines(df_funnel_edge2$x, df_funnel_edge2$y, lty = lty, col = col_line)

  # plot points
  graphics::points(df_points$x, df_points$y, pch = pch, col = col, bg = bg, cex = cex)

  return(invisible(NULL))
}


# ---------------------------------------------------------------------------- #
# .funnel_plot_ggplot
# ---------------------------------------------------------------------------- #
#
# Create funnel plot using ggplot2.
#
# @param data list of funnel plot data from .funnel_data_* functions
# @param dots list of graphical parameters
#
# @return ggplot object
#
# ---------------------------------------------------------------------------- #
.funnel_plot_ggplot <- function(data, dots) {
  # extract graphical parameters
  pch   <- dots[["pch"]]
  col   <- dots[["col"]]
  bg    <- dots[["bg"]]
  size  <- dots[["size"]]
  back  <- dots[["back"]]
  shade <- dots[["shade"]]
  lty   <- dots[["lty"]]
  col_line <- dots[["col.line"]]
  col_refline <- dots[["col.refline"]]
  main <- dots[["main"]]

  # extract data components
  df_points <- data$points
  df_funnel <- data$funnel
  df_funnel_edge1 <- data$funnel_edge1
  df_funnel_edge2 <- data$funnel_edge2
  df_background <- data$background
  x_range   <- data$x_range
  se_range  <- data$y_range
  xlab      <- data$xlab
  ylab      <- data$ylab
  # extract data components
  df_points <- data$points
  df_funnel <- data$funnel
  df_funnel_edge1 <- data$funnel_edge1
  df_funnel_edge2 <- data$funnel_edge2
  df_background   <- data$background

  # use limits from data object or dots
  if (!is.null(dots[["xlim"]])) xlim_plot <- dots[["xlim"]] else xlim_plot <- data$xlim
  if (!is.null(dots[["ylim"]])) ylim_plot <- dots[["ylim"]] else ylim_plot <- data$ylim

  # Generate pretty ticks
  x_ticks <- pretty(xlim_plot)
  y_ticks <- pretty(ylim_plot)

  xlab <- data$xlab
  ylab <- data$ylab
  df_refline <- data$refline

  out <- ggplot2::ggplot()

  # add background polygon (if not suppressed)
  if (!is.na(back)) {
    out <- out +
      ggplot2::geom_polygon(
        mapping = ggplot2::aes(
          x = df_background$x,
          y = df_background$y
        ),
        fill = back
      )
  }

  # add funnel polygon (if not suppressed)
  if (!is.na(shade)) {
    out <- out +
      ggplot2::geom_polygon(
        mapping = ggplot2::aes(
          x = df_funnel$x,
          y = df_funnel$y
        ),
        fill = shade
      )
  }

  out <- .add_funnel_path_or_line(
    out    = out,
    data   = df_refline,
    lty    = lty,
    colour = col_refline
  )

  # add funnel edge lines
  out <- .add_funnel_path_or_line(
    out    = out,
    data   = df_funnel_edge1,
    lty    = lty,
    colour = col_line
  )
  out <- .add_funnel_path_or_line(
    out    = out,
    data   = df_funnel_edge2,
    lty    = lty,
    colour = col_line
  )

  # add points
  out <- out +
    ggplot2::geom_point(
      mapping = ggplot2::aes(
        x = df_points$x,
        y = df_points$y
      ),
      colour = col,
      fill = bg,
      shape = pch,
      size = size
    )

  # set axis scales and labels
  out <- out +
    ggplot2::scale_x_continuous(
      breaks = x_ticks,
      limits = range(xlim_plot),
      name   = xlab
    ) +
    ggplot2::scale_y_reverse(
      breaks = rev(y_ticks),
      limits = rev(range(ylim_plot)),
      name   = ylab
    )

  # add title if specified
  if (!is.null(main)) {
    out <- out + ggplot2::ggtitle(main)
  }

  return(out)
}


# ---------------------------------------------------------------------------- #
# .add_funnel_path_or_line
# ---------------------------------------------------------------------------- #
#
# Add a contour line to a ggplot object.
#
# PET/PEESE contours can be non-monotone in x because they are parameterized by
# standard error. ggplot2::geom_line() sorts by x, which can scramble such
# contours, so those cases must use geom_path().
#
# ---------------------------------------------------------------------------- #
.add_funnel_path_or_line <- function(out, data, lty, colour) {

  x_finite <- data[["x"]][is.finite(data[["x"]])]
  if (length(x_finite) < 3L) {
    is_monotone <- TRUE
  } else {
    x_diff      <- diff(x_finite)
    x_diff      <- x_diff[x_diff != 0]
    is_monotone <- length(x_diff) < 2L || all(x_diff > 0) || all(x_diff < 0)
  }

  if (is_monotone) {
    return(out +
      ggplot2::geom_line(
        mapping = ggplot2::aes(
          x = data[["x"]],
          y = data[["y"]]
        ),
        linetype = lty,
        colour   = colour
      )
    )
  }

  return(out +
    ggplot2::geom_path(
      data    = data,
      mapping = ggplot2::aes(x = .data[["x"]], y = .data[["y"]]),
      linetype = lty,
      colour   = colour
    )
  )
}


# ---------------------------------------------------------------------------- #
# .clip_line_x
# ---------------------------------------------------------------------------- #
#
# Clip a line (x, y) to x-limits, interpolating exact intersections.
# Assumes y is monotonic (funnel lines usually are monotonic with SE).
#
# @param x numeric vector of x coordinates
# @param y numeric vector of y coordinates
# @param xlim numeric vector of length 2 (min, max)
#
# @return data.frame with xe and ye
#
# ---------------------------------------------------------------------------- #
.clip_line_x <- function(x, y, xlim) {

  if (length(x) < 2) return(data.frame(x=x, y=y))

  xmin <- min(xlim)
  xmax <- max(xlim)

  x_out <- numeric(0)
  y_out <- numeric(0)

  for (i in 1:(length(x) - 1)) {
    x1 <- x[i]
    y1 <- y[i]
    x2 <- x[i+1]
    y2 <- y[i+1]

    # strictly inside
    if (x1 >= xmin && x1 <= xmax) {
      x_out <- c(x_out, x1)
      y_out <- c(y_out, y1)
    }

    # check intersection with xmin
    if ((x1 < xmin && x2 > xmin) || (x1 > xmin && x2 < xmin)) {
      y_int <- y1 + (y2 - y1) * (xmin - x1) / (x2 - x1)
      x_out <- c(x_out, xmin)
      y_out <- c(y_out, y_int)
    }

    # check intersection with xmax
    if ((x1 < xmax && x2 > xmax) || (x1 > xmax && x2 < xmax)) {
      y_int <- y1 + (y2 - y1) * (xmax - x1) / (x2 - x1)
      x_out <- c(x_out, xmax)
      y_out <- c(y_out, y_int)
    }
  }

  # handle last point
  xn <- x[length(x)]
  yn <- y[length(y)]
  if (xn >= xmin && xn <= xmax) {
    x_out <- c(x_out, xn)
    y_out <- c(y_out, yn)
  }

  return(data.frame(x = x_out, y = y_out))
}


# ---------------------------------------------------------------------------- #
# .get_funnel_tau
# ---------------------------------------------------------------------------- #
#
# Extract mean tau (heterogeneity) estimate from brma object for funnel plot.
#
# For multilevel models, returns total tau (combining within and between components).
#
# @param object       brma object
# @return numeric scalar representing the mean tau estimate
#
# ---------------------------------------------------------------------------- #
.get_funnel_tau <- function(object) {
  # use pooled_heterogeneity to get mean tau
  tau_samples <- pooled_heterogeneity(object)
  tau_summary <- summary(tau_samples)
  return(tau_summary["tau", "Mean"])
}


# ---------------------------------------------------------------------------- #
# .get_funnel_quantiles
# ---------------------------------------------------------------------------- #
#
# Compute quantiles for funnel plot contours based on the sampling distribution.
#
# When sampling_bias = TRUE, posterior rows dispatch to their active bias model
# and the contour is obtained by inverting the model-averaged CDF.
#
# When sampling_bias = FALSE: uses standard normal quantiles.
#
# @param x                      brma object
# @param se_sequence            numeric vector of SE values for funnel contours
# @param mu                     numeric scalar of pooled effect estimate
# @param tau                    numeric scalar of heterogeneity estimate
# @param sampling_heterogeneity logical; whether to incorporate tau
# @param sampling_bias          logical; whether to incorporate bias into sampling dist
# @param is_weightfunction      logical; whether model has selection model
# @param is_PET                 logical; whether model has PET adjustment
# @param is_PEESE               logical; whether model has PEESE adjustment
# @param effect_direction       character; "positive" or "negative"
#
# @return list with 'lower' and 'upper' quantile vectors
#
# ---------------------------------------------------------------------------- #
.get_funnel_quantiles <- function(x, se_sequence, mu, tau,
                                  sampling_heterogeneity, sampling_bias,
                                  is_weightfunction, is_PET, is_PEESE,
                                  effect_direction, max_samples) {
  n_se   <- length(se_sequence)
  sd_seq <- sqrt(se_sequence^2 + tau^2)

  # default: standard normal quantiles centered at mu
  if (!sampling_bias || (!is_weightfunction && !is_PET && !is_PEESE)) {
    return(list(
      lower = stats::qnorm(0.025, mean = mu, sd = sd_seq),
      upper = stats::qnorm(0.975, mean = mu, sd = sd_seq),
      mid   = rep(mu, n_se)
    ))
  }

  if (!BayesTools::is.prior.mixture(x[["priors"]][["outcome"]][["bias"]])) {
    if (is_PET || is_PEESE) {
      return(.get_funnel_quantiles_PETPEESE_plugin(
        x                = x,
        se_sequence      = se_sequence,
        mu               = mu,
        sd_seq           = sd_seq,
        is_PET           = is_PET,
        is_PEESE         = is_PEESE,
        effect_direction = effect_direction
      ))
    }
  }

  return(.get_funnel_quantiles_model_averaged(
    x                      = x,
    se_sequence            = se_sequence,
    sampling_heterogeneity = sampling_heterogeneity,
    effect_direction       = effect_direction,
    max_samples            = max_samples
  ))
}


# ---------------------------------------------------------------------------- #
# .get_funnel_quantiles_PETPEESE_plugin
# ---------------------------------------------------------------------------- #
#
# Fast plug-in normal contours for single PET/PEESE priors.
#
# ---------------------------------------------------------------------------- #
.get_funnel_quantiles_PETPEESE_plugin <- function(x, se_sequence, mu, sd_seq,
                                                  is_PET, is_PEESE,
                                                  effect_direction) {

  n_se              <- length(se_sequence)
  posterior_samples <- .get_posterior_samples(x[["fit"]])
  bias_shift        <- rep(0, n_se)
  direction         <- ifelse(effect_direction == "negative", -1, 1)

  if (is_PET && "PET" %in% colnames(posterior_samples)) {
    PET_mean   <- mean(posterior_samples[, "PET"])
    bias_shift <- bias_shift + direction * PET_mean * se_sequence
  }

  if (is_PEESE && "PEESE" %in% colnames(posterior_samples)) {
    PEESE_mean <- mean(posterior_samples[, "PEESE"])
    bias_shift <- bias_shift + direction * PEESE_mean * se_sequence^2
  }

  mid   <- mu + bias_shift
  lower <- stats::qnorm(0.025, mean = mid, sd = sd_seq)
  upper <- stats::qnorm(0.975, mean = mid, sd = sd_seq)

  return(list(lower = lower, upper = upper, mid = mid))
}


# ---------------------------------------------------------------------------- #
# .get_funnel_quantiles_model_averaged
# ---------------------------------------------------------------------------- #
#
# Compute model-averaged funnel contours for publication-bias models.
#
# Each posterior row dispatches to its active bias model:
# - no-bias rows use a normal CDF
# - PET/PEESE rows use a normal CDF with the row-specific bias offset
# - selection-model rows use the selected-normal CDF
#
# Quantiles are obtained from the averaged posterior-row CDF.
#
# @return list with 'lower', 'upper', and 'mid' quantile vectors
#
# ---------------------------------------------------------------------------- #
.get_funnel_quantiles_model_averaged <- function(x, se_sequence,
                                                 sampling_heterogeneity,
                                                 effect_direction,
                                                 max_samples = Inf) {

  setup <- .funnel_model_averaged_setup(
    x                      = x,
    sampling_heterogeneity = sampling_heterogeneity,
    max_samples            = max_samples
  )

  if (.has_native_funnel_model_averaged_quantiles(setup)) {
    return(.funnel_model_averaged_quantiles_native(
      se_sequence      = se_sequence,
      setup            = setup,
      effect_direction = effect_direction
    ))
  }

  lower <- vapply(
    se_sequence,
    .funnel_model_averaged_quantile,
    numeric(1),
    p                = 0.025,
    setup            = setup,
    effect_direction = effect_direction
  )
  upper <- vapply(
    se_sequence,
    .funnel_model_averaged_quantile,
    numeric(1),
    p                = 0.975,
    setup            = setup,
    effect_direction = effect_direction
  )
  mid <- vapply(
    se_sequence,
    .funnel_model_averaged_quantile,
    numeric(1),
    p                = 0.5,
    setup            = setup,
    effect_direction = effect_direction
  )

  return(list(lower = lower, upper = upper, mid = mid))
}

.has_native_funnel_model_averaged_quantiles <- function(setup) {

  if (is.null(setup[["selection"]])) {
    return(FALSE)
  }

  return(is.loaded("RoBMA_funnel_model_averaged_quantiles", PACKAGE = "RoBMA"))
}

.funnel_model_averaged_quantiles_native <- function(se_sequence, setup,
                                                    effect_direction) {

  selection <- setup[["selection"]]
  if (any(setup[["is_weightfunction"]])) {
    .selection_require_step_evaluable(
      selection,
      ".funnel_model_averaged_quantiles()"
    )
  }

  direction <- ifelse(effect_direction == "negative", -1L, 1L)

  return(.Call(
    "RoBMA_funnel_model_averaged_quantiles",
    .native_numeric_vector(se_sequence),
    .native_numeric_vector(setup[["mu"]]),
    .native_numeric_vector(setup[["tau"]]),
    .native_numeric_vector(setup[["PET"]]),
    .native_numeric_vector(setup[["PEESE"]]),
    .native_integer_vector(setup[["is_weightfunction"]]),
    .native_numeric_matrix(selection[["omega"]]),
    .native_numeric_vector(selection[["alpha"]]),
    .native_integer_vector(selection[["phack_kind"]]),
    .native_integer_vector(selection[["kernel_mode"]]),
    .native_numeric_vector(selection[["z_lower"]]),
    .native_numeric_vector(selection[["z_upper"]]),
    .native_integer_vector(selection[["sign"]]),
    .native_integer_vector(selection[["phack_q"]]),
    .native_numeric_vector(selection[["phack_z_source"]]),
    .native_numeric_vector(selection[["phack_z_dest"]]),
    .native_numeric_vector(selection[["segments"]][["bounds"]]),
    .native_integer_vector(selection[["segments"]][["step_bin"]]),
    .native_integer_vector(selection[["segments"]][["phack_region"]]),
    .native_integer_vector(direction),
    .selection_telescope_probabilities(selection),
    PACKAGE = "RoBMA"
  ))
}


# ---------------------------------------------------------------------------- #
# .funnel_model_averaged_setup
# ---------------------------------------------------------------------------- #
#
# Prepare posterior-row quantities for model-averaged funnel contours.
#
# ---------------------------------------------------------------------------- #
.funnel_model_averaged_setup <- function(x, sampling_heterogeneity,
                                         max_samples = Inf) {

  posterior_samples <- .get_posterior_samples(x[["fit"]])
  priors_bias       <- x[["priors"]][["outcome"]][["bias"]]

  if (!BayesTools::is.prior.mixture(priors_bias)) {
    priors_bias <- list(priors_bias)
  }

  branch_is_PET   <- vapply(priors_bias, BayesTools::is.prior.PET,   logical(1))
  branch_is_PEESE <- vapply(priors_bias, BayesTools::is.prior.PEESE, logical(1))
  bias_indicator  <- .extract_bias_indicator(x, posterior_samples = posterior_samples)

  if (any(is.na(bias_indicator)) ||
      any(bias_indicator < 1L | bias_indicator > length(priors_bias))) {
    stop("Invalid 'bias_indicator' values in posterior samples.", call. = FALSE)
  }

  selected_rows <- .funnel_subsample_rows(
    bias_indicator = bias_indicator,
    max_samples    = max_samples
  )
  if (!is.null(selected_rows)) {
    posterior_samples <- posterior_samples[selected_rows, , drop = FALSE]
    bias_indicator    <- bias_indicator[selected_rows]
  }
  S <- nrow(posterior_samples)

  mu_samples <- .funnel_mu_samples(x, posterior_samples)

  if (sampling_heterogeneity) {
    tau_samples <- .funnel_tau_samples(x, posterior_samples)
  } else {
    tau_samples <- rep(0, S)
  }

  PET_samples   <- .funnel_posterior_column(posterior_samples, "PET",   S)
  PEESE_samples <- .funnel_posterior_column(posterior_samples, "PEESE", S)

  PET_samples[!branch_is_PET[bias_indicator]]     <- 0
  PEESE_samples[!branch_is_PEESE[bias_indicator]] <- 0

  selection <- .selection_context(
    object            = x,
    posterior_samples = posterior_samples
  )
  use_normal <- if (is.null(selection)) {
    rep(TRUE, S)
  } else {
    selection[["use_normal"]]
  }

  return(list(
    mu                    = mu_samples,
    tau                   = tau_samples,
    PET                   = PET_samples,
    PEESE                 = PEESE_samples,
    bias_indicator        = bias_indicator,
    is_weightfunction     = !use_normal,
    selection             = selection
  ))
}


# ---------------------------------------------------------------------------- #
# .funnel_subsample_rows
# ---------------------------------------------------------------------------- #
#
# Deterministically thin posterior rows for plotting while preserving bias-model
# proportions. Model-averaged funnel contours are visual summaries, and using
# every row from large MCMC fits makes root finding unnecessarily expensive.
#
# ---------------------------------------------------------------------------- #
.funnel_subsample_rows <- function(bias_indicator, max_samples) {

  return(.thin_sample_rows_by_group(
    group       = bias_indicator,
    max_samples = max_samples
  ))
}


# ---------------------------------------------------------------------------- #
# .funnel_mu_samples
# ---------------------------------------------------------------------------- #
#
# Extract pooled location samples without PET/PEESE bias offsets.
#
# ---------------------------------------------------------------------------- #
.funnel_mu_samples <- function(x, posterior_samples) {

  if (!.is_mods(x)) {
    return(as.numeric(posterior_samples[, "mu"]))
  }

  mu_samples <- predict.brma(
    object        = x,
    newdata       = TRUE,
    type          = "terms",
    bias_adjusted = TRUE,
    quiet         = TRUE
  )

  return(as.numeric(as.matrix(mu_samples)[, 1]))
}


# ---------------------------------------------------------------------------- #
# .funnel_tau_samples
# ---------------------------------------------------------------------------- #
#
# Extract total heterogeneity samples for outcome-mode funnel contours.
#
# ---------------------------------------------------------------------------- #
.funnel_tau_samples <- function(x, posterior_samples) {

  tau_result <- .evaluate.brma.tau(
    fit               = x[["fit"]],
    scale_data        = x[["data"]][["scale"]],
    scale_formula     = if (.is_scale(x)) .create_fit_formula_list(data = x[["data"]], "scale") else NULL,
    scale_priors      = x[["priors"]][["scale"]],
    is_scale          = .is_scale(x),
    is_multilevel     = .is_multilevel(x),
    K                 = nrow(x[["data"]][["outcome"]]),
    posterior_samples = posterior_samples
  )

  return(rowMeans(tau_result[["tau_total"]]))
}


# ---------------------------------------------------------------------------- #
# .funnel_posterior_column
# ---------------------------------------------------------------------------- #
#
# Extract an optional posterior column, returning zeros when absent.
#
# ---------------------------------------------------------------------------- #
.funnel_posterior_column <- function(posterior_samples, column, S) {

  if (column %in% colnames(posterior_samples)) {
    return(as.numeric(posterior_samples[, column]))
  }

  return(rep(0, S))
}


# ---------------------------------------------------------------------------- #
# .funnel_model_averaged_quantile
# ---------------------------------------------------------------------------- #
#
# Invert the model-averaged posterior-row CDF at one SE value.
#
# ---------------------------------------------------------------------------- #
.funnel_model_averaged_quantile <- function(se, p, setup, effect_direction) {

  se_setup <- .funnel_model_averaged_se_setup(
    se               = se,
    setup            = setup,
    effect_direction = effect_direction
  )
  total_sd <- se_setup[["total_sd"]]
  location <- se_setup[["location"]]
  eps_sd   <- sqrt(.Machine$double.eps)

  if (all(total_sd < eps_sd)) {
    return(unname(stats::quantile(location, probs = p, names = FALSE, type = 8)))
  }

  spread <- pmax(total_sd, eps_sd)
  lower  <- min(location - 10 * spread, na.rm = TRUE)
  upper  <- max(location + 10 * spread, na.rm = TRUE)

  if (!is.finite(lower) || !is.finite(upper)) {
    return(NA_real_)
  }
  if (lower >= upper) {
    lower <- lower - 1
    upper <- upper + 1
  }

  obj_fun <- function(q) {
    .funnel_model_averaged_cdf_precomputed(q, se_setup) - p
  }

  lower_value <- obj_fun(lower)
  upper_value <- obj_fun(upper)
  step        <- max(spread, na.rm = TRUE)
  if (!is.finite(step) || step <= 0) {
    step <- max(1, abs(location), na.rm = TRUE)
  }

  for (i in seq_len(25)) {
    if (lower_value <= 0 && upper_value >= 0) {
      break
    }
    if (lower_value > 0) {
      lower       <- lower - step
      lower_value <- obj_fun(lower)
    }
    if (upper_value < 0) {
      upper       <- upper + step
      upper_value <- obj_fun(upper)
    }
    step <- step * 2
  }

  if (lower_value > 0 || upper_value < 0) {
    return(.funnel_grid_quantile_precomputed(p, lower, upper, se_setup))
  }

  out <- tryCatch(
    stats::uniroot(obj_fun, interval = c(lower, upper), tol = 1e-6)[["root"]],
    error = function(e) NA_real_
  )

  if (is.na(out)) {
    out <- .funnel_grid_quantile_precomputed(p, lower, upper, se_setup)
  }

  return(out)
}


# ---------------------------------------------------------------------------- #
# .funnel_model_averaged_se_setup
# ---------------------------------------------------------------------------- #
#
# Precompute row quantities that stay constant while inverting quantiles for
# one standard error.
#
# ---------------------------------------------------------------------------- #
.funnel_model_averaged_se_setup <- function(se, setup, effect_direction) {

  total_sd      <- sqrt(se^2 + setup[["tau"]]^2)
  location      <- .funnel_row_location(se, setup, effect_direction)
  eps_sd        <- sqrt(.Machine$double.eps)
  zero_sd       <- total_sd < eps_sd
  normal_rows   <- !setup[["is_weightfunction"]] & !zero_sd
  weighted_rows <- setup[["is_weightfunction"]] & !zero_sd
  selection     <- setup[["selection"]]

  out <- list(
    se               = se,
    total_sd         = total_sd,
    location         = location,
    zero_sd          = zero_sd,
    normal_rows      = normal_rows,
    weighted_rows    = weighted_rows,
    setup            = setup,
    effect_direction = effect_direction
  )

  if (any(weighted_rows) && !is.null(selection)) {
    .selection_require_step_evaluable(selection, ".funnel_model_averaged_cdf()")
  }

  if (any(weighted_rows) && !is.null(selection) &&
      se > 0 && isFALSE(selection[["has_phack"]])) {
    rows <- which(weighted_rows)
    out[["step_selection"]] <- .funnel_step_selection_precompute(
      rows      = rows,
      se        = se,
      location  = location,
      total_sd  = total_sd,
      selection = selection
    )
  }

  return(out)
}


# ---------------------------------------------------------------------------- #
# .funnel_step_selection_precompute
# ---------------------------------------------------------------------------- #
#
# Precompute selected-normal denominators for one standard error.
#
# ---------------------------------------------------------------------------- #
.funnel_step_selection_precompute <- function(rows, se, location, total_sd,
                                              selection) {

  sign          <- selection[["sign"]]
  omega         <- selection[["omega"]][rows, , drop = FALSE]
  mean          <- sign * location[rows]
  sd            <- total_sd[rows]
  n_bins        <- selection[["n_bins"]]
  use_telescope <- isTRUE(selection[["telescope_probabilities"]]) && n_bins > 1L

  if (use_telescope) {
    boundary <- selection[["z_lower"]][seq_len(n_bins - 1L)] * se
    tails    <- vapply(
      boundary,
      stats::pnorm,
      numeric(length(rows)),
      mean       = mean,
      sd         = sd,
      lower.tail = FALSE
    )
    tails      <- matrix(tails, nrow = length(rows), ncol = n_bins - 1L)
    omega_diff <- omega[, seq_len(n_bins - 1L), drop = FALSE] -
      omega[, seq_len(n_bins - 1L) + 1L, drop = FALSE]
    denom <- omega[, n_bins] + rowSums(omega_diff * tails)

    use_telescope <- all(is.finite(denom) & denom > 0)
  }

  if (!use_telescope) {
    denom <- rep(0, length(rows))

    for (b in seq_len(n_bins)) {
      lower <- selection[["z_lower"]][b] * se
      upper <- selection[["z_upper"]][b] * se
      denom <- denom + omega[, b] * .selection_interval_prob_vec(lower, upper, mean, sd)
    }
  }

  return(list(
    rows                    = rows,
    se                      = se,
    sign                    = sign,
    omega                   = omega,
    omega_diff              = if (use_telescope) omega_diff else NULL,
    mean                    = mean,
    sd                      = sd,
    denom                   = denom,
    boundary                = if (use_telescope) boundary else NULL,
    boundary_tail           = if (use_telescope) tails else NULL,
    telescope_probabilities = use_telescope,
    z_lower                 = selection[["z_lower"]],
    z_upper                 = selection[["z_upper"]],
    n_bins                  = n_bins
  ))
}


# ---------------------------------------------------------------------------- #
# .funnel_model_averaged_cdf
# ---------------------------------------------------------------------------- #
#
# Average CDF values across posterior rows at one x-coordinate and SE.
#
# ---------------------------------------------------------------------------- #
.funnel_model_averaged_cdf <- function(q, se, setup, effect_direction) {

  se_setup <- .funnel_model_averaged_se_setup(
    se               = se,
    setup            = setup,
    effect_direction = effect_direction
  )

  return(.funnel_model_averaged_cdf_precomputed(q, se_setup))
}


# ---------------------------------------------------------------------------- #
# .funnel_model_averaged_cdf_precomputed
# ---------------------------------------------------------------------------- #
#
# Average CDF values using quantities precomputed for one standard error.
#
# ---------------------------------------------------------------------------- #
.funnel_model_averaged_cdf_precomputed <- function(q, se_setup) {

  S          <- length(se_setup[["location"]])
  total_sd   <- se_setup[["total_sd"]]
  location   <- se_setup[["location"]]
  cdf_values <- rep(NA_real_, S)

  zero_sd <- se_setup[["zero_sd"]]
  if (any(zero_sd)) {
    cdf_values[zero_sd] <- as.numeric(q >= location[zero_sd])
  }

  normal_rows <- se_setup[["normal_rows"]]
  if (any(normal_rows)) {
    cdf_values[normal_rows] <- stats::pnorm(
      q,
      mean = location[normal_rows],
      sd   = total_sd[normal_rows]
    )
  }

  weighted_rows <- se_setup[["weighted_rows"]]
  if (any(weighted_rows)) {
    if (!is.null(se_setup[["step_selection"]])) {
      step_selection <- se_setup[["step_selection"]]
      cdf_values[step_selection[["rows"]]] <-
        .funnel_step_selected_cdf_precomputed(q, step_selection)
    } else {
      rows <- which(weighted_rows)
      cdf_values[rows] <- .funnel_selected_cdf(
        q                = q,
        rows             = rows,
        se               = se_setup[["se"]],
        total_sd         = total_sd,
        setup            = se_setup[["setup"]],
        effect_direction = se_setup[["effect_direction"]]
      )
    }
  }

  cdf_values <- pmin(pmax(cdf_values, 0), 1)
  return(mean(cdf_values))
}


# ---------------------------------------------------------------------------- #
# .funnel_step_selected_cdf_precomputed
# ---------------------------------------------------------------------------- #
#
# Selected-normal CDF for step-selection rows with precomputed denominators.
#
# ---------------------------------------------------------------------------- #
.funnel_step_selected_cdf_precomputed <- function(q, step_selection) {

  q_signed <- step_selection[["sign"]] * q

  if (isTRUE(step_selection[["telescope_probabilities"]])) {
    q_tail <- stats::pnorm(
      q_signed,
      mean       = step_selection[["mean"]],
      sd         = step_selection[["sd"]],
      lower.tail = FALSE
    )
    q_cdf <- stats::pnorm(
      q_signed,
      mean = step_selection[["mean"]],
      sd   = step_selection[["sd"]]
    )

    if (step_selection[["sign"]] == 1L) {
      mass <- step_selection[["omega"]][, step_selection[["n_bins"]]] * q_cdf

      for (b in seq_len(step_selection[["n_bins"]] - 1L)) {
        if (q_signed > step_selection[["boundary"]][b]) {
          mass <- mass + step_selection[["omega_diff"]][, b] *
            (step_selection[["boundary_tail"]][, b] - q_tail)
        }
      }
    } else {
      mass <- step_selection[["omega"]][, step_selection[["n_bins"]]] * q_tail

      for (b in seq_len(step_selection[["n_bins"]] - 1L)) {
        tail_part <- if (q_signed >= step_selection[["boundary"]][b]) {
          q_tail
        } else {
          step_selection[["boundary_tail"]][, b]
        }
        mass <- mass + step_selection[["omega_diff"]][, b] *
          tail_part
      }
    }

    out <- mass / step_selection[["denom"]]
    out <- pmin(pmax(out, 0), 1)
    return(out)
  }

  mass     <- rep(0, length(step_selection[["rows"]]))

  for (b in seq_len(step_selection[["n_bins"]])) {
    lower <- step_selection[["z_lower"]][b] * step_selection[["se"]]
    upper <- step_selection[["z_upper"]][b] * step_selection[["se"]]

    if (step_selection[["sign"]] == 1L) {
      selected_mass <- .selection_interval_prob_vec(
        lower,
        min(upper, q_signed),
        step_selection[["mean"]],
        step_selection[["sd"]]
      )
    } else {
      selected_mass <- .selection_interval_prob_vec(
        max(lower, q_signed),
        upper,
        step_selection[["mean"]],
        step_selection[["sd"]]
      )
    }

    mass <- mass + step_selection[["omega"]][, b] * selected_mass
  }

  return(mass / step_selection[["denom"]])
}


# ---------------------------------------------------------------------------- #
# .funnel_row_location
# ---------------------------------------------------------------------------- #
#
# Row-specific expected observed location on the original effect-size scale.
#
# ---------------------------------------------------------------------------- #
.funnel_row_location <- function(se, setup, effect_direction) {

  direction <- ifelse(effect_direction == "negative", -1, 1)

  return(
    setup[["mu"]] +
      direction * setup[["PET"]] * se +
      direction * setup[["PEESE"]] * se^2
  )
}


# ---------------------------------------------------------------------------- #
# .funnel_selected_cdf
# ---------------------------------------------------------------------------- #
#
# Selected-normal CDF for selection-model posterior rows.
#
# ---------------------------------------------------------------------------- #
.funnel_selected_cdf <- function(q, rows, se, total_sd, setup, effect_direction) {

  selection <- setup[["selection"]]

  if (is.null(selection)) {
    return(stats::pnorm(q, mean = setup[["mu"]][rows], sd = total_sd[rows]))
  }

  .selection_require_step_evaluable(selection, ".funnel_selected_cdf()")

  if (se <= 0) {
    return(.funnel_selected_cdf_zero_se(
      q        = q,
      rows     = rows,
      total_sd = total_sd,
      setup    = setup
    ))
  }

  local_context <- .selection_context_subset_rows(selection, rows)
  cdf           <- .selection_step_cdf_matrix(
    q                 = q,
    mean              = matrix(setup[["mu"]][rows], ncol = 1),
    sd                = matrix(total_sd[rows], ncol = 1),
    sei               = se,
    selection_context = local_context,
    lower.tail        = TRUE
  )

  return(as.numeric(cdf[, 1]))
}


# ---------------------------------------------------------------------------- #
# .funnel_selected_cdf_zero_se
# ---------------------------------------------------------------------------- #
#
# Limit of the selected-normal CDF as the standard error approaches zero.
#
# Selection bins are defined on signed z = sign * y / se. At se -> 0, finite
# z-bin intervals collapse to zero mass. Only the two tail intervals remain:
# the upper z tail maps to signed effects above 0, and the lower z tail maps to
# signed effects below 0. This explicit limit avoids evaluating the native
# selected-normal kernel at the singular se = 0 boundary.
#
# ---------------------------------------------------------------------------- #
.funnel_selected_cdf_zero_se <- function(q, rows, total_sd, setup) {

  selection <- setup[["selection"]]
  sign      <- selection[["sign"]]
  omega     <- selection[["omega"]][rows, , drop = FALSE]
  mean      <- sign * setup[["mu"]][rows]
  sd        <- total_sd[rows]
  q_signed  <- sign * q
  S         <- length(rows)

  denom <- rep(0, S)
  mass  <- rep(0, S)

  for (b in seq_len(selection[["n_bins"]])) {
    z_lower <- selection[["z_lower"]][b]
    z_upper <- selection[["z_upper"]][b]

    if (is.infinite(z_lower) && z_lower < 0 &&
        is.infinite(z_upper) && z_upper > 0) {
      lower <- -Inf
      upper <-  Inf
    } else if (is.infinite(z_upper) && z_upper > 0) {
      lower <- 0
      upper <- Inf
    } else if (is.infinite(z_lower) && z_lower < 0) {
      lower <- -Inf
      upper <- 0
    } else {
      next
    }

    bin_mass <- .selection_interval_prob_vec(lower, upper, mean, sd)
    denom    <- denom + omega[, b] * bin_mass

    if (sign == 1L) {
      selected_mass <- .selection_interval_prob_vec(lower, min(upper, q_signed), mean, sd)
    } else {
      selected_mass <- .selection_interval_prob_vec(max(lower, q_signed), upper, mean, sd)
    }
    mass <- mass + omega[, b] * selected_mass
  }

  out <- mass / denom
  out <- pmin(pmax(out, 0), 1)
  return(out)
}


# ---------------------------------------------------------------------------- #
# .funnel_grid_quantile
# ---------------------------------------------------------------------------- #
#
# Fallback inversion for discontinuous CDFs from degenerate posterior rows.
#
# ---------------------------------------------------------------------------- #
.funnel_grid_quantile <- function(p, lower, upper, se, setup, effect_direction) {

  se_setup <- .funnel_model_averaged_se_setup(
    se               = se,
    setup            = setup,
    effect_direction = effect_direction
  )

  return(.funnel_grid_quantile_precomputed(p, lower, upper, se_setup))
}


# ---------------------------------------------------------------------------- #
# .funnel_grid_quantile_precomputed
# ---------------------------------------------------------------------------- #
#
# Fallback inversion using a precomputed standard-error context.
#
# ---------------------------------------------------------------------------- #
.funnel_grid_quantile_precomputed <- function(p, lower, upper, se_setup) {

  grid <- seq(lower, upper, length.out = 1000)
  cdf  <- vapply(
    grid,
    .funnel_model_averaged_cdf_precomputed,
    numeric(1),
    se_setup = se_setup
  )
  index <- which(cdf >= p)[1]

  if (is.na(index)) {
    return(grid[length(grid)])
  }

  return(grid[index])
}


# ---------------------------------------------------------------------------- #
# .normalize_funnel_max_samples
# ---------------------------------------------------------------------------- #
#
# Validate max posterior samples used by model-averaged funnel contours.
#
# ---------------------------------------------------------------------------- #
.normalize_funnel_max_samples <- function(max_samples) {

  return(.normalize_max_samples(max_samples, "max_samples"))
}


# ---------------------------------------------------------------------------- #
# .set_dots_funnel
# ---------------------------------------------------------------------------- #
#
# Process dots arguments for funnel plot with sensible defaults.
#
# Sets defaults for all graphical parameters that can be customized in the
# funnel plot. Users can override any of these by passing them in ...
#
# @param ... additional graphical arguments
#
# @return list of graphical parameters with defaults applied
#
# ---------------------------------------------------------------------------- #
.set_dots_funnel <- function(dots) {

  # point styling (use metafor-style argument names)
  dots <- .plot_point_style_defaults(dots)

  # funnel region styling
  if (is.null(dots[["back"]])) dots[["back"]] <- "grey" # background fill

  if (is.null(dots[["shade"]])) dots[["shade"]] <- "white" # funnel fill

  # line styling
  if (is.null(dots[["lty"]])) dots[["lty"]] <- "dotted" # line type
  if (is.null(dots[["col.line"]])) dots[["col.line"]] <- "black" # funnel edge color

  # reference line (position set in data functions based on mode)
  # if user provides refline, it will be passed through to override
  if (is.null(dots[["col.refline"]])) dots[["col.refline"]] <- "black" # reference line color

  # axis labels (now set separately in data functions)
  # main title (NULL = no title by default)
  if (is.null(dots[["main"]])) dots[["main"]] <- NULL
  if (is.null(dots[["las"]]))  dots[["las"]]  <- 1

  # axis limits (NULL = auto)
  # xlim and ylim are left as NULL if not provided (auto-computed)

  # data return flag
  if (is.null(dots[["as_data"]])) dots[["as_data"]] <- FALSE

  return(dots)
}

context("Funnel plot")

# Load common test helpers
source(testthat::test_path("common-functions.R"))
source(testthat::test_path("helper-test-matrix.R"))
source(testthat::test_path("helper-visuals.R"))

# list cached fits lazily
skip_if_no_fits()
skip_if_not_installed("metafor")
fit_names <- list_fits()
fits      <- lazy_fits(fit_names, validate = FALSE)
info      <- lazy_infos(fit_names, validate = FALSE)

.test_funnel <- function(..., max_samples = 1000) {

  funnel(..., max_samples = max_samples)
}


# ============================================================================ #
# Test: Simple Meta-Analysis Funnel Plot
# ============================================================================ #

test_that("Funnel plot for simple meta-analysis matches metafor structure", {

  name        <- "bcg_meta-analysis"
  fit_metafor <- info[[name]][["metafor"]]
  fit_brma    <- fits[[name]]

  # --------------------------------------------------
  # Visual comparison: side-by-side plots
  # --------------------------------------------------

  expect_vdiffr_snapshot("funnel_simple_comparison_no_tau", function() {
    oldpar <- graphics::par(no.readonly = TRUE)
    on.exit(graphics::par(mfrow = oldpar[["mfrow"]], mar = oldpar[["mar"]]))
    par(mfrow = c(1, 2), mar = c(4, 4, 2, 1))
    metafor::funnel(fit_metafor, main = "metafor", xlim = c(-3, 3), ylim = c(0, 0.8))
    .test_funnel(fit_brma, plot_type = "base", xlim = c(-3, 3), ylim = c(0, 0.8), main = "brma", sampling_heterogeneity = FALSE)
  })

  expect_vdiffr_snapshot("funnel_simple_comparison", function() {
    oldpar <- graphics::par(no.readonly = TRUE)
    on.exit(graphics::par(mfrow = oldpar[["mfrow"]], mar = oldpar[["mar"]]))
    par(mfrow = c(1, 2), mar = c(4, 4, 2, 1))
    metafor::funnel(fit_metafor, main = "metafor", addtau2 = TRUE, xlim = c(-3, 3), ylim = c(0, 0.8))
    .test_funnel(fit_brma, plot_type = "base", xlim = c(-3, 3), ylim = c(0, 0.8), main = "brma")
  })

  expect_vdiffr_snapshot(
    "funnel_simple_brma_ggplot",
    .test_funnel(fit_brma, plot_type = "ggplot")
  )
})

# ============================================================================ #
# Test: Meta-Regression Funnel Plot
# ============================================================================ #

test_that("Funnel plot for meta-regression matches metafor residual views", {

  name        <- "bcg_meta-regression"
  fit_metafor <- info[[name]][["metafor"]]
  fit_brma    <- fits[[name]]

  # --------------------------------------------------
  # Visual comparison: side-by-side plots
  # --------------------------------------------------

  expect_vdiffr_snapshot("funnel_regression_comparison-standard", function() {
    oldpar <- graphics::par(no.readonly = TRUE)
    on.exit(graphics::par(mfrow = oldpar[["mfrow"]], mar = oldpar[["mar"]]))
    par(mfrow = c(1, 2), mar = c(4, 4, 2, 1))
    metafor::funnel(fit_metafor, main = "metafor", ylim = c(0, 0.8), xlim = c(-2, 2), type = "rstandard")
    .test_funnel(fit_brma, plot_type = "base", main = "brma", ylim = c(0, 0.8), xlim = c(-2, 2), type = "rstandard")
  })

  expect_vdiffr_snapshot("funnel_regression_comparison-rstudent", function() {
    oldpar <- graphics::par(no.readonly = TRUE)
    on.exit(graphics::par(mfrow = oldpar[["mfrow"]], mar = oldpar[["mar"]]))
    par(mfrow = c(1, 2), mar = c(4, 4, 2, 1))
    metafor::funnel(fit_metafor, main = "metafor", ylim = c(0, 0.8), xlim = c(-2, 2), type = "rstudent")
    suppressWarnings(.test_funnel(fit_brma, plot_type = "base", main = "brma", ylim = c(0, 0.8), xlim = c(-2, 2), type = "rstudent"))
  })

  expect_vdiffr_snapshot(
    "funnel_regression_brma_ggplot",
    suppressWarnings(.test_funnel(fit_brma, plot_type = "ggplot"))
  )
})

test_that("Funnel plot for interaction meta-regression renders residual views", {

  skip_if_not_full_visuals("Interaction funnel variants duplicate the core meta-regression visual.")

  name        <- "bcg_meta-regression4"
  fit_metafor <- info[[name]][["metafor"]]
  fit_brma    <- fits[[name]]

  # --------------------------------------------------
  # Visual comparison: side-by-side plots
  # --------------------------------------------------

  expect_vdiffr_snapshot("funnel_regression4_comparison-standard", function() {
    oldpar <- graphics::par(no.readonly = TRUE)
    on.exit(graphics::par(mfrow = oldpar[["mfrow"]], mar = oldpar[["mar"]]))
    par(mfrow = c(1, 2), mar = c(4, 4, 2, 1))
    metafor::funnel(fit_metafor, main = "metafor", ylim = c(0, 0.8), xlim = c(-2, 2), type = "rstandard")
    suppressWarnings(.test_funnel(fit_brma, plot_type = "base", main = "brma", ylim = c(0, 0.8), xlim = c(-2, 2), type = "rstandard"))
  })

  expect_vdiffr_snapshot("funnel_regression4_comparison-rstudent", function() {
    oldpar <- graphics::par(no.readonly = TRUE)
    on.exit(graphics::par(mfrow = oldpar[["mfrow"]], mar = oldpar[["mar"]]))
    par(mfrow = c(1, 2), mar = c(4, 4, 2, 1))
    metafor::funnel(fit_metafor, main = "metafor", ylim = c(0, 1.2), xlim = c(-2, 2), type = "rstudent")
    suppressWarnings(.test_funnel(fit_brma, plot_type = "base", main = "brma", ylim = c(0, 1.2), xlim = c(-2, 2), type = "rstudent"))
  })
})

# ============================================================================ #
# Test: Location-Scale Model Funnel Plot
# ============================================================================ #

test_that("Funnel plot for location-scale model matches metafor residual view", {

  skip_if_not_full_visuals("Location-scale funnel variants are gallery coverage.")

  name <- "bangertdrowns2004_location-scale"
  fit_metafor <- info[[name]][["metafor"]]
  fit_brma    <- fits[[name]]

  # --------------------------------------------------
  # Visual comparison: side-by-side plots
  # --------------------------------------------------

  expect_vdiffr_snapshot("funnel_scale_comparison", function() {
    oldpar <- graphics::par(no.readonly = TRUE)
    on.exit(graphics::par(mfrow = oldpar[["mfrow"]], mar = oldpar[["mar"]]))
    par(mfrow = c(1, 2), mar = c(4, 4, 2, 1))
    metafor::funnel(fit_metafor, main = "metafor", ylim = c(0, 0.6), type = "rstandard")
    .test_funnel(fit_brma, plot_type = "base", main = "brma", ylim = c(0, 0.6), sampling_heterogeneity = FALSE, type = "rstandard")
  })

  expect_vdiffr_snapshot(
    "funnel_scale_brma_ggplot",
    .test_funnel(fit_brma, plot_type = "ggplot")
  )
})

# ============================================================================ #
# Test: 3-Level Model Funnel Plot
# ============================================================================ #

test_that("Funnel plot for 3-level model matches metafor structure", {

  name <- "konstantopoulos2011_3lvl"
  fit_metafor <- info[[name]][["metafor"]]
  fit_brma    <- fits[[name]]

  # --------------------------------------------------
  # Visual comparison: side-by-side plots
  # --------------------------------------------------

  expect_vdiffr_snapshot("funnel_3lvl_comparison", function() {
    oldpar <- graphics::par(no.readonly = TRUE)
    on.exit(graphics::par(mfrow = oldpar[["mfrow"]], mar = oldpar[["mar"]]))
    par(mfrow = c(1, 2), mar = c(4, 4, 2, 1))
    metafor::funnel(fit_metafor, main = "metafor", xlim = c(-1, 1.5), ylim = c(0.4, 0))
    .test_funnel(fit_brma, plot_type = "base", main = "brma", sampling_heterogeneity = FALSE, xlim = c(-1, 1.5), ylim = c(0.4, 0))
  })

  expect_vdiffr_snapshot(
    "funnel_3lvl_brma_ggplot",
    .test_funnel(fit_brma, plot_type = "ggplot")
  )
})

test_that("Funnel plot for 3-level meta-regression renders residual views", {

  skip_if_not_full_visuals("3-level meta-regression duplicates the default multilevel funnel visual.")

  name <- "konstantopoulos2011_3lvl2"
  fit_metafor <- info[[name]][["metafor"]]
  fit_brma    <- fits[[name]]

  # --------------------------------------------------
  # Visual comparison: side-by-side plots
  # --------------------------------------------------

  expect_vdiffr_snapshot("funnel_3lvl2_comparison", function() {
    oldpar <- graphics::par(no.readonly = TRUE)
    on.exit(graphics::par(mfrow = oldpar[["mfrow"]], mar = oldpar[["mar"]]))
    par(mfrow = c(1, 2), mar = c(4, 4, 2, 1))
    metafor::funnel(fit_metafor, main = "metafor", xlim = c(-1, 1.5), ylim = c(0.5, 0), type = "rstandard")
    suppressWarnings(suppressWarnings(.test_funnel(fit_brma, plot_type = "base", type = "rstandard", conditioning_depth = "marginal", main = "brma", sampling_heterogeneity = FALSE, xlim = c(-1, 1.5), ylim = c(0.5, 0))))
  })

  expect_vdiffr_snapshot(
    "funnel_3lvl2_brma_ggplot",
    suppressWarnings(.test_funnel(fit_brma, plot_type = "ggplot"))
  )
})

# ============================================================================ #
# Test: GLMM Model Funnel Plot
# ============================================================================ #

test_that("Funnel plot for GLMM model renders ggplot output", {

  name <- "nielweise2008_glmm"
  fit_brma <- fits[[name]]

  # there is no funnel plot for metafor
  expect_vdiffr_snapshot(
    "funnel_glmm_ggplot",
    .test_funnel(fit_brma, plot_type = "ggplot")
  )
})

test_that("Funnel plot for GLMM meta-regression renders ggplot output", {

  skip_if_not_full_visuals("GLMM meta-regression duplicates the default GLMM funnel visual.")

  name <- "bcg_glmm_reg"
  fit_brma <- fits[[name]]

  # there is no funnel plot for metafor
  expect_vdiffr_snapshot(
    "funnel_glmm_reg_ggplot",
    suppressWarnings(.test_funnel(fit_brma, plot_type = "ggplot"))
  )
})

# ============================================================================ #
# Test: Selection Model Funnel Plot
# ============================================================================ #

test_that("Funnel plot for selection model matches metafor structure", {

  name <- "dat.lehmann2018-3PSM"
  fit_metafor <- info[[name]][["metafor"]]
  fit_brma    <- fits[[name]]

  # --------------------------------------------------
  # Visual comparison: side-by-side plots
  # --------------------------------------------------

  expect_vdiffr_snapshot("funnel_selmodel_pos_comparison", function() {
    oldpar <- graphics::par(no.readonly = TRUE)
    on.exit(graphics::par(mfrow = oldpar[["mfrow"]], mar = oldpar[["mar"]]))
    par(mfrow = c(1, 2), mar = c(4, 4, 2, 1))
    metafor::funnel(fit_metafor, main = "metafor", xlim = c(-2, 2), ylim = c(0.8, 0))
    .test_funnel(fit_brma, plot_type = "base", main = "brma", xlim = c(-2, 2), ylim = c(0.8, 0), sampling_bias = FALSE, sampling_heterogeneity = FALSE)
  })

  expect_vdiffr_snapshot(
    "funnel_selmodel_pos_brma_ggplot",
    .test_funnel(fit_brma, plot_type = "ggplot", sampling_bias = TRUE, sampling_heterogeneity = TRUE, xlim = c(-2, 2))
  )
})

test_that("Funnel plot for selection meta-regression renders residual view", {

  skip_if_not_full_visuals("Selection meta-regression duplicates the default selection funnel visual.")

  name <- "dat.lehmann2018-3PSMreg"
  fit_metafor <- info[[name]][["metafor"]]
  fit_brma    <- fits[[name]]

  # --------------------------------------------------
  # Visual comparison: side-by-side plots
  # --------------------------------------------------

  expect_vdiffr_snapshot("funnel_selmodelreg_comparison", function() {
    oldpar <- graphics::par(no.readonly = TRUE)
    on.exit(graphics::par(mfrow = oldpar[["mfrow"]], mar = oldpar[["mar"]]))
    # not available for metafor
    # par(mfrow = c(1, 2), mar = c(4, 4, 2, 1))
    # metafor::funnel(fit_metafor, main = "metafor", xlim = c(-2, 2), ylim = c(0.8, 0))
    suppressWarnings(.test_funnel(fit_brma, plot_type = "base", main = "brma", xlim = c(-2, 2), ylim = c(0.8, 0), sampling_bias = FALSE, sampling_heterogeneity = FALSE))
  })
})

test_that("Funnel plot for negative-direction selection model matches metafor structure", {

  skip_if_not_full_visuals("Negative-direction selection is gallery coverage.")

  name <- "dat.lehmann2018-3PSM_neg"
  fit_metafor <- info[[name]][["metafor"]]
  fit_brma    <- fits[[name]]

  # --------------------------------------------------
  # Visual comparison: side-by-side plots
  # --------------------------------------------------

  expect_vdiffr_snapshot("funnel_selmodel_neg_comparison", function() {
    oldpar <- graphics::par(no.readonly = TRUE)
    on.exit(graphics::par(mfrow = oldpar[["mfrow"]], mar = oldpar[["mar"]]))
    par(mfrow = c(1, 2), mar = c(4, 4, 2, 1))
    metafor::funnel(fit_metafor, main = "metafor", xlim = c(-2, 2), ylim = c(0.8, 0))
    .test_funnel(fit_brma, plot_type = "base", main = "brma", xlim = c(-2, 2), ylim = c(0.8, 0), sampling_bias = FALSE, sampling_heterogeneity = FALSE)
  })
})

# ============================================================================ #
# Test: PET Model Funnel Plot
# ============================================================================ #

test_that("Funnel plot for PET model matches metafor residual view", {

  name <- "dat.lehmann2018-PET"
  fit_metafor <- info[[name]][["metafor"]]
  fit_brma    <- fits[[name]]

  # --------------------------------------------------
  # Visual comparison: side-by-side plots
  # --------------------------------------------------

  expect_vdiffr_snapshot("funnel_PET_pos_comparison", function() {
    oldpar <- graphics::par(no.readonly = TRUE)
    on.exit(graphics::par(mfrow = oldpar[["mfrow"]], mar = oldpar[["mar"]]))
    par(mfrow = c(1, 2), mar = c(4, 4, 2, 1))
    # the residuals need to be selected specifically because bPET is not treated as a regression
    metafor::funnel(fit_metafor, main = "metafor", xlim = c(-2, 2), ylim = c(0.8, 0), type = "rstudent")
    suppressWarnings(.test_funnel(fit_brma, plot_type = "base", sampling_bias = FALSE, sampling_heterogeneity = FALSE, residual = TRUE, type = "rstudent", xlim = c(-2, 2), ylim = c(0.8, 0)))
  })

  expect_vdiffr_snapshot(
    "funnel_PET_pos_brma_ggplot",
    .test_funnel(fit_brma, plot_type = "ggplot")
  )
})

test_that("Funnel plot for PET meta-regression matches metafor residual view", {

  skip_if_not_full_visuals("PET meta-regression duplicates the default PET funnel visual.")

  name <- "dat.lehmann2018-PETreg"
  fit_metafor <- info[[name]][["metafor"]]
  fit_brma    <- fits[[name]]

  # --------------------------------------------------
  # Visual comparison: side-by-side plots
  # --------------------------------------------------

  expect_vdiffr_snapshot("funnel_PETreg_comparison", function() {
    oldpar <- graphics::par(no.readonly = TRUE)
    on.exit(graphics::par(mfrow = oldpar[["mfrow"]], mar = oldpar[["mar"]]))
    par(mfrow = c(1, 2), mar = c(4, 4, 2, 1))
    # the residuals need to be selected specifically because bPET is not treated as a regression
    metafor::funnel(fit_metafor, main = "metafor", xlim = c(-2, 2), ylim = c(0.8, 0), type = "rstudent")
    suppressWarnings(.test_funnel(fit_brma, plot_type = "base", sampling_bias = FALSE, sampling_heterogeneity = FALSE, residual = TRUE, type = "rstudent", xlim = c(-2, 2), ylim = c(0.8, 0)))
  })
})

test_that("Funnel plot for negative-direction PET model matches metafor residual view", {

  skip_if_not_full_visuals("Negative-direction PET is gallery coverage.")

  name <- "dat.lehmann2018-PET_neg"
  fit_metafor <- info[[name]][["metafor"]]
  fit_brma    <- fits[[name]]

  # --------------------------------------------------
  # Visual comparison: side-by-side plots
  # --------------------------------------------------

  expect_vdiffr_snapshot("funnel_PET_neg_comparison", function() {
    oldpar <- graphics::par(no.readonly = TRUE)
    on.exit(graphics::par(mfrow = oldpar[["mfrow"]], mar = oldpar[["mar"]]))
    par(mfrow = c(1, 2), mar = c(4, 4, 2, 1))
    # the residuals need to be selected specifically because bPET is not treated as a regression
    metafor::funnel(fit_metafor, main = "metafor", xlim = c(-2, 2), ylim = c(0.8, 0), type = "rstudent")
    suppressWarnings(.test_funnel(fit_brma, plot_type = "base", sampling_bias = FALSE, sampling_heterogeneity = FALSE, residual = TRUE, type = "rstudent", xlim = c(-2, 2), ylim = c(0.8, 0)))
  })

  expect_vdiffr_snapshot(
    "funnel_PET_neg_brma_ggplot",
    .test_funnel(fit_brma, plot_type = "ggplot")
  )
})

# ============================================================================ #
# Test: BMA.norm Model Funnel Plot
# ============================================================================ #

test_that("Funnel plot for BMA.norm model renders base output", {

  name     <- "dat.lehmann2018_BMA.norm"
  fit_brma <- fits[[name]]

  expect_vdiffr_snapshot("funnel_BMA", function() {
    suppressWarnings(.test_funnel(fit_brma, plot_type = "base", sampling_heterogeneity = TRUE))
  })
})

test_that("Funnel plot for BMA.norm meta-regression renders base output", {

  skip_if_not_full_visuals("BMA meta-regression duplicates the default BMA funnel smoke test.")

  name     <- "dat.lehmann2018_BMA.norm_mods"
  fit_brma <- fits[[name]]

  expect_vdiffr_snapshot("funnel_BMAreg", function() {
    suppressWarnings(.test_funnel(fit_brma, plot_type = "base", sampling_heterogeneity = TRUE))
  })
})

# ============================================================================ #
# Test: BMA.glmm Model Funnel Plot
# ============================================================================ #

test_that("Funnel plot for BMA.glmm model renders base output", {

  name     <- "bcg_BMA.glmm_3lvl_location_scale"
  fit_brma <- fits[[name]]

  expect_vdiffr_snapshot("funnel_BMA.glmm", function() {
    suppressWarnings(.test_funnel(fit_brma, plot_type = "base"))
  })
})

# ============================================================================ #
# Test: RoBMA Model Funnel Plot
# ============================================================================ #

test_that("Funnel plot for RoBMA model renders base output", {

  name     <- "dat.lehmann2018_RoBMA"
  fit_brma <- fits[[name]]

  expect_vdiffr_snapshot("funnel_RoBMA", function() {
    suppressWarnings(.test_funnel(fit_brma, plot_type = "base", sampling_heterogeneity = TRUE, sampling_bias = TRUE))
  })
})

test_that("Funnel plot for RoBMA meta-regression renders LOO-PIT output", {

  skip_if_not_full_visuals("RoBMA meta-regression duplicates the default RoBMA funnel smoke test.")

  name     <- "dat.lehmann2018_RoBMA_3lvl_mods_scale"
  fit_brma <- fits[[name]]

  expect_vdiffr_snapshot("funnel_RoBMA_complex", function() {
    suppressWarnings(.test_funnel(fit_brma, plot_type = "base", type = "LOO-PIT"))
  })
})

# ============================================================================ #
# Test: Funnel Plot Options
# ============================================================================ #

test_that("Funnel plot data and argument validation are stable", {

  name <- "bcg_meta-analysis"
  fit_brma <- fits[[name]]

  # --------------------------------------------------
  # Test as_data = TRUE returns list with expected components
  # --------------------------------------------------

  funnel_data <- .test_funnel(fit_brma, as_data = TRUE)

  expect_true(is.list(funnel_data),
    info = "as_data = TRUE returns a list"
  )

  expected_components <- c(
    "points", "funnel", "funnel_edge1", "funnel_edge2",
    "background", "x_range", "y_range"
  )
  expect_true(all(expected_components %in% names(funnel_data)),
    info = "funnel data contains all expected components"
  )

  # Check points data.frame structure
  expect_true(is.data.frame(funnel_data$points),
    info = "points are returned as a data.frame"
  )
  expect_true(all(c("x", "y") %in% names(funnel_data$points)),
    info = "points contain x and y columns"
  )

  # Check number of points matches number of studies
  n_studies <- nrow(fit_brma$data$outcome)
  expect_equal(nrow(funnel_data$points), n_studies,
    info = "number of points matches number of studies"
  )

  # --------------------------------------------------
  # Test error on invalid plot_type
  # --------------------------------------------------

  expect_error(.test_funnel(fit_brma, plot_type = "invalid"),
    info = "invalid plot_type is rejected"
  )

  # --------------------------------------------------
  # Test error on invalid sampling_heterogeneity
  # --------------------------------------------------

  expect_error(.test_funnel(fit_brma, sampling_heterogeneity = "yes"),
    info = "invalid sampling_heterogeneity is rejected"
  )

  expect_no_error(
    .test_funnel(
      fit_brma,
      residual           = FALSE,
      type               = "not-a-residual-type",
      unit               = "not-a-unit",
      conditioning_depth = "not-a-depth",
      as_data            = TRUE
    )
  )

  expect_error(
    .test_funnel(fit_brma, residual = TRUE, type = "not-a-residual-type", as_data = TRUE),
    info = "residual mode validates residual type"
  )
})

# ============================================================================ #
# Test: Funnel Plot Customization
# ============================================================================ #

test_that("Funnel plot customization snapshots are stable", {

  skip_if_not_full_visuals("Customization snapshots are visual-gallery coverage.")

  name <- "bcg_meta-analysis"
  fit_brma <- fits[[name]]

  # --------------------------------------------------
  # Test custom point aesthetics
  # --------------------------------------------------

  expect_vdiffr_snapshot("funnel_custom_points_base", function() {
    .test_funnel(fit_brma, plot_type = "base", pch = 21, col = "blue", bg = "lightblue", cex = 1.5)
  })

  expect_vdiffr_snapshot(
    "funnel_custom_points_ggplot",
    .test_funnel(fit_brma, plot_type = "ggplot", pch = 19, col = "blue", bg = "lightblue", size = 3)
  )

  # --------------------------------------------------
  # Test custom funnel region styling
  # --------------------------------------------------

  expect_vdiffr_snapshot("funnel_custom_regions_base", function() {
    .test_funnel(fit_brma, plot_type = "base", back = "lightgrey", shade = "lightyellow", lty = "dashed")
  })

  expect_vdiffr_snapshot(
    "funnel_custom_regions_ggplot",
    .test_funnel(fit_brma, plot_type = "ggplot", back = "lightblue", shade = "lightyellow", lty = "dashed")
  )

  # --------------------------------------------------
  # Test suppressing background/shade
  # --------------------------------------------------

  expect_vdiffr_snapshot("funnel_no_background_base", function() {
    .test_funnel(fit_brma, plot_type = "base", back = NA, shade = "white")
  })

  expect_vdiffr_snapshot(
    "funnel_no_shade_ggplot",
    .test_funnel(fit_brma, plot_type = "ggplot", back = "grey", shade = NA)
  )

  # --------------------------------------------------
  # Test custom axis labels and title
  # --------------------------------------------------

  expect_vdiffr_snapshot("funnel_custom_labels_base", function() {
    .test_funnel(fit_brma, plot_type = "base", xlab = "Effect Size Residual", ylab = "SE", main = "Funnel Plot")
  })

  expect_vdiffr_snapshot(
    "funnel_custom_labels_ggplot",
    .test_funnel(fit_brma, plot_type = "ggplot", xlab = "Effect Size Residual", ylab = "SE", main = "Funnel Plot")
  )

  # --------------------------------------------------
  # Test line color customization
  # --------------------------------------------------

  expect_vdiffr_snapshot("funnel_custom_lines_base", function() {
    .test_funnel(fit_brma, plot_type = "base", col.line = "darkgrey", col.refline = "red", lty = "solid")
  })

  expect_vdiffr_snapshot(
    "funnel_custom_lines_ggplot",
    .test_funnel(fit_brma, plot_type = "ggplot", col.line = "darkgrey", col.refline = "red", lty = "solid")
  )
})

context("Plot point defaults")

.expect_metafor_point_style <- function(dots) {

  expect_identical(dots[["pch"]], 21)
  expect_identical(dots[["col"]], "black")
  expect_identical(dots[["bg"]], "#A6A6A6")
  expect_identical(dots[["cex"]], 1)
  expect_identical(dots[["size"]], 2)
}

test_that("point plotting defaults use metafor-style markers", {

  .expect_metafor_point_style(.plot_point_style_defaults(list()))
  .expect_metafor_point_style(.set_dots_funnel(list()))
  .expect_metafor_point_style(.set_dots_qqnorm(list()))
  .expect_metafor_point_style(.set_dots_radial())
  .expect_metafor_point_style(.set_dots_regplot())
})

test_that("point plotting defaults respect overrides", {

  custom <- list(
    pch  = 19,
    col  = "blue",
    bg   = "lightblue",
    cex  = 1.5,
    size = 3
  )

  expect_equal(.plot_point_style_defaults(custom)[names(custom)], custom)
  expect_equal(.set_dots_funnel(custom)[names(custom)], custom)
  expect_equal(.set_dots_qqnorm(custom)[names(custom)], custom)
  expect_equal(do.call(.set_dots_radial, custom)[names(custom)], custom)
  expect_equal(do.call(.set_dots_regplot, custom)[names(custom)], custom)
})

# ============================================================================ #
# effect-transformations.R
# ============================================================================ #
#
# Effect-size transformations for posterior draws and plotting.
#
# Measures follow metafor naming: OR, RR, HR, and IRR are log-scale measures.
# Use transform = "EXP" to exponentiate a log-scale measure for display.
#
# Transformation objects use the BayesTools convention:
# - fun: forward transformation
# - inv: inverse transformation
# - jac: derivative of fun(x) with respect to x
#
# ============================================================================ #


.effect_known_measures <- function() {

  return(c("SMD", "COR", "ZCOR", "OR", "RR", "HR", "IRR", "RD", "GEN"))
}

.effect_core_measures <- function() {

  return(c("SMD", "COR", "ZCOR", "OR"))
}

.effect_log_measures <- function() {

  return(c("OR", "RR", "HR", "IRR"))
}

.effect_measure_aliases <- function() {

  return(c(
    "D"          = "SMD",
    "COHENS_D"   = "SMD",
    "COHENSD"    = "SMD",
    "R"          = "COR",
    "CORRELATION" = "COR",
    "FISHERS_Z"  = "ZCOR",
    "FISHERSZ"   = "ZCOR",
    "FISHER_Z"   = "ZCOR",
    "FISHERZ"    = "ZCOR",
    "Z"          = "ZCOR",
    "LOGOR"      = "OR",
    "LOGRR"      = "RR",
    "LOGHR"      = "HR",
    "LOGIRR"     = "IRR",
    "NONE"       = "GEN"
  ))
}

.normalize_effect_measure <- function(measure, argument, allow_NULL = FALSE) {

  if (is.null(measure)) {
    if (allow_NULL) {
      return(NULL)
    }
    stop("The '", argument, "' argument must not be NULL.", call. = FALSE)
  }

  BayesTools::check_char(measure, argument, check_length = 1, allow_NA = FALSE)

  measure_key <- toupper(gsub("[^A-Za-z0-9]", "", measure))
  aliases     <- .effect_measure_aliases()

  if (measure_key %in% names(aliases)) {
    measure_key <- aliases[[measure_key]]
  }

  if (!measure_key %in% .effect_known_measures()) {
    stop(
      "Unknown effect-size measure in '", argument, "'. Available measures are: ",
      paste(.effect_known_measures(), collapse = ", "), ".",
      call. = FALSE
    )
  }

  return(measure_key)
}

.normalize_effect_transform <- function(transform) {

  if (is.null(transform)) {
    return("identity")
  }

  BayesTools::check_char(transform, "transform", check_length = 1, allow_NA = FALSE)

  transform_key <- toupper(gsub("[^A-Za-z0-9]", "", transform))
  transform_map <- c(
    NONE     = "identity",
    IDENTITY = "identity",
    ID       = "identity",
    EXP      = "EXP"
  )
  if (!transform_key %in% names(transform_map)) {
    stop(
      "Unknown 'transform'. Available options are 'EXP' and 'identity'.",
      call. = FALSE
    )
  }

  return(unname(transform_map[[transform_key]]))
}

.identity_effect_transformation <- function() {

  return(list(
    fun = function(x) x,
    inv = function(x) x,
    jac = function(x) rep(1, length(x))
  ))
}

.exp_effect_transformation <- function() {

  return(list(
    fun = exp,
    inv = log,
    jac = exp
  ))
}

.d_to_cor <- function(x) {

  out <- x / sqrt(x^2 + 4)

  out[is.infinite(x) & x > 0] <-  1
  out[is.infinite(x) & x < 0] <- -1

  return(out)
}

.cor_to_z <- function(x) {

  x <- pmin(pmax(x, -1), 1)

  return(atanh(x))
}

.effect_basic_transformations <- function() {

  return(list(
    "SMD->COR" = list(
      fun = .d_to_cor,
      inv = function(x) 2 * x / sqrt(1 - x^2),
      jac = function(x) 4 / (x^2 + 4)^(3 / 2)
    ),
    "COR->SMD" = list(
      fun = function(x) 2 * x / sqrt(1 - x^2),
      inv = .d_to_cor,
      jac = function(x) 2 / (1 - x^2)^(3 / 2)
    ),
    "COR->ZCOR" = list(
      fun = .cor_to_z,
      inv = tanh,
      jac = function(x) 1 / (1 - x^2)
    ),
    "ZCOR->COR" = list(
      fun = tanh,
      inv = .cor_to_z,
      jac = function(x) 1 - tanh(x)^2
    ),
    "SMD->OR" = list(
      fun = function(x) x * pi / sqrt(3),
      inv = function(x) x * sqrt(3) / pi,
      jac = function(x) rep(pi / sqrt(3), length(x))
    ),
    "OR->SMD" = list(
      fun = function(x) x * sqrt(3) / pi,
      inv = function(x) x * pi / sqrt(3),
      jac = function(x) rep(sqrt(3) / pi, length(x))
    )
  ))
}

.compose_effect_transformations <- function(first, second) {

  force(first)
  force(second)

  return(list(
    fun = function(x) second[["fun"]](first[["fun"]](x)),
    inv = function(x) first[["inv"]](second[["inv"]](x)),
    jac = function(x) {

      first[["jac"]](x) * second[["jac"]](first[["fun"]](x))
    }
  ))
}

.effect_neighbor_edges <- function(edges, node) {

  edge_names <- names(edges)
  keep       <- startsWith(edge_names, paste0(node, "->"))

  return(edge_names[keep])
}

.effect_measure_transformation <- function(from, to) {

  if (from == to) {
    return(.identity_effect_transformation())
  }

  if (!all(c(from, to) %in% .effect_core_measures())) {
    stop(
      "Effect-size transformation from ", from, " to ", to,
      " is not available. Only SMD, COR, ZCOR, and OR can be converted. ",
      "RR, HR, IRR, RD, and GEN can only be shown on their fitted measure; ",
      "use transform = 'EXP' for RR, HR, IRR, and OR ratios.",
      call. = FALSE
    )
  }

  edges   <- .effect_basic_transformations()
  queue   <- list(list(
    node           = from,
    transformation = .identity_effect_transformation()
  ))
  visited <- from

  while (length(queue) > 0L) {

    current <- queue[[1]]
    queue   <- queue[-1]

    for (edge_name in .effect_neighbor_edges(edges, current[["node"]])) {

      next_node <- sub("^.*->", "", edge_name)
      if (next_node %in% visited) {
        next
      }

      next_transformation <- .compose_effect_transformations(
        first  = current[["transformation"]],
        second = edges[[edge_name]]
      )

      if (next_node == to) {
        return(next_transformation)
      }

      visited <- c(visited, next_node)
      queue[[length(queue) + 1L]] <- list(
        node           = next_node,
        transformation = next_transformation
      )
    }
  }

  stop("No effect-size transformation path found from ", from, " to ", to, ".",
       call. = FALSE)
}

.effect_output_setup <- function(object, output_measure = NULL, transform = NULL) {

  return(.effect_output_setup_measure(
    input_measure  = .measure(object),
    output_measure = output_measure,
    transform      = transform
  ))
}

.effect_output_setup_measure <- function(input_measure, output_measure = NULL,
                                         transform = NULL) {

  input_measure  <- .normalize_effect_measure(input_measure, "input_measure")
  output_measure <- .normalize_effect_measure(output_measure, "output_measure", allow_NULL = TRUE)
  transform      <- .normalize_effect_transform(transform)

  requested <- !is.null(output_measure) || transform != "identity"

  if (is.null(output_measure)) {
    output_measure <- input_measure
  }

  if (transform == "EXP" && !output_measure %in% .effect_log_measures()) {
    stop(
      "transform = 'EXP' is only available for log-scale measures ",
      "OR, RR, HR, and IRR.",
      call. = FALSE
    )
  }

  measure_transformation <- .effect_measure_transformation(
    from = input_measure,
    to   = output_measure
  )

  if (transform == "EXP") {
    transformation <- .compose_effect_transformations(
      first  = measure_transformation,
      second = .exp_effect_transformation()
    )
  } else {
    transformation <- measure_transformation
  }

  active <- input_measure != output_measure || transform != "identity"

  output <- list(
    input_measure  = input_measure,
    output_measure = output_measure,
    transform      = transform,
    requested      = requested,
    active         = active,
    transformation = transformation,
    label          = .effect_output_label(output_measure, transform),
    note           = .effect_output_note(input_measure, output_measure, transform)
  )

  return(output)
}

.effect_output_requested <- function(effect_transform) {

  if (is.null(effect_transform)) {
    return(FALSE)
  }

  return(isTRUE(effect_transform[["requested"]]))
}

.effect_output_active <- function(effect_transform) {

  if (is.null(effect_transform)) {
    return(FALSE)
  }

  return(isTRUE(effect_transform[["active"]]))
}

.effect_output_label <- function(output_measure, transform) {

  if (transform == "EXP") {
    return(switch(
      output_measure,
      "OR"  = "odds ratio",
      "RR"  = "risk ratio",
      "HR"  = "hazard ratio",
      "IRR" = "incidence rate ratio"
    ))
  }

  return(switch(
    output_measure,
    "SMD"  = "standardized mean difference",
    "COR"  = "correlation",
    "ZCOR" = "Fisher's z",
    "OR"   = "log odds ratio",
    "RR"   = "log risk ratio",
    "HR"   = "log hazard ratio",
    "IRR"  = "log incidence rate ratio",
    "RD"   = "risk difference",
    "GEN"  = "effect size"
  ))
}

.effect_measure_label <- function(measure) {

  return(.effect_output_label(measure, "identity"))
}

.effect_output_note <- function(input_measure, output_measure, transform) {

  if (input_measure == output_measure && transform == "identity") {
    return(NULL)
  }

  input_label  <- .effect_measure_label(input_measure)
  output_label <- .effect_output_label(output_measure, transform)

  if (transform == "EXP" && input_measure == output_measure) {
    return(paste0(
      "Effect estimates are summarized on the ", output_label,
      " scale using EXP on the ", input_label, " measure."
    ))
  }

  if (transform == "EXP") {
    return(paste0(
      "Effect estimates are transformed from ", input_label, " to ",
      .effect_measure_label(output_measure), " and summarized on the ",
      output_label, " scale using EXP."
    ))
  }

  return(paste0(
    "Effect estimates are transformed from ", input_label,
    " to ", output_label, "."
  ))
}

.effect_output_title <- function(title, effect_transform) {

  if (!.effect_output_active(effect_transform)) {
    return(title)
  }

  suffix <- paste0(" (", effect_transform[["label"]], ")")

  if (grepl(":$", title)) {
    title <- sub(":$", paste0(suffix, ":"), title)
  } else {
    title <- paste0(title, suffix)
  }

  return(title)
}

.effect_plot_transformation <- function(effect_transform) {

  if (!.effect_output_active(effect_transform)) {
    return(NULL)
  }

  return(effect_transform[["transformation"]])
}

.is_effect_location_parameter <- function(parameter) {

  return(parameter %in% c("mu", "mu_intercept"))
}

.transform_effect_matrix <- function(samples, effect_transform) {

  if (!.effect_output_active(effect_transform)) {
    return(samples)
  }

  dim_names <- dimnames(samples)
  samples   <- effect_transform[["transformation"]][["fun"]](samples)

  if (!is.matrix(samples)) {
    samples <- as.matrix(samples)
  }

  dimnames(samples) <- dim_names

  return(samples)
}

.transform_effect_vector <- function(samples, effect_transform) {

  if (!.effect_output_active(effect_transform)) {
    return(samples)
  }

  return(effect_transform[["transformation"]][["fun"]](samples))
}

.new_effect_brma_samples <- function(samples, n_chains, n_iter, title,
                                     probs = c(.025, .975), data = NULL,
                                     effect_transform = NULL) {

  if (!is.null(effect_transform)) {
    samples <- .transform_effect_matrix(samples, effect_transform)
    title   <- .effect_output_title(title, effect_transform)
  }

  return(.new_brma_samples(
    samples          = samples,
    n_chains         = n_chains,
    n_iter           = n_iter,
    title            = title,
    probs            = probs,
    data             = data,
    effect_transform = effect_transform
  ))
}

.transform_marginal_samples_effect <- function(samples, effect_transform) {

  if (!.effect_output_active(effect_transform)) {
    return(samples)
  }

  for (parameter in names(samples)) {
    samples[[parameter]] <- .transform_marginal_posterior_effect(
      samples          = samples[[parameter]],
      effect_transform = effect_transform
    )
  }

  return(samples)
}

.transform_marginal_posterior_effect <- function(samples, effect_transform) {

  sample_attributes <- attributes(samples)

  for (i in seq_along(samples)) {
    samples[[i]] <- .transform_effect_vector(
      samples           = samples[[i]],
      effect_transform  = effect_transform
    )
  }

  attributes(samples) <- sample_attributes

  return(samples)
}

context("Estimated marginal means")

# Load common test helpers
source(testthat::test_path("common-functions.R"))
source(testthat::test_path("helper-test-matrix.R"))
source(testthat::test_path("helper-visuals.R"))
REFERENCE_DIR <<- testthat::test_path("..", "results", "marginal_means")

.marginal_means_test_object <- function() {

  p <- stats::ppoints(200)

  levels <- list(
    alternate  = stats::qnorm(p, mean = -0.60, sd = 0.20),
    random     = stats::qnorm(p, mean = -0.20, sd = 0.25),
    systematic = stats::qnorm(p, mean =  0.20, sd = 0.30)
  )
  class(levels) <- c("marginal_posterior.factor", "marginal_posterior")

  samples   <- list(mu_alloc = levels)
  inference <- list(averaged = samples, conditional = samples, inference = list())
  class(inference) <- c("marginal_inference", "list")

  emm <- list(
    inference        = inference,
    parameters       = "mu_alloc",
    term_map         = data.frame(
      term             = "alloc",
      parameter        = "mu_alloc",
      label            = "alloc",
      stringsAsFactors = FALSE
    ),
    input_measure    = "GEN",
    effect_transform = .effect_output_setup_measure(input_measure = "GEN"),
    model_averaged   = FALSE,
    bf               = FALSE
  )
  class(emm) <- c("marginal_means.brma", "marginal_means")

  return(emm)
}


test_that("marginal_means base plot keeps separate level colors", {

  emm  <- .marginal_means_test_object()
  levels <- length(emm[["inference"]][["averaged"]][["mu_alloc"]])
  dots   <- .set_dots_plot(n_levels = levels)

  expect_gte(length(unique(dots[["col"]])), levels)
  .with_temp_plot_device(expect_silent(plot(emm, parameter = "alloc", legend = FALSE)))
})


# list cached fits lazily
skip_if_no_fits()
fit_names <- unique(c(
  "bcg_meta-analysis",
  marginal_means_cases()[["name"]],
  marginal_means_interaction_plot_cases()[["name"]]
))
fits <- lazy_fits(fit_names, validate = FALSE)


marginal_means_expected_stats <- function(emm, type = "averaged",
                                          effect_transform = emm[["effect_transform"]],
                                          probs = c(.025, .50, .975)) {

  samples <- .transform_marginal_samples_effect(
    samples          = emm[["inference"]][[type]],
    effect_transform = effect_transform
  )

  stats <- unlist(lapply(emm[["parameters"]], function(parameter) {

    lapply(samples[[parameter]], function(draws) {

      draws <- as.numeric(draws)
      c(
        Mean = mean(draws),
        SD   = stats::sd(draws),
        stats::quantile(draws, probs = probs, names = FALSE)
      )
    })
  }), recursive = FALSE)

  stats <- do.call(rbind, stats)
  colnames(stats) <- c("Mean", "SD", as.character(probs))

  return(stats)
}


test_that("marginal_means stores BayesTools marginal inference", {

  mm <- marginal_means(fits[["bcg_meta-regression2"]], n_samples = 1000)

  expect_s3_class(mm, "marginal_means.brma")
  expect_s3_class(mm[["inference"]], "marginal_inference")
  expect_equal(mm[["parameters"]], c("mu_intercept", "mu_alloc"))
  expect_equal(mm[["term_map"]][["term"]], c("intercept", "alloc"))
})


test_that("marginal_means hides BFs for non-RoBMA fits by default", {

  emm_brma <- marginal_means(fits[["bcg_meta-regression2"]], n_samples = 1000)
  emm_brma_bf <- marginal_means(
    fits[["bcg_meta-regression2"]],
    n_samples = 1000,
    bf        = TRUE
  )
  emm_robma <- marginal_means(
    fits[["dat.lehmann2018_RoBMA_mods"]],
    n_samples = 1000
  )

  expect_false("inclusion_BF" %in% attr(summary(emm_brma), "type"))
  expect_false(any(grepl(
    "Savage-Dickey",
    attr(summary(emm_brma), "warnings"),
    fixed = TRUE
  )))
  expect_true("inclusion_BF" %in% attr(summary(emm_brma_bf), "type"))
  expect_true("inclusion_BF" %in% attr(summary(emm_brma, bf = TRUE), "type"))
  expect_true("inclusion_BF" %in% attr(summary(emm_robma), "type"))
})


test_that("marginal_means plot errors show formula terms", {

  emm <- marginal_means(fits[["bcg_meta-regression2"]], n_samples = 1000)

  expect_error(
    plot(emm),
    "^The 'parameter' argument must be specified\\. Available terms are: 'intercept', 'alloc'\\.$"
  )
  expect_error(
    plot(emm, parameter = "missing"),
    "^Unknown marginal means parameter 'missing'\\. Available terms are: 'intercept', 'alloc'\\.$"
  )
})


test_that("marginal_means plot labels effect axis and term legend", {

  emm  <- marginal_means(fits[["bcg_meta-regression2"]], n_samples = 1000)
  plot <- plot(emm, parameter = "alloc", plot_type = "ggplot")
  colour_scale   <- plot[["scales"]][["get_scales"]]("colour")
  linetype_scale <- plot[["scales"]][["get_scales"]]("linetype")

  expect_identical(
    plot[["scales"]][["get_scales"]]("x")[["name"]],
    "Effect Size"
  )
  expect_identical(colour_scale[["name"]], "alloc")
  expect_identical(linetype_scale[["name"]], "alloc")

  plot_custom <- plot(
    emm,
    parameter = "alloc",
    plot_type = "ggplot",
    xlab      = "Custom Label"
  )
  expect_identical(
    plot_custom[["scales"]][["get_scales"]]("x")[["name"]],
    "Custom Label"
  )

  plot_custom_legend <- plot(
    emm,
    parameter    = "alloc",
    plot_type    = "ggplot",
    legend_title = "Custom Legend"
  )
  expect_identical(
    plot_custom_legend[["scales"]][["get_scales"]]("colour")[["name"]],
    "Custom Legend"
  )
})


test_that("marginal_means summaries transform effect-size scale", {

  emm <- marginal_means(
    fits[["bcg_meta-regression2"]],
    n_samples = 1000,
    transform = "EXP"
  )

  summary_rr <- summary(emm)
  expected   <- marginal_means_expected_stats(emm)
  actual     <- as.data.frame(summary_rr)[, colnames(expected)]

  expect_equal(
    unname(as.matrix(actual)),
    unname(expected),
    tolerance = sqrt(.Machine$double.eps)
  )
  expect_equal(attr(summary_rr, "title"), "Marginal Means (risk ratio):")
  expect_match(attr(summary_rr, "footnotes"), "risk ratio")
})


test_that("marginal_means summaries convert effect-size measures", {

  emm <- marginal_means(
    fits[["dat.lehmann2018_BMA.norm_mods"]],
    n_samples = 1000
  )
  effect_transform <- .effect_output_setup_measure(
    input_measure  = emm[["input_measure"]],
    output_measure = "COR"
  )

  summary_cor <- summary(emm, output_measure = "COR")
  expected    <- marginal_means_expected_stats(
    emm              = emm,
    effect_transform = effect_transform
  )
  actual      <- as.data.frame(summary_cor)[, colnames(expected)]

  expect_equal(
    unname(as.matrix(actual)),
    unname(expected),
    tolerance = sqrt(.Machine$double.eps)
  )
  expect_equal(
    attr(summary_cor, "title"),
    "Model-Averaged Marginal Means (correlation):"
  )
  expect_match(attr(summary_cor, "footnotes"), "correlation")
})


test_that("marginal_means summaries match reference tables", {

  cases <- marginal_means_cases()
  for (i in seq_len(nrow(cases))) {

    name <- cases[["name"]][i]
    fit  <- fits[[name]]
    emm  <- marginal_means(fit, n_samples = 1000)

    if (inherits(fit, "RoBMA")) {
      test_reference_table(
        summary(emm),
        paste0("summary_marginal-", name, "-averaged.txt"),
        paste0("Averaged marginal means reference table mismatch for '", name, "'")
      )

      test_reference_table(
        summary(emm, type = "conditional"),
        paste0("summary_marginal-", name, "-conditional.txt"),
        paste0("Conditional marginal means reference table mismatch for '", name, "'")
      )
    } else {
      test_reference_table(
        summary(emm),
        paste0("summary_marginal-", name, ".txt"),
        paste0("Marginal means reference table mismatch for '", name, "'")
      )
    }
  }
})


test_that("marginal_means core plot snapshots are stable", {

  for_each_case(marginal_means_cases(), function(case) {

    name      <- case_name(case)
    parameter <- case_value(case, "parameter")
    fit       <- fits[[name]]
    emm       <- marginal_means(fit, n_samples = 1000)

    expect_brma_plot_snapshot(
      paste0(name, "-marginal_means-baseplot-no-prior"),
      function() plot(emm, parameter = parameter, prior = FALSE)
    )

    expect_brma_plot_snapshot(
      paste0(name, "-marginal_means-ggplot-no-prior"),
      plot(emm, parameter = parameter, prior = FALSE, plot_type = "ggplot")
    )
  }, tier = visual_test_tier())
})

test_that("marginal_means prior plot gallery snapshots are stable", {

  skip_if_not_full_visuals("Marginal-means prior overlays are gallery coverage.")

  for_each_case(marginal_means_cases(), function(case) {

    name      <- case_name(case)
    parameter <- case_value(case, "parameter")
    fit       <- fits[[name]]
    emm       <- marginal_means(fit, n_samples = 1000)

    expect_brma_plot_snapshot(
      paste0(name, "-marginal_means-baseplot-prior"),
      function() plot(emm, parameter = parameter, prior = TRUE)
    )

    expect_brma_plot_snapshot(
      paste0(name, "-marginal_means-ggplot-prior"),
      plot(emm, parameter = parameter, prior = TRUE, plot_type = "ggplot")
    )
  }, tier = visual_test_tier())
})


test_that("marginal_means interaction plots render moderator type combinations", {

  skip_if_missing_fits(unique(marginal_means_interaction_plot_cases()[["name"]]))

  for_each_case(marginal_means_interaction_plot_cases(), function(case) {

    name      <- case_name(case)
    parameter <- case_value(case, "parameter")
    fit       <- fits[[name]]
    emm       <- marginal_means(fit, n_samples = 1000)

    expect_true(parameter %in% emm[["term_map"]][["term"]])

    for (show_prior in c(FALSE, TRUE)) {
      expect_s3_class(
        plot(
          emm,
          parameter = parameter,
          prior     = show_prior,
          plot_type = "ggplot"
        ),
        "ggplot"
      )

      .with_temp_plot_device(expect_silent(plot(
        emm,
        parameter = parameter,
        prior     = show_prior
      )))
    }
  }, tier = test_tier())
})


test_that("marginal_means requires moderators", {

  expect_error(
    marginal_means(fits[["bcg_meta-analysis"]], n_samples = 1000),
    "moderators"
  )
})


test_that("marginal_means conditional type is RoBMA-only", {

  emm <- marginal_means(fits[["bcg_meta-regression2"]], n_samples = 1000)

  expect_error(
    summary(emm, type = "conditional"),
    "RoBMA marginal means"
  )
  expect_error(
    plot(emm, parameter = "alloc", type = "conditional", plot_type = "ggplot"),
    "RoBMA marginal means"
  )

  emm_robma <- marginal_means(
    fits[["dat.lehmann2018_RoBMA_mods"]],
    n_samples = 1000
  )
  expect_s3_class(
    summary(emm_robma, type = "conditional"),
    "summary.marginal_means.brma"
  )
})

context("Effect-size transformations")

source(testthat::test_path("common-functions.R"))

skip_if_not_installed("metafor")

expect_effect_transform_matches_metafor <- function(input_measure, output_measure,
                                                    values, expected) {

  info   <- .effect_output_setup_measure(
    input_measure  = input_measure,
    output_measure = output_measure
  )
  actual <- .transform_effect_vector(values, info)

  expect_equal(
    actual,
    expected,
    tolerance = sqrt(.Machine$double.eps),
    info      = paste(input_measure, "to", output_measure, "matches metafor")
  )
}

expect_effect_jacobian_matches_numeric <- function(input_measure, values,
                                                   output_measure = NULL,
                                                   transform = NULL) {

  info           <- .effect_output_setup_measure(
    input_measure  = input_measure,
    output_measure = output_measure,
    transform      = transform
  )
  transformation <- info[["transformation"]]
  step           <- 1e-6
  expected       <- (
    transformation[["fun"]](values + step) -
    transformation[["fun"]](values - step)
  ) / (2 * step)

  expect_equal(
    transformation[["jac"]](values),
    expected,
    tolerance = 1e-5,
    info      = paste(input_measure, "to", info[["output_measure"]], "uses forward Jacobian")
  )
}

test_that("effect-size measure transformations match metafor", {

  d_values <- c(-1, 0, 1)
  r_values <- c(-0.4, 0, 0.4)
  z_values <- metafor::transf.rtoz(r_values)
  o_values <- c(-1, 0, 1)

  expect_effect_transform_matches_metafor(
    input_measure  = "SMD",
    output_measure = "COR",
    values         = d_values,
    expected       = metafor::transf.dtorpb(d_values)
  )
  expect_effect_transform_matches_metafor(
    input_measure  = "SMD",
    output_measure = "ZCOR",
    values         = d_values,
    expected       = metafor::transf.rtoz(metafor::transf.dtorpb(d_values))
  )
  expect_effect_transform_matches_metafor(
    input_measure  = "SMD",
    output_measure = "OR",
    values         = d_values,
    expected       = metafor::transf.dtolnor.logis(d_values)
  )

  expect_effect_transform_matches_metafor(
    input_measure  = "COR",
    output_measure = "SMD",
    values         = r_values,
    expected       = metafor::transf.rpbtod(r_values)
  )
  expect_effect_transform_matches_metafor(
    input_measure  = "COR",
    output_measure = "ZCOR",
    values         = r_values,
    expected       = metafor::transf.rtoz(r_values)
  )
  expect_effect_transform_matches_metafor(
    input_measure  = "COR",
    output_measure = "OR",
    values         = r_values,
    expected       = metafor::transf.dtolnor.logis(metafor::transf.rpbtod(r_values))
  )

  expect_effect_transform_matches_metafor(
    input_measure  = "ZCOR",
    output_measure = "SMD",
    values         = z_values,
    expected       = metafor::transf.rpbtod(metafor::transf.ztor(z_values))
  )
  expect_effect_transform_matches_metafor(
    input_measure  = "ZCOR",
    output_measure = "COR",
    values         = z_values,
    expected       = metafor::transf.ztor(z_values)
  )
  expect_effect_transform_matches_metafor(
    input_measure  = "ZCOR",
    output_measure = "OR",
    values         = z_values,
    expected       = metafor::transf.dtolnor.logis(
      metafor::transf.rpbtod(metafor::transf.ztor(z_values))
    )
  )

  expect_effect_transform_matches_metafor(
    input_measure  = "OR",
    output_measure = "SMD",
    values         = o_values,
    expected       = metafor::transf.lnortod.logis(o_values)
  )
  expect_effect_transform_matches_metafor(
    input_measure  = "OR",
    output_measure = "COR",
    values         = o_values,
    expected       = metafor::transf.dtorpb(metafor::transf.lnortod.logis(o_values))
  )
  expect_effect_transform_matches_metafor(
    input_measure  = "OR",
    output_measure = "ZCOR",
    values         = o_values,
    expected       = metafor::transf.rtoz(
      metafor::transf.dtorpb(metafor::transf.lnortod.logis(o_values))
    )
  )
})

test_that("effect-size transformations use forward Jacobians", {

  expect_effect_jacobian_matches_numeric(
    input_measure  = "SMD",
    output_measure = "COR",
    values         = c(-1, 0, 1)
  )
  expect_effect_jacobian_matches_numeric(
    input_measure  = "COR",
    output_measure = "SMD",
    values         = c(-0.4, 0, 0.4)
  )
  expect_effect_jacobian_matches_numeric(
    input_measure  = "COR",
    output_measure = "ZCOR",
    values         = c(-0.4, 0, 0.4)
  )
  expect_effect_jacobian_matches_numeric(
    input_measure  = "ZCOR",
    output_measure = "COR",
    values         = c(-0.5, 0, 0.5)
  )
  expect_effect_jacobian_matches_numeric(
    input_measure  = "SMD",
    output_measure = "OR",
    values         = c(-1, 0, 1)
  )
  expect_effect_jacobian_matches_numeric(
    input_measure  = "OR",
    output_measure = "SMD",
    values         = c(-1, 0, 1)
  )
  expect_effect_jacobian_matches_numeric(
    input_measure = "RR",
    transform     = "EXP",
    values        = c(log(0.5), 0, log(2))
  )
  expect_effect_jacobian_matches_numeric(
    input_measure  = "SMD",
    output_measure = "OR",
    transform      = "EXP",
    values         = c(-1, 0, 1)
  )
  expect_effect_jacobian_matches_numeric(
    input_measure  = "SMD",
    output_measure = "ZCOR",
    values         = c(-1, 0, 1)
  )
  expect_effect_jacobian_matches_numeric(
    input_measure  = "OR",
    output_measure = "COR",
    values         = c(-1, 0, 1)
  )
})

test_that("EXP is explicit for log-scale ratio output", {

  log_values <- c(log(0.5), 0, log(2))

  info <- .effect_output_setup_measure(
    input_measure  = "OR",
    output_measure = "OR",
    transform      = "EXP"
  )

  expect_equal(
    .transform_effect_vector(log_values, info),
    metafor::transf.exp.int(log_values, targs = list(tau2 = 0))
  )
  expect_equal(info[["label"]], "odds ratio")

  expect_error(
    .effect_output_setup_measure(
      input_measure = "SMD",
      transform     = "EXP"
    ),
    "EXP"
  )
})

test_that("non-core measures are not converted across measures", {

  expect_error(
    .effect_output_setup_measure(
      input_measure  = "RR",
      output_measure = "SMD"
    ),
    "not available"
  )

  info <- .effect_output_setup_measure(
    input_measure = "RR",
    transform     = "EXP"
  )
  expect_equal(
    .transform_effect_vector(c(0, log(2)), info),
    metafor::transf.exp.int(c(0, log(2)), targs = list(tau2 = 0))
  )
})

test_that("plot transformations use BayesTools forward Jacobian convention", {

  original_x    <- c(-1, 0, 1)
  transformed_x <- exp(original_x)
  info          <- .effect_output_setup_measure(
    input_measure = "RR",
    transform     = "EXP"
  )
  transformation <- .effect_plot_transformation(info)

  expect_equal(
    transformation[["jac"]](original_x),
    transformed_x,
    tolerance = sqrt(.Machine$double.eps)
  )
  expect_warning(
    density_y <- BayesTools:::.density.prior_transformation_y(
      x              = transformed_x,
      y              = rep(1, length(transformed_x)),
      transformation = transformation
    ),
    NA
  )
  expect_equal(
    density_y,
    1 / transformed_x,
    tolerance = sqrt(.Machine$double.eps)
  )
})

test_that("transformed brma_samples preserve posterior draw integration", {

  samples <- matrix(c(0, log(2), log(3)), ncol = 1)
  colnames(samples) <- "mu"

  info <- .effect_output_setup_measure(
    input_measure = "OR",
    transform     = "EXP"
  )

  out <- .new_effect_brma_samples(
    samples          = samples,
    n_chains         = 1,
    n_iter           = 3,
    title            = "Pooled Effect Size",
    effect_transform = info
  )

  expect_s3_class(out, "brma_samples")
  expect_equal(as.vector(as.matrix(out)), c(1, 2, 3))
  expect_equal(attr(out, "nchains"), 1)
  expect_equal(attr(out, "niter"), 3)

  skip_if_not_installed("posterior")
  draws <- posterior::as_draws_matrix(out)
  expect_equal(posterior::variables(draws), "mu")
  expect_equal(as.numeric(draws[, "mu"]), c(1, 2, 3))
})

test_that("marginal posterior samples are transformed without losing metadata", {

  marginal_sample <- list(intercept = c(0, log(2)))
  class(marginal_sample) <- c("marginal_posterior.formula", "marginal_posterior")
  attr(marginal_sample, "formula_parameter") <- "mu"

  samples <- list(mu_intercept = marginal_sample)
  info    <- .effect_output_setup_measure(
    input_measure = "OR",
    transform     = "EXP"
  )

  out <- .transform_marginal_samples_effect(
    samples          = samples,
    effect_transform = info
  )

  expect_s3_class(out[["mu_intercept"]], "marginal_posterior.formula")
  expect_equal(attr(out[["mu_intercept"]], "formula_parameter"), "mu")
  expect_equal(out[["mu_intercept"]][["intercept"]], c(1, 2))
})

#' @title Prior Distribution
#'
#' @description Create prior distribution objects used by RoBMA and brma
#' fitting functions.
#'
#' @param distribution character. Prior distribution name.
#' @param parameters list. Distribution parameters.
#' @param truncation list with \code{lower} and \code{upper} truncation bounds.
#' @param prior_weights numeric prior model weight.
#'
#' @details This is RoBMA's re-export of \code{BayesTools::prior()}; supported
#' distribution names and parameter lists follow BayesTools.
#'
#' @return An object inheriting from \code{prior}.
#'
#' @examples
#' prior("normal", list(mean = 0, sd = 0.5))
#' prior(
#'   "normal",
#'   list(mean = 0, sd = 0.5),
#'   truncation = list(lower = 0, upper = Inf)
#' )
#'
#' @export
prior <- BayesTools::prior

#' @title Empty Prior
#'
#' @description Create an empty prior used to omit a model component.
#'
#' @inheritParams prior
#'
#' @return An object inheriting from \code{prior} and \code{prior.none}.
#'
#' @export
prior_none <- BayesTools::prior_none

#' @title Factor Prior
#'
#' @description Create priors on factor contrasts.
#'
#' @inheritParams prior
#' @param contrast character. Contrast coding used for factor levels. Common
#' RoBMA model options are \code{"treatment"}, \code{"meandif"}, and
#' \code{"orthonormal"}; the standalone helper also accepts BayesTools contrast
#' aliases such as \code{"independent"}.
#'
#' @details Mean-difference and orthonormal contrasts require vector or
#' multivariate priors. Treatment/dummy and independent contrasts use univariate
#' simple priors per contrast coefficient.
#'
#' @return An object inheriting from \code{prior}, \code{prior.factor}, and a
#' contrast-specific marker class.
#'
#' @export
prior_factor <- BayesTools::prior_factor

#' @title PET Prior
#'
#' @description Create PET publication-bias regression priors.
#'
#' @inheritParams prior
#'
#' @details This forwards to \code{BayesTools::prior_PET()} and uses the same
#' distribution and parameter conventions as \code{prior()}. By default, PET
#' priors are truncated to the positive half-line.
#'
#' @return An object inheriting from \code{prior} with the \code{prior.PET}
#' marker class.
#'
#' @seealso \code{\link{publication_bias_prior_specification}}
#'
#' @export
prior_PET  <- BayesTools::prior_PET

#' @title PEESE Prior
#'
#' @description Create PEESE publication-bias regression priors.
#'
#' @inheritParams prior
#'
#' @details This forwards to \code{BayesTools::prior_PEESE()} and uses the same
#' distribution and parameter conventions as \code{prior()}.
#'
#' @return An object inheriting from \code{prior} with the \code{prior.PEESE}
#' marker class.
#'
#' @seealso \code{\link{publication_bias_prior_specification}}
#'
#' @export
prior_PEESE <- BayesTools::prior_PEESE

#' @title Weightfunction Prior
#'
#' @description Create weightfunction publication-bias priors and their
#' weight-prior helper objects.
#'
#' @param side character. Either \code{"one-sided"} or \code{"two-sided"}.
#' @param steps numeric vector of p-value cut points. These define
#' \code{length(steps) + 1} p-value bins and must be ordered values in
#' \code{(0, 1)}.
#' @param weights a weight-prior object created by \code{wf_cumulative()},
#' \code{wf_fixed()}, or \code{wf_independent()}.
#' @param reference character. Reference bin, currently
#' \code{"most_significant"}.
#' @inheritParams prior
#'
#' @details Fixed weights must have one value per p-value bin
#' (\code{length(steps) + 1}), and the reference bin must have weight 1.
#'
#' @return \code{prior_weightfunction()} returns an object inheriting from
#' \code{prior} and \code{prior.weightfunction}; the \code{wf_*()} helpers
#' return \code{weightfunction_weights} helper objects with subclass markers.
#'
#' @examples
#' prior_weightfunction("one-sided", steps = 0.025)
#' prior_weightfunction(
#'   side    = "one-sided",
#'   steps   = c(0.025, 0.5),
#'   weights = wf_fixed(c(1, 0.8, 0.6))
#' )
#'
#' @seealso \code{\link{publication_bias_prior_specification}}
#'
#' @export
prior_weightfunction <- BayesTools::prior_weightfunction

#' @rdname prior_weightfunction
#' @param alpha optional positive cumulative-Dirichlet concentration parameters,
#' one per p-value bin. If \code{NULL}, \code{prior_weightfunction()} uses
#' \code{rep(1, length(steps) + 1)}. Cumulative weights encode monotone
#' decreasing publication weights relative to the most-significant bin.
#' @export
wf_cumulative <- BayesTools::wf_cumulative

#' @rdname prior_weightfunction
#' @param omega fixed publication weights, one per bin; values must be
#' non-missing, nonnegative, and match \code{length(steps) + 1} when used in
#' \code{prior_weightfunction()}.
#' @export
wf_fixed <- BayesTools::wf_fixed

#' @rdname prior_weightfunction
#' @param prior continuous simple prior distribution for each non-reference
#' weight. Point, discrete, mixture, and other non-simple priors are invalid.
#' @param scale latent scale for independent weights; either \code{"omega"},
#' \code{"log_omega"}, or the \code{"log"} alias. Direct \code{"omega"} priors
#' need nonnegative support; \code{"log"} is normalized to \code{"log_omega"}.
#' @export
wf_independent <- BayesTools::wf_independent

#' @title Informed Prior
#'
#' @description Create empirical informed prior distributions.
#'
#' @param name character. Informed prior name. Supported examples include
#' \code{"van Erp"}, \code{"Oosterwijk"}, \code{"Cochrane"}, and medicine
#' subfields from \code{BayesTools::prior_informed_medicine_names}.
#' @param parameter character. Parameter subset. Required for medicine priors
#' and not needed for psychology priors.
#' @param type character. Effect-size type. Medicine priors use values such as
#' \code{"smd"}, \code{"logOR"}, \code{"logRR"}, \code{"RD"}, and
#' \code{"logHR"}.
#'
#' @return An object of class \code{prior}.
#'
#' @details Further details can be found in \insertCite{erp2017estimates;textual}{RoBMA},
#' \insertCite{gronau2017bayesian;textual}{RoBMA},
#' \insertCite{bartos2021bayesian;textual}{RoBMA}, and
#' \insertCite{bartos2023empirical;textual}{RoBMA}.
#' @references
#' \insertAllCited{}
#' @export
prior_informed <- BayesTools::prior_informed

# needs to be completely re-exported because of some odd roxygen issues
#' @title BayesTools Contrast Matrices
#'
#' @description BayesTools provides several contrast matrix functions for Bayesian factor analysis.
#' These functions create different types of contrast matrices that can be used with factor
#' variables in Bayesian models.
#'
#' @details
#' The package includes the following contrast functions:
#' \describe{
#'   \item{\code{contr.orthonormal}}{Return a matrix of orthonormal contrasts.
#'     Code is based on \code{stanova::contr.bayes} and corresponding to description
#'     by \insertCite{rouder2012default;textual}{BayesTools}. Returns a matrix with n rows and
#'     k columns, with k = n - 1 if \code{contrasts = TRUE} and k = n if \code{contrasts = FALSE}.}
#'   \item{\code{contr.meandif}}{Return a matrix of mean difference contrasts.
#'     This is an adjustment to the \code{contr.orthonormal} that ascertains that the prior
#'     distributions on difference between the grand mean and factor level are identical independent
#'     of the number of factor levels (which does not hold for the orthonormal contrast). Furthermore,
#'     the contrast is re-scaled so the specified prior distribution exactly corresponds to the prior
#'     distribution on difference between each factor level and the grand mean -- this is approximately
#'     twice the scale of \code{contr.orthonormal}. Returns a matrix with n rows and k columns,
#'     with k = n - 1 if \code{contrasts = TRUE} and k = n if \code{contrasts = FALSE}.}
#'   \item{\code{contr.independent}}{Return a matrix of independent contrasts -- a level for each term.
#'     Returns a matrix with n rows and k columns, with k = n if \code{contrasts = TRUE} and k = n
#'     if \code{contrasts = FALSE}.}
#' }
#'
#' @param n a vector of levels for a factor, or the number of levels
#' @param contrasts logical indicating whether contrasts should be computed
#'
#' @examples
#' # Orthonormal contrasts
#' contr.orthonormal(c(1, 2))
#' contr.orthonormal(c(1, 2, 3))
#'
#' # Mean difference contrasts
#' contr.meandif(c(1, 2))
#' contr.meandif(c(1, 2, 3))
#'
#' # Independent contrasts
#' contr.independent(c(1, 2))
#' contr.independent(c(1, 2, 3))
#'
#' @references
#' \insertAllCited{}
#'
#' @aliases contr.orthonormal contr.meandif contr.independent
#' @name contr.BayesTools
NULL

#' @rdname contr.BayesTools
#' @export
contr.orthonormal <- BayesTools::contr.orthonormal

#' @rdname contr.BayesTools
#' @export
contr.meandif <- BayesTools::contr.meandif

#' @rdname contr.BayesTools
#' @export
contr.independent <- BayesTools::contr.independent

# ============================================================================ #
# brma.regplot.R
# ============================================================================ #
#
# This file contains the regression plot (bubble plot) functions for brma
# objects with moderators.
#
# Bubble plots display observed effect sizes against a moderator variable,
# with point sizes proportional to study precision (inverse variance).
# For continuous moderators, a regression line with CI/PI bands is shown.
# For categorical moderators, grouped points with optional jitter are shown.
#
# ============================================================================ #


### regplot functions ----
#' @export
regplot <- function(x, ...) UseMethod("regplot")


#' @title Regression Plot (Bubble Plot) for brma Object
#'
#' @description \code{regplot.brma} creates a regression plot (also known as
#' bubble plot) for a fitted brma object with moderators. The plot displays
#' observed effect sizes against a moderator variable, with point sizes
#' proportional to study precision.
#'
#' @param x a fitted brma object with moderators
#' @param mod index or name of the moderator variable to plot on the x-axis.
#' If not specified and only one moderator is present, that moderator is used.
#' If multiple moderators are present, this argument is required.
#' @param pred logical; whether to show the prediction line. Defaults to \code{TRUE}.
#' @param ci logical; whether to show credible interval bands. Defaults to \code{TRUE}.
#' @param pi logical; whether to show prediction interval bands. Defaults to \code{FALSE}.
#' @param si logical; whether to show sampling interval bands. Defaults to \code{FALSE}.
#' The sampling interval shows the expected range of observed effect sizes,
#' incorporating both heterogeneity (tau) and a representative level of
#' sampling error (median SE across studies by default). When the model includes
#' publication bias adjustments and \code{sampling_bias = TRUE}, the
#' sampling interval incorporates the expected distortion from the
#' selection process.
#' @param level numeric; credible/prediction interval level in percent.
#' Defaults to \code{95}.
#' @param at numeric vector; for continuous moderators, values at which to
#' evaluate the prediction. If not specified, uses a sequence across the
#' observed range.
#' @param digits integer; number of decimal places for labels. Defaults to \code{2}.
#' @param transf function; transformation to apply to the y-axis (effect sizes).
#' Defaults to \code{NULL} (no transformation).
#' @param atransf reserved for axis-label transformations. Currently not
#' implemented and must be `NULL`.
#' @param targs reserved for additional transformation arguments. Currently not
#' implemented and must be `NULL`.
#' @param refline numeric; position of horizontal reference line.
#' Defaults to \code{NULL} (no reference line).
#' @param psize numeric vector or \code{NULL}; point sizes for each study.
#' If scalar, it is recycled to all studies. If \code{NULL} (default), sizes
#' are computed based on inverse sampling variance.
#' @param plim numeric vector of length 2; range for point sizes.
#' Defaults to \code{c(0.5, 3)}.
#' @param by character; name of a moderator variable to use for separate
#' lines/colors. Defaults to \code{NULL}. If omitted and the selected
#' moderator enters exactly one two-way interaction, the other variable in the
#' interaction is used automatically. Continuous \code{by} moderators are shown
#' at mean - SD, mean, and mean + SD.
#' @param legend logical; whether to show legend when \code{by} is specified.
#' Defaults to \code{TRUE}.
#' @param xlab character; x-axis label. Defaults to the moderator name.
#' @param ylab character; y-axis label. Defaults to "Observed Effect Size".
#' @param xlim numeric vector of length 2; x-axis limits. Defaults to data range.
#' @param ylim numeric vector of length 2; y-axis limits. Defaults to data range.
#' @param sampling_bias whether publication bias should be incorporated into
#' plotted predictions and sampling intervals. Defaults to \code{TRUE}. For
#' PET/PEESE models, this includes the expected regression bias in predictions.
#' For selection models, sampling-bias adjustment applies to sampling intervals
#' when \code{si = TRUE}; the mean prediction is unchanged.
#' @param sei single positive numeric value used as the reference standard
#' error for sampling-bias and sampling-interval calculations. Defaults to the
#' median observed standard error.
#' @param max_samples maximum number of posterior samples used for prediction
#' summaries and interval bands. Defaults to \code{10000}. Use \code{Inf} to
#' use all posterior samples.
#' @param plot_type character; whether to use base R graphics (\code{"base"})
#' or ggplot2 (\code{"ggplot"}). Defaults to \code{"base"}.
#' @param as_data logical; if \code{TRUE}, returns plot data instead of
#' creating plot. Defaults to \code{FALSE}.
#' @param ... additional graphical arguments:
#' \describe{
#'   \item{main}{character string for plot title}
#'   \item{pch}{point symbol (default: 21)}
#'   \item{col}{point border color (default: "black")}
#'   \item{bg}{point fill color (default: "#A6A6A6")}
#'   \item{lcol}{line color (default: "black")}
#'   \item{lwd}{line width (default: 2)}
#'   \item{shade}{CI/PI/SI band shading (default: TRUE)}
#'   \item{col.ci}{CI band color (default: "gray70")}
#'   \item{col.pi}{PI band color (default: "gray85")}
#'   \item{col.si}{SI band color (default: "gray92")}
#'   \item{alpha.ci}{CI band transparency (default: 0.4)}
#'   \item{alpha.pi}{PI band transparency (default: 0.2)}
#'   \item{alpha.si}{SI band transparency (default: 0.15)}
#'   \item{jitter}{jitter amount for categorical moderators (default: 0.2)}
#'   \item{box.width}{box width for categorical interval summaries (default: 0.5)}
#' }
#'
#' @details
#' The regression plot (bubble plot) is a standard visualization for
#' meta-regression results. It displays:
#' \itemize{
#'   \item Observed effect sizes (y-axis) against moderator values (x-axis)
#'   \item Point sizes proportional to study precision (inverse variance)
#'   \item Prediction line showing the estimated regression relationship
#'   \item Confidence bands showing uncertainty in the mean prediction
#'   \item Optional prediction bands showing expected range of true effects
#'   \item Optional sampling interval bands showing expected range of observed outcomes
#' }
#'
#' For continuous moderators, predictions are computed across the observed
#' range of the moderator. For categorical moderators (factors), predictions
#' are computed at each factor level with optional jittering of points.
#'
#' The \code{by} argument allows displaying separate regression lines for
#' different levels of a second moderator, useful for visualizing interactions.
#'
#' @return \code{regplot.brma} returns \code{NULL} invisibly if
#' \code{plot_type = "base"} or a ggplot object if \code{plot_type = "ggplot"}.
#' If \code{as_data = TRUE}, returns a list with plot data components.
#'
#' @examples \dontrun{
#' if (requireNamespace("metadat", quietly = TRUE) &&
#'     requireNamespace("metafor", quietly = TRUE)) {
#'   data(dat.bcg, package = "metadat")
#'   dat <- metafor::escalc(
#'     measure = "RR",
#'     ai      = tpos,
#'     bi      = tneg,
#'     ci      = cpos,
#'     di      = cneg,
#'     data    = dat.bcg
#'   )
#'
#'   fit <- brma(
#'     yi      = yi,
#'     vi      = vi,
#'     mods    = ~ ablat + year,
#'     data    = dat,
#'     measure = "RR"
#'   )
#'   regplot(fit, mod = "ablat")
#'   regplot(fit, mod = "year", pi = TRUE, si = TRUE)
#'   regplot(fit, mod = "ablat", plot_type = "ggplot")
#'
#'   fit_cat <- brma(yi = yi, vi = vi, mods = ~ alloc, data = dat, measure = "RR")
#'   regplot(fit_cat)
#' }
#' }
#'
#' @seealso [funnel.brma()], [predict.brma()]
#' @aliases regplot
#' @export
#' @exportS3Method metafor::regplot
#' @rdname regplot
regplot.brma <- function(x, mod = NULL, pred = TRUE, ci = TRUE, pi = FALSE, si = FALSE,
                         level = 95, at = NULL, digits = 2,
                         transf = NULL, atransf = NULL, targs = NULL,
                         refline = NULL, psize = NULL, plim = c(0.5, 3),
                         by = NULL, legend = TRUE,
                         xlab = NULL, ylab = NULL, xlim = NULL, ylim = NULL,
                         sampling_bias = TRUE, sei = NULL, max_samples = 10000,
                         plot_type = "base", as_data = FALSE, ...) {

  # input validation
  BayesTools::check_bool(pred, "pred")
  BayesTools::check_bool(ci, "ci")
  BayesTools::check_bool(pi, "pi")
  BayesTools::check_bool(si, "si")
  BayesTools::check_real(level, "level", lower = 50, upper = 99.9)
  .check_plot_numeric(at, "at", allow_null = TRUE)
  BayesTools::check_int(digits, "digits", lower = 0)
  .check_plot_function(transf, "transf")
  .check_plot_numeric(refline, "refline", check_length = 1, allow_null = TRUE)
  .check_plot_numeric(plim, "plim", check_length = 2, lower = 0, allow_null = FALSE)
  if (plim[1] >= plim[2]) {
    stop("'plim' must be an increasing numeric vector of length 2.", call. = FALSE)
  }
  BayesTools::check_bool(legend, "legend")
  .check_plot_label(xlab, "xlab")
  .check_plot_label(ylab, "ylab")
  .check_plot_limits(xlim, "xlim")
  .check_plot_limits(ylim, "ylim")
  BayesTools::check_bool(sampling_bias, "sampling_bias")
  max_samples <- .normalize_max_samples(max_samples, "max_samples")
  if (!is.null(sei)) {
    BayesTools::check_real(sei, "sei", lower = 0)
    if (sei <= 0) {
      stop("'sei' must be positive.", call. = FALSE)
    }
  }
  BayesTools::check_char(plot_type, "plot_type", allow_values = c("base", "ggplot"))
  BayesTools::check_bool(as_data, "as_data")

  # not yet implemented arguments
  if (!is.null(atransf))
    stop("The 'atransf' argument is not yet implemented.", call. = FALSE)
  if (!is.null(targs))
    stop("The 'targs' argument is not yet implemented.", call. = FALSE)

  # check that model has moderators
  if (!.is_mods(x)) {
    stop("regplot requires a model with moderators. ",
         "Use funnel() for intercept-only models.", call. = FALSE)
  }

  # set up graphical arguments with defaults
  dots <- .set_dots_regplot(...)

  # identify the moderator to plot
  mod_info <- .regplot_get_moderator(x, mod)
  mod_name <- mod_info$name
  mod_type <- mod_info$type
  mod_data <- mod_info$data
  if (mod_type == "categorical" && !is.null(at)) {
    stop("'at' is only available for continuous moderators.", call. = FALSE)
  }

  K <- length(.outcome_data_yi(x))
  if (!is.null(psize)) {
    .check_plot_numeric(psize, "psize", lower = 0, allow_null = FALSE)
    if (!length(psize) %in% c(1L, K)) {
      stop("'psize' must be either a scalar or have one value per study.", call. = FALSE)
    }
  }

  # identify grouping variable (explicit or auto-detected interaction)
  by_info <- .regplot_get_by(x, by, mod_name)

  # generate plot data
  regplot_data <- .regplot_data(
    x             = x,
    mod_name      = mod_name,
    mod_type      = mod_type,
    mod_data      = mod_data,
    by_info       = by_info,
    pred          = pred,
    ci            = ci,
    pi            = pi,
    si            = si,
    level         = level,
    at            = at,
    digits        = digits,
    psize         = psize,
    plim          = plim,
    transf        = transf,
    xlim          = xlim,
    ylim          = ylim,
    xlab          = xlab,
    ylab          = ylab,
    refline       = refline,
    sampling_bias = sampling_bias,
    max_samples   = max_samples,
    reference_sei = sei,
    dots          = dots
  )

  # allow data return for programmatic access
  if (isTRUE(as_data)) {
    return(regplot_data)
  }

  # create plot
  if (plot_type == "ggplot") {
    return(.regplot_plot_ggplot(regplot_data, dots, legend = legend))
  } else {
    .regplot_plot_base(regplot_data, dots, legend = legend)
    return(invisible(NULL))
  }
}


# ---------------------------------------------------------------------------- #
# .regplot_get_moderator
# ---------------------------------------------------------------------------- #
#
# Identify and validate the moderator variable for plotting.
#
# @param x     brma object
# @param mod   index or name of moderator (NULL for auto-detect)
#
# @return list with name, type, and data for the moderator
#
# ---------------------------------------------------------------------------- #
.regplot_get_moderator <- function(x, mod) {

  mods_data <- x[["data"]][["mods"]]
  design    <- .fitted_formula_design(x, "mu", required = TRUE)

  if (is.null(mods_data)) {
    stop("No moderator data found in the model.", call. = FALSE)
  }

  mod_names <- design[["predictors"]]

  if (is.null(mod)) {
    # auto-detect: use single moderator or error if multiple
    if (length(mod_names) == 1) {
      mod_name <- mod_names[1]
    } else {
      stop("Multiple moderators present. Please specify 'mod' argument. ",
           "Available moderators: ", paste(mod_names, collapse = ", "),
           call. = FALSE)
    }
  } else if (is.numeric(mod)) {
    # index-based selection
    if (mod < 1 || mod > length(mod_names)) {
      stop("Invalid moderator index. Must be between 1 and ",
           length(mod_names), ".", call. = FALSE)
    }
    mod_name <- mod_names[mod]
  } else if (is.character(mod)) {
    # name-based selection
    if (!mod %in% mod_names) {
      stop("Moderator '", mod, "' not found. ",
           "Available moderators: ", paste(mod_names, collapse = ", "),
           call. = FALSE)
    }
    mod_name <- mod
  } else {
    stop("'mod' must be NULL, a character name, or a numeric index.",
         call. = FALSE)
  }

  # extract moderator data and determine type
  mod_values <- mods_data[[mod_name]]
  predictor_type <- design[["predictor_types"]][[mod_name]]

  if (identical(predictor_type, "factor") ||
      is.factor(mod_values) || is.character(mod_values)) {
    mod_type <- "categorical"
    mod_values <- factor(
      mod_values,
      levels = if (!is.null(design[["xlevels"]][[mod_name]])) {
        design[["xlevels"]][[mod_name]]
      } else {
        levels(factor(mod_values))
      }
    )
  } else {
    mod_type <- "continuous"
  }

  return(list(
    name = mod_name,
    type = mod_type,
    data = mod_values
  ))
}


# ---------------------------------------------------------------------------- #
# .regplot_get_by
# ---------------------------------------------------------------------------- #
#
# Identify and validate a grouping moderator for interaction displays.
#
# ---------------------------------------------------------------------------- #
.regplot_get_by <- function(x, by, mod_name) {

  mods_data <- x[["data"]][["mods"]]
  design    <- .fitted_formula_design(x, "mu", required = TRUE)

  if (is.null(by)) {
    by <- .regplot_detect_interaction_by(x, mod_name)
  }

  if (is.null(by)) {
    return(NULL)
  }

  mod_names <- design[["predictors"]]

  if (!is.character(by) || length(by) != 1L) {
    stop("'by' must be NULL or a single moderator name.", call. = FALSE)
  }
  if (by == mod_name) {
    stop("'by' must be different from 'mod'.", call. = FALSE)
  }
  if (!by %in% mod_names) {
    stop("Moderator '", by, "' not found. ",
         "Available moderators: ", paste(mod_names, collapse = ", "),
         call. = FALSE)
  }

  by_values <- mods_data[[by]]
  predictor_type <- design[["predictor_types"]][[by]]

  if (identical(predictor_type, "factor") ||
      is.factor(by_values) || is.character(by_values)) {
    by_type <- "categorical"
    by_values <- factor(
      by_values,
      levels = if (!is.null(design[["xlevels"]][[by]])) {
        design[["xlevels"]][[by]]
      } else {
        levels(factor(by_values))
      }
    )
  } else {
    by_type <- "continuous"
  }

  return(list(
    name = by,
    type = by_type,
    data = by_values
  ))
}


# ---------------------------------------------------------------------------- #
# .regplot_detect_interaction_by
# ---------------------------------------------------------------------------- #
#
# Auto-select the second variable from a unique two-way interaction containing
# the plotted moderator.
#
# ---------------------------------------------------------------------------- #
.regplot_detect_interaction_by <- function(x, mod_name) {

  mods_data <- x[["data"]][["mods"]]

  if (is.null(mods_data)) {
    return(NULL)
  }

  term_labels <- .fitted_formula_terms(
    object            = x,
    parameter         = "mu",
    include_intercept = FALSE,
    display           = TRUE
  )
  term_labels <- term_labels[grepl(":", term_labels, fixed = TRUE)]

  if (length(term_labels) == 0L) {
    return(NULL)
  }

  candidates <- character()

  for (term in term_labels) {
    variables <- strsplit(term, ":", fixed = TRUE)[[1]]
    variables <- trimws(variables)

    if (length(variables) == 2L && mod_name %in% variables) {
      candidates <- c(candidates, setdiff(variables, mod_name))
    }
  }

  candidates <- unique(candidates[candidates %in% names(mods_data)])

  if (length(candidates) == 0L) {
    return(NULL)
  }
  if (length(candidates) > 1L) {
    stop("The selected moderator enters multiple interactions. ",
         "Please specify 'by'. Candidate moderators: ",
         paste(candidates, collapse = ", "), ".", call. = FALSE)
  }

  return(candidates)
}


# ---------------------------------------------------------------------------- #
# .regplot_band_data_continuous
# ---------------------------------------------------------------------------- #
#
# Build continuous interval-band data with one row per polygon vertex while
# preserving the underlying interval bounds for programmatic access.
#
# @param xpred numeric vector of prediction x values
# @param lower numeric vector of lower interval bounds
# @param upper numeric vector of upper interval bounds
#
# @return data.frame with consistent polygon and interval metadata
#
# ---------------------------------------------------------------------------- #
.regplot_band_data_continuous <- function(xpred, lower, upper, group = "All",
                                          group_id = 1L) {

  band_x     <- c(xpred, rev(xpred))
  band_lower <- c(lower, rev(lower))
  band_upper <- c(upper, rev(upper))

  data.frame(
    x     = unname(band_x),
    y     = unname(c(lower, rev(upper))),
    lower = unname(band_lower),
    upper = unname(band_upper),
    xpred = unname(band_x),
    group = rep(group, length(band_x)),
    group_id = rep(group_id, length(band_x))
  )
}


# ---------------------------------------------------------------------------- #
# .regplot_prediction_grid
# ---------------------------------------------------------------------------- #
#
# Build moderator values for predictions. The selected moderator forms the
# x-axis; an optional `by` moderator expands the grid into interaction lines.
#
# ---------------------------------------------------------------------------- #
.regplot_prediction_grid <- function(x, mod_name, mod_type, mod_data,
                                     by_info, at, digits) {

  mods_data <- x[["data"]][["mods"]]
  design    <- .fitted_formula_design(x, "mu", required = TRUE)

  if (mod_type == "continuous") {
    if (is.null(at)) {
      x_range <- range(mod_data)
      at_pred <- seq(x_range[1], x_range[2], length.out = 101)
    } else {
      at_pred <- at
    }
    x_values  <- at_pred
    x_numeric <- at_pred
    x_levels  <- NULL
  } else {
    x_levels  <- levels(mod_data)
    at_pred   <- x_levels
    x_values  <- factor(at_pred, levels = x_levels)
    x_numeric <- seq_along(at_pred)
  }

  if (is.null(by_info)) {
    by_values <- list(
      values = NULL,
      labels = "All",
      levels = NULL
    )
  } else {
    by_values <- .regplot_by_values(by_info, digits)
  }

  grid <- expand.grid(
    x_id     = seq_along(at_pred),
    group_id = seq_along(by_values[["labels"]]),
    KEEP.OUT.ATTRS = FALSE
  )
  grid[["group"]] <- by_values[["labels"]][grid[["group_id"]]]
  grid[["x"]]     <- x_numeric[grid[["x_id"]]]
  if (mod_type == "categorical") {
    grid[["level"]] <- at_pred[grid[["x_id"]]]
  }

  n_pred       <- nrow(grid)
  newdata_list <- list()

  for (nm in design[["predictors"]]) {
    if (nm == mod_name) {
      if (mod_type == "continuous") {
        newdata_list[[nm]] <- x_values[grid[["x_id"]]]
      } else {
        newdata_list[[nm]] <- factor(at_pred[grid[["x_id"]]], levels = x_levels)
      }
    } else if (!is.null(by_info) && nm == by_info[["name"]]) {
      if (by_info[["type"]] == "continuous") {
        newdata_list[[nm]] <- by_values[["values"]][grid[["group_id"]]]
      } else {
        newdata_list[[nm]] <- factor(
          by_values[["values"]][grid[["group_id"]]],
          levels = by_values[["levels"]]
        )
      }
    } else {
      other_vals <- mods_data[[nm]]
      if (is.factor(other_vals) || is.character(other_vals)) {
        other_levels <- if (!is.null(design[["xlevels"]][[nm]])) {
          design[["xlevels"]][[nm]]
        } else {
          levels(factor(other_vals))
        }
        newdata_list[[nm]] <- factor(
          rep(other_levels[1], n_pred),
          levels = other_levels
        )
      } else {
        newdata_list[[nm]] <- rep(mean(other_vals), n_pred)
      }
    }
  }

  return(list(
    newdata  = as.data.frame(newdata_list),
    grid     = grid,
    at_pred  = at_pred,
    groups   = by_values[["labels"]],
    x_levels = x_levels
  ))
}


# ---------------------------------------------------------------------------- #
# .regplot_by_values
# ---------------------------------------------------------------------------- #
#
# Values used for the grouping moderator.
#
# ---------------------------------------------------------------------------- #
.regplot_by_values <- function(by_info, digits) {

  by_data <- by_info[["data"]]

  if (by_info[["type"]] == "continuous") {
    by_mean <- mean(by_data)
    by_sd   <- stats::sd(by_data)
    values  <- c(by_mean - by_sd, by_mean, by_mean + by_sd)
    labels  <- c("Mean - 1 SD", "Mean", "Mean + 1 SD")

    if (!is.finite(by_sd) || by_sd == 0) {
      values <- by_mean
      labels <- format(round(by_mean, digits = digits), trim = TRUE)
    }

    labels <- paste0(by_info[["name"]], ": ", labels)
    return(list(values = values, labels = labels, levels = NULL))
  }

  by_factor <- factor(by_data)

  return(list(
    values = levels(by_factor),
    labels = levels(by_factor),
    levels = levels(by_factor)
  ))
}


# ---------------------------------------------------------------------------- #
# .regplot_observed_groups
# ---------------------------------------------------------------------------- #
#
# Assign observed studies to plotting groups.
#
# ---------------------------------------------------------------------------- #
.regplot_observed_groups <- function(by_info, groups, n) {

  if (is.null(by_info)) {
    return(list(
      group    = rep("All", n),
      group_id = rep(1L, n)
    ))
  }

  by_data <- by_info[["data"]]

  if (by_info[["type"]] == "categorical") {
    group <- as.character(factor(by_data, levels = groups))
  } else {
    by_mean <- mean(by_data)
    by_sd   <- stats::sd(by_data)

    if (!is.finite(by_sd) || by_sd == 0 || length(groups) == 1L) {
      group <- rep(groups[1], length(by_data))
    } else {
      by_values <- c(by_mean - by_sd, by_mean, by_mean + by_sd)
      breaks    <- c(-Inf, mean(by_values[1:2]), mean(by_values[2:3]), Inf)
      group     <- as.character(cut(by_data, breaks = breaks, labels = groups))
    }
  }

  return(list(
    group    = group,
    group_id = match(group, groups)
  ))
}


# ---------------------------------------------------------------------------- #
# .regplot_add_dummy_outcome
# ---------------------------------------------------------------------------- #
#
# Add outcome columns for prediction paths that use sampling bias inputs.
#
# ---------------------------------------------------------------------------- #
.regplot_add_dummy_outcome <- function(x, newdata, n_pred, sei) {

  outcome_type <- .outcome_type(x)

  if (outcome_type == "norm") {
    newdata[["yi"]]  <- rep(0, n_pred)
    newdata[["sei"]] <- rep(sei, n_pred)
  } else if (outcome_type == "bin") {
    newdata[["ai"]]  <- rep(0, n_pred)
    newdata[["ci"]]  <- rep(0, n_pred)
    newdata[["n1i"]] <- rep(0, n_pred)
    newdata[["n2i"]] <- rep(0, n_pred)
  } else if (outcome_type == "pois") {
    newdata[["x1i"]] <- rep(0, n_pred)
    newdata[["x2i"]] <- rep(0, n_pred)
    newdata[["t1i"]] <- rep(0, n_pred)
    newdata[["t2i"]] <- rep(0, n_pred)
  }

  return(newdata)
}


# ---------------------------------------------------------------------------- #
# .regplot_band_data_categorical
# ---------------------------------------------------------------------------- #
#
# Categorical intervals are rendered as box-style summaries.
#
# ---------------------------------------------------------------------------- #
.regplot_band_data_categorical <- function(grid, middle, lower, upper) {

  out <- data.frame(
    x        = grid[["x"]],
    y        = middle,
    middle   = middle,
    lower    = lower,
    upper    = upper,
    group    = grid[["group"]],
    group_id = grid[["group_id"]],
    level    = grid[["level"]],
    stringsAsFactors = FALSE
  )

  return(out)
}


# ---------------------------------------------------------------------------- #
# .regplot_data
# ---------------------------------------------------------------------------- #
#
# Generate data for regression plot.
#
# @param x        brma object
# @param mod_name name of moderator variable
# @param mod_type "continuous" or "categorical"
# @param mod_data moderator data values
# @param by_info  list with grouping variable info (or NULL)
# @param pred     logical; show prediction line
# @param ci       logical; show CI bands
# @param pi       logical; show PI bands
# @param si       logical; show SI bands
# @param level    confidence level (0-100)
# @param at       values at which to evaluate predictions
# @param psize    point sizes (or NULL for auto)
# @param plim     point size range
# @param transf   transformation function
# @param xlim     x-axis limits
# @param ylim     y-axis limits
# @param xlab     x-axis label
# @param ylab     y-axis label
# @param refline       reference line position
# @param sampling_bias logical; incorporate bias into predictions
# @param max_samples   maximum posterior samples for plot summaries
# @param reference_sei numeric; reference standard error for sampling paths
# @param dots          graphical parameters
#
# @return list with plot data components
#
# ---------------------------------------------------------------------------- #
.regplot_data <- function(x, mod_name, mod_type, mod_data, by_info,
                          pred, ci, pi, si, level, at, digits, psize, plim,
                          transf, xlim, ylim, xlab, ylab, refline,
                          sampling_bias, max_samples, reference_sei, dots) {

  yi      <- .outcome_data_yi(x)
  sei_obs <- .outcome_data_sei(x)
  vi      <- sei_obs^2
  K       <- length(yi)
  se_rep  <- if (is.null(reference_sei)) stats::median(sei_obs) else reference_sei

  alpha <- (100 - level) / 100
  probs <- c(alpha / 2, 1 - alpha / 2)

  if (is.null(psize)) {
    weights   <- 1 / vi
    weights_n <- (weights - min(weights)) / (max(weights) - min(weights) + 1e-10)
    psize     <- plim[1] + weights_n * (plim[2] - plim[1])
  } else if (length(psize) == 1) {
    psize <- rep(psize, K)
  }

  yi_plot <- yi
  if (!is.null(transf)) {
    yi_plot <- transf(yi)
  }

  prediction_grid <- .regplot_prediction_grid(
    x         = x,
    mod_name  = mod_name,
    mod_type  = mod_type,
    mod_data  = mod_data,
    by_info   = by_info,
    at        = at,
    digits    = digits
  )
  grid   <- prediction_grid[["grid"]]
  groups <- prediction_grid[["groups"]]
  n_pred <- nrow(grid)

  prediction_se <- if (sampling_bias) se_rep else 0
  newdata <- .regplot_add_dummy_outcome(
    x       = x,
    newdata = prediction_grid[["newdata"]],
    n_pred  = n_pred,
    sei     = prediction_se
  )

  posterior_samples <- .get_posterior_samples(x[["fit"]])
  selected_rows     <- .thin_sample_rows(nrow(posterior_samples), max_samples)
  if (!is.null(selected_rows)) {
    posterior_samples <- posterior_samples[selected_rows, , drop = FALSE]
  }

  pred_samples <- predict.brma(
    object             = x,
    newdata            = newdata,
    type               = "terms",
    bias_adjusted      = !sampling_bias,
    quiet              = TRUE,
    .posterior_samples = posterior_samples
  )
  pred_samples <- as.matrix(pred_samples)

  if (nrow(pred_samples) != nrow(posterior_samples)) {
    stop("Posterior sample count mismatch in regplot().", call. = FALSE)
  }

  pred_mean  <- colMeans(pred_samples)
  pred_lower <- apply(pred_samples, 2, stats::quantile, probs = probs[1], names = FALSE)
  pred_upper <- apply(pred_samples, 2, stats::quantile, probs = probs[2], names = FALSE)

  if (pi || si) {
    is_scale      <- .is_scale(x)
    is_multilevel <- .is_multilevel(x)

    scale_data    <- NULL
    scale_formula <- NULL
    if (is_scale) {
      new_data_prepared <- .prepare_newdata(object = x, newdata = newdata, type = "estimate")
      scale_data        <- new_data_prepared[["scale"]]
      scale_formula     <- .create_fit_formula_list(data = new_data_prepared, "scale")
    }

    tau_result <- .evaluate.brma.tau(
      fit               = x[["fit"]],
      scale_data        = scale_data,
      scale_formula     = scale_formula,
      scale_priors      = x[["priors"]][["scale"]],
      is_scale          = is_scale,
      is_multilevel     = is_multilevel,
      K                 = n_pred,
      posterior_samples = posterior_samples
    )
    tau_total <- sqrt(tau_result[["tau_within"]]^2 + tau_result[["tau_between"]]^2)
  }

  if (pi) {
    pi_bounds <- .regplot_mixture_interval_quantiles(
      mean_samples = pred_samples,
      sd_samples   = tau_total,
      probs        = probs
    )
    pi_lower <- pi_bounds[["lower"]]
    pi_upper <- pi_bounds[["upper"]]
  } else {
    pi_lower <- NULL
    pi_upper <- NULL
  }

  if (si) {
    sd_si <- sqrt(tau_total^2 + se_rep^2)

    if (sampling_bias && .is_weightfunction(x)) {
      si_bounds <- .regplot_selection_mixture_interval_quantiles(
        x                 = x,
        mean_samples      = pred_samples,
        sd_samples        = sd_si,
        se                = se_rep,
        probs             = probs,
        posterior_samples = posterior_samples
      )
    } else {
      si_bounds <- .regplot_mixture_interval_quantiles(
        mean_samples = pred_samples,
        sd_samples   = sd_si,
        probs        = probs
      )
    }

    si_lower <- si_bounds[["lower"]]
    si_upper <- si_bounds[["upper"]]
  } else {
    si_lower <- NULL
    si_upper <- NULL
  }

  if (!is.null(transf)) {
    pred_mean  <- transf(pred_mean)
    ci_lo      <- transf(pred_lower)
    ci_hi      <- transf(pred_upper)
    pred_lower <- pmin(ci_lo, ci_hi)
    pred_upper <- pmax(ci_lo, ci_hi)

    if (pi) {
      pi_lo    <- transf(pi_lower)
      pi_hi    <- transf(pi_upper)
      pi_lower <- pmin(pi_lo, pi_hi)
      pi_upper <- pmax(pi_lo, pi_hi)
    }
    if (si) {
      si_lo    <- transf(si_lower)
      si_hi    <- transf(si_upper)
      si_lower <- pmin(si_lo, si_hi)
      si_upper <- pmax(si_lo, si_hi)
    }
  }

  if (is.null(xlim)) {
    if (mod_type == "continuous") {
      xlim <- range(pretty(range(mod_data)))
    } else {
      xlim <- c(0.5, length(levels(mod_data)) + 0.5)
    }
  }

  if (is.null(ylim)) {
    all_y <- c(yi_plot, pred_mean)
    if (ci) all_y <- c(all_y, pred_lower, pred_upper)
    if (pi) all_y <- c(all_y, pi_lower, pi_upper)
    if (si) all_y <- c(all_y, si_lower, si_upper)
    ylim <- range(pretty(range(all_y, na.rm = TRUE)))
  }

  if (is.null(xlab)) xlab <- mod_name
  if (is.null(ylab)) ylab <- "Observed Effect Size"

  observed_groups <- .regplot_observed_groups(
    by_info = by_info,
    groups  = groups,
    n       = K
  )

  df_points <- data.frame(
    x        = if (mod_type == "continuous") mod_data else as.numeric(mod_data),
    y        = yi_plot,
    size     = psize,
    group    = observed_groups[["group"]],
    group_id = observed_groups[["group_id"]],
    stringsAsFactors = FALSE
  )

  if (mod_type == "categorical") {
    df_points$level <- mod_data
  }

  df_pred <- NULL
  if (pred) {
    df_pred <- data.frame(
      x        = grid[["x"]],
      y        = pred_mean,
      group    = grid[["group"]],
      group_id = grid[["group_id"]],
      stringsAsFactors = FALSE
    )
    if (mod_type == "categorical") {
      df_pred[["level"]] <- grid[["level"]]
    }
  }

  df_ci <- NULL
  if (ci) {
    if (mod_type == "continuous") {
      df_ci <- do.call(rbind, lapply(seq_along(groups), function(i) {
        rows <- grid[["group_id"]] == i
        .regplot_band_data_continuous(
          xpred    = grid[["x"]][rows],
          lower    = pred_lower[rows],
          upper    = pred_upper[rows],
          group    = groups[i],
          group_id = i
        )
      }))
      rownames(df_ci) <- NULL
    } else {
      df_ci <- .regplot_band_data_categorical(grid, pred_mean, pred_lower, pred_upper)
    }
  }

  df_pi <- NULL
  if (pi) {
    if (mod_type == "continuous") {
      df_pi <- do.call(rbind, lapply(seq_along(groups), function(i) {
        rows <- grid[["group_id"]] == i
        .regplot_band_data_continuous(
          xpred    = grid[["x"]][rows],
          lower    = pi_lower[rows],
          upper    = pi_upper[rows],
          group    = groups[i],
          group_id = i
        )
      }))
      rownames(df_pi) <- NULL
    } else {
      df_pi <- .regplot_band_data_categorical(grid, pred_mean, pi_lower, pi_upper)
    }
  }

  df_si <- NULL
  if (si) {
    if (mod_type == "continuous") {
      df_si <- do.call(rbind, lapply(seq_along(groups), function(i) {
        rows <- grid[["group_id"]] == i
        .regplot_band_data_continuous(
          xpred    = grid[["x"]][rows],
          lower    = si_lower[rows],
          upper    = si_upper[rows],
          group    = groups[i],
          group_id = i
        )
      }))
      rownames(df_si) <- NULL
    } else {
      df_si <- .regplot_band_data_categorical(grid, pred_mean, si_lower, si_upper)
    }
  }

  df_refline <- NULL
  if (!is.null(refline)) {
    df_refline <- data.frame(y = refline)
  }

  return(list(
    points   = df_points,
    pred     = df_pred,
    ci       = df_ci,
    pi       = df_pi,
    si       = df_si,
    refline  = df_refline,
    xlim     = xlim,
    ylim     = ylim,
    xlab     = xlab,
    ylab     = ylab,
    mod_type = mod_type,
    mod_name = mod_name,
    by_name  = if (!is.null(by_info)) by_info[["name"]] else NULL,
    by_type  = if (!is.null(by_info)) by_info[["type"]] else NULL,
    groups   = groups,
    levels   = if (mod_type == "categorical") levels(mod_data) else NULL,
    sei      = se_rep
  ))
}


# ---------------------------------------------------------------------------- #
# .regplot_mixture_interval_quantiles
# ---------------------------------------------------------------------------- #
#
# Deterministic quantiles of posterior mixtures of normal distributions.
#
# ---------------------------------------------------------------------------- #
.has_native_regplot_mixture <- function() {

  return(is.loaded("RoBMA_regplot_normal_mixture_interval", PACKAGE = "RoBMA"))
}

.has_native_regplot_selection_mixture <- function() {

  return(is.loaded("RoBMA_regplot_selnorm_mixture_interval", PACKAGE = "RoBMA"))
}


.regplot_mixture_interval_quantiles <- function(mean_samples, sd_samples, probs) {

  mean_samples <- as.matrix(mean_samples)
  sd_samples   <- as.matrix(sd_samples)

  if (.has_native_regplot_mixture()) {
    return(.Call(
      "RoBMA_regplot_normal_mixture_interval",
      .native_numeric_matrix(mean_samples),
      .native_numeric_matrix(sd_samples),
      .native_numeric_vector(probs),
      PACKAGE = "RoBMA"
    ))
  }

  return(.regplot_mixture_interval_quantiles_r(
    mean_samples = mean_samples,
    sd_samples   = sd_samples,
    probs        = probs
  ))
}


.regplot_mixture_interval_quantiles_r <- function(mean_samples, sd_samples, probs) {

  lower <- numeric(ncol(mean_samples))
  upper <- numeric(ncol(mean_samples))

  for (i in seq_len(ncol(mean_samples))) {
    cdf_fun <- function(q) {
      .regplot_normal_mixture_cdf(
        q    = q,
        mean = mean_samples[, i],
        sd   = sd_samples[, i]
      )
    }

    lower[i] <- .regplot_mixture_quantile(probs[1], mean_samples[, i], sd_samples[, i], cdf_fun)
    upper[i] <- .regplot_mixture_quantile(probs[2], mean_samples[, i], sd_samples[, i], cdf_fun)
  }

  return(list(lower = lower, upper = upper))
}


# ---------------------------------------------------------------------------- #
# .regplot_selection_mixture_interval_quantiles
# ---------------------------------------------------------------------------- #
#
# Deterministic quantiles of observed-effect mixtures with selection branches.
#
# ---------------------------------------------------------------------------- #
.regplot_selection_mixture_interval_quantiles <- function(x, mean_samples,
                                                          sd_samples, se,
                                                          probs,
                                                          posterior_samples = NULL) {

  mean_samples     <- as.matrix(mean_samples)
  sd_samples       <- as.matrix(sd_samples)
  setup            <- .regplot_selection_setup(
    x                 = x,
    posterior_samples = posterior_samples
  )
  effect_direction <- .effect_direction(x)

  if (!is.null(setup[["selection"]]) &&
      se > 0 &&
      .has_native_regplot_selection_mixture()) {
    return(.regplot_selnorm_mixture_interval_quantiles(
      mean_samples      = mean_samples,
      sd_samples        = sd_samples,
      se                = se,
      probs             = probs,
      selection_context = setup[["selection"]]
    ))
  }

  return(.regplot_selection_mixture_interval_quantiles_r(
    mean_samples     = mean_samples,
    sd_samples       = sd_samples,
    se               = se,
    probs            = probs,
    setup            = setup,
    effect_direction = effect_direction
  ))
}

.regplot_selnorm_mixture_interval_quantiles <- function(mean_samples,
                                                        sd_samples, se,
                                                        probs,
                                                        selection_context) {

  .selection_require_step_evaluable(
    selection_context,
    ".regplot_selection_mixture_interval_quantiles()"
  )

  return(.Call(
    "RoBMA_regplot_selnorm_mixture_interval",
    .native_numeric_matrix(mean_samples),
    .native_numeric_matrix(sd_samples),
    .native_numeric_vector(se),
    .native_numeric_vector(probs),
    .native_numeric_matrix(selection_context[["omega"]]),
    .native_numeric_vector(selection_context[["alpha"]]),
    .native_integer_vector(selection_context[["phack_kind"]]),
    .native_integer_vector(selection_context[["kernel_mode"]]),
    .native_numeric_vector(selection_context[["z_lower"]]),
    .native_numeric_vector(selection_context[["z_upper"]]),
    .native_integer_vector(selection_context[["sign"]]),
    .native_integer_vector(selection_context[["phack_q"]]),
    .native_numeric_vector(selection_context[["phack_z_source"]]),
    .native_numeric_vector(selection_context[["phack_z_dest"]]),
    .native_numeric_vector(selection_context[["segments"]][["bounds"]]),
    .native_integer_vector(selection_context[["segments"]][["step_bin"]]),
    .native_integer_vector(selection_context[["segments"]][["phack_region"]]),
    .selection_telescope_probabilities(selection_context),
    PACKAGE = "RoBMA"
  ))
}


.regplot_selection_mixture_interval_quantiles_r <- function(mean_samples,
                                                            sd_samples, se,
                                                            probs, setup,
                                                            effect_direction) {

  lower <- numeric(ncol(mean_samples))
  upper <- numeric(ncol(mean_samples))

  for (i in seq_len(ncol(mean_samples))) {
    cdf_fun <- function(q) {
      .regplot_selection_mixture_cdf(
        q                = q,
        mean             = mean_samples[, i],
        sd               = sd_samples[, i],
        se               = se,
        setup            = setup,
        effect_direction = effect_direction
      )
    }

    lower[i] <- .regplot_mixture_quantile(probs[1], mean_samples[, i], sd_samples[, i], cdf_fun)
    upper[i] <- .regplot_mixture_quantile(probs[2], mean_samples[, i], sd_samples[, i], cdf_fun)
  }

  return(list(lower = lower, upper = upper))
}


# ---------------------------------------------------------------------------- #
# .regplot_mixture_quantile
# ---------------------------------------------------------------------------- #
#
# Invert a posterior-mixture CDF.
#
# ---------------------------------------------------------------------------- #
.regplot_mixture_quantile <- function(p, mean, sd, cdf_fun) {

  eps_sd <- sqrt(.Machine$double.eps)

  if (all(sd < eps_sd)) {
    return(unname(stats::quantile(mean, probs = p, names = FALSE, type = 8)))
  }

  spread <- pmax(sd, eps_sd)
  lower  <- min(mean - 10 * spread, na.rm = TRUE)
  upper  <- max(mean + 10 * spread, na.rm = TRUE)

  if (!is.finite(lower) || !is.finite(upper)) {
    return(NA_real_)
  }
  if (lower >= upper) {
    lower <- lower - 1
    upper <- upper + 1
  }

  obj_fun     <- function(q) cdf_fun(q) - p
  lower_value <- obj_fun(lower)
  upper_value <- obj_fun(upper)
  step        <- max(spread, na.rm = TRUE)

  if (!is.finite(step) || step <= 0) {
    step <- max(1, abs(mean), na.rm = TRUE)
  }

  for (i in seq_len(25)) {
    if (lower_value <= 0 && upper_value >= 0) {
      break
    }
    if (lower_value > 0) {
      lower       <- lower - step
      lower_value <- obj_fun(lower)
    }
    if (upper_value < 0) {
      upper       <- upper + step
      upper_value <- obj_fun(upper)
    }
    step <- step * 2
  }

  if (lower_value > 0 || upper_value < 0) {
    return(.regplot_grid_quantile(p, lower, upper, cdf_fun))
  }

  out <- tryCatch(
    stats::uniroot(obj_fun, interval = c(lower, upper), tol = 1e-6)[["root"]],
    error = function(e) NA_real_
  )

  if (is.na(out)) {
    out <- .regplot_grid_quantile(p, lower, upper, cdf_fun)
  }

  return(out)
}


# ---------------------------------------------------------------------------- #
# .regplot_normal_mixture_cdf
# ---------------------------------------------------------------------------- #
.regplot_normal_mixture_cdf <- function(q, mean, sd) {

  eps_sd     <- sqrt(.Machine$double.eps)
  cdf_values <- rep(NA_real_, length(mean))
  zero_sd    <- sd < eps_sd

  if (any(zero_sd)) {
    cdf_values[zero_sd] <- as.numeric(q >= mean[zero_sd])
  }
  if (any(!zero_sd)) {
    cdf_values[!zero_sd] <- stats::pnorm(
      q,
      mean = mean[!zero_sd],
      sd   = sd[!zero_sd]
    )
  }

  cdf_values <- pmin(pmax(cdf_values, 0), 1)
  return(base::mean(cdf_values))
}


# ---------------------------------------------------------------------------- #
# .regplot_selection_mixture_cdf
# ---------------------------------------------------------------------------- #
.regplot_selection_mixture_cdf <- function(q, mean, sd, se, setup,
                                           effect_direction) {

  eps_sd     <- sqrt(.Machine$double.eps)
  cdf_values <- rep(NA_real_, length(mean))
  zero_sd    <- sd < eps_sd

  if (any(zero_sd)) {
    cdf_values[zero_sd] <- as.numeric(q >= mean[zero_sd])
  }

  normal_rows <- !setup[["is_weightfunction"]] & !zero_sd
  if (any(normal_rows)) {
    cdf_values[normal_rows] <- stats::pnorm(
      q,
      mean = mean[normal_rows],
      sd   = sd[normal_rows]
    )
  }

  selected_rows <- setup[["is_weightfunction"]] & !zero_sd
  if (any(selected_rows)) {
    setup[["mu"]] <- mean
    rows <- which(selected_rows)
    cdf_values[rows] <- .funnel_selected_cdf(
      q                = q,
      rows             = rows,
      se               = se,
      total_sd         = sd,
      setup            = setup,
      effect_direction = effect_direction
    )
  }

  cdf_values <- pmin(pmax(cdf_values, 0), 1)
  return(base::mean(cdf_values))
}


# ---------------------------------------------------------------------------- #
# .regplot_selection_setup
# ---------------------------------------------------------------------------- #
.regplot_selection_setup <- function(x, posterior_samples = NULL) {

  if (is.null(posterior_samples)) {
    posterior_samples <- .get_posterior_samples(x[["fit"]])
  }
  S              <- nrow(posterior_samples)
  bias_indicator <- .extract_bias_indicator(x, posterior_samples = posterior_samples)
  selection      <- .selection_context(
    object            = x,
    posterior_samples = posterior_samples
  )
  use_normal     <- if (is.null(selection)) {
    rep(TRUE, S)
  } else {
    selection[["use_normal"]]
  }

  return(list(
    bias_indicator    = bias_indicator,
    is_weightfunction = !use_normal,
    selection         = selection
  ))
}


# ---------------------------------------------------------------------------- #
# .regplot_grid_quantile
# ---------------------------------------------------------------------------- #
.regplot_grid_quantile <- function(p, lower, upper, cdf_fun) {

  grid <- seq(lower, upper, length.out = 1000)
  cdf  <- vapply(grid, cdf_fun, numeric(1))
  index <- which(cdf >= p)[1]

  if (is.na(index)) {
    return(grid[length(grid)])
  }

  return(grid[index])
}


# ---------------------------------------------------------------------------- #
# Plotting helpers
# ---------------------------------------------------------------------------- #
.regplot_has_shade <- function(shade) {
  return(isTRUE(shade))
}

.regplot_palette <- function(groups, default_col) {

  n <- length(groups)
  if (n == 1L) {
    return(stats::setNames(default_col[1], groups))
  }

  if (length(default_col) >= n && length(unique(default_col)) > 1L) {
    cols <- default_col[seq_len(n)]
  } else {
    cols <- grDevices::hcl.colors(n, palette = "Dark 3")
  }

  return(stats::setNames(cols, groups))
}

.regplot_dodge_x <- function(x, group_id, n_groups, dodge_width) {

  if (n_groups <= 1L) {
    return(x)
  }

  offset <- (group_id - (n_groups + 1) / 2) * dodge_width / n_groups
  return(x + offset)
}

.regplot_jitter_values <- function(n, amount) {

  if (amount <= 0) {
    return(rep(0, n))
  }

  old_seed <- if (exists(".Random.seed", envir = .GlobalEnv)) {
    get(".Random.seed", envir = .GlobalEnv)
  } else {
    NULL
  }
  on.exit({
    if (!is.null(old_seed)) {
      assign(".Random.seed", old_seed, envir = .GlobalEnv)
    } else if (exists(".Random.seed", envir = .GlobalEnv)) {
      rm(".Random.seed", envir = .GlobalEnv)
    }
  })

  set.seed(42)
  return(stats::runif(n, -amount, amount))
}

.regplot_band_edges <- function(df_band) {

  out <- do.call(rbind, lapply(split(df_band, df_band[["group"]]), function(df_group) {
    n <- nrow(df_group) / 2
    data.frame(
      x        = rep(df_group[["x"]][seq_len(n)], 2),
      y        = c(df_group[["lower"]][seq_len(n)], df_group[["upper"]][seq_len(n)]),
      edge     = rep(c("lower", "upper"), each = n),
      group    = rep(df_group[["group"]][1], 2 * n),
      group_id = rep(df_group[["group_id"]][1], 2 * n),
      stringsAsFactors = FALSE
    )
  }))

  rownames(out) <- NULL
  return(out)
}

.regplot_draw_continuous_band_base <- function(df_band, fill_col, alpha,
                                               border_col, shade, lwd) {

  if (is.null(df_band)) {
    return(invisible(NULL))
  }

  for (group in unique(df_band[["group"]])) {
    df_group <- df_band[df_band[["group"]] == group, , drop = FALSE]
    n        <- nrow(df_group) / 2

    if (.regplot_has_shade(shade)) {
      graphics::polygon(
        df_group[["x"]],
        df_group[["y"]],
        col    = grDevices::adjustcolor(fill_col[group], alpha.f = alpha),
        border = NA
      )
    }

    graphics::lines(
      df_group[["x"]][seq_len(n)],
      df_group[["lower"]][seq_len(n)],
      col = border_col[group],
      lwd = max(1, lwd / 2)
    )
    graphics::lines(
      df_group[["x"]][seq_len(n)],
      df_group[["upper"]][seq_len(n)],
      col = border_col[group],
      lwd = max(1, lwd / 2)
    )
  }

  return(invisible(NULL))
}

.regplot_draw_categorical_band_base <- function(df_band, fill_col, alpha,
                                                border_col, shade, width,
                                                n_groups, dodge_width, lwd) {

  if (is.null(df_band)) {
    return(invisible(NULL))
  }

  x <- .regplot_dodge_x(df_band[["x"]], df_band[["group_id"]], n_groups, dodge_width)
  w <- width / max(1, n_groups)

  for (i in seq_len(nrow(df_band))) {
    group <- df_band[["group"]][i]
    graphics::rect(
      xleft   = x[i] - w / 2,
      ybottom = df_band[["lower"]][i],
      xright  = x[i] + w / 2,
      ytop    = df_band[["upper"]][i],
      col     = if (.regplot_has_shade(shade)) grDevices::adjustcolor(fill_col[group], alpha.f = alpha) else NA,
      border  = border_col[group],
      lwd     = max(1, lwd / 2)
    )
    graphics::segments(
      x0  = x[i] - w / 2,
      y0  = df_band[["middle"]][i],
      x1  = x[i] + w / 2,
      y1  = df_band[["middle"]][i],
      col = border_col[group],
      lwd = max(1, lwd / 2)
    )
  }

  return(invisible(NULL))
}


.regplot_add_continuous_band_ggplot <- function(out, df_band, fill_col, alpha,
                                                border_col, shade, has_by) {

  if (is.null(df_band)) {
    return(out)
  }

  if (.regplot_has_shade(shade)) {
    if (has_by) {
      out <- out +
        ggplot2::geom_polygon(
          data    = df_band,
          mapping = ggplot2::aes(
            x     = .data[["x"]],
            y     = .data[["y"]],
            group = .data[["group"]],
            fill  = .data[["group"]]
          ),
          alpha   = alpha
        )
    } else {
      out <- out +
        ggplot2::geom_polygon(
          data    = df_band,
          mapping = ggplot2::aes(
            x     = .data[["x"]],
            y     = .data[["y"]],
            group = .data[["group"]]
          ),
          fill    = fill_col[1],
          alpha   = alpha
        )
    }
  }

  df_edges <- .regplot_band_edges(df_band)
  if (has_by) {
    out <- out +
      ggplot2::geom_line(
        data    = df_edges,
        mapping = ggplot2::aes(
          x      = .data[["x"]],
          y      = .data[["y"]],
          group  = interaction(.data[["group"]], .data[["edge"]]),
          colour = .data[["group"]]
        ),
        linewidth = 0.35
      )
  } else {
    out <- out +
      ggplot2::geom_line(
        data    = df_edges,
        mapping = ggplot2::aes(
          x     = .data[["x"]],
          y     = .data[["y"]],
          group = .data[["edge"]]
        ),
        colour  = border_col[1],
        linewidth = 0.35
      )
  }

  return(out)
}

.regplot_add_categorical_band_ggplot <- function(out, df_band, fill_col, alpha,
                                                 border_col, shade, has_by,
                                                 width, n_groups, dodge_width) {

  if (is.null(df_band)) {
    return(out)
  }

  df_band[["x_plot"]] <- .regplot_dodge_x(
    df_band[["x"]],
    df_band[["group_id"]],
    n_groups,
    dodge_width
  )
  width <- width / max(1, n_groups)

  if (has_by && .regplot_has_shade(shade)) {
    out <- out +
      ggplot2::geom_crossbar(
        data    = df_band,
        mapping = ggplot2::aes(
          x      = .data[["x_plot"]],
          y      = .data[["middle"]],
          ymin   = .data[["lower"]],
          ymax   = .data[["upper"]],
          colour = .data[["group"]],
          fill   = .data[["group"]]
        ),
        width   = width,
        alpha   = alpha,
        linewidth = 0.35
      )
  } else if (has_by) {
    out <- out +
      ggplot2::geom_crossbar(
        data    = df_band,
        mapping = ggplot2::aes(
          x      = .data[["x_plot"]],
          y      = .data[["middle"]],
          ymin   = .data[["lower"]],
          ymax   = .data[["upper"]],
          colour = .data[["group"]]
        ),
        width   = width,
        fill    = NA,
        alpha   = 1,
        linewidth = 0.35
      )
  } else {
    out <- out +
      ggplot2::geom_crossbar(
        data    = df_band,
        mapping = ggplot2::aes(
          x    = .data[["x_plot"]],
          y    = .data[["middle"]],
          ymin = .data[["lower"]],
          ymax = .data[["upper"]]
        ),
        width   = width,
        colour  = border_col[1],
        fill    = if (.regplot_has_shade(shade)) fill_col[1] else NA,
        alpha   = if (.regplot_has_shade(shade)) alpha else 1,
        linewidth = 0.35
      )
  }

  return(out)
}


# ---------------------------------------------------------------------------- #
# .regplot_plot_base
# ---------------------------------------------------------------------------- #
#
# Create regression plot using base R graphics.
#
# @param data   list of plot data from .regplot_data
# @param dots   list of graphical parameters
# @param legend logical; show legend
#
# @return NULL invisibly
#
# ---------------------------------------------------------------------------- #
.regplot_plot_base <- function(data, dots, legend = TRUE) {

  pch       <- dots[["pch"]]
  col       <- dots[["col"]]
  bg        <- dots[["bg"]]
  lcol      <- dots[["lcol"]]
  lwd       <- dots[["lwd"]]
  col_ci    <- dots[["col.ci"]]
  col_pi    <- dots[["col.pi"]]
  col_si    <- dots[["col.si"]]
  alpha_ci  <- dots[["alpha.ci"]]
  alpha_pi  <- dots[["alpha.pi"]]
  alpha_si  <- dots[["alpha.si"]]
  shade     <- dots[["shade"]]
  main      <- dots[["main"]]
  jitter_am <- dots[["jitter"]]
  box_width <- dots[["box.width"]]
  dodge_w   <- dots[["dodge.width"]]
  las       <- dots[["las"]]

  df_points  <- data$points
  df_pred    <- data$pred
  df_ci      <- data$ci
  df_pi      <- data$pi
  df_si      <- data$si
  df_refline <- data$refline
  mod_type   <- data$mod_type
  groups     <- data$groups
  n_groups   <- length(groups)
  has_by     <- n_groups > 1L

  line_cols <- .regplot_palette(groups, lcol)
  point_cols <- if (has_by) line_cols else stats::setNames(rep(col, n_groups), groups)
  point_bgs  <- if (has_by) {
    stats::setNames(grDevices::adjustcolor(line_cols, alpha.f = 0.45), groups)
  } else {
    stats::setNames(rep(bg, n_groups), groups)
  }

  if (has_by) {
    ci_cols <- pi_cols <- si_cols <- line_cols
  } else {
    ci_cols <- stats::setNames(rep(col_ci, n_groups), groups)
    pi_cols <- stats::setNames(rep(col_pi, n_groups), groups)
    si_cols <- stats::setNames(rep(col_si, n_groups), groups)
  }

  if (mod_type == "categorical") {
    graphics::plot(
      NA, NA,
      xlim = data$xlim,
      ylim = data$ylim,
      xlab = data$xlab,
      ylab = data$ylab,
      main = main,
      type = "n",
      xaxt = "n",
      las  = las
    )
    graphics::axis(
      1,
      at     = seq_along(data$levels),
      labels = data$levels,
      las    = las
    )
  } else {
    graphics::plot(
      NA, NA,
      xlim = data$xlim,
      ylim = data$ylim,
      xlab = data$xlab,
      ylab = data$ylab,
      main = main,
      type = "n",
      las  = las
    )
  }

  if (!is.null(df_refline)) {
    graphics::abline(h = df_refline$y, lty = "dashed", col = "gray50")
  }

  if (!is.null(df_si)) {
    if (mod_type == "continuous") {
      .regplot_draw_continuous_band_base(df_si, si_cols, alpha_si, si_cols, shade, lwd)
    } else {
      .regplot_draw_categorical_band_base(
        df_si, si_cols, alpha_si, si_cols, shade,
        box_width * 1.4, n_groups, dodge_w, lwd
      )
    }
  }

  if (!is.null(df_pi)) {
    if (mod_type == "continuous") {
      .regplot_draw_continuous_band_base(df_pi, pi_cols, alpha_pi, pi_cols, shade, lwd)
    } else {
      .regplot_draw_categorical_band_base(
        df_pi, pi_cols, alpha_pi, pi_cols, shade,
        box_width * 1.15, n_groups, dodge_w, lwd
      )
    }
  }

  if (!is.null(df_ci)) {
    if (mod_type == "continuous") {
      .regplot_draw_continuous_band_base(df_ci, ci_cols, alpha_ci, ci_cols, shade, lwd)
    } else {
      .regplot_draw_categorical_band_base(
        df_ci, ci_cols, alpha_ci, ci_cols, shade,
        box_width, n_groups, dodge_w, lwd
      )
    }
  }

  if (!is.null(df_pred)) {
    if (mod_type == "continuous") {
      for (group in groups) {
        df_group <- df_pred[df_pred[["group"]] == group, , drop = FALSE]
        graphics::lines(df_group$x, df_group$y, col = line_cols[group], lwd = lwd)
      }
    } else {
      x_pred <- .regplot_dodge_x(df_pred$x, df_pred$group_id, n_groups, dodge_w)
      graphics::points(
        x_pred,
        df_pred$y,
        pch = 18,
        col = line_cols[df_pred$group],
        cex = 1.5
      )
    }
  }

  if (mod_type == "categorical") {
    x_points <- .regplot_dodge_x(df_points$x, df_points$group_id, n_groups, dodge_w)
    x_points <- x_points + .regplot_jitter_values(
      n      = nrow(df_points),
      amount = jitter_am / max(1, n_groups)
    )
    graphics::points(
      x_points,
      df_points$y,
      pch = pch,
      col = point_cols[df_points$group],
      bg  = point_bgs[df_points$group],
      cex = df_points$size
    )
  } else {
    graphics::points(
      df_points$x,
      df_points$y,
      pch = pch,
      col = point_cols[df_points$group],
      bg  = point_bgs[df_points$group],
      cex = df_points$size
    )
  }

  if (has_by && legend) {
    graphics::legend(
      "topright",
      legend = groups,
      col    = line_cols,
      pt.bg  = point_bgs,
      pch    = 21,
      lty    = 1,
      lwd    = lwd,
      bty    = "n"
    )
  }

  return(invisible(NULL))
}


# ---------------------------------------------------------------------------- #
# .regplot_plot_ggplot
# ---------------------------------------------------------------------------- #
#
# Create regression plot using ggplot2.
#
# @param data   list of plot data from .regplot_data
# @param dots   list of graphical parameters
# @param legend logical; show legend
#
# @return ggplot object
#
# ---------------------------------------------------------------------------- #
.regplot_plot_ggplot <- function(data, dots, legend = TRUE) {

  pch       <- dots[["pch"]]
  col       <- dots[["col"]]
  bg        <- dots[["bg"]]
  lcol      <- dots[["lcol"]]
  lwd       <- dots[["lwd"]]
  col_ci    <- dots[["col.ci"]]
  col_pi    <- dots[["col.pi"]]
  col_si    <- dots[["col.si"]]
  alpha_ci  <- dots[["alpha.ci"]]
  alpha_pi  <- dots[["alpha.pi"]]
  alpha_si  <- dots[["alpha.si"]]
  shade     <- dots[["shade"]]
  main      <- dots[["main"]]
  jitter_am <- dots[["jitter"]]
  box_width <- dots[["box.width"]]
  dodge_w   <- dots[["dodge.width"]]

  df_points  <- data$points
  df_pred    <- data$pred
  df_ci      <- data$ci
  df_pi      <- data$pi
  df_si      <- data$si
  df_refline <- data$refline
  mod_type   <- data$mod_type
  groups     <- data$groups
  n_groups   <- length(groups)
  has_by     <- n_groups > 1L

  line_cols <- .regplot_palette(groups, lcol)
  if (has_by) {
    ci_cols <- pi_cols <- si_cols <- line_cols
  } else {
    ci_cols <- stats::setNames(rep(col_ci, n_groups), groups)
    pi_cols <- stats::setNames(rep(col_pi, n_groups), groups)
    si_cols <- stats::setNames(rep(col_si, n_groups), groups)
  }

  out <- ggplot2::ggplot()

  if (!is.null(df_refline)) {
    out <- out +
      ggplot2::geom_hline(
        yintercept = df_refline$y,
        linetype   = "dashed",
        colour     = "gray50"
      )
  }

  if (!is.null(df_si)) {
    if (mod_type == "continuous") {
      out <- .regplot_add_continuous_band_ggplot(out, df_si, si_cols, alpha_si, si_cols, shade, has_by)
    } else {
      out <- .regplot_add_categorical_band_ggplot(
        out, df_si, si_cols, alpha_si, si_cols, shade, has_by,
        box_width * 1.4, n_groups, dodge_w
      )
    }
  }

  if (!is.null(df_pi)) {
    if (mod_type == "continuous") {
      out <- .regplot_add_continuous_band_ggplot(out, df_pi, pi_cols, alpha_pi, pi_cols, shade, has_by)
    } else {
      out <- .regplot_add_categorical_band_ggplot(
        out, df_pi, pi_cols, alpha_pi, pi_cols, shade, has_by,
        box_width * 1.15, n_groups, dodge_w
      )
    }
  }

  if (!is.null(df_ci)) {
    if (mod_type == "continuous") {
      out <- .regplot_add_continuous_band_ggplot(out, df_ci, ci_cols, alpha_ci, ci_cols, shade, has_by)
    } else {
      out <- .regplot_add_categorical_band_ggplot(
        out, df_ci, ci_cols, alpha_ci, ci_cols, shade, has_by,
        box_width, n_groups, dodge_w
      )
    }
  }

  if (!is.null(df_pred)) {
    if (mod_type == "continuous") {
      if (has_by) {
        out <- out +
          ggplot2::geom_line(
            data    = df_pred,
            mapping = ggplot2::aes(
              x      = .data[["x"]],
              y      = .data[["y"]],
              colour = .data[["group"]],
              group  = .data[["group"]]
            ),
            linewidth = lwd / 2
          )
      } else {
        out <- out +
          ggplot2::geom_line(
            data    = df_pred,
            mapping = ggplot2::aes(x = .data[["x"]], y = .data[["y"]]),
            colour  = lcol[1],
            linewidth = lwd / 2
          )
      }
    } else {
      df_pred[["x_plot"]] <- .regplot_dodge_x(df_pred$x, df_pred$group_id, n_groups, dodge_w)
      if (has_by) {
        out <- out +
          ggplot2::geom_point(
            data    = df_pred,
            mapping = ggplot2::aes(
              x      = .data[["x_plot"]],
              y      = .data[["y"]],
              colour = .data[["group"]]
            ),
            shape   = 18,
            size    = 4
          )
      } else {
        out <- out +
          ggplot2::geom_point(
            data    = df_pred,
            mapping = ggplot2::aes(x = .data[["x_plot"]], y = .data[["y"]]),
            shape   = 18,
            colour  = lcol[1],
            size    = 4
          )
      }
    }
  }

  if (mod_type == "categorical") {
    df_points[["x_plot"]] <- .regplot_dodge_x(df_points$x, df_points$group_id, n_groups, dodge_w)
    df_points[["x_plot"]] <- df_points[["x_plot"]] + .regplot_jitter_values(
      n      = nrow(df_points),
      amount = jitter_am / max(1, n_groups)
    )
  } else {
    df_points[["x_plot"]] <- df_points[["x"]]
  }

  if (has_by) {
    out <- out +
      ggplot2::geom_point(
        data    = df_points,
        mapping = ggplot2::aes(
          x      = .data[["x_plot"]],
          y      = .data[["y"]],
          size   = .data[["size"]],
          colour = .data[["group"]],
          fill   = .data[["group"]]
        ),
        shape   = pch
      )
  } else {
    out <- out +
      ggplot2::geom_point(
        data    = df_points,
        mapping = ggplot2::aes(
          x    = .data[["x_plot"]],
          y    = .data[["y"]],
          size = .data[["size"]]
        ),
        shape   = pch,
        colour  = col,
        fill    = bg
      )
  }

  out <- out +
    ggplot2::scale_size_identity()

  if (mod_type == "categorical") {
    out <- out +
      ggplot2::scale_x_continuous(
        breaks = seq_along(data$levels),
        labels = data$levels,
        limits = data$xlim,
        name   = data$xlab
      )
  } else {
    out <- out +
      ggplot2::scale_x_continuous(
        limits = data$xlim,
        name   = data$xlab
      )
  }

  out <- out +
    ggplot2::scale_y_continuous(
      limits = data$ylim,
      name   = data$ylab
    )

  if (has_by) {
    out <- out +
      ggplot2::scale_colour_manual(values = line_cols, name = data$by_name) +
      ggplot2::scale_fill_manual(values = line_cols, name = data$by_name)
  }

  if (!is.null(main)) {
    out <- out + ggplot2::ggtitle(main)
  }

  out <- out + ggplot2::guides(size = "none")
  if (has_by && !legend) {
    out <- out + ggplot2::guides(colour = "none", fill = "none")
  }

  return(out)
}


# ---------------------------------------------------------------------------- #
# .set_dots_regplot
# ---------------------------------------------------------------------------- #
#
# Process dots arguments for regression plot with sensible defaults.
#
# @param ... additional graphical arguments
#
# @return list of graphical parameters with defaults applied
#
# ---------------------------------------------------------------------------- #
.set_dots_regplot <- function(...) {

  dots <- list(...)

  # point styling
  dots <- .plot_point_style_defaults(dots)

  # line styling
  if (is.null(dots[["lcol"]]))     dots[["lcol"]]     <- "black"
  if (is.null(dots[["lwd"]]))      dots[["lwd"]]      <- 2
  .check_plot_positive_scalar(dots[["lwd"]], "lwd")

  # band styling
  if (is.null(dots[["shade"]]))    dots[["shade"]]    <- TRUE
  if (is.null(dots[["col.ci"]]))   dots[["col.ci"]]   <- "gray70"
  if (is.null(dots[["col.pi"]]))   dots[["col.pi"]]   <- "gray85"
  if (is.null(dots[["col.si"]]))   dots[["col.si"]]   <- "gray92"
  if (is.null(dots[["alpha.ci"]])) dots[["alpha.ci"]] <- 0.4
  if (is.null(dots[["alpha.pi"]])) dots[["alpha.pi"]] <- 0.2
  if (is.null(dots[["alpha.si"]])) dots[["alpha.si"]] <- 0.15
  BayesTools::check_bool(dots[["shade"]], "shade")
  .check_plot_numeric(dots[["alpha.ci"]], "alpha.ci", check_length = 1, lower = 0, upper = 1, allow_null = FALSE)
  .check_plot_numeric(dots[["alpha.pi"]], "alpha.pi", check_length = 1, lower = 0, upper = 1, allow_null = FALSE)
  .check_plot_numeric(dots[["alpha.si"]], "alpha.si", check_length = 1, lower = 0, upper = 1, allow_null = FALSE)

  # categorical moderator jitter
  if (is.null(dots[["jitter"]]))      dots[["jitter"]]      <- 0.2
  if (is.null(dots[["box.width"]]))   dots[["box.width"]]   <- 0.5
  if (is.null(dots[["dodge.width"]])) dots[["dodge.width"]] <- 0.75
  .check_plot_numeric(dots[["jitter"]], "jitter", check_length = 1, lower = 0, allow_null = FALSE)
  .check_plot_positive_scalar(dots[["box.width"]], "box.width")
  .check_plot_numeric(dots[["dodge.width"]], "dodge.width", check_length = 1, lower = 0, allow_null = FALSE)

  # title (NULL = no title by default)
  if (is.null(dots[["main"]]))     dots[["main"]]     <- NULL
  if (is.null(dots[["las"]]))      dots[["las"]]      <- 1
  .check_plot_label(dots[["main"]], "main")
  BayesTools::check_int(dots[["las"]], "las", lower = 0)
  if (!dots[["las"]] %in% 0:3) {
    stop("'las' must be one of 0, 1, 2, or 3.", call. = FALSE)
  }
  .check_plot_positive_scalar(dots[["cex"]], "cex")
  .check_plot_positive_scalar(dots[["size"]], "size")

  return(dots)
}

context("Regression plot (bubble plot)")

# Load common test helpers
source(testthat::test_path("common-functions.R"))
source(testthat::test_path("helper-test-matrix.R"))
source(testthat::test_path("helper-visuals.R"))

# list cached fits lazily
skip_if_no_fits()
skip_if_not_installed("metafor")
fit_names <- list_fits()
fits      <- lazy_fits(fit_names, validate = FALSE)
info      <- lazy_infos(fit_names, validate = FALSE)

.test_regplot <- function(..., max_samples = 1000) {

  regplot(..., max_samples = max_samples)
}

# ============================================================================ #
# Test: Continuous Moderator Regression Plot
# ============================================================================ #

test_that("Regression plot for continuous moderator matches metafor", {

  name        <- "bcg_meta-regression"
  fit_metafor <- info[[name]][["metafor"]]
  fit_brma    <- fits[[name]]

  # --------------------------------------------------
  # Visual comparison: side-by-side plots
  # --------------------------------------------------

  expect_vdiffr_snapshot("regplot_continuous_comparison", function() {
    oldpar <- graphics::par(no.readonly = TRUE)
    on.exit(graphics::par(mfrow = oldpar[["mfrow"]], mar = oldpar[["mar"]]))
    par(mfrow = c(1, 2), mar = c(4, 4, 2, 1))
    metafor::regplot(fit_metafor, mod = "year", main = "metafor", ylim = c(-2, 0.5))
    .test_regplot(fit_brma, mod = "year", plot_type = "base", main = "brma", ylim = c(-2, 0.5))
  })

  expect_vdiffr_snapshot(
    "regplot_continuous_brma_ggplot",
    .test_regplot(fit_brma, mod = "year", plot_type = "ggplot")
  )
})

test_that("Regression plot renders prediction intervals", {

  skip_if_not_full_visuals("Prediction-interval regplot variants are gallery coverage.")

  name     <- "bcg_meta-regression"
  fit_brma <- fits[[name]]

  # --------------------------------------------------
  # Visual tests with PI bands
  # --------------------------------------------------

  expect_vdiffr_snapshot("regplot_continuous_pi_base", function() {
    .test_regplot(fit_brma, mod = "year", plot_type = "base", pi = TRUE, main = "with PI")
  })

  expect_vdiffr_snapshot(
    "regplot_continuous_pi_ggplot",
    .test_regplot(fit_brma, mod = "year", plot_type = "ggplot", pi = TRUE)
  )
})

test_that("Regression plot renders categorical dummy moderator", {

  name        <- "bcg_meta-regression2"
  fit_metafor <- info[[name]][["metafor"]]
  fit_brma    <- fits[[name]]

  expect_vdiffr_snapshot("regplot_categorical_brma", function() {
    oldpar <- graphics::par(no.readonly = TRUE)
    on.exit(graphics::par(mfrow = oldpar[["mfrow"]], mar = oldpar[["mar"]]))
    .test_regplot(fit_brma, mod = "alloc", plot_type = "base", main = "brma")
  })

  expect_vdiffr_snapshot(
    "regplot_categorical_brma_ggplot",
    .test_regplot(fit_brma, plot_type = "ggplot")
  )
})

test_that("Regression plot renders alternative categorical coding", {

  skip_if_not_full_visuals("Alternative coding snapshots duplicate categorical regplot coverage.")

  fit_brma1 <- fits[["bcg_meta-regression2"]]
  fit_brma2 <- fits[["bcg_meta-regression2b"]]

  expect_vdiffr_snapshot("regplot_categorical_coding_comparison", function() {
    oldpar <- graphics::par(no.readonly = TRUE)
    on.exit(graphics::par(mfrow = oldpar[["mfrow"]], mar = oldpar[["mar"]]))
    par(mfrow = c(1, 2), mar = c(4, 4, 2, 1))
    .test_regplot(fit_brma1, mod = "alloc", plot_type = "base", main = "brma", si = TRUE)
    .test_regplot(fit_brma2, mod = "alloc", plot_type = "base", main = "brma", si = TRUE)
  })
})

test_that("Regression plot renders interaction moderators", {

  skip_if_not_full_visuals("Interaction regplot variants are gallery coverage.")

  fit_brma3 <- fits[["bcg_meta-regression3"]]
  fit_brma4 <- fits[["bcg_meta-regression4"]]

  expect_vdiffr_snapshot("regplot_interaction_coding_comparison", function() {
    oldpar <- graphics::par(no.readonly = TRUE)
    on.exit(graphics::par(mfrow = oldpar[["mfrow"]], mar = oldpar[["mar"]]))
    par(mfrow = c(1, 2), mar = c(4, 4, 2, 1))
    .test_regplot(fit_brma3, mod = "year",            plot_type = "base", by = "alloc")
    .test_regplot(fit_brma4, mod = "year_before1969", plot_type = "base", by = "alloc")
  })

  expect_vdiffr_snapshot(
    "regplot_interaction_coding_ggplot",
    .test_regplot(fit_brma4, mod = "year_before1969", plot_type = "ggplot", by = "alloc")
  )
})

# ============================================================================ #
# Test: Location-Scale Model Regression Plot
# ============================================================================ #

test_that("Regression plot for location-scale model matches metafor view", {

  skip_if_not_full_visuals("Location-scale regplot variants are gallery coverage.")

  name        <- "bangertdrowns2004_location-scale"
  fit_metafor <- info[[name]][["metafor"]]
  fit_brma    <- fits[[name]]

  # --------------------------------------------------
  # Visual tests
  # --------------------------------------------------

  expect_vdiffr_snapshot("regplot_scale_brma_base", function() {
    oldpar <- graphics::par(no.readonly = TRUE)
    on.exit(graphics::par(mfrow = oldpar[["mfrow"]], mar = oldpar[["mar"]]))
    par(mfrow = c(1, 2), mar = c(4, 4, 2, 1))
    metafor::regplot(fit_metafor, mod = "ni100", main = "metafor", ylim = c(-1, 1.5))
    .test_regplot(fit_brma, mod = "ni100", plot_type = "base", main = "brma", ylim = c(-1, 1.5))
  })

  expect_vdiffr_snapshot(
    "regplot_scale_brma_ggplot",
    .test_regplot(fit_brma, plot_type = "ggplot", mod = "ni100")
  )
})

test_that("Regression plot thins location-scale prediction intervals", {

  name <- "bangertdrowns2004_location-scale"
  skip_if_missing_fits(name)

  fit_brma     <- fits[[name]]
  regplot_data <- NULL

  expect_no_error({
    regplot_data <- .test_regplot(
      fit_brma,
      mod         = "ni100",
      pi          = TRUE,
      as_data     = TRUE,
      at          = c(0, 3, 6),
      max_samples = 200
    )
  })

  expect_equal(nrow(regplot_data[["pi"]]), 2 * nrow(regplot_data[["pred"]]))
  expect_true(all(is.finite(regplot_data[["pi"]][["lower"]])))
  expect_true(all(is.finite(regplot_data[["pi"]][["upper"]])))
})

# ============================================================================ #
# Test: 3-Level Model Regression Plot
# ============================================================================ #

test_that("Regression plot for multilevel model renders fitted moderator", {

  name        <- "konstantopoulos2011_3lvl2"
  fit_metafor <- info[[name]][["metafor"]]
  fit_brma    <- fits[[name]]
  mod         <- "vi"

  expect_no_error({
    oldpar <- graphics::par(no.readonly = TRUE)
    on.exit(graphics::par(mfrow = oldpar[["mfrow"]], mar = oldpar[["mar"]]))
    par(mfrow = c(1, 2), mar = c(4, 4, 2, 1))
    metafor::regplot(fit_metafor, mod = mod, main = "metafor", ylim = c(-1, 1.5))
    .test_regplot(fit_brma, mod = mod, plot_type = "base", main = "brma", ylim = c(-1, 1.5))
  })

  expect_s3_class(.test_regplot(fit_brma, plot_type = "ggplot", mod = mod), "ggplot")
})

# ============================================================================ #
# Test: Selection Regression Plot
# ============================================================================ #

test_that("Regression plot for selection model renders fitted moderator", {

  name        <- "dat.lehmann2018-3PSMreg"
  fit_metafor <- info[[name]][["metafor"]]
  fit_brma    <- fits[[name]]
  mod         <- "Preregistered"

  expect_no_error({
    oldpar <- graphics::par(no.readonly = TRUE)
    on.exit(graphics::par(mfrow = oldpar[["mfrow"]], mar = oldpar[["mar"]]))
    par(mfrow = c(1, 2), mar = c(4, 4, 2, 1))
    metafor::regplot(fit_metafor, mod = mod, main = "metafor", ylim = c(-1, 1.5))
    .test_regplot(fit_brma, mod = mod, plot_type = "base", main = "brma", ylim = c(-1, 1.5))
  })

  expect_s3_class(.test_regplot(fit_brma, plot_type = "ggplot", mod = mod), "ggplot")
})

# ============================================================================ #
# Test: PET Model Regression Plot
# ============================================================================ #

test_that("Regression plot for PET model renders fitted moderator", {

  name        <- "dat.lehmann2018-PETreg"
  fit_metafor <- info[[name]][["metafor"]]
  fit_brma    <- fits[[name]]
  mod         <- "Preregistered"

  expect_no_error({
    oldpar <- graphics::par(no.readonly = TRUE)
    on.exit(graphics::par(mfrow = oldpar[["mfrow"]], mar = oldpar[["mar"]]))
    par(mfrow = c(1, 2), mar = c(4, 4, 2, 1))
    metafor::regplot(fit_metafor, mod = mod, main = "metafor", ylim = c(-1, 1.5))
    .test_regplot(fit_brma, mod = mod, plot_type = "base", main = "brma", ylim = c(-1, 1.5))
  })

  expect_s3_class(.test_regplot(fit_brma, plot_type = "ggplot", mod = mod), "ggplot")
})

# ============================================================================ #
# Test: BMA.norm Model Regression Plot
# ============================================================================ #

test_that("Regression plot for BMA.norm meta-regression renders base output", {

  name     <- "dat.lehmann2018_BMA.norm_mods"
  fit_brma <- fits[[name]]

  expect_vdiffr_snapshot("regplot_BMAreg", function() {
    suppressWarnings(.test_regplot(fit_brma, plot_type = "base"))
  })
})

# ============================================================================ #
# Test: BMA.glmm Model Regression Plot
# ============================================================================ #

test_that("Regression plot for BMA.glmm model renders base output", {

  skip_if_not_full_visuals("BMA.glmm regplot smoke is gallery coverage.")

  name     <- "bcg_BMA.glmm_3lvl_location_scale"
  fit_brma <- fits[[name]]

  expect_vdiffr_snapshot("regplot_BMA.glmm", function() {
    suppressWarnings(.test_regplot(fit_brma, mod = "year", plot_type = "base"))
  })
})

# ============================================================================ #
# Test: RoBMA Model Regression Plot
# ============================================================================ #

test_that("Regression plot for RoBMA meta-regression renders base output", {

  name     <- "dat.lehmann2018_RoBMA_3lvl_mods_scale"
  fit_brma <- fits[[name]]

  expect_vdiffr_snapshot("regplot_RoBMA_complex", function() {
    suppressWarnings(.test_regplot(fit_brma, mod = "Preregistered", plot_type = "base"))
  })
})

# ============================================================================ #
# Test: Regression Plot Options
# ============================================================================ #

test_that("Regression plot data and argument validation are stable", {

  name     <- "bcg_meta-regression"
  fit_brma <- fits[[name]]

  # --------------------------------------------------
  # Test as_data = TRUE returns list with expected components
  # --------------------------------------------------

  regplot_data <- .test_regplot(fit_brma, mod = "year", pi = TRUE, si = TRUE, as_data = TRUE)
  expect_equal(regplot_data[["sei"]], stats::median(RoBMA:::.outcome_data_sei(fit_brma)))

  regplot_data_custom_sei <- .test_regplot(
    fit_brma,
    mod     = "year",
    si      = TRUE,
    sei     = .123,
    as_data = TRUE
  )
  expect_equal(regplot_data_custom_sei[["sei"]], .123)

  expect_true(is.list(regplot_data),
    info = "as_data = TRUE returns a list"
  )

  expected_components <- c(
    "points", "pred", "ci", "xlim", "ylim", "xlab", "ylab", "mod_type"
  )
  expect_true(all(expected_components %in% names(regplot_data)),
    info = "regplot data contains all expected components"
  )

  # Check points data.frame structure
  expect_true(is.data.frame(regplot_data$points),
    info = "points are returned as a data.frame"
  )
  expect_true(all(c("x", "y", "size") %in% names(regplot_data$points)),
    info = "points contain x, y, and size columns"
  )

  # Check continuous interval-band structure
  band_columns <- c("x", "y", "lower", "upper", "xpred")
  n_pred       <- nrow(regplot_data$pred)

  for (band_name in c("ci", "pi", "si")) {
    band_data <- regplot_data[[band_name]]

    expect_true(all(band_columns %in% names(band_data)),
      info = paste0(band_name, " contains x, y, lower, upper, and xpred columns")
    )
    expect_equal(nrow(band_data), 2 * n_pred,
      info = paste0(band_name, " contains one row per polygon vertex")
    )
    expect_true(all(vapply(band_data[band_columns], length, integer(1)) == nrow(band_data)),
      info = paste0(band_name, " columns all have matching lengths")
    )
    expect_equal(band_data$x, band_data$xpred,
      info = paste0(band_name, " x coordinates match the stored prediction x values")
    )
  }

  # Check number of points matches number of studies
  n_studies <- nrow(fit_brma$data$outcome)
  expect_equal(nrow(regplot_data$points), n_studies,
    info = "number of points matches number of studies"
  )

  # --------------------------------------------------
  # Test error on intercept-only model
  # --------------------------------------------------

  name_simple <- "bcg_meta-analysis"
  fit_simple  <- fits[[name_simple]]

  expect_error(.test_regplot(fit_simple),
    regexp = "regplot requires a model with moderators",
    info = "intercept-only model is rejected"
  )

  # --------------------------------------------------
  # Test error on invalid plot_type
  # --------------------------------------------------

  expect_error(.test_regplot(fit_brma, plot_type = "invalid"),
    info = "invalid plot_type is rejected"
  )

  expect_error(.test_regplot(fit_brma, sei = 0),
    info = "non-positive reference sei is rejected"
  )

  expect_error(.test_regplot(fit_brma, mod = "year", at = "bad"),
    regexp = "at",
    info = "non-numeric prediction grid is rejected"
  )
  expect_error(.test_regplot(fit_brma, mod = "year", plim = c(3, 1)),
    regexp = "plim",
    info = "non-increasing point-size limits are rejected"
  )
  expect_error(.test_regplot(fit_brma, mod = "year", psize = 1:2),
    regexp = "psize",
    info = "wrong-length point sizes are rejected"
  )
  expect_error(.test_regplot(fit_brma, mod = "year", lwd = 0),
    regexp = "lwd",
    info = "non-positive line width is rejected"
  )

  fit_categorical <- fits[["bcg_meta-regression2"]]
  expect_error(.test_regplot(fit_categorical, mod = "alloc", at = 1),
    regexp = "continuous moderators",
    info = "at is rejected for categorical moderators"
  )
})

test_that("Regression plot transforms effect-size data", {

  skip_if_not_full_visuals("Transformation regplot snapshots are gallery coverage.")

  name     <- "bcg_meta-regression"
  fit_brma <- fits[[name]]

  expect_vdiffr_snapshot("regplot_transform_exp_comparison", function() {
    oldpar <- graphics::par(no.readonly = TRUE)
    on.exit(graphics::par(mfrow = oldpar[["mfrow"]], mar = oldpar[["mar"]]))
    par(mfrow = c(1, 2), mar = c(4, 4, 2, 1))
    .test_regplot(fit_brma, mod = "year", plot_type = "base", main = "log RR")
    .test_regplot(fit_brma, mod = "year", transf = exp, plot_type = "base", main = "RR")
  })
})

# ============================================================================ #
# Test: Regression Plot Customization
# ============================================================================ #

test_that("Regression plot customization snapshots are stable", {

  skip_if_not_full_visuals("Customization snapshots are visual-gallery coverage.")

  name     <- "bcg_meta-regression"
  fit_brma <- fits[[name]]

  # --------------------------------------------------
  # Test custom point aesthetics
  # --------------------------------------------------

  expect_vdiffr_snapshot("regplot_custom_points_base", function() {
    .test_regplot(fit_brma, mod = "year", plot_type = "base", pch = 21, col = "blue", bg = "lightblue")
  })

  expect_vdiffr_snapshot(
    "regplot_custom_points_ggplot",
    .test_regplot(fit_brma, mod = "year", plot_type = "ggplot", pch = 21, col = "blue", bg = "lightblue")
  )

  # --------------------------------------------------
  # Test custom line and band styling
  # --------------------------------------------------

  expect_vdiffr_snapshot("regplot_custom_bands_base", function() {
    .test_regplot(fit_brma, mod = "year", plot_type = "base", lcol = "darkblue", col.ci = "lightblue",
            pi = TRUE, col.pi = "lightyellow")
  })

  expect_vdiffr_snapshot(
    "regplot_custom_bands_ggplot",
    .test_regplot(fit_brma, mod = "year", plot_type = "ggplot", lcol = "darkblue", col.ci = "lightblue",
            pi = TRUE, col.pi = "lightyellow")
  )

  # --------------------------------------------------
  # Test reference line
  # --------------------------------------------------

  expect_vdiffr_snapshot("regplot_refline_base", function() {
    .test_regplot(fit_brma, mod = "year", plot_type = "base", refline = 0)
  })

  expect_vdiffr_snapshot(
    "regplot_refline_ggplot",
    .test_regplot(fit_brma, mod = "year", plot_type = "ggplot", refline = 0)
  )

  # --------------------------------------------------
  # Test custom axis labels and title
  # --------------------------------------------------

  expect_vdiffr_snapshot("regplot_custom_labels_base", function() {
    .test_regplot(fit_brma, mod = "ablat", plot_type = "base",
            xlab = "Absolute Latitude", ylab = "Log Risk Ratio",
            main = "BCG Meta-Regression")
  })

  expect_vdiffr_snapshot(
    "regplot_custom_labels_ggplot",
    .test_regplot(fit_brma, mod = "ablat", plot_type = "ggplot",
            xlab = "Absolute Latitude", ylab = "Log Risk Ratio",
            main = "BCG Meta-Regression")
  )

  # --------------------------------------------------
  # Test no shading
  # --------------------------------------------------

  expect_vdiffr_snapshot("regplot_no_shade_base", function() {
    .test_regplot(fit_brma, mod = "ablat", plot_type = "base", shade = FALSE)
  })

  expect_vdiffr_snapshot(
    "regplot_no_shade_ggplot",
    .test_regplot(fit_brma, mod = "ablat", plot_type = "ggplot", shade = FALSE)
  )

  # --------------------------------------------------
  # Test custom axis limits
  # --------------------------------------------------

  expect_vdiffr_snapshot("regplot_custom_limits_base", function() {
    .test_regplot(fit_brma, mod = "ablat", plot_type = "base", xlim = c(0, 60), ylim = c(-3, 1))
  })

  expect_vdiffr_snapshot(
    "regplot_custom_limits_ggplot",
    .test_regplot(fit_brma, mod = "ablat", plot_type = "ggplot", xlim = c(0, 60), ylim = c(-3, 1))
  )

  # --------------------------------------------------
  # Test CI only (no PI)
  # --------------------------------------------------

  expect_vdiffr_snapshot("regplot_ci_only_base", function() {
    .test_regplot(fit_brma, mod = "ablat", plot_type = "base", ci = TRUE, pi = FALSE)
  })

  # --------------------------------------------------
  # Test neither CI nor PI
  # --------------------------------------------------

  expect_vdiffr_snapshot("regplot_no_bands_base", function() {
    .test_regplot(fit_brma, mod = "ablat", plot_type = "base", ci = FALSE, pi = FALSE)
  })
})

# ============================================================================ #
# Test: Regression Plot Confidence Level
# ============================================================================ #

test_that("Regression plot confidence-level snapshots are stable", {

  skip_if_not_full_visuals("Confidence-level snapshots are visual-gallery coverage.")

  name     <- "bcg_meta-regression"
  fit_brma <- fits[[name]]

  # --------------------------------------------------
  # Test different confidence levels
  # --------------------------------------------------

  expect_vdiffr_snapshot("regplot_level_90_base", function() {
    .test_regplot(fit_brma, mod = "ablat", plot_type = "base", level = 90, main = "90% CI")
  })

  expect_vdiffr_snapshot("regplot_level_99_base", function() {
    .test_regplot(fit_brma, mod = "ablat", plot_type = "base", level = 99, main = "99% CI")
  })
})

test_that("native regplot mixture intervals match R fallback", {

  skip_if_not(RoBMA:::.has_native_regplot_mixture())

  set.seed(1024)

  S     <- 200
  K     <- 4
  probs <- c(.025, .975)

  mean_samples <- matrix(stats::rnorm(S * K), nrow = S, ncol = K)
  sd_samples   <- matrix(stats::runif(S * K, .05, 1.25), nrow = S, ncol = K)

  expected <- RoBMA:::.regplot_mixture_interval_quantiles_r(
    mean_samples = mean_samples,
    sd_samples   = sd_samples,
    probs        = probs
  )
  actual <- RoBMA:::.regplot_mixture_interval_quantiles(
    mean_samples = mean_samples,
    sd_samples   = sd_samples,
    probs        = probs
  )

  expect_equal(actual, expected, tolerance = 1e-4)
})

#' @title 23 experimental studies from
#' \insertCite{anderson2010violent;textual}{RoBMA} that meet the best practice criteria
#'
#' @description The data set contains correlation coefficients, sample
#' sizes, and labels for 23 experimental studies focusing on the effect of
#' violent video games on aggressive behavior.
#'
#'
#' @format A data.frame with 3 columns and 23 observations:
#' \describe{
#'   \item{\code{r}}{Correlation coefficient.}
#'   \item{\code{n}}{Sample size.}
#'   \item{\code{name}}{Study label.}
#' }
#' @source \url{https://github.com/Joe-Hilgard/Anderson-meta}
#'
#' @references
#' \insertAllCited{}
"Anderson2010"


#' @title 9 experimental studies from
#' \insertCite{bem2011feeling;textual}{RoBMA} as described in
#' \insertCite{bem2011must;textual}{RoBMA}
#'
#' @description The data set contains Cohen's d effect sizes, standard errors,
#' and labels for 9 experimental studies of precognition from the infamous
#' \insertCite{bem2011feeling;textual}{RoBMA} as analyzed in his later meta-analysis
#' \insertCite{bem2011must}{RoBMA}.
#'
#' @format A data.frame with 3 columns and 9 observations:
#' \describe{
#'   \item{\code{d}}{Cohen's d effect size.}
#'   \item{\code{se}}{Standard error of d.}
#'   \item{\code{study}}{Study label.}
#' }
#' @references
#' \insertAllCited{}
"Bem2011"


#' @title 5 studies with a tactile outcome assessment from
#' \insertCite{poulsen2006potassium;textual}{RoBMA} of the effect of potassium-containing toothpaste
#' on dentine hypersensitivity
#'
#' @description The data set contains Cohen's d effect sizes, standard errors,
#' and labels for 5 studies assessing the tactile outcome from a meta-analysis of
#' the effect of potassium-containing toothpaste on dentine hypersensitivity
#' \insertCite{poulsen2006potassium}{RoBMA} which was used as an example in
#' \insertCite{bartos2021bayesian;textual}{RoBMA}.
#'
#' @format A data.frame with 3 columns and 5 observations:
#' \describe{
#'   \item{\code{d}}{Cohen's d effect size.}
#'   \item{\code{se}}{Standard error of d.}
#'   \item{\code{study}}{Study label.}
#' }
#' @references
#' \insertAllCited{}
"Poulsen2006"

#' @title 881 estimates from 69 studies of a relationship between employment and
#' educational outcomes collected by \insertCite{kroupova2021student;textual}{RoBMA}
#'
#' @description The data set contains partial correlation coefficients, standard errors,
#' study labels, sample sizes, type of the educational outcome, intensity of the
#' employment, gender of the student population, study location, study design, whether
#' the study controlled for endogeneity, and whether the study controlled for motivation.
#' The original data set including additional variables and the publication are
#' available from the project page.
#' (Note that some standard errors and employment intensities are missing.)
#'
#' @format A data.frame with 11 columns and 881 observations:
#' \describe{
#'   \item{\code{r}}{Partial correlation coefficient.}
#'   \item{\code{se}}{Standard error of r.}
#'   \item{\code{study}}{Study label.}
#'   \item{\code{sample_size}}{Sample size.}
#'   \item{\code{education_outcome}}{Type of educational outcome.}
#'   \item{\code{employment_intensity}}{Employment intensity.}
#'   \item{\code{students_gender}}{Gender composition of the student sample.}
#'   \item{\code{location}}{Study location.}
#'   \item{\code{design}}{Study design.}
#'   \item{\code{endogenity_control}}{Whether endogeneity was controlled.}
#'   \item{\code{motivation_control}}{Whether motivation was controlled.}
#' }
#' @source \url{http://meta-analysis.cz/students}
#'
#' @references
#' \insertAllCited{}
"Kroupova2021"

#' @title 18 studies of a relationship between acculturation mismatch and
#' intergenerational cultural conflict collected by
#' \insertCite{lui2015intergenerational;textual}{RoBMA}
#'
#' @description The data set contains correlation coefficients r,
#' sample sizes n, and labels for each study assessing the
#' relationship between acculturation mismatch (that is the result of the contrast
#' between the collectivist cultures of Asian and Latin immigrant groups
#' and the individualist culture in the United States) and intergenerational cultural
#' conflict \insertCite{lui2015intergenerational}{RoBMA} which was used as an
#' example in \insertCite{bartos2020adjusting;textual}{RoBMA}.
#'
#' @format A data.frame with 3 columns and 18 observations:
#' \describe{
#'   \item{\code{r}}{Correlation coefficient.}
#'   \item{\code{n}}{Sample size.}
#'   \item{\code{study}}{Study label.}
#' }
#' @references
#' \insertAllCited{}
"Lui2015"

#' @title 39 study rows on household chaos and child executive functions
#' from a meta-analysis by
#' \insertCite{andrews2021examining;textual}{RoBMA}
#'
#' @description The data set contains raw correlation coefficients and
#' sampling information using \pkg{metafor}-compatible column names
#' (\code{ri}, \code{vi}, \code{sei}, \code{ni}, and \code{slab}), together with
#' study-level moderators from the original data set. Three source rows have
#' missing effect-size inputs and are omitted automatically by model-fitting
#' functions that require complete outcomes. The original data set assessed the
#' effect of household chaos on child executive functions
#' \insertCite{andrews2021examining}{RoBMA} and was used as an example in
#' \insertCite{bartos2023robust;textual}{RoBMA}.
#'
#' @format A data.frame with 15 columns and 39 observations:
#' \describe{
#'   \item{\code{study_id}}{Source row identifier.}
#'   \item{\code{slab}}{Study label.}
#'   \item{\code{ri}}{Raw correlation coefficient.}
#'   \item{\code{vi}}{Sampling variance of the raw correlation coefficient.}
#'   \item{\code{sei}}{Sampling standard error of the raw correlation coefficient.}
#'   \item{\code{ni}}{Sample size.}
#'   \item{\code{year}}{Publication year.}
#'   \item{\code{percent_female}}{Percentage of female children in the sample.}
#'   \item{\code{age}}{Mean age of the children in years.}
#'   \item{\code{assessment_interval_months}}{Time between household-chaos and executive-function assessments in months.}
#'   \item{\code{dissertation}}{Whether the study was a dissertation.}
#'   \item{\code{percent_minority}}{Percentage of minority participants in the sample.}
#'   \item{\code{percent_low_parental_education}}{Percentage of parents with high school, GED, or lower education.}
#'   \item{\code{hc_dimension}}{Household-chaos dimension using the source coding.}
#'   \item{\code{measure}}{Executive-function assessment type, direct or informant.}
#' }
#' @references
#' \insertAllCited{}
"Andrews2021"

#' @title 70 effect sizes from a meta-analysis of ChatGPT's impact on student learning
#' by \insertCite{wang2025effect;textual}{RoBMA}
#'
#' @description The data set contains Hedges' g effect sizes, standard errors,
#' sample sizes for experimental and control groups, and various study characteristics
#' including grade level, type of course, duration, learning model, role of ChatGPT,
#' and area of ChatGPT application. The meta-analysis examined the effect of ChatGPT
#' on students' learning performance, learning perception, and higher-order thinking
#' \insertCite{wang2025effect}{RoBMA}.
#'
#' @format A data.frame with 12 columns and 70 observations:
#' \describe{
#'   \item{\code{Learning_effect}}{Learning-effect outcome category.}
#'   \item{\code{Author_year}}{Author-year study label.}
#'   \item{\code{N_EG}}{Experimental-group sample size.}
#'   \item{\code{N_CG}}{Control-group sample size.}
#'   \item{\code{g}}{Hedges' g effect size.}
#'   \item{\code{Grade_level}}{Grade level.}
#'   \item{\code{Type_of_course}}{Course type.}
#'   \item{\code{Duration}}{Study duration.}
#'   \item{\code{Learning_model}}{Learning model.}
#'   \item{\code{Role_of_ChatGPT}}{Role of ChatGPT in the intervention.}
#'   \item{\code{Area_of_ChatGPT_application}}{Application area.}
#'   \item{\code{se}}{Standard error of g.}
#' }
#' @references
#' \insertAllCited{}
"Wang2025"

#' @title 55 effect sizes from Many Labs 2 replication studies of
#' \insertCite{tversky1981framing;textual}{RoBMA} framing effects
#'
#' @description The data set contains standardized mean differences (y) and standard errors (se)
#' from 55 replication studies of the framing effect originally described by
#' \insertCite{tversky1981framing;textual}{RoBMA}. These studies were part of the
#' Many Labs 2 project examining variation in replicability across samples and settings
#' \insertCite{klein2018many;textual}{RoBMA}.
#'
#' @format A data.frame with 2 columns and 55 observations:
#' \describe{
#'   \item{\code{y}}{Standardized mean difference.}
#'   \item{\code{se}}{Standard error of y.}
#' }
#' @references
#' \insertAllCited{}
"ManyLabs16"

#' @title 37 studies from a meta-analysis of social comparison as a behavior change technique
#' by \insertCite{hoppen2025meta;textual}{RoBMA}
#'
#' @description The data set contains Cohen's d effect sizes, variances, and study
#' characteristics including outcome type, feedback level, social comparison type,
#' number of sessions, sample type, sample size, and country. The meta-analysis
#' examined social comparison as a behavior change technique across the behavioral
#' sciences \insertCite{hoppen2025meta}{RoBMA}.
#'
#' @format A data.frame with 9 columns and 37 observations:
#' \describe{
#'   \item{\code{d}}{Cohen's d effect size.}
#'   \item{\code{v}}{Sampling variance of d.}
#'   \item{\code{outcome}}{Outcome type.}
#'   \item{\code{feedback_level}}{Feedback level.}
#'   \item{\code{social_comparison_type}}{Social-comparison type.}
#'   \item{\code{sessions}}{Number of sessions.}
#'   \item{\code{sample_type}}{Sample type.}
#'   \item{\code{sample_size}}{Sample size.}
#'   \item{\code{country}}{Country.}
#' }
#' @references
#' \insertAllCited{}
"Hoppen2025"

#' @title 582 effect sizes examining the ease-of-retrieval effect
#' from a meta-analysis by \insertCite{weingarten2018does;textual}{RoBMA}
#'
#' @description The data set contains correlation coefficients between the manipulation
#' and outcome variable (r_xy), sample sizes (N), and various study characteristics
#' including publication status, country (USA vs. other), number of few and many
#' examples requested, whether the trial targeted episodic memory, paradigm type (standard vs. other),
#' dataset type (proximal vs. distal), and mediation variables (r_xm, r_my).
#' The meta-analysis examined whether subjective ease mediates the ease-of-retrieval effect,
#' where participants list either few or many examples and then make judgments
#' \insertCite{weingarten2018does}{RoBMA}.
#'
#' @format A data.frame with 12 columns and 582 observations:
#' \describe{
#'   \item{\code{r_xy}}{Correlation between manipulation and outcome.}
#'   \item{\code{N}}{Sample size.}
#'   \item{\code{paper_id}}{Paper identifier.}
#'   \item{\code{published}}{Publication-status indicator.}
#'   \item{\code{USA}}{Whether the study was conducted in the USA.}
#'   \item{\code{number_of_few}}{Number of examples in the few-examples condition.}
#'   \item{\code{number_of_many}}{Number of examples in the many-examples condition.}
#'   \item{\code{episodic_memory}}{Whether the trial targeted episodic memory.}
#'   \item{\code{standard_paradigm}}{Whether the standard paradigm was used.}
#'   \item{\code{proximal_dataset}}{Whether the data set was proximal.}
#'   \item{\code{r_xm}}{Correlation between manipulation and mediator.}
#'   \item{\code{r_my}}{Correlation between mediator and outcome.}
#' }
#' @references
#' \insertAllCited{}
"Weingarten2018"


#' @title 412 effect sizes from a meta-analysis of secondary benefits of family-based treatments
#' by \insertCite{johnides2025secondary;textual}{RoBMA}
#'
#' @description The data set contains Cohen's d effect sizes, standard errors, and study labels
#' from a meta-analysis investigating the extent to which family-based treatments for children
#' with mental health, physical health, and developmental disorders provide secondary benefits
#' to the children's siblings and caregivers \insertCite{johnides2025secondary}{RoBMA}.
#'
#' @format A data.frame with 3 columns and 412 observations:
#' \describe{
#'   \item{\code{study}}{Study label.}
#'   \item{\code{d}}{Cohen's d effect size.}
#'   \item{\code{se}}{Standard error of d.}
#' }
#' @references
#' \insertAllCited{}
"Johnides2025"


#' @title 1159 effect sizes from a meta-analysis of beauty and professional success
#' by \insertCite{havrankova2025beauty;textual}{RoBMA}
#'
#' @description The data set contains effect sizes (percent increase in earnings), standard errors,
#' study identifiers, sample sizes, and the type of customer contact (no, some, or direct).
#' The meta-analysis examined the relationship between perceived beauty and professional success
#' \insertCite{havrankova2025beauty}{RoBMA}.
#'
#' @format A data.frame with 5 columns and 1159 observations:
#' \describe{
#'   \item{\code{y}}{Effect size: percent increase in earnings.}
#'   \item{\code{se}}{Standard error of y.}
#'   \item{\code{facing_customer}}{Type of customer contact.}
#'   \item{\code{study_id}}{Study identifier.}
#'   \item{\code{N}}{Sample size.}
#' }
#' @references
#' \insertAllCited{}
"Havrankova2025"

# ---------------------------------------------------------------------------- #
# as_draws methods for posterior package integration
# ---------------------------------------------------------------------------- #

#' @title Convert brma_samples to posterior Draws Formats
#'
#' @description Provides an interface to the \pkg{posterior} package
#' for \code{brma_samples} objects. These functions convert the posterior
#' samples to various draws formats supported by the \pkg{posterior} package.
#'
#' @param x a \code{brma_samples} object
#' @param ... additional arguments passed to the corresponding
#' \pkg{posterior} function
#'
#' @details
#' The conversion reconstructs the MCMC chain structure from the stored
#' \code{nchains} and \code{niter} attributes. The samples are assumed
#' to be ordered with chains concatenated (i.e., all iterations from chain 1,
#' then all from chain 2, etc.). Conditional RoBMA samples are intentionally
#' stored as one flattened chain because conditioning subsets posterior rows
#' across chains.
#'
#' @return An object of the corresponding \pkg{posterior} draws class.
#'
#' @seealso \code{\link[posterior]{draws}}, \code{\link{as_draws.brma}}
#'
#' @name as_draws.brma_samples
NULL

#' @rdname as_draws.brma_samples
#' @export
as_draws.brma_samples <- function(x, ...) {

  .check_posterior_package()

  # reconstruct mcmc.list from samples, convert to draws_array first
  # (to properly set chain dimensions), then to draws_matrix
  # this matches posterior::as_draws.matrix behavior but preserves chain info
  mcmc.list <- .brma_samples_to_mcmc.list(x)

  return(posterior::as_draws_matrix(mcmc.list, ...))
}

#' @rdname as_draws.brma_samples
#' @export
as_draws_array.brma_samples <- function(x, ...) {

  .check_posterior_package()

  # reconstruct mcmc.list from samples
  mcmc.list <- .brma_samples_to_mcmc.list(x)

  return(posterior::as_draws_array(mcmc.list, ...))
}

#' @rdname as_draws.brma_samples
#' @export
as_draws_df.brma_samples <- function(x, ...) {

  .check_posterior_package()

  # reconstruct mcmc.list from samples
  mcmc.list <- .brma_samples_to_mcmc.list(x)

  return(posterior::as_draws_df(mcmc.list, ...))
}

#' @rdname as_draws.brma_samples
#' @export
as_draws_list.brma_samples <- function(x, ...) {

  .check_posterior_package()

  # reconstruct mcmc.list from samples
  mcmc.list <- .brma_samples_to_mcmc.list(x)

  return(posterior::as_draws_list(mcmc.list, ...))
}

#' @rdname as_draws.brma_samples
#' @export
as_draws_matrix.brma_samples <- function(x, ...) {

  .check_posterior_package()

  # reconstruct mcmc.list from samples
  mcmc.list <- .brma_samples_to_mcmc.list(x)

  return(posterior::as_draws_matrix(mcmc.list, ...))
}

#' @rdname as_draws.brma_samples
#' @export
as_draws_rvars.brma_samples <- function(x, ...) {

  .check_posterior_package()

  # reconstruct mcmc.list from samples
  mcmc.list <- .brma_samples_to_mcmc.list(x)

  return(posterior::as_draws_rvars(mcmc.list, ...))
}


# ---------------------------------------------------------------------------- #
# Helper: Convert brma_samples to mcmc.list
# ---------------------------------------------------------------------------- #

#' @title Convert brma_samples to mcmc.list
#'
#' @description Internal helper to reconstruct an \code{mcmc.list} object
#' from a \code{brma_samples} matrix by splitting samples back into chains.
#'
#' @param x a \code{brma_samples} object
#'
#' @return An \code{mcmc.list} object suitable for conversion to posterior
#' draws formats.
#'
#' @noRd
.brma_samples_to_mcmc.list <- function(x) {

  n_chains <- attr(x, "nchains")
  n_iter   <- attr(x, "niter")

  # convert to plain matrix
  samples_mat <- as.matrix(x)

  if (length(n_chains) != 1L || length(n_iter) != 1L ||
      is.na(n_chains) || is.na(n_iter) ||
      n_chains < 1L || n_iter < 1L ||
      n_chains * n_iter != nrow(samples_mat)) {
    stop("Invalid brma_samples chain metadata: 'nchains * niter' must equal the number of sample rows.",
         call. = FALSE)
  }

  # split into chains
  # samples are ordered: chain1_iter1, ..., chain1_iterN, chain2_iter1, ..., etc.
  chains <- vector("list", n_chains)
  for (i in seq_len(n_chains)) {
    start_idx <- (i - 1) * n_iter + 1
    end_idx   <- i * n_iter
    chains[[i]] <- coda::mcmc(samples_mat[start_idx:end_idx, , drop = FALSE])
  }

  return(coda::mcmc.list(chains))
}

# ============================================================================ #
# Outcome RNG Functions
# ============================================================================ #
#
# These are prediction-specific RNG functions for sampling from the posterior
# predictive distribution of observed effect sizes. They are kept separate from
# the .evaluate.brma.* functions because they involve random number generation
# and are specific to the prediction task.
#
# ============================================================================ #


# ---------------------------------------------------------------------------- #
# .outcome_rng.norm
# ---------------------------------------------------------------------------- #
#
# Sample observed effect sizes from normal posterior predictive distribution.
#
# For normal outcome models, the observed effect y_i is:
#   y_i ~ N(mu_i, tau_within^2 + se_i^2)
#
# This function samples from this distribution for each posterior sample.
#
# @param mu_samples       S x K matrix of location samples (with cluster effects if multilevel)
# @param tau_within       S x K matrix of within-cluster heterogeneity samples
# @param sei              numeric vector of length K; standard errors
#
# @return S x K matrix of sampled observed effect sizes
#
# ---------------------------------------------------------------------------- #
.outcome_rng.norm <- function(mu_samples, tau_within, sei) {

  S <- nrow(mu_samples)
  K <- ncol(mu_samples)

  # replicate sei across samples: matrix(sei, S, K, byrow = TRUE)
  # each row contains the same se values
  sei_mat <- matrix(sei, nrow = S, ncol = K, byrow = TRUE)

  # compute total SD: sqrt(tau^2 + se^2)
  total_sd <- sqrt(tau_within^2 + sei_mat^2)

  # sample from N(mu, total_sd) for each cell
  # matrix(rnorm(S*K), S, K) * total_sd + mu is vectorized sampling
  # equivalent: for each s,k: y[s,k] ~ N(mu[s,k], total_sd[s,k])
  response_samples <- mu_samples + matrix(stats::rnorm(S * K), nrow = S, ncol = K) * total_sd

  return(response_samples)
}


# ---------------------------------------------------------------------------- #
# .outcome_rng.selnorm
# ---------------------------------------------------------------------------- #
#
# Sample observed effect sizes from the selected-normal kernel.
#
# ---------------------------------------------------------------------------- #
.outcome_rng.selnorm <- function(mu_samples, tau_within, sei,
                                 selection_context) {

  S          <- nrow(mu_samples)
  K          <- ncol(mu_samples)
  sei_mat    <- matrix(sei, nrow = S, ncol = K, byrow = TRUE)
  total_sd   <- sqrt(tau_within^2 + sei_mat^2)
  use_normal <- selection_context[["use_normal"]]

  if (length(use_normal) == 1L) {
    use_normal <- rep(use_normal, S)
  }

  if (all(use_normal)) {
    return(.outcome_rng.norm(
      mu_samples = mu_samples,
      tau_within = tau_within,
      sei        = sei
    ))
  }

  if (any(use_normal)) {
    out <- matrix(NA_real_, nrow = S, ncol = K)

    normal_rows <- which(use_normal)
    step_rows   <- which(!use_normal)

    out[normal_rows, ] <- .outcome_rng.norm(
      mu_samples = mu_samples[normal_rows, , drop = FALSE],
      tau_within = tau_within[normal_rows, , drop = FALSE],
      sei        = sei
    )

    out[step_rows, ] <- .selection_step_rng_matrix(
      mean              = mu_samples[step_rows, , drop = FALSE],
      sd                = total_sd[step_rows, , drop = FALSE],
      sei               = sei,
      selection_context = .selection_context_subset_rows(selection_context, step_rows)
    )

    return(out)
  }

  return(.selection_step_rng_matrix(
    mean              = mu_samples,
    sd                = total_sd,
    sei               = sei,
    selection_context = selection_context
  ))
}


# ---------------------------------------------------------------------------- #
# .outcome_rng.binom
# ---------------------------------------------------------------------------- #
#
# Sample observed counts from binomial posterior predictive distribution.
#
# For binomial outcome models (log-odds ratio), the observed counts are:
#   ai ~ Binom(n1i, p1) where logit(p1) = logit(pi) + 0.5 * mu (treatment)
#   ci ~ Binom(n2i, p2) where logit(p2) = logit(pi) - 0.5 * mu (control)
#
# The 0.5 factor arises from the log-OR parameterization where mu represents
# the log-odds ratio between treatment and control groups.
#
# @param mu_samples       S x K matrix of log-odds ratio samples
# @param logit_baserate   S x K matrix of logit(pi) base-rate samples
# @param n1i              numeric vector of length K; treatment group sizes
# @param n2i              numeric vector of length K; control group sizes
#
# @return S x (2*K) matrix with columns interleaved: ai[1], ci[1], ai[2], ci[2], ...
#
# ---------------------------------------------------------------------------- #
.outcome_rng.binom <- function(mu_samples, logit_baserate, n1i, n2i) {

  S <- nrow(mu_samples)
  K <- ncol(mu_samples)

  # compute logit probabilities for each group
  # group 1: logit(p1) = logit(pi) + 0.5 * mu (treatment/exposed)
  # group 2: logit(p2) = logit(pi) - 0.5 * mu (control/unexposed)
  logit_p1 <- logit_baserate + 0.5 * mu_samples
  logit_p2 <- logit_baserate - 0.5 * mu_samples

  # sample from binomial for each group
  outcome_samples_ai <- matrix(NA_integer_, nrow = S, ncol = K)
  outcome_samples_ci <- matrix(NA_integer_, nrow = S, ncol = K)

  for (k in seq_len(K)) {
    outcome_samples_ai[, k] <- stats::rbinom(
      n    = S,
      size = n1i[k],
      prob = .inv_logit(logit_p1[, k])
    )
    outcome_samples_ci[, k] <- stats::rbinom(
      n    = S,
      size = n2i[k],
      prob = .inv_logit(logit_p2[, k])
    )
  }

  # name columns
  colnames(outcome_samples_ai) <- paste0("ai[", seq_len(K), "]")
  colnames(outcome_samples_ci) <- paste0("ci[", seq_len(K), "]")

  # merge samples with interleaved columns: ai[1], ci[1], ai[2], ci[2], ...
  outcome_samples <- matrix(NA_integer_, nrow = S, ncol = 2 * K)
  outcome_samples[, seq(1, 2 * K, by = 2)] <- outcome_samples_ai
  outcome_samples[, seq(2, 2 * K, by = 2)] <- outcome_samples_ci
  colnames(outcome_samples) <- as.vector(rbind(
    colnames(outcome_samples_ai),
    colnames(outcome_samples_ci)
  ))

  return(outcome_samples)
}


# ---------------------------------------------------------------------------- #
# .outcome_rng.pois
# ---------------------------------------------------------------------------- #
#
# Sample observed counts from Poisson posterior predictive distribution.
#
# For Poisson outcome models (log incidence rate ratio), the observed counts are:
#   x1i ~ Pois(t1i * exp(phi + 0.5 * mu))  (treatment/exposed)
#   x2i ~ Pois(t2i * exp(phi - 0.5 * mu))  (control/unexposed)
#
# where phi is the midpoint log rate and t1i, t2i are exposure times.
#
# @param mu_samples       S x K matrix of log incidence rate ratio samples
# @param log_phi          S x K matrix of midpoint log-rate samples
# @param t1i              numeric vector of length K; treatment exposure times
# @param t2i              numeric vector of length K; control exposure times
#
# @return S x (2*K) matrix with columns interleaved: x1i[1], x2i[1], x1i[2], x2i[2], ...
#
# ---------------------------------------------------------------------------- #
.outcome_rng.pois <- function(mu_samples, log_phi, t1i, t2i) {

  S <- nrow(mu_samples)
  K <- ncol(mu_samples)

  # replicate exposure times to S x K matrices for vectorized ops
  log_t1i <- matrix(log(t1i), nrow = S, ncol = K, byrow = TRUE)
  log_t2i <- matrix(log(t2i), nrow = S, ncol = K, byrow = TRUE)

  # compute log rates for each group
  # group 1: log(lambda1) = phi + 0.5 * mu + log(t1i)
  # group 2: log(lambda2) = phi - 0.5 * mu + log(t2i)
  log_lambda1 <- log_phi + 0.5 * mu_samples + log_t1i
  log_lambda2 <- log_phi - 0.5 * mu_samples + log_t2i

  # sample from Poisson for each group
  outcome_samples_x1i <- matrix(NA_integer_, nrow = S, ncol = K)
  outcome_samples_x2i <- matrix(NA_integer_, nrow = S, ncol = K)

  for (k in seq_len(K)) {
    outcome_samples_x1i[, k] <- stats::rpois(
      n      = S,
      lambda = exp(log_lambda1[, k])
    )
    outcome_samples_x2i[, k] <- stats::rpois(
      n      = S,
      lambda = exp(log_lambda2[, k])
    )
  }

  # name columns
  colnames(outcome_samples_x1i) <- paste0("x1i[", seq_len(K), "]")
  colnames(outcome_samples_x2i) <- paste0("x2i[", seq_len(K), "]")

  # merge samples with interleaved columns: x1i[1], x2i[1], x1i[2], x2i[2], ...
  outcome_samples <- matrix(NA_integer_, nrow = S, ncol = 2 * K)
  outcome_samples[, seq(1, 2 * K, by = 2)] <- outcome_samples_x1i
  outcome_samples[, seq(2, 2 * K, by = 2)] <- outcome_samples_x2i
  colnames(outcome_samples) <- as.vector(rbind(
    colnames(outcome_samples_x1i),
    colnames(outcome_samples_x2i)
  ))

  return(outcome_samples)
}

# ============================================================================ #
# brma.covratio.R
# ============================================================================ #
#
# This file contains the covratio method for brma model fits. COVRATIO
# measures the impact of an observation on the precision of the estimates.
#
# ============================================================================ #


#' @export
covratio <- function(model, ...) UseMethod("covratio")

#' @title COVRATIO for brma Objects
#'
#' @description Computes COVRATIO for a fitted brma object. COVRATIO measures
#' the change in the determinant of the covariance matrix of the estimates
#' when observation \eqn{i} is removed.
#'
#' @param model a fitted brma object.
#' @param type type of parameters to be summarized. Defaults to \code{"mods"}
#' (for the effect size and meta-regression coefficients). Use \code{"scale"}
#' for heterogeneity and scale-regression coefficients.
#' @param ... additional arguments. The internal \code{.weights} argument can
#' supply precomputed PSIS weights for callers that already extracted them.
#'
#' @details
#' COVRATIO is computed using importance sampling weights to approximate the
#' leave-one-out covariance matrices without refitting the model.
#' Estimate-unit LOO must first be computed with
#' \code{model <- add_loo(model, unit = "estimate")}, unless internal weights
#' are supplied.
#'
#' \deqn{COVRATIO_i = \frac{\det(Cov(\beta)_{-i})}{\det(Cov(\beta))}}
#'
#' Values > 1 indicate that the observation improves precision (decreases
#' variance), while values < 1 indicate that the observation decreases precision
#' (increases variance).
#' If any included parameter has zero posterior variance, or if a full or LOO
#' covariance determinant is zero or non-finite, COVRATIO is undefined. In that
#' case, values are reported as \code{NaN} with a printed note when available.
#'
#' @return A named numeric vector of COVRATIO values, one for each observation.
#'
#' @examples \dontrun{
#' if (requireNamespace("metadat", quietly = TRUE)) {
#'   data(dat.lehmann2018, package = "metadat")
#'   fit <- bPET(yi = yi, vi = vi, data = dat.lehmann2018, measure = "SMD")
#'   fit <- add_loo(fit)
#'
#'   covratio(fit)
#' }
#' }
#'
#' @seealso \code{\link{influence.brma}}, \code{\link{dffits.brma}}, \code{\link{cooks.distance.brma}}
#' @aliases covratio
#' @exportS3Method
covratio.brma <- function(model, type = "mods", ...) {

  dots <- list(...)
  .weights <- dots[[".weights"]]
  BayesTools::check_char(type, "type", allow_values = c("mods", "scale"))

  # Get PSIS weights (S x K matrix)
  weights <- .diagnostic_psis_weights(model, .weights)

  # determine whether to extract formula (for meta-regression) or parameter (for intercept-only)
  is_mods  <- .is_mods(model)
  is_scale <- .is_scale(model)

  # We follow dfbetas logic here:
  if (type == "mods") {
    if (is_mods) {
      # meta-regression: means mu is a formula
      samples_table <- BayesTools::JAGS_estimates_table(
        fit                = model[["fit"]],
        keep_formulas      = "mu",
        remove_diagnostics = TRUE,
        return_samples     = TRUE
      )
    } else {
      # random/fixed effects: mu is a parameter (intercept)
      samples_table <- BayesTools::JAGS_estimates_table(
        fit                = model[["fit"]],
        keep_parameters    = "mu",
        remove_diagnostics = TRUE,
        return_samples     = TRUE
      )
    }
  } else if (type == "scale") {
    if (is_scale) {
      # scale-regression: tau is modeled via log_tau formula
      samples_table <- BayesTools::JAGS_estimates_table(
        fit                = model[["fit"]],
        keep_formulas      = "log_tau",
        remove_diagnostics = TRUE,
        return_samples     = TRUE
      )
    } else {
      # random/fixed effects: tau is a parameter (intercept)
      samples_table <- BayesTools::JAGS_estimates_table(
        fit                = model[["fit"]],
        keep_parameters    = "tau",
        remove_diagnostics = TRUE,
        return_samples     = TRUE
      )
    }
  }

  # Ensure matrix (S x P)
  beta_samples <- as.matrix(samples_table)

  S <- nrow(beta_samples)
  K <- ncol(weights)

  undefined <- apply(
    beta_samples,
    2,
    function(x) diff(range(x, na.rm = TRUE)) <= sqrt(.Machine$double.eps)
  )
  if (any(undefined)) {
    return(.diagnostic_with_note(
      .diagnostic_set_names(rep(NaN, K), model),
      class = "covratio.brma",
      note  = .diagnostic_zero_variance_note(
        diagnostic = "COVRATIO",
        parameters = colnames(beta_samples)[undefined],
        variance   = "posterior"
      )
    ))
  }

  # 1. Full Covariance (Method = "ML" for consistency)
  # We construct uniform weights 1/S to use cov.wt with "ML"
  # This avoids the (S-1)/S mismatch with the LOO calculation
  w_full <- rep(1/S, S)
  cov_full_res <- stats::cov.wt(beta_samples, wt = w_full, method = "ML")

  # Log-Determinant of Full Covariance
  # determinant() returns $modulus which is the log-abs-determinant
  val_full <- as.numeric(determinant(cov_full_res$cov, logarithm = TRUE)$modulus)
  if (!is.finite(val_full)) {
    return(.diagnostic_with_note(
      .diagnostic_set_names(rep(NaN, K), model),
      class = "covratio.brma",
      note  = "COVRATIO could not be computed because the full posterior covariance determinant is zero or non-finite; values are reported as NaN."
    ))
  }

  out <- numeric(K)

  # 2. Loop over Studies for LOO Covariance
  for (i in seq_len(K)) {

    # Extract weights for observation i
    # (These should be pre-normalized, but cov.wt handles normalization too)
    w_i <- weights[, i]

    # STABILITY FIX: Use method = "ML"
    # "unbiased" uses factor 1/(1 - sum(w^2)), which explodes if ESS is low.
    # "ML" uses factor 1/sum(w) = 1, which is stable for distribution variance.
    cov_loo_res <- stats::cov.wt(beta_samples, wt = w_i, method = "ML")
    val_loo     <- as.numeric(determinant(cov_loo_res$cov, logarithm = TRUE)$modulus)

    # Compute Ratio
    out[i] <- if (is.finite(val_loo)) exp(val_loo - val_full) else NaN
  }

  out <- .diagnostic_set_names(out, model)
  if (any(is.nan(out))) {
    out <- .diagnostic_with_note(
      out,
      class = "covratio.brma",
      note  = "COVRATIO could not be computed for one or more observations because the LOO covariance determinant is zero or non-finite; affected values are reported as NaN."
    )
  }

  return(out)
}


#' @exportS3Method
print.covratio.brma <- function(x, ...) {

  note <- attr(x, "note")
  attr(x, "note") <- NULL
  class(x) <- NULL
  print(x, ...)
  .print_diagnostic_note(note)

  return(invisible(x))
}

# ============================================================================ #
# brma.wrappers.R
# ============================================================================ #
#
# This file contains user-friendly wrapper functions for common prediction
# tasks with brma model fits. These wrappers simplify the interface to
# predict.brma() for typical use cases:
#
# - fitted.brma(): In-sample fitted values
# - pooled_effect.brma(): Aggregated pooled effect size (mu)
# - pooled_heterogeneity.brma(): Aggregated pooled heterogeneity (tau)
# - blup.brma(): Best Linear Unbiased Predictions (true effects)
#
# Design principles:
# - Simple interface: minimal required arguments
# - Consistent with predict.brma: use same underlying machinery
# - Quiet operation: suppress aggregation messages for cleaner output
#
# ============================================================================ #


# ---------------------------------------------------------------------------- #
# nobs and coef methods
# ---------------------------------------------------------------------------- #

#' @title Number of Observations for brma Objects
#'
#' @description Extract the number of observations (studies/estimates) from
#' a fitted brma object.
#'
#' @param object a fitted brma object
#' @param ... additional arguments (currently ignored)
#'
#' @return An integer giving the number of observations.
#'
#' @seealso [brma()], [brma.glmm()]
#' @export
nobs.brma <- function(object, ...) {
  return(length(.outcome_data_yi(object)))
}


#' @title Extract Model Coefficients for brma Objects
#'
#' @description Extract model coefficients (posterior means) from a
#' fitted brma object.
#'
#' @param object a fitted brma object
#' @param ... additional arguments (currently ignored)
#'
#' @return A named numeric vector of posterior mean coefficients.
#'
#' @seealso [summary.brma()]
#' @export
coef.brma <- function(object, ...) {
  return(object[["coefficients"]])
}


#' @title Fitted Values for brma Objects
#'
#' @description Extract in-sample fitted values from a fitted \code{brma}
#' object.
#'
#' @param object a fitted brma object.
#' @param unit output unit. Only \code{"estimate"} is implemented currently.
#' @param conditioning_depth conditioning depth for location fitted values.
#' \code{"marginal"} uses fixed effects only, \code{"cluster"} conditions on
#' cluster-level random effects, and \code{"estimate"} conditions on the full
#' estimate-level fitted value.
#' @param component fitted component to return. \code{"location"} returns
#' location fitted values, \code{"scale"} returns fitted heterogeneity
#' \eqn{\tau_i}, and \code{"all"} returns both as a named list.
#' @param bias_adjusted whether location fitted values should adjust for
#' publication bias. Defaults to \code{FALSE}.
#' @param conditional whether to return fitted values from conditional posterior
#' predictions for RoBMA product-space objects.
#' @inheritParams predict.brma
#' @param ... additional arguments. Currently only \code{quiet} is honored.
#'
#' @details
#' This method is a compact adapter around \code{\link{predict.brma}}. It
#' summarizes posterior prediction draws by their column means and returns a
#' base numeric vector, matching the usual \code{\link[stats]{fitted}} contract.
#' Use \code{predict()} directly when posterior draws or intervals are needed.
#'
#' The default \code{conditioning_depth = "marginal"} corresponds to
#' \code{predict(object, type = "terms")} and matches the usual fitted-value
#' convention for meta-regression. For normal models,
#' \code{conditioning_depth = "estimate"} corresponds to BLUP means for the
#' observed estimates.
#'
#' For \code{component = "all"}, \code{conditioning_depth},
#' \code{output_measure}, and \code{transform} apply only to the
#' \code{location} component. The \code{scale} component always returns
#' fitted \eqn{\tau_i} values.
#'
#' @return A numeric vector of fitted values, or a named list with
#' \code{location} and \code{scale} components when \code{component = "all"}.
#'
#' @seealso [predict.brma()], [residuals.brma()], [blup.brma()]
#' @export
fitted.brma <- function(object, unit = "estimate",
                        conditioning_depth = "marginal",
                        component = "location",
                        bias_adjusted = FALSE,
                        output_measure = NULL,
                        transform = NULL,
                        conditional = FALSE, ...) {

  dots                         <- list(...)
  quiet                        <- .fitted_brma_quiet(dots)
  conditioning_depth_specified <- !missing(conditioning_depth)

  unit                         <- .normalize_unit(unit)
  conditioning_depth           <- .normalize_conditioning_depth(conditioning_depth)
  component                    <- match.arg(component, c("location", "scale", "all"))

  BayesTools::check_bool(bias_adjusted, "bias_adjusted")
  BayesTools::check_bool(conditional, "conditional")

  .check_unit_conditioning_depth(
    object             = object,
    unit               = unit,
    conditioning_depth = conditioning_depth,
    caller             = "fitted()"
  )

  if (unit == "cluster") {
    .check_cluster_unit_deferred("fitted()")
  }

  if (component == "scale" && conditioning_depth_specified &&
      conditioning_depth != "marginal") {
    stop(
      "'conditioning_depth' is not available for component = 'scale'.",
      call. = FALSE
    )
  }

  if (component == "scale" &&
      (!is.null(output_measure) || !is.null(transform))) {
    stop(
      "'output_measure' and 'transform' are only available for location ",
      "fitted values.",
      call. = FALSE
    )
  }

  if (component == "location") {
    return(.fitted_brma_location(
      object             = object,
      conditioning_depth = conditioning_depth,
      bias_adjusted      = bias_adjusted,
      output_measure     = output_measure,
      transform          = transform,
      conditional        = conditional,
      quiet              = quiet
    ))
  }

  if (component == "scale") {
    return(.fitted_brma_scale(
      object      = object,
      conditional = conditional,
      quiet       = quiet
    ))
  }

  return(list(
    location = .fitted_brma_location(
      object             = object,
      conditioning_depth = conditioning_depth,
      bias_adjusted      = bias_adjusted,
      output_measure     = output_measure,
      transform          = transform,
      conditional        = conditional,
      quiet              = quiet
    ),
    scale = .fitted_brma_scale(
      object      = object,
      conditional = conditional,
      quiet       = quiet
    )
  ))
}


.fitted_brma_quiet <- function(dots) {

  if (is.null(dots[["quiet"]])) {
    return(TRUE)
  }

  BayesTools::check_bool(dots[["quiet"]], "quiet")
  return(dots[["quiet"]])
}


.fitted_brma_location <- function(object, conditioning_depth,
                                  bias_adjusted, output_measure,
                                  transform, conditional, quiet) {

  pred_type <- switch(conditioning_depth,
    "marginal" = "terms",
    "cluster"  = "cluster",
    "estimate" = "estimate"
  )

  samples <- predict.brma(
    object         = object,
    newdata        = NULL,
    type           = pred_type,
    output_measure = output_measure,
    transform      = transform,
    bias_adjusted  = bias_adjusted,
    conditional    = conditional,
    quiet          = quiet
  )

  out <- colMeans(as.matrix(samples))
  out <- .diagnostic_set_names(out, object)

  return(out)
}


.fitted_brma_scale <- function(object, conditional, quiet) {

  samples <- predict.brma(
    object      = object,
    newdata     = NULL,
    type        = "terms.scale",
    conditional = conditional,
    quiet       = quiet
  )

  out <- colMeans(as.matrix(samples))
  out <- .diagnostic_set_names(out, object)

  return(out)
}


# ---------------------------------------------------------------------------- #
# pooled_effect generic and brma method
# ---------------------------------------------------------------------------- #

#' @title Pooled Effect Size
#'
#' @description Computes the pooled (aggregated) effect size estimate
#' from a fitted model.
#'
#' @param object a fitted model object
#' @param ... additional arguments passed to methods
#'
#' @return Method-specific return value, typically a summary table or
#' posterior samples of the pooled effect size.
#'
#' @seealso [predict.brma()]
#' @export
pooled_effect <- function(object, ...) {
  UseMethod("pooled_effect")
}


#' @title Pooled Effect Size for brma Objects
#'
#' @description Computes the pooled (aggregated) effect size estimate
#' from a fitted brma object by averaging across the moderator model matrix.
#'
#' @param object a fitted brma object
#' @param bias_adjusted whether to adjust for publication bias. Defaults to
#' \code{TRUE}. For PET/PEESE models this removes the regression bias term from
#' the pooled location effect. Selection-model weighting affects response
#' predictions, not this \code{type = "terms"} wrapper.
#' @param probs quantiles of the posterior distribution to be displayed.
#' Defaults to \code{c(.025, .975)} for 95% credible intervals.
#' @param conditional whether to return the pooled effect conditional on the
#' effect component for RoBMA product-space objects. Defaults to \code{FALSE}.
#' @inheritParams predict.brma
#' @param ... additional arguments passed to \code{\link{predict.brma}}; wrapper
#' arguments such as \code{newdata}, \code{type}, and \code{quiet} are fixed.
#'
#' @details
#' This function is a convenience wrapper around \code{predict.brma(...,
#' type = "terms", newdata = TRUE, bias_adjusted = TRUE, quiet = TRUE)}.
#'
#' For meta-regression models, the pooled effect averages the effect size
#' estimate across moderator levels proportionately to the levels observed
#' in the data. This provides an estimate representative of the sample of
#' studies.
#'
#' For models without moderators, this returns the single mu parameter.
#'
#' @return A \code{brma_samples} object containing posterior samples. When printed,
#' displays a summary table. Use \code{summary()} to obtain the summary table directly.
#' The samples can be converted to \pkg{posterior} draws formats using \code{as_draws()}.
#'
#' @examples \dontrun{
#' if (requireNamespace("metadat", quietly = TRUE)) {
#'   data(dat.lehmann2018, package = "metadat")
#'   fit <- brma(
#'     yi      = yi,
#'     vi      = vi,
#'     data    = dat.lehmann2018,
#'     measure = "SMD",
#'     seed    = 1,
#'     silent  = TRUE
#'   )
#'
#'   pooled_effect(fit)
#' }
#' }
#'
#' @seealso [predict.brma()], [pooled_heterogeneity()], [blup()]
#' @export
pooled_effect.brma <- function(object, bias_adjusted = TRUE,
                               output_measure = NULL, transform = NULL,
                               probs = c(.025, .975),
                               conditional = FALSE, ...) {
  out <- predict.brma(
    object         = object,
    newdata        = TRUE,
    type           = "terms",
    output_measure = output_measure,
    transform      = transform,
    probs          = probs,
    bias_adjusted  = bias_adjusted,
    conditional    = conditional,
    quiet          = TRUE,
    ...
  )
  attr(out, "title") <- .effect_output_title(
    title            = if (conditional) {
      "Conditional Pooled Effect Size"
    } else {
      "Pooled Effect Size"
    },
    effect_transform = attr(out, "effect_transform")
  )
  return(out)
}


# ---------------------------------------------------------------------------- #
# pooled_heterogeneity generic and brma method
# ---------------------------------------------------------------------------- #

#' @title Pooled Heterogeneity
#'
#' @description Computes the pooled (aggregated) heterogeneity estimate (tau)
#' from a fitted model.
#'
#' @param object a fitted model object
#' @param ... additional arguments passed to methods
#'
#' @return Method-specific return value, typically a summary table or
#' posterior samples of the pooled heterogeneity.
#'
#' @seealso [predict.brma()]
#' @export
pooled_heterogeneity <- function(object, ...) {
  UseMethod("pooled_heterogeneity")
}


#' @title Pooled Heterogeneity for brma Objects
#'
#' @description Computes the pooled (aggregated) heterogeneity estimate (tau)
#' from a fitted brma object by averaging across the scale model matrix.
#'
#' @param object a fitted brma object
#' @param probs quantiles of the posterior distribution to be displayed.
#' Defaults to \code{c(.025, .975)} for 95% credible intervals.
#' @param conditional whether to return the pooled heterogeneity conditional on
#' the heterogeneity component for RoBMA product-space objects. Defaults to
#' \code{FALSE}.
#' @param ... additional arguments passed to \code{\link{predict.brma}}; wrapper
#' arguments such as \code{newdata}, \code{type}, and \code{quiet} are fixed.
#'
#' @details
#' This function is a convenience wrapper around \code{predict.brma(...,
#' type = "terms.scale", newdata = TRUE)}.
#'
#' For location-scale models (with scale regression), the pooled heterogeneity
#' averages tau across the scale model matrix proportionately to the levels
#' observed in the data.
#'
#' For models without scale regression, this returns the single tau parameter.
#'
#' For multilevel (3-level) models, the returned tau is the total heterogeneity:
#' \code{tau = sqrt(tau_within^2 + tau_between^2)}.
#'
#' @return A \code{brma_samples} object containing posterior samples. When printed,
#' displays a summary table. Use \code{summary()} to obtain the summary table directly.
#' The samples can be converted to \pkg{posterior} draws formats using \code{as_draws()}.
#'
#' @examples \dontrun{
#' if (requireNamespace("metadat", quietly = TRUE)) {
#'   data(dat.lehmann2018, package = "metadat")
#'   fit <- brma(
#'     yi      = yi,
#'     vi      = vi,
#'     data    = dat.lehmann2018,
#'     measure = "SMD",
#'     seed    = 1,
#'     silent  = TRUE
#'   )
#'
#'   pooled_heterogeneity(fit)
#' }
#' }
#'
#' @seealso [predict.brma()], [pooled_effect()], [blup()]
#' @export
pooled_heterogeneity.brma <- function(object, probs = c(.025, .975),
                                      conditional = FALSE, ...) {
  out <- predict.brma(
    object      = object,
    newdata     = TRUE,
    type        = "terms.scale",
    probs       = probs,
    conditional = conditional,
    quiet       = TRUE,
    ...
  )
  attr(out, "title") <- if (conditional) {
    "Conditional Pooled Heterogeneity"
  } else {
    "Pooled Heterogeneity"
  }
  return(out)
}


# ---------------------------------------------------------------------------- #
# blup generic and brma method
# ---------------------------------------------------------------------------- #

#' @title Best Linear Unbiased Predictions (BLUPs)
#'
#' @description Computes the estimated true effects (theta) from a
#' fitted model. These correspond to Best Linear Unbiased Predictions (BLUPs)
#' or empirical Bayes estimates.
#'
#' @param object a fitted model object
#' @param ... additional arguments passed to methods
#'
#' @return Method-specific return value, typically a summary table or
#' posterior samples of BLUP or empirical-Bayes true-effect summaries.
#'
#' @seealso [predict.brma()]
#' @export
blup <- function(object, ...) {
  UseMethod("blup")
}


#' @title Best Linear Unbiased Predictions for brma Objects
#'
#' @description Computes the estimated true effects (theta) for a
#' fitted brma object. These correspond to Best Linear Unbiased Predictions
#' (BLUPs) or empirical Bayes estimates.
#'
#' @param object a fitted brma object
#' @param bias_adjusted whether to adjust for publication bias. Defaults to
#' \code{FALSE}, which returns estimates including publication bias effects
#' (i.e., what we expect the true effects to be given the biased observations).
#' Set to \code{TRUE} to obtain bias-corrected estimates.
#' @param probs quantiles of the posterior distribution to be displayed.
#' Defaults to \code{c(.025, .975)} for 95% credible intervals.
#' @inheritParams predict.brma
#' @param ... additional arguments passed to \code{\link{predict.brma}}; wrapper
#' arguments such as \code{newdata}, \code{type}, \code{quiet},
#' \code{output_measure}, and \code{transform} are fixed by this method.
#'
#' @details
#' This function is a convenience wrapper around \code{predict.brma(...,
#' type = "effect", newdata = NULL)}.
#'
#' For unweighted two-level normal models, true effects are computed using
#' empirical Bayes shrinkage:
#' \deqn{\theta_i = \lambda_i \cdot y_i + (1 - \lambda_i) \cdot \mu_i}
#' where \eqn{\lambda_i = \tau^2 / (\tau^2 + se_i^2)}.
#' With likelihood weights, \eqn{se_i^2} is replaced by the weighted sampling
#' variance \eqn{se_i^2 / w_i}.
#'
#' For GLMM models (binomial, Poisson), the estimate-level random effects
#' are extracted directly from the posterior samples.
#'
#' For multilevel (3-level) normal models, cluster-level effects are estimated
#' jointly within cluster blocks and estimate-level effects are then shrunk
#' conditional on those cluster effects.
#'
#' @return A \code{brma_samples} object containing posterior draws of BLUP or
#' empirical-Bayes true-effect summaries with one column per estimate. For
#' existing normal data, these are conditional BLUP means, not simulated
#' latent-effect draws. When printed, displays a summary table. Use
#' \code{summary()} to obtain the summary table directly. The samples can be
#' converted to \pkg{posterior} draws formats using \code{as_draws()}.
#'
#' @examples \dontrun{
#' if (requireNamespace("metadat", quietly = TRUE)) {
#'   data(dat.lehmann2018, package = "metadat")
#'   fit <- brma(
#'     yi      = yi,
#'     vi      = vi,
#'     data    = dat.lehmann2018,
#'     measure = "SMD",
#'     seed    = 1,
#'     silent  = TRUE
#'   )
#'
#'   blup(fit)
#' }
#' }
#'
#' @seealso [predict.brma()], [pooled_effect()], [pooled_heterogeneity()],
#' [true_effects()]
#' @export
blup.brma <- function(object, bias_adjusted = FALSE,
                      output_measure = NULL, transform = NULL,
                      probs = c(.025, .975), ...) {
  out <- predict.brma(
    object         = object,
    newdata        = NULL,
    type           = "effect",
    output_measure = output_measure,
    transform      = transform,
    probs          = probs,
    bias_adjusted  = bias_adjusted,
    quiet          = TRUE,
    ...
  )
  attr(out, "title") <- .effect_output_title(
    title            = "True Effects (BLUP Means)",
    effect_transform = attr(out, "effect_transform")
  )
  return(out)
}


# ---------------------------------------------------------------------------- #
# true_effects generic and brma method (alias for blup)
# ---------------------------------------------------------------------------- #

#' @title True Effects
#'
#' @description Computes the estimated true effects (theta) from a
#' fitted model. This is a separate S3 generic whose \code{brma} method
#' delegates to \code{\link{blup.brma}}.
#'
#' @param object a fitted model object
#' @param ... additional arguments passed to methods
#'
#' @return Method-specific return value, typically a summary table or
#' posterior samples of BLUP or empirical-Bayes true-effect summaries.
#'
#' @seealso [blup()], [predict.brma()]
#' @export
true_effects <- function(object, ...) {
  UseMethod("true_effects")
}


#' @title True Effects for brma Objects
#'
#' @description Computes the estimated true effects (theta) for a
#' fitted brma object. This is an alias for \code{\link{blup.brma}}.
#'
#' @inheritParams blup.brma
#'
#' @details
#' This function is identical to \code{\link{blup.brma}}. See that function
#' for full details on how true effects are computed.
#'
#' @return A \code{brma_samples} object containing posterior draws of BLUP or
#' empirical-Bayes true-effect summaries with one column per estimate. For
#' existing normal data, these are conditional BLUP means, not simulated
#' latent-effect draws. When printed, displays a summary table. Use
#' \code{summary()} to obtain the summary table directly. The samples can be
#' converted to \pkg{posterior} draws formats using \code{as_draws()}.
#'
#' @examples \dontrun{
#' if (requireNamespace("metadat", quietly = TRUE)) {
#'   data(dat.lehmann2018, package = "metadat")
#'   fit <- brma(
#'     yi      = yi,
#'     vi      = vi,
#'     data    = dat.lehmann2018,
#'     measure = "SMD",
#'     seed    = 1,
#'     silent  = TRUE
#'   )
#'
#'   true_effects(fit)
#' }
#' }
#'
#' @seealso [blup.brma()], [predict.brma()], [pooled_effect()],
#' [pooled_heterogeneity()]
#' @export
true_effects.brma <- function(object, bias_adjusted = FALSE,
                              output_measure = NULL, transform = NULL,
                              probs = c(.025, .975), ...) {
  blup.brma(
    object         = object,
    bias_adjusted  = bias_adjusted,
    output_measure = output_measure,
    transform      = transform,
    probs          = probs,
    ...
  )
}


# ---------------------------------------------------------------------------- #
# ranef generic and brma method
# ---------------------------------------------------------------------------- #

#' @title Random Effects
#'
#' @description Extracts the estimated random effects (deviations from the
#' fixed-effect predictions) from a fitted model.
#'
#' @param object a fitted model object
#' @param ... additional arguments passed to methods
#'
#' @return Method-specific return value, typically posterior samples of the
#' random effect deviations.
#'
#' @seealso [blup()], [predict.brma()]
#' @export
ranef <- function(object, ...) {
  UseMethod("ranef")
}


#' @title Random Effects for brma Objects
#'
#' @description Extracts random effect deviations from a fitted brma object.
#' These are posterior-sample offsets from the fixed-effect predictions,
#' analogous to random-effect deviations returned by \code{metafor::ranef()}.
#'
#' @param object a fitted brma object
#' @param bias_adjusted whether to adjust for publication bias. Defaults to
#' \code{FALSE}. See \code{\link{blup.brma}} for details.
#' @param probs quantiles of the posterior distribution to be displayed.
#' Defaults to \code{c(.025, .975)} for 95% credible intervals.
#' @param ... additional arguments forwarded to \code{\link{predict.brma}} for
#' supported options such as \code{conditional}. \code{newdata}, \code{type},
#' \code{quiet}, \code{output_measure}, and \code{transform} are controlled by
#' this method.
#'
#' @details
#' Random effects are computed as the difference between the true
#' effects (BLUPs) and the fixed-effect predictions:
#' \deqn{u_i = \hat{\theta}_i - \hat{\mu}_i}
#'
#' For standard (2-level) models, returns a single \code{brma_samples}
#' object with the estimate-level random effects.
#'
#' For multilevel (3-level) models, returns a list with two observation-aligned
#' \code{brma_samples} matrices, one column per estimate row:
#' \describe{
#'   \item{\code{cluster}}{Cluster-level random effects
#'     (\eqn{\gamma_j \cdot \tau_{between}}), representing between-cluster
#'     deviations from the fixed effects.}
#'   \item{\code{estimate}}{Estimate-level random effects
#'     (\eqn{\theta_i - \mu_i - \gamma_j \cdot \tau_{between}}),
#'     representing within-cluster deviations from the cluster means.}
#' }
#'
#' @return For 2-level models, a \code{brma_samples} object. For 3-level
#' models, a named list of \code{brma_samples} objects (one per variance
#' component).
#'
#' @examples \dontrun{
#' if (requireNamespace("metadat", quietly = TRUE)) {
#'   data(dat.lehmann2018, package = "metadat")
#'   fit <- brma(
#'     yi      = yi,
#'     vi      = vi,
#'     data    = dat.lehmann2018,
#'     measure = "SMD",
#'     seed    = 1,
#'     silent  = TRUE
#'   )
#'
#'   ranef(fit)
#' }
#' }
#'
#' @seealso [blup.brma()], [predict.brma()], [pooled_effect()]
#' @export
ranef.brma <- function(object, bias_adjusted = FALSE,
                       probs = c(.025, .975), ...) {

  dots <- list(...)
  if (!is.null(dots[["output_measure"]]) || !is.null(dots[["transform"]])) {
    stop(
      "'output_measure' and 'transform' are not available for random-effect ",
      "deviations.",
      call. = FALSE
    )
  }

  is_multilevel <- .is_multilevel(object)

  # get BLUPs (fixed + all random effects)
  blup_samples <- predict.brma(
    object        = object,
    newdata       = NULL,
    type          = "estimate",
    probs         = probs,
    bias_adjusted = bias_adjusted,
    quiet         = TRUE,
    ...
  )

  # extract MCMC chain info from the returned samples
  n_chains <- attr(blup_samples, "nchains")
  n_iter   <- attr(blup_samples, "niter")
  data     <- object[["data"]]
  labels   <- .get_estimate_labels(object)

  # get fixed-effect predictions only
  terms_samples <- predict.brma(
    object        = object,
    newdata       = NULL,
    type          = "terms",
    probs         = probs,
    bias_adjusted = bias_adjusted,
    quiet         = TRUE,
    ...
  )

  if (!is_multilevel) {

    # 2-level: random effects = BLUP - fixed effects
    ranef_mat <- unclass(blup_samples) - unclass(terms_samples)
    K         <- ncol(ranef_mat)
    colnames(ranef_mat) <- paste0("u[", labels[seq_len(K)], "]")

    return(.new_brma_samples(
      samples  = ranef_mat,
      n_chains = n_chains,
      n_iter   = n_iter,
      title    = "Random Effects:",
      probs    = probs,
      data     = data
    ))

  } else {

    # 3-level: decompose into cluster-level and estimate-level
    cluster_samples <- predict.brma(
      object        = object,
      newdata       = NULL,
      type          = "cluster",
      probs         = probs,
      bias_adjusted = bias_adjusted,
      quiet         = TRUE,
      ...
    )

    # cluster-level random effects: cluster predictions - fixed effects
    cluster_ranef_mat <- unclass(cluster_samples) - unclass(terms_samples)
    K                 <- ncol(cluster_ranef_mat)
    cluster          <- data[["outcome"]][["cluster"]]
    cluster_labels   <- .get_cluster_labels(object)[as.character(cluster)]
    cluster_labels   <- unname(cluster_labels)
    cluster_missing  <- is.na(cluster_labels)
    cluster_labels[cluster_missing] <- as.character(cluster[cluster_missing])
    if (any(duplicated(cluster_labels))) {
      cluster_names <- paste0(cluster_labels, "|", labels[seq_len(K)])
    } else {
      cluster_names <- cluster_labels
    }
    colnames(cluster_ranef_mat) <- paste0("u_cluster[", cluster_names, "]")

    cluster_ranef <- .new_brma_samples(
      samples  = cluster_ranef_mat,
      n_chains = n_chains,
      n_iter   = n_iter,
      title    = "Cluster-Level Random Effects:",
      probs    = probs,
      data     = data
    )

    # estimate-level random effects: BLUP - cluster predictions
    estimate_ranef_mat <- unclass(blup_samples) - unclass(cluster_samples)
    colnames(estimate_ranef_mat) <- paste0("u_estimate[", labels[seq_len(K)], "]")

    estimate_ranef <- .new_brma_samples(
      samples  = estimate_ranef_mat,
      n_chains = n_chains,
      n_iter   = n_iter,
      title    = "Estimate-Level Random Effects:",
      probs    = probs,
      data     = data
    )

    out <- list(
      cluster  = cluster_ranef,
      estimate = estimate_ranef
    )
    return(out)
  }
}

.test_object <- function(name) {

  frames <- sys.frames()
  for (i in rev(seq_along(frames))) {
    if (exists(name, envir = frames[[i]], inherits = FALSE)) {
      return(get(name, envir = frames[[i]], inherits = FALSE))
    }
  }

  return(get(name, envir = parent.frame(), inherits = TRUE))
}

.test_fits <- function() {

  return(.test_object("fits"))
}

.test_info <- function() {

  return(.test_object("info"))
}

.case_fit <- function(case) {

  return(.test_fits()[[case_name(case)]])
}

.case_metafor <- function(case) {

  return(.test_info()[[case_name(case)]][["metafor"]])
}

.metafor_dfbetas <- function(fit_metafor) {

  if (inherits(fit_metafor, "rma.mv")) {
    return(as.data.frame(metafor::dfbetas.rma.mv(fit_metafor)))
  }

  return(as.data.frame(metafor::dfbetas.rma.uni(fit_metafor)))
}

.expect_residuals_equal <- function(name, actual, expected, tolerance, label) {

  testthat::expect_equal(
    actual,
    as.vector(expected),
    tolerance = tolerance,
    info      = paste(name, label, "matches metafor")
  )
}

.expect_residual_table_equal <- function(name, actual, expected, tolerance,
                                         cols = c("resid", "se", "z"),
                                         rows = NULL, z_tolerance = tolerance,
                                         label = "rstandard") {

  if (is.null(rows)) {
    rows <- seq_len(nrow(actual))
  }

  for (col in cols) {
    col_tolerance <- if (col == "z") z_tolerance else tolerance
    testthat::expect_equal(
      actual[[col]][rows],
      expected[[col]][rows],
      tolerance = col_tolerance,
      info      = paste(name, label, col, "matches metafor")
    )
  }
}

.expect_residual_rstudent_rank <- function(name, brma_rstudent, metafor_rstudent,
                                           resid_cor = 0.90, se_cor = NULL,
                                           z_cor = 0.90) {

  testthat::expect_true(all(is.finite(brma_rstudent$resid)),
                        info = paste(name, "LOO-PIT residual means finite"))
  testthat::expect_true(all(is.finite(brma_rstudent$se)) && all(brma_rstudent$se > 0),
                        info = paste(name, "LOO predictive SEs finite and positive"))
  testthat::expect_true(
    stats::cor(brma_rstudent$resid, metafor_rstudent$resid,
               method = "spearman", use = "complete.obs") > resid_cor,
    info = paste(name, "LOO-PIT residual ranks align with metafor")
  )
  if (!is.null(se_cor)) {
    testthat::expect_true(
      stats::cor(brma_rstudent$se, metafor_rstudent$se,
                 method = "spearman", use = "complete.obs") > se_cor,
      info = paste(name, "LOO predictive SE ranks align with metafor")
    )
  }
  testthat::expect_true(
    stats::cor(brma_rstudent$z, metafor_rstudent$z,
               method = "spearman", use = "complete.obs") > z_cor,
    info = paste(name, "LOO-PIT z ranks align with metafor")
  )
}

expect_residuals_match_metafor <- function(case) {

  name        <- case_name(case)
  fit_brma    <- .case_fit(case)
  fit_metafor <- .case_metafor(case)
  kind        <- case_value(case, "kind", "standard")
  tolerance   <- case_value(case, "tolerance", 0.05)

  if (kind %in% c("standard", "regression", "multilevel", "multilevel_no_loo",
                  "pet", "pet_reg", "peese", "peese_reg")) {
    .expect_residuals_equal(name, residuals(fit_brma), residuals(fit_metafor),
                            tolerance, "outcome residuals")
    .expect_residuals_equal(name, residuals(fit_brma, type = "pearson"),
                            residuals(fit_metafor, type = "pearson"),
                            tolerance, "pearson residuals")
    .expect_residual_table_equal(name, rstandard(fit_brma), rstandard(fit_metafor),
                                 tolerance)
  }

  if (kind == "regression") {
    .expect_residual_table_equal(
      name,
      rstandard(fit_brma, conditioning_depth = "estimate"),
      rstandard(fit_metafor, type = "conditional"),
      tolerance,
      label = "conditional rstandard"
    )
  }

  if (kind == "interaction") {
    .expect_residual_table_equal(
      name,
      rstandard(fit_brma),
      rstandard(fit_metafor),
      tolerance,
      cols = c("resid", "se")
    )

    metafor_conditional <- suppressWarnings(rstandard(fit_metafor, type = "conditional"))
    brma_conditional    <- rstandard(fit_brma, conditioning_depth = "estimate")
    valid <- is.finite(metafor_conditional$se) & is.finite(metafor_conditional$z)

    .expect_residual_table_equal(
      name,
      brma_conditional,
      metafor_conditional,
      tolerance,
      cols = c("resid", "se"),
      rows = which(valid),
      label = "conditional rstandard"
    )
    testthat::expect_true(all(abs(brma_conditional$resid[!valid]) < 1e-8),
                          info = paste(name, "saturated residuals are zero"))
    testthat::expect_true(all(brma_conditional$se[!valid] < 1e-8),
                          info = paste(name, "saturated SEs collapse"))
    testthat::expect_true(all(abs(brma_conditional$z[!valid]) < 1e-8),
                          info = paste(name, "saturated z values are zero"))
  }

  if (kind %in% c("multilevel", "multilevel_no_loo")) {
    metafor_conditional <- metafor_rstandard_conditional_mv(fit_metafor)
    brma_conditional    <- rstandard(fit_brma, conditioning_depth = "estimate")
    brma_theta          <- colMeans(as.matrix(predict(fit_brma, type = "estimate", quiet = TRUE)))
    metafor_theta       <- fit_metafor[["yi"]] - metafor_conditional$resid

    testthat::expect_equal(as.vector(brma_theta), as.vector(metafor_theta),
                           tolerance = 0.11,
                           info = paste(name, "estimate-level BLUP oracle"))
    .expect_residual_table_equal(
      name,
      brma_conditional,
      metafor_conditional,
      tolerance = 0.11,
      cols = "resid",
      label = "multilevel conditional rstandard"
    )
    .expect_residual_table_equal(
      name,
      brma_conditional,
      metafor_conditional,
      tolerance = 0.05,
      cols = "se",
      label = "multilevel conditional rstandard"
    )
  }

  if (kind %in% c("selection_pos", "selection_reg", "selection_neg")) {
    .expect_residuals_equal(name, residuals(fit_brma), residuals(fit_metafor),
                            tolerance, "outcome residuals")
    testthat::expect_error(
      residuals(fit_brma, type = "pearson"),
      "Pearson residuals are not available for selection models",
      info = paste(name, "pearson residuals are rejected")
    )
    testthat::expect_error(
      residuals(fit_brma, type = "rstandard"),
      "Standardized residuals.*are not available for selection models",
      info = paste(name, "rstandard residuals are rejected")
    )
  }

  if (kind %in% c("glmm", "glmm_reg")) {
    .expect_residuals_equal(name, residuals(fit_brma), residuals(fit_metafor),
                            tolerance, "outcome residuals")
    testthat::expect_error(
      residuals(fit_brma, type = "pearson"),
      "Pearson residuals are only available for normal outcome models",
      info = paste(name, "pearson residuals are rejected")
    )
    testthat::expect_error(
      residuals(fit_brma, type = "rstandard"),
      "Standardized residuals.*are only available for normal outcome models",
      info = paste(name, "rstandard residuals are rejected")
    )
  }

  rstudent_kind <- case_value(case, "rstudent", NULL)
  if (identical(rstudent_kind, "equal")) {
    .expect_residual_table_equal(
      name,
      suppressWarnings(rstudent(fit_brma)),
      rstudent(fit_metafor),
      tolerance,
      z_tolerance = if (kind == "pet") 0.10 else tolerance,
      label = "rstudent"
    )
  } else if (identical(rstudent_kind, "rank")) {
    .expect_residual_rstudent_rank(
      name,
      suppressWarnings(rstudent(fit_brma)),
      rstudent(fit_metafor),
      resid_cor = if (kind == "multilevel") 0.65 else 0.90,
      se_cor    = if (kind == "multilevel") 0.80 else NULL,
      z_cor     = if (kind == "multilevel") 0.65 else 0.90
    )
  } else if (identical(rstudent_kind, "selection_pos") ||
             identical(rstudent_kind, "selection_neg")) {
    brma_resid    <- residuals(fit_brma)
    brma_rstudent <- suppressWarnings(rstudent(fit_brma))
    if (rstudent_kind == "selection_pos") {
      testthat::expect_true(brma_rstudent$z[4] > 4,
                            info = paste(name, "large positive outlier"))
    } else {
      testthat::expect_true(brma_rstudent$z[4] < -4,
                            info = paste(name, "large negative outlier"))
    }
    testthat::expect_true(all(abs(brma_rstudent$z[-4]) < 4),
                          info = paste(name, "single large outlier"))
    testthat::expect_true(stats::cor(brma_resid, brma_rstudent$z) > 0.9,
                          info = paste(name, "rstudent and residuals align"))
    testthat::expect_true(all(is.finite(brma_rstudent$se)),
                          info = paste(name, "selection LOO SEs finite"))
    testthat::expect_true(all(brma_rstudent$se > 0),
                          info = paste(name, "selection LOO SEs positive"))
  } else if (identical(rstudent_kind, "glmm_align")) {
    brma_resid    <- residuals(fit_brma)
    brma_rstudent <- suppressWarnings(rstudent(fit_brma, type = "estimate"))
    testthat::expect_true(stats::cor(brma_resid, brma_rstudent$z) > 0.9,
                          info = paste(name, "rstudent and residuals align"))
  }
}

.sample_means <- function(x) {

  return(summary(x)[, "Mean"])
}

.sample_mean <- function(x, row, column = "Mean") {

  return(summary(x)[row, column])
}

.expect_blup_tracks_observed <- function(name, fit_brma, theta) {

  testthat::expect_true(
    stats::cor(fit_brma$data$outcome$yi, theta, method = "spearman") > 0.8,
    info = paste(name, "BLUPs correlate with observed estimates")
  )
}

.expect_blup_shrinks <- function(name, fit_brma, theta, mu) {

  yi <- fit_brma$data$outcome$yi
  testthat::expect_true(
    mean(abs(theta - mu)) < mean(abs(yi - mu)),
    info = paste(name, "BLUPs shrink toward pooled estimate")
  )
  .expect_blup_tracks_observed(name, fit_brma, theta)
}

expect_prediction_matches_metafor <- function(case) {

  name        <- case_name(case)
  fit_brma    <- .case_fit(case)
  fit_metafor <- .case_metafor(case)
  kind        <- case_value(case, "kind", "simple")
  tolerance   <- case_value(case, "tolerance", 0.05)
  tau_tol     <- case_value(case, "tau_tolerance", tolerance)
  fitted_tol  <- case_value(case, "fitted_tolerance", tolerance)

  if (kind == "simple") {
    mu_predict <- .sample_mean(predict(fit_brma, type = "terms", newdata = TRUE), "mu")
    mu_wrapper <- .sample_mean(pooled_effect(fit_brma), "mu")
    testthat::expect_equal(mu_predict, mu_wrapper,
                           info = paste(name, "pooled_effect wrapper"))
    testthat::expect_equal(mu_predict, fit_metafor$beta[[1]], tolerance = tolerance,
                           info = paste(name, "pooled effect matches metafor"))

    tau_predict <- .sample_mean(predict(fit_brma, type = "terms.scale", newdata = TRUE), "tau")
    tau_wrapper <- .sample_mean(pooled_heterogeneity(fit_brma), "tau")
    testthat::expect_equal(tau_predict, tau_wrapper,
                           info = paste(name, "pooled_heterogeneity wrapper"))
    testthat::expect_equal(tau_predict, sqrt(fit_metafor$tau2), tolerance = tau_tol,
                           info = paste(name, "tau matches metafor"))

    theta_predict <- .sample_means(predict(fit_brma, type = "effect"))
    theta_blup    <- .sample_means(blup(fit_brma))
    theta_true    <- .sample_means(true_effects(fit_brma))
    testthat::expect_equal(theta_predict, theta_blup,
                           info = paste(name, "predict effect matches blup"))
    testthat::expect_equal(theta_blup, theta_true,
                           info = paste(name, "true_effects match blup"))
    testthat::expect_equal(theta_blup, metafor::blup(fit_metafor)$pred,
                           tolerance = 0.05,
                           info = paste(name, "BLUPs match metafor"))
    .expect_blup_tracks_observed(name, fit_brma, theta_true)
    return(invisible(TRUE))
  }

  if (kind %in% c("regression", "interaction")) {
    metafor_fitted <- predict(fit_metafor)$pred
    brma_fitted    <- .sample_means(predict(fit_brma, type = "terms"))

    if (kind == "interaction") {
      testthat::expect_true(stats::cor(brma_fitted, metafor_fitted) > 0.90,
                            info = paste(name, "fitted values align"))
    } else {
      testthat::expect_equal(brma_fitted, metafor_fitted, tolerance = fitted_tol,
                             info = paste(name, "fitted values match metafor"))
    }

    testthat::expect_equal(
      .sample_mean(pooled_effect(fit_brma), "mu"),
      mean(metafor_fitted),
      tolerance = tolerance,
      info      = paste(name, "pooled effect matches mean metafor fitted")
    )

    theta_brma    <- .sample_means(blup(fit_brma))
    theta_metafor <- metafor::blup(fit_metafor)$pred
    if (kind == "interaction") {
      testthat::expect_true(stats::cor(theta_brma, theta_metafor) > 0.90,
                            info = paste(name, "BLUPs align with metafor"))
    } else {
      testthat::expect_equal(theta_brma, theta_metafor, tolerance = 0.05,
                             info = paste(name, "BLUPs match metafor"))
    }
    .expect_blup_tracks_observed(name, fit_brma, theta_brma)
    return(invisible(TRUE))
  }

  if (kind == "scale") {
    metafor_fitted <- predict(fit_metafor)$pred
    brma_fitted    <- .sample_means(predict(fit_brma, type = "terms"))
    testthat::expect_equal(brma_fitted, metafor_fitted, tolerance = fitted_tol,
                           info = paste(name, "location predictions"))

    metafor_tau <- sqrt(as.vector(exp(fit_metafor$Z %*% fit_metafor$alpha)))
    brma_tau    <- .sample_means(predict(fit_brma, type = "terms.scale"))
    testthat::expect_equal(brma_tau, metafor_tau, tolerance = tau_tol,
                           info = paste(name, "study tau predictions"))
    testthat::expect_equal(.sample_mean(pooled_heterogeneity(fit_brma), "tau"),
                           mean(metafor_tau), tolerance = tau_tol,
                           info = paste(name, "pooled tau"))

    theta_brma <- .sample_means(blup(fit_brma))
    testthat::expect_equal(theta_brma, metafor::blup(fit_metafor)$pred,
                           tolerance = 0.10,
                           info = paste(name, "BLUPs match metafor"))
    .expect_blup_tracks_observed(name, fit_brma, theta_brma)
    return(invisible(TRUE))
  }

  if (kind == "multilevel") {
    brma_mu  <- .sample_mean(pooled_effect(fit_brma), "mu")
    brma_tau <- .sample_mean(pooled_heterogeneity(fit_brma), "tau")

    testthat::expect_equal(brma_mu, fit_metafor$beta[[1]], tolerance = tolerance,
                           info = paste(name, "pooled effect"))
    testthat::expect_equal(brma_tau, sqrt(fit_metafor$tau2), tolerance = tau_tol,
                           info = paste(name, "total tau"))

    theta <- .sample_means(blup(fit_brma))
    testthat::expect_equal(length(theta), nrow(fit_brma$data$outcome),
                           info = paste(name, "BLUP length"))
    .expect_blup_shrinks(name, fit_brma, theta, brma_mu)
    return(invisible(TRUE))
  }

  if (kind %in% c("glmm", "glmm_reg")) {
    brma_mu  <- .sample_mean(pooled_effect(fit_brma), "mu")
    brma_tau <- .sample_mean(pooled_heterogeneity(fit_brma), "tau")
    testthat::expect_equal(brma_mu, fit_metafor$beta[[1]], tolerance = tolerance,
                           info = paste(name, "pooled effect"))
    testthat::expect_equal(brma_tau, sqrt(fit_metafor$tau2), tolerance = tau_tol,
                           info = paste(name, "pooled tau"))

    theta <- .sample_means(blup(fit_brma))
    testthat::expect_equal(length(theta), nrow(fit_brma$data$outcome),
                           info = paste(name, "BLUP length"))
    return(invisible(TRUE))
  }

  if (kind == "selection") {
    brma_mu  <- .sample_mean(pooled_effect(fit_brma), "mu")
    brma_tau <- .sample_mean(pooled_heterogeneity(fit_brma), "tau")
    testthat::expect_equal(brma_mu, fit_metafor$beta[[1]], tolerance = tolerance,
                           info = paste(name, "pooled effect"))
    testthat::expect_equal(brma_tau, sqrt(fit_metafor$tau2), tolerance = tau_tol,
                           info = paste(name, "pooled tau"))

    theta <- .sample_means(blup(fit_brma))
    testthat::expect_equal(length(theta), nrow(fit_brma$data$outcome),
                           info = paste(name, "BLUP length"))
    .expect_blup_shrinks(name, fit_brma, theta, brma_mu)
    return(invisible(TRUE))
  }

  if (kind %in% c("pet", "pet_reg", "peese", "peese_reg")) {
    is_regression_bias <- kind %in% c("pet_reg", "peese_reg")
    bias_column        <- if (kind %in% c("pet", "pet_reg")) "sqrt(vi)" else "vi"

    if (is_regression_bias) {
      bias_newmods <- matrix(
        0,
        nrow     = nrow(fit_metafor$X),
        ncol     = 1,
        dimnames = list(NULL, bias_column)
      )
      metafor_newmods_bias_adjusted <- cbind(
        bias_newmods,
        fit_metafor$X[, -c(1, 2), drop = FALSE]
      )
      metafor_mu <- mean(predict(fit_metafor, newmods = metafor_newmods_bias_adjusted)$pred)
    } else {
      metafor_newmods_bias_adjusted <- rep(0, nrow(fit_brma$data$outcome))
      metafor_mu <- predict(fit_metafor, newmods = 0)$pred
    }

    brma_mu <- .sample_mean(pooled_effect(fit_brma), "mu")
    testthat::expect_equal(brma_mu, metafor_mu, tolerance = tolerance,
                           info = paste(name, "bias-adjusted pooled effect"))
    testthat::expect_equal(.sample_mean(pooled_heterogeneity(fit_brma), "tau"),
                           sqrt(fit_metafor$tau2), tolerance = tau_tol,
                           info = paste(name, "pooled tau"))

    metafor_fitted <- predict(fit_metafor)$pred
    brma_fitted    <- .sample_means(predict(fit_brma, type = "terms"))
    testthat::expect_equal(brma_fitted, metafor_fitted, tolerance = fitted_tol,
                           info = paste(name, "unadjusted fitted values"))

    theta <- .sample_means(blup(fit_brma))
    testthat::expect_equal(theta, metafor::blup(fit_metafor)$pred,
                           tolerance = 0.05,
                           info = paste(name, "BLUPs match metafor"))
    .expect_blup_tracks_observed(name, fit_brma, theta)

    adjusted_metafor <- predict(fit_metafor, newmods = metafor_newmods_bias_adjusted)$pred
    adjusted_brma    <- .sample_means(predict(fit_brma, type = "terms", bias_adjusted = TRUE))
    testthat::expect_equal(adjusted_brma, adjusted_metafor, tolerance = 0.05,
                           info = paste(name, "bias-adjusted fitted values"))
    return(invisible(TRUE))
  }

  stop(paste0("Unknown prediction case kind '", kind, "'."), call. = FALSE)
}

.metafor_linear_predictor <- function(fit_metafor, formula, newdata) {

  x_new <- stats::model.matrix(formula, newdata)
  return(as.vector(x_new %*% fit_metafor[["beta"]]))
}

.newdata_factor <- function(fit_brma, variable, values = NULL) {

  levels_variable <- levels(fit_brma[["data"]][["mods"]][[variable]])
  if (is.null(values)) {
    values <- levels_variable
  }

  return(factor(values, levels = levels_variable))
}

.newdata_prediction_case_data <- function(kind, fit_brma) {

  if (kind == "normal_continuous") {
    return(data.frame(
      ablat = c(10, 40, 70),
      year  = c(1950, 1968, 1980)
    ))
  }

  if (kind == "normal_factor") {
    alloc <- .newdata_factor(
      fit_brma,
      "alloc",
      values = c("random", "alternate", "systematic")
    )
    return(data.frame(alloc = alloc))
  }

  if (kind == "location_scale") {
    meta_levels <- levels(fit_brma[["data"]][["mods"]][["meta"]])
    return(data.frame(
      meta  = factor(c(meta_levels[1], meta_levels[2], meta_levels[1]),
                     levels = meta_levels),
      ni100 = c(0.5, 1.0, 2.0)
    ))
  }

  if (kind == "multilevel_mod") {
    return(data.frame(vi = c(0.02, 0.08, 0.20)))
  }

  if (kind == "glmm_factor") {
    alloc <- .newdata_factor(fit_brma, "alloc")
    return(data.frame(alloc = alloc))
  }

  if (kind == "pet_reg") {
    preregistered <- .newdata_factor(fit_brma, "Preregistered")
    return(data.frame(
      Preregistered = preregistered,
      sei           = c(0.10, 0.20),
      vi            = c(0.01, 0.04)
    ))
  }

  if (kind == "peese_reg") {
    preregistered <- .newdata_factor(fit_brma, "Preregistered")
    return(data.frame(
      Preregistered = preregistered,
      vi            = c(0.01, 0.04)
    ))
  }

  if (kind == "selection_reg") {
    preregistered <- .newdata_factor(fit_brma, "Preregistered")
    return(data.frame(Preregistered = preregistered))
  }

  stop(paste0("Unknown newdata prediction case kind '", kind, "'."), call. = FALSE)
}

.metafor_newdata_location <- function(kind, fit_metafor, newdata,
                                      bias_adjusted = FALSE) {

  if (kind == "normal_continuous") {
    return(.metafor_linear_predictor(fit_metafor, ~ ablat + year, newdata))
  }

  if (kind %in% c("normal_factor", "glmm_factor")) {
    return(.metafor_linear_predictor(fit_metafor, ~ alloc, newdata))
  }

  if (kind == "location_scale") {
    return(.metafor_linear_predictor(fit_metafor, ~ meta + ni100, newdata))
  }

  if (kind == "multilevel_mod") {
    return(.metafor_linear_predictor(fit_metafor, ~ vi, newdata))
  }

  if (kind %in% c("pet_reg", "peese_reg")) {
    metafor_newdata <- newdata
    if (bias_adjusted) {
      metafor_newdata[["vi"]] <- 0
    }
    formula <- if (kind == "pet_reg") {
      ~ sqrt(vi) + Preregistered
    } else {
      ~ vi + Preregistered
    }
    return(.metafor_linear_predictor(
      fit_metafor,
      formula,
      metafor_newdata
    ))
  }

  if (kind == "selection_reg") {
    return(.metafor_linear_predictor(fit_metafor, ~ Preregistered, newdata))
  }

  stop(paste0("Unknown newdata prediction case kind '", kind, "'."), call. = FALSE)
}

.metafor_newdata_tau <- function(kind, fit_metafor, newdata) {

  if (kind == "location_scale") {
    z_new <- stats::model.matrix(~ ni100, newdata)
    return(sqrt(as.vector(exp(z_new %*% fit_metafor[["alpha"]]))))
  }

  return(NULL)
}

expect_newdata_prediction_matches_metafor <- function(case) {

  name        <- case_name(case)
  fit_brma    <- .case_fit(case)
  fit_metafor <- .case_metafor(case)
  kind        <- case_value(case, "kind")
  tolerance   <- case_value(case, "tolerance", 0.05)
  tau_tol     <- case_value(case, "tau_tolerance", tolerance)
  newdata     <- .newdata_prediction_case_data(kind, fit_brma)

  brma_terms <- .sample_means(predict(
    fit_brma,
    newdata = newdata,
    type    = "terms",
    quiet   = TRUE
  ))
  metafor_terms <- .metafor_newdata_location(kind, fit_metafor, newdata)

  testthat::expect_equal(
    brma_terms,
    metafor_terms,
    tolerance = tolerance,
    info      = paste(name, "newdata location predictions match metafor")
  )

  if (kind == "pet_reg") {
    brma_adjusted <- .sample_means(predict(
      fit_brma,
      newdata       = newdata,
      type          = "terms",
      bias_adjusted = TRUE,
      quiet         = TRUE
    ))
    metafor_adjusted <- .metafor_newdata_location(
      kind          = kind,
      fit_metafor   = fit_metafor,
      newdata       = newdata,
      bias_adjusted = TRUE
    )
    testthat::expect_equal(
      brma_adjusted,
      metafor_adjusted,
      tolerance = tolerance,
      info      = paste(name, "bias-adjusted newdata predictions match metafor")
    )
  }

  metafor_tau <- .metafor_newdata_tau(kind, fit_metafor, newdata)
  if (!is.null(metafor_tau)) {
    brma_tau <- .sample_means(predict(
      fit_brma,
      newdata = newdata,
      type    = "terms.scale",
      quiet   = TRUE
    ))
    testthat::expect_equal(
      brma_tau,
      metafor_tau,
      tolerance = tau_tol,
      info      = paste(name, "newdata scale predictions match metafor")
    )
  }
}

expect_hatvalues_match_metafor <- function(case) {

  name        <- case_name(case)
  fit_brma    <- .case_fit(case)
  fit_metafor <- .case_metafor(case)
  tolerance   <- case_value(case, "tolerance", 0.05)

  testthat::expect_equal(
    unname(hatvalues(fit_brma)),
    as.vector(hatvalues(fit_metafor)),
    tolerance = tolerance,
    info      = paste(name, "hatvalues match metafor")
  )
}

brma_term_btt <- function(fit_brma) {

  X      <- RoBMA:::.get_model_matrix(fit_brma)
  assign <- attr(X, "assign")

  if (is.null(assign)) {
    stop("Model matrix is missing the 'assign' attribute.", call. = FALSE)
  }

  term_indices <- sort(unique(assign[assign != 0]))
  btt          <- lapply(term_indices, function(i) which(assign == i))

  return(btt)
}

metafor_vif_value <- function(fit_metafor, btt) {

  out       <- metafor::vif(fit_metafor, btt = btt)
  component <- if (!is.null(out[["beta"]])) out[["beta"]] else out
  row       <- component[["vif"]][[1]]

  return(c(
    GVIF = unname(row[["vif"]]),
    GSIF = unname(row[["sif"]])
  ))
}

metafor_vif_table <- function(fit_brma, fit_metafor, btt = NULL) {

  brma_vif <- vif(fit_brma, posterior_correlation = FALSE)[["vif"]]

  if (is.null(btt)) {
    btt <- brma_term_btt(fit_brma)
  }

  expected <- do.call(rbind, lapply(btt, function(x) metafor_vif_value(fit_metafor, x)))

  return(data.frame(
    term              = brma_vif[["term"]],
    df                = brma_vif[["df"]],
    GVIF              = unname(expected[, "GVIF"]),
    "GVIF^(1/(2*df))" = unname(expected[, "GSIF"]),
    stringsAsFactors  = FALSE,
    check.names       = FALSE
  ))
}

expect_vif_matches_metafor <- function(case) {

  name        <- case_name(case)
  fit_brma    <- .case_fit(case)
  fit_metafor <- .case_metafor(case)
  tolerance   <- case_value(case, "tolerance", 0.10)
  btt         <- case_value(case, "btt", NULL)

  brma_vif    <- vif(fit_brma, posterior_correlation = FALSE)[["vif"]]
  metafor_vif <- metafor_vif_table(fit_brma, fit_metafor, btt = btt)

  expect_vif_table(brma_vif, nrow(metafor_vif), info = name)
  testthat::expect_equal(
    brma_vif[["df"]],
    metafor_vif[["df"]],
    info = paste(name, "VIF degrees of freedom match metafor")
  )
  testthat::expect_equal(
    brma_vif[["GVIF"]],
    metafor_vif[["GVIF"]],
    tolerance = tolerance,
    info      = paste(name, "GVIF matches metafor")
  )
  testthat::expect_equal(
    brma_vif[["GVIF^(1/(2*df))"]],
    metafor_vif[["GVIF^(1/(2*df))"]],
    tolerance = tolerance,
    info      = paste(name, "adjusted GVIF matches metafor")
  )
}

expect_dfbetas_match_metafor <- function(case) {

  name        <- case_name(case)
  fit_brma    <- .case_fit(case)
  fit_metafor <- .case_metafor(case)
  oracle      <- case_value(case, "oracle", "equal")
  tolerance   <- case_value(case, "tolerance", 0.10)

  brma_dfbetas <- dfbetas(fit_brma)
  if (oracle == "structure") {
    expect_dfbetas_table(brma_dfbetas, nobs(fit_brma), info = name)
    return(invisible(TRUE))
  }

  metafor_dfbetas <- .metafor_dfbetas(fit_metafor)
  skip_rows       <- case_value(case, "skip_rows", integer())
  metafor_cols    <- case_value(case, "metafor_cols", NULL)
  brma_cols       <- case_value(case, "brma_cols", NULL)

  if (is.null(metafor_cols)) {
    metafor_cols <- seq_len(ncol(metafor_dfbetas))
  }
  if (is.null(brma_cols)) {
    brma_cols <- metafor_cols
  }

  rows <- seq_len(nrow(metafor_dfbetas))
  rows <- setdiff(rows, skip_rows)

  for (i in seq_along(metafor_cols)) {
    testthat::expect_equal(
      metafor_dfbetas[rows, metafor_cols[[i]]],
      brma_dfbetas[rows, brma_cols[[i]]],
      tolerance = tolerance,
      info      = paste(name, "DFBETAS match metafor")
    )
  }
}

expect_summary_heterogeneity_matches_metafor <- function(case) {

  name        <- case_name(case)
  fit_brma    <- .case_fit(case)
  fit_metafor <- .case_metafor(case)
  brma_het    <- summary_heterogeneity(fit_brma)
  kind        <- case_value(case, "kind", "standard")
  tolerance   <- case_value(case, "tolerance", 0.05)
  i2_tol      <- case_value(case, "i2_tolerance", tolerance)
  h2_tol      <- case_value(case, "h2_tolerance", tolerance)

  if (kind == "multilevel") {
    testthat::expect_equal(
      brma_het$estimates["tau", "Mean"],
      sqrt(fit_metafor$tau2),
      tolerance = tolerance,
      info      = paste(name, "total tau matches metafor")
    )
    testthat::expect_equal(
      brma_het$estimates["tau [within]", "Mean"],
      sqrt(fit_metafor$tau2 * (1 - fit_metafor$rho)),
      tolerance = tolerance,
      info      = paste(name, "within tau matches metafor")
    )
    testthat::expect_equal(
      brma_het$estimates["tau [between]", "Mean"],
      sqrt(fit_metafor$tau2 * fit_metafor$rho),
      tolerance = tolerance,
      info      = paste(name, "between tau matches metafor")
    )
    testthat::expect_equal(
      brma_het$estimates["rho", "Mean"],
      fit_metafor$rho,
      tolerance = tolerance,
      info      = paste(name, "rho matches metafor")
    )

    W         <- diag(1 / fit_metafor$vi)
    X         <- model.matrix(fit_metafor)
    P         <- W - W %*% X %*% solve(t(X) %*% W %*% X) %*% t(X) %*% W
    typical_v <- (fit_metafor$k - fit_metafor$p) / sum(diag(P))
    total_var <- fit_metafor$tau2

    i2_total   <- 100 * total_var / (total_var + typical_v)
    i2_within  <- 100 * (total_var * (1 - fit_metafor$rho)) / (total_var + typical_v)
    i2_between <- 100 * (total_var * fit_metafor$rho) / (total_var + typical_v)

    testthat::expect_equal(brma_het$estimates["I2", "Median"], i2_total,
                           tolerance = 0.05, info = paste(name, "I2 total"))
    testthat::expect_equal(brma_het$estimates["I2 [within]", "Median"], i2_within,
                           tolerance = 0.10, info = paste(name, "I2 within"))
    testthat::expect_equal(brma_het$estimates["I2 [between]", "Median"], i2_between,
                           tolerance = 0.10, info = paste(name, "I2 between"))
    return(invisible(TRUE))
  }

  if (kind == "scale") {
    testthat::expect_equal(brma_het$estimates["I2", "Median"], mean(fit_metafor$I2),
                           tolerance = i2_tol, info = paste(name, "I2"))
    testthat::expect_equal(brma_het$estimates["H2", "Median"], mean(fit_metafor$H2),
                           tolerance = h2_tol, info = paste(name, "H2"))
    return(invisible(TRUE))
  }

  testthat::expect_equal(brma_het$estimates["tau", "Mean"], sqrt(fit_metafor$tau2),
                         tolerance = tolerance, info = paste(name, "tau"))
  testthat::expect_equal(brma_het$estimates["tau2", "Mean"], fit_metafor$tau2,
                         tolerance = tolerance, info = paste(name, "tau2"))

  if (kind != "selection") {
    testthat::expect_equal(brma_het$estimates["I2", "Median"], fit_metafor$I2,
                           tolerance = i2_tol, info = paste(name, "I2"))
    testthat::expect_equal(brma_het$estimates["H2", "Median"], fit_metafor$H2,
                           tolerance = h2_tol, info = paste(name, "H2"))
  }
}

expect_influence_matches_metafor <- function(case) {

  name        <- case_name(case)
  fit_brma    <- .case_fit(case)
  fit_metafor <- .case_metafor(case)
  oracle      <- case_value(case, "oracle", "equal")

  if (oracle == "cooks_finite") {
    testthat::expect_true(
      all(is.finite(suppressWarnings(cooks.distance(fit_brma)))),
      info = paste(name, "cooks.distance is finite")
    )
    return(invisible(TRUE))
  }

  inf_metafor <- influence(fit_metafor)
  inf_brma    <- suppressWarnings(influence(fit_brma))

  if (oracle == "rank") {
    testthat::expect_true(
      stats::cor(inf_metafor$inf$rstudent, inf_brma$inf$rstudent,
                 method = "spearman", use = "pairwise") > 0.9,
      info = paste(name, "rstudent rank matches metafor")
    )
    testthat::expect_true(all(is.finite(inf_brma$inf$dffits)),
                          info = paste(name, "dffits finite"))
    testthat::expect_true(all(is.finite(inf_brma$inf$cook.d)),
                          info = paste(name, "cook.d finite"))
    testthat::expect_true(
      stats::cor(inf_metafor$inf$cov.r, inf_brma$inf$cov.r,
                 method = "spearman", use = "pairwise") > 0.9,
      info = paste(name, "cov.r rank matches metafor")
    )
    testthat::expect_true(
      stats::cor(inf_metafor$inf$tau2.del, inf_brma$inf$tau.del,
                 method = "spearman", use = "pairwise") > 0.9,
      info = paste(name, "tau.del rank matches metafor")
    )
    testthat::expect_true(
      stats::cor(inf_metafor$inf$hat, inf_brma$inf$hat,
                 method = "spearman", use = "pairwise") > 0.9,
      info = paste(name, "hat rank matches metafor")
    )
    return(invisible(TRUE))
  }

  skip_rows <- case_value(case, "skip_rows", integer())
  rows      <- setdiff(seq_len(nrow(inf_brma$inf)), skip_rows)
  tol       <- if (oracle == "equal") 0.05 else 0.10

  testthat::expect_equal(inf_metafor$inf$rstudent[rows], inf_brma$inf$rstudent[rows],
                         tolerance = if (oracle == "equal") 0.05 else 0.15,
                         info = paste(name, "rstudent"))
  testthat::expect_true(all(is.finite(inf_brma$inf$dffits[rows])),
                        info = paste(name, "dffits finite"))
  testthat::expect_true(all(is.finite(inf_brma$inf$cook.d[rows])),
                        info = paste(name, "cook.d finite"))

  if (oracle == "equal") {
    testthat::expect_equal(inf_metafor$inf$cov.r[rows], inf_brma$inf$cov.r[rows],
                           tolerance = tol, info = paste(name, "cov.r"))
  }
  testthat::expect_equal(inf_metafor$inf$tau2.del[rows], inf_brma$inf$tau.del[rows]^2,
                         tolerance = tol, info = paste(name, "tau2.del"))
  testthat::expect_equal(inf_metafor$inf$hat[rows], inf_brma$inf$hat[rows],
                         tolerance = tol, info = paste(name, "hat"))

  if (oracle == "equal") {
    testthat::expect_equal(inf_metafor$dfbs$intrcpt, inf_brma$dfbs$mu,
                           tolerance = 0.10, info = paste(name, "dfbetas"))
    testthat::expect_equal(inf_brma$inf$dffits, unname(dffits(fit_brma)),
                           tolerance = 1e-12)
    testthat::expect_equal(inf_brma$inf$cook.d, unname(cooks.distance(fit_brma)),
                           tolerance = 1e-12)
    testthat::expect_equal(inf_metafor$inf$cov.r, unname(covratio(fit_brma)),
                           tolerance = 0.05)
  }
}

context("Predict and wrappers")

source(testthat::test_path("common-functions.R"))
source(testthat::test_path("helper-contracts.R"))
source(testthat::test_path("helper-test-matrix.R"))
source(testthat::test_path("helper-metafor.R"))

REFERENCE_DIR <<- testthat::test_path("..", "results", "predict")

skip_if_no_fits()
skip_if_not_installed("metafor")

fit_names <- list_fits()
fits      <- lazy_fits(fit_names, validate = FALSE)
info      <- lazy_infos(fit_names, validate = FALSE)

for_each_case(prediction_metafor_cases(), function(case) {
  test_that_case("Predictions match metafor", case, {
    expect_prediction_matches_metafor(case)
  })
})

for_each_case(prediction_newdata_metafor_cases(), function(case) {
  test_that_case("Newdata predictions match metafor", case, {
    expect_newdata_prediction_matches_metafor(case)
  })
})

test_that("Predictions for equivalent interaction parameterizations match", {

  model_names <- c("bcg_meta-regression4", "bcg_meta-regression4b")
  skip_if_missing_fits(model_names)

  fit_brma1 <- fits[["bcg_meta-regression4"]]
  fit_brma2 <- fits[["bcg_meta-regression4b"]]

  expect_equal(
    .sample_means(predict(fit_brma1, type = "terms")),
    .sample_means(predict(fit_brma2, type = "terms")),
    tolerance = 0.15,
    info      = "per-study predictions match"
  )
  expect_equal(
    .sample_mean(pooled_effect(fit_brma1), "mu"),
    .sample_mean(pooled_effect(fit_brma2), "mu"),
    tolerance = 0.05,
    info      = "pooled effects match"
  )
  expect_equal(
    .sample_means(blup(fit_brma1)),
    .sample_means(blup(fit_brma2)),
    tolerance = 0.05,
    info      = "BLUPs match"
  )
})

test_that("Newdata prediction preserves duplicate rows and rejects novel factor levels", {

  skip_if_missing_fits(c("bcg_meta-regression", "bcg_meta-regression2"))

  fit_cont <- fits[["bcg_meta-regression"]]
  newdata_cont <- data.frame(
    ablat = c(10, 10, 70),
    year  = c(1950, 1950, 1980)
  )
  pred_cont <- predict(
    fit_cont,
    newdata = newdata_cont,
    type    = "terms",
    quiet   = TRUE
  )
  pred_single <- predict(
    fit_cont,
    newdata = newdata_cont[3, , drop = FALSE],
    type    = "terms",
    quiet   = TRUE
  )

  expect_equal(unname(pred_cont[, 1]), unname(pred_cont[, 2]),
               info = "duplicated newdata rows produce identical predictions")
  expect_equal(unname(pred_cont[, 3]), unname(pred_single[, 1]),
               info = "newdata row order is preserved")

  fit_factor <- fits[["bcg_meta-regression2"]]
  old_levels <- levels(fit_factor[["data"]][["mods"]][["alloc"]])
  newdata_factor <- data.frame(
    alloc = factor("not-in-training", levels = c(old_levels, "not-in-training"))
  )

  expect_error(
    predict(fit_factor, newdata = newdata_factor, type = "terms", quiet = TRUE),
    info = "novel factor levels are rejected"
  )
})

test_that("Predictions can use an already thinned posterior", {

  name <- "bcg_meta-regression"
  skip_if_missing_fits(name)

  fit_brma          <- fits[[name]]
  posterior_samples <- .get_posterior_samples(fit_brma[["fit"]])
  selected_rows     <- unique(round(seq(
    from       = 1,
    to         = nrow(posterior_samples),
    length.out = min(1000L, nrow(posterior_samples))
  )))

  full_prediction <- as.matrix(predict(
    fit_brma,
    type  = "terms",
    quiet = TRUE
  ))
  thinned_prediction <- as.matrix(predict(
    fit_brma,
    type               = "terms",
    quiet              = TRUE,
    .posterior_samples = posterior_samples[selected_rows, , drop = FALSE]
  ))

  expect_equal(
    thinned_prediction,
    full_prediction[selected_rows, , drop = FALSE],
    tolerance = 1e-12,
    info      = "formula-based predictions use the caller-supplied posterior rows"
  )
})

test_that("Wrapper functions have correct interface", {

  name <- "bcg_meta-analysis"
  skip_if_missing_fits(name)

  fit_brma <- fits[[name]]

  samples_effect <- pooled_effect(fit_brma)
  expect_brma_samples_matrix(samples_effect, 1, "pooled_effect")

  samples_het <- pooled_heterogeneity(fit_brma)
  expect_s3_class(samples_het, "brma_samples")
  expect_true(is.matrix(samples_het), info = "pooled_heterogeneity matrix")

  samples_blup <- blup(fit_brma)
  expect_brma_samples_matrix(samples_blup, nrow(fit_brma$data$outcome), "blup")

  pred_90ci <- pooled_effect(fit_brma, probs = c(0.05, 0.95))
  ci_names  <- colnames(summary(pred_90ci))
  expect_true(any(grepl("5%|0\\.05", ci_names)) &&
                any(grepl("95%|0\\.95", ci_names)),
              info = "probs argument controls CI quantiles")

  expect_equal(summary(blup(fit_brma)), summary(true_effects(fit_brma)),
               info = "true_effects are identical to blup")
})

test_that("fitted returns in-sample posterior means", {

  model_names <- c(
    "bcg_meta-regression",
    "bangertdrowns2004_location-scale",
    "konstantopoulos2011_3lvl",
    "dat.lehmann2018_RoBMA"
  )
  skip_if_missing_fits(model_names)

  fit_reg <- fits[["bcg_meta-regression"]]
  n_reg   <- nobs(fit_reg)

  fitted_reg <- fitted(fit_reg)
  terms_reg  <- predict(fit_reg, type = "terms", quiet = TRUE)
  expect_finite_vector(fitted_reg, n_reg, "fitted vector")
  expect_equal(
    unname(fitted_reg),
    unname(.sample_means(terms_reg)),
    tolerance = 1e-12,
    info      = "default fitted values wrap marginal predictions"
  )
  expect_equal(
    unname(fitted_reg),
    unname(stats::fitted(info[["bcg_meta-regression"]][["metafor"]])),
    tolerance = 0.05,
    info      = "default fitted values match metafor"
  )
  expect_equal(
    unname(residuals(fit_reg)),
    unname(.outcome_data_yi(fit_reg) - fitted_reg),
    tolerance = 1e-12,
    info      = "default residuals use default fitted values"
  )

  fit_3lvl <- fits[["konstantopoulos2011_3lvl"]]
  expect_equal(
    unname(fitted(fit_3lvl, conditioning_depth = "cluster")),
    unname(.sample_means(predict(fit_3lvl, type = "cluster", quiet = TRUE))),
    tolerance = 1e-12,
    info      = "cluster fitted values wrap cluster predictions"
  )
  expect_equal(
    unname(fitted(fit_3lvl, conditioning_depth = "estimate")),
    unname(.sample_means(blup(fit_3lvl))),
    tolerance = 1e-12,
    info      = "estimate fitted values wrap BLUPs"
  )

  fit_scale <- fits[["bangertdrowns2004_location-scale"]]
  fitted_scale <- fitted(fit_scale, component = "scale")
  expect_equal(
    unname(fitted_scale),
    unname(.sample_means(predict(fit_scale, type = "terms.scale", quiet = TRUE))),
    tolerance = 1e-12,
    info      = "scale fitted values wrap scale predictions"
  )
  fitted_all <- fitted(fit_scale, component = "all")
  expect_equal(names(fitted_all), c("location", "scale"),
               info = "component all returns location and scale")
  expect_equal(unname(fitted_all[["location"]]), unname(fitted(fit_scale)),
               tolerance = 1e-12)
  expect_equal(unname(fitted_all[["scale"]]), unname(fitted_scale),
               tolerance = 1e-12)

  fit_robma <- fits[["dat.lehmann2018_RoBMA"]]
  expect_equal(
    unname(fitted(fit_robma, bias_adjusted = TRUE, conditional = TRUE)),
    unname(.sample_means(predict(
      fit_robma,
      type          = "terms",
      bias_adjusted = TRUE,
      conditional   = TRUE,
      quiet         = TRUE
    ))),
    tolerance = 1e-12,
    info      = "conditional fitted values wrap conditional predictions"
  )

  expect_error(fitted(fit_reg, unit = "cluster"), "multilevel models")
  expect_error(
    fitted(fit_scale, component = "scale", conditioning_depth = "estimate"),
    "component = 'scale'"
  )
  expect_error(
    fitted(fit_scale, component = "scale", transform = "EXP"),
    "location fitted values"
  )
})

test_that("Conditional pooled wrappers condition RoBMA draws", {

  product_names <- c("dat.lehmann2018_RoBMA", "dat.lehmann2018_RoBMA_mods")
  skip_if_missing_fits(c(product_names, "bcg_meta-analysis"))

  for (name in product_names) {
    fit_brma <- fits[[name]]

    effect_rows <- .conditional_parameter_rows(
      object     = fit_brma,
      parameters = .conditional_effect_parameters(fit_brma)
    )
    pooled_effect_averaged <- pooled_effect(fit_brma)
    pooled_effect_cond     <- expect_silent(
      pooled_effect(fit_brma, conditional = TRUE)
    )
    pooled_effect_predict  <- expect_silent(
      predict(
        fit_brma,
        newdata       = TRUE,
        type          = "terms",
        bias_adjusted = TRUE,
        conditional   = TRUE,
        quiet         = TRUE
      )
    )
    expect_message(
      predict(
        fit_brma,
        newdata       = TRUE,
        type          = "terms",
        bias_adjusted = TRUE,
        conditional   = TRUE,
        quiet         = FALSE
      ),
      "flattened"
    )

    expect_equal(
      unname(as.matrix(pooled_effect_cond)),
      unname(as.matrix(pooled_effect_averaged)[effect_rows, , drop = FALSE]),
      info = paste(name, "pooled_effect conditional rows")
    )
    expect_equal(
      unname(as.matrix(pooled_effect_cond)),
      unname(as.matrix(pooled_effect_predict)),
      info = paste(name, "pooled_effect matches conditional predict")
    )
    expect_match(attr(pooled_effect_cond, "title"), "Conditional")
  }

  fit_brma       <- fits[["dat.lehmann2018_RoBMA"]]
  summary_brma   <- summary(fit_brma, conditional = TRUE)
  summary_effect <- summary(suppressMessages(pooled_effect(fit_brma, conditional = TRUE)))
  expect_equal(
    unname(summary_effect["mu", "Mean"]),
    unname(summary_brma[["estimates_conditional"]]["mu", "Mean"]),
    info = "conditional pooled_effect matches conditional summary mu"
  )

  heterogeneity_rows <- .conditional_parameter_rows(
    object     = fit_brma,
    parameters = .conditional_heterogeneity_parameters(fit_brma)
  )
  pooled_het_averaged <- pooled_heterogeneity(fit_brma)
  pooled_het_cond     <- expect_silent(
    pooled_heterogeneity(fit_brma, conditional = TRUE)
  )
  pooled_het_predict  <- expect_silent(
    predict(
      fit_brma,
      newdata     = TRUE,
      type        = "terms.scale",
      conditional = TRUE,
      quiet       = TRUE
    )
  )
  expect_message(
    predict(
      fit_brma,
      newdata     = TRUE,
      type        = "terms.scale",
      conditional = TRUE,
      quiet       = FALSE
    ),
    "flattened"
  )

  expect_equal(
    unname(as.matrix(pooled_het_cond)),
    unname(as.matrix(pooled_het_averaged)[heterogeneity_rows, , drop = FALSE]),
    info = "pooled_heterogeneity conditional rows"
  )
  expect_equal(
    unname(as.matrix(pooled_het_cond)),
    unname(as.matrix(pooled_het_predict)),
    info = "pooled_heterogeneity matches conditional predict"
  )
  expect_equal(
    unname(summary(pooled_het_cond)["tau", "Mean"]),
    unname(summary_brma[["estimates_conditional"]]["tau", "Mean"]),
    info = "conditional pooled_heterogeneity matches conditional summary tau"
  )
  expect_match(attr(pooled_het_cond, "title"), "Conditional")

  expect_error(pooled_effect(fits[["bcg_meta-analysis"]], conditional = TRUE),
               "RoBMA objects")
  expect_error(pooled_heterogeneity(fits[["bcg_meta-analysis"]], conditional = TRUE),
               "RoBMA objects")
})

test_that("Model-averaged predictions cover BMA.norm, BMA.glmm, and RoBMA", {

  product_names <- c(
    "dat.lehmann2018_BMA.norm",
    "bcg_BMA.glmm",
    "nielweise2008_BMA.glmm",
    "dat.lehmann2018_RoBMA"
  )
  skip_if_missing_fits(product_names)

  for (name in product_names) {
    fit_brma  <- fits[[name]]
    n_studies <- nobs(fit_brma)

    terms  <- predict(fit_brma, type = "terms")
    scale  <- predict(fit_brma, type = "terms.scale")
    effect <- predict(fit_brma, type = "effect")
    set.seed(1)
    response <- predict(fit_brma, type = "response")

    expect_brma_samples_matrix(terms, n_studies, paste(name, "terms"))
    expect_brma_samples_matrix(scale, n_studies, paste(name, "terms.scale"))
    expect_brma_samples_matrix(effect, n_studies, paste(name, "effect"))
    expect_brma_samples_matrix(response, n_studies, paste(name, "response"))

    pooled <- pooled_effect(fit_brma)
    pooled_predict <- predict(
      fit_brma,
      newdata       = TRUE,
      type          = "terms",
      bias_adjusted = TRUE,
      quiet         = TRUE
    )
    expect_brma_samples_matrix(pooled, 1, paste(name, "pooled_effect"))
    expect_equal(unname(as.matrix(pooled)), unname(as.matrix(pooled_predict)))

    pooled_het <- pooled_heterogeneity(fit_brma)
    pooled_het_predict <- predict(
      fit_brma,
      newdata = TRUE,
      type    = "terms.scale",
      quiet   = TRUE
    )
    expect_brma_samples_matrix(pooled_het, 1, paste(name, "pooled_heterogeneity"))
    expect_equal(unname(as.matrix(pooled_het)), unname(as.matrix(pooled_het_predict)))

    blup_samples <- blup(fit_brma)
    true_samples <- true_effects(fit_brma)
    expect_brma_samples_matrix(blup_samples, n_studies, paste(name, "blup"))
    expect_brma_samples_matrix(true_samples, n_studies, paste(name, "true_effects"))
    expect_equal(unname(as.matrix(blup_samples)), unname(as.matrix(true_samples)))

    ranef_samples <- ranef(fit_brma)
    expect_brma_samples_matrix(ranef_samples, n_studies, paste(name, "ranef"))
    expect_equal(
      unname(as.matrix(ranef_samples)),
      unname(as.matrix(blup_samples) - as.matrix(terms))
    )
  }
})

test_that("Model-averaged multilevel ranef decomposes cluster and estimate effects", {

  product_names <- c(
    "bcg_BMA.glmm_3lvl_location_scale",
    "dat.lehmann2018_RoBMA_3lvl_mods_scale"
  )
  skip_if_missing_fits(product_names)

  for (name in product_names) {
    fit_brma  <- fits[[name]]
    n_studies <- nobs(fit_brma)
    out       <- ranef(fit_brma)

    expect_type(out, "list")
    expect_equal(names(out), c("cluster", "estimate"), info = name)
    expect_brma_samples_matrix(out[["cluster"]], n_studies, paste(name, "cluster ranef"))
    expect_brma_samples_matrix(out[["estimate"]], n_studies, paste(name, "estimate ranef"))
  }
})

test_that("Wrappers suppress aggregation messages", {

  name <- "bcg_meta-regression"
  skip_if_missing_fits(name)

  fit_brma <- fits[[name]]
  expect_silent(pooled_effect(fit_brma))
  expect_silent(pooled_heterogeneity(fit_brma))
  expect_silent(blup(fit_brma))
  expect_silent(true_effects(fit_brma))
})

# ============================================================================ #
# CONFIGURATION: Set to TRUE to regenerate reference files, FALSE to run tests
# ============================================================================ #
if (!exists("GENERATE_REFERENCE_FILES")) {
  GENERATE_REFERENCE_FILES <- FALSE
}
if (!exists("FIT_CACHE_VERSION")) {
  FIT_CACHE_VERSION <- 3L
}

.common_functions_dir <- function() {

  frames <- sys.frames()
  ofiles <- vapply(frames, function(frame) {
    ofile <- frame[["ofile"]]
    if (is.null(ofile)) "" else ofile
  }, character(1))
  ofiles <- ofiles[nzchar(ofiles)]

  if (length(ofiles) > 0L) {
    return(dirname(normalizePath(ofiles[length(ofiles)], winslash = "/", mustWork = FALSE)))
  }

  fallback <- file.path("tests", "testthat", "common-functions.R")
  if (file.exists(fallback)) {
    return(dirname(normalizePath(fallback, winslash = "/", mustWork = TRUE)))
  }

  return(normalizePath(getwd(), winslash = "/", mustWork = FALSE))
}

# Get the directory where prefitted models are stored. Local development uses
# a persistent ignored folder; CRAN checks fall back to tempdir().
test_files_dir <- Sys.getenv("ROBMA_TEST_FILES_DIR")
if (test_files_dir == "") {
  on_cran <- get("on_cran", envir = asNamespace("testthat"), inherits = FALSE)
  test_files_dir <- if (on_cran()) {
    file.path(tempdir(), "RoBMA_test_files")
  } else {
    file.path(.common_functions_dir(), "test_files")
  }
}
test_files_dir <- normalizePath(test_files_dir, winslash = "/", mustWork = FALSE)

# Setup directory for saving fitted models
temp_fits_dir     <- file.path(test_files_dir, "fits")
temp_info_dir     <- file.path(test_files_dir, "info")
temp_metadata_dir <- file.path(test_files_dir, "metadata")
temp_temp_dir     <- file.path(test_files_dir, "temp")
if (!dir.exists(test_files_dir)) dir.create(test_files_dir, showWarnings = FALSE, recursive = TRUE)
if (!dir.exists(temp_fits_dir)) dir.create(temp_fits_dir, showWarnings = FALSE, recursive = TRUE)
if (!dir.exists(temp_info_dir)) dir.create(temp_info_dir, showWarnings = FALSE, recursive = TRUE)
if (!dir.exists(temp_metadata_dir)) dir.create(temp_metadata_dir, showWarnings = FALSE, recursive = TRUE)
if (!dir.exists(temp_temp_dir)) dir.create(temp_temp_dir, showWarnings = FALSE, recursive = TRUE)

# Set environment variable so other test files can locate pre-fitted models
Sys.setenv(ROBMA_TEST_FILES_DIR = test_files_dir)

# Use skip_if_no_fits() for tests that need pre-fitted models.

# ============================================================================ #
# HELPER FUNCTIONS: Reference File Testing
# ============================================================================ #

# Process reference file: save if GENERATE_REFERENCE_FILES=TRUE, test otherwise
test_reference_table <- function(table, filename, info_msg = NULL,
                                 print_dir = REFERENCE_DIR) {
  if (GENERATE_REFERENCE_FILES) {
    # Save mode
    if (!dir.exists(print_dir)) {
      dir.create(print_dir, recursive = TRUE)
    }
    writeLines(
      capture_output_lines(table, print = TRUE, width = 150),
      file.path(print_dir, filename)
    )
  } else {
    # Test mode
    ref_file <- file.path(print_dir, filename)
    if (file.exists(ref_file)) {
      expected_output <- readLines(ref_file, warn = FALSE)
      actual_output <- capture_output_lines(table, print = TRUE, width = 150)
      expect_equal(actual_output, expected_output, info = info_msg)
    } else {
      skip(paste("Reference file", filename, "not found."))
    }
  }
}

test_reference_text <- function(text, filename, info_msg = NULL,
                                print_dir = REFERENCE_DIR) {
  if (GENERATE_REFERENCE_FILES) {
    # Save mode
    if (!dir.exists(print_dir)) {
      dir.create(print_dir, recursive = TRUE)
    }
    writeLines(text, file.path(print_dir, filename))
  } else {
    # Test mode
    ref_file <- file.path(print_dir, filename)
    if (file.exists(ref_file)) {
      expected_output <- readLines(ref_file, warn = FALSE)
      expected_output <- paste0(expected_output, collapse = "\n")
      expect_equal(text, expected_output, info = info_msg)
    } else {
      skip(paste("Reference file", filename, "not found."))
    }
  }
}

vdiffr_snapshots_available <- function() {

  snap_dir <- testthat::test_path("_snaps")

  return(
    dir.exists(snap_dir) &&
      length(list.files(snap_dir, pattern = "\\.svg$", recursive = TRUE)) > 0L
  )
}

skip_if_no_vdiffr_snapshots <- function() {

  if (!vdiffr_snapshots_available() &&
      !is_true_env("ROBMA_TEST_ALLOW_MISSING_SNAPSHOTS")) {
    testthat::skip(paste(
      "vdiffr snapshots are not available.",
      "`tests/testthat/_snaps` is intentionally excluded from source builds.",
      "Set ROBMA_TEST_ALLOW_MISSING_SNAPSHOTS=TRUE only when regenerating snapshots."
    ))
  }

  return(invisible(FALSE))
}

# ============================================================================ #
# HELPER FUNCTIONS: Cached Fit Catalog
# ============================================================================ #

fit_catalog <- function() {

  catalog <- data.frame(
    name = c(
      "bcg_meta-analysis",
      "bcg_meta-regression",
      "bcg_meta-regression2",
      "bcg_meta-regression2b",
      "bcg_meta-regression3",
      "bcg_meta-regression3b",
      "bcg_meta-regression4",
      "bcg_meta-regression4b",
      "bangertdrowns2004_location-scale",
      "konstantopoulos2011_3lvl",
      "konstantopoulos2011_3lvl2",
      "bcg_glmm",
      "bcg_glmm_reg",
      "nielweise2008_glmm",
      "bcg_glmm_3lvl_scale",
      "dat.lehmann2018-PET",
      "dat.lehmann2018-PET_neg",
      "dat.lehmann2018-PETreg",
      "dat.lehmann2018-PEESE",
      "dat.lehmann2018-PEESE_neg",
      "dat.lehmann2018-PEESEreg",
      "dat.lehmann2018-3PSM",
      "dat.lehmann2018-4PSM",
      "dat.lehmann2018-3PSM_neg",
      "dat.lehmann2018-3PSMreg",
      "dat.lehmann2018_BMA.norm",
      "dat.lehmann2018_BMA.norm_custom",
      "dat.lehmann2018_BMA.norm_mods",
      "dat.lehmann2018_BMA.norm_scale",
      "bcg_BMA.glmm",
      "nielweise2008_BMA.glmm",
      "bcg_BMA.glmm_custom",
      "bcg_BMA.glmm_3lvl_location_scale",
      "dat.lehmann2018_RoBMA",
      "dat.lehmann2018_RoBMA_custom",
      "dat.lehmann2018_RoBMA_mods",
      "dat.lehmann2018_RoBMA_mods2",
      "dat.lehmann2018_RoBMA_3lvl_mods_scale"
    ),
    class = c(
      rep("brma.norm", 11),
      rep("brma.glmm", 4),
      rep("bPET", 3),
      rep("bPEESE", 3),
      rep("bselmodel", 4),
      rep("BMA.norm", 4),
      rep("BMA.glmm", 4),
      rep("RoBMA", 5)
    ),
    family = c(
      rep("norm", 11),
      rep("glmm", 4),
      rep("norm", 3),
      rep("norm", 3),
      rep("norm", 4),
      rep("norm", 4),
      rep("glmm", 4),
      rep("norm", 5)
    ),
    source_file = c(
      rep("test-01-brma.norm.R", 11),
      rep("test-01-brma.glmm.R", 4),
      rep("test-01-bPET.R", 3),
      rep("test-01-bPEESE.R", 3),
      rep("test-01-bselmodel.R", 4),
      rep("test-01-BMA.norm.R", 4),
      rep("test-01-BMA.glmm.R", 4),
      rep("test-01-RoBMA.R", 5)
    ),
    has_metafor = c(
      TRUE, TRUE, TRUE, FALSE, TRUE, FALSE, TRUE, FALSE, TRUE, TRUE, TRUE,
      TRUE, TRUE, TRUE, FALSE,
      TRUE, TRUE, TRUE,
      TRUE, TRUE, TRUE,
      TRUE, TRUE, TRUE, TRUE,
      rep(FALSE, 13)
    ),
    has_loo = TRUE,
    has_waic = FALSE,
    has_marglik = c(
      rep(TRUE, 25),
      rep(FALSE, 13)
    ),
    tier = c(
      "core", "core", "extended", "extended", "core", "extended", "extended", "extended",
      "core", "core", "extended",
      "core", "core", "core", "extended",
      "core", "extended", "extended",
      "core", "extended", "core",
      "core", "extended", "extended", "extended",
      "core", "extended", "core", "extended",
      "core", "core", "extended", "extended",
      "core", "extended", "core", "extended", "extended"
    ),
    stringsAsFactors = FALSE
  )

  catalog[["features"]] <- I(list(
    c("metafor", "normal", "simple"),
    c("metafor", "normal", "mods"),
    c("metafor", "normal", "factor_mods"),
    c("normal", "factor_mods", "meandif"),
    c("metafor", "normal", "interaction"),
    c("normal", "interaction", "meandif"),
    c("metafor", "normal", "interaction"),
    c("normal", "interaction", "meandif"),
    c("metafor", "normal", "mods", "scale"),
    c("metafor", "normal", "multilevel"),
    c("metafor", "normal", "multilevel", "mods"),
    c("metafor", "glmm", "binomial"),
    c("metafor", "glmm", "binomial", "mods"),
    c("metafor", "glmm", "poisson"),
    c("glmm", "binomial", "multilevel", "scale"),
    c("metafor", "PET"),
    c("metafor", "PET", "negative"),
    c("metafor", "PET", "mods"),
    c("metafor", "PEESE"),
    c("metafor", "PEESE", "negative"),
    c("metafor", "PEESE", "mods"),
    c("metafor", "selection"),
    c("metafor", "selection", "steps"),
    c("metafor", "selection", "negative"),
    c("metafor", "selection", "mods"),
    c("BMA.norm", "normal", "simple"),
    c("BMA.norm", "normal", "custom_priors"),
    c("BMA.norm", "normal", "mods"),
    c("BMA.norm", "normal", "scale"),
    c("BMA.glmm", "glmm", "binomial"),
    c("BMA.glmm", "glmm", "poisson"),
    c("BMA.glmm", "glmm", "binomial", "custom_priors"),
    c("BMA.glmm", "glmm", "binomial", "multilevel", "mods", "scale"),
    c("RoBMA", "normal", "simple", "selection", "PET"),
    c("RoBMA", "normal", "custom_priors"),
    c("RoBMA", "normal", "mods"),
    c("RoBMA", "normal", "interaction"),
    c("RoBMA", "normal", "multilevel", "mods", "scale")
  ))

  return(catalog)
}

fit_catalog_entry <- function(name) {

  catalog <- fit_catalog()
  index   <- match(name, catalog[["name"]])

  if (is.na(index)) {
    return(NULL)
  }

  return(catalog[index, , drop = FALSE])
}

catalog_group_fits <- function(name) {

  catalog <- fit_catalog()

  if (name %in% catalog[["name"]]) {
    return(name)
  }

  if (name %in% catalog[["class"]]) {
    return(catalog[catalog[["class"]] == name, "name"])
  }

  source_file <- paste0("test-01-", name, ".R")
  if (source_file %in% catalog[["source_file"]]) {
    return(catalog[catalog[["source_file"]] == source_file, "name"])
  }

  return(character())
}

catalog_fits <- function(feature, class, family, has_metafor, has_loo,
                         has_waic, has_marglik, tier) {

  catalog <- fit_catalog()

  if (!missing(feature) && !is.null(feature)) {
    catalog <- catalog[vapply(catalog[["features"]], function(x) all(feature %in% x), TRUE), , drop = FALSE]
  }
  if (!missing(class) && !is.null(class)) {
    catalog <- catalog[catalog[["class"]] %in% class, , drop = FALSE]
  }
  if (!missing(family) && !is.null(family)) {
    catalog <- catalog[catalog[["family"]] %in% family, , drop = FALSE]
  }
  if (!missing(has_metafor) && !is.null(has_metafor)) {
    catalog <- catalog[catalog[["has_metafor"]] %in% has_metafor, , drop = FALSE]
  }
  if (!missing(has_loo) && !is.null(has_loo)) {
    catalog <- catalog[catalog[["has_loo"]] %in% has_loo, , drop = FALSE]
  }
  if (!missing(has_waic) && !is.null(has_waic)) {
    catalog <- catalog[catalog[["has_waic"]] %in% has_waic, , drop = FALSE]
  }
  if (!missing(has_marglik) && !is.null(has_marglik)) {
    catalog <- catalog[catalog[["has_marglik"]] %in% has_marglik, , drop = FALSE]
  }
  if (!missing(tier) && !is.null(tier)) {
    catalog <- catalog[catalog[["tier"]] %in% tier, , drop = FALSE]
  }

  return(catalog[["name"]])
}

.source_hash_call_name <- function(expr) {

  if (!is.call(expr)) {
    return("")
  }

  fun <- expr[[1]]
  if (is.symbol(fun)) {
    return(as.character(fun))
  }
  if (is.call(fun) && identical(as.character(fun[[1]]), "::")) {
    return(as.character(fun[[3]]))
  }

  return("")
}

.source_hash_skip_call <- function(expr) {

  call_name <- .source_hash_call_name(expr)

  return(
    call_name %in% c("context", "source", "skip_on_cran",
                     "skip_if_not_installed", "skip_refit_if_cached") ||
      grepl("^expect_", call_name)
  )
}

.source_hash_normalize_expr <- function(expr) {

  if (is.call(expr)) {
    call_name <- .source_hash_call_name(expr)

    if (.source_hash_skip_call(expr)) {
      return(character())
    }

    if (identical(call_name, "test_that")) {
      body <- if (length(expr) >= 3L) expr[[3]] else NULL
      return(.source_hash_normalize_expr(body))
    }

    if (identical(call_name, "{")) {
      normalized <- unlist(
        lapply(as.list(expr)[-1], .source_hash_normalize_expr),
        use.names = FALSE
      )
      return(normalized[nzchar(normalized)])
    }
  }

  return(paste(deparse(expr, width.cutoff = 500L), collapse = "\n"))
}

source_file_md5 <- function(source_file) {

  if (is.null(source_file) || is.na(source_file) || !nzchar(source_file)) {
    return(NA_character_)
  }

  path <- testthat::test_path(source_file)
  if (!file.exists(path)) {
    return(NA_character_)
  }

  parsed <- try(parse(path), silent = TRUE)
  if (inherits(parsed, "try-error")) {
    lines <- readLines(path, warn = FALSE)
    lines <- sub("#.*$", "", lines)
    lines <- trimws(lines)
    lines <- lines[nzchar(lines)]
  } else {
    lines <- unlist(lapply(parsed, .source_hash_normalize_expr), use.names = FALSE)
    lines <- trimws(lines)
    lines <- lines[nzchar(lines)]
  }

  normalized <- tempfile("robma-fit-source-", fileext = ".R")
  writeLines(lines, normalized, useBytes = TRUE)
  on.exit(unlink(normalized), add = TRUE)

  return(unname(tools::md5sum(normalized)))
}

.package_source_md5_cache <- new.env(parent = emptyenv())

.fit_cache_required_source_files <- function() {

  # Keep this scoped to code that can change saved fit objects or cached fit
  # extensions. Post-fit methods are tested against the cached objects.
  r_files   <- c(
    "R/BMA.glmm.R",
    "R/BMA.norm.R",
    "R/RoBMA.R",
    "R/bPEESE.R",
    "R/bPET.R",
    "R/brma.glmm.R",
    "R/brma.norm.R",
    "R/bselmodel.R",
    "R/fit.R",
    "R/input-data.R",
    "R/input-object.R",
    "R/input-priors.R",
    "R/loo.R",
    "R/marglik.R",
    "R/pdf.R",
    "R/priors.R",
    "R/selection-mapping.R",
    "R/utilities.R",
    "R/zzz.R"
  )
  src_files <- c(
    "src/RoBMA.cc",
    "src/init.c",
    "src/distributions/DSELNORMKERNEL.cc",
    "src/distributions/DSELNORMKERNEL.h",
    "src/distributions/DSELNORMSTEP.cc",
    "src/distributions/DSELNORMSTEP.h",
    "src/distributions/DSELNORMSTEPSWITCH.cc",
    "src/distributions/DSELNORMSTEPSWITCH.h",
    "src/distributions/DWB.cc",
    "src/distributions/DWB.h",
    "src/distributions/DWN.cc",
    "src/distributions/DWN.h",
    "src/distributions/DWP.cc",
    "src/distributions/DWP.h",
    "src/r-glmm.cc",
    "src/r-selnorm.cc",
    "src/selnorm/selnorm.cc",
    "src/selnorm/selnorm.h"
  )

  return(sort(c(r_files, src_files)))
}

.path_ancestors <- function(path) {

  path <- normalizePath(path, winslash = "/", mustWork = FALSE)
  out  <- character()

  repeat {
    out    <- c(out, path)
    parent <- dirname(path)
    if (identical(parent, path)) {
      break
    }
    path <- parent
  }

  return(unique(out))
}

.fit_cache_source_root_candidates <- function() {

  package_name <- "RoBMA"
  roots        <- c(
    testthat::test_path("..", ".."),
    getwd(),
    system.file(package = package_name)
  )
  roots <- roots[nzchar(roots)]
  roots <- normalizePath(roots, winslash = "/", mustWork = FALSE)

  bases      <- unique(unlist(lapply(roots, .path_ancestors), use.names = FALSE))
  candidates <- unique(c(
    roots,
    file.path(bases, package_name),
    file.path(bases, "00_pkg_src", package_name)
  ))

  return(normalizePath(candidates, winslash = "/", mustWork = FALSE))
}

.fit_cache_source_tree_available <- function(package_root) {

  package_root <- normalizePath(package_root, winslash = "/", mustWork = FALSE)
  source_files <- .fit_cache_required_source_files()

  return(all(file.exists(file.path(package_root, source_files))))
}

.fit_cache_source_root <- function() {

  if (exists("source_root", envir = .package_source_md5_cache, inherits = FALSE)) {
    return(get("source_root", envir = .package_source_md5_cache, inherits = FALSE))
  }

  candidates <- .fit_cache_source_root_candidates()
  for (candidate in candidates) {
    if (.fit_cache_source_tree_available(candidate)) {
      assign("source_root", candidate, envir = .package_source_md5_cache)
      return(candidate)
    }
  }

  assign("source_root", NA_character_, envir = .package_source_md5_cache)
  return(NA_character_)
}

.fit_cache_source_files <- function(package_root = NULL, relative = FALSE) {

  if (is.null(package_root)) {
    package_root <- .fit_cache_source_root()
    if (is.na(package_root)) {
      return(character())
    }
  } else {
    package_root <- normalizePath(package_root, winslash = "/", mustWork = FALSE)
  }

  source_files <- .fit_cache_required_source_files()
  source_paths <- file.path(package_root, source_files)
  if (!all(file.exists(source_paths))) {
    return(character())
  }

  if (relative) {
    return(source_files)
  }

  return(sort(normalizePath(
    file.path(package_root, source_files),
    winslash = "/",
    mustWork = TRUE
  )))
}

.fit_cache_source_file_md5 <- function(path) {

  extension <- tolower(tools::file_ext(path))
  if (identical(extension, "r")) {
    parsed <- try(parse(file = path, keep.source = FALSE), silent = TRUE)
    if (!inherits(parsed, "try-error")) {
      lines <- unlist(lapply(parsed, function(expr) {
        paste(deparse(expr, width.cutoff = 500L), collapse = "\n")
      }), use.names = FALSE)

      normalized <- tempfile("robma-package-source-", fileext = ".R")
      writeLines(lines, normalized, useBytes = TRUE)
      on.exit(unlink(normalized), add = TRUE)

      return(unname(tools::md5sum(normalized)))
    }
  }

  return(unname(tools::md5sum(path)))
}

package_source_md5 <- function() {

  if (exists("value", envir = .package_source_md5_cache, inherits = FALSE)) {
    return(get("value", envir = .package_source_md5_cache, inherits = FALSE))
  }

  package_root <- .fit_cache_source_root()
  if (is.na(package_root)) {
    assign("value", NA_character_, envir = .package_source_md5_cache)
    return(NA_character_)
  }

  source_files <- .fit_cache_source_files(package_root = package_root)
  if (length(source_files) == 0L) {
    assign("value", NA_character_, envir = .package_source_md5_cache)
    return(NA_character_)
  }

  file_hashes    <- vapply(source_files, .fit_cache_source_file_md5, character(1))
  relative_files <- substring(source_files, nchar(package_root) + 2L)
  hash_input     <- paste(
    paste0(relative_files, ":", unname(file_hashes)),
    collapse = "\n"
  )

  normalized <- tempfile("robma-package-source-", fileext = ".txt")
  writeLines(hash_input, normalized, useBytes = TRUE)
  on.exit(unlink(normalized), add = TRUE)

  value <- unname(tools::md5sum(normalized))
  assign("value", value, envir = .package_source_md5_cache)

  return(value)
}

fit_cache_metadata <- function(name, fit, info = NULL) {

  entry       <- fit_catalog_entry(name)
  source_file <- if (is.null(entry)) NA_character_ else entry[["source_file"]]

  metadata <- list(
    version            = FIT_CACHE_VERSION,
    name               = name,
    saved_at           = format(Sys.time(), usetz = TRUE),
    fit_class          = class(fit),
    source_file        = source_file,
    source_file_md5    = source_file_md5(source_file),
    package_source_md5 = package_source_md5(),
    has_loo            = !is.null(fit[["loo"]]),
    has_waic           = !is.null(fit[["waic"]]),
    has_marglik        = !is.null(fit[["marglik"]]),
    has_metafor_info   = !is.null(info) && "metafor" %in% names(info) && !is.null(info[["metafor"]])
  )

  return(metadata)
}

fit_cache_paths <- function(name) {

  return(list(
    fit      = file.path(temp_fits_dir, paste0(name, ".RDS")),
    info     = file.path(temp_info_dir, paste0(name, ".RDS")),
    metadata = file.path(temp_metadata_dir, paste0(name, ".RDS")),
    marker   = file.path(temp_temp_dir, paste0(name, ".txt"))
  ))
}

load_fit_metadata <- function(name) {

  path <- fit_cache_paths(name)[["metadata"]]
  metadata <- suppressWarnings(try(readRDS(path), silent = TRUE))

  if (inherits(metadata, "try-error")) {
    return(NULL)
  }

  return(metadata)
}

read_cached_fit <- function(name) {

  path <- fit_cache_paths(name)[["fit"]]
  fit  <- suppressWarnings(try(readRDS(path), silent = TRUE))

  if (inherits(fit, "try-error")) {
    return(NULL)
  }

  return(fit)
}

.fit_object_cache  <- new.env(parent = emptyenv())
.info_object_cache <- new.env(parent = emptyenv())

is_true_env <- function(name) {

  value <- Sys.getenv(name)
  value <- tolower(value)

  return(value %in% c("1", "true", "yes", "y"))
}

is_false_env <- function(name) {

  value <- Sys.getenv(name)
  value <- tolower(value)

  return(value %in% c("0", "false", "no", "n"))
}

validate_cached_fit <- function(name, fit = NULL, info = NULL,
                                metadata = NULL, check_source = TRUE,
                                check_files = TRUE, deep = FALSE) {

  messages <- character()
  paths    <- fit_cache_paths(name)
  entry    <- fit_catalog_entry(name)

  if (check_files && !file.exists(paths[["fit"]]) && is.null(fit)) {
    messages <- c(messages, "fit file is missing")
    return(messages)
  }
  if (check_files && !file.exists(paths[["fit"]])) {
    messages <- c(messages, "fit file is missing")
  }

  if (is.null(entry)) {
    messages <- c(messages, "fit is missing from fit_catalog()")
  }

  if (is.null(metadata)) {
    metadata <- load_fit_metadata(name)
  }
  if (is.null(metadata)) {
    messages <- c(messages, "metadata file is missing")
  }

  if (!is.null(entry) && !is.null(metadata)) {
    if (is.null(metadata[["version"]]) || !identical(metadata[["version"]], FIT_CACHE_VERSION)) {
      messages <- c(messages, "cache version changed")
    }
    if (!is.null(metadata[["name"]]) && !identical(metadata[["name"]], name)) {
      messages <- c(messages, "metadata name mismatch")
    }
    if (is.null(metadata[["fit_class"]]) || !entry[["class"]] %in% metadata[["fit_class"]]) {
      messages <- c(messages, paste0("metadata fit class does not include '", entry[["class"]], "'"))
    }

    if (isTRUE(entry[["has_loo"]]) && !isTRUE(metadata[["has_loo"]])) {
      messages <- c(messages, "metadata reports missing LOO")
    }
    if (isFALSE(entry[["has_loo"]]) && isTRUE(metadata[["has_loo"]])) {
      messages <- c(messages, "metadata reports unexpected LOO")
    }

    if (isTRUE(entry[["has_waic"]]) && !isTRUE(metadata[["has_waic"]])) {
      messages <- c(messages, "metadata reports missing WAIC")
    }
    if (isFALSE(entry[["has_waic"]]) && isTRUE(metadata[["has_waic"]])) {
      messages <- c(messages, "metadata reports unexpected WAIC")
    }

    if (isTRUE(entry[["has_marglik"]]) && !isTRUE(metadata[["has_marglik"]])) {
      messages <- c(messages, "metadata reports missing marginal likelihood")
    }
    if (isFALSE(entry[["has_marglik"]]) && isTRUE(metadata[["has_marglik"]])) {
      messages <- c(messages, "metadata reports unexpected marginal likelihood")
    }

    if (isTRUE(entry[["has_metafor"]])) {
      if (!isTRUE(metadata[["has_metafor_info"]])) {
        messages <- c(messages, "metadata reports missing metafor reference")
      }
      if (check_files && !file.exists(paths[["info"]]) && is.null(info)) {
        messages <- c(messages, "metafor info file is missing")
      }
    }

    expected_md5 <- source_file_md5(entry[["source_file"]])
    if (check_source && !is.na(expected_md5) &&
        (is.null(metadata[["source_file_md5"]]) || !identical(metadata[["source_file_md5"]], expected_md5))) {
      messages <- c(messages, "source file hash changed")
    }

    expected_package_md5 <- package_source_md5()
    if (check_source && !is.na(expected_package_md5) &&
        (is.null(metadata[["package_source_md5"]]) ||
         !identical(metadata[["package_source_md5"]], expected_package_md5))) {
      messages <- c(messages, "cache source hash changed")
    }
  }

  if (!deep && is.null(fit) && is.null(info)) {
    return(messages)
  }

  if (is.null(fit)) {
    fit <- read_cached_fit(name)
  }
  if (is.null(fit)) {
    messages <- c(messages, "fit file cannot be read")
    return(messages)
  }

  if (is.null(info)) {
    info <- load_info(name, validate = FALSE)
  }

  if (!is.null(entry) && !inherits(fit, entry[["class"]])) {
    messages <- c(messages, paste0("fit does not inherit from '", entry[["class"]], "'"))
  }

  if (!is.null(entry) && isTRUE(entry[["has_loo"]]) && is.null(fit[["loo"]])) {
    messages <- c(messages, "LOO is missing")
  }
  if (!is.null(entry) && isFALSE(entry[["has_loo"]]) && !is.null(fit[["loo"]])) {
    messages <- c(messages, "LOO is unexpectedly present")
  }

  if (!is.null(entry) && isTRUE(entry[["has_waic"]]) && is.null(fit[["waic"]])) {
    messages <- c(messages, "WAIC is missing")
  }
  if (!is.null(entry) && isFALSE(entry[["has_waic"]]) && !is.null(fit[["waic"]])) {
    messages <- c(messages, "WAIC is unexpectedly present")
  }

  if (!is.null(entry) && isTRUE(entry[["has_marglik"]]) && is.null(fit[["marglik"]])) {
    messages <- c(messages, "marginal likelihood is missing")
  }
  if (!is.null(entry) && isFALSE(entry[["has_marglik"]]) && !is.null(fit[["marglik"]])) {
    messages <- c(messages, "marginal likelihood is unexpectedly present")
  }

  if (!is.null(entry) && isTRUE(entry[["has_metafor"]])) {
    if (is.null(info) || !"metafor" %in% names(info) || is.null(info[["metafor"]])) {
      messages <- c(messages, "metafor reference is missing")
    }
  }

  return(messages)
}

is_cached_fit_valid <- function(name, check_source = TRUE, deep = FALSE) {

  return(length(validate_cached_fit(name, check_source = check_source, deep = deep)) == 0)
}

# Skip if pre-fitted models are not available
skip_if_no_fits <- function() {

  if (length(list_fits(validate = TRUE)) == 0) {
    skip("No valid cached fits available. Run `devtools::test(filter = '01-')` first.")
  }
}

skip_if_missing_fits <- function(names) {

  missing <- setdiff(names, list_fits(validate = TRUE))
  if (length(missing) > 0) {
    skip(paste0("Required pre-fitted models missing or stale: ", paste(missing, collapse = ", ")))
  }
}

skip_if_not_full_diagnostics <- function(reason) {

  if (!is_true_env("ROBMA_TEST_FULL_DIAGNOSTICS")) {
    skip(paste(
      "Skipping extended diagnostic redundancy check by default.",
      reason,
      "Set ROBMA_TEST_FULL_DIAGNOSTICS=TRUE to run it."
    ))
  }
}

# Skip model fitting if a valid cached fit exists. Refit by setting
# ROBMA_TEST_FORCE_REFIT=TRUE or ROBMA_TEST_SKIP_REFIT=FALSE.
skip_refit_if_cached <- function(name) {

  if (is_true_env("ROBMA_TEST_FORCE_REFIT")) {
    return(invisible(FALSE))
  }

  skip_refit_env <- Sys.getenv("ROBMA_TEST_SKIP_REFIT")
  skip_refit     <- if (skip_refit_env == "") TRUE else !is_false_env("ROBMA_TEST_SKIP_REFIT")
  fit_names      <- catalog_group_fits(name)

  if (skip_refit && length(fit_names) > 0 &&
      all(vapply(fit_names, is_cached_fit_valid, TRUE))) {
    skip("Skipping model refit: valid cached fit exists.")
  }

  return(invisible(FALSE))
}

# ============================================================================ #
# HELPER FUNCTIONS: Model Fit Saving / Loading
# ============================================================================ #

save_fit <- function(name, fit, info = NULL) {
  metadata <- fit_cache_metadata(name, fit, info)
  problems <- validate_cached_fit(
    name         = name,
    fit          = fit,
    info         = info,
    metadata     = metadata,
    check_source = FALSE
  )
  problems <- problems[problems != "fit file is missing"]

  if (length(problems) > 0) {
    stop(
      paste0("Cached fit '", name, "' is invalid: ", paste(problems, collapse = "; ")),
      call. = FALSE
    )
  }

  paths <- fit_cache_paths(name)

  # Save model fit (marglik and loo are stored in the fit object)
  saveRDS(fit, file = paths[["fit"]])

  # Save info if provided
  if (!is.null(info)) {
    saveRDS(info, file = paths[["info"]])
  }

  # Save cache metadata last so interrupted fits do not validate later.
  saveRDS(metadata, file = paths[["metadata"]])

  return(invisible(TRUE))
}

load_fit <- function(name, validate = TRUE) {

  if (exists(name, envir = .fit_object_cache, inherits = FALSE)) {
    return(get(name, envir = .fit_object_cache, inherits = FALSE))
  }

  if (validate) {
    problems <- validate_cached_fit(name)
    if (length(problems) > 0) {
      skip(paste0(
        "Cached fit '", name, "' is missing, stale, or invalid: ",
        paste(problems, collapse = "; "),
        ". Run `test(filter = 'test-01')`."
      ))
    }
  }

  fit <- read_cached_fit(name)
  if (is.null(fit)) {
    stop(paste0("Cached fit '", name, "' could not be read."), call. = FALSE)
  }

  assign(name, fit, envir = .fit_object_cache)

  return(fit)
}

load_info <- function(name, validate = TRUE) {

  if (exists(name, envir = .info_object_cache, inherits = FALSE)) {
    return(get(name, envir = .info_object_cache, inherits = FALSE))
  }

  if (validate) {
    problems <- validate_cached_fit(name)
    if (length(problems) > 0) {
      skip(paste0(
        "Cached fit '", name, "' is missing, stale, or invalid: ",
        paste(problems, collapse = "; "),
        ". Run `test(filter = 'test-01')`."
      ))
    }
  }

  # load model info
  info <- suppressWarnings(try(readRDS(file = file.path(temp_info_dir, paste0(name, ".RDS"))), silent = TRUE))
  if (inherits(info, "try-error")) {
    info <- list()
  }

  assign(name, info, envir = .info_object_cache)

  return(info)
}

list_fits <- function(name, feature, class, family, has_metafor, has_loo,
                      has_waic, has_marglik, tier, validate = TRUE,
                      deep = FALSE) {

  files <- suppressWarnings(list.files(temp_fits_dir, pattern = "\\.RDS$"))
  files <- sub("\\.RDS$", "", files)

  if (!missing(name)) {
    files <- intersect(files, name)
  }

  if (!missing(feature) || !missing(class) || !missing(family) ||
      !missing(has_metafor) || !missing(has_loo) || !missing(has_waic) ||
      !missing(has_marglik) || !missing(tier)) {
    selected <- catalog_fits(
      feature     = if (!missing(feature)) feature else NULL,
      class       = if (!missing(class)) class else NULL,
      family      = if (!missing(family)) family else NULL,
      has_metafor = if (!missing(has_metafor)) has_metafor else NULL,
      has_loo     = if (!missing(has_loo)) has_loo else NULL,
      has_waic    = if (!missing(has_waic)) has_waic else NULL,
      has_marglik = if (!missing(has_marglik)) has_marglik else NULL,
      tier        = if (!missing(tier)) tier else NULL
    )
    files <- intersect(files, selected)
  }

  if (validate) {
    files <- files[vapply(files, is_cached_fit_valid, TRUE, deep = deep)]
  }

  return(files)
}

lazy_cached_objects <- function(names, loader) {

  names <- unique(names)

  return(structure(
    list(
      .names  = names,
      .loader = loader,
      .cache  = new.env(parent = emptyenv())
    ),
    class = "lazy_cached_objects"
  ))
}

lazy_fits <- function(names = list_fits(), validate = TRUE) {

  if (validate) {
    skip_if_missing_fits(names)
  }

  return(lazy_cached_objects(
    names,
    function(name) load_fit(name, validate = FALSE)
  ))
}

lazy_infos <- function(names = list_fits(), validate = TRUE) {

  if (validate) {
    skip_if_missing_fits(names)
  }

  return(lazy_cached_objects(
    names,
    function(name) load_info(name, validate = FALSE)
  ))
}

load_fits <- function(names = list_fits(), validate = TRUE) {

  out <- lapply(names, load_fit, validate = validate)
  names(out) <- names

  return(out)
}

load_infos <- function(names = list_fits(), validate = TRUE) {

  out <- lapply(names, load_info, validate = validate)
  names(out) <- names

  return(out)
}

names.lazy_cached_objects <- function(x) {

  return(unclass(x)[[".names"]])
}

length.lazy_cached_objects <- function(x) {

  return(length(names(x)))
}

`[[.lazy_cached_objects` <- function(x, i, exact = TRUE) {

  object_names <- names(x)

  if (is.numeric(i)) {
    i <- object_names[[i]]
  }
  if (!is.character(i) || length(i) != 1L || !i %in% object_names) {
    return(NULL)
  }

  cache <- unclass(x)[[".cache"]]
  if (!exists(i, envir = cache, inherits = FALSE)) {
    assign(i, unclass(x)[[".loader"]](i), envir = cache)
  }

  return(get(i, envir = cache, inherits = FALSE))
}

`[.lazy_cached_objects` <- function(x, i) {

  object_names <- names(x)
  selected     <- object_names[i]

  return(lazy_cached_objects(
    selected,
    unclass(x)[[".loader"]]
  ))
}

as.list.lazy_cached_objects <- function(x, ...) {

  out <- lapply(names(x), function(name) x[[name]])
  names(out) <- names(x)

  return(out)
}

clean_cached_fits <- function(name) {

  if (!missing(name)) {
    fit_names <- catalog_group_fits(name)
    if (length(fit_names) == 0) {
      fit_names <- name
    }
    for (fit_name in fit_names) {
      paths <- fit_cache_paths(fit_name)
      file.remove(unlist(paths, use.names = FALSE))
    }
  } else {
    # Remove all cached files from test directories
    unlink(temp_fits_dir, recursive = TRUE)
    unlink(temp_info_dir, recursive = TRUE)
    unlink(temp_metadata_dir, recursive = TRUE)
    unlink(temp_temp_dir, recursive = TRUE)

    # Recreate empty directories
    dir.create(temp_fits_dir, showWarnings = FALSE, recursive = TRUE)
    dir.create(temp_info_dir, showWarnings = FALSE, recursive = TRUE)
    dir.create(temp_metadata_dir, showWarnings = FALSE, recursive = TRUE)
    dir.create(temp_temp_dir, showWarnings = FALSE, recursive = TRUE)
  }

  message("Cleaned cached fits in: ", test_files_dir)

  return(invisible(TRUE))
}


# Construct conditional standardized residuals for multilevel metafor::rma.mv fits.
# metafor::rstandard() only provides marginal residuals for rma.mv, so the
# conditional oracle must be built from the marginal fit, BLUP random effects,
# and the GLS residual variance matrix.
metafor_rstandard_conditional_mv <- function(model) {
  if (!inherits(model, "rma.mv")) {
    stop("'model' must inherit from 'rma.mv'.", call. = FALSE)
  }
  if (is.null(model[["yi"]]) || is.null(model[["X"]]) ||
      is.null(model[["M"]])  || is.null(model[["V"]])) {
    stop("Model object does not contain the information needed for conditional residuals.", call. = FALSE)
  }

  fitted_values <- as.vector(stats::fitted(model))

  ranef_components <- metafor::ranef(model, expand = TRUE)
  if (length(ranef_components) == 0L) {
    stop("No random-effect BLUPs available for this 'rma.mv' object.", call. = FALSE)
  }

  random_effects <- Reduce(
    `+`,
    lapply(ranef_components, function(component) component[["intrcpt"]])
  )

  residuals <- as.vector(model[["yi"]] - (fitted_values + random_effects))

  hat_matrix <- stats::hatvalues(model, type = "matrix")
  marginal_v <- model[["M"]]
  sampling_v <- model[["V"]]
  weights    <- chol2inv(chol(marginal_v))
  k          <- model[["k"]]
  identity   <- diag(k)
  imh        <- identity - hat_matrix

  residual_v <- sampling_v %*% weights %*% imh %*% marginal_v %*%
    t(imh) %*% weights %*% sampling_v
  se <- sqrt(diag(residual_v))

  out <- data.frame(
    resid = residuals,
    se    = as.vector(se),
    z     = residuals / se
  )
  rownames(out) <- NULL

  return(out)
}

# ============================================================================ #
# brma.loo.R
# ============================================================================ #
#
# This file implements LOO-PSIS (Leave-One-Out Cross-Validation via Pareto
# Smoothed Importance Sampling) diagnostics for brma class objects.
#
# LOO can be computed at the estimate unit or, for multilevel models, at the
# cluster unit.
#
# Important: LOO-PSIS evaluates how well the model predicts new observations.
# This is different from evaluating how well the model fits the observed data.
# Model comparison via loo_compare() selects models based on their expected
# out-of-sample predictive performance, not based on in-sample fit.
#
# ============================================================================ #


# ---------------------------------------------------------------------------- #
# add_loo generic and method
# ---------------------------------------------------------------------------- #

#' @export
add_loo <- function(object, ...) UseMethod("add_loo")


#' @title Add LOO-PSIS to brma Objects
#'
#' @description Compute approximate leave-one-out cross-validation (LOO-CV)
#' using Pareto smoothed importance sampling (PSIS) for brma model objects
#' and store the result in the object.
#'
#' @param object a brma model object.
#' @param unit output/deletion unit. \code{"estimate"} computes one contribution
#' per effect-size estimate. \code{"cluster"} computes one contribution per
#' cluster and is available only for multilevel models.
#' @param r_eff optional vector of relative effective sample sizes. If not
#' provided, it is computed from the log-likelihood values.
#' @param parallel Logical. If \code{TRUE}, \code{loo::relative_eff()} and
#' \code{loo::loo()} use \code{RoBMA.get_option("max_cores")}. Log-likelihood
#' construction is unchanged. If \code{FALSE}, those computations use one core.
#' @param ... additional arguments (currently ignored).
#'
#' @details
#' With \code{unit = "estimate"}, LOO-CV is computed with one contribution per
#' effect-size estimate. For binomial and Poisson models, each pair of counts
#' (ai/ci or x1i/x2i) that defines a single effect size estimate is treated as
#' one contribution.
#'
#' With \code{unit = "cluster"}, LOO-CV is computed with one joint contribution
#' per cluster. For unweighted normal models without selection this uses the
#' analytic cluster block covariance. Selection, data-weighted normal, and
#' GLMM models integrate the held-out cluster effect with Gauss-Hermite
#' quadrature.
#'
#' For selection models, the LOO evaluates the weighted likelihood, conditioning
#' on the posterior omega samples.
#'
#' The PSIS object is essential for model comparison via
#' \code{\link[loo]{loo_compare}} and is automatically saved in the loo result.
#' RoBMA stores target metadata so comparisons can reject mismatched data,
#' unit, or conditioning-depth targets.
#'
#' \strong{Important for model comparison:} When comparing models via
#' \code{\link[loo]{loo_compare}}, the selection is based on expected
#' out-of-sample predictive performance. This evaluates how well models predict
#' \emph{new} observations, not how well they fit the observed data.
#'
#' @return The brma object with the LOO result stored in
#' \code{object[["loo"]][[unit]]}.
#'
#' @seealso \code{\link{loo.brma}}, \code{\link[loo]{loo}},
#' \code{\link[loo]{loo_compare}}, \code{\link[loo]{pareto_k_ids}}
#'
#' @examples \dontrun{
#' if (requireNamespace("metadat", quietly = TRUE)) {
#'   data(dat.lehmann2018, package = "metadat")
#'   fit <- bPET(yi = yi, vi = vi, data = dat.lehmann2018, measure = "SMD")
#'
#'   fit <- add_loo(fit)
#'   loo_fit <- loo(fit)
#'   print(loo_fit)
#'   plot(loo_fit)
#' }
#' }
#'
#' @references
#' \insertCite{vehtari2017practical}{RoBMA}
#' \insertCite{vehtari2024pareto}{RoBMA}
#'
#' @aliases add_loo
#' @export
add_loo.brma <- function(object, unit = "estimate", r_eff = NULL, parallel = FALSE, ...) {
  unit <- .normalize_unit(unit)
  BayesTools::check_bool(parallel, "parallel")

  conditioning_depth <- .loo_conditioning_depth_from_unit(unit)
  .check_unit_conditioning_depth(
    object             = object,
    unit               = unit,
    conditioning_depth = conditioning_depth,
    caller             = "add_loo()"
  )

  # compute the log-likelihood matrix (S x K)
  log_lik <- .log_lik.brma(object, unit = unit)
  target  <- attr(log_lik, "RoBMA_target", exact = TRUE)

  # determine number of cores based on `parallel` and package options
  cores <- if (parallel) max(1, RoBMA.get_option("max_cores")) else 1

  # compute relative effective sample sizes if not provided
  if (is.null(r_eff)) {
    # loo::relative_eff expects exp(log_lik) with chain_id for matrix input
    chain_id <- .loo_chain_id(object[["fit"]], n_samples = nrow(log_lik))

    r_eff <- loo::relative_eff(exp(log_lik), chain_id = chain_id, cores = cores)
  }

  # call loo on the log-likelihood matrix
  loo_result <- loo::loo(log_lik, r_eff = r_eff, save_psis = TRUE, cores = cores)
  loo_result <- .add_loo_target_metadata(
    object             = loo_result,
    unit               = target[["unit"]],
    conditioning_depth = .get_target_conditioning_depth(target),
    targets            = target[["targets"]],
    data_hash          = target[["data_hash"]]
  )

  # store in object and return
  if (is.null(object[["loo"]])) {
    object[["loo"]] <- list()
  }
  object[["loo"]][[unit]] <- loo_result
  return(object)
}


# ---------------------------------------------------------------------------- #
# .loo_chain_id
# ---------------------------------------------------------------------------- #
#
# Derive chain IDs from the retained MCMC samples.
#
# @param fit       runjags fit object.
# @param n_samples integer; expected number of posterior rows.
#
# @return integer vector of length n_samples.
#
# ---------------------------------------------------------------------------- #
.loo_chain_id <- function(fit, n_samples) {

  if (!is.null(fit[["mcmc"]])) {
    chain_lengths <- vapply(fit[["mcmc"]], NROW, integer(1))
    chain_id      <- rep(seq_along(chain_lengths), times = chain_lengths)
  } else {
    n_chains <- length(fit[["mcmc"]])
    n_iter   <- fit[["sample"]]
    chain_id <- rep(seq_len(n_chains), each = n_iter)
  }

  if (length(chain_id) != n_samples) {
    stop(
      "Could not derive valid chain IDs for relative_eff(): expected ",
      n_samples, " posterior rows but got ", length(chain_id), ". ",
      "Supply 'r_eff' explicitly.",
      call. = FALSE
    )
  }

  return(chain_id)
}


# ---------------------------------------------------------------------------- #
# add_waic generic and method
# ---------------------------------------------------------------------------- #

#' @export
add_waic <- function(object, ...) UseMethod("add_waic")


#' @title Add WAIC to brma Objects
#'
#' @description Compute the Widely Applicable Information Criterion (WAIC)
#' for brma model objects and store the result in the object.
#'
#' @param object a brma model object.
#' @param unit output/deletion unit. See \code{\link{add_loo}}; the same
#' accepted values and multilevel constraint apply.
#' @param ... additional arguments passed to \code{\link[loo]{waic}}.
#'
#' @details
#' WAIC is an alternative to LOO-CV for estimating out-of-sample predictive
#' accuracy. Like LOO, it evaluates expected predictive performance for new
#' observations.
#'
#' In most cases, LOO-PSIS (via \code{\link{add_loo}}) is preferred over WAIC
#' because it provides better estimates and includes diagnostics (Pareto k
#' values) that indicate when the approximation may be unreliable.
#'
#' @return The brma object with the WAIC result stored in
#' \code{object[["waic"]][[unit]]}.
#'
#' @seealso \code{\link{waic.brma}}, \code{\link{add_loo}}, \code{\link[loo]{waic}}
#'
#' @aliases add_waic
#' @export
add_waic.brma <- function(object, unit = "estimate", ...) {
  unit <- .normalize_unit(unit)
  conditioning_depth <- .loo_conditioning_depth_from_unit(unit)
  .check_unit_conditioning_depth(
    object             = object,
    unit               = unit,
    conditioning_depth = conditioning_depth,
    caller             = "add_waic()"
  )

  # compute the log-likelihood matrix (S x K)
  log_lik <- .log_lik.brma(object, unit = unit)
  target  <- attr(log_lik, "RoBMA_target", exact = TRUE)

  # call waic on the log-likelihood matrix
  waic_result <- loo::waic(log_lik, ...)
  waic_result <- .add_loo_target_metadata(
    object             = waic_result,
    unit               = target[["unit"]],
    conditioning_depth = .get_target_conditioning_depth(target),
    targets            = target[["targets"]],
    data_hash          = target[["data_hash"]]
  )

  # store in object and return
  if (is.null(object[["waic"]])) {
    object[["waic"]] <- list()
  }
  object[["waic"]][[unit]] <- waic_result
  return(object)
}


# ---------------------------------------------------------------------------- #
# loo generic and extraction method
# ---------------------------------------------------------------------------- #

#' @export
loo <- function(x, ...) UseMethod("loo")


#' @title LOO-PSIS for brma Objects
#'
#' @description Extract the LOO-PSIS object from a brma model object.
#' The LOO must first be computed using \code{\link{add_loo}}.
#'
#' @param x a brma model object.
#' @param unit output/deletion unit. See \code{\link{add_loo}}.
#' @param ... additional arguments (currently unused).
#'
#' @details
#' This function extracts the LOO object that was previously computed and
#' stored using \code{object <- add_loo(object, unit = unit)}. If LOO has not
#' been computed for the requested unit, an error is thrown.
#'
#' This is the RoBMA S3 generic and \code{brma} method. Use
#' \code{\link[loo]{loo}} directly for raw log-likelihood arrays or matrices.
#'
#' @return An object of class \code{c("psis_loo", "loo")} as returned by
#' \code{\link[loo]{loo}}.
#'
#' @seealso \code{\link{add_loo}}, \code{\link[loo]{loo}},
#' \code{\link[loo]{loo_compare}}, \code{\link[loo]{pareto_k_ids}}
#'
#' @examples \dontrun{
#' if (requireNamespace("metadat", quietly = TRUE)) {
#'   data(dat.lehmann2018, package = "metadat")
#'   fit <- bPET(yi = yi, vi = vi, data = dat.lehmann2018, measure = "SMD")
#'   fit <- add_loo(fit)
#'
#'   loo_fit <- loo(fit)
#'   print(loo_fit)
#' }
#' }
#'
#' @aliases loo
#' @export
loo.brma <- function(x, unit = "estimate", ...) {
  return(.check_loo_target(x, unit = unit))
}


#' @title Extract Log-Likelihood Matrix from brma Object
#'
#' @description Extract the pointwise log-likelihood matrix from a brma model
#' object. This is an S x K or S x G matrix where S is the number of posterior
#' samples, K is the number of estimates, and G is the number of clusters.
#' This method implements the S3
#' \'logLik\' generic for \code{brma} objects and returns the matrix of
#' pointwise log-likelihoods (one column per observation, one row per sample).
#'
#' @param object a brma model object.
#' @param unit output unit. See \code{\link{add_loo}}.
#' @param ... currently unused.
#'
#' @details
#' The log-likelihood is computed for each observation at each posterior sample.
#' For binomial and Poisson models, each observation consists of a pair of
#' counts (ai/ci or x1i/x2i) that together define a single effect size estimate.
#'
#' @return An S x K or S x G matrix of log-likelihood values.
#'
#' @seealso \code{\link{loo.brma}}
#'
#' @export
logLik.brma <- function(object, unit = "estimate", ...) {
  out <- .log_lik.brma(object, unit = unit)
  class(out) <- c("logLik.brma", class(out))
  return(out)
}


#' @export
print.logLik.brma <- function(x, ...) {
  S <- if (is.matrix(x)) nrow(x) else length(x)
  K <- if (is.matrix(x)) ncol(x) else 1
  cat(sprintf("%d*%d pointwise log-likelihood matrix\n", S, K))
  invisible(x)
}


# ---------------------------------------------------------------------------- #
# loo_compare generic and methods
# ---------------------------------------------------------------------------- #

#' @export
loo_compare <- function(x, ...) UseMethod("loo_compare")


#' @title Compare brma Models Using LOO
#'
#' @description Compare multiple brma models using LOO-PSIS cross-validation.
#' This is a convenience wrapper around \code{\link[loo]{loo_compare}}.
#'
#' @param x a brma model object (the first model to compare).
#' @param ... additional brma model objects or \code{loo} objects to compare.
#' @param unit output/deletion unit used when extracting LOO from brma objects.
#'
#' @details
#' This function compares models based on their expected out-of-sample
#' predictive performance (ELPD).
#'
#' \strong{Important for model comparison:} When comparing models via
#' \code{\link[loo]{loo_compare}}, the selection is based on expected
#' out-of-sample predictive performance. This evaluates how well models predict
#' \emph{new} observations, not how well they fit the observed data.
#' RoBMA rejects comparisons with different outcome targets/data, \code{unit},
#' or implied \code{conditioning_depth}.
#'
#' @return A matrix of class \code{"compare.loo"} as returned by
#' \code{\link[loo]{loo_compare}}.
#'
#' @seealso \code{\link{loo.brma}}, \code{\link[loo]{loo_compare}}
#'
#' @examples \dontrun{
#' if (requireNamespace("metadat", quietly = TRUE)) {
#'   data(dat.lehmann2018, package = "metadat")
#'
#'   fit_bias <- RoBMA(yi = yi, vi = vi, data = dat.lehmann2018, measure = "SMD")
#'   fit_nobias <- BMA(yi = yi, vi = vi, data = dat.lehmann2018, measure = "SMD")
#'
#'   fit_bias <- add_loo(fit_bias)
#'   fit_nobias <- add_loo(fit_nobias)
#'
#'   loo_compare(fit_bias, fit_nobias)
#'   loo_compare(loo(fit_bias), loo(fit_nobias))
#' }
#' }
#'
#' @aliases loo_compare
#' @export
loo_compare.brma <- function(x, ..., unit = "estimate") {
  unit <- .normalize_unit(unit)

  # collect all models: x plus any in ...
  models <- c(list(x), list(...))

  if (length(models) < 2) {
    stop("At least two models are required for comparison.", call. = FALSE)
  }

  # convert brma objects to loo objects if necessary
  loo_objects <- lapply(models, function(m) {
    if (inherits(m, "brma")) {
      loo.brma(m, unit = unit)
    } else if (inherits(m, "loo")) {
      m
    } else {
      stop("All arguments must be brma or loo objects.", call. = FALSE)
    }
  })

  # call the 'loo' package's default implementation to avoid dispatch recursion
  .check_loo_compare_targets(loo_objects)
  loo_compare_fun <- get("loo_compare.default", envir = asNamespace("loo"), inherits = FALSE)
  result <- do.call(loo_compare_fun, loo_objects)

  return(result)
}


#' @title Compare loo Objects Using LOO
#'
#' @description Method for comparing RoBMA-targeted \code{loo} or \code{waic}
#' objects directly.
#'
#' @param x a RoBMA-targeted \code{loo} or \code{waic} object (the first model
#' to compare).
#' @param ... additional RoBMA-targeted \code{loo}/\code{waic} or \code{brma}
#' objects to compare.
#' @param unit output/deletion unit used when extracting LOO from brma objects.
#'
#' @return A matrix of class \code{"compare.loo"} as returned by
#' \code{\link[loo]{loo_compare}}.
#'
#' @seealso \code{\link{loo.brma}}, \code{\link[loo]{loo_compare}}
#'
#' @export
loo_compare.loo <- function(x, ..., unit = "estimate") {
  unit <- .normalize_unit(unit)

  # collect all models: x plus any in ...
  models <- c(list(x), list(...))

  if (length(models) < 2) {
    stop("At least two models are required for comparison.", call. = FALSE)
  }

  # convert brma objects to loo objects if necessary
  loo_objects <- lapply(models, function(m) {
    if (inherits(m, "brma")) {
      loo.brma(m, unit = unit)
    } else if (inherits(m, "loo")) {
      m
    } else {
      stop("All arguments must be brma or loo objects.", call. = FALSE)
    }
  })

  # call the 'loo' package's default implementation to avoid dispatch recursion
  .check_loo_compare_targets(loo_objects)
  loo_compare_fun <- get("loo_compare.default", envir = asNamespace("loo"), inherits = FALSE)
  result <- do.call(loo_compare_fun, loo_objects)

  return(result)
}


# ---------------------------------------------------------------------------- #
# waic generic and extraction method
# ---------------------------------------------------------------------------- #

#' @export
waic <- function(x, ...) UseMethod("waic")


#' @title WAIC for brma Objects
#'
#' @description Extract the WAIC object from a brma model object.
#' The WAIC must first be computed using \code{\link{add_waic}}.
#'
#' @param x a brma model object.
#' @param unit output/deletion unit. See \code{\link{add_loo}}.
#' @param ... additional arguments (currently unused).
#'
#' @details
#' This function extracts the WAIC object that was previously computed and
#' stored using \code{object <- add_waic(object, unit = unit)}. If WAIC has not
#' been computed for the requested unit, an error is thrown.
#'
#' This is the RoBMA S3 generic and \code{brma} method. The method is also
#' registered for \code{\link[loo]{waic}}, so \code{loo::waic(fit)} extracts
#' the cached WAIC object for \code{brma} fits. Use \code{\link[loo]{waic}}
#' directly for raw log-likelihood arrays or matrices.
#'
#' In most cases, LOO-PSIS (via \code{\link{loo.brma}}) is preferred over WAIC
#' because it provides better estimates and includes diagnostics (Pareto k
#' values) that indicate when the approximation may be unreliable.
#'
#' @return An object of class \code{"waic"} as returned by
#' \code{\link[loo]{waic}}.
#'
#' @seealso \code{\link{add_waic}}, \code{\link{loo.brma}}, \code{\link[loo]{waic}}
#'
#' @aliases waic
#' @export
#' @exportS3Method loo::waic
waic.brma <- function(x, unit = "estimate", ...) {
  return(.check_waic_target(x, unit = unit))
}


# ---------------------------------------------------------------------------- #
# loo_weights generic and methods
# ---------------------------------------------------------------------------- #

#' @export
loo_weights <- function(object, ...) UseMethod("loo_weights")


#' @title Extract Normalized PSIS Weights from brma Object
#'
#' @description Extract the normalized Pareto smoothed importance sampling
#' (PSIS) weights from a brma model object.
#'
#' @param object a brma model object.
#' @param unit output/deletion unit. See \code{\link{add_loo}}.
#' @param ... currently unused.
#'
#' @details LOO must first be computed with
#' \code{object <- add_loo(object, unit = unit)}. This method extracts the
#' stored PSIS object and does not compute LOO.
#'
#' @return An \code{S x K} matrix for estimate-unit LOO, or \code{S x G}
#' matrix for cluster-unit LOO, with posterior samples in rows and LOO targets
#' in columns. Columns are normalized to sum to one.
#'
#' @aliases loo_weights
#' @exportS3Method
loo_weights.brma <- function(object, unit = "estimate", ...) {
  # extract loo
  loo_result <- loo.brma(object, unit = unit)

  # extract PSIS object and get normalized weights
  psis_object <- loo_result[["psis_object"]]
  psis_weights <- loo::weights.importance_sampling(psis_object, log = FALSE, normalize = TRUE)

  return(psis_weights)
}


# ---------------------------------------------------------------------------- #
# check_loo generic and methods
# ---------------------------------------------------------------------------- #

#' @export
check_loo <- function(object, ...) UseMethod("check_loo")


#' @title Check LOO Diagnostics for brma Objects
#'
#' @description Check Pareto k diagnostics for a brma model object and
#' warn if any values are high.
#'
#' @param object a brma model object.
#' @param unit output/deletion unit. See \code{\link{add_loo}}.
#' @param ... currently unused.
#'
#' @details LOO must first be computed with
#' \code{object <- add_loo(object, unit = unit)}. The method warns when any
#' Pareto \eqn{k} diagnostic is greater than 0.7.
#'
#' @return NULL (throws warning if diagnostics are unreliable).
#'
#' @aliases check_loo
#' @exportS3Method
check_loo.brma <- function(object, unit = "estimate", ...) {
  # extract loo
  unit       <- .normalize_unit(unit)
  loo_result <- loo.brma(object, unit = unit)

  # check Pareto k diagnostics and warn if unreliable
  pareto_k <- loo_result[["diagnostics"]][["pareto_k"]]
  bad_k <- which(pareto_k > 0.7)
  if (length(bad_k) > 0) {
    warning(
      "Some Pareto k values are high (> 0.7), indicating potentially unreliable ",
      "LOO diagnostics for ", unit, "s: ", paste(bad_k, collapse = ", "), ". ",
      "Inspect the loo fit by using loo(object).",
      call. = FALSE
    )
  }

  return()
}

context("Model fitting for RoBMA")

# Load common test helpers
source(testthat::test_path("common-functions.R"))
skip_on_cran()
skip_if_not_installed("metadat")
skip_refit_if_cached("RoBMA")


test_that("RoBMA handles default model", {
  data(dat.lehmann2018, package = "metadat")
  fit <- RoBMA(
    yi = yi, vi = vi,
    data = dat.lehmann2018, measure = "SMD",
    chains = 2, sample = 1000, burnin = 500, adapt = 500,
    seed = 1, silent = TRUE
  )
  fit <- suppressWarnings(add_loo(fit))
  save_fit("dat.lehmann2018_RoBMA", fit)

  expect_s3_class(fit, "RoBMA")
  expect_true(BayesTools::is.prior.mixture(fit$priors$outcome$mu))
  expect_true(BayesTools::is.prior.mixture(fit$priors$outcome$tau))
  expect_true(BayesTools::is.prior.mixture(fit$priors$outcome$bias))
  expect_no_error(summary(fit))
})

test_that("RoBMA handles custom priors", {
  data(dat.lehmann2018, package = "metadat")
  fit <- RoBMA(
    yi = yi, vi = vi,
    data = dat.lehmann2018, measure = "SMD",
    prior_effect = prior("normal", list(mean = 0, sd = 0.5)),
    prior_effect_null = prior("spike", list(location = 0)),
    prior_heterogeneity = prior("normal", list(mean = 0, sd = 0.25), truncation = list(lower = 0)),
    prior_heterogeneity_null = NULL,  # no null hypothesis for heterogeneity
    chains = 2, sample = 1000, burnin = 500, adapt = 500,
    seed = 1, silent = TRUE
  )
  fit <- suppressWarnings(add_loo(fit))
  save_fit("dat.lehmann2018_RoBMA_custom", fit)

  expect_s3_class(fit, "RoBMA")
  expect_equal(length(fit$priors$outcome$mu), 2)
  expect_equal(length(fit$priors$outcome$tau), 1)
  expect_equal(fit$priors$outcome$mu[[2]]$parameters, list(mean = 0, sd = 0.5))
  expect_equal(fit$priors$outcome$tau[[1]]$parameters, list(mean = 0, sd = 0.25))
})

test_that("RoBMA handles meta-regression", {
  data(dat.lehmann2018, package = "metadat")
  fit <- suppressWarnings(RoBMA(
    yi = yi, vi = vi, mods = ~ Preregistered,
    data = dat.lehmann2018, measure = "SMD",
    chains = 2, sample = 1000, burnin = 500, adapt = 500,
    seed = 1, silent = TRUE
  ))
  fit <- suppressWarnings(add_loo(fit))
  save_fit("dat.lehmann2018_RoBMA_mods", fit, info = list(mods = c("Preregistered")))

  expect_s3_class(fit, "RoBMA")
  expect_equal(length(fit$priors$mods), 2)
})

test_that("RoBMA handles meta-regression with interaction", {
  data(dat.lehmann2018, package = "metadat")
  fit <- RoBMA(
    yi = yi, vi = vi, mods = ~ Preregistered * Gender,
    data = dat.lehmann2018, measure = "SMD",
    chains = 2, sample = 1000, burnin = 500, adapt = 500,
    seed = 1, silent = TRUE
  )
  fit <- suppressWarnings(add_loo(fit))
  save_fit("dat.lehmann2018_RoBMA_mods2", fit, info = list(mods = c("Preregistered", "Gender", "Preregistered:Gender")))

  expect_s3_class(fit, "RoBMA")
  expect_equal(length(fit$priors$mods), 4)
})

test_that("RoBMA handles multilevel location-scale meta-regression", {
  data(dat.lehmann2018, package = "metadat")
  fit <- suppressWarnings(RoBMA(
    yi = yi, vi = vi, mods = ~ Preregistered, scale = ~ Preregistered, cluster = Full_Citation,
    data = dat.lehmann2018, measure = "SMD",
    chains = 2, sample = 1000, burnin = 500, adapt = 500,
    seed = 1, silent = TRUE
  ))
  fit <- suppressWarnings(add_loo(fit))
  save_fit("dat.lehmann2018_RoBMA_3lvl_mods_scale", fit, info = list(mods = c("Preregistered"), scale = c("Preregistered")))

  expect_s3_class(fit, "RoBMA")
  expect_equal(length(fit$priors$mods), 2)
  expect_equal(length(fit$priors$scale), 2)
  expect_equal(length(fit$priors$outcome$rho), 1)
})

context("Model fitting for bPEESE")

# Load common test helpers
source(testthat::test_path("common-functions.R"))
skip_on_cran()
skip_if_not_installed("metadat")
skip_if_not_installed("metafor")
skip_refit_if_cached("bPEESE")

### Uses examples from the metafor package
test_that("bPEESE fits PEESE metafor-reference model", {
  ### fit PEESE model
  data(dat.lehmann2018, package = "metadat")
  fit_PEESE.metafor <- metafor::rma(yi, vi, mods = ~ vi, data = dat.lehmann2018)

  # using RoBMA package
  fit.bPEESE <- bPEESE(yi, vi, data = dat.lehmann2018, measure = "SMD", seed = 1, silent = TRUE)
  fit.bPEESE <- add_marglik(fit.bPEESE)
  fit.bPEESE <- suppressWarnings(add_loo(fit.bPEESE))
  save_fit("dat.lehmann2018-PEESE", fit.bPEESE, info = list(metafor = fit_PEESE.metafor))

  expect_s3_class(fit.bPEESE, "bPEESE")
})

test_that("bPEESE fits negative-direction PEESE metafor-reference model", {
  ### fit PEESE model
  data(dat.lehmann2018, package = "metadat")
  dat.lehmann2018$yi <- -dat.lehmann2018$yi
  fit_PEESE.metafor <- metafor::rma(yi, vi, mods = ~ vi, data = dat.lehmann2018)

  # using RoBMA package
  fit.bPEESE <- bPEESE(yi, vi, data = dat.lehmann2018, measure = "SMD", seed = 1, silent = TRUE)
  fit.bPEESE <- add_marglik(fit.bPEESE)
  fit.bPEESE <- suppressWarnings(add_loo(fit.bPEESE))
  save_fit("dat.lehmann2018-PEESE_neg", fit.bPEESE, info = list(metafor = fit_PEESE.metafor))

  expect_s3_class(fit.bPEESE, "bPEESE")
})

test_that("bPEESE fits PEESE meta-regression model", {
  ### fit PEESE model
  data(dat.lehmann2018, package = "metadat")
  fit_PEESE.metafor <- metafor::rma(yi, vi, mods = ~ vi + Preregistered, data = dat.lehmann2018)

  # using RoBMA package
  fit.bPEESE <- bPEESE(yi, vi, mods = ~ Preregistered, data = dat.lehmann2018, measure = "SMD", seed = 1, silent = TRUE)
  fit.bPEESE <- add_marglik(fit.bPEESE)
  fit.bPEESE <- suppressWarnings(add_loo(fit.bPEESE))
  save_fit("dat.lehmann2018-PEESEreg", fit.bPEESE, info = list(metafor = fit_PEESE.metafor, mods = "Preregistered"))

  expect_s3_class(fit.bPEESE, "bPEESE")
})

context("Model fitting for bselmodel")

# Load common test helpers
source(testthat::test_path("common-functions.R"))
skip_on_cran()
skip_if_not_installed("metadat")
skip_if_not_installed("metafor")
skip_refit_if_cached("bselmodel")

### Uses examples from the metafor package
test_that("bselmodel fits one-step metafor-reference selection model", {
  ### fit selection model
  data(dat.lehmann2018, package = "metadat")
  fit_rma.metafor <- metafor::rma(yi, vi, data = dat.lehmann2018, method = "ML")
  fit_selmodel.metafor <- metafor::selmodel(fit_rma.metafor, type = "stepfun", alternative = "greater", steps = .025)

  # using RoBMA package
  fit.bselmodel <- bselmodel(yi, vi, data = dat.lehmann2018, measure = "SMD", seed = 1, silent = TRUE)
  fit.bselmodel <- add_marglik(fit.bselmodel)
  fit.bselmodel <- suppressWarnings(add_loo(fit.bselmodel))
  save_fit("dat.lehmann2018-3PSM", fit.bselmodel, info = list(metafor = fit_selmodel.metafor))

  expect_s3_class(fit.bselmodel, "bselmodel")
})

test_that("bselmodel fits two-step metafor-reference selection model", {
  ### fit selection model
  data(dat.lehmann2018, package = "metadat")
  fit_rma.metafor <- metafor::rma(yi, vi, data = dat.lehmann2018, method = "ML")
  fit_selmodel.metafor <- metafor::selmodel(fit_rma.metafor, type = "stepfun", alternative = "greater", steps = c(.025, 0.50))

  # using RoBMA package
  fit.bselmodel <- bselmodel(yi, vi, data = dat.lehmann2018, measure = "SMD", seed = 1, silent = TRUE, steps = c(0.025, 0.50))
  fit.bselmodel <- add_marglik(fit.bselmodel)
  fit.bselmodel <- suppressWarnings(add_loo(fit.bselmodel))
  save_fit("dat.lehmann2018-4PSM", fit.bselmodel, info = list(metafor = fit_selmodel.metafor))

  expect_s3_class(fit.bselmodel, "bselmodel")
})

test_that("bselmodel fits negative-direction metafor-reference selection model", {
  ### fit selection model
  data(dat.lehmann2018, package = "metadat")
  dat.lehmann2018$yi <- -dat.lehmann2018$yi
  fit_rma.metafor <- metafor::rma(yi, vi, data = dat.lehmann2018, method = "ML")
  fit_selmodel.metafor <- metafor::selmodel(fit_rma.metafor, type = "stepfun", alternative = "less", steps = .025)

  # using RoBMA package
  fit.bselmodel <- bselmodel(yi, vi, data = dat.lehmann2018, measure = "SMD", seed = 1, silent = TRUE)
  fit.bselmodel <- add_marglik(fit.bselmodel)
  fit.bselmodel <- suppressWarnings(add_loo(fit.bselmodel))
  save_fit("dat.lehmann2018-3PSM_neg", fit.bselmodel, info = list(metafor = fit_selmodel.metafor))

  expect_s3_class(fit.bselmodel, "bselmodel")
})

test_that("bselmodel fits selection meta-regression model", {
  ### fit selection model
  data(dat.lehmann2018, package = "metadat")
  fit_rma.metafor <- metafor::rma(yi, vi, mods = ~ Preregistered, data = dat.lehmann2018, method = "ML")
  fit_selmodel.metafor <- metafor::selmodel(fit_rma.metafor, type = "stepfun", steps = .025)

  # using RoBMA package
  fit.bselmodel <- bselmodel(yi, vi, mods = ~ Preregistered, data = dat.lehmann2018, measure = "SMD", seed = 1, silent = TRUE, steps = .025)
  fit.bselmodel <- add_marglik(fit.bselmodel)
  fit.bselmodel <- suppressWarnings(add_loo(fit.bselmodel))
  save_fit("dat.lehmann2018-3PSMreg", fit.bselmodel, info = list(metafor = fit_selmodel.metafor, mods = "Preregistered"))

  expect_s3_class(fit.bselmodel, "bselmodel")
})

context("Model fitting for bPET")

# Load common test helpers
source(testthat::test_path("common-functions.R"))
skip_on_cran()
skip_if_not_installed("metadat")
skip_if_not_installed("metafor")
skip_refit_if_cached("bPET")

### Uses examples from the metafor package
test_that("bPET fits PET metafor-reference model", {
  ### fit PET model
  data(dat.lehmann2018, package = "metadat")
  fit_PET.metafor <- metafor::rma(yi, vi, mods = ~ sqrt(vi), data = dat.lehmann2018)

  # using RoBMA package
  fit.bPET <- bPET(yi, vi, data = dat.lehmann2018, measure = "SMD", seed = 1, silent = TRUE)
  fit.bPET <- add_marglik(fit.bPET)
  fit.bPET <- suppressWarnings(add_loo(fit.bPET))
  save_fit("dat.lehmann2018-PET", fit.bPET, info = list(metafor = fit_PET.metafor))

  expect_s3_class(fit.bPET, "bPET")
})


test_that("bPET fits negative-direction PET metafor-reference model", {
  ### fit PET model
  data(dat.lehmann2018, package = "metadat")
  dat.lehmann2018$yi <- -dat.lehmann2018$yi
  fit_PET.metafor <- metafor::rma(yi, vi, mods = ~ sqrt(vi), data = dat.lehmann2018)

  # using RoBMA package
  fit.bPET <- bPET(yi, vi, data = dat.lehmann2018, measure = "SMD", seed = 1, silent = TRUE)
  fit.bPET <- add_marglik(fit.bPET)
  fit.bPET <- suppressWarnings(add_loo(fit.bPET))
  save_fit("dat.lehmann2018-PET_neg", fit.bPET, info = list(metafor = fit_PET.metafor))

  expect_s3_class(fit.bPET, "bPET")
})

test_that("bPET fits PET meta-regression model", {
  ### fit PET model
  data(dat.lehmann2018, package = "metadat")
  fit_PET.metafor <- metafor::rma(yi, vi, mods = ~ sqrt(vi) + Preregistered, data = dat.lehmann2018)

  # using RoBMA package
  fit.bPET <- bPET(yi, vi, mods = ~ Preregistered, data = dat.lehmann2018, measure = "SMD", seed = 1, silent = TRUE)
  fit.bPET <- add_marglik(fit.bPET)
  fit.bPET <- suppressWarnings(add_loo(fit.bPET))
  save_fit("dat.lehmann2018-PETreg", fit.bPET, info = list(metafor = fit_PET.metafor, mods = "Preregistered"))

  expect_s3_class(fit.bPET, "bPET")
})

context("Model fitting for brma.norm")

# Load common test helpers
source(testthat::test_path("common-functions.R"))
skip_on_cran()
skip_if_not_installed("metadat")
skip_if_not_installed("metafor")
skip_refit_if_cached("brma.norm")

### Uses examples from the metafor package
test_that("brma.norm fits metafor-reference normal models", {
  ### fit simple meta-analytic model
  data(dat.bcg, package = "metadat")
  dat <- metafor::escalc(measure = "RR", ai = tpos, bi = tneg, ci = cpos, di = cneg, data = dat.bcg)
  fit_simple.metafor <- metafor::rma(yi = yi, vi = vi, data = dat, method = "REML")

  # using RoBMA package
  fit_simple.brma <- brma(yi = yi, vi = vi, data = dat, measure = "RR", seed = 1, silent = TRUE)
  fit_simple.brma <- add_marglik(fit_simple.brma)
  fit_simple.brma <- add_loo(fit_simple.brma)
  save_fit("bcg_meta-analysis", fit_simple.brma, info = list(metafor = fit_simple.metafor))
  expect_s3_class(fit_simple.brma, "brma.norm")


  ### fit meta-regression (continuous predictors)
  fit_mods.metafor <- metafor::rma(yi, vi, mods = ~ ablat + year, data = dat)

  # using RoBMA package (rescale priors to reduce the shrinkage)
  fit_mods.brma <- brma(yi, vi, mods = ~ ablat + year, data = dat, measure = "RR", seed = 1, silent = TRUE)
  fit_mods.brma <- add_marglik(fit_mods.brma)
  fit_mods.brma <- suppressWarnings(add_loo(fit_mods.brma))
  save_fit("bcg_meta-regression", fit_mods.brma, info = list(mods = c("ablat", "year"), metafor = fit_mods.metafor))
  expect_s3_class(fit_mods.brma, "brma.norm")


  ### fit meta-regression (factor predictor)
  fit_mods2.metafor <- metafor::rma(yi, vi, mods = ~ alloc, data = dat)

  # using RoBMA package (dummy coding)
  fit_mods2.brma <- brma(yi, vi, mods = ~ alloc, data = dat, measure = "RR", seed = 1, silent = TRUE)
  fit_mods2.brma <- add_marglik(fit_mods2.brma)
  fit_mods2.brma <- suppressWarnings(add_loo(fit_mods2.brma))
  save_fit("bcg_meta-regression2", fit_mods2.brma, info = list(mods = c("alloc"), metafor = fit_mods2.metafor))
  expect_s3_class(fit_mods2.brma, "brma.norm")

  # using RoBMA package (meandif coding)
  fit_mods2b.brma <- brma(
    yi, vi, mods = ~ alloc, data = dat, measure = "RR",
    chains = 2, sample = 1000, burnin = 500, adapt = 500,
    seed = 1, silent = TRUE, rescale_priors = 2,
    set_contrast_factor_predictors = "meandif"
  )
  fit_mods2b.brma <- add_marglik(fit_mods2b.brma)
  fit_mods2b.brma <- suppressWarnings(add_loo(fit_mods2b.brma))
  save_fit("bcg_meta-regression2b", fit_mods2b.brma, info = list(mods = c("alloc"), metafor = NULL))
  expect_s3_class(fit_mods2b.brma, "brma.norm")


  ### fit meta-regression with interaction (fact * cont)
  fit_mods3.metafor <- metafor::rma(yi, vi, mods = ~ alloc * year, data = dat)

  # using RoBMA package (increase scale because of shrinkage, dummy coding)
  fit_mods3.brma <- brma(yi, vi, mods = ~ alloc * year, data = dat, measure = "RR", seed = 1, silent = TRUE)
  fit_mods3.brma <- add_marglik(fit_mods3.brma)
  fit_mods3.brma <- suppressWarnings(add_loo(fit_mods3.brma))
  save_fit("bcg_meta-regression3", fit_mods3.brma, info = list(mods = c("alloc", "year", "alloc:year"), metafor = fit_mods3.metafor))
  expect_s3_class(fit_mods3.brma, "brma.norm")

  # using RoBMA package (meandif coding)
  fit_mods3b.brma <- brma(
    yi, vi, mods = ~ alloc * year, data = dat, measure = "RR",
    chains = 2, sample = 1000, burnin = 500, adapt = 500,
    seed = 1, silent = TRUE,
    set_contrast_factor_predictors = "meandif"
  )
  fit_mods3b.brma <- add_marglik(fit_mods3b.brma)
  fit_mods3b.brma <- suppressWarnings(add_loo(fit_mods3b.brma))
  save_fit("bcg_meta-regression3b", fit_mods3b.brma, info = list(mods = c("alloc", "year", "alloc:year"), metafor = NULL))
  expect_s3_class(fit_mods3b.brma, "brma.norm")


  ### fit meta-regression with interaction (fact * fact)
  dat$year_before1969 <- factor(dat$year < 1969)
  fit_mods4.metafor <- metafor::rma(yi, vi, mods = ~ alloc * year_before1969, data = dat)

  # using RoBMA package (dummy coding)
  fit_mods4.brma <- brma(yi, vi, mods = ~ alloc * year_before1969, data = dat, measure = "RR", seed = 1, silent = TRUE)
  fit_mods4.brma <- add_marglik(fit_mods4.brma)
  fit_mods4.brma <- suppressWarnings(add_loo(fit_mods4.brma))
  save_fit("bcg_meta-regression4", fit_mods4.brma, info = list(mods = c("alloc", "year_before1969", "alloc:year_before1969"), metafor = fit_mods4.metafor))
  expect_s3_class(fit_mods4.brma, "brma.norm")

  # using RoBMA package (increase scale because of shrinkage, meandif coding)
  fit_mods4b.brma <- brma(
    yi, vi, mods = ~ alloc * year_before1969, data = dat, measure = "RR",
    chains = 2, sample = 1000, burnin = 500, adapt = 500,
    seed = 1, silent = TRUE,
    set_contrast_factor_predictors = "meandif"
  )
  fit_mods4b.brma <- add_marglik(fit_mods4b.brma)
  fit_mods4b.brma <- suppressWarnings(add_loo(fit_mods4b.brma))
  save_fit("bcg_meta-regression4b", fit_mods4b.brma, info = list(mods = c("alloc", "year_before1969", "alloc:year_before1969"), metafor = NULL))
  expect_s3_class(fit_mods4b.brma, "brma.norm")
})

test_that("brma.norm fits location-scale metafor-reference model", {
  ### fit location-scale model
  data(dat.bangertdrowns2004, package = "metadat")
  dat <- dat.bangertdrowns2004
  dat$ni100 <- dat$ni / 100
  dat$meta <- as.factor(dat$meta)
  dat$imag <- as.factor(dat$imag)
  fit_scale.metafor <- suppressWarnings(metafor::rma(yi, vi, mods = ~ meta + ni100, scale = ~ni100, data = dat, method = "REML"))

  # using RoBMA package (using wide priors for scale to remove shrinkage -- the likelihood is very wide)
  fit_scale.brma <- suppressWarnings(brma(yi, vi, mods = ~ meta + ni100, scale = ~ni100, data = dat, measure = "SMD", seed = 1, silent = TRUE))
  fit_scale.brma <- add_marglik(fit_scale.brma)
  fit_scale.brma <- add_loo(fit_scale.brma)
  save_fit("bangertdrowns2004_location-scale", fit_scale.brma, info = list(mods = c("meta", "ni100"), scale = "ni100", metafor = fit_scale.metafor))
  expect_s3_class(fit_scale.brma, "brma.norm")
})

test_that("brma.norm fits 3-level metafor-reference models", {
  ### fit 3lvl model
  data(dat.konstantopoulos2011, package = "metadat")
  fit_3lvl.metafor <- metafor::rma.mv(yi, vi, random = ~ school | district, data = dat.konstantopoulos2011)

  # using RoBMA package
  fit_3lvl.brma <- brma(yi, vi, cluster = district, data = dat.konstantopoulos2011, measure = "SMD", seed = 1, silent = TRUE)
  fit_3lvl.brma <- add_marglik(fit_3lvl.brma)
  fit_3lvl.brma <- suppressWarnings(add_loo(fit_3lvl.brma))
  save_fit("konstantopoulos2011_3lvl", fit_3lvl.brma, info = list(metafor = fit_3lvl.metafor))
  expect_s3_class(fit_3lvl.brma, "brma.norm")

  ### fit 3lvl model with a predictor
  data(dat.konstantopoulos2011, package = "metadat")
  fit_3lvl2.metafor <- metafor::rma.mv(yi, vi, mods = ~ vi, random = ~ school | district, data = dat.konstantopoulos2011)

  # using RoBMA package
  fit_3lvl2.brma <- brma(yi, vi, mods = ~ vi, cluster = district, data = dat.konstantopoulos2011, measure = "SMD", seed = 1, silent = TRUE)
  fit_3lvl2.brma <- add_marglik(fit_3lvl2.brma)
  fit_3lvl2.brma <- suppressWarnings(add_loo(fit_3lvl2.brma))
  save_fit("konstantopoulos2011_3lvl2", fit_3lvl2.brma, info = list(metafor = fit_3lvl2.metafor))
  expect_s3_class(fit_3lvl2.brma, "brma.norm")
})

context("Model fitting for BMA.glmm")

# Load common test helpers
source(testthat::test_path("common-functions.R"))
skip_on_cran()
skip_if_not_installed("metadat")
skip_refit_if_cached("BMA.glmm")


test_that("BMA.glmm fits binomial model (OR)", {
  data(dat.bcg, package = "metadat")

  fit <- BMA.glmm(
    ai = tpos, bi = tneg, ci = cpos, di = cneg,
    data = dat.bcg, measure = "OR",
    chains = 2, sample = 1000, burnin = 500, adapt = 500,
    seed = 1, silent = TRUE
  )
  fit <- suppressWarnings(add_loo(fit))
  save_fit("bcg_BMA.glmm", fit)

  expect_s3_class(fit, "BMA.glmm")
  expect_true(BayesTools::is.prior.mixture(fit$priors$outcome$mu))
  expect_true(BayesTools::is.prior.mixture(fit$priors$outcome$tau))
  expect_null(fit$priors$outcome$bias)
  expect_no_error(summary(fit))
})

test_that("BMA.glmm fits Poisson model (IRR)", {
  data(dat.nielweise2008, package = "metadat")

  fit <- BMA.glmm(
    x1i = x1i, t1i = t1i, x2i = x2i, t2i = t2i,
    data = dat.nielweise2008, measure = "IRR",
    chains = 2, sample = 1000, burnin = 500, adapt = 500,
    seed = 1, silent = TRUE
  )
  fit <- suppressWarnings(add_loo(fit))
  save_fit("nielweise2008_BMA.glmm", fit)

  expect_s3_class(fit, "BMA.glmm")
  expect_true(BayesTools::is.prior.mixture(fit$priors$outcome$mu))
  expect_true(BayesTools::is.prior.mixture(fit$priors$outcome$tau))
  expect_null(fit$priors$outcome$bias)
  expect_no_error(summary(fit))
})


test_that("BMA.glmm handles custom priors", {
  data(dat.bcg, package = "metadat")

  fit <- BMA.glmm(
    ai = tpos, bi = tneg, ci = cpos, di = cneg,
    data = dat.bcg, measure = "OR",
    prior_effect = prior("normal", list(mean = 0, sd = 1)),
    prior_effect_null = prior("spike", list(location = 0)),
    prior_heterogeneity = prior("normal", list(mean = 0, sd = 0.5), truncation = list(lower = 0)),
    prior_heterogeneity_null = NULL,  # no null hypothesis for heterogeneity
    chains = 2, sample = 1000, burnin = 500, adapt = 500,
    seed = 1, silent = TRUE
  )
  fit <- suppressWarnings(add_loo(fit))
  save_fit("bcg_BMA.glmm_custom", fit)

  expect_true(BayesTools::is.prior.mixture(fit$priors$outcome$mu))
  expect_true(BayesTools::is.prior.mixture(fit$priors$outcome$tau))
  expect_equal(length(fit$priors$outcome$tau), 1)
})

test_that("BMA.glmm handles 3lvl location-scale meta-regression", {
  data(dat.bcg, package = "metadat")

  fit <- BMA.glmm(
    ai = tpos, bi = tneg, ci = cpos, di = cneg,
    mods = ~ year, scale = ~ year, cluster = alloc,
    data = dat.bcg, measure = "OR",
    prior_effect = prior("normal", list(mean = 0, sd = 1)),
    prior_effect_null = prior("spike", list(location = 0)),
    prior_heterogeneity = prior("normal", list(mean = 0, sd = 0.5), truncation = list(lower = 0)),
    prior_heterogeneity_null = NULL,  # no null hypothesis for heterogeneity
    chains = 2, sample = 1000, burnin = 500, adapt = 500,
    seed = 1, silent = TRUE
  )
  fit <- suppressWarnings(add_loo(fit))
  save_fit("bcg_BMA.glmm_3lvl_location_scale", fit, info = list(mods = "year", scale = "year"))

  expect_equal(length(fit$priors$outcome$rho), 1)
  expect_equal(length(fit$priors$mods), 2)
  expect_equal(length(fit$priors$scale), 2)
})

# ============================================================================ #
# brma.summary_heterogeneity.R
# ============================================================================ #
#
# This file contains the summary_heterogeneity method for brma model fits.
# It computes absolute and relative heterogeneity measures:
# - tau, tau^2: Absolute heterogeneity (SD and variance)
# - I^2: Percentage of total variance due to heterogeneity
# - H^2: Ratio of total variance to sampling variance
#
# For multilevel (3-level) models, heterogeneity is partitioned into:
# - tau [within]: Estimate-level heterogeneity
# - tau [between]: Cluster-level heterogeneity
# - I^2 [within], I^2 [between]: Partitioned relative heterogeneity
#
# Formulas follow metafor documentation and Higgins & Thompson (2002).
#
# ============================================================================ #


# ---------------------------------------------------------------------------- #
# summary_heterogeneity generic
# ---------------------------------------------------------------------------- #

#' @title Summary of Heterogeneity
#'
#' @description Computes the absolute heterogeneity (tau, tau^2) and
#' relative measures of heterogeneity (I^2, H^2) for a fitted model.
#'
#' @param object a fitted model object
#' @param ... additional arguments passed to methods
#'
#' @return Method-specific return value containing heterogeneity estimates.
#'
#' @seealso [pooled_heterogeneity()], [summary.brma()]
#' @export
summary_heterogeneity <- function(object, ...) {

  UseMethod("summary_heterogeneity")
}


# ---------------------------------------------------------------------------- #
# summary_heterogeneity.brma
# ---------------------------------------------------------------------------- #

#' @title Summary of Heterogeneity for brma Objects
#'
#' @description Computes the absolute heterogeneity (tau, tau^2) and
#' relative measures of heterogeneity (I^2, H^2) for a fitted brma object.
#'
#' @param object a fitted brma object
#' @param probs quantiles of the posterior distribution to be displayed.
#' Defaults to \code{c(.025, .975)} for 95% credible intervals.
#' @param ... additional arguments (currently ignored)
#'
#' @details
#' For standard (2-level) random-effects models, the function reports:
#' \itemize{
#'   \item \code{tau}: Between-study standard deviation
#'   \item \code{tau2}: Between-study variance
#'   \item \code{I2}: Percentage of total variance due to heterogeneity
#'   \item \code{H2}: Ratio of total to sampling variance
#' }
#'
#' For multilevel (3-level) models with nested effects, the function additionally
#' partitions heterogeneity into estimate-level and cluster-level components:
#' \itemize{
#'   \item \code{rho}: Proportion of heterogeneity variance allocated to clusters
#'   \item \code{tau [within]}: Estimate-level standard deviation
#'   \item \code{tau [between]}: Cluster-level standard deviation
#'   \item \code{tau2 [within]}: Estimate-level variance
#'   \item \code{tau2 [between]}: Cluster-level variance
#'   \item \code{I2 [within]}: Percentage of variance due to estimate-level heterogeneity
#'   \item \code{I2 [between]}: Percentage of variance due to cluster-level heterogeneity
#' }
#'
#' For location-scale models, \code{tau2} aggregates the observation-specific
#' heterogeneity variances \eqn{\tau_i^2}; the corresponding \code{tau} summary
#' is the square root of this aggregate variance. The relative \eqn{I^2} and
#' \eqn{H^2} measures average the observation-specific indices.
#'
#' The I^2 and H^2 statistics are computed following the metafor package
#' implementation, using the "typical" sampling variance formula from
#' \insertCite{higgins2002quantifying;textual}{RoBMA}. For multilevel models,
#' the partitioned I^2 follows the approach described in the metafor documentation.
#'
#' @return A list of class \code{summary_heterogeneity.brma} containing:
#' \itemize{
#'   \item \code{estimates}: A \code{BayesTools_table} with heterogeneity statistics
#' }
#'
#' @examples \dontrun{
#' if (requireNamespace("metadat", quietly = TRUE)) {
#'   data(dat.lehmann2018, package = "metadat")
#'   fit <- brma(
#'     yi      = yi,
#'     vi      = vi,
#'     data    = dat.lehmann2018,
#'     measure = "SMD",
#'     seed    = 1,
#'     silent  = TRUE
#'   )
#'
#'   summary_heterogeneity(fit)
#' }
#' }
#'
#' @references
#' \insertAllCited{}
#'
#' @seealso [pooled_heterogeneity()], [summary.brma()]
#' @export
summary_heterogeneity.brma <- function(object, probs = c(.025, .975), ...) {

  # input validation
  BayesTools::check_real(probs, "probs", allow_NULL = TRUE, check_length = 0)

  # extract model characteristics
  is_multilevel <- .is_multilevel(object)
  is_scale      <- .is_scale(object)

  # extract sampling variances using helper
  vi <- .outcome_data_vi(object)
  K  <- length(vi)

  # compute typical sampling variance (v_tilde)
  # following metafor formula (Higgins & Thompson, 2002, eq. 9)
  # using the generalized formula with projection matrix for meta-regression
  X <- .get_model_matrix(object)
  p <- ncol(X)
  W <- diag(1/vi, nrow = K)
  P <- W - W %*% X %*% solve(t(X) %*% W %*% X) %*% t(X) %*% W

  v_tilde <- (K - p) / sum(diag(P))

  posterior_samples <- .get_posterior_samples(object[["fit"]])

  # extract tau samples using the evaluate helper
  tau_result <- .evaluate.brma.tau(
    fit               = object[["fit"]],
    scale_data        = object[["data"]][["scale"]],
    scale_formula     = if (is_scale) .create_fit_formula_list(data = object[["data"]], "scale") else NULL,
    scale_priors      = object[["priors"]][["scale"]],
    is_scale          = is_scale,
    is_multilevel     = is_multilevel,
    K                 = K,
    posterior_samples = posterior_samples
  )

  samples_list <- .summary_heterogeneity_samples(
    tau_within_samples  = tau_result[["tau_within"]],
    tau_between_samples = tau_result[["tau_between"]],
    rho_samples         = if (is_multilevel) posterior_samples[, "rho"] else NULL,
    v_tilde             = v_tilde,
    is_multilevel       = is_multilevel
  )

  # generate summary table
  estimates <- BayesTools::ensemble_estimates_table(
    samples    = samples_list,
    parameters = names(samples_list),
    probs      = probs,
    title      = "Heterogeneity Estimates:"
  )

  # create output object
  output <- list(
    estimates = estimates
  )

  class(output) <- "summary_heterogeneity.brma"

  return(output)
}


# ---------------------------------------------------------------------------- #
# .summary_heterogeneity_samples
# ---------------------------------------------------------------------------- #
#
# Compute heterogeneity summaries from observation-level tau samples.
#
.summary_heterogeneity_samples <- function(tau_within_samples, tau_between_samples,
                                           v_tilde, is_multilevel,
                                           rho_samples = NULL) {

  if (is.null(tau_between_samples)) {
    tau_between_samples <- matrix(0, nrow = nrow(tau_within_samples),
                                  ncol = ncol(tau_within_samples))
  }

  if (!is.matrix(tau_within_samples) || !is.matrix(tau_between_samples)) {
    stop("'tau_within_samples' and 'tau_between_samples' must be matrices.",
         call. = FALSE)
  }
  if (!identical(dim(tau_within_samples), dim(tau_between_samples))) {
    stop("'tau_within_samples' and 'tau_between_samples' must have matching dimensions.",
         call. = FALSE)
  }
  if (!is.numeric(v_tilde) || length(v_tilde) != 1 || !is.finite(v_tilde) || v_tilde <= 0) {
    stop("'v_tilde' must be a positive finite number.", call. = FALSE)
  }
  if (is_multilevel && !is.null(rho_samples) &&
      (!is.numeric(rho_samples) || length(rho_samples) != nrow(tau_within_samples))) {
    stop("'rho_samples' must be a numeric vector matching posterior sample rows.",
         call. = FALSE)
  }

  sigma2_within_matrix  <- tau_within_samples^2
  sigma2_between_matrix <- tau_between_samples^2
  sigma2_total_matrix   <- sigma2_within_matrix + sigma2_between_matrix

  sigma2_within  <- rowMeans(sigma2_within_matrix)
  sigma2_between <- rowMeans(sigma2_between_matrix)
  sigma2_total   <- sigma2_within + sigma2_between

  denominator_matrix <- sigma2_total_matrix + v_tilde

  I2_total <- rowMeans(100 * sigma2_total_matrix / denominator_matrix)
  H2       <- rowMeans(denominator_matrix / v_tilde)

  if (is_multilevel) {

    I2_within  <- rowMeans(100 * sigma2_within_matrix / denominator_matrix)
    I2_between <- rowMeans(100 * sigma2_between_matrix / denominator_matrix)
    if (is.null(rho_samples)) {
      rho_samples <- ifelse(sigma2_total > 0, sigma2_between / sigma2_total, 0)
    } else {
      rho_samples <- pmin(pmax(rho_samples, 0), 1)
    }

    return(list(
      "tau"             = sqrt(sigma2_total),
      "rho"             = rho_samples,
      "tau [within]"    = sqrt(sigma2_within),
      "tau [between]"   = sqrt(sigma2_between),
      "tau2"            = sigma2_total,
      "tau2 [within]"   = sigma2_within,
      "tau2 [between]"  = sigma2_between,
      "I2"              = I2_total,
      "I2 [within]"     = I2_within,
      "I2 [between]"    = I2_between,
      "H2"              = H2
    ))
  }

  return(list(
    "tau"  = sqrt(sigma2_total),
    "tau2" = sigma2_total,
    "I2"   = I2_total,
    "H2"   = H2
  ))
}


# ---------------------------------------------------------------------------- #
# print method for summary.brma_heterogeneity
# ---------------------------------------------------------------------------- #

#' @title Print Summary of Heterogeneity
#'
#' @description Prints the heterogeneity summary table.
#'
#' @param x a \code{summary_heterogeneity.brma} object
#' @param ... additional arguments (currently ignored)
#'
#' @return Returns \code{x} invisibly.
#'
#' @export
print.summary_heterogeneity.brma <- function(x, ...) {

  cat("\n")
  print(x$estimates)
  cat("\n")

  return(invisible(x))
}

context("Summary heterogeneity helpers")

test_that("scale heterogeneity summaries aggregate variances before tau", {

  tau_within <- matrix(
    c(1, 3,
      2, 4),
    nrow = 2,
    byrow = TRUE
  )

  samples <- RoBMA:::.summary_heterogeneity_samples(
    tau_within_samples  = tau_within,
    tau_between_samples = matrix(0, nrow = 2, ncol = 2),
    v_tilde             = 1,
    is_multilevel       = FALSE
  )

  expected_tau2 <- rowMeans(tau_within^2)
  old_tau2      <- rowMeans(tau_within)^2

  expect_equal(samples[["tau2"]], expected_tau2)
  expect_equal(samples[["tau"]], sqrt(expected_tau2))
  expect_false(isTRUE(all.equal(samples[["tau2"]], old_tau2)))
  expect_equal(samples[["I2"]], rowMeans(100 * tau_within^2 / (tau_within^2 + 1)))
  expect_equal(samples[["H2"]], rowMeans(tau_within^2 + 1))
})

test_that("multilevel scale heterogeneity partitions variance and I2", {

  tau_within <- matrix(
    c(1, 2,
      3, 4),
    nrow = 2,
    byrow = TRUE
  )
  tau_between <- matrix(
    c(2, 1,
      1, 3),
    nrow = 2,
    byrow = TRUE
  )

  samples <- RoBMA:::.summary_heterogeneity_samples(
    tau_within_samples  = tau_within,
    tau_between_samples = tau_between,
    rho_samples         = c(0.8, 0.2),
    v_tilde             = 2,
    is_multilevel       = TRUE
  )

  sigma2_within  <- tau_within^2
  sigma2_between <- tau_between^2
  sigma2_total   <- sigma2_within + sigma2_between
  denominator    <- sigma2_total + 2

  expect_equal(samples[["tau2 [within]"]], rowMeans(sigma2_within))
  expect_equal(samples[["tau2 [between]"]], rowMeans(sigma2_between))
  expect_equal(samples[["tau2"]], rowMeans(sigma2_total))
  expect_equal(samples[["rho"]], c(0.8, 0.2))
  expect_equal(samples[["I2"]], rowMeans(100 * sigma2_total / denominator))
  expect_equal(samples[["I2 [within]"]], rowMeans(100 * sigma2_within / denominator))
  expect_equal(samples[["I2 [between]"]], rowMeans(100 * sigma2_between / denominator))
  expect_equal(samples[["I2"]], samples[["I2 [within]"]] + samples[["I2 [between]"]])
})

context("Summary heterogeneity")

source(testthat::test_path("common-functions.R"))
source(testthat::test_path("helper-contracts.R"))
source(testthat::test_path("helper-test-matrix.R"))
source(testthat::test_path("helper-metafor.R"))

REFERENCE_DIR <<- testthat::test_path("..", "results", "summary_heterogeneity")

skip_if_no_fits()
skip_if_not_installed("metafor")

fit_names <- list_fits()
fits      <- lazy_fits(fit_names, validate = FALSE)
info      <- lazy_infos(fit_names, validate = FALSE)

for_each_case(summary_heterogeneity_cases(), function(case) {
  test_that_case("Heterogeneity summary matches metafor", case, {
    expect_summary_heterogeneity_matches_metafor(case)
  })
})

test_that("Heterogeneity for model-averaged 2-level models is summarized", {

  model_averaged_classes <- c("BMA.norm", "BMA.glmm", "RoBMA")
  model_names <- setdiff(
    catalog_fits(class = model_averaged_classes),
    catalog_fits(class = model_averaged_classes, feature = "multilevel")
  )
  skip_if_missing_fits(model_names)

  expected_rows <- c("tau", "tau2", "I2", "H2")

  for (name in model_names) {
    expect_summary_heterogeneity_structure(
      summary_heterogeneity(fits[[name]]),
      expected_rows,
      name
    )
  }
})

test_that("Heterogeneity for model-averaged 3-level models is partitioned", {

  model_averaged_classes <- c("BMA.norm", "BMA.glmm", "RoBMA")
  model_names <- catalog_fits(
    class   = model_averaged_classes,
    feature = "multilevel"
  )
  skip_if_missing_fits(model_names)

  expected_rows <- c(
    "tau",
    "rho",
    "tau [within]",
    "tau [between]",
    "tau2",
    "tau2 [within]",
    "tau2 [between]",
    "I2",
    "I2 [within]",
    "I2 [between]",
    "H2"
  )

  for (name in model_names) {
    heterogeneity <- summary_heterogeneity(fits[[name]])
    expect_summary_heterogeneity_structure(heterogeneity, expected_rows, name)
    expect_equal(
      heterogeneity$estimates["tau2", "Mean"],
      heterogeneity$estimates["tau2 [within]", "Mean"] +
        heterogeneity$estimates["tau2 [between]", "Mean"],
      tolerance = 0.01,
      info      = paste0("tau2 partitions sum to total for '", name, "'")
    )
    expect_equal(
      heterogeneity$estimates["I2", "Mean"],
      heterogeneity$estimates["I2 [within]", "Mean"] +
        heterogeneity$estimates["I2 [between]", "Mean"],
      tolerance = 0.01,
      info      = paste0("I2 partitions sum to total for '", name, "'")
    )
  }
})

test_that("summary_heterogeneity print keeps leading blank line", {

  skip_if_missing_fits("bcg_meta-analysis")

  output <- capture.output(print(summary_heterogeneity(fits[["bcg_meta-analysis"]])))

  expect_equal(output[[1]], "")
})

if (!exists("skip_if_no_vdiffr_snapshots", mode = "function")) {
  source(testthat::test_path("common-functions.R"))
}

expect_vdiffr_snapshot <- function(title, fig, ...) {

  skip_if_no_vdiffr_snapshots()
  testthat::skip_if_not_installed("vdiffr")
  vdiffr::expect_doppelganger(title, fig, ...)
}

expect_brma_plot_snapshot <- function(name, plot) {

  expect_vdiffr_snapshot(name, plot)
}

.with_temp_plot_device <- function(expr) {

  file <- tempfile(fileext = ".png")
  grDevices::png(filename = file)
  on.exit({
    grDevices::dev.off()
    unlink(file)
  }, add = TRUE)

  force(expr)
}

.is_ggplot <- function(x) {

  inherits(x, "ggplot")
}

ggplot_geom_layer_data <- function(plot, geom_class, index = 1L) {

  layer_index <- which(vapply(
    plot[["layers"]],
    function(layer) inherits(layer[["geom"]], geom_class),
    logical(1)
  ))

  if (length(layer_index) < index) {
    stop("Missing ggplot layer for geom class '", geom_class, "'.", call. = FALSE)
  }

  return(plot[["layers"]][[layer_index[[index]]]][["data"]])
}

expect_side_by_side_plot <- function(name, metafor, brma) {

  expect_vdiffr_snapshot(name, function() {
    old_par <- graphics::par(mfrow = c(1, 2))
    on.exit(graphics::par(old_par), add = TRUE)
    metafor()
    brma()
  })
}

context("Q-Q normal plot")

# Load common test helpers
source(testthat::test_path("common-functions.R"))
source(testthat::test_path("helper-test-matrix.R"))
source(testthat::test_path("helper-visuals.R"))

# list cached fits lazily
skip_if_no_fits()
skip_if_not_installed("metafor")
fit_names <- list_fits()
fits      <- lazy_fits(fit_names, validate = FALSE)
info      <- lazy_infos(fit_names, validate = FALSE)


# ============================================================================ #
# Test: Simple Meta-Analysis Q-Q Plot
# ============================================================================ #

test_that("Q-Q plot for simple meta-analysis matches metafor structure", {

  name        <- "bcg_meta-analysis"
  fit_metafor <- info[[name]][["metafor"]]
  fit_brma    <- fits[[name]]
  set.seed(1)

  # --------------------------------------------------
  # Visual comparison: side-by-side (rstandard for comparability)
  # --------------------------------------------------

  expect_vdiffr_snapshot("qqnorm_simple_comparison_rstandard", function() {
    oldpar <- graphics::par(no.readonly = TRUE)
    on.exit(graphics::par(mfrow = oldpar[["mfrow"]], mar = oldpar[["mar"]]))
    par(mfrow = c(1, 2), mar = c(4, 4, 2, 1))
    qqnorm(fit_metafor, main = "metafor", type = "rstandard", ylim = c(-3, 3))
    qqnorm(fit_brma, plot_type = "base", main = "brma", type = "rstandard", ylim = c(-3, 3))
  })

  # --------------------------------------------------
  # Default (rstudent) - brma only
  # --------------------------------------------------

  expect_vdiffr_snapshot("qqnorm_simple_rstudent_base", function() {
    qqnorm(fit_brma, plot_type = "base")
  })

  expect_vdiffr_snapshot(
    "qqnorm_simple_rstudent_ggplot",
    qqnorm(fit_brma, plot_type = "ggplot")
  )
})

# ============================================================================ #
# Test: Meta-Regression Q-Q Plot
# ============================================================================ #

test_that("Q-Q plot for meta-regression matches metafor residual quantiles", {

  name        <- "bcg_meta-regression"
  fit_metafor <- info[[name]][["metafor"]]
  fit_brma    <- fits[[name]]
  set.seed(1)

  # --------------------------------------------------
  # Visual comparison: side-by-side (rstandard for comparability)
  # --------------------------------------------------

  expect_vdiffr_snapshot("qqnorm_regression_comparison_rstandard", function() {
    oldpar <- graphics::par(no.readonly = TRUE)
    on.exit(graphics::par(mfrow = oldpar[["mfrow"]], mar = oldpar[["mar"]]))
    par(mfrow = c(1, 2), mar = c(4, 4, 2, 1))
    qqnorm(fit_metafor, main = "metafor", type = "rstandard", ylim = c(-3, 3))
    qqnorm(fit_brma, plot_type = "base", main = "brma", type = "rstandard", ylim = c(-3, 3))
  })

  expect_vdiffr_snapshot(
    "qqnorm_regression_rstudent_ggplot",
    suppressWarnings(qqnorm(fit_brma, plot_type = "ggplot"))
  )
})

test_that("Q-Q plot for interaction meta-regression renders residual quantiles", {

  skip_if_not_full_visuals("Interaction Q-Q variants duplicate the core meta-regression visual.")

  name        <- "bcg_meta-regression4"
  fit_metafor <- info[[name]][["metafor"]]
  fit_brma    <- fits[[name]]
  set.seed(1)

  # --------------------------------------------------
  # Visual comparison: side-by-side (rstandard for comparability)
  # --------------------------------------------------

  expect_vdiffr_snapshot("qqnorm_regression4_comparison_rstandard", function() {
    oldpar <- graphics::par(no.readonly = TRUE)
    on.exit(graphics::par(mfrow = oldpar[["mfrow"]], mar = oldpar[["mar"]]))
    par(mfrow = c(1, 2), mar = c(4, 4, 2, 1))
    qqnorm(fit_metafor, main = "metafor", type = "rstandard", ylim = c(-3, 3))
    qqnorm(fit_brma, plot_type = "base", main = "brma", type = "rstandard", ylim = c(-3, 3))
  })

  expect_vdiffr_snapshot(
    "qqnorm_regression4_rstudent_ggplot",
    suppressWarnings(qqnorm(fit_brma, plot_type = "ggplot"))
  )
})

# ============================================================================ #
# Test: Location-Scale Model Q-Q Plot
# ============================================================================ #

test_that("Q-Q plot for location-scale model renders residual quantiles", {

  skip_if_not_full_visuals("Location-scale Q-Q variants are gallery coverage.")

  name     <- "bangertdrowns2004_location-scale"
  fit_brma <- fits[[name]]
  set.seed(1)

  expect_vdiffr_snapshot("qqnorm_scale_rstandard_base", function() {
    qqnorm(fit_brma, plot_type = "base", type = "rstandard")
  })

  expect_vdiffr_snapshot(
    "qqnorm_scale_rstudent_ggplot",
    suppressWarnings(qqnorm(fit_brma, plot_type = "ggplot"))
  )
})

# ============================================================================ #
# Test: 3-Level Model Q-Q Plot
# ============================================================================ #

test_that("Q-Q plot for 3-level model renders residual quantiles", {

  name     <- "konstantopoulos2011_3lvl"
  fit_brma <- fits[[name]]
  set.seed(1)

  expect_vdiffr_snapshot("qqnorm_3lvl_base", function() {
    suppressWarnings(qqnorm(fit_brma, plot_type = "base"))
  })

  expect_vdiffr_snapshot(
    "qqnorm_3lvl_ggplot",
    suppressWarnings(qqnorm(fit_brma, plot_type = "ggplot"))
  )
})

test_that("Q-Q plot for 3-level meta-regression renders residual quantiles", {

  skip_if_not_full_visuals("3-level meta-regression duplicates the default multilevel Q-Q visual.")

  name     <- "konstantopoulos2011_3lvl2"
  fit_brma <- fits[[name]]
  set.seed(1)

  expect_vdiffr_snapshot("qqnorm_3lvl2_base", function() {
    suppressWarnings(qqnorm(fit_brma, plot_type = "base"))
  })

  expect_vdiffr_snapshot(
    "qqnorm_3lvl2_ggplot",
    suppressWarnings(qqnorm(fit_brma, plot_type = "ggplot"))
  )
})

# ============================================================================ #
# Test: GLMM Model Q-Q Plot
# ============================================================================ #

test_that("Q-Q plot for GLMM model renders residual quantiles", {

  name     <- "nielweise2008_glmm"
  fit_brma <- fits[[name]]
  set.seed(1)

  # rstudent only (rstandard not available for GLMM)
  expect_vdiffr_snapshot("qqnorm_glmm_base", function() {
    suppressWarnings(qqnorm(fit_brma, plot_type = "base"))
  })

  expect_vdiffr_snapshot(
    "qqnorm_glmm_ggplot",
    suppressWarnings(qqnorm(fit_brma, plot_type = "ggplot"))
  )

  # rstandard should error for GLMM models
  expect_error(
    qqnorm(fit_brma, type = "rstandard"),
    info = "rstandard residuals are rejected for GLMM models"
  )
})

test_that("Q-Q plot for GLMM meta-regression renders residual quantiles", {

  skip_if_not_full_visuals("GLMM meta-regression duplicates the default GLMM Q-Q visual.")

  name     <- "bcg_glmm_reg"
  fit_brma <- fits[[name]]
  set.seed(1)

  # rstudent only (rstandard not available for GLMM)
  expect_vdiffr_snapshot("qqnorm_glmm_reg_base", function() {
    suppressWarnings(qqnorm(fit_brma, plot_type = "base"))
  })

  expect_vdiffr_snapshot(
    "qqnorm_glmm_reg_ggplot",
    suppressWarnings(qqnorm(fit_brma, plot_type = "ggplot"))
  )

  # rstandard should error for GLMM models
  expect_error(
    qqnorm(fit_brma, type = "rstandard"),
    info = "rstandard residuals are rejected for GLMM models"
  )
})

# ============================================================================ #
# Test: PET Model Q-Q Plot
# ============================================================================ #

test_that("Q-Q plot for PET model matches metafor residual quantiles", {

  name        <- "dat.lehmann2018-PET"
  fit_metafor <- info[[name]][["metafor"]]
  fit_brma    <- fits[[name]]
  set.seed(1)

  # --------------------------------------------------
  # Visual comparison: side-by-side (rstandard for comparability)
  # --------------------------------------------------

  expect_vdiffr_snapshot("qqnorm_PET_base", function() {
    oldpar <- graphics::par(no.readonly = TRUE)
    on.exit(graphics::par(mfrow = oldpar[["mfrow"]], mar = oldpar[["mar"]]))
    par(mfrow = c(1, 2), mar = c(4, 4, 2, 1))
    qqnorm(fit_metafor, main = "metafor", type = "rstudent", ylim = c(-3, 3))
    suppressWarnings(qqnorm(fit_brma, plot_type = "base", main = "brma", type = "rstudent", ylim = c(-3, 3)))
  })
})

test_that("Q-Q plot for PET meta-regression matches metafor residual quantiles", {

  skip_if_not_full_visuals("PET meta-regression duplicates the default PET Q-Q visual.")

  name        <- "dat.lehmann2018-PETreg"
  fit_metafor <- info[[name]][["metafor"]]
  fit_brma    <- fits[[name]]
  set.seed(1)

  # --------------------------------------------------
  # Visual comparison: side-by-side (rstandard for comparability)
  # --------------------------------------------------

  expect_vdiffr_snapshot("qqnorm_PETreg_base", function() {
    oldpar <- graphics::par(no.readonly = TRUE)
    on.exit(graphics::par(mfrow = oldpar[["mfrow"]], mar = oldpar[["mar"]]))
    par(mfrow = c(1, 2), mar = c(4, 4, 2, 1))
    qqnorm(fit_metafor, main = "metafor", type = "rstudent", ylim = c(-3, 3))
    suppressWarnings(qqnorm(fit_brma, plot_type = "base", main = "brma", type = "rstudent", ylim = c(-3, 3)))
  })
})

# ============================================================================ #
# Test: Selection Model Q-Q Plot
# ============================================================================ #

test_that("Q-Q plot for selection model renders residual quantiles", {

  name     <- "dat.lehmann2018-3PSM"
  fit_brma <- fits[[name]]
  set.seed(1)

  expect_vdiffr_snapshot("qqnorm_selmodel_base", function() {
    suppressWarnings(qqnorm(fit_brma, plot_type = "base"))
  })

  # skip the ggplot version for to save test time
  # expect_vdiffr_snapshot(
  #   "qqnorm_selmodel_ggplot",
  #   suppressWarnings(qqnorm(fit_brma, plot_type = "ggplot"))
  # )

  # rstandard should error for selection models
  expect_error(
    qqnorm(fit_brma, type = "rstandard"),
    info = "rstandard residuals are rejected for selection models"
  )
})

test_that("Q-Q plot for selection meta-regression renders residual quantiles", {

  skip_if_not_full_visuals("Selection meta-regression duplicates the default selection Q-Q visual.")

  name     <- "dat.lehmann2018-3PSMreg"
  fit_brma <- fits[[name]]
  set.seed(1)

  expect_vdiffr_snapshot("qqnorm_selmodel_reg_base", function() {
    suppressWarnings(qqnorm(fit_brma, plot_type = "base"))
  })

  # skip the ggplot version for to save test time
  # expect_vdiffr_snapshot(
  #   "qqnorm_selmodel_ggplot",
  #   suppressWarnings(qqnorm(fit_brma, plot_type = "ggplot"))
  # )

  # rstandard should error for selection models
  expect_error(
    qqnorm(fit_brma, type = "rstandard"),
    info = "rstandard residuals are rejected for selection models"
  )
})

# ============================================================================ #
# Test: BMA.norm Model Q-Q Plot
# ============================================================================ #

test_that("Q-Q plot for BMA.norm model renders base output", {

  name     <- "dat.lehmann2018_BMA.norm"
  fit_brma <- fits[[name]]
  set.seed(1)

  expect_vdiffr_snapshot("qqnorm_BMA", function() {
    suppressWarnings(qqnorm(fit_brma, plot_type = "base"))
  })
})

test_that("Q-Q plot for BMA.norm meta-regression renders base output", {

  skip_if_not_full_visuals("BMA meta-regression duplicates the default BMA Q-Q smoke test.")

  name     <- "dat.lehmann2018_BMA.norm_mods"
  fit_brma <- fits[[name]]
  set.seed(1)

  expect_vdiffr_snapshot("qqnorm_BMAreg", function() {
    suppressWarnings(qqnorm(fit_brma, plot_type = "base"))
  })
})

# ============================================================================ #
# Test: BMA.glmm Model Q-Q Plot
# ============================================================================ #

test_that("Q-Q plot for BMA.glmm model renders base output", {

  name     <- "bcg_BMA.glmm_3lvl_location_scale"
  fit_brma <- fits[[name]]
  set.seed(1)

  expect_vdiffr_snapshot("qqnorm_BMA.glmm", function() {
    suppressWarnings(qqnorm(fit_brma, plot_type = "base"))
  })
})

# ============================================================================ #
# Test: RoBMA Model Q-Q Plot
# ============================================================================ #

test_that("Q-Q plot for RoBMA model renders base output", {

  name     <- "dat.lehmann2018_RoBMA"
  fit_brma <- fits[[name]]
  set.seed(1)

  expect_vdiffr_snapshot("qqnorm_RoBMA", function() {
    suppressWarnings(qqnorm(fit_brma, plot_type = "base"))
  })
})

test_that("Q-Q plot for RoBMA meta-regression renders LOO-PIT output", {

  skip_if_not_full_visuals("RoBMA meta-regression duplicates the default RoBMA Q-Q smoke test.")

  name     <- "dat.lehmann2018_RoBMA_3lvl_mods_scale"
  fit_brma <- fits[[name]]
  set.seed(1)

  expect_vdiffr_snapshot("qqnorm_RoBMA_complex", function() {
    suppressWarnings(qqnorm(fit_brma, plot_type = "base", type = "LOO-PIT"))
  })
})


# ============================================================================ #
# Test: Q-Q Plot Interface
# ============================================================================ #

test_that("Q-Q plot data and argument validation are stable", {

  name     <- "bcg_meta-analysis"
  fit_brma <- fits[[name]]
  set.seed(1)

  # --------------------------------------------------
  # Test as_data = TRUE returns list with expected components
  # --------------------------------------------------

  qq_data <- qqnorm(fit_brma, as_data = TRUE, type = "rstandard")

  expect_true(is.list(qq_data),
    info = "as_data = TRUE returns a list"
  )

  expected_components <- c(
    "x", "y", "points", "refline", "envelope",
    "xlim", "ylim", "xlab", "ylab"
  )
  expect_true(all(expected_components %in% names(qq_data)),
    info = "Q-Q data contains all expected components"
  )

  # Check points data.frame structure
  expect_true(is.data.frame(qq_data$points),
    info = "points are returned as a data.frame"
  )
  expect_true(all(c("x", "y") %in% names(qq_data$points)),
    info = "points contain x and y columns"
  )

  # Check number of points matches number of observations
  n_studies <- nrow(fit_brma$data$outcome)
  expect_equal(nrow(qq_data$points), n_studies,
    info = "number of points matches number of observations"
  )

  # Check x values are sorted (theoretical quantiles)
  expect_equal(qq_data$x, sort(qq_data$x),
    info = "theoretical quantiles are sorted"
  )

  # Check y values are sorted (sorted residuals)
  expect_equal(qq_data$y, sort(qq_data$y),
    info = "sample quantiles are sorted"
  )

  # --------------------------------------------------
  # Test envelope = FALSE suppresses envelope
  # --------------------------------------------------

  qq_data_no_env <- qqnorm(fit_brma, as_data = TRUE, envelope = FALSE, type = "rstandard")
  expect_null(qq_data_no_env$envelope,
    info = "envelope is NULL when envelope = FALSE"
  )

  # --------------------------------------------------
  # Test envelope = TRUE includes envelope
  # --------------------------------------------------

  expect_true(is.data.frame(qq_data$envelope),
    info = "envelope is a data.frame when envelope = TRUE"
  )
  expect_true(all(c("x", "lower", "upper") %in% names(qq_data$envelope)),
    info = "envelope contains x, lower, upper columns"
  )

  set.seed(1)
  qq_env_1 <- qqnorm(
    fit_brma,
    as_data  = TRUE,
    type     = "rstandard",
    reps     = 10,
    envelope = TRUE
  )[["envelope"]]
  set.seed(999)
  qq_env_2 <- qqnorm(
    fit_brma,
    as_data  = TRUE,
    type     = "rstandard",
    reps     = 100,
    envelope = TRUE
  )[["envelope"]]
  expect_equal(qq_env_1, qq_env_2,
    info = "closed-form QQ envelope is deterministic and ignores reps"
  )

  # --------------------------------------------------
  # Test error on invalid plot_type
  # --------------------------------------------------

  expect_error(qqnorm(fit_brma, plot_type = "invalid"),
    info = "invalid plot_type is rejected"
  )

  # --------------------------------------------------
  # Test error on invalid type
  # --------------------------------------------------

  expect_error(qqnorm(fit_brma, type = "invalid"),
    info = "invalid type is rejected"
  )
})

# ============================================================================ #
# Test: Q-Q Plot Customization
# ============================================================================ #

test_that("Q-Q plot customization snapshots are stable", {

  skip_if_not_full_visuals("Customization snapshots are visual-gallery coverage.")

  name     <- "bcg_meta-analysis"
  fit_brma <- fits[[name]]
  set.seed(1)

  # --------------------------------------------------
  # Test custom point aesthetics
  # --------------------------------------------------

  expect_vdiffr_snapshot("qqnorm_custom_points_base", function() {
    qqnorm(fit_brma, plot_type = "base", pch = 21, col = "blue",
           bg = "lightblue", cex = 1.5, type = "rstandard")
  })

  expect_vdiffr_snapshot(
    "qqnorm_custom_points_ggplot",
    qqnorm(fit_brma, plot_type = "ggplot", pch = 21, col = "blue",
           bg = "lightblue", size = 3, type = "rstandard")
  )

  # --------------------------------------------------
  # Test custom axis labels and title
  # --------------------------------------------------

  expect_vdiffr_snapshot("qqnorm_custom_labels_base", function() {
    qqnorm(fit_brma, plot_type = "base", xlab = "Expected", ylab = "Observed",
           main = "QQ Plot", type = "rstandard")
  })

  expect_vdiffr_snapshot(
    "qqnorm_custom_labels_ggplot",
    qqnorm(fit_brma, plot_type = "ggplot", xlab = "Expected", ylab = "Observed",
           main = "QQ Plot", type = "rstandard")
  )

  # --------------------------------------------------
  # Test without envelope
  # --------------------------------------------------

  expect_vdiffr_snapshot("qqnorm_no_envelope_base", function() {
    qqnorm(fit_brma, plot_type = "base", envelope = FALSE, type = "rstandard")
  })

  expect_vdiffr_snapshot(
    "qqnorm_no_envelope_ggplot",
    qqnorm(fit_brma, plot_type = "ggplot", envelope = FALSE, type = "rstandard")
  )

  # --------------------------------------------------
  # Test custom envelope level
  # --------------------------------------------------

  expect_vdiffr_snapshot("qqnorm_custom_level_base", function() {
    qqnorm(fit_brma, plot_type = "base", level = 99, type = "rstandard")
  })

  expect_vdiffr_snapshot(
    "qqnorm_custom_level_ggplot",
    qqnorm(fit_brma, plot_type = "ggplot", level = 99, type = "rstandard")
  )

  # --------------------------------------------------
  # Test Bonferroni correction
  # --------------------------------------------------

  expect_vdiffr_snapshot("qqnorm_bonferroni_base", function() {
    qqnorm(fit_brma, plot_type = "base", bonferroni = TRUE, type = "rstandard")
  })

  expect_vdiffr_snapshot(
    "qqnorm_bonferroni_ggplot",
    qqnorm(fit_brma, plot_type = "ggplot", bonferroni = TRUE, type = "rstandard")
  )

  # --------------------------------------------------
  # Test suppressing envelope shading
  # --------------------------------------------------

  expect_vdiffr_snapshot("qqnorm_no_shade_base", function() {
    qqnorm(fit_brma, plot_type = "base", shade = NA, type = "rstandard")
  })

  expect_vdiffr_snapshot(
    "qqnorm_no_shade_ggplot",
    qqnorm(fit_brma, plot_type = "ggplot", shade = NA, type = "rstandard")
  )
})

context("Prior plotting")

test_data <- data.frame(
  effect     = c(0.10, 0.25, 0.15, 0.30, 0.05),
  std_err    = sqrt(c(0.04, 0.06, 0.05, 0.08, 0.03)),
  mod_cont   = c(1.5, 2.3, 1.8, 3.1, 0.9),
  mod_factor = factor(c("A", "B", "A", "B", "A")),
  scale_var  = c(0.5, 1.0, 0.8, 1.2, 0.6),
  stringsAsFactors = FALSE
)

.with_temp_plot_device <- function(expr) {

  file <- tempfile(fileext = ".png")
  grDevices::png(filename = file)
  on.exit({
    grDevices::dev.off()
    unlink(file)
  }, add = TRUE)

  force(expr)
}

.is_ggplot <- function(x) {

  inherits(x, "ggplot")
}

.ggplot_x_range <- function(x) {

  build_data <- ggplot2::ggplot_build(x)[["data"]]
  x_values   <- unlist(lapply(build_data, function(layer) layer[["x"]]))

  range(x_values, finite = TRUE)
}

test_that("plot_prior plots outcome priors from only_priors objects", {

  skip_on_cran()

  priors <- BMA(
    yi = effect, sei = std_err, data = test_data,
    measure = "SMD", only_priors = TRUE
  )

  expect_true(.is_ggplot(plot_prior(priors, parameter = "mu",  plot_type = "ggplot")))
  expect_true(.is_ggplot(plot_prior(priors, parameter = "tau", plot_type = "ggplot")))

  plots <- plot_prior(priors, parameter = c("mu", "tau"), plot_type = "ggplot")
  expect_named(plots, c("mu", "tau"))
  expect_true(all(vapply(plots, .is_ggplot, logical(1))))

  .with_temp_plot_device(
    expect_silent(plot_prior(priors, parameter = "mu"))
  )
})

test_that("plot_prior selects moderator and scale priors", {

  skip_on_cran()

  reg_priors <- BMA(
    yi = effect, sei = std_err,
    mods = ~ mod_cont + mod_factor,
    data = test_data, measure = "SMD", only_priors = TRUE
  )

  expect_true(.is_ggplot(plot_prior(reg_priors, parameter = "mu", plot_type = "ggplot")))
  expect_true(.is_ggplot(plot_prior(reg_priors, parameter_mods = "mod_cont", plot_type = "ggplot")))
  expect_true(.is_ggplot(plot_prior(reg_priors, parameter = "mod_factor", plot_type = "ggplot")))
  expect_true(.is_ggplot(plot_prior(reg_priors, parameter = "mod_factor", standardized_coefficients = FALSE, plot_type = "ggplot")))

  brma_factor_priors <- brma(
    yi = effect, sei = std_err,
    mods = ~ mod_factor,
    data = test_data, measure = "SMD",
    set_contrast_factor_predictors = "meandif",
    only_priors = TRUE
  )

  .with_temp_plot_device(
    expect_silent(plot_prior(brma_factor_priors, parameter_mods = "mod_factor"))
  )

  expect_error(
    plot_prior(reg_priors, parameter_mods = "missing", plot_type = "ggplot"),
    regexp = "not available"
  )
  expect_error(
    plot_prior(reg_priors, parameter_mods = "mod_cont", standardized_coefficients = "no", plot_type = "ggplot"),
    regexp = "standardized_coefficients"
  )

  default_plot      <- plot_prior(reg_priors, parameter_mods = "mod_cont", plot_type = "ggplot")
  standardized_plot <- plot_prior(reg_priors, parameter_mods = "mod_cont", standardized_coefficients = TRUE,  plot_type = "ggplot")
  raw_plot          <- plot_prior(reg_priors, parameter_mods = "mod_cont", standardized_coefficients = FALSE, plot_type = "ggplot")

  expect_equal(.ggplot_x_range(default_plot), .ggplot_x_range(standardized_plot))
  expect_gt(diff(.ggplot_x_range(raw_plot)), diff(.ggplot_x_range(standardized_plot)))

  scale_priors <- suppressWarnings(BMA(
    yi = effect, sei = std_err,
    scale = ~ scale_var,
    data = test_data, measure = "SMD", only_priors = TRUE
  ))

  expect_true(.is_ggplot(plot_prior(scale_priors, parameter = "tau", plot_type = "ggplot")))
  expect_true(.is_ggplot(plot_prior(scale_priors, parameter_scale = "scale_var", plot_type = "ggplot")))
  expect_true(.is_ggplot(plot_prior(scale_priors, parameter_scale = "scale_var", standardized_coefficients = FALSE, plot_type = "ggplot")))
})

test_that("plot_prior supports direct prior objects", {

  skip_on_cran()

  prior_object <- prior("normal", parameters = list(mean = 0, sd = 1))

  expect_true(.is_ggplot(plot_prior(prior_object, plot_type = "ggplot")))
})

test_that("print_prior prints selected priors", {

  skip_on_cran()

  priors <- BMA(
    yi = effect, sei = std_err,
    mods = ~ mod_cont + mod_factor,
    data = test_data, measure = "SMD", only_priors = TRUE
  )

  expect_true(BayesTools::is.prior(print_prior(priors, parameter = "mu", silent = TRUE)))
  expect_true(BayesTools::is.prior(print_prior(priors, parameter_mods = "mod_cont", silent = TRUE)))
  expect_true(BayesTools::is.prior(print_prior(priors, parameter = "mod_factor", silent = TRUE)))

  full_output <- capture.output(full_selected <- print_prior(priors, silent = TRUE))
  expect_identical(full_output, character(0))
  expect_named(full_selected, c("mu_intercept", "mu_mod_cont", "mu_mod_factor", "tau"))
  expect_true(all(vapply(full_selected, BayesTools::is.prior, logical(1))))

  full_printed   <- capture.output(print_prior(priors))
  parameter_rows <- grep("^(mu_|tau:)", full_printed, value = TRUE)
  expect_equal(sub(":.*$", "", parameter_rows), names(full_selected))

  printed <- capture.output(print_prior(priors, parameter = "tau"))
  expect_equal(printed[1:3], c(
    "tau:",
    "  alternative:",
    "    (1/2) * Normal(0, 0.35)[0, Inf]"
  ))
  expect_true("  null:" %in% printed)

  selected <- print_prior(priors, parameter = c("mu", "tau"), silent = TRUE)
  expect_named(selected, c("mu", "tau"))
  expect_true(all(vapply(selected, BayesTools::is.prior, logical(1))))

  prior_object <- prior("normal", parameters = list(mean = 0, sd = 1))
  expect_true(BayesTools::is.prior(print_prior(prior_object, silent = TRUE)))
})

test_that("plot_prior handles publication-bias prior components", {

  skip_on_cran()

  priors <- RoBMA(
    yi = effect, sei = std_err, data = test_data,
    measure = "SMD", only_priors = TRUE
  )

  expect_error(
    plot_prior(priors, parameter = "bias", plot_type = "ggplot"),
    regexp = "mixes weight-function and PET-PEESE"
  )
  omega_output <- capture.output(
    omega_plot <- plot_prior(priors, parameter = "omega", plot_type = "ggplot")
  )
  expect_identical(omega_output, character(0))
  expect_true(.is_ggplot(omega_plot))
  expect_true(.is_ggplot(plot_prior(priors, parameter = "PET",   plot_type = "ggplot")))
  expect_true(.is_ggplot(plot_prior(priors, parameter = "PEESE", plot_type = "ggplot")))

  expect_true(BayesTools::is.prior(print_prior(priors, parameter = "bias",  silent = TRUE)))
  expect_true(BayesTools::is.prior(print_prior(priors, parameter = "omega", silent = TRUE)))
  expect_true(BayesTools::is.prior(print_prior(priors, parameter = "PET",   silent = TRUE)))
})

test_that("only_priors objects print and plot via prior methods", {

  skip_on_cran()

  priors <- RoBMA(
    yi = effect, sei = std_err, data = test_data,
    measure = "SMD", model_type = "PP", only_priors = TRUE
  )

  expect_s3_class(priors, "only_priors.brma")

  printed <- capture.output(print(priors))
  expect_true(any(grepl("^mu:", printed)))
  expect_true(any(grepl("^bias:", printed)))

  .with_temp_plot_device(
    expect_silent(plot(priors))
  )
  expect_true(.is_ggplot(plot(priors, plot_type = "ggplot")))
})

context("Prior and posterior plots")

# Load common test helpers
source(testthat::test_path("common-functions.R"))
source(testthat::test_path("helper-test-matrix.R"))
source(testthat::test_path("helper-visuals.R"))

# list cached fits lazily
skip_if_no_fits()
fit_names <- list_fits()
fits      <- lazy_fits(fit_names, validate = FALSE)


test_that("plot.brma uses parameter x-axis labels by default", {

  temp_fit <- fits[["bcg_meta-analysis"]]

  x_label <- function(plot) {

    return(plot$scales$get_scales("x")$name)
  }

  expect_identical(
    x_label(plot(temp_fit, "mu", plot_type = "ggplot")),
    "Effect Size"
  )
  expect_identical(
    x_label(plot(temp_fit, "tau", plot_type = "ggplot")),
    "Heterogeneity"
  )
  expect_identical(
    x_label(plot(temp_fit, "mu", plot_type = "ggplot", xlab = "Custom Label")),
    "Custom Label"
  )
})


test_that("Prior and posterior distributions for brma.norm models", {

  ### simple meta-analysis ----
  name <- "bcg_meta-analysis"
  temp_fit <- fits[[name]]

  ### effect size
  expect_vdiffr_snapshot(paste0(name, "-mu_baseplot_pp_no_prior"), function() plot(temp_fit, "mu", prior = FALSE))
  expect_vdiffr_snapshot(paste0(name, "-mu_ggplot_pp_no_prior"), plot(temp_fit, "mu", prior = FALSE, plot_type = "ggplot"))
  expect_vdiffr_snapshot(paste0(name, "-mu_baseplot_pp_prior"), function() plot(temp_fit, "mu", prior = TRUE))
  expect_vdiffr_snapshot(paste0(name, "-mu_ggplot_pp_prior"), plot(temp_fit, "mu", prior = TRUE, plot_type = "ggplot"))

  # change range
  expect_vdiffr_snapshot(paste0(name, "-mu_baseplot_pp_range"), function() plot(temp_fit, "mu", prior = TRUE, xlim = c(-1, 1), ylim = c(0, 5)))
  expect_vdiffr_snapshot(paste0(name, "-mu_ggplot_pp_range"), plot(temp_fit, "mu", prior = TRUE, plot_type = "ggplot", xlim = c(-1, 1), ylim = c(0, 5)))

  # change aesthetics
  expect_vdiffr_snapshot(paste0(name, "-mu_baseplot_pp_aesthetics"), function() plot(temp_fit, "mu", prior = TRUE, lwd = 3, lty = 3, col = "blue", dots_prior = list(lwd = 3, lty = 1, col = "red")))
  expect_vdiffr_snapshot(paste0(name, "-mu_ggplot_pp_aesthetics"), plot(temp_fit, "mu", prior = TRUE, plot_type = "ggplot", lwd = 2, lty = 3, col = "blue", dots_prior = list(lwd = 3, lty = 1, col = "red")))

  ### heterogeneity
  expect_vdiffr_snapshot(paste0(name, "-tau_baseplot_tau_pp_prior"), function() plot(temp_fit, "tau", prior = TRUE))
  expect_vdiffr_snapshot(paste0(name, "-tau_ggplot_pp_prior"), plot(temp_fit, "tau", prior = TRUE, plot_type = "ggplot"))

  ### meta-regression (continuous)
  name <- "bcg_meta-regression"
  temp_fit <- fits[[name]]

  expect_vdiffr_snapshot(paste0(name, "-mu_baseplot_orig"), function() plot(temp_fit, "mu", prior = TRUE))
  expect_vdiffr_snapshot(paste0(name, "-mu_baseplot_std"), function() plot(temp_fit, "mu", prior = TRUE, standardized_coefficients = TRUE))

  expect_vdiffr_snapshot(paste0(name, "-reg_baseplot_orig"), function() plot(temp_fit, parameter_mods = "year", prior = TRUE))
  expect_vdiffr_snapshot(paste0(name, "-reg_baseplot_std"), function() plot(temp_fit, parameter_mods = "year", prior = TRUE, standardized_coefficients = TRUE))

  ### meta-regression (categorical: dummy)
  name <- "bcg_meta-regression2"
  temp_fit <- fits[[name]]

  expect_vdiffr_snapshot(paste0(name, "-reg_baseplot_orig"), function() plot(temp_fit, parameter_mods = "alloc", prior = TRUE))
  expect_vdiffr_snapshot(paste0(name, "-reg_baseplot_std"), function() plot(temp_fit, parameter_mods = "alloc", prior = TRUE, standardized_coefficients = TRUE))

  ### meta-regression (categorical: meandif)
  name <- "bcg_meta-regression2b"
  temp_fit <- fits[[name]]

  expect_vdiffr_snapshot(paste0(name, "-reg_baseplot_orig"), function() plot(temp_fit, parameter_mods = "alloc", prior = TRUE))
  expect_vdiffr_snapshot(paste0(name, "-reg_baseplot_std"), function() plot(temp_fit, parameter_mods = "alloc", prior = TRUE, standardized_coefficients = TRUE))

  ### meta-regression with interactions (categorical: dummy)
  name <- "bcg_meta-regression3"
  temp_fit <- fits[[name]]

  expect_vdiffr_snapshot(paste0(name, "-reg0_baseplot_orig"), function() plot(temp_fit, parameter_mods = "intercept", prior = TRUE))
  expect_vdiffr_snapshot(paste0(name, "-reg0_baseplot_std"), function() plot(temp_fit, parameter_mods = "intercept", prior = TRUE, standardized_coefficients = TRUE))
  expect_vdiffr_snapshot(paste0(name, "-reg1_baseplot_orig"), function() plot(temp_fit, parameter_mods = "alloc", prior = TRUE))
  expect_vdiffr_snapshot(paste0(name, "-reg1_baseplot_std"), function() plot(temp_fit, parameter_mods = "alloc", prior = TRUE, standardized_coefficients = TRUE))
  expect_vdiffr_snapshot(paste0(name, "-reg2_baseplot_orig"), function() plot(temp_fit, parameter_mods = "year", prior = TRUE))
  expect_vdiffr_snapshot(paste0(name, "-reg2_baseplot_std"), function() plot(temp_fit, parameter_mods = "year", prior = TRUE, standardized_coefficients = TRUE))
  expect_vdiffr_snapshot(paste0(name, "-reg3_baseplot_orig"), function() plot(temp_fit, parameter_mods = "alloc:year", prior = TRUE))
  expect_vdiffr_snapshot(paste0(name, "-reg3_baseplot_std"), function() plot(temp_fit, parameter_mods = "alloc:year", prior = TRUE, standardized_coefficients = TRUE))

  ### meta-regression with interactions (categorical: meandif)
  name <- "bcg_meta-regression3b"
  temp_fit <- fits[[name]]

  expect_vdiffr_snapshot(paste0(name, "-reg0_baseplot_orig"), function() plot(temp_fit, parameter_mods = "intercept", prior = TRUE))
  expect_vdiffr_snapshot(paste0(name, "-reg0_baseplot_std"), function() plot(temp_fit, parameter_mods = "intercept", prior = TRUE, standardized_coefficients = TRUE))
  expect_vdiffr_snapshot(paste0(name, "-reg1_baseplot_orig"), function() plot(temp_fit, parameter_mods = "alloc", prior = TRUE))
  expect_vdiffr_snapshot(paste0(name, "-reg1_baseplot_std"), function() plot(temp_fit, parameter_mods = "alloc", prior = TRUE, standardized_coefficients = TRUE))
  expect_vdiffr_snapshot(paste0(name, "-reg2_baseplot_orig"), function() plot(temp_fit, parameter_mods = "year", prior = TRUE))
  expect_vdiffr_snapshot(paste0(name, "-reg2_baseplot_std"), function() plot(temp_fit, parameter_mods = "year", prior = TRUE, standardized_coefficients = TRUE))
  expect_vdiffr_snapshot(paste0(name, "-reg3_baseplot_orig"), function() plot(temp_fit, parameter_mods = "alloc:year", prior = TRUE))
  expect_vdiffr_snapshot(paste0(name, "-reg3_baseplot_std"), function() plot(temp_fit, parameter_mods = "alloc:year", prior = TRUE, standardized_coefficients = TRUE))

  ### location-scale model
  name <- "bangertdrowns2004_location-scale"
  temp_fit <- fits[[name]]

  expect_vdiffr_snapshot(paste0(name, "-scale0_baseplot_orig"), function() plot(temp_fit, parameter_scale = "intercept", prior = TRUE))
  expect_vdiffr_snapshot(paste0(name, "-scale0_baseplot_std"), function() plot(temp_fit, parameter_scale = "intercept", prior = TRUE, standardized_coefficients = TRUE))
  expect_vdiffr_snapshot(paste0(name, "-scale1_baseplot_orig"), function() plot(temp_fit, parameter_scale = "ni100", prior = TRUE))
  expect_vdiffr_snapshot(paste0(name, "-scale1_baseplot_std"), function() plot(temp_fit, parameter_scale = "ni100", prior = TRUE, standardized_coefficients = TRUE))

  ### between-study heterogeneity and multilevel models
  name <- "konstantopoulos2011_3lvl"
  temp_fit <- fits[[name]]

  expect_vdiffr_snapshot(paste0(name, "-rho_baseplot"), function() plot(temp_fit, parameter = "rho", prior = TRUE))
  expect_vdiffr_snapshot(paste0(name, "-tau_baseplot"), function() plot(temp_fit, parameter = "tau", prior = TRUE))
})

test_that("Transformed effect-size plots render", {

  name     <- "bcg_meta-analysis"
  temp_fit <- fits[[name]]

  .with_temp_plot_device(
    expect_silent(plot(temp_fit, "mu", transform = "EXP", plot_type = "base"))
  )
  expect_true(.is_ggplot(
    plot(temp_fit, "mu", transform = "EXP", plot_type = "ggplot")
  ))
})

test_that("Prior and posterior plots transform effect-size axis", {

  skip_if_not_full_visuals("Effect-size transform snapshots are visual-gallery coverage.")

  name     <- "bcg_meta-analysis"
  temp_fit <- fits[[name]]

  expect_vdiffr_snapshot(paste0(name, "-mu_transform_exp_comparison"), function() {
    oldpar <- graphics::par(no.readonly = TRUE)
    on.exit(graphics::par(mfrow = oldpar[["mfrow"]], mar = oldpar[["mar"]]))
    par(mfrow = c(1, 2), mar = c(4, 4, 2, 1))
    plot(temp_fit, "mu", plot_type = "base", main = "log RR")
    plot(temp_fit, "mu", transform = "EXP", plot_type = "base", main = "RR")
  })
})

test_that("Prior and posterior distributions for bPET / bPEESE objects", {

  skip_if_not_full_visuals("PET/PEESE prior-posterior snapshots are visual-gallery coverage.")

  ### PET model
  name <- "dat.lehmann2018-PET"
  temp_fit <- fits[[name]]
  set.seed(1)

  # no prior
  expect_vdiffr_snapshot("baseplot_pp_PETPEESE_no_prior", function() plot_pet_peese(temp_fit, prior = FALSE))
  expect_vdiffr_snapshot("ggplot_pp_PETPEESE_no_prior", plot_pet_peese(temp_fit, prior = FALSE, plot_type = "ggplot"))

  # change range
  expect_vdiffr_snapshot("baseplot_pp_PETPEESE_range", function() plot_pet_peese(temp_fit, prior = TRUE, xlim = c(0, 0.5), ylim = c(-3, 3)))
  expect_vdiffr_snapshot("ggplot_pp_PETPEESE_range", plot_pet_peese(temp_fit, prior = TRUE, plot_type = "ggplot", xlim = c(0, 0.5), ylim = c(-3, 3)))

  # change aesthetics
  expect_vdiffr_snapshot("baseplot_pp_PETPEESE_aesthetics", function() plot_pet_peese(temp_fit, prior = TRUE, lwd = 3, lty = 3, col = "blue", col.fill = scales::alpha("blue", 0.20),
                                                                                          dots_prior = list(lwd = 3, lty = 1, col = "red", col.fill = scales::alpha("red", 0.20))))
  expect_vdiffr_snapshot("ggplot_pp_PETPEESE_aesthetics", plot_pet_peese(temp_fit, prior = TRUE, plot_type = "ggplot", lwd = 2, lty = 3, col = "blue", col.fill = scales::alpha("blue", 0.20),
                                                                             dots_prior = list(lwd = 3, lty = 1, col = "red", col.fill = scales::alpha("red", 0.20))))

})

test_that("Prior and posterior distributions for bselmodel objects", {

  skip_if_not_full_visuals("Selection prior-posterior snapshots are visual-gallery coverage.")


  ### weight function
  name <- "dat.lehmann2018-3PSM"
  temp_fit <- fits[[name]]
  set.seed(1)

  # no prior
  expect_vdiffr_snapshot("baseplot_pp_weightfunction_no_prior", function() plot_weightfunction(temp_fit, prior = FALSE))
  expect_vdiffr_snapshot("ggplot_pp_weightfunction_no_prior", plot_weightfunction(temp_fit, prior = FALSE, plot_type = "ggplot"))

  # change range
  expect_vdiffr_snapshot("baseplot_pp_weightfunction_range", function() plot_weightfunction(temp_fit, prior = TRUE, rescale_p_values = FALSE))
  expect_vdiffr_snapshot("ggplot_pp_weightfunction_range", plot_weightfunction(temp_fit, prior = TRUE, plot_type = "ggplot", rescale_p_values = FALSE))

  # change aesthetics
  expect_vdiffr_snapshot("baseplot_pp_weightfunction_aesthetics", function() plot_weightfunction(temp_fit, prior = TRUE, lwd = 3, lty = 3, col = "blue", col.fill = scales::alpha("blue", 0.20),
                                                                                                      dots_prior = list(lwd = 3, lty = 1, col = "red", col.fill = scales::alpha("red", 0.20))))
  expect_vdiffr_snapshot("ggplot_pp_weightfunction_aesthetics", plot_weightfunction(temp_fit, prior = TRUE, plot_type = "ggplot", lwd = 2, lty = 3, col = "blue", col.fill = scales::alpha("blue", 0.20),
                                                                                         dots_prior = list(lwd = 3, lty = 1, col = "red", col.fill = scales::alpha("red", 0.20))))

})

test_that("Weightfunction plot supports observed p-value rug", {

  name <- "dat.lehmann2018-3PSM"
  skip_if_missing_fits(name)

  .with_temp_plot_device(
    expect_silent(plot_weightfunction(fits[[name]], show_data = TRUE))
  )
  expect_true(.is_ggplot(
    plot_weightfunction(
      fits[[name]],
      show_data = TRUE,
      plot_type = "ggplot",
      dots_data = list(color = "red", alpha = .5, linewidth = .4, rug_side = "top")
    )
  ))
  expect_silent(
    plot_weightfunction(
      fits[[name]],
      show_data        = TRUE,
      rescale_p_values = FALSE,
      dots_data        = list(col = "blue", lwd = 1, side = "bottom")
    )
  )

  rug_plot <- plot_weightfunction(
    fits[[name]],
    show_data        = TRUE,
    plot_type        = "ggplot",
    rescale_p_values = FALSE
  )
  rug_data <- ggplot_geom_layer_data(rug_plot, "GeomRug")

  expect_equal(
    rug_data[["p"]],
    .weightfunction_observed_p_values(fits[[name]]),
    tolerance = sqrt(.Machine$double.eps)
  )
})

test_that("Prior and posterior distributions for BMA.norm objects", {

  skip_if_not_full_visuals("BMA prior-posterior snapshots are visual-gallery coverage.")

  name <- "dat.lehmann2018_BMA.norm_mods"
  temp_fit <- fits[[name]]
  set.seed(1)

  # effect
  expect_vdiffr_snapshot(paste0(name, "-mu_baseplot_pp_no_prior"), function() plot(temp_fit, "mu", prior = FALSE))
  expect_vdiffr_snapshot(paste0(name, "-mu_ggplot_pp_no_prior"), plot(temp_fit, "mu", prior = FALSE, plot_type = "ggplot"))
  expect_vdiffr_snapshot(paste0(name, "-mu_baseplot_pp_prior"), function() plot(temp_fit, "mu", prior = TRUE))
  expect_vdiffr_snapshot(paste0(name, "-mu_ggplot_pp_prior"), plot(temp_fit, "mu", prior = TRUE, plot_type = "ggplot"))

  # moderation
  expect_vdiffr_snapshot(paste0(name, "-mods_ggplot_pp_no_prior"), plot(temp_fit, parameter_mods = "Preregistered", prior = FALSE, plot_type = "ggplot"))
  expect_vdiffr_snapshot(paste0(name, "-mods_ggplot_pp_prior"), plot(temp_fit, parameter_mods = "Preregistered", prior = TRUE, plot_type = "ggplot"))

  # heterogeneity
  expect_vdiffr_snapshot(paste0(name, "-tau_baseplot_pp_no_prior"), function() plot(temp_fit, "tau", prior = FALSE))
  expect_vdiffr_snapshot(paste0(name, "-tau_baseplot_pp_prior"), function() plot(temp_fit, "tau", prior = TRUE))

})

test_that("Prior and posterior distributions for RoBMA objects", {

  skip_if_not_full_visuals("RoBMA prior-posterior snapshots are visual-gallery coverage.")

  name <- "dat.lehmann2018_RoBMA_3lvl_mods_scale"
  temp_fit <- fits[[name]]
  set.seed(1)

  # effect
  expect_vdiffr_snapshot(paste0(name, "-mu_baseplot_pp_no_prior"), function() plot(temp_fit, "mu", prior = FALSE))
  expect_vdiffr_snapshot(paste0(name, "-mu_baseplot_pp_prior"), function() plot(temp_fit, "mu", prior = TRUE))

  # moderation
  expect_vdiffr_snapshot(paste0(name, "-mu_ggplot_pp_no_prior"), plot(temp_fit, parameter_mods = "Preregistered", prior = FALSE, plot_type = "ggplot"))
  expect_vdiffr_snapshot(paste0(name, "-mu_ggplot_pp_prior"), plot(temp_fit, parameter_mods = "Preregistered", prior = TRUE, plot_type = "ggplot"))

  # heterogeneity
  expect_vdiffr_snapshot(paste0(name, "-tau_baseplot_pp_no_prior"), function() plot(temp_fit, "tau", prior = TRUE))
  expect_vdiffr_snapshot(paste0(name, "-rho_baseplot_pp_no_prior"), function() plot(temp_fit, "rho", prior = TRUE))

})

vignette_cache <- function(name, objects, packages = character(), root = "../models") {

  stopifnot(
    is.character(name), length(name) == 1L,
    is.character(objects), length(objects) > 0L,
    is.character(packages),
    is.character(root), length(root) == 1L
  )

  object_names <- names(objects)
  file_names   <- unname(objects)

  if (is.null(object_names)) {
    object_names <- rep("", length(objects))
  }

  unnamed_objects <- !nzchar(object_names)
  object_names[unnamed_objects] <- sub(
    "\\.[Rr][Dd][Ss]$",
    "",
    basename(file_names[unnamed_objects])
  )

  missing_extension <- !grepl("\\.[Rr][Dd][Ss]$", file_names)
  file_names[missing_extension] <- paste0(file_names[missing_extension], ".RDS")

  paths <- file.path(root, name, file_names)
  names(paths) <- object_names

  structure(
    paths,
    class    = "vignette_cache",
    name     = name,
    packages = packages,
    root     = root
  )
}

vignette_cache_checking <- function() {

  ("CheckExEnv" %in% search()) ||
    any(c("_R_CHECK_TIMINGS_", "_R_CHECK_LICENSE_") %in% names(Sys.getenv()))
}

vignette_cache_missing <- function(cache) {

  names(cache)[!file.exists(unname(cache))]
}

vignette_cache_missing_packages <- function(cache) {

  packages <- attr(cache, "packages")
  if (is.null(packages)) {
    packages <- character()
  }

  has_package <- vapply(
    packages,
    requireNamespace,
    logical(1),
    quietly = TRUE
  )

  packages[!has_package]
}

vignette_cache_eval <- function(cache) {

  !vignette_cache_checking() &&
    !length(vignette_cache_missing_packages(cache)) &&
    !length(vignette_cache_missing(cache))
}

vignette_cache_load <- function(cache, envir = parent.frame()) {

  missing <- vignette_cache_missing(cache)
  if (length(missing)) {
    stop(
      "Missing cached vignette fits: ",
      paste(missing, collapse = ", "),
      call. = FALSE
    )
  }

  for (i in seq_along(cache)) {
    assign(
      names(cache)[i],
      readRDS(file = unname(cache[i])),
      envir = envir
    )
  }

  invisible(cache)
}

vignette_cache_save <- function(cache, envir = parent.frame(), compress = "xz") {

  for (cache_dir in unique(dirname(unname(cache)))) {
    dir.create(cache_dir, recursive = TRUE, showWarnings = FALSE)
  }

  has_object <- vapply(
    names(cache),
    exists,
    logical(1),
    envir    = envir,
    inherits = FALSE
  )

  if (any(!has_object)) {
    stop(
      "Missing objects to cache: ",
      paste(names(cache)[!has_object], collapse = ", "),
      call. = FALSE
    )
  }

  for (i in seq_along(cache)) {
    saveRDS(
      get(names(cache)[i], envir = envir, inherits = FALSE),
      file     = unname(cache[i]),
      compress = compress
    )
  }

  invisible(cache)
}

# ============================================================================ #
# Selection-Model Mapping Helpers
# ============================================================================ #
#
# Selected-normal kernels use BayesTools p-bin order directly. R code prepares
# compact kernel metadata; likelihood probability loops stay in native code.
#
# ============================================================================ #


SELKERNEL_NORMAL           <- 0L
SELKERNEL_STEP             <- 1L
SELKERNEL_PHACK_POWER      <- 2L
SELKERNEL_STEP_PHACK_POWER <- 3L


.is_prior_weightfunction <- function(x) {

  return(!is.null(x) && BayesTools::is.prior.weightfunction(x))
}

.is_prior_phacking <- function(x) {

  return(!is.null(x) && BayesTools::is_prior_phacking(x))
}

.is_prior_bias_kernel <- function(x) {

  return(!is.null(x) && BayesTools::is_prior_bias(x))
}

.prior_is_selection_kernel <- function(prior) {

  return(
    .is_prior_weightfunction(prior) ||
      .is_prior_phacking(prior) ||
      .is_prior_bias_kernel(prior)
  )
}

.prior_has_phacking <- function(prior) {

  if (is.null(prior)) {
    return(FALSE)
  }
  if (BayesTools::is.prior.mixture(prior)) {
    return(any(vapply(as.list(prior), .prior_has_phacking, logical(1))))
  }
  if (.is_prior_weightfunction(prior)) {
    return(FALSE)
  }
  if (!.prior_is_selection_kernel(prior)) {
    return(FALSE)
  }

  return(.selection_backend_has_phacking(.selection_backend_spec(list(prior))))
}

.prior_has_selection <- function(prior) {

  if (is.null(prior)) {
    return(FALSE)
  }
  if (BayesTools::is.prior.mixture(prior)) {
    return(any(vapply(as.list(prior), .prior_has_selection, logical(1))))
  }
  if (.is_prior_weightfunction(prior)) {
    return(TRUE)
  }
  if (!.prior_is_selection_kernel(prior)) {
    return(FALSE)
  }

  backend <- .selection_backend_spec(list(prior))
  return(backend[["mode"]] %in% c("step", "step_phack_power"))
}

.selection_stop_phacking_deferred <- function() {

  stop(
    "P-hacking selection kernels are not enabled in this RoBMA release.",
    call. = FALSE
  )
}

.selection_bias_priors <- function(priors) {

  priors_bias <- priors[["outcome"]][["bias"]]
  if (is.null(priors_bias)) {
    return(list())
  }
  if (BayesTools::is.prior.mixture(priors_bias)) {
    out <- as.list(priors_bias)
    class(out) <- "list"
    return(out)
  }
  return(list(priors_bias))
}

.selection_backend_spec <- function(priors_bias) {

  return(BayesTools::selection_backend_spec(priors_bias, backend = "jags"))
}

.selection_backend_has_phacking <- function(backend) {

  return(
    backend[["mode"]] %in% c("phack_power", "step_phack_power")
  )
}

.selection_kernel_breaks <- function(priors) {

  if (is.null(priors)) {
    return(NULL)
  }

  priors_list <- if (BayesTools::is.prior.mixture(priors)) {
    as.list(priors)
  } else if (BayesTools::is.prior(priors)) {
    list(priors)
  } else {
    priors
  }

  keep <- vapply(priors_list, .prior_has_selection, logical(1))
  if (!any(keep)) {
    return(NULL)
  }

  backend <- .selection_backend_spec(priors_list[keep])
  breaks  <- backend[["step"]][["breaks"]]
  if (is.null(breaks)) {
    breaks <- c(0, 1)
  }
  return(.selection_assert_p_cuts(breaks))
}

.selection_kernel_mode <- function(mode) {

  switch(
    mode,
    "normal"            = SELKERNEL_NORMAL,
    "step"              = SELKERNEL_STEP,
    "phack_power"       = SELKERNEL_PHACK_POWER,
    "step_phack_power"  = SELKERNEL_STEP_PHACK_POWER,
    stop("Unsupported selection kernel mode.", call. = FALSE)
  )
}

.selection_branch_kernel_modes <- function(backend) {

  branch_type <- backend[["branch_type"]]
  if (is.null(branch_type)) {
    return(.selection_kernel_mode(backend[["mode"]]))
  }

  kernel_mode <- rep(SELKERNEL_NORMAL, length(branch_type))
  has_step  <- branch_type %in% c("weightfunction", "combined")
  has_phack <- branch_type %in% c("phack", "phacking", "combined")

  if (backend[["mode"]] == "step") {
    kernel_mode[has_step] <- SELKERNEL_STEP
  } else if (backend[["mode"]] == "phack_power") {
    kernel_mode[has_phack] <- SELKERNEL_PHACK_POWER
  } else if (backend[["mode"]] == "step_phack_power") {
    kernel_mode[has_step & !has_phack]  <- SELKERNEL_STEP
    kernel_mode[has_phack & !has_step]  <- SELKERNEL_PHACK_POWER
    kernel_mode[has_step & has_phack]   <- SELKERNEL_STEP_PHACK_POWER
  }

  return(as.integer(kernel_mode))
}

.selection_jags_kernel_mode_expression <- function(branch_kernel_mode) {

  if (length(branch_kernel_mode) == 1L) {
    return(as.character(branch_kernel_mode[1L]))
  }

  terms <- paste0(
    branch_kernel_mode,
    " * equals(bias_indicator, ",
    seq_along(branch_kernel_mode),
    ")"
  )
  return(paste(terms, collapse = " + "))
}

.selection_obs_bin <- function(yi, sei, p_cuts, sign) {

  p_value <- stats::pnorm(sign * yi / sei, lower.tail = FALSE)
  return(.selection_p_bin(p_value, p_cuts))
}

.selection_step_bin_from_z <- function(z, p_cuts) {

  p_value <- stats::pnorm(z, lower.tail = FALSE)
  return(.selection_p_bin(p_value, p_cuts))
}

.selection_p_bin <- function(p_value, p_cuts) {

  p_cuts <- .selection_assert_p_cuts(p_cuts)

  for (cut in p_cuts) {
    close <- !is.na(p_value) & abs(p_value - cut) <= 1e-12
    p_value[close] <- cut
  }
  bin     <- findInterval(p_value, p_cuts, rightmost.closed = TRUE, left.open = TRUE)
  bin     <- pmin(pmax(bin, 1L), length(p_cuts) - 1L)
  return(as.integer(bin))
}

.selection_segment_midpoint <- function(lower, upper) {

  if (is.infinite(lower) && lower < 0) {
    return(upper - 1)
  }
  if (is.infinite(upper) && upper > 0) {
    return(lower + 1)
  }
  return((lower + upper) / 2)
}

.selection_segments <- function(p_cuts, phack_z_source = c(0, 0),
                                phack_z_dest = c(0, 0), has_phacking = FALSE) {

  p_cuts <- .selection_assert_p_cuts(p_cuts)

  z_lower <- stats::qnorm(1 - p_cuts[-1])
  z_upper <- stats::qnorm(1 - p_cuts[-length(p_cuts)])
  bounds  <- c(-Inf, Inf, z_lower[is.finite(z_lower)], z_upper[is.finite(z_upper)])

  if (has_phacking) {
    bounds <- c(bounds, phack_z_source, phack_z_dest)
  }

  bounds <- sort(unique(bounds))
  n_segments <- length(bounds) - 1L
  step_bin   <- integer(n_segments)
  region     <- integer(n_segments)

  for (i in seq_len(n_segments)) {
    mid <- .selection_segment_midpoint(bounds[i], bounds[i + 1L])
    step_bin[i] <- .selection_step_bin_from_z(mid, p_cuts)

    if (has_phacking) {
      if (mid >= phack_z_source[1] && mid <= phack_z_source[2]) {
        region[i] <- 1L
      } else if (mid > phack_z_dest[1] && mid <= phack_z_dest[2]) {
        region[i] <- 2L
      }
    }
  }

  return(list(
    bounds       = bounds,
    step_bin     = step_bin,
    phack_region = region
  ))
}

.selection_jags_bounds <- function(x) {

  x[is.infinite(x) & x < 0] <- -1e300
  x[is.infinite(x) & x > 0] <-  1e300
  return(x)
}

.selection_assert_p_cuts <- function(p_cuts) {

  if (!is.numeric(p_cuts) || length(p_cuts) < 2L ||
      anyNA(p_cuts) || any(!is.finite(p_cuts))) {
    stop("Selection p-value cut breaks must be finite numeric values.", call. = FALSE)
  }
  if (abs(p_cuts[1L]) > 1e-12 || abs(p_cuts[length(p_cuts)] - 1) > 1e-12) {
    stop("Selection p-value cut breaks must include endpoints 0 and 1.", call. = FALSE)
  }
  if (any(diff(p_cuts) <= 0)) {
    stop("Selection p-value cut breaks must be strictly increasing.", call. = FALSE)
  }

  p_cuts[1L]             <- 0
  p_cuts[length(p_cuts)] <- 1
  return(as.numeric(p_cuts))
}

.selection_prior_quantile <- function(prior, probability) {

  distribution <- prior[["distribution"]]
  parameters   <- prior[["parameters"]]
  truncation   <- prior[["truncation"]]
  lower        <- truncation[["lower"]]
  upper        <- truncation[["upper"]]

  q <- switch(
    distribution,
    "beta" = stats::qbeta(
      probability,
      shape1 = parameters[["alpha"]],
      shape2 = parameters[["beta"]]
    ),
    "normal" = stats::qnorm(
      probability,
      mean = parameters[["mean"]],
      sd   = parameters[["sd"]]
    ),
    Inf
  )

  if (is.finite(lower)) {
    q <- max(q, lower)
  }
  if (is.finite(upper)) {
    q <- min(q, upper)
  }

  return(q)
}

.selection_weightfunction_omega_bound <- function(prior, probability = .999) {

  weights <- prior[["weights"]]
  if (is.null(weights[["type"]])) {
    return(Inf)
  }

  if (identical(weights[["type"]], "fixed")) {
    return(max(weights[["omega"]], 1, na.rm = TRUE))
  }
  if (identical(weights[["type"]], "cumulative")) {
    return(1)
  }
  if (!identical(weights[["type"]], "independent")) {
    return(Inf)
  }

  prior <- weights[["prior"]]
  if (identical(weights[["scale"]], "omega")) {
    upper <- prior[["truncation"]][["upper"]]
    if (is.finite(upper)) {
      return(max(upper, 1))
    }
    return(max(.selection_prior_quantile(prior, probability), 1))
  }
  if (identical(weights[["scale"]], "log_omega")) {
    q <- .selection_prior_quantile(prior, probability)
    if (!is.finite(q) || q >= log(.Machine$double.xmax)) {
      return(Inf)
    }
    return(max(exp(q), 1))
  }

  return(Inf)
}

.selection_telescope_probability_check <- function(priors_bias, z_lower, z_upper) {

  z_threshold     <- 8
  omega_threshold <- 100
  finite_z        <- c(z_lower, z_upper)
  finite_z        <- finite_z[is.finite(finite_z)]
  reasons         <- character()

  if (any(abs(finite_z) > z_threshold)) {
    reasons <- c(reasons, "extreme selection cutpoints")
  }

  step_priors <- priors_bias[vapply(priors_bias, .prior_has_selection, logical(1))]
  omega_bound <- vapply(
    step_priors,
    .selection_weightfunction_omega_bound,
    numeric(1)
  )
  if (any(!is.finite(omega_bound) | omega_bound > omega_threshold)) {
    reasons <- c(reasons, "wide selection-weight priors")
  }

  if (length(reasons) > 0L) {
    warning(
      "Selection-model probability telescoping disabled: ",
      paste(unique(reasons), collapse = " and "),
      ".",
      call. = FALSE
    )
    return(FALSE)
  }

  return(TRUE)
}

.selection_spec <- function(priors, yi, sei, effect_direction, signed_data = FALSE) {

  priors_bias <- .selection_bias_priors(priors)
  if (!any(vapply(priors_bias, .prior_is_selection_kernel, logical(1)))) {
    return(NULL)
  }
  if (any(vapply(priors_bias, .prior_has_phacking, logical(1)))) {
    .selection_stop_phacking_deferred()
  }

  backend <- .selection_backend_spec(priors_bias)
  if (is.null(backend)) {
    return(NULL)
  }

  branch_kernel_mode   <- .selection_branch_kernel_modes(backend)
  jags_use_step_switch <- backend[["mode"]] == "step" &&
    any(branch_kernel_mode == SELKERNEL_NORMAL) &&
    any(branch_kernel_mode == SELKERNEL_STEP)

  p_cuts      <- backend[["step"]][["breaks"]]
  if (is.null(p_cuts)) {
    p_cuts <- c(0, 1)
  }
  p_cuts      <- .selection_assert_p_cuts(p_cuts)
  n_bins      <- length(p_cuts) - 1L
  has_step    <- backend[["mode"]] %in% c("step", "step_phack_power")
  has_phack   <- backend[["mode"]] %in% c("phack_power", "step_phack_power")
  phack       <- backend[["phacking"]]
  sign        <- if (signed_data || effect_direction != "negative") 1L else -1L
  z_lower     <- stats::qnorm(1 - p_cuts[-1])
  z_upper     <- stats::qnorm(1 - p_cuts[-length(p_cuts)])
  telescope_probabilities <- has_step &&
    .selection_telescope_probability_check(priors_bias, z_lower, z_upper)
  phack_q_values <- if (has_phack) .selection_phack_q_values(phack[["q"]]) else 1L
  phack_q        <- .selection_phack_static_q(phack_q_values)
  phack_source <- if (has_phack) .selection_phack_static_geometry(phack[["z_source"]], "source") else c(0, 0)
  phack_dest   <- if (has_phack) .selection_phack_static_geometry(phack[["z_destination"]], "destination") else c(0, 0)
  segments     <- .selection_segments(
    p_cuts          = p_cuts,
    phack_z_source  = phack_source,
    phack_z_dest    = phack_dest,
    has_phacking    = has_phack
  )

  jags_omega      <- if (has_step || has_phack) backend[["step"]][["coefficient"]] else "sel_omega"
  jags_alpha      <- if (has_phack) phack[["coefficient"]] else "sel_phack_alpha"
  jags_phack_kind <- if (has_phack) {
    kind_coefficient <- phack[["kind_coefficient"]]
    if (is.null(kind_coefficient)) "phack_kind" else kind_coefficient
  } else {
    "sel_phack_kind"
  }

  jags_data <- list(
    sel_z_lower             = .selection_jags_bounds(z_lower),
    sel_z_upper             = .selection_jags_bounds(z_upper),
    sel_obs_bin             = .selection_obs_bin(yi, sei, p_cuts, sign),
    sel_sign                = sign,
    sel_telescope_probabilities = as.integer(telescope_probabilities)
  )

  if (!identical(backend[["mode"]], "step")) {
    jags_data <- c(jags_data, list(
      sel_kernel_mode          = .selection_kernel_mode(backend[["mode"]]),
      phack_z_source           = phack_source,
      phack_z_dest             = phack_dest,
      sel_segment_bounds       = .selection_jags_bounds(segments[["bounds"]]),
      sel_segment_step_bin     = segments[["step_bin"]],
      sel_segment_phack_region = segments[["phack_region"]]
    ))

    if (!has_step && identical(jags_omega, "sel_omega")) {
      jags_data[[jags_omega]] <- rep(1, n_bins)
    }
    if (!has_phack) {
      jags_data[[jags_alpha]] <- 0
    } else {
      jags_data[["phack_z_source"]] <- NULL
      jags_data[["phack_z_dest"]]   <- NULL
    }
    if (identical(jags_phack_kind, "sel_phack_kind")) {
      jags_data[[jags_phack_kind]] <- if (has_phack) phack_q else 0L
    }
  }
  return(list(
    mode            = backend[["mode"]],
    kernel_mode     = .selection_kernel_mode(backend[["mode"]]),
    p_rule          = "signed_one_sided",
    p_cuts          = p_cuts,
    z_lower         = z_lower,
    z_upper         = z_upper,
    obs_bin         = jags_data[["sel_obs_bin"]],
    sign            = sign,
    n_bins          = n_bins,
    telescope_probabilities = telescope_probabilities,
    has_step        = has_step,
    has_phack       = has_phack,
    phack_q         = phack_q,
    phack_q_values  = phack_q_values,
    mixed_phack_q   = length(phack_q_values) > 1L,
    phack_z_source  = phack_source,
    phack_z_dest    = phack_dest,
    segments        = segments,
    branch_kernel_mode    = branch_kernel_mode,
    jags_use_step_switch  = jags_use_step_switch,
    jags_kernel_mode      = if (jags_use_step_switch) "sel_kernel_mode_active" else "sel_kernel_mode",
    jags_kernel_mode_expr = .selection_jags_kernel_mode_expression(branch_kernel_mode),
    jags_omega      = jags_omega,
    jags_alpha      = jags_alpha,
    jags_phack_kind = jags_phack_kind,
    jags_code       = paste(
      c(backend[["prior_code"]], backend[["transform_code"]]),
      collapse = "\n"
    ),
    backend_data    = backend[["data"]],
    jags_data       = jags_data
  ))
}

.selection_phack_static_q <- function(q) {

  q <- as.integer(q)
  if (length(q) == 0L) {
    return(1L)
  }
  if (length(unique(q)) == 1L) {
    return(q[1L])
  }

  # The active JAGS node phack_kind selects the form for fitted mixtures. The
  # native R wrapper still needs one fallback q for inactive/default calls.
  return(1L)
}

.selection_phack_q_values <- function(q) {

  q <- sort(unique(as.integer(q)))
  q <- q[!is.na(q)]
  if (length(q) == 0L) {
    q <- 1L
  }
  return(q)
}

.selection_phack_static_geometry <- function(z, label) {

  if (is.null(z)) {
    return(c(0, 0))
  }

  if (is.matrix(z)) {
    if (ncol(z) != 2L) {
      stop("P-hacking ", label, " geometry must have two z cut points.", call. = FALSE)
    }
    unique_rows <- unique(z)
    if (nrow(unique_rows) > 1L) {
      stop(
        "RoBMA currently requires p-hacking mixture components to share ",
        label, " geometry.",
        call. = FALSE
      )
    }
    return(as.numeric(unique_rows[1L, ]))
  }

  z <- as.numeric(z)
  if (length(z) != 2L) {
    stop("P-hacking ", label, " geometry must have two z cut points.", call. = FALSE)
  }

  return(z)
}

.selection_prepare_native_args <- function(selection_spec, S, alpha = NULL,
                                           phack_kind = NULL,
                                           kernel_mode = NULL) {

  if (is.null(alpha)) {
    alpha <- rep(0, S)
  }
  if (is.null(phack_kind)) {
    if (isTRUE(selection_spec[["mixed_phack_q"]])) {
      stop(
        "'phack_kind' is required for mixed linear/quadratic p-hacking forms.",
        call. = FALSE
      )
    }
    phack_kind <- rep(if (selection_spec[["has_phack"]]) selection_spec[["phack_q"]] else 0L, S)
  }
  if (is.null(kernel_mode)) {
    kernel_mode <- rep(selection_spec[["kernel_mode"]], S)
  }

  return(list(
    alpha       = alpha,
    phack_kind  = phack_kind,
    kernel_mode = kernel_mode
  ))
}

.selection_telescope_probabilities <- function(selection_spec) {

  return(isTRUE(selection_spec[["telescope_probabilities"]]))
}

.has_native_selnorm_kernel <- function() {

  symbols <- c(
    "RoBMA_selnorm_kernel_loglik_matrix",
    "RoBMA_selnorm_kernel_log_norm_matrix",
    "RoBMA_selnorm_kernel_cdf_matrix",
    "RoBMA_selnorm_kernel_moments_matrix",
    "RoBMA_selnorm_kernel_rng_matrix",
    "RoBMA_selnorm_kernel_weighted_summary"
  )

  return(all(vapply(symbols, is.loaded, logical(1), PACKAGE = "RoBMA")))
}

.selnorm_kernel_loglik_matrix <- function(yi, mu_num, sigma_num,
                                          mu_norm = mu_num,
                                          sigma_norm = sigma_num,
                                          sei, omega, selection_spec,
                                          alpha = NULL,
                                          phack_kind = NULL,
                                          kernel_mode = NULL,
                                          weights = NULL) {

  S <- nrow(mu_num)
  K <- ncol(mu_num)

  if (is.null(weights)) {
    weights <- rep(1, K)
  }
  native_args <- .selection_prepare_native_args(selection_spec, S, alpha, phack_kind, kernel_mode)
  alpha       <- native_args[["alpha"]]
  phack_kind  <- native_args[["phack_kind"]]
  kernel_mode <- native_args[["kernel_mode"]]

  if (!.has_native_selnorm_kernel()) {
    stop("The selected-normal native kernel is not loaded.", call. = FALSE)
  }

  return(.Call(
    "RoBMA_selnorm_kernel_loglik_matrix",
    .native_numeric_vector(yi),
    .native_numeric_matrix(mu_num),
    .native_numeric_matrix(sigma_num),
    .native_numeric_matrix(mu_norm),
    .native_numeric_matrix(sigma_norm),
    .native_numeric_vector(sei),
    .native_numeric_vector(weights),
    .native_numeric_matrix(omega),
    .native_numeric_vector(alpha),
    .native_integer_vector(phack_kind),
    .native_integer_vector(kernel_mode),
    .native_numeric_vector(selection_spec[["z_lower"]]),
    .native_numeric_vector(selection_spec[["z_upper"]]),
    .native_integer_vector(selection_spec[["obs_bin"]]),
    .native_integer_vector(selection_spec[["sign"]]),
    .native_integer_vector(selection_spec[["phack_q"]]),
    .native_numeric_vector(selection_spec[["phack_z_source"]]),
    .native_numeric_vector(selection_spec[["phack_z_dest"]]),
    .native_numeric_vector(selection_spec[["segments"]][["bounds"]]),
    .native_integer_vector(selection_spec[["segments"]][["step_bin"]]),
    .native_integer_vector(selection_spec[["segments"]][["phack_region"]]),
    .selection_telescope_probabilities(selection_spec),
    PACKAGE = "RoBMA"
  ))
}

.selnorm_kernel_log_norm_matrix <- function(mean, sd, sei, omega,
                                            selection_spec, alpha = NULL,
                                            phack_kind = NULL,
                                            kernel_mode = NULL) {

  S <- nrow(mean)

  native_args <- .selection_prepare_native_args(selection_spec, S, alpha, phack_kind, kernel_mode)
  alpha       <- native_args[["alpha"]]
  phack_kind  <- native_args[["phack_kind"]]
  kernel_mode <- native_args[["kernel_mode"]]

  if (!.has_native_selnorm_kernel()) {
    stop("The selected-normal native kernel is not loaded.", call. = FALSE)
  }

  return(.Call(
    "RoBMA_selnorm_kernel_log_norm_matrix",
    .native_numeric_matrix(mean),
    .native_numeric_matrix(sd),
    .native_numeric_vector(sei),
    .native_numeric_matrix(omega),
    .native_numeric_vector(alpha),
    .native_integer_vector(phack_kind),
    .native_integer_vector(kernel_mode),
    .native_numeric_vector(selection_spec[["z_lower"]]),
    .native_numeric_vector(selection_spec[["z_upper"]]),
    .native_integer_vector(selection_spec[["sign"]]),
    .native_integer_vector(selection_spec[["phack_q"]]),
    .native_numeric_vector(selection_spec[["phack_z_source"]]),
    .native_numeric_vector(selection_spec[["phack_z_dest"]]),
    .native_numeric_vector(selection_spec[["segments"]][["bounds"]]),
    .native_integer_vector(selection_spec[["segments"]][["step_bin"]]),
    .native_integer_vector(selection_spec[["segments"]][["phack_region"]]),
    .selection_telescope_probabilities(selection_spec),
    PACKAGE = "RoBMA"
  ))
}

.selnorm_kernel_cdf_matrix <- function(q, mean, sd, sei, omega,
                                       selection_spec, alpha = NULL,
                                       phack_kind = NULL,
                                       kernel_mode = NULL,
                                       lower.tail = TRUE) {

  S <- nrow(mean)

  native_args <- .selection_prepare_native_args(selection_spec, S, alpha, phack_kind, kernel_mode)
  alpha       <- native_args[["alpha"]]
  phack_kind  <- native_args[["phack_kind"]]
  kernel_mode <- native_args[["kernel_mode"]]
  .selection_require_native_no_active_phack(
    alpha       = alpha,
    phack_kind  = phack_kind,
    kernel_mode = kernel_mode,
    S           = S,
    caller      = ".selnorm_kernel_cdf_matrix()"
  )

  if (!.has_native_selnorm_kernel()) {
    stop("The selected-normal native kernel is not loaded.", call. = FALSE)
  }

  return(.Call(
    "RoBMA_selnorm_kernel_cdf_matrix",
    .native_numeric_vector(q),
    .native_numeric_matrix(mean),
    .native_numeric_matrix(sd),
    .native_numeric_vector(sei),
    .native_numeric_matrix(omega),
    .native_numeric_vector(alpha),
    .native_integer_vector(phack_kind),
    .native_integer_vector(kernel_mode),
    .native_numeric_vector(selection_spec[["z_lower"]]),
    .native_numeric_vector(selection_spec[["z_upper"]]),
    .native_integer_vector(selection_spec[["sign"]]),
    .native_integer_vector(selection_spec[["phack_q"]]),
    .native_numeric_vector(selection_spec[["phack_z_source"]]),
    .native_numeric_vector(selection_spec[["phack_z_dest"]]),
    .native_numeric_vector(selection_spec[["segments"]][["bounds"]]),
    .native_integer_vector(selection_spec[["segments"]][["step_bin"]]),
    .native_integer_vector(selection_spec[["segments"]][["phack_region"]]),
    as.logical(lower.tail),
    .selection_telescope_probabilities(selection_spec),
    PACKAGE = "RoBMA"
  ))
}

.selnorm_kernel_moments_matrix <- function(mean, sd, sei, omega,
                                           selection_spec, alpha = NULL,
                                           phack_kind = NULL,
                                           kernel_mode = NULL) {

  S <- nrow(mean)

  native_args <- .selection_prepare_native_args(selection_spec, S, alpha, phack_kind, kernel_mode)
  alpha       <- native_args[["alpha"]]
  phack_kind  <- native_args[["phack_kind"]]
  kernel_mode <- native_args[["kernel_mode"]]
  .selection_require_native_no_active_phack(
    alpha       = alpha,
    phack_kind  = phack_kind,
    kernel_mode = kernel_mode,
    S           = S,
    caller      = ".selnorm_kernel_moments_matrix()"
  )

  if (!.has_native_selnorm_kernel()) {
    stop("The selected-normal native kernel is not loaded.", call. = FALSE)
  }

  return(.Call(
    "RoBMA_selnorm_kernel_moments_matrix",
    .native_numeric_matrix(mean),
    .native_numeric_matrix(sd),
    .native_numeric_vector(sei),
    .native_numeric_matrix(omega),
    .native_numeric_vector(alpha),
    .native_integer_vector(phack_kind),
    .native_integer_vector(kernel_mode),
    .native_numeric_vector(selection_spec[["z_lower"]]),
    .native_numeric_vector(selection_spec[["z_upper"]]),
    .native_integer_vector(selection_spec[["sign"]]),
    .native_integer_vector(selection_spec[["phack_q"]]),
    .native_numeric_vector(selection_spec[["phack_z_source"]]),
    .native_numeric_vector(selection_spec[["phack_z_dest"]]),
    .native_numeric_vector(selection_spec[["segments"]][["bounds"]]),
    .native_integer_vector(selection_spec[["segments"]][["step_bin"]]),
    .native_integer_vector(selection_spec[["segments"]][["phack_region"]]),
    .selection_telescope_probabilities(selection_spec),
    PACKAGE = "RoBMA"
  ))
}

.selnorm_kernel_rng_matrix <- function(mean, sd, sei, omega,
                                       selection_spec, alpha = NULL,
                                       phack_kind = NULL,
                                       kernel_mode = NULL) {

  S <- nrow(mean)

  native_args <- .selection_prepare_native_args(selection_spec, S, alpha, phack_kind, kernel_mode)
  alpha       <- native_args[["alpha"]]
  phack_kind  <- native_args[["phack_kind"]]
  kernel_mode <- native_args[["kernel_mode"]]
  .selection_require_native_no_active_phack(
    alpha       = alpha,
    phack_kind  = phack_kind,
    kernel_mode = kernel_mode,
    S           = S,
    caller      = ".selnorm_kernel_rng_matrix()"
  )

  if (!.has_native_selnorm_kernel()) {
    stop("The selected-normal native kernel is not loaded.", call. = FALSE)
  }

  return(.Call(
    "RoBMA_selnorm_kernel_rng_matrix",
    .native_numeric_matrix(mean),
    .native_numeric_matrix(sd),
    .native_numeric_vector(sei),
    .native_numeric_matrix(omega),
    .native_numeric_vector(alpha),
    .native_integer_vector(phack_kind),
    .native_integer_vector(kernel_mode),
    .native_numeric_vector(selection_spec[["z_lower"]]),
    .native_numeric_vector(selection_spec[["z_upper"]]),
    .native_integer_vector(selection_spec[["sign"]]),
    .native_integer_vector(selection_spec[["phack_q"]]),
    .native_numeric_vector(selection_spec[["phack_z_source"]]),
    .native_numeric_vector(selection_spec[["phack_z_dest"]]),
    .native_numeric_vector(selection_spec[["segments"]][["bounds"]]),
    .native_integer_vector(selection_spec[["segments"]][["step_bin"]]),
    .native_integer_vector(selection_spec[["segments"]][["phack_region"]]),
    .selection_telescope_probabilities(selection_spec),
    PACKAGE = "RoBMA"
  ))
}

.selnorm_kernel_weighted_summary <- function(yi, mean, sd, sei, psis_weights,
                                             omega, selection_spec,
                                             alpha = NULL, phack_kind = NULL,
                                             kernel_mode = NULL) {

  S <- nrow(mean)

  native_args <- .selection_prepare_native_args(selection_spec, S, alpha, phack_kind, kernel_mode)
  alpha       <- native_args[["alpha"]]
  phack_kind  <- native_args[["phack_kind"]]
  kernel_mode <- native_args[["kernel_mode"]]
  .selection_require_native_no_active_phack(
    alpha       = alpha,
    phack_kind  = phack_kind,
    kernel_mode = kernel_mode,
    S           = S,
    caller      = ".selnorm_kernel_weighted_summary()"
  )

  if (!.has_native_selnorm_kernel()) {
    stop("The selected-normal native kernel is not loaded.", call. = FALSE)
  }

  return(.Call(
    "RoBMA_selnorm_kernel_weighted_summary",
    .native_numeric_vector(yi),
    .native_numeric_matrix(mean),
    .native_numeric_matrix(sd),
    .native_numeric_vector(sei),
    .native_numeric_matrix(psis_weights),
    .native_numeric_matrix(omega),
    .native_numeric_vector(alpha),
    .native_integer_vector(phack_kind),
    .native_integer_vector(kernel_mode),
    .native_numeric_vector(selection_spec[["z_lower"]]),
    .native_numeric_vector(selection_spec[["z_upper"]]),
    .native_integer_vector(selection_spec[["sign"]]),
    .native_integer_vector(selection_spec[["phack_q"]]),
    .native_numeric_vector(selection_spec[["phack_z_source"]]),
    .native_numeric_vector(selection_spec[["phack_z_dest"]]),
    .native_numeric_vector(selection_spec[["segments"]][["bounds"]]),
    .native_integer_vector(selection_spec[["segments"]][["step_bin"]]),
    .native_integer_vector(selection_spec[["segments"]][["phack_region"]]),
    .selection_telescope_probabilities(selection_spec),
    PACKAGE = "RoBMA"
  ))
}

.extract_selection_omega_samples <- function(posterior_samples, selection_spec) {

  omega_name <- selection_spec[["jags_omega"]]
  omega <- NULL
  if (!is.null(omega_name) && !identical(omega_name, "sel_omega")) {
    omega <- .extract_indexed_parameter_samples(
      posterior_samples,
      parameter  = omega_name,
      n_expected = selection_spec[["n_bins"]],
      required   = FALSE
    )
  }

  if (is.null(omega)) {
    omega <- .extract_indexed_parameter_samples(
      posterior_samples,
      parameter  = "omega",
      n_expected = selection_spec[["n_bins"]],
      required   = FALSE
    )
  }

  if (is.null(omega)) {
    omega <- matrix(1, nrow = nrow(posterior_samples), ncol = selection_spec[["n_bins"]])
    colnames(omega) <- paste0("sel_omega[", seq_len(selection_spec[["n_bins"]]), "]")
  }

  storage.mode(omega) <- "double"
  return(omega)
}

.extract_selection_alpha_samples <- function(posterior_samples, selection_spec) {

  if (!selection_spec[["has_phack"]]) {
    return(rep(0, nrow(posterior_samples)))
  }

  alpha_name <- selection_spec[["jags_alpha"]]
  if (alpha_name %in% colnames(posterior_samples)) {
    return(as.numeric(posterior_samples[, alpha_name]))
  }

  alpha_cols <- grep("^alpha(\\[|$)", colnames(posterior_samples), value = TRUE)
  if (length(alpha_cols) > 0L) {
    return(as.numeric(posterior_samples[, alpha_cols[1L]]))
  }

  return(rep(0, nrow(posterior_samples)))
}

.extract_selection_phack_kind <- function(posterior_samples, selection_spec) {

  if (!selection_spec[["has_phack"]]) {
    return(rep(0L, nrow(posterior_samples)))
  }
  kind_name <- selection_spec[["jags_phack_kind"]]
  if (!is.null(kind_name) && kind_name %in% colnames(posterior_samples)) {
    return(as.integer(posterior_samples[, kind_name]))
  }
  if ("phack_kind" %in% colnames(posterior_samples)) {
    return(as.integer(posterior_samples[, "phack_kind"]))
  }
  if (isTRUE(selection_spec[["mixed_phack_q"]])) {
    stop(
      "Missing posterior 'phack_kind' column for mixed linear/quadratic p-hacking forms.",
      call. = FALSE
    )
  }
  return(rep(selection_spec[["phack_q"]], nrow(posterior_samples)))
}

.selection_context <- function(object, posterior_samples = NULL, newdata = NULL) {

  if (!.is_weightfunction(object)) {
    return(NULL)
  }

  posterior_samples <- .get_posterior_samples(object[["fit"]], posterior_samples)
  outcome_data      <- if (is.null(newdata)) {
    object[["data"]][["outcome"]]
  } else if (is.list(newdata) && !is.null(newdata[["outcome"]])) {
    newdata[["outcome"]]
  } else {
    newdata
  }
  yi                <- outcome_data[["yi"]]
  sei               <- outcome_data[["sei"]]

  selection_spec <- .selection_spec(
    priors           = object[["priors"]],
    yi               = yi,
    sei              = sei,
    effect_direction = .effect_direction(object),
    signed_data      = FALSE
  )

  if (is.null(selection_spec)) {
    return(NULL)
  }

  bias_indicator <- .extract_bias_indicator(object, posterior_samples = posterior_samples)
  use_normal     <- .extract_use_normal(object, posterior_samples = posterior_samples)
  kernel_mode    <- rep(selection_spec[["kernel_mode"]], nrow(posterior_samples))
  kernel_mode[use_normal] <- SELKERNEL_NORMAL

  selection_context <- selection_spec
  selection_context[["family"]]         <- selection_spec[["mode"]]
  selection_context[["yi"]]             <- yi
  selection_context[["sei"]]            <- sei
  selection_context[["omega"]]          <- .extract_selection_omega_samples(posterior_samples, selection_spec)
  selection_context[["alpha"]]          <- .extract_selection_alpha_samples(posterior_samples, selection_spec)
  selection_context[["phack_kind"]]     <- .extract_selection_phack_kind(posterior_samples, selection_spec)
  selection_context[["kernel_mode"]]    <- kernel_mode
  selection_context[["bias_indicator"]] <- bias_indicator
  selection_context[["use_normal"]]     <- use_normal

  return(selection_context)
}

.selection_context_subset_rows <- function(selection_context, rows) {

  out <- selection_context
  S   <- nrow(selection_context[["omega"]])

  for (name in c("omega", "alpha", "phack_kind", "kernel_mode",
                 "bias_indicator", "use_normal")) {
    value <- out[[name]]
    if (is.null(value)) {
      next
    }
    if (is.matrix(value) && nrow(value) == S) {
      out[[name]] <- value[rows, , drop = FALSE]
    } else if (!is.matrix(value) && length(value) == S) {
      out[[name]] <- value[rows]
    }
  }

  return(out)
}

.selection_active_phack <- function(selection_context) {

  S           <- nrow(selection_context[["omega"]])
  alpha       <- selection_context[["alpha"]]
  phack_kind  <- selection_context[["phack_kind"]]
  kernel_mode <- selection_context[["kernel_mode"]]

  if (length(alpha) == 1L) {
    alpha <- rep(alpha, S)
  }
  if (length(phack_kind) == 1L) {
    phack_kind <- rep(phack_kind, S)
  }
  if (length(kernel_mode) == 1L) {
    kernel_mode <- rep(kernel_mode, S)
  }

  return(
    kernel_mode %in% c(SELKERNEL_PHACK_POWER, SELKERNEL_STEP_PHACK_POWER) &
      phack_kind > 0L &
      alpha > 0
  )
}

.selection_row_arg <- function(x, S, name) {

  if (length(x) == 1L) {
    return(rep(x, S))
  }
  if (length(x) == S) {
    return(x)
  }

  stop("'", name, "' must have length 1 or one value per posterior sample.",
       call. = FALSE)
}

.selection_active_phack_inputs <- function(alpha, phack_kind, kernel_mode, S) {

  alpha       <- .selection_row_arg(alpha, S, "alpha")
  phack_kind  <- .selection_row_arg(phack_kind, S, "phack_kind")
  kernel_mode <- .selection_row_arg(kernel_mode, S, "kernel_mode")

  return(
    kernel_mode %in% c(SELKERNEL_PHACK_POWER, SELKERNEL_STEP_PHACK_POWER) &
      phack_kind > 0L &
      alpha > 0
  )
}

.selection_require_native_no_active_phack <- function(alpha, phack_kind,
                                                      kernel_mode, S, caller) {

  if (any(.selection_active_phack_inputs(alpha, phack_kind, kernel_mode, S))) {
    stop(
      caller, " does not yet support active p-hacking selection kernels.",
      call. = FALSE
    )
  }

  return(invisible(TRUE))
}

.selection_require_step_evaluable <- function(selection_context, caller) {

  if (any(.selection_active_phack(selection_context))) {
    stop(
      caller, " does not yet support active p-hacking selection kernels.",
      call. = FALSE
    )
  }

  return(invisible(TRUE))
}

.selection_interval_prob_vec <- function(lower, upper, mean, sd) {

  out <- stats::pnorm(upper, mean = mean, sd = sd) -
    stats::pnorm(lower, mean = mean, sd = sd)
  out[lower >= upper] <- 0
  out[out < 0 & out > -1e-12] <- 0
  return(out)
}

.selection_step_log_norm_matrix <- function(mean, sd, sei, selection_context) {

  .selection_require_step_evaluable(selection_context, ".selection_step_log_norm_matrix()")

  return(.selnorm_kernel_log_norm_matrix(
    mean           = mean,
    sd             = sd,
    sei            = sei,
    omega          = selection_context[["omega"]],
    selection_spec = selection_context,
    alpha          = selection_context[["alpha"]],
    phack_kind     = selection_context[["phack_kind"]],
    kernel_mode    = selection_context[["kernel_mode"]]
  ))
}

.selection_step_cdf_matrix <- function(q, mean, sd, sei, selection_context,
                                       lower.tail = TRUE) {

  .selection_require_step_evaluable(selection_context, ".selection_step_cdf_matrix()")

  return(.selnorm_kernel_cdf_matrix(
    q              = q,
    mean           = mean,
    sd             = sd,
    sei            = sei,
    omega          = selection_context[["omega"]],
    selection_spec = selection_context,
    alpha          = selection_context[["alpha"]],
    phack_kind     = selection_context[["phack_kind"]],
    kernel_mode    = selection_context[["kernel_mode"]],
    lower.tail     = lower.tail
  ))
}

.selection_step_logpdf_matrix <- function(y, mean, sd, sei, selection_context) {

  .selection_require_step_evaluable(selection_context, ".selection_step_logpdf_matrix()")

  mean <- as.matrix(mean)
  sd   <- as.matrix(sd)

  if (length(y) == 1L) {
    y <- rep(y, ncol(mean))
  }

  selection_context[["obs_bin"]] <- .selection_obs_bin(
    y,
    sei,
    selection_context[["p_cuts"]],
    selection_context[["sign"]]
  )

  return(.selnorm_kernel_loglik_matrix(
    yi             = y,
    mu_num         = mean,
    sigma_num      = sd,
    sei            = sei,
    omega          = selection_context[["omega"]],
    selection_spec = selection_context,
    alpha          = selection_context[["alpha"]],
    phack_kind     = selection_context[["phack_kind"]],
    kernel_mode    = selection_context[["kernel_mode"]]
  ))
}

.selection_step_moments_matrix <- function(mean, sd, sei, selection_context) {

  .selection_require_step_evaluable(selection_context, ".selection_step_moments_matrix()")

  return(.selnorm_kernel_moments_matrix(
    mean           = mean,
    sd             = sd,
    sei            = sei,
    omega          = selection_context[["omega"]],
    selection_spec = selection_context,
    alpha          = selection_context[["alpha"]],
    phack_kind     = selection_context[["phack_kind"]],
    kernel_mode    = selection_context[["kernel_mode"]]
  ))
}

.selection_step_rng_matrix <- function(mean, sd, sei, selection_context) {

  .selection_require_step_evaluable(selection_context, ".selection_step_rng_matrix()")

  mean <- as.matrix(mean)
  sd   <- as.matrix(sd)

  return(.selnorm_kernel_rng_matrix(
    mean           = mean,
    sd             = sd,
    sei            = sei,
    omega          = selection_context[["omega"]],
    selection_spec = selection_context,
    alpha          = selection_context[["alpha"]],
    phack_kind     = selection_context[["phack_kind"]],
    kernel_mode    = selection_context[["kernel_mode"]]
  ))
}

.selection_step_weighted_summary <- function(yi, mean, sd, sei, psis_weights,
                                             selection_context) {

  .selection_require_step_evaluable(selection_context, ".selection_step_weighted_summary()")

  mean         <- as.matrix(mean)
  sd           <- as.matrix(sd)
  psis_weights <- as.matrix(psis_weights)

  return(.selnorm_kernel_weighted_summary(
    yi             = yi,
    mean           = mean,
    sd             = sd,
    sei            = sei,
    psis_weights   = psis_weights,
    omega          = selection_context[["omega"]],
    selection_spec = selection_context,
    alpha          = selection_context[["alpha"]],
    phack_kind     = selection_context[["phack_kind"]],
    kernel_mode    = selection_context[["kernel_mode"]]
  ))
}

context("Selection kernel")
skip_on_cran()

.test_step_spec <- function(yi, sei, effect_direction = "positive") {

  prior <- BayesTools::prior_weightfunction(
    side    = "one-sided",
    steps   = c(.025, .05, .50),
    weights = BayesTools::wf_fixed(c(1, .7, .35, .2))
  )

  .selection_spec(
    priors           = list(outcome = list(bias = prior)),
    yi               = yi,
    sei              = sei,
    effect_direction = effect_direction,
    signed_data      = FALSE
  )
}

.test_two_sided_step_spec <- function(yi, sei, effect_direction = "positive") {

  prior <- BayesTools::prior_weightfunction(
    side    = "two-sided",
    steps   = c(.05, .10),
    weights = BayesTools::wf_fixed(c(1, .7, .35))
  )

  .selection_spec(
    priors           = list(outcome = list(bias = prior)),
    yi               = yi,
    sei              = sei,
    effect_direction = effect_direction,
    signed_data      = FALSE
  )
}

.test_step_log_norm_reference <- function(mean, sd, sei, omega, spec) {

  S   <- nrow(mean)
  K   <- ncol(mean)
  out <- matrix(NA_real_, nrow = S, ncol = K)

  for (s in seq_len(S)) {
    for (k in seq_len(K)) {
      mean_z   <- spec[["sign"]] * mean[s, k] / sei[k]
      sd_z     <- sd[s, k] / sei[k]
      log_mass <- vapply(seq_len(spec[["n_bins"]]), function(b) {
        log(omega[s, b]) + .test_interval_log_prob(
          spec[["z_lower"]][b],
          spec[["z_upper"]][b],
          mean_z,
          sd_z
        )
      }, numeric(1))

      out[s, k] <- .test_logsumexp(log_mass)
    }
  }

  return(out)
}

test_that("selection model fit data and syntax use only the selected-normal kernel", {

  prior_bias <- BayesTools::prior_weightfunction(
    side    = "one-sided",
    steps   = c(.025, .05),
    weights = BayesTools::wf_fixed(c(1, .5, .25))
  )

  object <- bselmodel(
    yi                        = c(.1, .2, .3),
    sei                       = c(.1, .1, .1),
    measure                   = "SMD",
    prior_bias                = prior_bias,
    prior_unit_information_sd = 1,
    only_priors               = TRUE,
    silent                    = TRUE
  )

  fit_data <- .create_fit_data(object[["data"]], object[["priors"]])
  syntax   <- .create_model_syntax(object[["data"]], object[["priors"]])

  expect_true(all(c(
    "sel_z_lower", "sel_z_upper", "sel_obs_bin", "sel_sign"
  ) %in% names(fit_data)))
  expect_false(any(grepl("sel_phack|phack_z|sel_segment|sel_kernel_mode", names(fit_data))))
  expect_match(syntax, "dselnorm_step", fixed = TRUE)
  expect_match(syntax, "sel_total_sd\\[i\\]")
  expect_false(grepl("dselnorm_kernel|sel_phack|phack_z|sel_segment|sel_kernel_mode", syntax))
})

test_that("mixed normal-step bias syntax uses scalar step switch", {

  prior_bias <- BayesTools::prior_weightfunction(
    side    = "one-sided",
    steps   = c(.025, .05),
    weights = BayesTools::wf_fixed(c(1, .5, .25))
  )

  object <- RoBMA(
    yi                        = c(.1, .2, .3),
    sei                       = c(.1, .1, .1),
    measure                   = "SMD",
    prior_bias                = prior_bias,
    prior_bias_null           = BayesTools::prior_none(),
    prior_effect              = BayesTools::prior("normal", parameters = list(0, 1)),
    prior_effect_null         = NULL,
    prior_heterogeneity       = BayesTools::prior("invgamma", parameters = list(1, .15)),
    prior_heterogeneity_null  = NULL,
    prior_unit_information_sd = 1,
    only_priors               = TRUE,
    silent                    = TRUE
  )

  fit_data <- .create_fit_data(object[["data"]], object[["priors"]])
  syntax   <- .create_model_syntax(object[["data"]], object[["priors"]])

  expect_true(all(c(
    "sel_z_lower", "sel_z_upper", "sel_obs_bin", "sel_sign"
  ) %in% names(fit_data)))
  expect_false(any(grepl("sel_phack|phack_z|sel_segment", names(fit_data))))
  expect_match(syntax, "sel_kernel_mode_active", fixed = TRUE)
  expect_match(syntax, "dselnorm_step_switch", fixed = TRUE)
  expect_false(grepl("dselnorm_kernel", syntax, fixed = TRUE))
})

test_that("selection spec sets probability telescoping flag once", {

  yi  <- c(.1, .2, .3)
  sei <- c(.1, .1, .1)
  safe_prior <- BayesTools::prior_weightfunction(
    side    = "one-sided",
    steps   = .05,
    weights = BayesTools::wf_fixed(c(1, .5))
  )
  safe_spec <- .selection_spec(
    priors           = list(outcome = list(bias = safe_prior)),
    yi               = yi,
    sei              = sei,
    effect_direction = "positive",
    signed_data      = TRUE
  )

  expect_true(safe_spec[["telescope_probabilities"]])
  expect_identical(safe_spec[["jags_data"]][["sel_telescope_probabilities"]], 1L)

  wide_prior <- BayesTools::prior_weightfunction(
    side    = "one-sided",
    steps   = .05,
    weights = BayesTools::wf_fixed(c(1, 200))
  )
  expect_warning(
    wide_spec <- .selection_spec(
      priors           = list(outcome = list(bias = wide_prior)),
      yi               = yi,
      sei              = sei,
      effect_direction = "positive",
      signed_data      = TRUE
    ),
    "probability telescoping disabled"
  )

  expect_false(wide_spec[["telescope_probabilities"]])
  expect_identical(wide_spec[["jags_data"]][["sel_telescope_probabilities"]], 0L)
})

test_that("branch kernel modes follow BayesTools phack and combined labels", {

  selection <- BayesTools::prior_weightfunction(
    side    = "one-sided",
    steps   = c(.025),
    weights = BayesTools::wf_fixed(c(1, .5))
  )
  phacking <- BayesTools::prior_phacking(form = "linear")
  bias <- BayesTools::prior_mixture(list(
    BayesTools::prior_none(),
    selection,
    phacking,
    BayesTools::prior_bias(selection, phacking)
  ))
  backend <- BayesTools::selection_backend_spec(bias)

  expect_equal(
    backend[["branch_type"]],
    c("none", "weightfunction", "phack", "combined")
  )
  expect_equal(
    .selection_branch_kernel_modes(backend),
    c(
      SELKERNEL_NORMAL,
      SELKERNEL_STEP,
      SELKERNEL_PHACK_POWER,
      SELKERNEL_STEP_PHACK_POWER
    )
  )
})

test_that("single two-sided bselmodel uses active full-grid omega in JAGS", {

  prior_bias <- BayesTools::prior_weightfunction(
    side    = "two-sided",
    steps   = c(.05, .10),
    weights = BayesTools::wf_fixed(c(1, .5, .25))
  )

  object <- bselmodel(
    yi                        = c(.24, .31, -.18, .05),
    sei                       = c(.10, .12, .09, .20),
    measure                   = "SMD",
    prior_bias                = prior_bias,
    prior_unit_information_sd = 1,
    only_priors               = TRUE,
    silent                    = TRUE
  )

  fit_data <- .create_fit_data(object[["data"]], object[["priors"]])
  priors   <- .create_fit_priors(object[["data"]], object[["priors"]])
  syntax   <- BayesTools::JAGS_add_priors(
    .create_model_syntax(object[["data"]], object[["priors"]]),
    priors
  )

  expect_equal(length(fit_data[["sel_z_lower"]]), 5L)
  expect_equal(length(fit_data[["sel_z_upper"]]), 5L)
  expect_match(syntax, "omega_local\\[2\\] <- 0.5")
  expect_match(syntax, "omega\\[4\\] <- omega_local\\[2\\]")
  expect_match(syntax, "omega\\[5\\] <- omega_local\\[1\\]")
  expect_false(grepl("omega\\[6\\]", syntax))
  expect_true("omega" %in% BayesTools::JAGS_to_monitor(priors))

  skip_if_not_installed("rjags")

  fit <- suppressWarnings(BayesTools::JAGS_fit(
    model_syntax      = .create_model_syntax(object[["data"]], object[["priors"]]),
    data              = fit_data,
    prior_list        = priors,
    chains            = 1,
    adapt             = 50,
    burnin            = 50,
    sample            = 100,
    required_packages = "RoBMA",
    silent            = TRUE,
    seed              = 91
  ))

  expect_false(inherits(fit, "error"))
  posterior <- as.matrix(coda::as.mcmc(fit))
  expect_true(all(paste0("omega[", 1:5, "]") %in% colnames(posterior)))
  expect_equal(unname(posterior[, "omega[2]"]), rep(.5, nrow(posterior)))
  expect_equal(unname(posterior[, "omega[4]"]), rep(.5, nrow(posterior)))
})

test_that("JAGS permits fixed zero weights for empty p-value bins", {

  prior_bias <- BayesTools::prior_weightfunction(
    side    = "one-sided",
    steps   = c(.025, .05),
    weights = BayesTools::wf_fixed(c(1, 0, .5))
  )
  sei <- c(.10, .10)
  yi  <- stats::qnorm(c(.01, .20), lower.tail = FALSE) * sei

  object <- bselmodel(
    yi                        = yi,
    sei                       = sei,
    measure                   = "SMD",
    prior_bias                = prior_bias,
    prior_unit_information_sd = 1,
    only_priors               = TRUE,
    silent                    = TRUE
  )

  fit_data <- .create_fit_data(object[["data"]], object[["priors"]])
  priors   <- .create_fit_priors(object[["data"]], object[["priors"]])
  syntax   <- BayesTools::JAGS_add_priors(
    .create_model_syntax(object[["data"]], object[["priors"]]),
    priors
  )

  expect_equal(fit_data[["sel_obs_bin"]], c(1L, 3L))
  expect_match(syntax, "omega[2] <- 0", fixed = TRUE)

  skip_if_not_installed("rjags")

  fit <- suppressWarnings(BayesTools::JAGS_fit(
    model_syntax      = .create_model_syntax(object[["data"]], object[["priors"]]),
    data              = fit_data,
    prior_list        = priors,
    chains            = 1,
    adapt             = 50,
    burnin            = 50,
    sample            = 100,
    required_packages = "RoBMA",
    silent            = TRUE,
    seed              = 92
  ))

  posterior <- as.matrix(coda::as.mcmc(fit))
  expect_equal(unname(posterior[, "omega[2]"]), rep(0, nrow(posterior)))
})

test_that("selection p-value bins are upper-closed at step boundaries", {

  p_cuts <- c(0, .025, .50, 1)
  sei    <- rep(1, 5)
  p_val  <- c(.025, .0250001, .50, .5000001, .0249999)
  yi     <- stats::qnorm(p_val, lower.tail = FALSE) * sei

  expect_equal(
    .selection_obs_bin(yi, sei, p_cuts, sign = 1L),
    c(1L, 2L, 2L, 3L, 1L)
  )
  expect_equal(.selection_step_bin_from_z(0, p_cuts), 2L)
  expect_equal(.selection_step_bin_from_z(stats::qnorm(.025, lower.tail = FALSE), p_cuts), 1L)
})

test_that("selection p-value cut breaks include endpoints and are sorted", {

  expect_error(
    .selection_p_bin(c(.01, .20), c(.025, .05, 1)),
    regexp = "endpoints 0 and 1"
  )
  expect_error(
    .selection_p_bin(c(.01, .20), c(0, .05, .025, 1)),
    regexp = "strictly increasing"
  )
})

test_that("selection omega extraction orders indexed posterior columns numerically", {

  posterior_samples <- matrix(
    seq_len(30),
    nrow = 3,
    dimnames = list(NULL, c("omega[10]", "mu", "omega[2]", "omega[1]", "omega[3]",
                            "omega[4]", "omega[5]", "omega[6]", "omega[7]", "omega[8]"))
  )
  posterior_samples <- cbind(posterior_samples, "omega[9]" = 31:33)

  selection_spec <- list(jags_omega = "omega", n_bins = 10L)
  omega          <- .extract_selection_omega_samples(posterior_samples, selection_spec)

  expect_equal(colnames(omega), paste0("omega[", 1:10, "]"))
  expect_equal(omega[, 1], posterior_samples[, "omega[1]"])
  expect_equal(omega[, 10], posterior_samples[, "omega[10]"])

  custom_samples            <- posterior_samples
  colnames(custom_samples) <- sub("^omega", "custom_omega", colnames(custom_samples))
  selection_spec            <- list(jags_omega = "custom_omega", n_bins = 10L)
  custom_omega              <- .extract_selection_omega_samples(custom_samples, selection_spec)

  expect_equal(colnames(custom_omega), paste0("custom_omega[", 1:10, "]"))
  expect_equal(custom_omega[, 2], custom_samples[, "custom_omega[2]"])
})

test_that("step selected-normal kernel matches an independent p-bin reference", {

  skip_if_not(.has_native_selnorm_kernel())

  yi    <- c(-.15, 0, .196, .35, .9)
  sei   <- c(.10, .12, .10, .25, .50)
  mu    <- matrix(c(
    -.05, .02, .10, .20, .25,
     .01, .04, .12, .18, .30,
     .08, .06, .14, .24, .35
  ), nrow = 3, byrow = TRUE)
  sigma <- matrix(c(
    .11, .18, .20, .30, .55,
    .15, .22, .25, .35, .60,
    .18, .25, .30, .40, .70
  ), nrow = 3, byrow = TRUE)
  omega <- matrix(c(
    1, .70, .35, .20,
    1, .20, .80, .05,
    1, .95, .40, .10
  ), nrow = 3, byrow = TRUE)
  spec <- .test_step_spec(yi, sei)

  out <- .selnorm_kernel_loglik_matrix(
    yi             = yi,
    mu_num         = mu,
    sigma_num      = sigma,
    sei            = sei,
    omega          = omega,
    selection_spec = spec
  )

  expect_equal(out, .test_step_reference(yi, mu, sigma, sei, omega, spec),
               tolerance = 1e-12)
})

test_that("unit weights reduce one- and two-sided step kernels to normal", {

  skip_if_not(.has_native_selnorm_kernel())

  yi    <- c(-.20, .00, .15, .55)
  sei   <- c(.08, .10, .17, .26)
  mu    <- matrix(c(
    -.05, .02, .10, .30,
     .04, .01, .12, .22
  ), nrow = 2, byrow = TRUE)
  sigma <- matrix(c(
    .12, .18, .28, .34,
    .16, .20, .25, .40
  ), nrow = 2, byrow = TRUE)

  for (spec in list(
    .test_step_spec(yi, sei),
    .test_two_sided_step_spec(yi, sei)
  )) {
    omega <- matrix(1, nrow = nrow(mu), ncol = spec[["n_bins"]])

    log_pdf <- .selnorm_kernel_loglik_matrix(
      yi             = yi,
      mu_num         = mu,
      sigma_num      = sigma,
      sei            = sei,
      omega          = omega,
      selection_spec = spec
    )
    cdf <- .selnorm_kernel_cdf_matrix(
      q              = yi,
      mean           = mu,
      sd             = sigma,
      sei            = sei,
      omega          = omega,
      selection_spec = spec
    )
    moments <- .selnorm_kernel_moments_matrix(
      mean           = mu,
      sd             = sigma,
      sei            = sei,
      omega          = omega,
      selection_spec = spec
    )

    expect_equal(log_pdf, stats::dnorm(
      matrix(yi, nrow = nrow(mu), ncol = length(yi), byrow = TRUE),
      mean = mu,
      sd   = sigma,
      log  = TRUE
    ), tolerance = 1e-12)
    expect_equal(cdf, stats::pnorm(
      matrix(yi, nrow = nrow(mu), ncol = length(yi), byrow = TRUE),
      mean = mu,
      sd   = sigma
    ), tolerance = 1e-12)
    expect_equal(moments[["mean"]], mu, tolerance = 1e-12)
    expect_equal(moments[["second"]], sigma^2 + mu^2, tolerance = 1e-12)
  }
})

test_that("default RoBMA step normalizer matches independent Bem2011 reference", {

  skip_if_not(.has_native_selnorm_kernel())

  object <- RoBMA(
    yi           = Bem2011[["d"]],
    sei          = Bem2011[["se"]],
    measure      = "SMD",
    only_priors  = TRUE,
    silent       = TRUE
  )
  spec <- .selection_spec(
    priors           = object[["priors"]],
    yi               = object[["data"]][["outcome"]][["yi"]],
    sei              = object[["data"]][["outcome"]][["sei"]],
    effect_direction = attr(object[["data"]], "effect_direction")
  )

  set.seed(107)
  S     <- 4L
  K     <- length(Bem2011[["d"]])
  mean  <- matrix(stats::rnorm(S * K, mean = .12, sd = .08), nrow = S)
  sd    <- matrix(stats::runif(S * K, min = .09, max = .42), nrow = S)
  omega <- matrix(stats::runif(S * spec[["n_bins"]], min = .05, max = 1.50), nrow = S)

  log_norm <- .selnorm_kernel_log_norm_matrix(
    mean           = mean,
    sd             = sd,
    sei            = Bem2011[["se"]],
    omega          = omega,
    selection_spec = spec
  )

  expect_equal(
    log_norm,
    .test_step_log_norm_reference(mean, sd, Bem2011[["se"]], omega, spec),
    tolerance = 1e-12
  )
  expect_true(all(is.finite(log_norm)))
})

test_that("trusted step normalizer handles nonmonotone weights and tail underflow", {

  skip_if_not(.has_native_selnorm_kernel())

  yi    <- Bem2011[["d"]][1:3]
  sei   <- Bem2011[["se"]][1:3]
  spec  <- .test_step_spec(yi, sei)
  mean  <- matrix(c(.07, .15, .23, -.04, .05, .17), nrow = 2, byrow = TRUE)
  sd    <- matrix(c(.12, .19, .31, .10, .21, .28), nrow = 2, byrow = TRUE)
  omega <- matrix(c(.2, 1.4, .3, 1.1, 1.3, .1, 1.2, .2), nrow = 2, byrow = TRUE)

  expect_equal(
    .selnorm_kernel_log_norm_matrix(
      mean           = mean,
      sd             = sd,
      sei            = sei,
      omega          = omega,
      selection_spec = spec
    ),
    .test_step_log_norm_reference(mean, sd, sei, omega, spec),
    tolerance = 1e-12
  )

  tail_spec              <- .test_step_spec(0, 1)
  tail_spec[["n_bins"]]  <- 3L
  tail_spec[["z_lower"]] <- c(40, 39, -Inf)
  tail_spec[["z_upper"]] <- c(Inf, 40, 39)
  tail_spec[["obs_bin"]] <- 1L
  tail_spec[["telescope_probabilities"]] <- FALSE
  tail_omega             <- matrix(c(1e300, 0, 1e-300), nrow = 1)
  tail_mean              <- matrix(0, nrow = 1, ncol = 1)
  tail_sd                <- matrix(1, nrow = 1, ncol = 1)

  expect_equal(
    .selnorm_kernel_log_norm_matrix(
      mean           = tail_mean,
      sd             = tail_sd,
      sei            = 1,
      omega          = tail_omega,
      selection_spec = tail_spec
    )[1, 1],
    .test_step_log_norm_reference(tail_mean, tail_sd, 1, tail_omega, tail_spec)[1, 1],
    tolerance = 1e-10
  )

  cancellation_spec              <- .test_step_spec(0, 1)
  cancellation_spec[["n_bins"]]  <- 3L
  cancellation_spec[["z_lower"]] <- c(1e-8, 0, -Inf)
  cancellation_spec[["z_upper"]] <- c(Inf, 1e-8, 0)
  cancellation_spec[["obs_bin"]] <- 2L
  cancellation_omega             <- matrix(c(1, 1e16, 1), nrow = 1)

  expect_equal(
    .selnorm_kernel_log_norm_matrix(
      mean           = tail_mean,
      sd             = tail_sd,
      sei            = 1,
      omega          = cancellation_omega,
      selection_spec = cancellation_spec
    )[1, 1],
    .test_step_log_norm_reference(
      tail_mean, tail_sd, 1, cancellation_omega, cancellation_spec
    )[1, 1],
    tolerance = 1e-10
  )
})

test_that("step normalizer handles normal mode and full-support unit weights", {

  skip_if_not(.has_native_selnorm_kernel())

  yi    <- Bem2011[["d"]][1:4]
  sei   <- Bem2011[["se"]][1:4]
  spec  <- .test_step_spec(yi, sei)
  mean  <- matrix(c(.05, .12, .18, .24, .03, .10, .20, .30), nrow = 2, byrow = TRUE)
  sd    <- matrix(c(.12, .16, .20, .28, .10, .18, .22, .35), nrow = 2, byrow = TRUE)
  omega <- matrix(c(1, .6, .6, .2, 1, 1.4, .7, .7), nrow = 2, byrow = TRUE)

  expect_equal(
    .selnorm_kernel_log_norm_matrix(
      mean           = mean,
      sd             = sd,
      sei            = sei,
      omega          = omega,
      selection_spec = spec,
      kernel_mode    = SELKERNEL_NORMAL
    ),
    matrix(0, nrow = nrow(mean), ncol = ncol(mean)),
    tolerance = 1e-12
  )
  expect_equal(
    .selnorm_kernel_log_norm_matrix(
      mean           = mean,
      sd             = sd,
      sei            = sei,
      omega          = matrix(1, nrow = nrow(mean), ncol = spec[["n_bins"]]),
      selection_spec = spec
    ),
    matrix(0, nrow = nrow(mean), ncol = ncol(mean)),
    tolerance = 1e-12
  )
  expect_false(all(abs(.selnorm_kernel_log_norm_matrix(
    mean           = mean,
    sd             = sd,
    sei            = sei,
    omega          = omega,
    selection_spec = spec
  )) < 1e-12))
})

test_that("step normalizer is unchanged by sentinels and adjacent repeated bins", {

  skip_if_not(.has_native_selnorm_kernel())

  yi      <- Bem2011[["d"]][1:3]
  sei     <- Bem2011[["se"]][1:3]
  spec    <- .test_step_spec(yi, sei)
  mean    <- matrix(c(.06, .12, .18, -.03, .04, .10), nrow = 2, byrow = TRUE)
  sd      <- matrix(c(.11, .18, .25, .14, .20, .30), nrow = 2, byrow = TRUE)
  omega   <- matrix(c(1, .5, .5, .2, 1, .8, .8, .3), nrow = 2, byrow = TRUE)
  sentinel_spec <- spec

  sentinel_spec[["z_lower"]] <- .selection_jags_bounds(sentinel_spec[["z_lower"]])
  sentinel_spec[["z_upper"]] <- .selection_jags_bounds(sentinel_spec[["z_upper"]])

  expect_equal(
    .selnorm_kernel_log_norm_matrix(
      mean           = mean,
      sd             = sd,
      sei            = sei,
      omega          = omega,
      selection_spec = sentinel_spec
    ),
    .selnorm_kernel_log_norm_matrix(
      mean           = mean,
      sd             = sd,
      sei            = sei,
      omega          = omega,
      selection_spec = spec
    ),
    tolerance = 1e-12
  )

  collapsed_spec              <- spec
  collapsed_spec[["n_bins"]]  <- 3L
  collapsed_spec[["z_lower"]] <- c(spec[["z_lower"]][1L], spec[["z_lower"]][3L], spec[["z_lower"]][4L])
  collapsed_spec[["z_upper"]] <- c(spec[["z_upper"]][1L], spec[["z_upper"]][2L], spec[["z_upper"]][4L])
  collapsed_omega             <- omega[, c(1L, 2L, 4L), drop = FALSE]

  expect_equal(
    .selnorm_kernel_log_norm_matrix(
      mean           = mean,
      sd             = sd,
      sei            = sei,
      omega          = collapsed_omega,
      selection_spec = collapsed_spec
    ),
    .selnorm_kernel_log_norm_matrix(
      mean           = mean,
      sd             = sd,
      sei            = sei,
      omega          = omega,
      selection_spec = spec
    ),
    tolerance = 1e-12
  )
})

test_that("step normalizer sign handling is symmetric for negative effects", {

  skip_if_not(.has_native_selnorm_kernel())

  yi_pos <- Bem2011[["d"]][1:4]
  yi_neg <- -yi_pos
  sei    <- Bem2011[["se"]][1:4]
  mean   <- matrix(c(.04, .10, .16, .22, .02, .08, .14, .20), nrow = 2, byrow = TRUE)
  sd     <- matrix(c(.10, .14, .18, .22, .12, .16, .20, .24), nrow = 2, byrow = TRUE)
  omega  <- matrix(c(1, .7, .4, .2, 1, .9, .5, .3), nrow = 2, byrow = TRUE)

  spec_pos <- .test_step_spec(yi_pos, sei, effect_direction = "positive")
  spec_neg <- .test_step_spec(yi_neg, sei, effect_direction = "negative")

  expect_equal(
    .selnorm_kernel_log_norm_matrix(
      mean           = -mean,
      sd             = sd,
      sei            = sei,
      omega          = omega,
      selection_spec = spec_neg
    ),
    .selnorm_kernel_log_norm_matrix(
      mean           = mean,
      sd             = sd,
      sei            = sei,
      omega          = omega,
      selection_spec = spec_pos
    ),
    tolerance = 1e-12
  )
})

test_that("trusted step CDF, moments, and RNG match exact fallback paths", {

  skip_if_not(.has_native_selnorm_kernel())

  yi    <- c(.04, .18, .31, .47)
  sei   <- c(.08, .12, .17, .24)
  mean  <- matrix(c(
    .02, .10, .18, .26,
    .05, .08, .20, .33,
    .01, .14, .22, .29
  ), nrow = 3, byrow = TRUE)
  sd    <- matrix(c(
    .11, .19, .25, .31,
    .13, .22, .28, .36,
    .16, .20, .30, .42
  ), nrow = 3, byrow = TRUE)
  omega <- matrix(c(
    1, .35, 1.60, .20,
    .80, .80, .25, 1.10,
    1.40, .10, .95, .30
  ), nrow = 3, byrow = TRUE)

  for (effect_direction in c("positive", "negative")) {
    sign <- if (effect_direction == "positive") 1 else -1
    spec <- .test_step_spec(sign * yi, sei, effect_direction = effect_direction)
    spec_fallback <- spec
    spec_fallback[["telescope_probabilities"]] <- FALSE

    cdf_lower <- .selnorm_kernel_cdf_matrix(
      q              = sign * yi,
      mean           = sign * mean,
      sd             = sd,
      sei            = sei,
      omega          = omega,
      selection_spec = spec
    )
    cdf_upper <- .selnorm_kernel_cdf_matrix(
      q              = sign * yi,
      mean           = sign * mean,
      sd             = sd,
      sei            = sei,
      omega          = omega,
      selection_spec = spec,
      lower.tail     = FALSE
    )
    moments <- .selnorm_kernel_moments_matrix(
      mean           = sign * mean,
      sd             = sd,
      sei            = sei,
      omega          = omega,
      selection_spec = spec
    )

    expect_equal(
      cdf_lower,
      .test_step_cdf_reference(sign * yi, sign * mean, sd, sei, omega, spec),
      tolerance = 1e-12
    )
    expect_equal(
      cdf_upper,
      .test_step_cdf_reference(
        sign * yi, sign * mean, sd, sei, omega, spec, lower.tail = FALSE
      ),
      tolerance = 1e-12
    )
    expect_equal(
      moments,
      .test_step_moments_reference(sign * mean, sd, sei, omega, spec),
      tolerance = 1e-12
    )

    set.seed(511)
    rng_fast <- .selnorm_kernel_rng_matrix(
      mean           = sign * mean,
      sd             = sd,
      sei            = sei,
      omega          = omega,
      selection_spec = spec
    )
    set.seed(511)
    rng_fallback <- .selnorm_kernel_rng_matrix(
      mean           = sign * mean,
      sd             = sd,
      sei            = sei,
      omega          = omega,
      selection_spec = spec_fallback
    )
    expect_equal(rng_fast, rng_fallback, tolerance = 1e-12)
  }
})

test_that("step normalizer and CDF use stable tail interval probabilities", {

  skip_if_not(.has_native_selnorm_kernel())

  sei  <- 1
  yi   <- 9.5
  spec <- .test_step_spec(yi, sei)
  spec[["z_lower"]] <- c(9, 10, -Inf)
  spec[["z_upper"]] <- c(10, Inf, 9)
  spec[["obs_bin"]] <- 1L
  spec[["n_bins"]]  <- 3L

  omega <- matrix(c(1, 1e-300, 1e-300), nrow = 1)
  mean  <- matrix(0, nrow = 1, ncol = 1)
  sd    <- matrix(1, nrow = 1, ncol = 1)

  log_mass <- log(omega[1, ]) + vapply(
    seq_len(spec[["n_bins"]]),
    function(b) .test_interval_log_prob(
      spec[["z_lower"]][b],
      spec[["z_upper"]][b],
      0,
      1
    ),
    numeric(1)
  )
  expected_log_norm <- .test_logsumexp(log_mass)

  expected_log_cdf <- .test_logsumexp(log(omega[1, ]) + c(
    .test_interval_log_prob(9, yi, 0, 1),
    .test_interval_log_prob(10, yi, 0, 1),
    .test_interval_log_prob(-Inf, 9, 0, 1)
  ))
  expected_cdf <- exp(expected_log_cdf - expected_log_norm)

  log_norm <- .selnorm_kernel_log_norm_matrix(
    mean           = mean,
    sd             = sd,
    sei            = sei,
    omega          = omega,
    selection_spec = spec
  )
  log_pdf <- .selnorm_kernel_loglik_matrix(
    yi             = yi,
    mu_num         = mean,
    sigma_num      = sd,
    sei            = sei,
    omega          = omega,
    selection_spec = spec
  )
  cdf <- .selnorm_kernel_cdf_matrix(
    q              = yi,
    mean           = mean,
    sd             = sd,
    sei            = sei,
    omega          = omega,
    selection_spec = spec
  )

  expect_true(is.finite(log_norm[1, 1]))
  expect_equal(log_norm[1, 1], expected_log_norm, tolerance = 1e-10)
  expect_equal(
    log_pdf[1, 1],
    stats::dnorm(yi, mean = 0, sd = 1, log = TRUE) - expected_log_norm,
    tolerance = 1e-10
  )
  expect_equal(cdf[1, 1], expected_cdf, tolerance = 1e-10)
})

test_that("negative effect direction uses the same signed kernel", {

  skip_if_not(.has_native_selnorm_kernel())

  yi_pos <- c(.08, .20, .45)
  yi_neg <- -yi_pos
  sei    <- c(.10, .16, .22)
  mu_pos <- matrix(c(.04, .12, .30, .02, .10, .25), nrow = 2, byrow = TRUE)
  sigma  <- matrix(c(.16, .22, .28, .18, .24, .30), nrow = 2, byrow = TRUE)
  omega  <- matrix(c(1, .6, .3, .2, 1, .8, .4, .1), nrow = 2, byrow = TRUE)

  spec_pos <- .test_step_spec(yi_pos, sei, effect_direction = "positive")
  spec_neg <- .test_step_spec(yi_neg, sei, effect_direction = "negative")

  out_pos <- .selnorm_kernel_loglik_matrix(
    yi_pos, mu_pos, sigma, sei = sei, omega = omega,
    selection_spec = spec_pos
  )
  out_neg <- .selnorm_kernel_loglik_matrix(
    yi_neg, -mu_pos, sigma, sei = sei, omega = omega,
    selection_spec = spec_neg
  )

  expect_equal(out_neg, out_pos, tolerance = 1e-12)
})

test_that("inactive kernel rows reduce to ordinary normal likelihood", {

  skip_if_not(.has_native_selnorm_kernel())

  yi     <- c(.10, .40)
  sei    <- c(.10, .20)
  mu     <- matrix(c(.02, .20, .05, .22), nrow = 2, byrow = TRUE)
  sigma  <- matrix(c(.15, .30, .18, .35), nrow = 2, byrow = TRUE)
  omega  <- matrix(c(.01, .02, .03, .04, 1, .6, .3, .1), nrow = 2, byrow = TRUE)
  weights <- c(1.5, .25)
  spec   <- .test_step_spec(yi, sei)

  out <- .selnorm_kernel_loglik_matrix(
    yi             = yi,
    mu_num         = mu,
    sigma_num      = sigma,
    sei            = sei,
    omega          = omega,
    selection_spec = spec,
    kernel_mode    = c(SELKERNEL_NORMAL, SELKERNEL_STEP),
    weights        = weights
  )
  ref <- .test_step_reference(yi, mu, sigma, sei, omega, spec, weights = weights)
  ref[1, ] <- weights * stats::dnorm(yi, mean = mu[1, ], sd = sigma[1, ], log = TRUE)

  expect_equal(out, ref, tolerance = 1e-12)
})

test_that("step selected-normal RNG follows exact bin masses and moments", {

  skip_if_not(.has_native_selnorm_kernel())

  set.seed(101)
  yi    <- 0
  sei   <- .11
  spec  <- .test_step_spec(yi, sei)
  S     <- 30000L
  mu    <- matrix(.05, nrow = S, ncol = 1)
  sigma <- matrix(.23, nrow = S, ncol = 1)
  omega <- matrix(
    rep(c(1, .15, 2.5, .05), each = S),
    nrow = S,
    ncol = spec[["n_bins"]]
  )

  context <- spec
  context[["omega"]]       <- omega
  context[["alpha"]]       <- rep(0, S)
  context[["phack_kind"]]  <- rep(0L, S)
  context[["kernel_mode"]] <- rep(SELKERNEL_STEP, S)

  out  <- .selection_step_rng_matrix(mu, sigma, sei, context)
  bins <- .selection_step_bin_from_z(out[, 1] / sei, spec[["p_cuts"]])

  mass <- numeric(spec[["n_bins"]])
  for (b in seq_len(spec[["n_bins"]])) {
    mass[b] <- omega[1, b] * .test_interval_prob(
      spec[["z_lower"]][b],
      spec[["z_upper"]][b],
      spec[["sign"]] * mu[1, 1] / sei,
      sigma[1, 1] / sei
    )
  }
  expected_prob <- mass / sum(mass)
  observed_prob <- as.numeric(prop.table(table(
    factor(bins, levels = seq_len(spec[["n_bins"]]))
  )))

  moments <- .test_step_moments_reference(
    mu    = matrix(mu[1, 1], nrow = 1),
    sigma = matrix(sigma[1, 1], nrow = 1),
    sei   = sei,
    omega = matrix(omega[1, ], nrow = 1),
    spec  = spec
  )

  expect_equal(observed_prob, expected_prob, tolerance = .01)
  expect_equal(mean(out[, 1]), moments[["mean"]][1, 1], tolerance = .006)
  expect_equal(mean(out[, 1]^2), moments[["second"]][1, 1], tolerance = .006)
})

test_that("step selected-normal RNG handles signed extreme tail truncation", {

  skip_if_not(.has_native_selnorm_kernel())

  set.seed(102)
  sei   <- .10
  spec  <- .test_step_spec(-.10, sei, effect_direction = "negative")
  S     <- 128L
  omega <- matrix(
    rep(c(1, 1e-12, 1e-12, 1e-12), each = S),
    nrow = S,
    ncol = spec[["n_bins"]]
  )

  context <- spec
  context[["omega"]]       <- omega
  context[["alpha"]]       <- rep(0, S)
  context[["phack_kind"]]  <- rep(0L, S)
  context[["kernel_mode"]] <- rep(SELKERNEL_STEP, S)

  out <- .selection_step_rng_matrix(
    mean              = matrix(-.35, nrow = S, ncol = 1),
    sd                = matrix(.05, nrow = S, ncol = 1),
    sei               = sei,
    selection_context = context
  )
  signed_z <- spec[["sign"]] * out[, 1] / sei

  expect_true(all(is.finite(out)))
  expect_true(all(signed_z >= spec[["z_lower"]][1] - 1e-10))
  expect_true(all(out[, 1] < 0))
})

test_that("selected-normal native boundary handles structural zero omega", {

  skip_if_not(.has_native_selnorm_kernel())

  yi     <- c(.10, .20)
  sei    <- c(.10, .10)
  spec   <- .test_step_spec(yi, sei)
  mu     <- matrix(0, nrow = 1, ncol = 2)
  sigma  <- matrix(.20, nrow = 1, ncol = 2)
  omega_empty_bin <- matrix(c(1, 0, .5, .25), nrow = 1)
  omega_obs_bin   <- matrix(c(1, .5, 0, .25), nrow = 1)

  empty_bin_log_lik <- .selnorm_kernel_loglik_matrix(
    yi             = yi,
    mu_num         = mu,
    sigma_num      = sigma,
    sei            = sei,
    omega          = omega_empty_bin,
    selection_spec = spec
  )
  obs_bin_log_lik <- .selnorm_kernel_loglik_matrix(
    yi             = yi,
    mu_num         = mu,
    sigma_num      = sigma,
    sei            = sei,
    omega          = omega_obs_bin,
    selection_spec = spec
  )

  expect_true(all(is.finite(empty_bin_log_lik)))
  expect_equal(obs_bin_log_lik[1, 1], -Inf)
  expect_true(is.finite(obs_bin_log_lik[1, 2]))
})

test_that("active p-hacking kernels error on unsupported non-likelihood paths", {

  skip_if_not(.has_native_selnorm_kernel())

  yi    <- .10
  sei   <- .10
  spec  <- .test_step_spec(yi, sei)
  mean  <- matrix(c(0, .02), nrow = 2, ncol = 1)
  sd    <- matrix(.20, nrow = 2, ncol = 1)
  omega <- matrix(1, nrow = 2, ncol = spec[["n_bins"]])
  alpha <- c(.50, 0)
  phack <- c(1L, 0L)
  mode  <- c(SELKERNEL_STEP_PHACK_POWER, SELKERNEL_STEP)

  log_lik <- .selnorm_kernel_loglik_matrix(
    yi             = yi,
    mu_num         = mean,
    sigma_num      = sd,
    sei            = sei,
    omega          = omega,
    selection_spec = spec,
    alpha          = alpha,
    phack_kind     = phack,
    kernel_mode    = mode
  )
  expect_true(all(is.finite(log_lik)))

  expect_error(
    .selnorm_kernel_cdf_matrix(
      q              = yi,
      mean           = mean,
      sd             = sd,
      sei            = sei,
      omega          = omega,
      selection_spec = spec,
      alpha          = alpha,
      phack_kind     = phack,
      kernel_mode    = mode
    ),
    "active p-hacking"
  )
  expect_error(
    .selnorm_kernel_moments_matrix(
      mean           = mean,
      sd             = sd,
      sei            = sei,
      omega          = omega,
      selection_spec = spec,
      alpha          = alpha,
      phack_kind     = phack,
      kernel_mode    = mode
    ),
    "active p-hacking"
  )
  expect_error(
    .selnorm_kernel_rng_matrix(
      mean           = mean,
      sd             = sd,
      sei            = sei,
      omega          = omega,
      selection_spec = spec,
      alpha          = alpha,
      phack_kind     = phack,
      kernel_mode    = mode
    ),
    "active p-hacking"
  )
  expect_error(
    .selnorm_kernel_weighted_summary(
      yi             = yi,
      mean           = mean,
      sd             = sd,
      sei            = sei,
      psis_weights   = matrix(c(.5, .5), nrow = 2, ncol = 1),
      omega          = omega,
      selection_spec = spec,
      alpha          = alpha,
      phack_kind     = phack,
      kernel_mode    = mode
    ),
    "active p-hacking"
  )

  context <- spec
  context[["omega"]]          <- omega
  context[["alpha"]]          <- alpha
  context[["phack_kind"]]     <- phack
  context[["kernel_mode"]]    <- mode
  context[["bias_indicator"]] <- c(1L, 1L)
  context[["use_normal"]]     <- c(FALSE, FALSE)

  setup <- list(
    mu        = c(0, .02),
    selection = context
  )

  expect_error(
    .funnel_selected_cdf(
      q                = yi,
      rows             = 1:2,
      se               = 0,
      total_sd         = c(.20, .20),
      setup            = setup,
      effect_direction = "positive"
    ),
    "active p-hacking"
  )
})

test_that("funnel selected-normal CDF is finite at zero standard error", {

  skip_if_not(.has_native_selnorm_kernel())

  spec <- .test_step_spec(yi = c(.10, .25), sei = c(.10, .15))

  selection_context <- spec
  selection_context[["omega"]]          <- matrix(c(1, .70, .35, .20), nrow = 1)
  selection_context[["alpha"]]          <- 0
  selection_context[["phack_kind"]]     <- 0L
  selection_context[["kernel_mode"]]    <- SELKERNEL_STEP
  selection_context[["bias_indicator"]] <- 1L
  selection_context[["use_normal"]]     <- FALSE

  setup <- list(
    mu                = 0,
    tau               = .20,
    PET               = 0,
    PEESE             = 0,
    is_weightfunction = TRUE,
    selection         = selection_context
  )

  out <- .funnel_model_averaged_cdf(
    q                = 0,
    se               = 0,
    setup            = setup,
    effect_direction = "positive"
  )

  expected <- (.20 * stats::pnorm(0, mean = 0, sd = .20)) /
    (stats::pnorm(0, mean = 0, sd = .20, lower.tail = FALSE) +
      .20 * stats::pnorm(0, mean = 0, sd = .20))

  expect_equal(out, expected, tolerance = 1e-12)
})

test_that("funnel and regplot selected CDF paths match non-telescoped fallback", {

  skip_if_not(.has_native_selnorm_kernel())
  skip_if_not(.has_native_regplot_selection_mixture())

  set.seed(109)
  S     <- 11L
  K     <- 3L
  yi    <- c(.08, .20, .42)
  sei   <- c(.09, .14, .22)
  spec  <- .test_step_spec(yi, sei)
  omega <- matrix(runif(S * spec[["n_bins"]], .10, 1.80), nrow = S)

  selection <- spec
  selection[["omega"]]       <- omega
  selection[["alpha"]]       <- rep(0, S)
  selection[["phack_kind"]]  <- rep(0L, S)
  selection[["kernel_mode"]] <- rep(SELKERNEL_STEP, S)
  selection[["use_normal"]]  <- rep(FALSE, S)
  selection[["has_phack"]]   <- FALSE

  selection_fallback <- selection
  selection_fallback[["telescope_probabilities"]] <- FALSE

  setup <- list(
    mu                = seq(-.12, .20, length.out = S),
    tau               = seq(.03, .18, length.out = S),
    PET               = rep(0, S),
    PEESE             = rep(0, S),
    is_weightfunction = rep(TRUE, S),
    selection         = selection
  )
  setup_fallback <- setup
  setup_fallback[["selection"]] <- selection_fallback

  expect_equal(
    .funnel_model_averaged_cdf(.14, .13, setup, "positive"),
    .funnel_model_averaged_cdf(.14, .13, setup_fallback, "positive"),
    tolerance = 1e-12
  )

  mean_samples <- matrix(rnorm(S * K, .05, .18), nrow = S)
  sd_samples   <- matrix(runif(S * K, .08, .35), nrow = S)

  expect_equal(
    .regplot_selnorm_mixture_interval_quantiles(
      mean_samples      = mean_samples,
      sd_samples        = sd_samples,
      se                = .13,
      probs             = c(.10, .90),
      selection_context = selection
    ),
    .regplot_selnorm_mixture_interval_quantiles(
      mean_samples      = mean_samples,
      sd_samples        = sd_samples,
      se                = .13,
      probs             = c(.10, .90),
      selection_context = selection_fallback
    ),
    tolerance = 1e-8
  )
})

test_that("native funnel model-averaged quantiles match R fallback", {

  skip_if_not(.has_native_selnorm_kernel())
  skip_if_not(.has_native_funnel_model_averaged_quantiles(list(
    selection = list()
  )))

  set.seed(127)
  S     <- 19L
  yi    <- c(.08, .20, .42)
  sei   <- c(.09, .14, .22)
  omega <- matrix(runif(S * 4L, .10, 1.80), nrow = S)

  for (effect_direction in c("positive", "negative")) {
    sign <- if (effect_direction == "positive") 1 else -1
    spec <- .test_step_spec(sign * yi, sei, effect_direction = effect_direction)

    for (telescope_probabilities in c(TRUE, FALSE)) {
      selection <- spec
      selection[["omega"]] <- omega[
        , seq_len(spec[["n_bins"]]), drop = FALSE
      ]
      selection[["alpha"]]       <- rep(0, S)
      selection[["phack_kind"]]  <- rep(0L, S)
      selection[["kernel_mode"]] <- rep(SELKERNEL_STEP, S)
      selection[["kernel_mode"]][seq(3, S, by = 5)] <- SELKERNEL_NORMAL
      selection[["use_normal"]] <- selection[["kernel_mode"]] == SELKERNEL_NORMAL
      selection[["has_phack"]]  <- FALSE
      selection[["telescope_probabilities"]] <- telescope_probabilities

      setup <- list(
        mu                = sign * seq(-.12, .20, length.out = S),
        tau               = seq(.00, .18, length.out = S),
        PET               = rnorm(S, 0, .03),
        PEESE             = rnorm(S, 0, .02),
        is_weightfunction = !selection[["use_normal"]],
        selection         = selection
      )
      se_sequence <- c(0, .05, .13, .27)

      native <- .funnel_model_averaged_quantiles_native(
        se_sequence      = se_sequence,
        setup            = setup,
        effect_direction = effect_direction
      )
      ref <- list(
        lower = vapply(
          se_sequence,
          .funnel_model_averaged_quantile,
          numeric(1),
          p                = .025,
          setup            = setup,
          effect_direction = effect_direction
        ),
        upper = vapply(
          se_sequence,
          .funnel_model_averaged_quantile,
          numeric(1),
          p                = .975,
          setup            = setup,
          effect_direction = effect_direction
        ),
        mid = vapply(
          se_sequence,
          .funnel_model_averaged_quantile,
          numeric(1),
          p                = .5,
          setup            = setup,
          effect_direction = effect_direction
        )
      )

      expect_equal(native[["lower"]], ref[["lower"]], tolerance = 1e-5)
      expect_equal(native[["upper"]], ref[["upper"]], tolerance = 1e-5)
      expect_equal(native[["mid"]],   ref[["mid"]],   tolerance = 1e-5)
    }
  }
})

test_that("native funnel model-averaged quantiles reject active p-hacking", {

  skip_if_not(.has_native_selnorm_kernel())
  skip_if_not(.has_native_funnel_model_averaged_quantiles(list(
    selection = list()
  )))

  S    <- 3L
  spec <- .test_step_spec(c(.08, .20), c(.09, .14))

  selection <- spec
  selection[["omega"]]       <- matrix(
    runif(S * spec[["n_bins"]], .10, 1.80),
    nrow = S
  )
  selection[["alpha"]]       <- rep(.20, S)
  selection[["phack_kind"]]  <- rep(1L, S)
  selection[["kernel_mode"]] <- rep(SELKERNEL_STEP_PHACK_POWER, S)
  selection[["use_normal"]] <- rep(FALSE, S)
  selection[["has_phack"]]  <- TRUE

  setup <- list(
    mu                = seq(-.12, .20, length.out = S),
    tau               = seq(.00, .18, length.out = S),
    PET               = rep(0, S),
    PEESE             = rep(0, S),
    is_weightfunction = rep(TRUE, S),
    selection         = selection
  )

  expect_error(
    .funnel_model_averaged_quantiles_native(
      se_sequence      = c(.05, .13),
      setup            = setup,
      effect_direction = "positive"
    ),
    "active p-hacking"
  )
})

test_that("funnel max_samples helpers validate and stratify rows", {

  bias_indicator <- c(rep(1L, 5000), rep(2L, 3000), rep(3L, 2000))

  expect_equal(.normalize_max_samples(10), 10L)
  expect_equal(.normalize_max_samples(Inf), Inf)
  expect_error(.normalize_max_samples(9), "at least 10")
  expect_error(.normalize_max_samples(10.5), "positive integer or Inf")
  expect_equal(.normalize_funnel_max_samples(10), 10L)
  expect_equal(.normalize_funnel_max_samples(Inf), Inf)
  expect_error(.normalize_funnel_max_samples(9), "at least 10")
  expect_error(.normalize_funnel_max_samples(10.5), "positive integer or Inf")

  plain_rows <- .thin_sample_rows(10000, 1000)
  expect_equal(plain_rows, round(seq(from = 1, to = 10000, length.out = 1000)))
  expect_true(max(diff(plain_rows)) - min(diff(plain_rows)) <= 1)
  expect_null(.thin_sample_rows(1000, 1000))
  expect_null(.thin_sample_rows(100, Inf))

  rows <- .funnel_subsample_rows(
    bias_indicator = bias_indicator,
    max_samples    = 1000
  )

  expect_equal(rows, .funnel_subsample_rows(bias_indicator, 1000))
  expect_equal(length(rows), 1000)
  expect_equal(as.integer(table(bias_indicator[rows])), c(500L, 300L, 200L))
  expect_equal(rows, .thin_sample_rows_by_group(bias_indicator, 1000))
  expect_null(.funnel_subsample_rows(bias_indicator, Inf))

  rare_group <- c(1L, rep(2L, 9999))
  rare_rows  <- .thin_sample_rows_by_group(rare_group, 1000)
  expect_true(1L %in% rare_group[rare_rows])
})

test_that("selection-only prior_bias wrapper is delegated to BayesTools priors", {

  selection <- BayesTools::prior_weightfunction(
    "one-sided", c(.025), BayesTools::wf_fixed(c(1, .5))
  )

  object <- bselmodel(
    yi                        = c(.1, .2),
    sei                       = c(.1, .1),
    measure                   = "SMD",
    prior_bias                = BayesTools::prior_bias(selection = selection),
    prior_unit_information_sd = 1,
    only_priors               = TRUE,
    silent                    = TRUE
  )
  priors <- .create_fit_priors(object[["data"]], object[["priors"]])

  expect_true("bias" %in% names(priors))
  expect_true(BayesTools::is_prior_bias(priors[["bias"]]))
  expect_equal(BayesTools::JAGS_to_monitor(priors), c("mu", "tau", "omega"))
})

test_that("selection-only prior_bias wrapper is visible to zplot thresholds", {

  selection <- BayesTools::prior_weightfunction(
    "one-sided", c(.025), BayesTools::wf_fixed(c(1, .5))
  )

  object <- bselmodel(
    yi                        = c(.1, .2),
    sei                       = c(.1, .1),
    measure                   = "SMD",
    prior_bias                = BayesTools::prior_bias(selection = selection),
    prior_unit_information_sd = 1,
    only_priors               = TRUE,
    silent                    = TRUE
  )

  expect_equal(.zplot_threshold(object[["priors"]]), stats::qnorm(1 - .025))
})

test_that("native zplot density matches selected-normal R reference", {

  skip_if_not(.has_native_zplot_density(selection = TRUE))

  S        <- 4L
  K        <- 3L
  sei      <- c(.07, .13, .29)
  mu       <- matrix(c(
    -.20, .05, .30,
     .10, .12, .18,
     .40, .00, -.15,
     .05, -.25, .20
  ), nrow = S, byrow = TRUE)
  tau      <- matrix(c(
    .05, .07, .10,
    .08, .06, .11,
    .03, .09, .12,
    .04, .10, .15
  ), nrow = S, byrow = TRUE)
  z_values <- c(
    -3,
    0,
    stats::qnorm(.05, lower.tail = FALSE),
    stats::qnorm(.025, lower.tail = FALSE),
    3
  )
  spec     <- .test_step_spec(yi = c(.10, .25, .40), sei = sei)

  selection <- spec
  selection[["omega"]]       <- matrix(c(
    1, .7, .2, .1,
    .2, .4, .8, 1.5,
    1.5, .2, 2.0, .05,
    1, 1, 1, 1
  ), nrow = S, byrow = TRUE)
  selection[["alpha"]]       <- rep(0, S)
  selection[["phack_kind"]]  <- rep(0L, S)
  selection[["kernel_mode"]] <- c(
    SELKERNEL_STEP,
    SELKERNEL_NORMAL,
    SELKERNEL_STEP,
    SELKERNEL_NORMAL
  )
  selection[["use_normal"]] <- selection[["kernel_mode"]] == SELKERNEL_NORMAL

  reference_density <- function(extrapolate) {

    sei_mat     <- matrix(sei, nrow = S, ncol = K, byrow = TRUE)
    total_sd    <- sqrt(tau^2 + sei_mat^2)
    densities   <- matrix(NA_real_, nrow = S, ncol = length(z_values))
    const_rows  <- if (extrapolate) which(!selection[["use_normal"]]) else integer(0)
    weight_rows <- if (extrapolate) integer(0) else which(!selection[["use_normal"]])

    log_norm <- NULL
    if (length(const_rows) > 0L) {
      log_norm <- .selection_step_log_norm_matrix(
        mean              = mu[const_rows, , drop = FALSE],
        sd                = total_sd[const_rows, , drop = FALSE],
        sei               = sei,
        selection_context = .selection_context_subset_rows(selection, const_rows)
      )
    }

    for (ell in seq_along(z_values)) {
      y_seq <- z_values[ell] * sei
      dens  <- stats::dnorm(
        matrix(y_seq, nrow = S, ncol = K, byrow = TRUE),
        mean = mu,
        sd   = total_sd
      ) * sei_mat

      if (length(const_rows) > 0L) {
        dens[const_rows, ] <- dens[const_rows, , drop = FALSE] / exp(log_norm)
      }
      if (length(weight_rows) > 0L) {
        selection_weight <- .selection_context_subset_rows(selection, weight_rows)
        log_pdf <- .selection_step_logpdf_matrix(
          y                 = y_seq,
          mean              = mu[weight_rows, , drop = FALSE],
          sd                = total_sd[weight_rows, , drop = FALSE],
          sei               = sei,
          selection_context = selection_weight
        )
        dens[weight_rows, ] <- exp(log_pdf) * sei_mat[weight_rows, , drop = FALSE]
      }

      densities[, ell] <- rowMeans(dens)
    }

    return(densities)
  }

  for (extrapolate in c(FALSE, TRUE)) {
    expect_equal(
      .zplot_density_vectorized(
        z_sequence       = z_values,
        mu_samples       = mu,
        tau_within       = tau,
        sei              = sei,
        selection        = selection,
        extrapolate      = extrapolate,
        effect_direction = "positive"
      ),
      reference_density(extrapolate),
      tolerance = 1e-12
    )
  }
})


test_that("zplot threshold separates fitted, extrapolated, and missing-study targets", {

  skip_if_not(.has_native_selnorm_kernel())

  S         <- 4L
  K         <- 3L
  sei       <- c(.07, .13, .29)
  mu        <- matrix(c(
    -.20, .05, .30,
     .10, .12, .18,
     .40, .00, -.15,
     .05, -.25, .20
  ), nrow = S, byrow = TRUE)
  tau       <- matrix(c(
    .05, .07, .10,
    .08, .06, .11,
    .03, .09, .12,
    .04, .10, .15
  ), nrow = S, byrow = TRUE)
  threshold <- stats::qnorm(.025, lower.tail = FALSE)
  spec      <- .test_step_spec(yi = c(.10, .25, .40), sei = sei)

  selection <- spec
  selection[["omega"]]       <- matrix(c(
    1, .7, .2, .1,
    .2, .4, .8, 1.5,
    1.5, .2, 2.0, .05,
    1, 1, 1, 1
  ), nrow = S, byrow = TRUE)
  selection[["alpha"]]       <- rep(0, S)
  selection[["phack_kind"]]  <- rep(0L, S)
  selection[["kernel_mode"]] <- c(
    SELKERNEL_STEP,
    SELKERNEL_NORMAL,
    SELKERNEL_STEP,
    SELKERNEL_NORMAL
  )
  selection[["use_normal"]] <- selection[["kernel_mode"]] == SELKERNEL_NORMAL

  sei_mat  <- matrix(sei, nrow = S, ncol = K, byrow = TRUE)
  total_sd <- sqrt(tau^2 + sei_mat^2)
  q_upper  <- threshold * sei
  q_lower  <- -threshold * sei

  unselected_thresholds <- stats::pnorm(
    matrix(q_upper, nrow = S, ncol = K, byrow = TRUE),
    mean       = mu,
    sd         = total_sd,
    lower.tail = FALSE
  ) + stats::pnorm(
    matrix(q_lower, nrow = S, ncol = K, byrow = TRUE),
    mean       = mu,
    sd         = total_sd,
    lower.tail = TRUE
  )

  step_rows      <- which(!selection[["use_normal"]])
  selection_step <- .selection_context_subset_rows(selection, step_rows)
  log_norm       <- .selection_step_log_norm_matrix(
    mean              = mu[step_rows, , drop = FALSE],
    sd                = total_sd[step_rows, , drop = FALSE],
    sei               = sei,
    selection_context = selection_step
  )
  expected_weights <- matrix(1, nrow = S, ncol = K)
  expected_weights[step_rows, ] <- exp(-log_norm)

  selected_thresholds <- unselected_thresholds
  selected_thresholds[step_rows, ] <- .selection_step_cdf_matrix(
    q                 = q_upper,
    mean              = mu[step_rows, , drop = FALSE],
    sd                = total_sd[step_rows, , drop = FALSE],
    sei               = sei,
    selection_context = selection_step,
    lower.tail        = FALSE
  ) + .selection_step_cdf_matrix(
    q                 = q_lower,
    mean              = mu[step_rows, , drop = FALSE],
    sd                = total_sd[step_rows, , drop = FALSE],
    sei               = sei,
    selection_context = selection_step,
    lower.tail        = TRUE
  )

  extrapolated <- .zplot_threshold_vectorized(
    z_threshold      = threshold,
    mu_samples       = mu,
    tau_within       = tau,
    sei              = sei,
    selection        = selection,
    extrapolate      = TRUE,
    effect_direction = "positive"
  )
  fitted <- .zplot_threshold_vectorized(
    z_threshold      = threshold,
    mu_samples       = mu,
    tau_within       = tau,
    sei              = sei,
    selection        = selection,
    extrapolate      = FALSE,
    effect_direction = "positive"
  )

  expect_equal(
    extrapolated[["EDR"]],
    rowSums(unselected_thresholds * expected_weights) / rowSums(expected_weights),
    tolerance = 1e-12
  )
  expect_equal(extrapolated[["weights"]], rowMeans(expected_weights), tolerance = 1e-12)
  expect_equal(fitted[["EDR"]], rowMeans(selected_thresholds), tolerance = 1e-12)
  expect_equal(fitted[["weights"]], rowMeans(expected_weights), tolerance = 1e-12)
})


test_that("native regplot selection intervals match R reference", {

  skip_if_not(.has_native_regplot_selection_mixture())

  set.seed(131)
  S            <- 15L
  K            <- 4L
  se           <- .14
  probs        <- c(.025, .975)
  mean_samples <- matrix(rnorm(S * K, 0, .35), nrow = S, ncol = K)
  sd_samples   <- matrix(runif(S * K, .02, .30), nrow = S, ncol = K)
  sd_samples[c(3, 11)] <- sqrt(.Machine$double.eps) / 10
  spec         <- .test_step_spec(yi = seq_len(K) / 10, sei = rep(se, K))

  selection <- spec
  selection[["omega"]]       <- matrix(
    runif(S * spec[["n_bins"]], .02, 3),
    nrow = S,
    ncol = spec[["n_bins"]]
  )
  selection[["alpha"]]       <- rep(0, S)
  selection[["phack_kind"]]  <- rep(0L, S)
  selection[["kernel_mode"]] <- rep(SELKERNEL_STEP, S)
  selection[["kernel_mode"]][seq(2, S, by = 4)] <- SELKERNEL_NORMAL
  selection[["use_normal"]] <- selection[["kernel_mode"]] == SELKERNEL_NORMAL

  setup <- list(
    is_weightfunction = !selection[["use_normal"]],
    selection         = selection
  )

  native <- .regplot_selnorm_mixture_interval_quantiles(
    mean_samples      = mean_samples,
    sd_samples        = sd_samples,
    se                = se,
    probs             = probs,
    selection_context = selection
  )
  ref <- .regplot_selection_mixture_interval_quantiles_r(
    mean_samples     = mean_samples,
    sd_samples       = sd_samples,
    se               = se,
    probs            = probs,
    setup            = setup,
    effect_direction = "positive"
  )

  expect_equal(native[["lower"]], ref[["lower"]], tolerance = 1e-5)
  expect_equal(native[["upper"]], ref[["upper"]], tolerance = 1e-5)
})

test_that("native weighted selected-normal summary matches matrix reductions", {

  skip_if_not(.has_native_selnorm_kernel())

  set.seed(141)
  S    <- 17L
  K    <- 5L
  yi   <- -seq_len(K) / 20
  sei  <- seq(.06, .21, length.out = K)
  spec <- .test_step_spec(yi = yi, sei = sei, effect_direction = "negative")
  mean <- matrix(rnorm(S * K, 0, .25), nrow = S, ncol = K)
  sd   <- matrix(runif(S * K, .04, .35), nrow = S, ncol = K)
  psis <- matrix(rexp(S * K), nrow = S, ncol = K)
  psis <- sweep(psis, 2, colSums(psis), "/")

  selection <- spec
  selection[["omega"]]       <- matrix(
    runif(S * spec[["n_bins"]], .03, 2),
    nrow = S,
    ncol = spec[["n_bins"]]
  )
  selection[["alpha"]]       <- rep(0, S)
  selection[["phack_kind"]]  <- rep(0L, S)
  selection[["kernel_mode"]] <- rep(SELKERNEL_STEP, S)
  selection[["kernel_mode"]][seq(3, S, by = 5)] <- SELKERNEL_NORMAL
  selection[["use_normal"]] <- selection[["kernel_mode"]] == SELKERNEL_NORMAL

  native <- .selection_step_weighted_summary(
    yi                = yi,
    mean              = mean,
    sd                = sd,
    sei               = sei,
    psis_weights      = psis,
    selection_context = selection
  )
  cdf     <- .selection_step_cdf_matrix(yi, mean, sd, sei, selection)
  moments <- .selection_step_moments_matrix(mean, sd, sei, selection)

  expect_equal(native[["cdf"]], colSums(psis * cdf), tolerance = 1e-12)
  expect_equal(native[["mean"]], colSums(psis * moments[["mean"]]), tolerance = 1e-12)
  expect_equal(native[["second"]], colSums(psis * moments[["second"]]), tolerance = 1e-12)
})

test_that("p-hacking priors are deferred before RoBMA model construction", {

  phacking <- BayesTools::prior_phacking(form = "linear")
  selection <- BayesTools::prior_weightfunction(
    "one-sided", c(.025), BayesTools::wf_fixed(c(1, .5))
  )

  expect_false("prior_phacking" %in% getNamespaceExports("RoBMA"))
  expect_false("prior_bias" %in% getNamespaceExports("RoBMA"))
  expect_error(
    .use_plot_prior_list_dispatch(phacking),
    "not enabled"
  )
  expect_error(
    bselmodel(
      yi                        = c(.1, .2),
      sei                       = c(.1, .1),
      measure                   = "SMD",
      prior_bias                = phacking,
      prior_unit_information_sd = 1,
      only_priors               = TRUE,
      silent                    = TRUE
    ),
    "not enabled"
  )
  expect_error(
    bselmodel(
      yi                        = c(.1, .2),
      sei                       = c(.1, .1),
      measure                   = "SMD",
      prior_bias                = BayesTools::prior_bias(
        selection = selection,
        phacking  = phacking
      ),
      prior_unit_information_sd = 1,
      only_priors               = TRUE,
      silent                    = TRUE
    ),
    "not enabled"
  )
  expect_error(
    RoBMA(
      yi                        = c(.1, .2),
      sei                       = c(.1, .1),
      measure                   = "SMD",
      prior_bias                = phacking,
      prior_unit_information_sd = 1,
      only_priors               = TRUE,
      silent                    = TRUE
    ),
    "not enabled"
  )
  expect_error(
    RoBMA(
      yi                        = c(.1, .2),
      sei                       = c(.1, .1),
      measure                   = "SMD",
      prior_bias                = BayesTools::prior_bias(
        selection = selection,
        phacking  = phacking
      ),
      prior_unit_information_sd = 1,
      only_priors               = TRUE,
      silent                    = TRUE
    ),
    "not enabled"
  )
  expect_error(
    .selection_spec(
      priors           = list(outcome = list(bias = phacking)),
      yi               = c(.1, .2),
      sei              = c(.1, .1),
      effect_direction = "positive"
    ),
    "not enabled"
  )
  expect_error(
    .selection_spec(
      priors = list(outcome = list(
        bias = BayesTools::prior_bias(selection = selection, phacking = phacking)
      )),
      yi               = c(.1, .2),
      sei              = c(.1, .1),
      effect_direction = "positive"
    ),
    "not enabled"
  )
})

#' @title Interpret brma Results
#'
#' @description Creates a concise textual interpretation of fitted RoBMA
#'   \code{brma} objects.
#'
#' @param object a fitted model object.
#' @param ... additional arguments passed to methods. The \code{brma} method
#' reserves \code{...}; unused arguments error, except deprecated
#' \code{output_scale}, which is accepted with a warning.
#'
#' @return A character vector with class \code{"interpret.brma"}. The
#'   normalized BayesTools interpretation records are stored in the
#'   \code{"records"} attribute; the \code{brma} method also stores
#'   \code{"scope"}, \code{"conditional"}, and optionally \code{"priors"}.
#'
#' @export
interpret <- function(object, ...) {

  UseMethod("interpret")
}

#' @rdname interpret
#' @export
interpret.default <- function(object, ...) {

  stop("'interpret' is available only for fitted brma objects.", call. = FALSE)
}

#' @rdname interpret
#' @param output_measure optional effect-size measure used for the pooled
#'   effect only. Supported conversion targets include \code{"SMD"},
#'   \code{"COR"}, \code{"ZCOR"}, and \code{"OR"} when the input scale allows
#'   conversion.
#' @param transform optional display transformation. \code{"EXP"} exponentiates
#'   log-scale \code{"OR"}, \code{"RR"}, \code{"HR"}, and \code{"IRR"}
#'   pooled effects; aliases for no transform are accepted.
#' @param conditional whether to summarize conditional estimates for RoBMA
#'   product-space objects. Defaults to \code{FALSE}.
#' @param scope character vector specifying sections to include. Use
#'   \code{"core"} for the default concise interpretation, \code{"all"} for all
#'   sections, or any of \code{"components"}, \code{"estimates"},
#'   \code{"moderators"}, \code{"scale"}, and \code{"bias"}.
#' @param probs two quantiles used for credible intervals. Defaults to
#'   \code{c(.025, .975)}.
#' @param central whether estimates are described by posterior mean or median.
#'   Defaults to \code{NULL}, which uses the posterior mean, except for
#'   \code{transform = "EXP"} pooled effects where the posterior mode is used.
#' @param priors whether to print prior distributions after the interpretation.
#'   Defaults to \code{FALSE}.
#' @param digits number of digits after the decimal point.
#'
#' @export
interpret.brma <- function(object, output_measure = NULL, transform = NULL,
                           conditional = FALSE, scope = "core",
                           probs = c(.025, .975),
                           central = NULL,
                           priors = FALSE,
                           digits = 3, ...) {

  if (is.null(object[["fit"]]) || length(object[["fit"]]) == 0L) {
    stop("'interpret' requires a fitted brma object.", call. = FALSE)
  }

  dots <- list(...)
  if ("output_scale" %in% names(dots)) {
    if (!is.null(output_measure)) {
      stop("Use only one of 'output_measure' and deprecated 'output_scale'.",
           call. = FALSE)
    }
    output_measure <- .interpret_output_scale(dots[["output_scale"]])
    dots[["output_scale"]] <- NULL
    warning(
      "'output_scale' is deprecated; use 'output_measure' instead.",
      call. = FALSE
    )
  }
  .check_unused_dots(dots, allowed = character(), caller = "interpret.brma()")

  BayesTools::check_bool(conditional, "conditional")
  if (conditional && !.is_RoBMA(object)) {
    stop("'conditional' interpretations are available only for RoBMA objects.",
         call. = FALSE)
  }

  scope <- .interpret_scope(scope)
  BayesTools::check_real(probs, "probs", lower = 0, upper = 1,
                         check_length = 2, allow_NA = FALSE)
  probs <- sort(probs)
  BayesTools::check_bool(priors, "priors")
  BayesTools::check_int(digits, "digits", lower = 0)

  effect_transform <- .effect_output_setup(
    object         = object,
    output_measure = output_measure,
    transform      = transform
  )
  central_explicit <- !is.null(central)
  if (is.null(central)) {
    central <- "mean"
  } else {
    central <- match.arg(central, c("mean", "median"))
  }
  effect_central <- if (
    !central_explicit && identical(effect_transform[["transform"]], "EXP")
  ) {
    "mode"
  } else {
    central
  }

  summary_object <- summary(
    object,
    probs                    = unique(c(probs[1], .5, probs[2])),
    include_mcmc_diagnostics = FALSE,
    conditional              = conditional
  )

  records <- .interpret_brma_records(
    object           = object,
    summary_object   = summary_object,
    effect_transform = effect_transform,
    conditional      = conditional,
    scope            = scope,
    probs            = probs,
    central          = central,
    effect_central   = effect_central
  )
  text <- .interpret_records_text(
    records  = records,
    digits   = digits,
    averaged = .is_RoBMA(object)
  )

  text <- text[nzchar(text)]
  class(text) <- c("interpret.brma", "character")
  attr(text, "scope")       <- scope
  attr(text, "conditional") <- conditional
  attr(text, "priors")      <- if (priors) .interpret_prior_text(object)
  attr(text, "records")     <- records

  return(text)
}

#' @rdname interpret
#' @param x an \code{interpret.brma} object.
#'
#' @export
print.interpret.brma <- function(x, ...) {

  cat(paste(unclass(x), collapse = "\n"))
  prior_text <- attr(x, "priors")
  if (length(prior_text) > 0L) {
    cat("\n\n")
    cat(paste(prior_text, collapse = "\n"))
  }
  cat("\n")

  return(invisible(x))
}

.interpret_prior_text <- function(object) {

  text <- capture.output(print_prior(object))
  text <- text[nzchar(text)]

  if (length(text) == 0L) {
    return(character())
  }

  return(c("Prior distributions:", text))
}

.interpret_scope <- function(scope) {

  choices <- c(
    "core", "all", "components", "estimates", "moderators", "scale", "bias"
  )
  BayesTools::check_char(scope, "scope", check_length = 0, allow_NA = FALSE)
  scope <- match.arg(scope, choices, several.ok = TRUE)

  if ("all" %in% scope) {
    scope <- c("components", "estimates", "moderators", "scale", "bias")
  }
  if ("core" %in% scope) {
    scope <- c(setdiff(scope, "core"), "components", "estimates")
  }

  return(unique(scope))
}

.interpret_output_scale <- function(output_scale) {

  if (is.null(output_scale)) {
    return(NULL)
  }

  BayesTools::check_char(output_scale, "output_scale", check_length = 1,
                         allow_NA = FALSE)
  output_scale <- toupper(gsub("[^A-Za-z0-9]", "", output_scale))
  aliases <- c(
    D          = "SMD",
    COHENSD    = "SMD",
    SMD        = "SMD",
    R          = "COR",
    COR        = "COR",
    CORRELATION = "COR",
    Z          = "ZCOR",
    ZCOR       = "ZCOR",
    FISHERSZ   = "ZCOR",
    LOGOR      = "OR",
    OR         = "OR",
    Y          = "GEN",
    NONE       = "GEN"
  )

  if (!output_scale %in% names(aliases)) {
    stop("Unknown deprecated 'output_scale'.", call. = FALSE)
  }

  return(unname(aliases[[output_scale]]))
}

.interpret_brma_records <- function(object, summary_object, effect_transform,
                                    conditional, scope, probs, central,
                                    effect_central) {

  sources <- .interpret_brma_sources(
    object           = object,
    summary_object   = summary_object,
    effect_transform = effect_transform,
    conditional      = conditional,
    scope            = scope,
    probs            = probs,
    central          = central,
    effect_central   = effect_central
  )
  plan <- .interpret_brma_plan(
    summary_object   = summary_object,
    effect_transform = effect_transform,
    scope            = scope,
    sources          = sources
  )

  .interpret_records_standard(
    sources = sources,
    plan    = plan
  )
}

.interpret_records_standard <- function(sources, plan) {

  if (!exists("interpret_records", envir = asNamespace("BayesTools"),
              inherits = FALSE)) {
    stop(
      "'interpret' requires a BayesTools version with interpret_records().",
      call. = FALSE
    )
  }

  return(BayesTools::interpret_records(
    sources = sources,
    plan    = plan,
    output  = "records",
    missing = "error"
  ))
}

.interpret_brma_sources <- function(object, summary_object, effect_transform,
                                    conditional, scope, probs, central,
                                    effect_central) {

  sources <- stats::setNames(list(), character())

  if (length(summary_object[["inclusion_components"]]) > 0L) {
    sources[["component_tests"]] <- .interpret_evidence_source(
      summary_object[["inclusion_components"]]
    )
  }
  if (length(summary_object[["inclusion_mods"]]) > 0L) {
    sources[["moderator_tests"]] <- .interpret_evidence_source(
      table = summary_object[["inclusion_mods"]],
      label = paste0("the location component '",
                     rownames(summary_object[["inclusion_mods"]]), "'")
    )
  }
  if (length(summary_object[["inclusion_scale"]]) > 0L) {
    sources[["scale_tests"]] <- .interpret_evidence_source(
      table = summary_object[["inclusion_scale"]],
      label = paste0("the scale component '",
                     rownames(summary_object[["inclusion_scale"]]), "'")
    )
  }

  if ("estimates" %in% scope) {
    effect <- tryCatch(
      pooled_effect(
        object         = object,
        output_measure = effect_transform[["output_measure"]],
        transform      = effect_transform[["transform"]],
        probs          = probs,
        conditional    = conditional
      ),
      error = function(e) NULL
    )
    if (!is.null(effect)) {
      sources[["pooled_effect"]] <- .interpret_record_source(
        .interpret_samples_record(
          samples      = effect,
          parameter    = effect_transform[["label"]],
          conditioning = if (conditional) {
            "conditional on effect inclusion"
          } else {
            NULL
          },
          probs        = probs,
          central      = effect_central
        )
      )
    }

    heterogeneity <- tryCatch(
      pooled_heterogeneity(
        object      = object,
        probs       = probs,
        conditional = conditional
      ),
      error = function(e) NULL
    )
    if (!is.null(heterogeneity)) {
      sources[["pooled_heterogeneity"]] <- .interpret_record_source(
        .interpret_samples_record(
          samples      = heterogeneity,
          parameter    = "tau",
          conditioning = if (conditional) {
            "conditional on heterogeneity inclusion"
          } else {
            NULL
          },
          probs        = probs,
          central      = central
        )
      )
    }

    common_estimates <- .interpret_summary_estimates(summary_object, conditional)
    if (length(common_estimates) > 0L && "rho" %in% rownames(common_estimates)) {
      sources[["common_estimates"]] <- .interpret_estimate_source(
        table   = common_estimates,
        probs   = probs,
        central = central,
        units   = NULL
      )
    }
  }

  if ("moderators" %in% scope) {
    moderator_estimates <- .interpret_summary_estimates_mods(
      summary_object = summary_object,
      conditional    = conditional
    )
    if (length(moderator_estimates) > 0L) {
      sources[["moderator_estimates"]] <- .interpret_estimate_source(
        table   = moderator_estimates,
        probs   = probs,
        central = central,
        units   = if (.effect_output_requested(effect_transform)) {
          NULL
        } else {
          effect_transform[["label"]]
        }
      )
    }
  }

  if ("scale" %in% scope) {
    scale_estimates <- .interpret_summary_estimates_scale(
      summary_object = summary_object,
      conditional    = conditional
    )
    if (length(scale_estimates) > 0L) {
      sources[["scale_estimates"]] <- .interpret_estimate_source(
        table   = scale_estimates,
        probs   = probs,
        central = central,
        units   = NULL
      )
    }
  }

  if ("bias" %in% scope) {
    bias_estimates <- .interpret_summary_estimates_bias(
      summary_object = summary_object,
      conditional    = conditional
    )
    if (length(bias_estimates) > 0L) {
      sources[["bias_estimates"]] <- .interpret_estimate_source(
        table   = bias_estimates,
        probs   = probs,
        central = central,
        units   = NULL
      )
    }
  }

  return(sources)
}

.interpret_evidence_source <- function(table, label = NULL) {

  schema <- list(
    BF             = "inclusion_BF",
    BF_scale       = "linear",
    BF_orientation = "inclusion_over_exclusion"
  )
  if (!is.null(label)) {
    schema[["label"]] <- label
  }

  return(list(
    data   = table,
    schema = schema
  ))
}

.interpret_estimate_source <- function(table, probs, central, units) {

  row <- table[1, , drop = FALSE]
  schema <- list(
    central        = .interpret_central_column(row, central),
    central_name   = central,
    lower          = .interpret_probability_column(row, probs[1]),
    upper          = .interpret_probability_column(row, probs[2]),
    lower_prob     = probs[1],
    upper_prob     = probs[2],
    interval_level = diff(range(probs))
  )
  if (!is.null(units)) {
    schema[["units"]] <- units
  }

  return(list(
    data   = table,
    schema = schema
  ))
}

.interpret_record_source <- function(record) {

  return(list(
    type = "record",
    data = record
  ))
}

.interpret_samples_record <- function(samples, parameter, conditioning,
                                      probs, central) {

  table <- summary(samples, probs = probs)
  if (central == "mode") {
    table <- cbind(
      Mode = .interpret_posterior_mode(as.numeric(samples[, 1L])),
      table
    )
  }
  row <- table[1, , drop = FALSE]

  central_col <- .interpret_central_column(row, central)
  lower_col   <- .interpret_probability_column(row, probs[1])
  upper_col   <- .interpret_probability_column(row, probs[2])

  return(list(
    kind           = "estimate",
    parameter      = parameter,
    central_name   = central,
    central_value  = as.numeric(row[1, central_col]),
    lower_value    = as.numeric(row[1, lower_col]),
    upper_value    = as.numeric(row[1, upper_col]),
    lower_prob     = probs[1],
    upper_prob     = probs[2],
    interval_level = diff(range(probs)),
    conditioning   = conditioning
  ))
}

.interpret_summary_estimates <- function(summary_object, conditional) {

  if (conditional) {
    return(summary_object[["estimates_conditional"]])
  }

  return(summary_object[["estimates"]])
}

.interpret_summary_estimates_mods <- function(summary_object, conditional) {

  if (conditional) {
    return(summary_object[["estimates_mods_conditional"]])
  }

  return(summary_object[["estimates_mods"]])
}

.interpret_summary_estimates_scale <- function(summary_object, conditional) {

  if (conditional) {
    return(summary_object[["estimates_scale_conditional"]])
  }

  return(summary_object[["estimates_scale"]])
}

.interpret_summary_estimates_bias <- function(summary_object, conditional) {

  if (conditional) {
    return(summary_object[["estimates_bias_conditional"]])
  }

  return(summary_object[["estimates_bias"]])
}

.interpret_brma_plan <- function(summary_object, effect_transform, scope,
                                 sources) {

  plan <- list(list(
    kind    = "header",
    section = "model",
    item_id = "header",
    order   = 0,
    text    = paste0(summary_object[["name"]], ".")
  ))

  if ("components" %in% scope && "component_tests" %in% names(sources)) {
    plan <- .interpret_add_component_plan(
      plan  = plan,
      table = summary_object[["inclusion_components"]],
      order = 10
    )
  }

  if ("estimates" %in% scope) {
    if ("pooled_effect" %in% names(sources)) {
      plan[[length(plan) + 1L]] <- list(
        kind    = "estimate",
        section = "estimates",
        item_id = "effect",
        order   = 100,
        source  = "pooled_effect",
        label   = "pooled effect"
      )
    }
    if ("pooled_heterogeneity" %in% names(sources)) {
      plan[[length(plan) + 1L]] <- list(
        kind    = "estimate",
        section = "estimates",
        item_id = "heterogeneity",
        order   = 110,
        source  = "pooled_heterogeneity",
        label   = "pooled heterogeneity"
      )
    }
    if ("common_estimates" %in% names(sources)) {
      plan[[length(plan) + 1L]] <- list(
        kind      = "estimate",
        section   = "estimates",
        item_id   = "rho",
        order     = 120,
        source    = "common_estimates",
        row       = "rho",
        label     = "cluster correlation",
        parameter = "rho"
      )
    }
  }

  if ("moderators" %in% scope) {
    if ("moderator_tests" %in% names(sources)) {
      plan[[length(plan) + 1L]] <- list(
        kind     = "for_each",
        section  = "moderators",
        item_id  = "location_inclusion",
        order    = 200,
        source   = "moderator_tests",
        rows     = "source_order",
        template = "evidence",
        BF_name  = "BF"
      )
    }
    if ("moderator_estimates" %in% names(sources)) {
      plan[[length(plan) + 1L]] <- list(
        kind     = "for_each",
        section  = "moderators",
        item_id  = "location_estimate",
        order    = 300,
        source   = "moderator_estimates",
        rows     = "source_order",
        template = "estimate"
      )
    }
  }

  if ("scale" %in% scope) {
    if ("scale_tests" %in% names(sources)) {
      plan[[length(plan) + 1L]] <- list(
        kind     = "for_each",
        section  = "scale",
        item_id  = "scale_inclusion",
        order    = 400,
        source   = "scale_tests",
        rows     = "source_order",
        template = "evidence",
        BF_name  = "BF"
      )
    }
    if ("scale_estimates" %in% names(sources)) {
      plan[[length(plan) + 1L]] <- list(
        kind     = "for_each",
        section  = "scale",
        item_id  = "scale_estimate",
        order    = 500,
        source   = "scale_estimates",
        rows     = "source_order",
        template = "estimate"
      )
    }
  }

  if ("bias" %in% scope) {
    if (!"components" %in% scope && "component_tests" %in% names(sources) &&
        length(summary_object[["inclusion_components"]]) > 0L &&
        "Publication Bias" %in%
        rownames(summary_object[["inclusion_components"]])) {
      plan[[length(plan) + 1L]] <- .interpret_component_plan_item(
        row   = "Publication Bias",
        order = 600
      )
      plan[[length(plan)]][["section"]] <- "bias"
    }
    if ("bias_estimates" %in% names(sources)) {
      plan[[length(plan) + 1L]] <- list(
        kind     = "for_each",
        section  = "bias",
        item_id  = "publication_bias_estimate",
        order    = 700,
        source   = "bias_estimates",
        rows     = "source_order",
        template = "estimate"
      )
    }
  }

  if (.effect_output_requested(effect_transform) &&
      !is.null(effect_transform[["note"]])) {
    plan[[length(plan) + 1L]] <- list(
      kind    = "note",
      section = "notes",
      item_id = "effect_transform",
      order   = 900,
      text    = .interpret_effect_transform_note(effect_transform)
    )
  }

  return(plan)
}

.interpret_effect_transform_note <- function(effect_transform) {

  note <- effect_transform[["note"]]
  if (is.null(note)) {
    return(NULL)
  }

  return(sub("^Effect estimates", "Pooled effect estimates", note))
}

.interpret_add_component_plan <- function(plan, table, order) {

  rows <- rownames(table)
  for (i in seq_along(rows)) {
    plan[[length(plan) + 1L]] <- .interpret_component_plan_item(
      row   = rows[[i]],
      order = order + i
    )
  }

  return(plan)
}

.interpret_component_plan_item <- function(row, order) {

  BF_names <- c(
    "Effect"           = "BF10",
    "Heterogeneity"    = "BFrf",
    "Publication Bias" = "BFpb"
  )
  labels <- c(
    "Effect"           = "the effect",
    "Heterogeneity"    = "heterogeneity",
    "Publication Bias" = "publication bias"
  )
  item_ids <- c(
    "Effect"           = "effect",
    "Heterogeneity"    = "heterogeneity",
    "Publication Bias" = "publication_bias"
  )

  return(list(
    kind    = "evidence",
    section = "components",
    item_id = if (!is.null(item_ids[[row]])) unname(item_ids[[row]]) else row,
    order   = order,
    source  = "component_tests",
    row     = row,
    label   = if (!is.null(labels[[row]])) unname(labels[[row]]) else tolower(row),
    BF_name = if (!is.null(BF_names[[row]])) unname(BF_names[[row]]) else "BF"
  ))
}

.interpret_records_text <- function(records, digits, averaged) {

  if (nrow(records) == 0L) {
    return(character())
  }

  output <- vapply(seq_len(nrow(records)), function(i) {

    record <- records[i, , drop = FALSE]
    kind   <- record[["kind"]]

    switch(
      kind,
      header   = record[["text"]],
      note     = record[["text"]],
      prior    = record[["text"]],
      footnote = record[["text"]],
      evidence = .interpret_evidence_record_sentence(record, digits),
      estimate = .interpret_estimate_record_sentence(record, digits, averaged),
      ""
    )
  }, character(1))

  names(output) <- vapply(seq_len(nrow(records)), function(i) {
    .interpret_record_output_name(records[i, , drop = FALSE])
  }, character(1))

  return(output)
}

.interpret_evidence_record_sentence <- function(record, digits) {

  BF <- record[["BF_canonical_value"]]
  BF_bound_operator <- .interpret_record_value(
    record[["BF_canonical_bound_operator"]]
  )
  BF_name <- .interpret_record_value(record[["BF_name"]])
  if (is.null(BF_name)) {
    BF_name <- "BF"
  }
  display_BF <- .interpret_display_BF(BF, BF_name, BF_bound_operator)

  text <- paste0(
    .interpret_evidence_prefix(record), ": ",
    .interpret_evidence(
      BF             = BF,
      item           = record[["label"]],
      bound_operator = BF_bound_operator
    ),
    " (", display_BF[["name"]], " ",
    if (is.null(display_BF[["operator"]])) "=" else display_BF[["operator"]],
    " ",
    .interpret_number(display_BF[["value"]], digits),
    ")."
  )

  return(text)
}

.interpret_estimate_record_sentence <- function(record, digits, averaged) {

  central_name <- record[["central_name"]]
  if (is.na(central_name) || !nzchar(central_name)) {
    central_name <- "estimate"
  }
  central_label <- paste(
    c(if (averaged) "model-averaged", central_name),
    collapse = " "
  )

  text <- paste0(
    .interpret_estimate_prefix(record), ": ",
    central_label, " ", record[["parameter"]], " = ",
    .interpret_number(record[["central_value"]], digits),
    ", ", .interpret_interval_label(c(
      record[["lower_prob"]],
      record[["upper_prob"]]
    )),
    " CrI [",
    .interpret_number(record[["lower_value"]], digits),
    ", ",
    .interpret_number(record[["upper_value"]], digits),
    "]"
  )

  units <- .interpret_record_value(record[["units"]])
  if (!is.null(units)) {
    text <- paste0(text, " on the ", units, " scale")
  }

  conditioning <- .interpret_record_value(record[["conditioning"]])
  if (!is.null(conditioning)) {
    text <- paste0(text, " (", conditioning, ")")
  }

  return(paste0(text, "."))
}

.interpret_evidence_prefix <- function(record) {

  section <- record[["section"]]
  row     <- record[["row"]]

  if (section == "components" || section == "bias") {
    return(paste0(row, " inclusion"))
  }
  if (section == "moderators") {
    return(paste0("Location inclusion (", row, ")"))
  }
  if (section == "scale") {
    return(paste0("Scale inclusion (", row, ")"))
  }

  return(record[["label"]])
}

.interpret_estimate_prefix <- function(record) {

  section <- record[["section"]]
  item_id <- record[["item_id"]]
  row     <- record[["row"]]

  if (section == "estimates" && item_id == "effect") {
    return("Pooled effect")
  }
  if (section == "estimates" && item_id == "heterogeneity") {
    return("Pooled heterogeneity")
  }
  if (section == "estimates" && item_id == "rho") {
    return("Cluster correlation")
  }
  if (section == "moderators") {
    return(paste0("Location estimate (", row, ")"))
  }
  if (section == "scale") {
    return(paste0("Scale estimate (", row, ")"))
  }
  if (section == "bias") {
    return(paste0("Publication-bias estimate (", row, ")"))
  }

  return(record[["label"]])
}

.interpret_record_output_name <- function(record) {

  kind    <- record[["kind"]]
  section <- record[["section"]]
  item_id <- record[["item_id"]]
  row     <- record[["row"]]

  if (kind == "header" && section == "model") {
    return("model")
  }
  if (kind == "note") {
    return("note")
  }
  if (section == "components" || (section == "bias" && kind == "evidence")) {
    return(paste0("components.", make.names(tolower(row))))
  }
  if (section == "estimates") {
    return(item_id)
  }
  if (section == "moderators" && kind == "evidence") {
    return(paste0("location.inclusion.", make.names(row)))
  }
  if (section == "moderators" && kind == "estimate") {
    return(paste0("location.estimate.", make.names(row)))
  }
  if (section == "scale" && kind == "evidence") {
    return(paste0("scale.inclusion.", make.names(row)))
  }
  if (section == "scale" && kind == "estimate") {
    return(paste0("scale.estimate.", make.names(row)))
  }
  if (section == "bias" && kind == "estimate") {
    return(paste0("publication.bias.estimate.", make.names(row)))
  }

  return(record[["record_id"]])
}

.interpret_record_value <- function(x) {

  if (length(x) == 0L || is.null(x) || is.na(x) || !nzchar(x)) {
    return(NULL)
  }

  return(x)
}

.interpret_display_BF <- function(BF, BF_name, BF_bound_operator = NULL) {

  if (is.finite(BF) && BF > 0 && BF < 1) {
    return(list(
      name     = .interpret_inverse_BF_name(BF_name),
      value    = 1 / BF,
      operator = .interpret_inverse_BF_bound_operator(BF_bound_operator)
    ))
  }
  if (identical(BF, 0)) {
    return(list(
      name     = .interpret_inverse_BF_name(BF_name),
      value    = Inf,
      operator = .interpret_inverse_BF_bound_operator(BF_bound_operator)
    ))
  }

  return(list(
    name     = BF_name,
    value    = BF,
    operator = BF_bound_operator
  ))
}

.interpret_inverse_BF_name <- function(BF_name) {

  if (BF_name == "BF") {
    return(BF_name)
  }
  if (BF_name == "Inclusion BF") {
    return("Exclusion BF")
  }
  if (BF_name == "Exclusion BF") {
    return("Inclusion BF")
  }
  if (nchar(BF_name) == 4L && substr(BF_name, 1L, 2L) == "BF") {
    return(paste0("BF", substr(BF_name, 4L, 4L), substr(BF_name, 3L, 3L)))
  }

  return(paste0(BF_name, "^-1"))
}

.interpret_inverse_BF_bound_operator <- function(bound_operator) {

  if (is.null(bound_operator)) {
    return(NULL)
  }
  if (bound_operator == "<") {
    return(">")
  }
  if (bound_operator == ">") {
    return("<")
  }

  return(bound_operator)
}

.interpret_evidence <- function(BF, item, bound_operator = NULL) {

  if (!is.finite(BF) && !is.infinite(BF)) {
    return(paste0("evidence for ", item, " unavailable"))
  }
  if (isTRUE(all.equal(BF, 1)) && is.null(bound_operator)) {
    return(paste0("no evidence for or against ", item))
  }

  strength <- if (abs(log(BF)) > log(10)) {
    "strong"
  } else if (abs(log(BF)) > log(3)) {
    "moderate"
  } else {
    "weak"
  }
  direction <- if (BF > 1 || (isTRUE(all.equal(BF, 1)) &&
                              identical(bound_operator, ">"))) {
    "in favor of"
  } else {
    "against"
  }

  return(paste(strength, "evidence", direction, item))
}

.interpret_central_column <- function(row, central) {

  if (central == "mode" && "Mode" %in% colnames(row)) {
    return("Mode")
  }
  if (central == "mean" && "Mean" %in% colnames(row)) {
    return("Mean")
  }
  if (central == "median" && "Median" %in% colnames(row)) {
    return("Median")
  }
  if (central == "median" && "0.5" %in% colnames(row)) {
    return("0.5")
  }

  stop("Could not find requested central estimate in summary table.",
       call. = FALSE)
}

.interpret_posterior_mode <- function(x) {

  x <- x[is.finite(x)]
  if (length(x) == 0L) {
    return(NA_real_)
  }

  unique_x <- unique(x)
  if (length(unique_x) == 1L) {
    return(unique_x)
  }

  density <- stats::density(x)
  return(density[["x"]][which.max(density[["y"]])])
}

.interpret_probability_column <- function(row, prob) {

  column <- as.character(prob)
  if (column %in% colnames(row)) {
    return(column)
  }

  numeric_columns <- suppressWarnings(as.numeric(colnames(row)))
  index <- which(
    !is.na(numeric_columns) &
      abs(numeric_columns - prob) < sqrt(.Machine$double.eps)
  )
  if (length(index) == 1L) {
    return(colnames(row)[index])
  }

  stop("Could not find requested credible interval column in summary table.",
       call. = FALSE)
}

.interpret_interval_label <- function(probs) {

  level <- round(100 * diff(range(probs)), 1)
  if (isTRUE(all.equal(level, round(level)))) {
    level <- round(level)
  }

  return(paste0(level, "%"))
}

.interpret_number <- function(x, digits) {

  if (is.na(x)) {
    return("NA")
  }
  if (is.infinite(x)) {
    return(if (x > 0) "Inf" else "-Inf")
  }
  if (abs(x) < 0.5 * 10^(-digits)) {
    x <- 0
  }

  return(formatC(x, format = "f", digits = digits))
}

context("Interpret")

source(testthat::test_path("common-functions.R"))

REFERENCE_DIR <<- testthat::test_path("..", "results", "interpret")

skip_if_no_fits()

interpret_fit_names <- c(
  "bcg_meta-analysis",
  "bcg_BMA.glmm",
  "dat.lehmann2018_RoBMA",
  "dat.lehmann2018_RoBMA_mods2",
  "dat.lehmann2018_RoBMA_3lvl_mods_scale",
  "bangertdrowns2004_location-scale",
  "konstantopoulos2011_3lvl"
)
skip_if_missing_fits(interpret_fit_names)

fits <- lazy_fits(interpret_fit_names, validate = FALSE)

interpret_output_text <- function(object, ...) {

  return(paste(capture.output(print(interpret(object, ...))), collapse = "\n"))
}

expect_interpret_contract <- function(output, name, priors = FALSE) {

  expect_s3_class(output, "interpret.brma")
  expect_type(unclass(output), "character")
  expect_true(length(output) > 0L, info = paste0("non-empty output for ", name))
  expect_true(all(nzchar(output)), info = paste0("non-empty sections for ", name))
  expect_false(any(grepl("__xXx__", output, fixed = TRUE)),
               info = paste0("decoded labels for ", name))
  expect_true("model" %in% names(output), info = paste0("model section for ", name))

  printed <- capture.output(print(output))
  expect_true(any(grepl("Bayesian", printed, fixed = TRUE)),
              info = paste0("printed model name for ", name))
  expect_false(any(grepl("__xXx__", printed, fixed = TRUE)),
               info = paste0("printed decoded labels for ", name))
  if (priors) {
    expect_true(any(printed == ""), info = paste0("prior separator for ", name))
  } else {
    expect_false(any(printed == ""), info = paste0("no empty printed lines for ", name))
  }
  expect_false(any(grepl("prior =|posterior =|inverse =", printed)),
               info = paste0("compact BF reporting for ", name))

  records <- attr(output, "records")
  expect_s3_class(records, "BayesTools_interpret_records")
  expect_true(all(c(
    "record_id", "kind", "section", "item_id", "source", "row",
    "BF_canonical_value", "BF_canonical_bound_operator",
    "central_name", "central_value"
  ) %in% colnames(records)), info = paste0("records schema for ", name))
  expect_true("model.header.header" %in% records[["record_id"]],
              info = paste0("records header for ", name))
}

test_that("interpret.brma returns a stable object contract", {

  for (name in names(fits)) {
    output <- interpret(fits[[name]])
    expect_interpret_contract(output, name)
  }
})

test_that("interpret.brma handles options and errors", {

  output <- interpret(
    fits[["dat.lehmann2018_RoBMA_mods2"]],
    scope       = c("components", "moderators"),
    central     = "mean",
    probs       = c(.10, .90),
    digits      = 2
  )
  expect_interpret_contract(output, "dat.lehmann2018_RoBMA_mods2 options")
  expect_true(any(grepl("80% CrI", output, fixed = TRUE)))
  expect_true(any(grepl("mean", output, fixed = TRUE)))
  expect_true(any(grepl("Location inclusion", output, fixed = TRUE)))

  brma_output <- interpret(fits[["bcg_meta-analysis"]])
  expect_true(any(grepl("Pooled effect: mean", brma_output, fixed = TRUE)))
  expect_false(any(grepl("model-averaged", brma_output, fixed = TRUE)))

  robma_output <- interpret(fits[["dat.lehmann2018_RoBMA"]])
  expect_true(any(grepl("Pooled effect: model-averaged mean", robma_output,
                        fixed = TRUE)))
  expect_true(any(grepl("BF01 =", robma_output, fixed = TRUE)))
  robma_records <- attr(robma_output, "records")
  heterogeneity_record <- robma_records[
    robma_records[["record_id"]] == "components.heterogeneity.evidence",
    ,
    drop = FALSE
  ]
  expect_equal(heterogeneity_record[["BF_canonical_bound_operator"]], ">")
  expect_true(is.finite(heterogeneity_record[["BF_canonical_value"]]))

  exp_output <- interpret(fits[["bcg_BMA.glmm"]], transform = "EXP")
  expect_true(any(grepl("Pooled effect: model-averaged mode", exp_output,
                        fixed = TRUE)))
  expect_true(any(grepl("Pooled effect estimates are summarized", exp_output,
                        fixed = TRUE)))

  transformed_mods <- interpret(
    fits[["dat.lehmann2018_RoBMA_mods2"]],
    scope          = "moderators",
    output_measure = "COR"
  )
  expect_false(any(grepl("on the correlation scale", transformed_mods,
                         fixed = TRUE)))

  prior_output <- interpret(fits[["dat.lehmann2018_RoBMA"]], priors = TRUE)
  expect_interpret_contract(prior_output, "dat.lehmann2018_RoBMA priors",
                            priors = TRUE)
  prior_text <- attr(prior_output, "priors")
  expect_true(any(grepl("Prior distributions", prior_text, fixed = TRUE)))
  expect_true(any(grepl("mu:", prior_text, fixed = TRUE)))

  prior_printed <- capture.output(print(prior_output))
  expect_true(any(prior_printed == ""))
  expect_equal(sum(prior_printed == ""), 1L)
  expect_true(any(grepl("Prior distributions", prior_printed, fixed = TRUE)))

  expect_warning(
    legacy <- interpret(fits[["dat.lehmann2018_RoBMA"]], output_scale = "r"),
    "deprecated"
  )
  expect_true(any(grepl("correlation", legacy, fixed = TRUE)))

  expect_error(interpret(1), "brma objects")
  expect_error(
    interpret(fits[["bcg_meta-analysis"]], conditional = TRUE),
    "RoBMA objects"
  )
  expect_error(
    interpret(fits[["bcg_meta-analysis"]], unused = TRUE),
    "Unused argument"
  )
})

test_that("interpret.brma printed output matches reference text", {

  cases <- list(
    list(
      name     = "bcg_meta-analysis",
      filename = "interpret-bcg_meta-analysis-core.txt",
      args     = list()
    ),
    list(
      name     = "bcg_BMA.glmm",
      filename = "interpret-bcg_BMA.glmm-exp.txt",
      args     = list(transform = "EXP")
    ),
    list(
      name     = "dat.lehmann2018_RoBMA",
      filename = "interpret-dat.lehmann2018_RoBMA-core.txt",
      args     = list()
    ),
    list(
      name     = "dat.lehmann2018_RoBMA",
      filename = "interpret-dat.lehmann2018_RoBMA-core-priors.txt",
      args     = list(priors = TRUE)
    ),
    list(
      name     = "dat.lehmann2018_RoBMA",
      filename = "interpret-dat.lehmann2018_RoBMA-bias.txt",
      args     = list(scope = "bias")
    ),
    list(
      name     = "dat.lehmann2018_RoBMA_mods2",
      filename = "interpret-dat.lehmann2018_RoBMA_mods2-all-conditional.txt",
      args     = list(scope = "all", conditional = TRUE)
    ),
    list(
      name     = "dat.lehmann2018_RoBMA_3lvl_mods_scale",
      filename = "interpret-dat.lehmann2018_RoBMA_3lvl_mods_scale-core.txt",
      args     = list()
    ),
    list(
      name     = "bangertdrowns2004_location-scale",
      filename = "interpret-bangertdrowns2004_location-scale-all.txt",
      args     = list(scope = "all")
    ),
    list(
      name     = "konstantopoulos2011_3lvl",
      filename = "interpret-konstantopoulos2011_3lvl-core.txt",
      args     = list()
    )
  )

  for (case in cases) {
    output <- do.call(
      interpret_output_text,
      c(list(object = fits[[case[["name"]]]]), case[["args"]])
    )

    test_reference_text(
      text     = output,
      filename = case[["filename"]],
      info_msg = paste0("Interpret reference mismatch for ", case[["name"]])
    )
  }
})

#' @title Options for the RoBMA package
#'
#' @description Inspect or change package-level defaults used by RoBMA.
#'
#' @param name a single non-missing character string matching one public option
#' exactly; for available options, see details below.
#' @param ... named option(s) to change. Names must be exact public option names,
#' nonempty, and unique.
#'
#' @details
#' The available options are:
#' \describe{
#'   \item{\code{max_cores}}{number of cores to use for parallel computing (default is one fewer than detected logical cores, with a minimum/fallback of 1)}
#'   \item{\code{check_scaling}}{whether to check scaling of predictors (default \code{TRUE})}
#'   \item{\code{silent}}{whether to suppress output (default \code{FALSE})}
#'   \item{\code{autocompute.loo}}{whether to automatically compute LOO (default \code{FALSE})}
#'   \item{\code{autocompute.waic}}{whether to automatically compute WAIC (default \code{FALSE})}
#'   \item{\code{autocompute.marglik}}{whether to automatically compute marginal likelihood (default \code{FALSE})}
#'   \item{\code{cluster_likelihood.n_gamma}}{number of Gauss-Hermite nodes used for cluster-unit log-likelihoods (default \code{15})}
#'   \item{\code{default_UISD.effect}}{default scaling of the unit information standard deviation for the effect size parameter (default \code{0.5})}
#'   \item{\code{default_UISD.heterogeneity}}{default scaling of the unit information standard deviation for the heterogeneity parameter (default \code{0.25})}
#'   \item{\code{default_UISD.mods}}{default scaling of the unit information standard deviation for the moderators (default \code{0.25})}
#'   \item{\code{default_UISD.scale}}{default scaling of the unit information standard deviation for the scale parameter (default \code{0.5})}
#'   \item{\code{default_informed_priors.mods}}{default scaling of informed priors for moderators (default \code{0.5})}
#'   \item{\code{default_informed_priors.scale}}{default scaling of informed priors for the scale parameter (default \code{0.5})}
#'   \item{\code{default_bias_weightfunction.alpha}}{default alpha for the weightfunction (default \code{1})}
#'   \item{\code{default_bias_PET.scale}}{default scale for the PET (default \code{1})}
#'   \item{\code{default_bias_PEESE.scale}}{default scale for the PEESE (default \code{5})}
#' }
#'
#' Boolean options require scalar \code{TRUE} or \code{FALSE} values.
#' \code{max_cores} must be integer-like and at least 1;
#' \code{cluster_likelihood.n_gamma} must be integer-like and at least 3.
#' Scale and alpha defaults must be finite positive numbers.
#'
#' @return `RoBMA.options()` invisibly returns a named list with all current
#' options after applying any changes. `RoBMA.get_option()` returns the current
#' value of the requested option.
#'
#' @export RoBMA.options
#' @export RoBMA.get_option
#' @name RoBMA_options
#' @aliases RoBMA_options RoBMA.options RoBMA.get_option
NULL


#' @rdname RoBMA_options
RoBMA.options    <- function(...) {

  opts <- list(...)

  if (length(opts) > 0 && (is.null(names(opts)) || any(names(opts) == ""))) {
    stop("All options must be named.", call. = FALSE)
  }

  option_names <- RoBMA.private[["option_names"]]

  if (length(opts) > 0 && anyDuplicated(names(opts))) {
    stop("Option names must be unique.", call. = FALSE)
  }

  for (i in seq_along(opts)) {

    if (!names(opts)[i] %in% option_names) {
      stop(paste("Unmatched or ambiguous option '", names(opts)[i], "'", sep = ""), call. = FALSE)
    }

    value <- .RoBMA_validate_option(names(opts)[i], opts[[i]])
    assign(names(opts)[i], value, envir = RoBMA.private)
  }

  return(invisible(.RoBMA_current_options()))
}

#' @rdname RoBMA_options
RoBMA.get_option <- function(name) {

  if (!is.character(name) || length(name) != 1 || is.na(name)) {
    stop("Only 1 option can be retrieved at a time", call. = FALSE)
  }

  option_names <- RoBMA.private[["option_names"]]

  if (!name %in% option_names) {
    stop(paste("Unmatched or ambiguous option '", name, "'", sep = ""), call. = FALSE)
  }

  return(RoBMA.private[[name]])
}

# export the function directly to suppress import warnings
.runjags__findjags <- function() runjags::findjags()

# adapted from the runjags package version 2.2.0
RoBMA.private <- new.env()

assign("defaultoptions",  list(
  jagspath = expression(.runjags__findjags())
), envir = RoBMA.private)

assign("options",         RoBMA.private$defaultoptions,   envir = RoBMA.private)
assign("RoBMA_version",   "notset",                       envir = RoBMA.private)
assign("min_jags_major",  4,                              envir = RoBMA.private)
assign("max_jags_major",  4,                              envir = RoBMA.private)

.RoBMA_default_max_cores <- function() {

  cores <- parallel::detectCores(logical = TRUE)
  if (length(cores) != 1 || is.na(cores) || !is.finite(cores)) {
    return(1L)
  }

  return(max(1L, as.integer(cores) - 1L))
}

.RoBMA_check_option_bool <- function(value, name) {

  if (!is.logical(value) || length(value) != 1 || is.na(value)) {
    stop(paste0("Option '", name, "' must be TRUE or FALSE."), call. = FALSE)
  }

  return(value)
}

.RoBMA_check_option_int <- function(value, name, lower = -Inf) {

  if (!is.numeric(value) || length(value) != 1 || is.na(value) ||
      !is.finite(value) || value != as.integer(value) || value < lower) {
    stop(paste0("Option '", name, "' must be an integer >= ", lower, "."), call. = FALSE)
  }

  return(as.integer(value))
}

.RoBMA_check_option_real <- function(value, name, lower = -Inf, allow_bound = TRUE) {

  if (!is.numeric(value) || length(value) != 1 || is.na(value) || !is.finite(value)) {
    stop(paste0("Option '", name, "' must be a finite number."), call. = FALSE)
  }

  below_bound <- if (allow_bound) value < lower else value <= lower
  if (below_bound) {
    bound_text <- if (allow_bound) " >= " else " > "
    stop(paste0("Option '", name, "' must be", bound_text, lower, "."), call. = FALSE)
  }

  return(as.numeric(value))
}

.RoBMA_option_schema <- list(
  "max_cores" = list(
    default  = .RoBMA_default_max_cores,
    validate = function(value, name) .RoBMA_check_option_int(value, name, lower = 1L)
  ),
  "check_scaling" = list(
    default  = TRUE,
    validate = .RoBMA_check_option_bool
  ),
  "silent" = list(
    default  = FALSE,
    validate = .RoBMA_check_option_bool
  ),
  "autocompute.loo" = list(
    default  = FALSE,
    validate = .RoBMA_check_option_bool
  ),
  "autocompute.waic" = list(
    default  = FALSE,
    validate = .RoBMA_check_option_bool
  ),
  "autocompute.marglik" = list(
    default  = FALSE,
    validate = .RoBMA_check_option_bool
  ),
  "cluster_likelihood.n_gamma" = list(
    default  = 15L,
    validate = function(value, name) .RoBMA_check_option_int(value, name, lower = 3L)
  ),
  "default_UISD.effect" = list(
    default  = 1/2,
    validate = function(value, name) .RoBMA_check_option_real(value, name, lower = 0, allow_bound = FALSE)
  ),
  "default_UISD.heterogeneity" = list(
    default  = 1/4,
    validate = function(value, name) .RoBMA_check_option_real(value, name, lower = 0, allow_bound = FALSE)
  ),
  "default_UISD.mods" = list(
    default  = 1/4,
    validate = function(value, name) .RoBMA_check_option_real(value, name, lower = 0, allow_bound = FALSE)
  ),
  "default_UISD.scale" = list(
    default  = 1/2,
    validate = function(value, name) .RoBMA_check_option_real(value, name, lower = 0, allow_bound = FALSE)
  ),
  "default_informed_priors.mods" = list(
    default  = 1/2,
    validate = function(value, name) .RoBMA_check_option_real(value, name, lower = 0, allow_bound = FALSE)
  ),
  "default_informed_priors.scale" = list(
    default  = 1/2,
    validate = function(value, name) .RoBMA_check_option_real(value, name, lower = 0, allow_bound = FALSE)
  ),
  "default_bias_weightfunction.alpha" = list(
    default  = 1,
    validate = function(value, name) .RoBMA_check_option_real(value, name, lower = 0, allow_bound = FALSE)
  ),
  "default_bias_PET.scale" = list(
    default  = 1,
    validate = function(value, name) .RoBMA_check_option_real(value, name, lower = 0, allow_bound = FALSE)
  ),
  "default_bias_PEESE.scale" = list(
    default  = 5,
    validate = function(value, name) .RoBMA_check_option_real(value, name, lower = 0, allow_bound = FALSE)
  )
)

assign("option_names", names(.RoBMA_option_schema), envir = RoBMA.private)

for (option_name in names(.RoBMA_option_schema)) {

  default <- .RoBMA_option_schema[[option_name]][["default"]]
  value   <- if (is.function(default)) default() else default
  assign(option_name, value, envir = RoBMA.private)
}

.RoBMA_validate_option <- function(name, value) {

  validator <- .RoBMA_option_schema[[name]][["validate"]]
  return(validator(value, name))
}


.RoBMA_current_options <- function() {

  option_names <- RoBMA.private[["option_names"]]
  out          <- mget(option_names, envir = RoBMA.private, inherits = FALSE)

  return(out)
}


# check and fix number of threads (sometimes bugs out during installation)
.check_max_cores <- function() {

  max_cores <- .RoBMA_default_max_cores()
  if (RoBMA.private$max_cores > max_cores || RoBMA.private$max_cores < 1) {
    RoBMA.options(max_cores = max_cores)
  }
}


#' @title Control MCMC fitting process
#'
#' @description \code{set_autofit_control} and \code{set_convergence_checks} control settings for the
#' autofit process of the MCMC JAGS sampler and specify termination criteria and convergence checks.
#'
#' @param max_Rhat maximum value of the R-hat diagnostic.
#' Defaults to \code{1.05}.
#' @param min_ESS minimum estimated sample size.
#' Defaults to \code{500}.
#' @param max_error maximum value of the MCMC error.
#' Defaults to \code{NULL}. Be aware that PEESE publication bias
#' adjustment can have estimates on different scale than the rest of
#' the output, resulting in relatively large max MCMC error.
#' @param max_SD_error maximum value of the proportion of MCMC error
#' of the estimated SD of the parameter.
#' Defaults to \code{NULL}.
#' @param max_time list with the time and unit specifying the maximum
#' autofitting process per model. Passed to \link[base]{difftime} function
#' (possible units are \code{"secs"}, \code{"mins"}, \code{"hours"},
#' \code{"days"}, and \code{"weeks"}).
#' Defaults to \code{list(time = 60, unit = "mins")}.
#' @param sample_extend number of samples to extend the fitting process if
#' the criteria are not satisfied.
#' Defaults to \code{1000}.
#' @param restarts number of times new initial values should be generated in case a
#' model fails to initialize. Defaults to \code{10}.
#' @param max_extend number of times after which the automatic fitting function is stopped.
#' Defaults to \code{10}.
#'
#' @details
#' \code{set_autofit_control} controls the automatic model fitting process, determining when to
#' stop iterating based on convergence criteria. \code{set_convergence_checks} sets thresholds
#' for determining whether a model has converged adequately.
#'
#' The autofit control manages computational resources by setting maximum time limits and
#' determining how many additional samples to draw if convergence criteria are not met.
#' The convergence checks determine the quality standards that fitted models must meet.
#' Autofit thresholds \code{max_Rhat}, \code{min_ESS}, \code{max_error},
#' \code{max_SD_error}, \code{max_time}, \code{restarts}, and
#' \code{max_extend} can be set to \code{NULL}; \code{sample_extend} must be a
#' positive integer.
#' Thresholds can be disabled with \code{NULL}; otherwise \code{max_Rhat} must be
#' at least 1, \code{min_ESS} and \code{max_error} must be nonnegative, and
#' \code{max_SD_error} must be between 0 and 1.
#'
#' @examples
#' # Set custom autofit control with shorter time limit
#' autofit_ctrl <- set_autofit_control(
#'   max_time = list(time = 30, unit = "mins"),
#'   sample_extend = 2000
#' )
#'
#' # Set custom convergence checks with stricter criteria
#' conv_checks <- set_convergence_checks(
#'   max_Rhat = 1.01,
#'   min_ESS = 1000
#' )
#'
#' \dontrun{
#' if (requireNamespace("metadat", quietly = TRUE)) {
#'   data(dat.lehmann2018, package = "metadat")
#'
#'   fit <- RoBMA(
#'     yi                 = yi,
#'     vi                 = vi,
#'     data               = dat.lehmann2018,
#'     measure            = "SMD",
#'     autofit_control    = autofit_ctrl,
#'     convergence_checks = conv_checks
#'   )
#' }
#' }
#'
#' @return \code{set_autofit_control} returns a list of autofit control settings
#' including \code{max_Rhat}, \code{min_ESS}, \code{max_error},
#' \code{max_SD_error}, \code{max_time}, \code{sample_extend},
#' \code{restarts}, \code{max_extend}, and delegated
#' \code{check_indicators}.
#' \code{set_convergence_checks} returns a list with convergence thresholds
#' \code{max_Rhat}, \code{min_ESS}, \code{max_error}, \code{max_SD_error},
#' and delegated \code{check_indicators}.
#'
#' @export set_autofit_control
#' @export set_convergence_checks
#' @name RoBMA_control
#' @aliases set_autofit_control set_convergence_checks
#'
#' @seealso [RoBMA()]
NULL

#' @rdname RoBMA_control
set_autofit_control     <- function(max_Rhat = 1.05, min_ESS = 500, max_error = NULL, max_SD_error = NULL, max_time = list(time = 60, unit = "mins"), sample_extend = 1000, restarts = 10, max_extend = 10){

  autofit_settings <- list(
    max_Rhat      = max_Rhat,
    min_ESS       = min_ESS,
    max_error     = max_error,
    max_SD_error  = max_SD_error,
    max_time      = max_time,
    sample_extend = sample_extend,
    restarts      = restarts,
    max_extend    = max_extend
  )
  autofit_settings <- BayesTools::JAGS_check_and_list_autofit_settings(autofit_settings, call = "Checking 'autofit_control':\n\t")

  return(autofit_settings)
}
#' @rdname RoBMA_control
set_convergence_checks  <- function(max_Rhat = 1.05, min_ESS = 500, max_error = NULL, max_SD_error = NULL){

  convergence_checks <- list(
    max_Rhat     = max_Rhat,
    min_ESS      = min_ESS,
    max_error    = max_error,
    max_SD_error = max_SD_error
  )
  # allows NULL arguments so it can be used in this way too
  convergence_checks <- .check_and_list_convergence_checks(convergence_checks)

  return(convergence_checks)
}


# Internal control parameter update functions:
# Purpose: Merge user-specified parameters with existing control settings
# Used for updating model fitting, convergence checking, and autofit parameters
# LLM Note: These functions implement parameter inheritance for model updates

# Update MCMC fitting control parameters:
# Handles chains, adaptation, burnin, sampling parameters for JAGS
.update_fit_control     <- function(old_fit_control, chains, adapt, burnin, sample, thin, autofit, parallel, cores, silent, seed){

  if(is.null(chains)){
    chains <- old_fit_control[["chains"]]
  }
  if(is.null(adapt)){
    adapt  <- old_fit_control[["adapt"]]
  }
  if(is.null(burnin)){
    burnin <- old_fit_control[["burnin"]]
  }
  if(is.null(sample)){
    sample <- old_fit_control[["sample"]]
  }
  if(is.null(thin)){
    thin  <- old_fit_control[["thin"]]
  }
  if(is.null(autofit)){
    autofit  <- old_fit_control[["autofit"]]
  }
  if(is.null(parallel)){
    parallel <- old_fit_control[["parallel"]]
  }
  if(is.null(cores)){
    cores <- old_fit_control[["cores"]]
  }
  if(is.null(silent)){
    silent <- old_fit_control[["silent"]]
  }
  if(is.null(seed)){
    seed   <- old_fit_control[["seed"]]
  }

  new_fit_control <- BayesTools::JAGS_check_and_list_fit_settings(chains = chains, adapt = adapt, burnin = burnin, sample = sample, thin = thin, autofit = autofit, parallel = parallel, cores = cores, silent = silent, seed = seed)

  return(new_fit_control)
}
.update_autofit_control <- function(old_autofit_control, autofit_control){

  if(!is.null(autofit_control[["max_Rhat"]])){
    max_Rhat <- autofit_control[["max_Rhat"]]
  }else{
    max_Rhat <- old_autofit_control[["max_Rhat"]]
  }
  if(!is.null(autofit_control[["min_ESS"]])){
    min_ESS <- autofit_control[["min_ESS"]]
  }else{
    min_ESS <- old_autofit_control[["min_ESS"]]
  }
  if(!is.null(autofit_control[["max_error"]])){
    max_error <- autofit_control[["max_error"]]
  }else{
    max_error <- old_autofit_control[["max_error"]]
  }
  if(!is.null(autofit_control[["max_SD_error"]])){
    max_SD_error <- autofit_control[["max_SD_error"]]
  }else{
    max_SD_error <- old_autofit_control[["max_SD_error"]]
  }
  if(!is.null(autofit_control[["max_time"]])){
    max_time <- autofit_control[["max_time"]]
  }else{
    max_time <- old_autofit_control[["max_time"]]
  }
  if(!is.null(autofit_control[["sample_extend"]])){
    sample_extend <- autofit_control[["sample_extend"]]
  }else{
    sample_extend <- old_autofit_control[["sample_extend"]]
  }
  if(!is.null(autofit_control[["restarts"]])){
    restarts <- autofit_control[["restarts"]]
  }else{
    restarts <- old_autofit_control[["restarts"]]
  }
  if(!is.null(autofit_control[["max_extend"]])){
    max_extend <- autofit_control[["max_extend"]]
  }else{
    max_extend <- old_autofit_control[["max_extend"]]
  }

  new_autofit_control <- set_autofit_control(max_Rhat = max_Rhat, min_ESS = min_ESS, max_error = max_error, max_SD_error = max_SD_error, max_time = max_time, sample_extend = sample_extend, restarts = restarts, max_extend = max_extend)
  new_autofit_control <- BayesTools::JAGS_check_and_list_autofit_settings(autofit_control = new_autofit_control)

  return(new_autofit_control)
}
.update_convergence_checks <- function(old_convergence_checks, convergence_checks){

  if(!is.null(convergence_checks[["max_Rhat"]])){
    max_Rhat <- convergence_checks[["max_Rhat"]]
  }else{
    max_Rhat <- old_convergence_checks[["max_Rhat"]]
  }
  if(!is.null(convergence_checks[["min_ESS"]])){
    min_ESS <- convergence_checks[["min_ESS"]]
  }else{
    min_ESS <- old_convergence_checks[["min_ESS"]]
  }
  if(!is.null(convergence_checks[["max_error"]])){
    max_error <- convergence_checks[["max_error"]]
  }else{
    max_error <- old_convergence_checks[["max_error"]]
  }
  if(!is.null(convergence_checks[["max_SD_error"]])){
    max_SD_error <- convergence_checks[["max_SD_error"]]
  }else{
    max_SD_error <- old_convergence_checks[["max_SD_error"]]
  }
  new_convergence_checks <- set_convergence_checks(max_Rhat = max_Rhat, min_ESS = min_ESS, max_error = max_error, max_SD_error = max_SD_error)
  new_convergence_checks <- .check_and_list_convergence_checks(new_convergence_checks)

  return(new_convergence_checks)
}


.check_and_list_convergence_checks <- function(convergence_checks){

  convergence_checks <- BayesTools::JAGS_check_and_list_autofit_settings(convergence_checks, skip_sample_extend = TRUE, call = "Checking 'convergence_checks':\n\t")

  return(convergence_checks)
}

# Apply shared metafor-style defaults for observed-study point markers.
.plot_point_style_defaults <- function(dots) {

  if (is.null(dots[["pch"]]))  dots[["pch"]]  <- 21
  if (is.null(dots[["col"]]))  dots[["col"]]  <- "black"
  if (is.null(dots[["bg"]]))   dots[["bg"]]   <- "#A6A6A6"
  if (is.null(dots[["cex"]]))  dots[["cex"]]  <- 1
  if (is.null(dots[["size"]])) dots[["size"]] <- 2

  return(dots)
}

.normalize_max_samples <- function(max_samples, argument = "max_samples",
                                   minimum = 10L) {

  valid <- is.numeric(max_samples) &&
    length(max_samples) == 1L &&
    !is.na(max_samples) &&
    max_samples > 0

  if (!valid || (!is.infinite(max_samples) && max_samples != floor(max_samples))) {
    stop("'", argument, "' must be a single positive integer or Inf.", call. = FALSE)
  }

  if (!is.infinite(max_samples) && max_samples < minimum) {
    stop("'", argument, "' must be at least ", minimum, ".", call. = FALSE)
  }

  if (is.infinite(max_samples)) {
    return(Inf)
  }

  return(as.integer(max_samples))
}

.thin_sample_rows <- function(n_samples, max_samples) {

  if (is.infinite(max_samples) || n_samples <= max_samples) {
    return(NULL)
  }

  return(unique(round(seq(
    from       = 1,
    to         = n_samples,
    length.out = as.integer(max_samples)
  ))))
}

.thin_sample_rows_by_group <- function(group, max_samples) {

  n_samples <- length(group)
  if (is.infinite(max_samples) || n_samples <= max_samples) {
    return(NULL)
  }
  if (anyNA(group)) {
    stop("'group' must not contain missing values.", call. = FALSE)
  }

  max_samples <- as.integer(max_samples)
  group       <- factor(group)
  group_rows  <- split(seq_len(n_samples), group)
  group_n     <- as.integer(vapply(group_rows, length, integer(1)))
  raw_n       <- group_n / n_samples * max_samples
  selected_n  <- floor(raw_n)
  remaining   <- max_samples - sum(selected_n)

  while (remaining > 0L && any(selected_n < group_n)) {
    available  <- which(selected_n < group_n)
    remainders <- raw_n - selected_n
    add_to     <- available[which.max(remainders[available])]
    selected_n[add_to] <- selected_n[add_to] + 1L
    remaining          <- remaining - 1L
  }

  if (max_samples >= length(group_n) && any(selected_n == 0L)) {
    for (i in which(selected_n == 0L)) {
      donors <- which(selected_n > 1L)
      if (length(donors) == 0L) {
        break
      }
      donor <- donors[which.max(selected_n[donors])]
      selected_n[donor] <- selected_n[donor] - 1L
      selected_n[i]     <- 1L
    }
  }

  rows <- integer(0)
  for (i in seq_along(group_rows)) {
    if (selected_n[i] == 0L) {
      next
    }
    selected <- group_rows[[i]][round(seq(
      from       = 1,
      to         = length(group_rows[[i]]),
      length.out = selected_n[i]
    ))]
    rows     <- c(rows, selected)
  }

  return(sort(unique(rows)))
}

.check_plot_numeric <- function(x, argument, check_length = NULL,
                                lower = -Inf, upper = Inf,
                                allow_null = TRUE) {

  if (is.null(x)) {
    if (allow_null) {
      return(invisible(TRUE))
    }
    stop("'", argument, "' must not be NULL.", call. = FALSE)
  }

  if (!is.numeric(x) || length(x) == 0L || anyNA(x) || any(!is.finite(x))) {
    stop("'", argument, "' must be a finite numeric vector.", call. = FALSE)
  }
  if (!is.null(check_length) && length(x) != check_length) {
    stop("'", argument, "' must have length ", check_length, ".", call. = FALSE)
  }
  if (any(x < lower) || any(x > upper)) {
    stop("'", argument, "' must be between ", lower, " and ", upper, ".", call. = FALSE)
  }

  return(invisible(TRUE))
}

.check_plot_limits <- function(x, argument) {

  .check_plot_numeric(x, argument, check_length = 2, allow_null = TRUE)
  if (is.null(x)) {
    return(invisible(TRUE))
  }
  if (x[1] >= x[2]) {
    stop("'", argument, "' must be an increasing numeric vector of length 2.", call. = FALSE)
  }

  return(invisible(TRUE))
}

.check_plot_label <- function(x, argument) {

  if (is.null(x)) {
    return(invisible(TRUE))
  }
  if (is.character(x) && length(x) == 1L && !is.na(x)) {
    return(invisible(TRUE))
  }
  if (is.expression(x) && length(x) == 1L) {
    return(invisible(TRUE))
  }

  stop("'", argument, "' must be a single character string or expression.", call. = FALSE)
}

.check_plot_function <- function(x, argument) {

  if (!is.null(x) && !is.function(x)) {
    stop("'", argument, "' must be a function or NULL.", call. = FALSE)
  }

  return(invisible(TRUE))
}

.check_plot_list <- function(x, argument) {

  if (!is.null(x) && !is.list(x)) {
    stop("'", argument, "' must be a list or NULL.", call. = FALSE)
  }

  return(invisible(TRUE))
}

.check_plot_positive_scalar <- function(x, argument, allow_zero = FALSE) {

  .check_plot_numeric(x, argument, check_length = 1, lower = 0, allow_null = FALSE)
  if (!allow_zero && x <= 0) {
    stop("'", argument, "' must be positive.", call. = FALSE)
  }

  return(invisible(TRUE))
}

context("update.brma")

mock_update_object <- function() {

  object <- brma.norm(
    yi        = c(0.10, 0.20, 0.05),
    sei       = c(0.05, 0.06, 0.07),
    only_data = TRUE,
    silent    = TRUE
  )

  object[["priors"]]       <- list(outcome = list())
  object[["fit"]]          <- structure(list(has_posterior = TRUE), class = "BayesTools_fit")
  object[["summary"]]      <- data.frame(Mean = 1, row.names = "mu")
  object[["coefficients"]] <- c(mu = 1)

  return(object)
}

test_that("update.brma updates slab labels without fitting", {

  object <- brma.norm(
    yi        = c(0.10, 0.20, 0.05),
    sei       = c(0.05, 0.06, 0.07),
    only_data = TRUE,
    silent    = TRUE
  )

  updated <- update(object, slab = c("A", "B", NA))

  expect_equal(updated[["data"]][["outcome"]][["slab"]], c("A", "B", NA))
  expect_true(attr(updated[["data"]], "slab"))
  expect_null(updated[["fit"]])
})

test_that("update.brma rejects structural update arguments", {

  object <- mock_update_object()

  expect_error(update(object, formula. = ~ x), "Updating formulas")
  expect_error(update(object, evaluate = FALSE), "evaluate = FALSE")
  expect_error(update(object, sample = 10), "Unused argument")
  expect_error(update(object, slab = c("A", "B")), "same as the fitted data")
})

test_that("update.brma extends one sample chunk and recomputes stored caches", {

  object <- mock_update_object()
  object[["loo"]]     <- list(estimate = "old-loo")
  object[["waic"]]    <- list(estimate = "old-waic")
  object[["marglik"]] <- "old-marglik"

  calls <- list()

  testthat::local_mocked_bindings(
    .fit = function(object, extend = FALSE) {

      calls[["extend"]]          <<- extend
      calls[["sample_extend"]]   <<- object[["autofit_control"]][["sample_extend"]]
      calls[["max_extend"]]      <<- object[["autofit_control"]][["max_extend"]]
      object[["fit"]][["mock"]] <- "extended"
      return(object[["fit"]])
    },
    .object_summary = function(object) {

      return(data.frame(Mean = 2, row.names = "mu"))
    },
    .object_coefficients = function(object) {

      return(c(mu = 2))
    },
    add_loo = function(object, unit = "estimate", ...) {

      calls[["loo_units"]] <<- c(calls[["loo_units"]], unit)
      if (is.null(object[["loo"]])) {
        object[["loo"]] <- list()
      }
      object[["loo"]][[unit]] <- paste0("new-loo-", unit)
      return(object)
    },
    add_waic = function(object, unit = "estimate", ...) {

      calls[["waic_units"]] <<- c(calls[["waic_units"]], unit)
      if (is.null(object[["waic"]])) {
        object[["waic"]] <- list()
      }
      object[["waic"]][[unit]] <- paste0("new-waic-", unit)
      return(object)
    },
    add_marglik = function(object, ...) {

      calls[["marglik"]] <<- TRUE
      object[["marglik"]] <- "new-marglik"
      return(object)
    },
    .package = "RoBMA"
  )

  updated <- update(
    object,
    sample_extend = 7,
    autofit_control = set_autofit_control(sample_extend = 3, max_extend = 4)
  )

  expect_true(calls[["extend"]])
  expect_equal(calls[["sample_extend"]], 7)
  expect_equal(calls[["max_extend"]], 1)
  expect_equal(updated[["autofit_control"]][["sample_extend"]], 7)
  expect_equal(updated[["autofit_control"]][["max_extend"]], 4)
  expect_equal(updated[["fit"]][["mock"]], "extended")
  expect_equal(updated[["summary"]][["Mean"]], 2)
  expect_equal(updated[["coefficients"]], c(mu = 2))
  expect_equal(updated[["loo"]][["estimate"]], "new-loo-estimate")
  expect_equal(updated[["waic"]][["estimate"]], "new-waic-estimate")
  expect_equal(updated[["marglik"]], "new-marglik")
  expect_equal(calls[["loo_units"]], "estimate")
  expect_equal(calls[["waic_units"]], "estimate")
  expect_true(calls[["marglik"]])
})

test_that("update.brma can drop stored caches after extension", {

  object <- mock_update_object()
  object[["loo"]]     <- list(estimate = "old-loo")
  object[["waic"]]    <- list(estimate = "old-waic")
  object[["marglik"]] <- "old-marglik"

  testthat::local_mocked_bindings(
    .fit = function(object, extend = FALSE) {

      object[["fit"]][["mock"]] <- "extended"
      return(object[["fit"]])
    },
    .object_summary = function(object) {

      return(data.frame(Mean = 2, row.names = "mu"))
    },
    .object_coefficients = function(object) {

      return(c(mu = 2))
    },
    .package = "RoBMA"
  )

  updated <- expect_warning(
    update(object, sample_extend = 1, recompute = "drop"),
    "Dropping cached"
  )

  expect_null(updated[["loo"]])
  expect_null(updated[["waic"]])
  expect_null(updated[["marglik"]])
})

test_that("update.brma rejects adversarial extension inputs before fitting", {

  object <- mock_update_object()

  testthat::local_mocked_bindings(
    .fit = function(object, extend = FALSE) {

      stop(".fit should not be called", call. = FALSE)
    },
    .package = "RoBMA"
  )

  expect_error(update(object, sample_extend = 0), "sample_extend")
  expect_error(update(object, sample_extend = 1.5), "sample_extend")
  expect_error(update(object, sample_extend = c(1, 2)), "sample_extend")
  expect_error(update(object, sample_extend = NA_integer_), "sample_extend")
  expect_error(update(object, sample_extend = 1, recompute = "none"), "all")
})

test_that("update.brma rejects extension when no fit is stored", {

  object <- mock_update_object()
  object[["fit"]] <- NULL

  expect_error(
    update(object, sample_extend = 1),
    "does not contain a fitted model"
  )
})

test_that("update.brma recomputes all named cached units", {

  object <- mock_update_object()
  object[["loo"]]  <- list(estimate = "old-loo-estimate", cluster = "old-loo-cluster")
  object[["waic"]] <- list(estimate = "old-waic-estimate", cluster = "old-waic-cluster")

  calls <- list()

  testthat::local_mocked_bindings(
    .fit = function(object, extend = FALSE) {

      return(object[["fit"]])
    },
    .object_summary = function(object) {

      return(data.frame(Mean = 2, row.names = "mu"))
    },
    .object_coefficients = function(object) {

      return(c(mu = 2))
    },
    add_loo = function(object, unit = "estimate", ...) {

      calls[["loo_units"]] <<- c(calls[["loo_units"]], unit)
      if (is.null(object[["loo"]])) {
        object[["loo"]] <- list()
      }
      object[["loo"]][[unit]] <- paste0("new-loo-", unit)
      return(object)
    },
    add_waic = function(object, unit = "estimate", ...) {

      calls[["waic_units"]] <<- c(calls[["waic_units"]], unit)
      if (is.null(object[["waic"]])) {
        object[["waic"]] <- list()
      }
      object[["waic"]][[unit]] <- paste0("new-waic-", unit)
      return(object)
    },
    .package = "RoBMA"
  )

  updated <- update(object, sample_extend = 1)

  expect_equal(calls[["loo_units"]], c("estimate", "cluster"))
  expect_equal(calls[["waic_units"]], c("estimate", "cluster"))
  expect_equal(updated[["loo"]][["cluster"]], "new-loo-cluster")
  expect_equal(updated[["waic"]][["cluster"]], "new-waic-cluster")
})

test_that("update.brma rejects malformed multi-unit cache names", {

  object <- mock_update_object()
  object[["loo"]] <- list("old-loo-1", "old-loo-2")

  testthat::local_mocked_bindings(
    .fit = function(object, extend = FALSE) {

      return(object[["fit"]])
    },
    .object_summary = function(object) {

      return(data.frame(Mean = 2, row.names = "mu"))
    },
    .object_coefficients = function(object) {

      return(c(mu = 2))
    },
    .package = "RoBMA"
  )

  expect_error(
    update(object, sample_extend = 1),
    "must be named by unit"
  )
})

test_that("update.brma does not recompute marginal likelihood for RoBMA objects", {

  object <- mock_update_object()
  class(object) <- c("RoBMA", "brma")
  object[["marglik"]] <- "stale-marglik"

  testthat::local_mocked_bindings(
    .fit = function(object, extend = FALSE) {

      return(object[["fit"]])
    },
    .object_summary = function(object) {

      return(data.frame(Mean = 2, row.names = "mu"))
    },
    .object_coefficients = function(object) {

      return(c(mu = 2))
    },
    add_marglik = function(object, ...) {

      stop("add_marglik should not be called", call. = FALSE)
    },
    .package = "RoBMA"
  )

  updated <- update(object, sample_extend = 1)

  expect_null(updated[["marglik"]])
  expect_s3_class(updated, "RoBMA")
})

context("Autocompute options")

source(testthat::test_path("common-functions.R"))
skip_on_cran()
skip_if_not_installed("metadat")
skip_if_not_installed("metafor")

test_that("autocompute options are applied for brma()", {

  old_options <- list(
    autocompute.loo     = RoBMA.get_option("autocompute.loo"),
    autocompute.waic    = RoBMA.get_option("autocompute.waic"),
    autocompute.marglik = RoBMA.get_option("autocompute.marglik")
  )
  on.exit(do.call(RoBMA.options, old_options), add = TRUE)

  # Using a very simple model to keep the test-01 fit cheap.
  data(dat.bcg, package = "metadat")
  dat <- metafor::escalc(
    measure = "RR",
    ai      = tpos,
    bi      = tneg,
    ci      = cpos,
    di      = cneg,
    data    = dat.bcg
  )
  dat <- dat[1:3, ]

  RoBMA.options(
    autocompute.loo     = FALSE,
    autocompute.waic    = FALSE,
    autocompute.marglik = FALSE
  )
  fit_default <- suppressWarnings(brma(
    yi      = yi,
    vi      = vi,
    data    = dat,
    measure = "RR",
    seed    = 1,
    silent  = TRUE,
    chains  = 2,
    sample  = 250,
    burnin  = 50,
    adapt   = 100
  ))

  expect_null(fit_default$loo)
  expect_null(fit_default$waic)
  expect_null(fit_default$marglik)

  RoBMA.options(
    autocompute.loo     = TRUE,
    autocompute.waic    = TRUE,
    autocompute.marglik = TRUE
  )
  fit_auto <- suppressWarnings(brma(
    yi      = yi,
    vi      = vi,
    data    = dat,
    measure = "RR",
    seed    = 1,
    silent  = TRUE,
    chains  = 2,
    sample  = 250,
    burnin  = 50,
    adapt   = 100
  ))

  expect_s3_class(fit_auto$loo$estimate, "loo")
  expect_s3_class(fit_auto$waic$estimate, "waic")
  expect_s3_class(fit_auto$marglik, "bridge")
})

test_that("product-space constructors autocompute LOO and WAIC only", {

  old_options <- list(
    autocompute.loo     = RoBMA.get_option("autocompute.loo"),
    autocompute.waic    = RoBMA.get_option("autocompute.waic"),
    autocompute.marglik = RoBMA.get_option("autocompute.marglik")
  )
  on.exit(do.call(RoBMA.options, old_options), add = TRUE)

  data(dat.bcg, package = "metadat")
  dat <- metafor::escalc(
    measure = "RR",
    ai      = tpos,
    bi      = tneg,
    ci      = cpos,
    di      = cneg,
    data    = dat.bcg
  )
  dat <- dat[1:3, ]

  RoBMA.options(
    autocompute.loo     = TRUE,
    autocompute.waic    = TRUE,
    autocompute.marglik = TRUE
  )
  fit_auto <- suppressWarnings(BMA.norm(
    yi      = yi,
    vi      = vi,
    data    = dat,
    measure = "RR",
    seed    = 1,
    silent  = TRUE,
    chains  = 2,
    sample  = 250,
    burnin  = 50,
    adapt   = 100
  ))

  expect_s3_class(fit_auto$loo$estimate, "loo")
  expect_s3_class(fit_auto$waic$estimate, "waic")
  expect_null(fit_auto$marglik)
})

context("Options and brma_samples metadata")

test_that("RoBMA options expose only public options", {

  old_options <- RoBMA.options()
  on.exit(do.call(RoBMA.options, old_options), add = TRUE)

  expect_true("silent" %in% names(old_options))
  expect_false("RoBMA_version" %in% names(old_options))
  expect_false("module_location" %in% names(old_options))

  updated <- RoBMA.options(silent = TRUE)
  expect_true(updated[["silent"]])
  expect_true(RoBMA.get_option("silent"))

  expect_error(RoBMA.options(RoBMA_version = "bad"), "Unmatched or ambiguous option")
  expect_error(RoBMA.get_option("RoBMA_version"), "Unmatched or ambiguous option")
  expect_error(RoBMA.options(TRUE), "All options must be named")
  expect_error(RoBMA.options(silent = 1), "must be TRUE or FALSE")
  expect_error(RoBMA.options(max_cores = 0), "must be an integer")
  expect_error(RoBMA.options(default_bias_PET.scale = expression(stop("boom"))), "finite number")

  updated <- RoBMA.options(max_cores = 1L)
  expect_equal(updated[["max_cores"]], 1L)
})


test_that("convergence checks expose only active thresholds", {

  checks <- set_convergence_checks()

  expect_true(all(c("max_Rhat", "min_ESS", "max_error", "max_SD_error") %in% names(checks)))
  expect_false("remove_failed" %in% names(checks))
  expect_false("balance_probability" %in% names(checks))
  expect_error(set_convergence_checks(remove_failed = TRUE), "unused argument")
  expect_error(set_convergence_checks(balance_probability = FALSE), "unused argument")
})


test_that("brma_samples metadata must match sample rows", {

  expect_error(
    RoBMA:::.new_brma_samples(
      samples  = matrix(seq_len(3), ncol = 1),
      n_chains = 2,
      n_iter   = 2,
      title    = "Invalid"
    ),
    "Invalid brma_samples chain metadata"
  )

  samples <- RoBMA:::.new_brma_samples(
    samples  = matrix(seq_len(3), ncol = 1),
    n_chains = 1,
    n_iter   = 3,
    title    = "Valid"
  )
  attr(samples, "nchains") <- 2
  attr(samples, "niter")   <- 2

  expect_error(
    RoBMA:::.brma_samples_to_mcmc.list(samples),
    "Invalid brma_samples chain metadata"
  )
})


test_that("brma_samples chain info preserves only balanced chains", {

  balanced_fit <- list(
    mcmc = list(
      matrix(seq_len(4), nrow = 2),
      matrix(seq_len(4), nrow = 2)
    )
  )
  unequal_fit <- list(
    mcmc = list(
      matrix(seq_len(2), nrow = 1),
      matrix(seq_len(4), nrow = 2)
    )
  )

  expect_equal(
    RoBMA:::.brma_samples_chain_info(balanced_fit, n_samples = 4),
    list(n_chains = 2L, n_iter = 2L)
  )
  expect_equal(
    RoBMA:::.brma_samples_chain_info(unequal_fit, n_samples = 3),
    list(n_chains = 1L, n_iter = 3L)
  )
})

context("Prior input handling for RoBMA (mixture priors)")

skip_on_cran()
source(testthat::test_path("helper-contracts.R"))

# Helper function to get is_null logical vector from mixture prior
# BayesTools stores this as "components" attribute with values "null" and "alternative"
.get_is_null <- function(prior) {
  components <- attr(prior, "components")
  if (is.null(components)) return(NULL)
  return(components == "null")
}

expect_mixture_component_counts <- function(prior, n_null, n_alt) {

  expect_true(BayesTools::is.prior.mixture(prior))

  is_null <- .get_is_null(prior)
  expect_equal(sum(is_null), n_null)
  expect_equal(sum(!is_null), n_alt)
}

expect_bias_mixture_shape <- function(prior, n_null, n_wf, n_pet, n_peese) {

  expect_mixture_component_counts(prior, n_null = n_null, n_alt = n_wf + n_pet + n_peese)

  has_wf    <- sapply(prior, BayesTools::is.prior.weightfunction)
  has_pet   <- sapply(prior, BayesTools::is.prior.PET)
  has_peese <- sapply(prior, BayesTools::is.prior.PEESE)

  expect_equal(sum(has_wf), n_wf)
  expect_equal(sum(has_pet), n_pet)
  expect_equal(sum(has_peese), n_peese)
}

# Test data for prior specification tests
test_data <- data.frame(
  effect     = c(0.10, 0.25, 0.15, 0.30, 0.05),
  std_err    = sqrt(c(0.04, 0.06, 0.05, 0.08, 0.03)),
  n          = c(50L, 75L, 60L, 40L, 90L),
  mod_cont   = c(1.5, 2.3, 1.8, 3.1, 0.9),
  mod_factor = factor(c("A", "B", "A", "B", "A")),
  scale_var  = c(0.5, 1.0, 0.8, 1.2, 0.6),
  cluster    = c("g1", "g1", "g2", "g2", "g3"),
  stringsAsFactors = FALSE
)


# ============================================================================
# Tests for mixture prior creation for effect and heterogeneity
# ============================================================================

test_that("Default priors create effect and heterogeneity mixtures", {

  result <- RoBMA(
    yi = effect, sei = std_err, data = test_data,
    measure = "SMD", only_priors = TRUE
  )[["priors"]]

  expect_mixture_component_counts(result$outcome$mu, n_null = 1, n_alt = 1)
  expect_true(.get_is_null(result$outcome$mu)[1])
  expect_false(.get_is_null(result$outcome$mu)[2])
  expect_true(BayesTools::is.prior.point(result$outcome$mu[[1]]))
  expect_equal(mean(result$outcome$mu[[1]]), 0)

  expected_effect_sd <- sqrt(2) * RoBMA.get_option("default_UISD.effect")
  expect_equal(result$outcome$mu[[2]]$parameters$sd, expected_effect_sd)

  expect_mixture_component_counts(result$outcome$tau, n_null = 1, n_alt = 1)
  expect_true(.get_is_null(result$outcome$tau)[1])
  expect_false(.get_is_null(result$outcome$tau)[2])
  expect_true(BayesTools::is.prior.point(result$outcome$tau[[1]]))
  expect_equal(mean(result$outcome$tau[[1]]), 0)
  expect_equal(result$outcome$tau[[2]]$truncation$lower, 0)

  expected_het_sd <- sqrt(2) * RoBMA.get_option("default_UISD.heterogeneity")
  expect_equal(result$outcome$tau[[2]]$parameters$sd, expected_het_sd)
})


test_that("Mixture priors scale with different measures", {

  measures_uisd <- list(SMD = sqrt(2), ZCOR = 1, OR = 2, RR = 2, HR = 2)

  for (measure in names(measures_uisd)) {
    result <- RoBMA(
      yi = effect, sei = std_err, data = test_data,
      measure = measure, only_priors = TRUE
    )[["priors"]]

    expected_effect_sd <- measures_uisd[[measure]] * RoBMA.get_option("default_UISD.effect")
    expected_het_sd    <- measures_uisd[[measure]] * RoBMA.get_option("default_UISD.heterogeneity")

    # Check alternative component (second in mixture)
    expect_equal(result$outcome$mu[[2]]$parameters$sd, expected_effect_sd)
    expect_equal(result$outcome$tau[[2]]$parameters$sd, expected_het_sd)
  }
})


test_that("rescale_priors applies to alternative components in mixture", {

  result_base <- RoBMA(
    yi = effect, sei = std_err, data = test_data,
    measure = "SMD", only_priors = TRUE
  )[["priors"]]

  for (scale_factor in c(0.5, 2)) {
    result_rescaled <- RoBMA(
      yi = effect, sei = std_err, data = test_data,
      measure = "SMD", rescale_priors = scale_factor, only_priors = TRUE
    )[["priors"]]

    # Alternative component (index 2) should be rescaled
    expect_equal(
      result_rescaled$outcome$mu[[2]]$parameters$sd,
      result_base$outcome$mu[[2]]$parameters$sd * scale_factor
    )
    expect_equal(
      result_rescaled$outcome$tau[[2]]$parameters$sd,
      result_base$outcome$tau[[2]]$parameters$sd * scale_factor
    )
  }
})


# ============================================================================
# Tests for custom null/alternative prior lists
# ============================================================================

test_that("Custom prior_effect replaces alternative component", {

  custom_effect <- BayesTools::prior("normal", parameters = list(mean = 0, sd = 0.5))

  result <- RoBMA(
    yi = effect, sei = std_err, data = test_data,
    prior_effect = custom_effect, measure = "SMD", only_priors = TRUE
  )[["priors"]]

  expect_true(BayesTools::is.prior.mixture(result$outcome$mu))
  expect_equal(result$outcome$mu[[2]]$parameters$sd, 0.5)
})


test_that("Custom prior_effect_null replaces null component", {

  custom_null <- BayesTools::prior("spike", parameters = list(location = 0.1))

  result <- RoBMA(
    yi = effect, sei = std_err, data = test_data,
    prior_effect_null = custom_null, measure = "SMD", only_priors = TRUE
  )[["priors"]]

  expect_true(BayesTools::is.prior.mixture(result$outcome$mu))
  expect_true(BayesTools::is.prior.point(result$outcome$mu[[1]]))
  expect_equal(mean(result$outcome$mu[[1]]), 0.1)
})


test_that("List of priors creates mixture with multiple components", {

  alt_priors <- list(
    BayesTools::prior("normal", parameters = list(mean = 0, sd = 0.5)),
    BayesTools::prior("normal", parameters = list(mean = 0, sd = 1.0))
  )

  result <- RoBMA(
    yi = effect, sei = std_err, data = test_data,
    prior_effect = alt_priors, measure = "SMD", only_priors = TRUE
  )[["priors"]]

  expect_true(BayesTools::is.prior.mixture(result$outcome$mu))
  expect_equal(length(result$outcome$mu), 3)  # 1 null + 2 alternatives

  # Check is_null attribute
  is_null <- .get_is_null(result$outcome$mu)
  expect_equal(sum(is_null), 1)   # 1 null
  expect_equal(sum(!is_null), 2)  # 2 alternatives
})


test_that("NULL/FALSE omit effect mixture components", {

  result_null <- RoBMA(
    yi = effect, sei = std_err, data = test_data,
    prior_effect_null = NULL, measure = "SMD", only_priors = TRUE
  )[["priors"]]

  result_alt <- RoBMA(
    yi = effect, sei = std_err, data = test_data,
    prior_effect = NULL, measure = "SMD", only_priors = TRUE
  )[["priors"]]

  result_false <- RoBMA(
    yi = effect, sei = std_err, data = test_data,
    prior_effect_null = FALSE, measure = "SMD", only_priors = TRUE
  )[["priors"]]

  expect_mixture_component_counts(result_null$outcome$mu, n_null = 0, n_alt = 1)
  expect_mixture_component_counts(result_alt$outcome$mu, n_null = 1, n_alt = 0)
  expect_equal(length(result_null$outcome$mu), length(result_false$outcome$mu))
  expect_equal(.get_is_null(result_null$outcome$mu), .get_is_null(result_false$outcome$mu))
})


# ============================================================================
# Tests for publication bias model types
# ============================================================================

test_that("model_type presets create expected bias mixtures", {

  preset_cases <- list(
    list(model_type = "2w",   n_null = 1, n_wf = 2, n_pet = 0, n_peese = 0),
    list(model_type = "6w",   n_null = 1, n_wf = 6, n_pet = 0, n_peese = 0),
    list(model_type = "PP",   n_null = 1, n_wf = 0, n_pet = 1, n_peese = 1),
    list(model_type = "PSMA", n_null = 1, n_wf = 6, n_pet = 1, n_peese = 1)
  )

  for (case in preset_cases) {
    result <- RoBMA(
      yi          = effect,
      sei         = std_err,
      data        = test_data,
      measure     = "SMD",
      model_type  = case$model_type,
      only_priors = TRUE
    )[["priors"]]

    expect_bias_mixture_shape(
      result$outcome$bias,
      n_null  = case$n_null,
      n_wf    = case$n_wf,
      n_pet   = case$n_pet,
      n_peese = case$n_peese
    )
  }
})


test_that("Custom prior_bias and prior_bias_null override model_type", {

  custom_bias <- BayesTools::prior_weightfunction(
    side    = "two-sided",
    steps   = c(0.05),
    weights = BayesTools::wf_cumulative(c(1, 1))
  )

  result <- RoBMA(
    yi = effect, sei = std_err, data = test_data,
    prior_bias = custom_bias, measure = "SMD", only_priors = TRUE
  )[["priors"]]

  expect_true(BayesTools::is.prior.mixture(result$outcome$bias))

  # Should have the custom weight function as alternative
  has_wf <- sapply(result$outcome$bias, function(p) BayesTools::is.prior.weightfunction(p))
  expect_equal(sum(has_wf), 1)
})


test_that("prior_bias_null = NULL omits null bias hypothesis", {

  custom_bias <- BayesTools::prior_weightfunction(
    side    = "two-sided",
    steps   = c(0.05),
    weights = BayesTools::wf_cumulative(c(1, 1))
  )

  result <- RoBMA(
    yi = effect, sei = std_err, data = test_data,
    prior_bias = custom_bias, prior_bias_null = NULL,
    measure = "SMD", only_priors = TRUE
  )[["priors"]]

  is_null <- .get_is_null(result$outcome$bias)
  expect_equal(sum(is_null), 0)  # no null component
})


test_that("NULL/FALSE omit bias mixture components", {

  result_alt_null <- RoBMA(
    yi          = effect,
    sei         = std_err,
    data        = test_data,
    prior_bias  = NULL,
    measure     = "SMD",
    only_priors = TRUE
  )[["priors"]]

  result_alt_false <- RoBMA(
    yi          = effect,
    sei         = std_err,
    data        = test_data,
    prior_bias  = FALSE,
    measure     = "SMD",
    only_priors = TRUE
  )[["priors"]]

  result_null_null <- RoBMA(
    yi              = effect,
    sei             = std_err,
    data            = test_data,
    measure         = "SMD",
    model_type      = "PP",
    prior_bias_null = NULL,
    only_priors     = TRUE
  )[["priors"]]

  result_null_false <- RoBMA(
    yi              = effect,
    sei             = std_err,
    data            = test_data,
    measure         = "SMD",
    model_type      = "PP",
    prior_bias_null = FALSE,
    only_priors     = TRUE
  )[["priors"]]

  expect_true(BayesTools::is.prior.none(result_alt_null$outcome$bias))
  expect_true(BayesTools::is.prior.none(result_alt_false$outcome$bias))

  expect_bias_mixture_shape(result_null_null$outcome$bias, n_null = 0, n_wf = 0, n_pet = 1, n_peese = 1)
  expect_equal(length(result_null_null$outcome$bias), length(result_null_false$outcome$bias))
  expect_equal(.get_is_null(result_null_null$outcome$bias), .get_is_null(result_null_false$outcome$bias))
})


test_that("Bias presets only require UISD when PEESE is present", {

  custom_effect <- BayesTools::prior("normal", parameters = list(mean = 0, sd = 0.5))
  custom_het    <- BayesTools::prior("normal", parameters = list(mean = 0, sd = 0.3), truncation = list(0, Inf))

  expect_no_error(
    RoBMA(
      yi = effect, sei = std_err, data = test_data,
      measure = "GEN", model_type = "6w",
      prior_effect = custom_effect, prior_heterogeneity = custom_het,
      only_priors = TRUE
    )
  )
})


test_that("Partial bias customization respects model_type alternatives", {

  result <- RoBMA(
    yi = effect, sei = std_err, data = test_data,
    measure = "SMD", model_type = "PP",
    prior_bias_null = NULL, only_priors = TRUE
  )[["priors"]]

  has_wf    <- sapply(result$outcome$bias, BayesTools::is.prior.weightfunction)
  has_pet   <- sapply(result$outcome$bias, BayesTools::is.prior.PET)
  has_peese <- sapply(result$outcome$bias, BayesTools::is.prior.PEESE)

  expect_equal(sum(has_wf), 0)
  expect_equal(sum(has_pet), 1)
  expect_equal(sum(has_peese), 1)
})


test_that("Empty top-level mixtures are rejected", {

  expect_error_cases(list(
    list(
      label  = "empty effect mixture",
      expr   = quote(RoBMA(
        yi = effect, sei = std_err, data = test_data,
        measure = "SMD",
        prior_effect = NULL, prior_effect_null = NULL,
        only_priors = TRUE
      )),
      regexp = "effect"
    ),
    list(
      label  = "empty bias mixture",
      expr   = quote(RoBMA(
        yi = effect, sei = std_err, data = test_data,
        measure = "SMD",
        prior_bias = NULL, prior_bias_null = NULL,
        only_priors = TRUE
      )),
      regexp = "bias"
    )
  ))
})


# ============================================================================
# Tests for heterogeneity allocation mixture (multilevel models)
# ============================================================================

test_that("Heterogeneity allocation creates mixture with cluster", {

  result <- RoBMA(
    yi = effect, sei = std_err, cluster = cluster, data = test_data,
    measure = "SMD", only_priors = TRUE
  )[["priors"]]

  expect_true("rho" %in% names(result$outcome))
  expect_true(BayesTools::is.prior.mixture(result$outcome$rho))

  # Default: no null hypothesis for rho (only alternative Beta(1,1))
  is_null <- .get_is_null(result$outcome$rho)
  expect_equal(sum(is_null), 0)

  # Check alternative is Beta(1,1)
  expect_equal(result$outcome$rho[[1]]$distribution, "beta")
  expect_equal(result$outcome$rho[[1]]$parameters$alpha, 1)
  expect_equal(result$outcome$rho[[1]]$parameters$beta, 1)
})


test_that("Custom prior_heterogeneity_allocation is accepted", {

  custom_prior <- BayesTools::prior("beta", parameters = list(alpha = 2, beta = 2))

  result <- RoBMA(
    yi = effect, sei = std_err, cluster = cluster, data = test_data,
    prior_heterogeneity_allocation = custom_prior,
    measure = "SMD", only_priors = TRUE
  )[["priors"]]

  expect_equal(result$outcome$rho[[1]]$parameters$alpha, 2)
  expect_equal(result$outcome$rho[[1]]$parameters$beta, 2)
})


test_that("prior_heterogeneity_allocation_null adds null hypothesis", {

  null_prior <- BayesTools::prior("spike", parameters = list(location = 0.5))

  result <- RoBMA(
    yi = effect, sei = std_err, cluster = cluster, data = test_data,
    prior_heterogeneity_allocation_null = null_prior,
    measure = "SMD", only_priors = TRUE
  )[["priors"]]

  is_null <- .get_is_null(result$outcome$rho)
  expect_equal(sum(is_null), 1)  # now has null component
  expect_true(BayesTools::is.prior.point(result$outcome$rho[[1]]))
})


# ============================================================================
# Tests for moderator and scale priors with mixtures
# ============================================================================

test_that("Moderator priors create mixture with intercept from effect prior", {

  result <- RoBMA(
    yi = effect, sei = std_err, mods = ~ mod_cont,
    data = test_data, measure = "SMD", only_priors = TRUE
  )[["priors"]]

  expect_true(!is.null(result$mods))
  expect_true("intercept" %in% names(result$mods))
  expect_null(result$outcome$mu)  # mu forwarded to intercept

  # Intercept should be a mixture prior
  expect_true(BayesTools::is.prior.mixture(result$mods$intercept))
})


test_that("No-intercept RoBMA moderator formulas use fixed-zero intercept", {

  result <- RoBMA(
    yi = effect, sei = std_err, mods = ~ 0 + mod_cont,
    data = test_data, measure = "SMD", only_priors = TRUE
  )[["priors"]]

  expect_null(result$outcome$mu)
  expect_true(BayesTools::is.prior.point(result$mods$intercept))
  expect_equal(mean(result$mods$intercept), 0)
  expect_true(BayesTools::is.prior.mixture(result$mods$mod_cont))
})


test_that("Moderator term priors include null and alternative", {

  result <- RoBMA(
    yi = effect, sei = std_err, mods = ~ mod_cont,
    data = test_data, measure = "SMD", only_priors = TRUE
  )[["priors"]]

  # mod_cont should have a mixture prior
  expect_true(BayesTools::is.prior.mixture(result$mods$mod_cont))

  is_null <- .get_is_null(result$mods$mod_cont)
  expect_true(any(is_null))   # has null component
  expect_true(any(!is_null))  # has alternative component
})


test_that("Custom prior_mods replaces alternative for terms", {

  custom_mod <- list(mod_cont = BayesTools::prior("normal", parameters = list(mean = 0, sd = 0.3)))

  result <- RoBMA(
    yi = effect, sei = std_err, mods = ~ mod_cont,
    prior_mods = custom_mod, data = test_data, measure = "SMD", only_priors = TRUE
  )[["priors"]]

  # Alternative component should have custom prior
  is_null <- .get_is_null(result$mods$mod_cont)
  alt_idx <- which(!is_null)[1]
  expect_equal(result$mods$mod_cont[[alt_idx]]$parameters$sd, 0.3)
})


test_that("Custom prior_mods_null replaces null for terms", {

  custom_null <- list(mod_cont = BayesTools::prior("spike", parameters = list(location = 0.1)))

  result <- RoBMA(
    yi = effect, sei = std_err, mods = ~ mod_cont,
    prior_mods_null = custom_null, data = test_data, measure = "SMD", only_priors = TRUE
  )[["priors"]]

  # Null component should have custom prior
  is_null <- .get_is_null(result$mods$mod_cont)
  null_idx <- which(is_null)[1]
  expect_true(BayesTools::is.prior.point(result$mods$mod_cont[[null_idx]]))
  expect_equal(mean(result$mods$mod_cont[[null_idx]]), 0.1)
})


# ============================================================================
# Tests for informed priors with mixtures
# ============================================================================

test_that("Informed priors are accepted with mixture structure", {

  result <- RoBMA(
    yi = effect, sei = std_err, data = test_data,
    measure = "OR", prior_informed_field = "medicine", only_priors = TRUE
  )[["priors"]]

  # Check mixture structure preserved
  expect_true(BayesTools::is.prior.mixture(result$outcome$mu))
  expect_true(BayesTools::is.prior.mixture(result$outcome$tau))

  # Alternative component should be informed prior - compare key properties
  expected_mu  <- BayesTools::prior_informed(name = "cochrane", parameter = "effect", type = "logOR")
  expected_tau <- BayesTools::prior_informed(name = "cochrane", parameter = "heterogeneity", type = "logOR")

  expect_equal(result$outcome$mu[[2]]$distribution, expected_mu$distribution)
  expect_equal(result$outcome$mu[[2]]$parameters, expected_mu$parameters)
  expect_equal(result$outcome$tau[[2]]$distribution, expected_tau$distribution)
  expect_equal(result$outcome$tau[[2]]$parameters, expected_tau$parameters)
})


test_that("Informed priors with rescaling apply to alternative", {

  result_base <- RoBMA(
    yi = effect, sei = std_err, data = test_data,
    measure = "SMD", prior_informed_field = "medicine", only_priors = TRUE
  )[["priors"]]

  result_rescaled <- RoBMA(
    yi = effect, sei = std_err, data = test_data,
    measure = "SMD", prior_informed_field = "medicine",
    rescale_priors = 2, only_priors = TRUE
  )[["priors"]]

  scale_param <- if ("sd" %in% names(result_base$outcome$mu[[2]]$parameters)) "sd" else "scale"
  expect_equal(
    result_rescaled$outcome$mu[[2]]$parameters[[scale_param]],
    result_base$outcome$mu[[2]]$parameters[[scale_param]] * 2
  )
})


# ============================================================================
# Tests for error conditions
# ============================================================================

test_that("Invalid model_type is rejected", {

  expect_error(
    RoBMA(
      yi = effect, sei = std_err, data = test_data, measure = "SMD",
      model_type = "invalid", only_priors = TRUE
    ),
    regexp = "model_type"
  )
})


test_that("Conflicting prior specifications are rejected", {

  # Both UISD and informed priors
  expect_error(
    RoBMA(
      yi = effect, sei = std_err, data = test_data, measure = "SMD",
      prior_unit_information_sd = 1.5, prior_informed_field = "medicine", only_priors = TRUE
    ),
    regexp = "prior_unit_information_sd|prior_informed_field|mutually exclusive"
  )
})


test_that("Invalid prior types are rejected", {

  # Non-prior object
  expect_error(
    RoBMA(
      yi = effect, sei = std_err, data = test_data, measure = "SMD",
      prior_effect = "not a prior", only_priors = TRUE
    ),
    regexp = "prior_effect"
  )
})


test_that("Missing measure is rejected before GEN UISD setup", {

  expect_error(
    RoBMA(yi = effect, sei = std_err, data = test_data, only_priors = TRUE),
    regexp = "requires explicit 'measure'"
  )
})


test_that("GEN measure requires ni or prior_unit_information_sd for RoBMA", {

  expect_error(
    RoBMA(yi = effect, sei = std_err, data = test_data, measure = "GEN", only_priors = TRUE),
    regexp = "ni|unit_information_sd|UISD|Sample size"
  )

  # Should work with ni
  expect_no_error(
    RoBMA(yi = effect, sei = std_err, ni = n, data = test_data, measure = "GEN", only_priors = TRUE)
  )

  # Should work with prior_unit_information_sd
  expect_no_error(
    RoBMA(yi = effect, sei = std_err, data = test_data, measure = "GEN", prior_unit_information_sd = 1.5, only_priors = TRUE)
  )
})


# ============================================================================
# Tests for scale priors (simple priors, not mixtures for terms)
# ============================================================================

test_that("Scale priors create mixture with intercept from heterogeneity prior", {

  result <- suppressWarnings(RoBMA(
    yi = effect, sei = std_err, scale = ~ scale_var,
    data = test_data, measure = "SMD", only_priors = TRUE
  )[["priors"]])

  expect_true(!is.null(result$scale))
  expect_true("intercept" %in% names(result$scale))
  expect_null(result$outcome$tau)  # tau forwarded to intercept

  # Intercept should be a mixture prior
  expect_true(BayesTools::is.prior.mixture(result$scale$intercept))

  # Note: spike(0) null is removed for scale since log(0) is undefined
  # Check that remaining components are truncated (positive heterogeneity)
  for (i in seq_along(result$scale$intercept)) {
    if (!BayesTools::is.prior.point(result$scale$intercept[[i]])) {
      expect_equal(result$scale$intercept[[i]]$truncation$lower, 0)
    }
  }
})


test_that("Scale term priors are mixture priors (same as mods)", {

  # Scale terms use .assign_prior_list.terms_mixture() like mods
  # returning mixture priors for model-averaging

  result <- suppressWarnings(RoBMA(
    yi = effect, sei = std_err, scale = ~ scale_var,
    data = test_data, measure = "SMD", only_priors = TRUE
  )[["priors"]])

  # scale_var should be a mixture prior (like mods terms)
  expect_true(BayesTools::is.prior.mixture(result$scale$scale_var))
  # Check it has both null and alternative components
  is_null <- .get_is_null(result$scale$scale_var)
  expect_true(any(is_null))   # has null component(s)
  expect_true(any(!is_null))  # has alternative component(s)
})


test_that("Custom prior_scale replaces alternative for scale terms", {

  custom_scale <- list(scale_var = BayesTools::prior("normal", parameters = list(mean = 0, sd = 0.4)))

  result <- suppressWarnings( RoBMA(
    yi = effect, sei = std_err, scale = ~ scale_var,
    prior_scale = custom_scale, data = test_data, measure = "SMD", only_priors = TRUE
  )[["priors"]])

  # Result should be a mixture
  expect_true(BayesTools::is.prior.mixture(result$scale$scale_var))
  # Custom prior should be in the alternative component
  is_null <- .get_is_null(result$scale$scale_var)
  alt_indices <- which(!is_null)
  expect_equal(result$scale$scale_var[[alt_indices[1]]]$parameters$sd, 0.4)
})


# ============================================================================
# Tests for custom heterogeneity prior lists
# ============================================================================

test_that("Custom prior_heterogeneity replaces alternative component", {

  custom_het <- BayesTools::prior("normal", parameters = list(mean = 0, sd = 0.3), truncation = list(0, Inf))

  result <- RoBMA(
    yi = effect, sei = std_err, data = test_data,
    prior_heterogeneity = custom_het, measure = "SMD", only_priors = TRUE
  )[["priors"]]

  expect_true(BayesTools::is.prior.mixture(result$outcome$tau))
  expect_equal(result$outcome$tau[[2]]$parameters$sd, 0.3)
})


test_that("Custom prior_heterogeneity_null replaces null component", {

  custom_null <- BayesTools::prior("spike", parameters = list(location = 0.1))

  result <- RoBMA(
    yi = effect, sei = std_err, data = test_data,
    prior_heterogeneity_null = custom_null, measure = "SMD", only_priors = TRUE
  )[["priors"]]

  expect_true(BayesTools::is.prior.mixture(result$outcome$tau))
  expect_true(BayesTools::is.prior.point(result$outcome$tau[[1]]))
  expect_equal(mean(result$outcome$tau[[1]]), 0.1)
})


test_that("List of heterogeneity priors creates mixture with multiple alternatives", {

  alt_priors <- list(
    BayesTools::prior("normal", parameters = list(mean = 0, sd = 0.2), truncation = list(0, Inf)),
    BayesTools::prior("normal", parameters = list(mean = 0, sd = 0.5), truncation = list(0, Inf))
  )

  result <- RoBMA(
    yi = effect, sei = std_err, data = test_data,
    prior_heterogeneity = alt_priors, measure = "SMD", only_priors = TRUE
  )[["priors"]]

  expect_true(BayesTools::is.prior.mixture(result$outcome$tau))
  expect_equal(length(result$outcome$tau), 3)  # 1 null + 2 alternatives

  is_null <- .get_is_null(result$outcome$tau)
  expect_equal(sum(is_null), 1)
  expect_equal(sum(!is_null), 2)
})


test_that("prior_heterogeneity = NULL creates null-only mixture", {

  result <- RoBMA(
    yi = effect, sei = std_err, data = test_data,
    prior_heterogeneity = NULL, measure = "SMD", only_priors = TRUE
  )[["priors"]]

  expect_true(BayesTools::is.prior.mixture(result$outcome$tau))
  expect_equal(length(result$outcome$tau), 1)

  is_null <- .get_is_null(result$outcome$tau)
  expect_equal(sum(is_null), 1)
})


# ============================================================================
# Tests for factor moderator priors
# ============================================================================

test_that("Factor moderator creates mixture prior", {

  result <- RoBMA(
    yi = effect, sei = std_err, mods = ~ mod_factor,
    data = test_data, measure = "SMD", only_priors = TRUE
  )[["priors"]]

  # Factor moderator is stored as mod_factor (not individual levels)
  expect_true("mod_factor" %in% names(result$mods))
  expect_true(BayesTools::is.prior.mixture(result$mods$mod_factor))

  is_null <- .get_is_null(result$mods$mod_factor)
  expect_true(any(is_null))
  expect_true(any(!is_null))
})


test_that("Multiple moderators each get mixture priors", {

  result <- RoBMA(
    yi = effect, sei = std_err, mods = ~ mod_cont + mod_factor,
    data = test_data, measure = "SMD", only_priors = TRUE
  )[["priors"]]

  # Both moderators should have mixture priors
  expect_true(BayesTools::is.prior.mixture(result$mods$mod_cont))
  expect_true(BayesTools::is.prior.mixture(result$mods$mod_factor))
})


# ============================================================================
# Tests for mixed moderator specifications (some with null, some without)
# ============================================================================

test_that("Different moderators can have different null specifications", {

  # mod_cont: omit null (alternative only)
  # mod_factor: keep default (null + alternative)
  result <- RoBMA(
    yi = effect, sei = std_err, mods = ~ mod_cont + mod_factor,
    prior_mods_null = list(mod_cont = NULL),  # omit null for mod_cont only
    data = test_data, measure = "SMD", only_priors = TRUE
  )[["priors"]]

  # mod_cont should have no null component
  is_null_cont <- .get_is_null(result$mods$mod_cont)
  expect_equal(sum(is_null_cont), 0)

  # mod_factor should still have null component
  is_null_factor <- .get_is_null(result$mods$mod_factor)
  expect_equal(sum(is_null_factor), 1)
})


test_that("Different moderators can have different alternative specifications", {

  # Custom alternative for mod_cont, default for mod_factor
  custom_mod <- list(mod_cont = BayesTools::prior("normal", parameters = list(mean = 0, sd = 0.2)))

  result <- RoBMA(
    yi = effect, sei = std_err, mods = ~ mod_cont + mod_factor,
    prior_mods = custom_mod, data = test_data, measure = "SMD", only_priors = TRUE
  )[["priors"]]

  # mod_cont should have custom alternative
  is_null_cont <- .get_is_null(result$mods$mod_cont)
  alt_idx <- which(!is_null_cont)[1]
  expect_equal(result$mods$mod_cont[[alt_idx]]$parameters$sd, 0.2)

  # mod_factor should have default alternative
  is_null_factor <- .get_is_null(result$mods$mod_factor)
  expect_true(any(!is_null_factor))  # has alternative component
})


# ============================================================================
# Tests for list of bias priors
# ============================================================================

test_that("List of bias priors creates mixture with multiple alternatives", {

  bias_priors <- list(
    BayesTools::prior_weightfunction("two-sided", c(0.05), BayesTools::wf_cumulative(c(1, 1))),
    BayesTools::prior_PET("normal", parameters = list(mean = 0, sd = 1))
  )

  result <- RoBMA(
    yi = effect, sei = std_err, data = test_data,
    prior_bias = bias_priors, measure = "SMD", only_priors = TRUE
  )[["priors"]]

  expect_true(BayesTools::is.prior.mixture(result$outcome$bias))

  # Should have 1 null (prior_none) + 2 custom alternatives
  expect_equal(length(result$outcome$bias), 3)

  has_wf  <- sapply(result$outcome$bias, BayesTools::is.prior.weightfunction)
  has_pet <- sapply(result$outcome$bias, BayesTools::is.prior.PET)
  expect_equal(sum(has_wf), 1)
  expect_equal(sum(has_pet), 1)
})


# ============================================================================
# Tests for heterogeneity allocation list priors
# ============================================================================

test_that("List of heterogeneity allocation priors creates mixture", {

  rho_priors <- list(
    BayesTools::prior("beta", parameters = list(alpha = 1, beta = 1)),
    BayesTools::prior("beta", parameters = list(alpha = 2, beta = 2))
  )

  result <- RoBMA(
    yi = effect, sei = std_err, cluster = cluster, data = test_data,
    prior_heterogeneity_allocation = rho_priors,
    measure = "SMD", only_priors = TRUE
  )[["priors"]]

  expect_true(BayesTools::is.prior.mixture(result$outcome$rho))
  expect_equal(length(result$outcome$rho), 2)  # 2 alternatives (no null by default)

  # Both should be beta distributions with different parameters
  expect_equal(result$outcome$rho[[1]]$parameters$alpha, 1)
  expect_equal(result$outcome$rho[[2]]$parameters$alpha, 2)
})


test_that("Heterogeneity allocation with both null and alternative lists", {

  alt_priors  <- list(BayesTools::prior("beta", parameters = list(alpha = 2, beta = 2)))
  null_priors <- list(BayesTools::prior("spike", parameters = list(location = 0.5)))

  result <- RoBMA(
    yi = effect, sei = std_err, cluster = cluster, data = test_data,
    prior_heterogeneity_allocation = alt_priors,
    prior_heterogeneity_allocation_null = null_priors,
    measure = "SMD", only_priors = TRUE
  )[["priors"]]

  expect_true(BayesTools::is.prior.mixture(result$outcome$rho))
  expect_equal(length(result$outcome$rho), 2)  # 1 null + 1 alternative

  is_null <- .get_is_null(result$outcome$rho)
  expect_equal(sum(is_null), 1)
  expect_equal(sum(!is_null), 1)
})


# ============================================================================
# Tests for multiple null priors (list of nulls)
# ============================================================================

test_that("List of null priors creates mixture with multiple null components", {

  null_priors <- list(
    BayesTools::prior("spike", parameters = list(location = 0)),
    BayesTools::prior("spike", parameters = list(location = 0.1))
  )

  result <- RoBMA(
    yi = effect, sei = std_err, data = test_data,
    prior_effect_null = null_priors, measure = "SMD", only_priors = TRUE
  )[["priors"]]

  expect_true(BayesTools::is.prior.mixture(result$outcome$mu))
  expect_equal(length(result$outcome$mu), 3)  # 2 nulls + 1 default alternative

  is_null <- .get_is_null(result$outcome$mu)
  expect_equal(sum(is_null), 2)
  expect_equal(sum(!is_null), 1)
})


# ============================================================================
# Tests for interaction between effect/heterogeneity and mods/scale
# ============================================================================

test_that("Effect prior mixture is forwarded to mods intercept", {

  custom_effect <- BayesTools::prior("normal", parameters = list(mean = 0, sd = 0.8))
  custom_null   <- BayesTools::prior("spike", parameters = list(location = 0.1))

  result <- suppressWarnings(RoBMA(
    yi = effect, sei = std_err, mods = ~ mod_cont,
    prior_effect = custom_effect, prior_effect_null = custom_null,
    data = test_data, measure = "SMD", only_priors = TRUE
  )[["priors"]])

  # Intercept should have the custom priors
  expect_true(BayesTools::is.prior.mixture(result$mods$intercept))

  is_null  <- .get_is_null(result$mods$intercept)
  null_idx <- which(is_null)[1]
  alt_idx  <- which(!is_null)[1]

  expect_equal(mean(result$mods$intercept[[null_idx]]), 0.1)
  expect_equal(result$mods$intercept[[alt_idx]]$parameters$sd, 0.8)
})


test_that("Both mods and scale use mixture priors together", {

  result <- suppressWarnings(RoBMA(
    yi = effect, sei = std_err, mods = ~ mod_cont, scale = ~ scale_var,
    data = test_data, measure = "SMD", only_priors = TRUE
  )[["priors"]])

  expect_true(!is.null(result$mods))
  expect_true(!is.null(result$scale))
  expect_true("intercept" %in% names(result$mods))
  expect_true("intercept" %in% names(result$scale))
  expect_null(result$outcome$mu)
  expect_null(result$outcome$tau)

  # mods intercept should be mixture
  expect_true(BayesTools::is.prior.mixture(result$mods$intercept))
})

context("CRAN smoke test")

on_cran <- get("on_cran", envir = asNamespace("testthat"), inherits = FALSE)

skip_if_not(on_cran(), "CRAN smoke test only")

test_that("package loads and fits a minimal brma model", {

  expect_true("RoBMA" %in% loadedNamespaces())

  yi  <- c(-0.12, 0.05, 0.18, 0.31)
  sei <- c(0.20, 0.18, 0.22, 0.19)

  fit <- brma(
    yi = yi, sei = sei, measure = "GEN",
    prior_unit_information_sd = 1,
    sample = 100, burnin = 50, adapt = 500,
    chains = 1, seed = 1, silent = TRUE,
    convergence_checks = set_convergence_checks(
      max_Rhat = NULL,
      min_ESS  = NULL
    )
  )

  fit_summary <- summary(fit)

  expect_s3_class(fit, "brma.norm")
  expect_s3_class(fit_summary, "summary.brma")
})

# ============================================================================ #
# marglik.R
# ============================================================================ #
#
# This file implements marginal likelihood computation for brma class objects
# via bridge sampling. The marginal likelihood is used for Bayesian model
# comparison via Bayes factors.
#
# The implementation reuses the evaluate and pdf helper functions from
# evaluate.R and pdf.R by converting single-sample parameters
# to 1-row matrices.
#
# ============================================================================ #


### RoBMA 4.0.0


# ---------------------------------------------------------------------------- #
# add_marglik generic and method
# ---------------------------------------------------------------------------- #

#' @export
add_marglik <- function(object, ...) UseMethod("add_marglik")


#' @title Add Marginal Likelihood to brma Objects
#'
#' @description Compute the marginal likelihood of a brma model using
#' bridge sampling and store the result in the object.
#'
#' @param object a brma model object.
#' @param ... additional arguments (currently not used).
#'
#' @details
#' The marginal likelihood is computed using the \code{bridgesampling} package
#' via the \code{BayesTools::JAGS_bridgesampling} wrapper. The result is stored
#' in the object and can be extracted using \code{\link{bridge_sampler.brma}}.
#'
#' Product-space model-averaging objects (\code{BMA.norm}, \code{BMA.glmm},
#' and \code{RoBMA}) do not expose a bridge-sampling marginal likelihood;
#' use predictive comparison methods such as \code{\link{loo.brma}} instead.
#'
#' @return The brma object with the marginal likelihood result stored in
#' \code{object[["marglik"]]}.
#'
#' @seealso \code{\link{bridge_sampler.brma}}, \code{\link{logml.brma}},
#' \code{\link{bf.brma}}, \code{\link{post_prob.brma}}
#'
#' @examples
#' \dontrun{
#' if (requireNamespace("metadat", quietly = TRUE)) {
#'   data(dat.lehmann2018, package = "metadat")
#'   fit <- brma(yi = yi, vi = vi, data = dat.lehmann2018, measure = "SMD")
#'
#'   fit <- add_marglik(fit)
#'
#'   bridge <- bridge_sampler(fit)
#'   print(bridge)
#'
#'   logml(fit)
#' }
#' }
#'
#' @aliases add_marglik
#' @export
add_marglik.brma <- function(object, ...) {
  if (inherits(object, "RoBMA")) {
    stop(
      "Marginal likelihood is not available for product-space ",
      "model-averaging objects (BMA.norm, BMA.glmm, RoBMA).",
      call. = FALSE
    )
  }

  marglik <- .marglik(object)
  object[["marglik"]] <- marglik
  return(object)
}


# ---------------------------------------------------------------------------- #
# .marglik internal function
# ---------------------------------------------------------------------------- #

# Compute marginal likelihood for brma objects using bridge sampling.
# This is an internal function called by add_marglik.brma.
#
# @param object a brma model object.
#
# @return An object of class "bridge" as returned by bridgesampling::bridge_sampler.
#
# @keywords internal
.marglik <- function(object) {

  data   <- object[["data"]]
  priors <- object[["priors"]]
  fit    <- object[["fit"]]

  # Public constructors reject p-hacking/composed selection kernels earlier.
  if (.marglik_has_composed_bias(priors[["outcome"]][["bias"]])) {
    stop(
      "Marginal likelihood is not available for combined ",
      "prior_bias(selection, phacking) models yet.",
      call. = FALSE
    )
  }

  ### create arguments to be passed to BayesTools::JAGS_bridgesampling
  fit_formula_list        <- list()
  fit_formula_data_list   <- list()
  fit_formula_prior_list  <- list()
  fit_formula_scale_list  <- list()

  ### create model base
  fit_priors <- .create_fit_priors(data = data, priors = priors)
  fit_data   <- .create_fit_data(data = data, priors = priors)
  fit_data   <- .marglik_add_selection_bridge_data(
    fit_data         = fit_data,
    priors           = priors,
    effect_direction = .data_effect_direction(data)
  )

  ### add effect regressions
  if (.is_data_mods(data)) {
    fit_formula_list[["mu"]]       <- .create_fit_formula_list(data = data, parameter = "mods")
    fit_formula_data_list[["mu"]]  <- .create_fit_formula_data_list(data = data, parameter = "mods")
    fit_formula_prior_list[["mu"]] <- .create_fit_formula_prior_list(priors = priors, parameter = "mods")
    fit_formula_scale_list[["mu"]] <- .data_standardize_continuous_predictors(data)
  }

  ### add heterogeneity regressions
  if (.is_data_scale(data)) {
    fit_formula_list[["log_tau"]]       <- .create_fit_formula_list(data = data, parameter = "scale")
    fit_formula_data_list[["log_tau"]]  <- .create_fit_formula_data_list(data = data, parameter = "scale")
    fit_formula_prior_list[["log_tau"]] <- .create_fit_formula_prior_list(priors = priors, parameter = "scale")
    fit_formula_scale_list[["log_tau"]] <- .data_standardize_continuous_predictors(data)
  }

  ### compute marginal likelihood
  marglik <- BayesTools::JAGS_bridgesampling(
    fit                = fit,
    log_posterior      = .log_posterior,
    data               = fit_data,
    prior_list         = fit_priors,
    formula_list       = if (length(fit_formula_list)       > 0) fit_formula_list       else NULL,
    formula_data_list  = if (length(fit_formula_data_list)  > 0) fit_formula_data_list  else NULL,
    formula_prior_list = if (length(fit_formula_prior_list) > 0) fit_formula_prior_list else NULL,
    formula_scale      = if (length(fit_formula_scale_list) > 0) fit_formula_scale_list else NULL,
    # additional arguments passed to .log_posterior via ...
    is_mods            = .is_data_mods(data),
    is_scale           = .is_data_scale(data),
    is_multilevel      = .is_data_multilevel(data),
    is_weights         = .is_data_weights(data),
    is_PET             = .is_priors_PET(priors),
    is_PEESE           = .is_priors_PEESE(priors),
    is_weightfunction  = .is_priors_weightfunction(priors),
    effect_direction   = .data_effect_direction(data),
    outcome_type       = .data_outcome_type(data)
  )

  return(marglik)
}


.marglik_has_composed_bias <- function(prior) {

  if (is.null(prior)) {
    return(FALSE)
  }
  if (BayesTools::is.prior.mixture(prior)) {
    return(any(vapply(as.list(prior), .marglik_has_composed_bias, logical(1))))
  }
  if (!.is_prior_bias_kernel(prior)) {
    return(FALSE)
  }

  return(.prior_has_phacking(prior))
}


.marglik_add_selection_bridge_data <- function(fit_data, priors, effect_direction) {

  if (!.is_priors_weightfunction(priors) || is.null(fit_data[["yi"]])) {
    return(fit_data)
  }

  selection_spec <- .selection_spec(
    priors           = priors,
    yi               = fit_data[["yi"]],
    sei              = fit_data[["sei"]],
    effect_direction = effect_direction,
    signed_data      = TRUE
  )

  if (is.null(selection_spec)) {
    return(fit_data)
  }

  fit_data[["sel_kernel_mode"]]           <- selection_spec[["kernel_mode"]]
  fit_data[["sel_segment_bounds"]]        <- .selection_jags_bounds(selection_spec[["segments"]][["bounds"]])
  fit_data[["sel_segment_step_bin"]]      <- selection_spec[["segments"]][["step_bin"]]
  fit_data[["sel_segment_phack_region"]]  <- selection_spec[["segments"]][["phack_region"]]
  fit_data[["sel_phack_q"]]               <- selection_spec[["phack_q"]]
  fit_data[["sel_phack_z_source"]]        <- selection_spec[["phack_z_source"]]
  fit_data[["sel_phack_z_dest"]]          <- selection_spec[["phack_z_dest"]]

  return(fit_data)
}


# ---------------------------------------------------------------------------- #
# .log_posterior
# ---------------------------------------------------------------------------- #
#
# Compute the log-likelihood of the data given a single posterior sample.
# This function is called by BayesTools::JAGS_bridgesampling for each
# posterior sample during bridge sampling.
#
# The function converts single-sample parameters to 1-row matrices to reuse
# the existing evaluate and pdf helper functions from evaluate.R and pdf.R.
# (in contrast to .pdf.brma, this function does not add likelihood contributions from 
#  estimate-level effects / baserate/ lograte for GLMMs since those are automatically 
#  handled via JAGS_bridgesampling function -- i.e., all parameters generated directly 
#  by BayesTools are automatically evaluated in the likelihood calculation, the only 
#  part missing is the final p(data | parameters) term) 
#
# @param parameters       named list of parameter values (preprocessed by BayesTools)
#                         - mu: scalar (no mods) or vector of length K (with mods)
#                         - tau: scalar (no scale regression)
#                         - log_tau: vector of length K (scale regression)
#                         - rho: scalar (if multilevel)
#                         - gamma: vector of length n_clusters (if multilevel)
#                         - PET: scalar (if PET model)
#                         - PEESE: scalar (if PEESE model)
#                         - omega: vector of weights (if weightfunction)
#                         - pi: vector of baserates (if binomial)
#                         - phi: vector of log-rates (if Poisson)
#                         - theta: vector of random effects (if GLMM)
# @param data             list containing fit_data (from .create_fit_data)
# @param is_mods          logical; whether model has moderators
# @param is_scale         logical; whether model has scale regression
# @param is_multilevel    logical; whether model is multilevel
# @param is_weights       logical; whether model uses weights (currently unused)
# @param is_PET           logical; whether model includes PET
# @param is_PEESE         logical; whether model includes PEESE
# @param is_weightfunction logical; whether model includes selection weights
# @param effect_direction character; "positive" or "negative"
# @param outcome_type     character; "norm", "bin", or "pois"
#
# @return scalar; sum of log-likelihoods across all observations
#
# ---------------------------------------------------------------------------- #
.log_posterior <- function(
    parameters, data,
    is_mods, is_scale, is_multilevel, is_weights,
    is_PET, is_PEESE, is_weightfunction, effect_direction,
    outcome_type) {

  ### extract number of observations
  K <- data[["K"]]

  ### compute mu samples as 1 x K matrix
  # BayesTools evaluates the formula and returns:
  # - scalar mu (no mods) -> replicate to K columns

  # - vector mu of length K (with mods) -> use directly
  mu_samples <- .marglik_get_mu_samples(
    parameters       = parameters,
    is_mods          = is_mods,
    is_PET           = is_PET,
    is_PEESE         = is_PEESE,
    effect_direction = effect_direction,
    sei              = if (outcome_type == "norm") data[["sei"]] else NULL,
    K                = K
  )

  ### compute tau samples as 1 x K matrix
  # BayesTools returns:
  # - scalar tau (no scale regression) -> replicate to K columns
  # - vector log_tau of length K (scale regression) -> exponentiate, use directly
  tau_result <- .marglik_get_tau_samples(
    parameters    = parameters,
    is_scale      = is_scale,
    is_multilevel = is_multilevel,
    K             = K
  )
  tau_within_samples  <- tau_result[["tau_within"]]
  tau_between_samples <- tau_result[["tau_between"]]

  ### add cluster-level (gamma) contribution for multilevel models
  if (is_multilevel) {
    cluster_contribution <- .marglik_get_cluster_effects(
      parameters       = parameters,
      tau_between      = tau_between_samples,
      cluster          = data[["cluster"]],
      effect_direction = effect_direction
    )
    mu_samples <- mu_samples + cluster_contribution
  }

  ### dispatch to appropriate log-likelihood computation based on outcome type
  if (outcome_type == "norm") {

    # for selection models, PET, and PEESE, the outcome is in "positive" space
    # (yi flipped for negative effect direction in .create_fit_data)
    # so we need to flip mu_samples to match
    if (effect_direction == "negative") {
      mu_samples <- -mu_samples
    }

    if (is_weightfunction) {

      selection_context <- .marglik_selection_context(parameters, data)
      log_lik <- .outcome_pdf.selnorm(
        yi                = data[["yi"]],
        mu_samples        = mu_samples,
        tau_within        = tau_within_samples,
        sei               = data[["sei"]],
        selection_context = selection_context
      )

    } else {

      # standard normal likelihood
      log_lik <- .outcome_pdf.norm(
        yi         = data[["yi"]],
        mu_samples = mu_samples,
        tau_within = tau_within_samples,
        sei        = data[["sei"]]
      )

    }

  } else if (outcome_type == "bin") {

    # add GLMM random effects (theta * tau_within) for binomial models
    theta_contribution <- .marglik_get_theta_samples(
      parameters = parameters,
      tau_within = tau_within_samples,
      K          = K
    )
    mu_samples <- mu_samples + theta_contribution

    # get baserate samples
    logit_baserate <- .marglik_get_baserate_samples(
      parameters = parameters,
      K          = K
    )

    log_lik <- .outcome_pdf.binom_conditional(
      ai             = data[["ai"]],
      ci             = data[["ci"]],
      n1i            = data[["n1i"]],
      n2i            = data[["n2i"]],
      mu_samples     = mu_samples,
      logit_baserate = logit_baserate
    )

  } else if (outcome_type == "pois") {

    # add GLMM random effects (theta * tau_within) for Poisson models
    theta_contribution <- .marglik_get_theta_samples(
      parameters = parameters,
      tau_within = tau_within_samples,
      K          = K
    )
    mu_samples <- mu_samples + theta_contribution

    # get log-rate samples
    log_phi <- .marglik_get_lograte_samples(
      parameters = parameters,
      K          = K
    )

    log_lik <- .outcome_pdf.pois_conditional(
      x1i        = data[["x1i"]],
      x2i        = data[["x2i"]],
      t1i        = data[["t1i"]],
      t2i        = data[["t2i"]],
      mu_samples = mu_samples,
      log_phi    = log_phi
    )

  }

  if (is_weights) {
    log_lik <- .apply_log_lik_weights(log_lik, data[["weight"]])
  }

  ### return sum of log-likelihoods (scalar)
  return(sum(log_lik))
}


# ---------------------------------------------------------------------------- #
# Helper functions for .log_posterior
# ---------------------------------------------------------------------------- #
#
# These functions convert single-sample parameters from BayesTools to 1-row
# matrices compatible with the evaluate and pdf helper functions.
# ---------------------------------------------------------------------------- #


#' @keywords internal
.marglik_get_mu_samples <- function(parameters, is_mods, is_PET, is_PEESE,
                                    effect_direction, sei, K) {

  # BayesTools returns:
  # - scalar mu (no mods)
  #   => replicate to 1 x K matrix
  # - vector mu of length K (with mods, formula already evaluated)
  #   => vector of length K -> 1 x K matrix
  mu_samples <- matrix(parameters[["mu"]], nrow = 1, ncol = K)

  # add PET adjustment (PET * sei)
  # direction multiplier: +1 for positive, -1 for negative
  if (is_PET) {
    direction <- ifelse(effect_direction == "negative", -1, 1)
    mu_samples <- mu_samples + direction * parameters[["PET"]] * sei
  }

  # add PEESE adjustment (PEESE * sei^2)
  if (is_PEESE) {
    direction <- ifelse(effect_direction == "negative", -1, 1)
    mu_samples <- mu_samples + direction * parameters[["PEESE"]] * sei^2
  }

  return(mu_samples)
}


#' @keywords internal
.marglik_get_tau_samples <- function(parameters, is_scale, is_multilevel, K) {

  # BayesTools returns:
  # - scalar tau (no scale regression)
  # - vector log_tau of length K (scale regression, need to exponentiate)
  if (is_scale) {
    # log_tau vector -> exponentiate and convert to 1 x K matrix
    tau_samples <- matrix(exp(parameters[["log_tau"]]), nrow = 1, ncol = K)
  } else {
    # scalar tau -> replicate to 1 x K matrix
    tau_samples <- matrix(parameters[["tau"]], nrow = 1, ncol = K)
  }

  # split tau into within/between components for multilevel models
  if (is_multilevel) {
    # extract rho (proportion of variance at cluster-level)
    rho <- parameters[["rho"]]

    # clamp rho to [0, 1] to handle numerical precision issues
    rho <- min(max(rho, 0), 1)

    # tau_within  = tau * sqrt(1 - rho)  (estimate-level heterogeneity)
    # tau_between = tau * sqrt(rho)      (cluster-level heterogeneity)
    tau_within_samples  <- tau_samples * sqrt(1 - rho)
    tau_between_samples <- tau_samples * sqrt(rho)

  } else {

    # non-multilevel: all heterogeneity is at estimate-level
    tau_within_samples  <- tau_samples
    tau_between_samples <- matrix(0, nrow = 1, ncol = K)

  }

  return(list(
    tau_total   = tau_samples,
    tau_within  = tau_within_samples,
    tau_between = tau_between_samples
  ))
}


#' @keywords internal
.marglik_get_cluster_effects <- function(parameters, tau_between, cluster,
                                         effect_direction) {

  K <- ncol(tau_between)

  # extract gamma samples (vector of length n_clusters)
  # BayesTools returns gamma as a vector
  gamma <- parameters[["gamma"]]

  # gamma[cluster] maps each observation to its cluster-level effect
  # multiply by tau_between to get the contribution
  cluster_contribution <- matrix(gamma[cluster] * tau_between, nrow = 1, ncol = K)

  return(cluster_contribution)
}


.marglik_selection_context <- function(parameters, data) {

  n_bins <- length(data[["sel_z_lower"]])
  phack_z_source <- if (!is.null(data[["phack_z_source"]])) {
    data[["phack_z_source"]]
  } else if (!is.null(data[["sel_phack_z_source"]])) {
    data[["sel_phack_z_source"]]
  } else {
    c(0, 0)
  }
  phack_z_dest <- if (!is.null(data[["phack_z_dest"]])) {
    data[["phack_z_dest"]]
  } else if (!is.null(data[["sel_phack_z_dest"]])) {
    data[["sel_phack_z_dest"]]
  } else {
    c(0, 0)
  }
  omega <- if (!is.null(parameters[["omega"]])) {
    matrix(parameters[["omega"]], nrow = 1)
  } else if (!is.null(data[["sel_omega"]])) {
    matrix(data[["sel_omega"]], nrow = 1)
  } else {
    matrix(1, nrow = 1, ncol = n_bins)
  }

  alpha <- if (!is.null(parameters[["alpha"]])) parameters[["alpha"]] else {
    if (!is.null(data[["sel_phack_alpha"]])) data[["sel_phack_alpha"]] else 0
  }
  phack_kind <- if (!is.null(parameters[["phack_kind"]])) {
    as.integer(parameters[["phack_kind"]])
  } else {
    if (!is.null(data[["sel_phack_kind"]])) {
      as.integer(data[["sel_phack_kind"]])
    } else if (!is.null(data[["sel_phack_q"]]) &&
               data[["sel_kernel_mode"]] %in% c(SELKERNEL_PHACK_POWER, SELKERNEL_STEP_PHACK_POWER)) {
      as.integer(data[["sel_phack_q"]])
    } else {
      0L
    }
  }

  return(list(
    kernel_mode    = data[["sel_kernel_mode"]],
    z_lower        = data[["sel_z_lower"]],
    z_upper        = data[["sel_z_upper"]],
    obs_bin        = data[["sel_obs_bin"]],
    sign           = data[["sel_sign"]],
    n_bins         = n_bins,
    has_step       = data[["sel_kernel_mode"]] %in% c(SELKERNEL_STEP, SELKERNEL_STEP_PHACK_POWER),
    has_phack      = data[["sel_kernel_mode"]] %in% c(SELKERNEL_PHACK_POWER, SELKERNEL_STEP_PHACK_POWER),
    phack_q        = if (phack_kind > 0L) phack_kind else 1L,
    phack_z_source = phack_z_source,
    phack_z_dest   = phack_z_dest,
    segments       = list(
      bounds       = data[["sel_segment_bounds"]],
      step_bin     = data[["sel_segment_step_bin"]],
      phack_region = data[["sel_segment_phack_region"]]
    ),
    omega          = omega,
    alpha          = alpha,
    phack_kind     = phack_kind
  ))
}


#' @keywords internal
.marglik_get_theta_samples <- function(parameters, tau_within, K) {

  # theta is a vector of length K (estimate-level random effects for GLMM)
  theta <- parameters[["theta"]]

  # theta contribution = theta[k] * tau_within[k]
  theta_contribution <- matrix(theta * tau_within, nrow = 1, ncol = K)

  return(theta_contribution)
}


#' @keywords internal
.marglik_get_baserate_samples <- function(parameters, K) {

  # pi is a vector of length K (baserates for binomial models)
  pi <- parameters[["pi"]]

  # return logit(pi) as 1 x K matrix
  logit_baserate <- matrix(.logit(pi), nrow = 1, ncol = K)

  return(logit_baserate)
}


#' @keywords internal
.marglik_get_lograte_samples <- function(parameters, K) {

  # phi is a vector of length K (log-rates for Poisson models)
  phi <- parameters[["phi"]]

  # return log(phi) as 1 x K matrix
  # Note: phi is already on log scale in JAGS model
  log_phi <- matrix(phi, nrow = 1, ncol = K)

  return(log_phi)
}

# ============================================================================ #
#
# Test bridge sampling marginal likelihood computation for brma objects.
# This tests the logml, bf, bayes_factor, and post_prob S3 methods.
# The bridge_sampler function is tested directly alongside the fit functions
#
# ============================================================================ #
context("bridgesampling methods for brma objects")

source(testthat::test_path("common-functions.R"))
skip_on_cran()
skip_if_no_fits()
skip_if_not_installed("metafor")

# list cached marginal-likelihood fits lazily
marglik_names <- list_fits(has_marglik = TRUE)
fits          <- lazy_fits(marglik_names, validate = FALSE)
info          <- lazy_infos(marglik_names, validate = FALSE)

# ---------------------------------------------------------------------------- #
# bridge_sampler function works with all model types
# ---------------------------------------------------------------------------- #

test_that("bridge_sampler extracts bridge sampling object", {

  for (name in marglik_names) {
    # the marginal likelihood inherits the correct class
    expect_s3_class(bridge_sampler(fits[[name]]), "bridge")
  }
})

test_that("bridge sampling marginal likelihood is close to BIC for metafor-reference fits", {

  metafor_names <- list_fits(has_marglik = TRUE, has_metafor = TRUE)

  for (name in metafor_names) {

    # the marginal likelihood and BIC-based marglik are close
    # (the analyses don't use unit information priors but the
    #  difference should not be large)
    BIC_metafor <- BIC(info[[name]][["metafor"]])
    marglik     <- logml(fits[[name]])

    if (name %in% c("bcg_glmm", "nielweise2008_glmm", "bcg_glmm_reg")) {
      next
    } # skip glmm because of marginalization differences

    if (name %in% c("bcg_meta-regression", "bcg_meta-regression2", "bcg_meta-regression3", "bcg_meta-regression3b", "bcg_meta-regression4", "bcg_meta-regression4b")) {
      next
    } # skip because of scaled priors differences

    expect_equal(-BIC_metafor / 2, marglik, tolerance = 0.15)
  }
})


test_that("effect-direction reversals preserve marginal likelihood", {

  fit1 <- fits[["dat.lehmann2018-PET"]]
  fit2 <- fits[["dat.lehmann2018-PET_neg"]]

  expect_equal(bridgesampling::bf(bridge_sampler(fit2), bridge_sampler(fit1))$bf, 1, tolerance = 0.01)

  fit1 <- fits[["dat.lehmann2018-3PSM"]]
  fit2 <- fits[["dat.lehmann2018-3PSM_neg"]]

  expect_equal(bridgesampling::bf(bridge_sampler(fit2), bridge_sampler(fit1))$bf, 1, tolerance = 0.01)
})

test_that("logml returns scalar log marginal likelihood, can be applied to both bridge and brma", {

  skip_if_missing_fits("bcg_meta-analysis")

  name     <- "bcg_meta-analysis"
  fit_brma <- fits[[name]]

  logml_result1 <- bridgesampling::logml(fit_brma)
  logml_result2 <- RoBMA::logml(fit_brma)

  # consistency check
  expect_equal(logml_result1, logml_result2, tolerance = 0.01)
})


test_that("bf computes Bayes factor between two models, can be applied to both bridge and brma", {

  skip_if_missing_fits(c("bcg_meta-analysis", "bcg_meta-regression"))

  fit_brma1 <- fits[["bcg_meta-analysis"]]
  fit_brma2 <- fits[["bcg_meta-regression"]]

  bf_result1 <- RoBMA::bf(fit_brma1, fit_brma2)
  bf_result2 <- bridgesampling::bf(fit_brma1, fit_brma2)
  bf_result3 <- RoBMA::bayes_factor(fit_brma1, fit_brma2)
  bf_result4 <- bridgesampling::bayes_factor(fit_brma1, fit_brma2)

  # consistency check
  expect_equal(bf_result1, bf_result2, tolerance = 0.01)
  expect_equal(bf_result1, bf_result3, tolerance = 0.01)
  expect_equal(bf_result1, bf_result4, tolerance = 0.01)
})


test_that("post_prob computes posterior model probabilities", {

  skip_if_missing_fits(c("bcg_meta-analysis", "bcg_meta-regression"))

  fit_brma1 <- fits[["bcg_meta-analysis"]]
  fit_brma2 <- fits[["bcg_meta-regression"]]

  pp_result1 <- RoBMA::post_prob(fit_brma1, fit_brma2)
  pp_result2 <- bridgesampling::post_prob(fit_brma1, fit_brma2)

  # consistency check
  expect_equal(pp_result1, pp_result2, tolerance = 0.01)
})


test_that("post_prob respects prior model probabilities", {

  skip_if_missing_fits(c("bcg_meta-analysis", "bcg_meta-regression"))

  fit_brma  <- fits[["bcg_meta-analysis"]]
  fit_brma2 <- fits[["bcg_meta-regression"]]

  # equal priors (default)
  pp_equal <- RoBMA::post_prob(fit_brma, fit_brma2)

  # unequal priors
  pp_unequal <- RoBMA::post_prob(fit_brma, fit_brma2, prior_prob = c(0.9, 0.1))

  # with stronger prior on first model, its posterior should increase
  expect_lt(unname(pp_equal[1]), unname(pp_unequal[1]))
  expect_equal(sum(pp_unequal), 1)
})


# ---------------------------------------------------------------------------- #
# error handling tests
# ---------------------------------------------------------------------------- #

test_that("bf rejects non-brma object", {

  skip_if_missing_fits("bcg_meta-analysis")

  fit_brma <- fits[["bcg_meta-analysis"]]
  expect_error(
    bf(fit_brma, "not a brma"),
    "needs to be of class 'brma'"
  )
})


test_that("post_prob rejects single model", {

  skip_if_missing_fits("bcg_meta-analysis")

  fit_brma <- fits[["bcg_meta-analysis"]]
  expect_error(
    post_prob(fit_brma),
    "Only one object"
  )
})


test_that("bridge_sampler rejects fits without marginal likelihood", {

  skip_if_missing_fits("bcg_meta-analysis")

  # create a fit without marglik
  fit_brma <- fits[["bcg_meta-analysis"]]
  fit_brma[["marglik"]] <- NULL
  expect_error(
    bridge_sampler(fit_brma),
    "add_marglik"
  )
})

test_that("add_marglik rejects product-space model-averaging objects", {

  product_names <- c(
    "dat.lehmann2018_BMA.norm",
    "bcg_BMA.glmm",
    "dat.lehmann2018_RoBMA"
  )
  skip_if_missing_fits(product_names)

  for (name in product_names) {
    expect_error(
      add_marglik(load_fit(name, validate = FALSE)),
      "Marginal likelihood is not available for product-space"
    )
    expect_error(
      bridge_sampler(load_fit(name, validate = FALSE)),
      "Marginal likelihood is not available for product-space"
    )
    expect_error(
      logml(load_fit(name, validate = FALSE)),
      "Marginal likelihood is not available for product-space"
    )
  }
})

# ============================================================================ #
# brma.radial.R
# ============================================================================ #
#
# Radial (Galbraith) plot functions for brma objects.
#
# The radial plot displays effect sizes on a transformed scale where:
# - x-axis: precision (1/sqrt(vi + tau^2))
# - z-axis: standardized effect (yi/sqrt(vi + tau^2))
#
# A line from the origin through any point has slope equal to the
# observed effect size. An arc on the right side maps z-values back
# to the effect size scale.
#
# Only supported for intercept-only models (no moderators, no scale).
#
# ============================================================================ #


### radial plot functions ----
#' @export
radial <- function(x, ...) UseMethod("radial")

#' @export
galbraith <- function(x, ...) UseMethod("galbraith")


#' @title Radial (Galbraith) Plot for brma Object
#'
#' @description \code{radial.brma} creates a radial (Galbraith) plot for a
#' fitted brma object. The plot displays each study as a point with
#' precision on the x-axis and the standardized effect size on the z-axis.
#' Without centering, a line from the origin through any point has slope equal
#' to the observed effect size. With centering, slopes are relative to the
#' pooled estimate. An arc on the right side maps standardized values back to
#' the effect size scale.
#'
#' @param x a fitted brma object. Must be an intercept-only model
#' (no moderators or scale regression).
#' @param center logical indicating whether to center the plot at the
#' pooled estimate. When \code{TRUE}, the pooled estimate is shifted to
#' zero on the z-axis. Defaults to \code{FALSE}.
#' @param xlim x-axis limits. If not specified, limits are computed from data.
#' @param zlim z-axis limits. If not specified, limits are computed from data.
#' @param xlab title for the x-axis. If not specified, a default label is used.
#' @param zlab title for the z-axis. If not specified, a default label is used.
#' @param atz numeric vector of positions for tick marks on the z-axis (left
#' vertical axis). If not specified, positions are chosen automatically.
#' @param aty numeric vector of effect size values to mark on the y-axis arc
#' (right side). If not specified, values are chosen automatically.
#' @param steps integer specifying the number of tick marks on the y-axis arc
#' when \code{aty} is not specified. Defaults to \code{7}.
#' @param level numeric value between 0 and 100 specifying the confidence level
#' for the confidence band and pooled-effect CI arc. Defaults to \code{95}.
#' @param digits integer specifying the number of decimal places for y-axis
#' arc labels. Defaults to \code{2}.
#' @param transf optional transformation function applied to the y-axis arc
#' labels (e.g., \code{transf = exp} when effect sizes are on a log scale).
#' @param targs optional list of additional arguments passed to the
#' \code{transf} function.
#' @param plot_type whether to use a base plot \code{"base"} or ggplot2
#' \code{"ggplot"} for plotting. Defaults to \code{"base"}.
#' @param ... additional graphical arguments to customize the plot appearance:
#' \describe{
#'   \item{pch}{point symbol (default: 21, filled circle)}
#'   \item{col}{point border color (default: "black")}
#'   \item{bg}{point fill/background color (default: "#A6A6A6")}
#'   \item{cex}{point size multiplier for base graphics (default: 1)}
#'   \item{size}{point size for ggplot2 (default: 2)}
#'   \item{las}{axis-label style for base graphics (default: 1)}
#'   \item{back}{background color for the confidence band (default: "grey90").
#'     Set to \code{NA} to suppress.}
#'   \item{arc.res}{integer specifying the number of line segments used to
#'     draw arcs (default: 100)}
#'   \item{main}{character string for plot title (default: no title)}
#'   \item{as_data}{if \code{TRUE}, returns plot data instead of creating
#'     the plot}
#' }
#'
#' @details
#' The radial (Galbraith) plot transforms each study's effect size and
#' standard error into a point in precision-standardized space:
#'
#' \itemize{
#'   \item x-axis: precision = \eqn{1/\sqrt{v_i + \hat{\tau}^2}}
#'   \item z-axis: standardized effect = \eqn{y_i/\sqrt{v_i + \hat{\tau}^2}}
#' }
#'
#' Under the random-effects model, studies consistent with the pooled effect
#' should fall within the sloped parallelogram confidence band around the
#' pooled-effect line. The arc on the right side allows reading individual
#' effect sizes by projecting from the origin through a point to the arc; when
#' \code{center = TRUE}, the plotted slope is relative to the pooled effect.
#'
#' This function requires an intercept-only model; radial plots are not
#' meaningful for meta-regression or location-scale models where the
#' pooled effect varies across studies.
#'
#' \code{galbraith()} is a same-argument alias for \code{radial()}.
#'
#' @return \code{radial.brma} returns \code{NULL} invisibly if
#' \code{plot_type = "base"} or a ggplot object if \code{plot_type = "ggplot"}.
#'
#' If \code{as_data = TRUE}, returns a list with the computed plot data
#' including: \code{points}, \code{refline}, \code{band}, \code{arc},
#' \code{arc_ticks}, \code{ci_arc}, \code{ci_ticks}, \code{xlim}, and
#' \code{zlim}.
#'
#' @references
#' Galbraith, R. F. (1988). Graphical display of estimates having differing
#' standard errors. \emph{Technometrics}, 30(3), 271-281.
#'
#' Galbraith, R. F. (1994). Some applications of radial plots. \emph{Journal
#' of the American Statistical Association}, 89(428), 1232-1242.
#'
#' @examples \dontrun{
#' if (requireNamespace("metadat", quietly = TRUE) &&
#'     requireNamespace("metafor", quietly = TRUE)) {
#'   data(dat.bcg, package = "metadat")
#'   dat <- metafor::escalc(
#'     measure = "RR",
#'     ai      = tpos,
#'     bi      = tneg,
#'     ci      = cpos,
#'     di      = cneg,
#'     data    = dat.bcg
#'   )
#'
#'   fit <- brma(yi = yi, vi = vi, data = dat, measure = "RR")
#'   radial(fit)
#'   radial(fit, center = TRUE)
#'   radial(fit, plot_type = "ggplot")
#'   galbraith(fit)
#' }
#' }
#'
#' @seealso [funnel.brma()], [pooled_effect.brma()], [pooled_heterogeneity.brma()]
#' @aliases radial galbraith
#' @export
#' @rdname radial
radial.brma <- function(x, center = FALSE, xlim, zlim, xlab, zlab,
                        atz, aty, steps = 7, level = 95, digits = 2,
                        transf, targs, plot_type = "base", ...) {

  # input validation
  BayesTools::check_bool(center, "center")
  BayesTools::check_char(plot_type, "plot_type", allow_values = c("base", "ggplot"))
  .check_plot_limits(if (missing(xlim)) NULL else xlim, "xlim")
  .check_plot_limits(if (missing(zlim)) NULL else zlim, "zlim")
  .check_plot_label(if (missing(xlab)) NULL else xlab, "xlab")
  .check_plot_label(if (missing(zlab)) NULL else zlab, "zlab")
  .check_plot_numeric(if (missing(atz)) NULL else atz, "atz", allow_null = TRUE)
  .check_plot_numeric(if (missing(aty)) NULL else aty, "aty", allow_null = TRUE)
  BayesTools::check_int(steps, "steps", lower = 1)
  BayesTools::check_real(level, "level", check_length = 1, allow_NA = FALSE)
  if (level <= 0 || level >= 100) {
    stop("'level' must be greater than 0 and less than 100.", call. = FALSE)
  }
  BayesTools::check_int(digits, "digits", lower = 0)
  .check_plot_function(if (missing(transf)) NULL else transf, "transf")
  .check_plot_list(if (missing(targs)) NULL else targs, "targs")

  # check model type - error on models with moderators or scale
  if (.is_mods(x)) {
    stop("Radial plots cannot be drawn for models that contain moderators.", call. = FALSE)
  }
  if (.is_scale(x)) {
    stop("Radial plots cannot be drawn for models with scale regression.", call. = FALSE)
  }

  # set up graphical arguments with defaults
  dots <- .set_dots_radial(...)

  # generate radial plot data
  radial_data <- .radial_data(
    x       = x,
    center  = center,
    xlim    = if (missing(xlim)) NULL else xlim,
    zlim    = if (missing(zlim)) NULL else zlim,
    xlab    = if (missing(xlab)) NULL else xlab,
    zlab    = if (missing(zlab)) NULL else zlab,
    atz     = if (missing(atz)) NULL else atz,
    aty     = if (missing(aty)) NULL else aty,
    steps   = steps,
    level   = level,
    digits  = digits,
    transf  = if (missing(transf)) NULL else transf,
    targs   = if (missing(targs)) NULL else targs,
    dots    = dots
  )

  # allow data return for programmatic access
  if (isTRUE(dots[["as_data"]])) {
    return(radial_data)
  }

  # create plot
  if (plot_type == "ggplot") {
    return(.radial_plot_ggplot(radial_data, dots))
  } else {
    .radial_plot_base(radial_data, dots)
    return(invisible(NULL))
  }
}


#' @export
#' @rdname radial
galbraith.brma <- function(x, ...) {

  radial.brma(x, ...)
}


# ---------------------------------------------------------------------------- #
# .radial_data
# ---------------------------------------------------------------------------- #
#
# Generate data for radial (Galbraith) plot.
#
# Computes precision (x-axis), standardized effects (z-axis),
# confidence band, reference line, and arc data.
#
# Arc coordinates are computed using trigonometry (for ggplot and as_data).
# The base R plotting function recomputes arcs using the aspect ratio
# for correct circular appearance (matching metafor::radial).
#
# @param x       brma object
# @param center  logical; center at pooled estimate
# @param xlim    x-axis limits (NULL for auto)
# @param zlim    z-axis limits (NULL for auto)
# @param xlab    x-axis label (NULL for default)
# @param zlab    z-axis label (NULL for default)
# @param atz     z-axis tick positions (NULL for auto)
# @param aty     arc tick values (NULL for auto)
# @param steps   number of arc ticks when aty is NULL
# @param level   confidence level (0-100)
# @param digits  decimal places for arc labels
# @param transf  transformation for arc labels (NULL for none)
# @param targs   additional args for transf (NULL for none)
# @param dots    list of graphical parameters
#
# @return list with radial plot data components
#
# ---------------------------------------------------------------------------- #
.radial_data <- function(x, center, xlim, zlim, xlab, zlab,
                         atz, aty, steps, level, digits, transf, targs, dots) {

  # get observed effect sizes and variances
  yi <- .outcome_data_yi(x)
  vi <- .outcome_data_vi(x)
  K  <- length(yi)

  # confidence level
  alpha  <- 1 - level / 100
  z_crit <- stats::qnorm(1 - alpha / 2)
  probs  <- c(alpha / 2, 1 - alpha / 2)

  # get pooled effect estimate
  mu_samples <- pooled_effect(x, probs = probs)
  mu_summary <- summary(mu_samples)
  beta       <- mu_summary["mu", "Mean"]
  ci.lb      <- mu_summary["mu", as.character(probs[1])]
  ci.ub      <- mu_summary["mu", as.character(probs[2])]

  # get heterogeneity estimate
  tau_samples <- pooled_heterogeneity(x)
  tau_summary <- summary(tau_samples)
  tau         <- tau_summary["tau", "Mean"]
  tau2        <- tau^2

  # compute precision and standardized values
  wi <- vi + tau2
  xi <- 1 / sqrt(wi)
  zi <- yi / sqrt(wi)

  # save uncentered yi for arc label computation
  yi.c <- yi

  # arc range tracking (on original scale before centering)
  if (is.null(aty)) {
    atyis <- range(yi)
  } else {
    atyis <- range(aty)
    aty.c <- aty
  }

  # center if requested
  if (center) {
    yi         <- yi - beta
    zi         <- yi / sqrt(wi)
    beta_plot  <- 0
    ci.lb_plot <- ci.lb - beta
    ci.ub_plot <- ci.ub - beta
    atyis      <- atyis - beta
    if (!is.null(aty)) aty <- aty - beta
  } else {
    beta_plot  <- beta
    ci.lb_plot <- ci.lb
    ci.ub_plot <- ci.ub
  }

  # ---- x-axis limits (matching metafor: 30% wider) ----
  if (is.null(xlim)) {
    xlim <- c(0, 1.3 * max(xi))
  }
  xaxismax <- xlim[2]

  # arc positions (outside xlim, matching metafor)
  ci.xpos <- xlim[2] + 0.12 * diff(xlim)
  ya.xpos <- xlim[2] + 0.14 * diff(xlim)

  arc_res <- dots[["arc.res"]]

  # ---- axis labels ----
  if (is.null(xlab)) {
    xlab <- expression(x[i] == 1 / sqrt(v[i] + tau^2))
  }

  # zlab: fraction expression for base R mtext, simpler for ggplot
  zlab_auto <- is.null(zlab)
  if (zlab_auto) {
    if (center) {
      zlab <- expression(z[i] == frac(y[i] - hat(mu), sqrt(v[i] + tau^2)))
    } else {
      zlab <- expression(z[i] == frac(y[i], sqrt(v[i] + tau^2)))
    }
  }

  # ---- arc tick values (effect size scale) ----
  if (is.null(aty)) {
    aty   <- pretty(atyis, n = steps)
    aty   <- unique(aty)
    aty.c <- if (center) aty + beta else aty
  }
  # else: aty (centered if needed) and aty.c (original) already set

  # arc tick labels (optionally transformed)
  if (!is.null(transf)) {
    if (!is.null(targs)) {
      label_values <- do.call(transf, c(list(aty.c), targs))
    } else {
      label_values <- transf(aty.c)
    }
    aty_label_text <- formatC(label_values, digits = digits, format = "f")
  } else {
    aty_label_text <- formatC(aty.c, digits = digits, format = "f")
  }

  # ---- z-axis limits (matching metafor) ----
  if (is.null(zlim)) {
    zlim <- c(
      min(-5, 1.1 * min(zi),
          1.1 * ci.lb_plot * ci.xpos,
          1.1 * min(aty) * ya.xpos,
          1.1 * min(yi) * ya.xpos,
          -1.1 * z_crit + xaxismax * beta_plot),
      max(5, 1.1 * max(zi),
          1.1 * ci.ub_plot * ci.xpos,
          1.1 * max(aty) * ya.xpos,
          1.1 * max(yi) * ya.xpos,
          1.1 * z_crit + xaxismax * beta_plot)
    )
  }

  # z-axis tick positions
  if (is.null(atz)) {
    atz <- pretty(zlim)
  }

  # ---- plot data ----
  df_points <- data.frame(x = xi, z = zi)

  # reference line (within axis range only)
  df_refline <- data.frame(
    x = c(0, xaxismax),
    z = c(0, xaxismax * beta_plot)
  )

  # confidence band (parallelogram, within axis range)
  df_band <- data.frame(
    x = c(0, xaxismax, xaxismax, 0),
    z = c(z_crit, z_crit + xaxismax * beta_plot,
          -z_crit + xaxismax * beta_plot, -z_crit)
  )
  df_band_upper <- data.frame(
    x = c(0, xaxismax),
    z = c(z_crit, z_crit + xaxismax * beta_plot)
  )
  df_band_lower <- data.frame(
    x = c(0, xaxismax),
    z = c(-z_crit, -z_crit + xaxismax * beta_plot)
  )

  # ---- arc coordinates (trigonometric, for ggplot and as_data) ----
  # outer arc
  arc_slopes <- seq(min(aty), max(aty), length.out = arc_res)
  arc_theta  <- atan(arc_slopes)
  df_arc <- data.frame(
    x = ya.xpos * cos(arc_theta),
    z = ya.xpos * sin(arc_theta)
  )

  # outer arc tick marks
  tick_theta <- atan(aty)
  tick_len   <- 0.015 * diff(xlim)
  df_arc_ticks <- data.frame(
    x     = ya.xpos * cos(tick_theta),
    z     = ya.xpos * sin(tick_theta),
    x_end = (ya.xpos + tick_len) * cos(tick_theta),
    z_end = (ya.xpos + tick_len) * sin(tick_theta),
    value = aty,
    label = aty_label_text
  )

  # CI arc
  ci_values    <- c(ci.lb_plot, beta_plot, ci.ub_plot)
  ci_arc_slopes <- seq(ci.lb_plot, ci.ub_plot, length.out = ceiling(arc_res / 4))
  ci_arc_theta  <- atan(ci_arc_slopes)
  df_ci_arc <- data.frame(
    x = ci.xpos * cos(ci_arc_theta),
    z = ci.xpos * sin(ci_arc_theta)
  )

  # CI arc tick marks
  ci_tick_theta <- atan(ci_values)
  ci_tick_len   <- 0.007 * diff(xlim)
  df_ci_ticks <- data.frame(
    x     = (ci.xpos - ci_tick_len) * cos(ci_tick_theta),
    z     = (ci.xpos - ci_tick_len) * sin(ci_tick_theta),
    x_end = (ci.xpos + ci_tick_len) * cos(ci_tick_theta),
    z_end = (ci.xpos + ci_tick_len) * sin(ci_tick_theta),
    value = ci_values,
    type  = c("ci.lb", "beta", "ci.ub")
  )

  return(list(
    # plot elements
    points     = df_points,
    refline    = df_refline,
    band       = df_band,
    band_upper = df_band_upper,
    band_lower = df_band_lower,
    # arcs (trigonometric, for ggplot/as_data)
    arc        = df_arc,
    arc_ticks  = df_arc_ticks,
    ci_arc     = df_ci_arc,
    ci_ticks   = df_ci_ticks,
    # limits and labels
    xlim       = xlim,
    zlim       = zlim,
    xlab       = xlab,
    zlab       = zlab,
    zlab_auto  = zlab_auto,
    atz        = atz,
    # raw values for base R arc computation (aspect-ratio-aware)
    aty        = aty,
    aty_labels = aty_label_text,
    ci_values  = ci_values,
    beta_plot  = beta_plot,
    z_crit     = z_crit,
    xaxismax   = xaxismax,
    ci.xpos    = ci.xpos,
    ya.xpos    = ya.xpos,
    arc_res    = arc_res,
    level      = level,
    center     = center
  ))
}


# ---------------------------------------------------------------------------- #
# .radial_plot_base
# ---------------------------------------------------------------------------- #
#
# Create radial plot using base R graphics, matching metafor::radial.
#
# The arc is drawn outside the plot axis range using xpd = TRUE.
# Arc coordinates are computed using the plot's aspect ratio so the
# arc appears circular on screen (matching metafor's approach).
#
# @param data list of radial plot data from .radial_data()
# @param dots list of graphical parameters
#
# @return NULL invisibly
#
# ---------------------------------------------------------------------------- #
.radial_plot_base <- function(data, dots) {

  # extract graphical parameters
  pch  <- dots[["pch"]]
  col  <- dots[["col"]]
  bg   <- dots[["bg"]]
  cex  <- dots[["cex"]]
  back <- dots[["back"]]
  main <- dots[["main"]]
  las  <- dots[["las"]]

  # extract data components
  df_points  <- data$points
  df_refline <- data$refline
  df_band    <- data$band
  df_band_up <- data$band_upper
  df_band_lo <- data$band_lower
  xlim       <- data$xlim
  zlim       <- data$zlim
  xlab       <- data$xlab
  zlab       <- data$zlab
  zlab_auto  <- data$zlab_auto
  atz        <- data$atz
  xaxismax   <- data$xaxismax

  # raw arc parameters (recomputed with aspect ratio below)
  aty        <- data$aty
  aty_labels <- data$aty_labels
  ci_values  <- data$ci_values
  ci.xpos    <- data$ci.xpos
  ya.xpos    <- data$ya.xpos
  arc_res    <- data$arc_res

  # ---- save and restore graphical parameters ----
  par_mar <- graphics::par("mar")
  par_pty <- graphics::par("pty")
  par_xpd <- graphics::par("xpd")

  # wider margins: +4 left (for fraction zlab), +6 right (for arc labels)
  par_mar_adj <- par_mar + c(0, 4, 0, 6)
  par_mar_adj[par_mar_adj < 1] <- 1
  graphics::par(mar = par_mar_adj, pty = "s")

  on.exit(graphics::par(mar = par_mar, pty = par_pty, xpd = par_xpd), add = TRUE)

  # ---- empty plot: no box, no auto axes, exact limits ----
  graphics::plot(
    NA, NA,
    xlim = xlim,
    ylim = zlim,
    bty  = "n",
    xaxt = "n",
    yaxt = "n",
    xlab = xlab,
    ylab = "",
    xaxs = "i",
    yaxs = "i",
    las  = las
  )

  # title
  if (!is.null(main)) {
    graphics::title(main = main)
  }

  # ---- confidence band (parallelogram) ----
  if (!is.na(back)) {
    graphics::polygon(
      x      = df_band$x,
      y      = df_band$z,
      col    = back,
      border = NA
    )
  }

  # reference line
  graphics::segments(
    df_refline$x[1], df_refline$z[1],
    df_refline$x[2], df_refline$z[2],
    lty = "solid"
  )

  # band edges
  graphics::segments(
    df_band_up$x[1], df_band_up$z[1],
    df_band_up$x[2], df_band_up$z[2],
    lty = "dotted"
  )
  graphics::segments(
    df_band_lo$x[1], df_band_lo$z[1],
    df_band_lo$x[2], df_band_lo$z[2],
    lty = "dotted"
  )

  # ---- axes (manual) ----
  graphics::axis(side = 1, las = las)

  graphics::axis(side = 2, at = atz, labels = atz, las = las)

  # z-axis label via mtext (fraction expression renders properly)
  if (zlab_auto) {
    graphics::mtext(
      zlab, side = 2, line = par_mar_adj[2] - 1,
      at = 0, adj = 0, las = 1
    )
  } else {
    graphics::mtext(
      zlab, side = 2, line = par_mar_adj[2] - 4,
      at = 0
    )
  }

  # ---- arcs (outside plot region, using aspect ratio) ----
  graphics::par(xpd = TRUE)

  par_usr <- graphics::par("usr")
  asp_rat <- (par_usr[4] - par_usr[3]) / (par_usr[2] - par_usr[1])

  # helper: compute (x, z) on circle of radius `len` at slope `s`
  # accounting for aspect ratio (matching metafor)
  .arc_x <- function(len, s) sqrt(len^2 / (1 + (s / asp_rat)^2))
  .arc_z <- function(len, s) .arc_x(len, s) * s

  # ---- outer arc line ----
  arc_slopes <- seq(min(aty), max(aty), length.out = arc_res)
  arc_xi     <- .arc_x(ya.xpos, arc_slopes)
  arc_zi     <- .arc_z(ya.xpos, arc_slopes)

  valid <- arc_zi > zlim[1] & arc_zi < zlim[2]
  if (any(valid)) {
    graphics::lines(arc_xi[valid], arc_zi[valid])
  }

  # ---- outer arc tick marks ----
  tick_len  <- 0.015 * diff(xlim)
  tick_xi_l <- .arc_x(ya.xpos, aty)
  tick_zi_l <- .arc_z(ya.xpos, aty)
  tick_xi_u <- .arc_x(ya.xpos + tick_len, aty)
  tick_zi_u <- .arc_z(ya.xpos + tick_len, aty)

  valid <- tick_zi_l > zlim[1] & tick_zi_u > zlim[1] &
           tick_zi_l < zlim[2] & tick_zi_u < zlim[2]
  if (any(valid)) {
    graphics::segments(
      tick_xi_l[valid], tick_zi_l[valid],
      tick_xi_u[valid], tick_zi_u[valid]
    )
  }

  # ---- outer arc labels ----
  label_len <- ya.xpos + 0.02 * diff(xlim)
  label_xi  <- .arc_x(label_len, aty)
  label_zi  <- .arc_z(label_len, aty)

  valid <- label_zi > zlim[1] & label_zi < zlim[2]
  if (any(valid)) {
    graphics::text(
      label_xi[valid], label_zi[valid],
      aty_labels[valid],
      pos = 4, cex = 0.85, offset = 0.25
    )
  }

  # ---- CI arc line ----
  ci_arc_slopes <- seq(ci_values[1], ci_values[3],
                       length.out = ceiling(arc_res / 4))
  ci_arc_xi     <- .arc_x(ci.xpos, ci_arc_slopes)
  ci_arc_zi     <- .arc_z(ci.xpos, ci_arc_slopes)

  valid <- ci_arc_zi > zlim[1] & ci_arc_zi < zlim[2]
  if (any(valid)) {
    graphics::lines(ci_arc_xi[valid], ci_arc_zi[valid])
  }

  # ---- CI arc tick marks ----
  ci_tick_len  <- 0.007 * diff(xlim)
  ci_tick_xi_l <- .arc_x(ci.xpos - ci_tick_len, ci_values)
  ci_tick_zi_l <- .arc_z(ci.xpos - ci_tick_len, ci_values)
  ci_tick_xi_u <- .arc_x(ci.xpos + ci_tick_len, ci_values)
  ci_tick_zi_u <- .arc_z(ci.xpos + ci_tick_len, ci_values)

  valid <- ci_tick_zi_l > zlim[1] & ci_tick_zi_u > zlim[1] &
           ci_tick_zi_l < zlim[2] & ci_tick_zi_u < zlim[2]
  if (any(valid)) {
    graphics::segments(
      ci_tick_xi_l[valid], ci_tick_zi_l[valid],
      ci_tick_xi_u[valid], ci_tick_zi_u[valid]
    )
  }

  # ---- restore xpd and draw points last (on top) ----
  graphics::par(xpd = par_xpd)

  graphics::points(
    df_points$x, df_points$z,
    pch = pch, col = col, bg = bg, cex = cex
  )

  return(invisible(NULL))
}


# ---------------------------------------------------------------------------- #
# .radial_plot_ggplot
# ---------------------------------------------------------------------------- #
#
# Create radial plot using ggplot2.
#
# Uses pre-computed trigonometric arc coordinates from .radial_data().
# Arcs are drawn outside the panel via coord_cartesian(clip = "off").
#
# @param data list of radial plot data from .radial_data()
# @param dots list of graphical parameters
#
# @return ggplot object
#
# ---------------------------------------------------------------------------- #
.radial_plot_ggplot <- function(data, dots) {

  # extract graphical parameters
  pch  <- dots[["pch"]]
  col  <- dots[["col"]]
  bg   <- dots[["bg"]]
  size <- dots[["size"]]
  back <- dots[["back"]]
  main <- dots[["main"]]

  # extract data components
  df_points    <- data$points
  df_refline   <- data$refline
  df_band      <- data$band
  df_band_up   <- data$band_upper
  df_band_lo   <- data$band_lower
  xlim         <- data$xlim
  zlim         <- data$zlim
  xlab         <- data$xlab
  zlab         <- data$zlab
  atz          <- data$atz

  # raw arc parameters (recomputed with aspect ratio, matching base R)
  aty        <- data$aty
  aty_labels <- data$aty_labels
  ci_values  <- data$ci_values
  ci.xpos    <- data$ci.xpos
  ya.xpos    <- data$ya.xpos
  arc_res    <- data$arc_res

  # ---- compute aspect-ratio-corrected arc coordinates ----
  # matching the base R approach: square panel with data aspect ratio
  asp_rat <- diff(zlim) / diff(xlim)

  # helper: compute (x, z) on circle of radius `len` at slope `s`
  # accounting for aspect ratio (matching metafor / base R)
  .arc_x <- function(len, s) sqrt(len^2 / (1 + (s / asp_rat)^2))
  .arc_z <- function(len, s) .arc_x(len, s) * s

  # ---- outer arc line ----
  arc_slopes <- seq(min(aty), max(aty), length.out = arc_res)
  df_arc <- data.frame(
    x = .arc_x(ya.xpos, arc_slopes),
    z = .arc_z(ya.xpos, arc_slopes)
  )
  valid <- df_arc$z > zlim[1] & df_arc$z < zlim[2]
  df_arc <- df_arc[valid, ]

  # ---- outer arc tick marks ----
  tick_len <- 0.015 * diff(xlim)
  df_arc_ticks <- data.frame(
    x     = .arc_x(ya.xpos, aty),
    z     = .arc_z(ya.xpos, aty),
    x_end = .arc_x(ya.xpos + tick_len, aty),
    z_end = .arc_z(ya.xpos + tick_len, aty),
    value = aty,
    label = aty_labels
  )
  valid <- df_arc_ticks$z > zlim[1] & df_arc_ticks$z_end > zlim[1] &
           df_arc_ticks$z < zlim[2] & df_arc_ticks$z_end < zlim[2]
  df_arc_ticks <- df_arc_ticks[valid, ]

  # ---- outer arc labels ----
  label_len <- ya.xpos + 0.02 * diff(xlim)
  df_arc_labels <- data.frame(
    x     = .arc_x(label_len, aty),
    z     = .arc_z(label_len, aty),
    label = aty_labels
  )
  valid <- df_arc_labels$z > zlim[1] & df_arc_labels$z < zlim[2]
  df_arc_labels <- df_arc_labels[valid, ]

  # ---- CI arc line ----
  ci_arc_slopes <- seq(ci_values[1], ci_values[3],
                       length.out = ceiling(arc_res / 4))
  df_ci_arc <- data.frame(
    x = .arc_x(ci.xpos, ci_arc_slopes),
    z = .arc_z(ci.xpos, ci_arc_slopes)
  )
  valid <- df_ci_arc$z > zlim[1] & df_ci_arc$z < zlim[2]
  df_ci_arc <- df_ci_arc[valid, ]

  # ---- CI arc tick marks ----
  ci_tick_len <- 0.007 * diff(xlim)
  df_ci_ticks <- data.frame(
    x     = .arc_x(ci.xpos - ci_tick_len, ci_values),
    z     = .arc_z(ci.xpos - ci_tick_len, ci_values),
    x_end = .arc_x(ci.xpos + ci_tick_len, ci_values),
    z_end = .arc_z(ci.xpos + ci_tick_len, ci_values),
    value = ci_values,
    type  = c("ci.lb", "beta", "ci.ub")
  )
  valid <- df_ci_ticks$z > zlim[1] & df_ci_ticks$z_end > zlim[1] &
           df_ci_ticks$z < zlim[2] & df_ci_ticks$z_end < zlim[2]
  df_ci_ticks <- df_ci_ticks[valid, ]

  # ---- build plot ----
  out <- ggplot2::ggplot()

  # confidence band (parallelogram)
  if (!is.na(back)) {
    out <- out +
      ggplot2::geom_polygon(
        data    = df_band,
        mapping = ggplot2::aes(x = .data$x, y = .data$z),
        fill    = back
      )
  }

  # confidence band edges
  out <- out +
    ggplot2::geom_line(
      data     = df_band_up,
      mapping  = ggplot2::aes(x = .data$x, y = .data$z),
      linetype = "dotted",
      colour   = "grey50"
    ) +
    ggplot2::geom_line(
      data     = df_band_lo,
      mapping  = ggplot2::aes(x = .data$x, y = .data$z),
      linetype = "dotted",
      colour   = "grey50"
    )

  # reference line
  out <- out +
    ggplot2::geom_line(
      data    = df_refline,
      mapping = ggplot2::aes(x = .data$x, y = .data$z),
      colour  = "black"
    )

  # points
  out <- out +
    ggplot2::geom_point(
      data    = df_points,
      mapping = ggplot2::aes(x = .data$x, y = .data$z),
      shape   = pch,
      colour  = col,
      fill    = bg,
      size    = size
    )

  # outer arc (drawn outside panel via clip = "off")
  out <- out +
    ggplot2::geom_path(
      data    = df_arc,
      mapping = ggplot2::aes(x = .data$x, y = .data$z),
      colour  = "black"
    )

  # arc tick marks
  if (nrow(df_arc_ticks) > 0) {
    out <- out +
      ggplot2::geom_segment(
        data    = df_arc_ticks,
        mapping = ggplot2::aes(
          x    = .data$x,
          y    = .data$z,
          xend = .data$x_end,
          yend = .data$z_end
        ),
        colour = "black"
      )
  }

  # arc tick labels
  if (nrow(df_arc_labels) > 0) {
    out <- out +
      ggplot2::geom_text(
        data    = df_arc_labels,
        mapping = ggplot2::aes(x = .data$x, y = .data$z, label = .data$label),
        hjust   = 0,
        size    = 3
      )
  }

  # CI arc
  out <- out +
    ggplot2::geom_path(
      data      = df_ci_arc,
      mapping   = ggplot2::aes(x = .data$x, y = .data$z),
      colour    = "black",
      linewidth = 0.8
    )

  # CI arc ticks
  if (nrow(df_ci_ticks) > 0) {
    out <- out +
      ggplot2::geom_segment(
        data      = df_ci_ticks,
        mapping   = ggplot2::aes(
          x    = .data$x,
          y    = .data$z,
          xend = .data$x_end,
          yend = .data$z_end
        ),
        colour    = "black",
        linewidth = ifelse(df_ci_ticks$type == "beta", 0.8, 0.5)
      )
  }

  # scales
  out <- out +
    ggplot2::scale_x_continuous(name = xlab) +
    ggplot2::scale_y_continuous(breaks = atz, name = zlab)

  # coord: square panel matching base R pty="s"; arcs outside via clip = "off"
  out <- out +
    ggplot2::coord_cartesian(
      xlim = xlim,
      ylim = zlim,
      clip = "off"
    ) +
    ggplot2::theme_bw() +
    ggplot2::theme(
      aspect.ratio  = 1,
      panel.border  = ggplot2::element_blank(),
      axis.line.x.bottom = ggplot2::element_line(),
      axis.line.y.left   = ggplot2::element_line(),
      plot.margin   = ggplot2::margin(t = 10, r = 60, b = 10, l = 10)
    )

  # title
  if (!is.null(main)) {
    out <- out + ggplot2::ggtitle(main)
  }

  return(out)
}


# ---------------------------------------------------------------------------- #
# .set_dots_radial
# ---------------------------------------------------------------------------- #
#
# Process dots arguments for radial plot with sensible defaults.
#
# @param ... additional graphical arguments
#
# @return list of graphical parameters with defaults applied
#
# ---------------------------------------------------------------------------- #
.set_dots_radial <- function(...) {

  dots <- list(...)

  # point styling
  dots <- .plot_point_style_defaults(dots)

  # confidence band
  if (is.null(dots[["back"]])) dots[["back"]] <- "grey90"
  if (length(dots[["back"]]) != 1L ||
      (!is.na(dots[["back"]]) &&
       !(is.character(dots[["back"]]) || is.numeric(dots[["back"]])))) {
    stop("'back' must be a single color value or NA.", call. = FALSE)
  }

  # arc resolution
  if (is.null(dots[["arc.res"]])) dots[["arc.res"]] <- 100
  BayesTools::check_int(dots[["arc.res"]], "arc.res", lower = 2)

  # title
  if (is.null(dots[["main"]])) dots[["main"]] <- NULL
  if (is.null(dots[["las"]]))  dots[["las"]]  <- 1
  .check_plot_label(dots[["main"]], "main")
  BayesTools::check_int(dots[["las"]], "las", lower = 0)
  if (!dots[["las"]] %in% 0:3) {
    stop("'las' must be one of 0, 1, 2, or 3.", call. = FALSE)
  }

  # data return flag
  if (is.null(dots[["as_data"]])) dots[["as_data"]] <- FALSE
  BayesTools::check_bool(dots[["as_data"]], "as_data")
  .check_plot_positive_scalar(dots[["cex"]], "cex")
  .check_plot_positive_scalar(dots[["size"]], "size")

  return(dots)
}

context("Radial (Galbraith) plot")

# Load common test helpers
source(testthat::test_path("common-functions.R"))
source(testthat::test_path("helper-test-matrix.R"))
source(testthat::test_path("helper-visuals.R"))

# list cached fits lazily
skip_if_no_fits()
skip_if_not_installed("metafor")
fit_names <- list_fits()
fits      <- lazy_fits(fit_names, validate = FALSE)
info      <- lazy_infos(fit_names, validate = FALSE)


# ============================================================================ #
# Test: Simple Meta-Analysis Radial Plot
# ============================================================================ #

test_that("Radial plot for simple meta-analysis matches metafor structure", {

  name        <- "bcg_meta-analysis"
  fit_metafor <- info[[name]][["metafor"]]
  fit_brma    <- fits[[name]]

  # --------------------------------------------------
  # Visual comparison: side-by-side plots
  # --------------------------------------------------

  expect_vdiffr_snapshot("radial_simple_comparison", function() {
    oldpar <- graphics::par(no.readonly = TRUE)
    on.exit(graphics::par(mfrow = oldpar[["mfrow"]], mar = oldpar[["mar"]]))
    par(mfrow = c(1, 2), mar = c(4, 4, 2, 5))
    metafor::radial(fit_metafor, main = "metafor")
    radial(fit_brma, plot_type = "base", main = "brma")
  })

  expect_vdiffr_snapshot("radial_simple_centered_comparison", function() {
    oldpar <- graphics::par(no.readonly = TRUE)
    on.exit(graphics::par(mfrow = oldpar[["mfrow"]], mar = oldpar[["mar"]]))
    par(mfrow = c(1, 2), mar = c(4, 4, 2, 5))
    metafor::radial(fit_metafor, center = TRUE, main = "metafor")
    radial(fit_brma, center = TRUE, plot_type = "base", main = "brma")
  })

  expect_vdiffr_snapshot(
    "radial_simple_brma_ggplot",
    radial(fit_brma, plot_type = "ggplot")
  )

  expect_vdiffr_snapshot(
    "radial_simple_centered_brma_ggplot",
    radial(fit_brma, center = TRUE, plot_type = "ggplot")
  )
})

# ============================================================================ #
# Test: 3-Level Model Radial Plot
# ============================================================================ #

test_that("Radial plot for 3-level model renders brma outputs", {

  name     <- "konstantopoulos2011_3lvl"
  fit_brma <- fits[[name]]

  # metafor does not support radial for rma.mv, brma-only tests

  expect_vdiffr_snapshot("radial_3lvl_brma_base", function() {
    radial(fit_brma, plot_type = "base")
  })

  expect_vdiffr_snapshot(
    "radial_3lvl_brma_ggplot",
    radial(fit_brma, plot_type = "ggplot")
  )
})

# ============================================================================ #
# Test: GLMM Model Radial Plot
# ============================================================================ #

test_that("Radial plot for GLMM model renders brma outputs", {

  name     <- "nielweise2008_glmm"
  fit_brma <- fits[[name]]

  expect_vdiffr_snapshot("radial_glmm_brma_base", function() {
    radial(fit_brma, plot_type = "base")
  })

  expect_vdiffr_snapshot(
    "radial_glmm_brma_ggplot",
    radial(fit_brma, plot_type = "ggplot")
  )
})

# ============================================================================ #
# Test: Selection Model Radial Plot
# ============================================================================ #

test_that("Radial plot for selection model renders brma outputs", {

  name     <- "dat.lehmann2018-3PSM"
  fit_brma <- fits[[name]]

  expect_vdiffr_snapshot("radial_selmodel_brma_base", function() {
    radial(fit_brma, plot_type = "base")
  })

  expect_vdiffr_snapshot(
    "radial_selmodel_brma_ggplot",
    radial(fit_brma, plot_type = "ggplot")
  )
})

# ============================================================================ #
# Test: PET Model Radial Plot
# ============================================================================ #

test_that("Radial plot for PET model renders brma outputs", {

  name     <- "dat.lehmann2018-PET"
  fit_brma <- fits[[name]]

  expect_vdiffr_snapshot("radial_PET_brma_base", function() {
    radial(fit_brma, plot_type = "base")
  })

  expect_vdiffr_snapshot(
    "radial_PET_brma_ggplot",
    radial(fit_brma, plot_type = "ggplot")
  )
})

# ============================================================================ #
# Test: BMA.norm Model Radial Plot
# ============================================================================ #

test_that("Radial plot for BMA.norm model renders base output", {

  name     <- "dat.lehmann2018_BMA.norm"
  fit_brma <- fits[[name]]

  expect_vdiffr_snapshot("radial_BMA", function() {
    suppressWarnings(radial(fit_brma, plot_type = "base"))
  })
})

# ============================================================================ #
# Test: BMA.glmm Model Radial Plot
# ============================================================================ #

test_that("Radial plot for BMA.glmm model renders base output", {

  name     <- "bcg_BMA.glmm"
  fit_brma <- fits[[name]]

  expect_vdiffr_snapshot("radial_BMA.glmm", function() {
    suppressWarnings(radial(fit_brma, plot_type = "base"))
  })
})

# ============================================================================ #
# Test: RoBMA Model Radial Plot
# ============================================================================ #

test_that("Radial plot for RoBMA model renders base output", {

  name     <- "dat.lehmann2018_RoBMA"
  fit_brma <- fits[[name]]

  expect_vdiffr_snapshot("radial_RoBMA", function() {
    suppressWarnings(radial(fit_brma, plot_type = "base"))
  })
})

# ============================================================================ #
# Test: Error on Unsupported Models
# ============================================================================ #

test_that("Radial plot errors on unsupported model types", {

  # meta-regression (has moderators)
  expect_error(
    radial(fits[["bcg_meta-regression"]]),
    "moderators",
    info = "meta-regression model is rejected"
  )

  # location-scale model (has both mods and scale, moderator check fires first)
  expect_error(
    radial(fits[["bangertdrowns2004_location-scale"]]),
    "moderators",
    info = "location-scale model is rejected"
  )
})


# ============================================================================ #
# Test: Radial Plot Interface
# ============================================================================ #

test_that("Radial plot data and alias interface are stable", {

  name     <- "bcg_meta-analysis"
  fit_brma <- fits[[name]]

  # --------------------------------------------------
  # Test as_data = TRUE returns list with expected components
  # --------------------------------------------------

  radial_data <- radial(fit_brma, as_data = TRUE)

  expect_true(is.list(radial_data),
    info = "as_data = TRUE returns a list"
  )

  expected_components <- c(
    "points", "refline", "band", "arc", "arc_ticks",
    "ci_arc", "ci_ticks", "xlim", "zlim"
  )
  expect_true(all(expected_components %in% names(radial_data)),
    info = "radial data contains all expected components"
  )

  # Check points data.frame structure
  expect_true(is.data.frame(radial_data$points),
    info = "points are returned as a data.frame"
  )
  expect_true(all(c("x", "z") %in% names(radial_data$points)),
    info = "points contain x and z columns"
  )

  # Check number of points matches number of studies
  n_studies <- nrow(fit_brma$data$outcome)
  expect_equal(nrow(radial_data$points), n_studies,
    info = "number of points matches number of studies"
  )

  # --------------------------------------------------
  # Test center changes data values
  # --------------------------------------------------

  radial_data_centered <- radial(fit_brma, center = TRUE, as_data = TRUE)

  expect_true(radial_data_centered$center,
    info = "center flag is TRUE when centering"
  )

  # centered z-values should differ from non-centered
  expect_false(
    all(abs(radial_data$points$z - radial_data_centered$points$z) < 1e-10),
    info = "centering changes z-axis values"
  )

  # --------------------------------------------------
  # Test error on invalid plot_type
  # --------------------------------------------------

  expect_error(radial(fit_brma, plot_type = "invalid"),
    info = "invalid plot_type is rejected"
  )

  expect_error(radial(fit_brma, level = 100),
    regexp = "level",
    info = "invalid confidence level is rejected"
  )
  expect_error(radial(fit_brma, xlim = c(1, 1)),
    regexp = "xlim",
    info = "non-increasing xlim is rejected"
  )
  expect_error(radial(fit_brma, arc.res = 1),
    regexp = "arc.res",
    info = "invalid arc resolution is rejected"
  )
  expect_error(radial(fit_brma, transf = 1),
    regexp = "transf",
    info = "invalid transformation is rejected"
  )

  # --------------------------------------------------
  # Test galbraith alias works
  # --------------------------------------------------

  galbraith_data <- galbraith(fit_brma, as_data = TRUE)
  expect_equal(radial_data, galbraith_data,
    info = "galbraith() produces same output as radial()"
  )
})


# ============================================================================ #
# Test: Radial Plot Customization
# ============================================================================ #

test_that("Radial plot customization snapshots are stable", {

  skip_if_not_full_visuals("Customization snapshots are visual-gallery coverage.")

  name     <- "bcg_meta-analysis"
  fit_brma <- fits[[name]]

  # --------------------------------------------------
  # Test custom point aesthetics
  # --------------------------------------------------

  expect_vdiffr_snapshot("radial_custom_points_base", function() {
    radial(fit_brma, plot_type = "base", pch = 21, col = "blue", bg = "lightblue", cex = 1.5)
  })

  expect_vdiffr_snapshot(
    "radial_custom_points_ggplot",
    radial(fit_brma, plot_type = "ggplot", pch = 21, col = "blue", bg = "lightblue", size = 3)
  )

  # --------------------------------------------------
  # Test custom axis labels and title
  # --------------------------------------------------

  expect_vdiffr_snapshot("radial_custom_labels_base", function() {
    radial(fit_brma, plot_type = "base", xlab = "Precision", zlab = "z-score", main = "Radial Plot")
  })

  expect_vdiffr_snapshot(
    "radial_custom_labels_ggplot",
    radial(fit_brma, plot_type = "ggplot", xlab = "Precision", zlab = "z-score", main = "Radial Plot")
  )

  # --------------------------------------------------
  # Test suppressing background
  # --------------------------------------------------

  expect_vdiffr_snapshot("radial_no_background_base", function() {
    radial(fit_brma, plot_type = "base", back = NA)
  })

  expect_vdiffr_snapshot(
    "radial_no_background_ggplot",
    radial(fit_brma, plot_type = "ggplot", back = NA)
  )

  # --------------------------------------------------
  # Test custom level
  # --------------------------------------------------

  expect_vdiffr_snapshot("radial_custom_level_base", function() {
    radial(fit_brma, plot_type = "base", level = 99)
  })

  expect_vdiffr_snapshot(
    "radial_custom_level_ggplot",
    radial(fit_brma, plot_type = "ggplot", level = 99)
  )
})

## ----vignette-nowrap-options, include = FALSE, eval = [encoded data, 9944 characters]
options(width = 10000)

## ----setup, include = [encoded data, 9976 characters]
source("vignette-cache.R", local = knitr::knit_global())

cached_fits <- vignette_cache(
  name     = "v13-metafor-parity-glmm",
  objects  = c("fit1_brma"),
  packages = c("metafor", "metadat")
)

knitr::opts_chunk$set(
  collapse    = TRUE,
  comment     = "#>",
  eval        = vignette_cache_eval(cached_fits),
  message     = FALSE,
  warning     = FALSE,
  fig.width   = 7,
  fig.height  = 3.5,
  dev         = "png",
  fig.retina  = 3
)
if (.Platform$OS.type == "windows") {
  knitr::opts_chunk$set(dev.args = list(type = "cairo"))
}

## ----load-models, include = [encoded data, 9970 characters]
library("metafor")
library("RoBMA")

data("dat.bcg", package = "metadat")

vignette_cache_load(cached_fits)

## ----precompute-models, include = FALSE, eval = [encoded data, 9950 characters]
# library("metafor")
# library("RoBMA")
# 
# data("dat.bcg", package = "metadat")
# 
# fit1_brma <- brma.glmm(
#   ai = tpos, bi = tneg, ci = cpos, di = cneg, measure = "OR",
#   data = dat.bcg, seed = 1
# )
# 
# vignette_cache_save(cached_fits)

## [encoded data, 9997 characters]
data("dat.bcg", package = "metadat")
head(dat.bcg)

## [encoded data, 9997 characters]
fit1_metafor <- metafor::rma.glmm(
  ai = tpos, bi = tneg, ci = cpos, di = cneg, measure = "OR", 
  model = "UM.FS", data = dat.bcg
)
fit1_metafor

## ----fit1-brma, eval = [encoded data, 9975 characters]
# fit1_brma <- brma.glmm(
#   ai = tpos, bi = tneg, ci = cpos, di = cneg, measure = "OR",
#   data = dat.bcg, seed = 1
# )

## [encoded data, 9997 characters]
summary(fit1_brma, include_mcmc_diagnostics = FALSE)

## [encoded data, 9997 characters]
pooled_effect(fit1_brma, transform = "EXP")

#' @title Update a brma Fit
#'
#' @description
#' Extends an existing fitted \code{brma} object by additional MCMC samples,
#' updates study labels, and optionally recomputes cached fit-dependent
#' quantities.
#'
#' @param object a fitted \code{brma} object.
#' @param formula. unsupported; included for compatibility with
#'   \code{\link[stats]{update}}.
#' @param ... unsupported additional arguments.
#' @param sample_extend integer. Number of additional samples per chain.
#' @param slab optional character vector of study labels. Updating labels does
#'   not refit or extend the model.
#' @param autofit_control list of autofit control settings. Values are merged
#'   with the existing settings before extending.
#' @param convergence_checks list of convergence check settings. Values are
#'   merged with the existing settings and used to re-check the fit.
#' @param recompute whether cached \code{loo}, \code{waic}, and
#'   \code{marglik} values already stored in \code{object} should be recomputed
#'   after extension (\code{"all"}) or dropped with a warning (\code{"drop"}).
#' @param parallel logical. Whether to extend chains in parallel.
#' @param cores integer. Number of cores to use when \code{parallel = TRUE}.
#' @param silent logical. Whether to suppress JAGS output during extension.
#' @param seed optional seed used before extending.
#' @param evaluate unsupported; included for compatibility with
#'   \code{\link[stats]{update}}.
#'
#' @return The updated \code{brma} object.
#'
#' @details
#' Extending a fit adds posterior samples only. It does not rerun adaptation or
#' burn-in. Prior, data, and model-structure updates are intentionally not
#' supported by this method.
#'
#' @examples \dontrun{
#' fit <- update(fit, sample_extend = 1000)
#' fit <- update(fit, slab = paste("Study", seq_len(nobs(fit))))
#' }
#'
#' @export
update.brma <- function(
    object, formula. = NULL, ...,
    sample_extend = NULL, slab = NULL,
    autofit_control = NULL, convergence_checks = NULL,
    recompute = c("all", "drop"),
    parallel = NULL, cores = NULL, silent = NULL, seed = NULL,
    evaluate = TRUE) {

  if (!inherits(object, "brma")) {
    stop("'object' must be a 'brma' object.", call. = FALSE)
  }
  if (!is.null(formula.)) {
    stop("Updating formulas is not supported by update.brma().", call. = FALSE)
  }
  if (!isTRUE(evaluate)) {
    stop("update.brma() does not support 'evaluate = FALSE'.", call. = FALSE)
  }

  dots <- list(...)
  .check_unused_dots(
    dots    = dots,
    allowed = character(0),
    caller  = "update.brma()"
  )

  recompute <- match.arg(recompute)

  if (!is.null(sample_extend)) {
    BayesTools::check_int(
      sample_extend,
      "sample_extend",
      lower     = 1,
      allow_NA  = FALSE
    )
  }

  if (!is.null(slab)) {
    object <- .update_brma_slab(object = object, slab = slab)
  }

  object[["fit_control"]] <- .update_fit_control(
    old_fit_control = object[["fit_control"]],
    chains          = NULL,
    adapt           = NULL,
    burnin          = NULL,
    sample          = NULL,
    thin            = NULL,
    autofit         = NULL,
    parallel        = parallel,
    cores           = cores,
    silent          = silent,
    seed            = seed
  )
  object[["autofit_control"]] <- .update_autofit_control(
    old_autofit_control = object[["autofit_control"]],
    autofit_control     = .update_brma_autofit_input(
      autofit_control = autofit_control,
      sample_extend   = sample_extend
    )
  )
  object[["convergence_checks"]] <- .update_convergence_checks(
    old_convergence_checks = object[["convergence_checks"]],
    convergence_checks     = convergence_checks
  )

  fit_extended <- !is.null(sample_extend)
  cached       <- .collect_brma_fit_cache(object)

  if (fit_extended) {
    object <- .extend_brma_fit_once(object)
    object[["summary"]]      <- .object_summary(object)
    object[["coefficients"]] <- .object_coefficients(object)
    object <- .refresh_brma_fit_cache(
      object    = object,
      cached    = cached,
      recompute = recompute
    )
  } else if (!is.null(convergence_checks)) {
    object[["fit"]] <- .recheck_brma_fit(object)
  }

  return(object)
}


# Update study labels without touching the posterior fit.
.update_brma_slab <- function(object, slab) {

  outcome <- object[["data"]][["outcome"]]

  if (length(slab) != nrow(outcome)) {
    stop(
      "The 'slab' argument must have length ",
      nrow(outcome),
      " (same as the fitted data).",
      call. = FALSE
    )
  }

  object[["data"]][["outcome"]][["slab"]] <- as.character(slab)
  attr(object[["data"]], "slab")          <- TRUE

  return(object)
}


# Merge update-time autofit settings before validation.
.update_brma_autofit_input <- function(autofit_control, sample_extend) {

  if (is.null(autofit_control)) {
    autofit_control <- list()
  }
  if (!is.null(sample_extend)) {
    autofit_control[["sample_extend"]] <- sample_extend
  }

  return(autofit_control)
}


# Extend exactly one sample_extend chunk, even when stored max_extend is larger.
.extend_brma_fit_once <- function(object) {

  if (is.null(object[["fit"]]) || length(object[["fit"]]) == 0L) {
    stop("'object' does not contain a fitted model to extend.", call. = FALSE)
  }

  stored_autofit_control <- object[["autofit_control"]]
  extend_autofit_control <- stored_autofit_control
  extend_autofit_control[["max_extend"]] <- 1
  extend_autofit_control <- BayesTools::JAGS_check_and_list_autofit_settings(
    autofit_control = extend_autofit_control
  )

  object[["autofit_control"]] <- extend_autofit_control
  object[["fit"]]             <- .fit(object, extend = TRUE)
  object[["autofit_control"]] <- stored_autofit_control

  return(object)
}


# Re-run convergence checks on the current posterior samples.
.recheck_brma_fit <- function(object) {

  fit <- object[["fit"]]

  if (is.null(fit) || length(fit) == 0L) {
    stop("'object' does not contain a fitted model to check.", call. = FALSE)
  }

  prior_list <- attr(fit, "prior_list")
  if (is.null(prior_list)) {
    prior_list <- .create_fit_priors(
      data   = object[["data"]],
      priors = object[["priors"]]
    )
  }

  check_fit <- BayesTools::JAGS_check_convergence(
    fit          = fit,
    prior_list   = prior_list,
    max_Rhat     = object[["convergence_checks"]][["max_Rhat"]],
    min_ESS      = object[["convergence_checks"]][["min_ESS"]],
    max_error    = object[["convergence_checks"]][["max_error"]],
    max_SD_error = object[["convergence_checks"]][["max_SD_error"]]
  )

  fit[["converged"]]     <- check_fit
  fit[["has_posterior"]] <- TRUE
  fit[["warnings"]]      <- c(attr(fit, "warnings"), attr(check_fit, "errors"))

  return(fit)
}


# Record which fit-dependent quantities need cache handling after extension.
.collect_brma_fit_cache <- function(object) {

  cached <- list(
    loo     = object[["loo"]],
    waic    = object[["waic"]],
    marglik = object[["marglik"]]
  )

  return(cached)
}


# Recompute or drop fit-dependent cached quantities after extension.
.refresh_brma_fit_cache <- function(object, cached, recompute) {

  cached_names <- .brma_cached_names(cached)

  if (length(cached_names) == 0L) {
    return(object)
  }

  object <- .drop_brma_fit_cache(object)

  if (recompute == "drop") {
    warning(
      "Dropping cached ",
      paste(cached_names, collapse = ", "),
      " because the fit was extended.",
      call. = FALSE
    )
    return(object)
  }

  object <- .recompute_brma_fit_cache(
    object = object,
    cached = cached
  )

  return(object)
}


# Names of cached fit-dependent quantities present in the original object.
.brma_cached_names <- function(cached) {

  cached_names <- character(0)

  if (!is.null(cached[["loo"]])) {
    cached_names <- c(cached_names, "loo")
  }
  if (!is.null(cached[["waic"]])) {
    cached_names <- c(cached_names, "waic")
  }
  if (!is.null(cached[["marglik"]])) {
    cached_names <- c(cached_names, "marglik")
  }

  return(cached_names)
}


# Remove cached fit-dependent quantities.
.drop_brma_fit_cache <- function(object) {

  object[["loo"]]     <- NULL
  object[["waic"]]    <- NULL
  object[["marglik"]] <- NULL

  return(object)
}


# Recompute the same cached quantities and units that existed before extension.
.recompute_brma_fit_cache <- function(object, cached) {

  loo_units <- .brma_cache_units(cached[["loo"]])
  for (unit in loo_units) {
    object <- add_loo(object, unit = unit)
  }

  waic_units <- .brma_cache_units(cached[["waic"]])
  for (unit in waic_units) {
    object <- add_waic(object, unit = unit)
  }

  if (!is.null(cached[["marglik"]]) && !inherits(object, "RoBMA")) {
    object <- add_marglik(object)
  }

  return(object)
}


# Extract cached LOO/WAIC unit names, accepting a single unnamed legacy cache.
.brma_cache_units <- function(cache) {

  if (is.null(cache)) {
    return(character(0))
  }

  units <- names(cache)
  if (is.null(units)) {
    if (length(cache) == 1L) {
      return("estimate")
    }
    stop("Cached LOO/WAIC objects must be named by unit.", call. = FALSE)
  }

  units <- units[nzchar(units)]

  return(units)
}

# ============================================================================ #
# outcome-helpers.R
# ============================================================================ #
#
# Shared outcome-data extraction helpers used by residuals, diagnostics, plots,
# likelihood evaluation, and zplot code.
#
# ============================================================================ #


# ---------------------------------------------------------------------------- #
# .outcome_data_yi
# ---------------------------------------------------------------------------- #
#
# Extract or compute observed effect sizes from brma object.
#
# For normal models, yi is directly available in the data.
# For GLMM models, yi is computed from raw frequency data using formulas
# equivalent to metafor::escalc with default zero-cell adjustment (add = 0.5).
#
# @param object brma object
#
# @return numeric vector of observed effect sizes
#
# ---------------------------------------------------------------------------- #
.outcome_data_yi <- function(object) {

  outcome_type <- .outcome_type(object)
  outcome_data <- object[["data"]][["outcome"]]

  if (outcome_type == "norm") {
    # normal models: yi is directly available
    return(outcome_data[["yi"]])
  } else if (outcome_type == "bin") {
    # binomial models: compute log odds ratio from 2x2 table
    # using metafor::escalc formula with default zero-cell adjustment (add = 0.5)
    ai <- outcome_data[["ai"]]
    bi <- outcome_data[["n1i"]] - outcome_data[["ai"]] # bi = n1i - ai
    ci <- outcome_data[["ci"]]
    di <- outcome_data[["n2i"]] - outcome_data[["ci"]] # di = n2i - ci
    add <- 0.5

    # vectorized computation with conditional zero-cell adjustment
    needs_adj <- (ai == 0 | bi == 0 | ci == 0 | di == 0)
    ai_adj <- ai + add * needs_adj
    bi_adj <- bi + add * needs_adj
    ci_adj <- ci + add * needs_adj
    di_adj <- di + add * needs_adj

    return(log((ai_adj * di_adj) / (bi_adj * ci_adj)))
  } else if (outcome_type == "pois") {
    # Poisson models: compute log incidence rate ratio
    x1i <- outcome_data[["x1i"]]
    t1i <- outcome_data[["t1i"]]
    x2i <- outcome_data[["x2i"]]
    t2i <- outcome_data[["t2i"]]
    add <- 0.5

    # vectorized computation with conditional zero-cell adjustment
    needs_adj <- (x1i == 0 | x2i == 0)
    x1i_adj <- x1i + add * needs_adj
    x2i_adj <- x2i + add * needs_adj

    return(log((x1i_adj / t1i) / (x2i_adj / t2i)))
  }
}


# ---------------------------------------------------------------------------- #
# .outcome_data_sei
# ---------------------------------------------------------------------------- #
#
# Extract or compute sampling standard errors from brma object.
#
# For normal models, sei is directly available in the data.
# For GLMM models, sei is computed from raw frequency data using formulas
# equivalent to metafor::escalc with default zero-cell adjustment (add = 0.5).
#
# @param object brma object
#
# @return numeric vector of sampling standard errors
#
# ---------------------------------------------------------------------------- #
.outcome_data_sei <- function(object) {

  outcome_type <- .outcome_type(object)
  outcome_data <- object[["data"]][["outcome"]]

  if (outcome_type == "norm") {
    # normal models: sei is directly available
    return(outcome_data[["sei"]])
  } else if (outcome_type == "bin") {
    # binomial models: compute approximate SE for log odds ratio
    # SE(logOR) = sqrt(1/ai + 1/bi + 1/ci + 1/di) with zero-cell adjustment
    ai <- outcome_data[["ai"]]
    bi <- outcome_data[["n1i"]] - outcome_data[["ai"]]
    ci <- outcome_data[["ci"]]
    di <- outcome_data[["n2i"]] - outcome_data[["ci"]]
    add <- 0.5

    # vectorized computation with conditional zero-cell adjustment
    needs_adj <- (ai == 0 | bi == 0 | ci == 0 | di == 0)
    ai_adj <- ai + add * needs_adj
    bi_adj <- bi + add * needs_adj
    ci_adj <- ci + add * needs_adj
    di_adj <- di + add * needs_adj

    return(sqrt(1 / ai_adj + 1 / bi_adj + 1 / ci_adj + 1 / di_adj))
  } else if (outcome_type == "pois") {
    # Poisson models: compute approximate SE for log IRR
    # SE(logIRR) = sqrt(1/x1i + 1/x2i)
    x1i <- outcome_data[["x1i"]]
    x2i <- outcome_data[["x2i"]]
    add <- 0.5

    # vectorized computation with conditional zero-cell adjustment
    needs_adj <- (x1i == 0 | x2i == 0)
    x1i_adj <- x1i + add * needs_adj
    x2i_adj <- x2i + add * needs_adj

    return(sqrt(1 / x1i_adj + 1 / x2i_adj))
  }
}


# ---------------------------------------------------------------------------- #
# .outcome_data_vi
# ---------------------------------------------------------------------------- #
#
# Extract or compute sampling variances from brma object.
#
# Convenience wrapper that returns sei^2 from .outcome_data_sei().
#
# @param object brma object
#
# @return numeric vector of sampling variances
#
# ---------------------------------------------------------------------------- #
.outcome_data_vi <- function(object) {

  return(.outcome_data_sei(object)^2)
}


# ---------------------------------------------------------------------------- #
# .outcome_data_weights
# ---------------------------------------------------------------------------- #
#
# Extract likelihood weights from the outcome data, defaulting to unit weights.
#
# ---------------------------------------------------------------------------- #
.outcome_data_weights <- function(object) {

  outcome_data <- object[["data"]][["outcome"]]

  if (.is_weights(object)) {
    return(as.numeric(outcome_data[["weights"]]))
  }

  return(rep(1, nrow(outcome_data)))
}

#' @title Bayesian Meta-Analysis
#'
#' @description Function for fitting normal-likelihood/effect-size
#' random-effects, meta-regression, multilevel, and location-scale
#' meta-analytic models. `brma.norm()` is an alias for `brma()`; raw-count
#' GLMM models use `brma.glmm()`.
#'
#' @inheritParams data_input
#' @inheritParams prior_specification
#' @inheritParams fitting_specification
#' @details
#' ## Prior distributions
#' Prior distributions must be specified for all model parameters.
#' This typically includes the pooled effect \eqn{\mu} and between-study heterogeneity
#' \eqn{\tau}. In the case of meta-regression, the pooled effect \eqn{\mu} corresponds to
#' the intercept, and additional prior distributions for the regression coefficients
#' are required. In the case of a location-scale model, the between-study heterogeneity
#' corresponds to the intercept of the scale regression, and additional prior
#' distributions for the scale regression coefficients are required.
#'
#' There are several ways to specify the prior distributions: \enumerate{
#'    \item{via a standardized effect size `measure` with known unit information standard deviation,}
#'    \item{by estimating unit information standard deviation using sample sizes `ni`,}
#'    \item{by manually setting `prior_unit_information_sd`,}
#'    \item{by specifying informed empirical prior distributions via `prior_informed_field`
#'    and `prior_informed_subfield`,}
#'    \item{or via fully custom specification using the `prior_effect`, `prior_heterogeneity`,
#'    `prior_mods`, `prior_scale`, and `prior_heterogeneity_allocation` arguments.}
#' }
#' In all cases, the prior behavior can be further modified by the `rescale_priors`,
#' `standardize_continuous_predictors`, and `set_contrast_factor_predictors` arguments.
#'
#' ### (1) Specifying prior distributions via standardized effect size measures with known
#' unit information standard deviation
#' This is the easiest way to specify prior distributions. The width of prior
#' distributions is based on a fraction of the known unit information standard deviation (`UISD`)
#' \insertCite{rover2021weakly}{RoBMA}. The default prior distributions for the parameters
#' are set as follows:\tabular{ll}{
#'    effect size:               \tab Normal(0, \eqn{\frac{1}{2}} `UISD`)  \cr
#'    heterogeneity:             \tab Normal+(0, \eqn{\frac{1}{4}} `UISD`) \cr
#'    effect moderation:         \tab Normal(0, \eqn{\frac{1}{4}} `UISD`)  \cr
#'    heterogeneity moderation:  \tab Normal(0, \eqn{\frac{1}{2}})
#' }
#' The heterogeneity moderation parameters are multiplicative, as such they are independent of `UISD`.
#'
#' The default fraction of the `UISD` can be changed via `RoBMA.options()` using one of
#' the following arguments: `"default_UISD.effect"`, `"default_UISD.heterogeneity"`,
#' `"default_UISD.mods"`, `"default_UISD.scale"`.
#'
#' The known `UISD` for standardized effect size measures (`measure`) are set as follows:
#' \tabular{ll}{
#'   `"SMD"`:   \tab \eqn{\sqrt{2}} \cr
#'   `"ZCOR"`:  \tab \eqn{1}        \cr
#'   `"RR"`:    \tab \eqn{\sqrt{4}} \cr
#'   `"OR"`:    \tab \eqn{\sqrt{4}} \cr
#'   `"HR"`:    \tab \eqn{\sqrt{4}} \cr
#'   `"IRR"`:   \tab \eqn{\sqrt{4}}
#' }
#' See Chapter 2.4 in \insertCite{spiegelhalter2004bayesian;textual}{RoBMA} and
#' Chapter 1 in \insertCite{grieve2022hybrid;textual}{RoBMA}.
#'
#' ### (2) Estimating unit information standard deviation using sample sizes
#' When effect sizes are on a non-standardized scale (`measure = "GEN"`) or use a
#' standardized effect size without known `UISD`, the `UISD` can be estimated from
#' sample sizes (`ni`) and standard errors (`sei`) following Equation 6 in
#' \insertCite{rover2021weakly;textual}{RoBMA}. The estimated `UISD` is then used to
#' scale the default prior distributions as described in section (1).
#'
#' Note that the known `UISD` for standardized effect size measures (section (1)) is
#' used if available, even when `ni` is provided.
#'
#' ### (3) Manually setting unit information standard deviation
#' Alternatively, the `UISD` can be specified directly via the `prior_unit_information_sd`
#' argument. This is useful when the appropriate scale for prior distributions is known
#' a priori or when multiple analyses are to be performed on subsets of the same data
#' (re-estimating `UISD` on different subsets of the data can lead to slightly different
#' prior distributions for different subsets; see [estimate_unit_information_sd()]).
#' The specified `prior_unit_information_sd` is then used to scale the default prior
#' distributions as described in section (1).
#'
#' Note that the manually specified `prior_unit_information_sd` takes precedence over
#' the estimated `UISD` from `ni` (section (2)) and the known `UISD` from `measure`
#' (section (1)). It cannot be combined with `prior_informed_field`.
#'
#' ### (4) Specifying informed empirical prior distributions
#' Informed prior distributions can be specified via the `prior_informed_field` and
#' `prior_informed_subfield` arguments. Currently, only `prior_informed_field = "medicine"`
#' with subfields defined in `BayesTools::prior_informed_medicine_names` is supported,
#' which uses empirically derived prior distributions from medical meta-analyses as described in
#' \insertCite{bartos2021bayesian;textual}{RoBMA} and
#' \insertCite{bartos2023empirical;textual}{RoBMA}.
#'
#' When `prior_informed_field = "medicine"`, the default `prior_informed_subfield` is
#' `"Cochrane"` (i.e., using the whole CDSR database as a reference). The informed
#' prior distributions are available for the following effect size measures:
#' \tabular{ll}{
#'   `"SMD"`:  \tab standardized mean difference \cr
#'   `"OR"`:   \tab log odds ratio               \cr
#'   `"RR"`:   \tab log risk ratio               \cr
#'   `"RD"`:   \tab risk difference              \cr
#'   `"HR"`:   \tab log hazard ratio
#' }
#'
#' Note that informed prior distributions are only available for the effect size (\eqn{\mu})
#' and heterogeneity (\eqn{\tau}) parameters. For effect moderation (`prior_mods`), the
#' informed effect prior is scaled by a factor of `RoBMA.get_option("default_informed_priors.mods")`.
#' For heterogeneity moderation (`prior_scale`), a normal prior with standard deviation
#' specified by `RoBMA.get_option("default_informed_priors.scale")` is used.
#'
#' ### (5) Fully custom prior distributions
#' Prior distributions can be fully customized by directly specifying the `prior_effect`,
#' `prior_heterogeneity`, `prior_mods`, `prior_scale`, and `prior_heterogeneity_allocation`
#' arguments. These should be prior distribution objects created via `BayesTools::prior()`
#' or related functions (e.g., `prior_factor()`).
#'
#' ### Rescaling prior distributions
#' The `rescale_priors` argument allows rescaling supported prior distributions by a
#' multiplicative factor. For example, `rescale_priors = 2` doubles the standard
#' deviations/scales of normal, Cauchy, t, and inverse-gamma prior distributions,
#' making them more diffuse. Point and none priors are unchanged.
#'
#' ### Handling of continuous and factor predictors
#' When `standardize_continuous_predictors = TRUE`, continuous moderator and
#' scale predictors are internally standardized before fitting. Reported
#' summaries are transformed back to the original predictor scale by default;
#' use `standardized_coefficients = TRUE` in `summary()` or related methods to
#' inspect coefficients on the standardized scale.
#'
#' Factor predictors use the contrast specified by
#' `set_contrast_factor_predictors`. The default `"treatment"` follows standard
#' treatment coding. Model-averaging functions default to `"meandif"` so
#' inclusion priors for factor levels are centered on deviations from the grand
#' mean.
#'
#' @return A fitted object of class `c("brma.norm", "brma")`. The object
#' contains checked `data`, checked `priors`, the JAGS `fit`, cached `summary`,
#' and cached `coefficients`. If the corresponding package options are enabled,
#' it can also contain cached LOO, WAIC, or marginal likelihood results. The
#' advanced internal `only_data = TRUE` and `only_priors = TRUE` paths return
#' partially constructed objects.
#'
#' @examples \dontrun{
#' if (requireNamespace("metadat", quietly = TRUE) &&
#'     requireNamespace("metafor", quietly = TRUE)) {
#'   data(dat.bcg, package = "metadat")
#'   dat <- metafor::escalc(
#'     measure = "RR",
#'     ai      = tpos,
#'     bi      = tneg,
#'     ci      = cpos,
#'     di      = cneg,
#'     data    = dat.bcg
#'   )
#'
#'   fit <- brma(
#'     yi      = yi,
#'     vi      = vi,
#'     mods    = ~ ablat + year,
#'     data    = dat,
#'     measure = "RR",
#'     seed    = 1,
#'     silent  = TRUE
#'   )
#'
#'   summary(fit)
#'   predict(fit, type = "terms")
#' }
#' }
#'
#' @references
#' \insertAllCited{}
#'
#' @seealso [RoBMA()], [BMA()], [brma.glmm()], [summary.brma()],
#' [plot.brma()], [predict.brma()]
#'
#' @aliases brma.norm
#' @export brma.norm
#' @export
brma <- brma.norm <- function(
    # input specification
    yi, vi, sei, weights, ni,
    mods, scale, cluster,
    data, slab, subset,
    measure,

    # prior specification
    prior_effect, prior_heterogeneity, prior_mods, prior_scale,
    prior_heterogeneity_allocation,
    standardize_continuous_predictors = TRUE,
    set_contrast_factor_predictors = "treatment",
    prior_unit_information_sd, rescale_priors = 1,
    prior_informed_field, prior_informed_subfield,

    # MCMC fitting settings
    sample = 5000, burnin = 2000, adapt = 500,
    chains = 3, thin = 1, parallel = FALSE,
    autofit = FALSE, autofit_control = set_autofit_control(),
    convergence_checks = set_convergence_checks(),

    # additional settings
    seed = NULL, silent, ...
    ) {

  ### create the output object
  dots            <- list(...)
  missing_measure <- missing(measure)
  if (missing_measure && !isTRUE(dots[["only_data"]])) {
    .stop_missing_measure("brma.norm()")
  }
  if (missing_measure) {
    measure <- "GEN"
  }
  dots            <- .validate_constructor_dots(dots, caller = "brma.norm()")
  object          <- .createObject(
    dots = dots, class = c("brma.norm", "brma"),
    # MCMC and fitting settings
    chains = chains, adapt = adapt, burnin = burnin, sample = sample, thin = thin,
    autofit = autofit, parallel = parallel, silent = silent, seed = seed,
    autofit_control = autofit_control, convergence_checks = convergence_checks
  )

  ### check and store the data
  object$data <- .check_and_list_data(
    .call = match.call(), .envir = parent.frame(), class = "norm",
    set_contrast_factor_predictors = set_contrast_factor_predictors,
    standardize_continuous_predictors = standardize_continuous_predictors,
    measure = measure)
  if (isTRUE(dots[["only_data"]]))
    return(object)

  ### check and store priors
  object$priors <- .check_and_list_priors.brma(
    prior_effect = prior_effect, prior_heterogeneity = prior_heterogeneity,
    prior_mods = prior_mods, prior_scale = prior_scale,
    prior_heterogeneity_allocation = prior_heterogeneity_allocation,
    rescale_priors                    = rescale_priors,
    prior_unit_information_sd         = prior_unit_information_sd,
    prior_informed_field              = prior_informed_field,
    prior_informed_subfield           = prior_informed_subfield,
    data = object[["data"]])
  if (isTRUE(dots[["only_priors"]]))
    return(.set_only_priors_class(object))

  ### fit the model
  object$fit <- .fit(object)

  ### store simple summary & coefficients
  object$summary       <- .object_summary(object)
  object$coefficients  <- .object_coefficients(object)

  object               <- .autocompute_brma(object)

  return(object)
}

# ============================================================================ #
# Outcome PDF Functions (Log-Likelihood)
# ============================================================================ #
#
# These functions compute pointwise log-likelihoods for each observation and
# posterior sample. They are used for LOO-PSIS diagnostics and model comparison.
#
# Parallels the structure of rng.R but returns density values instead of
# random samples.
#
# Note: For binomial and Poisson models, each "observation" consists of a pair
# of data points (ai+ci or x1i+x2i) that together define a single effect size
# estimate. The log-likelihood for the observation is the sum of log-likelihoods
# for both components.
#
# ============================================================================ #


# ---------------------------------------------------------------------------- #
# .rowLogSumExps
# ---------------------------------------------------------------------------- #
#
# Compute log(sum(exp(x))) for each row of a matrix, using the log-sum-exp
# trick for numerical stability.
#
# @param lx  numeric matrix of log-values
#
# @return numeric vector of length nrow(lx)
#
# ---------------------------------------------------------------------------- #
.rowLogSumExps <- function(lx) {
  if (!is.matrix(lx)) {
    lx <- as.matrix(lx)
  }

  if (!anyNA(lx)) {
    row_max <- lx[cbind(seq_len(nrow(lx)), max.col(lx, ties.method = "first"))]
    out     <- row_max

    finite_rows <- is.finite(row_max)
    if (any(finite_rows)) {
      out[finite_rows] <- row_max[finite_rows] +
        log(rowSums(exp(lx[finite_rows, , drop = FALSE] - row_max[finite_rows])))
    }

    return(out)
  }

  has_value <- rowSums(!is.na(lx)) > 0
  lx[is.na(lx)] <- -Inf

  row_max <- lx[cbind(seq_len(nrow(lx)), max.col(lx, ties.method = "first"))]
  out     <- rep(NA_real_, nrow(lx))

  infinite_rows <- has_value & is.infinite(row_max)
  out[infinite_rows] <- row_max[infinite_rows]

  finite_rows <- has_value & is.finite(row_max)
  if (any(finite_rows)) {
    out[finite_rows] <- row_max[finite_rows] +
      log(rowSums(exp(lx[finite_rows, , drop = FALSE] - row_max[finite_rows])))
  }

  return(out)
}


# ---------------------------------------------------------------------------- #
# .apply_log_lik_weights
# ---------------------------------------------------------------------------- #
#
# Match JAGS weighted-density semantics by multiplying log-likelihood
# contributions by the user-supplied data weights.
#
# @param log_lik S x K matrix of log-likelihood values.
# @param weights numeric vector of length K, or NULL.
#
# @return weighted log-likelihood matrix.
#
# ---------------------------------------------------------------------------- #
.apply_log_lik_weights <- function(log_lik, weights) {

  if (is.null(weights)) {
    return(log_lik)
  }

  if (length(weights) != ncol(log_lik)) {
    stop("Data weights length does not match the log-likelihood matrix.",
         call. = FALSE)
  }

  return(sweep(log_lik, 2, weights, "*"))
}


# ---------------------------------------------------------------------------- #
# .get_log_lik_data_weights
# ---------------------------------------------------------------------------- #
#
# @param object brma object.
#
# @return numeric vector of data weights, or NULL.
#
# ---------------------------------------------------------------------------- #
.get_log_lik_data_weights <- function(object) {

  if (!.is_weights(object)) {
    return(NULL)
  }

  return(object[["data"]][["outcome"]][["weights"]])
}


# ---------------------------------------------------------------------------- #
# .glmm_likelihood_weights
# ---------------------------------------------------------------------------- #
#
# Validate and default GLMM pair likelihood weights.
#
# ---------------------------------------------------------------------------- #
.glmm_likelihood_weights <- function(weights, K) {

  if (is.null(weights)) {
    return(rep(1, K))
  }

  BayesTools::check_real(
    weights, "weights",
    check_length = K, allow_NULL = FALSE, allow_NA = FALSE,
    lower = 0, allow_bound = FALSE
  )

  return(as.numeric(weights))
}


# ---------------------------------------------------------------------------- #
# .outcome_pdf.norm
# ---------------------------------------------------------------------------- #
#
# Compute pointwise log-likelihoods for normal outcome models.
#
# For normal outcome models, the observed effect y_i has likelihood:
#   y_i ~ N(mu_i, tau_within^2 + se_i^2)
#
# This function computes the log-density for each observation at each posterior
# sample.
#
# @param yi               numeric vector of length K; observed effect sizes
# @param mu_samples       S x K matrix of location samples (with cluster effects if multilevel)
# @param tau_within       S x K matrix of estimate-level heterogeneity samples
# @param sei              numeric vector of length K; standard errors
#
# @return S x K matrix of log-likelihood values
#
# ---------------------------------------------------------------------------- #
.outcome_pdf.norm <- function(yi, mu_samples, tau_within, sei) {

  S <- nrow(mu_samples)
  K <- ncol(mu_samples)

  # replicate yi and sei across samples: matrix(vec, S, K, byrow = TRUE)
  yi_mat  <- matrix(yi,  nrow = S, ncol = K, byrow = TRUE)
  sei_mat <- matrix(sei, nrow = S, ncol = K, byrow = TRUE)

  # compute total SD: sqrt(tau^2 + se^2)
  total_sd <- sqrt(tau_within^2 + sei_mat^2)

  # compute log-likelihood for each cell
  log_lik <- stats::dnorm(yi_mat, mean = mu_samples, sd = total_sd, log = TRUE)

  return(log_lik)
}


# ---------------------------------------------------------------------------- #
# .outcome_pdf.selnorm
# ---------------------------------------------------------------------------- #
#
# Selected-normal likelihood using the compiled p-bin kernel.
#
# ---------------------------------------------------------------------------- #
.outcome_pdf.selnorm <- function(yi, mu_samples, tau_within, sei,
                                 selection_context, weights = NULL) {

  S       <- nrow(mu_samples)
  K       <- ncol(mu_samples)
  sei_mat <- matrix(sei, nrow = S, ncol = K, byrow = TRUE)
  total_sd <- sqrt(tau_within^2 + sei_mat^2)

  return(.selnorm_kernel_loglik_matrix(
    yi             = yi,
    mu_num         = mu_samples,
    sigma_num      = total_sd,
    mu_norm        = mu_samples,
    sigma_norm     = total_sd,
    sei            = sei,
    omega          = selection_context[["omega"]],
    selection_spec = selection_context,
    alpha          = selection_context[["alpha"]],
    phack_kind     = selection_context[["phack_kind"]],
    kernel_mode    = selection_context[["kernel_mode"]],
    weights        = weights
  ))
}


# ---------------------------------------------------------------------------- #
# .has_native_glmm
# ---------------------------------------------------------------------------- #
#
# @param outcome_type character; GLMM outcome type, either "bin" or "pois".
#
# @return TRUE when native GLMM likelihood kernels are available for the
# requested outcome type.
#
# ---------------------------------------------------------------------------- #
.has_native_glmm <- function(outcome_type) {

  if (outcome_type == "bin") {
    return(is.loaded("RoBMA_glmm_binom_marginal_loglik", PACKAGE = "RoBMA"))
  }
  if (outcome_type == "pois") {
    return(is.loaded("RoBMA_glmm_pois_marginal_loglik", PACKAGE = "RoBMA"))
  }

  stop("Unknown GLMM outcome type.", call. = FALSE)
}

.has_native_glmm_cluster <- function() {

  return(
    is.loaded("RoBMA_glmm_binom_cluster_loglik", PACKAGE = "RoBMA") &&
      is.loaded("RoBMA_glmm_pois_cluster_loglik",  PACKAGE = "RoBMA")
  )
}


# ---------------------------------------------------------------------------- #
# .cluster_indices_flatten
# ---------------------------------------------------------------------------- #
#
# Flatten a cluster index list while preserving list order.
#
# ---------------------------------------------------------------------------- #
.cluster_indices_flatten <- function(cluster_indices) {

  return(list(
    index = as.integer(unlist(cluster_indices, use.names = FALSE)),
    size  = as.integer(lengths(cluster_indices))
  ))
}


# ---------------------------------------------------------------------------- #
# .gauss_legendre_nodes
# ---------------------------------------------------------------------------- #
#
# Compute Gauss-Legendre quadrature nodes and weights on (0, 1).
#
# ---------------------------------------------------------------------------- #
.gauss_legendre_nodes_cache <- new.env(parent = emptyenv())

.gauss_legendre_nodes <- function(n) {

  key <- as.character(n)
  if (exists(key, envir = .gauss_legendre_nodes_cache, inherits = FALSE)) {
    return(get(key, envir = .gauss_legendre_nodes_cache, inherits = FALSE))
  }

  i <- seq_len(n - 1)
  b <- i / sqrt(4 * i^2 - 1)

  J <- diag(0, n)
  J[cbind(i, i + 1)] <- b
  J[cbind(i + 1, i)] <- b

  eig <- eigen(J, symmetric = TRUE)

  nodes_raw   <- eig$values
  weights_raw <- 2 * eig$vectors[1, ]^2

  nodes   <- (nodes_raw + 1) / 2
  weights <- weights_raw / 2

  ord <- order(nodes)

  out <- list(
    nodes       = nodes[ord],
    weights     = weights[ord],
    log_weights = log(weights[ord])
  )
  assign(key, out, envir = .gauss_legendre_nodes_cache)

  return(out)
}


# ---------------------------------------------------------------------------- #
# .glmm_prior_quantile_grid
# ---------------------------------------------------------------------------- #
#
# Construct quadrature nodes under a prior using the probability integral
# transform. The returned weights integrate with respect to the prior measure.
#
# ---------------------------------------------------------------------------- #
.glmm_prior_quantile_grid <- function(prior, n) {

  gl      <- .gauss_legendre_nodes(n)
  q_grid  <- as.numeric(BayesTools::quant(prior, gl[["nodes"]]))

  if (anyNA(q_grid) || any(!is.finite(q_grid))) {
    stop("The GLMM nuisance prior produced non-finite quadrature nodes.",
         call. = FALSE)
  }

  return(list(
    grid        = q_grid,
    log_weights = gl[["log_weights"]]
  ))
}


# ---------------------------------------------------------------------------- #
# .glmm_binom_logit_pi_grid
# ---------------------------------------------------------------------------- #
#
# Construct prior-scale nuisance grids for binomial GLMM likelihoods.
#
# ---------------------------------------------------------------------------- #
.glmm_binom_logit_pi_grid <- function(ai, ci, n1i, n2i, prior_pi, n_pi) {

  K       <- length(ai)
  pi_grid <- .glmm_prior_quantile_grid(prior = prior_pi, n = n_pi)

  if (any(pi_grid[["grid"]] < 0 | pi_grid[["grid"]] > 1)) {
    stop("The binomial GLMM baserate prior produced nodes outside [0, 1].",
         call. = FALSE)
  }

  pi_nodes <- pmin(pmax(pi_grid[["grid"]], .Machine$double.eps),
                   1 - .Machine$double.eps)

  logit_pi_grid  <- matrix(qlogis(pi_nodes), nrow = n_pi, ncol = K)
  log_pi_weights <- matrix(pi_grid[["log_weights"]], nrow = n_pi, ncol = K)

  return(list(
    grid        = logit_pi_grid,
    log_weights = log_pi_weights
  ))
}


# ---------------------------------------------------------------------------- #
# .glmm_pois_log_phi_grid
# ---------------------------------------------------------------------------- #
#
# Construct prior-scale nuisance grids for Poisson GLMM likelihoods.
#
# ---------------------------------------------------------------------------- #
.glmm_pois_log_phi_grid <- function(x1i, x2i, t1i, t2i, prior_phi, n_phi) {

  K        <- length(x1i)
  phi_grid <- .glmm_prior_quantile_grid(prior = prior_phi, n = n_phi)

  log_phi_grid    <- matrix(phi_grid[["grid"]], nrow = n_phi, ncol = K)
  log_phi_weights <- matrix(phi_grid[["log_weights"]], nrow = n_phi, ncol = K)

  return(list(
    grid        = log_phi_grid,
    log_weights = log_phi_weights
  ))
}


# ---------------------------------------------------------------------------- #
# .outcome_pdf.binom
# ---------------------------------------------------------------------------- #
#
# Native wrapper for binomial GLMM marginal log-likelihoods.
#
# ---------------------------------------------------------------------------- #
.outcome_pdf.binom <- function(ai, ci, n1i, n2i, mu_samples, tau_within, prior_pi,
                               weights = NULL, n_theta = 15, n_pi = 30) {

  if (!.has_native_glmm("bin")) {
    return(.outcome_pdf.binom_r(
      ai         = ai,
      ci         = ci,
      n1i        = n1i,
      n2i        = n2i,
      mu_samples = mu_samples,
      tau_within = tau_within,
      prior_pi   = prior_pi,
      weights    = weights,
      n_theta    = n_theta,
      n_pi       = n_pi
    ))
  }

  gh      <- .gauss_hermite_nodes(n_theta)
  pi_grid <- .glmm_binom_logit_pi_grid(
    ai       = ai,
    ci       = ci,
    n1i      = n1i,
    n2i      = n2i,
    prior_pi = prior_pi,
    n_pi     = n_pi
  )
  weights_arg <- if (is.null(weights)) NULL else .native_numeric_vector(
    .glmm_likelihood_weights(weights, ncol(mu_samples))
  )

  return(.Call(
    "RoBMA_glmm_binom_marginal_loglik",
    .native_integer_vector(ai),
    .native_integer_vector(ci),
    .native_integer_vector(n1i),
    .native_integer_vector(n2i),
    .native_numeric_matrix(mu_samples),
    .native_numeric_matrix(tau_within),
    weights_arg,
    .native_numeric_vector(gh[["nodes"]]),
    .native_numeric_vector(gh[["log_weights"]]),
    .native_numeric_matrix(pi_grid[["grid"]]),
    .native_numeric_matrix(pi_grid[["log_weights"]]),
    PACKAGE = "RoBMA"
  ))
}


# ---------------------------------------------------------------------------- #
# .outcome_pdf.binom
# ---------------------------------------------------------------------------- #
#
# Compute pointwise marginal log-likelihoods for binomial outcome models.
#
# For binomial GLMM models, we need to compute the marginal likelihood by
# integrating over the latent parameters (theta, pi):
#   log p(data) = log ∫∫ p(ai, ci | theta, pi, mu) p(theta) p(pi) d(theta) d(pi)
#
# This function uses Gauss-Hermite quadrature for the theta dimension (N(0,1)
# prior) which requires far fewer points than quantile spacing for similar
# accuracy. The pi dimension uses Gauss-Legendre quadrature on the prior CDF
# scale, so the integration covers the full baserate prior support.
#
# The integration over samples is vectorized for efficiency: instead of looping
# over S samples, we compute all samples simultaneously using matrix operations.
#
# Performance optimizations:
# - Gauss-Hermite quadrature for theta (15 points vs 30 for same accuracy)
# - Pre-computed log-binomial coefficients (avoids repeated lgamma calls)
# - Direct log-likelihood calculation without dbinom overhead
# - Log-sum-exp trick for numerical stability
# - Minimized memory allocations in inner loop
#
# @param ai               integer vector of length K; events in treatment group
# @param ci               integer vector of length K; events in control group
# @param n1i              integer vector of length K; treatment group sizes
# @param n2i              integer vector of length K; control group sizes
# @param mu_samples       S x K matrix of log-odds ratio samples (without theta contribution)
# @param tau_within       S x K matrix of estimate-level heterogeneity samples
# @param prior_pi         BayesTools prior object for pi
# @param n_theta          integer; number of Gauss-Hermite points for theta (default: 15)
# @param n_pi             integer; number of grid points for pi dimension (default: 30)
#
# @return S x K matrix of marginal log-likelihood values (one per estimate)
#
# ---------------------------------------------------------------------------- #
.outcome_pdf.binom_r <- function(ai, ci, n1i, n2i, mu_samples, tau_within, prior_pi,
                                 weights = NULL, n_theta = 15, n_pi = 30) {

  S <- nrow(mu_samples)
  K <- ncol(mu_samples)
  weights <- .glmm_likelihood_weights(weights, K)

  # initialize output matrix
  log_lik <- matrix(NA_real_, nrow = S, ncol = K)

  # --- 1. SETUP THETA GRID (Gauss-Hermite Quadrature) ---
  gh <- .gauss_hermite_nodes(n_theta)
  theta_grid        <- gh$nodes
  log_theta_weights <- gh$log_weights

  pi_quadrature <- .glmm_binom_logit_pi_grid(
    ai       = ai,
    ci       = ci,
    n1i      = n1i,
    n2i      = n2i,
    prior_pi = prior_pi,
    n_pi     = n_pi
  )

  # Expand theta grid for vectorized computation
  theta_expanded <- rep(theta_grid, times = n_pi)  # length G

  # --- 2. PRE-COMPUTE BINOMIAL COEFFICIENTS (Outside Loop) ---
  log_binom_coef_a <- lgamma(n1i + 1) - lgamma(ai + 1) - lgamma(n1i - ai + 1)
  log_binom_coef_c <- lgamma(n2i + 1) - lgamma(ci + 1) - lgamma(n2i - ci + 1)

  # --- 3. LOOP OVER OBSERVATIONS ---
  for (k in seq_len(K)) {

    lpi_grid       <- pi_quadrature[["grid"]][, k]
    log_pi_weights <- pi_quadrature[["log_weights"]][, k]

    # Create log-weight vector for all grid points
    log_weights <- as.vector(outer(log_theta_weights, log_pi_weights, "+"))

    # Expand pi grid
    lpi_expanded <- rep(lpi_grid, each = n_theta)

    # --- 4. VECTORIZED INTEGRATION ---
    effect_mat <- mu_samples[, k] + outer(tau_within[, k], theta_expanded)

    half_effect <- 0.5 * effect_mat
    logit_p1    <- sweep(half_effect, 2, lpi_expanded, "+")
    logit_p2    <- sweep(-half_effect, 2, lpi_expanded, "+")

    log_p1   <- plogis(logit_p1, log.p = TRUE)
    log_1mp1 <- plogis(logit_p1, lower.tail = FALSE, log.p = TRUE)
    log_p2   <- plogis(logit_p2, log.p = TRUE)
    log_1mp2 <- plogis(logit_p2, lower.tail = FALSE, log.p = TRUE)

    log_lik_grid <- log_binom_coef_a[k] + ai[k] * log_p1 + (n1i[k] - ai[k]) * log_1mp1 +
                    log_binom_coef_c[k] + ci[k] * log_p2 + (n2i[k] - ci[k]) * log_1mp2
    log_lik_grid <- weights[k] * log_lik_grid

    log_terms <- sweep(log_lik_grid, 2, log_weights, "+")
    log_lik[, k] <- .rowLogSumExps(log_terms)
  }

  return(log_lik)
}


# ---------------------------------------------------------------------------- #
# .outcome_pdf.binom_conditional
# ---------------------------------------------------------------------------- #
#
# Compute conditional log-likelihoods for binomial outcome models.
#
# This path is used by bridge sampling where pi and theta are sampled model
# parameters and should not be integrated out.
#
# @param ai             integer vector of length K; events in treatment group
# @param ci             integer vector of length K; events in control group
# @param n1i            integer vector of length K; treatment group sizes
# @param n2i            integer vector of length K; control group sizes
# @param mu_samples     S x K matrix of log-odds ratio samples
# @param logit_baserate S x K matrix of logit base-rate samples
#
# @return S x K matrix of conditional log-likelihood values
#
# ---------------------------------------------------------------------------- #
.outcome_pdf.binom_conditional <- function(ai, ci, n1i, n2i, mu_samples,
                                           logit_baserate) {

  S <- nrow(mu_samples)
  K <- ncol(mu_samples)

  logit_p1 <- logit_baserate + 0.5 * mu_samples
  logit_p2 <- logit_baserate - 0.5 * mu_samples

  log_lik <- matrix(
    stats::dbinom(
      x    = rep(ai, each = S),
      size = rep(n1i, each = S),
      prob = as.vector(.inv_logit(logit_p1)),
      log  = TRUE
    ) + stats::dbinom(
      x    = rep(ci, each = S),
      size = rep(n2i, each = S),
      prob = as.vector(.inv_logit(logit_p2)),
      log  = TRUE
    ),
    nrow = S,
    ncol = K
  )

  return(log_lik)
}


# ---------------------------------------------------------------------------- #
# .gauss_hermite_nodes
# ---------------------------------------------------------------------------- #
#
# Compute Gauss-Hermite quadrature nodes and weights for standard normal.
#
# Gauss-Hermite quadrature integrates ∫ f(x) exp(-x²) dx exactly for polynomials
# up to degree 2n-1. For the standard normal N(0,1), we transform:
#   ∫ f(z) * (1/√(2π)) * exp(-z²/2) dz
#
# Using substitution u = z/√2, this becomes:
#   (1/√π) ∫ f(√2 * u) * exp(-u²) du
#
# So for N(0,1): nodes = √2 * GH_nodes, and we need to normalize weights
# so they sum to 1 (representing probability mass).
#
# This function uses the Golub-Welsch algorithm (eigenvalue method).
#
# @param n integer; number of quadrature points
#
# @return list with components:
#   - nodes: n quadrature points (transformed for N(0,1))
#   - weights: n quadrature weights (sum to 1 for N(0,1))
#   - log_weights: log of weights for numerical stability
#
# ---------------------------------------------------------------------------- #
.gauss_hermite_nodes_cache <- new.env(parent = emptyenv())

.gauss_hermite_nodes <- function(n) {

  key <- as.character(n)
  if (exists(key, envir = .gauss_hermite_nodes_cache, inherits = FALSE)) {
    return(get(key, envir = .gauss_hermite_nodes_cache, inherits = FALSE))
  }

  # Golub-Welsch algorithm: eigenvalues of symmetric tridiagonal matrix

  # For Gauss-Hermite (physicist's), the recurrence coefficients are:
  #   α_k = 0, β_k = k/2 for k = 1, ..., n-1
  # The Jacobi matrix has off-diagonal elements √β_k = √(k/2)

  i <- seq_len(n - 1)
  b <- sqrt(i / 2)

  # Build symmetric tridiagonal matrix
  H <- diag(0, n)
  H[cbind(i, i + 1)] <- b
  H[cbind(i + 1, i)] <- b

  # Eigendecomposition
  eig <- eigen(H, symmetric = TRUE)

  # Nodes are eigenvalues
  # Weights: w_i = μ₀ * (first component of eigenvector i)²
  # where μ₀ = ∫ exp(-x²) dx = √π
  nodes_raw   <- eig$values
  weights_raw <- sqrt(pi) * eig$vectors[1, ]^2

  # Transform for standard normal N(0,1):
  # ∫ f(z) φ(z) dz = (1/√π) ∫ f(√2 u) exp(-u²) du ≈ (1/√π) Σ w_i f(√2 u_i)
  # So: nodes = √2 * raw_nodes, weights = raw_weights / √π
  nodes   <- sqrt(2) * nodes_raw
  weights <- weights_raw / sqrt(pi)

  # Verify: weights should sum to 1
  # (The raw weights sum to √π because μ₀ = √π and Σ v_{1i}² = 1 for orthonormal eigenvectors)

  # Sort by nodes (eigen returns in decreasing order of eigenvalue)
  ord <- order(nodes)

  out <- list(
    nodes       = nodes[ord],
    weights     = weights[ord],
    log_weights = log(weights[ord])
  )
  assign(key, out, envir = .gauss_hermite_nodes_cache)

  return(out)
}


# ---------------------------------------------------------------------------- #
# .outcome_pdf.pois
# ---------------------------------------------------------------------------- #
#
# Native wrapper for Poisson GLMM marginal log-likelihoods.
#
# ---------------------------------------------------------------------------- #
.outcome_pdf.pois <- function(x1i, x2i, t1i, t2i, mu_samples, tau_within, prior_phi,
                              weights = NULL, n_theta = 15, n_phi = 30) {

  if (!.has_native_glmm("pois")) {
    return(.outcome_pdf.pois_r(
      x1i        = x1i,
      x2i        = x2i,
      t1i        = t1i,
      t2i        = t2i,
      mu_samples = mu_samples,
      tau_within = tau_within,
      prior_phi  = prior_phi,
      weights    = weights,
      n_theta    = n_theta,
      n_phi      = n_phi
    ))
  }

  gh       <- .gauss_hermite_nodes(n_theta)
  phi_grid <- .glmm_pois_log_phi_grid(
    x1i      = x1i,
    x2i      = x2i,
    t1i      = t1i,
    t2i      = t2i,
    prior_phi = prior_phi,
    n_phi    = n_phi
  )
  weights_arg <- if (is.null(weights)) NULL else .native_numeric_vector(
    .glmm_likelihood_weights(weights, ncol(mu_samples))
  )

  return(.Call(
    "RoBMA_glmm_pois_marginal_loglik",
    .native_integer_vector(x1i),
    .native_integer_vector(x2i),
    .native_numeric_vector(t1i),
    .native_numeric_vector(t2i),
    .native_numeric_matrix(mu_samples),
    .native_numeric_matrix(tau_within),
    weights_arg,
    .native_numeric_vector(gh[["nodes"]]),
    .native_numeric_vector(gh[["log_weights"]]),
    .native_numeric_matrix(phi_grid[["grid"]]),
    .native_numeric_matrix(phi_grid[["log_weights"]]),
    PACKAGE = "RoBMA"
  ))
}


# ---------------------------------------------------------------------------- #
# .outcome_pdf.pois
# ---------------------------------------------------------------------------- #
#
# Compute pointwise marginal log-likelihoods for Poisson outcome models.
#
# For Poisson GLMM models, we need to compute the marginal likelihood by
# integrating over the latent parameters (theta, phi):
#   log p(data) = log ∫∫ p(x1i, x2i | theta, phi, mu) p(theta) p(phi) d(theta) d(phi)
#
# This function uses Gauss-Hermite quadrature for the theta dimension (N(0,1)
# prior) which requires far fewer points than quantile spacing for similar
# accuracy. The phi dimension uses Gauss-Legendre quadrature on the prior CDF
# scale, so the integration covers the full log-rate prior support.
#
# The integration over samples is vectorized for efficiency: instead of looping
# over S samples, we compute all samples simultaneously using matrix operations.
#
# Performance optimizations:
# - Gauss-Hermite quadrature for theta (15 points vs 30 for same accuracy)
# - Pre-computed log-factorial terms (avoids repeated lgamma calls)
# - Direct log-likelihood calculation without dpois overhead
# - Log-sum-exp trick for numerical stability
# - Minimized memory allocations in inner loop
#
# @param x1i              integer vector of length K; events in treatment group
# @param x2i              integer vector of length K; events in control group
# @param t1i              numeric vector of length K; treatment exposure times
# @param t2i              numeric vector of length K; control exposure times
# @param mu_samples       S x K matrix of log incidence rate ratio samples (without theta contribution)
# @param tau_within       S x K matrix of estimate-level heterogeneity samples
# @param prior_phi        BayesTools prior object for phi
# @param n_theta          integer; number of Gauss-Hermite points for theta (default: 15)
# @param n_phi            integer; number of grid points for phi dimension (default: 30)
#
# @return S x K matrix of marginal log-likelihood values (one per estimate)
#
# ---------------------------------------------------------------------------- #
.outcome_pdf.pois_r <- function(x1i, x2i, t1i, t2i, mu_samples, tau_within, prior_phi,
                                weights = NULL, n_theta = 15, n_phi = 30) {

  S <- nrow(mu_samples)
  K <- ncol(mu_samples)
  weights <- .glmm_likelihood_weights(weights, K)

  # initialize output matrix
  log_lik <- matrix(NA_real_, nrow = S, ncol = K)

  # --- 1. SETUP THETA GRID (Gauss-Hermite Quadrature) ---
  gh <- .gauss_hermite_nodes(n_theta)
  theta_grid        <- gh$nodes
  log_theta_weights <- gh$log_weights

  phi_quadrature <- .glmm_pois_log_phi_grid(
    x1i       = x1i,
    x2i       = x2i,
    t1i       = t1i,
    t2i       = t2i,
    prior_phi = prior_phi,
    n_phi     = n_phi
  )

  # Expand theta grid for vectorized computation
  theta_expanded <- rep(theta_grid, times = n_phi)  # length G

  # --- 2. PRE-COMPUTE LOG-FACTORIAL TERMS (Outside Loop) ---
  log_factorial_x1 <- lgamma(x1i + 1)
  log_factorial_x2 <- lgamma(x2i + 1)
  log_t1i <- log(t1i)
  log_t2i <- log(t2i)

  # --- 3. LOOP OVER OBSERVATIONS ---
  for (k in seq_len(K)) {

    phi_grid        <- phi_quadrature[["grid"]][, k]
    log_phi_weights <- phi_quadrature[["log_weights"]][, k]

    # Create log-weight vector for all grid points
    log_weights <- as.vector(outer(log_theta_weights, log_phi_weights, "+"))

    # Expand phi grid
    phi_expanded <- rep(phi_grid, each = n_theta)

    # --- 4. VECTORIZED INTEGRATION ---
    effect_mat <- mu_samples[, k] + outer(tau_within[, k], theta_expanded)

    half_effect <- 0.5 * effect_mat
    log_lambda1 <- sweep(half_effect, 2, phi_expanded + log_t1i[k], "+")
    log_lambda2 <- sweep(-half_effect, 2, phi_expanded + log_t2i[k], "+")

    log_lik_grid <- x1i[k] * log_lambda1 - exp(log_lambda1) - log_factorial_x1[k] +
                    x2i[k] * log_lambda2 - exp(log_lambda2) - log_factorial_x2[k]
    log_lik_grid <- weights[k] * log_lik_grid

    log_terms <- sweep(log_lik_grid, 2, log_weights, "+")
    log_lik[, k] <- .rowLogSumExps(log_terms)
  }

  return(log_lik)
}


# ---------------------------------------------------------------------------- #
# .outcome_pdf.pois_conditional
# ---------------------------------------------------------------------------- #
#
# Compute conditional log-likelihoods for Poisson outcome models.
#
# This path is used by bridge sampling where phi and theta are sampled model
# parameters and should not be integrated out.
#
# @param x1i        integer vector of length K; treatment events
# @param x2i        integer vector of length K; control events
# @param t1i        numeric vector of length K; treatment exposure times
# @param t2i        numeric vector of length K; control exposure times
# @param mu_samples S x K matrix of log incidence rate ratio samples
# @param log_phi    S x K matrix of midpoint log-rate samples
#
# @return S x K matrix of conditional log-likelihood values
#
# ---------------------------------------------------------------------------- #
.outcome_pdf.pois_conditional <- function(x1i, x2i, t1i, t2i, mu_samples,
                                          log_phi) {

  S <- nrow(mu_samples)
  K <- ncol(mu_samples)

  log_t1i <- matrix(log(t1i), nrow = S, ncol = K, byrow = TRUE)
  log_t2i <- matrix(log(t2i), nrow = S, ncol = K, byrow = TRUE)

  log_lambda1 <- log_phi + 0.5 * mu_samples + log_t1i
  log_lambda2 <- log_phi - 0.5 * mu_samples + log_t2i

  log_lik <- matrix(
    stats::dpois(
      x      = rep(x1i, each = S),
      lambda = as.vector(exp(log_lambda1)),
      log    = TRUE
    ) + stats::dpois(
      x      = rep(x2i, each = S),
      lambda = as.vector(exp(log_lambda2)),
      log    = TRUE
    ),
    nrow = S,
    ncol = K
  )

  return(log_lik)
}


# ---------------------------------------------------------------------------- #
# .get_cluster_likelihood_n_gamma
# ---------------------------------------------------------------------------- #
#
# @return integer; number of Gauss-Hermite nodes for cluster likelihoods.
#
# ---------------------------------------------------------------------------- #
.get_cluster_likelihood_n_gamma <- function() {

  n_gamma <- RoBMA.get_option("cluster_likelihood.n_gamma")
  BayesTools::check_int(n_gamma, "cluster_likelihood.n_gamma", lower = 3)

  return(as.integer(n_gamma))
}


# ---------------------------------------------------------------------------- #
# .log_lik_cluster_gamma_quadrature
# ---------------------------------------------------------------------------- #
#
# Integrate the held-out cluster effect gamma_g with Gauss-Hermite quadrature.
#
# @param cluster_indices     list of cluster index vectors.
# @param mu_samples          S x K matrix of fixed-effect means.
# @param tau_between_samples S x K matrix of cluster-level SDs.
# @param log_lik_fun         function(idx, mu_node) returning S x length(idx)
#                            conditional log-likelihood matrix.
# @param weights             optional data weights.
# @param n_gamma             number of Gauss-Hermite nodes.
#
# @return S x G cluster-unit log-likelihood matrix.
#
# ---------------------------------------------------------------------------- #
.log_lik_cluster_gamma_quadrature <- function(cluster_indices, mu_samples,
                                              tau_between_samples, log_lik_fun,
                                              weights = NULL,
                                              n_gamma = .get_cluster_likelihood_n_gamma()) {

  gh      <- .gauss_hermite_nodes(n_gamma)
  S       <- nrow(mu_samples)
  G       <- length(cluster_indices)
  log_lik <- matrix(NA_real_, nrow = S, ncol = G)

  for (g in seq_along(cluster_indices)) {
    idx       <- cluster_indices[[g]]
    log_terms <- matrix(NA_real_, nrow = S, ncol = n_gamma)

    for (j in seq_len(n_gamma)) {
      mu_node <- mu_samples[, idx, drop = FALSE] +
        gh$nodes[j] * tau_between_samples[, idx, drop = FALSE]

      point_log_lik <- log_lik_fun(idx, mu_node)
      point_log_lik <- .apply_log_lik_weights(point_log_lik, weights[idx])

      log_terms[, j] <- rowSums(point_log_lik) + gh$log_weights[j]
    }

    log_lik[, g] <- .rowLogSumExps(log_terms)
  }

  return(log_lik)
}


# ---------------------------------------------------------------------------- #
# .log_lik.brma
# ---------------------------------------------------------------------------- #
#
# Compute the log-likelihood matrix for a brma object.
#
# @param object brma object.
# @param unit   character; output/deletion unit.
#
# @return S x K or S x G matrix of log-likelihood values.
#
# ---------------------------------------------------------------------------- #
.log_lik.brma <- function(object, unit = "estimate") {

  unit <- .normalize_unit(unit)

  if (unit == "estimate") {
    return(.log_lik_estimate.brma(object))
  } else {
    return(.log_lik_cluster.brma(object))
  }
}


# ---------------------------------------------------------------------------- #
# .pdf.brma
# ---------------------------------------------------------------------------- #
#
# Back-end wrapper kept for internal callers that use the old helper name.
#
# @param object brma object.
# @param unit   character; output/deletion unit.
#
# @return S x K or S x G matrix of log-likelihood values.
#
# ---------------------------------------------------------------------------- #
.pdf.brma <- function(object, unit = "estimate") {

  return(.log_lik.brma(object = object, unit = unit))
}


# ---------------------------------------------------------------------------- #
# .estimate_likelihood_setup.brma
# ---------------------------------------------------------------------------- #
#
# Extract common posterior quantities for estimate-unit predictive likelihoods.
#
# The estimate-unit target conditions on fitted cluster effects for multilevel
# models and integrates over estimate-level heterogeneity.
#
# @param object brma object.
#
# @return list with model metadata and posterior matrices.
#
# ---------------------------------------------------------------------------- #
.estimate_likelihood_setup.brma <- function(object) {

  priors            <- object[["priors"]]
  data              <- object[["data"]]
  is_multilevel     <- .is_multilevel(object)
  is_mods           <- .is_mods(object)
  is_scale          <- .is_scale(object)
  is_PET            <- .is_PET(object)
  is_PEESE          <- .is_PEESE(object)
  is_weightfunction <- .is_weightfunction(object)
  outcome_type      <- .outcome_type(object)
  effect_direction  <- .effect_direction(object)
  posterior_samples <- .get_posterior_samples(object[["fit"]])

  yi  <- .outcome_data_yi(object)
  sei <- .outcome_data_sei(object)
  K   <- length(yi)

  tau_result <- .evaluate.brma.tau(
    fit               = object[["fit"]],
    scale_data        = data[["scale"]],
    scale_formula     = if (is_scale) .create_fit_formula_list(data = data, "scale") else NULL,
    scale_priors      = priors[["scale"]],
    is_scale          = is_scale,
    is_multilevel     = is_multilevel,
    K                 = K,
    posterior_samples = posterior_samples
  )

  mu_samples <- .evaluate.brma.mu(
    fit               = object[["fit"]],
    outcome_data      = data[["outcome"]],
    mods_data         = data[["mods"]],
    mods_formula      = if (is_mods) .create_fit_formula_list(data = data, "mods") else NULL,
    mods_priors       = priors[["mods"]],
    is_mods           = is_mods,
    is_PET            = is_PET,
    is_PEESE          = is_PEESE,
    effect_direction  = effect_direction,
    bias_adjusted     = FALSE,
    K                 = K,
    posterior_samples = posterior_samples
  )

  fit_data <- NULL
  if (is_multilevel || is_weightfunction) {
    fit_data <- .create_fit_data(data = data, priors = priors)
  }

  if (is_multilevel) {
    mu_samples <- mu_samples + .evaluate.brma.cluster_effects(
      fit               = object[["fit"]],
      tau_between       = tau_result[["tau_between"]],
      cluster           = fit_data[["cluster"]],
      same_data         = TRUE,
      effect_direction  = effect_direction,
      posterior_samples = posterior_samples
    )
  }

  return(list(
    priors            = priors,
    data              = data,
    fit_data          = fit_data,
    yi                = yi,
    sei               = sei,
    K                 = K,
    S                 = nrow(mu_samples),
    mu                = mu_samples,
    tau_within        = tau_result[["tau_within"]],
    tau_between       = tau_result[["tau_between"]],
    is_weightfunction = is_weightfunction,
    outcome_type      = outcome_type,
    effect_direction  = effect_direction,
    posterior_samples = posterior_samples
  ))
}


# ---------------------------------------------------------------------------- #
# .log_lik_estimate.brma
# ---------------------------------------------------------------------------- #
#
# Estimate-unit likelihood, one contribution per effect-size estimate.
#
# For normal multilevel models this is the model likelihood factor conditional
# on the cluster effect and marginal over estimate-level heterogeneity.
#
# @param object brma object.
#
# @return S x K matrix of log-likelihood values.
#
# ---------------------------------------------------------------------------- #
.log_lik_estimate.brma <- function(object) {

  setup <- .estimate_likelihood_setup.brma(object)

  priors            <- setup[["priors"]]
  data              <- setup[["data"]]
  yi                <- setup[["yi"]]
  sei               <- setup[["sei"]]
  K                 <- setup[["K"]]
  mu_samples         <- setup[["mu"]]
  tau_within_samples <- setup[["tau_within"]]
  is_weightfunction <- setup[["is_weightfunction"]]
  outcome_type      <- setup[["outcome_type"]]
  effect_direction  <- setup[["effect_direction"]]
  data_weights      <- .get_log_lik_data_weights(object)

  ### extract posterior samples once if needed
  if (is_weightfunction) {
    posterior_samples <- setup[["posterior_samples"]]
  }

  ### compute log-likelihood based on outcome type
  if (outcome_type == "norm") {

    # flip for negative effect direction
    if (effect_direction == "negative") {
      mu_samples_pdf <- -mu_samples
      yi_pdf         <- -yi
    } else {
      mu_samples_pdf <- mu_samples
      yi_pdf         <- yi
    }

    # dispatch between weighted and standard normal
    if (is_weightfunction) {
      selection_context <- .selection_context(
        object            = object,
        posterior_samples = posterior_samples
      )

      log_lik <- .outcome_pdf.selnorm(
        yi                = yi,
        mu_samples        = mu_samples,
        tau_within        = tau_within_samples,
        sei               = sei,
        selection_context = selection_context
      )

    } else {

      # standard normal likelihood
      log_lik <- .outcome_pdf.norm(
        yi         = yi_pdf,
        mu_samples = mu_samples_pdf,
        tau_within = tau_within_samples,
        sei        = sei
      )

    }

  } else if (outcome_type == "bin") {

    # compute marginal log-likelihood via integration over theta and pi
    # the new .outcome_pdf.binom handles integration internally
    log_lik <- .outcome_pdf.binom(
      ai         = data[["outcome"]][["ai"]],
      ci         = data[["outcome"]][["ci"]],
      n1i        = data[["outcome"]][["n1i"]],
      n2i        = data[["outcome"]][["n2i"]],
      mu_samples = mu_samples,
      tau_within = tau_within_samples,
      prior_pi   = priors[["outcome"]][["pi"]],
      weights    = data_weights
    )


  } else if (outcome_type == "pois") {

    # compute marginal log-likelihood via integration over theta and phi
    # the new .outcome_pdf.pois handles integration internally
    log_lik <- .outcome_pdf.pois(
      x1i        = data[["outcome"]][["x1i"]],
      x2i        = data[["outcome"]][["x2i"]],
      t1i        = data[["outcome"]][["t1i"]],
      t2i        = data[["outcome"]][["t2i"]],
      mu_samples = mu_samples,
      tau_within = tau_within_samples,
      prior_phi  = priors[["outcome"]][["phi"]],
      weights    = data_weights
    )

  } else {

    stop("Unsupported outcome type for estimate-unit log-likelihood.",
         call. = FALSE)

  }

  # add column names and target metadata
  if (outcome_type == "norm") {
    log_lik <- .apply_log_lik_weights(log_lik, data_weights)
  }

  colnames(log_lik) <- paste0("log_lik[", seq_len(K), "]")
  attr(log_lik, "RoBMA_target") <- list(
    unit               = "estimate",
    conditioning_depth = "estimate",
    n                  = K,
    targets            = seq_len(K),
    data_hash          = .get_outcome_hash(object)
  )

  return(log_lik)
}


# ---------------------------------------------------------------------------- #
# .log_lik_cluster.brma
# ---------------------------------------------------------------------------- #
#
# Cluster-unit likelihood for multilevel models.
#
# Each column is the joint held-out-cluster log-likelihood contribution.
#
# @param object brma object.
#
# @return S x G matrix of log-likelihood values.
#
# ---------------------------------------------------------------------------- #
.log_lik_cluster.brma <- function(object) {

  if (!.is_multilevel(object)) {
    stop("Cluster-unit log-likelihood is only available for multilevel models.",
         call. = FALSE)
  }

  outcome_type      <- .outcome_type(object)
  is_weightfunction <- .is_weightfunction(object)
  is_weights        <- .is_weights(object)

  if (outcome_type == "norm" && !is_weightfunction && !is_weights) {
    return(.log_lik_cluster_norm.brma(object))
  } else if (outcome_type == "norm") {
    return(.log_lik_cluster_norm_quadrature.brma(object))
  } else if (outcome_type %in% c("bin", "pois")) {
    return(.log_lik_cluster_glmm.brma(object))
  } else {
    stop("Unsupported outcome type for cluster-unit log-likelihood.",
         call. = FALSE)
  }
}


# ---------------------------------------------------------------------------- #
# .log_lik_cluster_setup.brma
# ---------------------------------------------------------------------------- #
#
# Extract common posterior matrices for cluster-unit likelihoods.
#
# @param object brma object.
#
# @return list with model metadata and posterior matrices.
#
# ---------------------------------------------------------------------------- #
.log_lik_cluster_setup.brma <- function(object) {

  priors           <- object[["priors"]]
  data             <- object[["data"]]
  is_mods          <- .is_mods(object)
  is_scale          <- .is_scale(object)
  is_PET            <- .is_PET(object)
  is_PEESE          <- .is_PEESE(object)
  effect_direction  <- .effect_direction(object)
  K                 <- nrow(data[["outcome"]])
  posterior_samples <- .get_posterior_samples(object[["fit"]])

  mu_samples <- .evaluate.brma.mu(
    fit               = object[["fit"]],
    outcome_data      = data[["outcome"]],
    mods_data         = data[["mods"]],
    mods_formula      = if (is_mods) .create_fit_formula_list(data = data, "mods") else NULL,
    mods_priors       = priors[["mods"]],
    is_mods           = is_mods,
    is_PET            = is_PET,
    is_PEESE          = is_PEESE,
    effect_direction  = effect_direction,
    bias_adjusted     = FALSE,
    K                 = K,
    posterior_samples = posterior_samples
  )

  tau_result <- .evaluate.brma.tau(
    fit               = object[["fit"]],
    scale_data        = data[["scale"]],
    scale_formula     = if (is_scale) .create_fit_formula_list(data = data, "scale") else NULL,
    scale_priors      = priors[["scale"]],
    is_scale          = is_scale,
    is_multilevel     = TRUE,
    K                 = K,
    posterior_samples = posterior_samples
  )

  return(list(
    priors            = priors,
    data              = data,
    K                 = K,
    S                 = nrow(mu_samples),
    mu                = mu_samples,
    tau_within        = tau_result[["tau_within"]],
    tau_between       = tau_result[["tau_between"]],
    cluster           = .get_cluster_indices(object),
    weights           = .get_log_lik_data_weights(object),
    data_hash         = .get_outcome_hash(object),
    effect_direction  = effect_direction,
    posterior_samples = posterior_samples
  ))
}


# ---------------------------------------------------------------------------- #
# .add_cluster_log_lik_metadata
# ---------------------------------------------------------------------------- #
#
# @param log_lik         S x G log-likelihood matrix.
# @param cluster_indices named list of cluster index vectors.
# @param data_hash       character; hash of the outcome target.
#
# @return log-likelihood matrix with names and metadata.
#
# ---------------------------------------------------------------------------- #
.add_cluster_log_lik_metadata <- function(log_lik, cluster_indices, data_hash) {

  cluster_labels    <- names(cluster_indices)
  colnames(log_lik) <- paste0("log_lik_cluster[", cluster_labels, "]")
  attr(log_lik, "RoBMA_target") <- list(
    unit               = "cluster",
    conditioning_depth = "cluster",
    n                  = length(cluster_indices),
    targets            = cluster_labels,
    data_hash          = data_hash
  )

  return(log_lik)
}


# ---------------------------------------------------------------------------- #
# .log_lik_cluster_norm.brma
# ---------------------------------------------------------------------------- #
#
# Analytic cluster-unit likelihood for unweighted normal multilevel models
# without selection. The normal cluster effect is integrated by the block
# covariance.
#
# @param object brma object.
#
# @return S x G cluster-unit log-likelihood matrix.
#
# ---------------------------------------------------------------------------- #
.log_lik_cluster_norm.brma <- function(object) {

  setup <- .log_lik_cluster_setup.brma(object)
  yi    <- .outcome_data_yi(object)
  vi    <- .outcome_data_vi(object)

  if (setup[["effect_direction"]] == "negative") {
    mu_samples <- -setup[["mu"]]
    yi         <- -yi
  } else {
    mu_samples <- setup[["mu"]]
  }

  cluster_indices <- setup[["cluster"]]
  log_lik         <- matrix(NA_real_, nrow = setup[["S"]], ncol = length(cluster_indices))

  for (g in seq_along(cluster_indices)) {
    idx <- cluster_indices[[g]]

    for (s in seq_len(setup[["S"]])) {
      log_lik[s, g] <- .log_dmvnorm_diag_rank_one(
        x        = yi[idx],
        mean     = mu_samples[s, idx],
        diagonal = vi[idx] + setup[["tau_within"]][s, idx]^2,
        rank_one = setup[["tau_between"]][s, idx]
      )
    }
  }

  return(.add_cluster_log_lik_metadata(log_lik, cluster_indices, setup[["data_hash"]]))
}


# ---------------------------------------------------------------------------- #
# .log_dmvnorm_diag_rank_one
# ---------------------------------------------------------------------------- #
#
# Log-density for N(mean, diag(diagonal) + rank_one %*% t(rank_one)).
#
# ---------------------------------------------------------------------------- #
.log_dmvnorm_diag_rank_one <- function(x, mean, diagonal, rank_one) {

  residual <- x - mean

  if (!all(is.finite(diagonal)) || any(diagonal <= 0)) {
    covariance <- diag(diagonal, nrow = length(diagonal), ncol = length(diagonal)) +
      tcrossprod(rank_one)
    return(mvtnorm::dmvnorm(
      x     = x,
      mean  = mean,
      sigma = covariance,
      log   = TRUE
    ))
  }

  inv_diag <- 1 / diagonal
  inv_rank <- rank_one * inv_diag
  denom    <- 1 + sum(rank_one * inv_rank)

  if (!is.finite(denom) || denom <= .Machine$double.eps) {
    covariance <- diag(diagonal, nrow = length(diagonal), ncol = length(diagonal)) +
      tcrossprod(rank_one)
    return(mvtnorm::dmvnorm(
      x     = x,
      mean  = mean,
      sigma = covariance,
      log   = TRUE
    ))
  }

  log_det <- sum(log(diagonal)) + log(denom)
  quad    <- sum(residual^2 * inv_diag) -
    sum(rank_one * residual * inv_diag)^2 / denom

  return(-0.5 * (length(x) * log(2 * pi) + log_det + quad))
}


# ---------------------------------------------------------------------------- #
# .log_lik_cluster_norm_quadrature.brma
# ---------------------------------------------------------------------------- #
#
# Gamma-quadrature cluster-unit likelihood for selected-normal models and
# data-weighted normal models. Conditional on gamma, per-estimate likelihood
# contributions factorize.
#
# @param object brma object.
#
# @return S x G cluster-unit log-likelihood matrix.
#
# ---------------------------------------------------------------------------- #
.log_lik_cluster_norm_quadrature.brma <- function(object) {

  setup             <- .log_lik_cluster_setup.brma(object)
  is_weightfunction <- .is_weightfunction(object)
  yi                <- .outcome_data_yi(object)
  sei               <- .outcome_data_sei(object)

  if (setup[["effect_direction"]] == "negative" && !is_weightfunction) {
    mu_samples <- -setup[["mu"]]
    yi         <- -yi
  } else {
    mu_samples <- setup[["mu"]]
  }

  if (is_weightfunction) {
    posterior_samples <- setup[["posterior_samples"]]
    selection_context <- .selection_context(
      object            = object,
      posterior_samples = posterior_samples
    )
  }

  log_lik_fun <- function(idx, mu_node) {
    if (is_weightfunction) {
      local_context <- selection_context
      local_context[["obs_bin"]] <- selection_context[["obs_bin"]][idx]

      return(.outcome_pdf.selnorm(
        yi                = yi[idx],
        mu_samples        = mu_node,
        tau_within        = setup[["tau_within"]][, idx, drop = FALSE],
        sei               = sei[idx],
        selection_context = local_context
      ))
    } else {
      return(.outcome_pdf.norm(
        yi         = yi[idx],
        mu_samples = mu_node,
        tau_within = setup[["tau_within"]][, idx, drop = FALSE],
        sei        = sei[idx]
      ))
    }
  }

  log_lik <- .log_lik_cluster_gamma_quadrature(
    cluster_indices     = setup[["cluster"]],
    mu_samples          = mu_samples,
    tau_between_samples = setup[["tau_between"]],
    log_lik_fun         = log_lik_fun,
    weights             = setup[["weights"]]
  )

  return(.add_cluster_log_lik_metadata(log_lik, setup[["cluster"]], setup[["data_hash"]]))
}


# ---------------------------------------------------------------------------- #
# .log_lik_cluster_glmm.brma
# ---------------------------------------------------------------------------- #
#
# Gamma-quadrature cluster-unit likelihood for binomial and Poisson multilevel
# models. Conditional on gamma, existing estimate-level GLMM quadrature helpers
# are reused.
#
# @param object brma object.
#
# @return S x G cluster-unit log-likelihood matrix.
#
# ---------------------------------------------------------------------------- #
.log_lik_cluster_glmm.brma <- function(object) {

  setup       <- .log_lik_cluster_setup.brma(object)
  data        <- setup[["data"]]
  priors      <- setup[["priors"]]
  outcome_type <- .outcome_type(object)

  if (.has_native_glmm_cluster()) {
    log_lik <- .log_lik_cluster_glmm_native(
      setup        = setup,
      data         = data,
      priors       = priors,
      outcome_type = outcome_type
    )
  } else {
    log_lik <- .log_lik_cluster_glmm_r(
      setup        = setup,
      data         = data,
      priors       = priors,
      outcome_type = outcome_type
    )
  }

  return(.add_cluster_log_lik_metadata(log_lik, setup[["cluster"]], setup[["data_hash"]]))
}


# ---------------------------------------------------------------------------- #
# .log_lik_cluster_glmm_r
# ---------------------------------------------------------------------------- #
#
# R-composed reference implementation for GLMM cluster likelihoods.
#
# ---------------------------------------------------------------------------- #
.log_lik_cluster_glmm_r <- function(setup, data, priors, outcome_type,
                                    n_theta = 15,
                                    n_gamma = .get_cluster_likelihood_n_gamma(),
                                    n_pi = 30, n_phi = 30) {

  log_lik_fun <- switch(
    outcome_type,
    "bin" = function(idx, mu_node) {
      .outcome_pdf.binom(
        ai         = data[["outcome"]][["ai"]][idx],
        ci         = data[["outcome"]][["ci"]][idx],
        n1i        = data[["outcome"]][["n1i"]][idx],
        n2i        = data[["outcome"]][["n2i"]][idx],
        mu_samples = mu_node,
        tau_within = setup[["tau_within"]][, idx, drop = FALSE],
        prior_pi   = priors[["outcome"]][["pi"]],
        weights    = setup[["weights"]][idx],
        n_theta    = n_theta,
        n_pi       = n_pi
      )
    },
    "pois" = function(idx, mu_node) {
      .outcome_pdf.pois(
        x1i        = data[["outcome"]][["x1i"]][idx],
        x2i        = data[["outcome"]][["x2i"]][idx],
        t1i        = data[["outcome"]][["t1i"]][idx],
        t2i        = data[["outcome"]][["t2i"]][idx],
        mu_samples = mu_node,
        tau_within = setup[["tau_within"]][, idx, drop = FALSE],
        prior_phi  = priors[["outcome"]][["phi"]],
        weights    = setup[["weights"]][idx],
        n_theta    = n_theta,
        n_phi      = n_phi
      )
    }
  )

  log_lik <- .log_lik_cluster_gamma_quadrature(
    cluster_indices     = setup[["cluster"]],
    mu_samples          = setup[["mu"]],
    tau_between_samples = setup[["tau_between"]],
    log_lik_fun         = log_lik_fun,
    n_gamma             = n_gamma
  )

  return(log_lik)
}


# ---------------------------------------------------------------------------- #
# .log_lik_cluster_glmm_native
# ---------------------------------------------------------------------------- #
#
# Native batched implementation for GLMM cluster likelihoods.
#
# ---------------------------------------------------------------------------- #
.log_lik_cluster_glmm_native <- function(setup, data, priors, outcome_type,
                                         n_theta = 15,
                                         n_gamma = .get_cluster_likelihood_n_gamma(),
                                         n_pi = 30, n_phi = 30) {

  gh_theta <- .gauss_hermite_nodes(n_theta)
  gh_gamma <- .gauss_hermite_nodes(n_gamma)
  cluster  <- .cluster_indices_flatten(setup[["cluster"]])
  weights  <- if (is.null(setup[["weights"]])) NULL else .native_numeric_vector(setup[["weights"]])

  if (outcome_type == "bin") {
    pi_grid <- .glmm_binom_logit_pi_grid(
      ai       = data[["outcome"]][["ai"]],
      ci       = data[["outcome"]][["ci"]],
      n1i      = data[["outcome"]][["n1i"]],
      n2i      = data[["outcome"]][["n2i"]],
      prior_pi = priors[["outcome"]][["pi"]],
      n_pi     = n_pi
    )

    return(.Call(
      "RoBMA_glmm_binom_cluster_loglik",
      .native_integer_vector(data[["outcome"]][["ai"]]),
      .native_integer_vector(data[["outcome"]][["ci"]]),
      .native_integer_vector(data[["outcome"]][["n1i"]]),
      .native_integer_vector(data[["outcome"]][["n2i"]]),
      .native_numeric_matrix(setup[["mu"]]),
      .native_numeric_matrix(setup[["tau_within"]]),
      .native_numeric_matrix(setup[["tau_between"]]),
      cluster[["index"]],
      cluster[["size"]],
      weights,
      .native_numeric_vector(gh_theta[["nodes"]]),
      .native_numeric_vector(gh_theta[["log_weights"]]),
      .native_numeric_matrix(pi_grid[["grid"]]),
      .native_numeric_matrix(pi_grid[["log_weights"]]),
      .native_numeric_vector(gh_gamma[["nodes"]]),
      .native_numeric_vector(gh_gamma[["log_weights"]]),
      PACKAGE = "RoBMA"
    ))
  }

  if (outcome_type == "pois") {
    phi_grid <- .glmm_pois_log_phi_grid(
      x1i       = data[["outcome"]][["x1i"]],
      x2i       = data[["outcome"]][["x2i"]],
      t1i       = data[["outcome"]][["t1i"]],
      t2i       = data[["outcome"]][["t2i"]],
      prior_phi = priors[["outcome"]][["phi"]],
      n_phi     = n_phi
    )

    return(.Call(
      "RoBMA_glmm_pois_cluster_loglik",
      .native_integer_vector(data[["outcome"]][["x1i"]]),
      .native_integer_vector(data[["outcome"]][["x2i"]]),
      .native_numeric_vector(data[["outcome"]][["t1i"]]),
      .native_numeric_vector(data[["outcome"]][["t2i"]]),
      .native_numeric_matrix(setup[["mu"]]),
      .native_numeric_matrix(setup[["tau_within"]]),
      .native_numeric_matrix(setup[["tau_between"]]),
      cluster[["index"]],
      cluster[["size"]],
      weights,
      .native_numeric_vector(gh_theta[["nodes"]]),
      .native_numeric_vector(gh_theta[["log_weights"]]),
      .native_numeric_matrix(phi_grid[["grid"]]),
      .native_numeric_matrix(phi_grid[["log_weights"]]),
      .native_numeric_vector(gh_gamma[["nodes"]]),
      .native_numeric_vector(gh_gamma[["log_weights"]]),
      PACKAGE = "RoBMA"
    ))
  }

  stop("Unsupported outcome type for GLMM cluster-unit log-likelihood.",
       call. = FALSE)
}

# ============================================================================ #
#
# Test LOO-PSIS functionality for brma objects
#
# ============================================================================
context("loo methods for brma objects")

# The LOO-PSIS functionality is necessary for the residuals and funnel plot
# functionality. Due to the computational costs (and the possibility to test)
# against other metafor output) it is primary tested therein.

source(testthat::test_path("common-functions.R"))
source(testthat::test_path("helper-contracts.R"))
skip_on_cran()
skip_if_no_fits()
skip_if_not_installed("loo")

# list cached fits lazily
fit_names <- list_fits()
fits      <- lazy_fits(fit_names, validate = FALSE)
info      <- lazy_infos(fit_names, validate = FALSE)


test_that("logLik returns finite pointwise matrices for cached fits", {

  for (name in c("bcg_meta-analysis", "bcg_meta-regression")) {
    fit_brma <- fits[[name]]
    log_lik  <- logLik(fit_brma)

    expect_s3_class(log_lik, "logLik.brma")
    expect_true(is.matrix(log_lik), info = name)
    expect_equal(ncol(log_lik), nobs(fit_brma), info = name)
    expect_true(all(is.finite(log_lik)), info = name)
    expect_true(stats::var(rowSums(log_lik)) > 0, info = name)
  }
})

test_that("loo and WAIC expose cached diagnostics and missing-cache errors", {

  for (name in c("bcg_meta-analysis", "bcg_meta-regression")) {
    fit_brma <- fits[[name]]
    loo_brma <- suppressWarnings(loo(fit_brma))

    expect_s3_class(loo_brma, "loo")
    expect_true(all(c("elpd_loo", "p_loo", "looic") %in% rownames(loo_brma[["estimates"]])),
                info = name)
    expect_true(all(is.finite(loo_brma[["estimates"]][, "Estimate"])), info = name)

    fit_missing <- fit_brma
    fit_missing[["loo"]] <- NULL
    expect_error(loo(fit_missing), "LOO has not been computed", info = name)
    expect_error(waic(fit_missing), "WAIC has not been computed", info = name)

    fit_waic <- suppressWarnings(add_waic(fit_missing))
    waic_brma <- waic(fit_waic)
    waic_loo  <- loo::waic(fit_waic)

    expect_s3_class(waic_brma, "waic")
    expect_s3_class(waic_loo, "waic")
    expect_equal(waic_loo, waic_brma)
    expect_true(all(c("elpd_waic", "p_waic", "waic") %in% rownames(waic_brma[["estimates"]])),
                info = name)
    expect_true(all(is.finite(waic_brma[["estimates"]][, "Estimate"])), info = name)
  }
})

test_that("loo_compare compares two brma models", {

  # get two brma fits
  fit_brma  <- fits[["bcg_meta-analysis"]]
  fit_brma2 <- fits[["bcg_meta-regression"]]

  # compare
  out <- suppressWarnings(loo_compare(fit_brma, fit_brma2))

  # check structure
  expect_true(is.matrix(out))
  expect_equal(nrow(out), 2)
  expect_true("elpd_diff" %in% colnames(out))
  expect_true("se_diff" %in% colnames(out))
})

test_that("loo_compare accepts loo objects", {

  # get two brma fits
  fit_brma  <- fits[["bcg_meta-analysis"]]
  fit_brma2 <- fits[["bcg_meta-regression"]]

  # get two loo brma fits
  loo_brma  <- loo(fits[["bcg_meta-analysis"]])
  loo_brma2 <- suppressWarnings(loo(fits[["bcg_meta-regression"]]))

  # compare
  out <- loo_compare(loo_brma, loo_brma2)

  # check structure
  expect_true(is.matrix(out))
  expect_equal(nrow(out), 2)
  expect_true("elpd_diff" %in% colnames(out))
  expect_true("se_diff" %in% colnames(out))
})

test_that("loo_compare rejects fewer than two models", {

  # get one brma fit
  fit_brma <- fits[["bcg_meta-analysis"]]

  expect_error(loo_compare(fit_brma, "At least two models"))
})

test_that("logLik, LOO, weights, diagnostics, and WAIC are available for product-space fits", {

  product_names <- c(
    "dat.lehmann2018_BMA.norm",
    "bcg_BMA.glmm",
    "dat.lehmann2018_RoBMA"
  )
  skip_if_missing_fits(product_names)

  for (name in product_names) {

    fit_brma <- fits[[name]]

    log_lik <- logLik(fit_brma)
    expect_s3_class(log_lik, "logLik.brma")
    expect_true(is.matrix(log_lik), info = name)
    expect_equal(ncol(log_lik), nobs(fit_brma), info = name)
    expect_true(all(is.finite(log_lik)), info = name)

    loo_result <- loo(fit_brma)
    expect_s3_class(loo_result, "loo")

    weights <- loo_weights(fit_brma)
    expect_true(is.matrix(weights), info = name)
    expect_equal(dim(weights), dim(log_lik), info = name)
    expect_equal(colSums(weights), rep(1, ncol(weights)), tolerance = 1e-10,
                 info = name)

    expect_no_error(suppressWarnings(check_loo(fit_brma)))

    fit_waic <- fit_brma
    fit_waic[["waic"]] <- NULL
    fit_waic <- suppressWarnings(add_waic(fit_waic))
    waic_result <- waic(fit_waic)
    expect_s3_class(waic_result, "waic")
  }
})

test_that("loo_compare compares BMA and RoBMA product-space fits on the same data", {

  product_names <- c("dat.lehmann2018_BMA.norm", "dat.lehmann2018_RoBMA")
  skip_if_missing_fits(product_names)

  out <- suppressWarnings(loo_compare(
    fits[["dat.lehmann2018_BMA.norm"]],
    fits[["dat.lehmann2018_RoBMA"]]
  ))

  expect_true(is.matrix(out))
  expect_equal(nrow(out), 2)
  expect_true("elpd_diff" %in% colnames(out))
  expect_true("se_diff" %in% colnames(out))
})


# ---------------------------------------------------------------------------- #
# Outcome type specific tests
# ---------------------------------------------------------------------------- #

test_that(".outcome_pdf.norm computes correct log-likelihood", {

  set.seed(123)

  yi  <- c(0.1, 0.2, 0.3)
  sei <- c(0.1, 0.1, 0.1)
  K   <- length(yi)
  S   <- 10

  mu_samples <- matrix(0.15, nrow = S, ncol = K)
  tau_within <- matrix(0.05, nrow = S, ncol = K)

  log_lik <- .outcome_pdf.norm(yi, mu_samples, tau_within, sei)

  expect_equal(dim(log_lik), c(S, K))

  # compute expected log-likelihood directly for first observation
  total_sd <- sqrt(0.05^2 + 0.1^2)
  expected_ll <- dnorm(0.1, mean = 0.15, sd = total_sd, log = TRUE)
  expect_equal(log_lik[1, 1], expected_ll, tolerance = 1e-10)
})

test_that(".outcome_cdf.norm keeps negative-direction tail precision", {

  yi         <- 10
  sei        <- 1
  mu_samples <- matrix(0, nrow = 1, ncol = 1)
  tau_within <- matrix(0, nrow = 1, ncol = 1)

  cdf_vals <- .outcome_cdf.norm(
    yi         = yi,
    mu_samples = mu_samples,
    tau_within = tau_within,
    sei        = sei,
    lower.tail = FALSE
  )

  expect_equal(cdf_vals[1, 1], stats::pnorm(yi, lower.tail = FALSE))
  expect_gt(cdf_vals[1, 1], 0)
})

test_that(".outcome_cdf.selnorm forwards lower.tail through step kernel", {

  yi         <- 3
  sei        <- 1
  mu_samples <- matrix(0, nrow = 1, ncol = 1)
  tau_within <- matrix(0, nrow = 1, ncol = 1)
  prior      <- BayesTools::prior_weightfunction(
    "one-sided", c(.05), BayesTools::wf_fixed(c(1, 1))
  )
  spec <- .selection_spec(
    priors           = list(outcome = list(bias = prior)),
    yi               = yi,
    sei              = sei,
    effect_direction = "positive"
  )
  selection_context <- c(spec, list(
    omega       = matrix(c(1, 1), nrow = 1),
    alpha       = 0,
    phack_kind  = 0L,
    kernel_mode = SELKERNEL_STEP
  ))

  cdf_vals <- .outcome_cdf.selnorm(
    yi                = yi,
    mu_samples        = mu_samples,
    tau_within        = tau_within,
    sei               = sei,
    selection_context = selection_context,
    lower.tail        = FALSE
  )

  expect_equal(cdf_vals[1, 1], stats::pnorm(yi, lower.tail = FALSE),
               tolerance = 1e-12)
  expect_gt(cdf_vals[1, 1], 0)
})

test_that("GLMM nuisance grids integrate over prior support", {

  prior_pi <- BayesTools::prior("beta", list(1, 1))
  pi_grid  <- .glmm_binom_logit_pi_grid(
    ai       = c(0L, 50L),
    ci       = c(50L, 0L),
    n1i      = c(100L, 100L),
    n2i      = c(100L, 100L),
    prior_pi = prior_pi,
    n_pi     = 17
  )

  expect_equal(pi_grid[["grid"]][, 1], pi_grid[["grid"]][, 2])
  expect_equal(sum(exp(pi_grid[["log_weights"]][, 1])), 1, tolerance = 1e-12)

  prior_phi <- BayesTools::prior("normal", list(0, 1))
  phi_grid  <- .glmm_pois_log_phi_grid(
    x1i       = c(0L, 100L),
    x2i       = c(100L, 0L),
    t1i       = c(10, 1000),
    t2i       = c(1000, 10),
    prior_phi = prior_phi,
    n_phi     = 17
  )

  expect_equal(phi_grid[["grid"]][, 1], phi_grid[["grid"]][, 2])
  expect_equal(sum(exp(phi_grid[["log_weights"]][, 1])), 1, tolerance = 1e-12)
})

test_that(".outcome_pdf.binom computes correct log-likelihood", {

  ai  <- c(10, 20)
  ci  <- c(15, 25)
  n1i <- c(100, 100)
  n2i <- c(100, 100)
  K   <- length(ai)
  S   <- 5

  mu_samples <- matrix(0.0, nrow = S, ncol = K) # log-OR = 0
  tau_within <- matrix(0.0, nrow = S, ncol = K)
  prior_pi   <- BayesTools::prior("beta", list(1, 1))

  log_lik <- .outcome_pdf.binom(
    ai         = ai,
    ci         = ci,
    n1i        = n1i,
    n2i        = n2i,
    mu_samples = mu_samples,
    tau_within = tau_within,
    prior_pi   = prior_pi,
    n_theta    = 3,
    n_pi       = 120
  )

  expect_equal(dim(log_lik), c(S, K))

  expected <- lchoose(n1i[1], ai[1]) + lchoose(n2i[1], ci[1]) +
    lbeta(ai[1] + ci[1] + 1, n1i[1] + n2i[1] - ai[1] - ci[1] + 1)

  expect_equal(log_lik[1, 1], expected, tolerance = 1e-8)
})

test_that(".outcome_pdf.binom handles boundary cell studies", {

  prior_pi   <- BayesTools::prior("beta", list(1, 1))
  mu_samples <- matrix(0, nrow = 2, ncol = 4)
  tau_within <- matrix(0, nrow = 2, ncol = 4)

  log_lik <- .outcome_pdf.binom(
    ai         = c(10, 0, 10, 0),
    ci         = c(10, 0, 0, 10),
    n1i        = c(10, 10, 10, 10),
    n2i        = c(10, 10, 10, 10),
    mu_samples = mu_samples,
    tau_within = tau_within,
    prior_pi   = prior_pi
  )

  expect_equal(dim(log_lik), c(2, 4))
  expect_true(all(is.finite(log_lik)))
})

test_that(".outcome_pdf.binom matches R reference", {

  prior_pi <- BayesTools::prior("beta", list(1.5, 2.5))

  ai  <- c(3L, 0L, 8L, 12L)
  ci  <- c(2L, 4L, 0L, 10L)
  n1i <- c(20L, 18L, 15L, 30L)
  n2i <- c(22L, 17L, 16L, 31L)

  mu_samples <- matrix(
    c(-0.4, 0.1, 0.7, 0.2,
      -0.2, 0.3, 0.5, 0.0,
       0.1, 0.5, 0.3, -0.2),
    nrow = 3,
    ncol = 4
  )
  tau_within <- matrix(
    c(0.0, 0.1, 0.4, 0.2,
      0.2, 0.3, 0.1, 0.5,
      0.4, 0.2, 0.0, 0.3),
    nrow = 3,
    ncol = 4
  )

  expect_equal(
    .outcome_pdf.binom(ai, ci, n1i, n2i, mu_samples, tau_within, prior_pi, n_theta = 7, n_pi = 9),
    .outcome_pdf.binom_r(ai, ci, n1i, n2i, mu_samples, tau_within, prior_pi, n_theta = 7, n_pi = 9),
    tolerance = 1e-10
  )

  weights <- c(1, 0.5, 1.25, 2)
  weighted <- .outcome_pdf.binom(
    ai, ci, n1i, n2i, mu_samples, tau_within, prior_pi,
    weights = weights, n_theta = 7, n_pi = 9
  )

  expect_equal(
    weighted,
    .outcome_pdf.binom_r(
      ai, ci, n1i, n2i, mu_samples, tau_within, prior_pi,
      weights = weights, n_theta = 7, n_pi = 9
    ),
    tolerance = 1e-10
  )
  expect_false(isTRUE(all.equal(
    weighted,
    sweep(.outcome_pdf.binom(
      ai, ci, n1i, n2i, mu_samples, tau_within, prior_pi,
      n_theta = 7, n_pi = 9
    ), 2, weights, "*"),
    tolerance = 1e-8
  )))
})

test_that(".outcome_pdf.pois computes correct log-likelihood", {

  x1i <- c(10, 20)
  x2i <- c(15, 25)
  t1i <- c(100, 100)
  t2i <- c(100, 100)
  K   <- length(x1i)
  S   <- 5

  mu_samples <- matrix(0.0, nrow = S, ncol = K) # log-IRR = 0
  tau_within <- matrix(0.0, nrow = S, ncol = K)
  prior_phi  <- BayesTools::prior("normal", list(0, 1))

  log_lik <- .outcome_pdf.pois(
    x1i        = x1i,
    x2i        = x2i,
    t1i        = t1i,
    t2i        = t2i,
    mu_samples = mu_samples,
    tau_within = tau_within,
    prior_phi  = prior_phi,
    n_theta    = 3,
    n_phi      = 80
  )

  expect_equal(dim(log_lik), c(S, K))

  expected <- log(stats::integrate(
    function(phi) {
      lambda <- t1i[1] * exp(phi)
      exp(stats::dpois(x1i[1], lambda = lambda, log = TRUE) +
          stats::dpois(x2i[1], lambda = lambda, log = TRUE) +
          stats::dnorm(phi, log = TRUE))
    },
    lower   = -Inf,
    upper   = Inf,
    rel.tol = 1e-10
  )[["value"]])

  expect_equal(log_lik[1, 1], expected, tolerance = 1e-5)
})

test_that(".outcome_pdf.pois matches R reference", {

  prior_phi <- BayesTools::prior("normal", list(-1, 1.5))

  x1i <- c(0L, 3L, 10L, 12L)
  x2i <- c(1L, 0L, 8L, 9L)
  t1i <- c(12, 30, 45, 50)
  t2i <- c(10, 28, 43, 48)

  mu_samples <- matrix(
    c(-0.4, 0.1, 0.7, 0.2,
      -0.2, 0.3, 0.5, 0.0,
       0.1, 0.5, 0.3, -0.2),
    nrow = 3,
    ncol = 4
  )
  tau_within <- matrix(
    c(0.0, 0.1, 0.4, 0.2,
      0.2, 0.3, 0.1, 0.5,
      0.4, 0.2, 0.0, 0.3),
    nrow = 3,
    ncol = 4
  )

  expect_equal(
    .outcome_pdf.pois(x1i, x2i, t1i, t2i, mu_samples, tau_within, prior_phi, n_theta = 7, n_phi = 9),
    .outcome_pdf.pois_r(x1i, x2i, t1i, t2i, mu_samples, tau_within, prior_phi, n_theta = 7, n_phi = 9),
    tolerance = 1e-10
  )

  weights <- c(1, 0.5, 1.25, 2)
  weighted <- .outcome_pdf.pois(
    x1i, x2i, t1i, t2i, mu_samples, tau_within, prior_phi,
    weights = weights, n_theta = 7, n_phi = 9
  )

  expect_equal(
    weighted,
    .outcome_pdf.pois_r(
      x1i, x2i, t1i, t2i, mu_samples, tau_within, prior_phi,
      weights = weights, n_theta = 7, n_phi = 9
    ),
    tolerance = 1e-10
  )
  expect_false(isTRUE(all.equal(
    weighted,
    sweep(.outcome_pdf.pois(
      x1i, x2i, t1i, t2i, mu_samples, tau_within, prior_phi,
      n_theta = 7, n_phi = 9
    ), 2, weights, "*"),
    tolerance = 1e-8
  )))
})

test_that("native GLMM cluster likelihood matches R composition", {

  skip_if_not(.has_native_glmm_cluster(), "Native GLMM cluster kernels unavailable.")

  set.seed(2024)
  S <- 5
  K <- 5

  setup <- list(
    mu          = matrix(rnorm(S * K, 0, 0.25), nrow = S, ncol = K),
    tau_within  = matrix(runif(S * K, 0.05, 0.25), nrow = S, ncol = K),
    tau_between = matrix(runif(S * K, 0.02, 0.18), nrow = S, ncol = K),
    cluster     = list(a = c(1L, 3L), b = c(2L, 4L, 5L)),
    weights     = c(1, 0.5, 1.25, 2, 0.75)
  )

  bin_data <- list(outcome = data.frame(
    ai  = c(3L, 0L, 8L, 12L, 2L),
    ci  = c(2L, 4L, 0L, 10L, 1L),
    n1i = c(20L, 18L, 15L, 30L, 22L),
    n2i = c(22L, 17L, 16L, 31L, 21L)
  ))
  bin_priors <- list(outcome = list(
    pi = BayesTools::prior("beta", list(1.5, 2.5))
  ))

  expect_equal(
    .log_lik_cluster_glmm_native(
      setup        = setup,
      data         = bin_data,
      priors       = bin_priors,
      outcome_type = "bin",
      n_theta      = 5,
      n_gamma      = 5,
      n_pi         = 7
    ),
    .log_lik_cluster_glmm_r(
      setup        = setup,
      data         = bin_data,
      priors       = bin_priors,
      outcome_type = "bin",
      n_theta      = 5,
      n_gamma      = 5,
      n_pi         = 7
    ),
    tolerance = 1e-10
  )

  pois_data <- list(outcome = data.frame(
    x1i = c(0L, 3L, 10L, 12L, 1L),
    x2i = c(1L, 0L, 8L, 9L, 2L),
    t1i = c(12, 30, 45, 50, 25),
    t2i = c(10, 28, 43, 48, 24)
  ))
  pois_priors <- list(outcome = list(
    phi = BayesTools::prior("normal", list(-1, 1.5))
  ))

  expect_equal(
    .log_lik_cluster_glmm_native(
      setup        = setup,
      data         = pois_data,
      priors       = pois_priors,
      outcome_type = "pois",
      n_theta      = 5,
      n_gamma      = 5,
      n_phi        = 7
    ),
    .log_lik_cluster_glmm_r(
      setup        = setup,
      data         = pois_data,
      priors       = pois_priors,
      outcome_type = "pois",
      n_theta      = 5,
      n_gamma      = 5,
      n_phi        = 7
    ),
    tolerance = 1e-10
  )
})


# ---------------------------------------------------------------------------- #
# loo_weights and check_loo S3 tests
# ---------------------------------------------------------------------------- #

test_that("loo_weights and check_loo return stable diagnostics", {

  fit_brma <- fits[["bcg_meta-analysis"]]

  # check loo_weights
  weights <- loo_weights(fit_brma)
  expect_true(is.matrix(weights))
  expect_equal(dim(weights), dim(logLik(fit_brma)))
  expect_equal(colSums(weights), rep(1, ncol(weights)), tolerance = 1e-10)

  # check check_loo (should not throw anything for this clean fit)
  expect_silent(check_loo(fit_brma))

  # simulate bad k
  fit_bad <- fit_brma
  fit_bad[["loo"]][["estimate"]][["diagnostics"]][["pareto_k"]][1] <- 0.8
  expect_warning(check_loo(fit_bad), "Some Pareto k values are high")
})

context("Evaluate product-space posterior structure")

source(testthat::test_path("common-functions.R"))

skip_if_no_fits()
fit_names <- list_fits()
fits      <- lazy_fits(fit_names, validate = FALSE)

test_that("RoBMA mixed-bias branch PDF, CDF, and RNG use selected-normal branches", {

  name <- "dat.lehmann2018_RoBMA"
  skip_if_missing_fits(name)

  object            <- fits[[name]]
  setup             <- .estimate_likelihood_setup.brma(object)
  posterior_samples <- setup[["posterior_samples"]]
  bias_indicator    <- .extract_bias_indicator(object, posterior_samples = posterior_samples)
  use_normal        <- .extract_use_normal(object, posterior_samples = posterior_samples)

  normal_rows   <- head(which(use_normal), 100)
  weighted_rows <- head(which(!use_normal), 100)
  if (length(normal_rows) == 0L || length(weighted_rows) == 0L) {
    skip("Cached RoBMA fit lacks both normal and weighted publication-bias branches.")
  }

  selected_rows <- sort(c(normal_rows, weighted_rows))
  selection_context <- .selection_context(
    object            = object,
    posterior_samples = posterior_samples
  )
  selected_context <- .selection_context_subset_rows(selection_context, selected_rows)

  log_lik <- .outcome_pdf.selnorm(
    yi                = setup[["yi"]],
    mu_samples        = setup[["mu"]][selected_rows, , drop = FALSE],
    tau_within        = setup[["tau_within"]][selected_rows, , drop = FALSE],
    sei               = setup[["sei"]],
    selection_context = selected_context
  )
  cdf_vals <- .outcome_cdf.selnorm(
    yi                = setup[["yi"]],
    mu_samples        = setup[["mu"]][selected_rows, , drop = FALSE],
    tau_within        = setup[["tau_within"]][selected_rows, , drop = FALSE],
    sei               = setup[["sei"]],
    selection_context = selected_context
  )

  expect_equal(dim(log_lik), c(length(selected_rows), setup[["K"]]))
  expect_equal(dim(cdf_vals), c(length(selected_rows), setup[["K"]]))
  expect_true(all(is.finite(log_lik)))
  expect_true(all(is.finite(cdf_vals)))
  expect_true(all(cdf_vals > 0 & cdf_vals < 1))

  normal_branch   <- sort(unique(bias_indicator[normal_rows]))[1]
  weighted_branch <- sort(unique(bias_indicator[weighted_rows]))[1]
  normal_branch_rows   <- normal_rows[bias_indicator[normal_rows] == normal_branch]
  weighted_branch_rows <- weighted_rows[bias_indicator[weighted_rows] == weighted_branch]

  normal_pdf <- .outcome_pdf.norm(
    yi         = if (setup[["effect_direction"]] == "negative") -setup[["yi"]] else setup[["yi"]],
    mu_samples = if (setup[["effect_direction"]] == "negative") {
      -setup[["mu"]][normal_branch_rows, , drop = FALSE]
    } else {
      setup[["mu"]][normal_branch_rows, , drop = FALSE]
    },
    tau_within = setup[["tau_within"]][normal_branch_rows, , drop = FALSE],
    sei        = setup[["sei"]]
  )
  normal_cdf <- .outcome_cdf.norm(
    yi         = if (setup[["effect_direction"]] == "negative") -setup[["yi"]] else setup[["yi"]],
    mu_samples = if (setup[["effect_direction"]] == "negative") {
      -setup[["mu"]][normal_branch_rows, , drop = FALSE]
    } else {
      setup[["mu"]][normal_branch_rows, , drop = FALSE]
    },
    tau_within = setup[["tau_within"]][normal_branch_rows, , drop = FALSE],
    sei        = setup[["sei"]],
    lower.tail = if (setup[["effect_direction"]] == "negative") FALSE else TRUE
  )

  expect_equal(
    log_lik[match(normal_branch_rows, selected_rows), , drop = FALSE],
    normal_pdf,
    tolerance = 1e-10
  )
  expect_equal(
    cdf_vals[match(normal_branch_rows, selected_rows), , drop = FALSE],
    normal_cdf,
    tolerance = 1e-10
  )

  weighted_pdf_norm <- .outcome_pdf.norm(
    yi         = if (setup[["effect_direction"]] == "negative") -setup[["yi"]] else setup[["yi"]],
    mu_samples = if (setup[["effect_direction"]] == "negative") {
      -setup[["mu"]][weighted_branch_rows, , drop = FALSE]
    } else {
      setup[["mu"]][weighted_branch_rows, , drop = FALSE]
    },
    tau_within = setup[["tau_within"]][weighted_branch_rows, , drop = FALSE],
    sei        = setup[["sei"]]
  )
  expect_false(isTRUE(all.equal(
    log_lik[match(weighted_branch_rows, selected_rows), , drop = FALSE],
    weighted_pdf_norm,
    tolerance = 1e-8
  )))

  set.seed(11)
  rng_normal <- .outcome_rng.selnorm(
    mu_samples        = setup[["mu"]][normal_branch_rows, , drop = FALSE],
    tau_within        = setup[["tau_within"]][normal_branch_rows, , drop = FALSE],
    sei               = setup[["sei"]],
    selection_context = .selection_context_subset_rows(selection_context, normal_branch_rows)
  )
  set.seed(11)
  rng_normal_expected <- .outcome_rng.norm(
    mu_samples = setup[["mu"]][normal_branch_rows, , drop = FALSE],
    tau_within = setup[["tau_within"]][normal_branch_rows, , drop = FALSE],
    sei        = setup[["sei"]]
  )
  expect_equal(dim(rng_normal), dim(rng_normal_expected))
  expect_equal(as.vector(rng_normal), as.vector(rng_normal_expected),
               tolerance = 1e-12)

  set.seed(12)
  rng_mixed <- .outcome_rng.selnorm(
    mu_samples        = setup[["mu"]][selected_rows, , drop = FALSE],
    tau_within        = setup[["tau_within"]][selected_rows, , drop = FALSE],
    sei               = setup[["sei"]],
    selection_context = selected_context
  )

  expect_equal(dim(rng_mixed), c(length(selected_rows), setup[["K"]]))
  expect_true(all(is.finite(rng_mixed)))
})

.expect_indicator_for_prior <- function(posterior_samples, prior, column, info) {

  if (is.null(prior) || !BayesTools::is.prior.mixture(prior)) {
    return(invisible(TRUE))
  }

  expect_true(column %in% colnames(posterior_samples),
              info = paste(info, "missing posterior indicator", column))
  if (!column %in% colnames(posterior_samples)) {
    return(invisible(FALSE))
  }

  values <- posterior_samples[, column]

  expect_true(all(is.finite(values)),
              info = paste(info, column, "contains non-finite values"))
  expect_true(all(values == as.integer(values)),
              info = paste(info, column, "contains non-integer values"))
  expect_true(all(values >= 1 & values <= length(prior)),
              info = paste(info, column, "outside prior component range"))
}

test_that("product-space posterior indicators have expected columns and valid ranges", {

  product_names <- c(
    "dat.lehmann2018_BMA.norm",
    "dat.lehmann2018_BMA.norm_mods",
    "dat.lehmann2018_BMA.norm_scale",
    "bcg_BMA.glmm",
    "nielweise2008_BMA.glmm",
    "bcg_BMA.glmm_3lvl_location_scale",
    "dat.lehmann2018_RoBMA",
    "dat.lehmann2018_RoBMA_3lvl_mods_scale"
  )
  skip_if_missing_fits(product_names)

  for (name in product_names) {

    object            <- fits[[name]]
    posterior_samples <- suppressWarnings(coda::as.mcmc(object[["fit"]]))
    priors            <- object[["priors"]]

    .expect_indicator_for_prior(
      posterior_samples = posterior_samples,
      prior             = priors[["outcome"]][["mu"]],
      column            = "mu_indicator",
      info              = name
    )
    .expect_indicator_for_prior(
      posterior_samples = posterior_samples,
      prior             = priors[["outcome"]][["tau"]],
      column            = "tau_indicator",
      info              = name
    )
    .expect_indicator_for_prior(
      posterior_samples = posterior_samples,
      prior             = priors[["outcome"]][["rho"]],
      column            = "rho_indicator",
      info              = name
    )
    .expect_indicator_for_prior(
      posterior_samples = posterior_samples,
      prior             = priors[["outcome"]][["bias"]],
      column            = "bias_indicator",
      info              = name
    )

    for (term in names(priors[["mods"]])) {
      column <- paste0(
        BayesTools::JAGS_parameter_names(term, formula_parameter = "mu"),
        "_indicator"
      )
      .expect_indicator_for_prior(
        posterior_samples = posterior_samples,
        prior             = priors[["mods"]][[term]],
        column            = column,
        info              = name
      )
    }

    for (term in names(priors[["scale"]])) {
      column <- paste0(
        BayesTools::JAGS_parameter_names(term, formula_parameter = "log_tau"),
        "_indicator"
      )
      .expect_indicator_for_prior(
        posterior_samples = posterior_samples,
        prior             = priors[["scale"]][[term]],
        column            = column,
        info              = name
      )
    }
  }
})

test_that("RoBMA inactive bias branches expose zeros or neutral weights", {

  product_names <- c("dat.lehmann2018_RoBMA", "dat.lehmann2018_RoBMA_3lvl_mods_scale")
  skip_if_missing_fits(product_names)

  for (name in product_names) {

    object            <- fits[[name]]
    posterior_samples <- suppressWarnings(coda::as.mcmc(object[["fit"]]))
    priors_bias       <- object[["priors"]][["outcome"]][["bias"]]
    if (is.null(priors_bias) || !"bias_indicator" %in% colnames(posterior_samples)) {
      next
    }
    if (!BayesTools::is.prior.mixture(priors_bias)) {
      priors_bias <- list(priors_bias)
    }

    bias_indicator      <- as.integer(posterior_samples[, "bias_indicator"])
    branch_is_PET       <- vapply(priors_bias, BayesTools::is.prior.PET, logical(1))
    branch_is_PEESE     <- vapply(priors_bias, BayesTools::is.prior.PEESE, logical(1))
    branch_is_weightfun <- vapply(priors_bias, BayesTools::is.prior.weightfunction, logical(1))

    if ("PET" %in% colnames(posterior_samples)) {
      inactive <- which(!branch_is_PET[bias_indicator])
      expect_equal(as.numeric(posterior_samples[inactive, "PET"]), rep(0, length(inactive)),
                   tolerance = 1e-12, info = paste(name, "inactive PET branch"))
    }
    if ("PEESE" %in% colnames(posterior_samples)) {
      inactive <- which(!branch_is_PEESE[bias_indicator])
      expect_equal(as.numeric(posterior_samples[inactive, "PEESE"]), rep(0, length(inactive)),
                   tolerance = 1e-12, info = paste(name, "inactive PEESE branch"))
    }

    omega_cols <- grep("^omega(\\[|$)", colnames(posterior_samples), value = TRUE)
    if (length(omega_cols) > 0L) {
      inactive <- which(!branch_is_weightfun[bias_indicator])
      omega_inactive <- as.matrix(posterior_samples[inactive, omega_cols, drop = FALSE])
      expect_equal(dim(omega_inactive), c(length(inactive), length(omega_cols)),
                   info = paste(name, "inactive weightfunction branch"))
      expect_equal(as.vector(omega_inactive), rep(1, length(omega_inactive)),
                   tolerance = 1e-12,
                   info = paste(name, "inactive weightfunction branch"))
    }
  }
})

#' @title Robust Bayesian Model-Averaged Meta-Analysis
#'
#' @description Fits a robust Bayesian model-averaged meta-analysis. The
#' default ensemble averages across models with and without an effect,
#' heterogeneity, and publication-bias adjustment.
#'
#' @inheritParams data_input
#' @inheritParams RoBMA_prior_specification
#' @inheritParams prior_specification
#' @inheritParams fitting_specification
#'
#' @details
#' `RoBMA()` uses product-space Bayesian model averaging. Inclusion Bayes
#' factors and model-averaged estimates are obtained from mixture priors for
#' effect, heterogeneity, moderators, scale regression, and publication-bias
#' components.
#'
#' By default, `model_type = "PSMA"` includes selection-model weight functions
#' together with PET and PEESE publication-bias adjustments. Use `BMA()` for
#' model averaging without publication-bias adjustment, or `brma()` for fitting
#' a single meta-analytic model.
#'
#' `RoBMA()` uses normal/effect-size input (`yi` with `vi` or `sei`). Raw-count
#' GLMM model averaging is provided by `BMA.glmm()`.
#'
#' Product-space objects support predictive comparison with `add_loo()` and
#' `add_waic()`. Bridge-sampling marginal likelihood via `add_marglik()` is
#' not available for product-space model-averaging objects.
#'
#' @return A fitted object of class `c("RoBMA", "brma")`. The object contains
#' checked `data`, checked `priors`, the JAGS `fit`, cached `summary`, and
#' cached `coefficients`. It can be passed to `summary()`, `plot()`,
#' `predict()`, `funnel()`, `add_loo()`, and related methods.
#'
#' @examples \dontrun{
#' if (requireNamespace("metadat", quietly = TRUE)) {
#'   data(dat.lehmann2018, package = "metadat")
#'
#'   fit <- RoBMA(
#'     yi      = yi,
#'     vi      = vi,
#'     data    = dat.lehmann2018,
#'     measure = "SMD",
#'     seed    = 1,
#'     silent  = TRUE
#'   )
#'
#'   summary(fit)
#'   plot(fit)
#' }
#' }
#'
#' @seealso [publication_bias_prior_specification], [BMA()], [brma()],
#' [bselmodel()], [bPET()], [bPEESE()], [summary.brma()], [plot.brma()]
#' @export
RoBMA <- function(
  # input specification
  yi, vi, sei, weights, ni,
  mods, scale, cluster,
  data, slab, subset,
  measure, effect_direction = "detect",

  # prior specification
  prior_effect, prior_heterogeneity, prior_mods, prior_scale, prior_heterogeneity_allocation, prior_bias,
  prior_effect_null, prior_heterogeneity_null, prior_mods_null, prior_scale_null, prior_heterogeneity_allocation_null, prior_bias_null,
  standardize_continuous_predictors = TRUE,
  set_contrast_factor_predictors = "meandif",
  prior_unit_information_sd, rescale_priors = 1,
  prior_informed_field, prior_informed_subfield,
  model_type = "PSMA",

  # MCMC fitting settings
  sample = 5000, burnin = 2000, adapt = 500,
  chains = 3, thin = 1, parallel = FALSE,
  autofit = FALSE, autofit_control = set_autofit_control(),
  convergence_checks = set_convergence_checks(),

  # additional settings
  seed = NULL, silent = TRUE, ...) {

  ### create the output object
  dots            <- list(...)
  missing_measure <- missing(measure)
  if (missing_measure && !isTRUE(dots[["only_data"]])) {
    .stop_missing_measure("RoBMA()")
  }
  if (missing_measure) {
    measure <- "GEN"
  }
  dots            <- .validate_constructor_dots(dots, caller = "RoBMA()")
  object          <- .createObject(
    dots = dots, class = c("RoBMA", "brma"),
    # MCMC and fitting settings
    chains = chains, adapt = adapt, burnin = burnin, sample = sample, thin = thin,
    autofit = autofit, parallel = parallel, silent = silent, seed = seed,
    autofit_control = autofit_control, convergence_checks = convergence_checks
  )

  ### check and store the data
  object$data <- .check_and_list_data(
    .call = match.call(), .envir = parent.frame(), class = "norm",
    set_contrast_factor_predictors = set_contrast_factor_predictors,
    standardize_continuous_predictors = standardize_continuous_predictors,
    effect_direction = effect_direction, measure = measure)
  if (isTRUE(dots[["only_data"]]))
    return(object)


  ### check and store priors
  # checks and store the base priors
  object$priors <- .check_and_list_priors.RoBMA(
    prior_effect = prior_effect, prior_heterogeneity = prior_heterogeneity,
    prior_mods = prior_mods, prior_scale = prior_scale,
    prior_heterogeneity_allocation = prior_heterogeneity_allocation, prior_bias = prior_bias,
    prior_effect_null = prior_effect_null, prior_heterogeneity_null = prior_heterogeneity_null,
    prior_mods_null = prior_mods_null, prior_scale_null = prior_scale_null,
    prior_heterogeneity_allocation_null = prior_heterogeneity_allocation_null, prior_bias_null = prior_bias_null,
    rescale_priors                    = rescale_priors,
    prior_unit_information_sd         = prior_unit_information_sd,
    prior_informed_field              = prior_informed_field,
    prior_informed_subfield           = prior_informed_subfield,
    data = object[["data"]], model_type = model_type)
  if (isTRUE(dots[["only_priors"]]))
    return(.set_only_priors_class(object))

  ### fit the model
  object$fit <- .fit(object)

  ### store simple summary & coefficients
  object$summary       <- .object_summary(object)
  object$coefficients  <- .object_coefficients(object)
  object               <- .autocompute_brma(object)

  return(object)
}

test_that("brma.norm is exported alias for brma", {

  expect_true("brma.norm" %in% getNamespaceExports("RoBMA"))
  expect_identical(brma.norm, brma)
})

test_that("top-level fitting constructors reject stale algorithm argument", {

  norm_args <- list(
    yi         = c(0.10, 0.20, 0.15),
    sei        = c(0.05, 0.06, 0.07),
    only_data  = TRUE,
    algorithm  = "ss"
  )
  glmm_args <- list(
    ai         = c(5, 8, 6),
    bi         = c(45, 42, 44),
    ci         = c(4, 7, 5),
    di         = c(46, 43, 45),
    only_data  = TRUE,
    algorithm  = "ss"
  )

  norm_constructors <- list(
    brma      = brma,
    brma.norm = brma.norm,
    RoBMA     = RoBMA,
    BMA       = BMA,
    BMA.norm  = BMA.norm,
    bPET      = bPET,
    bPEESE    = bPEESE,
    bselmodel = bselmodel
  )
  glmm_constructors <- list(
    brma.glmm = brma.glmm,
    BMA.glmm  = BMA.glmm
  )

  for (constructor in norm_constructors) {
    expect_error(
      do.call(constructor, norm_args),
      "Unused argument.*algorithm"
    )
  }
  for (constructor in glmm_constructors) {
    expect_error(
      do.call(constructor, glmm_args),
      "Unused argument.*algorithm"
    )
  }
})

test_that("top-level fitting constructors reject other named dots", {

  expect_error(
    RoBMA(
      yi               = c(0.10, 0.20, 0.15),
      sei              = c(0.05, 0.06, 0.07),
      legacy_algorithm = "ss",
      only_data        = TRUE
    ),
    "Unused argument.*legacy_algorithm"
  )
})

test_that("top-level fitting constructors validate internal dot types", {

  expect_error(
    RoBMA(
      yi        = c(0.10, 0.20, 0.15),
      sei       = c(0.05, 0.06, 0.07),
      measure   = "SMD",
      only_data = 1
    ),
    "only_data"
  )

  expect_error(
    brma.glmm(
      ai             = c(5, 8, 6),
      bi             = c(45, 42, 44),
      ci             = c(4, 7, 5),
      di             = c(46, 43, 45),
      only_data      = TRUE,
      is_JASP_prefix = TRUE
    ),
    "is_JASP_prefix"
  )
})

# ============================================================================ #
# brma_samples.R
# ============================================================================ #
#
# This file defines the brma_samples class for posterior samples returned by
# predict.brma() and related wrapper functions. The class:
#
# - Stores posterior samples as a matrix (S x K)
# - Retains MCMC metadata (nchains, niter) for posterior package integration
# - Provides print/summary methods using BayesTools::ensemble_estimates_table
# - Supports as_draws conversion for posterior package compatibility
#
# ============================================================================ #


# ---------------------------------------------------------------------------- #
# Constructor
# ---------------------------------------------------------------------------- #

#' @title Create a brma_samples Object
#'
#' @description Internal constructor for creating \code{brma_samples} objects
#' that store posterior samples with metadata for printing and conversion
#' to \pkg{posterior} draws formats.
#'
#' @param samples a matrix of posterior samples with dimensions S x K
#' (samples x parameters/observations)
#' @param n_chains number of MCMC chains used in sampling
#' @param n_iter number of iterations per chain (after thinning)
#' @param title title for the summary table output
#' @param probs default quantiles for credible intervals. Defaults to
#' \code{c(.025, .975)}
#' @param data optional data associated with the predictions (e.g., for
#' non-aggregated predictions)
#' @param effect_transform optional effect-size transformation metadata
#'
#' @return An object of class \code{brma_samples} which inherits from
#' \code{matrix}.
#'
#' @noRd
.new_brma_samples <- function(samples, n_chains, n_iter, title,
                              probs = c(.025, .975), data = NULL,
                              effect_transform = NULL) {

  # ensure samples is a matrix with proper column names
  if (!is.matrix(samples)) {
    samples <- as.matrix(samples)
  }

  n_chains <- as.integer(n_chains)
  n_iter   <- as.integer(n_iter)

  if (length(n_chains) != 1L || length(n_iter) != 1L ||
      is.na(n_chains) || is.na(n_iter) ||
      n_chains < 1L || n_iter < 1L ||
      n_chains * n_iter != nrow(samples)) {
    stop("Invalid brma_samples chain metadata: 'n_chains * n_iter' must equal the number of sample rows.",
         call. = FALSE)
  }

  # add attributes for MCMC structure
  attr(samples, "nchains") <- n_chains
  attr(samples, "niter")   <- n_iter

  # add attributes for display
  attr(samples, "title")    <- title
  attr(samples, "probs")    <- probs
  attr(samples, "data")     <- data

  if (!is.null(effect_transform)) {
    attr(samples, "effect_transform") <- effect_transform
    attr(samples, "footnotes")        <- effect_transform[["note"]]
  }

  # set class (inherits from matrix for backward compatibility)
  class(samples) <- c("brma_samples", "matrix", "array")

  return(samples)
}


# Derive valid chain metadata for brma_samples objects.
.brma_samples_chain_info <- function(fit = NULL, n_samples) {

  n_samples <- as.integer(n_samples)

  if (!is.null(fit) && !is.null(fit[["mcmc"]]) && length(fit[["mcmc"]]) > 0L) {
    chain_lengths <- vapply(fit[["mcmc"]], function(x) NROW(x), integer(1))
    if (sum(chain_lengths) == n_samples && length(unique(chain_lengths)) == 1L) {
      return(list(
        n_chains = length(chain_lengths),
        n_iter   = chain_lengths[[1]]
      ))
    }
  }

  return(list(
    n_chains = 1L,
    n_iter   = n_samples
  ))
}


# ---------------------------------------------------------------------------- #
# Print method
# ---------------------------------------------------------------------------- #

#' @title Print brma_samples Object
#'
#' @description Prints a summary table of posterior samples using
#' \code{BayesTools::ensemble_estimates_table}. Returns the samples
#' invisibly for method chaining.
#'
#' @param x a \code{brma_samples} object
#' @param probs quantiles for credible intervals. If \code{NULL}, uses
#' the default stored in the object (typically \code{c(.025, .975)})
#' @param ... additional arguments passed to
#' \code{BayesTools::ensemble_estimates_table}
#'
#' @return Returns \code{x} invisibly.
#'
#' @export
print.brma_samples <- function(x, probs = NULL, ...) {

  summary <- summary.brma_samples(x, probs = probs, ...)
  print.summary.brma_samples(summary)

  return(invisible(x))
}


# ---------------------------------------------------------------------------- #
# Summary method
# ---------------------------------------------------------------------------- #

#' @title Summarize brma_samples Object
#'
#' @description Creates and returns a summary table of posterior samples
#' using \code{BayesTools::ensemble_estimates_table}.
#'
#' @param object a \code{brma_samples} object
#' @param probs quantiles for credible intervals. If \code{NULL}, uses
#' the default stored in the object (typically \code{c(.025, .975)})
#' @param ... additional arguments passed to
#' \code{BayesTools::ensemble_estimates_table}
#'
#' @return A \code{BayesTools_table} object containing the summary statistics.
#'
#' @export
summary.brma_samples <- function(object, probs = NULL, ...) {

  # use stored probs if not provided
  if (is.null(probs)) {
    probs <- attr(object, "probs")
  }

  dots <- list(...)
  if (is.null(dots[["footnotes"]]) && !is.null(attr(object, "footnotes"))) {
    dots[["footnotes"]] <- attr(object, "footnotes")
  }

  # create summary table
  summary_table <- do.call(
    BayesTools::ensemble_estimates_table,
    c(
      list(
        samples    = asplit(object, 2),
        parameters = colnames(object),
        probs      = probs,
        title      = attr(object, "title")
      ),
      dots
    )
  )

  class(summary_table) <- c("summary.brma_samples", class(summary_table))
 
  return(summary_table)
}


# ---------------------------------------------------------------------------- #
# Print summary method
# ---------------------------------------------------------------------------- #

#' @title Print summary.brma_samples Object
#'
#' @description Prints a summary table of posterior samples using
#' \code{BayesTools::ensemble_estimates_table}. Returns the summary table
#' invisibly.
#'
#' @param x a \code{summary.brma_samples} object
#' @param probs quantiles for credible intervals. If \code{NULL}, uses
#' the default stored in the object (typically \code{c(.025, .975)})
#' @param ... additional arguments passed to
#' \code{BayesTools::ensemble_estimates_table}
#'
#' @return Returns the summary table invisibly.
#'
#' @export
print.summary.brma_samples <- function(x, probs = NULL, ...) {

  class(x) <- setdiff(class(x), "summary.brma_samples")

  cat("\n")
  print(x)
  cat("\n")

  return(invisible(x))
}


# ---------------------------------------------------------------------------- #
# as.matrix method (for backward compatibility)
# ---------------------------------------------------------------------------- #

#' @title Convert brma_samples to Matrix
#'
#' @description Converts a \code{brma_samples} object to a plain matrix,
#' removing all brma_samples-specific attributes.
#'
#' @param x a \code{brma_samples} object
#' @param ... additional arguments (ignored)
#'
#' @return A plain matrix of posterior samples.
#'
#' @export
as.matrix.brma_samples <- function(x, ...) {
  # remove class and custom attributes
  attributes(x) <- list(
    dim      = dim(x),
    dimnames = dimnames(x)
  )
  return(x)
}

#' @title Bayesian Generalized Meta-Analysis
#'
#' @description Function for fitting random-effects, meta-regression, multilevel,
#' and location-scale meta-analytic models directly to either binary or count data.
#'
#' @inheritParams data_input
#' @inheritParams prior_specification
#' @inheritParams fitting_specification
#' @details
#' Model for odds ratios (`measure = "OR"`) corresponds to Model 4 described in
#' \insertCite{jackson2018comparison;textual}{RoBMA}.
#' `logit(pi[i])` is the study-specific midpoint of the two arm logits.
#' `prior_baserate` defines the estimate-specific prior distribution on `pi[i]`.
#'
#' Model for incidence rate ratios (`measure = "IRR"`) corresponds to
#' \insertCite{bagos2009mixed;textual}{RoBMA}.
#' `phi[i]` is the study-specific midpoint of the two arm log incidence rates.
#' `prior_lograte` defines the estimate-specific prior distribution on `phi[i]`.
#' If unspecified, a unit-information prior is based on the data and used
#' independently for each estimate.
#'
#' When `weights` are supplied, they are treated as likelihood weights on the
#' paired two-arm study contribution.
#'
#' @return A fitted object of class `c("brma.glmm", "brma")`. The object
#' contains checked `data`, checked `priors`, the JAGS `fit`, cached `summary`,
#' and cached `coefficients`. If the corresponding package options are enabled,
#' it can also contain cached LOO, WAIC, or marginal likelihood results.
#'
#' @examples \dontrun{
#' if (requireNamespace("metadat", quietly = TRUE)) {
#'   data(dat.bcg, package = "metadat")
#'
#'   fit <- brma.glmm(
#'     ai      = tpos,
#'     bi      = tneg,
#'     ci      = cpos,
#'     di      = cneg,
#'     mods    = ~ alloc,
#'     data    = dat.bcg,
#'     measure = "OR",
#'     seed    = 1,
#'     silent  = TRUE
#'   )
#'
#'   summary(fit)
#' }
#' }
#'
#' @seealso [brma()], [BMA.glmm()], [summary.brma()], [predict.brma()]
#' @export
brma.glmm <- function(
  # input specification
  ai, bi, ci, di, n1i, n2i, x1i, x2i, t1i, t2i, weights,
  mods, scale, cluster,
  data, slab, subset,
  measure = "OR",

  # prior specification
  prior_effect, prior_heterogeneity, prior_mods, prior_scale,
  prior_heterogeneity_allocation, prior_baserate, prior_lograte,
  standardize_continuous_predictors = TRUE,
  set_contrast_factor_predictors = "treatment",
  prior_unit_information_sd, rescale_priors = 1,
  prior_informed_field, prior_informed_subfield,

  # MCMC fitting settings
  sample = 5000, burnin = 2000, adapt = 500,
  chains = 3, thin = 1, parallel = FALSE,
  autofit = FALSE, autofit_control = set_autofit_control(),
  convergence_checks = set_convergence_checks(),

  # additional settings
  seed = NULL, silent, ...
){

  ### create the output object
  dots         <- list(...)
  dots         <- .validate_constructor_dots(dots, caller = "brma.glmm()")
  object       <- .createObject(
    dots = dots, class = c("brma.glmm", "brma"),
    # MCMC and fitting settings
    chains = chains, adapt = adapt, burnin = burnin, sample = sample, thin = thin,
    autofit = autofit, parallel = parallel, silent = silent, seed = seed,
    autofit_control = autofit_control, convergence_checks = convergence_checks
  )

  ### check and store the data
  object$data <- .check_and_list_data(
    .call = match.call(), .envir = parent.frame(), class = "glmm",
    set_contrast_factor_predictors = set_contrast_factor_predictors,
    measure = measure, standardize_continuous_predictors = standardize_continuous_predictors)
  if (isTRUE(dots[["only_data"]]))
    return(object)

  ### check and store priors
  object$priors <- .check_and_list_priors.brma(
    prior_effect = prior_effect, prior_heterogeneity = prior_heterogeneity,
    prior_mods = prior_mods, prior_scale = prior_scale,
    prior_heterogeneity_allocation = prior_heterogeneity_allocation,
    prior_baserate = prior_baserate, prior_lograte = prior_lograte,
    rescale_priors                    = rescale_priors,
    prior_unit_information_sd         = prior_unit_information_sd,
    prior_informed_field              = prior_informed_field,
    prior_informed_subfield           = prior_informed_subfield,
    data = object[["data"]])
  if (isTRUE(dots[["only_priors"]]))
    return(.set_only_priors_class(object))

  ### fit the model
  object$fit <- .fit(object)

  ### store simple summary & coefficients
  object$summary       <- .object_summary(object)
  object$coefficients  <- .object_coefficients(object)

  object               <- .autocompute_brma(object)

  return(object)
}

context("Input handling for brma.glmm")

skip_on_cran()
source(testthat::test_path("helper-contracts.R"))

# Test data for GLMM input specification tests (2x2 contingency tables)
test_data_glmm <- data.frame(
  # Cell counts (ai = events in treatment, bi = non-events in treatment,
  #              ci = events in control, di = non-events in control)
  ai  = c(10L, 15L, 12L,  8L, 20L),
  bi  = c(40L, 35L, 38L, 42L, 30L),
  ci  = c( 5L, 10L,  8L,  4L, 12L),
  di  = c(45L, 40L, 42L, 46L, 38L),
  # Marginal totals
  n1i = c(50L, 50L, 50L, 50L, 50L),  # ai + bi
  n2i = c(50L, 50L, 50L, 50L, 50L),  # ci + di
  # Optional variables
  wgt     = c(1.0, 1.5, 1.2, 0.8, 1.3),
  study   = c("Study A", "Study B", "Study C", "Study D", "Study E"),
  cluster = c("g1", "g1", "g2", "g2", "g3"),
  stringsAsFactors = FALSE
)

expect_glmm_outcome <- function(result, n, ai, ci, n1i = NULL, n2i = NULL) {

  expect_type(result, "list")
  expect_true("outcome" %in% names(result))
  expect_equal(nrow(result$outcome), n)
  expect_equal(result$outcome$ai, ai)
  expect_equal(result$outcome$ci, ci)

  if (!is.null(n1i)) {
    expect_equal(result$outcome$n1i, n1i)
  }
  if (!is.null(n2i)) {
    expect_equal(result$outcome$n2i, n2i)
  }
}


# ============================================================================
# Basic input specification tests
# ============================================================================

test_that("GLMM input rejects unsupported effect-size measures directly", {

  expect_error(
    brma.glmm(
      ai        = c(10L, 15L),
      bi        = c(40L, 35L),
      ci        = c(5L, 10L),
      di        = c(45L, 40L),
      measure   = "SMD",
      only_data = TRUE
    ),
    "OR.*IRR"
  )

  expect_error(
    BMA.glmm(
      ai        = c(10L, 15L),
      bi        = c(40L, 35L),
      ci        = c(5L, 10L),
      di        = c(45L, 40L),
      measure   = "RR",
      only_data = TRUE
    ),
    "OR.*IRR"
  )
})

test_that("GLMM accepts supported 2x2 input specifications", {

  input_cases <- list(
    list(
      label = "direct cell vectors",
      expr  = quote(brma.glmm(
        ai        = c(10L, 15L, 12L),
        bi        = c(40L, 35L, 38L),
        ci        = c(5L, 10L, 8L),
        di        = c(45L, 40L, 42L),
        only_data = TRUE
      )[["data"]]),
      n    = 3,
      ai   = c(10L, 15L, 12L),
      ci   = c(5L, 10L, 8L),
      n1i  = c(50L, 50L, 50L),
      n2i  = c(50L, 50L, 50L)
    ),
    list(
      label = "events with arm totals",
      expr  = quote(brma.glmm(
        ai        = c(10L, 15L, 12L),
        ci        = c(5L, 10L, 8L),
        n1i       = c(50L, 50L, 50L),
        n2i       = c(50L, 50L, 50L),
        only_data = TRUE
      )[["data"]]),
      n    = 3,
      ai   = c(10L, 15L, 12L),
      ci   = c(5L, 10L, 8L),
      n1i  = c(50L, 50L, 50L),
      n2i  = c(50L, 50L, 50L)
    ),
    list(
      label = "unquoted data columns",
      expr  = quote(brma.glmm(
        ai        = ai,
        bi        = bi,
        ci        = ci,
        di        = di,
        data      = test_data_glmm,
        only_data = TRUE
      )[["data"]]),
      n    = 5,
      ai   = test_data_glmm$ai,
      ci   = test_data_glmm$ci,
      n1i  = test_data_glmm$n1i,
      n2i  = test_data_glmm$n2i
    ),
    list(
      label = "quoted data columns",
      expr  = quote(brma.glmm(
        ai        = "ai",
        bi        = "bi",
        ci        = "ci",
        di        = "di",
        data      = test_data_glmm,
        only_data = TRUE
      )[["data"]]),
      n    = 5,
      ai   = test_data_glmm$ai,
      ci   = test_data_glmm$ci,
      n1i  = test_data_glmm$n1i,
      n2i  = test_data_glmm$n2i
    ),
    list(
      label = "mixed quoted and unquoted columns",
      expr  = quote(brma.glmm(
        ai        = ai,
        bi        = "bi",
        ci        = ci,
        di        = di,
        data      = test_data_glmm,
        only_data = TRUE
      )[["data"]]),
      n    = 5,
      ai   = test_data_glmm$ai,
      ci   = test_data_glmm$ci,
      n1i  = test_data_glmm$n1i,
      n2i  = test_data_glmm$n2i
    )
  )

  for (case in input_cases) {
    expect_glmm_outcome(
      eval(case$expr),
      n    = case$n,
      ai   = case$ai,
      ci   = case$ci,
      n1i  = case$n1i,
      n2i  = case$n2i
    )
  }
})


# ============================================================================
# Optional arguments tests
# ============================================================================

test_that("GLMM handles slab and cluster metadata", {

  skip_on_cran()

  default_result <- brma.glmm(
    ai        = ai,
    bi        = bi,
    ci        = ci,
    di        = di,
    data      = test_data_glmm,
    only_data = TRUE
  )[["data"]]

  result <- brma.glmm(
    ai        = ai,
    bi        = bi,
    ci        = ci,
    di        = di,
    cluster   = cluster,
    slab      = study,
    data      = test_data_glmm,
    only_data = TRUE
  )[["data"]]

  expect_equal(default_result$outcome$slab, paste0("Study ", 1:5))
  expect_equal(result$outcome$slab, test_data_glmm$study)
  expect_true(is.numeric(result$outcome$cluster))
  expect_equal(result$outcome$cluster, c(1, 1, 2, 2, 3))
})


# ============================================================================
# Input validation tests
# ============================================================================

test_that("GLMM rejects missing required inputs", {

  skip_on_cran()

  expect_error_cases(list(
    list(
      label  = "missing control arm",
      expr   = quote(brma.glmm(
        ai        = c(10L, 15L),
        bi        = c(40L, 35L),
        only_data = TRUE
      )),
      regexp = "provide either"
    ),
    list(
      label  = "missing ci",
      expr   = quote(brma.glmm(
        ai        = c(10L, 15L),
        bi        = c(40L, 35L),
        di        = c(45L, 40L),
        only_data = TRUE
      )),
      regexp = "provide either"
    )
  ))
})


test_that("GLMM validates cell counts are non-negative integers", {

  skip_on_cran()

  expect_error_cases(list(
    list(
      label  = "negative ai",
      expr   = quote(brma.glmm(
        ai        = c(10L, -5L),
        bi        = c(40L, 35L),
        ci        = c(5L, 10L),
        di        = c(45L, 40L),
        only_data = TRUE
      )),
      regexp = "ai"
    ),
    list(
      label  = "fractional ai",
      expr   = quote(brma.glmm(
        ai        = c(10, 15.5),
        bi        = c(40L, 35L),
        ci        = c(5L, 10L),
        di        = c(45L, 40L),
        only_data = TRUE
      )),
      regexp = "ai"
    )
  ))
})


test_that("GLMM validates redundant 2x2 totals against supplied cells", {

  skip_on_cran()

  result <- brma.glmm(
    ai        = c(10L, 15L),
    bi        = c(40L, 35L),
    ci        = c(5L, 10L),
    di        = c(45L, 40L),
    n1i       = c(50L, 50L),
    n2i       = c(50L, 50L),
    only_data = TRUE
  )[["data"]]

  expect_equal(result$outcome$n1i, c(50L, 50L))
  expect_equal(result$outcome$n2i, c(50L, 50L))

  expect_error_cases(list(
    list(
      label  = "bad supplied n1i",
      expr   = quote(brma.glmm(
        ai        = c(10L, 15L),
        bi        = c(40L, 35L),
        ci        = c(5L, 10L),
        di        = c(45L, 40L),
        n1i       = c(50L, 100L),
        n2i       = c(50L, 50L),
        only_data = TRUE
      )),
      regexp = "n1i.*ai [+] bi"
    ),
    list(
      label  = "bad supplied n2i",
      expr   = quote(brma.glmm(
        ai        = c(10L, 15L),
        bi        = c(40L, 35L),
        ci        = c(5L, 10L),
        di        = c(45L, 40L),
        n1i       = c(50L, 50L),
        n2i       = c(50L, 100L),
        only_data = TRUE
      )),
      regexp = "n2i.*ci [+] di"
    ),
    list(
      label  = "bad inferred bi",
      expr   = quote(brma.glmm(
        ai        = c(10L, 15L),
        bi        = c(40L, 99L),
        ci        = c(5L, 10L),
        n1i       = c(50L, 50L),
        n2i       = c(50L, 50L),
        only_data = TRUE
      )),
      regexp = "n1i.*ai [+] bi"
    ),
    list(
      label  = "bad inferred di",
      expr   = quote(brma.glmm(
        ai        = c(10L, 15L),
        ci        = c(5L, 10L),
        di        = c(45L, 99L),
        n1i       = c(50L, 50L),
        n2i       = c(50L, 50L),
        only_data = TRUE
      )),
      regexp = "n2i.*ci [+] di"
    )
  ))
})


test_that("GLMM input preserves likelihood weights", {

  skip_on_cran()

  bin_result <- brma.glmm(
    ai        = ai,
    bi        = bi,
    ci        = ci,
    di        = di,
    weights   = wgt,
    data      = test_data_glmm,
    only_data = TRUE
  )[["data"]]

  expect_true(attr(bin_result, "weights"))
  expect_equal(bin_result$outcome$weights, test_data_glmm$wgt)

  pois_result <- brma.glmm(
    x1i       = c(3L, 5L),
    x2i       = c(2L, 4L),
    t1i       = c(10, 12),
    t2i       = c(11, 13),
    weights   = c(1, 2),
    measure   = "IRR",
    only_data = TRUE
  )[["data"]]

  expect_true(attr(pois_result, "weights"))
  expect_equal(pois_result$outcome$weights, c(1, 2))

  expect_error(
    brma.glmm(
      ai        = ai,
      bi        = bi,
      ci        = ci,
      di        = di,
      weights   = c(1, 0, 1, 1, 1),
      data      = test_data_glmm,
      only_data = TRUE
    ),
    regexp = "weights"
  )
})


test_that("GLMM validates inferred cells and argument lengths", {

  skip_on_cran()

  expect_error_cases(list(
    list(
      label  = "computed bi is negative",
      expr   = quote(brma.glmm(
        ai        = c(10L, 60L),
        ci        = c(5L, 10L),
        n1i       = c(50L, 50L),
        n2i       = c(50L, 50L),
        only_data = TRUE
      )),
      regexp = "bi.*negative"
    ),
    list(
      label  = "computed di is negative",
      expr   = quote(brma.glmm(
        ai        = c(10L, 15L),
        ci        = c(5L, 60L),
        n1i       = c(50L, 50L),
        n2i       = c(50L, 50L),
        only_data = TRUE
      )),
      regexp = "di.*negative"
    ),
    list(
      label  = "cell-count length mismatch",
      expr   = quote(brma.glmm(
        ai        = c(10L, 15L, 12L),
        bi        = c(40L, 35L),
        ci        = c(5L, 10L, 8L),
        di        = c(45L, 40L, 42L),
        only_data = TRUE
      )),
      regexp = "bi"
    ),
    list(
      label  = "redundant-total length mismatch",
      expr   = quote(brma.glmm(
        ai        = c(10L, 15L, 12L),
        bi        = c(40L, 35L),
        ci        = c(5L, 10L, 8L),
        n1i       = c(50L, 50L, 50L),
        n2i       = c(50L, 50L, 50L),
        only_data = TRUE
      )),
      regexp = "bi"
    )
  ))
})


# ============================================================================
# Subset tests
# ============================================================================

test_that("GLMM applies and validates subsets", {

  skip_on_cran()

  subset_cases <- list(
    list(subset = c(TRUE, TRUE, FALSE, TRUE, FALSE), rows = c(1, 2, 4)),
    list(subset = c(1, 3, 5), rows = c(1, 3, 5))
  )

  for (case in subset_cases) {
    result <- brma.glmm(
      ai        = ai,
      bi        = bi,
      ci        = ci,
      di        = di,
      subset    = case$subset,
      data      = test_data_glmm,
      only_data = TRUE
    )[["data"]]

    expect_equal(nrow(result$outcome), length(case$rows))
    expect_equal(result$outcome$ai, test_data_glmm$ai[case$rows])
    expect_equal(result$outcome$ci, test_data_glmm$ci[case$rows])
  }

  expect_error(
    brma.glmm(
      ai     = ai,
      bi     = bi,
      ci     = ci,
      di     = di,
      subset = c(TRUE, NA, TRUE, TRUE, TRUE),
      data   = test_data_glmm,
      only_data = TRUE
    ),
    regexp = "subset.*NA"
  )
})


# ============================================================================
# NA handling tests
# ============================================================================

test_that("GLMM drops NA cell rows and refreshes derived metadata", {

  skip_on_cran()

  test_data_na <- data.frame(
    ai = c(10L, 15L, 12L, 8L),
    bi = c(40L, 35L, 38L, 42L),
    ci = c(5L, 10L, 8L, 4L),
    di = c(45L, 40L, 42L, 46L)
  )
  test_data_na[["ai"]][2] <- NA_integer_

  expect_warning(
    result <- brma.glmm(
      ai        = ai,
      bi        = bi,
      ci        = ci,
      di        = di,
      data      = test_data_na,
      only_data = TRUE
    )[["data"]],
    regexp = "1 observation.*removed"
  )

  expect_equal(nrow(result$outcome), 3)
  expect_equal(result$outcome$ai, c(10L, 12L, 8L))

  test_data_multi_na <- data.frame(
    ai = c(10L, 15L, 12L, 8L),
    bi = c(40L, 35L, 38L, 42L),
    ci = c(5L, 10L, 8L, 4L),
    di = c(45L, 40L, 42L, 46L)
  )
  test_data_multi_na[["bi"]][2] <- NA_integer_
  test_data_multi_na[["ci"]][3] <- NA_integer_
  test_data_multi_na[["di"]][4] <- NA_integer_

  expect_warning(
    result <- brma.glmm(
      ai        = ai,
      bi        = bi,
      ci        = ci,
      di        = di,
      data      = test_data_multi_na,
      only_data = TRUE
    )[["data"]],
    regexp = "3 observation.*removed"
  )

  expect_equal(nrow(result$outcome), 1)
  expect_equal(result$outcome$ai, 10L)

  test_data_na <- data.frame(
    ai  = c(10L, 15L, 12L),
    bi  = c(40L, NA, 38L),
    ci  = c(5L, 10L, 8L),
    di  = c(45L, 40L, 42L),
    n1i = c(50L, 50L, 50L),
    n2i = c(50L, 50L, 50L)
  )

  expect_warning(
    result <- brma.glmm(
      ai        = ai,
      bi        = bi,
      ci        = ci,
      di        = di,
      n1i       = n1i,
      n2i       = n2i,
      data      = test_data_na,
      only_data = TRUE
    )[["data"]],
    regexp = "1 observation.*removed"
  )

  expect_equal(nrow(result$outcome), 2)
  expect_equal(result$outcome$ai, c(10L, 12L))
  expect_equal(result$outcome$n1i, c(50L, 50L))

  test_data_na <- data.frame(
    ai = c(10L, 15L, 12L),
    bi = c(40L, 35L, 38L),
    ci = c(5L, 10L, 8L),
    di = c(45L, 40L, 42L)
  )
  test_data_na[["ai"]][2] <- NA_integer_

  expect_warning(
    result <- brma.glmm(
      ai   = ai,
      bi   = bi,
      ci   = ci,
      di   = di,
      data = test_data_na,
      only_data = TRUE
    )[["data"]],
    regexp = "removed"
  )

  expect_equal(result$outcome$slab, c("Study 1", "Study 2"))

  test_data_na <- data.frame(
    ai      = c(10L, 15L, 12L, 8L),
    bi      = c(40L, 35L, 38L, 42L),
    ci      = c(5L, 10L, 8L, 4L),
    di      = c(45L, 40L, 42L, 46L),
    cluster = c("g1", "g1", "g2", "g2")
  )
  test_data_na[["ai"]][2] <- NA_integer_

  expect_warning(
    result <- brma.glmm(
      ai        = ai,
      bi        = bi,
      ci        = ci,
      di        = di,
      cluster   = cluster,
      data      = test_data_na,
      only_data = TRUE
    )[["data"]],
    regexp = "removed"
  )

  expect_equal(result$outcome$cluster, c(1, 2, 2))
})


test_that("GLMM rejects NA cluster, but preserves NA slab", {

  skip_on_cran()

  test_data_na <- data.frame(
    ai      = c(10L, 15L, 12L, 8L),
    bi      = c(40L, 35L, 38L, 42L),
    ci      = c(5L, 10L, 8L, 4L),
    di      = c(45L, 40L, 42L, 46L),
    cluster = c("g1", "g1", "g2", "g3"),
    slab    = c("A", NA, "C", "D")
  )

  test_data_na[["cluster"]][2] <- NA_character_

  expect_error(
    brma.glmm(
      ai        = ai,
      bi        = bi,
      ci        = ci,
      di        = di,
      cluster   = cluster,
      data      = test_data_na,
      only_data = TRUE
    ),
    regexp = "argument must not contain missing values"
  )

  test_data_na[["cluster"]][2] <- "g1"

  result <- brma.glmm(
    ai        = ai,
    bi        = bi,
    ci        = ci,
    di        = di,
    cluster   = cluster,
    slab      = slab,
    data      = test_data_na,
    only_data = TRUE
  )[["data"]]

  expect_equal(nrow(result$outcome), 4)
  expect_true(is.na(result$outcome$slab[2]))
})


test_that("GLMM rejects all-NA observations", {

  skip_on_cran()

  test_data_all_na <- data.frame(
    ai = c(1L, 2L),
    bi = c(40L, 35L),
    ci = c(5L, 10L),
    di = c(45L, 40L)
  )
  test_data_all_na[["ai"]] <- c(NA_integer_, NA_integer_)

  expect_error(
    suppressWarnings(
      brma.glmm(
        ai   = ai,
        bi   = bi,
        ci   = ci,
        di   = di,
        data = test_data_all_na,
        only_data = TRUE
      )
    ),
    regexp = "No observations remaining"
  )
})


# ============================================================================
# Moderator (mods) tests
# ============================================================================

test_data_glmm_mods <- data.frame(
  ai         = c(10L, 15L, 12L, 8L, 20L),
  bi         = c(40L, 35L, 38L, 42L, 30L),
  ci         = c(5L, 10L, 8L, 4L, 12L),
  di         = c(45L, 40L, 42L, 46L, 38L),
  mod_cont   = c(1.5, 2.3, 1.8, 3.1, 0.9),
  mod_factor = factor(c("A", "B", "A", "B", "A")),
  scale_var  = c(0.5, 1.0, 0.8, 1.2, 0.6),
  stringsAsFactors = FALSE
)


test_that("GLMM parses moderator and scale formulas", {

  skip_on_cran()

  formula_cases <- list(
    list(
      expr        = quote(brma.glmm(
        ai        = ai,
        bi        = bi,
        ci        = ci,
        di        = di,
        mods      = ~ mod_cont + mod_factor,
        data      = test_data_glmm_mods,
        only_data = TRUE
      )[["data"]]),
      mods_names  = c("mod_cont", "mod_factor"),
      scale_names = character()
    ),
    list(
      expr        = quote(brma.glmm(
        ai        = ai,
        bi        = bi,
        ci        = ci,
        di        = di,
        scale     = ~ scale_var,
        data      = test_data_glmm_mods,
        only_data = TRUE
      )[["data"]]),
      mods_names  = character(),
      scale_names = "scale_var"
    ),
    list(
      expr        = quote(brma.glmm(
        ai        = ai,
        bi        = bi,
        ci        = ci,
        di        = di,
        mods      = ~ mod_cont + mod_factor,
        scale     = ~ scale_var,
        data      = test_data_glmm_mods,
        only_data = TRUE
      )[["data"]]),
      mods_names  = c("mod_cont", "mod_factor"),
      scale_names = "scale_var"
    )
  )

  for (case in formula_cases) {
    result <- eval(case$expr)

    if (length(case$mods_names)) {
      expect_equal(nrow(result$mods), 5)
      expect_true(all(case$mods_names %in% names(result$mods)))
    } else {
      expect_null(result$mods)
    }

    if (length(case$scale_names)) {
      expect_equal(nrow(result$scale), 5)
      expect_true(all(case$scale_names %in% names(result$scale)))
    } else {
      expect_null(result$scale)
    }
  }
})


test_that("GLMM drops rows with NA in mods", {

  skip_on_cran()

  test_data_na <- data.frame(
    ai       = c(10L, 15L, 12L, 8L),
    bi       = c(40L, 35L, 38L, 42L),
    ci       = c(5L, 10L, 8L, 4L),
    di       = c(45L, 40L, 42L, 46L),
    mod_cont = c(1.5, NA, 1.8, 3.1)
  )

  expect_warning(
    result <- brma.glmm(
      ai   = ai,
      bi   = bi,
      ci   = ci,
      di   = di,
      mods = ~ mod_cont,
      data = test_data_na,
      only_data = TRUE
    )[["data"]],
    regexp = "1 observation.*removed"
  )

  expect_equal(nrow(result$outcome), 3)
  expect_equal(nrow(result$mods), 3)
})


# ============================================================================
# Predictor validation tests
# ============================================================================

test_that("GLMM rejects degenerate moderator design matrices", {

  skip_on_cran()

  test_data_const <- data.frame(
    ai        = c(10L, 15L, 12L),
    bi        = c(40L, 35L, 38L),
    ci        = c(5L, 10L, 8L),
    di        = c(45L, 40L, 42L),
    const_var = c(5, 5, 5)
  )

  test_data_single <- data.frame(
    ai         = c(10L, 15L, 12L),
    bi         = c(40L, 35L, 38L),
    ci         = c(5L, 10L, 8L),
    di         = c(45L, 40L, 42L),
    single_fac = factor(c("A", "A", "A"))
  )

  expect_error_cases(list(
    list(
      label  = "constant numeric moderator",
      expr   = quote(brma.glmm(
        ai        = ai,
        bi        = bi,
        ci        = ci,
        di        = di,
        mods      = ~ const_var,
        data      = test_data_const,
        only_data = TRUE
      )),
      regexp = "zero variance"
    ),
    list(
      label  = "single-level factor moderator",
      expr   = quote(brma.glmm(
        ai        = ai,
        bi        = bi,
        ci        = ci,
        di        = di,
        mods      = ~ single_fac,
        data      = test_data_single,
        only_data = TRUE
      )),
      regexp = "only one level"
    ),
    list(
      label  = "subset drops all but one factor level",
      expr   = quote(brma.glmm(
        ai        = ai,
        bi        = bi,
        ci        = ci,
        di        = di,
        mods      = ~ mod_factor,
        subset    = c(TRUE, FALSE, TRUE, FALSE, TRUE),
        data      = test_data_glmm_mods,
        only_data = TRUE
      )),
      regexp = "only one level"
    )
  ))
})


# ============================================================================
# RoBMA_data class and attributes tests
# ============================================================================

test_that("GLMM returns RoBMA_data class with correct attributes", {

  skip_on_cran()

  result <- brma.glmm(
    ai        = ai,
    bi        = bi,
    ci        = ci,
    di        = di,
    cluster   = cluster,
    slab      = study,
    data      = test_data_glmm,
    only_data = TRUE
  )[["data"]]

  expect_s3_class(result, "RoBMA_data")
  expect_equal(attr(result, "k_final"), 5)
  expect_equal(attr(result, "n_dropped"), 0)
  expect_false(attr(result, "mods"))
  expect_false(attr(result, "scale"))
  expect_false(attr(result, "weights"))
  expect_true(attr(result, "slab"))
  expect_true(attr(result, "cluster"))
})


test_that("GLMM attributes reflect NA dropping", {

  skip_on_cran()

  test_data_na <- data.frame(
    ai = c(10L, 15L, 12L, 8L),
    bi = c(40L, 35L, 38L, 42L),
    ci = c(5L, 10L, 8L, 4L),
    di = c(45L, 40L, 42L, 46L)
  )
  test_data_na[["ai"]][2] <- NA_integer_

  expect_warning(
    result <- brma.glmm(
      ai   = ai,
      bi   = bi,
      ci   = ci,
      di   = di,
      data = test_data_na,
      only_data = TRUE
    )[["data"]],
    regexp = "removed"
  )

  expect_equal(attr(result, "k_final"), 3)
  expect_equal(attr(result, "n_dropped"), 1)
})

context("Prior input handling for brma.glmm")

skip_on_cran()

test_data_bin <- data.frame(
  ai = c(4L, 6L, 8L),
  bi = c(16L, 14L, 12L),
  ci = c(3L, 5L, 7L),
  di = c(17L, 15L, 13L)
)

test_data_pois <- data.frame(
  x1i = c(4L, 6L, 8L),
  x2i = c(3L, 5L, 7L),
  t1i = c(20, 25, 30),
  t2i = c(21, 24, 31)
)


test_that("Binomial GLMM baserate priors are assigned", {

  result_default <- brma.glmm(
    ai = ai, bi = bi, ci = ci, di = di,
    data = test_data_bin, measure = "OR",
    prior_baserate = NULL, only_priors = TRUE
  )[["priors"]]

  expect_equal(result_default$outcome$pi$distribution, "beta")
  expect_equal(result_default$outcome$pi$parameters$alpha, 1)
  expect_equal(result_default$outcome$pi$parameters$beta,  1)

  custom_prior <- BayesTools::prior("beta", parameters = list(alpha = 2, beta = 3))
  result_custom <- brma.glmm(
    ai = ai, bi = bi, ci = ci, di = di,
    data = test_data_bin, measure = "OR",
    prior_baserate = custom_prior, only_priors = TRUE
  )[["priors"]]

  expect_equal(result_custom$outcome$pi$parameters$alpha, 2)
  expect_equal(result_custom$outcome$pi$parameters$beta,  3)
})


test_that("Poisson GLMM lograte priors are assigned", {

  result_default <- brma.glmm(
    x1i = x1i, x2i = x2i, t1i = t1i, t2i = t2i,
    data = test_data_pois, measure = "IRR",
    prior_lograte = NULL, only_priors = TRUE
  )[["priors"]]

  expect_equal(result_default$outcome$phi$distribution, "normal")

  custom_prior <- BayesTools::prior("normal", parameters = list(mean = 0, sd = 2))
  result_custom <- brma.glmm(
    x1i = x1i, x2i = x2i, t1i = t1i, t2i = t2i,
    data = test_data_pois, measure = "IRR",
    prior_lograte = custom_prior, only_priors = TRUE
  )[["priors"]]

  expect_equal(result_custom$outcome$phi$parameters$sd, 2)
})


test_that("Poisson GLMM default lograte prior requires observed events", {

  expect_error(
    brma.glmm(
      x1i           = c(0L, 0L),
      x2i           = c(0L, 0L),
      t1i           = c(10, 12),
      t2i           = c(11, 13),
      measure       = "IRR",
      prior_lograte = NULL,
      only_priors   = TRUE
    ),
    regexp = "prior_lograte.*observed Poisson event|all-zero"
  )
})


test_that("Poisson GLMM validates custom lograte prior before transformation", {

  expect_error(
    brma.glmm(
      x1i           = x1i,
      x2i           = x2i,
      t1i           = t1i,
      t2i           = t2i,
      data           = test_data_pois,
      measure        = "IRR",
      prior_lograte  = list(distribution = "normal"),
      only_priors    = TRUE
    ),
    regexp = "prior_lograte.*prior distribution"
  )
})

context("Input handling for predict")

source(testthat::test_path("helper-contracts.R"))

skip_on_cran()

test_data_norm <- data.frame(
  yi        = c(0.2, 0.5, -0.1, 0.3, 0.4),
  sei       = c(0.1, 0.15, 0.12, 0.08, 0.11),
  mod_cont  = c(1.5, 2.3, 1.8, 3.1, 0.9),
  mod_fac   = factor(c("A", "B", "A", "B", "A")),
  scale_var = c(0.5, 1.0, 0.8, 1.2, 0.6),
  stringsAsFactors = FALSE
)

test_data_norm_vi <- data.frame(
  yi       = c(0.2, 0.5, -0.1, 0.3, 0.4),
  vi       = c(0.01, 0.0225, 0.0144, 0.0064, 0.0121),
  mod_cont = c(1.5, 2.3, 1.8, 3.1, 0.9),
  mod_fac  = factor(c("A", "B", "A", "B", "A")),
  stringsAsFactors = FALSE
)

test_data_glmm <- data.frame(
  ai       = c(10L, 15L, 12L, 8L, 20L),
  ci       = c(5L, 10L, 8L, 4L, 12L),
  n1i      = c(50L, 50L, 50L, 50L, 50L),
  n2i      = c(50L, 50L, 50L, 50L, 50L),
  mod_cont = c(1.5, 2.3, 1.8, 3.1, 0.9),
  mod_fac  = factor(c("A", "B", "A", "B", "A")),
  stringsAsFactors = FALSE
)

compare_data_lists <- function(data1, data2, check_slab = FALSE,
                               structure_only = FALSE, info = NULL) {

  expect_s3_class(data1, "RoBMA_data")
  expect_s3_class(data2, "RoBMA_data")

  if (!structure_only) {
    expect_equal(nrow(data1$outcome), nrow(data2$outcome), info = info)
  }
  expect_equal(ncol(data1$outcome), ncol(data2$outcome), info = info)
  expect_equal(names(data1$outcome), names(data2$outcome), info = info)

  if (!structure_only) {
    outcome_cols <- setdiff(names(data1$outcome), if (check_slab) character() else "slab")
    for (col in outcome_cols) {
      expect_equal(data1$outcome[[col]], data2$outcome[[col]],
                   info = paste(info, "outcome", col))
    }
  }

  for (slot in c("mods", "scale")) {
    expect_equal(is.null(data1[[slot]]), is.null(data2[[slot]]), info = info)
    if (!is.null(data1[[slot]])) {
      if (!structure_only) {
        expect_equal(nrow(data1[[slot]]), nrow(data2[[slot]]), info = info)
      }
      expect_equal(names(data1[[slot]]), names(data2[[slot]]), info = info)
      if (!structure_only) {
        for (col in names(data1[[slot]])) {
          expect_equal(data1[[slot]][[col]], data2[[slot]][[col]],
                       info = paste(info, slot, col))
        }
      }
      expect_equal(attr(data1[[slot]], "formula"), attr(data2[[slot]], "formula"),
                   info = paste(info, slot, "formula"))
    }
  }

  attrs <- c(
    "outcome_type", "mods", "scale",
    "standardize_continuous_predictors",
    "set_contrast_factor_predictors", "effect_direction"
  )
  for (attr_name in attrs) {
    expect_equal(attr(data1, attr_name), attr(data2, attr_name),
                 info = paste(info, "attribute", attr_name))
  }
  if (!structure_only) {
    expect_equal(attr(data1, "k_final"), attr(data2, "k_final"),
                 info = paste(info, "attribute k_final"))
  }
}

prepare_newdata_cases <- list(
  list(
    label = "normal same data with sei",
    fit = quote(brma.norm(
      yi = yi, sei = sei, data = test_data_norm, only_data = TRUE
    )),
    newdata = quote(test_data_norm)
  ),
  list(
    label = "normal same data with vi",
    fit = quote(brma.norm(
      yi = yi, vi = vi, data = test_data_norm_vi, only_data = TRUE
    )),
    newdata = quote(test_data_norm_vi)
  ),
  list(
    label = "normal different outcome rows",
    fit = quote(brma.norm(
      yi = yi, sei = sei, data = test_data_norm, only_data = TRUE
    )),
    newdata = quote(data.frame(yi = c(0.1, 0.6, 0.0), sei = c(0.2, 0.1, 0.15))),
    structure_only = TRUE
  ),
  list(
    label = "normal moderators same data",
    fit = quote(brma.norm(
      yi = yi, sei = sei, mods = ~ mod_cont + mod_fac,
      data = test_data_norm, only_data = TRUE
    )),
    newdata = quote(test_data_norm)
  ),
  list(
    label = "normal moderators different rows",
    fit = quote(brma.norm(
      yi = yi, sei = sei, mods = ~ mod_cont + mod_fac,
      data = test_data_norm, only_data = TRUE
    )),
    newdata = quote(data.frame(
      yi       = c(0.1, 0.3),
      sei      = c(0.1, 0.2),
      mod_cont = c(2.0, 1.0),
      mod_fac  = factor(c("A", "B"), levels = c("A", "B"))
    )),
    structure_only = TRUE
  ),
  list(
    label = "normal scale same data",
    fit = quote(brma.norm(
      yi = yi, sei = sei, scale = ~ scale_var,
      data = test_data_norm, only_data = TRUE
    )),
    newdata = quote(test_data_norm)
  ),
  list(
    label = "normal moderators and scale",
    fit = quote(brma.norm(
      yi = yi, sei = sei, mods = ~ mod_cont, scale = ~ scale_var,
      data = test_data_norm, only_data = TRUE
    )),
    newdata = quote(test_data_norm)
  ),
  list(
    label = "GLMM same data",
    fit = quote(brma.glmm(
      ai = ai, ci = ci, n1i = n1i, n2i = n2i,
      data = test_data_glmm, only_data = TRUE
    )),
    newdata = quote(test_data_glmm)
  ),
  list(
    label = "GLMM moderators same data",
    fit = quote(brma.glmm(
      ai = ai, ci = ci, n1i = n1i, n2i = n2i, mods = ~ mod_cont + mod_fac,
      data = test_data_glmm, only_data = TRUE
    )),
    newdata = quote(test_data_glmm)
  )
)

test_that(".prepare_newdata reconstructs response, moderator, and scale data", {

  for (case in prepare_newdata_cases) {
    fit <- eval(case[["fit"]])
    result <- RoBMA:::.prepare_newdata(
      object  = fit,
      newdata = eval(case[["newdata"]]),
      type    = "terms"
    )

    compare_data_lists(
      fit[["data"]],
      result,
      structure_only = isTRUE(case[["structure_only"]]),
      info           = case[["label"]]
    )
  }
})

test_that(".prepare_newdata inserts dummy outcomes only when the response is unused", {

  fit <- brma.norm(
    yi = yi, sei = sei, mods = ~ mod_cont,
    data = test_data_norm, only_data = TRUE
  )

  result <- RoBMA:::.prepare_newdata(
    object  = fit,
    newdata = data.frame(mod_cont = c(1.2, 2.4)),
    type    = "terms"
  )

  expect_equal(nrow(result[["outcome"]]), 2L)
  expect_equal(result[["outcome"]][["yi"]], c(0, 0))
  expect_equal(result[["outcome"]][["sei"]], c(0, 0))

  response_fit <- brma.norm(
    yi = yi, sei = sei, data = test_data_norm, only_data = TRUE
  )
  response <- RoBMA:::.prepare_newdata(
    object  = response_fit,
    newdata = data.frame(sei = c(0.1, 0.2)),
    type    = "response"
  )
  expect_equal(response[["outcome"]][["yi"]], c(0, 0))
  expect_equal(response[["outcome"]][["sei"]], c(0.1, 0.2))
})

test_that(".prepare_newdata rejects missing required variables", {

  fit_norm <- brma.norm(
    yi = yi, sei = sei, data = test_data_norm, only_data = TRUE
  )
  fit_mods <- brma.norm(
    yi = yi, sei = sei, mods = ~ mod_cont + mod_fac,
    data = test_data_norm, only_data = TRUE
  )
  fit_scale <- brma.norm(
    yi = yi, sei = sei, scale = ~ scale_var,
    data = test_data_norm, only_data = TRUE
  )
  fit_glmm <- brma.glmm(
    ai = ai, ci = ci, n1i = n1i, n2i = n2i,
    data = test_data_glmm, only_data = TRUE
  )

  expect_error_cases(list(
    list(
      label  = "normal response missing sei/vi",
      expr   = quote(RoBMA:::.prepare_newdata(
        object = fit_norm, newdata = data.frame(yi = c(0.1, 0.2)),
        type = "response"
      )),
      regexp = "sei.*vi"
    ),
    list(
      label  = "missing moderator",
      expr   = quote(RoBMA:::.prepare_newdata(
        object = fit_mods,
        newdata = data.frame(
          yi = c(0.1, 0.2), sei = c(0.1, 0.2),
          mod_fac = factor(c("A", "B"))
        ),
        type = "terms"
      )),
      regexp = "mod_cont"
    ),
    list(
      label  = "missing scale predictor",
      expr   = quote(RoBMA:::.prepare_newdata(
        object = fit_scale,
        newdata = data.frame(yi = c(0.1, 0.2), sei = c(0.1, 0.2)),
        type = "terms"
      )),
      regexp = "scale_var"
    ),
    list(
      label  = "missing GLMM response size",
      expr   = quote(RoBMA:::.prepare_newdata(
        object = fit_glmm,
        newdata = data.frame(ai = c(10L, 15L), ci = c(5L, 10L), n1i = c(50L, 50L)),
        type = "response"
      )),
      regexp = "n2i"
    )
  ))
})

test_that(".prepare_newdata permits bias-adjusted PET/PEESE terms without new standard errors", {

  fit <- bPET(
    yi = yi, sei = sei, data = test_data_norm, only_data = TRUE
  )
  fit[["priors"]] <- list(outcome = list(
    bias = BayesTools::prior_PET("normal", list(mean = 0, sd = 1))
  ))
  new_df <- data.frame(row = 1:2)

  expect_error(
    RoBMA:::.prepare_newdata(
      object = fit, newdata = new_df, type = "terms"
    ),
    regexp = "sei.*vi"
  )
  expect_no_error(RoBMA:::.prepare_newdata(
    object = fit, newdata = new_df, type = "terms", bias_adjusted = TRUE
  ))
})

test_that(".prepare_newdata preserves predictor transformation settings", {

  fit_std <- brma.norm(
    yi = yi, sei = sei, mods = ~ mod_cont,
    data = test_data_norm,
    standardize_continuous_predictors = TRUE,
    only_data = TRUE
  )
  fit_no_std <- brma.norm(
    yi = yi, sei = sei, mods = ~ mod_cont,
    data = test_data_norm,
    standardize_continuous_predictors = FALSE,
    only_data = TRUE
  )
  result_std <- RoBMA:::.prepare_newdata(fit_std, test_data_norm, type = "terms")
  result_no_std <- RoBMA:::.prepare_newdata(fit_no_std, test_data_norm, type = "terms")

  expect_true(attr(result_std, "standardize_continuous_predictors"))
  expect_false(attr(result_no_std, "standardize_continuous_predictors"))

  fit_treatment <- brma.norm(
    yi = yi, sei = sei, mods = ~ mod_fac,
    data = test_data_norm,
    set_contrast_factor_predictors = "treatment",
    only_data = TRUE
  )
  fit_meandif <- brma.norm(
    yi = yi, sei = sei, mods = ~ mod_fac,
    data = test_data_norm,
    set_contrast_factor_predictors = "meandif",
    only_data = TRUE
  )

  result_treatment <- RoBMA:::.prepare_newdata(fit_treatment, test_data_norm, type = "terms")
  result_meandif <- RoBMA:::.prepare_newdata(fit_meandif, test_data_norm, type = "terms")

  expect_equal(attr(result_treatment, "set_contrast_factor_predictors"), "treatment")
  expect_equal(attr(result_meandif, "set_contrast_factor_predictors"), "meandif")
})

test_that(".prepare_newdata drops rows with missing outcome or moderator values", {

  fit_norm <- brma.norm(
    yi = yi, sei = sei, data = test_data_norm, only_data = TRUE
  )
  expect_warning(
    result_norm <- RoBMA:::.prepare_newdata(
      object = fit_norm,
      newdata = data.frame(yi = c(0.1, NA, 0.3), sei = c(0.1, 0.2, 0.15)),
      type = "terms"
    ),
    regexp = "removed"
  )
  expect_equal(nrow(result_norm[["outcome"]]), 2L)

  fit_mods <- brma.norm(
    yi = yi, sei = sei, mods = ~ mod_cont,
    data = test_data_norm, only_data = TRUE
  )
  expect_warning(
    result_mods <- RoBMA:::.prepare_newdata(
      object = fit_mods,
      newdata = data.frame(
        yi = c(0.1, 0.2, 0.3), sei = c(0.1, 0.2, 0.15),
        mod_cont = c(1.5, NA, 2.0)
      ),
      type = "terms"
    ),
    regexp = "removed"
  )
  expect_equal(nrow(result_mods[["outcome"]]), 2L)
  expect_equal(nrow(result_mods[["mods"]]), 2L)
})

test_that(".prepare_newdata accepts prediction-only edge cases", {

  edge_cases <- list(
    list(
      label = "single row",
      fit = quote(brma.norm(
        yi = yi, sei = sei, data = test_data_norm, only_data = TRUE
      )),
      newdata = quote(data.frame(yi = 0.5, sei = 0.1)),
      check = function(result) {
        expect_equal(nrow(result[["outcome"]]), 1L)
        expect_equal(result[["outcome"]][["yi"]], 0.5)
      }
    ),
    list(
      label = "extra columns",
      fit = quote(brma.norm(
        yi = yi, sei = sei, data = test_data_norm, only_data = TRUE
      )),
      newdata = quote(data.frame(
        yi = c(0.1, 0.2), sei = c(0.1, 0.2),
        extra_col = c("a", "b"), another = c(1, 2)
      )),
      check = function(result) expect_equal(nrow(result[["outcome"]]), 2L)
    ),
    list(
      label = "single-level factor",
      fit = quote(brma.norm(
        yi = yi, sei = sei, mods = ~ mod_fac,
        data = test_data_norm, only_data = TRUE
      )),
      newdata = quote(data.frame(
        yi = c(0.1, 0.2, 0.3), sei = c(0.1, 0.1, 0.1),
        mod_fac = factor(c("A", "A", "A"), levels = c("A", "B"))
      )),
      check = function(result) {
        expect_equal(nrow(result[["mods"]]), 3L)
        expect_equal(as.character(result[["mods"]][["mod_fac"]]), rep("A", 3))
      }
    ),
    list(
      label = "zero-variance moderator",
      fit = quote(brma.norm(
        yi = yi, sei = sei, mods = ~ mod_cont,
        data = test_data_norm, only_data = TRUE
      )),
      newdata = quote(data.frame(
        yi = c(0.1, 0.2, 0.3), sei = c(0.1, 0.1, 0.1),
        mod_cont = c(2, 2, 2)
      )),
      check = function(result) expect_equal(result[["mods"]][["mod_cont"]], c(2, 2, 2))
    ),
    list(
      label = "single observation with mixed moderators",
      fit = quote(brma.norm(
        yi = yi, sei = sei, mods = ~ mod_cont + mod_fac,
        data = test_data_norm, only_data = TRUE
      )),
      newdata = quote(data.frame(
        yi = 0.25, sei = 0.1, mod_cont = 2.5,
        mod_fac = factor("B", levels = c("A", "B"))
      )),
      check = function(result) {
        expect_equal(nrow(result[["mods"]]), 1L)
        expect_equal(as.character(result[["mods"]][["mod_fac"]]), "B")
      }
    ),
    list(
      label = "zero-variance scale predictor",
      fit = quote(brma.norm(
        yi = yi, sei = sei, scale = ~ scale_var,
        data = test_data_norm, only_data = TRUE
      )),
      newdata = quote(data.frame(
        yi = c(0.1, 0.2), sei = c(0.1, 0.1), scale_var = c(1, 1)
      )),
      check = function(result) expect_equal(result[["scale"]][["scale_var"]], c(1, 1))
    ),
    list(
      label = "zero sei",
      fit = quote(brma.norm(
        yi = yi, sei = sei, data = test_data_norm, only_data = TRUE
      )),
      newdata = quote(data.frame(yi = c(0.1, 0.2, 0.3), sei = c(0, 0.1, 0))),
      check = function(result) expect_equal(result[["outcome"]][["sei"]], c(0, 0.1, 0))
    ),
    list(
      label = "zero vi",
      fit = quote(brma.norm(
        yi = yi, vi = vi, data = test_data_norm_vi, only_data = TRUE
      )),
      newdata = quote(data.frame(yi = c(0.1, 0.2), vi = c(0, 0))),
      check = function(result) expect_equal(result[["outcome"]][["sei"]], c(0, 0))
    ),
    list(
      label = "optional ni omitted",
      fit = quote(brma.norm(
        yi = yi, sei = sei, ni = ni,
        data = data.frame(yi = c(0.2, 0.5, -0.1),
                          sei = c(0.1, 0.15, 0.12),
                          ni = c(50, 100, 75)),
        only_data = TRUE
      )),
      newdata = quote(data.frame(yi = c(0.1, 0.2), sei = c(0, 0.1))),
      check = function(result) expect_equal(result[["outcome"]][["sei"]], c(0, 0.1))
    )
  )

  for (case in edge_cases) {
    fit <- eval(case[["fit"]])
    result <- RoBMA:::.prepare_newdata(
      object  = fit,
      newdata = eval(case[["newdata"]]),
      type    = "terms"
    )
    case[["check"]](result)
  }
})

# Fitting functions -----
.create_fit_priors <- function(data, priors) {

  # extract the common prior list
  prior_list <- priors[["outcome"]]

  # add levels to the baseline prior for outcomes
  if (.data_outcome_type(data) == "bin") {
    # encode number of levels for the baserate and random-effects prior
    attr(prior_list[["pi"]], "levels")    <- nrow(data[["outcome"]])
    attr(prior_list[["theta"]], "levels") <- nrow(data[["outcome"]])
  } else if (.data_outcome_type(data) == "pois") {
    # encode number of levels for the baserate and random-effects prior
    attr(prior_list[["phi"]], "levels")   <- nrow(data[["outcome"]])
    attr(prior_list[["theta"]], "levels") <- nrow(data[["outcome"]])
  }

  # add cluster-level indicators
  if (.is_data_multilevel(data)) {
    # encode number of levels for the random-effects prior
    attr(prior_list[["gamma"]], "levels") <- length(unique(data[["outcome"]][["cluster"]]))
  }

  ### deal with non-prior mixture distributions (bPET, bPEESE, and bselmodel)
  # the prior name must match the output parameter for sample extraction in plots etc
  if (.is_priors_bias(priors)) {
    if (!BayesTools::is.prior.mixture(prior_list[["bias"]])) {
      if (BayesTools::is.prior.weightfunction(prior_list[["bias"]])) {
        names(prior_list)[names(prior_list) == "bias"] <- "omega"
      } else if (.prior_has_phacking(prior_list[["bias"]])) {
        .selection_stop_phacking_deferred()
      } else if (BayesTools::is.prior.PET(prior_list[["bias"]])) {
        names(prior_list)[names(prior_list) == "bias"] <- "PET"
      } else if (BayesTools::is.prior.PEESE(prior_list[["bias"]])) {
        names(prior_list)[names(prior_list) == "bias"] <- "PEESE"
      }
    }
  }


  return(prior_list)
}
.create_fit_data   <- function(data, priors) {

  ### add outcome specific data
  if (.data_outcome_type(data) == "norm") {
    # always include yi and sei
    fit_data <- list(
      yi  = data[["outcome"]][["yi"]],
      sei = data[["outcome"]][["sei"]]
    )

    # flip effect size direction (needed for selection models, PET, and PEESE models)
    # (done for everything for consistency)
    effect_direction  <- .data_effect_direction(data)
    if (effect_direction == "negative") {
      fit_data[["yi"]] <- -1 * fit_data[["yi"]]
    }

    # add selection-kernel data for selection models
    if (.is_priors_weightfunction(priors)) {
      selection_spec <- .selection_spec(
        priors           = priors,
        yi               = fit_data[["yi"]],
        sei              = fit_data[["sei"]],
        effect_direction = effect_direction,
        signed_data      = TRUE
      )

      fit_data <- c(fit_data, selection_spec[["jags_data"]])

    }

  } else if (.data_outcome_type(data) == "bin") {
    # always include ai, ci, n1i, and n2i
    fit_data <- list(
      ai  = data[["outcome"]][["ai"]],
      ci  = data[["outcome"]][["ci"]],
      n1i = data[["outcome"]][["n1i"]],
      n2i = data[["outcome"]][["n2i"]]
    )
  } else if (.data_outcome_type(data) == "pois") {
    # always include x1i, x2i, t1i, and t2i
    fit_data <- list(
      x1i = data[["outcome"]][["x1i"]],
      x2i = data[["outcome"]][["x2i"]],
      t1i = data[["outcome"]][["t1i"]],
      t2i = data[["outcome"]][["t2i"]]
    )
  }

  ### add common data
  # add number of estimates
  fit_data[["K"]] <- nrow(data[["outcome"]])

  # add cluster for 3lvl models
  if (.is_data_multilevel(data)) {
    fit_data[["cluster"]] <- data[["outcome"]][["cluster"]]
  }

  # add weights for weighted models
  if (.is_data_weights(data)) {
    fit_data[["weight"]] <- data[["outcome"]][["weights"]]
  }



  return(fit_data)
}
.create_fit_formula_list       <- function(data, parameter) {

  formula <- attr(data[[parameter]], "formula")

  # for scale regression - modify formula for exponential parameterization
  # (required for the exponential parameterization trick in BayesTools::JAGS_fit)
  if (parameter == "scale") {

    # check whether the intercept was removed from the formula
    # if so, warn the user and return it back (intercepts cannot be omitted from scale models)
    if (attr(terms(formula), "intercept") == 0) {
      warning("Intercept cannot be omitted from scale models (the regression estimates a multiplicative constant for the intercept). The intercept removal term has been ignored.", call. = FALSE)
      formula <- BayesTools::formula_add_intercept(formula)
    }

    # add the corresponding attribute
    attr(formula, "log(intercept)") <- TRUE
  }

  return(formula)
}
.create_fit_formula_data_list  <- function(data, parameter) {
  return(data[[parameter]])
}
.create_fit_formula_prior_list <- function(priors, parameter) {
  return(priors[[parameter]])
}
.create_model_syntax           <- function(data, priors) {

  ### extract structural information about the model
  is_mods           <- .is_data_mods(data)
  is_scale          <- .is_data_scale(data)
  is_multilevel     <- .is_data_multilevel(data)
  is_weights        <- .is_data_weights(data)
  is_PET            <- .is_priors_PET(priors)
  is_PEESE          <- .is_priors_PEESE(priors)
  is_weightfunction <- .is_priors_weightfunction(priors)
  outcome_type      <- .data_outcome_type(data)
  effect_direction  <- .data_effect_direction(data)

  ### create the model syntax
  model_syntax <- "model{\n"

  selection_spec <- NULL
  if (is_weightfunction) {
    selection_spec <- .selection_spec(
      priors           = priors,
      yi               = data[["outcome"]][["yi"]],
      sei              = data[["outcome"]][["sei"]],
      effect_direction = effect_direction,
      signed_data      = TRUE
    )
  }

  ### the main model parameters are created automatically via BayesTools::JAGS_fit
  # - mu  (!is_mods)  / mu[i]  (is_mods)
  # - tau (!is_scale) / tau[i] (is_scale)
  # - rho (is_multilevel)
  # for publication bias
  # - PET            (is_PET)
  # - PEESE          (is_PEESE)
  # - weightfunction (is_weightfunction)

  ### prepare multilevel heterogeneity parameter (non-regression = defined outside of the for loop)
  # for multilevel: specify heterogeneity allocation
  # tau_within  - within-cluster variance (estimate-level variance)
  # tau_between - between-cluster variance (cluster-level variance)
  if (is_multilevel & !is_scale) {
    model_syntax <- paste0(model_syntax, "tau_within  = tau * sqrt(1-rho)\n")
    model_syntax <- paste0(model_syntax, "tau_between = tau * sqrt(rho)\n")
  }

  tau_total_node   <- if (is_scale) "tau[i]" else "tau"
  tau_within_node  <- if (is_multilevel) {
    if (is_scale) "tau_within[i]" else "tau_within"
  } else {
    tau_total_node
  }
  tau_between_node <- if (is_scale) "tau_between[i]" else "tau_between"

  if (is_weightfunction && isTRUE(selection_spec[["jags_use_step_switch"]])) {
    model_syntax <- paste0(
      model_syntax,
      "sel_kernel_mode_active = ",
      selection_spec[["jags_kernel_mode_expr"]],
      "\n"
    )
  }

  ### enter the main block
  model_syntax <- paste0(model_syntax, "for(i in 1:K){\n")

  ### prepare effect size parameter
  # flip effect size direction (needed for selection models, PET, and PEESE models)
  # (done for everything for consistency, data are flipped within `.create_fit_data()`)
  if (is_mods) {
    mu_estimate <- ifelse(effect_direction == "negative", paste0("- mu[i]"), "mu[i]")
  } else {
    mu_estimate <- ifelse(effect_direction == "negative", "- mu", "mu")
  }
  # add cluster-level effects
  if (is_multilevel) {
    mu_estimate <- paste0(mu_estimate, ifelse(effect_direction == "negative", " - ", " + "),  "gamma[cluster[i]] * ", tau_between_node)
  }
  # add PET/PEESE
  if (is_PET) {
    mu_estimate <- paste0(mu_estimate, " + PET * sei[i]")
  }
  if (is_PEESE) {
    mu_estimate <- paste0(mu_estimate, " + PEESE * pow(sei[i],2)")
  }


  ### prepare heterogeneity parameter (regression)
  # in the case of a regression: the scale model is specified on multiplicative scale:
  # tau = intercept * exp(sum(beta_i * x_i))
  # the model uses a trick to separately pass the tau_intercept and multiply it with
  # log_tau scale regression with zero intercept
  if (is_scale) {
    model_syntax <- paste0(model_syntax, "  tau[i] = exp(log_tau[i])\n")
  }
  # for multilevel: specify heterogeneity allocation & dispatch estimate-specific/common parameter
  # tau_within  - within-cluster variance (estimate-level variance)
  # tau_between - between-cluster variance (cluster-level variance)
  # for rest: dispatch estimate-specific/common parameter
  if (is_multilevel) {
    if (is_scale) {
      model_syntax <- paste0(model_syntax, "  tau_within[i]  = tau[i] * sqrt(1-rho)\n")
      model_syntax <- paste0(model_syntax, "  tau_between[i] = tau[i] * sqrt(rho)\n")
      tau_estimate <- tau_within_node
    } else {
      # the variance allocation performed outside of the loop
      tau_estimate <- tau_within_node
    }
  } else {
    tau_estimate <- tau_total_node
  }
  # compute the total marginal variance of the observed estimate
  total_var_expr <- paste0("( pow(sei[i],2) + pow(", tau_estimate, ",2) )")


  ### specify model likelihood
  # separate likelihoods for normal / selection / binomial models
  # subversion for weighted/unweighted likelihoods
  if (outcome_type == "norm") {

    # selection and normal models for norm data outcome
    if (is_weightfunction) {
      likelihood_weight_expr <- if (is_weights) "weight[i]" else "1"
      total_sd_node          <- "sel_total_sd[i]"
      model_syntax           <- paste0(
        model_syntax,
        "  ", total_sd_node, " = ", paste0("sqrt", total_var_expr), "\n"
      )
      if (identical(selection_spec[["mode"]], "step") &&
          isTRUE(selection_spec[["jags_use_step_switch"]])) {
        model_syntax <- paste0(
          model_syntax,
          "  yi[i] ~ dselnorm_step_switch(",
          mu_estimate, ",", total_sd_node, ",",
          "sei[i],", likelihood_weight_expr, ",",
          selection_spec[["jags_omega"]], ",",
          "sel_z_lower,sel_z_upper,sel_obs_bin[i],sel_sign,",
          selection_spec[["jags_kernel_mode"]], ",",
          "sel_telescope_probabilities)\n"
        )
      } else if (identical(selection_spec[["mode"]], "step")) {
        model_syntax <- paste0(
          model_syntax,
          "  yi[i] ~ dselnorm_step(",
          mu_estimate, ",", total_sd_node, ",",
          "sei[i],", likelihood_weight_expr, ",",
          selection_spec[["jags_omega"]], ",",
          "sel_z_lower,sel_z_upper,sel_obs_bin[i],sel_sign,",
          "sel_telescope_probabilities)\n"
        )
      } else {
        model_syntax <- paste0(
          model_syntax,
          "  yi[i] ~ dselnorm_kernel(",
          mu_estimate, ",", total_sd_node, ",",
          mu_estimate, ",", total_sd_node, ",",
          "sei[i],", likelihood_weight_expr, ",",
          selection_spec[["jags_omega"]], ",",
          "sel_z_lower,sel_z_upper,sel_obs_bin[i],sel_sign,",
          selection_spec[["jags_alpha"]], ",",
          selection_spec[["jags_phack_kind"]], ",",
          "phack_z_source,phack_z_dest,",
          "sel_segment_bounds,sel_segment_step_bin,sel_segment_phack_region,",
          "sel_kernel_mode)\n"
        )
      }
    } else {
      if (is_weights) {
        model_syntax <- paste0(model_syntax, "  yi[i] ~ dwnorm(", mu_estimate, ",", "1/", total_var_expr, ", weight[i])\n")
      } else {
        model_syntax <- paste0(model_syntax, "  yi[i] ~ dnorm(",  mu_estimate, ",", "1/", total_var_expr, ")\n")
      }
    }

  } else if (outcome_type %in% c("bin", "pois")) {

    # estimate-level variance is not marginalized in binomial-normal/poisson-normal models
    mu_estimate <- paste0(mu_estimate, " + theta[i] * ", tau_estimate)

    # specify appropriate link for each type of outcome
    if (outcome_type == "bin") {
      # transform the parameters to the probability scale
      model_syntax <- paste0(model_syntax, "  logit(p1[i]) = logit(pi[i]) + 0.5 * (", mu_estimate, ")\n")
      model_syntax <- paste0(model_syntax, "  logit(p2[i]) = logit(pi[i]) - 0.5 * (", mu_estimate, ")\n")
    } else if (outcome_type == "pois") {
      # transform the parameters to the probability scale
      model_syntax <- paste0(model_syntax, "  log(r1[i]) = phi[i] + 0.5 * (", mu_estimate, ") + log(t1i[i])\n")
      model_syntax <- paste0(model_syntax, "  log(r2[i]) = phi[i] - 0.5 * (", mu_estimate, ") + log(t2i[i])\n")
    }

    # specify appropriate likelihoods
    if (outcome_type == "bin") {
      # the observed data
      if (is_weights) {
        model_syntax <- paste0(model_syntax, "  ai[i] ~ dwbinom(p1[i], n1i[i], weight[i])\n")
        model_syntax <- paste0(model_syntax, "  ci[i] ~ dwbinom(p2[i], n2i[i], weight[i])\n")
      } else {
        model_syntax <- paste0(model_syntax, "  ai[i] ~ dbinom(p1[i], n1i[i])\n")
        model_syntax <- paste0(model_syntax, "  ci[i] ~ dbinom(p2[i], n2i[i])\n")
      }
    } else if (outcome_type == "pois") {
      # the observed data
      if (is_weights) {
        model_syntax <- paste0(model_syntax, "  x1i[i] ~ dwpois(r1[i], weight[i])\n")
        model_syntax <- paste0(model_syntax, "  x2i[i] ~ dwpois(r2[i], weight[i])\n")
      } else {
        model_syntax <- paste0(model_syntax, "  x1i[i] ~ dpois(r1[i])\n")
        model_syntax <- paste0(model_syntax, "  x2i[i] ~ dpois(r2[i])\n")
      }
    }



  }


  model_syntax <- paste0(model_syntax, "}\n")
  model_syntax <- paste0(model_syntax, "}")

  return(model_syntax)
}

.fit <- function(object, extend = FALSE) {

  fit_control        <- object[["fit_control"]]
  autofit_control    <- object[["autofit_control"]]
  convergence_checks <- object[["convergence_checks"]]
  data               <- object[["data"]]
  priors             <- object[["priors"]]

  errors   <- NULL
  warnings <- NULL

  ### create arguments to be passed to BayesTools::JAGS_fit
  fit_formula_list        <- list()
  fit_formula_data_list   <- list()
  fit_formula_prior_list  <- list()
  fit_formula_scale_list  <- list()

  ### create model base
  fit_priors <- .create_fit_priors(data = data, priors = priors)
  fit_data   <- .create_fit_data(data = data, priors = priors)

  ### add effect regressions
  if (.is_data_mods(data)) {
    fit_formula_list[["mu"]]       <- .create_fit_formula_list(data = data, parameter = "mods")
    fit_formula_data_list[["mu"]]  <- .create_fit_formula_data_list(data = data, parameter = "mods")
    fit_formula_prior_list[["mu"]] <- .create_fit_formula_prior_list(priors = priors, parameter = "mods")
    fit_formula_scale_list[["mu"]] <- .data_standardize_continuous_predictors(data)
  }

  ### add heterogeneity regressions
  if (.is_data_scale(data)) {
    fit_formula_list[["log_tau"]]       <- .create_fit_formula_list(data = data, parameter = "scale")
    fit_formula_data_list[["log_tau"]]  <- .create_fit_formula_data_list(data = data, parameter = "scale")
    fit_formula_prior_list[["log_tau"]] <- .create_fit_formula_prior_list(priors = priors, parameter = "scale")
    fit_formula_scale_list[["log_tau"]] <- .data_standardize_continuous_predictors(data)
  }

  ### generate the model syntax
  model_syntax <- .create_model_syntax(data = data, priors = priors)

  ### fit the model
  if (!extend || length(object[["fit"]]) == 0) {

    fit <- BayesTools::JAGS_fit(
      model_syntax          = model_syntax,
      data                  = fit_data,
      prior_list            = fit_priors,
      formula_list          = if (length(fit_formula_list)       > 0) fit_formula_list,
      formula_data_list     = if (length(fit_formula_data_list)  > 0) fit_formula_data_list,
      formula_prior_list    = if (length(fit_formula_prior_list) > 0) fit_formula_prior_list,
      formula_scale         = if (length(fit_formula_scale_list) > 0) fit_formula_scale_list,
      chains                = fit_control[["chains"]],
      adapt                 = fit_control[["adapt"]],
      burnin                = fit_control[["burnin"]],
      sample                = fit_control[["sample"]],
      thin                  = fit_control[["thin"]],
      autofit               = fit_control[["autofit"]],
      autofit_control       = autofit_control,
      parallel              = fit_control[["parallel"]],
      cores                 = fit_control[["cores"]],
      silent                = fit_control[["silent"]],
      seed                  = fit_control[["seed"]],
      required_packages     = "RoBMA",
      is_JASP               = object[["is_JASP"]],
      is_JASP_prefix        = object[["is_JASP_prefix"]]
    )

  } else {

    fit <- BayesTools::JAGS_extend(
      fit                = object[["fit"]],
      autofit_control    = autofit_control,
      parallel           = fit_control[["parallel"]],
      cores              = fit_control[["cores"]],
      silent             = fit_control[["silent"]],
      seed               = fit_control[["seed"]]
    )

  }


  # assess the model fit and deal with errors
  if (inherits(fit, "error")) {

    if(grepl("Unknown function", fit$message))
      stop("The RoBMA module could not be loaded. Check whether the RoBMA package was installed correctly and whether 'RoBMA::RoBMA.private$module_location' contains path to the RoBMA JAGS module.", call. = FALSE)

    fit_error_message       <- fit$message
    fit                     <- list()
    attr(fit, "prior_list") <- fit_priors

    converged      <- FALSE
    has_posterior  <- FALSE
    errors         <- c(errors, fit_error_message)

  } else {

    has_posterior <- TRUE
    check_fit     <- BayesTools::JAGS_check_convergence(
      fit          = fit,
      prior_list   = attr(fit, "prior_list"),
      max_Rhat     = convergence_checks[["max_Rhat"]],
      min_ESS      = convergence_checks[["min_ESS"]],
      max_error    = convergence_checks[["max_error"]],
      max_SD_error = convergence_checks[["max_SD_error"]]
    )
    warnings  <- c(warnings, attr(fit, "warnings"), attr(check_fit, "errors"))
    converged <- check_fit

  }

  # add results
  fit$errors        <- errors
  fit$warnings      <- warnings
  fit$converged     <- converged
  fit$has_posterior <- has_posterior

  return(fit)
}

.is_priors_PET            <- function(priors) {

  if (is.null(priors[["outcome"]][["bias"]]))
    return(FALSE)

  if (is.prior.mixture(priors[["outcome"]][["bias"]]))
    return(any(sapply(priors[["outcome"]][["bias"]], is.prior.PET)))

  return(is.prior.PET(priors[["outcome"]][["bias"]]))
}
.is_priors_PEESE          <- function(priors) {

  if (is.null(priors[["outcome"]][["bias"]]))
    return(FALSE)

  if (is.prior.mixture(priors[["outcome"]][["bias"]]))
    return(any(sapply(priors[["outcome"]][["bias"]], is.prior.PEESE)))

  return(is.prior.PEESE(priors[["outcome"]][["bias"]]))
}
.is_priors_weightfunction <- function(priors) {

  if (is.null(priors[["outcome"]][["bias"]]))
    return(FALSE)

  if (is.prior.mixture(priors[["outcome"]][["bias"]]))
    return(any(sapply(priors[["outcome"]][["bias"]], .prior_is_selection_kernel)))

  return(.prior_is_selection_kernel(priors[["outcome"]][["bias"]]))
}
.is_priors_bias           <- function(priors) {
  return(.is_priors_PET(priors) || .is_priors_PEESE(priors) || .is_priors_weightfunction(priors))
}
.is_data_multilevel       <- function(data) {
  return(isTRUE(attr(data, "cluster")))
}
.is_data_mods             <- function(data) {
  return(isTRUE(attr(data, "mods")))
}
.is_data_scale            <- function(data) {
  return(isTRUE(attr(data, "scale")))
}
.is_data_weights          <- function(data) {
  return(isTRUE(attr(data, "weights")))
}
.data_outcome_type        <- function(data) {
  return(attr(data, "outcome_type"))
}
.data_measure             <- function(data) {
  return(attr(data, "measure"))
}
.data_effect_direction    <- function(data) {
  return(attr(data, "effect_direction"))
}
.data_standardize_continuous_predictors <- function(data) {
  return(attr(data, "standardize_continuous_predictors"))
}

.is_PET            <- function(object) .is_priors_PET(object[["priors"]])
.is_PEESE          <- function(object) .is_priors_PEESE(object[["priors"]])
.is_weightfunction <- function(object) .is_priors_weightfunction(object[["priors"]])
.is_bias           <- function(object) .is_priors_bias(object[["priors"]])
.is_RoBMA          <- function(object) inherits(object, "RoBMA")
.is_BMA            <- function(object) inherits(object, "BMA.norm") || inherits(object, "BMA.glmm")
.is_robust_RoBMA   <- function(object) .is_RoBMA(object) && !.is_BMA(object)
.is_multilevel     <- function(object) .is_data_multilevel(object[["data"]])
.is_mods           <- function(object) .is_data_mods(object[["data"]])
.is_scale          <- function(object) .is_data_scale(object[["data"]])
.is_weights        <- function(object) .is_data_weights(object[["data"]])
.outcome_type      <- function(object) .data_outcome_type(object[["data"]])
.measure           <- function(object) .data_measure(object[["data"]])
.effect_direction  <- function(object) .data_effect_direction(object[["data"]])
.standardize_continuous_predictors  <- function(object) .data_standardize_continuous_predictors(object[["data"]])

context("Model fitting for brma.glmm")

# Load common test helpers
source(testthat::test_path("common-functions.R"))
skip_on_cran()
skip_if_not_installed("metadat")
skip_if_not_installed("metafor")
skip_refit_if_cached("brma.glmm")

### Uses examples from the metafor package
test_that("brma.glmm fits binomial and Poisson metafor-reference models", {
  ### fit generalized meta-analytic model to difference in two proportions
  data(dat.bcg, package = "metadat")
  fit_simple.metafor <- metafor::rma.glmm(measure = "OR", ai = tpos, bi = tneg, ci = cpos, di = cneg, data = dat.bcg, model = "UM.FS")

  # using RoBMA package
  fit_simple.brma <- brma.glmm(ai = tpos, bi = tneg, ci = cpos, di = cneg, data = dat.bcg, measure = "OR", seed = 1, silent = TRUE)
  fit_simple.brma <- add_marglik(fit_simple.brma)
  fit_simple.brma <- suppressWarnings(add_loo(fit_simple.brma))
  save_fit("bcg_glmm", fit_simple.brma, info = list(metafor = fit_simple.metafor))
  expect_s3_class(fit_simple.brma, "brma.glmm")


  ### fit generalized meta-regression
  fit_reg.metafor <- suppressWarnings(metafor::rma.glmm(measure = "OR", ai = tpos, bi = tneg, ci = cpos, di = cneg, mods = ~ alloc, data = dat.bcg, model = "UM.FS"))

  # using RoBMA package
  fit_reg.brma <- brma.glmm(ai = tpos, bi = tneg, ci = cpos, di = cneg, mods = ~ alloc, data = dat.bcg, measure = "OR", seed = 1, silent = TRUE)
  fit_reg.brma <- add_marglik(fit_reg.brma)
  fit_reg.brma <- suppressWarnings(add_loo(fit_reg.brma))
  save_fit("bcg_glmm_reg", fit_reg.brma, info = list(mods = c("alloc"), metafor = fit_reg.metafor))
  expect_s3_class(fit_reg.brma, "brma.glmm")


  ### fit generalized meta-analytic model to difference in two rations
  data(dat.nielweise2008, package = "metadat")
  fit_simple.metafor <- metafor::rma.glmm(measure = "IRR", x1i = x1i, t1i = t1i, x2i = x2i, t2i = t2i, data = dat.nielweise2008, model = "UM.FS")

  fit_simple.brma <- brma.glmm(x1i = x1i, t1i = t1i, x2i = x2i, t2i = t2i, data = dat.nielweise2008, measure = "IRR", seed = 1, silent = TRUE)
  fit_simple.brma <- add_marglik(fit_simple.brma)
  fit_simple.brma <- suppressWarnings(add_loo(fit_simple.brma))
  save_fit("nielweise2008_glmm", fit_simple.brma, info = list(metafor = fit_simple.metafor))
  expect_s3_class(fit_simple.brma, "brma.glmm")
})

test_that("brma.glmm handles multilevel scale regression model", {
  # using RoBMA package
  data(dat.bcg, package = "metadat")
  fit_simple.brma <- brma.glmm(
    ai = tpos, bi = tneg, ci = cpos, di = cneg,
    scale = ~ year, cluster = alloc,
    data = dat.bcg, measure = "OR",
    chains = 2, sample = 1000, burnin = 500, adapt = 500,
    seed = 1, silent = TRUE
  )
  fit_simple.brma <- add_marglik(fit_simple.brma)
  fit_simple.brma <- suppressWarnings(add_loo(fit_simple.brma))
  save_fit("bcg_glmm_3lvl_scale", fit_simple.brma, info = list(scale = c("year")))

  expect_s3_class(fit_simple.brma, "brma.glmm")
  expect_true(.is_multilevel(fit_simple.brma))
  expect_true(.is_scale(fit_simple.brma))
  expect_no_error(summary(fit_simple.brma))
})

context("Prior input handling for brma.norm")

source(testthat::test_path("helper-contracts.R"))

# Test data for prior specification tests
test_data <- data.frame(
  effect     = c(0.10, 0.25, 0.15, 0.30, 0.05),
  std_err    = sqrt(c(0.04, 0.06, 0.05, 0.08, 0.03)),
  n          = c(50L, 75L, 60L, 40L, 90L),
  mod_cont   = c(1.5, 2.3, 1.8, 3.1, 0.9),
  mod_factor = factor(c("A", "B", "A", "B", "A")),
  scale_var  = c(0.5, 1.0, 0.8, 1.2, 0.6),
  cluster    = c("g1", "g1", "g2", "g2", "g3"),
  stringsAsFactors = FALSE
)


# ============================================================================
# Tests for measure specification and UISD
# ============================================================================

test_that("Default priors scale with known UISD for each measure", {

  skip_on_cran()

  # Test all known UISD measures
  measures_uisd <- list(SMD = sqrt(2), ZCOR = 1, OR = 2, RR = 2, HR = 2)

  for (measure in names(measures_uisd)) {
    result <- brma.norm(
      yi = effect, sei = std_err, data = test_data,
      measure = measure, only_priors = TRUE
    )[["priors"]]

    expected_effect_sd <- measures_uisd[[measure]] * RoBMA.get_option("default_UISD.effect")
    expected_het_sd    <- measures_uisd[[measure]] * RoBMA.get_option("default_UISD.heterogeneity")

    expect_true(BayesTools::is.prior(result$outcome$mu))
    expect_true(BayesTools::is.prior(result$outcome$tau))
    expect_equal(result$outcome$mu$parameters$sd, expected_effect_sd)
    expect_equal(result$outcome$tau$parameters$sd, expected_het_sd)
    expect_equal(result$outcome$tau$truncation$lower, 0)
  }
})


test_that("UISD is estimated from sample sizes for GEN measure", {

  skip_on_cran()

  result <- brma.norm(
    yi = effect, sei = std_err, ni = n, data = test_data,
    measure = "GEN", only_priors = TRUE
  )[["priors"]]

  expected_uisd <- estimate_unit_information_sd(sei = test_data$std_err, ni = test_data$n)
  expected_effect_sd <- expected_uisd * RoBMA.get_option("default_UISD.effect")

  expect_equal(result$outcome$mu$parameters$sd, expected_effect_sd, tolerance = 1e-6)
})


test_that("GEN measure without ni is rejected", {

  skip_on_cran()

  expect_error(
    brma.norm(yi = effect, sei = std_err, data = test_data, measure = "GEN", only_priors = TRUE),
    regexp = "ni|unit_information_sd|UISD"
  )
})

test_that("GEN measure rejects incomplete ni for UISD defaults", {

  skip_on_cran()

  expect_error(
    brma.norm(
      yi          = effect,
      sei         = std_err,
      ni          = c(50L, NA, 60L, 40L, 90L),
      data        = test_data,
      measure     = "GEN",
      only_priors = TRUE
    ),
    regexp = "ni.*complete|unit information"
  )
})


test_that("Invalid measure is rejected", {

  skip_on_cran()

  expect_error(
    brma.norm(yi = effect, sei = std_err, data = test_data, measure = "INVALID", only_priors = TRUE),
    regexp = "measure"
  )
})


test_that("Normal constructors require explicit measure", {

  skip_on_cran()

  constructors <- list(
    brma      = brma,
    BMA       = BMA,
    RoBMA     = RoBMA,
    bPET      = bPET,
    bPEESE    = bPEESE,
    bselmodel = bselmodel
  )

  for (constructor in constructors) {
    expect_error(
      do.call(constructor, list(
        yi                        = test_data$effect,
        sei                       = test_data$std_err,
        prior_unit_information_sd = 1,
        only_priors               = TRUE
      )),
      regexp = "requires explicit 'measure'"
    )
  }
})


# ============================================================================
# Tests for prior_unit_information_sd and estimate_unit_information_sd
# ============================================================================

test_that("Manual prior_unit_information_sd is used and overrides known UISD", {

  skip_on_cran()

  custom_uisd <- 3.0

  # Should work for GEN measure
  result_gen <- brma.norm(
    yi = effect, sei = std_err, ni = n, data = test_data,
    measure = "GEN", prior_unit_information_sd = custom_uisd, only_priors = TRUE
  )[["priors"]]

  # Should override SMD's known UISD
  result_smd <- brma.norm(
    yi = effect, sei = std_err, ni = n, data = test_data,
    measure = "SMD", prior_unit_information_sd = custom_uisd, only_priors = TRUE
  )[["priors"]]

  expected_sd <- custom_uisd * RoBMA.get_option("default_UISD.effect")
  expect_equal(result_gen$outcome$mu$parameters$sd, expected_sd)
  expect_equal(result_smd$outcome$mu$parameters$sd, expected_sd)
})


test_that("estimate_unit_information_sd computes UISD and validates inputs", {

  skip_on_cran()

  sei <- c(0.2, 0.3, 0.25, 0.15)
  ni  <- c(50, 30, 40, 80)

  # Correct computation
  result <- estimate_unit_information_sd(sei = sei, ni = ni)
  expected <- sqrt(sum(ni) / sum(sei^(-2)))
  expect_equal(result, expected, tolerance = 1e-10)

  expect_error_cases(list(
    list(
      label  = "negative sei",
      expr   = quote(estimate_unit_information_sd(sei = c(-0.2, 0.3), ni = c(50, 30))),
      regexp = "sei"
    ),
    list(
      label  = "zero sei",
      expr   = quote(estimate_unit_information_sd(sei = c(0, 0.3), ni = c(50, 30))),
      regexp = "sei"
    ),
    list(
      label  = "negative ni",
      expr   = quote(estimate_unit_information_sd(sei = c(0.2, 0.3), ni = c(-50, 30))),
      regexp = "ni"
    ),
    list(
      label  = "length mismatch",
      expr   = quote(estimate_unit_information_sd(sei = c(0.2, 0.3, 0.4), ni = c(50, 30))),
      regexp = "length|ni"
    ),
    list(
      label  = "missing sei",
      expr   = quote(estimate_unit_information_sd(sei = c(0.2, NA), ni = c(50, 30))),
      regexp = "sei"
    )
  ))
})


test_that("Invalid prior_unit_information_sd is rejected", {

  skip_on_cran()

  expect_error_cases(list(
    list(
      label  = "negative UISD",
      expr   = quote(brma.norm(
        yi = effect, sei = std_err, data = test_data,
        measure = "GEN",
        prior_unit_information_sd = -1, only_priors = TRUE
      )),
      regexp = "prior_unit_information_sd"
    ),
    list(
      label  = "zero UISD",
      expr   = quote(brma.norm(
        yi = effect, sei = std_err, data = test_data,
        measure = "GEN",
        prior_unit_information_sd = 0, only_priors = TRUE
      )),
      regexp = "prior_unit_information_sd"
    )
  ))
})


# ============================================================================
# Tests for constructor aliases and bias prior defaults
# ============================================================================

test_that("brma.norm alias is not overwritten by bselmodel", {

  skip_on_cran()

  result_alias <- brma(
    yi = effect, sei = std_err, data = test_data,
    measure = "SMD", only_data = TRUE
  )
  result       <- brma.norm(
    yi = effect, sei = std_err, data = test_data,
    measure = "SMD", only_data = TRUE
  )

  expect_identical(class(result_alias), class(result))
  expect_s3_class(result, "brma.norm")
  expect_false("bselmodel" %in% class(result))
  expect_false("prior_bias" %in% names(formals(brma.norm)))
})


test_that("Explicit NULL prior_bias uses defaults for single bias models", {

  skip_on_cran()

  result_sel <- bselmodel(
    yi = effect, sei = std_err, data = test_data,
    measure = "SMD", prior_bias = NULL, only_priors = TRUE
  )[["priors"]]
  expect_true(BayesTools::is.prior.weightfunction(result_sel$outcome$bias))

  result_pet <- bPET(
    yi = effect, sei = std_err, data = test_data,
    measure = "SMD", prior_bias = NULL, only_priors = TRUE
  )[["priors"]]
  expect_true(BayesTools::is.prior.PET(result_pet$outcome$bias))
})

test_that("bselmodel keeps only steps shortcut for default weightfunction", {

  skip_on_cran()

  result <- bselmodel(
    yi = effect, sei = std_err, data = test_data,
    measure = "SMD", steps = c(0.05, 0.10), only_priors = TRUE
  )[["priors"]]

  expect_true(BayesTools::is.prior.weightfunction(result$outcome$bias))
  expect_equal(result$outcome$bias$side, "one-sided")
  expect_equal(result$outcome$bias$steps, c(0.05, 0.10))
  expect_equal(result$outcome$bias$weights$type, "cumulative")
  expect_equal(result$outcome$bias$weights$alpha, rep(RoBMA.get_option("default_bias_weightfunction.alpha"), 3))

  prior_bias <- prior_weightfunction("two-sided", c(0.05), wf_fixed(c(1, 0.5)))
  result <- bselmodel(
    yi = effect, sei = std_err, data = test_data,
    measure = "SMD", prior_bias = prior_bias, only_priors = TRUE
  )[["priors"]]

  expect_equal(result$outcome$bias$side, "two-sided")
  expect_equal(result$outcome$bias$weights$type, "fixed")
  expect_equal(result$outcome$bias$weights$omega, c(1, 0.5))
})

test_that("RoBMA accepts fixed zero weightfunctions at prior setup", {

  skip_on_cran()

  zero_prior <- BayesTools::prior_weightfunction(
    "one-sided", c(.05), BayesTools::wf_fixed(c(1, 0))
  )
  valid_prior <- BayesTools::prior_weightfunction(
    "one-sided", c(.05), BayesTools::wf_fixed(c(1, .5))
  )

  expect_equal(zero_prior$weights$omega, c(1, 0))
  expect_silent(
    bselmodel(
      yi = effect, sei = std_err, data = test_data,
      measure = "SMD", prior_bias = zero_prior, only_priors = TRUE
    )
  )
  expect_silent(
    bselmodel(
      yi = effect, sei = std_err, data = test_data,
      measure = "SMD",
      prior_bias = BayesTools::prior_bias(selection = zero_prior),
      only_priors = TRUE
    )
  )
  expect_silent(
    RoBMA(
      yi = effect, sei = std_err, data = test_data,
      measure = "SMD",
      prior_bias = list(valid_prior, zero_prior),
      prior_unit_information_sd = 1,
      only_priors = TRUE,
      silent = TRUE
    )
  )
})

test_that("selection backend consumes BayesTools omega p-order directly", {

  skip_on_cran()

  prior_bias <- prior_weightfunction(
    "one-sided", c(0.025, 0.05), wf_fixed(c(1, 0.5, 0.1))
  )
  object <- bselmodel(
    yi = effect, sei = std_err, data = test_data,
    measure = "SMD", prior_bias = prior_bias, only_priors = TRUE
  )

  fit_data <- .create_fit_data(object[["data"]], object[["priors"]])
  syntax   <- .create_model_syntax(object[["data"]], object[["priors"]])

  expect_true(all(c(
    "sel_z_lower", "sel_z_upper", "sel_obs_bin", "sel_sign"
  ) %in% names(fit_data)))
  expect_false(any(grepl("sel_phack|phack_z|sel_segment|sel_kernel_mode", names(fit_data))))
  expect_match(syntax, "dselnorm_step", fixed = TRUE)
  expect_false(grepl("dselnorm_kernel|sel_phack|phack_z|sel_segment|sel_kernel_mode", syntax))

  omega_p <- matrix(c(1, 0.5, 0.1), nrow = 1)
  sei     <- object[["data"]][["outcome"]][["sei"]][1]
  yi_eval <- stats::qnorm(.04, lower.tail = FALSE) * sei

  selection_spec <- .selection_spec(
    priors           = object[["priors"]],
    yi               = yi_eval,
    sei              = sei,
    effect_direction = .effect_direction(object)
  )
  log_lik <- .selnorm_kernel_loglik_matrix(
    yi             = yi_eval,
    mu_num         = matrix(0, nrow = 1, ncol = 1),
    sigma_num      = matrix(sei, nrow = 1, ncol = 1),
    sei            = sei,
    omega          = omega_p,
    selection_spec = selection_spec
  )
  prob <- stats::pnorm(
    selection_spec[["z_upper"]],
    mean = 0, sd = sei / sei
  ) - stats::pnorm(
    selection_spec[["z_lower"]],
    mean = 0, sd = sei / sei
  )
  ref_log_lik <- stats::dnorm(yi_eval, mean = 0, sd = sei, log = TRUE) +
    log(omega_p[selection_spec[["obs_bin"]]]) -
    log(sum(omega_p * prob))

  expect_equal(
    as.numeric(log_lik), as.numeric(ref_log_lik),
    tolerance = 1e-12
  )
})


test_that("bPEESE default respects supplied prior_unit_information_sd", {

  skip_on_cran()

  result <- bPEESE(
    yi = effect, sei = std_err, data = test_data,
    measure = "SMD", prior_unit_information_sd = 10,
    only_priors = TRUE
  )[["priors"]]

  expect_equal(
    result$outcome$bias$parameters$scale,
    RoBMA.get_option("default_bias_PEESE.scale") * sqrt(2) / 10
  )
})


# ============================================================================
# Tests for rescale_priors
# ============================================================================

test_that("rescale_priors scales prior distributions", {

  skip_on_cran()

  result_base <- brma.norm(yi = effect, sei = std_err, data = test_data, measure = "SMD", only_priors = TRUE)[["priors"]]

  for (scale_factor in c(0.5, 2)) {
    result_rescaled <- brma.norm(
      yi = effect, sei = std_err, data = test_data,
      measure = "SMD", rescale_priors = scale_factor, only_priors = TRUE
    )[["priors"]]
    expect_equal(result_rescaled$outcome$mu$parameters$sd, result_base$outcome$mu$parameters$sd * scale_factor)
    expect_equal(result_rescaled$outcome$tau$parameters$sd, result_base$outcome$tau$parameters$sd * scale_factor)
  }

  expect_error_cases(list(
    list(
      label  = "negative rescale_priors",
      expr   = quote(brma.norm(
        yi = effect, sei = std_err, data = test_data,
        measure = "SMD",
        rescale_priors = -1, only_priors = TRUE
      )),
      regexp = "rescale_priors"
    ),
    list(
      label  = "zero rescale_priors",
      expr   = quote(brma.norm(
        yi = effect, sei = std_err, data = test_data,
        measure = "SMD",
        rescale_priors = 0, only_priors = TRUE
      )),
      regexp = "rescale_priors"
    )
  ))
})


# ============================================================================
# Tests for custom priors
# ============================================================================

test_that("Custom prior_effect and prior_heterogeneity rescale", {

  skip_on_cran()

  custom_effect <- BayesTools::prior("normal", parameters = list(mean = 0, sd = 0.5))
  custom_het    <- BayesTools::prior("normal", parameters = list(mean = 0, sd = 0.3), truncation = list(0, Inf))

  result <- brma.norm(
    yi = effect, sei = std_err, data = test_data,
    measure = "SMD",
    prior_effect = custom_effect, prior_heterogeneity = custom_het,
    only_priors = TRUE
  )[["priors"]]

  expect_equal(result$outcome$mu$parameters$sd, 0.5)
  expect_equal(result$outcome$tau$parameters$sd, 0.3)

  # With rescaling
  result_rescaled <- brma.norm(
    yi = effect, sei = std_err, data = test_data,
    measure = "SMD",
    prior_effect = custom_effect, prior_heterogeneity = custom_het, rescale_priors = 2, only_priors = TRUE
  )[["priors"]]
  expect_equal(result_rescaled$outcome$mu$parameters$sd,  0.5 * 2)
  expect_equal(result_rescaled$outcome$tau$parameters$sd, 0.3 * 2)
})


test_that("Different prior distribution types are accepted", {

  skip_on_cran()

  # Spike prior
  spike <- BayesTools::prior("spike", parameters = list(location = 0))
  result_spike <- brma.norm(yi = effect, sei = std_err, data = test_data, prior_effect = spike, measure = "OR", only_priors = TRUE)[["priors"]]
  expect_true(BayesTools::is.prior.point(result_spike$outcome$mu))

  # t/Cauchy prior (rescalable via scale parameter)
  cauchy <- BayesTools::prior("cauchy", parameters = list(location = 0, scale = 0.707))
  result_cauchy <- brma.norm(yi = effect, sei = std_err, data = test_data, prior_effect = cauchy, measure = "RR", only_priors = TRUE)[["priors"]]
  expect_equal(result_cauchy$outcome$mu$distribution, "t")

  result_cauchy_rescaled <- brma.norm(
    yi = effect, sei = std_err, data = test_data,
    prior_effect = cauchy, rescale_priors = 2, measure = "HR", only_priors = TRUE
  )[["priors"]]
  expect_equal(result_cauchy_rescaled$outcome$mu$parameters$scale, 0.707 * 2)
})


# ============================================================================
# Tests for informed priors
# ============================================================================

test_that("Informed priors are accepted for medicine field with rescaling", {

  skip_on_cran()

  # Test general priors
  result <- brma.norm(
    yi = effect, sei = std_err, data = test_data,
    measure = "OR", prior_informed_field = "medicine", only_priors = TRUE
  )[["priors"]]
  expect_equal(result$outcome$mu, BayesTools::prior_informed(
    name = "cochrane", parameter = "effect", type = "logOR"))
  expect_equal(result$outcome$tau, BayesTools::prior_informed(
    name = "cochrane", parameter = "heterogeneity", type = "logOR"))

  # With subfield
  result_subfield <- brma.norm(
    yi = effect, sei = std_err, data = test_data,
    measure = "SMD", prior_informed_field = "medicine", prior_informed_subfield = "neonatal",
    only_priors = TRUE
  )[["priors"]]
  expect_equal(result_subfield$outcome$mu, BayesTools::prior_informed(
    name = "neonatal", parameter = "effect", type = "smd"))
  expect_equal(result_subfield$outcome$tau, BayesTools::prior_informed(
    name = "neonatal", parameter = "heterogeneity", type = "smd"))

  # With rescaling
  result_base <- brma.norm(
    yi = effect, sei = std_err, data = test_data,
    measure = "SMD", prior_informed_field = "medicine", only_priors = TRUE
  )[["priors"]]
  result_rescaled <- brma.norm(
    yi = effect, sei = std_err, data = test_data,
    measure = "SMD", prior_informed_field = "medicine", rescale_priors = 2, only_priors = TRUE
  )[["priors"]]

  scale_param <- if ("sd" %in% names(result_base$outcome$mu$parameters)) "sd" else "scale"
  expect_equal(
    result_rescaled$outcome$mu$parameters[[scale_param]],
    result_base$outcome$mu$parameters[[scale_param]] * 2
  )
})


test_that("Conflicting prior specifications are rejected", {

  skip_on_cran()

  expect_error_cases(list(
    list(
      label  = "invalid informed-prior field",
      expr   = quote(brma.norm(
        yi = effect, sei = std_err, data = test_data,
        measure = "SMD",
        prior_informed_field = "invalid", only_priors = TRUE
      )),
      regexp = "prior_informed_field"
    ),
    list(
      label  = "UISD conflicts with informed priors",
      expr   = quote(brma.norm(
        yi = effect, sei = std_err, data = test_data,
        measure = "SMD",
        prior_unit_information_sd = 1.5, prior_informed_field = "medicine",
        only_priors = TRUE
      )),
      regexp = "prior_unit_information_sd|prior_informed_field|mutually exclusive"
    )
  ))
})


test_that("Informed scale priors are assigned", {

  skip_on_cran()

  result <- brma.norm(
    yi = effect, sei = std_err, scale = ~ scale_var,
    data = test_data, measure = "SMD",
    prior_informed_field = "medicine", only_priors = TRUE
  )[["priors"]]

  expect_equal(
    result$scale$scale_var$parameters$sd,
    RoBMA.get_option("default_informed_priors.scale")
  )
})


# ============================================================================
# Tests for heterogeneity allocation prior (cluster)
# ============================================================================

test_that("Heterogeneity allocation priors are assigned", {

  skip_on_cran()

  # Default Beta(1,1) when cluster specified
  result <- brma.norm(
    yi = effect, sei = std_err, cluster = cluster, data = test_data,
    measure = "SMD", only_priors = TRUE
  )[["priors"]]
  expect_true("rho" %in% names(result$outcome))
  expect_equal(result$outcome$rho$distribution, "beta")
  expect_equal(result$outcome$rho$parameters$alpha, 1)
  expect_equal(result$outcome$rho$parameters$beta, 1)

  # Custom prior
  custom_prior <- BayesTools::prior("beta", parameters = list(alpha = 2, beta = 2))
  result_custom <- brma.norm(
    yi = effect, sei = std_err, cluster = cluster, data = test_data,
    prior_heterogeneity_allocation = custom_prior, measure = "OR", only_priors = TRUE
  )[["priors"]]
  expect_equal(result_custom$outcome$rho$parameters$alpha, 2)

  # No rho when cluster not specified
  result_no_ids <- brma.norm(yi = effect, sei = std_err, data = test_data, measure = "SMD", only_priors = TRUE)
  expect_false("rho" %in% names(result_no_ids$outcome))
})


test_that("Constrained point priors outside support are rejected", {

  skip_on_cran()

  expect_error_cases(list(
    list(
      label  = "negative heterogeneity spike",
      expr   = quote(brma.norm(
        yi = effect, sei = std_err, data = test_data,
        measure = "SMD",
        prior_heterogeneity = BayesTools::prior("spike", parameters = list(location = -0.1)),
        only_priors = TRUE
      )),
      regexp = "prior_heterogeneity"
    ),
    list(
      label  = "heterogeneity-allocation spike above one",
      expr   = quote(brma.norm(
        yi = effect, sei = std_err, cluster = cluster, data = test_data,
        measure = "SMD",
        prior_heterogeneity_allocation = BayesTools::prior("spike", parameters = list(location = 1.1)),
        only_priors = TRUE
      )),
      regexp = "heterogeneity_allocation"
    )
  ))
})


# ============================================================================
# Tests for moderator and scale priors
# ============================================================================

test_that("Moderator priors are assigned", {

  skip_on_cran()

  result <- brma.norm(
    yi = effect, sei = std_err, mods = ~ mod_cont + mod_factor,
    data = test_data, measure = "SMD", only_priors = TRUE
  )[["priors"]]

  expect_true(!is.null(result$mods))
  expect_true("intercept" %in% names(result$mods))
  expect_false("mu" %in% names(result$outcome))  # mu forwarded to intercept

  # Intercept inherits from custom effect prior
  custom_prior <- BayesTools::prior("normal", parameters = list(mean = 0, sd = 0.8))
  result_custom <- brma.norm(
    yi = effect, sei = std_err, mods = ~ mod_cont, data = test_data,
    prior_effect = custom_prior, measure = "SMD", only_priors = TRUE
  )[["priors"]]
  expect_equal(result_custom$mods$intercept$parameters$sd, 0.8)

  # Moderator priors scale with UISD
  expected_mod_sd <- sqrt(2) * RoBMA.get_option("default_UISD.mods")
  expect_equal(result$mods$mod_cont$parameters$sd, expected_mod_sd)

  # Custom prior_mods
  custom_mod <- list(mod_cont = BayesTools::prior("normal", parameters = list(mean = 0, sd = 0.3)))
  result_mod <- brma.norm(
    yi = effect, sei = std_err, mods = ~ mod_cont, data = test_data,
    prior_mods = custom_mod, measure = "SMD", only_priors = TRUE
  )[["priors"]]
  expect_equal(result_mod$mods$mod_cont$parameters$sd, 0.3)

  # Single prior_mods applies to all terms
  result_mod_default <- brma.norm(
    yi = effect, sei = std_err, mods = ~ mod_cont,
    data = test_data, prior_mods = custom_mod[[1]],
    measure = "SMD", only_priors = TRUE
  )[["priors"]]
  expect_equal(result_mod_default$mods$mod_cont$parameters$sd, 0.3)
})


test_that("No-intercept moderator formulas use fixed-zero intercept", {

  object <- brma.norm(
    yi = effect, sei = std_err, mods = ~ 0 + mod_cont,
    data = test_data, measure = "SMD", only_priors = TRUE
  )
  result <- object[["priors"]]

  expect_null(result$outcome$mu)
  expect_true(BayesTools::is.prior.point(result$mods$intercept))
  expect_equal(mean(result$mods$intercept), 0)
  expect_false(BayesTools::is.prior.point(result$mods$mod_cont))

  formula_output <- BayesTools::JAGS_formula(
    parameter  = "mu",
    data       = object[["data"]][["mods"]],
    formula    = .create_fit_formula_list(data = object[["data"]], parameter = "mods"),
    prior_list = .create_fit_formula_prior_list(priors = object[["priors"]], parameter = "mods")
  )

  expect_true(BayesTools::is.prior.point(formula_output[["prior_list"]][["mu_intercept"]]))
  expect_equal(mean(formula_output[["prior_list"]][["mu_intercept"]]), 0)
  expect_match(formula_output[["formula_syntax"]], "mu_intercept", fixed = TRUE)
})

test_that("Scale formulas repair no-intercept terms through BayesTools public API", {

  env <- new.env(parent = emptyenv())

  data <- list(
    scale = test_data
  )
  formula_minus <- ~ scale_var - 1
  environment(formula_minus) <- env
  attr(data[["scale"]], "formula") <- formula_minus

  expect_warning(
    repaired <- .create_fit_formula_list(data = data, parameter = "scale"),
    "Intercept cannot be omitted"
  )
  expect_equal(attr(stats::terms(repaired), "intercept"), 1)
  expect_equal(attr(stats::terms(repaired), "term.labels"), "scale_var")
  expect_true(identical(environment(repaired), env))
  expect_true(isTRUE(attr(repaired, "log(intercept)")))

  formula_zero <- ~ 0 + scale_var
  environment(formula_zero) <- env
  attr(data[["scale"]], "formula") <- formula_zero
  expect_warning(
    repaired_zero <- .create_fit_formula_list(data = data, parameter = "scale"),
    "Intercept cannot be omitted"
  )
  expect_equal(attr(stats::terms(repaired_zero), "intercept"), 1)
  expect_equal(attr(stats::terms(repaired_zero), "term.labels"), "scale_var")

  formula_special <- ~ I(scale_var - 1) + offset(n - 1)
  environment(formula_special) <- env
  attr(data[["scale"]], "formula") <- formula_special
  expect_silent(
    repaired_special <- .create_fit_formula_list(data = data, parameter = "scale")
  )
  expect_equal(
    attr(stats::terms(repaired_special), "term.labels"),
    "I(scale_var - 1)"
  )
  expect_equal(attr(stats::terms(repaired_special), "offset"), 2)
})


test_that("Scale priors are assigned", {

  skip_on_cran()

  result <- brma.norm(
    yi = effect, sei = std_err, scale = ~ scale_var,
    data = test_data, measure = "SMD", only_priors = TRUE
  )[["priors"]]

  expect_true(!is.null(result$scale))
  expect_true("intercept" %in% names(result$scale))
  expect_false("tau" %in% names(result$outcome))  # tau forwarded to intercept

  # Intercept inherits from custom heterogeneity prior
  custom_prior <- BayesTools::prior("normal", parameters = list(mean = 0, sd = 0.5), truncation = list(0, Inf))
  result_custom <- brma.norm(
    yi = effect, sei = std_err, scale = ~ scale_var, data = test_data,
    measure = "SMD",
    prior_heterogeneity = custom_prior, prior_unit_information_sd = 10, only_priors = TRUE
  )[["priors"]]
  expect_equal(result_custom$scale$intercept$parameters$sd, 0.5)

  # Scale priors use fixed UISD (not measure-dependent)
  result_smd  <- brma.norm(yi = effect, sei = std_err, scale = ~ scale_var, data = test_data, measure = "SMD",  only_priors = TRUE)[["priors"]]
  result_zcor <- brma.norm(yi = effect, sei = std_err, scale = ~ scale_var, data = test_data, measure = "ZCOR", only_priors = TRUE)[["priors"]]

  if (!is.null(result_smd$scale$scale_var) && !is.null(result_zcor$scale$scale_var)) {
    expect_equal(result_smd$scale$scale_var$parameters$sd, result_zcor$scale$scale_var$parameters$sd)
  }
})


test_that("Both mods and scale priors are assigned together", {

  skip_on_cran()

  result <- brma.norm(
    yi = effect, sei = std_err, mods = ~ mod_cont + mod_factor, scale = ~ scale_var,
    data = test_data, measure = "SMD", only_priors = TRUE
  )[["priors"]]

  expect_true(!is.null(result$mods))
  expect_true(!is.null(result$scale))
  expect_true("intercept" %in% names(result$mods))
  expect_true("intercept" %in% names(result$scale))
  expect_false("mu" %in% names(result$outcome))
  expect_false("tau" %in% names(result$outcome))
})


# ============================================================================
# Tests for contrast
# ============================================================================

test_that("set_contrast_factor_predictors options are applied", {

  skip_on_cran()

  for (contrast in c("treatment", "meandif", "orthonormal")) {
    result <- brma.norm(
      yi = effect, sei = std_err, mods = ~ mod_factor, data = test_data,
      measure = "SMD", set_contrast_factor_predictors = contrast, only_priors = TRUE
    )[["priors"]]
    expect_true(any(grepl(contrast, class(result$mods[["mod_factor"]]))))
  }

  expect_error(
    brma.norm(
      yi = effect, sei = std_err, mods = ~ mod_factor, data = test_data,
      measure = "SMD",
      set_contrast_factor_predictors = "invalid", only_priors = TRUE
    ),
    regexp = "set_contrast_factor_predictors"
  )
})

context("Summary")

source(testthat::test_path("common-functions.R"))

skip_if_no_fits()
fit_names <- list_fits()
fits      <- lazy_fits(fit_names, validate = FALSE)

summary_sections <- c(
  "name",
  "inclusion_components",
  "inclusion_mods",
  "inclusion_scale",
  "estimates",
  "estimates_conditional",
  "estimates_mods",
  "estimates_mods_conditional",
  "estimates_scale",
  "estimates_scale_conditional",
  "estimates_bias",
  "estimates_bias_conditional"
)

summary_table_sections <- setdiff(summary_sections, "name")

summary_common_parameters <- function(fit) {

  prior_names <- names(attr(fit[["fit"]], "prior_list"))
  return(intersect(c("mu", "tau", "rho"), prior_names))
}

expect_summary_subtable <- function(table, name, section) {

  if (length(table) == 0L) {
    return(invisible(NULL))
  }

  info <- paste0("summary section '", section, "' for '", name, "'")
  expect_true(is.matrix(table) || is.data.frame(table), info = info)
  expect_true(nrow(table) > 0L, info = info)
  expect_true(ncol(table) > 0L, info = info)
  expect_false(is.null(rownames(table)), info = info)
  expect_true(all(nzchar(rownames(table))), info = info)

  values <- unlist(
    as.data.frame(table)[vapply(as.data.frame(table), is.numeric, TRUE)],
    use.names = FALSE
  )
  if (length(values) > 0L) {
    expect_true(all(is.na(values) | !is.nan(values)), info = info)
  }

  title <- attr(table, "title")
  expect_true(
    is.character(title) && length(title) == 1L && nzchar(title),
    info = paste0(info, " title")
  )
}

expect_summary_sections <- function(summary_object, fit, name, conditional = FALSE) {

  common_parameters <- summary_common_parameters(fit)

  expect_equal(
    length(summary_object[["estimates"]]) > 0L,
    length(common_parameters) > 0L,
    info = paste0("common estimate section for '", name, "'")
  )
  expect_equal(
    length(summary_object[["estimates_mods"]]) > 0L,
    .is_mods(fit),
    info = paste0("moderator estimate section for '", name, "'")
  )
  expect_equal(
    length(summary_object[["estimates_scale"]]) > 0L,
    .is_scale(fit),
    info = paste0("scale estimate section for '", name, "'")
  )
  expect_equal(
    length(summary_object[["estimates_bias"]]) > 0L,
    .is_bias(fit),
    info = paste0("bias estimate section for '", name, "'")
  )
  expect_equal(
    length(summary_object[["inclusion_components"]]) > 0L,
    .is_RoBMA(fit),
    info = paste0("component inclusion section for '", name, "'")
  )

  conditional_sections <- c(
    "estimates_conditional",
    "estimates_mods_conditional",
    "estimates_scale_conditional",
    "estimates_bias_conditional"
  )
  conditional_present <- vapply(
    summary_object[conditional_sections],
    function(x) length(x) > 0L,
    TRUE
  )

  if (conditional) {
    expect_true(
      any(conditional_present),
      info = paste0("conditional sections for '", name, "'")
    )
  } else {
    expect_false(
      any(conditional_present),
      info = paste0("conditional sections for '", name, "'")
    )
  }
}

expect_summary_contract <- function(summary_object, fit, name,
                                    conditional = FALSE) {

  expect_s3_class(summary_object, "summary.brma")
  expect_named(summary_object, summary_sections)
  expect_type(summary_object[["name"]], "character")
  expect_true(length(summary_object[["name"]]) == 1L)
  expect_true(nzchar(summary_object[["name"]]))

  expect_identical(attr(summary_object, "mods"), .is_mods(fit))
  expect_identical(attr(summary_object, "scale"), .is_scale(fit))
  expect_identical(attr(summary_object, "multilevel"), .is_multilevel(fit))
  expect_identical(attr(summary_object, "bias"), .is_bias(fit))
  expect_identical(attr(summary_object, "RoBMA"), .is_RoBMA(fit))
  expect_identical(attr(summary_object, "outcome_type"), .outcome_type(fit))

  for (section in summary_table_sections) {
    expect_summary_subtable(summary_object[[section]], name, section)
  }
  expect_summary_sections(summary_object, fit, name, conditional = conditional)
}

expect_printed_summary <- function(summary_object, name) {

  output <- capture.output(print(summary_object))
  expect_true(any(nzchar(output)), info = paste0("printed summary for '", name, "'"))
  expect_true(any(grepl("Bayesian", output, fixed = TRUE)),
              info = paste0("printed summary model name for '", name, "'"))
  expect_false(any(grepl("__xXx__", output, fixed = TRUE)),
               info = paste0("printed summary labels for '", name, "'"))
}


test_that("summary.brma returns a stable object contract", {

  for (name in names(fits)) {
    fit <- fits[[name]]
    out <- summary(fit)

    expect_summary_contract(out, fit, name)
    expect_printed_summary(out, name)
  }
})

test_that("summary.brma options change table schema", {

  name <- "bcg_meta-analysis"
  skip_if_missing_fits(name)

  out <- summary(
    fits[[name]],
    probs                    = c(0.01, 0.99),
    include_mcmc_diagnostics = FALSE
  )
  expect_summary_contract(out, fits[[name]], name)

  cols <- colnames(out[["estimates"]])
  expect_true(all(c("Mean", "SD", "0.01", "0.99") %in% cols))
  expect_false(any(grepl("error\\(MCMC\\)|ESS|R-hat", cols)))
})

test_that("summary.brma controls BayesTools diagnostic columns", {

  name <- "dat.lehmann2018_RoBMA"
  skip_if_missing_fits(name)

  old_options <- options(
    BayesTools.JAGS_estimates_diagnostic_columns = c("ESS", "R_hat"),
    BayesTools.JAGS_BF_diagnostic_columns        = c("ESS", "MCMC_error", "BF_error_percent")
  )
  on.exit(options(old_options), add = TRUE)

  out <- summary(fits[[name]], include_mcmc_diagnostics = TRUE)
  estimate_diagnostics <- c("MCMC_error", "MCMC_SD_error", "ESS", "R_hat")
  BF_diagnostics       <- c("ESS", "MCMC_error", "BF_error_percent")

  expect_true(all(estimate_diagnostics %in% colnames(out[["estimates"]])))
  expect_true("BF_error_percent" %in% colnames(out[["inclusion_components"]]))
  expect_false(any(c("ESS", "MCMC_error") %in%
                     colnames(out[["inclusion_components"]])))

  out_none <- summary(fits[[name]], include_mcmc_diagnostics = FALSE)
  expect_false(any(estimate_diagnostics %in% colnames(out_none[["estimates"]])))
  expect_false(any(BF_diagnostics %in%
                     colnames(out_none[["inclusion_components"]])))
})

test_that("summary.brma inclusion summaries support BF direction and log scale", {

  name <- "dat.lehmann2018_RoBMA"
  skip_if_missing_fits(name)

  bf_column <- function(x) {

    return(as.data.frame(x[["inclusion_components"]])[["inclusion_BF"]])
  }

  out_default <- summary(fits[[name]], include_mcmc_diagnostics = FALSE)
  out_log     <- summary(fits[[name]], include_mcmc_diagnostics = FALSE, logBF = TRUE)
  out_BF01    <- summary(fits[[name]], include_mcmc_diagnostics = FALSE, BF01 = TRUE)
  out_both    <- summary(
    fits[[name]],
    include_mcmc_diagnostics = FALSE,
    logBF                    = TRUE,
    BF01                     = TRUE
  )

  expect_false(isTRUE(attr(bf_column(out_default), "logBF")))
  expect_false(isTRUE(attr(bf_column(out_default), "BF01")))
  expect_true(isTRUE(attr(bf_column(out_log), "logBF")))
  expect_false(isTRUE(attr(bf_column(out_log), "BF01")))
  expect_false(isTRUE(attr(bf_column(out_BF01), "logBF")))
  expect_true(isTRUE(attr(bf_column(out_BF01), "BF01")))
  expect_true(isTRUE(attr(bf_column(out_both), "logBF")))
  expect_true(isTRUE(attr(bf_column(out_both), "BF01")))

  default_effect <- as.numeric(bf_column(out_default)["Effect"])
  expect_equal(
    as.numeric(bf_column(out_log)["Effect"]),
    log(default_effect),
    tolerance = sqrt(.Machine$double.eps)
  )
  expect_equal(
    as.numeric(bf_column(out_BF01)["Effect"]),
    1 / default_effect,
    tolerance = sqrt(.Machine$double.eps)
  )
  expect_equal(
    as.numeric(bf_column(out_both)["Effect"]),
    log(1 / default_effect),
    tolerance = sqrt(.Machine$double.eps)
  )
})

test_that("summary.brma inclusion subtables preserve row-level BF bounds", {

  inclusion_BF <- BayesTools::format_BF(
    c(14999, 1.257, 0.588),
    inclusion = TRUE
  )
  attr(inclusion_BF, "bound_operator") <- c(">", NA_character_, NA_character_)
  class(inclusion_BF) <- unique(c("BayesTools_BF", class(inclusion_BF)))

  table <- data.frame(
    prior_prob   = c(0.5, 0.5, 0.5),
    post_prob    = c(1.0, 0.557, 0.370),
    inclusion_BF = inclusion_BF,
    row.names    = c("(mu) intercept", "(mu) tailor", "tau"),
    check.names  = FALSE
  )
  class(table)              <- c("BayesTools_table", class(table))
  attr(table, "type")       <- c("prior_prob", "post_prob", "inclusion_BF")
  attr(table, "parameters") <- c("mu_intercept", "mu_tailor", "tau")
  attr(table, "rownames")   <- TRUE

  mod_table <- .summary.inclusion_subtable(
    table      = table,
    indices    = 2L,
    row_labels = "tailor",
    title      = "Meta-Regression Inclusion"
  )

  expect_identical(
    attr(mod_table[["inclusion_BF"]], "bound_operator"),
    NA_character_
  )
  expect_false(any(grepl(">1.257", capture.output(print(mod_table)),
                         fixed = TRUE)))
})

test_that("summary.brma standardized coefficients use the standardized scale", {

  name <- "bangertdrowns2004_location-scale"
  skip_if_missing_fits(name)

  fit             <- fits[[name]]
  out_default     <- summary(fit)
  out_standardized <- summary(fit, standardized_coefficients = TRUE)

  expect_summary_contract(out_standardized, fit, name)
  expect_true("ni100" %in% rownames(out_default[["estimates_mods"]]))
  expect_true("ni100" %in% rownames(out_standardized[["estimates_mods"]]))
  expect_gt(
    abs(out_default[["estimates_mods"]]["ni100", "Mean"] -
          out_standardized[["estimates_mods"]]["ni100", "Mean"]),
    sqrt(.Machine$double.eps)
  )

  expect_equal(
    capture.output(print(fit)),
    capture.output(print(out_default)),
    info = "print.brma delegates to summary.brma"
  )
})

test_that("RoBMA conditional summaries expose conditional sections", {

  name <- "dat.lehmann2018_RoBMA_mods"
  skip_if_missing_fits(c("bcg_meta-analysis", name))

  out <- summary(fits[[name]], conditional = TRUE)
  expect_summary_contract(out, fits[[name]], name, conditional = TRUE)

  conditional_titles <- vapply(
    c(
      "estimates_conditional",
      "estimates_mods_conditional",
      "estimates_scale_conditional",
      "estimates_bias_conditional"
    ),
    function(section) {
      title <- attr(out[[section]], "title")
      if (is.null(title)) "" else title
    },
    character(1)
  )
  expect_true(any(grepl("Conditional", conditional_titles, fixed = TRUE)))

  expect_error(
    summary(fits[["bcg_meta-analysis"]], conditional = TRUE),
    "RoBMA objects"
  )
})

test_that("RoBMA inclusion summaries use user-facing labels", {

  skip_if_missing_fits(c("dat.lehmann2018_RoBMA", "dat.lehmann2018_RoBMA_mods2"))

  out_simple <- summary(fits[["dat.lehmann2018_RoBMA"]])
  expect_true(all(c("Effect", "Heterogeneity", "Publication Bias") %in%
                    rownames(out_simple[["inclusion_components"]])))

  out_mods2 <- summary(fits[["dat.lehmann2018_RoBMA_mods2"]])
  expect_false(any(grepl("__xXx__", rownames(out_mods2[["inclusion_mods"]]),
                         fixed = TRUE)))
  expect_true("Preregistered:Gender" %in%
                rownames(out_mods2[["inclusion_mods"]]))
})

# ============================================================================ #
# brma.bridgesampling.R
# ============================================================================ #
#
# This file implements S3 methods for bridgesampling package generics
# for brma class objects. The marginal likelihood computed via bridge
# sampling is used for Bayesian model comparison via Bayes factors.
#
# The implementation follows the same pattern as brma.loo.R and brma.as_draws.R:
# generics are defined locally so they work regardless of whether the
# bridgesampling package is loaded first, last, or not at all.
#
# ============================================================================ #


# ---------------------------------------------------------------------------- #
# Re-export bridgesampling generics
# ---------------------------------------------------------------------------- #

#' @importFrom bridgesampling bridge_sampler
#' @export
bridgesampling::bridge_sampler

#' @importFrom bridgesampling logml
#' @export
bridgesampling::logml

#' @importFrom bridgesampling post_prob
#' @export
bridgesampling::post_prob

#' @importFrom bridgesampling bf
#' @export
bridgesampling::bf

#' @importFrom bridgesampling bayes_factor
#' @export
bridgesampling::bayes_factor


# ---------------------------------------------------------------------------- #
# S3 Methods for brma Objects
# ---------------------------------------------------------------------------- #

.log_sum_exp <- function(x) {

  max_x <- max(x)
  return(max_x + log(sum(exp(x - max_x))))
}

.post_prob_from_logml <- function(logml, prior_prob = NULL, model_names = NULL) {

  BayesTools::check_real(logml, "logml", check_length = 0, allow_NA = FALSE)
  if (any(!is.finite(logml))) {
    stop("'logml' must contain only finite values.", call. = FALSE)
  }

  if (length(logml) < 2L) {
    stop("At least two log marginal likelihoods are required.", call. = FALSE)
  }

  if (is.null(prior_prob)) {
    prior_prob <- rep(1 / length(logml), length(logml))
  } else {
    BayesTools::check_real(prior_prob, "prior_prob", lower = 0, check_length = 0, allow_NA = FALSE)
    if (any(!is.finite(prior_prob))) {
      stop("'prior_prob' must contain only finite values.", call. = FALSE)
    }
    if (length(prior_prob) != length(logml)) {
      stop("'prior_prob' must have the same length as the number of models.", call. = FALSE)
    }
    if (sum(prior_prob) <= 0) {
      stop("'prior_prob' must contain at least one positive value.", call. = FALSE)
    }
    prior_prob <- prior_prob / sum(prior_prob)
  }

  if (!is.null(model_names)) {
    BayesTools::check_char(model_names, "model_names", check_length = 0, allow_NA = FALSE)
    if (length(model_names) != length(logml)) {
      stop("'model_names' must have the same length as the number of models.", call. = FALSE)
    }
  }

  log_posterior <- logml + log(prior_prob)
  posterior     <- exp(log_posterior - .log_sum_exp(log_posterior))

  if (!is.null(model_names)) {
    names(posterior) <- model_names
  }

  return(posterior)
}

.bf_from_logml <- function(logml1, logml2, log = FALSE) {

  BayesTools::check_real(logml1, "logml1", check_length = 1, allow_NA = FALSE)
  BayesTools::check_real(logml2, "logml2", check_length = 1, allow_NA = FALSE)
  if (!is.finite(logml1) || !is.finite(logml2)) {
    stop("'logml1' and 'logml2' must be finite.", call. = FALSE)
  }
  BayesTools::check_bool(log, "log")

  log_bf <- logml1 - logml2
  out    <- list(
    bf  = if (log) log_bf else exp(log_bf),
    log = log
  )

  class(out) <- "bf_default"
  return(out)
}

#' @title Bridge Sampling for brma Objects
#'
#' @description Extract the marginal likelihood bridge sampling object from
#' a brma model. The marginal likelihood must first be computed using
#' \code{\link{add_marglik}}.
#'
#' @param samples a brma model object.
#' @param ... additional arguments (currently not used).
#'
#' @details
#' This function extracts the bridge sampling object that was previously
#' computed and stored using \code{\link{add_marglik}}. If the marginal
#' likelihood has not been computed, an error is thrown.
#' Product-space model-averaging objects (\code{BMA.norm}, \code{BMA.glmm},
#' and \code{RoBMA}) do not expose bridge-sampling marginal likelihoods.
#'
#' The returned object can be used for Bayesian model comparison via
#' \code{\link{bf}} and \code{\link{post_prob}}.
#'
#' @return An object of class \code{"bridge"} as returned by
#' \code{\link[bridgesampling]{bridge_sampler}}.
#'
#' @seealso \code{\link{add_marglik}}, \code{\link[bridgesampling]{bridge_sampler}},
#' \code{\link{logml.brma}}, \code{\link{bf.brma}}, \code{\link{post_prob.brma}}
#'
#' @examples
#' \dontrun{
#' if (requireNamespace("metadat", quietly = TRUE)) {
#'   data(dat.lehmann2018, package = "metadat")
#'   fit <- brma(yi = yi, vi = vi, data = dat.lehmann2018, measure = "SMD")
#'
#'   fit <- add_marglik(fit)
#'
#'   bridge <- bridge_sampler(fit)
#'   print(bridge)
#' }
#' }
#'
#' @export
#' @exportS3Method bridgesampling::bridge_sampler
bridge_sampler.brma <- function(samples, ...) {
  if (inherits(samples, "RoBMA")) {
    stop(
      "Marginal likelihood is not available for product-space ",
      "model-averaging objects (BMA.norm, BMA.glmm, RoBMA).",
      call. = FALSE
    )
  }

  if (is.null(samples[["marglik"]])) {
    stop("Marginal likelihood has not been computed. Call 'object <- add_marglik(object)' first.",
      call. = FALSE
    )
  }

  return(samples[["marglik"]])
}


#' @title Log Marginal Likelihood for brma Objects
#'
#' @description Extract the log marginal likelihood from a brma model.
#' The marginal likelihood must first be computed using \code{\link{add_marglik}}.
#'
#' @param x a brma model object.
#' @param ... additional arguments (currently not used).
#'
#' @details
#' This function extracts the log marginal likelihood from the bridge sampling
#' object that was previously computed and stored using \code{\link{add_marglik}}.
#' Product-space model-averaging objects (\code{BMA.norm}, \code{BMA.glmm},
#' and \code{RoBMA}) do not expose bridge-sampling marginal likelihoods.
#'
#' @return A scalar numeric value representing the log marginal likelihood.
#'
#' @seealso \code{\link{add_marglik}}, \code{\link{bridge_sampler.brma}},
#' \code{\link{bf.brma}}, \code{\link{post_prob.brma}}
#'
#' @examples
#' \dontrun{
#' if (requireNamespace("metadat", quietly = TRUE)) {
#'   data(dat.lehmann2018, package = "metadat")
#'   fit <- brma(yi = yi, vi = vi, data = dat.lehmann2018, measure = "SMD")
#'
#'   fit <- add_marglik(fit)
#'
#'   logml(fit)
#' }
#' }
#'
#' @export
#' @exportS3Method bridgesampling::logml
logml.brma <- function(x, ...) {
  bridge <- bridge_sampler.brma(x)
  return(bridge$logml)
}


#' @title Posterior Model Probabilities for brma Objects
#'
#' @description Compute posterior model probabilities from marginal
#' likelihoods of brma models.
#'
#' @param x a brma model object.
#' @param ... additional brma model objects.
#' @param prior_prob numeric vector with prior model probabilities or weights.
#' Values must be finite, nonnegative, have the same length as the retained
#' models, and have a positive total. If omitted, a uniform prior is used.
#' Supplied values are normalized internally.
#' @param model_names character vector with model names. If \code{NULL}
#' (the default), names will be derived from deparsing the call.
#'
#' @details
#' The marginal likelihoods must first be computed using \code{\link{add_marglik}}.
#' \code{x} and at least one additional \code{brma} model must be supplied.
#' Non-\code{brma} objects in \code{...} are ignored with a warning. All retained
#' models must be fitted to the same outcome target/data, including outcome type
#' and, when present, weights and cluster identifiers.
#'
#' @return A named numeric vector with posterior model probabilities
#' (i.e., which sum to one).
#'
#' @seealso \code{\link{add_marglik}}, \code{\link{bridge_sampler.brma}},
#' \code{\link{bf.brma}}, \code{\link{logml.brma}}
#'
#' @examples
#' \dontrun{
#' if (requireNamespace("metadat", quietly = TRUE)) {
#'   data(dat.lehmann2018, package = "metadat")
#'   fit1 <- brma(yi = yi, vi = vi, data = dat.lehmann2018, measure = "SMD")
#'   fit2 <- brma(
#'     yi           = yi,
#'     vi           = vi,
#'     data         = dat.lehmann2018,
#'     measure      = "SMD",
#'     prior_effect = FALSE
#'   )
#'
#'   fit1 <- add_marglik(fit1)
#'   fit2 <- add_marglik(fit2)
#'
#'   post_prob(fit1, fit2)
#' }
#' }
#'
#' @export
#' @exportS3Method bridgesampling::post_prob
post_prob.brma <- function(x, ..., prior_prob = NULL, model_names = NULL) {
  dots <- list(...)
  mc <- match.call()

  # check that all objects in ... are brma objects
  modb <- vapply(dots, inherits, NA, what = "brma")
  if (sum(modb) == 0) {
    stop("Only one object of class 'brma' passed.", call. = FALSE)
  }
  if (sum(modb) != length(dots)) {
    warning("Objects not of class 'brma' are ignored.", call. = FALSE)
  }
  models <- c(list(x), dots[modb])
  .check_brma_compare_targets(models, "post_prob()")

  # get model names
  if (is.null(model_names)) {
    model_names <- c(
      deparse(mc[["x"]]),
      vapply(which(modb), function(i) deparse(mc[[i + 2]]), "")
    )
  }

  # compute log marginal likelihoods (this will error if add_marglik not called)
  logml_values <- c(
    bridge_sampler.brma(x)$logml,
    vapply(models[-1], function(obj) bridge_sampler.brma(obj)$logml, 0)
  )

  return(.post_prob_from_logml(
    logml       = logml_values,
    prior_prob = prior_prob,
    model_names = model_names
  ))
}


#' @title Bayes Factor for brma Objects
#'
#' @description Compute the Bayes factor comparing two brma models.
#'
#' @param x1 a brma model object (numerator).
#' @param x2 a brma model object (denominator).
#' @param log logical; if \code{TRUE}, the log Bayes factor is returned.
#' Default is \code{FALSE}.
#' @param ... additional arguments (currently not used).
#'
#' @details
#' Computes the Bayes factor in favor of the model \code{x1} over the
#' model \code{x2}. The marginal likelihoods must first be computed using
#' \code{\link{add_marglik}}. Both models must be fitted to the same
#' outcome target/data, including outcome type and, when present, weights and
#' cluster identifiers.
#'
#' @return A list of class \code{"bf_default"} with components:
#' \itemize{
#'   \item \code{bf}: (scalar) value of the Bayes factor in favor of
#'   \code{x1} over \code{x2}.
#'   \item \code{log}: Boolean indicating whether \code{bf} corresponds
#'   to the log Bayes factor.
#' }
#'
#' @seealso \code{\link{add_marglik}}, \code{\link{bridge_sampler.brma}},
#' \code{\link{post_prob.brma}}, \code{\link{logml.brma}}
#'
#' @examples
#' \dontrun{
#' if (requireNamespace("metadat", quietly = TRUE)) {
#'   data(dat.lehmann2018, package = "metadat")
#'   fit1 <- brma(yi = yi, vi = vi, data = dat.lehmann2018, measure = "SMD")
#'   fit2 <- brma(
#'     yi           = yi,
#'     vi           = vi,
#'     data         = dat.lehmann2018,
#'     measure      = "SMD",
#'     prior_effect = FALSE
#'   )
#'
#'   fit1 <- add_marglik(fit1)
#'   fit2 <- add_marglik(fit2)
#'
#'   bf(fit1, fit2)
#' }
#' }
#'
#' @export
#' @exportS3Method bridgesampling::bf
bf.brma <- function(x1, x2, log = FALSE, ...) {
  if (!inherits(x2, "brma")) {
    stop("x2 needs to be of class 'brma'.", call. = FALSE)
  }
  .check_brma_compare_targets(list(x1, x2), "bf()")

  # compute log marginal likelihoods
  logml1 <- bridge_sampler.brma(x1)$logml
  logml2 <- bridge_sampler.brma(x2)$logml

  return(.bf_from_logml(
    logml1 = logml1,
    logml2 = logml2,
    log    = log
  ))
}


#' @rdname bf.brma
#' @export
#' @exportS3Method bridgesampling::bayes_factor
bayes_factor.brma <- function(x1, x2, log = FALSE, ...) {
  bf.brma(x1, x2, log = log, ...)
}

#' @title Input Data Specification
#' @name data_input
#' @description
#' Shared data-input arguments used by the RoBMA fitting functions.
#'
#' Normal models use approximate effect-size estimates supplied through
#' `yi` with either `vi` or `sei`. GLMM models use the raw two-arm count
#' arguments for binomial (`measure = "OR"`) or Poisson (`measure = "IRR"`)
#' outcomes.
#'
#' @param yi a vector of effect sizes, or a formula with the effect size on the
#' left-hand side and location moderators on the right-hand side (for example
#' `yi ~ x1 + x2`). If a formula is supplied, `mods` must not be specified.
#' @param vi a vector of sampling variances. Either `vi` or `sei` must be
#' supplied for normal models.
#' @param sei a vector of standard errors. Either `vi` or `sei` must be
#' supplied for normal models.
#' @param weights an optional vector of positive likelihood weights. For
#' normal/effect-size models, each weight powers the estimate likelihood. For
#' constructors with GLMM raw-count input, each weight powers the paired
#' two-arm likelihood for one study.
#' @param ni an optional vector of sample sizes. Used for `measure = "GEN"`
#' or when estimating `"UISD"`).
#' @param mods an optional matrix, data.frame, or formula specifying
#' location moderators (meta-regressors). Formula input is evaluated in `data`.
#' @param scale an optional matrix, data.frame, or formula specifying
#' scale predictors for location-scale models. Formula input is evaluated in
#' `data`.
#' @param cluster an optional vector of cluster identifiers for multilevel
#' meta-analysis.
#' @param data an optional data frame containing the variables.
#' @param slab an optional vector of study labels.
#' @param subset an optional logical or numeric vector specifying a subset of
#' data to be used.
#' @param measure a character string specifying the effect size measure.
#' Normal/effect-size constructors require an explicit value and support
#' `"SMD"`, `"ZCOR"`, `"RR"`, `"OR"`, `"HR"`, `"RD"`, `"IRR"`, and `"GEN"`.
#' Use `"GEN"` only for general effect sizes without a known unit information
#' standard deviation. GLMM raw-count constructors support only `"OR"` and
#' `"IRR"` and default to `"OR"`.
#' @param effect_direction direction used by publication-bias adjustments.
#' `"positive"` assumes statistically significant positive estimates are more
#' likely to be selected; `"negative"` mirrors the selection direction;
#' `"detect"` infers the direction from the fitted data.
#' @param ai a vector of the number of events in the treatment or experimental
#' group for binomial GLMM models.
#' @param bi a vector of the number of non-events in the treatment or
#' experimental group for binomial GLMM models.
#' @param ci a vector of the number of events in the control group for binomial
#' GLMM models.
#' @param di a vector of the number of non-events in the control group for
#' binomial GLMM models.
#' @param n1i a vector of the sample size in the treatment or experimental
#' group. If omitted for binomial GLMMs, it is computed as `ai + bi`.
#' @param n2i a vector of the sample size in the control group. If omitted for
#' binomial GLMMs, it is computed as `ci + di`.
#' @param x1i a vector of the number of events in the treatment/experimental group
#' (for Poisson data).
#' @param x2i a vector of the number of events in the control group (for Poisson data).
#' @param t1i a vector of the person-time in the treatment/experimental group.
#' @param t2i a vector of the person-time in the control group.
#'
NULL



# Internal helper function to extract a variable from a data frame or environment
# Supports three input formats:
#   1. Direct vector: yi = c(0.1, 0.2, 0.3)
#   2. Variable name (unquoted): yi = effect_size (where effect_size is a column in data)
#   3. String name (quoted): yi = "effect_size" (where "effect_size" is a column name in data)
#
# @param mf The matched call from the parent function
# @param data The data frame to search in (can be NULL)
# @param enclos The enclosing environment for evaluation
# @param name The name of the argument (for error messages)
# @param allow_NULL Logical; if TRUE, NULL values are allowed
# @return The extracted vector or NULL
.get_variable <- function(mf, data, enclos, name, allow_NULL = TRUE) {

  # Check if argument is in the call
  arg_index <- match(name, names(mf))

  # If argument was not specified in the call, return NULL
  if (is.na(arg_index)) {
    return(NULL)
  }

  # Get the unevaluated expression from the matched call
  mf_x <- mf[[arg_index]]

  # If the argument value itself is NULL, return NULL
  if (is.null(mf_x)) {
    return(NULL)
  }

  # First, try to evaluate the expression directly in the data frame (if provided)
  if (!is.null(data) && is.data.frame(data)) {
    out <- try(eval(mf_x, data, enclos), silent = TRUE)
  } else {
    out <- try(eval(mf_x, enclos), silent = TRUE)
  }

  # If evaluation succeeded and result is a valid vector, return it
  if (!inherits(out, "try-error") && !is.function(out)) {

    # If result is a single string matching a column name in data, extract that column
    if (is.character(out) && length(out) == 1 && !is.null(data) && is.data.frame(data) && out %in% names(data)) {
      out <- data[[out]]
    }

    # Strip attributes (e.g., from metafor::escalc output) from atomic vectors only
    # Do not strip from formulas, data.frames, lists, or other complex objects
    if (is.atomic(out) && !inherits(out, "formula")) {
      out <- as.vector(out)
    }

    return(out)
  }

  # If direct evaluation failed, check if it's a string referring to a column
  if (is.character(mf_x) && length(mf_x) == 1 && !is.null(data) && is.data.frame(data) && mf_x %in% names(data)) {
    return(as.vector(data[[mf_x]]))
  }

  # If still not found, report error
  if (!allow_NULL) {
    stop(paste0("Cannot find the object/variable ('", deparse(mf_x),
                "') specified for the '", name, "' argument."), call. = FALSE)
  }

  return(NULL)
}


# Internal function to extract common optional variables shared by all outcome handlers
# Handles weights, cluster, and slab extraction and validation
#
# @param .call Matched call from the calling function
# @param data The data frame (can be NULL)
# @param .envir The enclosing environment
# @param k The expected number of observations
# @param primary_var The name of the primary variable (for error messages, e.g., "yi" or "ai")
#
# @return A list with:
#   - weights: numeric vector or NULL
#   - cluster: vector or NULL
#   - slab: character vector or NULL
#   - slab_provided: logical
#   - cluster_provided: logical
.check_and_list_data.optional_vars <- function(.call, data, .envir, k, primary_var) {

  # Extract optional variables
  weights   <- .get_variable(.call, data, .envir, "weights",   allow_NULL = TRUE)
  cluster   <- .get_variable(.call, data, .envir, "cluster",   allow_NULL = TRUE)
  slab      <- .get_variable(.call, data, .envir, "slab",      allow_NULL = TRUE)

  # Track which optional fields were provided
  weights_provided   <- !is.null(weights)
  slab_provided      <- !is.null(slab)
  cluster_provided   <- !is.null(cluster)

  # Validate weights
  if (!is.null(weights) && anyNA(weights))
    stop("The 'weights' argument must not contain missing values.", call. = FALSE)
  if (!is.null(weights))
    BayesTools::check_real(weights, "weights", check_length = k, allow_NULL = TRUE, allow_NA = FALSE, lower = 0, allow_bound = FALSE)

  # Validate cluster
  if (!is.null(cluster) && length(cluster) != k)
    stop(paste0("The 'cluster' argument must have length ", k, " (same as '", primary_var, "')."), call. = FALSE)
  if (!is.null(cluster) && anyNA(cluster))
    stop("The 'cluster' argument must not contain missing values.", call. = FALSE)

  # Validate slab (study labels)
  if (!is.null(slab) && length(slab) != k)
    stop(paste0("The 'slab' argument must have length ", k, " (same as '", primary_var, "')."), call. = FALSE)

  return(list(
    weights            = weights,
    cluster            = cluster,
    slab               = slab,
    weights_provided   = weights_provided,
    slab_provided      = slab_provided,
    cluster_provided   = cluster_provided
  ))
}


# Internal function to check and list input data for RoBMA
# This is the main entry point for processing input data. It dispatches to
# the appropriate outcome handler based on the class argument.
#
# @param .call Matched call from the calling function (required for NSE)
# @param .envir The environment where the calling function was invoked (required for NSE)
# @param class The model class: "norm" for normal likelihood (yi, sei) or
#              "glmm" for generalized linear mixed models (ai, bi, ci, di, etc.)
# @param measure The effect size measure (used for "glmm" class to determine
#                outcome type: "OR" for binomial, "IRR" for Poisson)
# @param standardize_continuous_predictors Whether to standardize continuous predictors
# @param skip_validation Whether to skip predictor validation checks (variability/levels).
#        Should be TRUE when processing newdata for prediction. (validation done in BayesTools::JAGS_formula())
#
# @return A list with components:
#   - outcome: data.frame with outcome variables (columns depend on class)
#   - mods: moderator information (data.frame or NULL)
#   - scale: scale information (data.frame or NULL)
.check_and_list_data <- function(.call, .envir, class = "norm", measure,
                                 set_contrast_factor_predictors, standardize_continuous_predictors,
                                 effect_direction = "positive", skip_validation = FALSE) {

  # check additional input
  .check_measure(measure, class = class)
  BayesTools::check_bool(standardize_continuous_predictors, "standardize_continuous_predictors", allow_NA = FALSE)
  BayesTools::check_char(set_contrast_factor_predictors, "set_contrast_factor_predictors", allow_values = c("treatment", "meandif", "orthonormal"), allow_NA = FALSE)
  if (missing(effect_direction)) {
    effect_direction <- "positive"
  }
  BayesTools::check_char(effect_direction, "effect_direction", allow_values = c("positive", "negative", "detect"))
  BayesTools::check_bool(skip_validation, "skip_validation")

  ### Extract the data argument first - other variables may reference columns within it
  data <- .get_variable(.call, NULL, .envir, "data", allow_NULL = TRUE)

  ### Step 1: Extract and validate outcome variables (dispatch based on class)
  outcome_result <- switch(
    class,
    "norm" = .check_and_list_data.outcome.norm(.call, data, .envir, effect_direction, skip_validation),
    "glmm" = switch(
      measure,
      "OR"  = .check_and_list_data.outcome.bin(.call, data, .envir, skip_validation),
      "IRR" = .check_and_list_data.outcome.pois(.call, data, .envir, skip_validation)
    )
  )

  data_outcome       <- outcome_result$data_outcome
  k                  <- outcome_result$k
  mods_from_yi       <- outcome_result$mods_from_yi
  formula_yi         <- outcome_result$formula_yi
  weights_provided   <- outcome_result$weights_provided
  slab_provided      <- outcome_result$slab_provided
  cluster_provided   <- outcome_result$cluster_provided
  na_check_cols      <- outcome_result$na_check_cols
  outcome_type       <- outcome_result$outcome_type
  effect_direction   <- outcome_result$effect_direction

  ### Step 2: Extract moderator variables (mods and scale)
  if (!is.null(mods_from_yi)) {
    # Mods already extracted from yi formula
    data_mods <- as.data.frame(mods_from_yi)
    attr(data_mods, "formula")    <- formula_yi[-2]  # RHS only
    attr(data_mods, "formula_yi") <- formula_yi
  } else {
    data_mods <- .check_and_list_data.predictors(
      .call  = .call,
      data   = data,
      .envir = .envir,
      name   = "mods",
      k      = k
    )
  }

  data_scale <- .check_and_list_data.predictors(
    .call  = .call,
    data   = data,
    .envir = .envir,
    name   = "scale",
    k      = k
  )

    data_mods  <- .check_and_list_data.coerce_character_predictors(data_mods)
    data_scale <- .check_and_list_data.coerce_character_predictors(data_scale)

  ### Step 3: Apply subset (before NA handling)
  subset <- .get_variable(.call, data, .envir, "subset", allow_NULL = TRUE)

  if (!is.null(subset)) {
    # Validate and convert subset to logical
    subset <- .check_and_list_data.validate_subset(subset, k)

    # Apply subsetting to all data frames
    data_outcome <- .check_and_list_data.subset(data_outcome, subset)
    data_mods    <- .check_and_list_data.subset(data_mods, subset)
    data_scale   <- .check_and_list_data.subset(data_scale, subset)
  }

  ### Step 4: Handle NA dropping
  # Build list of data frames to check for NAs
  # Use only the columns specified by the outcome handler for NA checking
  data_list_for_na <- list(
    outcome_required = data_outcome[, na_check_cols, drop = FALSE]
  )
  if (!is.null(data_mods)) {
    data_list_for_na$mods <- data_mods
  }
  if (!is.null(data_scale)) {
    data_list_for_na$scale <- data_scale
  }

  # Get rows with NAs
  na_rows <- .check_and_list_data.check_na(data_list_for_na)

  # Drop NA rows if any
  n_dropped <- sum(na_rows)
  if (n_dropped > 0) {

    warning(paste0(n_dropped, " observation(s) removed due to missing values."), call. = FALSE, immediate. = TRUE)

    keep_rows <- !na_rows
    data_outcome <- .check_and_list_data.subset(data_outcome, keep_rows)
    data_mods    <- .check_and_list_data.subset(data_mods, keep_rows)
    data_scale   <- .check_and_list_data.subset(data_scale, keep_rows)
  }

  ### Step 5: Final validation and processing
  k_final <- nrow(data_outcome)

  if (k_final == 0) {
    stop("No observations remaining after removing missing values.", call. = FALSE)
  }

  if (outcome_type == "norm" && effect_direction == "detect") {
    effect_direction <- if (stats::median(data_outcome[["yi"]], na.rm = TRUE) >= 0) "positive" else "negative"
  }

  # Validate predictor variables (after subsetting and NA dropping)
  # Skip validation when processing newdata (skip_validation = TRUE)
  .check_and_list_data.validate_predictors(data_mods, "mods", skip_validation)
  .check_and_list_data.validate_predictors(data_scale, "scale", skip_validation)

  # Generate default study labels if not provided (after NA dropping)
  if (!slab_provided) {
    data_outcome$slab <- paste0("Study ", seq_len(k_final))
  }

  # Convert cluster to numeric factor for use as indices in fitting (after NA dropping)
  if (cluster_provided) {
    if (anyNA(data_outcome[["cluster"]])) {
      stop("The 'cluster' argument must not contain missing values.", call. = FALSE)
    }

    cluster_factor <- factor(data_outcome[["cluster"]], levels = unique(data_outcome[["cluster"]]))
    data_outcome[["cluster_label"]] <- as.character(cluster_factor)
    data_outcome[["cluster"]]       <- as.integer(cluster_factor)
  }

  ### Create output object
  data_list <- list(
    outcome = data_outcome,
    mods    = data_mods,
    scale   = data_scale
  )

  class(data_list) <- "RoBMA_data"
  attr(data_list, "outcome_type")                       <- outcome_type
  attr(data_list, "measure")                            <- measure
  attr(data_list, "n_dropped")                          <- n_dropped
  attr(data_list, "k_final")                            <- k_final
  attr(data_list, "mods")                               <- !is.null(data_mods)
  attr(data_list, "scale")                              <- !is.null(data_scale)
  attr(data_list, "weights")                            <- weights_provided
  attr(data_list, "slab")                               <- slab_provided
  attr(data_list, "cluster")                            <- cluster_provided
  attr(data_list, "standardize_continuous_predictors")  <- standardize_continuous_predictors
  attr(data_list, "set_contrast_factor_predictors")     <- set_contrast_factor_predictors
  attr(data_list, "effect_direction")                   <- effect_direction
  return(data_list)
}


# Internal function to extract and validate outcome variables for normal likelihood models
# Handles yi, vi/sei, ni, weights, cluster, slab, and yi as formula
#
# @param .call Matched call from the calling function
# @param data The data frame (can be NULL)
# @param .envir The enclosing environment
# @param effect_direction The effect direction ("positive", "negative", or "detect")
# @param skip_validation Whether to skip strict validation checks (for newdata).
#        When TRUE, allows sei/vi/ni = 0 (useful for prediction at zero SE).
#
# @return A list with:
#   - data_outcome: data.frame with yi, sei, ni, cluster, slab, weights
#   - k: number of observations
#   - mods_from_yi: moderators extracted from yi formula (or NULL)
#   - formula_yi: the original yi formula (or NULL)
#   - slab_provided: logical, whether slab was provided by user
#   - cluster_provided: logical, whether cluster was provided by user
#   - na_check_cols: character vector of column names to check for NAs
.check_and_list_data.outcome.norm <- function(.call, data, .envir, effect_direction, skip_validation = FALSE) {

  # Extract yi (may be a formula like yi ~ mod1 + mod2)
  yi <- .get_variable(.call, data, .envir, "yi", allow_NULL = FALSE)

  # Handle yi as formula (e.g., yi ~ mod1 + mod2)
  formula_yi   <- NULL
  mods_from_yi <- NULL

  if (inherits(yi, "formula")) {

    formula_yi <- yi

    # Check that mods is not also specified (would be ambiguous)
    mods_check <- .get_variable(.call, data, .envir, "mods", allow_NULL = TRUE)
    if (!is.null(mods_check)) {
      stop("Cannot specify 'mods' when 'yi' is a formula. Use either 'yi ~ mod1 + mod2' or 'yi = effect, mods = ~ mod1 + mod2', but not both.", call. = FALSE)
    }

    # Extract the model frame from the formula
    na_act <- getOption("na.action")
    options(na.action = "na.pass")
    on.exit(options(na.action = na_act), add = TRUE)

    full_mf <- stats::model.frame(yi, data = data)

    # Extract response (LHS)
    yi <- stats::model.response(full_mf)
    names(yi) <- NULL

    # Extract predictors (RHS)
    if (ncol(full_mf) > 1) {
      mods_from_yi <- full_mf[, -1, drop = FALSE]
    }
  }

  # Extract variance/standard error
  vi  <- .get_variable(.call, data, .envir, "vi",  allow_NULL = TRUE)
  sei <- .get_variable(.call, data, .envir, "sei", allow_NULL = TRUE)

  # Extract ni (sample sizes) - specific to normal likelihood models
  ni <- .get_variable(.call, data, .envir, "ni", allow_NULL = TRUE)

  ### Input validation

  # Validate yi
  BayesTools::check_real(yi, "yi", check_length = 0, allow_NULL = FALSE, allow_NA = TRUE)
  if (all(is.na(yi)))
    stop("The 'yi' argument must contain at least one non-NA value.", call. = FALSE)

  k <- length(yi)

  # Validate vi (sampling variance)
  # When skip_validation = TRUE (newdata), allow vi = 0 for prediction at zero variance
  if (!is.null(vi))
    BayesTools::check_real(vi, "vi", check_length = k, allow_NULL = TRUE, allow_NA = TRUE, lower = 0, allow_bound = skip_validation)

  # Validate sei (standard error)
  # When skip_validation = TRUE (newdata), allow sei = 0 for prediction at zero SE
  if (!is.null(sei))
    BayesTools::check_real(sei, "sei", check_length = k, allow_NULL = TRUE, allow_NA = TRUE, lower = 0, allow_bound = skip_validation)

  # Convert vi to sei or vice versa
  if (is.null(sei) && !is.null(vi)) {
    sei <- sqrt(vi)
  } else if (is.null(vi) && !is.null(sei)) {
    vi <- sei^2
  } else if (is.null(vi) && is.null(sei)) {
    stop("Either 'vi' (variance) or 'sei' (standard error) must be provided.", call. = FALSE)
  } else {
    # Both provided - check consistency
    if (any(abs(vi - sei^2) > 1e-10, na.rm = TRUE))
      stop("The provided 'vi' and 'sei' values are inconsistent.", call. = FALSE)
  }

  # Validate ni (sample sizes)
  # When skip_validation = TRUE (newdata), allow ni = 0
  if (!is.null(ni))
    BayesTools::check_real(ni, "ni", check_length = k, allow_NULL = TRUE, allow_NA = TRUE, lower = 0, allow_bound = skip_validation)

  # Extract and validate common optional variables (weights, cluster, slab)
  optional <- .check_and_list_data.optional_vars(.call, data, .envir, k, "yi")

  ### Construct output data frame
  data_outcome <- data.frame(
    yi            = yi,
    sei           = sei,
    ni            = if (!is.null(ni))                  ni               else rep(NA_integer_, k),
    cluster       = if (!is.null(optional$cluster))    optional$cluster else rep(NA_character_, k),
    cluster_label = if (!is.null(optional$cluster))    as.character(optional$cluster) else rep(NA_character_, k),
    slab          = if (!is.null(optional$slab))       optional$slab    else rep(NA_character_, k),
    weights       = if (!is.null(optional$weights))    optional$weights else rep(NA, k),
    stringsAsFactors = FALSE
  )

  return(list(
    data_outcome       = data_outcome,
    k                  = k,
    mods_from_yi       = mods_from_yi,
    formula_yi         = formula_yi,
    weights_provided   = optional$weights_provided,
    slab_provided      = optional$slab_provided,
    cluster_provided   = optional$cluster_provided,
    na_check_cols      = c("yi", "sei"),  # Only check these columns for NAs
    effect_direction   = effect_direction,
    outcome_type       = "norm"
  ))
}


# Internal function to extract and validate outcome variables for binomial GLMM models
# Handles ai, bi, ci, di, n1i, n2i, weights, cluster, slab (for measure = "OR")
#
# @param .call Matched call from the calling function
# @param data The data frame (can be NULL)
# @param .envir The enclosing environment
#
# @return A list with:
#   - data_outcome: data.frame with ai, ci, n1i, n2i, cluster, slab, weights
#   - k: number of observations
#   - mods_from_yi: NULL (GLMM does not support formula syntax for outcome)
#   - formula_yi: NULL
#   - slab_provided: logical, whether slab was provided by user
#   - cluster_provided: logical, whether cluster was provided by user
#   - na_check_cols: character vector of column names to check for NAs
# @param skip_validation Whether to skip strict validation checks (for newdata).
.check_and_list_data.outcome.bin <- function(.call, data, .envir, skip_validation = FALSE) {

  # Extract cell counts for 2x2 tables
  ai  <- .get_variable(.call, data, .envir, "ai",  allow_NULL = TRUE)
  bi  <- .get_variable(.call, data, .envir, "bi",  allow_NULL = TRUE)
  ci  <- .get_variable(.call, data, .envir, "ci",  allow_NULL = TRUE)
  di  <- .get_variable(.call, data, .envir, "di",  allow_NULL = TRUE)
  n1i <- .get_variable(.call, data, .envir, "n1i", allow_NULL = TRUE)
  n2i <- .get_variable(.call, data, .envir, "n2i", allow_NULL = TRUE)

  ### Determine input format and validate

  # Determine k from first available variable
  if (!is.null(ai)) {
    k <- length(ai)
  } else {
    stop("For GLMM models, provide either (ai, bi, ci, di) or (ai, ci, n1i, n2i).", call. = FALSE)
  }

  count_args <- list(ai = ai, bi = bi, ci = ci, di = di, n1i = n1i, n2i = n2i)
  for (arg_name in names(count_args)) {
    if (!is.null(count_args[[arg_name]])) {
      BayesTools::check_int(
        count_args[[arg_name]], arg_name,
        check_length = k, allow_NULL = FALSE, allow_NA = TRUE, lower = 0
      )
    }
  }

  has_cells  <- !is.null(ci) && !is.null(bi) && !is.null(di)
  has_totals <- !is.null(ci) && !is.null(n1i) && !is.null(n2i)

  if (!has_cells && !has_totals) {
    stop("For GLMM models, provide either (ai, bi, ci, di) or (ai, ci, n1i, n2i).", call. = FALSE)
  }

  if (!is.null(n1i) && any(ai > n1i, na.rm = TRUE))
    stop("Invalid data: some values of 'bi' (= n1i - ai) are negative.", call. = FALSE)
  if (!is.null(n2i) && any(ci > n2i, na.rm = TRUE))
    stop("Invalid data: some values of 'di' (= n2i - ci) are negative.", call. = FALSE)

  if (!is.null(bi) && !is.null(n1i) && any(n1i != ai + bi, na.rm = TRUE))
    stop("The provided 'n1i' values must equal 'ai + bi' when both are supplied.", call. = FALSE)
  if (!is.null(di) && !is.null(n2i) && any(n2i != ci + di, na.rm = TRUE))
    stop("The provided 'n2i' values must equal 'ci + di' when both are supplied.", call. = FALSE)

  if (!is.null(bi)) {
    n1i <- ai + bi
  }
  if (!is.null(di)) {
    n2i <- ci + di
  }

  bi <- n1i - ai
  di <- n2i - ci

  # Extract and validate common optional variables (weights, cluster, slab)
  optional <- .check_and_list_data.optional_vars(.call, data, .envir, k, "ai")

  ### Construct output data frame
  data_outcome <- data.frame(
    ai            = ai,
    ci            = ci,
    n1i           = n1i,
    n2i           = n2i,
    cluster       = if (!is.null(optional$cluster)) optional$cluster else rep(NA_character_, k),
    cluster_label = if (!is.null(optional$cluster)) as.character(optional$cluster) else rep(NA_character_, k),
    slab          = if (!is.null(optional$slab))      optional$slab      else rep(NA_character_, k),
    weights       = if (!is.null(optional$weights))   optional$weights   else rep(NA, k),
    stringsAsFactors = FALSE
  )

  return(list(
    data_outcome       = data_outcome,
    k                  = k,
    mods_from_yi       = NULL,  # GLMM does not support formula syntax
    formula_yi         = NULL,
    weights_provided   = optional$weights_provided,
    slab_provided      = optional$slab_provided,
    cluster_provided   = optional$cluster_provided,
    na_check_cols      = c("ai", "ci", "n1i", "n2i"),  # Check cell counts for NAs
    effect_direction   = "positive", # Non-consequential, for consistency with .norm
    outcome_type       = "bin"
  ))
}


# Internal function to extract and validate outcome variables for Poisson GLMM models
# Handles x1i, x2i, t1i, t2i, weights, cluster, slab (for measure = "IRR")
#
# @param .call Matched call from the calling function
# @param data The data frame (can be NULL)
# @param .envir The enclosing environment
# @param skip_validation Whether to skip strict validation checks (for newdata).
#        When TRUE, allows t1i/t2i = 0 (useful for prediction).
#
# @return A list with:
#   - data_outcome: data.frame with x1i, x2i, t1i, t2i, cluster, slab, weights
#   - k: number of observations
#   - mods_from_yi: NULL (GLMM does not support formula syntax for outcome)
#   - formula_yi: NULL
#   - slab_provided: logical, whether slab was provided by user
#   - cluster_provided: logical, whether cluster was provided by user
#   - na_check_cols: character vector of column names to check for NAs
.check_and_list_data.outcome.pois <- function(.call, data, .envir, skip_validation = FALSE) {

  # Extract event counts and person-time for Poisson models
  x1i <- .get_variable(.call, data, .envir, "x1i", allow_NULL = TRUE)
  x2i <- .get_variable(.call, data, .envir, "x2i", allow_NULL = TRUE)
  t1i <- .get_variable(.call, data, .envir, "t1i", allow_NULL = TRUE)
  t2i <- .get_variable(.call, data, .envir, "t2i", allow_NULL = TRUE)

  ### Validate that all required variables are provided
  if (is.null(x1i) || is.null(x2i) || is.null(t1i) || is.null(t2i)) {
    stop("For Poisson models (measure = 'IRR'), all of 'x1i', 'x2i', 't1i', and 't2i' must be provided.", call. = FALSE)
  }

  # Determine k from x1i
  k <- length(x1i)

  ### Validate inputs
  # x1i: event counts in treatment group
  BayesTools::check_int(x1i, "x1i", check_length = k, allow_NULL = FALSE, allow_NA = TRUE, lower = 0)

  # x2i: event counts in control group
  BayesTools::check_int(x2i, "x2i", check_length = k, allow_NULL = FALSE, allow_NA = TRUE, lower = 0)

  # t1i: person-time in treatment group
  # When skip_validation = TRUE (newdata), allow t1i = 0
  BayesTools::check_real(t1i, "t1i", check_length = k, allow_NULL = FALSE, allow_NA = TRUE, lower = 0, allow_bound = skip_validation)

  # t2i: person-time in control group
  # When skip_validation = TRUE (newdata), allow t2i = 0
  BayesTools::check_real(t2i, "t2i", check_length = k, allow_NULL = FALSE, allow_NA = TRUE, lower = 0, allow_bound = skip_validation)

  # Extract and validate common optional variables (weights, cluster, slab)
  optional <- .check_and_list_data.optional_vars(.call, data, .envir, k, "x1i")

  ### Construct output data frame
  data_outcome <- data.frame(
    x1i           = x1i,
    x2i           = x2i,
    t1i           = t1i,
    t2i           = t2i,
    cluster       = if (!is.null(optional$cluster)) optional$cluster else rep(NA_character_, k),
    cluster_label = if (!is.null(optional$cluster)) as.character(optional$cluster) else rep(NA_character_, k),
    slab          = if (!is.null(optional$slab))      optional$slab      else rep(NA_character_, k),
    weights       = if (!is.null(optional$weights))   optional$weights   else rep(NA, k),
    stringsAsFactors = FALSE
  )

  return(list(
    data_outcome       = data_outcome,
    k                  = k,
    mods_from_yi       = NULL,  # GLMM does not support formula syntax
    formula_yi         = NULL,
    weights_provided   = optional$weights_provided,
    slab_provided      = optional$slab_provided,
    cluster_provided   = optional$cluster_provided,
    na_check_cols      = c("x1i", "x2i", "t1i", "t2i"),  # Check Poisson data for NAs
    effect_direction   = "positive", # Non-consequential, for consistency with .norm
    outcome_type       = "pois"
  ))
}


# Internal function to extract and validate predictor variables (mods or scale)
# Handles formulas, matrices, and data.frames
#
# @param .call Matched call from the calling function
# @param data The data frame (can be NULL)
# @param .envir The enclosing environment
# @param name The name of the argument ("mods" or "scale")
# @param k The expected number of rows
#
# @return A data.frame with predictor variables (with "formula" attribute) or NULL
.check_and_list_data.predictors <- function(.call, data, .envir, name, k) {

  # Check if argument is in the call
  arg_index <- match(name, names(.call))

  if (is.na(arg_index))
    return(NULL)

  mods_expr <- .call[[arg_index]]

  if (is.null(mods_expr))
    return(NULL)

  # Evaluate the expression to get the formula, matrix, or data.frame
  if (!is.null(data) && is.data.frame(data)) {
    mods <- try(eval(mods_expr, data, .envir), silent = TRUE)
  } else {
    mods <- try(eval(mods_expr, .envir), silent = TRUE)
  }

  if (inherits(mods, "try-error"))
    stop(paste0("Cannot evaluate the '", name, "' argument: ",
                conditionMessage(attr(mods, "condition"))), call. = FALSE)

  # Handle formula input
  if (inherits(mods, "formula")) {

    original_formula <- mods

    # Ensure formula has no LHS (response)
    if (length(mods) == 3) {
      warning(paste0("The '", name, "' formula should not have a left-hand side. ",
                     "The LHS will be ignored."), call. = FALSE)
      mods <- mods[-2]
      original_formula <- mods
    }

    # Create model frame from formula
    if (!is.null(data) && is.data.frame(data)) {
      mf <- try(
        stats::model.frame(mods, data = data, na.action = stats::na.pass),
        silent = TRUE
      )
    } else {
      mf <- try(
        stats::model.frame(mods, na.action = stats::na.pass),
        silent = TRUE
      )
    }

    if (inherits(mf, "try-error"))
      stop(paste0("Cannot create model frame from '", name, "' formula: ",
                  conditionMessage(attr(mf, "condition"))), call. = FALSE)

    if (nrow(mf) != k)
      stop(paste0("The number of rows in '", name, "' (", nrow(mf),
                  ") must equal the number of effect sizes (", k, ")."), call. = FALSE)

    rownames(mf) <- NULL
    attr(mf, "formula") <- original_formula

    return(mf)

  } else if (is.matrix(mods)) {

    mods_df <- as.data.frame(mods)

    if (nrow(mods_df) != k)
      stop(paste0("The number of rows in '", name, "' (", nrow(mods_df),
                  ") must equal the number of effect sizes (", k, ")."), call. = FALSE)

    rownames(mods_df) <- NULL
    attr(mods_df, "formula") <- .create_formula_from_names(names(mods_df))

    return(mods_df)

  } else if (is.data.frame(mods)) {

    mods_df <- mods

    if (nrow(mods_df) != k)
      stop(paste0("The number of rows in '", name, "' (", nrow(mods_df),
                  ") must equal the number of effect sizes (", k, ")."), call. = FALSE)

    rownames(mods_df) <- NULL
    attr(mods_df, "formula") <- .create_formula_from_names(names(mods_df))

    return(mods_df)

  } else {
    stop(paste0("The '", name, "' argument must be a formula, matrix, or data.frame."),
         call. = FALSE)
  }
}


# Internal function to coerce character predictors to factors
# Character moderators and scale predictors must be factors before downstream
# contrast handling builds design matrices.
#
# @param df A data.frame with predictor variables (can be NULL)
#
# @return The input data.frame with character columns converted to factors
.check_and_list_data.coerce_character_predictors <- function(df) {

  if (is.null(df))
    return(NULL)

  character_columns <- vapply(df, is.character, logical(1))

  if (!any(character_columns))
    return(df)

  for (col_name in names(df)[character_columns]) {
    df[[col_name]] <- factor(df[[col_name]])
  }

  return(df)
}


# Internal function to validate and convert subset argument to logical vector
#
# @param subset The subset argument (logical or numeric vector)
# @param k The expected length (number of observations)
#
# @return A logical vector of length k
.check_and_list_data.validate_subset <- function(subset, k) {

  if (is.logical(subset)) {

    if (length(subset) != k)
      stop(paste0("The 'subset' argument must have length ", k, " (same as the outcome) when logical."), call. = FALSE)

    # Check for NAs in subset
    if (any(is.na(subset)))
      stop("The 'subset' argument must not contain NA values.", call. = FALSE)

    return(subset)

  } else if (is.numeric(subset)) {

    # Check for NAs in subset
    if (any(is.na(subset)))
      stop("The 'subset' argument must not contain NA values.", call. = FALSE)

    if (any(!is.finite(subset)) || any(subset != floor(subset)))
      stop("The 'subset' argument must contain integer numeric indices.", call. = FALSE)

    if (any(subset < 1) || any(subset > k))
      stop(paste0("The 'subset' argument must contain values between 1 and ", k, "."), call. = FALSE)

    # Convert numeric indices to logical
    subset_logical <- rep(FALSE, k)
    subset_logical[subset] <- TRUE
    return(subset_logical)

  } else {
    stop("The 'subset' argument must be a logical or numeric vector.", call. = FALSE)
  }
}


# Internal function to apply subsetting to a data.frame
# Preserves formula attributes and drops unused factor levels
#
# @param df A data.frame (can be NULL)
# @param subset A logical vector indicating which rows to keep
#
# @return The subsetted data.frame (or NULL if input was NULL)
.check_and_list_data.subset <- function(df, subset) {

  if (is.null(df))
    return(NULL)

  # Preserve attributes
  saved_formula    <- attr(df, "formula")
  saved_formula_yi <- attr(df, "formula_yi")

  # Apply subset
  df <- df[subset, , drop = FALSE]

  # Drop unused factor levels
  df <- droplevels(df)

  # Reset row names
  rownames(df) <- NULL

  # Restore attributes
  if (!is.null(saved_formula))
    attr(df, "formula") <- saved_formula
  if (!is.null(saved_formula_yi))
    attr(df, "formula_yi") <- saved_formula_yi

  return(df)
}


# Internal function to validate predictor variables (mods or scale)
# Checks that:
#   - Continuous variables have sd > 0 (are not constant)
#   - Factor variables have more than one level
#
# These checks are appropriate for original data but should be skipped for
# newdata used in prediction, where single-value predictions are legitimate.
#
# @param df A data.frame with predictor variables (can be NULL)
# @param name The name of the predictor type ("mods" or "scale") for error messages
# @param skip_validation Logical; if TRUE, skip all validation checks.
#        Should be TRUE when processing newdata for prediction.
#
# @return NULL (invisibly); stops with error if validation fails
.check_and_list_data.validate_predictors <- function(df, name, skip_validation = FALSE) {

  if (is.null(df))
    return(invisible(NULL))

  # Skip validation for newdata (e.g., predicting at a single factor level or

  # single continuous value is valid for prediction)
  if (skip_validation)
    return(invisible(NULL))

  for (col_name in names(df)) {
    col <- df[[col_name]]

    if (is.factor(col) || is.character(col)) {
      # Factor/character: must have more than one unique level
      n_levels <- length(unique(col))
      if (n_levels < 2) {
        stop(paste0("The '", name, "' variable '", col_name,
                    "' has only one level. Factor predictors must have at least two levels."),
             call. = FALSE)
      }
    } else if (is.numeric(col)) {
      # Numeric: must have sd > 0 (not constant)
      col_sd <- stats::sd(col, na.rm = TRUE)
      if (is.na(col_sd) || col_sd == 0) {
        stop(paste0("The '", name, "' variable '", col_name,
                    "' has zero variance (all values are identical). ",
                    "Continuous predictors must have variation."),
             call. = FALSE)
      }
    }
    # For other types (logical, etc.), no specific check
  }

  return(invisible(NULL))
}


# Internal function to check for NA values across a list of data.frames
# Returns a logical vector indicating which rows have at least one NA
#
# @param data_list A named list of data.frames to check
#
# @return A logical vector where TRUE indicates the row has at least one NA
.check_and_list_data.check_na <- function(data_list) {

  if (length(data_list) == 0)
    return(logical(0))

  # Combine all data.frames into one for vectorized NA check
  combined <- do.call(cbind, Filter(function(df) !is.null(df) && is.data.frame(df), data_list))

  # Vectorized NA check: TRUE if any column in that row has NA
  na_rows <- rowSums(is.na(combined)) > 0

  return(na_rows)
}


# Internal helper to create a formula from column names
# Creates a one-sided formula like ~ var1 + var2 + var3
#
# @param col_names Character vector of column names
# @return A formula object
.create_formula_from_names <- function(col_names) {

  if (length(col_names) == 0) {
    return(~ 1)
  }

  # Backtick names that need protection (contain spaces, special chars, etc.)
  col_names_safe <- vapply(col_names, function(nm) {
    if (make.names(nm) != nm) {
      paste0("`", nm, "`")
    } else {
      nm
    }
  }, character(1), USE.NAMES = FALSE)

  # Create formula string and convert to formula
  formula_str <- paste("~", paste(col_names_safe, collapse = " + "))
  return(stats::as.formula(formula_str))
}


#' @title Print method for RoBMA_data objects
#'
#' @description Prints a concise summary of an RoBMA_data object.
#'
#' @param x an RoBMA_data object
#' @param n maximum number of rows to display for each component. Defaults to 6.
#' @param ... additional arguments (ignored)
#'
#' @return Invisibly returns the input object.
#'
#' @export
print.RoBMA_data <- function(x, n = 6, ...) {

  k_final   <- attr(x, "k_final")
  n_dropped <- attr(x, "n_dropped")

  # Outcome data header
  if (n_dropped > 0) {
    cat(sprintf("Outcome data (k = %d, dropped = %d)\n", k_final, n_dropped))
  } else {
    cat(sprintf("Outcome data (k = %d)\n", k_final))
  }
  print(utils::head(x$outcome, n = n))

  # Moderators (effect)
  if (!is.null(x$mods)) {
    cat("\nModerators (effect)\n")
    print(utils::head(x$mods, n = n))
  }

  # Moderators (scale)
  if (!is.null(x$scale)) {
    cat("\nModerators (scale)\n")
    print(utils::head(x$scale, n = n))
  }

  # Indicate if more rows exist

  if (k_final > n) {
    cat(sprintf("\n... with %d more rows\n", k_final - n))
  }

  invisible(x)
}


# Internal helper to normalize data.frame/list newdata inputs.
.prepare_newdata_as_data_frame <- function(newdata) {

  if (is.list(newdata) && !is.data.frame(newdata)) {
    newdata <- as.data.frame(newdata, stringsAsFactors = FALSE)
  }

  return(newdata)
}


# Internal helper to report missing newdata columns.
.prepare_newdata_stop_missing <- function(newdata, cols) {

  missing_cols <- setdiff(cols, names(newdata))
  if (length(missing_cols) == 0L) {
    return(invisible(TRUE))
  }

  stop(
    "The 'newdata' must contain columns: ",
    paste(missing_cols, collapse = ", "), ".",
    call. = FALSE
  )
}


# Internal helper to add placeholder columns that are not used by prediction.
.prepare_newdata_add_missing <- function(newdata, cols, value, n) {

  for (col in cols) {
    if (!col %in% names(newdata)) {
      newdata[[col]] <- rep(value, n)
    }
  }

  return(newdata)
}


# Internal helper to decide whether normal newdata needs sampling SE.
.prepare_newdata_needs_norm_sei <- function(object, type, bias_adjusted) {

  return(
    type == "response" ||
      (
        !bias_adjusted &&
          type %in% c("terms", "cluster", "estimate") &&
          (.is_PET(object) || .is_PEESE(object))
      )
  )
}


# Internal helper to add outcome placeholders for the shared data parser.
.prepare_newdata_outcome <- function(object, newdata, type, bias_adjusted) {

  outcome_type <- .outcome_type(object)
  n_new        <- nrow(newdata)

  if (outcome_type == "norm") {
    newdata <- .prepare_newdata_add_missing(
      newdata = newdata,
      cols    = "yi",
      value   = 0,
      n       = n_new
    )

    if (!("sei" %in% names(newdata) || "vi" %in% names(newdata))) {
      if (.prepare_newdata_needs_norm_sei(object, type, bias_adjusted)) {
        stop(
          "The 'newdata' must contain either 'sei' (standard error) or ",
          "'vi' (variance) column.",
          call. = FALSE
        )
      }
      newdata[["sei"]] <- rep(0, n_new)
    }

  } else if (outcome_type == "bin") {
    if (type == "response") {
      .prepare_newdata_stop_missing(newdata, c("n1i", "n2i"))
    }
    newdata <- .prepare_newdata_add_missing(
      newdata = newdata,
      cols    = c("ai", "ci", "n1i", "n2i"),
      value   = 0L,
      n       = n_new
    )

  } else if (outcome_type == "pois") {
    if (type == "response") {
      .prepare_newdata_stop_missing(newdata, c("t1i", "t2i"))
    }
    newdata <- .prepare_newdata_add_missing(
      newdata = newdata,
      cols    = c("x1i", "x2i", "t1i", "t2i"),
      value   = 0,
      n       = n_new
    )
  }

  return(newdata)
}


# Internal helper to construct outcome arguments for `.check_and_list_data`.
.prepare_newdata_outcome_call_args <- function(outcome_type, newdata) {

  if (outcome_type == "norm") {
    call_args <- list(yi = quote(yi))

    if ("sei" %in% names(newdata)) {
      call_args[["sei"]] <- quote(sei)
    } else {
      call_args[["vi"]] <- quote(vi)
    }

  } else if (outcome_type == "bin") {
    call_args <- list(
      ai  = quote(ai),
      ci  = quote(ci),
      n1i = quote(n1i),
      n2i = quote(n2i)
    )

  } else if (outcome_type == "pois") {
    call_args <- list(
      x1i = quote(x1i),
      x2i = quote(x2i),
      t1i = quote(t1i),
      t2i = quote(t2i)
    )
  }

  return(call_args)
}


# Internal helper to validate formula variables before evaluation.
.prepare_newdata_validate_formula_vars <- function(newdata, formula, label) {

  missing_vars <- setdiff(all.vars(formula), names(newdata))
  if (length(missing_vars) == 0L) {
    return(invisible(TRUE))
  }

  stop(
    "The 'newdata' must contain all ", label, " variables. Missing: ",
    paste(missing_vars, collapse = ", "), ".",
    call. = FALSE
  )
}


# Internal helper function to prepare newdata for prediction
# Reuses `.check_and_list_data` by constructing appropriate call and environment
#
# The newdata data.frame must contain all variables used by the requested
# prediction. Moderator and scale variables are always required when referenced
# by the fitted formulas. Outcome columns are required only when the requested
# prediction uses the sampling distribution or PET/PEESE bias terms; otherwise
# internal dummy values are inserted to keep the shared parser path.
#
# @param object A fitted brma object
# @param newdata A data.frame with new data for prediction.
# @param type Prediction type: "terms", "effect", or "response"
# @param bias_adjusted Whether PET/PEESE terms should be omitted.
#
# @return A data list equivalent to `object[["data"]]` but for `newdata`
.prepare_newdata <- function(object, newdata, type, bias_adjusted = FALSE) {

  # extract settings from the original fitted object's data attributes
  original_data <- object[["data"]]
  set_contrast_factor_predictors    <- attr(original_data, "set_contrast_factor_predictors")
  standardize_continuous_predictors <- attr(original_data, "standardize_continuous_predictors")
  effect_direction                  <- .effect_direction(object)
  outcome_type                      <- .outcome_type(object)

  newdata <- .prepare_newdata_as_data_frame(newdata)
  newdata <- .prepare_newdata_outcome(
    object        = object,
    newdata       = newdata,
    type          = type,
    bias_adjusted = bias_adjusted
  )

  # build the synthetic call expression
  # start with the base call structure
  call_args <- list(quote(.prepare_newdata_call))

  # add data argument
  call_args[["data"]] <- quote(data)

  # add outcome arguments based on outcome_type
  call_args <- c(
    call_args,
    .prepare_newdata_outcome_call_args(outcome_type, newdata)
  )

  # add mods formula if this is a regression model
  if (.is_mods(object)) {
    call_args[["mods"]] <- attr(original_data[["mods"]], "formula")
  }

  # add scale formula if present
  if (.is_scale(object)) {
    call_args[["scale"]] <- attr(original_data[["scale"]], "formula")
  }

  # add cluster structure for multilevel predictions
  if (.is_multilevel(object)) {
    if ("cluster" %in% names(newdata)) {
      call_args[["cluster"]] <- quote(cluster)
    } else {
      n_new <- nrow(newdata)
      newdata[[".RoBMA_cluster"]] <- seq_len(n_new)
      call_args[["cluster"]] <- quote(.RoBMA_cluster)
    }
  }

  # construct the call object
  .call <- as.call(call_args)

  # pre-validate that all formula variables exist in newdata
  # this prevents formulas from finding variables in parent environments
  if (.is_mods(object)) {
    mods_formula <- attr(original_data[["mods"]], "formula")
    .prepare_newdata_validate_formula_vars(newdata, mods_formula, "moderator")
  }

  if (.is_scale(object)) {
    scale_formula <- attr(original_data[["scale"]], "formula")
    .prepare_newdata_validate_formula_vars(newdata, scale_formula, "scale")
  }

  # create environment with newdata as "data"
  # use baseenv() as parent to prevent variable leakage from calling context
  .envir           <- new.env(parent = baseenv())
  .envir[["data"]] <- newdata

  # determine class and measure for .check_and_list_data
  measure    <- .measure(object)
  data_class <- switch(
    outcome_type,
    "norm" = "norm",
    "bin"  = "glmm",
    "pois" = "glmm"
  )

  # call .check_and_list_data with extracted settings
  # Use skip_validation = TRUE because:
  # 1. Newdata may have single-level factors (predicting for one group)
  # 2. Newdata may have zero-variance continuous predictors (predicting at one value)
  # 3. Factor levels are validated by BayesTools during formula evaluation
  # 4. Scaling uses original data's mean/sd stored in the fit object
  new_data <- .check_and_list_data(
    .call  = .call,
    .envir = .envir,
    class  = data_class,
    measure = measure,
    set_contrast_factor_predictors    = set_contrast_factor_predictors,
    standardize_continuous_predictors = standardize_continuous_predictors,
    effect_direction                  = effect_direction,
    skip_validation                   = TRUE
  )

  return(new_data)
}

# ============================================================================ #
# brma.as_draws.R
# ============================================================================ #
#
# This file provides an interface to the posterior package for brma objects.
# It defines:
#
# - Generic S3 methods for as_draws family (for use when posterior is not loaded)
# - Methods for brma class that convert MCMC samples to posterior draws formats
# - Helper function to extract mcmc.list from brma objects
#
# ============================================================================ #


# ---------------------------------------------------------------------------- #
# Documentation
# ---------------------------------------------------------------------------- #

#' @title Convert brma Objects to posterior Draws Formats
#'
#' @description Provides an interface to the \pkg{posterior} package
#' for \code{brma} objects. These functions convert the MCMC samples
#' from a fitted \code{brma} model to various draws formats supported
#' by the \pkg{posterior} package, enabling the use of posterior's
#' rich set of diagnostics and summary functions.
#'
#' @param x an object to convert. The \code{brma} methods expect a fitted
#' \code{brma} object; default methods forward non-\code{brma} objects to the
#' corresponding \pkg{posterior} conversion function.
#' @param ... additional arguments passed to the corresponding
#' \pkg{posterior} function.
#'
#' @details
#' These functions are S3 generics. Their default methods forward to the
#' corresponding \pkg{posterior} generics so attaching \pkg{RoBMA} preserves
#' the usual \pkg{posterior} behavior for non-\code{brma} objects.
#'
#' The following conversion functions are available:
#' \itemize{
#'   \item \code{as_draws}: converts to the default draws format
#'   \item \code{as_draws_array}: converts to a 3-D array (iteration x chain x variable)
#'   \item \code{as_draws_df}: converts to a data frame with columns for iterations, chains, and variables
#'   \item \code{as_draws_list}: converts to a list of lists
#'   \item \code{as_draws_matrix}: converts to a 2-D matrix (draw x variable)
#'   \item \code{as_draws_rvars}: converts to random variable objects
#' }
#'
#' These methods require the \pkg{posterior} package to be installed.
#' For \code{brma} methods, conversion is performed by first extracting the MCMC
#' samples as a \code{mcmc.list} object and then using the corresponding
#' \pkg{posterior} conversion function. \code{brma_samples} objects have
#' separate methods documented at \code{\link{as_draws.brma_samples}}.
#'
#' @return An object of the corresponding \pkg{posterior} draws class.
#'
#' @seealso \code{\link[posterior]{draws}}, \code{\link{brma}},
#' \code{\link{as_draws.brma_samples}}
#'
#' @name as_draws.brma
NULL


# ---------------------------------------------------------------------------- #
# Helper: Extract mcmc.list from brma objects
# ---------------------------------------------------------------------------- #

#' @title Extract mcmc.list from brma Objects
#'
#' @description Internal helper to extract the MCMC samples from a fitted
#' \code{brma} object as a \code{mcmc.list} object suitable for conversion
#' to \pkg{posterior} draws formats.
#'
#' @param x a fitted \code{brma} object
#'
#' @return A \code{mcmc.list} object containing the MCMC samples.
#'
#' @noRd
.brma_to_mcmc.list <- function(x) {

  if (!inherits(x, "brma")) {
    stop("'x' must be a 'brma' object", call. = FALSE)
  }

  if (is.null(x[["fit"]]) || length(x[["fit"]]) == 0) {
    stop("The 'brma' object does not contain a valid fit", call. = FALSE)
  }

  return(coda::as.mcmc.list(x[["fit"]]))
}


# ---------------------------------------------------------------------------- #
# Helper: Check posterior package availability
# ---------------------------------------------------------------------------- #

.check_posterior_package <- function() {
  if (!requireNamespace("posterior", quietly = TRUE)) {
    stop(
      "Package 'posterior' is required for as_draws conversion.\n",
      "Install it with: install.packages('posterior')",
      call. = FALSE
    )
  }
}

.posterior_default_method <- function(generic) {

  return(getS3method(
    f     = generic,
    class = "default",
    envir = asNamespace("posterior")
  ))
}

#' @rdname as_draws.brma
#' @export
as_draws <- function(x, ...) {
  UseMethod("as_draws")
}

#' @rdname as_draws.brma
#' @export
as_draws_array <- function(x, ...) {
  UseMethod("as_draws_array")
}

#' @rdname as_draws.brma
#' @export
as_draws_df <- function(x, ...) {
  UseMethod("as_draws_df")
}

#' @rdname as_draws.brma
#' @export
as_draws_list <- function(x, ...) {
  UseMethod("as_draws_list")
}

#' @rdname as_draws.brma
#' @export
as_draws_matrix <- function(x, ...) {
  UseMethod("as_draws_matrix")
}

#' @rdname as_draws.brma
#' @export
as_draws_rvars <- function(x, ...) {
  UseMethod("as_draws_rvars")
}

#' @rdname as_draws.brma
#' @export
as_draws.default <- function(x, ...) {

  .check_posterior_package()

  return(.posterior_default_method("as_draws")(x, ...))
}

#' @rdname as_draws.brma
#' @export
as_draws_array.default <- function(x, ...) {

  .check_posterior_package()

  return(.posterior_default_method("as_draws_array")(x, ...))
}

#' @rdname as_draws.brma
#' @export
as_draws_df.default <- function(x, ...) {

  .check_posterior_package()

  return(.posterior_default_method("as_draws_df")(x, ...))
}

#' @rdname as_draws.brma
#' @export
as_draws_list.default <- function(x, ...) {

  .check_posterior_package()

  return(.posterior_default_method("as_draws_list")(x, ...))
}

#' @rdname as_draws.brma
#' @export
as_draws_matrix.default <- function(x, ...) {

  .check_posterior_package()

  return(.posterior_default_method("as_draws_matrix")(x, ...))
}

#' @rdname as_draws.brma
#' @export
as_draws_rvars.default <- function(x, ...) {

  .check_posterior_package()

  return(.posterior_default_method("as_draws_rvars")(x, ...))
}


# ---------------------------------------------------------------------------- #
# as_draws methods for brma class
# ---------------------------------------------------------------------------- #

#' @rdname as_draws.brma
#' @export
as_draws.brma <- function(x, ...) {

  .check_posterior_package()

  mcmc.list <- .brma_to_mcmc.list(x)

  return(posterior::as_draws(mcmc.list, ...))
}

#' @rdname as_draws.brma
#' @export
as_draws_array.brma <- function(x, ...) {

  .check_posterior_package()

  mcmc.list <- .brma_to_mcmc.list(x)

  return(posterior::as_draws_array(mcmc.list, ...))
}

#' @rdname as_draws.brma
#' @export
as_draws_df.brma <- function(x, ...) {

  .check_posterior_package()

  mcmc.list <- .brma_to_mcmc.list(x)

  return(posterior::as_draws_df(mcmc.list, ...))
}

#' @rdname as_draws.brma
#' @export
as_draws_list.brma <- function(x, ...) {

  .check_posterior_package()

  mcmc.list <- .brma_to_mcmc.list(x)

  return(posterior::as_draws_list(mcmc.list, ...))
}

#' @rdname as_draws.brma
#' @export
as_draws_matrix.brma <- function(x, ...) {

  .check_posterior_package()

  mcmc.list <- .brma_to_mcmc.list(x)

  return(posterior::as_draws_matrix(mcmc.list, ...))
}

#' @rdname as_draws.brma
#' @export
as_draws_rvars.brma <- function(x, ...) {

  .check_posterior_package()

  mcmc.list <- .brma_to_mcmc.list(x)

  return(posterior::as_draws_rvars(mcmc.list, ...))
}

context("as_draws methods for brma_samples objects")

# Load common test helpers
source(testthat::test_path("common-functions.R"))

# load prefitted model
skip_if_no_fits()
skip_if_not_installed("posterior")

fit <- load_fit("bcg_meta-analysis")

test_that("as_draws methods return posterior draws for brma_samples", {

  # obtain brma_samples
  brma_samples <- predict(fit)
  n_draws       <- nrow(brma_samples)
  n_chains      <- attr(brma_samples, "nchains")
  n_iter        <- attr(brma_samples, "niter")

  # test as_draws and consistency between methods
  draws <- RoBMA::as_draws(brma_samples)
  expect_s3_class(draws, "draws")
  expect_true(posterior::ndraws(draws)      == n_draws)
  expect_true(posterior::nchains(draws)     == n_chains)
  expect_true(posterior::niterations(draws) == n_iter)

  draws <- posterior::as_draws(brma_samples)
  expect_s3_class(draws, "draws")
  expect_true(posterior::ndraws(draws)      == n_draws)
  expect_true(posterior::nchains(draws)     == n_chains)
  expect_true(posterior::niterations(draws) == n_iter)

  # test as_draws_array
  draws <- RoBMA::as_draws_array(brma_samples)
  expect_s3_class(draws, "draws_array")
  expect_true(posterior::ndraws(draws)      == n_draws)
  expect_true(posterior::nchains(draws)     == n_chains)
  expect_true(posterior::niterations(draws) == n_iter)

  draws <- posterior::as_draws_array(brma_samples)
  expect_s3_class(draws, "draws_array")
  expect_true(posterior::ndraws(draws)      == n_draws)
  expect_true(posterior::nchains(draws)     == n_chains)
  expect_true(posterior::niterations(draws) == n_iter)

  # test as_draws_df
  draws <- RoBMA::as_draws_df(brma_samples)
  expect_s3_class(draws, "draws_df")
  expect_true(posterior::ndraws(draws)      == n_draws)
  expect_true(posterior::nchains(draws)     == n_chains)
  expect_true(posterior::niterations(draws) == n_iter)

  draws <- posterior::as_draws_df(brma_samples)
  expect_s3_class(draws, "draws_df")
  expect_true(posterior::ndraws(draws)      == n_draws)
  expect_true(posterior::nchains(draws)     == n_chains)
  expect_true(posterior::niterations(draws) == n_iter)

  # test as_draws_list
  draws <- RoBMA::as_draws_list(brma_samples)
  expect_s3_class(draws, "draws_list")
  expect_true(posterior::ndraws(draws)      == n_draws)
  expect_true(posterior::nchains(draws)     == n_chains)
  expect_true(posterior::niterations(draws) == n_iter)

  draws <- posterior::as_draws_list(brma_samples)
  expect_s3_class(draws, "draws_list")
  expect_true(posterior::ndraws(draws)      == n_draws)
  expect_true(posterior::nchains(draws)     == n_chains)
  expect_true(posterior::niterations(draws) == n_iter)

  # test as_draws_rvars
  draws <- RoBMA::as_draws_rvars(brma_samples)
  expect_s3_class(draws, "draws_rvars")
  expect_true(posterior::ndraws(draws)      == n_draws)
  expect_true(posterior::nchains(draws)     == n_chains)
  expect_true(posterior::niterations(draws) == n_iter)
  expect_true("mu" == posterior::variables(draws))

  draws <- posterior::as_draws_rvars(brma_samples)
  expect_s3_class(draws, "draws_rvars")
  expect_true(posterior::ndraws(draws)      == n_draws)
  expect_true(posterior::nchains(draws)     == n_chains)
  expect_true(posterior::niterations(draws) == n_iter)
  expect_true("mu" == posterior::variables(draws))
})

test_that("as_draws methods preserve brma_samples values and chain order", {

  brma_samples    <- predict(fit)
  expected_matrix <- unclass(brma_samples)
  draws_matrix    <- as.matrix(RoBMA::as_draws_matrix(brma_samples))
  selected_rows   <- unique(round(seq(1, nrow(expected_matrix), length.out = 7)))
  selected_vars   <- colnames(expected_matrix)

  expect_equal(
    as.vector(draws_matrix[selected_rows, selected_vars, drop = FALSE]),
    as.vector(expected_matrix[selected_rows, selected_vars, drop = FALSE]),
    tolerance = 0
  )

  draws_array <- RoBMA::as_draws_array(brma_samples)
  first_var   <- selected_vars[[1]]
  expect_equal(
    as.numeric(draws_array[1, 1, first_var]),
    as.numeric(expected_matrix[1, first_var]),
    tolerance = 0,
    info      = "first brma_samples chain and iteration are preserved"
  )
})

context("as_draws methods for brma objects")

# Load common test helpers
source(testthat::test_path("common-functions.R"))

skip_if_not_installed("posterior")

test_that("RoBMA as_draws defaults forward to posterior", {

  x <- matrix(seq_len(6), ncol = 2)
  colnames(x) <- c("alpha", "beta")

  expect_s3_class(RoBMA::as_draws(x), "draws")
  expect_s3_class(RoBMA::as_draws_array(x), "draws_array")
  expect_s3_class(RoBMA::as_draws_df(x), "draws_df")
  expect_s3_class(RoBMA::as_draws_list(x), "draws_list")
  expect_s3_class(RoBMA::as_draws_matrix(x), "draws_matrix")
  expect_s3_class(RoBMA::as_draws_rvars(x), "draws_rvars")
})

# load prefitted model
skip_if_no_fits()
fit <- load_fit("bcg_meta-analysis")

.fit_draw_dimensions <- function(fit) {

  mcmc_list <- RoBMA:::.brma_to_mcmc.list(fit)
  n_iter    <- vapply(mcmc_list, function(x) nrow(as.matrix(x)), integer(1))

  return(list(
    n_draws  = sum(n_iter),
    n_chains = length(mcmc_list),
    n_iter   = unique(n_iter)
  ))
}

fit_dims <- .fit_draw_dimensions(fit)

test_that("as_draws methods return posterior draws for fits", {

  # test as_draws and consistency between methods
  draws <- RoBMA::as_draws(fit)
  expect_s3_class(draws, "draws")
  expect_true(posterior::ndraws(draws)      == fit_dims[["n_draws"]])
  expect_true(posterior::nchains(draws)     == fit_dims[["n_chains"]])
  expect_true(posterior::niterations(draws) == fit_dims[["n_iter"]])

  draws <- posterior::as_draws(fit)
  expect_s3_class(draws, "draws")
  expect_true(posterior::ndraws(draws)      == fit_dims[["n_draws"]])
  expect_true(posterior::nchains(draws)     == fit_dims[["n_chains"]])
  expect_true(posterior::niterations(draws) == fit_dims[["n_iter"]])

  # test as_draws_array
  draws <- RoBMA::as_draws_array(fit)
  expect_s3_class(draws, "draws_array")
  expect_true(posterior::ndraws(draws)      == fit_dims[["n_draws"]])
  expect_true(posterior::nchains(draws)     == fit_dims[["n_chains"]])
  expect_true(posterior::niterations(draws) == fit_dims[["n_iter"]])

  draws <- posterior::as_draws_array(fit)
  expect_s3_class(draws, "draws_array")
  expect_true(posterior::ndraws(draws)      == fit_dims[["n_draws"]])
  expect_true(posterior::nchains(draws)     == fit_dims[["n_chains"]])
  expect_true(posterior::niterations(draws) == fit_dims[["n_iter"]])

  # test as_draws_df
  draws <- RoBMA::as_draws_df(fit)
  expect_s3_class(draws, "draws_df")
  expect_true(posterior::ndraws(draws)      == fit_dims[["n_draws"]])
  expect_true(posterior::nchains(draws)     == fit_dims[["n_chains"]])
  expect_true(posterior::niterations(draws) == fit_dims[["n_iter"]])

  draws <- posterior::as_draws_df(fit)
  expect_s3_class(draws, "draws_df")
  expect_true(posterior::ndraws(draws)      == fit_dims[["n_draws"]])
  expect_true(posterior::nchains(draws)     == fit_dims[["n_chains"]])
  expect_true(posterior::niterations(draws) == fit_dims[["n_iter"]])

  # test as_draws_list
  draws <- RoBMA::as_draws_list(fit)
  expect_s3_class(draws, "draws_list")
  expect_true(posterior::ndraws(draws)      == fit_dims[["n_draws"]])
  expect_true(posterior::nchains(draws)     == fit_dims[["n_chains"]])
  expect_true(posterior::niterations(draws) == fit_dims[["n_iter"]])

  draws <- posterior::as_draws_list(fit)
  expect_s3_class(draws, "draws_list")
  expect_true(posterior::ndraws(draws)      == fit_dims[["n_draws"]])
  expect_true(posterior::nchains(draws)     == fit_dims[["n_chains"]])
  expect_true(posterior::niterations(draws) == fit_dims[["n_iter"]])

  # test as_draws_rvars
  draws <- RoBMA::as_draws_rvars(fit)
  expect_s3_class(draws, "draws_rvars")
  expect_true(posterior::ndraws(draws)      == fit_dims[["n_draws"]])
  expect_true(posterior::nchains(draws)     == fit_dims[["n_chains"]])
  expect_true(posterior::niterations(draws) == fit_dims[["n_iter"]])
  expect_true("mu" %in% posterior::variables(draws))

  draws <- posterior::as_draws_rvars(fit)
  expect_s3_class(draws, "draws_rvars")
  expect_true(posterior::ndraws(draws)      == fit_dims[["n_draws"]])
  expect_true(posterior::nchains(draws)     == fit_dims[["n_chains"]])
  expect_true(posterior::niterations(draws) == fit_dims[["n_iter"]])
  expect_true("mu" %in% posterior::variables(draws))
})

test_that("as_draws methods preserve raw MCMC values and order", {

  mcmc_list       <- RoBMA:::.brma_to_mcmc.list(fit)
  expected_matrix <- do.call(rbind, lapply(mcmc_list, as.matrix))
  draws_matrix    <- as.matrix(RoBMA::as_draws_matrix(fit))
  variables       <- intersect(colnames(expected_matrix), colnames(draws_matrix))
  selected_rows   <- unique(round(seq(1, nrow(expected_matrix), length.out = 7)))
  selected_vars   <- head(variables, 8)

  expect_true(length(selected_vars) > 0L)
  expect_equal(
    as.vector(draws_matrix[selected_rows, selected_vars, drop = FALSE]),
    as.vector(expected_matrix[selected_rows, selected_vars, drop = FALSE]),
    tolerance = 0
  )

  draws_array <- RoBMA::as_draws_array(fit)
  first_var   <- selected_vars[[1]]
  expect_equal(
    as.numeric(draws_array[1, 1, first_var]),
    as.numeric(as.matrix(mcmc_list[[1]])[1, first_var]),
    tolerance = 0,
    info      = "first chain and iteration are preserved"
  )
})

test_that("as_draws methods preserve product-space BMA and RoBMA indicators", {

  product_names <- c(
    "dat.lehmann2018_BMA.norm",
    "bcg_BMA.glmm",
    "nielweise2008_BMA.glmm",
    "dat.lehmann2018_RoBMA"
  )
  skip_if_missing_fits(product_names)

  for (name in product_names) {

    fit_product <- load_fit(name, validate = FALSE)
    n_chains    <- length(fit_product[["fit"]][["mcmc"]])
    n_iter      <- fit_product[["fit"]][["sample"]]
    n_draws     <- n_chains * n_iter

    draws <- RoBMA::as_draws(fit_product)
    expect_s3_class(draws, "draws")
    expect_equal(posterior::ndraws(draws), n_draws)
    expect_equal(posterior::nchains(draws), n_chains)
    expect_equal(posterior::niterations(draws), n_iter)

    draws_array <- RoBMA::as_draws_array(fit_product)
    expect_s3_class(draws_array, "draws_array")
    expect_equal(posterior::ndraws(draws_array), n_draws)
    expect_equal(posterior::nchains(draws_array), n_chains)
    expect_equal(posterior::niterations(draws_array), n_iter)

    draws_df <- RoBMA::as_draws_df(fit_product)
    expect_s3_class(draws_df, "draws_df")
    expect_equal(posterior::ndraws(draws_df), n_draws)
    expect_equal(posterior::nchains(draws_df), n_chains)
    expect_equal(posterior::niterations(draws_df), n_iter)

    draws_matrix <- RoBMA::as_draws_matrix(fit_product)
    expect_s3_class(draws_matrix, "draws_matrix")
    expect_equal(posterior::ndraws(draws_matrix), n_draws)

    draws_rvars <- RoBMA::as_draws_rvars(fit_product)
    expect_s3_class(draws_rvars, "draws_rvars")
    expect_equal(posterior::ndraws(draws_rvars), n_draws)

    variables <- posterior::variables(draws_df)
    expect_true(any(grepl("_indicator$", variables)),
                info = paste(name, "exposes product-space indicators"))

    if (!is.null(fit_product[["priors"]][["outcome"]][["mu"]]) &&
        BayesTools::is.prior.mixture(fit_product[["priors"]][["outcome"]][["mu"]])) {
      expect_true("mu_indicator" %in% variables, info = name)
    }
    if (!is.null(fit_product[["priors"]][["outcome"]][["tau"]]) &&
        BayesTools::is.prior.mixture(fit_product[["priors"]][["outcome"]][["tau"]])) {
      expect_true("tau_indicator" %in% variables, info = name)
    }
    if (!is.null(fit_product[["priors"]][["outcome"]][["bias"]])) {
      expect_true("bias_indicator" %in% variables, info = name)
    }
  }
})

#' @title Hat Values for brma Objects
#'
#' @description Computes hat values (leverages) from a fitted brma object.
#' Returns posterior mean leverages, one for each observation.
#'
#' @param model a fitted brma object
#' @param ... additional arguments (currently ignored)
#'
#' @details
#' Hat values (leverages) measure the influence of each observation on the
#' fitted values. In a Bayesian meta-analysis, the random effects variance \eqn{\tau^2}
#' is uncertain, so the hat matrix depends on the posterior samples of \eqn{\tau^2}.
#'
#' This function computes the diagonal elements of the hat matrix:
#' \deqn{h_i = (X (X^T W X)^{-1} X^T W)_{ii}}
#' where \eqn{W} is the weight matrix inverse to the marginal variance matrix.
#'
#' The hat matrix is computed for each posterior draw and then averaged over
#' draws, matching the vector output shape used by `metafor`.
#'
#' This method is available only for normal outcome models without
#' weightfunction selection.
#'
#' @return A numeric vector of length `K`, one posterior mean leverage per
#' observation.
#'
#' @examples \dontrun{
#' if (requireNamespace("metadat", quietly = TRUE) &&
#'     requireNamespace("metafor", quietly = TRUE)) {
#'   data(dat.bcg, package = "metadat")
#'   dat <- metafor::escalc(
#'     measure = "RR",
#'     ai      = tpos,
#'     bi      = tneg,
#'     ci      = cpos,
#'     di      = cneg,
#'     data    = dat.bcg
#'   )
#'
#'   fit <- brma(yi = yi, vi = vi, data = dat, measure = "RR")
#'   hatvalues(fit)
#' }
#' }
#'
#' @exportS3Method
hatvalues.brma <- function(model, ...) {

  # the function relies on normal-normal approximations
  # as such it is sensible only sensible for normal models
  outcome_type       <- .outcome_type(model)
  is_weightfunction  <- .is_weightfunction(model)

  if (outcome_type != "norm") {
    stop("hatvalues is only available for normal outcome models.", call. = FALSE)
  }
  if (is_weightfunction) {
    stop("hatvalues is not available for selection models (weightfunction).", call. = FALSE)
  }

  # compute hat matrix samples
  # returns list(H_diag, M_diag, ...)
  # we only need H_diag
  res <- .compute_hat_matrix_samples(
    object             = model,
    conditioning_depth = "marginal",
    return_full_H      = FALSE,
    return_se          = FALSE
  )

  # extract the diagonal
  res <- res[["H_diag"]]

  # return only the posterior mean for consistency with metafor
  res <- colMeans(res)
  res <- .diagnostic_set_names(res, model)

  return(res)
}

# ============================================================================ #
# brma.influence.R
# ============================================================================ #
#
# This file contains the influence method for brma model fits. It computes
# various diagnostics (DFFITS, Cook's distance, COVRATIO, etc.) to identify
# influential observations.
#
# ============================================================================ #


#' @title Measure Influence for brma Objects
#'
#' @description Computes DFFITS, Cook's distance, COVRATIO, and other
#' influence diagnostics for a fitted brma object.
#'
#' @param model a fitted brma object.
#' @param ... additional arguments (currently ignored).
#'
#' @return An object of class \code{"infl.brma"}, which corresponds to the
#' structure of \code{metafor::influence} objects. It is a list containing:
#' \item{inf}{A data frame with columns: \code{rstudent}, \code{dffits},
#' \code{cook.d}, \code{cov.r}, \code{tau.del}, and \code{hat}, where available.
#' The \code{tau.del} column is the PSIS leave-one-out posterior mean of the
#' aggregate heterogeneity. For scale models, aggregation is over the remaining
#' scale-model rows after each deletion.}
#' \item{dfbs}{A data frame with DFBETAS values for the model coefficients.}
#' Undefined determinant- or variance-standardized diagnostics are reported as
#' \code{NaN} and printed with an explanatory note.
#'
#' @examples \dontrun{
#' if (requireNamespace("metadat", quietly = TRUE)) {
#'   data(dat.lehmann2018, package = "metadat")
#'   fit <- brma(
#'     yi      = yi,
#'     vi      = vi,
#'     data    = dat.lehmann2018,
#'     measure = "SMD",
#'     seed    = 1,
#'     silent  = TRUE
#'   )
#'   fit <- add_loo(fit)
#'
#'   inf <- influence(fit)
#'   print(inf)
#' }
#' }
#'
#' @seealso \code{\link{dffits.brma}}, \code{\link{cooks.distance.brma}},
#' \code{\link{covratio.brma}}, \code{\link{dfbetas.brma}}, \code{\link{hatvalues.brma}}
#' @exportS3Method
influence.brma <- function(model, ...) {

  # object information
  # hatvalues, dffit, and cooks distance are possible only for normal-normal models
  outcome_type       <- .outcome_type(model)
  is_weightfunction  <- .is_weightfunction(model)
  psis_context       <- .diagnostic_psis_context(model)
  loo_wts            <- psis_context[["psis_weights"]]


  ### Precomputed Shared Components
  # rstudent (LOO-PIT residuals)
  rstud_df     <- rstudent.brma(model, .psis_context = psis_context)
  rstudent_val <- rstud_df[["z"]]

  # Hat Matrix Samples (S x K)
  if (outcome_type == "norm" && !is_weightfunction) {
    hat_res <- .compute_hat_matrix_samples(
      object             = model,
      conditioning_depth = "marginal",
      return_full_H      = FALSE,
      return_se          = FALSE
    )
    hat_samples <- hat_res[["H_diag"]]

    # Hat Values (Posterior Means)
    hat_val <- colMeans(hat_samples)

    fit_samples <- .influence_fit_samples(model)
  }

  ### Compute Individual Diagnostics using Shared Components
  # DFFITS
  if (outcome_type == "norm" && !is_weightfunction) {
    dffits_val <- .dffits_internal(fit_samples, loo_wts)
  }

  # Cook's Distance
  # Need P (number of coefficients)
  if (outcome_type == "norm" && !is_weightfunction) {
    X        <- .get_model_matrix(model)
    P        <- qr(X)[["rank"]]
    cook_val <- .cooks.distance_internal(fit_samples, loo_wts, P)
  }

  # COVRATIO
  # covratio uses weights and beta samples, not hat_samples directly.
  cov_val  <- covratio.brma(model, .weights = loo_wts)
  cov_note <- attr(cov_val, "note")
  cov_val <- as.numeric(cov_val)

  # DFBETAS
  # dfbetas uses weights and beta samples.
  dfb_val  <- dfbetas.brma(model, .weights = loo_wts)
  dfb_note <- attr(dfb_val, "note")

  # Compute tau.del (LOO aggregate heterogeneity estimate)
  tau_del_val <- .influence_tau_del(model, weights = loo_wts)

  # Construct 'inf' data frame
  if (outcome_type == "norm" && !is_weightfunction) {
    inf_df <- data.frame(
      rstudent = rstudent_val,
      dffits   = dffits_val,
      cook.d   = cook_val,
      cov.r    = cov_val,
      tau.del  = tau_del_val,
      hat      = hat_val
    )
  } else {
    inf_df <- data.frame(
      rstudent = rstudent_val,
      cov.r    = cov_val,
      tau.del  = tau_del_val
    )
  }
  inf_df <- .diagnostic_set_rownames(inf_df, model)


  ### Return list structure matching metafor
  res <- list(
    inf  = inf_df,
    dfbs = dfb_val
  )
  note <- .diagnostic_collect_notes(cov_note, dfb_note)
  if (!is.null(note)) {
    attr(res, "note") <- note
  }

  class(res) <- "infl.brma"
  return(res)
}


# ---------------------------------------------------------------------------- #
# .influence_tau_del
# ---------------------------------------------------------------------------- #
#
# PSIS leave-one-out aggregate tau estimates for influence diagnostics.
#
.influence_tau_del <- function(model, weights = NULL) {

  weights           <- .diagnostic_psis_weights(model, weights)
  is_scale          <- .is_scale(model)
  is_multilevel     <- .is_multilevel(model)
  K                 <- ncol(weights)
  posterior_samples <- .get_posterior_samples(model[["fit"]])

  tau_result <- .evaluate.brma.tau(
    fit               = model[["fit"]],
    scale_data        = model[["data"]][["scale"]],
    scale_formula     = if (is_scale) .create_fit_formula_list(data = model[["data"]], "scale") else NULL,
    scale_priors      = model[["priors"]][["scale"]],
    is_scale          = is_scale,
    is_multilevel     = is_multilevel,
    K                 = K,
    posterior_samples = posterior_samples
  )
  tau_total <- sqrt(tau_result[["tau_within"]]^2 + tau_result[["tau_between"]]^2)

  return(.influence_tau_del_from_samples(tau_total, weights))
}


# ---------------------------------------------------------------------------- #
# .influence_tau_del_from_samples
# ---------------------------------------------------------------------------- #
#
# @param tau_samples S x K matrix of observation-level total tau samples.
# @param weights S x K matrix of normalized PSIS weights.
#
.influence_tau_del_from_samples <- function(tau_samples, weights) {

  if (!is.matrix(tau_samples)) {
    tau_samples <- as.matrix(tau_samples)
  }
  if (!is.matrix(weights)) {
    weights <- as.matrix(weights)
  }
  if (!identical(dim(tau_samples), dim(weights))) {
    stop("'tau_samples' and 'weights' must have the same dimensions.",
         call. = FALSE)
  }

  K <- ncol(tau_samples)
  if (K > 1L) {
    tau_deleted <- (rowSums(tau_samples) - tau_samples) / (K - 1L)
  } else {
    tau_deleted <- tau_samples
  }

  return(unname(colSums(weights * tau_deleted)))
}

#' @exportS3Method
print.infl.brma <- function(x, digits = 3, ...) {
  cat("Influence diagnostics:\n")
  print(x$inf, digits = digits, ...)
  cat("\nDFBETAS:\n")
  print(x$dfbs, digits = digits, ...)
  .print_diagnostic_note(attr(x, "note"))
  invisible(x)
}

### basic plotting functions ----
#' @title Plots brma Object
#'
#' @description \code{plot.brma} visualizes posterior
#' (and prior) distribution a brma object.
#'
#' @param x a fitted \code{brma}, \code{BMA}, or \code{RoBMA} object.
#' @param parameter a parameter to be plotted. Defaults to \code{"mu"} for
#' the effect size, or to the meta-regression intercept when moderators are
#' present. Additional options are \code{"tau"}, \code{"rho"} for multilevel
#' models, \code{"PET"}, \code{"PEESE"}, and \code{"omega"} or
#' \code{"weightfunction"} for selection models. Use \code{plot_pet_peese()}
#' for PET/PEESE regression plots.
#' @param parameter_mods character. Moderator term to plot. Use
#' \code{"intercept"} for the adjusted effect in meta-regression models.
#' @param parameter_scale character. Scale-regression term to plot. Use
#' \code{"intercept"} for the heterogeneity intercept in location-scale models.
#' @param plot_type whether to use a base plot \code{"base"}
#' or ggplot2 \code{"ggplot"} for plotting. Defaults to
#' \code{"base"}.
#' @param prior whether prior distribution should be added to
#' figure. Defaults to \code{FALSE}.
#' @param standardized_coefficients whether to plot moderator and
#' scale-regression coefficients on the standardized predictor scale. Defaults
#' to \code{FALSE}.
#' @param conditional whether to plot the conditional posterior distribution
#' for RoBMA product-space objects. Defaults to \code{FALSE}.
#' @param dots_prior list of additional graphical arguments
#' to be passed to the plotting function of the prior
#' distribution. Supported arguments are \code{lwd},
#' \code{lty}, \code{col}, and \code{col.fill}, to adjust
#' the line thickness, line type, line color, and fill color
#' of the prior distribution respectively.
#' @inheritParams predict.brma
#' @param ... list of additional graphical arguments
#' to be passed to the plotting function. Supported arguments
#' are \code{lwd}, \code{lty}, \code{col}, \code{col.fill},
#' \code{xlab}, \code{ylab}, \code{main}, \code{xlim}, \code{ylim}
#' to adjust the line thickness, line type, line color, fill color,
#' x-label, y-label, title, x-axis range, and y-axis range
#' respectively.
#'
#' @examples \dontrun{
#' if (requireNamespace("metadat", quietly = TRUE)) {
#'   data(dat.lehmann2018, package = "metadat")
#'   fit <- bPET(yi = yi, vi = vi, data = dat.lehmann2018, measure = "SMD")
#'
#'   plot(fit, parameter = "mu")
#'   plot(fit, parameter = "tau", prior = TRUE)
#'   plot(fit, parameter = "PET")
#' }
#' }
#'
#'
#' @return \code{plot.brma} returns either \code{NULL} if \code{plot_type = "base"}
#' or a \code{ggplot2} object if \code{plot_type = "ggplot"}.
#'
#' @seealso [RoBMA()]
#' @export
plot.brma  <- function(
    x, parameter, parameter_mods, parameter_scale,
    prior = FALSE, standardized_coefficients = FALSE,
    conditional = FALSE,
    output_measure = NULL, transform = NULL,
    plot_type = "base", dots_prior = NULL, ...) {


  ### check user input
  BayesTools::check_char(plot_type, "plot_type", allow_values = c("base", "ggplot"))
  BayesTools::check_bool(prior, "prior")
  BayesTools::check_bool(standardized_coefficients, "standardized_coefficients")
  BayesTools::check_bool(conditional, "conditional")
  if (conditional && !.is_RoBMA(x)) {
    stop("'conditional' plots are available only for RoBMA objects.", call. = FALSE)
  }

  ### select and validate the parameter to be plotted
  parameter <- .check_and_select_plot_parameter(
    parameter       = parameter,
    parameter_mods  = parameter_mods,
    parameter_scale = parameter_scale,
    object          = x
  )
  effect_transform <- .effect_output_setup(
    object         = x,
    output_measure = output_measure,
    transform      = transform
  )
  if (.effect_output_requested(effect_transform) &&
      !.is_effect_location_parameter(parameter)) {
    stop(
      "'output_measure' and 'transform' are only available for effect-size ",
      "location parameters ('mu' or the meta-regression intercept).",
      call. = FALSE
    )
  }

  ### obtain posterior samples in the plotting format
  sample_parameter <- .as_mixed_posteriors_parameters(x, parameter)
  samples <- BayesTools::as_mixed_posteriors(
    model            = x[["fit"]],
    parameters       = sample_parameter,
    conditional      = if (conditional) parameter else NULL,
    transform_scaled = !standardized_coefficients
  )

  ### set up plotting arguments
  n_levels   <- .get_samples_n_levels(samples, parameter)
  dots       <- .set_dots_plot(..., n_levels = n_levels)
  dots_prior <- .set_dots_prior(dots_prior)
  if (is.null(dots[["par_name"]])) {
    dots[["par_name"]] <- .plot_parameter_label(parameter, effect_transform)
  }

  # prepare the argument call
  args                          <- dots
  args$samples                  <- samples
  args$parameter                <- parameter
  args$plot_type                <- plot_type
  args$prior                    <- prior
  args$n_points                 <- 1000
  args$n_samples                <- 10000
  args$force_samples            <- FALSE
  args$dots_prior               <- dots_prior
  args$individual               <- TRUE
  args$show_figures             <- NULL
  if (.effect_output_requested(effect_transform)) {
    args$transformation           <- .effect_plot_transformation(effect_transform)
    args$transformation_arguments <- NULL
    args$transformation_settings  <- FALSE
  }

  # suppress messages about transformations
  plot <- suppressMessages(do.call(BayesTools::plot_posterior, args))

  # return the plots
  if(plot_type == "base"){
    return(invisible(plot))
  }else if(plot_type == "ggplot"){
    return(plot)
  }
}


#' @title Plot Prior Distributions
#'
#' @description \code{plot_prior} visualizes prior distributions stored in
#' \code{brma}, \code{BMA}, and \code{RoBMA} objects.
#' This is especially useful for objects created with \code{only_priors = TRUE}.
#'
#' @param x a \code{brma}, \code{BMA}, or \code{RoBMA} object, or a prior
#' distribution object.
#' @param parameter character. Base parameter to plot. Defaults to \code{"mu"}.
#' Common options are \code{"mu"}, \code{"tau"}, \code{"rho"}, \code{"PET"},
#' \code{"PEESE"}, \code{"omega"}, and \code{"bias"}, with aliases
#' \code{"effect"} = \code{"mu"}, \code{"heterogeneity"} = \code{"tau"},
#' and \code{"weightfunction"} = \code{"omega"}. \code{"bias"} plots only
#' non-mixed or homogeneous bias priors; for mixed weightfunction and PET/PEESE
#' mixtures use \code{"omega"}, \code{"PET"}, or \code{"PEESE"}. Moderator and
#' scale terms can also be selected by name when unambiguous. A character vector
#' requests multiple base parameters.
#' @param parameter_mods character. Moderator term to plot.
#' Use \code{"intercept"} for the adjusted effect in meta-regression models.
#' @param parameter_scale character. Scale-regression term to plot.
#' Use \code{"intercept"} for the heterogeneity intercept in location-scale models.
#' @param standardized_coefficients whether to plot moderator and scale-regression
#' priors on the standardized predictor scale. Defaults to \code{TRUE}, which
#' shows the priors as specified. Set to \code{FALSE} to transform them to the
#' original predictor scale when continuous predictors were standardized.
#' @inheritParams predict.brma
#' @param plot_type whether to use a base plot \code{"base"} or ggplot2
#' \code{"ggplot"} for plotting. Defaults to \code{"base"}.
#' @param ... additional arguments passed to the prior plotting method.
#'
#' @details \code{output_measure} and \code{transform} transform the prior
#' plotting scale only for effect-size location priors (\code{"mu"} or the
#' meta-regression intercept).
#'
#' @return \code{plot_prior} returns either \code{NULL} invisibly if
#' \code{plot_type = "base"} or a \code{ggplot2} object if
#' \code{plot_type = "ggplot"}. If multiple parameters are requested, a named
#' list is returned, invisibly for base plots.
#'
#' @examples \dontrun{
#' if (requireNamespace("metadat", quietly = TRUE)) {
#'   data(dat.lehmann2018, package = "metadat")
#'
#'   priors <- BMA(
#'     yi           = yi,
#'     vi           = vi,
#'     mods         = ~ Preregistered,
#'     data         = dat.lehmann2018,
#'     measure      = "SMD",
#'     only_priors  = TRUE
#'   )
#'
#'   plot_prior(priors, parameter = "mu")
#'   plot_prior(priors, parameter = "tau")
#'   plot_prior(priors, parameter_mods = "Preregistered")
#' }
#' }
#'
#' @seealso [BMA()] [RoBMA()] [brma()] [prior()]
#' @export
plot_prior <- function(x, ...) UseMethod("plot_prior")

#' @export
#' @rdname plot_prior
plot_prior.prior <- function(x, plot_type = "base", ...) {

  BayesTools::check_char(plot_type, "plot_type", allow_values = c("base", "ggplot"))

  plot <- plot(x, plot_type = plot_type, ...)

  if (plot_type == "base") {
    return(invisible(plot))
  } else if (plot_type == "ggplot") {
    return(plot)
  }
}

#' @export
#' @rdname plot_prior
plot_prior.brma <- function(
    x, parameter = "mu", parameter_mods, parameter_scale,
    standardized_coefficients = TRUE,
    output_measure = NULL, transform = NULL,
    plot_type = "base", ...) {

  BayesTools::check_char(plot_type, "plot_type", allow_values = c("base", "ggplot"))
  BayesTools::check_bool(standardized_coefficients, "standardized_coefficients")

  if (is.null(x[["priors"]])) {
    stop("The object does not contain prior distributions.", call. = FALSE)
  }

  has_parameter       <- !missing(parameter)       && !is.null(parameter)
  has_parameter_mods  <- !missing(parameter_mods)  && !is.null(parameter_mods)
  has_parameter_scale <- !missing(parameter_scale) && !is.null(parameter_scale)

  n_specified <- sum(c(has_parameter, has_parameter_mods, has_parameter_scale))
  if (n_specified > 1) {
    stop("Only one of 'parameter', 'parameter_mods', or 'parameter_scale' can be specified.", call. = FALSE)
  }

  if (has_parameter && length(parameter) > 1) {
    if (!is.character(parameter) || anyNA(parameter) || any(!nzchar(parameter))) {
      stop("The 'parameter' argument must be a character vector without missing or empty values.", call. = FALSE)
    }

    plots <- lapply(parameter, function(parameter_i) {
      plot_prior(
        x                         = x,
        parameter                 = parameter_i,
        standardized_coefficients = standardized_coefficients,
        output_measure            = output_measure,
        transform                 = transform,
        plot_type                 = plot_type,
        ...
      )
    })
    names(plots) <- parameter

    if (plot_type == "base") {
      return(invisible(plots))
    } else if (plot_type == "ggplot") {
      return(plots)
    }
  }

  selected <- .select_plot_prior_parameter(
    object          = x,
    parameter       = if (has_parameter) parameter else NULL,
    parameter_mods  = if (has_parameter_mods) parameter_mods else NULL,
    parameter_scale = if (has_parameter_scale) parameter_scale else NULL
  )
  effect_transform <- .effect_output_setup(
    object         = x,
    output_measure = output_measure,
    transform      = transform
  )
  if (.effect_output_requested(effect_transform) &&
      !.is_effect_location_parameter(selected[["label"]])) {
    stop(
      "'output_measure' and 'transform' are only available for effect-size ",
      "location priors ('mu' or the meta-regression intercept).",
      call. = FALSE
    )
  }

  dots <- list(...)
  if (is.null(dots[["par_name"]])) {
    dots[["par_name"]] <- selected[["label"]]
  }
  if (.effect_output_requested(effect_transform)) {
    dots[["transformation"]]           <- .effect_plot_transformation(effect_transform)
    dots[["transformation_arguments"]] <- NULL
    dots[["transformation_settings"]]  <- FALSE
    if (.effect_output_active(effect_transform)) {
      dots[["par_name"]] <- effect_transform[["label"]]
    }
  }

  if (!standardized_coefficients) {
    prior_plot <- .plot_prior_unstandardized(
      object    = x,
      selected  = selected,
      plot_type = plot_type,
      dots      = dots
    )
    if (!is.null(prior_plot)) {
      if (plot_type == "base") {
        return(invisible(prior_plot))
      } else if (plot_type == "ggplot") {
        return(prior_plot)
      }
    }
  }

  if (BayesTools::is.prior.mixture(selected[["prior"]])) {
    plot <- do.call(
      BayesTools::plot_prior_list,
      c(
        list(
          prior_list = .prior_mixture_plot_list(selected[["prior"]]),
          plot_type  = plot_type
        ),
        dots
      )
    )
  } else if (.use_plot_prior_list_dispatch(selected[["prior"]])) {
    plot <- do.call(
      BayesTools::plot_prior_list,
      c(
        list(prior_list = list(selected[["prior"]]), plot_type = plot_type),
        dots
      )
    )
  } else {
    plot <- do.call(
      "plot",
      c(
        list(x = selected[["prior"]], plot_type = plot_type),
        dots
      )
    )
  }

  if (plot_type == "base") {
    return(invisible(plot))
  } else if (plot_type == "ggplot") {
    return(plot)
  }
}

#' @export
plot.only_priors.brma <- function(x, ...) {

  plot_prior.brma(x = x, ...)
}


#' @title Print Prior Distributions
#'
#' @description \code{print_prior} prints prior distributions stored in
#' \code{brma}, \code{BMA}, and \code{RoBMA} objects.
#' This is a user-facing helper for inspecting priors without extracting them
#' from the internal \code{$priors} list.
#'
#' @param x a \code{brma}, \code{BMA}, or \code{RoBMA} object, or a prior
#' distribution object.
#' @param parameter character. Base parameter to print. If omitted, all stored
#' prior distributions are printed.
#' Common options are \code{"mu"}, \code{"tau"}, \code{"rho"}, \code{"PET"},
#' \code{"PEESE"}, \code{"omega"}, and \code{"bias"}, with aliases
#' \code{"effect"} = \code{"mu"}, \code{"heterogeneity"} = \code{"tau"},
#' and \code{"weightfunction"} = \code{"omega"}. GLMM outcome priors
#' \code{"pi"} and \code{"phi"} are available when present. Moderator and
#' scale terms can also be selected by name when unambiguous. A character vector
#' requests multiple base parameters.
#' @param parameter_mods character. Moderator term to print.
#' Use \code{"intercept"} for the adjusted effect in meta-regression models.
#' @param parameter_scale character. Scale-regression term to print.
#' Use \code{"intercept"} for the heterogeneity intercept in location-scale models.
#' @param ... additional arguments passed to the prior printing method. Use
#' \code{silent = TRUE} for programmatic inspection without console output.
#'
#' @return \code{print_prior} invisibly returns the selected prior distribution.
#' If multiple parameters are requested, a named list of prior distributions is
#' returned invisibly.
#'
#' @examples \dontrun{
#' if (requireNamespace("metadat", quietly = TRUE)) {
#'   data(dat.lehmann2018, package = "metadat")
#'   priors <- BMA(
#'     yi          = yi,
#'     vi          = vi,
#'     mods        = ~ Preregistered,
#'     data        = dat.lehmann2018,
#'     measure     = "SMD",
#'     only_priors = TRUE
#'   )
#'
#'   print_prior(priors)
#'   print_prior(priors, parameter = "mu")
#'   print_prior(priors, parameter = "tau")
#'   print_prior(priors, parameter_mods = "Preregistered")
#' }
#' }
#'
#' @seealso [plot_prior()] [BMA()] [RoBMA()] [brma()] [prior()]
#' @export
print_prior <- function(x, ...) UseMethod("print_prior")

#' @export
#' @rdname print_prior
print_prior.prior <- function(x, ...) {

  .print_prior_object(x, ...)

  return(invisible(x))
}

#' @export
#' @rdname print_prior
print_prior.brma <- function(
    x, parameter, parameter_mods, parameter_scale, ...) {

  if (is.null(x[["priors"]])) {
    stop("The object does not contain prior distributions.", call. = FALSE)
  }

  parameter_missing   <- missing(parameter)
  has_parameter       <- !parameter_missing        && !is.null(parameter)
  has_parameter_mods  <- !missing(parameter_mods)  && !is.null(parameter_mods)
  has_parameter_scale <- !missing(parameter_scale) && !is.null(parameter_scale)

  n_specified <- sum(c(has_parameter, has_parameter_mods, has_parameter_scale))
  if (n_specified > 1) {
    stop("Only one of 'parameter', 'parameter_mods', or 'parameter_scale' can be specified.", call. = FALSE)
  }

  if (parameter_missing && !has_parameter_mods && !has_parameter_scale) {
    selected <- .select_print_prior_all(x)
    priors   <- lapply(selected, `[[`, "prior")
    names(priors) <- vapply(selected, `[[`, character(1), "label")

    for (i in seq_along(selected)) {
      .print_prior_object(priors[[i]], label = names(priors)[i], ...)
    }

    return(invisible(priors))
  }

  if (has_parameter && length(parameter) > 1) {
    if (!is.character(parameter) || anyNA(parameter) || any(!nzchar(parameter))) {
      stop("The 'parameter' argument must be a character vector without missing or empty values.", call. = FALSE)
    }

    selected <- lapply(parameter, function(parameter_i) {
      .select_plot_prior_parameter(
        object           = x,
        parameter        = parameter_i,
        parameter_mods   = NULL,
        parameter_scale  = NULL,
        allow_mixed_bias = TRUE
      )
    })
    names(selected) <- parameter

    priors <- lapply(selected, `[[`, "prior")
    for (i in seq_along(selected)) {
      .print_prior_object(priors[[i]], label = selected[[i]][["label"]], ...)
    }

    return(invisible(priors))
  }

  selected <- .select_plot_prior_parameter(
    object           = x,
    parameter        = if (has_parameter) parameter else NULL,
    parameter_mods   = if (has_parameter_mods) parameter_mods else NULL,
    parameter_scale  = if (has_parameter_scale) parameter_scale else NULL,
    allow_mixed_bias = TRUE
  )

  .print_prior_object(selected[["prior"]], label = selected[["label"]], ...)

  return(invisible(selected[["prior"]]))
}

#' @export
print.only_priors.brma <- function(x, ...) {

  print_prior.brma(x = x, ...)
}


#' @title Plots Weight Function of brma Object
#'
#' @description \code{plot_weightfunction.brma} visualizes the posterior
#' (and optionally prior) publication-bias weight function of a brma object.
#'
#' @param x a fitted \code{brma} object with a weightfunction/selection
#' component, such as \code{bselmodel()} or a \code{RoBMA()} object with
#' weightfunction priors.
#' @param plot_type whether to use a base plot \code{"base"}
#' or ggplot2 \code{"ggplot"} for plotting. Defaults to
#' \code{"base"}.
#' @param prior whether prior distribution should be added to
#' figure. Defaults to \code{FALSE}.
#' @param rescale_p_values whether to rescale p-values to the interval shown
#' by the weightfunction plot. Defaults to \code{TRUE}.
#' @param show_data whether observed one-sided p-values should be shown as rug
#' marks on the weightfunction axis. Defaults to \code{TRUE}.
#' @param dots_data list of additional graphical arguments for observed
#' p-value rug marks. Supported arguments include \code{col}/\code{color},
#' \code{alpha}, \code{lwd}/\code{linewidth}/\code{size},
#' \code{side}/\code{rug_side}, and
#' \code{height}/\code{rug_height}/\code{ticksize}.
#' @param dots_prior list of additional graphical arguments
#' to be passed to the plotting function of the prior
#' distribution. Supported arguments are \code{lwd},
#' \code{lty}, \code{col}, and \code{col.fill}, to adjust
#' the line thickness, line type, line color, and fill color
#' of the prior distribution respectively.
#' @param ... list of additional graphical arguments
#' to be passed to the plotting function. Supported arguments
#' are \code{lwd}, \code{lty}, \code{col}, \code{col.fill},
#' \code{xlab}, \code{ylab}, \code{main}, \code{xlim}, \code{ylim}
#' to adjust the line thickness, line type, line color, fill color,
#' x-label, y-label, title, x-axis range, and y-axis range
#' respectively.
#'
#' @examples \dontrun{
#' if (requireNamespace("metadat", quietly = TRUE)) {
#'   data(dat.lehmann2018, package = "metadat")
#'   fit <- bselmodel(
#'     yi      = yi,
#'     vi      = vi,
#'     data    = dat.lehmann2018,
#'     measure = "SMD",
#'     seed    = 1,
#'     silent  = TRUE
#'   )
#'
#'   plot_weightfunction(fit)
#'   plot_weightfunction(fit, prior = TRUE)
#' }
#' }
#'
#'
#' @return \code{plot_weightfunction.brma} returns either \code{NULL} if
#' \code{plot_type = "base"} or a \code{ggplot2} object if
#' \code{plot_type = "ggplot"}.
#' The method errors for fitted objects without a weightfunction component.
#'
#' @seealso [RoBMA()]
#' @export
plot_weightfunction <- function(x, ...)  UseMethod("plot_weightfunction")

#' @export
#' @rdname plot_weightfunction
plot_weightfunction.brma  <- function(
    x, rescale_p_values = TRUE,
    prior = FALSE, plot_type = "base", show_data = TRUE,
    dots_data = NULL, dots_prior = NULL, ...) {

  ### check user input
  BayesTools::check_char(plot_type, "plot_type", allow_values = c("base", "ggplot"))
  BayesTools::check_bool(prior, "prior")
  BayesTools::check_bool(rescale_p_values, "rescale_p_values")
  BayesTools::check_bool(show_data, "show_data")
  BayesTools::check_list(dots_data, "dots_data", allow_NULL = TRUE)
  dots_raw <- list(...)
  if (is.null(dots_data)) {
    dots_data <- list()
  }

  if (!.is_weightfunction(x)) {
    stop("'plot_weightfunction' is available only for models with a weightfunction component.", call. = FALSE)
  }

  ### obtain posterior samples in the plotting format
  samples <- BayesTools::as_mixed_posteriors(
    model      = x[["fit"]],
    parameters = .as_mixed_posteriors_parameters(x, "omega")
  )

  ### set up plotting arguments
  dots       <- do.call(.set_dots_plot, dots_raw)
  dots_prior <- .set_dots_prior(dots_prior)
  data_p     <- if (show_data) .weightfunction_observed_p_values(x) else NULL

  # prepare the argument call
  args                          <- dots
  args$samples                  <- samples
  args$parameter                <- "weightfunction"
  args$plot_type                <- plot_type
  args$prior                    <- prior
  args$n_points                 <- 1000
  args$n_samples                <- 10000
  args$force_samples            <- FALSE
  args$dots_prior               <- dots_prior
  args$individual               <- FALSE
  args$rescale_x                <- rescale_p_values
  args$show_data                <- show_data
  args$data                     <- data_p
  args$dots_data                <- dots_data

  # suppress messages about transformations
  plot <- suppressMessages(do.call(BayesTools::plot_posterior, args))

  # return the plots
  if(plot_type == "base"){
    return(invisible(plot))
  }else if(plot_type == "ggplot"){
    return(plot)
  }
}

.weightfunction_observed_p_values <- function(x) {

  outcome_data <- x[["data"]][["outcome"]]
  selection    <- .selection_spec(
    priors           = x[["priors"]],
    yi               = outcome_data[["yi"]],
    sei              = outcome_data[["sei"]],
    effect_direction = .effect_direction(x),
    signed_data      = FALSE
  )

  if (is.null(selection)) {
    return(NULL)
  }

  return(stats::pnorm(
    selection[["sign"]] * outcome_data[["yi"]] / outcome_data[["sei"]],
    lower.tail = FALSE
  ))
}


#' @title Plot PET-PEESE Fit of brma Object
#'
#' @description \code{plot_pet_peese} visualizes posterior
#' (and prior) PET-PEESE fit of a brma object.
#'
#' @param x a fitted brma object with a PET or PEESE component.
#' @param show_data whether the observed effect sizes should be added to
#' figure. Defaults to \code{TRUE}.
#' @param plot_type whether to use a base plot \code{"base"}
#' or ggplot2 \code{"ggplot"} for plotting. Defaults to
#' \code{"base"}.
#' @param prior whether prior distribution should be added to
#' figure. Defaults to \code{FALSE}.
#' @param dots_prior list of additional graphical arguments
#' to be passed to the plotting function of the prior
#' distribution. Supported arguments are \code{lwd},
#' \code{lty}, \code{col}, and \code{col.fill}, to adjust
#' the line thickness, line type, line color, and fill color
#' of the prior distribution respectively.
#' @param ... list of additional graphical arguments
#' to be passed to the plotting function. Supported arguments
#' are \code{lwd}, \code{lty}, \code{col}, \code{col.fill},
#' \code{xlab}, \code{ylab}, \code{main}, \code{xlim}, \code{ylim},
#' \code{pch}, \code{bg}, \code{cex}, and \code{size}
#' to adjust the line thickness, line type, line color, fill color,
#' x-label, y-label, title, x-axis range, y-axis range, and observed data
#' point style respectively.
#'
#' @details The plot shows observed \code{yi} values against \code{sei}. PET
#' regression uses \eqn{\mu + PET \cdot se_i}; PEESE regression uses
#' \eqn{\mu + PEESE \cdot se_i^2}, with the fitted effect direction.
#'
#' @examples \dontrun{
#' if (requireNamespace("metadat", quietly = TRUE)) {
#'   data(dat.lehmann2018, package = "metadat")
#'
#'   fit <- bPET(
#'     yi      = yi,
#'     vi      = vi,
#'     data    = dat.lehmann2018,
#'     measure = "SMD",
#'     seed    = 1,
#'     silent  = TRUE
#'   )
#'
#'   plot_pet_peese(fit)
#' }
#' }
#'
#'
#' @return \code{plot_pet_peese} returns either \code{NULL} invisibly if
#' \code{plot_type = "base"} or a ggplot2 object if
#' \code{plot_type = "ggplot"}.
#'
#' @seealso [RoBMA()]
#' @export
plot_pet_peese <- function(x, ...)  UseMethod("plot_pet_peese")

#' @export
#' @rdname plot_pet_peese
plot_pet_peese.brma  <- function(
    x, show_data = TRUE,
    prior = FALSE, plot_type = "base", dots_prior = NULL, ...) {


  ### check user input
  BayesTools::check_char(plot_type, "plot_type", allow_values = c("base", "ggplot"))
  BayesTools::check_bool(prior, "prior")
  BayesTools::check_bool(show_data, "show_data")

  if (!(.is_PET(x) || .is_PEESE(x))) {
    stop("'plot_pet_peese' is available only for models with a PET or PEESE component.", call. = FALSE)
  }

  ### obtain posterior samples in the plotting format
  location_parameter <- .check_and_select_plot_parameter(
    parameter       = "mu",
    parameter_mods  = NULL,
    parameter_scale = NULL,
    object          = x
  )
  samples <- BayesTools::as_mixed_posteriors(
    model      = x[["fit"]],
    parameters = .as_mixed_posteriors_parameters(
      x,
      c(location_parameter, if (.is_PET(x)) "PET", if (.is_PEESE(x)) "PEESE")
    )
  )

  ### set up plotting arguments
  dots       <- .set_dots_plot(...)
  dots_prior <- .set_dots_prior(dots_prior)

  # set plotting range (make sure all effect sizes are within the plotting range)
  data_outcome <- x[["data"]][["outcome"]]
  if (is.null(dots[["ylim"]])) {
    dots[["ylim"]] <- range(pretty(c(0, range(data_outcome$yi))))
  }

  # prepare the argument call
  args                          <- dots
  args$samples                  <- samples
  args$parameter                <- "PETPEESE"
  args$plot_type                <- plot_type
  args$prior                    <- prior
  args$n_points                 <- 1000
  args$n_samples                <- 10000
  args$force_samples            <- FALSE
  args$dots_prior               <- dots_prior
  args$individual               <- FALSE
  args$effect_direction         <- .effect_direction(x)

  # suppress messages about transformations
  plot <- suppressMessages(do.call(BayesTools::plot_posterior, args))

  # include data
  if (show_data) {
    data_dots <- .plot_point_style_defaults(dots)

    if (plot_type == "ggplot") {
      plot <- plot + ggplot2::geom_point(
        ggplot2::aes(
          x  = data_outcome$sei,
          y  = data_outcome$yi),
        size   = data_dots[["size"]],
        shape  = data_dots[["pch"]],
        colour = data_dots[["col"]],
        fill   = data_dots[["bg"]]
      )
    } else if (plot_type == "base") {
      graphics::points(
        data_outcome$sei, data_outcome$yi,
        cex = data_dots[["cex"]],
        pch = data_dots[["pch"]],
        col = data_dots[["col"]],
        bg  = data_dots[["bg"]]
      )
    }
  }

  # return the plots
  if (plot_type == "base") {
    return(invisible(plot))
  } else if (plot_type == "ggplot") {
    return(plot)
  }
}



### basic MCMC diagnostics functions ----

#' @title Plot MCMC Diagnostics
#'
#' @description \code{plot_diagnostic} creates visual MCMC diagnostics for a
#' fitted brma object. Convenience wrappers are available for trace, density,
#' and autocorrelation plots.
#'
#' @param x a fitted brma object
#' @param parameter base parameter to plot. Defaults to \code{NULL}, which uses
#'   \code{"mu"} or the meta-regression intercept. Valid values include
#'   \code{"mu"}, \code{"tau"}, \code{"rho"}, \code{"PET"}, \code{"PEESE"},
#'   and \code{"omega"} or \code{"weightfunction"} when present.
#' @param parameter_mods moderator term for location regression.
#' @param parameter_scale term for scale regression.
#' @param type diagnostic plot type. Convenience wrappers set a type-specific
#'   default but still forward this argument to \code{plot_diagnostic.brma()}.
#' @param plot_type whether to use a base plot \code{"base"} or ggplot2
#'   \code{"ggplot"} for plotting. Defaults to \code{"base"}.
#' @param lags number of lags for autocorrelation plots. Defaults to 30.
#' @param ... additional graphical arguments passed through RoBMA's diagnostic
#'   setup to \code{BayesTools::JAGS_diagnostics()}.
#'
#' @return \code{plot_diagnostic} returns the object returned by
#'   \code{BayesTools::JAGS_diagnostics()}, invisibly for base graphics.
#'
#' @seealso [summary.brma()]
#'
#' @export
plot_diagnostic <- function(x, ...) UseMethod("plot_diagnostic")

#' @export
#' @rdname plot_diagnostic
plot_diagnostic.brma <- function(
    x, parameter = NULL, parameter_mods = NULL, parameter_scale = NULL,
    type, plot_type = "base", lags = 30, ...) {

  ### check user input
  BayesTools::check_char(plot_type, "plot_type", allow_values = c("base", "ggplot"))
  BayesTools::check_char(type, "type", allow_values = c("autocorrelation", "trace", "density"))

  ### select and validate the parameter to be plotted
  parameter <- .check_and_select_plot_parameter(
    parameter       = parameter,
    parameter_mods  = parameter_mods,
    parameter_scale = parameter_scale,
    object          = x
  )

  # a message with info about multiple plots
  if(plot_type == "base" && parameter == "omega") {
    message("Multiple plots will be produced. See '?layout' for help with setting multiple plots.")
  }

  dots  <- .set_dots_diagnostics(..., type = type, chains = x[["fit_control"]][["chains"]])

  # get the parameter name
  args                   <- dots
  args$fit               <- x[["fit"]]
  args$parameter         <- parameter
  args$type              <- type
  args$plot_type         <- plot_type
  args$lags              <- lags
  args$transform_factors <- TRUE
  args$short_name        <- FALSE
  args$parameter_names   <- FALSE
  args$formula_prefix    <- FALSE

  plots <- do.call(BayesTools::JAGS_diagnostics, args)

  # return the plots
  if(plot_type == "base"){
    return(invisible(plots))
  }else if(plot_type == "ggplot"){
    return(plots)
  }
}

#' @export
#' @rdname plot_diagnostic
plot_diagnostic_autocorrelation <- function(x, ...) UseMethod("plot_diagnostic_autocorrelation")

#' @export
#' @rdname plot_diagnostic
plot_diagnostic_autocorrelation.brma <- function(
    x, parameter = NULL, parameter_mods = NULL, parameter_scale = NULL,
    type = "autocorrelation", plot_type = "base", lags = 30, ...) {

  plot_diagnostic(
    x               = x,
    parameter       = parameter,
    parameter_mods  = parameter_mods,
    parameter_scale = parameter_scale,
    type            = type,
    plot_type       = plot_type,
    lags            = lags,
    ...
  )
}

#' @export
#' @rdname plot_diagnostic
plot_diagnostic_trace <- function(x, ...) UseMethod("plot_diagnostic_trace")

#' @export
#' @rdname plot_diagnostic
plot_diagnostic_trace.brma           <- function(
    x, parameter = NULL, parameter_mods = NULL, parameter_scale = NULL,
    type = "trace", plot_type = "base", lags = 30, ...) {

  plot_diagnostic(
    x               = x,
    parameter       = parameter,
    parameter_mods  = parameter_mods,
    parameter_scale = parameter_scale,
    type            = type,
    plot_type       = plot_type,
    lags            = lags,
    ...
  )
}

#' @export
#' @rdname plot_diagnostic
plot_diagnostic_density <- function(x, ...) UseMethod("plot_diagnostic_density")

#' @export
#' @rdname plot_diagnostic
plot_diagnostic_density.brma         <- function(
    x, parameter = NULL, parameter_mods = NULL, parameter_scale = NULL,
    type = "density", plot_type = "base", lags = 30, ...) {

  plot_diagnostic(
    x               = x,
    parameter       = parameter,
    parameter_mods  = parameter_mods,
    parameter_scale = parameter_scale,
    type            = type,
    plot_type       = plot_type,
    lags            = lags,
    ...
  )
}


### Common plotting helper functions ----
# Internal helper function for selecting and validating a plot parameter
#
# Processes parameter selection for plotting based on three mutually exclusive arguments:
# parameter, parameter_mods, parameter_scale. Only one should be specified.
#
# @param parameter Base parameter name ("mu", "tau", "rho")
# @param parameter_mods Moderator parameter for location regression
# @param parameter_scale Moderator parameter for scale regression
# @param object A brma object
#
# @return A character string with the validated parameter name (with appropriate prefix).
#   Attributes:
#   - parameter_samples: The JAGS samples name
#   - type: One of "base", "mods", "scale" indicating the parameter source
.check_and_select_plot_parameter <- function(parameter, parameter_mods, parameter_scale, object) {

  # Check for model characteristics
  is_mods           <- .is_mods(object)
  is_scale          <- .is_scale(object)
  is_multilevel     <- .is_multilevel(object)
  is_PET            <- .is_PET(object)
  is_PEESE          <- .is_PEESE(object)
  is_weightfunction <- .is_weightfunction(object)

  # Determine which arguments were provided
  has_parameter       <- !missing(parameter)       && !is.null(parameter)
  has_parameter_mods  <- !missing(parameter_mods)  && !is.null(parameter_mods)
  has_parameter_scale <- !missing(parameter_scale) && !is.null(parameter_scale)

  # Count how many were specified

  n_specified <- sum(c(has_parameter, has_parameter_mods, has_parameter_scale))

  # Error if more than one specified
  if (n_specified > 1) {
    stop("Only one of 'parameter', 'parameter_mods', or 'parameter_scale' can be specified.", call. = FALSE)
  }

  # Default behavior if none specified
  if (n_specified == 0) {
    if (is_mods) {
      # Default to intercept for meta-regression models
      return("mu_intercept")
    } else {
      # Default to mu for simple models
      return("mu")
    }
  }

  # Process based on which argument was specified
  if (has_parameter) {

    # Validate base parameters
    BayesTools::check_char(parameter, "parameter")

    # Handle mu parameter
    if (parameter == "mu") {
      if (is_mods) {
        # mu becomes mu_intercept when mods are present
        return("mu_intercept")
      } else {
        return("mu")
      }
    }

    # Handle tau parameter
    if (parameter == "tau") {
      if (is_scale) {
        # tau becomes log_tau_intercept when scale is present
        return("log_tau_intercept")
      } else {
        return("tau")
      }
    }

    # Handle rho parameter (only for multilevel models)
    if (parameter == "rho") {
      if (!is_multilevel) {
        stop("The 'rho' parameter is only available for multilevel models.", call. = FALSE)
      }
      return("rho")
    }

    ### Handle publication bias parameters
    if (parameter == "PET") {
      if (!is_PET) {
        stop("The 'PET' parameter is only available for PET models.", call. = FALSE)
      }
      return("PET")
    }

    if (parameter == "PEESE") {
      if (!is_PEESE) {
        stop("The 'PEESE' parameter is only available for PEESE models.", call. = FALSE)
      }
      return("PEESE")
    }

    if (parameter == "omega" || parameter == "weightfunction") {
      if (!is_weightfunction) {
        stop("The 'omega'/'weightfunction' parameter is only available for selection models.", call. = FALSE)
      }
      return("omega")
    }

    # Unknown base parameter
    # Build list of valid parameters dynamically
    valid_params <- c("'mu'", "'tau'")
    if (is_multilevel)     valid_params <- c(valid_params, "'rho'")
    if (is_PET)            valid_params <- c(valid_params, "'PET'")
    if (is_PEESE)          valid_params <- c(valid_params, "'PEESE'")
    if (is_weightfunction) valid_params <- c(valid_params, "'omega'/'weightfunction'")

    stop(sprintf(
      "Unknown parameter '%s'. Valid base parameters are: %s.",
      parameter,
      paste(valid_params, collapse = ", ")
    ), call. = FALSE)

  } else if (has_parameter_mods) {

    # Validate that mods are present
    if (!is_mods) {
      stop("The 'parameter_mods' argument can only be used for models with moderators.", call. = FALSE)
    }

    BayesTools::check_char(parameter_mods, "parameter_mods")

    available_terms <- .fitted_formula_terms(
      object            = object,
      parameter         = "mu",
      include_intercept = TRUE,
      display           = TRUE
    )

    # Validate the specified term exists
    if (!parameter_mods %in% available_terms) {
      stop(sprintf(
        "The specified 'parameter_mods' term '%s' is not available. Available terms are: %s.",
        parameter_mods,
        paste0("'", available_terms, "'", collapse = ", ")
      ), call. = FALSE)
    }

    return(BayesTools::JAGS_parameter_names(
      parameters        = parameter_mods,
      formula_parameter = "mu"
    ))

  } else if (has_parameter_scale) {

    # Validate that scale is present
    if (!is_scale) {
      stop("The 'parameter_scale' argument can only be used for location-scale models.", call. = FALSE)
    }

    BayesTools::check_char(parameter_scale, "parameter_scale")

    available_terms <- .fitted_formula_terms(
      object            = object,
      parameter         = "log_tau",
      include_intercept = TRUE,
      display           = TRUE
    )

    # Validate the specified term exists
    if (!parameter_scale %in% available_terms) {
      stop(sprintf(
        "The specified 'parameter_scale' term '%s' is not available. Available terms are: %s.",
        parameter_scale,
        paste0("'", available_terms, "'", collapse = ", ")
      ), call. = FALSE)
    }

    return(BayesTools::JAGS_parameter_names(
      parameters        = parameter_scale,
      formula_parameter = "log_tau"
    ))
  }
}

.set_dots_plot        <- function(..., n_levels = 1) {

  dots <- list(...)
  if (is.null(dots[["col"]]) & n_levels == 1) {
    dots[["col"]]      <- "black"
  }else if (is.null(dots[["col"]]) & n_levels > 1) {
    dots[["col"]]      <- .plot_level_palette(n_levels)
  }
  if (is.null(dots[["col.fill"]])) {
    dots[["col.fill"]] <- "#4D4D4D4C" # scales::alpha("grey30", .30)
  }

  return(dots)
}

.plot_level_palette   <- function(n_levels) {

  if (n_levels <= 0L) {
    return(character())
  }
  if (n_levels == 1L) {
    return("black")
  }

  okabe_levels <- min(n_levels + 1L, 9L)
  colors       <- grDevices::palette.colors(
    n       = okabe_levels,
    palette = "Okabe-Ito"
  )[-1L]

  if (length(colors) < n_levels) {
    colors <- c(
      colors,
      grDevices::hcl.colors(
        n       = n_levels - length(colors),
        palette = "Dark 3"
      )
    )
  }

  return(colors[seq_len(n_levels)])
}
.set_dots_prior       <- function(dots_prior) {

  if (is.null(dots_prior)) {
    dots_prior <- list()
  }

  if (is.null(dots_prior[["col"]])) {
    dots_prior[["col"]]      <- "grey60"
  }
  if (is.null(dots_prior[["lty"]])) {
    dots_prior[["lty"]]      <- 1
  }
  if (is.null(dots_prior[["col.fill"]])) {
    dots_prior[["col.fill"]] <- "#B3B3B34C" # scales::alpha("grey70", .30)
  }

  return(dots_prior)
}
.set_dots_diagnostics <- function(..., type, chains) {

  dots <- list(...)
  if (is.null(dots[["col"]])) {
    dots[["col"]] <- if(type == "autocorrelation") "black" else rev(scales::viridis_pal()(chains))
  }

  return(dots)
}
.get_samples_n_levels <- function(samples, parameter) {

  if (inherits(samples[[parameter]], "mixed_posteriors.factor")) {
    if (attr(samples[[parameter]], "orthonormal") || attr(samples[[parameter]], "meandif")) {
      n_levels <- length(attr(samples[[parameter]], "level_names"))
    } else if (attr(samples[[parameter]], "treatment")) {
      n_levels <- length(attr(samples[[parameter]], "level_names")) - 1
    } else {
      n_levels <- 1
    }
  } else {
    n_levels <- 1
  }

  return(n_levels)
}

.as_mixed_posteriors_parameters <- function(object, parameters) {

  fit_priors <- attr(object[["fit"]], "prior_list")

  if (!is.null(fit_priors[["bias"]])) {
    parameters[parameters %in% c("omega", "PET", "PEESE")] <- "bias"
    parameters <- unique(parameters)
  }

  return(parameters)
}

.plot_parameter_label <- function(parameter, effect_transform = NULL) {

  if (.is_effect_location_parameter(parameter)) {
    if (.effect_output_active(effect_transform)) {
      return(paste0("Effect Size (", effect_transform[["label"]], ")"))
    }

    return("Effect Size")
  }

  if (grepl("^mu_", parameter)) {
    return(paste0("Effect Size: ", .summary_parameter_label(sub("^mu_", "", parameter))))
  }

  if (parameter %in% c("tau", "log_tau_intercept")) {
    return("Heterogeneity")
  }

  if (grepl("^log_tau_", parameter)) {
    return(paste0("Heterogeneity: ", .summary_parameter_label(sub("^log_tau_", "", parameter))))
  }

  label <- switch(
    parameter,
    "rho"   = "Heterogeneity Allocation",
    "PET"   = "PET",
    "PEESE" = "PEESE",
    "omega" = "Publication Bias",
    "bias"  = "Publication Bias",
    parameter
  )

  return(label)
}

.select_plot_prior_parameter <- function(
    object, parameter, parameter_mods, parameter_scale, allow_mixed_bias = FALSE) {

  n_specified <- sum(!vapply(
    list(parameter, parameter_mods, parameter_scale),
    is.null,
    logical(1)
  ))

  if (n_specified == 0) {
    parameter <- "mu"
  }

  priors <- object[["priors"]]

  if (!is.null(parameter_mods)) {
    .check_plot_prior_name(parameter_mods, "parameter_mods")
    return(.select_plot_prior_term(
      prior_list = priors[["mods"]],
      term       = parameter_mods,
      argument   = "parameter_mods",
      prefix     = "mu",
      source     = "mods"
    ))
  }

  if (!is.null(parameter_scale)) {
    .check_plot_prior_name(parameter_scale, "parameter_scale")
    return(.select_plot_prior_term(
      prior_list = priors[["scale"]],
      term       = parameter_scale,
      argument   = "parameter_scale",
      prefix     = "log_tau",
      source     = "scale"
    ))
  }

  .check_plot_prior_name(parameter, "parameter")

  parameter <- switch(
    parameter,
    "effect"        = "mu",
    "heterogeneity" = "tau",
    "weightfunction" = "omega",
    parameter
  )

  if (parameter == "mu" && is.null(priors[["outcome"]][["mu"]]) && !is.null(priors[["mods"]][["intercept"]])) {
    return(list(
      prior  = priors[["mods"]][["intercept"]],
      label  = "mu_intercept",
      source = "mods",
      term   = "intercept"
    ))
  }

  if (parameter == "tau" && is.null(priors[["outcome"]][["tau"]]) && !is.null(priors[["scale"]][["intercept"]])) {
    return(list(
      prior  = priors[["scale"]][["intercept"]],
      label  = "log_tau_intercept",
      source = "scale",
      term   = "intercept"
    ))
  }

  if (parameter %in% c("omega", "PET", "PEESE", "bias") && !is.null(priors[["outcome"]][["bias"]])) {
    prior <- .select_plot_prior_bias(
      prior            = priors[["outcome"]][["bias"]],
      parameter        = parameter,
      allow_mixed_bias = allow_mixed_bias
    )
    return(list(prior = prior, label = parameter, source = "bias", term = parameter))
  }

  if (parameter %in% names(priors[["outcome"]]) && !is.null(priors[["outcome"]][[parameter]])) {
    return(list(prior = priors[["outcome"]][[parameter]], label = parameter, source = "outcome", term = parameter))
  }

  in_mods  <- !is.null(priors[["mods"]])  && parameter %in% names(priors[["mods"]])
  in_scale <- !is.null(priors[["scale"]]) && parameter %in% names(priors[["scale"]])

  if (in_mods && in_scale) {
    stop(sprintf(
      "The term '%s' is available in both moderator and scale priors. Use 'parameter_mods' or 'parameter_scale'.",
      parameter
    ), call. = FALSE)
  }

  if (in_mods) {
    return(list(
      prior  = priors[["mods"]][[parameter]],
      label  = paste0("mu_", parameter),
      source = "mods",
      term   = parameter
    ))
  }

  if (in_scale) {
    return(list(
      prior  = priors[["scale"]][[parameter]],
      label  = paste0("log_tau_", parameter),
      source = "scale",
      term   = parameter
    ))
  }

  stop(sprintf(
    "Unknown prior parameter '%s'. Available parameters are: %s.",
    parameter,
    .collapse_plot_prior_names(priors)
  ), call. = FALSE)
}

.select_plot_prior_term <- function(prior_list, term, argument, prefix, source) {

  if (is.null(prior_list)) {
    stop(sprintf("The '%s' argument can only be used when the object contains the corresponding priors.", argument), call. = FALSE)
  }

  if (!term %in% names(prior_list)) {
    stop(sprintf(
      "The specified '%s' term '%s' is not available. Available terms are: %s.",
      argument,
      term,
      paste0("'", names(prior_list), "'", collapse = ", ")
    ), call. = FALSE)
  }

  return(list(
    prior  = prior_list[[term]],
    label  = paste0(prefix, "_", term),
    source = source,
    term   = term
  ))
}

.select_plot_prior_bias <- function(prior, parameter, allow_mixed_bias = FALSE) {

  if (.prior_has_phacking(prior)) {
    .selection_stop_phacking_deferred()
  }

  if (!BayesTools::is.prior.mixture(prior)) {
    if (parameter == "bias" ||
        (parameter == "omega" && .prior_has_selection(prior)) ||
        (parameter == "PET" && BayesTools::is.prior.PET(prior)) ||
        (parameter == "PEESE" && BayesTools::is.prior.PEESE(prior))) {
      return(prior)
    }

    stop(sprintf("The publication-bias prior does not contain a '%s' component.", parameter), call. = FALSE)
  }

  has_weightfunction <- any(vapply(prior, .prior_has_selection, logical(1)))
  has_petpeese       <- any(vapply(prior, function(x) BayesTools::is.prior.PET(x) || BayesTools::is.prior.PEESE(x), logical(1)))

  if (parameter == "bias") {
    if (has_weightfunction && has_petpeese) {
      if (!allow_mixed_bias) {
        stop(
          "The publication-bias prior mixes weight-function and PET-PEESE components. ",
          "Use parameter = 'omega', 'PET', or 'PEESE' to plot one component type.",
          call. = FALSE
        )
      }
    }
    return(prior)
  }

  keep <- switch(
    parameter,
    "omega" = vapply(prior, .prior_has_selection, logical(1)),
    "PET"   = vapply(prior, BayesTools::is.prior.PET, logical(1)),
    "PEESE" = vapply(prior, BayesTools::is.prior.PEESE, logical(1))
  )

  if (!any(keep)) {
    stop(sprintf("The publication-bias prior does not contain a '%s' component.", parameter), call. = FALSE)
  }

  selected <- unclass(prior)[keep]
  if (length(selected) == 1) {
    return(selected[[1]])
  }

  class(selected) <- c("prior", "prior.mixture")
  attr(selected, "components")    <- rep("alternative", length(selected))
  attr(selected, "prior_weights") <- rep(1, length(selected))

  return(selected)
}

.append_print_prior_selection <- function(selected, prior, label, source, term) {

  if (is.null(prior)) {
    return(selected)
  }

  selected[[label]] <- list(
    prior  = prior,
    label  = label,
    source = source,
    term   = term
  )

  return(selected)
}

.select_print_prior_all <- function(object) {

  priors  <- object[["priors"]]
  outcome <- priors[["outcome"]]
  mods    <- priors[["mods"]]
  scale   <- priors[["scale"]]

  selected <- list()

  selected <- .append_print_prior_selection(
    selected = selected,
    prior    = outcome[["mu"]],
    label    = "mu",
    source   = "outcome",
    term     = "mu"
  )

  if (!is.null(mods)) {
    selected <- .append_print_prior_selection(
      selected = selected,
      prior    = mods[["intercept"]],
      label    = "mu_intercept",
      source   = "mods",
      term     = "intercept"
    )

    for (term in setdiff(names(mods), "intercept")) {
      selected <- .append_print_prior_selection(
        selected = selected,
        prior    = mods[[term]],
        label    = paste0("mu_", term),
        source   = "mods",
        term     = term
      )
    }
  }

  selected <- .append_print_prior_selection(
    selected = selected,
    prior    = outcome[["tau"]],
    label    = "tau",
    source   = "outcome",
    term     = "tau"
  )

  if (!is.null(scale)) {
    selected <- .append_print_prior_selection(
      selected = selected,
      prior    = scale[["intercept"]],
      label    = "log_tau_intercept",
      source   = "scale",
      term     = "intercept"
    )

    for (term in setdiff(names(scale), "intercept")) {
      selected <- .append_print_prior_selection(
        selected = selected,
        prior    = scale[[term]],
        label    = paste0("log_tau_", term),
        source   = "scale",
        term     = term
      )
    }
  }

  for (parameter in setdiff(names(outcome), c("mu", "tau"))) {
    selected <- .append_print_prior_selection(
      selected = selected,
      prior    = outcome[[parameter]],
      label    = parameter,
      source   = if (parameter == "bias") "bias" else "outcome",
      term     = parameter
    )
  }

  return(selected)
}

.print_prior_object <- function(x, label = NULL, ...) {

  dots <- list(...)
  silent <- isTRUE(dots[["silent"]])
  dots[["silent"]] <- TRUE

  output <- do.call(print, c(list(x = x), dots))

  if (!silent) {
    if (!is.null(label)) {
      output <- paste(output, collapse = "\n")
      output <- unlist(strsplit(output, "\n", fixed = TRUE), use.names = FALSE)
      output <- paste0("  ", output)
      output <- paste(output, collapse = "\n")
      cat(label, ":\n", sep = "")
    }
    cat(output, "\n", sep = "")
  }

  return(invisible(x))
}

.plot_prior_unstandardized <- function(object, selected, plot_type, dots) {

  if (!(selected[["source"]] %in% c("mods", "scale"))) {
    return(NULL)
  }

  if (!isTRUE(.standardize_continuous_predictors(object))) {
    return(NULL)
  }

  formula_info <- .plot_prior_formula_info(
    object = object,
    source = selected[["source"]]
  )

  if (is.null(formula_info) || is.null(formula_info[["formula_scale"]])) {
    return(NULL)
  }

  parameter <- gsub(":", "__xXx__", selected[["label"]], fixed = TRUE)
  if (!parameter %in% formula_info[["column_names"]]) {
    return(NULL)
  }

  n_points <- if (!is.null(dots[["n_points"]])) dots[["n_points"]] else 1000
  BayesTools::check_int(n_points, "n_points", lower = 2)

  par_name <- dots[["par_name"]]
  plot_dots <- dots
  plot_dots[c(
    "par_name", "n_points", "n_samples", "force_samples", "x_range_quant",
    "individual", "show_figures", "rescale_x", "transformation",
    "transformation_arguments", "transformation_settings"
  )] <- NULL

  plot <- do.call(
    BayesTools::plot_transformed_prior,
    c(
      list(
        prior_list               = formula_info[["prior_list"]],
        column_names             = formula_info[["column_names"]],
        formula_scale            = formula_info[["formula_scale"]],
        parameter                = parameter,
        n_points                 = n_points,
        x_range                  = dots[["xlim"]],
        transformation           = dots[["transformation"]],
        transformation_arguments = dots[["transformation_arguments"]],
        transformation_settings  = if (!is.null(dots[["transformation_settings"]])) {
          dots[["transformation_settings"]]
        } else {
          FALSE
        },
        plot_type                = plot_type,
        par_name                 = par_name
      ),
      plot_dots
    )
  )

  return(plot)
}

.plot_prior_formula_info <- function(object, source) {

  parameter <- .fitted_formula_parameter(source)
  design    <- .fitted_formula_design(
    object    = object,
    parameter = parameter,
    required  = TRUE
  )

  if (is.null(design[["formula_scale"]])) {
    return(NULL)
  }

  formula_scale <- list(design[["formula_scale"]])
  names(formula_scale) <- parameter

  return(list(
    prior_list    = design[["prior_list"]],
    column_names  = names(design[["prior_list"]]),
    formula_scale = formula_scale
  ))
}

.use_plot_prior_list_dispatch <- function(prior) {

  if (.is_prior_phacking(prior) || .prior_has_phacking(prior)) {
    .selection_stop_phacking_deferred()
  }

  return(
    (
      BayesTools::is.prior.factor(prior) &&
        !BayesTools::is.prior.mixture(prior) &&
        (BayesTools::is.prior.meandif(prior) || BayesTools::is.prior.orthonormal(prior))
    ) ||
      .is_prior_bias_kernel(prior)
  )
}

.prior_mixture_plot_list <- function(prior) {

  prior_list  <- unclass(prior)
  prior_names <- names(prior_list)

  attributes(prior_list) <- NULL
  names(prior_list)     <- prior_names

  return(prior_list)
}

.check_plot_prior_name <- function(x, argument) {

  if (!is.character(x) || length(x) != 1 || is.na(x) || !nzchar(x)) {
    stop(sprintf("The '%s' argument must be a single non-empty character string.", argument), call. = FALSE)
  }

  return(invisible(TRUE))
}

.collapse_plot_prior_names <- function(priors) {

  names_available <- names(priors[["outcome"]])

  if (!is.null(priors[["outcome"]][["bias"]])) {
    names_available <- unique(c(names_available, "omega", "PET", "PEESE"))
  }

  if (!is.null(priors[["mods"]])) {
    names_available <- c(names_available, paste0("parameter_mods = '", names(priors[["mods"]]), "'"))
  }

  if (!is.null(priors[["scale"]])) {
    names_available <- c(names_available, paste0("parameter_scale = '", names(priors[["scale"]]), "'"))
  }

  return(paste0("'", names_available, "'", collapse = ", "))
}

context("Publication-bias mixture plots")

mock_bias_mixture_fit <- function() {

  bias <- BayesTools::prior_mixture(
    prior_list = list(
      BayesTools::prior_none(),
      BayesTools::prior_weightfunction(
        side    = "two-sided",
        steps   = c(0.05),
        weights = BayesTools::wf_cumulative(c(1, 1))
      ),
      BayesTools::prior_PET("normal", list(mean = 0, sd = 1)),
      BayesTools::prior_PEESE("normal", list(mean = 0, sd = 1))
    ),
    is_null = c(TRUE, FALSE, FALSE, FALSE)
  )

  data <- list(outcome = data.frame(
    yi  = c(0.1, 0.2),
    sei = c(0.1, 0.2)
  ))
  attr(data, "measure")          <- "SMD"
  attr(data, "outcome_type")     <- "norm"
  attr(data, "effect_direction") <- "positive"

  fit <- structure(list(), class = "BayesTools_fit")
  attr(fit, "prior_list") <- list(
    mu   = BayesTools::prior("normal", list(mean = 0, sd = 1)),
    bias = bias
  )

  structure(
    list(
      data   = data,
      priors = list(outcome = list(bias = bias)),
      fit    = fit
    ),
    class = c("RoBMA", "brma")
  )
}

test_that("publication-bias plots request mixed bias posterior", {

  calls <- list()

  testthat::local_mocked_bindings(
    as_mixed_posteriors = function(model, parameters, ...) {
      calls[["as_mixed_posteriors"]] <<- list(parameters = parameters)

      return(list(bias = structure(
        matrix(1, nrow = 2, ncol = 1, dimnames = list(NULL, "omega[0,0.05]")),
        class = "mixed_posteriors"
      )))
    },
    plot_posterior = function(samples, parameter, ...) {
      calls[["plot_posterior"]] <<- list(parameter = parameter)

      return(structure(list(), class = "mock_plot"))
    },
    .package = "BayesTools"
  )

  out <- plot_weightfunction(
    mock_bias_mixture_fit(),
    show_data = FALSE,
    plot_type = "ggplot"
  )

  expect_s3_class(out, "mock_plot")
  expect_equal(calls[["as_mixed_posteriors"]][["parameters"]], "bias")
  expect_equal(calls[["plot_posterior"]][["parameter"]], "weightfunction")
})

test_that("PET-PEESE plot requests location and mixed bias posterior", {

  calls <- list()

  testthat::local_mocked_bindings(
    as_mixed_posteriors = function(model, parameters, ...) {
      calls[["as_mixed_posteriors"]] <<- list(parameters = parameters)

      return(list(
        mu   = structure(c(0, 0), class = "mixed_posteriors"),
        bias = structure(
          matrix(0, nrow = 2, ncol = 2, dimnames = list(NULL, c("PET", "PEESE"))),
          class = "mixed_posteriors"
        )
      ))
    },
    plot_posterior = function(samples, parameter, ...) {
      calls[["plot_posterior"]] <<- list(parameter = parameter)

      return(structure(list(), class = "mock_plot"))
    },
    .package = "BayesTools"
  )

  out <- plot_pet_peese(
    mock_bias_mixture_fit(),
    show_data = FALSE,
    plot_type = "ggplot"
  )

  expect_s3_class(out, "mock_plot")
  expect_equal(calls[["as_mixed_posteriors"]][["parameters"]], c("mu", "bias"))
  expect_equal(calls[["plot_posterior"]][["parameter"]], "PETPEESE")
})

test_that("generic posterior plot requests mixed bias posterior", {

  calls <- list()

  testthat::local_mocked_bindings(
    as_mixed_posteriors = function(model, parameters, conditional = NULL, ...) {
      calls[["as_mixed_posteriors"]] <<- list(
        parameters  = parameters,
        conditional = conditional
      )

      return(list(bias = structure(
        matrix(1, nrow = 2, ncol = 1, dimnames = list(NULL, "omega[0,0.05]")),
        class = "mixed_posteriors"
      )))
    },
    plot_posterior = function(samples, parameter, ...) {
      calls[["plot_posterior"]] <<- list(parameter = parameter)

      return(structure(list(), class = "mock_plot"))
    },
    .package = "BayesTools"
  )

  out <- plot(
    mock_bias_mixture_fit(),
    parameter = "weightfunction",
    plot_type = "ggplot"
  )

  expect_s3_class(out, "mock_plot")
  expect_equal(calls[["as_mixed_posteriors"]][["parameters"]], "bias")
  expect_null(calls[["as_mixed_posteriors"]][["conditional"]])
  expect_equal(calls[["plot_posterior"]][["parameter"]], "omega")
})

# ============================================================================ #
# brma.zplot.R
# ============================================================================ #
#
# Zplot diagnostics for brma class objects.
#
# Zplot analysis provides estimates of Expected Discovery Rate (EDR) and
# related metrics for assessing replicability of meta-analytic findings.
# The implementation follows the framework from Bartos & Schimmack (2022) and
# extends it to Bayesian meta-analytic models.
#
# Design principles:
# - as_zplot() transforms brma to zplot_brma, enabling summary/plot methods
# - Extrapolation mode removes publication bias to estimate "true" power
# - Fitted mode includes bias adjustments for observed distribution
# - All computations are sample-based to propagate posterior uncertainty
#
# ============================================================================ #


# ---------------------------------------------------------------------------- #
# as_zplot generic and method
# ---------------------------------------------------------------------------- #

#' @export
as_zplot <- function(object, ...) UseMethod("as_zplot")


#' @title Transform brma Object to Zplot
#'
#' @description Transforms an estimated brma model into a zplot object
#' that can be summarized and plotted to assess replicability.
#'
#' @param object a normal-outcome \code{brma} object.
#' @param significance_level z-value threshold for significance. Defaults
#' to \code{qnorm(0.975)} (two-sided alpha = 0.05).
#' @param max_samples maximum number of posterior samples for estimation.
#' Defaults to 10000. Use \code{Inf} to use all posterior samples.
#' @param ... additional arguments (currently unused).
#'
#' @details
#' Zplot analysis estimates the Expected Discovery Rate (EDR), the posterior
#' mean probability that an exact replication would be statistically significant
#' at the supplied threshold. This provides insight into the replicability of
#' findings in a literature.
#'
#' The implementation uses extrapolation mode by default, which removes
#' selection-model weights when evaluating the model-implied z-value density.
#' PET/PEESE regression terms remain part of the fitted location model.
#' Zplot diagnostics are available only for normal outcome models. GLMM
#' objects are rejected because their raw likelihood is on a count scale while
#' zplot diagnostics require observed effect-size z-statistics with standard
#' errors.
#'
#' The resulting object retains all original brma properties while adding
#' zplot results, enabling both standard meta-analytic summaries and
#' zplot diagnostics on the same object.
#'
#' @return The input object with added class \code{"zplot_brma"} and a new
#' \code{zplot} list component containing:
#' \describe{
#'   \item{estimates}{a list with posterior samples for \code{EDR} and
#'     missing-study \code{weights}}
#'   \item{data}{a list with \code{significance_level}, observed
#'     \code{z}-statistics, \code{N_significant}, and \code{N_observed}}
#' }
#'
#' @seealso [summary.zplot_brma()], [plot.zplot_brma()], [hist.zplot_brma()]
#'
#' @examples \dontrun{
#' if (requireNamespace("metadat", quietly = TRUE)) {
#'   data(dat.lehmann2018, package = "metadat")
#'   fit <- bPET(yi = yi, vi = vi, data = dat.lehmann2018, measure = "SMD")
#'
#'   zfit <- as_zplot(fit)
#'   summary(zfit)
#'   plot(zfit)
#' }
#' }
#'
#' @aliases as_zplot
#' @export
as_zplot.brma <- function(object, significance_level = stats::qnorm(0.975), max_samples = 10000, ...) {

  BayesTools::check_real(significance_level, "significance_level", lower = 0)
  max_samples <- .normalize_max_samples(max_samples, "max_samples")

  if (.outcome_type(object) != "norm") {
    stop(
      "as_zplot() is only available for normal outcome models; ",
      "GLMM objects are not supported.",
      call. = FALSE
    )
  }

  # compute the main estimates (extrapolate = TRUE: unbiased power)
  # This corresponds to "Expected Discovery Rate" if studies were exact replications
  # but without publication bias
  out_estimates <- .zplot_fun.brma(
    object      = object,
    z_threshold = significance_level,
    max_samples = max_samples,
    extrapolate = TRUE
  )

  # compute data summaries (observed z-stats)
  yi  <- .outcome_data_yi(object)
  sei <- .outcome_data_sei(object)
  z   <- yi / sei

  # store zplot results
  object[["zplot"]] <- list(
    estimates = list(
      EDR     = out_estimates[["EDR"]],
      weights = out_estimates[["weights"]]
    ),
    data = list(
      significance_level = significance_level,
      z                  = z,
      N_significant      = sum(abs(z) > significance_level),
      N_observed         = length(z)
    )
  )

  # add class
  class(object) <- c("zplot_brma", class(object))

  return(object)
}


# ---------------------------------------------------------------------------- #
# zplot generic and methods
# ---------------------------------------------------------------------------- #

#' @export
zplot <- function(object, ...) UseMethod("zplot")


#' @title Plot Zplot Diagnostics Directly
#'
#' @description Convenience wrapper for creating and plotting zplot diagnostics
#' from a fitted \code{brma} object.
#'
#' @param object a normal-outcome \code{brma} object, or a
#' \code{zplot_brma} object.
#' @param significance_level z-value threshold for significance. Defaults
#' to \code{qnorm(0.975)} (two-sided alpha = 0.05).
#' @param summary_max_samples maximum number of posterior samples used for the
#' EDR and missing-study summaries stored in the generated zplot object.
#' This is separate from the plot-density \code{max_samples} argument accepted
#' by \code{plot.zplot_brma()} through \code{...}. Defaults to 10000. Use
#' \code{Inf} to use all posterior samples.
#' @param ... arguments passed to \code{\link[=plot.zplot_brma]{plot.zplot_brma()}}.
#'
#' @details When \code{object} already inherits from \code{zplot_brma},
#' \code{zplot()} dispatches directly to \code{plot.zplot_brma()} without
#' recomputing stored summaries.
#'
#' @return \code{NULL} invisibly for base graphics, or a ggplot2 object.
#'
#' @seealso [as_zplot.brma()], [plot.zplot_brma()], [summary.zplot_brma()]
#'
#' @examples \dontrun{
#' if (requireNamespace("metadat", quietly = TRUE)) {
#'   data(dat.lehmann2018, package = "metadat")
#'   fit <- bPET(yi = yi, vi = vi, data = dat.lehmann2018, measure = "SMD")
#'
#'   zplot(fit)
#' }
#' }
#'
#' @aliases zplot
#' @export
zplot.brma <- function(object, significance_level = stats::qnorm(0.975),
                       summary_max_samples = 10000, ...) {

  zplot_object <- as_zplot(
    object             = object,
    significance_level = significance_level,
    max_samples        = summary_max_samples
  )

  return(plot(zplot_object, ...))
}


#' @rdname zplot.brma
#' @export
zplot.zplot_brma <- function(object, ...) {

  return(plot(object, ...))
}


# ---------------------------------------------------------------------------- #
# summary.zplot_brma
# ---------------------------------------------------------------------------- #

#' @title Summarize Zplot Results
#'
#' @description Creates summary tables for zplot estimates including
#' EDR, Soric FDR, and estimated missing studies.
#'
#' @param object a zplot_brma object.
#' @param probs quantiles of the posterior distribution to display.
#' Defaults to \code{c(.025, .975)}.
#' @param ... additional arguments (currently unused).
#'
#' @details
#' The summary includes:
#' \describe{
#'   \item{EDR}{Expected Discovery Rate - average power of significant studies}
#'   \item{Soric FDR}{Expected proportion of false discoveries among significant
#'     results, computed from EDR following Soric (1989)}
#'   \item{Missing N}{Estimated number of studies suppressed by publication bias,
#'     computed from the selection model weights}
#' }
#'
#' The footer reports the Observed Discovery Rate (ODR) with 95\% CI for
#' comparison with the model-estimated EDR.
#'
#' @return An object of class \code{"summary.zplot_brma"} containing the
#' estimates table.
#'
#' @seealso [as_zplot.brma()], [plot.zplot_brma()]
#'
#' @export
summary.zplot_brma <- function(object, probs = c(.025, .975), ...) {

  # get proportion of significant results
  obs_proportion <- stats::prop.test(
    object$zplot$data[["N_significant"]],
    object$zplot$data[["N_observed"]],
    conf.level = 0.95
  )

  info_text <- sprintf(
    "Estimated using %1$i estimates, %2$i significant (ODR = %3$.2f, 95%% CI [%4$.2f, %5$.2f]).",
    object$zplot$data[["N_observed"]],
    object$zplot$data[["N_significant"]],
    obs_proportion$estimate,
    obs_proportion$conf.int[1],
    obs_proportion$conf.int[2]
  )

  # compute estimates
  sig_level <- stats::pnorm(object$zplot$data[["significance_level"]], lower.tail = FALSE) * 2

  estimates <- cbind.data.frame(
    "EDR"       = object$zplot$estimates[["EDR"]],
    "Soric FDR" = .get_Soric_FDR(object$zplot$estimates[["EDR"]], sig_level),
    "Missing N" = (object$zplot$estimates[["weights"]] - 1) * object$zplot$data[["N_observed"]]
  )

  estimates_table <- BayesTools::ensemble_estimates_table(
    samples    = estimates,
    parameters = names(estimates),
    probs      = probs,
    title      = "Zplot Estimates:",
    footnotes  = info_text
  )

  output <- list(
    estimates = estimates_table
  )

  class(output) <- "summary.zplot_brma"
  return(output)
}


#' @title Print Zplot Summary
#'
#' @param x a summary.zplot_brma object.
#' @param ... additional arguments (currently unused).
#'
#' @return Invisibly returns the summary object.
#'
#' @export
print.summary.zplot_brma <- function(x, ...) {

  cat("\n")
  print(x[["estimates"]])
  cat("\n")

  return(invisible(x))
}


#' @export
print.zplot_brma <- function(x, ...) {
  print(summary(x, ...))
}


# ---------------------------------------------------------------------------- #
# plot.zplot_brma
# ---------------------------------------------------------------------------- #

#' @title Plot Zplot Results
#'
#' @description Plots a zplot visualization showing the histogram of
#' observed z-statistics overlaid with model-implied densities.
#'
#' @param x a zplot_brma object.
#' @param plot_type graphics system: \code{"base"} or \code{"ggplot"}.
#' Defaults to \code{"base"}.
#' @param probs quantiles for credible intervals. Defaults to \code{c(.025, .975)}.
#' @param max_samples maximum posterior samples for density estimation.
#' Defaults to 10000. Use \code{Inf} to use all posterior samples.
#' @param plot_fit whether to show fitted density (with bias adjustments).
#' Defaults to \code{TRUE}.
#' @param plot_extrapolation whether to show extrapolated density (bias removed).
#' Defaults to \code{TRUE}.
#' @param plot_ci whether to show credible interval bands. Defaults to \code{TRUE}.
#' @param plot_thresholds whether to show significance threshold lines.
#' Defaults to \code{TRUE}.
#' @param from,to z-value range for plotting. Defaults to \code{-6} and \code{6}.
#' @param by.hist bin width for histogram. Defaults to 0.5.
#' @param length.out.hist number of histogram bins (alternative to \code{by.hist}).
#' @param by.lines step size for density lines. Defaults to 0.05.
#' @param length.out.lines number of density points (alternative to \code{by.lines}).
#' @param dots_hist graphical parameters for histogram (list).
#' @param dots_fit graphical parameters for fit lines (list).
#' @param dots_extrapolation graphical parameters for extrapolation lines (list).
#' @param dots_thresholds graphical parameters for threshold lines (list).
#' @param ... additional graphical parameters passed to components.
#'
#' @details
#' The plot displays two density curves:
#' \describe{
#'   \item{Fit (black)}{Model-implied density including publication bias adjustments.
#'     This represents the expected distribution of z-statistics given the estimated
#'     selection process.}
#'   \item{Extrapolation (blue)}{Bias-corrected curve representing the hypothetical
#'     distribution without selective reporting, scaled by the expected suppressed
#'     studies under selection models. This curve is not normalized to integrate
#'     to one when selection implies missing studies.}
#' }
#'
#' @return \code{NULL} invisibly for base graphics, or a ggplot2 object.
#'
#' @seealso [hist.zplot_brma()], [lines.zplot_brma()], [summary.zplot_brma()]
#'
#' @export
plot.zplot_brma <- function(x, plot_type = "base",
                             probs = c(.025, .975), max_samples = 10000,
                             plot_fit = TRUE, plot_extrapolation = TRUE,
                             plot_ci = TRUE, plot_thresholds = TRUE,
                             from = -6, to = 6,
                             by.hist = 0.5, length.out.hist = NULL,
                             by.lines = 0.05, length.out.lines = NULL,
                             dots_hist = NULL, dots_fit = NULL,
                             dots_extrapolation = NULL, dots_thresholds = NULL, ...) {

  BayesTools::check_char(plot_type, "plot_type", allow_values = c("base", "ggplot"))
  BayesTools::check_bool(plot_fit, "plot_fit")
  BayesTools::check_bool(plot_extrapolation, "plot_extrapolation")
  BayesTools::check_bool(plot_ci, "plot_ci")
  BayesTools::check_bool(plot_thresholds, "plot_thresholds")
  BayesTools::check_real(probs, "probs", lower = 0, upper = 1, check_length = 2)
  max_samples <- .normalize_max_samples(max_samples, "max_samples")

  dots <- list(...)

  # get line values first so we can set ylim
  ymax <- 0

  # 1. Fit lines (Fitted model - with bias)
  if (plot_fit) {
    dots_fit <- if(!is.null(dots_fit)) dots_fit else list()
    lines_fit <- do.call(lines.zplot_brma, c(
      list(x = x, plot_type = plot_type, probs = probs, max_samples = max_samples,
           extrapolate = FALSE, plot_ci = plot_ci, from = from, to = to,
           by = by.lines, length.out = length.out.lines, as_data = TRUE),
      dots_fit, dots
    ))
    ymax <- max(c(ymax, lines_fit$y, if (plot_ci) lines_fit$y_uCI))
  }

  # 2. Extrapolation lines (unbiased zplot density)
  if (plot_extrapolation) {
    # prepare extrapolation dots with default blue color
    dots_extrapolation <- if(!is.null(dots_extrapolation)) dots_extrapolation else list()
    lines_extrapolation <- do.call(lines.zplot_brma, c(
      list(x = x, plot_type = plot_type, probs = probs, max_samples = max_samples,
           extrapolate = TRUE, plot_ci = plot_ci, from = from, to = to,
           by = by.lines, length.out = length.out.lines, as_data = TRUE,
           col = "blue"), # default color if not in dots
      dots_extrapolation, dots
    ))
    ymax <- max(c(ymax, lines_extrapolation$y, if (plot_ci) lines_extrapolation$y_uCI))
  }

  # 3. Create histogram base
  plot <- hist.zplot_brma(
    x               = x,
    plot_type       = plot_type,
    from            = from,
    to              = to,
    by              = by.hist,
    length.out      = length.out.hist,
    plot_thresholds = plot_thresholds,
    dots_thresholds = dots_thresholds,
    ylim            = if (ymax != 0) c(0, ymax * 1.05) else NULL,
    dots_hist       = dots_hist,
    dots_all        = dots
  )

  # 4. Add fit lines
  if (plot_fit) {
    dots_fit_lines <- .get_dots_lines_zplot(c(dots, dots_fit), plot_type = plot_type)

    if (plot_type == "base") {
      if (plot_ci) {
        graphics::polygon(
          c(lines_fit$x, rev(lines_fit$x)),
          c(lines_fit$y_lCI, rev(lines_fit$y_uCI)),
          border = NA,
          col    = scales::alpha(dots_fit_lines$col, dots_fit_lines$alpha)
        )
      }
      graphics::lines(
        lines_fit$x, lines_fit$y,
        lwd = dots_fit_lines$lwd,
        col = dots_fit_lines$col,
        lty = dots_fit_lines$lty
      )
    } else if (plot_type == "ggplot") {
      if (plot_ci) {
        plot <- plot + ggplot2::geom_ribbon(
          ggplot2::aes(
            x    = lines_fit$x,
            ymin = lines_fit$y_lCI,
            ymax = lines_fit$y_uCI
          ),
          fill  = dots_fit_lines$color,
          alpha = dots_fit_lines$alpha
        )
      }
      plot <- plot + ggplot2::geom_line(
        ggplot2::aes(
          x = lines_fit$x,
          y = lines_fit$y
        ),
        color     = dots_fit_lines$color,
        linewidth = dots_fit_lines$linewidth,
        linetype  = dots_fit_lines$linetype
      )
    }
  }

  # 5. Add extrapolation lines
  if (plot_extrapolation) {
     dots_ext_lines <- .get_dots_lines_zplot(c(dots, dots_extrapolation), plot_type = plot_type, col = "blue")

    if (plot_type == "base") {
      if (plot_ci) {
        graphics::polygon(
          c(lines_extrapolation$x, rev(lines_extrapolation$x)),
          c(lines_extrapolation$y_lCI, rev(lines_extrapolation$y_uCI)),
          border = NA,
          col    = scales::alpha(dots_ext_lines$col, dots_ext_lines$alpha)
        )
      }
      graphics::lines(
        lines_extrapolation$x, lines_extrapolation$y,
        lwd = dots_ext_lines$lwd,
        col = dots_ext_lines$col,
        lty = dots_ext_lines$lty
      )
    } else if (plot_type == "ggplot") {
      if (plot_ci) {
        plot <- plot + ggplot2::geom_ribbon(
          ggplot2::aes(
            x    = lines_extrapolation$x,
            ymin = lines_extrapolation$y_lCI,
            ymax = lines_extrapolation$y_uCI
          ),
          fill  = dots_ext_lines$color,
          alpha = dots_ext_lines$alpha
        )
      }
      plot <- plot + ggplot2::geom_line(
        ggplot2::aes(
          x = lines_extrapolation$x,
          y = lines_extrapolation$y
        ),
        color     = dots_ext_lines$color,
        linewidth = dots_ext_lines$linewidth,
        linetype  = dots_ext_lines$linetype
      )
    }
  }

  # return
  if (plot_type == "base") {
    return(invisible())
  } else if (plot_type == "ggplot") {
    return(plot)
  }
}


# ---------------------------------------------------------------------------- #
# hist.zplot_brma
# ---------------------------------------------------------------------------- #

#' @title Histogram of Z-Statistics
#'
#' @description Plots a histogram of observed z-values from the meta-analysis.
#'
#' @param x a zplot_brma object.
#' @param plot_type graphics system: \code{"base"} or \code{"ggplot"}.
#' Defaults to \code{"base"}.
#' @param from,to z-value range for plotting. Defaults to \code{-6} and \code{6}.
#' @param by bin width. Defaults to 0.5.
#' @param length.out number of bins (alternative to \code{by}).
#' @param add whether to add to existing plot. Defaults to \code{FALSE}.
#' @param plot_thresholds whether to show significance threshold lines.
#' Defaults to \code{TRUE}.
#' @param dots_thresholds graphical parameters for threshold lines (list).
#' @param dots_hist graphical parameters for histogram (list).
#' @param dots_all graphical parameters passed to all components (list).
#' @param ... additional graphical parameters.
#'
#' @details
#' Z-statistics are computed as effect size divided by standard error
#' (\code{yi / sei}). Histogram bins are adjusted to align with significance
#' thresholds when selection model priors are present.
#'
#' @return \code{NULL} invisibly for base graphics, or a ggplot2 object.
#'
#' @seealso [plot.zplot_brma()], [lines.zplot_brma()]
#'
#' @export
hist.zplot_brma <- function(x, plot_type = "base",
                             from = -6, to = 6, by = 0.5, length.out = NULL,
                             add = FALSE, plot_thresholds = TRUE,
                             dots_thresholds = NULL, dots_hist = NULL, dots_all = NULL, ...) {

  BayesTools::check_char(plot_type, "plot_type", allow_values = c("base", "ggplot"))
  BayesTools::check_real(from, "from")
  BayesTools::check_real(to, "to")
  BayesTools::check_real(by, "by", allow_NULL = TRUE)
  BayesTools::check_bool(add, "add")
  BayesTools::check_bool(plot_thresholds, "plot_thresholds")

  dots <- list(...)
  if (!is.null(dots_all)) {
    dots <- c(dots, dots_all[!names(dots_all) %in% names(dots)])
  }

  z <- x$zplot$data[["z"]]

  # create bin sequence using zplot helper (handles thresholds)
  z_sequence <- .zplot_bins(priors = x[["priors"]], from = from, to = to, by = by, length.out = length.out, type = "hist")
  z_in_range <- z >= min(z_sequence) & z <= max(z_sequence)

  # message about out-of-range values
  if (sum(!z_in_range) > 0) {
    message(sprintf("%1$i z-statistics are out of the plotting range", sum(!z_in_range)))
  }

  # create histogram data
  z_hist <- graphics::hist(z[z_in_range], breaks = z_sequence, plot = FALSE)
  z_hist$density <- z_hist$density * mean(z_in_range)

  df_hist <- data.frame(
    x       = z_hist$mids,
    density = z_hist$density,
    breaks  = diff(z_hist$breaks)
  )

  # return data if requested via dots
  if (isTRUE(dots[["as_data"]])) {
    return(df_hist)
  }

  # merge dots_hist
  if (!is.null(dots_hist)) {
    dots <- c(dots_hist, dots[!names(dots) %in% names(dots_hist)])
  }
  dots_hist_params <- .get_dots_hist_zplot(dots, plot_type = plot_type, max_density = max(z_hist$density))

  # create the plot
  if (plot_type == "ggplot") {
    out <- ggplot2::ggplot() +
      ggplot2::geom_col(
        ggplot2::aes(
          x = df_hist$x,
          y = df_hist$density
        ),
        fill  = dots_hist_params$fill,
        color = dots_hist_params$color,
        alpha = dots_hist_params$alpha,
        width = df_hist$breaks
      ) +
      ggplot2::labs(x = dots_hist_params$xlab, y = dots_hist_params$ylab) +
      ggplot2::ggtitle(dots_hist_params$main)
  } else {
    if (add) {
      graphics::rect(
        xleft   = df_hist$x - df_hist$breaks / 2,
        xright  = df_hist$x + df_hist$breaks / 2,
        ybottom = 0,
        ytop    = df_hist$density,
        border  = dots_hist_params$border,
        col     = dots_hist_params$col
      )
    } else {
      graphics::plot(
        z_hist,
        freq   = FALSE,
        border = dots_hist_params$border,
        col    = dots_hist_params$col,
        xlab   = dots_hist_params$xlab,
        ylab   = dots_hist_params$ylab,
        main   = dots_hist_params$main,
        ylim   = dots_hist_params$ylim,
        xaxt   = dots_hist_params$xaxt,
        yaxt   = dots_hist_params$yaxt,
        las    = dots_hist_params$las
      )
    }
  }

  # add threshold lines
  if (plot_thresholds) {
    thresholds <- .zplot_threshold(x[["priors"]])
    if (length(thresholds) > 0) {
      dots_thresholds <- if (!is.null(dots_thresholds)) dots_thresholds else list()
      dots_thresholds_params <- .get_dots_thresholds_zplot(dots_thresholds, plot_type = plot_type)

      if (plot_type == "base") {
        graphics::abline(
          v   = thresholds,
          col = dots_thresholds_params$col,
          lty = dots_thresholds_params$lty,
          lwd = dots_thresholds_params$lwd
        )
      } else if (plot_type == "ggplot") {
        out <- out + ggplot2::geom_vline(
          xintercept = thresholds,
          color      = dots_thresholds_params$color,
          linetype   = dots_thresholds_params$linetype,
          linewidth  = dots_thresholds_params$linewidth
        )
      }
    }
  }

  # return
  if (plot_type == "base") {
    return(invisible())
  } else if (plot_type == "ggplot") {
    return(out)
  }
}


# ---------------------------------------------------------------------------- #
# lines.zplot_brma
# ---------------------------------------------------------------------------- #

#' @title Add Zplot Density Lines
#'
#' @description Adds model-implied density lines to an existing zplot.
#'
#' @param x a zplot_brma object.
#' @param plot_type graphics system: \code{"base"} or \code{"ggplot"}.
#' Defaults to \code{"base"}.
#' @param probs quantiles for credible intervals. Defaults to \code{c(.025, .975)}.
#' @param max_samples maximum posterior samples for density. Defaults to 10000.
#' Use \code{Inf} to use all posterior samples.
#' @param plot_ci whether to show credible interval bands. Defaults to \code{TRUE}.
#' @param extrapolate whether to remove bias adjustments. Defaults to \code{FALSE}.
#' @param from,to z-value range for density. Defaults to \code{-6} and \code{6}.
#' @param by step size for density points. Defaults to 0.05.
#' @param length.out number of density points (alternative to \code{by}).
#' @param col line color. Defaults to \code{"black"}.
#' @param as_data whether to return data instead of plotting. Defaults to \code{FALSE}.
#' @param ... additional graphical parameters.
#'
#' @details
#' When \code{extrapolate = FALSE}, the density includes all bias adjustments
#' (PET/PEESE regression, selection weights) representing the fitted model.
#' When \code{extrapolate = TRUE}, bias adjustments are removed to show the
#' hypothetical distribution without publication bias. Under selection models,
#' the curve is scaled by inverse selection probability and need not integrate
#' to one.
#'
#' @return \code{NULL} invisibly for base graphics, ggplot2 layers for ggplot,
#' or a data frame with columns \code{x}, \code{y}, \code{y_lCI}, \code{y_uCI}
#' if \code{as_data = TRUE}.
#'
#' @seealso [plot.zplot_brma()], [hist.zplot_brma()]
#'
#' @export
lines.zplot_brma <- function(x, plot_type = "base",
                              probs = c(.025, .975), max_samples = 10000,
                              plot_ci = TRUE, extrapolate = FALSE,
                              from = -6, to = 6, by = 0.05, length.out = NULL,
                              col = "black", as_data = FALSE, ...) {

  BayesTools::check_char(plot_type, "plot_type", allow_values = c("base", "ggplot"))
  BayesTools::check_real(probs, "probs", lower = 0, upper = 1, check_length = 2)
  max_samples <- .normalize_max_samples(max_samples, "max_samples")
  BayesTools::check_bool(plot_ci, "plot_ci")
  BayesTools::check_bool(extrapolate, "extrapolate")
  BayesTools::check_real(from, "from")
  BayesTools::check_real(to, "to")
  BayesTools::check_real(by, "by", allow_NULL = TRUE)


  dots <- list(...)

  # prepare z sequence
  z_sequence <- .zplot_bins(priors = x[["priors"]], from = from, to = to, by = by, length.out = length.out, type = "dens")

  # get densities from model
  z_density <- .zplot_fun.brma(
     object      = x,
     z_sequence  = z_sequence,
     max_samples = max_samples,
     extrapolate = extrapolate
  )

  df_density <- data.frame(
    x     = z_sequence,
    y     = colMeans(z_density),
    y_lCI = apply(z_density, 2, stats::quantile, probs = probs[1]),
    y_uCI = apply(z_density, 2, stats::quantile, probs = probs[2])
  )

  # return data if requested
  if (as_data) {
    return(df_density)
  }

  # get line-specific parameters
  dots_lines <- .get_dots_lines_zplot(dots, plot_type = plot_type, col = col)

  if (plot_type == "ggplot") {
    out <- list()
    if (plot_ci) {
      out[[1]] <- ggplot2::geom_ribbon(
        ggplot2::aes(
          x    = df_density$x,
          ymin = df_density$y_lCI,
          ymax = df_density$y_uCI
        ),
        fill  = dots_lines$color,
        alpha = dots_lines$alpha
      )
    }
    out[[length(out) + 1]] <- ggplot2::geom_line(
      ggplot2::aes(
        x = df_density$x,
        y = df_density$y
      ),
      color     = dots_lines$color,
      linewidth = dots_lines$linewidth,
      linetype  = dots_lines$linetype
    )
  } else {
    if (plot_ci) {
      graphics::polygon(
        c(df_density$x, rev(df_density$x)),
        c(df_density$y_lCI, rev(df_density$y_uCI)),
        border = NA,
        col    = scales::alpha(dots_lines$col, dots_lines$alpha)
      )
    }
    graphics::lines(
      df_density$x, df_density$y,
      lwd = dots_lines$lwd,
      col = dots_lines$col,
      lty = dots_lines$lty
    )
  }

  # return
  if (plot_type == "base") {
    return(invisible())
  } else if (plot_type == "ggplot") {
    return(out)
  }
}


# ============================================================================ #
# Internal Helper Functions
# ============================================================================ #

# ---------------------------------------------------------------------------- #
# .zplot_fun.brma
# ---------------------------------------------------------------------------- #
#
# Core computation function for zplot estimates and densities.
#
# This function operates in two modes:
# 1. EDR mode (z_threshold set): Computes Expected Discovery Rate
# 2. Density mode (z_sequence set): Computes density over z-values
#
# The extrapolate parameter controls bias adjustment:
# - extrapolate=TRUE:  Removes publication bias (PET/PEESE/weights) to estimate
#                      "true" power distribution
# - extrapolate=FALSE: Includes all bias adjustments for fitted distribution
#
# @param object      brma object with fit and priors
# @param z_threshold z-value threshold for significance (EDR mode)
# @param z_sequence  vector of z-values for density evaluation (density mode)
# @param max_samples maximum posterior samples for computation
# @param extrapolate whether to remove bias adjustments
#
# @return EDR mode:     list with EDR (S-vector) and weights (S-vector)
#         Density mode: S x length(z_sequence) matrix of densities
#
# ---------------------------------------------------------------------------- #
.zplot_fun.brma <- function(object, z_threshold = NULL, z_sequence = NULL,
                             max_samples = 10000, extrapolate = FALSE) {

  max_samples <- .normalize_max_samples(max_samples, "max_samples")

  ### extract model info
  is_multilevel     <- .is_multilevel(object)
  is_weightfunction <- .is_weightfunction(object)
  outcome_type      <- .outcome_type(object)
  effect_direction  <- .effect_direction(object)
  priors            <- object[["priors"]]
  data              <- object[["data"]]

  ### get outcome data
  outcome_data <- object[["data"]][["outcome"]]
  yi           <- outcome_data[["yi"]]
  sei          <- outcome_data[["sei"]]
  K            <- length(yi)

  ### determine mu type
  # "cluster" for multilevel (mu + gamma), "terms" for marginal/single-level
  mu_type <- if (is_multilevel) "cluster" else "terms"

  ### 1. Predict mu (location) samples
  # predict.brma handles PET/PEESE via bias_adjusted argument
  # extrapolate=TRUE -> bias_adjusted=TRUE (unbiased predictions)
  mu_samples <- predict.brma(
    object        = object,
    type          = mu_type,
    bias_adjusted = extrapolate,
    quiet         = TRUE
  )

  ### 2. Get tau (heterogeneity) samples
  # specific handling to get tau_within
  # .evaluate.brma.tau handles ML splitting logic
  tau_result <- .evaluate.brma.tau(
    fit           = object[["fit"]],
    scale_data    = object[["data"]][["scale"]],
    scale_formula = if (.is_scale(object)) .create_fit_formula_list(data = object[["data"]], "scale") else NULL,
    scale_priors  = object[["priors"]][["scale"]],
    is_scale      = .is_scale(object),
    is_multilevel = is_multilevel,
    K             = K
  )
  tau_within <- tau_result[["tau_within"]]

  ### 3. Subsample for efficiency
  S            <- nrow(mu_samples)
  selected_ind <- .thin_sample_rows(S, max_samples)
  if (!is.null(selected_ind)) {
    mu_samples        <- mu_samples[selected_ind, , drop = FALSE]
    tau_within        <- tau_within[selected_ind, , drop = FALSE]
    posterior_samples <- suppressWarnings(coda::as.mcmc(object[["fit"]]))[selected_ind, ]
    S                 <- length(selected_ind)
  } else {
    posterior_samples <- suppressWarnings(coda::as.mcmc(object[["fit"]]))
  }

  ### 4. Prepare for Computation
  selection <- .zplot_selection_context(
    object            = object,
    data              = data,
    priors            = priors,
    posterior_samples = posterior_samples,
    is_weightfunction = is_weightfunction
  )


  ### 5. Dispatch: EDR (Threshold) vs Density (Sequence)

  if (!is.null(z_threshold)) {
    return(.zplot_threshold_vectorized(
      z_threshold      = z_threshold,
      mu_samples       = mu_samples,
      tau_within       = tau_within,
      sei              = sei,
      selection        = selection,
      extrapolate      = extrapolate,
      effect_direction = effect_direction
    ))
  }

  if (!is.null(z_sequence)) {
    return(.zplot_density_vectorized(
      z_sequence       = z_sequence,
      mu_samples       = mu_samples,
      tau_within       = tau_within,
      sei              = sei,
      selection        = selection,
      extrapolate      = extrapolate,
      effect_direction = effect_direction
    ))
  }

  return(NULL)
}


# ---------------------------------------------------------------------------- #
# .zplot_threshold_vectorized
# ---------------------------------------------------------------------------- #
#
# Vectorized EDR computation for normal and selected-normal rows.
#
# ---------------------------------------------------------------------------- #
.zplot_threshold_vectorized <- function(z_threshold, mu_samples, tau_within,
                                         sei, selection, extrapolate,
                                         effect_direction) {

  S         <- nrow(mu_samples)
  K         <- ncol(mu_samples)
  sei_mat   <- matrix(sei, nrow = S, ncol = K, byrow = TRUE)
  total_sd  <- sqrt(tau_within^2 + sei_mat^2)

  if (!is.null(selection) && .has_native_zplot_threshold()) {
    return(.zplot_selnorm_threshold_summary(
      z_threshold       = z_threshold,
      mean              = mu_samples,
      sd                = total_sd,
      sei               = sei,
      selection_context = selection,
      extrapolate       = extrapolate
    ))
  }

  q_upper   <- z_threshold * sei
  q_lower   <- -z_threshold * sei

  thresholds <- stats::pnorm(
    matrix(q_upper, nrow = S, ncol = K, byrow = TRUE),
    mean       = mu_samples,
    sd         = total_sd,
    lower.tail = FALSE
  ) + stats::pnorm(
    matrix(q_lower, nrow = S, ncol = K, byrow = TRUE),
    mean       = mu_samples,
    sd         = total_sd,
    lower.tail = TRUE
  )
  weights       <- .zplot_inverse_selection_weights(
    mean      = mu_samples,
    sd        = total_sd,
    sei       = sei,
    selection = selection
  )
  weighted_rows <- if (is.null(selection)) integer(0) else which(!selection[["use_normal"]])
  if (length(weighted_rows) > 0) {
    selection_weight <- .selection_context_subset_rows(selection, weighted_rows)
    mean_weight      <- mu_samples[weighted_rows, , drop = FALSE]
    sd_weight        <- total_sd[weighted_rows, , drop = FALSE]

    if (!extrapolate) {
      prob_upper <- .selection_step_cdf_matrix(
        q                 = q_upper,
        mean              = mean_weight,
        sd                = sd_weight,
        sei               = sei,
        selection_context = selection_weight,
        lower.tail        = FALSE
      )
      prob_lower <- .selection_step_cdf_matrix(
        q                 = q_lower,
        mean              = mean_weight,
        sd                = sd_weight,
        sei               = sei,
        selection_context = selection_weight,
        lower.tail        = TRUE
      )

      thresholds[weighted_rows, ] <- prob_upper + prob_lower
    }
  }

  if (!is.null(selection) && extrapolate) {
    EDR <- rowSums(thresholds * weights) / rowSums(weights)
  } else {
    EDR <- rowMeans(thresholds)
  }

  return(list(
    EDR     = EDR,
    weights = rowMeans(weights)
  ))
}


# ---------------------------------------------------------------------------- #
# .zplot_selnorm_threshold_summary
# ---------------------------------------------------------------------------- #
#
# Native EDR and inverse-selection-weight reductions for zplot thresholds.
#
# ---------------------------------------------------------------------------- #
.has_native_zplot_threshold <- function() {

  return(is.loaded(
    "RoBMA_selnorm_zcurve_threshold_summary",
    PACKAGE = "RoBMA"
  ))
}

.zplot_selnorm_threshold_summary <- function(z_threshold, mean, sd, sei,
                                             selection_context, extrapolate) {

  .selection_require_step_evaluable(selection_context, ".zplot_threshold_vectorized()")

  return(.Call(
    "RoBMA_selnorm_zcurve_threshold_summary",
    .native_numeric_vector(z_threshold),
    .native_numeric_matrix(mean),
    .native_numeric_matrix(sd),
    .native_numeric_vector(sei),
    .native_numeric_matrix(selection_context[["omega"]]),
    .native_numeric_vector(selection_context[["alpha"]]),
    .native_integer_vector(selection_context[["phack_kind"]]),
    .native_integer_vector(selection_context[["kernel_mode"]]),
    .native_numeric_vector(selection_context[["z_lower"]]),
    .native_numeric_vector(selection_context[["z_upper"]]),
    .native_integer_vector(selection_context[["sign"]]),
    .native_integer_vector(selection_context[["phack_q"]]),
    .native_numeric_vector(selection_context[["phack_z_source"]]),
    .native_numeric_vector(selection_context[["phack_z_dest"]]),
    .native_numeric_vector(selection_context[["segments"]][["bounds"]]),
    .native_integer_vector(selection_context[["segments"]][["step_bin"]]),
    .native_integer_vector(selection_context[["segments"]][["phack_region"]]),
    as.logical(extrapolate),
    .selection_telescope_probabilities(selection_context),
    PACKAGE = "RoBMA"
  ))
}


# ---------------------------------------------------------------------------- #
# .zplot_density_vectorized
# ---------------------------------------------------------------------------- #
#
# Vectorized z-density computation for normal and selected-normal rows.
#
# ---------------------------------------------------------------------------- #
.has_native_zplot_density <- function(selection = FALSE) {

  symbols <- "RoBMA_zcurve_normal_density_matrix"
  if (selection) {
    symbols <- c(symbols, "RoBMA_selnorm_zcurve_density_matrix")
  }

  return(all(vapply(symbols, is.loaded, logical(1), PACKAGE = "RoBMA")))
}

.zplot_normal_density_matrix <- function(z_sequence, mean, sd, sei) {

  if (!.has_native_zplot_density(selection = FALSE)) {
    stop("The native zplot density kernel is not loaded.", call. = FALSE)
  }

  return(.Call(
    "RoBMA_zcurve_normal_density_matrix",
    .native_numeric_vector(z_sequence),
    .native_numeric_matrix(mean),
    .native_numeric_matrix(sd),
    .native_numeric_vector(sei),
    PACKAGE = "RoBMA"
  ))
}

.zplot_selnorm_density_matrix <- function(z_sequence, mean, sd, sei,
                                           selection_context, extrapolate) {

  .selection_require_step_evaluable(selection_context, ".zplot_density_vectorized()")

  if (!.has_native_zplot_density(selection = TRUE)) {
    stop("The native selected-normal zplot density kernel is not loaded.", call. = FALSE)
  }

  return(.Call(
    "RoBMA_selnorm_zcurve_density_matrix",
    .native_numeric_vector(z_sequence),
    .native_numeric_matrix(mean),
    .native_numeric_matrix(sd),
    .native_numeric_vector(sei),
    .native_numeric_matrix(selection_context[["omega"]]),
    .native_numeric_vector(selection_context[["alpha"]]),
    .native_integer_vector(selection_context[["phack_kind"]]),
    .native_integer_vector(selection_context[["kernel_mode"]]),
    .native_numeric_vector(selection_context[["z_lower"]]),
    .native_numeric_vector(selection_context[["z_upper"]]),
    .native_integer_vector(selection_context[["sign"]]),
    .native_integer_vector(selection_context[["phack_q"]]),
    .native_numeric_vector(selection_context[["phack_z_source"]]),
    .native_numeric_vector(selection_context[["phack_z_dest"]]),
    .native_numeric_vector(selection_context[["segments"]][["bounds"]]),
    .native_integer_vector(selection_context[["segments"]][["step_bin"]]),
    .native_integer_vector(selection_context[["segments"]][["phack_region"]]),
    as.logical(extrapolate),
    .selection_telescope_probabilities(selection_context),
    PACKAGE = "RoBMA"
  ))
}

.zplot_density_vectorized <- function(z_sequence, mu_samples, tau_within,
                                       sei, selection, extrapolate,
                                       effect_direction) {

  S           <- nrow(mu_samples)
  K           <- ncol(mu_samples)
  sei_mat     <- matrix(sei, nrow = S, ncol = K, byrow = TRUE)
  total_sd    <- sqrt(tau_within^2 + sei_mat^2)

  if (is.null(selection)) {
    return(.zplot_normal_density_matrix(
      z_sequence = z_sequence,
      mean       = mu_samples,
      sd         = total_sd,
      sei        = sei
    ))
  }

  density <- .zplot_selnorm_density_matrix(
    z_sequence        = z_sequence,
    mean              = mu_samples,
    sd                = total_sd,
    sei               = sei,
    selection_context = selection,
    extrapolate       = extrapolate
  )

  return(density)
}


# ---------------------------------------------------------------------------- #
# .zplot_inverse_selection_weights
# ---------------------------------------------------------------------------- #
#
# Expected attempted studies represented by each observed study under the
# selection model. Normal/no-bias branches have weight one.
#
# ---------------------------------------------------------------------------- #
.zplot_inverse_selection_weights <- function(mean, sd, sei, selection) {

  weights <- matrix(1, nrow = nrow(mean), ncol = ncol(mean))

  if (is.null(selection)) {
    return(weights)
  }

  weighted_rows <- which(!selection[["use_normal"]])
  if (length(weighted_rows) == 0L) {
    return(weights)
  }

  selection_weight <- .selection_context_subset_rows(selection, weighted_rows)
  log_norm <- .selection_step_log_norm_matrix(
    mean              = mean[weighted_rows, , drop = FALSE],
    sd                = sd[weighted_rows, , drop = FALSE],
    sei               = sei,
    selection_context = selection_weight
  )

  weights[weighted_rows, ] <- exp(-log_norm)
  return(weights)
}


# ---------------------------------------------------------------------------- #
# .zplot_weighted_rows / .zplot_constant_rows
# ---------------------------------------------------------------------------- #
#
# Row selectors for fitted and extrapolated selection-model zplot densities.
#
# ---------------------------------------------------------------------------- #
.zplot_weighted_rows <- function(selection, extrapolate) {

  if (is.null(selection) || extrapolate) {
    return(integer(0))
  }

  return(which(!selection[["use_normal"]]))
}

.zplot_constant_rows <- function(selection, extrapolate) {

  return(integer(0))
}


# ---------------------------------------------------------------------------- #
# .zplot_selection_context
# ---------------------------------------------------------------------------- #
#
# Prepare posterior-row selection metadata for branch-aware zplot evaluation.
#
# ---------------------------------------------------------------------------- #
.zplot_selection_context <- function(object, data, priors, posterior_samples,
                                      is_weightfunction) {

  if (!is_weightfunction) {
    return(NULL)
  }

  return(.selection_context(
    object            = object,
    posterior_samples = posterior_samples
  ))
}


# ---------------------------------------------------------------------------- #
# .zplot_selection_args
# ---------------------------------------------------------------------------- #
#
# Extract active omega and cutpoints for one posterior row and estimate.
#
# ---------------------------------------------------------------------------- #
.zplot_selection_args <- function(selection, row, estimate, n = 1L) {

  .selection_require_step_evaluable(selection, ".zplot_selection_args()")

  omega <- selection[["omega"]][row, , drop = FALSE]

  if (n > 1L) {
    omega <- matrix(as.numeric(omega), nrow = n, ncol = ncol(omega), byrow = TRUE)
  }

  return(list(
    omega   = omega,
    crit_yi = stats::qnorm(rev(selection[["p_cuts"]])[-c(1L, length(selection[["p_cuts"]]))],
                           lower.tail = FALSE) *
      selection[["sei"]][estimate]
  ))
}


# ============================================================================ #
# Graphical Helper Functions
# ============================================================================ #
#
# These functions extract and set default graphical parameters for zplot
# plotting components. They handle the translation between base graphics
# and ggplot2 parameter naming conventions.
#
# ============================================================================ #


# ---------------------------------------------------------------------------- #
# .get_dots_hist_zplot
# ---------------------------------------------------------------------------- #
#
# Extracts histogram graphical parameters with defaults.
#
# @param dots      list of user-supplied graphical parameters
# @param plot_type "base" or "ggplot"
# @param max_density maximum density value for setting ylim
#
# @return list of graphical parameters appropriate for plot_type
#
# ---------------------------------------------------------------------------- #
.get_dots_hist_zplot <- function(dots, plot_type = "base", max_density = NULL) {
  if (plot_type == "base") {
    if (!is.null(dots[["ylim"]]) && !is.null(max_density)) {
      ylim <- range(dots[["ylim"]], max_density)
    } else if (!is.null(max_density)) {
      ylim <- c(0, max_density)
    } else {
      ylim <- NULL
    }
    hist_dots <- list(
      border = if (!is.null(dots[["border"]])) dots[["border"]] else "gray60",
      col    = if (!is.null(dots[["col"]]))    dots[["col"]]    else NA,
      xlab   = if (!is.null(dots[["xlab"]]))   dots[["xlab"]]   else "Z-Statistic",
      ylab   = if (!is.null(dots[["ylab"]]))   dots[["ylab"]]   else "Density",
      main   = if (!is.null(dots[["main"]]))   dots[["main"]]   else "",
      ylim   = ylim,
      xaxt   = if (!is.null(dots[["xaxt"]]))   dots[["xaxt"]]   else "s",
      yaxt   = if (!is.null(dots[["yaxt"]]))   dots[["yaxt"]]   else "s",
      las    = if (!is.null(dots[["las"]]))    dots[["las"]]    else 1
    )
  } else if (plot_type == "ggplot") {
    hist_dots <- list(
      color = if (!is.null(dots[["color"]])) dots[["color"]] else if (!is.null(dots[["col"]])) dots[["col"]] else "gray60",
      fill  = if (!is.null(dots[["fill"]]))  dots[["fill"]]  else NA,
      alpha = if (!is.null(dots[["alpha"]])) dots[["alpha"]] else 1,
      xlab  = if (!is.null(dots[["xlab"]]))  dots[["xlab"]]  else "Z-Statistic",
      ylab  = if (!is.null(dots[["ylab"]]))  dots[["ylab"]]  else "Density",
      main  = if (!is.null(dots[["main"]]))  dots[["main"]]  else ""
    )
  }
  return(hist_dots)
}


# ---------------------------------------------------------------------------- #
# .get_dots_lines_zplot
# ---------------------------------------------------------------------------- #
#
# Extracts density line graphical parameters with defaults.
#
# @param dots      list of user-supplied graphical parameters
# @param plot_type "base" or "ggplot"
# @param col       default line color
# @param alpha     default alpha for CI ribbons
#
# @return list of graphical parameters appropriate for plot_type
#
# ---------------------------------------------------------------------------- #
.get_dots_lines_zplot <- function(dots, plot_type = "base", col = "black", alpha = 0.40) {
  if (plot_type == "base") {
    line_dots <- list(
      lwd   = if (!is.null(dots[["lwd"]]))   dots[["lwd"]]   else 2,
      col   = if (!is.null(dots[["col"]]))   dots[["col"]]   else col,
      lty   = if (!is.null(dots[["lty"]]))   dots[["lty"]]   else 1,
      alpha = if (!is.null(dots[["alpha"]])) dots[["alpha"]] else alpha
    )
  } else if (plot_type == "ggplot") {
    line_dots <- list(
      linewidth = if (!is.null(dots[["linewidth"]])) dots[["linewidth"]] else if (!is.null(dots[["lwd"]])) dots[["lwd"]] else 1,
      color     = if (!is.null(dots[["color"]]))     dots[["color"]]     else if (!is.null(dots[["col"]])) dots[["col"]] else col,
      linetype  = if (!is.null(dots[["linetype"]])) dots[["linetype"]]  else if (!is.null(dots[["lty"]])) dots[["lty"]] else 1,
      alpha     = if (!is.null(dots[["alpha"]]))    dots[["alpha"]]     else alpha
    )
  }
  return(line_dots)
}


# ---------------------------------------------------------------------------- #
# .get_dots_thresholds_zplot
# ---------------------------------------------------------------------------- #
#
# Extracts threshold line graphical parameters with defaults.
#
# @param dots      list of user-supplied graphical parameters
# @param plot_type "base" or "ggplot"
#
# @return list of graphical parameters appropriate for plot_type
#
# ---------------------------------------------------------------------------- #
.get_dots_thresholds_zplot <- function(dots, plot_type = "base") {
  if (plot_type == "base") {
    threshold_dots <- list(
      col = if (!is.null(dots[["col"]])) dots[["col"]] else "red",
      lty = if (!is.null(dots[["lty"]])) dots[["lty"]] else 3,
      lwd = if (!is.null(dots[["lwd"]])) dots[["lwd"]] else 1
    )
  } else if (plot_type == "ggplot") {
    threshold_dots <- list(
      color     = if (!is.null(dots[["color"]])) dots[["color"]] else if (!is.null(dots[["col"]])) dots[["col"]] else "red",
      linetype  = if (!is.null(dots[["linetype"]])) dots[["linetype"]] else if (!is.null(dots[["lty"]])) dots[["lty"]] else "dashed",
      linewidth = if (!is.null(dots[["linewidth"]])) dots[["linewidth"]] else if (!is.null(dots[["lwd"]])) dots[["lwd"]] else 0.5
    )
  }
  return(threshold_dots)
}


# ---------------------------------------------------------------------------- #
# .get_Soric_FDR
# ---------------------------------------------------------------------------- #
#
# Computes Soric's (1989) False Discovery Rate estimate from EDR.
#
# @param EDR       Expected Discovery Rate (vector or scalar)
# @param sig_level two-sided significance level (e.g., 0.05)
#
# @return Soric FDR estimate
#
# ---------------------------------------------------------------------------- #
.get_Soric_FDR <- function(EDR, sig_level) {

  ((1 / EDR) - 1) * (sig_level / (1 - sig_level))
}


# ---------------------------------------------------------------------------- #
# .zplot_bins
# ---------------------------------------------------------------------------- #
#
# Creates bin sequence for zplot histograms and densities.
#
# For histograms, bins are shifted to align with significance thresholds
# when selection model priors are present. For densities, threshold
# values are added to the sequence to ensure accurate representation.
#
# @param priors     list of priors (used to extract thresholds)
# @param from       lower bound of z-value range
# @param to         upper bound of z-value range
# @param by         step size (NULL to use length.out)
# @param length.out number of bins (NULL to use by)
# @param type       "hist" or "dens"
#
# @return numeric vector of bin boundaries or density evaluation points
#
# ---------------------------------------------------------------------------- #
.zplot_bins      <- function(priors, from, to, by, length.out, type = "hist"){

  if(is.null(length.out)){
    bins <- seq(from = from, to = to, by = by)
  }else{
    bins <- seq(from = from, to = to, length.out = length.out)
  }

  # obtain tresholds from the specified priors
  z_bounds <- .zplot_threshold(priors)

  # return simple binning in case of no weightfunctions
  if(length(z_bounds) == 0){
    return(bins)
  }

  # retain bounds within the plotting range
  z_bounds <- z_bounds[z_bounds > from & z_bounds < to]

  if(type == "hist"){
    # for histogram, shift the specified bin boundaries to the closest z-treshold
    for(i in seq_along(z_bounds)){

      # get index of the first larger sequence
      i_larger <- which(bins > z_bounds[i])[1]

      # if there is none skip
      if(is.na(i_larger))
        next

      # get index of the closer one from below
      i_lower  <- i_larger - 1

      # replace the closer sequence with the boundary
      if(bins[i_larger] - z_bounds[i] < z_bounds[i] - bins[i_lower]){
        bins[i_larger] <- z_bounds[i]
      }else{
        bins[i_lower]  <- z_bounds[i]
      }
    }
  }else if(type == "dens"){
    # for density, extend the specified support at the threshold
    bins <- sort(unique(c(bins, z_bounds)))
  }

  return(bins)
}
.zplot_bias_priors <- function(priors){

  if(is.null(priors)){
    return(list())
  }

  if(!is.null(priors[["outcome"]])){
    return(.selection_bias_priors(priors))
  }

  priors_bias <- priors[["bias"]]
  if(is.null(priors_bias)){
    return(list())
  }
  if(BayesTools::is.prior.mixture(priors_bias)){
    out <- as.list(priors_bias)
    class(out) <- "list"
    return(out)
  }

  return(list(priors_bias))
}
.zplot_threshold <- function(priors){

  priors_bias <- .zplot_bias_priors(priors)

  # return simple binning in case of no bias
  if(length(priors_bias) == 0L){
    return()
  }

  p_bounds <- .selection_kernel_breaks(priors_bias)

  # return simple binning in case of no selection kernels
  if(is.null(p_bounds)){
    return()
  }

  # obtain tresholds from the specified priors
  z_bounds <- stats::qnorm(rev(p_bounds), 0, 1, lower.tail = FALSE)
  z_bounds <- z_bounds[!is.infinite(z_bounds)]

  return(z_bounds)
}

context("zplot")

# Load common test helpers
source(testthat::test_path("common-functions.R"))
source(testthat::test_path("helper-test-matrix.R"))
source(testthat::test_path("helper-visuals.R"))

# list cached fits lazily
skip_if_no_fits()
fit_names <- list_fits()
fits      <- lazy_fits(fit_names, validate = FALSE)
info      <- lazy_infos(fit_names, validate = FALSE)

.test_as_zplot <- function(..., max_samples = 1000) {

  as_zplot(..., max_samples = max_samples)
}

.test_zplot <- function(..., summary_max_samples = 1000, max_samples = 1000) {

  zplot(..., summary_max_samples = summary_max_samples, max_samples = max_samples)
}

.test_plot_zplot <- function(..., max_samples = 1000) {

  plot(..., max_samples = max_samples)
}

.test_lines_zplot <- function(..., max_samples = 1000) {

  lines(..., max_samples = max_samples)
}

.zplot_test_area <- function(df) {

  x <- df[["x"]]
  y <- df[["y"]]

  return(sum(diff(x) * (y[-length(y)] + y[-1L]) / 2))
}


test_that("zplot rejects GLMM fits early", {

  name <- "bcg_BMA.glmm"
  skip_if_missing_fits(name)

  expect_error(
    .test_as_zplot(fits[[name]]),
    "normal outcome models"
  )
})


test_that("zplot creates reusable objects and plots directly", {

  name <- "bcg_meta-analysis"
  skip_if_missing_fits(name)

  fit <- fits[[name]]
  zp  <- .test_as_zplot(fit, max_samples = 1000)

  expect_s3_class(zp, "zplot_brma")
  expect_named(zp[["zplot"]], c("estimates", "data"))
  expect_true(.is_ggplot(.test_zplot(
    fit,
    plot_type           = "ggplot",
    summary_max_samples = 1000,
    max_samples = 1000
  )))
  expect_true(.is_ggplot(.test_zplot(
    zp,
    plot_type   = "ggplot",
    max_samples = 1000
  )))
})


# ============================================================================ #
# Test: Simple Meta-Analysis Zplot
# ============================================================================ #

test_that("zplot for simple meta-analysis renders base and ggplot output", {

  name     <- "bcg_meta-analysis"
  fit_brma <- fits[[name]]
  zc       <- .test_as_zplot(fit_brma)

  # --------------------------------------------------
  # Visual tests
  # --------------------------------------------------

  expect_vdiffr_snapshot("zplot_simple_base", function() {
    suppressMessages(.test_plot_zplot(zc, plot_type = "base"))
  })

  expect_vdiffr_snapshot(
    "zplot_simple_ggplot",
    suppressMessages(.test_plot_zplot(zc, plot_type = "ggplot"))
  )
})


# ============================================================================ #
# Test: Zplot Customization
# ============================================================================ #

test_that("zplot customization snapshots are stable", {

  skip_if_not_full_visuals("Customization snapshots are visual-gallery coverage.")

  name     <- "bcg_meta-analysis"
  zc       <- .test_as_zplot(fits[[name]])

  # --------------------------------------------------
  # Test custom styles
  # --------------------------------------------------

  expect_vdiffr_snapshot("zplot_custom_base", function() {
    suppressMessages(.test_plot_zplot(
      zc, plot_type = "base",
      plot_fit = TRUE, plot_ci = TRUE,
      lwd = 2, lty = 2,           # line args
      dots_hist = list(col = "lightblue"),
      main = "Custom Zplot"
    ))
  })

  # For ggplot, we pass specific list args for hist/lines as per .get_dots_* in implementation
  # or global args that get mapped
  expect_vdiffr_snapshot(
    "zplot_custom_ggplot",
    suppressMessages(.test_plot_zplot(
      zc, plot_type = "ggplot",
      dots_hist = list(fill = "lightblue", color = "blue"),
      dots_thresholds = list(color = "red", linetype = "dashed"),
      main = "Custom Zplot GGplot"
    ))

  )

  # --------------------------------------------------
  # Test components only (hist / lines)
  # --------------------------------------------------

  expect_vdiffr_snapshot("zplot_hist_only_base", function() {
    suppressMessages(hist(zc, plot_type = "base", main = "Hist Only"))
  })

  expect_vdiffr_snapshot("zplot_lines_only_base", function() {
    # lines() adds to existing plot usually, but here we test the function
    # so we create an empty plot and add lines
    plot(0, 0, type = "n", xlim = c(-6, 6), ylim = c(0, 0.5), main = "Lines Only")
    .test_lines_zplot(zc, plot_type = "base", col = "purple")
  })

})


# ============================================================================ #
# Test: Meta-Regression Zplot
# ============================================================================ #

test_that("zplot for meta-regression renders base output", {

  name <- "bcg_meta-regression"
  zc   <- .test_as_zplot(fits[[name]])

  expect_vdiffr_snapshot("zplot_regression_base", function() {
    suppressMessages(.test_plot_zplot(zc, plot_type = "base", main = "Meta-Regression Zplot"))
  })

})

# ============================================================================ #
# Test: Selection Models Zplot
# ============================================================================ #

test_that("zplot for positive-direction selection model renders base output", {

  name <- "dat.lehmann2018-3PSM"
  zc   <- .test_as_zplot(fits[[name]])

  expect_vdiffr_snapshot("zplot_selection_pos_base", function() {
    suppressMessages(.test_plot_zplot(zc, plot_type = "base", main = "Selection Model (Pos) Zplot"))
  })

})

test_that("zplot for negative-direction selection model renders base output", {

  skip_if_not_full_visuals("Negative-direction selection zplot is gallery coverage.")

  name <- "dat.lehmann2018-3PSM_neg"
  zc   <- .test_as_zplot(fits[[name]])

  expect_vdiffr_snapshot("zplot_selection_neg_base", function() {
    suppressMessages(.test_plot_zplot(zc, plot_type = "base", main = "Selection Model (Neg) Zplot"))
  })

})

test_that("zplot handles RoBMA bias-mixture branches", {

  name <- "dat.lehmann2018_RoBMA"
  skip_if_not(name %in% names(fits), "RoBMA cached fit not available.")

  fit               <- fits[[name]]
  posterior_samples <- .get_posterior_samples(fit[["fit"]])
  if (nrow(posterior_samples) > 1000) {
    selected_ind      <- round(seq(from = 1, to = nrow(posterior_samples), length.out = 1000))
    posterior_samples <- posterior_samples[selected_ind, , drop = FALSE]
  }
  selection         <- .zplot_selection_context(
    object            = fit,
    data              = fit[["data"]],
    priors            = fit[["priors"]],
    posterior_samples = posterior_samples,
    is_weightfunction = .is_weightfunction(fit)
  )
  weighted_rows <- which(!selection[["use_normal"]])
  skip_if_not(length(weighted_rows) > 0, "No weightfunction posterior rows in cached fit.")

  selection_args <- .zplot_selection_args(
    selection = selection,
    row       = weighted_rows[1],
    estimate  = 1,
    n         = 3
  )
  active_cuts <- selection[["n_bins"]] - 1L

  expect_equal(nrow(selection_args[["omega"]]), 3)
  expect_equal(ncol(selection_args[["omega"]]), selection[["n_bins"]])
  expect_length(selection_args[["crit_yi"]], active_cuts)

  zc <- .test_as_zplot(fit, max_samples = 1000)
  expect_true(all(is.finite(zc[["zplot"]][["estimates"]][["EDR"]])))
  expect_true(all(zc[["zplot"]][["estimates"]][["EDR"]] >= 0))
  expect_true(all(zc[["zplot"]][["estimates"]][["EDR"]] <= 1))
  expect_true(all(is.finite(zc[["zplot"]][["estimates"]][["weights"]])))

  fitted_density <- .test_lines_zplot(
    zc, as_data = TRUE, max_samples = 1000, plot_ci = FALSE,
    extrapolate = FALSE, length.out = 25
  )
  extrapolated_density <- .test_lines_zplot(
    zc, as_data = TRUE, max_samples = 1000, plot_ci = FALSE,
    extrapolate = TRUE, length.out = 25
  )

  expect_true(all(is.finite(unlist(fitted_density[c("y", "y_lCI", "y_uCI")]))))
  expect_true(all(is.finite(unlist(extrapolated_density[c("y", "y_lCI", "y_uCI")]))))

  fitted_area       <- .zplot_test_area(.test_lines_zplot(
    zc,
    as_data     = TRUE,
    max_samples = 1000,
    plot_ci     = FALSE,
    extrapolate = FALSE,
    from        = -20,
    to          = 20,
    length.out  = 2001
  ))
  extrapolated_area <- .zplot_test_area(.test_lines_zplot(
    zc,
    as_data     = TRUE,
    max_samples = 1000,
    plot_ci     = FALSE,
    extrapolate = TRUE,
    from        = -20,
    to          = 20,
    length.out  = 2001
  ))
  expected_area     <- mean(zc[["zplot"]][["estimates"]][["weights"]])

  expect_equal(fitted_area, 1, tolerance = 0.01)
  expect_equal(extrapolated_area, expected_area, tolerance = 0.01)
  expect_gt(extrapolated_area, fitted_area)
})

# ============================================================================ #
# Test: PET Models Zplot
# ============================================================================ #

test_that("zplot for positive-direction PET model renders base output", {

  name <- "dat.lehmann2018-PET"
  zc   <- .test_as_zplot(fits[[name]])

  expect_vdiffr_snapshot("zplot_PET_pos_base", function() {
    suppressMessages(.test_plot_zplot(zc, plot_type = "base", main = "PET Model (Pos) Zplot"))
  })

})

test_that("zplot for negative-direction PET model renders base output", {

  skip_if_not_full_visuals("Negative-direction PET zplot is gallery coverage.")

  name <- "dat.lehmann2018-PET_neg"
  zc   <- .test_as_zplot(fits[[name]])

  expect_vdiffr_snapshot("zplot_PET_neg_base", function() {
    suppressMessages(.test_plot_zplot(zc, plot_type = "base", main = "PET Model (Neg) Zplot"))
  })

})

# ============================================================================ #
# Test: Multilevel Models Zplot
# ============================================================================ #

test_that("zplot for multilevel model renders base output", {

  skip_if_not_full_visuals("Multilevel zplot is gallery coverage.")

  name <- "konstantopoulos2011_3lvl"
  zc   <- .test_as_zplot(fits[[name]])

  expect_vdiffr_snapshot("zplot_multilevel_base", function() {
    suppressMessages(.test_plot_zplot(zc, plot_type = "base", main = "Multilevel Model Zplot"))
  })

})

# ============================================================================ #
# brma.hat_matrix.R
# ============================================================================ #
#
# This file contains internal functions for computing the hat matrix and related
# quantities for brma objects. These functions are shared between residuals()
# (for rstandard) and hatvalues().
#
# ============================================================================ #


# ---------------------------------------------------------------------------- #
# .compute_hat_matrix_samples
# ---------------------------------------------------------------------------- #
#
# Computes hat matrix components for each posterior sample.
#
# Returns a list containing:
# - H_diag: S x K matrix of hat matrix diagonals (leverages)
# - H: S x K x K array of full hat matrices (only if return_full_H = TRUE)
# - se_components: S x K matrix of standard errors components for residuals (if return_se = TRUE)
# - resid_components: S x K matrix of GLS residuals (if return_resid = TRUE)
# - M_diag: S x K matrix of marginal variance diagonals
#
# @param object brma object
# @param conditioning_depth character; conditioning depth for residual SEs:
#                           - "marginal" (default): Uses marginal variance M
#                           - "cluster": Uses within-cluster variance M_within
#                           - "estimate": Uses sampling variance V
# @param return_full_H logical; whether to return the full hat matrix for each sample
# @param return_se logical; whether to compute and return standard error components
# @param return_resid logical; whether to compute and return GLS residual components
#
# ---------------------------------------------------------------------------- #
.compute_hat_matrix_samples <- function(object, conditioning_depth = "marginal",
                                        return_full_H = FALSE,
                                        return_se = FALSE, return_resid = FALSE,
                                        summarize = FALSE) {
  # check inputs
  conditioning_depth <- .normalize_conditioning_depth(conditioning_depth)

  # get model characteristics
  is_multilevel <- .is_multilevel(object)
  is_scale      <- .is_scale(object)
  priors        <- object[["priors"]]

  # get observed data
  outcome_data <- object[["data"]][["outcome"]]
  yi           <- .outcome_data_yi(object)
  vi           <- .outcome_data_vi(object)
  weights      <- .outcome_data_weights(object)
  K            <- length(yi)

  # get design matrix X
  X <- .get_model_matrix(object)

  # get tau samples (heterogeneity)
  # tau_within: estimate-level heterogeneity (used for cluster and estimate residuals)
  # tau_between: cluster-level heterogeneity (only for multilevel models)
  tau_result <- .evaluate.brma.tau(
    fit           = object[["fit"]],
    scale_data    = object[["data"]][["scale"]],
    scale_formula = if (is_scale) .create_fit_formula_list(data = object[["data"]], "scale") else NULL,
    scale_priors  = priors[["scale"]],
    is_scale      = is_scale,
    is_multilevel = is_multilevel,
    K             = K
  )
  tau_within_samples  <- tau_result[["tau_within"]]
  tau_between_samples <- tau_result[["tau_between"]]

  S <- nrow(tau_within_samples)

  # get cluster for multilevel structure
  ids <- NULL
  if (is_multilevel) {
    ids <- outcome_data[["cluster"]]
  }

  # initialize outputs
  H_diag_samples <- matrix(0, nrow = S, ncol = K)
  H_samples      <- if (return_full_H) array(0, dim = c(S, K, K)) else NULL
  se_samples     <- if (return_se && !summarize) matrix(0, nrow = S, ncol = K) else NULL
  resid_samples  <- if (return_resid && !summarize) matrix(0, nrow = S, ncol = K) else NULL
  se_sum         <- if (return_se && summarize) rep(0, K) else NULL
  resid_sum      <- if (return_resid && summarize) rep(0, K) else NULL
  z_sum          <- if (return_se && return_resid && summarize) rep(0, K) else NULL
  M_diag_samples <- matrix(0, nrow = S, ncol = K) # useful debug/checking
  I_K            <- diag(K)
  residual_tol   <- 100 * .Machine$double.eps * max(1, max(abs(yi)))

  # Pre-calculate indices for multilevel blocks to avoid repeating inside loop
  block_indices <- list()
  if (is_multilevel) {
    block_indices <- .get_multilevel_block_indices(ids)
  }

  for (s in seq_len(S)) {
    tau_w_s             <- tau_within_samples[s, ]
    tau_b_s             <- tau_between_samples[s, ]
    sampling_diagonal_s <- vi / weights
    diagonal_s          <- (vi + tau_w_s^2) / weights
    M_diag_s            <- diagonal_s

    if (is_multilevel) {
      M_diag_s <- M_diag_s + tau_b_s^2
    }

    M_diag_samples[s, ] <- M_diag_s

    WX <- .hat_apply_precision(
      x             = X,
      diagonal      = diagonal_s,
      rank_one      = if (is_multilevel) tau_b_s else NULL,
      block_indices = block_indices
    )
    Wy <- .hat_apply_precision(
      x             = yi,
      diagonal      = diagonal_s,
      rank_one      = if (is_multilevel) tau_b_s else NULL,
      block_indices = block_indices
    )

    XtWX     <- crossprod(X, WX)
    XtWX_inv <- .hat_solve_crossprod(XtWX)
    XB       <- X %*% XtWX_inv
    Q_diag   <- rowSums(XB * X)

    H_diag_samples[s, ] <- rowSums(XB * WX)

    if (return_full_H) {
      H_samples[s, , ] <- X %*% XtWX_inv %*% t(WX)
    }

    if (return_se || return_resid) {
      residual_s <- NULL
      se_s       <- NULL

      beta_hat <- as.vector(XtWX_inv %*% crossprod(X, Wy))
      residual <- yi - as.vector(X %*% beta_hat)

      if (conditioning_depth == "estimate") {
        weighted_residual <- .hat_apply_precision(
          x             = residual,
          diagonal      = diagonal_s,
          rank_one      = if (is_multilevel) tau_b_s else NULL,
          block_indices = block_indices
        )
        residual <- sampling_diagonal_s * weighted_residual

      } else if (conditioning_depth == "cluster" && is_multilevel) {
        weighted_residual <- .hat_apply_precision(
          x             = residual,
          diagonal      = diagonal_s,
          rank_one      = tau_b_s,
          block_indices = block_indices
        )
        cluster_adjust <- rep(0, K)
        for (idx in block_indices) {
          cluster_adjust[idx] <- tau_b_s[idx] * sum(tau_b_s[idx] * weighted_residual[idx])
        }
        residual <- residual - cluster_adjust
      }

      if (return_resid) {
        residual[abs(residual) < residual_tol] <- 0
        residual_s <- residual
        if (summarize) {
          resid_sum <- resid_sum + residual
        } else {
          resid_samples[s, ] <- residual
        }
      }

      if (return_se) {
        if (conditioning_depth == "marginal" ||
            (conditioning_depth == "cluster" && !is_multilevel)) {
          se2 <- M_diag_s - Q_diag

        } else if (conditioning_depth == "estimate") {
          W_diag  <- .hat_precision_diag(
            diagonal      = diagonal_s,
            rank_one      = if (is_multilevel) tau_b_s else NULL,
            block_indices = block_indices
          )
          QW_diag <- rowSums((WX %*% XtWX_inv) * WX)
          se2     <- sampling_diagonal_s^2 * (W_diag - QW_diag)

        } else {
          se2 <- .hat_cluster_se2(
            X             = X,
            XtWX_inv      = XtWX_inv,
            diagonal      = diagonal_s,
            tau_between   = tau_b_s,
            vi            = sampling_diagonal_s,
            block_indices = block_indices,
            I_K           = I_K
          )
        }

        se_s <- sqrt(pmax(se2, 0))
        if (summarize) {
          se_sum <- se_sum + se_s
        } else {
          se_samples[s, ] <- se_s
        }
      }

      if (summarize && return_se && return_resid) {
        z_s <- residual_s / se_s
        z_s[residual_s == 0 & se_s == 0] <- 0
        z_sum <- z_sum + z_s
      }
    }
  }

  result <- list(
    H_diag = H_diag_samples,
    M_diag = M_diag_samples
  )

  if (return_full_H) {
    result[["H"]] <- H_samples
  }
  if (return_se) {
    result[["se"]] <- if (summarize) se_sum / S else se_samples
  }
  if (return_resid) {
    result[["resid"]] <- if (summarize) resid_sum / S else resid_samples
  }
  if (summarize && return_se && return_resid) {
    result[["z"]] <- z_sum / S
  }

  return(result)
}


# ---------------------------------------------------------------------------- #
# .hat_apply_precision
# ---------------------------------------------------------------------------- #
#
# Apply the inverse marginal covariance to a vector or matrix.
#
# ---------------------------------------------------------------------------- #
.hat_apply_precision <- function(x, diagonal, rank_one = NULL, block_indices = list()) {

  was_vector <- is.null(dim(x))
  if (was_vector) {
    x <- matrix(x, ncol = 1)
  }

  out <- matrix(0, nrow = nrow(x), ncol = ncol(x))

  if (is.null(rank_one)) {
    out <- x / diagonal
  } else {
    for (idx in block_indices) {
      inv_diag <- 1 / diagonal[idx]
      inv_x    <- x[idx, , drop = FALSE] * inv_diag
      inv_u    <- rank_one[idx] * inv_diag
      denom    <- 1 + sum(rank_one[idx] * inv_u)

      out[idx, ] <- inv_x -
        tcrossprod(inv_u, colSums(rank_one[idx] * inv_x) / denom)
    }
  }

  if (was_vector) {
    return(as.vector(out))
  }
  return(out)
}


# ---------------------------------------------------------------------------- #
# .hat_precision_diag
# ---------------------------------------------------------------------------- #
#
# Diagonal of the inverse marginal covariance.
#
# ---------------------------------------------------------------------------- #
.hat_precision_diag <- function(diagonal, rank_one = NULL, block_indices = list()) {

  if (is.null(rank_one)) {
    return(1 / diagonal)
  }

  out <- rep(NA_real_, length(diagonal))
  for (idx in block_indices) {
    inv_diag <- 1 / diagonal[idx]
    inv_u    <- rank_one[idx] * inv_diag
    denom    <- 1 + sum(rank_one[idx] * inv_u)

    out[idx] <- inv_diag - inv_u^2 / denom
  }

  return(out)
}


# ---------------------------------------------------------------------------- #
# .hat_precision_matrix
# ---------------------------------------------------------------------------- #
#
# Build the full inverse marginal covariance only for rare full-matrix paths.
#
# ---------------------------------------------------------------------------- #
.hat_precision_matrix <- function(diagonal, rank_one = NULL, block_indices = list()) {

  K <- length(diagonal)
  W <- matrix(0, nrow = K, ncol = K)

  if (is.null(rank_one)) {
    diag(W) <- 1 / diagonal
    return(W)
  }

  for (idx in block_indices) {
    inv_diag <- 1 / diagonal[idx]
    inv_u    <- rank_one[idx] * inv_diag
    denom    <- 1 + sum(rank_one[idx] * inv_u)

    W[idx, idx] <- diag(inv_diag, nrow = length(idx), ncol = length(idx)) -
      tcrossprod(inv_u) / denom
  }

  return(W)
}


# ---------------------------------------------------------------------------- #
# .hat_solve_crossprod
# ---------------------------------------------------------------------------- #
#
# Stable inverse for the small fixed-effect crossproduct.
#
# ---------------------------------------------------------------------------- #
.hat_solve_crossprod <- function(x) {

  chk <- try(chol(x), silent = TRUE)
  if (!inherits(chk, "try-error")) {
    return(chol2inv(chk))
  }

  return(tryCatch(solve(x), error = function(e) MASS::ginv(x)))
}


# ---------------------------------------------------------------------------- #
# .hat_cluster_se2
# ---------------------------------------------------------------------------- #
#
# Fallback cluster-level residual variance using block-structured W and M.
#
# ---------------------------------------------------------------------------- #
.hat_cluster_se2 <- function(X, XtWX_inv, diagonal, tau_between, vi,
                             block_indices, I_K) {

  K <- length(vi)
  W <- .hat_precision_matrix(
    diagonal      = diagonal,
    rank_one      = tau_between,
    block_indices = block_indices
  )
  M <- .build_multilevel_marginal_covariance(
    tau_within    = sqrt(pmax(diagonal - vi, 0)),
    tau_between   = tau_between,
    vi            = vi,
    block_indices = block_indices
  )

  between_cov <- matrix(0, nrow = K, ncol = K)
  for (idx in block_indices) {
    between_cov[idx, idx] <- tcrossprod(tau_between[idx])
  }

  Q <- X %*% XtWX_inv %*% t(X)
  A <- (I_K - between_cov %*% W)

  return(diag(A %*% (M - Q) %*% t(A)))
}

# ============================================================================ #
# Outcome CDF Functions (Cumulative Distribution)
# ============================================================================ #
#
# These functions compute pointwise CDF values F(yi | theta) for each
# observation and posterior sample. The target-specific estimate-unit CDF is
# used for LOO-PIT residuals via probability integral transformation.
#
# Parallels the structure of pdf.R but returns CDF values instead of
# density values.
#
# Note: For binomial and Poisson models, each "observation" consists of a pair
# of data points (ai+ci or x1i+x2i) that together define a single effect size
# estimate. The CDF is computed using the implied normal approximation for the
# effect size (log-OR or log-IRR).
#
# ============================================================================ #


# ---------------------------------------------------------------------------- #
# .outcome_cdf.norm
# ---------------------------------------------------------------------------- #
#
# Compute pointwise CDF values for normal outcome models.
#
# For normal outcome models, the observed effect y_i has distribution:
#   y_i ~ N(mu_i, tau_within^2 + se_i^2)
#
# This function computes F(y_i | mu_i, sigma_i) for each observation at each
# posterior sample.
#
# @param yi               numeric vector of length K; observed effect sizes
# @param mu_samples       S x K matrix of location samples (with cluster effects if multilevel)
# @param tau_within       S x K matrix of estimate-level heterogeneity samples
# @param sei              numeric vector of length K; standard errors
# @param lower.tail       logical; return P(Y <= yi) if TRUE, P(Y > yi) if FALSE
#
# @return S x K matrix of CDF values in (0, 1)
#
# ---------------------------------------------------------------------------- #
.outcome_cdf.norm <- function(yi, mu_samples, tau_within, sei,
                              lower.tail = TRUE) {

  S <- nrow(mu_samples)
  K <- ncol(mu_samples)

  # replicate yi and sei across samples: matrix(vec, S, K, byrow = TRUE)
  yi_mat  <- matrix(yi,  nrow = S, ncol = K, byrow = TRUE)
  sei_mat <- matrix(sei, nrow = S, ncol = K, byrow = TRUE)

  # compute total SD: sqrt(tau^2 + se^2)
  total_sd <- sqrt(tau_within^2 + sei_mat^2)

  # compute CDF value for each cell
  cdf_vals <- stats::pnorm(
    yi_mat,
    mean       = mu_samples,
    sd         = total_sd,
    lower.tail = lower.tail
  )

  return(cdf_vals)
}


# ---------------------------------------------------------------------------- #
# .outcome_cdf.selnorm
# ---------------------------------------------------------------------------- #
#
# Compute pointwise CDF values for the selected-normal kernel.
#
# ---------------------------------------------------------------------------- #
.outcome_cdf.selnorm <- function(yi, mu_samples, tau_within, sei,
                                 selection_context, lower.tail = TRUE) {

  S         <- nrow(mu_samples)
  K         <- ncol(mu_samples)
  sei_mat   <- matrix(sei, nrow = S, ncol = K, byrow = TRUE)
  total_sd  <- sqrt(tau_within^2 + sei_mat^2)

  return(.selection_step_cdf_matrix(
    q                 = yi,
    mean              = mu_samples,
    sd                = total_sd,
    sei               = sei,
    selection_context = selection_context,
    lower.tail        = lower.tail
  ))
}


# ---------------------------------------------------------------------------- #
# .outcome_cdf.binom
# ---------------------------------------------------------------------------- #
#
# Compute pointwise CDF values for binomial outcome models.
#
# For binomial outcome models, we use a normal approximation based on the
# implied log-odds ratio effect size and its approximate sampling variance.
# This approach is consistent with how metafor computes residuals for GLMM.
#
# The scalar residual is computed on the approximate log-odds-ratio scale:
#   y_i ~ N(mu_i, tau_within_i^2 + sigma_i^2)
# where sigma_i is the approximate sampling SE from the cell counts.
#
# @param yi               numeric vector of length K; approximate log-OR effect sizes
# @param sei              numeric vector of length K; approximate sampling SEs
# @param mu_samples       S x K matrix of log-odds ratio samples
# @param tau_within       S x K matrix of estimate-level heterogeneity samples
#
# @return S x K matrix of CDF values (one per estimate)
#
# ---------------------------------------------------------------------------- #
.outcome_cdf.binom <- function(yi, sei, mu_samples, tau_within) {
  return(.outcome_cdf.norm(
    yi         = yi,
    mu_samples = mu_samples,
    tau_within = tau_within,
    sei        = sei
  ))
}


# ---------------------------------------------------------------------------- #
# .outcome_cdf.pois
# ---------------------------------------------------------------------------- #
#
# Compute pointwise CDF values for Poisson outcome models.
#
# For Poisson outcome models, we use a normal approximation based on the
# implied log incidence rate ratio effect size and its approximate sampling
# variance. This is consistent with how metafor computes residuals for GLMM.
#
# The scalar residual is computed on the approximate log-incidence-rate-ratio
# scale:
#   y_i ~ N(mu_i, tau_within_i^2 + sigma_i^2)
# where sigma_i is the approximate sampling SE from the counts.
#
# @param yi               numeric vector of length K; approximate log-IRR effect sizes
# @param sei              numeric vector of length K; approximate sampling SEs
# @param mu_samples       S x K matrix of log-IRR samples
# @param tau_within       S x K matrix of estimate-level heterogeneity samples
#
# @return S x K matrix of CDF values (one per estimate)
#
# ---------------------------------------------------------------------------- #
.outcome_cdf.pois <- function(yi, sei, mu_samples, tau_within) {
  return(.outcome_cdf.norm(
    yi         = yi,
    mu_samples = mu_samples,
    tau_within = tau_within,
    sei        = sei
  ))
}


# ---------------------------------------------------------------------------- #
# .cdf.brma
# ---------------------------------------------------------------------------- #
#
# Compute the full CDF matrix for a brma object.
#
# This function computes pointwise CDF values F(yi | mu, tau) for each
# observation and posterior sample. It uses predict.brma to obtain the
# appropriate mu samples at the requested conditioning depth.
#
# @param object brma object
# @param conditioning_depth character; conditioning depth. Options are:
#                           - "marginal" (default): Fixed effects only (mu).
#                             CDF: yi ~ N(mu_i, tau^2 + sei^2)
#                           - "cluster": Fixed effects + cluster-level random
#                             effects (mu + gamma). Only available for
#                             multilevel models. CDF:
#                             yi ~ N(mu_i + gamma_j, tau_within^2 + sei^2)
#                           - "estimate": True estimate effects
#                             (mu + gamma + theta). CDF: yi ~ N(theta_i, sei^2)
#
# @return S x K matrix of CDF values in (0, 1)
#
# ---------------------------------------------------------------------------- #
.cdf.brma <- function(object, conditioning_depth = "marginal") {

  ### input validation
  conditioning_depth <- .normalize_conditioning_depth(conditioning_depth)

  ### extract structural information about the model
  priors            <- object[["priors"]]
  data              <- object[["data"]]
  is_multilevel     <- .is_multilevel(object)
  is_scale          <- .is_scale(object)
  is_weightfunction <- .is_weightfunction(object)
  outcome_type      <- .outcome_type(object)
  effect_direction  <- .effect_direction(object)
  posterior_samples <- .get_posterior_samples(object[["fit"]])

  .check_unit_conditioning_depth(
    object             = object,
    unit               = "estimate",
    conditioning_depth = conditioning_depth,
    caller             = ".cdf.brma()"
  )

  ### obtain observed effect sizes and sampling SEs
  yi  <- .outcome_data_yi(object)
  sei <- .outcome_data_sei(object)
  K   <- length(yi)

  ### get mu samples at the appropriate conditioning depth using predict.brma
  # map CDF conditioning depths to predict.brma types
  predict_type <- switch(conditioning_depth,
    "marginal"  = "terms",
    "cluster"   = "cluster",
    "estimate"  = "estimate"
  )

  mu_samples <- predict.brma(
    object  = object,
    newdata = NULL,
    type    = predict_type,
    quiet   = TRUE
  )
  S <- nrow(mu_samples)

  ### determine tau component for CDF computation
  tau_result <- .evaluate.brma.tau(
    fit               = object[["fit"]],
    scale_data        = object[["data"]][["scale"]],
    scale_formula     = if (is_scale) .create_fit_formula_list(data = object[["data"]], "scale") else NULL,
    scale_priors      = priors[["scale"]],
    is_scale          = is_scale,
    is_multilevel     = is_multilevel,
    K                 = K,
    posterior_samples = posterior_samples
  )

  if (conditioning_depth == "estimate") {
    tau_within_samples <- matrix(0, nrow = S, ncol = K)
  } else if (conditioning_depth == "cluster") {
    tau_within_samples <- tau_result[["tau_within"]]
  } else {
    tau_within_samples <- tau_result[["tau_total"]]
  }

  ### compute CDF based on outcome type
  if (outcome_type == "norm") {

    # flip for negative effect direction
    if (effect_direction == "negative") {
      mu_samples_cdf <- -mu_samples
      yi_cdf         <- -yi
      lower_tail     <- FALSE
    } else {
      mu_samples_cdf <- mu_samples
      yi_cdf         <- yi
      lower_tail     <- TRUE
    }

    # dispatch between weighted and standard normal
    if (is_weightfunction) {
      selection_context <- .selection_context(
        object            = object,
        posterior_samples = posterior_samples
      )

      cdf_vals <- .outcome_cdf.selnorm(
        yi                = yi,
        mu_samples        = mu_samples,
        tau_within        = tau_within_samples,
        sei               = sei,
        selection_context = selection_context,
        lower.tail        = TRUE
      )

    } else {

      # standard normal CDF
      cdf_vals <- .outcome_cdf.norm(
        yi         = yi_cdf,
        mu_samples = mu_samples_cdf,
        tau_within = tau_within_samples,
        sei        = sei,
        lower.tail = lower_tail
      )

    }

  } else if (outcome_type == "bin") {

    # binomial CDF using normal approximation
    cdf_vals <- .outcome_cdf.binom(
      yi         = yi,
      sei        = sei,
      mu_samples = mu_samples,
      tau_within = tau_within_samples
    )

  } else if (outcome_type == "pois") {

    # Poisson CDF using normal approximation
    cdf_vals <- .outcome_cdf.pois(
      yi         = yi,
      sei        = sei,
      mu_samples = mu_samples,
      tau_within = tau_within_samples
    )

  } else {

    stop("Unsupported outcome type for CDF computation.", call. = FALSE)

  }

  # add column names for observations
  colnames(cdf_vals) <- paste0("cdf[", seq_len(K), "]")

  return(cdf_vals)
}


# ---------------------------------------------------------------------------- #
# .cdf_lik_estimate.brma
# ---------------------------------------------------------------------------- #
#
# Compute CDF values for the estimate-unit LOO target.
#
# This mirrors `.log_lik_estimate.brma()`: fixed effects plus fitted cluster
# effects for multilevel models, marginal over estimate-level heterogeneity.
# It is used by LOO-PIT residuals so the PSIS weights and CDF target match.
#
# @param object brma object.
#
# @return S x K matrix of CDF values in (0, 1)
#
# ---------------------------------------------------------------------------- #
.cdf_lik_estimate.brma <- function(object) {

  setup             <- .estimate_likelihood_setup.brma(object)
  yi                <- setup[["yi"]]
  sei               <- setup[["sei"]]
  K                 <- setup[["K"]]
  mu_samples        <- setup[["mu"]]
  tau_within        <- setup[["tau_within"]]
  outcome_type      <- setup[["outcome_type"]]
  is_weightfunction <- setup[["is_weightfunction"]]
  effect_direction  <- setup[["effect_direction"]]
  posterior_samples <- setup[["posterior_samples"]]

  if (outcome_type == "norm") {

    if (effect_direction == "negative") {
      mu_samples_cdf <- -mu_samples
      yi_cdf         <- -yi
      lower_tail     <- FALSE
    } else {
      mu_samples_cdf <- mu_samples
      yi_cdf         <- yi
      lower_tail     <- TRUE
    }

    if (is_weightfunction) {
      selection_context <- .selection_context(
        object            = object,
        posterior_samples = posterior_samples
      )

      cdf_vals <- .outcome_cdf.selnorm(
        yi                = yi,
        mu_samples        = mu_samples,
        tau_within        = tau_within,
        sei               = sei,
        selection_context = selection_context,
        lower.tail        = TRUE
      )

    } else {

      cdf_vals <- .outcome_cdf.norm(
        yi         = yi_cdf,
        mu_samples = mu_samples_cdf,
        tau_within = tau_within,
        sei        = sei,
        lower.tail = lower_tail
      )
    }

  } else if (outcome_type == "bin") {

    cdf_vals <- .outcome_cdf.binom(
      yi         = yi,
      sei        = sei,
      mu_samples = mu_samples,
      tau_within = tau_within
    )

  } else if (outcome_type == "pois") {

    cdf_vals <- .outcome_cdf.pois(
      yi         = yi,
      sei        = sei,
      mu_samples = mu_samples,
      tau_within = tau_within
    )

  } else {

    stop("Unsupported outcome type for estimate-unit CDF computation.",
         call. = FALSE)
  }

  colnames(cdf_vals) <- paste0("cdf_lik[", seq_len(K), "]")

  return(cdf_vals)
}

# ============================================================================ #
# brma.qqnorm.R
# ============================================================================ #
#
# Normal QQ plot functions for brma objects.
#
# The QQ plot displays sorted standardized residuals against theoretical
# standard normal quantiles. Under a correctly specified model, residuals
# should fall approximately along the y = x diagonal reference line.
#
# Supports deterministic pointwise confidence envelopes from normal order
# statistic distributions.
#
# ============================================================================ #


#' @title Normal QQ Plot for brma Object
#'
#' @description \code{qqnorm.brma} creates a normal QQ plot for a fitted brma
#' object. The plot displays sorted standardized residuals on the vertical axis
#' against the theoretical quantiles of a standard normal distribution on the
#' horizontal axis. Under a correctly specified model, the points should fall
#' approximately along the diagonal reference line.
#'
#' @param y a fitted brma object.
#' @param type the type of standardized residuals to use. Options are:
#' \itemize{
#'   \item \code{"rstudent"} (alias: \code{"LOO-PIT"}; default): Leave-one-out
#'     probability integral transform residuals. Works for all model types
#'     including GLMMs and selection models. This is the recommended type for
#'     Bayesian models as it properly accounts for estimation uncertainty and
#'     leverage. Note: requires that loo has been computed (see
#'     \code{\link{add_loo}}).
#'   \item \code{"rstandard"}: Internally standardized residuals using the hat
#'     matrix. Only available for normal outcome models without selection
#'     (weightfunction) bias adjustment.
#' }
#' @param unit output unit. Only \code{"estimate"} is implemented currently.
#' @param conditioning_depth residual conditioning depth. Options are
#' \code{"marginal"}, \code{"cluster"}, and \code{"estimate"}.
#' @param envelope logical indicating whether to add a pseudo confidence
#' envelope to the plot. The envelope is created by simulating sets of
#' standard normal quantiles, providing a reference for the expected variability
#' under a correctly specified model. Defaults to \code{TRUE}.
#' @param conf_level numeric value between 0 and 100 specifying the confidence level
#' for the envelope bounds. Defaults to \code{95}.
#' @param bonferroni logical indicating whether to apply Bonferroni correction
#' to the envelope, so it can be regarded as a simultaneous confidence region
#' for all \eqn{k} residuals. Defaults to \code{FALSE}.
#' @param reps retained for compatibility with earlier simulation-based
#' envelopes. The current envelope is deterministic and does not use this
#' argument.
#' @param smooth logical indicating whether to smooth the envelope bounds using
#' Friedman's SuperSmoother (\code{\link[stats]{supsmu}}). Defaults to
#' \code{TRUE}.
#' @param xlim x-axis limits. If not specified, limits are computed from data.
#' @param ylim y-axis limits. If not specified, limits are computed from data.
#' @param xlab title for the x-axis. If not specified, defaults to
#' \code{"Theoretical Quantiles"}.
#' @param ylab title for the y-axis. If not specified, defaults to
#' \code{"Sample Quantiles"}.
#' @param plot_type whether to use a base plot \code{"base"} or ggplot2
#' \code{"ggplot"} for plotting. Defaults to \code{"base"}.
#' @param ... additional graphical arguments to customize the plot appearance:
#' \describe{
#'   \item{pch}{point symbol (default: 21, filled circle)}
#'   \item{col}{point border color (default: "black")}
#'   \item{bg}{point fill/background color (default: "#A6A6A6")}
#'   \item{cex}{point size multiplier for base graphics (default: 1)}
#'   \item{size}{point size for ggplot2 (default: 2)}
#'   \item{shade}{fill color for the envelope region (default: "grey90").
#'     Set to \code{NA} to suppress shading.}
#'   \item{col.envelope}{color for envelope boundary lines
#'     (default: "grey50")}
#'   \item{col.refline}{color of the diagonal reference line
#'     (default: "black")}
#'   \item{main}{character string for plot title (default: no title)}
#'   \item{as_data}{if \code{TRUE}, returns plot data instead of creating
#'     the plot}
#' }
#'
#' @details
#' The normal QQ plot is a diagnostic tool for assessing whether the
#' standardized residuals follow a standard normal distribution, as expected
#' under a correctly specified model. Each point represents one estimate,
#' plotted at coordinates \eqn{(\Phi^{-1}(p_i), z_{(i)})} where
#' \eqn{p_i} are the plotting positions and \eqn{z_{(i)}} are the sorted
#' standardized residuals.
#'
#' The default residual type is \code{"rstudent"} (LOO-PIT), which differs
#' from \code{metafor::qqnorm} (which defaults to \code{"rstandard"}).
#' LOO-PIT residuals are preferred because they are available for all model
#' types (including GLMMs and selection models) and properly account for
#' estimation uncertainty via leave-one-out cross-validation.
#'
#' The confidence envelope is computed in closed form from the distribution of
#' normal order statistics. For the \eqn{i}-th sorted residual,
#' \eqn{\Phi(Z_{(i)})} follows a beta distribution with shape parameters
#' \eqn{i} and \eqn{k + 1 - i}. Pointwise quantile bounds are transformed back
#' with \eqn{\Phi^{-1}}. When \code{smooth = TRUE}, bounds are smoothed using
#' Friedman's SuperSmoother. When \code{bonferroni = TRUE}, the confidence
#' level is adjusted for simultaneous inference across all \eqn{k} comparisons.
#'
#' For GLMM models, LOO-PIT residuals and the QQ plot are computed on the
#' approximate effect-size scale used by \code{loo}; they are not exact
#' count-scale PIT diagnostics.
#'
#' @return For \code{plot_type = "base"}, returns an invisible list with
#' components \code{x} (theoretical quantiles) and \code{y} (sorted
#' residuals). For \code{plot_type = "ggplot"}, returns a ggplot object.
#'
#' If \code{as_data = TRUE}, returns a list with all computed plot data
#' including: \code{x}, \code{y}, \code{points}, \code{refline},
#' \code{envelope}, \code{xlim}, \code{ylim}, \code{xlab}, and \code{ylab}.
#'
#'
#' @examples \dontrun{
#' if (requireNamespace("metadat", quietly = TRUE)) {
#'   data(dat.lehmann2018, package = "metadat")
#'   fit <- brma(
#'     yi      = yi,
#'     vi      = vi,
#'     data    = dat.lehmann2018,
#'     measure = "SMD",
#'     seed    = 1,
#'     silent  = TRUE
#'   )
#'   fit <- add_loo(fit)
#'
#'   qqnorm(fit)
#'   qqnorm(fit, type = "rstandard")
#'   qqnorm(fit, envelope = FALSE)
#'   qqnorm(fit, plot_type = "ggplot")
#'   qqnorm(fit, pch = 19, col = "blue", shade = "lightblue")
#' }
#' }
#'
#' @seealso [rstudent.brma()], [rstandard.brma()], [residuals.brma()],
#' [funnel.brma()], [radial.brma()]
#' @exportS3Method
qqnorm.brma <- function(y, type = "rstudent", unit = "estimate",
                         conditioning_depth = "marginal", envelope = TRUE,
                         conf_level = 95,
                         bonferroni = FALSE, reps = 1000, smooth = TRUE,
                         xlim, ylim, xlab, ylab, plot_type = "base", ...) {

  # input validation
  conditioning_depth_specified <- !missing(conditioning_depth)
  dots                         <- list(...)
  if ("level" %in% names(dots)) {
    if (is.numeric(dots[["level"]]) && length(dots[["level"]]) == 1L) {
      conf_level      <- dots[["level"]]
      dots[["level"]] <- NULL
    } else {
      .check_legacy_level_arg(dots, "qqnorm()")
    }
  }

  BayesTools::check_char(type, "type", allow_values = c("rstandard", "rstudent", "LOO-PIT"))
  unit               <- .normalize_unit(unit)
  conditioning_depth <- .normalize_conditioning_depth(conditioning_depth)
  BayesTools::check_bool(envelope, "envelope")
  BayesTools::check_real(conf_level, "conf_level", lower = 0, upper = 100)
  BayesTools::check_bool(bonferroni, "bonferroni")
  BayesTools::check_int(reps, "reps", lower = 1)
  BayesTools::check_bool(smooth, "smooth")
  BayesTools::check_char(plot_type, "plot_type", allow_values = c("base", "ggplot"))
  .check_unit_conditioning_depth(
    object             = y,
    unit               = unit,
    conditioning_depth = conditioning_depth,
    caller             = "qqnorm()"
  )

  if (unit == "cluster") {
    .check_cluster_unit_deferred("qqnorm()")
  }

  if ((type == "LOO-PIT" || type == "rstudent") && conditioning_depth_specified) {
    stop(
      "LOO-PIT residuals use the estimate-unit LOO target; ",
      "do not set 'conditioning_depth'.",
      call. = FALSE
    )
  }

  # set up graphical arguments with defaults
  dots <- .set_dots_qqnorm(dots)

  # generate QQ plot data
  qq_data <- .qqnorm_data(
    x                  = y,
    type               = type,
    unit               = unit,
    conditioning_depth = conditioning_depth,
    envelope           = envelope,
    conf_level         = conf_level,
    bonferroni         = bonferroni,
    reps               = reps,
    smooth             = smooth,
    xlim               = if (missing(xlim)) NULL else xlim,
    ylim               = if (missing(ylim)) NULL else ylim,
    xlab               = if (missing(xlab)) NULL else xlab,
    ylab               = if (missing(ylab)) NULL else ylab,
    dots               = dots
  )

  # allow data return for programmatic access
  if (isTRUE(dots[["as_data"]])) {
    return(qq_data)
  }

  # create plot
  if (plot_type == "ggplot") {
    return(.qqnorm_plot_ggplot(qq_data, dots))
  } else {
    .qqnorm_plot_base(qq_data, dots)
    return(invisible(qq_data[c("x", "y")]))
  }
}


# ---------------------------------------------------------------------------- #
# .qqnorm_data
# ---------------------------------------------------------------------------- #
#
# Generate data for normal QQ plot.
#
# Computes theoretical quantiles, sorted standardized residuals,
# reference line, and optional pseudo confidence envelope.
#
# @param x          brma object
# @param type       residual type ("rstudent", "LOO-PIT", or "rstandard")
# @param unit       output unit.
# @param conditioning_depth residual conditioning depth.
# @param envelope   logical; draw envelope
# @param conf_level confidence level (0-100)
# @param bonferroni logical; Bonferroni correction
# @param reps       number of simulation replications
# @param smooth     logical; smooth envelope bounds
# @param xlim       x-axis limits (NULL for auto)
# @param ylim       y-axis limits (NULL for auto)
# @param xlab       x-axis label (NULL for default)
# @param ylab       y-axis label (NULL for default)
# @param dots       list of graphical parameters
#
# @return list with QQ plot data components
#
# ---------------------------------------------------------------------------- #
.qqnorm_data <- function(x, type, unit, conditioning_depth, envelope, conf_level,
                          bonferroni, reps, smooth,
                          xlim, ylim, xlab, ylab, dots) {

  # get standardized residuals
  if (type == "rstandard") {
    res_obj <- rstandard.brma(
      model              = x,
      unit               = unit,
      conditioning_depth = conditioning_depth
    )
  } else {
    # type == "rstudent" or "LOO-PIT"
    res_obj <- rstudent.brma(x, unit = unit)
  }
  z <- res_obj$z
  K <- length(z)

  # theoretical standard normal quantiles
  qq_x <- stats::qnorm(stats::ppoints(K))

  # sorted standardized residuals
  sort_order <- order(z)
  qq_y       <- z[sort_order]

  # reference line: y = x (diagonal through origin)
  df_refline <- data.frame(
    x = range(qq_x),
    y = range(qq_x)
  )

  # compute envelope (if requested)
  env_data <- NULL
  if (envelope) {
    env_data <- .qqnorm_envelope(
      K          = K,
      level      = conf_level,
      bonferroni = bonferroni,
      reps       = reps,
      smooth     = smooth,
      qq_x       = qq_x
    )
  }

  # axis limits
  if (is.null(xlim)) {
    xlim <- range(qq_x) * 1.1
  }
  if (is.null(ylim)) {
    all_y <- qq_y
    if (!is.null(env_data)) {
      all_y <- c(all_y, env_data$lower, env_data$upper)
    }
    ylim <- range(all_y) * 1.1
  }

  # axis labels
  if (is.null(xlab)) xlab <- "Theoretical Quantiles"
  if (is.null(ylab)) ylab <- "Sample Quantiles"

  # points data frame
  df_points <- data.frame(
    x                  = qq_x,
    y                  = qq_y,
    unit               = rep(unit, K),
    conditioning_depth = rep(conditioning_depth, K)
  )
  if (is.data.frame(res_obj) && "cluster" %in% names(res_obj)) {
    df_points[["cluster"]]     <- res_obj[["cluster"]][sort_order]
    df_points[["n_estimates"]] <- res_obj[["n_estimates"]][sort_order]
  }

  return(list(
    # core QQ data (for invisible return, matching metafor)
    x        = qq_x,
    y        = qq_y,
    # plot elements
    points   = df_points,
    refline  = df_refline,
    envelope = env_data,
    # limits and labels
    xlim     = xlim,
    ylim     = ylim,
    xlab     = xlab,
    ylab     = ylab,
    # metadata
    type               = type,
    level              = conf_level,
    unit               = unit,
    conditioning_depth = conditioning_depth,
    K                  = K
  ))
}


# ---------------------------------------------------------------------------- #
# .qqnorm_envelope
# ---------------------------------------------------------------------------- #
#
# Deterministic pointwise confidence envelope for the QQ plot.
#
# Under a correctly specified model, standardized residuals should be
# approximately iid N(0,1). If U = Phi(Z), the i-th order statistic U_(i)
# has a Beta(i, K + 1 - i) distribution. This gives a deterministic envelope
# without RNG state or simulation error.
#
# @param K          number of observations
# @param level      confidence level (0-100)
# @param bonferroni logical; Bonferroni correction
# @param reps       number of simulation replications
# @param smooth     logical; smooth bounds with supsmu
# @param qq_x       theoretical quantiles
#
# @return data.frame with columns x, lower, upper
#
# ---------------------------------------------------------------------------- #
.qqnorm_envelope <- function(K, level, bonferroni, reps, smooth, qq_x) {

  # compute alpha (with optional Bonferroni correction)
  alpha <- 1 - level / 100
  if (bonferroni) {
    alpha <- alpha / K
  }

  index <- seq_len(K)
  lower <- stats::qnorm(stats::qbeta(
    alpha / 2,
    shape1 = index,
    shape2 = K + 1L - index
  ))
  upper <- stats::qnorm(stats::qbeta(
    1 - alpha / 2,
    shape1 = index,
    shape2 = K + 1L - index
  ))

  # smooth bounds (skip if too few points)
  if (smooth && K >= 4) {
    lower <- stats::supsmu(qq_x, lower)$y
    upper <- stats::supsmu(qq_x, upper)$y
  }

  data.frame(
    x     = qq_x,
    lower = lower,
    upper = upper
  )
}


# ---------------------------------------------------------------------------- #
# .qqnorm_plot_base
# ---------------------------------------------------------------------------- #
#
# Create QQ normal plot using base R graphics.
#
# @param data list of QQ plot data from .qqnorm_data()
# @param dots list of graphical parameters
#
# @return NULL invisibly
#
# ---------------------------------------------------------------------------- #
.qqnorm_plot_base <- function(data, dots) {

  # extract graphical parameters
  pch  <- dots[["pch"]]
  col  <- dots[["col"]]
  bg   <- dots[["bg"]]
  cex  <- dots[["cex"]]
  main <- dots[["main"]]
  las  <- dots[["las"]]

  # set up empty plot
  graphics::plot(
    NA, NA,
    xlim = data$xlim,
    ylim = data$ylim,
    xlab = data$xlab,
    ylab = data$ylab,
    main = if (!is.null(main)) main else "",
    bty  = "l",
    las  = las
  )

  # draw envelope (if present)
  if (!is.null(data$envelope)) {
    env <- data$envelope

    # shaded region
    if (!is.na(dots[["shade"]])) {
      graphics::polygon(
        x      = c(env$x, rev(env$x)),
        y      = c(env$lower, rev(env$upper)),
        col    = dots[["shade"]],
        border = NA
      )
    }

    # envelope boundary lines
    graphics::lines(env$x, env$lower, lty = "dotted", col = dots[["col.envelope"]])
    graphics::lines(env$x, env$upper, lty = "dotted", col = dots[["col.envelope"]])
  }

  # reference line y = x
  graphics::abline(a = 0, b = 1, lty = "dashed", col = dots[["col.refline"]])

  # points (drawn last, on top)
  graphics::points(
    data$points$x, data$points$y,
    pch = pch, col = col, bg = bg, cex = cex
  )

  return(invisible(NULL))
}


# ---------------------------------------------------------------------------- #
# .qqnorm_plot_ggplot
# ---------------------------------------------------------------------------- #
#
# Create QQ normal plot using ggplot2.
#
# @param data list of QQ plot data from .qqnorm_data()
# @param dots list of graphical parameters
#
# @return ggplot object
#
# ---------------------------------------------------------------------------- #
.qqnorm_plot_ggplot <- function(data, dots) {

  # extract graphical parameters
  pch  <- dots[["pch"]]
  col  <- dots[["col"]]
  bg   <- dots[["bg"]]
  size <- dots[["size"]]
  main <- dots[["main"]]

  out <- ggplot2::ggplot()

  # envelope (if present)
  if (!is.null(data$envelope)) {
    env <- data$envelope

    # shaded region
    if (!is.na(dots[["shade"]])) {
      out <- out +
        ggplot2::geom_ribbon(
          data    = env,
          mapping = ggplot2::aes(x = .data$x, ymin = .data$lower, ymax = .data$upper),
          fill    = dots[["shade"]]
        )
    }

    # envelope boundary lines
    out <- out +
      ggplot2::geom_line(
        data     = env,
        mapping  = ggplot2::aes(x = .data$x, y = .data$lower),
        linetype = "dotted",
        colour   = dots[["col.envelope"]]
      ) +
      ggplot2::geom_line(
        data     = env,
        mapping  = ggplot2::aes(x = .data$x, y = .data$upper),
        linetype = "dotted",
        colour   = dots[["col.envelope"]]
      )
  }

  # reference line y = x
  out <- out +
    ggplot2::geom_abline(
      intercept = 0,
      slope     = 1,
      linetype  = "dashed",
      colour    = dots[["col.refline"]]
    )

  # points
  out <- out +
    ggplot2::geom_point(
      data    = data$points,
      mapping = ggplot2::aes(x = .data$x, y = .data$y),
      shape   = pch,
      colour  = col,
      fill    = bg,
      size    = size
    )

  # scales and coord
  out <- out +
    ggplot2::scale_x_continuous(name = data$xlab) +
    ggplot2::scale_y_continuous(name = data$ylab) +
    ggplot2::coord_cartesian(xlim = data$xlim, ylim = data$ylim) +
    ggplot2::theme_bw()

  # title
  if (!is.null(main)) {
    out <- out + ggplot2::ggtitle(main)
  }

  return(out)
}


# ---------------------------------------------------------------------------- #
# .set_dots_qqnorm
# ---------------------------------------------------------------------------- #
#
# Process dots arguments for QQ normal plot with sensible defaults.
#
# @param ... additional graphical arguments
#
# @return list of graphical parameters with defaults applied
#
# ---------------------------------------------------------------------------- #
.set_dots_qqnorm <- function(dots) {

  # point styling
  dots <- .plot_point_style_defaults(dots)

  # envelope styling
  if (is.null(dots[["shade"]]))        dots[["shade"]]        <- "grey90"
  if (is.null(dots[["col.envelope"]])) dots[["col.envelope"]] <- "grey50"

  # reference line
  if (is.null(dots[["col.refline"]])) dots[["col.refline"]] <- "black"

  # title
  if (is.null(dots[["main"]])) dots[["main"]] <- NULL
  if (is.null(dots[["las"]]))  dots[["las"]]  <- 1

  # data return flag
  if (is.null(dots[["as_data"]])) dots[["as_data"]] <- FALSE

  return(dots)
}

.test_step_reference <- function(yi, mu, sigma, sei, omega, spec,
                                 weights = rep(1, length(yi))) {

  S    <- nrow(mu)
  K    <- ncol(mu)
  sign <- spec[["sign"]]
  out  <- matrix(NA_real_, nrow = S, ncol = K)

  for (s in seq_len(S)) {
    for (k in seq_len(K)) {
      signed_mean <- sign * mu[s, k]
      signed_y    <- sign * yi[k]
      denom       <- 0

      for (b in seq_len(spec[["n_bins"]])) {
        lower <- spec[["z_lower"]][b] * sei[k]
        upper <- spec[["z_upper"]][b] * sei[k]
        prob  <- stats::pnorm(upper, mean = signed_mean, sd = sigma[s, k]) -
          stats::pnorm(lower, mean = signed_mean, sd = sigma[s, k])
        denom <- denom + omega[s, b] * prob
      }

      out[s, k] <- weights[k] * (
        stats::dnorm(signed_y, mean = signed_mean, sd = sigma[s, k], log = TRUE) +
          log(omega[s, spec[["obs_bin"]][k]]) - log(denom)
      )
    }
  }

  return(out)
}

.test_interval_prob <- function(lower, upper, mean, sd) {

  if (!(sd > 0) || lower >= upper) {
    return(0)
  }
  if (is.infinite(lower) && lower < 0) {
    return(stats::pnorm(upper, mean = mean, sd = sd))
  }
  if (is.infinite(upper) && upper > 0) {
    return(stats::pnorm(lower, mean = mean, sd = sd, lower.tail = FALSE))
  }
  return(stats::pnorm(upper, mean = mean, sd = sd) -
    stats::pnorm(lower, mean = mean, sd = sd))
}

.test_logspace_sub <- function(log_a, log_b) {

  if (is.infinite(log_b) && log_b < 0) {
    return(log_a)
  }
  if (log_b > log_a) {
    return(NaN)
  }

  return(log_a + log1p(-exp(log_b - log_a)))
}

.test_interval_log_prob <- function(lower, upper, mean, sd) {

  if (!(sd > 0) || lower >= upper) {
    return(-Inf)
  }
  if (is.infinite(lower) && lower < 0 && is.infinite(upper) && upper > 0) {
    return(0)
  }

  if (lower >= mean) {
    log_s_lower <- if (is.infinite(lower) && lower > 0) {
      -Inf
    } else {
      stats::pnorm(lower, mean = mean, sd = sd, lower.tail = FALSE, log.p = TRUE)
    }
    log_s_upper <- if (is.infinite(upper) && upper > 0) {
      -Inf
    } else {
      stats::pnorm(upper, mean = mean, sd = sd, lower.tail = FALSE, log.p = TRUE)
    }
    return(.test_logspace_sub(log_s_lower, log_s_upper))
  }

  if (upper <= mean) {
    log_f_upper <- if (is.infinite(upper) && upper < 0) {
      -Inf
    } else {
      stats::pnorm(upper, mean = mean, sd = sd, lower.tail = TRUE, log.p = TRUE)
    }
    log_f_lower <- if (is.infinite(lower) && lower < 0) {
      -Inf
    } else {
      stats::pnorm(lower, mean = mean, sd = sd, lower.tail = TRUE, log.p = TRUE)
    }
    return(.test_logspace_sub(log_f_upper, log_f_lower))
  }

  prob <- .test_interval_prob(lower, upper, mean, sd)
  return(if (prob > 0) log(prob) else -Inf)
}

.test_logsumexp <- function(x) {

  max_x <- max(x)
  if (!is.finite(max_x)) {
    return(max_x)
  }

  return(max_x + log(sum(exp(x - max_x))))
}

.test_step_cdf_reference <- function(q, mu, sigma, sei, omega, spec,
                                     kernel_mode = rep(SELKERNEL_STEP, nrow(mu)),
                                     lower.tail = TRUE) {

  S <- nrow(mu)
  K <- ncol(mu)
  if (length(q) == 1L) {
    q <- rep(q, K)
  }

  out <- matrix(NA_real_, nrow = S, ncol = K)
  for (s in seq_len(S)) {
    for (k in seq_len(K)) {
      if (kernel_mode[s] == SELKERNEL_NORMAL) {
        out[s, k] <- stats::pnorm(
          q[k],
          mean       = mu[s, k],
          sd         = sigma[s, k],
          lower.tail = lower.tail
        )
        next
      }

      mean_z     <- spec[["sign"]] * mu[s, k] / sei[k]
      sd_z       <- sigma[s, k] / sei[k]
      q_z        <- spec[["sign"]] * q[k] / sei[k]
      denom      <- 0
      lower_mass <- 0
      upper_mass <- 0

      for (b in seq_len(spec[["n_bins"]])) {
        lower <- spec[["z_lower"]][b]
        upper <- spec[["z_upper"]][b]
        denom <- denom + omega[s, b] *
          .test_interval_prob(lower, upper, mean_z, sd_z)
        lower_mass <- lower_mass + omega[s, b] *
          .test_interval_prob(lower, min(upper, q_z), mean_z, sd_z)
        upper_mass <- upper_mass + omega[s, b] *
          .test_interval_prob(max(lower, q_z), upper, mean_z, sd_z)
      }

      if (spec[["sign"]] == 1L) {
        out[s, k] <- if (lower.tail) lower_mass / denom else upper_mass / denom
      } else {
        out[s, k] <- if (lower.tail) upper_mass / denom else lower_mass / denom
      }
    }
  }

  return(out)
}

.test_step_moments_reference <- function(mu, sigma, sei, omega, spec,
                                         kernel_mode = rep(SELKERNEL_STEP, nrow(mu))) {

  S             <- nrow(mu)
  K             <- ncol(mu)
  moment_mean   <- matrix(NA_real_, nrow = S, ncol = K)
  moment_second <- matrix(NA_real_, nrow = S, ncol = K)

  for (s in seq_len(S)) {
    for (k in seq_len(K)) {
      if (kernel_mode[s] == SELKERNEL_NORMAL) {
        moment_mean[s, k]   <- mu[s, k]
        moment_second[s, k] <- sigma[s, k]^2 + mu[s, k]^2
        next
      }

      signed_mean   <- spec[["sign"]] * mu[s, k]
      denom         <- 0
      signed_first  <- 0
      signed_second <- 0

      for (b in seq_len(spec[["n_bins"]])) {
        lower <- spec[["z_lower"]][b] * sei[k]
        upper <- spec[["z_upper"]][b] * sei[k]
        m0    <- .test_interval_prob(lower, upper, signed_mean, sigma[s, k])
        a     <- (lower - signed_mean) / sigma[s, k]
        b_z   <- (upper - signed_mean) / sigma[s, k]
        pa    <- stats::dnorm(a)
        pb    <- stats::dnorm(b_z)
        a_pa  <- a * pa
        b_pb  <- b_z * pb
        if (!is.finite(a_pa)) {
          a_pa <- 0
        }
        if (!is.finite(b_pb)) {
          b_pb <- 0
        }

        m1 <- signed_mean * m0 + sigma[s, k] * (pa - pb)
        m2 <- (signed_mean^2 + sigma[s, k]^2) * m0 +
          2 * signed_mean * sigma[s, k] * (pa - pb) +
          sigma[s, k]^2 * (a_pa - b_pb)

        denom         <- denom + omega[s, b] * m0
        signed_first  <- signed_first + omega[s, b] * m1
        signed_second <- signed_second + omega[s, b] * m2
      }

      moment_mean[s, k]   <- spec[["sign"]] * signed_first / denom
      moment_second[s, k] <- signed_second / denom
    }
  }

  return(list(
    mean   = moment_mean,
    second = moment_second
  ))
}

# ============================================================================ #
# brma.cooks.distance.R
# ============================================================================ #
#
# This file contains the cooks.distance method for brma model fits. Cook's
# distance measures the aggregate impact of an observation on all model
# coefficients.
#
# ============================================================================ #


#' @title Cook's Distance for brma Objects
#'
#' @description Computes Cook's distance for a fitted brma object. Cook's
#' distance measures the aggregate influence of each observation on the
#' model coefficients.
#'
#' @param model a fitted brma object.
#' @param ... additional arguments (currently ignored).
#'
#' @details
#' Cook's distance is computed as a PSIS leave-one-out deletion diagnostic. For
#' each observation \eqn{i}, normalized PSIS weights estimate the fitted values
#' under the leave-one-out posterior. The distance is the posterior Mahalanobis
#' distance between the full-data and leave-one-out fitted-value vectors:
#' \deqn{D_i = \frac{\Delta_i' V_\mu^+ \Delta_i}{P}}
#'
#' where \eqn{\Delta_i = \hat{\mu} - \hat{\mu}_{(-i)}}, \eqn{V_\mu^+} is the
#' generalized inverse of the full-posterior fitted-value covariance, and
#' \eqn{P} is the rank of the fixed-effect model matrix.
#'
#' @return A numeric vector of Cook's distance values, one for each observation.
#'
#' @examples \dontrun{
#' if (requireNamespace("metadat", quietly = TRUE)) {
#'   data(dat.lehmann2018, package = "metadat")
#'   fit <- bPET(yi = yi, vi = vi, data = dat.lehmann2018, measure = "SMD")
#'   fit <- add_loo(fit)
#'
#'   cooks.distance(fit)
#' }
#' }
#'
#' @seealso \code{\link{influence.brma}}, \code{\link{dffits.brma}}, \code{\link{hatvalues.brma}}
#' @importFrom stats cooks.distance
#' @exportS3Method
cooks.distance.brma <- function(model, ...) {

  # the function relies on hatvalues
  # as such it is sensible only sensible for normal models
  outcome_type       <- .outcome_type(model)
  is_weightfunction  <- .is_weightfunction(model)

  if (outcome_type != "norm") {
    stop("cooks.distance is only available for normal outcome models.", call. = FALSE)
  }
  if (is_weightfunction) {
    stop("cooks.distance is not available for selection models (weightfunction).", call. = FALSE)
  }

  fit_samples <- .influence_fit_samples(model)
  weights     <- .diagnostic_psis_weights(model)
  P           <- qr(.get_model_matrix(model))[["rank"]]
  d_vec       <- .cooks.distance_internal(fit_samples, weights, P)
  d_vec       <- .diagnostic_set_names(d_vec, model)

  return(d_vec)
}

.cooks.distance_internal <- function(fit_samples, weights, P) {

  summary <- .psis_fit_influence_summary(fit_samples, weights)
  delta   <- sweep(summary[["loo_fit"]], 2, summary[["full_fit"]], "-")
  delta   <- -delta

  vcov_fit <- stats::cov(fit_samples)
  vcov_inv <- .symmetric_ginv(vcov_fit)

  d_vec <- rowSums((delta %*% vcov_inv) * delta) / max(P, 1L)
  names(d_vec) <- colnames(fit_samples)

  return(d_vec)
}


# ---------------------------------------------------------------------------- #
# .symmetric_ginv
# ---------------------------------------------------------------------------- #
#
# Generalized inverse for symmetric positive semi-definite covariance matrices.
#
# ---------------------------------------------------------------------------- #
.symmetric_ginv <- function(x, tol = sqrt(.Machine$double.eps)) {

  x   <- (x + t(x)) / 2
  eig <- eigen(x, symmetric = TRUE)

  keep <- eig[["values"]] > max(abs(eig[["values"]]), 1) * tol
  if (!any(keep)) {
    return(matrix(0, nrow = nrow(x), ncol = ncol(x)))
  }

  vectors <- eig[["vectors"]][, keep, drop = FALSE]
  values  <- eig[["values"]][keep]

  return(vectors %*% diag(1 / values, nrow = length(values)) %*% t(vectors))
}

# ============================================================================ #
# brma.vif.R
# ============================================================================ #
#
# Variance Inflation Factor (VIF) and Generalized VIF (GVIF) for brma objects.
#
# Two complementary diagnostics for multicollinearity:
# 1. Classical VIF/GVIF from cov2cor(vcov) (matches metafor)
# 2. Posterior correlation of regression coefficients (Bayesian diagnostic)
#
# VIF is computed from the correlation matrix of the coefficient
# variance-covariance matrix: vcov = (X'WX)^{-1}. For simple models, W uses
# diag(weight_i/(vi + tau^2)); scale and multilevel models average this
# coefficient covariance over posterior heterogeneity draws.
#
# GVIF formula (Fox & Monette, 1992, for multi-df terms):
#   GVIF = det(R_11) * det(R_22) / det(R)
#
# GSIF (comparable across terms with different df):
#   GSIF = GVIF^(1/(2*m))
#
# ============================================================================ #


# ---------------------------------------------------------------------------- #
# vif generic
# ---------------------------------------------------------------------------- #

#' @title Variance Inflation Factors
#'
#' @description Computes variance inflation factors (VIF) for a fitted model
#' with moderators, to assess multicollinearity among predictors.
#'
#' @param object a fitted model object
#' @param ... additional arguments passed to methods
#'
#' @return Method-specific return value containing VIF diagnostics.
#'
#' @references
#' \insertCite{fox1992generalized}{RoBMA}
#'
#' @seealso [regplot()], [summary.brma()]
#' @export
vif <- function(object, ...) {

  UseMethod("vif")
}


# ---------------------------------------------------------------------------- #
# vif.brma
# ---------------------------------------------------------------------------- #

#' @title Variance Inflation Factors for brma Objects
#'
#' @description Computes variance inflation factors (VIF) and generalized
#' VIF (GVIF) for a fitted brma meta-regression model. Also optionally
#' returns the posterior correlation matrix of regression coefficients.
#'
#' @param object a fitted brma object with moderators
#' @param posterior_correlation logical; whether to also compute and return
#' the posterior correlation matrix of regression coefficients. Defaults
#' to \code{TRUE}.
#' @param ... additional arguments (currently ignored)
#'
#' @details
#' VIF is computed from the correlation matrix derived from the
#' coefficient variance-covariance matrix \eqn{(X'WX)^{-1}}. For standard
#' meta-regression models, \eqn{W = \mathrm{diag}(w_i/(v_i + \hat\tau^2))}
#' with \eqn{\hat\tau} equal to the posterior mean heterogeneity. For scale
#' and multilevel models, the coefficient covariance is averaged across
#' posterior heterogeneity draws, using observation-specific \eqn{\tau_i}
#' and block-structured multilevel covariance where applicable.
#'
#' A VIF of 1 indicates no collinearity; values above 5 or 10 are
#' commonly considered problematic.
#'
#' For multi-column terms, such as factor contrasts,
#' the Generalized VIF (GVIF) of \insertCite{fox1992generalized;textual}{RoBMA}
#' is reported. GVIF captures the joint inflation for all coefficients
#' belonging to the same term. To enable comparison across terms with different
#' degrees of freedom, \eqn{GVIF^{1/(2 \cdot df)}} is also reported; this
#' value can be compared against the usual VIF thresholds (after squaring).
#'
#' When \code{posterior_correlation = TRUE}, the function also returns the
#' posterior correlation matrix of the regression coefficients. This
#' Bayesian diagnostic complements VIF: while VIF diagnoses the
#' \emph{potential} for collinearity problems (a data property), the
#' posterior correlation shows the \emph{realized} identification
#' given the data and priors. Informative priors can mitigate
#' collinearity, reducing posterior correlations even when VIF is high.
#'
#' @return An object of class \code{vif.brma} containing:
#' \item{vif}{A data frame with columns \code{term}, \code{df}, \code{GVIF},
#' and \code{GVIF^(1/(2*df))} (= \eqn{GVIF^{1/(2 \cdot df)}}). For single-df
#' terms, GVIF equals the standard VIF.}
#' \item{posterior_correlation}{A correlation matrix of posterior regression
#' coefficient samples when requested and available; otherwise \code{NULL}.}
#'
#' @examples \dontrun{
#' if (requireNamespace("metadat", quietly = TRUE) &&
#'     requireNamespace("metafor", quietly = TRUE)) {
#'   data(dat.bcg, package = "metadat")
#'   dat <- metafor::escalc(
#'     measure = "RR",
#'     ai      = tpos,
#'     bi      = tneg,
#'     ci      = cpos,
#'     di      = cneg,
#'     data    = dat.bcg
#'   )
#'   fit <- brma(
#'     yi      = yi,
#'     vi      = vi,
#'     mods    = ~ ablat + year,
#'     data    = dat,
#'     measure = "RR",
#'     seed    = 1,
#'     silent  = TRUE
#'   )
#'
#'   vif(fit)
#' }
#' }
#'
#' @references
#' \insertAllCited{}
#'
#' @seealso [regplot()], [summary.brma()]
#' @exportS3Method
vif.brma <- function(object, posterior_correlation = TRUE, ...) {

  BayesTools::check_bool(posterior_correlation, "posterior_correlation")

  # require moderators
  if (!.is_mods(object)) {
    stop("VIF is only meaningful for models with moderators (meta-regression).", call. = FALSE)
  }

  # compute VIF from vcov correlation matrix
  vif_df <- .compute_vif(object)

  # posterior correlation of regression coefficients
  post_cor <- NULL
  if (posterior_correlation) {
    post_cor <- .compute_posterior_correlation(object)
  }

  output <- list(
    vif                  = vif_df,
    posterior_correlation = post_cor
  )

  class(output) <- "vif.brma"

  return(output)
}


# ---------------------------------------------------------------------------- #
# .compute_vif
# ---------------------------------------------------------------------------- #
#
# Compute VIF/GVIF from cov2cor(solve(X'WX)). Simple models use the posterior
# mean tau plug-in. Scale and multilevel models average solve(X'WX) over
# posterior heterogeneity draws so observation-specific and block covariances
# are preserved.
#
# @param object brma object with moderators
#
# @return data frame with columns: term, df, GVIF, GVIF^(1/(2*df))
#
# ---------------------------------------------------------------------------- #
.compute_vif <- function(object) {

  # extract model matrix
  X      <- .get_model_matrix(object)
  assign <- attr(X, "assign")

  # extract sampling variances
  vi      <- .outcome_data_vi(object)
  weights <- .outcome_data_weights(object)
  K       <- length(vi)

  # extract posterior heterogeneity
  # for simple models: use summary["tau", "Mean"] to preserve metafor-style VIF
  # for scale/multilevel models: use full posterior tau matrices
  is_scale      <- .is_scale(object)
  is_multilevel <- .is_multilevel(object)

  if (!is_scale && !is_multilevel) {
    tau_within_samples  <- matrix(object[["summary"]]["tau", "Mean"], nrow = 1, ncol = K)
    tau_between_samples <- matrix(0, nrow = 1, ncol = K)
  } else {
    tau_result <- .evaluate.brma.tau(
      fit           = object[["fit"]],
      scale_data    = object[["data"]][["scale"]],
      scale_formula = if (is_scale) .create_fit_formula_list(data = object[["data"]], "scale") else NULL,
      scale_priors  = object[["priors"]][["scale"]],
      is_scale      = is_scale,
      is_multilevel = is_multilevel,
      K             = K
    )
    tau_within_samples  <- tau_result[["tau_within"]]
    tau_between_samples <- tau_result[["tau_between"]]
  }

  # compute posterior-averaged vcov = (X'WX)^{-1} using meta-analytic weights
  vcov <- .vif_vcov_from_tau_samples(
    X                   = X,
    vi                  = vi,
    weights             = weights,
    tau_within_samples  = tau_within_samples,
    tau_between_samples = tau_between_samples,
    cluster             = if (is_multilevel) object[["data"]][["outcome"]][["cluster"]] else NULL
  )

  # identify and remove intercept (assign == 0)
  has_intercept <- 0 %in% assign
  if (has_intercept) {
    keep   <- assign != 0
    vcov   <- vcov[keep, keep, drop = FALSE]
    assign <- assign[keep]
  }

  p <- ncol(vcov)

  term_labels <- attr(X, "term_labels")
  if (is.null(term_labels)) {
    term_labels <- .fitted_formula_terms(
      object            = object,
      parameter         = "mu",
      include_intercept = FALSE,
      display           = TRUE
    )
  }

  # compute GVIF per term from vcov correlation matrix
  terms_unique <- sort(unique(assign))
  n_terms      <- length(terms_unique)

  gvif_values    <- numeric(n_terms)
  gvif_df_values <- numeric(n_terms)
  df_values      <- integer(n_terms)
  term_names     <- character(n_terms)

  if (p > 1) {
    R    <- stats::cov2cor(vcov)
    detR <- det(R)
  }

  for (k in seq_len(n_terms)) {

    term_index <- terms_unique[k]
    cols       <- which(assign == term_index)
    m          <- length(cols)

    df_values[k]  <- m
    term_names[k] <- if (term_index <= length(term_labels)) term_labels[term_index] else colnames(vcov)[cols[1]]

    if (p <= 1 || m == p) {
      # single predictor or all columns belong to one term
      gvif_values[k]    <- 1
      gvif_df_values[k] <- 1
    } else {
      # Fox & Monette (1992) formula
      gvif <- det(R[cols, cols, drop = FALSE]) *
              det(R[-cols, -cols, drop = FALSE]) / detR

      gvif_values[k]    <- gvif
      gvif_df_values[k] <- gvif^(1 / (2 * m))
    }
  }

  vif_df <- data.frame(
    term               = term_names,
    df                 = df_values,
    GVIF               = gvif_values,
    "GVIF^(1/(2*df))"  = gvif_df_values,
    stringsAsFactors   = FALSE,
    check.names        = FALSE
  )

  return(vif_df)
}


# ---------------------------------------------------------------------------- #
# .vif_vcov_from_tau_samples
# ---------------------------------------------------------------------------- #
#
# Compute the coefficient covariance used by VIF from one or more posterior
# heterogeneity draws.
#
# @param X design matrix.
# @param vi sampling variances.
# @param weights likelihood weights.
# @param tau_within_samples S x K matrix of estimate-level heterogeneity SDs.
# @param tau_between_samples optional S x K matrix of cluster-level SDs.
# @param cluster optional cluster identifiers for multilevel block covariance.
#
# @return posterior mean of solve(X'WX).
#
# ---------------------------------------------------------------------------- #
.vif_vcov_from_tau_samples <- function(X, vi, weights, tau_within_samples,
                                       tau_between_samples = NULL,
                                       cluster = NULL) {

  K <- length(vi)
  P <- ncol(X)

  tau_within_samples <- .vif_tau_matrix(tau_within_samples, K, "tau_within_samples")

  if (is.null(tau_between_samples)) {
    tau_between_samples <- matrix(0, nrow = nrow(tau_within_samples), ncol = K)
  } else {
    tau_between_samples <- .vif_tau_matrix(tau_between_samples, K, "tau_between_samples")
  }

  if (nrow(tau_between_samples) != nrow(tau_within_samples)) {
    stop("'tau_within_samples' and 'tau_between_samples' must have the same number of rows.",
         call. = FALSE)
  }

  block_indices <- list()
  if (!is.null(cluster)) {
    block_indices <- .get_multilevel_block_indices(cluster)
  }

  S        <- nrow(tau_within_samples)
  vcov_sum <- matrix(0, nrow = P, ncol = P)

  for (s in seq_len(S)) {
    diagonal_s <- (vi + tau_within_samples[s, ]^2) / weights
    WX <- .hat_apply_precision(
      x             = X,
      diagonal      = diagonal_s,
      rank_one      = if (!is.null(cluster)) tau_between_samples[s, ] else NULL,
      block_indices = block_indices
    )

    vcov_sum <- vcov_sum + .hat_solve_crossprod(crossprod(X, WX))
  }

  vcov <- vcov_sum / S
  if (!is.null(colnames(X))) {
    dimnames(vcov) <- list(colnames(X), colnames(X))
  }

  return(vcov)
}


# ---------------------------------------------------------------------------- #
# .vif_tau_matrix
# ---------------------------------------------------------------------------- #
#
# Normalize heterogeneity input to an S x K matrix.
#
# ---------------------------------------------------------------------------- #
.vif_tau_matrix <- function(x, K, name) {

  if (is.null(dim(x))) {
    if (length(x) == 1L) {
      return(matrix(x, nrow = 1, ncol = K))
    }
    if (length(x) == K) {
      return(matrix(x, nrow = 1, ncol = K))
    }
    stop("'", name, "' must have length 1, length K, or be an S x K matrix.",
         call. = FALSE)
  }

  x <- as.matrix(x)

  if (ncol(x) == 1L && K > 1L) {
    x <- matrix(x[, 1], nrow = nrow(x), ncol = K)
  }

  if (ncol(x) != K) {
    stop("'", name, "' must have K columns.", call. = FALSE)
  }

  return(x)
}


# ---------------------------------------------------------------------------- #
# .compute_posterior_correlation
# ---------------------------------------------------------------------------- #
#
# Extract posterior samples of regression coefficients and compute their
# correlation matrix. Caller must ensure the model has moderators.
#
# @param object brma object with moderators
#
# @return correlation matrix, or NULL if fewer than 2 coefficients
#
# ---------------------------------------------------------------------------- #
.compute_posterior_correlation <- function(object) {

  # extract coefficient samples via BayesTools
  # keep_formulas = "mu" extracts all mu-formula coefficients
  # return_samples = TRUE returns S x P matrix instead of summary
  samples_mat <- as.matrix(BayesTools::JAGS_estimates_table(
    fit                = object[["fit"]],
    keep_formulas      = "mu",
    remove_diagnostics = TRUE,
    transform_factors  = TRUE,
    transform_scaled   = TRUE,
    return_samples     = TRUE
  ))

  # need at least 2 coefficients for correlation
  if (ncol(samples_mat) < 2) {
    return(NULL)
  }

  return(stats::cor(samples_mat))
}


# ---------------------------------------------------------------------------- #
# print.vif.brma
# ---------------------------------------------------------------------------- #

#' @title Print VIF Results
#'
#' @description Prints variance inflation factors and optional posterior
#' correlation matrix.
#'
#' @param x a \code{vif.brma} object
#' @param digits number of decimal places. Defaults to 3.
#' @param ... additional arguments (currently ignored)
#'
#' @return Returns \code{x} invisibly.
#'
#' @exportS3Method
print.vif.brma <- function(x, digits = 3, ...) {

  vif_df <- x$vif

  cat("\nVariance Inflation Factors:\n")

  if (all(vif_df$df == 1)) {
    # simple VIF display (no GVIF columns needed)
    out        <- vif_df$GVIF
    names(out) <- vif_df$term
    print(round(out, digits))
  } else {
    # full GVIF display with term as rownames
    print_df           <- vif_df[, -1, drop = FALSE]
    rownames(print_df) <- vif_df$term
    print(round(print_df, digits))
  }

  if (!is.null(x$posterior_correlation)) {
    cat("\nPosterior Correlation of Coefficients:\n")
    print(round(x$posterior_correlation, digits))
  }

  cat("\n")

  return(invisible(x))
}

context("Weighted diagnostics")

skip_on_cran()


test_that("Outcome likelihood weights default to one and preserve supplied weights", {

  object_unweighted <- brma.norm(
    yi        = c(.10, .20, .30),
    sei       = c(.10, .20, .30),
    only_data = TRUE
  )

  expect_equal(.outcome_data_weights(object_unweighted), c(1, 1, 1))

  object_weighted <- brma.norm(
    yi        = c(.10, .20, .30),
    sei       = c(.10, .20, .30),
    weights   = c(.5, 1, 2),
    only_data = TRUE
  )

  expect_equal(.outcome_data_weights(object_weighted), c(.5, 1, 2))
})


test_that("VIF uses likelihood weights in the meta-regression precision", {

  dat <- data.frame(
    yi  = c(.10, .25, .05, .30, .18, .12),
    sei = c(.12, .18, .14, .20, .16, .22),
    w   = c(3.0, .4, 1.5, .7, 2.2, .9),
    x1  = c(-2, -1, 0, 1, 2, 3),
    x2  = c(3, 1, 2, 5, 4, 7)
  )

  object <- brma.norm(
    yi        = yi,
    sei       = sei,
    weights   = w,
    mods      = ~ x1 + x2,
    data      = dat,
    only_data = TRUE
  )
  object[["priors"]]  <- list(outcome = list(bias = NULL))
  object[["summary"]] <- matrix(
    .20,
    nrow     = 1,
    dimnames = list("tau", "Mean")
  )

  brma_vif <- .compute_vif(object)

  X       <- .get_model_matrix(object)
  assign  <- attr(X, "assign")
  keep    <- assign != 0
  vi      <- dat[["sei"]]^2
  tau2    <- .20^2
  W       <- diag(dat[["w"]] / (vi + tau2), nrow = nrow(dat))
  vcov    <- solve(crossprod(X, W) %*% X)
  vcov    <- vcov[keep, keep, drop = FALSE]
  assign  <- assign[keep]
  R       <- stats::cov2cor(vcov)
  det_R   <- det(R)
  terms_i <- sort(unique(assign))

  expected <- vapply(terms_i, function(term_i) {
    cols <- which(assign == term_i)
    det(R[cols, cols, drop = FALSE]) *
      det(R[-cols, -cols, drop = FALSE]) / det_R
  }, numeric(1))

  W_unweighted    <- diag(1 / (vi + tau2), nrow = nrow(dat))
  vcov_unweighted <- solve(crossprod(X, W_unweighted) %*% X)
  vcov_unweighted <- vcov_unweighted[keep, keep, drop = FALSE]
  R_unweighted    <- stats::cov2cor(vcov_unweighted)
  det_unweighted  <- det(R_unweighted)
  expected_unweighted <- vapply(terms_i, function(term_i) {
    cols <- which(assign == term_i)
    det(R_unweighted[cols, cols, drop = FALSE]) *
      det(R_unweighted[-cols, -cols, drop = FALSE]) / det_unweighted
  }, numeric(1))

  expect_equal(brma_vif[["GVIF"]], expected, tolerance = 1e-12)
  expect_false(isTRUE(all.equal(expected, expected_unweighted, tolerance = 1e-8)))
})


test_that("VIF covariance helper uses observation-specific tau samples", {

  X <- cbind(
    intercept = 1,
    x1        = c(-2, -1, 0, 1, 2),
    x2        = c(2, 1, 3, 5, 4)
  )
  vi <- c(.02, .04, .03, .05, .06)
  weights <- c(1.0, 2.0, .5, 1.5, .8)
  tau_within <- rbind(
    c(.05, .10, .15, .20, .25),
    c(.20, .18, .16, .14, .12)
  )

  brma_vcov <- .vif_vcov_from_tau_samples(
    X                  = X,
    vi                 = vi,
    weights            = weights,
    tau_within_samples = tau_within
  )

  expected <- Reduce(`+`, lapply(seq_len(nrow(tau_within)), function(s) {
    W <- diag(weights / (vi + tau_within[s, ]^2), nrow = length(vi))
    solve(crossprod(X, W) %*% X)
  })) / nrow(tau_within)

  scalar_tau <- sqrt(mean(tau_within^2))
  scalar_W <- diag(weights / (vi + scalar_tau^2), nrow = length(vi))
  scalar_vcov <- solve(crossprod(X, scalar_W) %*% X)

  expect_equal(unname(brma_vcov), unname(expected), tolerance = 1e-12)
  expect_false(isTRUE(all.equal(brma_vcov, scalar_vcov, tolerance = 1e-8)))
})


test_that("VIF covariance helper uses multilevel block precision", {

  X <- cbind(
    intercept = 1,
    x1        = c(-1, 0, 1, -2, 2),
    x2        = c(2, 4, 3, 1, 5)
  )
  vi <- c(.03, .05, .04, .06, .02)
  weights <- c(1.0, .7, 1.4, .9, 1.2)
  cluster <- c(1, 1, 2, 2, 2)
  tau_within <- rbind(
    c(.10, .12, .08, .11, .09),
    c(.15, .14, .13, .12, .11)
  )
  tau_between <- rbind(
    c(.20, .20, .15, .15, .15),
    c(.10, .10, .18, .18, .18)
  )

  brma_vcov <- .vif_vcov_from_tau_samples(
    X                   = X,
    vi                  = vi,
    weights             = weights,
    tau_within_samples  = tau_within,
    tau_between_samples = tau_between,
    cluster             = cluster
  )

  block_indices <- split(seq_along(cluster), cluster)
  expected <- Reduce(`+`, lapply(seq_len(nrow(tau_within)), function(s) {
    M <- diag((vi + tau_within[s, ]^2) / weights, nrow = length(vi))
    for (idx in block_indices) {
      M[idx, idx] <- M[idx, idx] + tcrossprod(tau_between[s, idx])
    }
    W <- solve(M)
    solve(crossprod(X, W) %*% X)
  })) / nrow(tau_within)

  diagonal_only <- .vif_vcov_from_tau_samples(
    X                   = X,
    vi                  = vi,
    weights             = weights,
    tau_within_samples  = tau_within,
    tau_between_samples = tau_between
  )

  expect_equal(unname(brma_vcov), unname(expected), tolerance = 1e-12)
  expect_false(isTRUE(all.equal(brma_vcov, diagonal_only, tolerance = 1e-8)))
})

context("VIF")

source(testthat::test_path("common-functions.R"))
source(testthat::test_path("helper-contracts.R"))
source(testthat::test_path("helper-test-matrix.R"))
source(testthat::test_path("helper-metafor.R"))

skip_if_no_fits()
skip_if_not_installed("metafor")

fit_names <- list_fits()
fits      <- lazy_fits(fit_names, validate = FALSE)
info      <- lazy_infos(fit_names, validate = FALSE)

for_each_case(vif_cases(), function(case) {
  test_that_case("VIF matches metafor", case, {
    expect_vif_matches_metafor(case)
  })
})

test_that("VIF returns optional posterior correlation diagnostics", {

  name <- "bcg_meta-regression"
  skip_if_missing_fits(name)

  fit_brma      <- fits[[name]]
  vif_with_post <- vif(fit_brma)
  post_cor      <- vif_with_post[["posterior_correlation"]]

  expect_s3_class(vif_with_post, "vif.brma")
  expect_vif_table(vif_with_post[["vif"]], 2)
  expect_type(post_cor, "double")
  expect_equal(nrow(post_cor), ncol(post_cor))
  expect_equal(unname(diag(post_cor)), rep(1, nrow(post_cor)), tolerance = 1e-12)
  expect_equal(post_cor, t(post_cor), tolerance = 1e-12)
  expect_true(all(is.finite(post_cor)))

  vif_without_post <- vif(fit_brma, posterior_correlation = FALSE)
  expect_null(vif_without_post[["posterior_correlation"]])
})

test_that("VIF rejects models without moderators", {

  model_names <- c(
    "bcg_meta-analysis",
    "konstantopoulos2011_3lvl",
    "nielweise2008_glmm",
    "dat.lehmann2018-PET",
    "dat.lehmann2018-3PSM"
  )
  skip_if_missing_fits(model_names)

  for (name in model_names) {
    expect_error(
      vif(fits[[name]]),
      "only meaningful for models with moderators",
      info = paste("vif rejects model without moderators for", name)
    )
  }
})

#' @title Fitting specification
#' @name fitting_specification
#'
#' @description
#' The `brma` family of functions uses the following arguments to specify the
#' MCMC sampling and fitting settings.
#'
#' @param sample numeric. Number of MCMC samples to save. Defaults to `5000`.
#' @param burnin numeric. Number of burn-in iterations. Defaults to `2000`.
#' @param adapt numeric. Number of adaptation iterations. Defaults to `500`.
#' @param chains numeric. Number of MCMC chains. Defaults to `3`.
#' @param thin numeric. Thinning interval. Defaults to `1`.
#' @param parallel logical. Whether to run MCMC chains in parallel. Defaults to `FALSE`.
#' @param autofit logical. Whether to automatically extend the MCMC chains if convergence is not met.
#' Defaults to `FALSE`.
#' @param autofit_control list of autofit control settings. See [set_autofit_control()] for details.
#' @param convergence_checks list of convergence check settings. See [set_convergence_checks()] for details.
#' @param seed numeric. Random seed for reproducibility. Defaults to `NULL`.
#' @param silent logical. Whether to suppress output. Constructors with no
#' explicit default use `RoBMA.get_option("silent")` when `silent` is omitted.
#' Model-averaging wrappers default to `TRUE` unless explicitly changed.
#' @param ... additional advanced arguments. Fitting functions reject unused
#' arguments; currently recognized internal arguments include `only_data`,
#' `only_priors`, `is_JASP`, and `is_JASP_prefix`.
#'
#' @seealso \code{\link{brma}}, \code{\link{set_autofit_control}}, \code{\link{set_convergence_checks}}
#' @aliases fitting_specification
NULL

### creates basic RoBMA object with
# - fit_control
# - autofit_control
# - convergence_checks
#
# @param dots additional arguments originally passed as ...
# @param chains Number of MCMC chains
# @param adapt Number of adaptation iterations
# @param burnin Number of burnin iterations
# @param sample Number of sampling iterations
# @param thin Thinning interval
# @param autofit Whether to use autofit
# @param parallel Whether to run chains in parallel
# @param silent Whether to suppress output
# @param seed Random seed
# @param autofit_control Autofit control settings
# @param convergence_checks Convergence check settings
#
# @return A list containing fit_control, autofit_control, convergence_checks
.createObject <- function(
    dots, class,
    chains, adapt, burnin, sample, thin,
    autofit, parallel, silent, seed,
    autofit_control, convergence_checks) {

  object       <- NULL

  ### input global settings if unspecified
  if (missing(silent)) {
    silent <- RoBMA.get_option("silent")
  }

  ### check and store MCMC settings
  object$fit_control <- BayesTools::JAGS_check_and_list_fit_settings(
    chains = chains, adapt = adapt, burnin = burnin, sample = sample,
    thin = thin, autofit = autofit, parallel = parallel, cores = chains,
    silent = silent, seed = seed
  )
  object$autofit_control    <- BayesTools::JAGS_check_and_list_autofit_settings(autofit_control = autofit_control)
  object$convergence_checks <- .check_and_list_convergence_checks(convergence_checks = convergence_checks)


  ### include JASP indicators for progress bars
  if (!is.null(dots[["is_JASP"]])) {
    object[["is_JASP"]]        <- dots[["is_JASP"]]
    object[["is_JASP_prefix"]] <- dots[["is_JASP_prefix"]]
  }


  ### add class
  class(object) <- class

  return(object)
}

.set_only_priors_class <- function(object) {

  class(object) <- unique(c("only_priors.brma", class(object)))

  return(object)
}

.validate_constructor_dots <- function(dots, caller) {

  allowed <- c("only_data", "only_priors", "is_JASP", "is_JASP_prefix")
  .check_unused_dots(
    dots    = dots,
    allowed = allowed,
    caller  = caller
  )

  bool_arguments <- intersect(
    c("only_data", "only_priors", "is_JASP"),
    names(dots)
  )
  for (argument in bool_arguments) {
    BayesTools::check_bool(
      dots[[argument]],
      argument,
      allow_NA = FALSE
    )
  }

  if ("is_JASP_prefix" %in% names(dots)) {
    BayesTools::check_char(
      dots[["is_JASP_prefix"]],
      "is_JASP_prefix",
      check_length = 1,
      allow_NA     = FALSE
    )
  }

  return(dots)
}

.check_unused_dots <- function(dots, allowed, caller) {

  if (length(dots) == 0L) {
    return(invisible(TRUE))
  }

  dot_names <- names(dots)
  if (is.null(dot_names)) {
    dot_names <- rep("", length(dots))
  }

  unused <- dot_names[!nzchar(dot_names) | !dot_names %in% allowed]
  if (length(unused) == 0L) {
    return(invisible(TRUE))
  }

  unused[!nzchar(unused)] <- "<unnamed>"
  stop(
    "Unused argument", if (length(unused) > 1L) "s" else "",
    " in ", caller, ": ",
    paste0("'", unused, "'", collapse = ", "),
    call. = FALSE
  )
}

.autocompute_brma <- function(object, marglik = !inherits(object, "RoBMA")) {

  if (RoBMA.get_option("autocompute.loo")) {
    object <- add_loo(object)
  }
  if (RoBMA.get_option("autocompute.waic")) {
    object <- add_waic(object)
  }
  if (marglik && RoBMA.get_option("autocompute.marglik")) {
    object <- add_marglik(object)
  }

  return(object)
}

.extract_posterior_indicator <- function(posterior_samples, parameter,
                                         prior = NULL, column = NULL) {

  if (is.null(column)) {
    column <- paste0(parameter, "_indicator")
  }
  if (!column %in% colnames(posterior_samples)) {
    stop("Missing posterior model indicator: '", column, "'.",
         call. = FALSE)
  }

  indicator <- posterior_samples[, column]
  if (!is.numeric(indicator) && !is.integer(indicator)) {
    stop("Invalid posterior model indicator: '", column, "'.",
         call. = FALSE)
  }
  if (any(!is.finite(indicator)) ||
      any(abs(indicator - round(indicator)) > sqrt(.Machine$double.eps))) {
    stop("Invalid posterior model indicator: '", column, "'.",
         call. = FALSE)
  }

  indicator <- as.integer(round(indicator))
  if (!is.null(prior)) {
    valid_values <- seq_len(length(prior))
    if (any(!indicator %in% valid_values)) {
      stop("Invalid posterior model indicator range: '", column, "'.",
           call. = FALSE)
    }
  }

  return(indicator)
}


### object tools options
# add simple summary and model coefficients to the object
# (this differ from more customizable user facing summary function)
.object_summary      <- function(object) {

  # provide a simple summary
  estimates <- BayesTools::JAGS_estimates_table(
    fit               = object[["fit"]],
    transform_factors = TRUE,
    transform_scaled  = TRUE,
    remove_parameters = c(
      "theta", # remove random-effects (estimate-level)
      "gamma", # remove random-effects (cluster-level)
      "pi",    # remove baserate for OR models
      "phi"    # remove lograte for IRR models
    )
  )

  return(estimates)
}
.object_coefficients <- function(object) {

  estimates        <- object[["summary"]][,"Mean"]
  names(estimates) <- rownames(object[["summary"]])

  return(estimates)
}

# ============================================================================ #
# diagnostics-helpers.R
# ============================================================================ #
#
# Shared diagnostic helpers used by residual, influence, and deletion methods.
#
# ============================================================================ #


.diagnostic_zero_variance_note <- function(diagnostic, parameters,
                                           variance = "LOO posterior") {

  parameters <- paste(parameters, collapse = ", ")
  return(paste0(
    diagnostic,
    " could not be computed for parameter(s) ",
    parameters,
    " because the ",
    variance,
    " variance is zero; values are reported as NaN."
  ))
}


.diagnostic_with_note <- function(x, class, note) {

  attr(x, "note") <- note
  class(x)        <- c(class, class(x))

  return(x)
}


.print_diagnostic_note <- function(note) {

  if (!is.null(note) && nzchar(note)) {
    cat("\nNote: ", note, "\n", sep = "")
  }

  return(invisible(NULL))
}


.diagnostic_psis_context <- function(model, context = NULL) {

  if (!is.null(context)) {
    return(context)
  }

  loo_result   <- loo.brma(model, unit = "estimate")
  psis_object  <- loo_result[["psis_object"]]
  psis_weights <- loo::weights.importance_sampling(
    psis_object,
    log       = FALSE,
    normalize = TRUE
  )

  return(list(
    loo_result   = loo_result,
    psis_object  = psis_object,
    psis_weights = psis_weights
  ))
}


.diagnostic_psis_weights <- function(model, weights = NULL) {

  if (!is.null(weights)) {
    return(weights)
  }

  return(loo_weights(model, unit = "estimate"))
}


.diagnostic_check_loo <- function(model, context = NULL, unit = "estimate") {

  if (is.null(context)) {
    return(check_loo(model, unit = unit))
  }

  loo_result <- context[["loo_result"]]
  pareto_k   <- loo_result[["diagnostics"]][["pareto_k"]]
  bad_k      <- which(pareto_k > 0.7)

  if (length(bad_k) > 0L) {
    warning(
      "Some Pareto k values are high (> 0.7), indicating potentially unreliable ",
      "LOO diagnostics for ", unit, "s: ", paste(bad_k, collapse = ", "), ". ",
      "Inspect the loo fit by using loo(object).",
      call. = FALSE
    )
  }

  return(invisible(NULL))
}


.diagnostic_study_labels <- function(model) {

  labels <- as.character(.get_estimate_labels(model))
  empty  <- is.na(labels) | !nzchar(labels)

  if (any(empty)) {
    labels[empty] <- as.character(which(empty))
  }
  if (anyDuplicated(labels)) {
    labels <- make.unique(labels)
  }

  return(labels)
}


.diagnostic_set_rownames <- function(x, model) {

  labels <- .diagnostic_study_labels(model)
  if (NROW(x) == length(labels)) {
    rownames(x) <- labels
  }

  return(x)
}


.diagnostic_set_names <- function(x, model) {

  labels <- .diagnostic_study_labels(model)
  if (length(x) == length(labels)) {
    names(x) <- labels
  }

  return(x)
}


.diagnostic_collect_notes <- function(...) {

  notes <- unlist(list(...), use.names = FALSE)
  notes <- notes[!is.na(notes) & nzchar(notes)]
  notes <- unique(notes)

  if (length(notes) == 0L) {
    return(NULL)
  }

  return(paste(notes, collapse = " "))
}

context("Hatvalues")

source(testthat::test_path("common-functions.R"))
source(testthat::test_path("helper-contracts.R"))
source(testthat::test_path("helper-test-matrix.R"))
source(testthat::test_path("helper-metafor.R"))

skip_if_no_fits()
skip_if_not_installed("metafor")

fit_names <- list_fits()
fits      <- lazy_fits(fit_names, validate = FALSE)
info      <- lazy_infos(fit_names, validate = FALSE)

for_each_case(hatvalue_cases(), function(case) {
  test_that_case("Hatvalues match metafor", case, {
    expect_hatvalues_match_metafor(case)
  })
})

test_that("Hatvalues reject unsupported single-model families", {

  model_names <- c("nielweise2008_glmm", "dat.lehmann2018-3PSM")
  skip_if_missing_fits(model_names)

  for (name in model_names) {
    expect_error(hatvalues(fits[[name]]), "available", info = name)
  }
})

test_that("Hatvalues for BMA.norm fits are internally consistent", {

  model_names <- c("dat.lehmann2018_BMA.norm", "dat.lehmann2018_BMA.norm_mods")
  skip_if_missing_fits(model_names)

  for (name in model_names) {
    fit <- fits[[name]]
    expect_hatvalues_vector(hatvalues(fit), nobs(fit), info = name)
  }
})

test_that("Hatvalues use study labels as names", {

  name <- "bcg_meta-analysis"
  skip_if_missing_fits(name)

  fit <- fits[[name]]
  expect_equal(names(hatvalues(fit)), RoBMA:::.diagnostic_study_labels(fit))
})

test_that("Hatvalues reject unsupported model-averaging families", {

  model_names <- c("bcg_BMA.glmm", "dat.lehmann2018_RoBMA")
  skip_if_missing_fits(model_names)

  expect_error(hatvalues(fits[["bcg_BMA.glmm"]]), "normal outcome models")
  expect_error(hatvalues(fits[["dat.lehmann2018_RoBMA"]]), "selection models")
})

context("Diagnostic notes")

test_that("DFBETAS note objects print and preserve data-frame contract", {

  x <- data.frame(mu = c(NaN, NaN))
  x <- RoBMA:::.diagnostic_with_note(
    x,
    class = "dfbetas.brma",
    note  = RoBMA:::.diagnostic_zero_variance_note(
      diagnostic = "DFBETAS",
      parameters = "mu"
    )
  )

  expect_s3_class(x, "data.frame")
  expect_s3_class(x, "dfbetas.brma")
  expect_true(all(is.nan(x[["mu"]])))
  expect_true(any(grepl("Note: DFBETAS could not be computed", capture.output(print(x)))))
})


test_that("COVRATIO note objects print and preserve numeric values", {

  x <- RoBMA:::.diagnostic_with_note(
    rep(NaN, 2),
    class = "covratio.brma",
    note  = RoBMA:::.diagnostic_zero_variance_note(
      diagnostic = "COVRATIO",
      parameters = "mu",
      variance   = "posterior"
    )
  )

  expect_s3_class(x, "covratio.brma")
  expect_true(all(is.nan(as.numeric(x))))
  expect_true(any(grepl("Note: COVRATIO could not be computed", capture.output(print(x)))))
})


test_that("Diagnostic label helpers preserve study labels and repair invalid row names", {

  object <- brma.norm(
    yi        = c(.10, .20, .30),
    sei       = c(.10, .20, .30),
    slab      = c("Study A", "Study A", "Study C"),
    only_data = TRUE
  )
  object[["data"]][["outcome"]][["slab"]][3] <- NA

  labels <- RoBMA:::.diagnostic_study_labels(object)

  expect_equal(labels, c("Study A", "Study A.1", "3"))

  out <- RoBMA:::.diagnostic_set_rownames(data.frame(x = 1:3), object)
  expect_equal(rownames(out), labels)

  vec <- RoBMA:::.diagnostic_set_names(1:3, object)
  expect_equal(names(vec), labels)
})


test_that("diagnostic plot wrappers expose the full selector API", {

  wrapper_formals <- list(
    autocorrelation = names(formals(plot_diagnostic_autocorrelation.brma)),
    trace           = names(formals(plot_diagnostic_trace.brma)),
    density         = names(formals(plot_diagnostic_density.brma))
  )

  for (formals_i in wrapper_formals) {
    expect_true(all(c("parameter_mods", "parameter_scale", "type", "lags") %in% formals_i))
  }
})

context("DFBETAS")

source(testthat::test_path("common-functions.R"))
source(testthat::test_path("helper-contracts.R"))
source(testthat::test_path("helper-test-matrix.R"))
source(testthat::test_path("helper-metafor.R"))

skip_if_no_fits()
skip_if_not_installed("metafor")

fit_names <- list_fits()
fits      <- lazy_fits(fit_names, validate = FALSE)
info      <- lazy_infos(fit_names, validate = FALSE)

for_each_case(dfbetas_metafor_cases(), function(case) {
  test_that_case("DFBETAS match metafor", case, {
    expect_dfbetas_match_metafor(case)
  })
})

test_that("DFBETAS for location-scale models expose location and scale terms", {

  name <- "bangertdrowns2004_location-scale"
  skip_if_missing_fits(name)

  fit_brma <- fits[[name]]

  expect_dfbetas_table(dfbetas(fit_brma), nobs(fit_brma), info = name)
  expect_dfbetas_table(dfbetas(fit_brma, type = "scale"), nobs(fit_brma),
                       info = paste(name, "scale"))
})

test_that("DFBETAS outputs use study-label row names", {

  name <- "bcg_meta-analysis"
  skip_if_missing_fits(name)

  fit_brma <- fits[[name]]
  labels   <- RoBMA:::.diagnostic_study_labels(fit_brma)

  expect_equal(rownames(dfbetas(fit_brma)), labels)
  expect_equal(
    rownames(dfbetas(fit_brma, return_loo_estimates = TRUE)),
    labels
  )
})

test_that("DFBETAS for selection models expose bias terms", {

  model_names <- c("dat.lehmann2018-3PSM", "dat.lehmann2018-3PSM_neg", "dat.lehmann2018-3PSMreg")
  skip_if_missing_fits(model_names)

  fit_pos <- fits[["dat.lehmann2018-3PSM"]]
  fit_neg <- fits[["dat.lehmann2018-3PSM_neg"]]
  dfb_pos <- dfbetas(fit_pos)
  dfb_neg <- dfbetas(fit_neg)

  expect_equal(dfb_pos[-4, 1], -dfb_neg[-4, 1], tolerance = 0.01,
               info = "positive and negative selection DFBETAS flip")

  for (name in c("dat.lehmann2018-3PSM", "dat.lehmann2018-3PSMreg")) {
    fit_brma <- fits[[name]]
    expect_dfbetas_table(dfbetas(fit_brma), nobs(fit_brma), info = name)

    bias_dfbetas <- dfbetas(fit_brma, type = "bias")
    expect_dfbetas_table(bias_dfbetas, nobs(fit_brma), info = paste(name, "bias"))
    expect_true(any(grepl("^omega", colnames(bias_dfbetas))), info = name)
  }
})

test_that("DFBETAS for PET and PEESE expose publication-bias terms", {

  bias_cases <- data.frame(
    name = c(
      "dat.lehmann2018-PET",
      "dat.lehmann2018-PETreg",
      "dat.lehmann2018-PET_neg",
      "dat.lehmann2018-PEESE",
      "dat.lehmann2018-PEESEreg",
      "dat.lehmann2018-PEESE_neg"
    ),
    column = c("PET", "PET", "PET", "PEESE", "PEESE", "PEESE"),
    stringsAsFactors = FALSE
  )
  skip_if_missing_fits(bias_cases[["name"]])

  for (i in seq_len(nrow(bias_cases))) {
    name          <- bias_cases[["name"]][[i]]
    expected_col  <- bias_cases[["column"]][[i]]
    fit_brma      <- fits[[name]]
    bias_dfbetas  <- dfbetas(fit_brma, type = "bias")

    expect_dfbetas_table(bias_dfbetas, nobs(fit_brma), info = paste(name, "bias"))
    expect_true(expected_col %in% colnames(bias_dfbetas), info = name)
  }
})

test_that("DFBETAS for GLMM fits are finite", {

  model_names <- c("nielweise2008_glmm", "bcg_glmm_reg")
  skip_if_missing_fits(model_names)

  for (name in model_names) {
    fit_brma <- fits[[name]]
    expect_dfbetas_table(dfbetas(fit_brma), nobs(fit_brma), info = name)
  }
})

test_that("DFBETAS for model-averaging fits are internally consistent", {

  cases <- data.frame(
    name = c(
      "dat.lehmann2018_BMA.norm",
      "dat.lehmann2018_BMA.norm_mods",
      "dat.lehmann2018_BMA.norm_scale",
      "bcg_BMA.glmm",
      "nielweise2008_BMA.glmm",
      "dat.lehmann2018_RoBMA_mods",
      "dat.lehmann2018_RoBMA_3lvl_mods_scale"
    ),
    type = c(NA, NA, "scale", NA, NA, NA, NA),
    min_cols = c(1, 2, 2, 1, 1, 1, 1),
    stringsAsFactors = FALSE
  )
  skip_if_missing_fits(c(cases[["name"]], "dat.lehmann2018_RoBMA"))

  for (i in seq_len(nrow(cases))) {
    name     <- cases[["name"]][[i]]
    type     <- cases[["type"]][[i]]
    fit_brma <- fits[[name]]
    dfb      <- if (is.na(type)) dfbetas(fit_brma) else dfbetas(fit_brma, type = type)

    expect_dfbetas_table(dfb, nobs(fit_brma), min_cols = cases[["min_cols"]][[i]],
                         info = name)
  }

  bias_dfbetas <- dfbetas(fits[["dat.lehmann2018_RoBMA"]], type = "bias")
  expect_dfbetas_table(bias_dfbetas, nobs(fits[["dat.lehmann2018_RoBMA"]]),
                       min_cols = 3, info = "RoBMA bias")
  expect_true(any(grepl("^omega", colnames(bias_dfbetas))))
  expect_true("PET" %in% colnames(bias_dfbetas))
  expect_true("PEESE" %in% colnames(bias_dfbetas))
})

#' @title Prior specification
#' @name prior_specification
#'
#' @description
#' Prior distributions can be supplied for all model parameters. When omitted,
#' the fitting functions construct defaults from the effect-size measure, sample
#' sizes, or informed-prior settings when available.
#'
#' This typically includes the pooled effect \eqn{\mu} and between-study heterogeneity
#' \eqn{\tau}. In the case of meta-regression, the pooled effect \eqn{\mu} corresponds to
#' the intercept, and additional prior distributions for the regression coefficients
#' are required. In the case of a location-scale model, the between-study heterogeneity
#' corresponds to the intercept of the scale regression, and additional prior
#' distributions for the scale regression coefficients are required.
#'
#' @param prior_effect prior distribution for the effect size (\eqn{\mu}) parameter
#' (the intercept). If omitted, a default prior is constructed. In single-model
#' functions, explicit `NULL` or `FALSE` sets a spike at zero.
#' @param prior_heterogeneity prior distribution for the heterogeneity (\eqn{\tau})
#' parameter. If omitted, a default prior is constructed. In single-model
#' functions, explicit `NULL` or `FALSE` sets a spike at zero.
#' @param prior_mods prior distribution for the moderators (\eqn{\beta}) parameters.
#' A single prior applies to all terms; a named list can specify term-specific
#' priors. If omitted or `NULL`, default priors are used.
#' @param prior_scale prior distribution for the scale (\eqn{\delta}) parameters.
#' A single prior applies to all terms; a named list can specify term-specific
#' priors. If omitted or `NULL`, default priors are used.
#' @param prior_heterogeneity_allocation prior distribution for the fraction of
#' heterogeneity allocated to the cluster-level component in multilevel models
#' (\eqn{\rho}). If omitted or `NULL`, defaults to `Beta(1, 1)`.
#' @param prior_baserate prior distribution for the estimate-specific midpoint
#' base-rate probability in binomial GLMM models. If omitted or `NULL`, defaults
#' to independent `Beta(1, 1)` priors.
#' @param prior_lograte prior distribution for the estimate-specific midpoint
#' log-rate in Poisson GLMM models. If omitted or `NULL`, a data-based
#' unit-information normal prior is used independently for each estimate.
#' @param prior_unit_information_sd numeric. The unit information standard deviation (\eqn{\sigma_{unit}}).
#' Cannot be used together with `prior_informed_field`.
#' @param rescale_priors numeric. A scaling factor for supported prior distributions.
#' Point and none priors are unchanged. For constructors with publication-bias
#' prior distributions, `rescale_priors` does not rescale them except for the
#' default PEESE prior's UISD adjustment. Defaults to 1.
#' @param standardize_continuous_predictors logical. Whether to standardize continuous predictors.
#' Defaults to `TRUE`.
#' @param set_contrast_factor_predictors character. How to set contrast for factor predictors.
#' Defaults are constructor-specific and shown in each function usage; single-model
#' constructors use `"treatment"`, while model-averaging constructors use `"meandif"`.
#' @param prior_informed_field character. The field of the informed prior distributions.
#' Omit to use the standard default prior specification; explicit `NULL` is invalid.
#' @param prior_informed_subfield character. The subfield of the informed prior distributions.
#' Omit to use the field-specific default, such as `"Cochrane"` for
#' `prior_informed_field = "medicine"`; explicit `NULL` is invalid.
#'
#' @details
#' There are several ways to specify the prior distributions: \enumerate{
#'    \item{via a standardized effect size `measure` with known unit information standard deviation,}
#'    \item{by estimating unit information standard deviation using sample sizes `ni`,}
#'    \item{by manually setting `prior_unit_information_sd`,}
#'    \item{by specifying informed empirical prior distributions via `prior_informed_field`
#'    and `prior_informed_subfield`,}
#'    \item{or via fully custom specification using the `prior_effect`, `prior_heterogeneity`,
#'    `prior_mods`, `prior_scale`, and `prior_heterogeneity_allocation` arguments.}
#' }
#' In all cases, the prior behavior can be further modified by the `rescale_priors`,
#' `standardize_continuous_predictors`, and `set_contrast_factor_predictors` arguments.
#'
#' ### (1) Specifying prior distributions via standardized effect size measures with known
#' unit information standard deviation
#' This is the easiest way to specify prior distributions. The width of prior
#' distributions is based on a fraction of the known unit information standard deviation (`UISD`)
#' \insertCite{rover2021weakly}{RoBMA}. The default prior distributions for the parameters
#' are set as follows:\tabular{ll}{
#'    effect size:               \tab Normal(0, \eqn{\frac{1}{2}} `UISD`)  \cr
#'    heterogeneity:             \tab Normal+(0, \eqn{\frac{1}{4}} `UISD`) \cr
#'    effect moderation:         \tab Normal(0, \eqn{\frac{1}{4}} `UISD`)  \cr
#'    heterogeneity moderation:  \tab Normal(0, \eqn{\frac{1}{2}})
#' }
#' The heterogeneity moderation parameters are multiplicative, as such they are independent of `UISD`.
#'
#' The default fraction of the `UISD` can be changed via `RoBMA.options()` using one of
#' the following arguments: `"default_UISD.effect"`, `"default_UISD.heterogeneity"`,
#' `"default_UISD.mods"`, `"default_UISD.scale"`.
#'
#' The known `UISD` for standardized effect size measures (`measure`) are set as follows:
#' \tabular{ll}{
#'   `"SMD"`:   \tab \eqn{\sqrt{2}} \cr
#'   `"ZCOR"`:  \tab \eqn{1}        \cr
#'   `"RR"`:    \tab \eqn{\sqrt{4}} \cr
#'   `"OR"`:    \tab \eqn{\sqrt{4}} \cr
#'   `"HR"`:    \tab \eqn{\sqrt{4}} \cr
#'   `"IRR"`:   \tab \eqn{\sqrt{4}}
#' }
#' See Chapter 2.4 in \insertCite{spiegelhalter2004bayesian;textual}{RoBMA} and
#' Chapter 1 in \insertCite{grieve2022hybrid;textual}{RoBMA}.
#'
#' ### (2) Estimating unit information standard deviation using sample sizes
#' When effect sizes are on a non-standardized scale (`measure = "GEN"`) or use a
#' standardized effect size without known `UISD`, the `UISD` can be estimated from
#' sample sizes (`ni`) and standard errors (`sei`) following Equation 6 in
#' \insertCite{rover2021weakly;textual}{RoBMA}. The estimated `UISD` is then used to
#' scale the default prior distributions as described in section (1).
#'
#' Note that the known `UISD` for standardized effect size measures (section (1)) is
#' used if available, even when `ni` is provided.
#'
#' ### (3) Manually setting unit information standard deviation
#' Alternatively, the `UISD` can be specified directly via the `prior_unit_information_sd`
#' argument. This is useful when the appropriate scale for prior distributions is known
#' a priori or when multiple analyses are to be performed on subsets of the same data
#' (re-estimating `UISD` on different subsets of the data can lead to slightly different
#' prior distributions for different subsets; see [estimate_unit_information_sd()]).
#' The specified `prior_unit_information_sd` is then used to scale the default prior
#' distributions as described in section (1).
#'
#' Note that the manually specified `prior_unit_information_sd` takes precedence over
#' the estimated `UISD` from `ni` (section (2)) and the known `UISD` from `measure`
#' (section (1)). It cannot be combined with `prior_informed_field`.
#'
#' ### (4) Specifying informed empirical prior distributions
#' Informed prior distributions can be specified via the `prior_informed_field` and
#' `prior_informed_subfield` arguments. Currently, only `prior_informed_field = "medicine"`
#' with subfields defined in `BayesTools::prior_informed_medicine_names` is supported,
#' which uses empirically derived prior distributions from medical meta-analyses as described in
#' \insertCite{bartos2021bayesian;textual}{RoBMA} and
#' \insertCite{bartos2023empirical;textual}{RoBMA}.
#'
#' When `prior_informed_field = "medicine"`, the default `prior_informed_subfield` is
#' `"Cochrane"` (i.e., using the whole CDSR database as a reference). The informed
#' prior distributions are available for the following effect size measures:
#' \tabular{ll}{
#'   `"SMD"`:  \tab standardized mean difference \cr
#'   `"OR"`:   \tab log odds ratio               \cr
#'   `"RR"`:   \tab log risk ratio               \cr
#'   `"RD"`:   \tab risk difference              \cr
#'   `"HR"`:   \tab log hazard ratio
#' }
#'
#' Note that informed prior distributions are only available for the effect size (\eqn{\mu})
#' and heterogeneity (\eqn{\tau}) parameters. For effect moderation (`prior_mods`), the
#' informed effect prior is scaled by a factor of `RoBMA.get_option("default_informed_priors.mods")`.
#' For heterogeneity moderation (`prior_scale`), a normal prior with standard deviation
#' specified by `RoBMA.get_option("default_informed_priors.scale")` is used.
#'
#' ### (5) Fully custom prior distributions
#' Prior distributions can be fully customized by directly specifying the `prior_effect`,
#' `prior_heterogeneity`, `prior_mods`, `prior_scale`, and `prior_heterogeneity_allocation`
#' arguments. These should be prior distribution objects created via `BayesTools::prior()`
#' or related functions (e.g., `prior_factor()`).
#'
#' ### Rescaling prior distributions
#' The `rescale_priors` argument allows rescaling supported prior distributions by a
#' multiplicative factor. For example, `rescale_priors = 2` doubles the standard
#' deviations/scales of normal, Cauchy, t, and inverse-gamma prior distributions,
#' making them more diffuse. Point and none priors are unchanged. For publication-bias
#' prior distributions, see \code{\link{publication_bias_prior_specification}}.
#'
#' @references
#' \insertAllCited{}
#'
#' @seealso \code{\link{publication_bias_prior_specification}},
#' \code{\link[BayesTools]{prior}}, \code{\link{RoBMA.options}},
#' \code{\link{brma}}
#' @aliases prior_specification
NULL


#' @title Publication-bias prior specification
#' @name publication_bias_prior_specification
#'
#' @description
#' Publication-bias prior distributions are specified separately from the
#' meta-analytic prior distributions documented in
#' \code{\link{prior_specification}}. They define selection-model weights,
#' PET/PEESE regression coefficients, or the publication-bias model space in
#' \code{\link{RoBMA}}.
#'
#' @param prior_bias prior distribution for publication-bias adjustment. For
#' \code{\link{bselmodel}}, this is usually a weightfunction prior created by
#' \code{\link{prior_weightfunction}}. For \code{\link{bPET}}, use
#' \code{\link{prior_PET}}. For \code{\link{bPEESE}}, use
#' \code{\link{prior_PEESE}}. For \code{\link{RoBMA}}, this can be a single
#' publication-bias prior distribution or a list of publication-bias prior
#' distributions. In the single-model bias-adjustment constructors, omitted or
#' \code{NULL} uses the corresponding default prior distribution.
#' @param prior_bias_null prior distribution(s) for null publication-bias
#' component(s) in \code{\link{RoBMA}}, usually \code{\link{prior_none}()}.
#' A single prior distribution object creates one null component, a list creates
#' multiple null components, and \code{NULL} or \code{FALSE} omits the null
#' publication-bias component.
#' @param model_type character string specifying predefined publication-bias
#' model ensembles for \code{\link{RoBMA}}. One of \code{"PSMA"},
#' \code{"6w"}, \code{"2w"}, or \code{"PP"}.
#' @param steps numeric vector of one-sided p-value cut points for the default
#' \code{\link{bselmodel}} weightfunction prior. If \code{prior_bias} is supplied,
#' the prior distribution carries its own p-value cut points.
#'
#' @details
#' ## Single-model bias-adjustment priors
#'
#' \code{\link{bselmodel}}, \code{\link{bPET}}, and \code{\link{bPEESE}} fit one
#' publication-bias adjustment at a time. The \code{prior_bias} argument must
#' match the fitted bias-adjustment type.
#'
#' \tabular{lll}{
#' Constructor \tab Prior constructor \tab Default prior distribution \cr
#' \code{bselmodel()} \tab \code{\link{prior_weightfunction}} \tab one-sided cumulative weightfunction with \code{steps = 0.025} \cr
#' \code{bPET()} \tab \code{\link{prior_PET}} \tab positive Cauchy centered at 0 \cr
#' \code{bPEESE()} \tab \code{\link{prior_PEESE}} \tab positive Cauchy centered at 0, with UISD-adjusted scale
#' }
#'
#' The default PET prior distribution uses
#' \code{RoBMA.get_option("default_bias_PET.scale")}. The default PEESE prior
#' distribution uses \code{RoBMA.get_option("default_bias_PEESE.scale")} after
#' rescaling to the analyzed effect-size measure. The default weightfunction
#' prior distribution uses \code{RoBMA.get_option("default_bias_weightfunction.alpha")}
#' for each cumulative-Dirichlet alpha parameter.
#'
#' ## Model-averaged publication-bias priors
#'
#' \code{\link{RoBMA}} averages over publication-bias components. By default,
#' \code{prior_bias_null} is \code{\link{prior_none}()} and \code{model_type}
#' determines the alternative components:
#'
#' \tabular{ll}{
#' \code{"PSMA"} \tab six weight functions, PET, and PEESE \cr
#' \code{"6w"} \tab six weight functions \cr
#' \code{"2w"} \tab two two-sided weight functions \cr
#' \code{"PP"} \tab PET and PEESE
#' }
#'
#' Custom \code{prior_bias} replaces the preset alternative components. If
#' \code{prior_bias} is omitted, \code{model_type} still supplies the default
#' alternative components even when \code{prior_bias_null} is customized or
#' omitted. Setting \code{prior_bias = NULL} or \code{prior_bias = FALSE}
#' omits all alternative bias-adjustment models. Setting
#' \code{prior_bias_null = NULL} or \code{prior_bias_null = FALSE} omits the
#' no-bias component.
#'
#' Each publication-bias component has a prior model weight. User-specified
#' components receive equal weights unless their prior objects set
#' \code{prior_weights}. The \code{"2w"} and \code{"6w"} presets split the
#' alternative bias mass equally across their weight functions, \code{"PP"}
#' splits it equally across PET and PEESE, and \code{"PSMA"} assigns half of the
#' alternative bias mass to the six weight functions combined and one quarter
#' each to PET and PEESE.
#'
#' Publication-bias prior distributions are not rescaled by \code{rescale_priors}.
#' The exception is the default PEESE prior distribution, whose scale is adjusted
#' to the effect-size measure via the unit-information standard deviation.
#'
#' @seealso
#' \code{\link{prior_weightfunction}}, \code{\link{prior_PET}},
#' \code{\link{prior_PEESE}}, \code{\link{prior_none}},
#' \code{\link{RoBMA_prior_specification}},
#' \code{\link{prior_specification}}
#'
#' @aliases publication_bias_prior_specification bias_prior_specification
NULL


#' @title Prior specification for model-averaging
#' @name RoBMA_prior_specification
#'
#' @description
#' The \code{\link{RoBMA}} function extends the prior specification described in
#' \code{\link{prior_specification}} by allowing mixture priors that define competing
#' hypotheses for Bayesian model-averaging. This enables multimodel inference via the
#' product space method \insertCite{carlin1995bayesian,lodewyckx2011tutorial}{RoBMA},
#' where a single MCMC chain explores a joint model space containing all competing models.
#'
#' @param prior_effect prior distribution(s) for the alternative effect
#' component(s).
#' @param prior_heterogeneity prior distribution(s) for the alternative
#' heterogeneity component(s).
#' @param prior_mods prior distribution(s) for alternative moderator
#' components. A single prior applies to all terms; a named list can specify
#' term-specific components.
#' @param prior_scale prior distribution(s) for alternative scale-regression
#' components. A single prior applies to all terms; a named list can specify
#' term-specific components.
#' @param prior_heterogeneity_allocation prior distribution(s) for the
#' alternative cluster-level heterogeneity allocation component(s).
#' @param prior_bias prior distribution(s) for alternative publication-bias
#' component(s), such as weight functions, PET, or PEESE.
#'
#' Alternative prior arguments can be:
#' \itemize{
#'   \item A single prior distribution object (creates a mixture with one alternative)
#'   \item A list of prior distributions (creates a mixture with multiple alternatives)
#'   \item `NULL` or `FALSE` (omits the alternative hypothesis component)
#' }
#' See \code{\link{publication_bias_prior_specification}} for details on
#' specifying publication-bias priors and \code{\link{prior_specification}} for
#' details on specifying meta-analytic parameter priors.
#'
#' @param prior_effect_null prior distribution(s) for the null effect
#' component(s).
#' @param prior_heterogeneity_null prior distribution(s) for the null
#' heterogeneity component(s).
#' @param prior_mods_null prior distribution(s) for null moderator components.
#' A single prior applies to all terms; a named list can specify term-specific
#' components.
#' @param prior_scale_null prior distribution(s) for null scale-regression
#' components. A single prior applies to all terms; a named list can specify
#' term-specific components.
#' @param prior_heterogeneity_allocation_null prior distribution(s) for the
#' null cluster-level heterogeneity allocation component(s).
#' @param prior_bias_null prior distribution(s) for null publication-bias
#' component(s), usually `prior_none()`. See
#' \code{\link{publication_bias_prior_specification}}.
#'
#' Null prior arguments can be:
#' \itemize{
#'   \item A single prior distribution object (creates a mixture with one null)
#'   \item A list of prior distributions (creates a mixture with multiple nulls)
#'   \item `NULL` or `FALSE` (omits the null hypothesis component)
#' }
#' Defaults to a point mass (spike) at zero for effect and heterogeneity parameters.
#'
#' @param model_type character string specifying predefined publication-bias model
#' ensembles. One of:
#' \itemize{
#'   \item `"PSMA"` (default): Full RoBMA-PSMA ensemble with 6 weight functions + PET + PEESE
#'   \item `"6w"`: Six weight function models
#'   \item `"2w"`: Two weight function models 
#'   \item `"PP"`: PET-PEESE models only
#' }
#' Custom `prior_bias` replaces the preset alternative bias components. If
#' `prior_bias` is omitted, `model_type` determines the default alternatives even
#' when `prior_bias_null` is customized or omitted.
#'
#' @details
#' ## The product space method for model-averaging
#'
#' RoBMA implements Bayesian model-averaging using the product space method
#' \insertCite{carlin1995bayesian,lodewyckx2011tutorial}{RoBMA}. Rather than fitting
#' each model separately and combining results after the fitting is complete, 
#' this approach embeds all competing models within a single joint model space. 
#' A discrete model indicator variable selects which model component is "active" 
#' at each MCMC iteration, and the posterior probability of each model is estimated 
#' from the proportion of iterations spent in that model's subspace.
#'
#' This is achieved by specifying **mixture priors** that combine:
#' \itemize{
#'   \item **Null hypothesis component(s)**: Typically point masses (spikes) representing
#'         the absence of an effect, heterogeneity, or publication bias
#'   \item **Alternative hypothesis component(s)**: Continuous distributions representing
#'         the presence of the parameter
#' }
#'
#' The mixture structure enables direct computation of:
#' \itemize{
#'   \item **Inclusion Bayes factors**: Evidence for the presence vs. absence of each
#'         parameter (e.g., effect, heterogeneity, bias)
#'   \item **Model-averaged estimates**: Posterior estimates that account for model
#'         uncertainty by weighting across all models
#'   \item **Conditional estimates**: Posterior estimates given that a particular
#'         hypothesis is true
#' }
#'
#' ## Default mixture prior structure
#'
#' By default, \code{\link{RoBMA}} creates the following mixture priors:
#'
#' \tabular{lll}{
#'   **Parameter**      \tab **Null component**         \tab **Alternative component**   \cr
#'   Effect (\eqn{\mu})        \tab Spike(0)                   \tab Normal(0, UISD-scaled)      \cr
#'   Heterogeneity (\eqn{\tau}) \tab Spike(0)                   \tab Normal+(0, UISD-scaled)     \cr
#'   Publication bias   \tab No bias (prior_none)       \tab Weight functions + PET + PEESE \cr
#'   Allocation (\eqn{\rho})   \tab (none by default)          \tab Beta(1, 1)
#' }
#'
#' See \code{\link{prior_specification}} for details on how the UISD scaling is determined.
#'
#' ## Specifying custom mixture priors
#'
#' ### Single prior vs. list of priors
#'
#' Each `prior_*` and `prior_*_null` argument accepts either a single prior or a list:
#' \itemize{
#'   \item **Single prior**: Creates one component in the mixture
#'   \item **List of priors**: Creates multiple components, enabling model-averaging
#'         across different prior specifications (e.g., different effect size scales)
#' }
#'
#' When multiple priors are specified in a list, they are all treated as alternative
#' (or null) hypothesis components for the main inclusion Bayes factor. The inclusion
#' Bayes factor tests the combined evidence for all alternative components against all
#' null components. The detailed summary output shows posterior probabilities and 
#' inclusion Bayes factors for each individual component separately, allowing finer-grained
#' inference about which specific prior specification is most supported by the data.
#'
#' ### Omitting hypothesis components
#'
#' For top-level mixture components, setting a prior argument to `NULL` or `FALSE`
#' removes that hypothesis component:
#' \itemize{
#'   \item `prior_effect_null = NULL`: No null hypothesis for effect (assumes effect exists)
#'   \item `prior_effect = NULL`: No alternative hypothesis (assumes no effect)
#'   \item `prior_bias_null = NULL` or `FALSE`: No "no bias" model (assumes some bias mechanism)
#'   \item `prior_bias = NULL` or `FALSE`: No bias model (assumes no bias mechanism)
#' }
#' For moderator and scale regression terms, an omitted or whole-argument `NULL`
#' `prior_mods` / `prior_scale` means "use defaults". To omit a component for a
#' specific term, use a named list entry such as `prior_mods_null = list(x1 = NULL)`.
#'
#' ### Parameter-specific customization for moderators
#'
#' For moderator priors (`prior_mods`, `prior_mods_null`), you can specify different
#' priors for different predictors using named lists. A single prior object applies
#' to all moderator terms:
#' \preformatted{
#' RoBMA(...,
#'   mods = ~ x1 + x2,
#'   prior_mods_null = list(x1 = NULL)  # No null for x1, default null for x2
#' )
#' }
#' This allows testing whether specific moderators have effects while assuming others
#' are always included or excluded.
#'
#' ## Mixture priors for different parameter types
#'
#' ### Effect size (\eqn{\mu}) and heterogeneity (\eqn{\tau})
#'
#' Effect and heterogeneity priors combine spike-and-slab components:
#' \preformatted{
#' # Default: spike(0) null + normal alternative
#' # Custom: multiple alternatives with different scales
#' RoBMA(...,
#'   prior_effect = list(
#'     prior("normal", list(mean = 0, sd = 0.5)),
#'     prior("normal", list(mean = 0, sd = 1.0))
#'   )
#' )
#' }
#'
#' ### Publication bias
#'
#' Bias priors combine different publication bias adjustment mechanisms:
#' \itemize{
#'   \item No publication bias (null hypothesis)
#'   \item Weight functions: Selection models with different step functions
#'   \item PET/PEESE: Regression-based bias adjustment
#' }
#'
#' The `model_type` argument provides convenient presets:
#' \preformatted{
#' RoBMA(..., model_type = "PSMA")  # Full ensemble (default)
#' RoBMA(..., model_type = "PP")    # Only PET-PEESE models
#' }
#' See \code{\link{publication_bias_prior_specification}} for the bias-prior
#' constructors and preset model spaces.
#'
#' ### Heterogeneity allocation (\eqn{\rho}) for multilevel models
#'
#' When `cluster` is specified, \eqn{\rho} controls the cluster-level variance
#' allocation. The decomposition is
#' \eqn{\tau_{between} = \tau\sqrt{\rho}} and
#' \eqn{\tau_{within} = \tau\sqrt{1 - \rho}}. By default, only an alternative
#' (Beta(1,1)) is specified, but null hypotheses can be added:
#' \preformatted{
#' RoBMA(...,
#'   cluster = study,
#'   prior_heterogeneity_allocation_null = prior("spike", list(location = 0.5))
#' )
#' }
#'
#' ### Moderator and scale regression terms
#'
#' When `mods` is specified, the effect prior becomes the intercept prior, and each
#' moderator receives a mixture prior. The **intercept** inherits the effect mixture
#' structure, while **regression coefficients** get separate spike-and-slab mixtures.
#'
#' Similarly, when `scale = ~ ...` is specified for location-scale models, the
#' heterogeneity prior becomes the scale intercept prior, and each scale predictor
#' receives a mixture prior following the same pattern as moderator terms.
#' Note that the scale intercept (i.e., baseline heterogeneity) is always positive and 
#' the multiplicative scale coefficients require non-zero prior distribution on the 
#' intercept.
#' 
#' ## Model weights and prior odds
#'
#' Each component in a mixture prior has an associated prior weight (prior model
#' probability). User-specified components receive equal weights unless their
#' prior objects set `prior_weights`. Publication-bias presets use fixed weights:
#' `"2w"` and `"6w"` split the alternative bias mass equally across their
#' weight functions, `"PP"` splits it equally across PET and PEESE, and `"PSMA"`
#' assigns half of the alternative bias mass to the six weight functions
#' combined and one quarter each to PET and PEESE. Custom weights can be
#' specified via the `prior_weights` argument in individual prior objects.
#'
#' The prior odds between null and alternative hypotheses affect the Bayes factor
#' interpretation. With equal prior weights on null and alternative, the posterior
#' odds equal the Bayes factor.
#'
#' @references
#' \insertAllCited{}
#'
#' @seealso
#' \code{\link{prior_specification}} for base prior specification options,
#' \code{\link{publication_bias_prior_specification}} for publication-bias priors,
#' \code{\link{RoBMA}} for the main model-averaging function,
#' \code{\link[BayesTools]{prior}} for creating prior distribution objects
#'
#' @examples
#' \dontrun{
#' if (requireNamespace("metadat", quietly = TRUE)) {
#'   data(dat.lehmann2018, package = "metadat")
#'
#'   # Default mixture priors (spike-and-slab for effect and heterogeneity)
#'   fit1 <- RoBMA(
#'     yi      = yi,
#'     vi      = vi,
#'     data    = dat.lehmann2018,
#'     measure = "SMD",
#'     seed    = 1,
#'     silent  = TRUE
#'   )
#'
#'   # Custom alternative priors (two effect size scales)
#'   fit2 <- RoBMA(
#'     yi           = yi,
#'     vi           = vi,
#'     data         = dat.lehmann2018,
#'     measure      = "SMD",
#'     prior_effect = list(
#'       prior("normal", list(mean = 0, sd = 0.35)),
#'       prior("normal", list(mean = 0, sd = 0.70))
#'     ),
#'     seed         = 1,
#'     silent       = TRUE
#'   )
#'
#'   # No null hypothesis for effect (assume effect exists)
#'   fit3 <- RoBMA(
#'     yi                = yi,
#'     vi                = vi,
#'     data              = dat.lehmann2018,
#'     measure           = "SMD",
#'     prior_effect_null = NULL,
#'     seed              = 1,
#'     silent            = TRUE
#'   )
#'
#'   # Only selection model bias adjustment (no PET-PEESE)
#'   fit4 <- RoBMA(
#'     yi         = yi,
#'     vi         = vi,
#'     data       = dat.lehmann2018,
#'     measure    = "SMD",
#'     model_type = "6w",
#'     seed       = 1,
#'     silent     = TRUE
#'   )
#'
#'   # Meta-regression with different null specifications per moderator
#'   fit5 <- RoBMA(
#'     yi              = yi,
#'     vi              = vi,
#'     mods            = ~ Preregistered,
#'     data            = dat.lehmann2018,
#'     measure         = "SMD",
#'     prior_mods_null = list(Preregistered = NULL),
#'     seed            = 1,
#'     silent          = TRUE
#'   )
#' }
#' }
#'
#' @aliases RoBMA_prior_specification
NULL




# stop if public fitting constructors did not receive an explicit measure
.stop_missing_measure <- function(caller) {

  stop(
    paste0(
      caller, " requires explicit 'measure'. Use 'GEN' only for general ",
      "effect sizes; otherwise specify one of 'SMD', 'ZCOR', 'RR', 'OR', ",
      "'HR', 'RD', or 'IRR'."
    ),
    call. = FALSE
  )
}

# verify that the the measure is available
.check_measure <- function(measure, class = "norm") {

  BayesTools::check_char(measure, "measure")

  if (identical(class, "glmm")) {
    if (!measure %in% c("OR", "IRR")) {
      stop(
        "GLMM models support only 'OR' and 'IRR' measures.",
        call. = FALSE
      )
    }
    return()
  }

  BayesTools::check_char(
    measure,
    "measure",
    allow_values = c("SMD", "ZCOR", "RR", "OR", "HR", "RD", "IRR", "GEN")
  )

  return()
}


### prior scale functions
.get_unit_information_sd <- function(data, measure) {

  if (measure %in% c("SMD", "ZCOR", "RR", "OR", "HR", "IRR")) {
    # use the pre-specified UISD for known effect sizes
    prior_unit_information_sd <- .get_unit_information_sd.known(measure)
  } else {
    # compute sd based on the sample sizes for the general effect sizes
    ni <- data[["outcome"]][["ni"]]
    if (all(is.na(ni))) {
      stop("Sample size 'ni' or unit information sd 'unit_information_sd' must be specified to set-up prior distributions without known UISD.", call. = FALSE)
    }
    if (any(is.na(ni))) {
      stop("Sample size 'ni' must be complete to estimate unit information sd for measures without known UISD.", call. = FALSE)
    }
    prior_unit_information_sd <- .get_unit_information_sd.norm(sei = data[["outcome"]][["sei"]], ni = ni)

  }

  return(prior_unit_information_sd)
}

# returns the sd of unit information
# based on Chapter 2.4 in Spiegelhalter, Abrams, and Myles 2004 and Chapter 1 in Grieve 2022
# as reported in Table 2 of Pawel, S., & Held, L. (2025). Closed-form power and sample size calculations for Bayes factors. The American Statistician 9(3), 330-344, 10.1080/00031305.2025.2467919
.get_unit_information_sd.known <- function(measure) {

  BayesTools::check_char(measure, "measure", allow_values = c("SMD", "ZCOR", "RR", "OR", "HR", "IRR"), allow_NA = FALSE)

  return(switch(
    measure,
    "SMD"  = sqrt(2),
    "ZCOR" = 1,
    "RR"   = sqrt(4),
    "OR"   = sqrt(4),
    "HR"   = sqrt(4),
    "IRR"  = sqrt(4)
  ))
}

# computes the unit information based on data
# based on Eq. 6 of Röver, C., Bender, R., Dias, S., Schmid, C. H., Schmidli, H., Sturtz, S., ... & Friede, T. (2021). On weakly informative prior distributions for the heterogeneity parameter in Bayesian random‐effects meta‐analysis. Research Synthesis Methods, 12(4), 448-474. 10.1002/jrsm.1475
.get_unit_information_sd.norm <- function(sei, ni) {

  UISD <- sqrt(sum(ni) / sum(sei^(-2)))

  return(UISD)
}

# computes the unit information based on Poisson (rate) data
# for IRR effect size with GLMM models
.get_unit_information_sd.pois <- function(x1i, x2i, t1i, t2i) {

  # Total events represents the Total Precision (Fisher Information) for the log-rate
  total_events <- sum(x1i + x2i)

  # Total person-time represents the Total Sample Size
  total_time <- sum(t1i + t2i)

  # UISD: sqrt(Total N / Total Precision)
  # Derived from Fisher Information I(alpha) = lambda * t
  # Unit variance approx 1/(lambda * 1_unit_time) -> SD = sqrt(1/lambda) -> sqrt(t/y)
  UISD <- sqrt(total_time / total_events)

  return(UISD)
}


#' @title Estimate Unit Information Standard Deviation
#'
#' @description Estimates the unit information standard deviation (UISD) from
#' sample sizes and standard errors. The UISD is used to scale weakly informative
#' prior distributions for meta-analytic parameters.
#'
#' @param sei a complete numeric vector of strictly positive standard errors
#' for each study.
#' @param ni a complete numeric vector of strictly positive sample sizes for
#' each study, with the same length as `sei`.
#'
#' @details
#' The unit information standard deviation is computed following Equation 6 in
#' \insertCite{rover2021weakly;textual}{RoBMA}:
#' \deqn{\text{UISD} = \sqrt{\frac{\sum n_i}{\sum \text{se}_i^{-2}}}}
#' where \eqn{n_i} is the sample size and \eqn{\text{se}_i} is the standard error
#' for each study.
#'
#' This function is useful when you want to compute the UISD once and pass it
#' to multiple analyses via the `prior_unit_information_sd` argument in
#' [brma()] or related functions. This ensures consistent prior scaling
#' across analyses performed on different subsets of the same data.
#'
#' @references
#' \insertAllCited{}
#'
#' @return Returns a single numeric value representing the estimated unit
#' information standard deviation.
#'
#' @examples
#' # Example with simulated data
#' sei <- c(0.2, 0.3, 0.25, 0.15)
#' ni  <- c(50, 30, 40, 80)
#' estimate_unit_information_sd(sei = sei, ni = ni)
#'
#' @export
estimate_unit_information_sd <- function(sei, ni) {

  BayesTools::check_real(sei, "sei", lower = 0, allow_bound = FALSE, allow_NULL = FALSE, allow_NA = FALSE, check_length = FALSE)
  BayesTools::check_real(ni,  "ni",  lower = 0, allow_bound = FALSE, allow_NULL = FALSE, allow_NA = FALSE, check_length = length(sei))

  return(.get_unit_information_sd.norm(sei = sei, ni = ni))
}

### assign prior distributions
.assign_prior.simple                   <- function(
    prior, parameter, measure, data, prior_unit_information_sd,
    prior_informed_field, prior_informed_subfield, rescale_priors) {


  ### always use the user-specified prior if possible
  if (!missing(prior)) {

    # allow using FALSE/NULL arguments to set spike(0) prior distributions
    if (is.null(prior) || isFALSE(prior)) {
      return(BayesTools::prior("spike", parameters = list(0)))
    }

    # apply rescaling if specified
    prior <- .rescale_prior_object(prior, rescale_priors)

    # check the specified prior distribution
    prior <- switch(
      parameter,
      "effect"        = .check_prior.effect(prior),
      "heterogeneity" = .check_prior.heterogeneity(prior)
    )

    return(prior)
  }

  ### use informed prior distributions if specified
  if (!missing(prior_informed_field)) {

    if (prior_informed_field == "medicine") {

      # input translation for BayesTools::prior_informed
      if (missing(prior_informed_subfield)) {
        prior_name <- "Cochrane"
      } else {
        prior_name <- prior_informed_subfield
      }

      type <- switch(
        tolower(measure),
        "smd"   = "smd",
        "or"    = "logOR",
        "rr"    = "logRR",
        "rd"    = "RD",
        "hr"    = "logHR"
      )
      if (is.null(type)) {
        stop("Informed medicine priors are only available for measures 'SMD', 'OR', 'RR', 'RD', and 'HR'.", call. = FALSE)
      }

      # simple informed priors from BayesTools package
      prior <- BayesTools::prior_informed(prior_name, parameter = parameter, type = type)
    }

    # apply rescaling if specified
    prior <- .rescale_prior_object(prior, rescale_priors)
    return(prior)
  }

  ### use prior_unit_information_sd otherwise
  if (missing(prior_unit_information_sd)) {
    # if not passed directly, obtain from data
    prior_unit_information_sd <- .get_unit_information_sd(data = data, measure = measure)
  }

  # get the prior standard deviation based on unit information
  prior_sd <- switch(
    parameter,
    "effect"        = prior_unit_information_sd * RoBMA.get_option("default_UISD.effect"),
    "heterogeneity" = prior_unit_information_sd * RoBMA.get_option("default_UISD.heterogeneity")
  )

  # create to corresponding prior
  prior <- switch(
    parameter,
    "effect"        = BayesTools::prior("normal", parameters = list("mean" = 0, "sd" = prior_sd)),
    "heterogeneity" = BayesTools::prior("normal", parameters = list("mean" = 0, "sd" = prior_sd), truncation = list(0, Inf))
  )

  # apply rescaling if specified
  prior <- .rescale_prior_object(prior, rescale_priors)

  return(prior)
}
.assign_prior.heterogeneity_allocation <- function(prior) {

  # use default settings if prior distribution is not specified
  if (missing(prior) || is.null(prior)) {
    prior <- BayesTools::prior("beta", parameters = list("alpha" = 1, "beta" = 1))
    return(prior)
  }

  # check the user specified prior distribution
  prior <- .check_prior.restricted_01(prior, prior_name = "heterogeneity_allocation")

  return(prior)
}
.assign_prior.baserate                 <- function(prior) {

  # use default settings if prior distribution is not specified
  if (missing(prior) || is.null(prior)) {
    prior <- BayesTools::prior_factor("beta", parameters = list("alpha" = 1, "beta" = 1), contrast = "independent")
    return(prior)
  }

  # check the user specified prior distribution
  prior <- .check_prior.restricted_01(prior, prior_name = "baserate")

  # transform the prior into independent factor prior
  prior <- BayesTools::prior_factor(prior[["distribution"]], parameters = prior[["parameters"]], truncation = prior[["truncation"]], contrast = "independent")

  return(prior)
}
.assign_prior.lograte                  <- function(prior, data) {
  # the log-rate is unbounded, so use a proper data-based default prior
  # rather than an improper flat nuisance-parameter prior

  if (missing(prior) || is.null(prior)) {
    return(.get_unit_information_lograte_prior(
      x1i = data[["outcome"]][["x1i"]],
      x2i = data[["outcome"]][["x2i"]],
      t1i = data[["outcome"]][["t1i"]],
      t2i = data[["outcome"]][["t2i"]])
    )
  }

  # check the user specified prior distribution
  prior <- .check_prior.simple_or_point(prior, prior_name = "prior_lograte")

  # transform the prior into independent factor prior
  prior <- BayesTools::prior_factor(prior[["distribution"]], parameters = prior[["parameters"]], truncation = prior[["truncation"]], contrast = "independent")

  return(prior)
}
.get_PEESE_UISD_ratio                  <- function(measure, data, prior_unit_information_sd) {

  UISD_SMD <- .get_unit_information_sd.known("SMD")
  if (missing(prior_unit_information_sd)) {
    UISD_measure <- .get_unit_information_sd(data = data, measure = measure)
  } else {
    UISD_measure <- prior_unit_information_sd
  }

  return(UISD_SMD / UISD_measure)
}
.assign_prior.bias                     <- function(prior, measure, data, prior_unit_information_sd, bias_type, steps) {

  # assigns either a weight function or PET/PEESE model based on bias_type
  # there is no scaling of the publication bias priors based on rescale_priors
  # (only PEESE prior is rescaled to the match the UISD of the measure)

  if (bias_type == "selmodel") {

    # specify default settings / check the prior distribution
    if (missing(prior) || is.null(prior)) {
      if (missing(steps)) {
        steps <- c(0.025)
      }
      prior <- BayesTools::prior_weightfunction(
        "one-sided",
        steps   = steps,
        weights = BayesTools::wf_cumulative(
          alpha = rep(RoBMA.get_option("default_bias_weightfunction.alpha"), length(steps) + 1)
        )
      )
    } else if (!.prior_is_selection_kernel(prior)) {
      stop(
        "'prior_bias' must be a `prior_weightfunction` or `prior_bias` object",
        call. = FALSE
      )
    } else if (.prior_has_phacking(prior)) {
      .selection_stop_phacking_deferred()
    }

    return(prior)
  }

  if (bias_type == "PET") {

    # specify default settings / check the prior distribution
    if (missing(prior) || is.null(prior)) {
      # PET prior is invariant to different effect size types
      prior <- BayesTools::prior_PET("cauchy", parameters = list("location" = 0, "scale" = RoBMA.get_option("default_bias_PET.scale")), truncation = list(0, Inf))
    } else if (!BayesTools::is.prior.PET(prior)) {
      stop("'prior_bias' must be a `prior_PET` object", call. = FALSE)
    }

    return(prior)
  }

  if (bias_type == "PEESE") {

    # specify default settings / check the prior distribution
    if (missing(prior) || is.null(prior)) {
      ### PEESE prior is inversely related to the specified effect size
      # for SMD we use Cauchy with scale 1
      # for other effect sizes, we re-scale by the corresponding UISD
      UISD_ratio <- .get_PEESE_UISD_ratio(
        measure = measure, data = data, prior_unit_information_sd = prior_unit_information_sd)
      prior <- BayesTools::prior_PEESE("cauchy", parameters = list("location" = 0, "scale" = RoBMA.get_option("default_bias_PEESE.scale") * UISD_ratio), truncation = list(0, Inf))
    } else if (!BayesTools::is.prior.PEESE(prior)) {
      stop("'prior_bias' must be a `prior_PEESE` object", call. = FALSE)
    }

    return(prior)
  }

  if (bias_type == "none") {
    # for simplified handling in RoBMA
    return(prior)
  }
}
.term_default_continuous_prior         <- function(prior) {

  if (!BayesTools::is.prior.factor(prior)) {
    return(prior)
  }

  distribution <- prior[["distribution"]]
  if (distribution %in% c("mnormal", "mcauchy", "mt")) {
    distribution <- substring(distribution, 2)
  }

  return(BayesTools::prior(
    distribution = distribution,
    parameters   = prior[["parameters"]],
    truncation   = prior[["truncation"]]
  ))
}
.term_default_factor_prior             <- function(prior, contrast) {

  if (BayesTools::is.prior.factor(prior)) {
    return(prior)
  }

  distribution <- prior[["distribution"]]
  if (contrast != "treatment" && !distribution %in% c("spike", "mnormal", "mcauchy", "mt")) {
    distribution <- paste0("m", distribution)
  }

  return(BayesTools::prior_factor(
    distribution = distribution,
    parameters   = prior[["parameters"]],
    truncation   = prior[["truncation"]],
    contrast     = contrast
  ))
}
.assign_prior_list.terms               <- function(
    prior_list, prior_intercept, parameter, measure, data, prior_unit_information_sd,
    prior_informed_field, prior_informed_subfield, rescale_priors,
    set_default_spike = FALSE) {

  # set_default_spike argument is used when calling from .assign_prior_list.terms_mixture
  # to set-up prior distributions under the null hypotheses

  ### for scale regression, the intercept cannot be spike(0) since it's used as log(intercept)
  # in exp( log(intercept) + sum(beta_i * x_i) )
  if (parameter == "scale") {
    if (BayesTools::is.prior.point(prior_intercept) && mean(prior_intercept) == 0) {
      stop("Intercept prior distribution for scale regression cannot be set to 0 since the model is parameterized as tau = exp( log(intercept) + sum(beta_i * x_i) ).", call. = FALSE)
    }
  }
  if (parameter == "mods" &&
      attr(stats::terms(attr(data[[parameter]], "formula")), "intercept") == 0) {
    prior_intercept <- BayesTools::prior("spike", parameters = list(0))
  }

  ### check the user-specified priors
  user_default_prior <- NULL
  if (!missing(prior_list) && BayesTools::is.prior(prior_list)) {
    .check_prior.term(prior_list)
    user_default_prior <- prior_list
    prior_list         <- list()
  } else if (!missing(prior_list) && is.null(prior_list)) {
    prior_list <- list()
  } else if (!missing(prior_list)) {
    # check the priors and apply rescaling if specified
    for (i in seq_along(prior_list)) {
      .check_prior.term(prior_list[[i]])
      prior_list[[i]] <- .rescale_prior_object(prior_list[[i]], rescale_priors)
    }
  } else {
    # create an empty placeholder instead
    prior_list <- list()
  }

  ### add intercept to be beginning of the list (parsed previously)
  prior_list <- c(list(intercept = prior_intercept), prior_list)

  ### add default priors for remaining terms if left unspecified
  # the assignment of missing priors is performed within BayesTools::JAGS_formula
  # (it's easier to check and fill in the formula on assignment when it's being parsed)

  # use informed prior distributions if specified
  # otherwise rely on unit information based priors
  if (set_default_spike) {

    # set_default_spike argument is used when calling from .assign_prior_list.terms_mixture
    # to set-up prior distributions under the null hypotheses
    default_prior_continuous <- BayesTools::prior("spike", parameters = list(0))
    default_prior_factor     <- BayesTools::prior_factor("spike", parameters = list(0), contrast = attr(data, "set_contrast_factor_predictors"))

  } else if (!missing(prior_informed_field)) {

    if (prior_informed_field == "medicine") {

      # input translation for BayesTools::prior_informed
      if (missing(prior_informed_subfield)) {
        prior_name <- "Cochrane"
      } else {
        prior_name <- prior_informed_subfield
      }

      type <- switch(
        tolower(measure),
        "smd"   = "smd",
        "or"    = "logOR",
        "rr"    = "logRR",
        "rd"    = "RD",
        "hr"    = "logHR"
      )
      if (is.null(type)) {
        stop("Informed medicine priors are only available for measures 'SMD', 'OR', 'RR', 'RD', and 'HR'.", call. = FALSE)
      }

      if (parameter == "mods") {
        # treat moderators by scaling them according to 1/2 of the effect
        default_prior_continuous <- BayesTools::prior_informed(prior_name, parameter = "effect", type = type)
        default_prior_continuous <- .rescale_prior_object(default_prior_continuous, RoBMA.get_option("default_informed_priors.mods"))
      } else if (parameter == "scale") {
        # scale parameters use the default scaling (no informed priors exist)
        default_prior_continuous <- BayesTools::prior("normal", parameters = list("mean" = 0, "sd" = RoBMA.get_option("default_informed_priors.scale")))
      }

      # transform the continuous prior distributions into prior distributions for factors
      if (attr(data, "set_contrast_factor_predictors") == "treatment") {
        default_prior_factor <- BayesTools::prior_factor(
          distribution = default_prior_continuous[["distribution"]],
          parameters   = default_prior_continuous[["parameters"]],
          contrast     = "treatment"
        )
      } else {
        default_prior_factor <- BayesTools::prior_factor(
          distribution = paste0("m", default_prior_continuous[["distribution"]]),
          parameters   = default_prior_continuous[["parameters"]],
          contrast     = attr(data, "set_contrast_factor_predictors")
        )
      }

    }

  } else {

    ### use prior_unit_information_sd otherwise
    if (missing(prior_unit_information_sd)) {
      # if not passed directly, obtain from data
      prior_unit_information_sd <- .get_unit_information_sd(data = data, measure = measure)
    }

    # get the prior standard deviation based on unit information
    prior_sd <- switch(
      parameter,
      "mods"          = prior_unit_information_sd * RoBMA.get_option("default_UISD.mods"),
      "scale"         = RoBMA.get_option("default_UISD.scale") # moderators for scale do not depend on UISD (multiplicative)
    )

    default_prior_continuous <- BayesTools::prior("normal", parameters = list("mean" = 0, "sd" = prior_sd))
    if (attr(data, "set_contrast_factor_predictors") == "treatment") {
      default_prior_factor <- BayesTools::prior_factor(
        distribution = "normal",
        parameters   = list("mean" = 0, "sd" = prior_sd),
        contrast     = "treatment"
      )
    } else {
      default_prior_factor <- BayesTools::prior_factor(
        distribution = "mnormal",
        parameters   = list("mean" = 0, "sd" = prior_sd),
        contrast     = attr(data, "set_contrast_factor_predictors")
      )
    }
  }

  if (!is.null(user_default_prior)) {
    default_prior_continuous <- .term_default_continuous_prior(user_default_prior)
    default_prior_factor     <- .term_default_factor_prior(
      prior    = user_default_prior,
      contrast = attr(data, "set_contrast_factor_predictors")
    )
  }

  # apply rescaling if specified
  default_prior_continuous <- .rescale_prior_object(default_prior_continuous, rescale_priors)
  default_prior_factor     <- .rescale_prior_object(default_prior_factor, rescale_priors)

  # list the priors into the default positions
  prior_list[["__default_continuous"]] <- default_prior_continuous
  prior_list[["__default_factor"]]     <- default_prior_factor

  ### use BayesTools::JAGS_formula to obtain the assigned list of priors
  prior_list <- BayesTools::JAGS_formula(
    parameter  = switch (parameter, "mods" = "mu", "scale" = "log_tau"),
    data       = data[[parameter]],
    formula    = attr(data[[parameter]], "formula"),
    prior_list = prior_list
  )[["prior_list"]]
  names(prior_list) <- BayesTools::format_parameter_names(
    parameters         = names(prior_list),
    formula_parameters = switch (parameter, "mods" = "mu", "scale" = "log_tau"),
    formula_prefix     = FALSE
  )

  return(prior_list)
}

# Remove invalid fixed-effect heterogeneity components for scale regression.
.remove_zero_scale_intercept_prior <- function(prior_intercept) {

  is_zero_point <- function(prior) {
    BayesTools::is.prior.point(prior) && mean(prior) == 0
  }

  if (!BayesTools::is.prior.mixture(prior_intercept)) {
    if (is_zero_point(prior_intercept)) {
      stop(
        "At least one non-zero between-study heterogeneity prior must be defined for scale regression.",
        call. = FALSE
      )
    }
    return(prior_intercept)
  }

  keep <- !vapply(prior_intercept, is_zero_point, logical(1))

  if (all(keep))
    return(prior_intercept)

  warning(
    "Spike(0) prior distribution for between-study heterogeneity tau was removed since scale regression requires non-zero between-study heterogeneity.",
    immediate. = TRUE
  )

  if (!any(keep)) {
    stop(
      "At least one non-zero between-study heterogeneity prior must be defined for scale regression.",
      call. = FALSE
    )
  }

  components <- attr(prior_intercept, "components")
  prior_list <- prior_intercept[keep]
  attributes(prior_list) <- NULL

  return(BayesTools::prior_mixture(
    prior_list = prior_list,
    components = components[keep]
  ))
}
### assign mixture prior distributions
.assign_prior.simple_mixture                   <- function(
    prior, prior_null, parameter, measure, data, prior_unit_information_sd,
    prior_informed_field, prior_informed_subfield, rescale_priors) {

  ### check the alternative prior input
  # prior can be either
  # - unspecified (set default as in brma)
  # - NULL/FALSE (omit prior distribution)
  # - single prior distribution
  # - list of prior distributions
  # each of these prior distributions needs to satisfy the same properties as in brma
  # if only a single distribution is specified, place it in a list for a simpler treatment
  if (missing(prior) || BayesTools::is.prior(prior)) {
    priors_alt <- list(.assign_prior.simple(
      prior = prior, parameter = parameter, measure = measure,
      data = data, prior_unit_information_sd = prior_unit_information_sd,
      prior_informed_field = prior_informed_field, prior_informed_subfield = prior_informed_subfield,
      rescale_priors = rescale_priors
    ))
  } else if (is.null(prior) || isFALSE(prior)) {
    priors_alt <- list()
  } else if (is.list(prior) && all(sapply(prior, BayesTools::is.prior))) {
    priors_alt <- prior
    for (i in seq_along(priors_alt)) {
      priors_alt[[i]] <- .assign_prior.simple(
        prior = priors_alt[[i]], parameter = parameter, measure = measure,
        data = data, prior_unit_information_sd = prior_unit_information_sd,
        prior_informed_field = prior_informed_field, prior_informed_subfield = prior_informed_subfield,
        rescale_priors = rescale_priors
      )
    }
  } else {
    stop(gettextf("The 'prior_%1$s' must be either prior distribution or a list of prior distributions.", parameter), call. = FALSE)
  }

  ### check the null prior input
  # prior can be either
  # - empty (set point prior as default)
  # - NULL/FALSE (omit prior distribution)
  # - single prior distribution
  # - list of prior distributions
  # each of these prior distributions needs to satisfy the same properties as in brma
  # if only a single distribution is specified, place it in a list for a simpler treatment
  if (missing(prior_null)) {
    priors_null <- list(BayesTools::prior("spike", parameters = list(0)))
  } else if (is.null(prior_null) || isFALSE(prior_null)) {
    priors_null <- list()
  } else if (BayesTools::is.prior(prior_null)) {
    priors_null <- list(.assign_prior.simple(
      prior = prior_null, parameter = parameter, measure = measure,
      data = data, prior_unit_information_sd = prior_unit_information_sd,
      prior_informed_field = prior_informed_field, prior_informed_subfield = prior_informed_subfield,
      rescale_priors = rescale_priors
    ))
  } else if (is.list(prior_null) && all(sapply(prior_null, BayesTools::is.prior))) {
    priors_null <- prior_null
    for (i in seq_along(priors_null)) {
      priors_null[[i]] <- .assign_prior.simple(
        prior = priors_null[[i]], parameter = parameter, measure = measure,
        data = data, prior_unit_information_sd = prior_unit_information_sd,
        prior_informed_field = prior_informed_field, prior_informed_subfield = prior_informed_subfield,
        rescale_priors = rescale_priors
      )
    }
  } else {
    stop(gettextf("The 'prior_%1$s_null' must be either prior distribution or a list of prior distributions.", parameter), call. = FALSE)
  }

  if (length(priors_null) + length(priors_alt) == 0) {
    stop(gettextf("At least one prior distribution needs to be defined for '%1$s'.", parameter), call. = FALSE)
  }

  ### specify the corresponding mixture prior
  prior <- BayesTools::prior_mixture(
    prior_list = c(priors_null, priors_alt),
    is_null    = c(rep(TRUE, length(priors_null)), rep(FALSE, length(priors_alt)))
  )

  return(prior)
}
.default_prior.bias_alt                         <- function(model_type, measure, data, prior_unit_information_sd) {

  if (model_type == "2w") {
    return(list(
      BayesTools::prior_weightfunction("two-sided", c(0.05),       BayesTools::wf_cumulative(c(1, 1)),    prior_weights = 1/2),
      BayesTools::prior_weightfunction("two-sided", c(0.05, 0.10), BayesTools::wf_cumulative(c(1, 1, 1)), prior_weights = 1/2)
    ))
  }

  if (model_type == "6w") {
    return(list(
      BayesTools::prior_weightfunction("two-sided", c(0.05),             BayesTools::wf_cumulative(c(1, 1)),       prior_weights = 1/6),
      BayesTools::prior_weightfunction("two-sided", c(0.05, 0.10),       BayesTools::wf_cumulative(c(1, 1, 1)),    prior_weights = 1/6),
      BayesTools::prior_weightfunction("one-sided", c(0.05),             BayesTools::wf_cumulative(c(1, 1)),       prior_weights = 1/6),
      BayesTools::prior_weightfunction("one-sided", c(0.025, 0.05),      BayesTools::wf_cumulative(c(1, 1, 1)),    prior_weights = 1/6),
      BayesTools::prior_weightfunction("one-sided", c(0.05, 0.5),        BayesTools::wf_cumulative(c(1, 1, 1)),    prior_weights = 1/6),
      BayesTools::prior_weightfunction("one-sided", c(0.025, 0.05, 0.5), BayesTools::wf_cumulative(c(1, 1, 1, 1)), prior_weights = 1/6)
    ))
  }

  UISD_ratio <- .get_PEESE_UISD_ratio(
    measure = measure, data = data, prior_unit_information_sd = prior_unit_information_sd)

  if (model_type == "PP") {
    return(list(
      BayesTools::prior_PET(distribution = "Cauchy",   parameters = list(0, RoBMA.get_option("default_bias_PET.scale")),                truncation = list(0, Inf), prior_weights = 1/2),
      BayesTools::prior_PEESE(distribution = "Cauchy", parameters = list(0, RoBMA.get_option("default_bias_PEESE.scale") * UISD_ratio), truncation = list(0, Inf), prior_weights = 1/2)
    ))
  }

  return(list(
    BayesTools::prior_weightfunction("two-sided", c(0.05),             BayesTools::wf_cumulative(c(1, 1)),       prior_weights = 1/12),
    BayesTools::prior_weightfunction("two-sided", c(0.05, 0.10),       BayesTools::wf_cumulative(c(1, 1, 1)),    prior_weights = 1/12),
    BayesTools::prior_weightfunction("one-sided", c(0.05),             BayesTools::wf_cumulative(c(1, 1)),       prior_weights = 1/12),
    BayesTools::prior_weightfunction("one-sided", c(0.025, 0.05),      BayesTools::wf_cumulative(c(1, 1, 1)),    prior_weights = 1/12),
    BayesTools::prior_weightfunction("one-sided", c(0.05, 0.5),        BayesTools::wf_cumulative(c(1, 1, 1)),    prior_weights = 1/12),
    BayesTools::prior_weightfunction("one-sided", c(0.025, 0.05, 0.5), BayesTools::wf_cumulative(c(1, 1, 1, 1)), prior_weights = 1/12),
    BayesTools::prior_PET(distribution = "Cauchy",   parameters = list(0, RoBMA.get_option("default_bias_PET.scale")),                truncation = list(0, Inf), prior_weights = 1/4),
    BayesTools::prior_PEESE(distribution = "Cauchy", parameters = list(0, RoBMA.get_option("default_bias_PEESE.scale") * UISD_ratio), truncation = list(0, Inf), prior_weights = 1/4)
  ))
}
.assign_prior.bias_mixture                     <- function(
    prior, prior_null, measure, data, prior_unit_information_sd, model_type) {

  ### use precanned publication bias adjustment models
  # if neither `prior` or `prior_null` input is specified use the `model_type`
  # there is no scaling of the publication bias priors based on rescale_priors
  # (only PEESE prior is rescaled to the match the UISD of the measure)

  if (missing(prior) && missing(prior_null)) {
    priors_null <- list(BayesTools::prior_none(prior_weights = 1))
    priors_alt  <- .default_prior.bias_alt(
      model_type = model_type, measure = measure, data = data,
      prior_unit_information_sd = prior_unit_information_sd)
    prior <- BayesTools::prior_mixture(
      prior_list = c(priors_null, priors_alt),
      is_null    = c(TRUE, rep(FALSE, length(priors_alt)))
    )

    return(prior)
  }


  ### use the user specified prior distributions if specified

  ### check the alternative prior input
  # prior can be either
  # - unspecified (set alternative part of PSMA)
  # - NULL/FALSE (omit prior distribution)
  # - single prior distribution
  # - list of prior distributions
  # each of these prior distributions needs to satisfy the same properties as in brma
  # if only a single distribution is specified, place it in a list for a simpler treatment
  if (missing(prior)) {
    priors_alt <- .default_prior.bias_alt(
      model_type = model_type, measure = measure, data = data,
      prior_unit_information_sd = prior_unit_information_sd)
  } else if (is.null(prior) || isFALSE(prior)) {
    priors_alt <- list()
  } else if (BayesTools::is.prior(prior)) {
    priors_alt <- list(.assign_prior.bias(
      prior = prior, measure = measure, data = data,
      prior_unit_information_sd = prior_unit_information_sd, bias_type = .get_prior_bias_type(prior)))
  } else if (is.list(prior) && all(sapply(prior, BayesTools::is.prior))) {
    priors_alt <- prior
    for (i in seq_along(priors_alt)) {
      priors_alt[[i]] <- .assign_prior.bias(
        prior = priors_alt[[i]], measure = measure, data = data,
        prior_unit_information_sd = prior_unit_information_sd, bias_type = .get_prior_bias_type(priors_alt[[i]]))
    }
  } else {
    stop("The 'prior_bias' must be either prior distribution or a list of prior distributions.", call. = FALSE)
  }

  ### check the null prior input
  # prior can be either
  # - empty (set point prior as default)
  # - NULL/FALSE (omit prior distribution)
  # - single prior distribution
  # - list of prior distributions
  # each of these prior distributions needs to satisfy the same properties as in brma
  # if only a single distribution is specified, place it in a list for a simpler treatment
  if (missing(prior_null)) {
    priors_null <- list(BayesTools::prior_none(prior_weights = 1))
  } else if (is.null(prior_null) || isFALSE(prior_null)) {
    priors_null <- list()
  } else if (BayesTools::is.prior(prior_null)) {
    priors_null <- list(.assign_prior.bias(
      prior = prior_null, measure = measure, data = data,
      prior_unit_information_sd = prior_unit_information_sd, bias_type = .get_prior_bias_type(prior_null)))
  } else if (is.list(prior_null) && all(sapply(prior_null, BayesTools::is.prior))) {
    priors_null <- prior_null
    for (i in seq_along(priors_null)) {
      priors_null[[i]] <- .assign_prior.bias(
        prior = priors_null[[i]], measure = measure, data = data,
        prior_unit_information_sd = prior_unit_information_sd, bias_type = .get_prior_bias_type(priors_null[[i]]))
    }
  } else {
    stop("The 'prior_bias_null' must be either prior distribution or a list of prior distributions.", call. = FALSE)
  }

  if (length(priors_null) + length(priors_alt) == 0) {
    stop("At least one prior distribution needs to be defined for 'bias'.", call. = FALSE)
  }

  if (length(priors_alt) == 0 &&
      length(priors_null) == 1 &&
      BayesTools::is.prior.none(priors_null[[1]])) {
    return(priors_null[[1]])
  }

  ### specify the corresponding mixture prior
  prior <- BayesTools::prior_mixture(
    prior_list = c(priors_null, priors_alt),
    is_null    = c(rep(TRUE, length(priors_null)), rep(FALSE, length(priors_alt)))
  )

  return(prior)
}
.assign_prior.heterogeneity_allocation_mixture <- function(prior, prior_null) {

  ### check the alternative prior input
  # prior can be either
  # - unspecified (set default as in brma)
  # - NULL/FALSE (omit prior distribution)
  # - single prior distribution
  # - list of prior distributions
  # each of these prior distributions needs to satisfy the same properties as in brma
  # if only a single distribution is specified, place it in a list for a simpler treatment
  if (missing(prior) || BayesTools::is.prior(prior)) {
    priors_alt <- list(.assign_prior.heterogeneity_allocation(prior = prior))
  } else if (is.null(prior) || isFALSE(prior)) {
    priors_alt <- list()
  } else if (is.list(prior) && all(sapply(prior, BayesTools::is.prior))) {
    priors_alt <- prior
    for (i in seq_along(priors_alt)) {
      priors_alt[[i]] <- .assign_prior.heterogeneity_allocation(prior = priors_alt[[i]])
    }
  } else {
    stop("The 'prior_heterogeneity_allocation' must be either prior distribution or a list of prior distributions.", call. = FALSE)
  }

  ### check the null prior input
  # prior can be either
  # - empty (omit prior as default)
  # - NULL/FALSE (omit prior distribution)
  # - single prior distribution
  # - list of prior distributions
  # each of these prior distributions needs to satisfy the same properties as in brma
  # if only a single distribution is specified, place it in a list for a simpler treatment
  if (missing(prior_null) || is.null(prior_null) || isFALSE(prior_null)) {
    priors_null <- list()
  } else if (BayesTools::is.prior(prior_null)) {
    priors_null <- list(.assign_prior.heterogeneity_allocation(prior = prior_null))
  } else if (is.list(prior_null) && all(sapply(prior_null, BayesTools::is.prior))) {
    priors_null <- prior_null
    for (i in seq_along(priors_null)) {
      priors_null[[i]] <- .assign_prior.heterogeneity_allocation(prior = priors_null[[i]])
    }
  } else {
    stop("The 'prior_heterogeneity_allocation_null' must be either prior distribution or a list of prior distributions.", call. = FALSE)
  }

  if (length(priors_null) + length(priors_alt) == 0) {
    stop("At least one prior distribution needs to be defined for 'heterogeneity_allocation'.", call. = FALSE)
  }

  ### specify the corresponding mixture prior
  prior <- BayesTools::prior_mixture(
    prior_list = c(priors_null, priors_alt),
    is_null    = c(rep(TRUE, length(priors_null)), rep(FALSE, length(priors_alt)))
  )

  return(prior)
}
.assign_prior_list.terms_mixture               <- function(
    prior_list, prior_list_null, prior_intercept, parameter, measure, data, prior_unit_information_sd,
    prior_informed_field, prior_informed_subfield, rescale_priors) {

  # in contrast to effect/heterogeneity prior list settings terms currently only allow a single prior for alternative
  # and a single prior for the null hypothesis for each term

  # one issue when re-using .assign_prior_list.terms is that we need to forward prior for the intercept which is
  # an already parsed mixture prior - when merging mixture priors for terms, we ought to not duplicate the intercept

  # we want to allow NULL/FALSE to omit prior distribution for a given component, .assign_prior_list.terms
  # fills all missing prior distributions with default priors
  # as such, we store names of parameters that should be omited and remove them manually
  if (!missing(prior_list) && !BayesTools::is.prior(prior_list) && length(prior_list) > 0 && is.list(prior_list)) {
    par_alt_omit <- names(prior_list)[sapply(prior_list, function(x) is.null(x) || isFALSE(x))]
    # filter out NULL/FALSE values before passing to .assign_prior_list.terms
    prior_list <- prior_list[!sapply(prior_list, function(x) is.null(x) || isFALSE(x))]
  } else {
    par_alt_omit <- NULL
  }
  if (!missing(prior_list_null) && !BayesTools::is.prior(prior_list_null) && length(prior_list_null) > 0 && is.list(prior_list_null)) {
    par_null_omit <- names(prior_list_null)[sapply(prior_list_null, function(x) is.null(x) || isFALSE(x))]
    # filter out NULL/FALSE values before passing to .assign_prior_list.terms
    prior_list_null <- prior_list_null[!sapply(prior_list_null, function(x) is.null(x) || isFALSE(x))]
  } else {
    par_null_omit <- NULL
  }


  if (parameter == "scale") {
    # in the case of scale moderation, spike at zero prior (i.e., fixed-effect model) is not possible since the intercept
    # is on a log scale
    # remove spike(0) priors and rebuild mixture metadata
    prior_intercept <- .remove_zero_scale_intercept_prior(prior_intercept)
  }
  if (parameter == "mods" &&
      attr(stats::terms(attr(data[[parameter]], "formula")), "intercept") == 0) {
    prior_intercept <- BayesTools::prior("spike", parameters = list(0))
  }

  prior_list_alt <- .assign_prior_list.terms(
    prior_list = prior_list, prior_intercept = prior_intercept, parameter = parameter,
    measure = measure, data = data,  prior_unit_information_sd = prior_unit_information_sd,
    prior_informed_field = prior_informed_field, prior_informed_subfield = prior_informed_subfield,
    rescale_priors = rescale_priors)

  prior_list_null <- .assign_prior_list.terms(
    prior_list = prior_list_null, prior_intercept = prior_intercept, parameter = parameter,
    measure = measure, data = data,  prior_unit_information_sd = prior_unit_information_sd,
    prior_informed_field = prior_informed_field, prior_informed_subfield = prior_informed_subfield,
    rescale_priors = rescale_priors, set_default_spike = TRUE)

  ### initiate the list of prior mixtures with the intercept
  prior_list <- list(intercept = prior_intercept)
  for (par in names(prior_list_alt)) {

    # skip the already assigned intercept
    if (par == "intercept") {
      next
    }

    # create temporal holder for the parameter
    temp_prior_list  <- list()
    temp_is_null     <- NULL

    if (!par %in% par_null_omit) {
      temp_prior_list <- c(temp_prior_list, list(prior_list_null[[par]]))
      temp_is_null    <- c(temp_is_null, TRUE)
    }

    if (!par %in% par_alt_omit) {
      temp_prior_list <- c(temp_prior_list, list(prior_list_alt[[par]]))
      temp_is_null    <- c(temp_is_null, FALSE)
    }

    if (length(temp_prior_list) == 0) {
      stop(sprintf("At least one prior distribution needs to be defined for '%1$s' parameter in the '%2$s' model.", par, parameter))
    }

    prior_list[[par]] <- BayesTools::prior_mixture(
      prior_list = temp_prior_list,
      is_null    = temp_is_null
    )
  }

  prior_list <- .repair_formula_prior_list(
    prior_list = prior_list,
    parameter  = switch(parameter, "mods" = "mu", "scale" = "log_tau")
  )

  return(prior_list)
}

# Restore BayesTools formula metadata lost when formula prior mixtures are rebuilt.
.repair_formula_prior_list <- function(prior_list, parameter) {

  if (is.null(prior_list)) {
    return(prior_list)
  }

  for (i in seq_along(prior_list)) {
    prior_list[[i]] <- .repair_formula_prior(
      prior    = prior_list[[i]],
      parameter = parameter
    )
  }

  return(prior_list)
}
.repair_formula_prior <- function(prior, parameter) {

  attr(prior, "parameter") <- parameter

  formula_attributes <- c(
    "interaction", "interaction_terms", "levels", "level_names",
    "term_components", "factor_terms", "factor_contrasts",
    "factor_design", "factor_cell_names", "multiply_by"
  )

  if (.has_formula_prior_components(prior)) {
    source_attributes <- NULL
    for (i in seq_along(prior)) {
      if (!is.null(attr(prior[[i]], "parameter"))) {
        source_attributes <- attributes(prior[[i]])
        break
      }
    }

    if (!is.null(source_attributes)) {
      for (attribute in formula_attributes) {
        if (is.null(attr(prior, attribute)) && !is.null(source_attributes[[attribute]])) {
          attr(prior, attribute) <- source_attributes[[attribute]]
        }
      }
    }

    for (i in seq_along(prior)) {
      attr(prior[[i]], "parameter") <- parameter
      for (attribute in formula_attributes) {
        if (is.null(attr(prior[[i]], attribute)) && !is.null(attr(prior, attribute))) {
          attr(prior[[i]], attribute) <- attr(prior, attribute)
        }
      }
    }
  }

  return(prior)
}
.has_formula_prior_components <- function(prior) {

  return(BayesTools::is.prior.mixture(prior) ||
    BayesTools::is.prior.spike_and_slab(prior))
}
### check specified prior distributions whether they follow basic requirements
.check_prior.effect                   <- function(prior) {

  prior <- .check_prior.simple_or_point(prior, prior_name = "prior_effect")

  return(prior)
}
.check_prior.heterogeneity            <- function(prior) {

  prior <- .check_prior.simple_or_point(prior, prior_name = "prior_heterogeneity")

  # check range restriction
  if (BayesTools::is.prior.point(prior) && mean(prior) < 0) {
    stop("Point prior distribution for 'prior_heterogeneity' must be at least 0.", call. = FALSE)
  }
  if (prior[["truncation"]][["lower"]] < 0) {
    warning("Lower truncation point for 'prior_heterogeneity' prior distribution must be at least 0. The prior distribution was modified.",
            immediate. = TRUE, call. = FALSE)
    prior[["truncation"]][["lower"]] <- 0
  }

  return(prior)
}
.check_prior.simple_or_point          <- function(prior, prior_name) {

  # check object type
  if (!BayesTools::is.prior(prior))
    stop(sprintf("The '%s' argument must be a prior distribution.", prior_name), call. = FALSE)
  if (!(BayesTools::is.prior.simple(prior) || BayesTools::is.prior.point(prior)))
    stop(sprintf("The '%s' argument must be a either a simple or point prior distribution.", prior_name), call. = FALSE)

  return(prior)
}
.check_prior.restricted_01            <- function(prior, prior_name) {

  prior <- .check_prior.simple_or_point(prior, prior_name)

  # check range restriction
  if (BayesTools::is.prior.point(prior) && (mean(prior) < 0 || mean(prior) > 1)) {
    stop(sprintf("Point prior distribution for '%s' must be within [0, 1].", prior_name), call. = FALSE)
  }
  if (prior[["truncation"]][["lower"]] < 0) {
    warning(sprintf("Lower truncation point for '%s' prior distribution must be at least 0. The prior distribution was modified.", prior_name),
            immediate. = TRUE, call. = FALSE)
    prior[["truncation"]][["lower"]] <- 0
  }

  if (prior[["truncation"]][["upper"]] > 1) {
    warning(sprintf("Upper truncation point for '%s' prior distribution must be at most 1. The prior distribution was modified.", prior_name),
            immediate. = TRUE, call. = FALSE)
    prior[["truncation"]][["upper"]] <- 1
  }

  return(prior)
}
.check_prior.term                     <- function(prior) {

  # check object type
  if (!BayesTools::is.prior(prior))
    stop("The 'prior_mods' / 'prior_scale' arguments must be a prior distribution.", call. = FALSE)
  if (!(BayesTools::is.prior.simple(prior) || BayesTools::is.prior.point(prior) || BayesTools::is.prior.factor(prior)))
    stop("The 'prior_mods' / 'prior_scale' arguments for predictors must be a either a simple, factor, or point prior distribution.", call. = FALSE)

  return()
}
.check_prior_specification_conflict   <- function(prior_unit_information_sd, prior_informed_field) {

  # check that both UISD and informed priors are not specified simultaneously
  if (!missing(prior_unit_information_sd) && !missing(prior_informed_field)) {
    stop(paste0(
      "Both 'prior_unit_information_sd' and 'prior_informed_field' were specified. ",
      "These arguments are mutually exclusive: 'prior_unit_information_sd' is used to scale default priors, ",
      "while 'prior_informed_field' uses pre-specified informed prior distributions that do not rely on UISD scaling."
    ), call. = FALSE)
  }

  return()
}
.get_prior_bias_type                  <- function(prior) {
  if (BayesTools::is.prior.none(prior)) {
    bias_type <- "none"
  } else if (.prior_is_selection_kernel(prior)) {
    bias_type <- "selmodel"
  } else if (BayesTools::is.prior.PET(prior)) {
    bias_type <- "PET"
  } else if (BayesTools::is.prior.PEESE(prior)) {
    bias_type <- "PEESE"
  } else {
    stop("The publication bias prior was not recognized", call. = FALSE)
  }
  return(bias_type)
}

# helper function for defining unit information prior on lograte for IRR models
.get_unit_information_lograte_prior   <- function(x1i, x2i, t1i, t2i) {

  if (anyNA(x1i) || anyNA(x2i) || anyNA(t1i) || anyNA(t2i)) {
    stop("Default 'prior_lograte' requires complete Poisson event counts and person-times.", call. = FALSE)
  }

  total_events <- sum(x1i + x2i)
  total_time   <- sum(t1i + t2i)
  if (!is.finite(total_events) || total_events <= 0) {
    stop("Default 'prior_lograte' requires at least one observed Poisson event. Specify 'prior_lograte' manually for all-zero event counts.", call. = FALSE)
  }
  if (!is.finite(total_time) || total_time <= 0) {
    stop("Default 'prior_lograte' requires positive finite total person-time.", call. = FALSE)
  }

  # compute the unit information standard deviation from Poisson data
  sigma_prior <- .get_unit_information_sd.pois(x1i, x2i, t1i, t2i)

  # compute the pooled crude log-rate for the mean
  mu_prior     <- log(total_events / total_time)
  if (!is.finite(mu_prior) || !is.finite(sigma_prior)) {
    stop("Default 'prior_lograte' could not be computed from the supplied Poisson data.", call. = FALSE)
  }

  # define the prior object
  prior <- BayesTools::prior_factor("normal", parameters = list(mean = mu_prior, sd = sigma_prior), contrast = "independent")

  return(prior)
}

# priors rescaling function
.rescale_prior_object <- function(prior, rescale_priors) {
  # re-scaling prior distributions

  # no need if re-scaling is not specified
  if (missing(rescale_priors) || rescale_priors == 1)
    return(prior)

  # no need for point / no prior distributions
  if (BayesTools::is.prior.point(prior) || BayesTools::is.prior.none(prior))
    return(prior)

  # only specific prior distributions can be rescaled
  # (UPGRADE: once BayesTools incorporates direct scaling, extend to all kinds of priors)
  can_be_rescaled <- c("normal", "mnormal", "cauchy", "mcauchy", "t", "mt", "invgamma")
  if (!is.element(prior[["distribution"]], can_be_rescaled))
    stop(sprintf(
      paste0("The '%1$s' prior distribution cannot be rescaled using the 'rescale_priors' argument. ",
             "Only one of the following distributions can be rescaled: %2$s"),
      prior[["distribution"]],
      paste0("'", can_be_rescaled, "'", sep = ", ")
    ), call. = FALSE)

  if (prior[["distribution"]] %in% c("normal", "mnormal")) {
    prior[["parameters"]][["sd"]] <- prior[["parameters"]][["sd"]] * rescale_priors
  } else if (prior[["distribution"]] %in% c("cauchy", "mcauchy", "t", "mt", "invgamma")) {
    prior[["parameters"]][["scale"]] <- prior[["parameters"]][["scale"]] * rescale_priors
  }

  return(prior)
}

### prior specification functions
.check_and_list_priors.brma <- function(
    prior_effect, prior_heterogeneity,
    prior_mods, prior_scale,
    prior_heterogeneity_allocation, prior_bias,
    prior_baserate, prior_lograte,
    rescale_priors,
    prior_unit_information_sd,
    prior_informed_field, prior_informed_subfield,
    data, measure,
    bias_type, steps) {

  ### check input
  if (!missing(rescale_priors))
    BayesTools::check_real(rescale_priors, "rescale_priors", lower = 0, allow_bound = FALSE, allow_NA = FALSE)
  if (!missing(prior_unit_information_sd))
    BayesTools::check_real(prior_unit_information_sd, "prior_unit_information_sd", lower = 0, allow_bound = FALSE, allow_NA = FALSE)
  if (!missing(prior_informed_field))
    BayesTools::check_char(prior_informed_field, "prior_informed_field", allow_values = c("medicine"), allow_NA = FALSE)
  if (!missing(prior_informed_subfield))
    BayesTools::check_char(prior_informed_subfield, "prior_informed_subfield", allow_NA = FALSE)
  .check_prior_specification_conflict(prior_unit_information_sd, prior_informed_field)
  if (!missing(steps))
    BayesTools::check_real(steps, "steps", lower = 0, upper = 1, allow_bound = FALSE, check_length = FALSE, allow_NA = FALSE)

  ### set prior distributions
  measure       <- .data_measure(data)
  prior_outcome <- list()

  prior_outcome[["mu"]] <- .assign_prior.simple(
      prior = prior_effect, parameter = "effect", measure = measure,
      data = data, prior_unit_information_sd = prior_unit_information_sd,
      prior_informed_field = prior_informed_field, prior_informed_subfield = prior_informed_subfield,
      rescale_priors = rescale_priors
    )
  prior_outcome[["tau"]] <- .assign_prior.simple(
    prior = prior_heterogeneity, parameter = "heterogeneity", measure = measure,
    data = data, prior_unit_information_sd = prior_unit_information_sd,
    prior_informed_field = prior_informed_field, prior_informed_subfield = prior_informed_subfield,
    rescale_priors = rescale_priors
  )

  # only for multilevel models
  if (.is_data_multilevel(data)) {
    prior_outcome[["rho"]]   <- .assign_prior.heterogeneity_allocation(prior = prior_heterogeneity_allocation)
    # gamma is an auxiliary parameter for non-centered random-effects parameterization (cluster-level)
    prior_outcome[["gamma"]] <- prior_factor("normal", parameters = list("mean" = 0, "sd" = 1), contrast = "independent")
  }
  # only for glmm models of OR
  if (.data_outcome_type(data) == "bin") {
    prior_outcome[["pi"]]    <- .assign_prior.baserate(prior = prior_baserate)
    # theta is an auxiliary parameter for non-centered random-effects parameterization (for estimate-level)
    # (these are marginalized in the normal model)
    prior_outcome[["theta"]] <- prior_factor("normal", parameters = list("mean" = 0, "sd" = 1), contrast = "independent")
  }
  # only for glmm models of IRR
  if (.data_outcome_type(data) == "pois") {
    prior_outcome[["phi"]] <- .assign_prior.lograte(prior = prior_lograte, data)
    # theta is an auxiliary parameter for non-centered random-effects parameterization (for estimate-level)
    # (these are marginalized in the normal model)
    prior_outcome[["theta"]] <- prior_factor("normal", parameters = list("mean" = 0, "sd" = 1), contrast = "independent")
  }
  # only for selection / PET / PEESE models
  if (!missing(bias_type)) {
    prior_outcome[["bias"]] <- .assign_prior.bias(
      prior = prior_bias, measure = measure, data = data, prior_unit_information_sd = prior_unit_information_sd,
      bias_type = bias_type, steps = steps)
  }


  if (.is_data_mods(data)) {
    prior_mods <- .assign_prior_list.terms(
      prior_list = prior_mods, prior_intercept = prior_outcome[["mu"]], parameter = "mods",
      measure = measure, data = data,  prior_unit_information_sd = prior_unit_information_sd,
      prior_informed_field = prior_informed_field, prior_informed_subfield = prior_informed_subfield,
      rescale_priors = rescale_priors)
    prior_outcome[["mu"]] <- NULL # the prior is forwarded through "mods" to the intercept
  } else {
    prior_mods <- NULL
  }

  if (.is_data_scale(data)) {
    prior_scale <- .assign_prior_list.terms(
      prior_list = prior_scale, prior_intercept = prior_outcome[["tau"]], parameter = "scale",
      measure = measure, data = data,  prior_unit_information_sd = prior_unit_information_sd,
      prior_informed_field = prior_informed_field, prior_informed_subfield = prior_informed_subfield,
      rescale_priors = rescale_priors)
    prior_outcome[["tau"]] <- NULL # the prior is forwarded through "scale" to the log(intercept)
  } else {
    prior_scale <- NULL
  }

  return(list(
    outcome  = prior_outcome,
    mods     = prior_mods,
    scale    = prior_scale
  ))
}


.check_and_list_priors.RoBMA <- function(
    prior_effect, prior_heterogeneity,
    prior_mods, prior_scale,
    prior_heterogeneity_allocation, prior_bias,
    prior_effect_null, prior_heterogeneity_null,
    prior_mods_null, prior_scale_null,
    prior_heterogeneity_allocation_null, prior_bias_null,
    prior_baserate, prior_lograte,
    rescale_priors,
    prior_unit_information_sd,
    prior_informed_field, prior_informed_subfield,
    data, model_type) {

  ### check input
  if (!missing(rescale_priors))
    BayesTools::check_real(rescale_priors, "rescale_priors", lower = 0, allow_bound = FALSE, allow_NA = FALSE)
  if (!missing(prior_unit_information_sd))
    BayesTools::check_real(prior_unit_information_sd, "prior_unit_information_sd", lower = 0, allow_bound = FALSE, allow_NA = FALSE)
  if (!missing(prior_informed_field))
    BayesTools::check_char(prior_informed_field, "prior_informed_field", allow_values = c("medicine"), allow_NA = FALSE)
  if (!missing(prior_informed_subfield))
    BayesTools::check_char(prior_informed_subfield, "prior_informed_subfield", allow_NA = FALSE)
  .check_prior_specification_conflict(prior_unit_information_sd, prior_informed_field)
  if (!missing(model_type))
    BayesTools::check_char(model_type, "model_type", allow_values = c("2w", "6w", "PP", "PSMA"))

  ### set prior distributions
  measure       <- .data_measure(data)
  prior_outcome <- list()

  prior_outcome[["mu"]] <- .assign_prior.simple_mixture(
    prior = prior_effect, prior_null = prior_effect_null, parameter = "effect", measure = measure,
    data = data, prior_unit_information_sd = prior_unit_information_sd,
    prior_informed_field = prior_informed_field, prior_informed_subfield = prior_informed_subfield,
    rescale_priors = rescale_priors
  )
  prior_outcome[["tau"]] <- .assign_prior.simple_mixture(
    prior = prior_heterogeneity, prior_null = prior_heterogeneity_null, parameter = "heterogeneity", measure = measure,
    data = data, prior_unit_information_sd = prior_unit_information_sd,
    prior_informed_field = prior_informed_field, prior_informed_subfield = prior_informed_subfield,
    rescale_priors = rescale_priors
  )
  # only for multilevel models
  if (.is_data_multilevel(data)) {
    prior_outcome[["rho"]]   <- .assign_prior.heterogeneity_allocation_mixture(
      prior = prior_heterogeneity_allocation, prior_null = prior_heterogeneity_allocation_null)
    # gamma is an auxiliary parameter for non-centered random-effects parameterization (cluster-level)
    prior_outcome[["gamma"]] <- prior_factor("normal", parameters = list("mean" = 0, "sd" = 1), contrast = "independent")
  }
  # only for glmm models of OR
  if (.data_outcome_type(data) == "bin") {
    prior_outcome[["pi"]]    <- .assign_prior.baserate(prior = prior_baserate)
    # theta is an auxiliary parameter for non-centered random-effects parameterization (for estimate-level)
    # (these are marginalized in the normal model)
    prior_outcome[["theta"]] <- prior_factor("normal", parameters = list("mean" = 0, "sd" = 1), contrast = "independent")
  }
  # only for glmm models of IRR
  if (.data_outcome_type(data) == "pois") {
    prior_outcome[["phi"]] <- .assign_prior.lograte(prior = prior_lograte, data)
    # theta is an auxiliary parameter for non-centered random-effects parameterization (for estimate-level)
    # (these are marginalized in the normal model)
    prior_outcome[["theta"]] <- prior_factor("normal", parameters = list("mean" = 0, "sd" = 1), contrast = "independent")
  }
  # only for RoBMA
  if (!missing(model_type)) {
    prior_outcome[["bias"]] <- .assign_prior.bias_mixture(
      prior = prior_bias, prior_null = prior_bias_null, measure = measure,
      data = data, prior_unit_information_sd = prior_unit_information_sd, model_type = model_type
    )
  }


  if (.is_data_mods(data)) {
    prior_mods <- .assign_prior_list.terms_mixture(
      prior_list = prior_mods, prior_list_null = prior_mods_null, prior_intercept = prior_outcome[["mu"]], parameter = "mods",
      measure = measure, data = data,  prior_unit_information_sd = prior_unit_information_sd,
      prior_informed_field = prior_informed_field, prior_informed_subfield = prior_informed_subfield,
      rescale_priors = rescale_priors)
    prior_outcome[["mu"]] <- NULL # the prior is forwarded through "mods" to the intercept
  } else {
    prior_mods <- NULL
  }

  if (.is_data_scale(data)) {
    prior_scale <- .assign_prior_list.terms_mixture(
      prior_list = prior_scale, prior_list_null = prior_scale_null, prior_intercept = prior_outcome[["tau"]], parameter = "scale",
      measure = measure, data = data,  prior_unit_information_sd = prior_unit_information_sd,
      prior_informed_field = prior_informed_field, prior_informed_subfield = prior_informed_subfield,
      rescale_priors = rescale_priors)
    prior_outcome[["tau"]] <- NULL # the prior is forwarded through "scale" to the log(intercept)
  } else {
    prior_scale <- NULL
  }

  return(list(
    outcome  = prior_outcome,
    mods     = prior_mods,
    scale    = prior_scale
  ))
}

expect_finite_vector <- function(x, n = NULL, info = NULL) {

  testthat::expect_type(x, "double")
  if (!is.null(n)) {
    testthat::expect_equal(length(x), n, info = info)
  }
  testthat::expect_true(all(is.finite(x)), info = info)
}

expect_finite_table <- function(x, cols = NULL, n = NULL, min_cols = NULL,
                                info = NULL) {

  testthat::expect_s3_class(x, "data.frame")
  if (!is.null(cols)) {
    testthat::expect_true(all(cols %in% names(x)), info = info)
  }
  if (!is.null(n)) {
    testthat::expect_equal(nrow(x), n, info = info)
  }
  if (!is.null(min_cols)) {
    testthat::expect_true(ncol(x) >= min_cols, info = info)
  }
  numeric_cols <- vapply(x, is.numeric, TRUE)
  if (any(numeric_cols)) {
    testthat::expect_true(all(is.finite(as.matrix(x[, numeric_cols, drop = FALSE]))),
                          info = info)
  }
}

expect_positive <- function(x, strict = TRUE, info = NULL) {

  testthat::expect_true(all(is.finite(x)), info = info)
  if (strict) {
    testthat::expect_true(all(x > 0), info = info)
  } else {
    testthat::expect_true(all(x >= 0), info = info)
  }
}

expect_probability <- function(x, open = FALSE, info = NULL) {

  testthat::expect_true(all(is.finite(x)), info = info)
  if (open) {
    testthat::expect_true(all(x > 0 & x < 1), info = info)
  } else {
    testthat::expect_true(all(x >= 0 & x <= 1), info = info)
  }
}

expect_monotone <- function(x, direction = "increasing", strict = FALSE,
                            info = NULL) {

  delta <- diff(x)
  if (direction == "decreasing") {
    delta <- -delta
  }
  if (strict) {
    testthat::expect_true(all(delta > 0), info = info)
  } else {
    testthat::expect_true(all(delta >= 0), info = info)
  }
}

expect_valid_indicator <- function(x, values, info = NULL) {

  testthat::expect_true(all(is.finite(x)), info = info)
  testthat::expect_true(all(x == as.integer(x)), info = info)
  testthat::expect_true(all(x %in% values), info = info)
}

expect_error_cases <- function(cases, envir = parent.frame()) {

  for (case in cases) {
    testthat::expect_error(
      eval(case[["expr"]], envir = envir),
      regexp = case[["regexp"]],
      info   = case[["label"]]
    )
  }

  invisible(NULL)
}

expect_residual_vector <- function(x, n, info = NULL) {

  expect_finite_vector(x, n = n, info = info)
}

expect_residual_table <- function(x, n, check_se = TRUE, info = NULL) {

  expect_finite_table(x, cols = c("resid", "se", "z"), n = n, info = info)
  if (check_se) {
    expect_positive(x[["se"]], info = info)
  }
}

expect_hatvalues_vector <- function(x, n, info = NULL) {

  expect_finite_vector(x, n = n, info = info)
  testthat::expect_true(all(x >= 0 & x <= 1 + sqrt(.Machine$double.eps)),
                        info = info)
}

expect_dfbetas_table <- function(x, n, min_cols = 1, info = NULL) {

  testthat::expect_s3_class(x, "data.frame")
  testthat::expect_equal(nrow(x), n, info = info)
  testthat::expect_true(ncol(x) >= min_cols, info = info)

  numeric_cols <- vapply(x, is.numeric, TRUE)
  if (any(numeric_cols)) {
    values <- as.matrix(x[, numeric_cols, drop = FALSE])
    testthat::expect_true(all(is.finite(values) | is.nan(values)), info = info)
    if (any(is.nan(values))) {
      testthat::expect_true(!is.null(attr(x, "note")), info = info)
      testthat::expect_true(nzchar(attr(x, "note")), info = info)
    }
  }
}

expect_vif_table <- function(x, n_terms = NULL, info = NULL) {

  cols <- c("term", "df", "GVIF", "GVIF^(1/(2*df))")
  expect_finite_table(x, cols = cols, n = n_terms, info = info)
  testthat::expect_true(all(nzchar(x[["term"]])), info = info)
  testthat::expect_true(all(x[["df"]] >= 1), info = info)
  testthat::expect_true(all(x[["GVIF"]] >= 1 - sqrt(.Machine$double.eps)),
                        info = info)
  testthat::expect_true(
    all(x[["GVIF^(1/(2*df))"]] >= 1 - sqrt(.Machine$double.eps)),
    info = info
  )
}

expect_influence_object <- function(x, n, inf_cols, min_dfbs_cols = 1,
                                    info = NULL) {

  testthat::expect_s3_class(x, "infl.brma")
  testthat::expect_true(all(inf_cols %in% names(x[["inf"]])), info = info)
  testthat::expect_equal(nrow(x[["inf"]]), n, info = info)
  inf_values <- as.matrix(x[["inf"]][, inf_cols, drop = FALSE])
  testthat::expect_true(all(is.finite(inf_values) | is.nan(inf_values)),
                        info = info)
  if (any(is.nan(inf_values))) {
    testthat::expect_true(!is.null(attr(x, "note")), info = info)
    testthat::expect_true(nzchar(attr(x, "note")), info = info)
  }
  expect_dfbetas_table(x[["dfbs"]], n = n, min_cols = min_dfbs_cols,
                       info = info)
}

expect_brma_samples_matrix <- function(x, n_col, info = NULL) {

  testthat::expect_s3_class(x, "brma_samples")
  testthat::expect_true(is.matrix(x), info = info)
  testthat::expect_equal(ncol(x), n_col, info = info)
  testthat::expect_true(all(is.finite(unclass(x))), info = info)
}

expect_summary_heterogeneity_structure <- function(heterogeneity, expected_rows,
                                                   name) {

  columns <- c("Mean", "Median", "0.025", "0.975")

  testthat::expect_true(
    inherits(heterogeneity, "summary_heterogeneity.brma"),
    info = paste0("summary_heterogeneity class for '", name, "'")
  )
  testthat::expect_equal(
    sort(rownames(heterogeneity$estimates)),
    sort(expected_rows),
    info = paste0("summary_heterogeneity rows for '", name, "'")
  )

  estimates <- heterogeneity$estimates[expected_rows, columns, drop = FALSE]
  values    <- as.matrix(estimates)

  testthat::expect_true(
    all(is.finite(values)),
    info = paste0("summary_heterogeneity finite estimates for '", name, "'")
  )
  testthat::expect_true(
    all(values >= 0),
    info = paste0("summary_heterogeneity non-negative estimates for '", name, "'")
  )

  i2_rows <- grep("^I2", expected_rows, value = TRUE)
  if (length(i2_rows) > 0) {
    i2_values <- as.matrix(heterogeneity$estimates[i2_rows, columns, drop = FALSE])
    testthat::expect_true(
      all(i2_values >= 0 & i2_values <= 100),
      info = paste0("summary_heterogeneity I2 bounds for '", name, "'")
    )
  }

  if ("rho" %in% expected_rows) {
    rho_values <- as.matrix(heterogeneity$estimates["rho", columns, drop = FALSE])
    testthat::expect_true(
      all(rho_values >= 0 & rho_values <= 1),
      info = paste0("summary_heterogeneity rho bounds for '", name, "'")
    )
  }

  testthat::expect_true(
    all(heterogeneity$estimates["H2", columns] >= 1),
    info = paste0("summary_heterogeneity H2 bounds for '", name, "'")
  )
}

context("Conditional posterior plots")

mock_plot_fit <- function(class = c("BMA.norm", "RoBMA", "brma")) {

  data <- structure(
    list(),
    measure      = "SMD",
    outcome_type = "norm"
  )

  fit <- structure(list(), class = "BayesTools_fit")

  structure(
    list(
      data   = data,
      priors = list(outcome = list()),
      fit    = fit
    ),
    class = class
  )
}

test_that("plot.brma forwards conditional parameter for RoBMA objects", {

  calls <- list()

  testthat::local_mocked_bindings(
    as_mixed_posteriors = function(model, parameters, conditional = NULL,
                                   transform_scaled = FALSE, ...) {
      calls[["as_mixed_posteriors"]] <<- list(
        parameters       = parameters,
        conditional      = conditional,
        transform_scaled = transform_scaled
      )

      samples <- list()
      samples[[parameters]] <- structure(c(0, 1), class = "mixed_posteriors")

      return(samples)
    },
    plot_posterior = function(samples, parameter, ...) {
      calls[["plot_posterior"]] <<- list(
        samples   = samples,
        parameter = parameter
      )

      return(structure(list(), class = "mock_plot"))
    },
    .package = "BayesTools"
  )

  out <- plot(
    mock_plot_fit(),
    parameter   = "mu",
    conditional = TRUE,
    plot_type   = "ggplot"
  )

  expect_s3_class(out, "mock_plot")
  expect_equal(calls[["as_mixed_posteriors"]][["parameters"]], "mu")
  expect_equal(calls[["as_mixed_posteriors"]][["conditional"]], "mu")
  expect_true(calls[["as_mixed_posteriors"]][["transform_scaled"]])
  expect_equal(calls[["plot_posterior"]][["parameter"]], "mu")
})

test_that("plot.brma rejects conditional parameter for non-RoBMA objects", {

  expect_error(
    plot(
      mock_plot_fit(class = c("brma.norm", "brma")),
      parameter   = "mu",
      conditional = TRUE
    ),
    "RoBMA objects"
  )
})

context("nobs and coef methods")

source(testthat::test_path("common-functions.R"))
source(testthat::test_path("helper-contracts.R"))

skip_if_no_fits()
skip_if_not_installed("metafor")

fit_names <- list_fits()
fits      <- lazy_fits(fit_names, validate = FALSE)
info      <- lazy_infos(fit_names, validate = FALSE)

test_that("nobs matches outcome data and metafor", {

  metafor_names <- list_fits(has_metafor = TRUE)
  skip_if_missing_fits(metafor_names)

  for (name in names(fits)) {
    fit <- fits[[name]]
    expect_true(is.numeric(nobs(fit)) && nobs(fit) > 0,
                info = paste0("nobs is positive for '", name, "'"))
    expect_equal(nobs(fit), nrow(fit$data$outcome),
                 info = paste0("nobs matches outcome rows for '", name, "'"))
  }

  for (name in metafor_names) {
    expect_equal(
      nobs(fits[[name]]),
      as.integer(nobs(info[[name]][["metafor"]])),
      info = paste0("nobs matches metafor for '", name, "'")
    )
  }
})

test_that("coef returns model coefficients across cached fits", {

  for (name in names(fits)) {
    fit <- fits[[name]]
    cf  <- coef(fit)
    expect_true(is.numeric(cf) && length(cf) > 0,
                info = paste0("coef is non-empty numeric for '", name, "'"))
    expect_true(!is.null(names(cf)),
                info = paste0("coef is named for '", name, "'"))
    expect_identical(cf, fit$coefficients,
                     info = paste0("coef matches $coefficients for '", name, "'"))
  }
})

context("Prior input handling for BMA product-space constructors")

skip_on_cran()

.get_is_null <- function(prior) {
  components <- attr(prior, "components")
  if (is.null(components)) return(NULL)
  return(components == "null")
}

.get_prior_weights <- function(prior) {
  return(vapply(prior, function(x) x[["prior_weights"]], numeric(1)))
}

bma_test_data <- data.frame(
  effect    = c(0.10, 0.25, 0.15, 0.30, 0.05),
  std_err   = sqrt(c(0.04, 0.06, 0.05, 0.08, 0.03)),
  n         = c(50L, 75L, 60L, 40L, 90L),
  mod_cont  = c(1.5, 2.3, 1.8, 3.1, 0.9),
  scale_var = c(0.5, 1.0, 0.8, 1.2, 0.6),
  stringsAsFactors = FALSE
)

bma_glmm_data <- data.frame(
  ai = c(4L, 7L, 3L, 9L, 6L),
  bi = c(46L, 68L, 57L, 31L, 84L),
  ci = c(8L, 6L, 5L, 7L, 9L),
  di = c(42L, 69L, 55L, 33L, 81L),
  stringsAsFactors = FALSE
)


test_that("BMA alias and staged normal constructor return product-space objects", {

  object_alias <- BMA(
    yi = effect, sei = std_err, data = bma_test_data,
    measure = "SMD", only_priors = TRUE
  )
  object_norm <- BMA.norm(
    yi = effect, sei = std_err, data = bma_test_data,
    measure = "SMD", only_priors = TRUE
  )

  expect_s3_class(object_alias, "BMA.norm")
  expect_s3_class(object_alias, "RoBMA")
  expect_s3_class(object_alias, "brma")
  expect_identical(class(object_alias), class(object_norm))
  expect_null(object_alias[["priors"]][["outcome"]][["bias"]])
  expect_true(BayesTools::is.prior.mixture(object_alias[["priors"]][["outcome"]][["mu"]]))
  expect_true(BayesTools::is.prior.mixture(object_alias[["priors"]][["outcome"]][["tau"]]))

  object_data <- BMA(
    yi = effect, sei = std_err, data = bma_test_data,
    measure = "SMD", only_data = TRUE
  )

  expect_s3_class(object_data, "BMA.norm")
  expect_true("data" %in% names(object_data))
  expect_null(object_data[["priors"]])
  expect_null(object_data[["fit"]])
})


test_that("BMA.glmm staged constructor returns product-space objects without bias priors", {

  object_data <- BMA.glmm(
    ai = ai, bi = bi, ci = ci, di = di,
    data = bma_glmm_data, measure = "OR",
    only_data = TRUE
  )
  object_priors <- BMA.glmm(
    ai = ai, bi = bi, ci = ci, di = di,
    data = bma_glmm_data, measure = "OR",
    only_priors = TRUE
  )

  expect_s3_class(object_data, "BMA.glmm")
  expect_s3_class(object_data, "RoBMA")
  expect_s3_class(object_data, "brma.glmm")
  expect_s3_class(object_priors, "BMA.glmm")
  expect_null(object_priors[["priors"]][["outcome"]][["bias"]])
  expect_true(BayesTools::is.prior.mixture(object_priors[["priors"]][["outcome"]][["mu"]]))
  expect_true(BayesTools::is.prior.mixture(object_priors[["priors"]][["outcome"]][["tau"]]))
})


test_that("BMA null prior NULL and FALSE omit null components", {

  result_null <- BMA(
    yi = effect, sei = std_err, data = bma_test_data,
    prior_effect_null = NULL,
    prior_heterogeneity_null = NULL,
    measure = "SMD", only_priors = TRUE
  )[["priors"]]
  result_false <- BMA(
    yi = effect, sei = std_err, data = bma_test_data,
    prior_effect_null = FALSE,
    prior_heterogeneity_null = FALSE,
    measure = "SMD", only_priors = TRUE
  )[["priors"]]

  expect_equal(length(result_null[["outcome"]][["mu"]]), 1)
  expect_equal(length(result_null[["outcome"]][["tau"]]), 1)
  expect_equal(sum(.get_is_null(result_null[["outcome"]][["mu"]])), 0)
  expect_equal(sum(.get_is_null(result_null[["outcome"]][["tau"]])), 0)

  expect_equal(length(result_false[["outcome"]][["mu"]]), 1)
  expect_equal(length(result_false[["outcome"]][["tau"]]), 1)
  expect_equal(sum(.get_is_null(result_false[["outcome"]][["mu"]])), 0)
  expect_equal(sum(.get_is_null(result_false[["outcome"]][["tau"]])), 0)
})


test_that("BMA keeps custom prior weights in mixture constructors", {

  effect_alt <- list(
    BayesTools::prior(
      "normal", parameters = list(mean = 0, sd = 0.25),
      prior_weights = 3
    ),
    BayesTools::prior(
      "normal", parameters = list(mean = 0, sd = 0.75),
      prior_weights = 2
    )
  )
  effect_null <- BayesTools::prior(
    "spike", parameters = list(location = 0),
    prior_weights = 5
  )

  result <- BMA(
    yi = effect, sei = std_err, data = bma_test_data,
    prior_effect = effect_alt,
    prior_effect_null = effect_null,
    measure = "SMD", only_priors = TRUE
  )[["priors"]][["outcome"]][["mu"]]

  expect_equal(length(result), 3)
  expect_equal(.get_prior_weights(result), c(5, 3, 2))
  expect_equal(.get_is_null(result), c(TRUE, FALSE, FALSE))
})


test_that("BMA formula null prior settings apply to moderator and scale terms", {

  result <- suppressWarnings(BMA(
    yi = effect, sei = std_err,
    mods = ~ mod_cont, scale = ~ scale_var,
    data = bma_test_data, measure = "SMD",
    prior_mods_null  = list(mod_cont = NULL),
    prior_scale_null = list(scale_var = FALSE),
    only_priors = TRUE
  )[["priors"]])

  expect_true(BayesTools::is.prior.mixture(result[["mods"]][["mod_cont"]]))
  expect_true(BayesTools::is.prior.mixture(result[["scale"]][["scale_var"]]))
  expect_equal(sum(.get_is_null(result[["mods"]][["mod_cont"]])), 0)
  expect_equal(sum(.get_is_null(result[["scale"]][["scale_var"]])), 0)
})


test_that("BMA constructors reject bias and model arguments passed through dots", {

  expect_error(
    BMA(
      yi = effect, sei = std_err, data = bma_test_data,
      measure = "SMD", model_type = "PSMA", only_priors = TRUE
    ),
    "Unused argument.*model_type"
  )

  expect_error(
    BMA(
      yi = effect, sei = std_err, data = bma_test_data,
      measure = "SMD", prior_bias = BayesTools::prior_none(),
      only_priors = TRUE
    ),
    "Unused argument.*prior_bias"
  )

  expect_error(
    BMA.glmm(
      ai = ai, bi = bi, ci = ci, di = di,
      data = bma_glmm_data, measure = "OR",
      prior_bias = BayesTools::prior_none(),
      only_priors = TRUE
    ),
    "Unused argument.*prior_bias"
  )
})

context("Residuals")

source(testthat::test_path("common-functions.R"))
source(testthat::test_path("helper-contracts.R"))
source(testthat::test_path("helper-test-matrix.R"))
source(testthat::test_path("helper-metafor.R"))

REFERENCE_DIR <<- testthat::test_path("..", "results", "residuals")

skip_if_no_fits()
skip_if_not_installed("metafor")

fit_names <- list_fits()
fits      <- lazy_fits(fit_names, validate = FALSE)
info      <- lazy_infos(fit_names, validate = FALSE)

for_each_case(residual_metafor_cases(), function(case) {
  test_that_case("Residuals match metafor", case, {
    expect_residuals_match_metafor(case)
  })
})

test_that("Residuals for equivalent interaction parameterizations match", {

  model_names <- c("bcg_meta-regression3", "bcg_meta-regression3b")
  skip_if_missing_fits(model_names)

  fit_brma1 <- fits[["bcg_meta-regression3"]]
  fit_brma2 <- fits[["bcg_meta-regression3b"]]

  expect_equal(residuals(fit_brma1), residuals(fit_brma2), tolerance = 0.10,
               info = "outcome residuals match")
  expect_equal(
    residuals(fit_brma1, type = "pearson"),
    residuals(fit_brma2, type = "pearson"),
    tolerance = 0.10,
    info      = "pearson residuals match"
  )
  expect_equal(rstandard(fit_brma1), rstandard(fit_brma2), tolerance = 0.10,
               info = "rstandard residuals match")
})

test_that("Extended rstudent residuals for interaction parameterizations match", {

  skip_if_not_full_diagnostics(
    "Interaction parameterization LOO-PIT checks are redundant with core residual coverage."
  )
  model_names <- c("bcg_meta-regression4", "bcg_meta-regression4b")
  skip_if_missing_fits(model_names)

  brma_rstudent1 <- suppressWarnings(rstudent(fits[["bcg_meta-regression4"]]))
  brma_rstudent2 <- suppressWarnings(rstudent(fits[["bcg_meta-regression4b"]]))

  expect_true(cor(brma_rstudent1[, "z"], brma_rstudent2[, "z"]) > 0.90,
              info = "rstudent residuals match")
})

test_that("Extended rstudent residuals align with metafor oracles", {

  skip_if_not_full_diagnostics(
    "Default residual coverage keeps one representative LOO-PIT check per family."
  )

  extended <- list(
    list(
      name      = "bcg_meta-regression4",
      resid_cor = 0.90,
      se_cor    = NULL,
      z_cor     = 0.90
    ),
    list(
      name      = "konstantopoulos2011_3lvl2",
      resid_cor = 0.60,
      se_cor    = 0.80,
      z_cor     = 0.60
    ),
    list(
      name      = "dat.lehmann2018-PETreg",
      equal     = TRUE
    )
  )
  skip_if_missing_fits(vapply(extended, `[[`, "", "name"))

  for (case in extended) {
    name <- case[["name"]]
    if (isTRUE(case[["equal"]])) {
      .expect_residual_table_equal(
        name,
        suppressWarnings(rstudent(fits[[name]])),
        rstudent(info[[name]][["metafor"]]),
        tolerance   = 0.05,
        z_tolerance = 0.10,
        label       = "rstudent"
      )
    } else {
      .expect_residual_rstudent_rank(
        name,
        suppressWarnings(rstudent(fits[[name]])),
        rstudent(info[[name]][["metafor"]]),
        resid_cor = case[["resid_cor"]],
        se_cor    = case[["se_cor"]],
        z_cor     = case[["z_cor"]]
      )
    }
  }
})

test_that("Extended rstudent residuals without direct metafor oracle are calibrated", {

  skip_if_not_full_diagnostics(
    "These checks have weaker or family-specific oracles and duplicate default branches."
  )

  scale_name <- "bangertdrowns2004_location-scale"
  sel_name   <- "dat.lehmann2018-3PSMreg"
  glmm_name  <- "bcg_glmm_reg"
  skip_if_missing_fits(c(scale_name, sel_name, glmm_name))

  scale_rstudent <- rstudent(fits[[scale_name]])
  scale_standard <- rstandard(info[[scale_name]][["metafor"]])
  expect_equal(mean(scale_rstudent$z), 0, 0.05)
  expect_equal(sd(scale_rstudent$z), 1, 0.05)
  expect_true(cor(scale_rstudent$z, scale_standard$z, method = "spearman") > 0.8)

  sel_resid    <- residuals(fits[[sel_name]])
  sel_rstudent <- suppressWarnings(rstudent(fits[[sel_name]]))
  expect_true(sel_rstudent$z[4] > 4)
  expect_true(all(abs(sel_rstudent$z[-4]) < 4))
  expect_true(cor(sel_resid, sel_rstudent$z) > 0.9)

  glmm_resid    <- residuals(fits[[glmm_name]])
  glmm_rstudent <- suppressWarnings(rstudent(fits[[glmm_name]], type = "estimate"))
  expect_true(cor(glmm_resid, glmm_rstudent$z) > 0.9)
})

test_that("Residuals for BMA.norm fits are internally consistent", {

  name <- "dat.lehmann2018_BMA.norm"
  skip_if_missing_fits(name)

  fit_brma <- fits[[name]]
  n        <- nobs(fit_brma)

  resid_marginal <- residuals(fit_brma, conditioning_depth = "marginal")
  resid_estimate <- residuals(fit_brma, conditioning_depth = "estimate")

  expect_residual_vector(resid_marginal, n)
  expect_residual_vector(resid_estimate, n)
  expect_error(residuals(fit_brma, conditioning_depth = "cluster"),
               "is only available for multilevel models")
  expect_true(cor(resid_marginal, resid_estimate) > 0.9)
  expect_true(sd(resid_marginal) > sd(resid_estimate))

  pearson_marginal <- residuals(
    fit_brma,
    type               = "pearson",
    conditioning_depth = "marginal"
  )
  pearson_estimate <- residuals(
    fit_brma,
    type               = "pearson",
    conditioning_depth = "estimate"
  )

  expect_residual_vector(pearson_marginal, n)
  expect_residual_vector(pearson_estimate, n)
  expect_true(cor(pearson_marginal, pearson_estimate) > 0.9)

  brma_rstandard <- rstandard(fit_brma)
  expect_residual_table(brma_rstandard, n)
  expect_equal(residuals(fit_brma, type = "rstandard"), brma_rstandard[["z"]],
               tolerance = 1e-12)

  brma_rstudent <- suppressWarnings(rstudent(fit_brma))
  expect_residual_table(brma_rstudent, n)
  expect_equal(suppressWarnings(residuals(fit_brma, type = "rstudent")),
               brma_rstudent[["z"]], tolerance = 1e-12)
  expect_error(
    residuals(fit_brma, type = "rstudent", conditioning_depth = "marginal"),
    "do not set 'conditioning_depth'"
  )
})

test_that("Residuals for BMA.glmm fits are internally consistent", {

  model_names <- c("bcg_BMA.glmm", "nielweise2008_BMA.glmm", "bcg_BMA.glmm_3lvl_location_scale")
  skip_if_missing_fits(model_names)

  for (name in c("bcg_BMA.glmm", "nielweise2008_BMA.glmm")) {
    fit_brma <- fits[[name]]
    n        <- nobs(fit_brma)

    expect_residual_vector(residuals(fit_brma), n)
    expect_error(residuals(fit_brma, type = "pearson"), "normal outcome models")
    expect_error(rstandard(fit_brma), "normal outcome models")
    expect_residual_table(suppressWarnings(rstudent(fit_brma)), n)
  }

  fit_3lvl <- fits[["bcg_BMA.glmm_3lvl_location_scale"]]
  n_3lvl   <- nrow(fit_3lvl[["data"]][["outcome"]])

  expect_null(fit_3lvl[["marglik"]])
  expect_residual_vector(residuals(fit_3lvl, conditioning_depth = "cluster"),
                         n_3lvl)
  expect_error(rstudent(fit_3lvl, unit = "cluster"), "Cluster-unit rstudent residuals")
})

test_that("Residuals for RoBMA fits are internally consistent", {

  model_names <- c("dat.lehmann2018_RoBMA", "dat.lehmann2018_RoBMA_3lvl_mods_scale")
  skip_if_missing_fits(model_names)

  fit_brma <- fits[["dat.lehmann2018_RoBMA"]]
  n        <- nobs(fit_brma)

  expect_residual_vector(residuals(fit_brma), n)
  expect_error(residuals(fit_brma, type = "pearson"), "selection models")
  expect_error(rstandard(fit_brma), "selection models")

  brma_rstudent <- suppressWarnings(rstudent(fit_brma))
  expect_residual_table(brma_rstudent, n)
  expect_equal(suppressWarnings(residuals(fit_brma, type = "rstudent")),
               brma_rstudent[["z"]], tolerance = 1e-12)

  fit_3lvl <- fits[["dat.lehmann2018_RoBMA_3lvl_mods_scale"]]
  n_3lvl   <- nrow(fit_3lvl[["data"]][["outcome"]])

  expect_null(fit_3lvl[["marglik"]])
  expect_residual_vector(residuals(fit_3lvl, conditioning_depth = "cluster"),
                         n_3lvl)
})

context("Cached fit catalog")

source(testthat::test_path("common-functions.R"))

.collect_save_fit_calls <- function(expr, source_file) {

  out <- data.frame(
    name        = character(),
    source_file = character(),
    stringsAsFactors = FALSE
  )

  if (is.call(expr) && identical(.source_hash_call_name(expr), "save_fit")) {
    fit_name <- expr[[2]]
    if (is.character(fit_name) && length(fit_name) == 1L) {
      out <- rbind(
        out,
        data.frame(
          name        = fit_name,
          source_file = source_file,
          stringsAsFactors = FALSE
        )
      )
    }
  }

  if (is.recursive(expr)) {
    children <- lapply(as.list(expr), .collect_save_fit_calls, source_file = source_file)
    children <- Filter(function(x) nrow(x) > 0L, children)
    if (length(children) > 0L) {
      out <- rbind(out, do.call(rbind, children))
    }
  }

  return(out)
}

.test_01_saved_fits <- function() {

  test_files <- list.files(
    testthat::test_path(),
    pattern    = "^test-01-.*\\.R$",
    full.names = FALSE
  )
  saved_fits <- lapply(test_files, function(source_file) {

    parsed <- parse(testthat::test_path(source_file))
    do.call(rbind, lapply(parsed, .collect_save_fit_calls, source_file = source_file))
  })
  saved_fits <- Filter(function(x) nrow(x) > 0L, saved_fits)

  if (length(saved_fits) == 0L) {
    return(data.frame(
      name        = character(),
      source_file = character(),
      stringsAsFactors = FALSE
    ))
  }

  return(do.call(rbind, saved_fits))
}

.fake_fit_for_catalog_entry <- function(name) {

  entry <- fit_catalog_entry(name)
  fit   <- list()

  if (isTRUE(entry[["has_loo"]])) {
    fit[["loo"]] <- list(estimate = "loo")
  }
  if (isTRUE(entry[["has_waic"]])) {
    fit[["waic"]] <- list(estimate = "waic")
  }
  if (isTRUE(entry[["has_marglik"]])) {
    fit[["marglik"]] <- list(estimate = "marglik")
  }

  class(fit) <- entry[["class"]]

  return(fit)
}

.fake_info_for_catalog_entry <- function(name) {

  entry <- fit_catalog_entry(name)

  if (isTRUE(entry[["has_metafor"]])) {
    return(list(metafor = list(reference = TRUE)))
  }

  return(list())
}

.valid_metadata_for_catalog_entry <- function(name) {

  entry <- fit_catalog_entry(name)

  return(list(
    version            = FIT_CACHE_VERSION,
    name               = name,
    saved_at           = "synthetic",
    fit_class          = entry[["class"]],
    source_file        = entry[["source_file"]],
    source_file_md5    = source_file_md5(entry[["source_file"]]),
    package_source_md5 = package_source_md5(),
    has_loo            = entry[["has_loo"]],
    has_waic           = entry[["has_waic"]],
    has_marglik        = entry[["has_marglik"]],
    has_metafor_info   = entry[["has_metafor"]]
  ))
}


test_that("cache source hash tracks only cache-affecting fitting sources", {

  source_files <- .fit_cache_source_files(relative = TRUE)
  skip_if_not(
    length(source_files) > 0L,
    "Complete package R/src source tree is not available under source-build checks."
  )

  expect_equal(anyDuplicated(source_files), 0L)
  expect_true(all(c(
    "R/fit.R",
    "R/input-data.R",
    "R/input-object.R",
    "R/input-priors.R",
    "R/selection-mapping.R",
    "R/marglik.R",
    "R/loo.R",
    "src/RoBMA.cc",
    "src/distributions/DWN.cc",
    "src/r-glmm.cc",
    "src/r-selnorm.cc"
  ) %in% source_files))
  expect_false(any(c(
    "DESCRIPTION",
    "NAMESPACE",
    "R/as_draws.R",
    "R/effect-transformations.R",
    "R/plot.R",
    "R/predict.R",
    "R/summary.R",
    "src/r-regplot.cc"
  ) %in% source_files))
})

test_that("cache source hash requires a complete source root", {

  package_root <- tempfile("robma-incomplete-source-")
  dir.create(file.path(package_root, "R"), recursive = TRUE)
  writeLines("Package: RoBMA", file.path(package_root, "DESCRIPTION"), useBytes = TRUE)
  writeLines("fit <- function() NULL", file.path(package_root, "R", "fit.R"), useBytes = TRUE)
  on.exit(unlink(package_root, recursive = TRUE), add = TRUE)

  expect_equal(
    .fit_cache_source_files(package_root = package_root, relative = TRUE),
    character()
  )
})

test_that("R source hashes ignore comments and formatting", {

  file_a <- tempfile("robma-source-a-", fileext = ".R")
  file_b <- tempfile("robma-source-b-", fileext = ".R")
  file_c <- tempfile("robma-source-c-", fileext = ".R")
  on.exit(unlink(c(file_a, file_b, file_c)), add = TRUE)

  writeLines(c(
    "# comment before code",
    "f <- function(x) {",
    "  x + 1",
    "}"
  ), file_a, useBytes = TRUE)
  writeLines(c(
    "f<-function(x){x+1} # trailing comment"
  ), file_b, useBytes = TRUE)
  writeLines(c(
    "f <- function(x) {",
    "  x + 2",
    "}"
  ), file_c, useBytes = TRUE)

  expect_identical(
    .fit_cache_source_file_md5(file_a),
    .fit_cache_source_file_md5(file_b)
  )
  expect_false(identical(
    .fit_cache_source_file_md5(file_a),
    .fit_cache_source_file_md5(file_c)
  ))
})


test_that("fit catalog is internally consistent", {

  catalog <- fit_catalog()

  expect_equal(anyDuplicated(catalog[["name"]]), 0L,
               info = "fit catalog names must be unique")
  expect_true(all(file.exists(testthat::test_path(catalog[["source_file"]]))),
              info = "all catalog source files must exist")
  expect_true(all(catalog[["has_loo"]]),
              info = "all cached test fits include pre-computed LOO")
  expect_true(all(!catalog[["has_waic"]]),
              info = "WAIC is not pre-computed for cached test fits")
  expect_true(all(!catalog[catalog[["class"]] %in% c("BMA.norm", "BMA.glmm", "RoBMA"), "has_marglik"]),
              info = "product-space model averaging fits must not cache marginal likelihoods")
  expect_true(all(catalog[!catalog[["class"]] %in% c("BMA.norm", "BMA.glmm", "RoBMA"), "has_marglik"]),
              info = "single-model cached fits include marginal likelihoods")
  expect_true(all(catalog[catalog[["class"]] %in% c("brma.glmm", "BMA.glmm"), "family"] == "glmm"),
              info = "GLMM classes must be tagged as glmm family")
  expect_true(all(catalog[!catalog[["class"]] %in% c("brma.glmm", "BMA.glmm"), "family"] == "norm"),
              info = "normal-outcome classes must be tagged as norm family")
  expect_equal(length(catalog[["features"]]), nrow(catalog),
               info = "each catalog row must have feature tags")
  expect_true(all(vapply(catalog[["features"]], length, integer(1)) > 0L),
              info = "feature tags cannot be empty")
})

test_that("fit catalog mirrors test-01 save_fit calls", {

  catalog    <- fit_catalog()
  saved_fits <- .test_01_saved_fits()

  expect_equal(anyDuplicated(saved_fits[["name"]]), 0L,
               info = "each cached fit name is saved once")
  expect_equal(setdiff(saved_fits[["name"]], catalog[["name"]]), character(),
               info = "all save_fit() names must be catalogued")
  expect_equal(setdiff(catalog[["name"]], saved_fits[["name"]]), character(),
               info = "all catalogued fits must be saved by a test-01 file")

  saved_source <- saved_fits[["source_file"]][match(catalog[["name"]], saved_fits[["name"]])]
  expect_equal(saved_source, catalog[["source_file"]],
               info = "catalog source_file must match the test-01 file calling save_fit()")
})

test_that("cache validation rejects corrupted synthetic metadata", {

  name     <- "bcg_meta-analysis"
  fit      <- .fake_fit_for_catalog_entry(name)
  info     <- .fake_info_for_catalog_entry(name)
  metadata <- .valid_metadata_for_catalog_entry(name)

  expect_equal(
    validate_cached_fit(
      name        = name,
      fit         = fit,
      info        = info,
      metadata    = metadata,
      check_files = FALSE,
      deep        = TRUE
    ),
    character(),
    info = "valid synthetic metadata passes without cache files"
  )

  stale_source <- metadata
  stale_source[["source_file_md5"]] <- "not-current"
  expect_true(
    "source file hash changed" %in% validate_cached_fit(
      name        = name,
      fit         = fit,
      info        = info,
      metadata    = stale_source,
      check_files = FALSE,
      deep        = TRUE
    ),
    info = "test source hash is enforced"
  )

  stale_package <- metadata
  stale_package[["package_source_md5"]] <- "not-current"
  if (!is.na(package_source_md5())) {
    expect_true(
      "cache source hash changed" %in% validate_cached_fit(
        name        = name,
        fit         = fit,
        info        = info,
        metadata    = stale_package,
        check_files = FALSE,
        deep        = TRUE
      ),
      info = "cache source hash is enforced"
    )
  }

  missing_marglik <- metadata
  missing_marglik[["has_marglik"]] <- FALSE
  expect_true(
    "metadata reports missing marginal likelihood" %in% validate_cached_fit(
      name        = name,
      fit         = fit,
      info        = info,
      metadata    = missing_marglik,
      check_files = FALSE,
      deep        = TRUE
    ),
    info = "required marginal likelihood metadata is enforced"
  )

  missing_metafor <- metadata
  missing_metafor[["has_metafor_info"]] <- FALSE
  expect_true(
    "metadata reports missing metafor reference" %in% validate_cached_fit(
      name        = name,
      fit         = fit,
      info        = info,
      metadata    = missing_metafor,
      check_files = FALSE,
      deep        = TRUE
    ),
    info = "required metafor metadata is enforced"
  )

  product_name     <- "dat.lehmann2018_RoBMA"
  product_fit      <- .fake_fit_for_catalog_entry(product_name)
  product_info     <- .fake_info_for_catalog_entry(product_name)
  product_metadata <- .valid_metadata_for_catalog_entry(product_name)
  product_metadata[["has_marglik"]] <- TRUE
  product_fit[["marglik"]] <- list(estimate = "unexpected")

  problems <- validate_cached_fit(
    name        = product_name,
    fit         = product_fit,
    info        = product_info,
    metadata    = product_metadata,
    check_files = FALSE,
    deep        = TRUE
  )
  expect_true(any(grepl("unexpected marginal likelihood", problems)),
              info = "product-space marginal likelihoods are forbidden")
})

test_that("cached fits are catalogued and valid", {

  skip_on_cran()

  cached_names <- list_fits(validate = FALSE)
  if (length(cached_names) == 0) {
    skip("No cached fits available for catalog validation.")
  }

  catalog_names <- fit_catalog()[["name"]]
  uncatalogued  <- setdiff(cached_names, catalog_names)

  expect_equal(uncatalogued, character(),
               info = "all cached fits must be listed in fit_catalog()")

  problems <- lapply(cached_names, validate_cached_fit, deep = TRUE)
  names(problems) <- cached_names
  problems <- problems[lengths(problems) > 0]
  problem_text <- if (length(problems) == 0) {
    ""
  } else {
    paste(
      paste0(names(problems), ": ", vapply(problems, paste, "", collapse = "; ")),
      collapse = "\n"
    )
  }

  expect_true(
    length(problems) == 0,
    info = paste(
      "cached fits must match catalog requirements; rerun test-01 after cache-affecting fitting changes",
      problem_text,
      sep = "\n"
    )
  )
})

test_tier <- function() {

  tier <- "core"
  if (is_true_env("ROBMA_TEST_EXTENDED")) {
    tier <- c(tier, "extended")
  }

  return(tier)
}

visual_test_tier <- function() {

  tier <- test_tier()
  if (is_true_env("ROBMA_TEST_FULL_VISUALS")) {
    tier <- c(tier, "visual-gallery")
  }

  return(tier)
}

case_value <- function(case, name, default = NULL) {

  if (!name %in% names(case)) {
    return(default)
  }

  value <- case[[name]]
  if (is.list(value)) {
    value <- value[[1]]
  }
  if (length(value) == 0 || (length(value) == 1 && is.na(value))) {
    return(default)
  }

  return(value)
}

case_name <- function(case) {

  return(case_value(case, "name"))
}

case_label <- function(case) {

  label <- case_value(case, "label", default = case_name(case))
  return(label)
}

case_has_check <- function(case, check) {

  return(check %in% case_value(case, "checks", character()))
}

filter_cases <- function(cases, tier = test_tier()) {

  if (!"tier" %in% names(cases)) {
    return(cases)
  }

  return(cases[cases[["tier"]] %in% tier, , drop = FALSE])
}

for_each_case <- function(cases, callback, tier = test_tier()) {

  cases <- filter_cases(cases, tier = tier)
  if (nrow(cases) == 0L) {
    return(invisible(NULL))
  }

  for (i in seq_len(nrow(cases))) {
    local({
      case <- cases[i, , drop = FALSE]
      callback(case)
    })
  }

  return(invisible(NULL))
}

test_that_case <- function(description, case, code) {

  testthat::test_that(paste0(description, " [", case_label(case), "]"), {
    code
  })
}

hatvalue_cases <- function() {

  data.frame(
    name = c(
      "bcg_meta-analysis",
      "bcg_meta-regression",
      "bcg_meta-regression4",
      "konstantopoulos2011_3lvl",
      "konstantopoulos2011_3lvl2",
      "dat.lehmann2018-PET",
      "dat.lehmann2018-PETreg",
      "dat.lehmann2018-PET_neg",
      "dat.lehmann2018-PEESE",
      "dat.lehmann2018-PEESEreg",
      "dat.lehmann2018-PEESE_neg"
    ),
    label = c(
      "normal simple",
      "normal meta-regression",
      "normal interaction",
      "normal multilevel",
      "normal multilevel meta-regression",
      "PET",
      "PET meta-regression",
      "PET negative",
      "PEESE",
      "PEESE meta-regression",
      "PEESE negative"
    ),
    tolerance = rep(0.05, 11),
    tier = c("core", "core", "core", "core", "extended", "core", "core", "extended",
             "core", "core", "extended"),
    stringsAsFactors = FALSE
  )
}

vif_cases <- function() {

  out <- data.frame(
    name = c(
      "bcg_meta-regression",
      "bcg_meta-regression2",
      "bcg_meta-regression3",
      "bcg_meta-regression4",
      "bangertdrowns2004_location-scale",
      "konstantopoulos2011_3lvl2",
      "dat.lehmann2018-PETreg",
      "dat.lehmann2018-3PSMreg",
      "bcg_glmm_reg"
    ),
    label = c(
      "normal continuous",
      "normal factor",
      "normal interaction year",
      "normal interaction factor",
      "normal location-scale",
      "normal multilevel meta-regression",
      "PET meta-regression",
      "selection meta-regression",
      "GLMM meta-regression"
    ),
    tolerance = rep(0.10, 9),
    tier = c("core", "extended", "extended", "core", "core", "extended", "core", "core", "core"),
    stringsAsFactors = FALSE
  )
  out[["btt"]] <- I(list(NULL, NULL, NULL, NULL, NULL, NULL, list(3, 2), NULL, NULL))

  return(out)
}

dfbetas_metafor_cases <- function() {

  out <- data.frame(
    name = c(
      "bcg_meta-analysis",
      "bcg_meta-regression",
      "bcg_meta-regression4",
      "konstantopoulos2011_3lvl",
      "konstantopoulos2011_3lvl2",
      "dat.lehmann2018-PET",
      "dat.lehmann2018-PETreg",
      "dat.lehmann2018-PET_neg",
      "dat.lehmann2018-PEESE",
      "dat.lehmann2018-PEESEreg",
      "dat.lehmann2018-PEESE_neg"
    ),
    label = c(
      "normal simple",
      "normal meta-regression",
      "normal interaction",
      "normal multilevel",
      "normal multilevel meta-regression",
      "PET",
      "PET meta-regression",
      "PET negative",
      "PEESE",
      "PEESE meta-regression",
      "PEESE negative"
    ),
    tolerance = rep(0.10, 11),
    oracle = c("equal", "equal", "structure", "equal", "equal", "mapped", "mapped", "mapped",
               "mapped", "mapped", "mapped"),
    tier = c("core", "core", "core", "core", "extended", "core", "core", "extended",
             "core", "core", "extended"),
    stringsAsFactors = FALSE
  )
  out[["skip_rows"]] <- I(list(integer(), c(4, 6, 13), integer(), integer(),
                               integer(), integer(), integer(), integer(),
                               integer(), integer(), integer()))
  out[["metafor_cols"]] <- I(list(NULL, NULL, NULL, NULL, NULL, 1, c(1, 3), 1,
                                  1, c(1, 3), 1))
  out[["brma_cols"]]    <- I(list(NULL, NULL, NULL, NULL, NULL, 1, c(1, 2), 1,
                                  1, c(1, 2), 1))

  return(out)
}

summary_heterogeneity_cases <- function() {

  data.frame(
    name = c(
      "bcg_meta-analysis",
      "konstantopoulos2011_3lvl",
      "bcg_meta-regression",
      "bangertdrowns2004_location-scale",
      "dat.lehmann2018-PET",
      "dat.lehmann2018-PET_neg",
      "dat.lehmann2018-PEESE",
      "dat.lehmann2018-PEESEreg",
      "dat.lehmann2018-PEESE_neg",
      "dat.lehmann2018-3PSM",
      "dat.lehmann2018-3PSM_neg",
      "nielweise2008_glmm"
    ),
    label = c(
      "normal simple",
      "normal multilevel",
      "normal meta-regression",
      "normal location-scale",
      "PET",
      "PET negative",
      "PEESE",
      "PEESE meta-regression",
      "PEESE negative",
      "selection",
      "selection negative",
      "GLMM"
    ),
    kind = c("standard", "multilevel", "standard", "scale", "standard",
             "standard", "standard", "standard", "standard", "selection",
             "selection", "standard"),
    tolerance = c(0.05, 0.05, 0.10, 0.05, 0.05, 0.05, 0.05, 0.05,
                  0.05, 0.05, 0.05, 0.10),
    h2_tolerance = c(0.05, NA, 0.20, 0.20, 0.05, 0.05, 0.05, 0.05,
                     0.05, NA, NA, 0.10),
    i2_tolerance = c(0.05, NA, 0.10, 0.05, 0.05, 0.05, 0.05, 0.05,
                     0.05, NA, NA, 0.20),
    tier = c("core", "core", "core", "core", "core", "extended", "core",
             "core", "extended", "core", "extended", "core"),
    stringsAsFactors = FALSE
  )
}

influence_metafor_cases <- function() {

  out <- data.frame(
    name = c(
      "bcg_meta-analysis",
      "bcg_meta-regression",
      "bcg_meta-regression4",
      "konstantopoulos2011_3lvl",
      "konstantopoulos2011_3lvl2"
    ),
    label = c(
      "normal simple",
      "normal meta-regression",
      "normal interaction",
      "normal multilevel",
      "normal multilevel meta-regression"
    ),
    oracle = c("equal", "equal_without_covratio", "rank", "cooks_finite", "cooks_finite"),
    tier = c("core", "core", "core", "core", "extended"),
    stringsAsFactors = FALSE
  )
  out[["skip_rows"]] <- I(list(integer(), c(4, 6, 13), c(4, 5, 6, 8, 10),
                               integer(), integer()))

  return(out)
}

residual_metafor_cases <- function() {

  out <- data.frame(
    name = c(
      "bcg_meta-analysis",
      "bcg_meta-regression",
      "bcg_meta-regression4",
      "bangertdrowns2004_location-scale",
      "konstantopoulos2011_3lvl",
      "konstantopoulos2011_3lvl2",
      "dat.lehmann2018-3PSM",
      "dat.lehmann2018-3PSMreg",
      "dat.lehmann2018-3PSM_neg",
      "dat.lehmann2018-PET",
      "dat.lehmann2018-PETreg",
      "dat.lehmann2018-PET_neg",
      "dat.lehmann2018-PEESE",
      "dat.lehmann2018-PEESEreg",
      "dat.lehmann2018-PEESE_neg",
      "nielweise2008_glmm",
      "bcg_glmm_reg"
    ),
    label = c(
      "normal simple",
      "normal meta-regression",
      "normal interaction",
      "normal location-scale",
      "normal multilevel",
      "normal multilevel meta-regression",
      "selection",
      "selection meta-regression",
      "selection negative",
      "PET",
      "PET meta-regression",
      "PET negative",
      "PEESE",
      "PEESE meta-regression",
      "PEESE negative",
      "GLMM",
      "GLMM meta-regression"
    ),
    kind = c(
      "standard",
      "regression",
      "interaction",
      "standard",
      "multilevel",
      "multilevel_no_loo",
      "selection_pos",
      "selection_reg",
      "selection_neg",
      "pet",
      "pet_reg",
      "pet",
      "peese",
      "peese_reg",
      "peese",
      "glmm",
      "glmm_reg"
    ),
    tolerance = c(0.05, 0.10, 0.10, 0.05, 0.05, 0.05, 0.05,
                  0.05, 0.05, 0.05, 0.05, 0.05, 0.05, 0.05,
                  0.05, 0.05, 0.10),
    tier = c("core", "core", "core", "core", "core", "extended", "core",
             "extended", "extended", "core", "core", "extended", "core",
             "core", "extended", "core", "core"),
    stringsAsFactors = FALSE
  )
  out[["rstudent"]] <- I(list(
    "equal", "rank", NULL, NULL, "rank", NULL, "selection_pos", NULL,
    "selection_neg", "equal", NULL, "equal", "equal", NULL, "equal",
    "glmm_align", NULL
  ))

  return(out)
}

prediction_metafor_cases <- function() {

  data.frame(
    name = c(
      "bcg_meta-analysis",
      "bcg_meta-regression",
      "bcg_meta-regression4",
      "bangertdrowns2004_location-scale",
      "konstantopoulos2011_3lvl",
      "konstantopoulos2011_3lvl2",
      "bcg_glmm",
      "bcg_glmm_reg",
      "dat.lehmann2018-3PSM",
      "dat.lehmann2018-3PSMreg",
      "dat.lehmann2018-3PSM_neg",
      "dat.lehmann2018-PET",
      "dat.lehmann2018-PETreg",
      "dat.lehmann2018-PET_neg",
      "dat.lehmann2018-PEESE",
      "dat.lehmann2018-PEESEreg",
      "dat.lehmann2018-PEESE_neg"
    ),
    label = c(
      "normal simple",
      "normal meta-regression",
      "normal interaction",
      "normal location-scale",
      "normal multilevel",
      "normal multilevel meta-regression",
      "GLMM",
      "GLMM meta-regression",
      "selection",
      "selection meta-regression",
      "selection negative",
      "PET",
      "PET meta-regression",
      "PET negative",
      "PEESE",
      "PEESE meta-regression",
      "PEESE negative"
    ),
    kind = c(
      "simple",
      "regression",
      "interaction",
      "scale",
      "multilevel",
      "multilevel",
      "glmm",
      "glmm_reg",
      "selection",
      "selection",
      "selection",
      "pet",
      "pet_reg",
      "pet",
      "peese",
      "peese_reg",
      "peese"
    ),
    tolerance = c(0.05, 0.05, 0.05, 0.05, 0.05, 0.10, 0.05,
                  0.15, 0.05, 0.05, 0.05, 0.05, 0.05, 0.05,
                  0.05, 0.05, 0.05),
    tau_tolerance = c(0.05, NA, NA, 0.05, 0.05, 0.05, 0.05,
                      0.10, 0.05, 0.05, 0.05, 0.05, 0.05, 0.05,
                      0.05, 0.05, 0.05),
    fitted_tolerance = c(NA, 0.05, NA, 0.05, NA, NA, NA, NA, NA,
                         NA, NA, 0.10, 0.10, 0.10, 0.10, 0.10, 0.10),
    tier = c("core", "core", "core", "core", "core", "extended", "core",
             "core", "core", "extended", "extended", "core", "core", "extended",
             "core", "core", "extended"),
    stringsAsFactors = FALSE
  )
}

prediction_newdata_metafor_cases <- function() {

  data.frame(
    name = c(
      "bcg_meta-regression",
      "bcg_meta-regression2",
      "bangertdrowns2004_location-scale",
      "konstantopoulos2011_3lvl2",
      "bcg_glmm_reg",
      "dat.lehmann2018-PETreg",
      "dat.lehmann2018-PEESEreg",
      "dat.lehmann2018-3PSMreg"
    ),
    label = c(
      "normal continuous moderators",
      "normal factor moderator",
      "normal location-scale",
      "normal multilevel moderator",
      "GLMM factor moderator",
      "PET moderator",
      "PEESE moderator",
      "selection moderator"
    ),
    kind = c(
      "normal_continuous",
      "normal_factor",
      "location_scale",
      "multilevel_mod",
      "glmm_factor",
      "pet_reg",
      "peese_reg",
      "selection_reg"
    ),
    tolerance = c(0.10, 0.10, 0.05, 0.05, 0.10, 0.07, 0.07, 0.05),
    tau_tolerance = c(NA, NA, 0.05, NA, NA, NA, NA, NA),
    tier = c("core", "core", "core", "extended", "core", "core", "core",
             "extended"),
    stringsAsFactors = FALSE
  )
}

marginal_means_cases <- function() {

  data.frame(
    name = c(
      "bcg_meta-regression",
      "bcg_meta-regression2",
      "bcg_meta-regression2b",
      "dat.lehmann2018_BMA.norm_mods",
      "dat.lehmann2018_RoBMA_mods"
    ),
    label = c(
      "continuous moderator",
      "factor moderator",
      "factor moderator transformed",
      "BMA moderator",
      "RoBMA moderator"
    ),
    parameter = c(
      "year",
      "alloc",
      "alloc",
      "Preregistered",
      "Preregistered"
    ),
    tier = c("core", "core", "visual-gallery", "core", "core"),
    stringsAsFactors = FALSE
  )
}

marginal_means_interaction_plot_cases <- function() {

  data.frame(
    name = c(
      "bcg_meta-regression3",
      "bcg_meta-regression4",
      "dat.lehmann2018_RoBMA_mods2"
    ),
    label = c(
      "factor by continuous interaction",
      "factor by factor interaction",
      "RoBMA factor by factor interaction"
    ),
    parameter = c(
      "alloc:year",
      "alloc:year_before1969",
      "Preregistered:Gender"
    ),
    tier = c("core", "core", "core"),
    stringsAsFactors = FALSE
  )
}

skip_if_not_full_visuals <- function(reason = NULL) {

  if (!is_true_env("ROBMA_TEST_FULL_VISUALS")) {
    detail <- if (is.null(reason)) "" else paste0(" ", reason)
    testthat::skip(paste0(
      "Skipping extended visual gallery by default.",
      detail,
      " Set ROBMA_TEST_FULL_VISUALS=TRUE to run it."
    ))
  }
}

context("Metafor coefficient references")

source(testthat::test_path("common-functions.R"))

skip_if_no_fits()
skip_if_not_installed("metafor")
skip_if_not_installed("metadat")

coef_fit_names <- c(
  "bcg_meta-analysis",
  "bcg_meta-regression",
  "bcg_meta-regression2",
  "bcg_meta-regression3",
  "bcg_meta-regression4",
  "bangertdrowns2004_location-scale",
  "konstantopoulos2011_3lvl",
  "konstantopoulos2011_3lvl2",
  "bcg_glmm",
  "bcg_glmm_reg",
  "nielweise2008_glmm",
  "dat.lehmann2018-PET",
  "dat.lehmann2018-PET_neg",
  "dat.lehmann2018-PETreg",
  "dat.lehmann2018-PEESE",
  "dat.lehmann2018-3PSM",
  "dat.lehmann2018-4PSM",
  "dat.lehmann2018-3PSM_neg",
  "dat.lehmann2018-3PSMreg"
)

fits <- lazy_fits(coef_fit_names, validate = TRUE)
info <- lazy_infos(coef_fit_names, validate = TRUE)

summary_mean <- function(name, row) {

  return(fits[[name]]$summary[row, "Mean"])
}

metafor_ref <- function(name) {

  return(info[[name]][["metafor"]])
}

expect_summary_mean_matches <- function(name, row, expected, tolerance) {

  expect_equal(
    summary_mean(name, row),
    expected,
    tolerance = tolerance,
    info      = paste0("Summary mean '", row, "' matches metafor for '", name, "'")
  )
}

coef_cases <- list(
  list("bcg_meta-analysis", "mu", function(m) m$beta[[1]], 0.05),
  list("bcg_meta-analysis", "tau", function(m) sqrt(m$tau2), 0.05),
  list("bcg_meta-regression", "(mu) ablat", function(m) m$beta[[2]], 0.10),
  list("bcg_meta-regression", "(mu) year", function(m) m$beta[[3]], 0.10),
  list("bcg_meta-regression2", "(mu) intercept", function(m) m$beta[[1]], 0.05),
  list("bcg_meta-regression2", "(mu) alloc[random]", function(m) m$beta[[2]], 0.15),
  list("bcg_meta-regression2", "(mu) alloc[systematic]", function(m) m$beta[[3]], 0.05),
  list("bcg_meta-regression3", "(mu) year", function(m) m$beta[[4]], 0.15),
  list("bcg_meta-regression3", "(mu) alloc[systematic]:year", function(m) m$beta[[5]], 0.15),
  list("bcg_meta-regression3", "(mu) alloc[random]:year", function(m) m$beta[[6]], 0.15),
  list("bcg_meta-regression4", "(mu) alloc[systematic]", function(m) m$beta[[3]], 0.15),
  list("bangertdrowns2004_location-scale", "(mu) intercept", function(m) m$beta[[1]], 0.05),
  list("bangertdrowns2004_location-scale", "(mu) meta[1]", function(m) m$beta[[2]], 0.05),
  list("bangertdrowns2004_location-scale", "(mu) ni100", function(m) m$beta[[3]], 0.05),
  list("bangertdrowns2004_location-scale", "(log_tau) ni100", function(m) 0.5 * m$alpha[[2]], 0.15),
  list("konstantopoulos2011_3lvl", "mu", function(m) m$beta[[1]], 0.01),
  list("konstantopoulos2011_3lvl", "tau", function(m) sqrt(m$tau2), 0.01),
  list("konstantopoulos2011_3lvl", "rho", function(m) m$rho, 0.05),
  list("konstantopoulos2011_3lvl2", "(mu) intercept", function(m) m$beta[[1]], 0.10),
  list("konstantopoulos2011_3lvl2", "(mu) vi", function(m) m$beta[[2]], 0.10),
  list("konstantopoulos2011_3lvl2", "tau", function(m) sqrt(m$tau2), 0.10),
  list("konstantopoulos2011_3lvl2", "rho", function(m) m$rho, 0.10),
  list("bcg_glmm", "mu", function(m) m$beta[[1]], 0.05),
  list("bcg_glmm", "tau", function(m) sqrt(m$tau2), 0.05),
  list("bcg_glmm_reg", "(mu) intercept", function(m) m$beta[[1]], 0.05),
  list("bcg_glmm_reg", "(mu) alloc[random]", function(m) m$beta[[2]], 0.15),
  list("bcg_glmm_reg", "(mu) alloc[systematic]", function(m) m$beta[[3]], 0.10),
  list("bcg_glmm_reg", "tau", function(m) sqrt(m$tau2), 0.10),
  list("nielweise2008_glmm", "mu", function(m) m$beta[[1]], 0.05),
  list("nielweise2008_glmm", "tau", function(m) sqrt(m$tau2), 0.10),
  list("dat.lehmann2018-PET", "mu", function(m) m$beta[[1]], 0.05),
  list("dat.lehmann2018-PET", "tau", function(m) sqrt(m$tau2), 0.05),
  list("dat.lehmann2018-PET", "PET", function(m) m$beta[[2]], 0.20),
  list("dat.lehmann2018-PET_neg", "mu", function(m) m$beta[[1]], 0.05),
  list("dat.lehmann2018-PET_neg", "tau", function(m) sqrt(m$tau2), 0.05),
  list("dat.lehmann2018-PET_neg", "PET", function(m) -m$beta[[2]], 0.20),
  list("dat.lehmann2018-PETreg", "(mu) intercept", function(m) m$beta[[1]], 0.05),
  list("dat.lehmann2018-PETreg", "(mu) Preregistered[Pre-Registered]", function(m) m$beta[[3]], 0.05),
  list("dat.lehmann2018-PETreg", "PET", function(m) m$beta[[2]], 0.20),
  list("dat.lehmann2018-PETreg", "tau", function(m) sqrt(m$tau2), 0.05),
  list("dat.lehmann2018-PEESE", "mu", function(m) m$beta[[1]], 0.05),
  list("dat.lehmann2018-PEESE", "tau", function(m) sqrt(m$tau2), 0.05),
  list("dat.lehmann2018-PEESE", "PEESE", function(m) m$beta[[2]], 0.05),
  list("dat.lehmann2018-3PSM", "mu", function(m) m$beta[[1]], 0.01),
  list("dat.lehmann2018-3PSM", "tau", function(m) sqrt(m$tau2), 0.01),
  list("dat.lehmann2018-3PSM", "omega[0.025,1]", function(m) m$delta[[2]], 0.10),
  list("dat.lehmann2018-4PSM", "mu", function(m) m$beta[[1]], 0.05),
  list("dat.lehmann2018-4PSM", "tau", function(m) sqrt(m$tau2), 0.05),
  list("dat.lehmann2018-4PSM", "omega[0.025,0.5]", function(m) m$delta[[2]], 0.20),
  list("dat.lehmann2018-4PSM", "omega[0.5,1]", function(m) m$delta[[3]], 0.20),
  list("dat.lehmann2018-3PSM_neg", "mu", function(m) m$beta[[1]], 0.01),
  list("dat.lehmann2018-3PSM_neg", "tau", function(m) sqrt(m$tau2), 0.01),
  list("dat.lehmann2018-3PSM_neg", "omega[0.025,1]", function(m) m$delta[[2]], 0.10),
  list("dat.lehmann2018-3PSMreg", "(mu) intercept", function(m) m$beta[[1]], 0.05),
  list("dat.lehmann2018-3PSMreg", "(mu) Preregistered[Pre-Registered]", function(m) m$beta[[2]], 0.05),
  list("dat.lehmann2018-3PSMreg", "tau", function(m) sqrt(m$tau2), 0.05),
  list("dat.lehmann2018-3PSMreg", "omega[0.025,1]", function(m) m$delta[[2]], 0.10)
)

test_that("cached summary means match metafor fit references", {

  for (case in coef_cases) {
    name      <- case[[1]]
    row       <- case[[2]]
    expected  <- case[[3]](metafor_ref(name))
    tolerance <- case[[4]]

    expect_summary_mean_matches(name, row, expected, tolerance)
  }
})

test_that("location-scale intercept matches metafor on the log-tau scale", {

  name <- "bangertdrowns2004_location-scale"

  expect_equal(
    log(summary_mean(name, "(log_tau) exp(intercept)")),
    0.5 * metafor_ref(name)$alpha[[1]],
    tolerance = 0.15,
    info      = paste0("Log-tau intercept matches metafor for '", name, "'")
  )
})

test_that("scaled predictor oracle for BCG meta-regression matches JAGS estimates", {

  data(dat.bcg, package = "metadat")
  dat <- metafor::escalc(
    measure = "RR",
    ai      = tpos,
    bi      = tneg,
    ci      = cpos,
    di      = cneg,
    data    = dat.bcg
  )

  fit_metafor <- metafor::rma(yi, vi, mods = ~ scale(ablat) + scale(year), data = dat)
  fit_brma    <- BayesTools::JAGS_estimates_table(fits[["bcg_meta-regression"]]$fit)

  expect_equal(fit_metafor$beta[[1]], fit_brma["(mu) intercept", "Mean"], tolerance = 0.10)
  expect_equal(fit_metafor$beta[[2]], fit_brma["(mu) ablat", "Mean"], tolerance = 0.10)
  expect_equal(fit_metafor$beta[[3]], fit_brma["(mu) year", "Mean"], tolerance = 0.10)
})

---
title: "Multilevel Meta-Analysis"
author: "František Bartoš"
date: "28th of April 2026"
output:
  rmarkdown::html_vignette:
    self_contained: yes
bibliography: ../inst/REFERENCES.bib
csl: ../inst/apa.csl
link-citations: true
vignette: >
  %\VignetteIndexEntry{Multilevel Meta-Analysis}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
  %\VignetteEngine{knitr::rmarkdown_notangle}
---

```{r child = "_vignette-nowrap.md", echo = FALSE, eval = TRUE}
```

```{r setup, include = FALSE}
source("vignette-cache.R", local = knitr::knit_global())

cached_fits <- vignette_cache(
  name     = "v10-metafor-parity-multilevel",
  objects  = c("fit1_brma"),
  packages = c("metafor", "metadat")
)

knitr::opts_chunk$set(
  collapse    = TRUE,
  comment     = "#>",
  eval        = vignette_cache_eval(cached_fits),
  message     = FALSE,
  warning     = FALSE,
  fig.width   = 7,
  fig.height  = 3.5,
  dev         = "png",
  fig.retina  = 3
)
if (.Platform$OS.type == "windows") {
  knitr::opts_chunk$set(dev.args = list(type = "cairo"))
}
```

```{r load-models, include = FALSE}
library("metafor")
library("RoBMA")

data("dat.konstantopoulos2011", package = "metadat")
dat <- dat.konstantopoulos2011

vignette_cache_load(cached_fits)
```

```{r precompute-models, include = FALSE, eval = FALSE}
library("metafor")
library("RoBMA")

data("dat.konstantopoulos2011", package = "metadat")
dat <- dat.konstantopoulos2011

fit1_brma <- brma(
  yi = yi, vi = vi, measure = "SMD",
  cluster = district,
  data = dat, seed = 1
)

vignette_cache_save(cached_fits)
```

This vignette illustrates how to use the `RoBMA` R package [@RoBMA] with multilevel meta-analytic data.
We walk through the school-calendar dataset from @konstantopoulos2011fixed, showing the `brma()` call with the `cluster` argument alongside its `metafor::rma.mv()` counterpart [@metafor].

All `brma()` calls keep the default prior distributions; see the [*Prior Distributions*](v01-prior-distributions.html) vignette for the prior definitions and customization options.
For the non-multilevel workflow showing more features (also applicable to the multilevel models) see the [*Bayesian Meta-Analysis*](v02-bayesian-meta-analysis.html) vignette.

The `RoBMA` R package currently supports the simple multilevel structure shown here, with one cluster grouping above the estimate level (parameterized by total heterogeneity $\tau^2$ and cluster-level allocation $\rho$). Broader support for multivariate models and more complex random-effects structures is planned for a future release.

## Multilevel Random-Effects Model

The school-calendar dataset from the `metadat` package contains 56 standardized mean differences nested in 11 school districts.
Effect sizes `yi` and sampling variances `vi` are already provided, so no `escalc()` step is needed.

```{r data}
data("dat.konstantopoulos2011", package = "metadat")
dat <- dat.konstantopoulos2011
head(dat)
```

`metafor::rma.mv()` fits a multilevel random-effects model with the compound-symmetric structure `random = ~ school | district`, parameterized by the total heterogeneity $\tau$ and the within-district correlation $\rho$.

```{r fit1-metafor}
fit1_metafor <- metafor::rma.mv(yi, vi, random = ~ school | district, data = dat)
fit1_metafor
```

The matching Bayesian fit replaces the `random = ...` argument with `cluster = district`.
`brma()` parameterizes the multilevel model in the same compound-symmetric form: a total heterogeneity $\tau$ and a cluster-level allocation $\rho$ where $\tau_{between} = \tau\sqrt{\rho}$ and $\tau_{within} = \tau\sqrt{1 - \rho}$.
`measure = "SMD"` selects the default prior distributions for standardized mean differences.

```{r fit1-brma, eval = FALSE}
fit1_brma <- brma(
  yi = yi, vi = vi, measure = "SMD",
  cluster = district,
  data = dat, seed = 1
)
```

`summary()` reports posterior means, credible intervals, and (suppressed here) MCMC convergence diagnostics for `mu`, `tau`, and `rho`.

```{r fit1-brma-summary}
summary(fit1_brma, include_mcmc_diagnostics = FALSE)
```

The posterior mean for the pooled effect tracks the REML estimate from the `metafor` package, with comparable cluster-level allocation `rho` and total heterogeneity `tau`.

## Heterogeneity: $\tau$ and $\rho$

`metafor::confint()` reports profile-likelihood confidence intervals for `tau^2`, `tau`, and `rho`.

```{r heterogeneity-confint}
confint(fit1_metafor)
```

`summary_heterogeneity()` is the Bayesian counterpart, reporting posterior summaries of the same quantities along with the partitioned within- and between-cluster components.

```{r heterogeneity-bayes}
summary_heterogeneity(fit1_brma)
```

`plot()` visualizes the posterior, and `prior = TRUE` overlays the prior distribution so the shift from prior to posterior is visible.
Side-by-side panels for `tau` and `rho` show the bulk of the posterior moving away from the weakly informative defaults.

```{r tau-rho-posterior, fig.width = 7, fig.height = 3.2}
par(mfrow = c(1, 2), mar = c(4, 4, 2, 1), cex.axis = 0.8, cex.lab = 0.8, cex.main = 0.75)
plot(fit1_brma, parameter = "tau", prior = TRUE, main = "tau", xlim = c(0, 1))
plot(fit1_brma, parameter = "rho", prior = TRUE, main = "rho", xlim = c(0, 1))
```

The `tau` posterior concentrates around 0.3 with credible interval comparable to the profile-likelihood bounds, while the `rho` posterior favors a moderate-to-high cluster-level allocation, broadly consistent with the REML point estimate of 0.67.


## Other Inference Helpers

All inference helpers available for any other `brma()` fit work the same way for the multilevel fit specified via `cluster`.
This includes meta-regression and location-scale models, model comparison, and fit diagnostics.
See the [*Bayesian Meta-Analysis*](v02-bayesian-meta-analysis.html) vignette for the full walkthrough; the only thing that changes here is the addition of `cluster = district`.


## References

---
title: "Feature Coverage"
author: "František Bartoš"
date: "29th of April 2026"
output:
  rmarkdown::html_vignette:
    self_contained: yes
bibliography: ../inst/REFERENCES.bib
csl: ../inst/apa.csl
link-citations: true
vignette: >
  %\VignetteIndexEntry{Feature Coverage}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
  %\VignetteEngine{knitr::rmarkdown_notangle}
---

```{r child = "_vignette-nowrap.md", echo = FALSE, eval = TRUE}
```

```{r setup, include = FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment  = "#>",
  message  = FALSE,
  warning  = FALSE
)

tick    <- '<span class="yes">&#10003;</span>'
limited <- '<span class="limited">limited</span>'
no      <- '<span class="no">No</span>'
na_cell <- '<span class="na-cell">n/a</span>'

model_columns <- c(
  "brma.norm", "brma.glmm", "bselmodel", "bPET/bPEESE",
  "BMA.norm", "BMA.glmm", "RoBMA"
)

reference_links <- c(
  "add_loo()"                            = "../reference/add_loo.brma.html",
  "add_marglik()"                        = "../reference/add_marglik.brma.html",
  "add_waic()"                           = "../reference/add_waic.brma.html",
  "as_draws()"                           = "../reference/as_draws.brma.html",
  "as_draws_array()"                     = "../reference/as_draws.brma.html",
  "as_draws_df()"                        = "../reference/as_draws.brma.html",
  "as_draws_list()"                      = "../reference/as_draws.brma.html",
  "as_draws_matrix()"                    = "../reference/as_draws.brma.html",
  "as_draws_rvars()"                     = "../reference/as_draws.brma.html",
  "as_zplot()"                           = "../reference/as_zplot.brma.html",
  "bayes_factor()"                       = "../reference/bf.brma.html",
  "bf()"                                 = "../reference/bf.brma.html",
  "blup()"                               = "../reference/blup.brma.html",
  "bPET()"                               = "../reference/bPET.html",
  "bPEESE()"                             = "../reference/bPEESE.html",
  "brma()"                               = "../reference/brma.html",
  "brma.glmm()"                          = "../reference/brma.glmm.html",
  "BMA()"                                = "../reference/BMA.html",
  "BMA.glmm()"                           = "../reference/BMA.glmm.html",
  "bridge_sampler()"                     = "../reference/bridge_sampler.brma.html",
  "bselmodel()"                          = "../reference/bselmodel.html",
  "check_loo()"                          = "../reference/check_loo.brma.html",
  "coef()"                               = "../reference/coef.brma.html",
  "contr.meandif()"                      = "../reference/contr.BayesTools.html",
  "contr.orthonormal()"                  = "../reference/contr.BayesTools.html",
  "contr.treatment()"                    = "https://stat.ethz.ch/R-manual/R-devel/library/stats/html/contrast.html",
  "cooks.distance()"                     = "../reference/cooks.distance.brma.html",
  "covratio()"                           = "../reference/covratio.brma.html",
  "dfbetas()"                            = "../reference/dfbetas.brma.html",
  "dffits()"                             = "../reference/dffits.brma.html",
  "estimate_unit_information_sd()"       = "../reference/estimate_unit_information_sd.html",
  "fitted()"                             = "../reference/fitted.brma.html",
  "funnel()"                             = "../reference/funnel.html",
  "galbraith()"                          = "../reference/radial.html",
  "hatvalues()"                          = "../reference/hatvalues.brma.html",
  "influence()"                          = "../reference/influence.brma.html",
  "interpret()"                          = "../reference/interpret.html",
  "logLik()"                             = "../reference/logLik.brma.html",
  "logml()"                              = "../reference/logml.brma.html",
  "loo()"                                = "../reference/loo.brma.html",
  "loo::pareto_k_ids()"                  = "https://mc-stan.org/loo/reference/pareto-k-diagnostic.html",
  "loo::pareto_k_table()"                = "https://mc-stan.org/loo/reference/pareto-k-diagnostic.html",
  "loo_compare()"                        = "../reference/loo_compare.brma.html",
  "loo_weights()"                        = "../reference/loo_weights.brma.html",
  "marginal_means()"                     = "../reference/marginal_means.brma.html",
  "nobs()"                               = "../reference/nobs.brma.html",
  "plot()"                               = "../reference/plot.brma.html",
  "plot_diagnostic()"                    = "../reference/plot_diagnostic.html",
  "plot_diagnostic_autocorrelation()"    = "../reference/plot_diagnostic.html",
  "plot_diagnostic_density()"            = "../reference/plot_diagnostic.html",
  "plot_diagnostic_trace()"              = "../reference/plot_diagnostic.html",
  "plot_pet_peese()"                     = "../reference/plot_pet_peese.html",
  "plot_prior()"                         = "../reference/plot_prior.html",
  "plot_weightfunction()"                = "../reference/plot_weightfunction.html",
  "pooled_effect()"                      = "../reference/pooled_effect.brma.html",
  "pooled_heterogeneity()"               = "../reference/pooled_heterogeneity.brma.html",
  "post_prob()"                          = "../reference/post_prob.brma.html",
  "predict()"                            = "../reference/predict.brma.html",
  "print_prior()"                        = "../reference/print_prior.html",
  "prior()"                              = "../reference/prior.html",
  "prior_factor()"                       = "../reference/prior_factor.html",
  "prior_informed()"                     = "../reference/prior_informed.html",
  "prior_none()"                         = "../reference/prior_none.html",
  "prior_PEESE()"                        = "../reference/prior_PEESE.html",
  "prior_PET()"                          = "../reference/prior_PET.html",
  "prior_weightfunction()"               = "../reference/prior_weightfunction.html",
  "qqnorm()"                             = "../reference/qqnorm.brma.html",
  "radial()"                             = "../reference/radial.html",
  "ranef()"                              = "../reference/ranef.brma.html",
  "regplot()"                            = "../reference/regplot.html",
  "residuals()"                          = "../reference/residuals.brma.html",
  "RoBMA()"                              = "../reference/RoBMA.html",
  "rstandard()"                          = "../reference/rstandard.brma.html",
  "rstudent()"                           = "../reference/rstudent.brma.html",
  "simulate()"                           = "https://stat.ethz.ch/R-manual/R-devel/library/stats/html/simulate.html",
  "summary()"                            = "../reference/summary.brma.html",
  "summary_heterogeneity()"              = "../reference/summary_heterogeneity.brma.html",
  "summary_models()"                     = "../reference/summary_models.html",
  "true_effects()"                       = "../reference/true_effects.brma.html",
  "update()"                             = "../reference/update.brma.html",
  "vcov()"                               = "https://stat.ethz.ch/R-manual/R-devel/library/stats/html/vcov.html",
  "vif()"                                = "../reference/vif.brma.html",
  "waic()"                               = "../reference/waic.brma.html",
  "weights()"                            = "https://stat.ethz.ch/R-manual/R-devel/library/stats/html/weights.html",
  "wf_cumulative()"                      = "../reference/prior_weightfunction.html",
  "wf_fixed()"                           = "../reference/prior_weightfunction.html",
  "wf_independent()"                     = "../reference/prior_weightfunction.html",
  "forest()"                             = "https://wviechtb.github.io/metafor/reference/forest.rma.html",
  "baujat()"                             = "https://wviechtb.github.io/metafor/reference/baujat.html",
  "gosh()"                               = "https://wviechtb.github.io/metafor/reference/gosh.html",
  "labbe()"                              = "https://wviechtb.github.io/metafor/reference/labbe.html"
)

model_column_links <- list(
  "brma.norm"   = "../reference/brma.html",
  "brma.glmm"   = "../reference/brma.glmm.html",
  "bselmodel"   = "../reference/bselmodel.html",
  "bPET/bPEESE" = c("../reference/bPET.html", "../reference/bPEESE.html"),
  "BMA.norm"    = "../reference/BMA.html",
  "BMA.glmm"    = "../reference/BMA.glmm.html",
  "RoBMA"       = "../reference/RoBMA.html"
)

status_map <- c(
  Y = tick,
  L = limited,
  N = no,
  A = na_cell
)

feature_row <- function(topic, feature, pattern) {

  values <- strsplit(pattern, "", fixed = TRUE)[[1]]
  if (length(values) != length(model_columns)) {
    stop("Status pattern has wrong length: ", pattern, call. = FALSE)
  }

  out <- c(
    list(Topic = topic, Feature = feature),
    stats::setNames(as.list(status_map[values]), model_columns)
  )

  return(out)
}

feature_table <- function(rows) {

  rows <- as.data.frame(do.call(rbind, rows), stringsAsFactors = FALSE, check.names = FALSE)
  rows[] <- lapply(rows, function(x) {

    if (is.list(x)) {
      return(vapply(x, function(y) paste0(as.character(y), collapse = ""), character(1)))
    }
    return(as.character(x))
  })

  cat('
<style>
.feature-matrix {
  border-collapse: collapse;
  width: 100%;
  font-size: 0.84em;
}
.feature-matrix th,
.feature-matrix td {
  border: 1px solid #d0d7de;
  padding: 4px 6px;
  vertical-align: top;
}
.feature-matrix th {
  background: #f6f8fa;
}
.feature-matrix .topic-head {
  min-width: 60px;
  width: 60px;
}
.feature-matrix .feature-head {
  min-width: 25em;
}
.feature-matrix .model-head {
  height: 118px;
  min-width: 34px;
  max-width: 34px;
  width: 34px;
  padding: 0;
  position: relative;
  vertical-align: bottom;
}
.feature-matrix .rotated-header {
  bottom: 6px;
  display: block;
  left: 50%;
  line-height: 1;
  position: absolute;
  transform: rotate(-90deg);
  transform-origin: left center;
  white-space: nowrap;
}
.feature-matrix .topic-cell {
  background: #f6f8fa;
  min-width: 60px;
  width: 60px;
  padding: 0;
  text-align: center;
  vertical-align: middle;
}
.feature-matrix .topic-cell span {
  display: inline-block;
  transform: rotate(-90deg);
  white-space: normal;
  width: 8em;
  text-align: center;
  font-weight: 700;
  line-height: 1.1;
}
.feature-matrix .feature-cell {
  min-width: 25em;
}
.feature-matrix .feature-cell a {
  color: #0969da;
  text-decoration: none;
}
.feature-matrix .feature-cell a:hover {
  text-decoration: underline;
}
.feature-matrix .feature-cell a code {
  color: inherit;
}
.feature-matrix .status-cell {
  text-align: center;
  vertical-align: middle;
  white-space: nowrap;
  width: 34px;
}
.feature-matrix .yes {
  color: #1a7f37;
  font-weight: 700;
}
.feature-matrix .limited {
  color: #9a6700;
}
.feature-matrix .no,
.feature-matrix .na-cell {
  color: #6e7781;
}
</style>
')

  escape_html_text <- function(x) {

    x <- as.character(x)
    x <- gsub("&", "&amp;", x, fixed = TRUE)
    x <- gsub("<", "&lt;", x, fixed = TRUE)
    x <- gsub(">", "&gt;", x, fixed = TRUE)
    x <- gsub('"', "&quot;", x, fixed = TRUE)
    return(x)
  }

  format_code_span <- function(code) {

    code_html <- paste0("<code>", escape_html_text(code), "</code>")
    link      <- if (code %in% names(reference_links)) reference_links[[code]] else NA_character_

    if (!is.na(link)) {
      return(paste0('<a href="', escape_html_text(link), '">', code_html, "</a>"))
    }

    return(code_html)
  }

  escape_html <- function(x) {

    x       <- as.character(x)
    matches <- gregexpr("`[^`]+`", x, perl = TRUE)[[1]]

    if (matches[1] == -1L) {
      return(escape_html_text(x))
    }

    match_lengths <- attr(matches, "match.length")
    pieces        <- character(0)
    position      <- 1L

    for (j in seq_along(matches)) {
      match_start <- matches[j]
      match_end   <- match_start + match_lengths[j] - 1L

      if (match_start > position) {
        pieces <- c(pieces, escape_html_text(substr(x, position, match_start - 1L)))
      }

      code <- substr(x, match_start + 1L, match_end - 1L)
      pieces <- c(pieces, format_code_span(code))
      position <- match_end + 1L
    }

    if (position <= nchar(x)) {
      pieces <- c(pieces, escape_html_text(substr(x, position, nchar(x))))
    }

    return(paste0(pieces, collapse = ""))
  }

  format_model_column <- function(model_column) {

    link <- model_column_links[[model_column]]

    if (length(link) == 2L) {
      return(paste0(
        '<a href="', escape_html_text(link[[1]]), '">bPET</a>/',
        '<a href="', escape_html_text(link[[2]]), '">bPEESE</a>'
      ))
    }

    return(paste0(
      '<a href="', escape_html_text(link), '">',
      escape_html_text(model_column),
      "</a>"
    ))
  }

  topic_runs <- rle(rows$Topic)
  topic_start <- cumsum(c(1L, head(topic_runs$lengths, -1L)))
  topic_span <- rep(0L, nrow(rows))
  topic_span[topic_start] <- topic_runs$lengths

  cat('<div style="overflow-x: auto;">\n')
  cat('<table class="feature-matrix">\n')
  cat("<thead><tr>")
  cat('<th class="topic-head">Topic</th>')
  cat('<th class="feature-head">Feature</th>')
  for (model_column in model_columns) {
    cat('<th class="model-head"><span class="rotated-header">', format_model_column(model_column), "</span></th>", sep = "")
  }
  cat("</tr></thead>\n")
  cat("<tbody>\n")
  for (i in seq_len(nrow(rows))) {
    cat("<tr>")
    if (topic_span[i] > 0L) {
      cat(
        '<td class="topic-cell" rowspan="', topic_span[i], '"><span>',
        escape_html(rows$Topic[i]),
        "</span></td>",
        sep = ""
      )
    }
    cat('<td class="feature-cell">', escape_html(rows$Feature[i]), "</td>", sep = "")
    for (model_column in model_columns) {
      cat('<td class="status-cell">', rows[[model_column]][i], "</td>", sep = "")
    }
    cat("</tr>\n")
  }
  cat("</tbody>\n")
  cat("</table>\n")
  cat("</div>\n")
}
```

This vignette overviews current features of the `RoBMA` R package [@RoBMA].
The table is organized by substantive capability.
Some rows point to additional features not (yet) available in the `RoBMA` R package but featured in the `metafor` package [@metafor] as reference points.

Green ticks mark available features.
Cells marked `limited` mean the feature exists but with narrower model coverage.



```{r feature-matrix, echo = FALSE, results = "asis"}
rows <- list(
  feature_row("Model / Structure", "Fixed- and random-effects models", "YYYYYYY"),
  feature_row("Model / Structure", "Moderation (`mods`)", "YYYYYYY"),
  feature_row("Model / Structure", "Location-scale models (`scale`)", "YYYYYYY"),
  feature_row("Model / Structure", "Multilevel models (`cluster`)", "YYYYYYY"),
  feature_row("Model / Structure", "Model averaging", "NNNNYYY"),
  feature_row("Model / Structure", "Inclusion Bayes factors", "NNNNYYY"),
  feature_row("Model / Structure", "General multivariate / covariance structures", "NNNNNNN"),
  feature_row("Model / Structure", "Robust / sandwich inference", "NNNNNNN"),
  feature_row("Model / Structure", "Refit / update (`update()`)", "YYYYYYY"),

  feature_row("Priors", "Default prior distributions", "YYYYYYY"),
  feature_row("Priors", "Empirical/informed prior distributions (`prior_informed()`)", "YYYYYYY"),
  feature_row("Priors", "Custom prior distributions (`prior()`, `prior_factor()`, `prior_weightfunction()`, `prior_none()`)", "YYYYYYY"),
  feature_row("Priors", "PET / PEESE prior distributions (`prior_PET()`, `prior_PEESE()`)", "NNNYNNY"),
  feature_row("Priors", "Prior-only inspection (`only_priors = TRUE`)", "YYYYYYY"),
  feature_row("Priors", "Weight-function shape (`wf_cumulative()`, `wf_fixed()`, `wf_independent()`)", "NNYNNNY"),
  feature_row("Priors", "Factor contrasts (`contr.treatment()`, `contr.meandif()`, `contr.orthonormal()`)", "YYYYYYY"),
  feature_row("Priors", "UISD estimation (`estimate_unit_information_sd()`)", "YYYYYYY"),

  feature_row("Publication Bias", "Selection models", "NNYNNNY"),
  feature_row("Publication Bias", "PET / PEESE models", "NNNYNNY"),
  feature_row("Publication Bias", "Model averaging over bias models", "NNNNNNY"),
  feature_row("Publication Bias", "Bias-adjusted summaries / predictions", "NNYYNNY"),
  feature_row("Publication Bias", "Trim-fill / fail-safe N", "NNNNNNN"),

  feature_row("Prediction", "Pooled effect / heterogeneity summaries (`pooled_effect()`, `pooled_heterogeneity()`)", "YYYYYYY"),
  feature_row("Prediction", "Heterogeneity decomposition (`summary_heterogeneity()`, tau², I², H²)", "YYYYYYY"),
  feature_row("Prediction", "Fitted-value extraction (`fitted()`)", "YYYYYYY"),
  feature_row("Prediction", "Prediction for new covariate values (`predict()`)", "YYYYYYY"),
  feature_row("Prediction", "Posterior predictive response summaries (`predict()`)", "YYYYYYY"),
  feature_row("Prediction", "True effects / BLUPs / random effects (`true_effects()`, `blup()`, `ranef()`)", "YYYYYYY"),
  feature_row("Prediction", "Estimated marginal means (`marginal_means()`)", "YYYYYYY"),

  feature_row("Plots", "Posterior plots (`plot()`)", "YYYYYYY"),
  feature_row("Plots", "Prior plots (`plot_prior()`)", "YYYYYYY"),
  feature_row("Plots", "Marginal means plots (`marginal_means()` + `plot()`)", "YYYYYYY"),
  feature_row("Plots", "Regression plots (`regplot()`)", "YYYYYYY"),
  feature_row("Plots", "Funnel plots (`funnel()`)", "YYYYYYY"),
  feature_row("Plots", "zplot diagnostics (`as_zplot()`/`zplot()`)", "YYYYYYY"),
  feature_row("Plots", "Weight-function plots (`plot_weightfunction()`)", "NNYNNNY"),
  feature_row("Plots", "PET-PEESE plots (`plot_pet_peese()`)", "NNNYNNY"),
  feature_row("Plots", "Radial / Galbraith plots (`radial()`, `galbraith()`)", "YLYYYLY"),
  feature_row("Plots", "Forest plots (`forest()`)", "NNNNNNN"),
  feature_row("Plots", "Baujat plots (`baujat()`)", "NNNNNNN"),
  feature_row("Plots", "GOSH plots (`gosh()`)", "NNNNNNN"),
  feature_row("Plots", "L'Abbe plots (`labbe()`)", "NNNNNNN"),

  feature_row("Residuals / Diagnostics", "Raw residuals (`residuals()`)", "YYYYYYY"),
  feature_row("Residuals / Diagnostics", "Pearson / standardized residuals (`rstandard()`)", "YNNYYNL"),
  feature_row("Residuals / Diagnostics", "Studentized residuals (LOO-PIT) (`rstudent()`)", "YYYYYYY"),
  feature_row("Residuals / Diagnostics", "Q-Q plots (`qqnorm()`)", "YYYYYYY"),
  feature_row("Residuals / Diagnostics", "Cook's distances (`cooks.distance()`)", "YNNYYNN"),
  feature_row("Residuals / Diagnostics", "DFBETAS (`dfbetas()`)", "YYYYYYY"),
  feature_row("Residuals / Diagnostics", "DFFITS (`dffits()`)", "YNNYYNN"),
  feature_row("Residuals / Diagnostics", "Covariance ratios (`covratio()`)", "YYYYYYY"),
  feature_row("Residuals / Diagnostics", "Hat values (`hatvalues()`)", "YNNYYNN"),
  feature_row("Residuals / Diagnostics", "Combined influence summary (`influence()`)", "YLLYYLL"),
  feature_row("Residuals / Diagnostics", "LOO / WAIC diagnostics (`check_loo()`, `loo::pareto_k_ids()`, `loo::pareto_k_table()`)", "YYYYYYY"),
  feature_row("Residuals / Diagnostics", "Moderator collinearity diagnostics (`vif()`)", "YYYYYYY"),
  feature_row("Residuals / Diagnostics", "Refit leave-one-out / permutation tests", "NNNNNNN"),

  feature_row("MCMC Diagnostics", "Posterior summaries (Rhat, ESS, MCMC error) (`summary()`)", "YYYYYYY"),
  feature_row("MCMC Diagnostics", "Trace plots (`plot_diagnostic_trace()`)", "YYYYYYY"),
  feature_row("MCMC Diagnostics", "Density plots (`plot_diagnostic_density()`)", "YYYYYYY"),
  feature_row("MCMC Diagnostics", "Autocorrelation plots (`plot_diagnostic_autocorrelation()`)", "YYYYYYY"),
  feature_row("MCMC Diagnostics", "Diagnostic-plot wrapper (`plot_diagnostic()`)", "YYYYYYY"),

  feature_row("Model Comparison", "Marginal likelihood (`add_marglik()`, `logml()`)", "YYYYNNN"),
  feature_row("Model Comparison", "WAIC/LOO (`add_waic()`, `add_loo()`, `waic()`, `loo()`)", "YYYYYYY"),
  feature_row("Model Comparison", "Bayes factors (`bf()`, `bayes_factor()`, `post_prob()`)", "YYYYYYY"),
  feature_row("Model Comparison", "WAIC comparison (`loo_compare()`, `loo_weights()`)", "YYYYYYY"),
  feature_row("Model Comparison", "LOO comparison (`loo_compare()`, `loo_weights()`)", "YYYYYYY"),
  feature_row("Model Comparison", "AIC / BIC", "NNNNNNN"),

  feature_row("Reporting", "Plain-text interpretation (`interpret()`)", "YYYYYYY"),
  feature_row("Reporting", "Sub-model summary (`summary_models()`)", "NNNNYYY"),
  feature_row("Reporting", "Prior inspection (`print_prior()`)", "YYYYYYY"),

  feature_row("Extraction", "Posterior draw extraction (`as_draws()`, `as_draws_array()`, `as_draws_df()`, `as_draws_list()`, `as_draws_matrix()`, `as_draws_rvars()`)", "YYYYYYY"),
  feature_row("Extraction", "Coefficients (`coef()`)", "YYYYYYY"),
  feature_row("Extraction", "Point-wise log-likelihood (`logLik()`)", "YYYYYYY"),
  feature_row("Extraction", "Sample size (`nobs()`)", "YYYYYYY"),
  feature_row("Extraction", "Variance-covariance matrix (`vcov()`)", "NNNNNNN"),
  feature_row("Extraction", "Credible intervals (`summary()`, `summary_heterogeneity()`, `pooled_effect()`, `pooled_heterogeneity()`)", "YYYYYYY"),
  feature_row("Extraction", "Model weights (`weights()`)", "NNNNNNN"),
  feature_row("Extraction", "Simulated responses (`simulate()`)", "NNNNNNN")
)

# Intentionally skipped exported functions in this feature table:
# bridge_sampler(), contr.independent(), RoBMA.options(), RoBMA.get_option(),
# set_autofit_control(), set_convergence_checks().

feature_table(rows)
```

## References

---
title: "Multilevel Robust Bayesian Meta-Analysis"
author: "František Bartoš"
date: "2025"
output:
  rmarkdown::html_vignette:
    self_contained: yes
bibliography: ../inst/REFERENCES.bib
csl: ../inst/apa.csl
vignette: >
  %\VignetteIndexEntry{Multilevel Robust Bayesian Meta-Analysis}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r setup, include = FALSE}
is_check <- ("CheckExEnv" %in% search()) ||
              any(c("_R_CHECK_TIMINGS_", "_R_CHECK_LICENSE_") %in% names(Sys.getenv())) ||
              !file.exists("../models/MultilevelRoBMA/fit_Johnides2025.RDS")
knitr::opts_chunk$set(
  collapse  = TRUE,
  comment   = "#>",
  eval      = !is_check,
  dev       = "png")
if(.Platform$OS.type == "windows"){
  knitr::opts_chunk$set(dev.args = list(type = "cairo"))
}
```

```{r include = FALSE}
library(RoBMA)
fit        <- readRDS(file = "../models/MultilevelRoBMA/fit_Johnides2025.RDS")
fit_simple <- readRDS(file = "../models/MultilevelRoBMA/fit_Johnides2025_single.RDS")
```

```{r include = FALSE, eval = FALSE}
# R package version updating
library(RoBMA)
data("Johnides2025", package = "RoBMA")

fit <- RoBMA(
  d         = Johnides2025$d, 
  se        = Johnides2025$se, 
  study_ids = Johnides2025$study, 
  algorithm = "ss",
  adapt     = 5000, 
  burnin    = 5000, 
  sample    = 10000, 
  parallel  = TRUE, 
  seed      = 1, 
  autofit   = FALSE
)
saveRDS(fit, file = "../models/MultilevelRoBMA/fit_Johnides2025.RDS", compress = "xz")

fit_simple <- RoBMA(
  d         = Johnides2025$d, 
  se        = Johnides2025$se, 
  algorithm = "ss",
  adapt     = 5000, 
  burnin    = 5000, 
  sample    = 10000, 
  parallel  = TRUE, 
  seed      = 1, 
  autofit   = FALSE
)
saveRDS(fit_simple, file = "../models/MultilevelRoBMA/fit_Johnides2025_single.RDS", compress = "xz")
```

**This vignette accompanies the manuscript [Robust Bayesian Multilevel Meta-Analysis: Adjusting for Publication Bias in the Presence of Dependent Effect Sizes](https://doi.org/10.31234/osf.io/9tgp2_v1) preprinted at *PsyArXiv* [@bartos2025robust].**

This vignette reproduces the first example from the manuscript. For multilevel meta-regression with moderators see [Multilevel Robust Bayesian Model-Averaged Meta-Regression](MultilevelRoBMARegression.html).

Multilevel Robust Bayesian Meta-Analysis (RoBMA) extends the standard RoBMA framework to datasets containing multiple effect sizes from the same studies. We demonstrate the method using data from @johnides2025secondary, who meta-analyzed 412 effect sizes from 128 studies investigating secondary benefits of family-based treatments for childhood disorders.

### When to Use Multilevel RoBMA

Multilevel RoBMA is appropriate for dataset that contain multiple effect sizes from the same studies. The multilevel approach explicitly models the clustering structure through the `study_ids` argument, which:

 - accounts for dependencies among effect sizes from the same study,
 - partitions heterogeneity into within-study and between-study components,
 - while simultaneously adjusting for publication bias at the study level (corresponding to selective reporting).

### Loading the Data

We load the package and examine the dataset structure.
```{r}
library(RoBMA)
data("Johnides2025", package = "RoBMA")
```

The dataset contains effect sizes (`d`), standard errors (`se`), and study identifiers (`study`).

```{r}
head(Johnides2025)
```

Multiple rows share the same study name, indicating that these effect sizes come from the same study. 
For example, the first five effect sizes are from "Price et al. (2012)" and might be more similar to each other than to effect sizes from other studies.

### Fitting the Multilevel Model

The `study_ids` argument specifies which effect sizes belong together. We use `algorithm = "ss"` (spike-and-slab), which is faster for complex models and neccessary for estimating multilevel selection models.
The `"ss"` algorithm estimates the models and parameters simultaneously, therefore we increase MCMC adaptation, iteration, and sampling to obtain more reliable estimates
(we disable automatic convergence checking with `autofit = FALSE` to speed up the fitting process here).

```{r, eval = FALSE}
fit <- RoBMA(
  d         = Johnides2025$d, 
  se        = Johnides2025$se, 
  study_ids = Johnides2025$study, 
  algorithm = "ss",
  adapt     = 5000, 
  burnin    = 5000, 
  sample    = 10000, 
  parallel  = TRUE, 
  seed      = 1, 
  autofit   = FALSE
)
```

*Note: This model takes 10-15 minutes with parallel processing enabled. Without parallel processing, expect over an hour.*

### Interpreting the Results

The `summary()` output contains two main sections.

```{r}
summary(fit)
```

#### Components Summary

This table shows Bayes factors testing the presence of effect, heterogeneity, and publication bias. Each row displays the prior probability, posterior probability after observing the data, and the inclusion Bayes factor comparing models with versus without that component.

In our example, the inclusion BF for the effect is 0.927, indicating weak evidence against an effect. The inclusion BF for heterogeneity and publication bias are reported as `Inf`, i.e., beyond the numerical precision of the `ss` algorithm (> 10⁶), indicating extreme evidence for both components.

We also notice warning about effective sample size (ESS) below the set target. The difference is however not substantial and we can safely ignore it. 

#### Model-Averaged Estimates

This section provides meta-analytic estimates averaged across all models, weighted by their posterior probabilities. The average effect size is `mu`, overall heterogeneity is `tau`, and the heterogeneity allocation parameter `rho`.
The heterogeneity allocation parameter `rho` indicates the proportion of heterogeneity within versus between studies: `rho = 0` means all heterogeneity is between studies, `rho = 1` means all is within studies, and `rho = 0.5` means equal split.
The remaining parameters summarize the weight function and PET/PEESE regression for publication bias.

In our example, we find small effect size (*d* = 0.050, 95% CI [0.000, 0.173]), substantial heterogeneity (τ = 0.40, 95% CI [0.348, 0.458]), and nearly balanced heterogeneity allocation (ρ = 0.461, 95% CI [0.306, 0.603]). The large heterogeneity combined with the small pooled effect suggests substantial variation in true effects. This distribution of true effect sizes can be obtain by the `summary_heterogeneity()` function:

```{r}
summary_heterogeneity(fit)
```

The wide prediction interval (*d* = -0.752 to 0.845) quantifies the degree of this heterogeneity: some studies may show benefits while others show harm.

### Model Types Summary

To understand which publication bias mechanisms the data support, we examine the posterior distribution across model types:

```{r}
summary(fit, type = "models")
```

Most (all) posterior probability is allocated to selection models (weight functions). The publication bias adjustment therefore reflects selective reporting rather than small study effects (PET/PEESE regression), specifically, one-sided selection operating on marginally significant $p$-values and direction of the effect.

### Visualizing the Weight Function

The weight function shows how publication probability varies across *p*-value ranges:

```{r, fig.width = 6, fig.height = 4}
plot(fit, parameter = "weightfunction", rescale_x = TRUE)
```

The plot displays one-sided *p*-value cutoffs (x-axis) against relative publication probability (y-axis), averaged across weight function specifications. A flat line at 1.0 indicates no publication bias; values below 1.0 indicate suppression. 

In our example, we notice a small drop in the relative publication probability at cutoff corresponding to p = 0.05 (one-sided, i.e., 0.10 two-sided: selection on marginally significant $p$-values) and a much sharper drop corresponding to the direction of the effect (p = 0.50).


### Comparison with Single-Level RoBMA

To understand the importance of accounting for nested structure, we compare our results with a standard single-level RoBMA that ignores dependencies among effect sizes from the same study.

```{r, eval = FALSE}
fit_simple <- RoBMA(
  d         = Johnides2025$d, 
  se        = Johnides2025$se, 
  algorithm = "ss",
  adapt     = 5000, 
  burnin    = 5000, 
  sample    = 10000, 
  parallel  = TRUE, 
  seed      = 1, 
  autofit   = FALSE
)
```

```{r}
summary(fit_simple)
```

The single-level model differs by:

 - omitting `study_ids`: it treats all effect sizes as independent,
 - estimating only one heterogeneity parameter: it cannot distinguish within-study from between-study variation,
 - potentially biased inference: ignoring dependencies can lead to overconfident conclusions.

For the Johnides et al. data, the single-level RoBMA finds strong evidence for the absence of an effect, while the multilevel model provides only weak evidence against it. 
Properly accounting for data structure leads to more conservative and appropriate inferences.


### References

---
title: "Fitting Custom Meta-Analytic Ensembles"
author: "František Bartoš"
date: "`r Sys.Date()`"
output:
  rmarkdown::html_vignette:
    self_contained: yes
bibliography: ../inst/REFERENCES.bib
csl: ../inst/apa.csl
vignette: >
  %\VignetteIndexEntry{Fitting Custom Meta-Analytic Ensembles}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
  %\VignetteEngine{knitr::rmarkdown_notangle}
---

```{r setup, include = FALSE}
is_check <- ("CheckExEnv" %in% search()) ||
             any(c("_R_CHECK_TIMINGS_", "_R_CHECK_LICENSE_") %in% names(Sys.getenv())) ||
             !file.exists("../models/CustomEnsembles/Bem_update1.RDS") 
knitr::opts_chunk$set(
  collapse = TRUE,
  comment  = "#>",
  eval     = !is_check,
  dev      = "png"
)
if(.Platform$OS.type == "windows"){
  knitr::opts_chunk$set(dev.args = list(type = "cairo"))
}
```

```{r include = FALSE, eval = FALSE}
# R package version updating
library(RoBMA)

data("Bem2011", package = "RoBMA")
fit <- RoBMA(d = Bem2011$d, se = Bem2011$se, study_names = Bem2011$study,
             priors_effect = NULL, priors_heterogeneity = NULL, priors_bias = NULL,
             priors_effect_null        = prior("spike", parameters = list(location = 0)),
             priors_heterogeneity_null = prior("spike", parameters = list(location = 0)),
             priors_bias_null          = prior_none(),
             seed = 1)

fit <- update(fit,
              prior_effect             = prior("normal", parameters = list(mean = .15, sd = .10), truncation = list(lower = 0)),
              prior_heterogeneity_null = prior("spike",  parameters = list(location = 0)),
              prior_bias_null          = prior_none())
saveRDS(fit, file = "../models/CustomEnsembles/Bem_update1.RDS")

fit <- update(fit,
              prior_effect        = prior("normal", parameters = list(mean = .15, sd = .10), truncation = list(lower = 0)),
              prior_heterogeneity = prior("invgamma", parameters = list(shape = 1, scale = .15)),
              prior_bias_null     = prior_none())
fit <- update(fit,
              prior_effect_null        = prior("spike",     parameters = list(location = 0)),
              prior_heterogeneity_null = prior("spike",     parameters = list(location = 0)),
              prior_bias               = prior_weightfunction("one.sided", parameters = list(alpha = c(1, 1), steps = c(0.05))))
fit <- update(fit,
              prior_effect_null        = prior("spike",     parameters = list(location = 0)),
              prior_heterogeneity_null = prior("spike",     parameters = list(location = 0)),
              prior_bias               = prior_weightfunction("one.sided", parameters = list(alpha = c(1, 1, 1), steps = c(0.05, 0.10))))
fit <- update(fit,
              prior_effect_null        = prior("spike",     parameters = list(location = 0)),
              prior_heterogeneity_null = prior("spike",     parameters = list(location = 0)),
              prior_bias               = prior_PET("Cauchy", parameters = list(0, 1),  truncation = list(lower = 0)))
fit <- update(fit,
              prior_effect_null        = prior("spike",     parameters = list(location = 0)),
              prior_heterogeneity_null = prior("spike",     parameters = list(location = 0)),
              prior_bias               = prior_PEESE("Cauchy", parameters = list(0, 5),  truncation = list(lower = 0)))
saveRDS(fit, file = "../models/CustomEnsembles/Bem_update2.RDS")
```

This vignette provides a step-by-step guide to fitting custom meta-analytic ensembles using the RoBMA R package. By the end of this guide, you will be able to construct and evaluate custom meta-analytic models.

By default, the `RoBMA()` function specifies models as a combination of all supplied prior distributions (across null and alternative specification), with their prior model weights being equal to the product of prior distributions' weights. This results in the 36 meta-analytic models using the default settings [@bartos2021no]$^1$. In [another vignette](ReproducingBMA.html), we illustrated that RoBMA can be also utilized for reproducing Bayesian Model-Averaged Meta-Analysis (BMA) [@gronau2017bayesian; @gronau2020primer; @bartos2021bayesian]. However, the package was built as a framework for estimating highly customized meta-analytic model ensembles. Here, we are going to illustrate how to do exactly that (see @bartos2020adjusting for a tutorial paper on customizing the model ensemble with JASP). 

Please keep in mind that all models should be justified by theory. Furthermore, the models should be tested to make sure that the ensemble can perform as intended a priori to drawing inference from it. The following sections are only for illustrating the functionality of the package. We provide a complete discussion with the relevant sources in the Example section of @bartos2021no.

### The Dataset

To illustrate the custom model building procedure, we use data from the infamous @bem2011feeling "Feeling the future" precognition study. We use coding of the results as summarized by Bem in one of his later replies [@bem2011must]. 

```{r}
library(RoBMA)

data("Bem2011", package = "RoBMA")
Bem2011
```

### The Custom Ensemble

We consider the following scenarios as plausible explanations for the data, and decide to include only those models into the meta-analytic ensemble:

  1) there is absolutely no precognition effect - a fixed effects model assuming the effect size to be zero ($H_{0}^f$),
  2) the experiments measured the same underlying precognition effect - a fixed effects model ($H_{1}^f$),
  3) each of the experiments measured a slightly different precognition effect - a random effects model ($H_{1}^r$),
  4) there is absolutely no precognition effect and the results can be explained by publication bias, modeled with one of the following publication bias adjustments:
    - 4.1) one-sided selection operating on significant *p*-values ($H_{1,\text{pb1}}^f$),
    - 4.2) one-sided selection operating on significant and marginally significant *p*-values ($H_{1,\text{pb2}}^f$),
    - 4.3) PET correction for publication bias which adjusts for the relationship between effect sizes and standard errors ($H_{1,\text{pb3}}^f$),
    - 4.4) PEESE correction for publication bias which adjusts for the relationship between effect sizes and standard errors squared ($H_{1,\text{pb4}}^f$).
  
If we were to fit the ensemble using the `RoBMA()` function and specifying all of the priors, we would have ended with 2 (effect or no effect) * 2 (heterogeneity or no heterogeneity) * 5 (no publication bias or 4 ways of adjusting for publication bias) = 20 models. That is 13 models more than requested. Furthermore, we could not specify different parameters for the prior distributions for each model. The following process allows this, though we do not utilize it here.

We start with fitting only the first model using the `RoBMA()` function and we will continuously update the fitted object to include all of the models.

#### Model 1
We initiate the model ensemble by specifying only the first model with the `RoBMA()` function. We explicitly specify prior distributions for all components and set the prior distributions to correspond to the null hypotheses and set the seed to ensure reproducibility of the results.

```{r}
fit <- RoBMA(d = Bem2011$d, se = Bem2011$se, study_names = Bem2011$study,
             priors_effect = NULL, priors_heterogeneity = NULL, priors_bias = NULL,
             priors_effect_null        = prior("spike", parameters = list(location = 0)),
             priors_heterogeneity_null = prior("spike", parameters = list(location = 0)),
             priors_bias_null          = prior_none(),
             seed = 1)
```

We verify that the ensemble contains only the single specified model with the `summary()` function by setting `type = "models"`.

```{r}
summary(fit, type = "models")
```

#### Model 2
Before we add the second model to the ensemble, we need to decide on the prior distribution for the mean parameter. If precognition were to exist, the effect would be small since all casinos would be bankrupted otherwise. The effect would also be positive, since any deviation from randomness could be characterized as an effect. Therefore, we decide to use a normal distribution with mean = 0.15, standard deviation 0.10, and truncated to the positive range. This sets the prior density around small effect sizes. To get a better grasp of the prior distribution, we visualize it using the `plot())` function (the figure can also be created using the ggplot2 package by adding `plot_type = "ggplot"` argument).

```{r fig_mu_prior, dpi = 300, fig.width = 4, fig.height = 3, out.width = "50%", fig.align = "center"}
plot(prior("normal", parameters = list(mean = .15, sd = .10), truncation = list(lower = 0)))
```

We add the second model to the ensemble using the `update.RoBMA()` function. The function can also be used for many other purposes - updating settings, prior model weights, and refitting failed models. Here, we supply the fitted ensemble object and add an argument specifying the prior distributions of each component for the additional model. Since we want to add Model 2 - we set the prior for the $\mu$ parameter to be treated as a prior belonging to the alternative hypothesis of the effect size component and the remaining priors treated as belonging to the null hypotheses. If we wanted, we could also specify `prior_weights` argument, to change the prior probability of the fitted model but we do not utilize this option here and keep the default value, which sets the prior weights for the new model to `1`. (Note that the arguments for specifying prior distributions in `update.RoBMA()` function are `prior_X` - in singular, in comparison to `RoBMA()` function that uses `priors_X` in plural.)

```{r include = FALSE}
# these fits are relatively fast, but we reduce the knitting time considerably
fit <- readRDS(file = "../models/CustomEnsembles/Bem_update1.RDS")
```
```r
fit <- update(fit,
              prior_effect             = prior("normal", parameters = list(mean = .15, sd = .10), truncation = list(lower = 0)),
              prior_heterogeneity_null = prior("spike",  parameters = list(location = 0)),
              prior_bias_null          = prior_none())
```

We can again inspect the updated ensemble to verify that it contains both models. We see that Model 2 notably outperformed the first model and attained all of the posterior model probability.

```{r}
summary(fit, type = "models")
```

#### Models 3-4.4

Before we add the remaining models to the ensemble using the `update()` function, we need to decide on the remaining prior distributions. Specifically, on the prior distribution for the heterogeneity parameter $\tau$, and the publication bias adjustment parameters $\omega$ (for the selection models' weightfunctions) and PET and PEESE for the PET and PEESE adjustment.

For Model 3, we use the usual inverse-gamma(1, .15) prior distribution based on empirical heterogeneity estimates [@erp2017estimates] for the heterogeneity parameter $\tau$. For Models 4.1-4.4 we use the default settings for the publication bias adjustments as outlined in the Appendix B of [@bartos2021no].

Now, we just need to add the remaining models to the ensemble using the `update()` function as already illustrated.

```r
### adding Model 3
fit <- update(fit,
              prior_effect        = prior("normal", parameters = list(mean = .15, sd = .10), truncation = list(lower = 0)),
              prior_heterogeneity = prior("invgamma", parameters = list(shape = 1, scale = .15)),
              prior_bias_null     = prior_none())

### adding Model 4.1
fit <- update(fit,
              prior_effect_null        = prior("spike",     parameters = list(location = 0)),
              prior_heterogeneity_null = prior("spike",     parameters = list(location = 0)),
              prior_bias               = prior_weightfunction("one.sided", parameters = list(alpha = c(1, 1), steps = c(0.05))))
              
### adding Model 4.2
fit <- update(fit,
              prior_effect_null        = prior("spike",     parameters = list(location = 0)),
              prior_heterogeneity_null = prior("spike",     parameters = list(location = 0)),
              prior_bias               = prior_weightfunction("one.sided", parameters = list(alpha = c(1, 1, 1), steps = c(0.05, 0.10))))
              
### adding Model 4.3
fit <- update(fit,
              prior_effect_null        = prior("spike",     parameters = list(location = 0)),
              prior_heterogeneity_null = prior("spike",     parameters = list(location = 0)),
              prior_bias               = prior_PET("Cauchy", parameters = list(0, 1),  truncation = list(lower = 0)))
              
### adding Model 4.4
fit <- update(fit,
              prior_effect_null        = prior("spike",     parameters = list(location = 0)),
              prior_heterogeneity_null = prior("spike",     parameters = list(location = 0)),
              prior_bias               = prior_PEESE("Cauchy", parameters = list(0, 5),  truncation = list(lower = 0)))
```

We again verify that all of the requested models are included in the ensemble using the `summary())` function with `type = "models"` argument.

```{r include = FALSE}
fit <- readRDS(file = "../models/CustomEnsembles/Bem_update2.RDS")
```
```{r}
summary(fit, type = "models")
```

### Using the Fitted Ensemble

Finally, we use the `summary()` function to inspect the model results. The results from our custom ensemble indicate weak evidence for the absence of the precognition effect, $\text{BF}_{10} = 0.584$ -> $\text{BF}_{01} = 1.71$, moderate evidence for the absence of heterogeneity, $\text{BF}_{\text{rf}} = 0.132$ -> $\text{BF}_{\text{fr}} = 7.58$, and moderate evidence for the presence of the publication bias, $\text{BF}_{\text{pb}} = 3.21$. 

```{r}
summary(fit)
```

The finalized ensemble can be treated as any other `RoBMA` ensemble using the `summary()`, `plot()`, `plot_models()`, `forest()`, and `diagnostics()` functions. For example, we can use the `plot.RoBMA()` with the `parameter = "mu", prior = TRUE` arguments to plot the prior (grey) and posterior distribution (black) for the effect size. The function visualizes the model-averaged estimates across all models by default. The arrows represent the probability mass at the value 0 (a spike at 0). The secondary y-axis (right) shows the probability mass at the zero effect size, which increased from the prior probability of 0.71 to the posterior the posterior probability of 0.81.

```{r fig_mu_posterior, dpi = 300, fig.width = 4, fig.height = 3.5, out.width = "50%", fig.align = "center"}
plot(fit, parameter = "mu", prior = TRUE)
```

We can also inspect the posterior distributions of the publication bias adjustments. To visualize the model-averaged weightfunction, we set `parameter = weightfunction` argument. The resulting figure shows the light gray prior distribution and the dark gray the posterior distribution.

```{r fig_weightfunction_posterior, dpi = 300, fig.width = 5, fig.height = 4, out.width = "75%", fig.align = "center"}
plot(fit, parameter = "weightfunction", prior = TRUE)
```

We can also inspect the posterior estimate of the regression relationship between the standard errors and effect sizes by setting `parameter = "PET-PEESE"`.

```{r fig_PETPEESE_posterior, dpi = 300, fig.width = 5, fig.height = 4, out.width = "75%", fig.align = "center"}
plot(fit, parameter = "PET-PEESE", prior = TRUE)
```


### Footnotes

$^1$ - The default setting used to produce 12 models in RoBMA versions < 2, which corresponded to an earlier an article by @maier2020robust in which we applied Bayesian model-averaging only across selection models.

### References

---
title: "Introduction to RoBMA"
author: "František Bartoš"
date: "28th of April 2026"
output:
  rmarkdown::html_vignette:
    self_contained: yes
bibliography: ../inst/REFERENCES.bib
csl: ../inst/apa.csl
link-citations: true
vignette: >
  %\VignetteIndexEntry{Introduction to RoBMA}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
  %\VignetteEngine{knitr::rmarkdown_notangle}
---

```{r child = "_vignette-nowrap.md", echo = FALSE, eval = TRUE}
```

```{r setup, include = FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment  = "#>",
  eval     = FALSE
)

article_url <- function(page) {
  paste0(page, ".html")
}

reference_url <- function(page) {
  paste0("../reference/", page, ".html")
}
```

```{r child = "_v00-introduction-body.md", eval = TRUE}
```

---
title: "Multilevel Robust Bayesian Model-Averaged Meta-Regression"
author: "František Bartoš"
date: "2025"
output:
  rmarkdown::html_vignette:
    self_contained: yes
bibliography: ../inst/REFERENCES.bib
csl: ../inst/apa.csl
vignette: >
  %\VignetteIndexEntry{Multilevel Robust Bayesian Model-Averaged Meta-Regression}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r setup, include = FALSE}
is_check <- ("CheckExEnv" %in% search()) ||
              any(c("_R_CHECK_TIMINGS_", "_R_CHECK_LICENSE_") %in% names(Sys.getenv())) ||
              !file.exists("../models/MultilevelRoBMARegression/zfit_Havrankova2025.RDS")
knitr::opts_chunk$set(
  collapse  = TRUE,
  comment   = "#>",
  eval      = !is_check,
  dev       = "png")
if(.Platform$OS.type == "windows"){
  knitr::opts_chunk$set(dev.args = list(type = "cairo"))
}
```

```{r include = FALSE}
library(RoBMA)
zfit_reg <- readRDS(file = "../models/MultilevelRoBMARegression/zfit_Havrankova2025.RDS")
fit_reg  <- zfit_reg
class(fit_reg) <- class(fit_reg)[!class(fit_reg) %in% "zcurve_RoBMA"]
```

```{r include = FALSE, eval = FALSE}
# R package version updating
library(RoBMA)
data("Havrankova2025", package = "RoBMA")

# Prior scaling
fit_fe        <- metafor::rma(yi = y, sei = se, data = Havrankova2025, method = "FE")
unti_scale    <- fit_fe$se * sqrt(sum(Havrankova2025$N))
prior_scale   <- unti_scale * 0.5

df_reg  <- data.frame(
  y               = Havrankova2025$y,
  se              = Havrankova2025$se, 
  facing_customer = Havrankova2025$facing_customer, 
  study_id        = Havrankova2025$study_id
)

fit_reg <- RoBMA.reg(
  ~ facing_customer, 
  study_ids      = df_reg$study_id,
  data           = df_reg, 
  rescale_priors = prior_scale,
  prior_scale = "none", transformation = "none",
  algorithm = "ss", sample = 20000, burnin = 10000, adapt = 10000, 
  thin = 5, parallel = TRUE, autofit = FALSE, seed = 1)
)

zfit_reg    <- as_zcurve(fit_reg)
saveRDS(zfit_reg, file = "../models/MultilevelRoBMARegression/zfit_Havrankova2025.RDS", compress = "xz")
```

**This vignette accompanies the manuscript [Robust Bayesian Multilevel Meta-Analysis: Adjusting for Publication Bias in the Presence of Dependent Effect Sizes](https://doi.org/10.31234/osf.io/9tgp2_v1) preprinted at *PsyArXiv* [@bartos2025robust].**

This vignette reproduces the second example from the manuscript, re-analyzing data from @havrankova2025beauty on the relationship between beauty and professional success (1,159 effect sizes from 67 studies). 
We investigate whether the effect depends on the type of customer contact ("no", "some", or "direct"). 
For the first example describing a publication bias-adjusted multilevel meta-analysis see [Multilevel Robust Bayesian Meta-Analysis](MultilevelRoBMA.html).

### Rescaling Priors for Non-Standardized Effect Sizes

The effect sizes in this dataset are not standardized (percent increase in earnings with one-standard deviation increase in beauty). 
Therefore, we must rescale the default prior distributions to avoid specifying overly narrow or wide priors. 
Overly narrow priors would make it difficult to find evidence for the effect (as both models would predict effects very close to zero--their predictions would overlap substantially), 
while overly wide priors would bias the results towards the null hypothesis (as alternative models would predict unrealistically large effects--their predictions would never be supported by the data).

We rescale the priors by matching the default prior scale to the unit information scale of the data [@rover2021weakly; @mulder2024bayesian]. 
The unit information scale can be approximated by the standard error of a fixed-effect model multiplied by the square root of the total sample size. 
Since the default prior for standardized mean differences has a standard deviation of one (corresponding to half the unit information), we multiply the unit information scale by 0.5. 

```{r}
library(RoBMA)
data("Havrankova2025", package = "RoBMA")

fit_fe        <- metafor::rma(yi = y, sei = se, data = Havrankova2025, method = "FE")
unti_scale    <- fit_fe$se * sqrt(sum(Havrankova2025$N))
prior_scale   <- unti_scale * 0.5
```

The resulting `prior_scale`, in our example approximately 26, can be used in the `rescale_priors` argument in the fitting functions of the `RoBMA` package,

```{r, eval = FALSE}
df_reg  <- data.frame(
  y               = Havrankova2025$y,
  se              = Havrankova2025$se, 
  facing_customer = Havrankova2025$facing_customer, 
  study_id        = Havrankova2025$study_id
)

fit_reg <- RoBMA.reg(
  ~ facing_customer, 
  study_ids      = df_reg$study_id,
  data           = df_reg, 
  rescale_priors = prior_scale,
  prior_scale    = "none", transformation = "none",
  algorithm = "ss", sample = 20000, burnin = 10000, adapt = 10000, 
  thin = 5, parallel = TRUE, autofit = FALSE, seed = 1)
)
```

where we first specify the input `data.frame` (with `y` and `se` denoting the effect size and standard error inputs) and `prior_scale = "none"` and `transformation = "none"` arguments disable prior distributions for standardized effect sizes. 
We also notably increase the number of adaptation, burn-in, and sampling iterations to ensure convergence of the more complex model (and use the `algorithm = "ss"` options since the bridge-sampling algorithm is unfeasible for multilevel models).

*Note that the meta-regression model can take more than an hour to run with parallel processing enabled.*

### Interpreting the Results

The `summary` and `marginal_summary` functions provide the overall model estimates and the estimated marginal means.

```{r}
summary(fit_reg)
```

Using the rescaled prior distributions, the multilevel RoBMA meta-regression finds extreme evidence for the presence of the average effect, moderation by the degree of consumer contact, between-study heterogeneity, and publication bias (all BFs > 10^6). 
The model-averaged average effect estimate of mu = 3.0, 95% CI [1.9, 4.2], is accompanied by considerable overall heterogeneity tau = 4.5, 95% CI [3.9, 5.3], 
and within-cluster allocation close to 1, rho = 0.82, 95% CI [0.75, 0.88], indicating a high degree of similarity of estimates from the same study.

The summary function also warns us about a large R-hat. Inspection of the parameter-specific MCMC diagnostics via `summary(fit_reg, type = "diagnostics")` shows that while some R-hats are inflated, most parameters were sampled to an acceptable level. The main source of issue seems to be the moderator coefficients `facing_customer`. We visually assess the chains `diagnostics_trace(fit_reg, "facing_customer")` but do not notice substantial issues.

The average effect is moderated by the degree of consumer contact, to examine the effect at each level of the moderator we can use the `marginal_summary()` function.

```{r}
marginal_summary(fit_reg)
```

The estimated marginal means at each level of the moderator reveals moderate evidence for the absence of the effect for no consumer contact jobs mu = 0.9, 95% CI [-0.5, 2.5], BF10 = 0.11, 
and extreme evidence for jobs with some consumer contact mu = 4.1, 95% CI [2.6, 5.5], BF10 > 10^6 and jobs with direct consumer contact mu = 4.1, 95% CI [2.7, 5.5], BF10 > 10^6.

### Visual Fit Assessment

We can visualize evalate the model fit using the meta-analytic z-curve plot [@bartos2025zcurve] (also see [Z-Curve Publication Bias Diagnostics](ZCurveDiagnostics.html) vignette for more details).
The z-curve plot shows the distribution of the observed $z$-statistics (computed as the effect size / standard error), with dotted red horizontal lines highlighting the typical steps on which publication bias operates 
($z = \pm 1.65$ and $z = \pm 1.96$ corresponding to statistically significant and marginally significant $p$-values, and $z = 0$ corresponding to the selection of the direction of the effect).

```{r, eval = FALSE}
zfit_reg <- as_zcurve(fit_reg)
```
```{r, fig.width = 6, fig.height = 4}
par(mar = c(4,4,0,0))
plot(zfit_reg, by.hist = 0.25, plot_extrapolation = FALSE, from = -4, to = 8)
```

We can notice clear discontinuities corresponding to the selection on the direction of the effect and marginal significance. 
The posterior predictive density from the multilevel RoBMA meta-regression indicates that the RoBMA meta-regression model approximates the observed data reasonably well, 
capturing the discontinuities and the proportion of statistically significant results. This reassures us that the model provides a good fit to the data while adjusting for publication bias.

### References

---
title: "Reproducing Bayesian Model-Averaged Meta-Analysis"
author: "František Bartoš"
date: "`r Sys.Date()`"
output:
  rmarkdown::html_vignette:
    self_contained: yes
bibliography: ../inst/REFERENCES.bib
csl: ../inst/apa.csl
vignette: >
  %\VignetteIndexEntry{Reproducing Bayesian Model-Averaged Meta-Analysis}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
  %\VignetteEngine{knitr::rmarkdown_notangle}
---

```{r setup, include = FALSE}
is_check <- ("CheckExEnv" %in% search()) ||
              any(c("_R_CHECK_TIMINGS_", "_R_CHECK_LICENSE_") %in% names(Sys.getenv())) ||
              !file.exists("../models/ReproducingBMA/BMA_PowerPoseTest.RDS")
knitr::opts_chunk$set(
  collapse  = TRUE,
  comment   = "#>",
  eval      = !is_check,
  dev      = "png")
if(.Platform$OS.type == "windows"){
  knitr::opts_chunk$set(dev.args = list(type = "cairo"))
}
```
```{r include = FALSE}
library(RoBMA)
# we pre-load the RoBMA models, the fitting time is around 2-5 minutes
fit_BMA_test   <- readRDS(file = "../models/ReproducingBMA/BMA_PowerPoseTest.RDS")
fit_BMA_est    <- readRDS(file = "../models/ReproducingBMA/BMA_PowerPoseEst.RDS")
fit_RoBMA_test <- readRDS(file = "../models/ReproducingBMA/PowerPoseTest.RDS")
fit_RoBMA_est  <- readRDS(file = "../models/ReproducingBMA/PowerPoseEst.RDS")
```

```{r include = FALSE, eval = FALSE}
# R package version updating
library(RoBMA)

data("power_pose", package = "metaBMA")

fit_RoBMA_test <- RoBMA(d = power_pose$effectSize, se = power_pose$SE, study_names = power_pose$study,
                        priors_effect  = prior(
                          distribution = "cauchy",
                          parameters = list(location = 0, scale = 1/sqrt(2)),
                          truncation = list(0, Inf)),
                        priors_heterogeneity = prior(
                          distribution = "invgamma",
                          parameters = list(shape = 1, scale = 0.15)),
                        priors_bias = NULL,
                        transformation = "cohens_d", seed = 1, parallel = TRUE)

fit_RoBMA_est  <- RoBMA(d = power_pose$effectSize, se = power_pose$SE, study_names = power_pose$study,
                        priors_effect  = prior(
                          distribution = "cauchy",
                          parameters = list(location = 0, scale = 1/sqrt(2))),
                        priors_heterogeneity = prior(
                          distribution = "invgamma",
                          parameters = list(shape = 1, scale = 0.15)),
                        priors_bias = NULL,
                        priors_effect_null = NULL,
                        transformation = "cohens_d", seed = 2, parallel = TRUE)

saveRDS(fit_RoBMA_test, file = "../models/ReproducingBMA/PowerPoseTest.RDS")
saveRDS(fit_RoBMA_est, file = "../models/ReproducingBMA/PowerPoseEst.RDS")
```

By default, the RoBMA package estimates an ensemble of 36 meta-analytic models and provides functions for convenient manipulation of the fitted object. However, the package has been designed so it can be used as a framework for estimating any combination of meta-analytic models (or a single model). Here, we illustrate how to build a custom ensemble of meta-analytic models - specifically the same ensemble that is used in 'classical' Bayesian Model-Averaged Meta-Analysis [@gronau2017bayesian; @gronau2020primer; @bartos2021bayesian]. See [this vignette](CustomEnsembles.html) if you are interested in building more customized ensembles or @bartos2020adjusting for a tutorial on fitting (custom) models in JASP.


### Reproducing Bayesian Model-Averaged Meta-Analysis (BMA)

We illustrate how to fit a classical BMA (not adjusting for publication bias) using `RoBMA`. For this purpose, we reproduce a meta-analysis of registered reports on Power posing by @gronau2017bayesian. We focus only on the analysis of all reported results using a Cauchy prior distribution with scale $1/\sqrt{2}$ for the effect size estimation (half-Cauchy for testing) and inverse-gamma distribution with shape = 1 and scale = 0.15 for the heterogeneity parameter. You can find the figure from the original publication [here](https://www.flickr.com/photos/149473606@N08/34086113581/in/dateposted-public/) and the paper's supplementary materials at https://osf.io/fxg32/.

First, we load the power posing data provided within the metaBMA package and reproduce the analysis performed by @gronau2017bayesian.

```{r}
data("power_pose", package = "metaBMA")
power_pose[,c("study", "effectSize", "SE")]
``` 
``` r
fit_BMA_test <- metaBMA::meta_bma(y   = power_pose$effectSize, SE = power_pose$SE,
                                  d   = metaBMA::prior(family = "halfcauchy", param = 1/sqrt(2)),
                                  tau = metaBMA::prior(family = "invgamma", param = c(1, .15)))
 
fit_BMA_est  <- metaBMA::meta_bma(y   = power_pose$effectSize, SE = power_pose$SE,
                                  d   = metaBMA::prior(family = "cauchy", param = c(0, 1/sqrt(2))),
                                  tau = metaBMA::prior(family = "invgamma", param = c(1, .15)))
``` 
```{r}
fit_BMA_test$inclusion

round(fit_BMA_est$estimates,2)
```

From the output, we can see the inclusion Bayes factor for the effect size was $BF_{10} = 33.14$ and the effect size estimate 0.22, 95% HDI [0.09, 0.34], which matches the reported results. Please note that the `metaBMA` package model-averages only across the $H_{1}$ models, whereas the `RoBMA` package model-averages across all models (assuming the presence and absence of the effect).

### Using RoBMA

Now we reproduce the analysis with `RoBMA`. We set the corresponding prior distributions for effect sizes ($\mu$) and heterogeneity ($\tau$), and remove the alternative prior distributions for the publication bias by setting `priors_bias = NULL`. To specify the half-Cauchy prior distribution with the `RoBMA::prior()` function we use a regular Cauchy distribution and truncate it at zero (note that both `metaBMA` and `RoBMA` export their own `prior()` functions that will clash when loading both packages simultaneously). The inverse-gamma prior distribution for the heterogeneity parameter is the default option (we specify it for completeness). We omit the specifications for the null prior distributions for the effect size, heterogeneity (both of which are set to a spike at 0 by default), and publication bias (which is set to no publication bias by default). Note that starting from version 3.1, the package includes the `NoBMA()` function, which allows users to skip publication bias adjustment directly.

Since `metaBMA` model-averages the effect size estimates only across the models assuming presence of the effect, we remove the models assuming absence of the effect from the estimation ensemble with `priors_effect_null = NULL`. Finally, we set `transformation = "cohens_d"` to estimate the models on Cohen's *d* scale. RoBMA uses Fisher's *z* scale by default and transforms the estimated coefficients back to the scale that is used for specifying the prior distributions. We speed up the computation by setting `parallel = TRUE`, and set a seed for reproducibility.

``` r
library(RoBMA)

fit_RoBMA_test <- RoBMA(d = power_pose$effectSize, se = power_pose$SE, study_names = power_pose$study,
                        priors_effect  = prior(
                          distribution = "cauchy",
                          parameters = list(location = 0, scale = 1/sqrt(2)),
                          truncation = list(0, Inf)),
                        priors_heterogeneity = prior(
                          distribution = "invgamma",
                          parameters = list(shape = 1, scale = 0.15)),
                        priors_bias = NULL,
                        transformation = "cohens_d", seed = 1, parallel = TRUE)

fit_RoBMA_est  <- RoBMA(d = power_pose$effectSize, se = power_pose$SE, study_names = power_pose$study,
                        priors_effect  = prior(
                          distribution = "cauchy",
                          parameters = list(location = 0, scale = 1/sqrt(2))),
                        priors_heterogeneity = prior(
                          distribution = "invgamma",
                          parameters = list(shape = 1, scale = 0.15)),
                        priors_bias = NULL,
                        priors_effect_null = NULL,
                        transformation = "cohens_d", seed = 2, parallel = TRUE)
```
```{r}
summary(fit_RoBMA_test)

summary(fit_RoBMA_est)
```


The output from the `summary.RoBMA()` function has 2 parts. The first one under the "Robust Bayesian Meta-Analysis" heading provides a basic summary of the fitted models by component types (presence of the Effect and Heterogeneity). The table summarizes the prior and posterior probabilities and the inclusion Bayes factors of the individual components. The results for the half-Cauchy model specified for testing show that the inclusion BF is nearly identical to the one computed by the `metaBMA` package, $\text{BF}_{10} = 33.11$. 

The second part under the 'Model-averaged estimates' heading displays the parameter estimates. The results for the unrestricted Cauchy model specified for estimation show the effect size estimate $\mu = 0.22$, 95% CI [0.10, 0.35] that also mirrors the one obtained from `metaBMA` package.


### Visualizing the Results

RoBMA provides extensive options for visualizing the results. Here, we visualize the prior (grey) and posterior (black) distribution for the mean parameter.

```{r fig_mu_est, dpi = 300, fig.width = 4, fig.height = 4, out.width = "50%", fig.align = "center"}
plot(fit_RoBMA_est, parameter = "mu", prior = TRUE, xlim = c(-1, 1))
```

If we visualize the effect size from the model specified for testing, we notice a few more things. The function plots the model-averaged estimates across all models by default, including models assuming the absence of the effect. The arrows represents the probability of a spike, here, at the value 0. The secondary y-axis (right) shows the probability of the value 0 decreased from 0.50, to 0.03 (also obtainable from the "Robust Bayesian Meta-Analysis" field in the `summary.RoBMA()` function). Furthermore, the continuous prior distributions for the effect size under the alternative hypothesis are truncated to only positive values, reflecting the assumption that the effect size cannot be negative. 
 
```{r fig_mu_test, dpi = 300, fig.width = 4, fig.height = 4, out.width = "50%", fig.align = "center"}
plot(fit_RoBMA_test, parameter = "mu", prior = TRUE, xlim = c(-.5, 1))
```


We can also visualize the estimates from the individual models used in the ensemble. We do that with the `plot_models()` function, which visualizes the effect size estimates and 95% CI of each of the specified models from the estimation ensemble (Model 1 corresponds to the fixed effect model and Model 2 to the random effect model). The size of the square representing the mean estimate reflects the posterior model probability of the model, which is also displayed in the right-hand side panel. The bottom part of the figure shows the model-averaged estimate that is a combination of the individual model posterior distributions weighted by the posterior model probabilities.

```{r fig_models, dpi = 300, fig.height = 4.5, fig.width = 7, out.width = '75%', fig.align = "center"}
plot_models(fit_RoBMA_est)
```

The last type of visualization that we show here is the forest plot. It displays both the effect sizes from the original studies and the overall meta-analytic estimate in a single figure. It can be requested by using the `forest()` function.

```{r fig_forest, dpi = 300, fig.height = 4.5, fig.width = 7, out.width = '75%', fig.align = "center"}
forest(fit_RoBMA_est)
```

For more options provided by the plotting function, see its documentation using `?plot.RoBMA()`, `?plot_models()`, and `?forest()`.

### References

---
title: "Generalized Linear Mixed-Effects Meta-Analysis"
author: "František Bartoš"
date: "28th of April 2026"
output:
  rmarkdown::html_vignette:
    self_contained: yes
bibliography: ../inst/REFERENCES.bib
csl: ../inst/apa.csl
link-citations: true
vignette: >
  %\VignetteIndexEntry{Generalized Linear Mixed-Effects Meta-Analysis}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
  %\VignetteEngine{knitr::rmarkdown_notangle}
---

```{r child = "_vignette-nowrap.md", echo = FALSE, eval = TRUE}
```

```{r setup, include = FALSE}
source("vignette-cache.R", local = knitr::knit_global())

cached_fits <- vignette_cache(
  name     = "v13-metafor-parity-glmm",
  objects  = c("fit1_brma"),
  packages = c("metafor", "metadat")
)

knitr::opts_chunk$set(
  collapse    = TRUE,
  comment     = "#>",
  eval        = vignette_cache_eval(cached_fits),
  message     = FALSE,
  warning     = FALSE,
  fig.width   = 7,
  fig.height  = 3.5,
  dev         = "png",
  fig.retina  = 3
)
if (.Platform$OS.type == "windows") {
  knitr::opts_chunk$set(dev.args = list(type = "cairo"))
}
```

```{r load-models, include = FALSE}
library("metafor")
library("RoBMA")

data("dat.bcg", package = "metadat")

vignette_cache_load(cached_fits)
```

```{r precompute-models, include = FALSE, eval = FALSE}
library("metafor")
library("RoBMA")

data("dat.bcg", package = "metadat")

fit1_brma <- brma.glmm(
  ai = tpos, bi = tneg, ci = cpos, di = cneg, measure = "OR",
  data = dat.bcg, seed = 1
)

vignette_cache_save(cached_fits)
```

This vignette illustrates how to use the `RoBMA` R package [@RoBMA] when the outcome is a count rather than an effect-size estimate.
We walk through the BCG vaccine dataset from the `metafor` package [@metafor], showing the `brma.glmm()` call alongside its `metafor::rma.glmm()` counterpart for log odds ratios.
The interface mirrors `metafor::rma.glmm()`: counts (`ai`, `bi`, `ci`, `di`) are passed in directly, and no `escalc()` step is needed.

All `brma.glmm()` calls keep the default prior distributions; see the [*Prior Distributions*](v01-prior-distributions.html) vignette for the prior definitions and customization options.
For the non-GLMM workflow showing more features (also applicable to GLMM models) see the [*Bayesian Meta-Analysis*](v02-bayesian-meta-analysis.html) vignette.

`brma.glmm()` currently supports binomial outcomes (log odds ratio, `measure = "OR"`) and Poisson outcomes (log incidence rate ratio, `measure = "IRR"`).

## Binomial Random-Effects Model

The BCG vaccine dataset from the `metadat` package contains 13 randomized trials on tuberculosis prevention.
Treated and control event counts (`tpos`, `tneg`, `cpos`, `cneg`) are passed in directly.

```{r data}
data("dat.bcg", package = "metadat")
head(dat.bcg)
```

`metafor::rma.glmm()` fits the random-effects logistic model.
We use `model = "UM.FS"` to obtain the unconditional model with fixed study-specific intercepts, the formulation that corresponds to the binomial GLMM in `brma.glmm()` (Model 4 in @jackson2018comparison).

```{r fit1-metafor}
fit1_metafor <- metafor::rma.glmm(
  ai = tpos, bi = tneg, ci = cpos, di = cneg, measure = "OR", 
  model = "UM.FS", data = dat.bcg
)
fit1_metafor
```

The matching Bayesian fit takes the same count columns and `measure = "OR"`.

```{r fit1-brma, eval = FALSE}
fit1_brma <- brma.glmm(
  ai = tpos, bi = tneg, ci = cpos, di = cneg, measure = "OR",
  data = dat.bcg, seed = 1
)
```

`summary()` reports posterior means, credible intervals, and (suppressed here) MCMC convergence diagnostics for the pooled log odds ratio `mu` and the heterogeneity `tau`.

```{r fit1-brma-summary}
summary(fit1_brma, include_mcmc_diagnostics = FALSE)
```

The posterior mean for `mu` tracks the maximum-likelihood estimate from the `metafor` package, with comparable heterogeneity `tau`.
`pooled_effect()` returns the same summary backtransformed to the odds-ratio scale via `transform = "EXP"`:

```{r fit1-brma-pooled}
pooled_effect(fit1_brma, transform = "EXP")
```

## Default Prior Distributions and the Auxiliary Parameter

The default prior distributions on `mu` and `tau` follow the unit-information set-up as elsewhere in the package, scaled by the known unit information standard deviation for the chosen `measure` (`sqrt(4)` for `OR` and `IRR`).
Beyond `mu` and `tau`, GLMM models include estimate-specific nuisance parameters for the per-study baseline level.
For binomial outcomes (`measure = "OR"`), `logit(pi[i])` is the midpoint of the two arm logits before applying half of the study-specific log odds ratio to each arm.
The default `prior_baserate = Beta(1, 1)` is placed independently on each `pi[i]`.
For Poisson outcomes (`measure = "IRR"`), `phi[i]` is the midpoint of the two arm log incidence rates before applying half of the study-specific log incidence rate ratio to each arm.
Because this midpoint log-rate is unbounded, there is no bounded flat default analogous to `Beta(1, 1)`, so the default `prior_lograte` is a proper data-based unit-information normal used independently for each `phi[i]`.
Thus the default nuisance-parameter prior distributions are exchangeable in the iid sense, but they are not hierarchical prior distributions with a learned common baseline-rate distribution.
Both defaults can be overridden via the corresponding `prior_*` argument; see the [*Prior Distributions*](v01-prior-distributions.html) vignette.
The Poisson model corresponds to @bagos2009mixed and uses event counts and exposure times (`x1i`, `t1i`, `x2i`, `t2i`) in place of the four-way count layout.

## Other Inference Helpers

All inference helpers available for any other `brma()` fit work the same way for a `brma.glmm()` fit.
Meta-regression via `mods`, multilevel structures via `cluster`, location-scale models via `scale`, posterior summaries, plots, predictions, residuals, influence, and LOO comparisons all carry over.
See the [*Bayesian Meta-Analysis*](v02-bayesian-meta-analysis.html) vignette for the full walkthrough and the [*Multilevel Meta-Analysis*](v10-metafor-parity-multilevel.html) vignette for the multilevel extension; the only thing that changes here is the count-based input and the auxiliary baseline-rate prior.


## References

---
title: "Robust Bayesian Model-Averaged Meta-Regression"
author: "František Bartoš"
date: "`r Sys.Date()`"
output:
  rmarkdown::html_vignette:
    self_contained: yes
bibliography: ../inst/REFERENCES.bib
csl: ../inst/apa.csl
vignette: >
  %\VignetteIndexEntry{Robust Bayesian Model-Averaged Meta-Regression}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
  %\VignetteEngine{knitr::rmarkdown_notangle}
---

```{r setup, include = FALSE}
is_check <- ("CheckExEnv" %in% search()) ||
              any(c("_R_CHECK_TIMINGS_", "_R_CHECK_LICENSE_") %in% names(Sys.getenv())) ||
              !file.exists("../models/MetaRegression/fit_BMA.RDS")
knitr::opts_chunk$set(
  collapse  = TRUE,
  comment   = "#>",
  eval      = !is_check,
  dev      = "png")
if(.Platform$OS.type == "windows"){
  knitr::opts_chunk$set(dev.args = list(type = "cairo"))
}
```
```{r include = FALSE}
library(RoBMA)
# we pre-load the RoBMA models, the fitting time is around 2-5 minutes
fit_BMA   <- readRDS(file = "../models/MetaRegression/fit_BMA.RDS")
fit_RoBMA <- readRDS(file = "../models/MetaRegression/fit_RoBMA.RDS")
```

```{r include = FALSE, eval = FALSE}
library(RoBMA)

fit_BMA    <- NoBMA.reg(~ measure + age, data = Andrews2021, parallel = TRUE, seed = 1, save = "min")
fit_RoBMA  <- RoBMA.reg(~ measure + age, data = Andrews2021, parallel = TRUE, seed = 1, chains = 1, save = "min")

saveRDS(fit_BMA,   file = "../models/MetaRegression/fit_BMA.RDS",   compress = "xz")
saveRDS(fit_RoBMA, file = "../models/MetaRegression/fit_RoBMA.RDS", compress = "xz")
```

Robust Bayesian model-averaged meta-regression (RoBMA-reg) extends the robust Bayesian model-averaged meta-analysis (RoBMA) by including covariates in the meta-analytic model. RoBMA-reg allows for estimating and testing the moderating effects of study-level covariates on the meta-analytic effect in a unified framework (e.g., accounting for uncertainty in the presence vs. absence of the effect, heterogeneity, and publication bias). This vignette illustrates how to fit a robust Bayesian model-averaged meta-regression using the `RoBMA` R package. We reproduce the example from @bartos2023robust, who re-analyzed a meta-analysis of the effect of household chaos on child executive functions with the mean age and assessment type covariates based on @andrews2021examining's meta-analysis. 

First, we fit a frequentist meta-regression using the `metafor` R package. Second, we explain the Bayesian meta-regression model specification, the default prior distributions for continuous and categorical moderators, and standardized effect sizes input specification. Third, we estimate Bayesian model-averaged meta-regression (without publication bias adjustment). Finally, we estimate the complete robust Bayesian model-averaged meta-regression.

### Data

We start by loading the `Andrews2021` dataset included in the `RoBMA` R package, which contains 36 estimates of the effect of household chaos on child executive functions with the mean age and assessment type covariates. The dataset includes correlation coefficients (`r`), standard errors of the correlation coefficients (`se`), the type of executive function assessment (`measure`), and the mean age of the children (`age`) in each study.

```{r}
library(RoBMA)
data("Andrews2021", package = "RoBMA")
head(Andrews2021)
```

### Frequentist Meta-Regression

We start by fitting a frequentist meta-regression using the `metafor` R package [@metafor]. While @andrews2021examining estimated univariate meta-regressions for each moderator, we directly proceed by analyzing both moderators simultaneously. For consistency with original reporting, we estimate the meta-regression using the correlation coefficients and the standard errors provided by [@andrews2021examining]; however, note that Fisher's z transformation is recommended for estimating meta-analytic models (e.g., @stanley2024correcting).

```{r}
fit_rma <- metafor::rma(yi = r, sei = se, mods = ~ measure + age, data = Andrews2021)
fit_rma
```

The results reveal a statistically significant moderation effect of the executive function assessment type on the effect of household chaos on child executive functions ($p = 0.0099$). To explore the moderation effect further, we estimate the estimated marginal means for the executive function assessment type using the `emmeans` R package [@emmeans].

```{r}
emmeans::emmeans(metafor::emmprep(fit_rma), specs = "measure")
```

Studies using the informant-completed questionnaires show a stronger effect of household chaos on child executive functions, r = 0.229, 95\% CI [0.161, 0.297], than the direct assessment, r = 0.109, 95\% CI [0.049, 0.169]; both types of studies show statistically significant effects.

The mean age of the children does not significantly moderate the effect ($p = 0.627$) with the estimated regression coefficient of b = 0.003, 95\% CI [-0.009, 0.015]. As usual, frequentist inference limits us to failing to reject the null hypothesis. Here, we try to overcome this limitation with Bayesian model-averaged meta-regression.


### Bayesian Meta-Regression Specification

Before we proceed with the Bayesian model-averaged meta-regression, we provide a quick overview of the regression model specification. In contrast to frequentist meta-regression, we need to specify prior distributions on the regression coefficients, which encode the tested hypotheses about the presence vs. absence of the moderation (specifying different prior distributions corresponds to different hypotheses and results in different conclusions). Importantly, the treatment of continuous and categorical covariates differs in the Bayesian model-averaged meta-regression.

#### Continuous vs. Categorical Moderators and Default Prior Distributions

The default prior distribution for continuous moderators is a normal prior distribution with mean of 0 and a standard deviation of 1/4. In other words, the default prior distribution assumes that the effect of the moderator is small and smaller moderation effects are more likely than larger effects. The default choice for continuous moderators can be overridden by the `prior_covariates` argument (for all continuous covariates) or by the `priors` argument (for specific covariates, see `?RoBMA.reg` for more information). The package automatically standardizes the continuous moderators. This achieves scale-invariance of the specified prior distributions and ensures that the prior distribution for the intercept correspond to the grand mean effect. This setting can be overridden by specifying the `standardize_predictors = FALSE` argument. 

The default prior distribution for the categorical moderators is a normal distribution with a mean of 0 and a standard deviation of 1/4, representing the deviation of each level from the grand mean effect. The package uses standardized orthonormal contrasts (`contrast = "meandif"`) to model deviations of each category from the grand mean effect. The default choice for categorical moderators can be overridden by the `prior_factors` argument (for all categorical covariates) or by the `priors` argument (for specific covariates, see `?RoBMA.reg` for more information). The `"meandif"` contrasts achieve label invariance (i.e., the coding of the categorical covariates does not affect the results) and the prior distribution for the intercept corresponds to the grand mean effect. Alternatively, the package also allows specifying `"treatment"` contrasts, which result in a prior distribution on the difference between the default level and the remaining levels of the categorical covariate (with the intercept corresponding to the effect in the default factor level).

#### Effect Size Input Specification

Prior distributions for Bayesian meta-analyses are calibrated for the standardized effect size measures. As such, the fitting function needs to know what kind of effect size was supplied as the input. In `RoBMA()` function, this is achieved by the `d`, `r`, `logOR`, `OR`, `z`, `se`, `v`, `n`, `lCI`, and `uCI` arguments. The input is passed to the `combine_data()` function in the background that combines the effect sizes and merges them into a single data.frame. The `RoBMA.reg()` (and `NoBMA.reg()`) function requires the dataset to be passed as a data.frame (without missing values) with column names identifying the
- moderators passed using the formula interface (i.e., `~ measure + age` in our example)
- and the effect sizes and standard errors (i.e., `r` and `se` in our example).

As such, it is crucial for the column names to correctly identify the standardized effect sizes, standard errors, sample sizes, and moderators.

### Bayesian Model-Averaged Meta-Regression

We fit the Bayesian model-averaged meta-regression using the `NoBMA.reg()` function (the `NoBMA.reg()` function is a wrapper around the `RoBMA.reg()` function that automatically removes models adjusting for publication bias). We specify the model formula with the `~` operator similarly to the `rma()` function and pass the dataset as a data.frame with named columns as outlined in the section above (the names need to identify the moderators and effect size measures). We also set the `parallel = TRUE` argument to speed up the computation by running the chains in parallel and `seed = 1` argument to ensure reproducibility.

```r
fit_BMA <- NoBMA.reg(~ measure + age, data = Andrews2021, parallel = TRUE, seed = 1)
```

Note that the `NoBMA.reg()` function specifies the combination of all models assuming presence vs. absence of the effect, heterogeneity, moderation by `measure`, and moderation by `age`, which corresponds to $2*2*2*2=16$ models. Including each additional moderator doubles the number of models, leading to an exponential increase in model count and significantly longer fitting times.

Once the ensemble is estimated, we can use the `summary()` functions with the `output_scale = "r"` argument, which produces meta-analytic estimates that are transformed to the correlation scale.

```{r}
summary(fit_BMA, output_scale = "r")
```

The summary function produces output with multiple sections The first section contains the `Components summary` with the hypothesis test results for the overall effect size and heterogeneity. We find overwhelming evidence for both with inclusion Bayes factors (`Inclusion BF`) above 10,000.

The second section contains the `Meta-regression components summary` with the hypothesis test results for the moderators. We find moderate evidence for the moderation by the executive function assessment type, $\text{BF}_{\text{measure}} = 4.75$. Furthermore, we find moderate evidence for the null hypothesis of no moderation by mean age of the children, $\text{BF}_{\text{age}} = 0.245$ (i.e., BF for the null is $1/0.245 = 4.08$). These findings extend the frequentist meta-regression by disentangling the absence of evidence from the evidence of absence.

The third section contains the `Model-averaged estimates` with the model-averaged estimates for mean effect $\rho = 0.16$, 95% CI [0.12, 0.21] and between-study heterogeneity $\tau_{\text{Fisher's z}} = 0.12$, 95% CI [0.09, 0.17].

The fourth section contains the `Model-averaged meta-regression estimates` with the model-averaged regression coefficient estimates. The main difference from the usual frequentist meta-regression output is that the categorical predictors are summarized as a difference from the grand mean for each factor level. Here, the `intercept` regression coefficient estimate corresponds to the grand mean effect and the `measure [dif: direct]` regression coefficient estimate of -0.047, 95\% CI [-0.099, 0.000] corresponds to the difference between the direct assessment and the grand mean. As such, the results suggest that the effect size in studies using direct assessment is lower in comparison to the grand mean of the studies. The `age` regression coefficient estimate is standardized, therefore, the increase of 0.003, 95\% CI [-0.011, 0.043] corresponds to the increase in the  mean effect when increasing mean age of children by one standard deviation.

Similarly to the frequentist meta-regression, we can use the `marginal_summary()` function to obtain the marginal estimates for each of the factor levels.
```{r}
marginal_summary(fit_BMA, output_scale = "r")
```
The estimated marginal means are similar to the frequentist results. Studies using the informant-completed questionnaires again show a stronger effect of household chaos on child executive functions, $\rho = 0.208$, 95\% CI [0.130, 0.280], than the direct assessment, $\rho = 0.117$, 95\% CI [0.052, 0.185]. 

The last column summarizes results from a test against a null hypothesis of marginal means equals 0. Here, we find very strong evidence for the effect size of studies using the informant-completed questionnaires differing from zero, $\text{BF}_{10} = 62.8$ and extreme evidence for the effect size of studies using the direct assessment differing from zero, $\text{BF}_{10} = \infty$. The test is performed using the change from prior to posterior distribution at 0 (i.e., the Savage-Dickey density ratio) assuming the presence of the overall effect or the presence of difference according to the tested factor. Because the tests use prior and posterior samples, calculating the Bayes factor can be problematic when the posterior distribution is far from the tested value. In such cases, warning messages are printed and $\text{BF}_{10} = \infty$ returned (like here)---while the actual Bayes factor is less than infinity, it is still too large to be computed precisely given the posterior samples.

The full model-averaged posterior marginal means distribution can be visualized by the `marginal_plot()` function.
```{r fig_BMA, dpi = 300, fig.width = 6, fig.height = 4, out.width = "75%", fig.align = "center"}
marginal_plot(fit_BMA, parameter = "measure", output_scale = "r", lwd = 2)
```

### Robust Bayesian Model-Averaged Meta-Regression

Finally, we adjust the Bayesian model-averaged meta-regression model by fitting the robust Bayesian model-averaged meta-regression. In contrast to the previous publication bias unadjusted model ensemble, RoBMA-reg extends the model ensemble by the publication bias component specified via 6 weight functions and PET-PEESE [@bartos2021no]. We use the `RoBMA.reg()` function with the same arguments as in the previous section. The estimation time further increases as the ensemble now contains 144 models.

```r
fit_RoBMA <- RoBMA.reg(~ measure + age, data = Andrews2021, parallel = TRUE, seed = 1)
```
```{r}
summary(fit_RoBMA, output_scale = "r")
```
All previously described functions for manipulating the fitted model work identically with the publication bias adjusted model. As such, we just briefly mention the main differences found after adjusting for publication bias.

RoBMA-reg reveals strong evidence of publication bias $\text{BF}_{\text{pb}} = 28.0$. Furthermore, accounting for publication bias turns the previously found evidence for the overall effect into a weak evidence against the effect $\text{BF}_{10} = 0.51$ and notably reduces the mean effect estimate $\rho = 0.032$, 95\% CI [0.000, 0.164].

```{r}
marginal_summary(fit_RoBMA, output_scale = "r")
```

The estimated marginal means now suggest that studies using the informant-completed questionnaires show a much smaller effect of household chaos on child executive functions, $\rho = 0.094$, 95\% CI [0.000, 0.222] with only moderate evidence against no effect, $\text{BF}_{10} = 8.23$, while studies using direct assessment even provide weak evidence against the effect of household chaos on child executive functions, $\text{BF}_{10} = 0.71$, with most likely effect sizes around zero, $\rho = -0.031$, 95\% CI [-0.105, 0.120].

A visual summary of the estimated marginal means highlights the considerably wider model-averaged posterior distributions of the marginal means---a consequence of accounting and adjusting for publication bias.

```{r fig_RoBMA, dpi = 300, fig.width = 6, fig.height = 4, out.width = "75%", fig.align = "center"}
marginal_plot(fit_RoBMA, parameter = "measure", output_scale = "r", lwd = 2)
```

The Bayesian model-averaged meta-regression models are compatible with the remaining custom specification, visualization, and summary functions included in the `RoBMA` R package, highlighted in other vignettes. E.g., custom model specification is demonstrated in the vignette [Fitting Custom Meta-Analytic Ensembles](CustomEnsembles.html) and visualizations and summaries are demonstrated in the [Reproducing BMA](ReproducingBMA.html) and [Informed Bayesian Model-Averaged Meta-Analysis in Medicine](MedicineBMA.html) vignettes.

### References

---
title: "Robust Bayesian Meta-Analysis"
author: "František Bartoš"
date: "4th of May 2026"
output:
  rmarkdown::html_vignette:
    self_contained: yes
bibliography: ../inst/REFERENCES.bib
csl: ../inst/apa.csl
link-citations: true
vignette: >
  %\VignetteIndexEntry{Robust Bayesian Meta-Analysis}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
  %\VignetteEngine{knitr::rmarkdown_notangle}
---

```{r child = "_vignette-nowrap.md", echo = FALSE, eval = TRUE}
```

```{r setup, include = FALSE}
source("vignette-cache.R", local = knitr::knit_global())

cached_fits <- vignette_cache(
  name     = "v21-robust-bayesian-meta-analysis",
  objects  = c("fit_RoBMA", "fit_RoBMA_custom"),
  packages = c("metadat")
)

knitr::opts_chunk$set(
  collapse    = TRUE,
  comment     = "#>",
  eval        = vignette_cache_eval(cached_fits),
  message     = FALSE,
  warning     = FALSE,
  fig.width   = 7,
  fig.height  = 3.5,
  dev         = "png",
  fig.retina  = 3
)
if (.Platform$OS.type == "windows") {
  knitr::opts_chunk$set(dev.args = list(type = "cairo"))
}
```

```{r load-models, include = FALSE}
library("RoBMA")

data("dat.lehmann2018", package = "metadat")
dat <- dat.lehmann2018

vignette_cache_load(cached_fits)
```

```{r precompute-models, include = FALSE, eval = FALSE}
library("RoBMA")

data("dat.lehmann2018", package = "metadat")
dat <- dat.lehmann2018

fit_RoBMA <- RoBMA(
  yi = yi, vi = vi, measure = "SMD",
  data = dat, seed = 1
)

fit_RoBMA_custom <- RoBMA(
  yi = yi, vi = vi, measure = "SMD",
  prior_bias = list(
    prior_PET(
      distribution = "Cauchy",
      parameters   = list(location = 0, scale = 1),
      truncation   = list(lower = 0, upper = Inf)
    ),
    prior_PEESE(
      distribution = "Cauchy",
      parameters   = list(location = 0, scale = 5),
      truncation   = list(lower = 0, upper = Inf)
    ),
    prior_weightfunction(
      "one-sided",
      steps   = c(0.025),
      weights = wf_cumulative(c(1, 1))
    )
  ),
  data = dat, seed = 1
)

vignette_cache_save(cached_fits)
```

In the [*Publication-Bias Adjustment*](v11-metafor-parity-publication-bias.html) vignette, we fit PET, PEESE, and a weight-function selection model on the ego-depletion dataset and obtained three different bias-corrected effect size estimates.
The methods encoded different bias mechanisms: PET and PEESE corrected for the relationship between effect sizes and standard errors or sampling variances, and the selection model adjusted for publication bias operating on p-values.
There is no consensus on which of the three is the right adjustment to apply.
The [*Bayesian Model Averaging*](v20-bayesian-model-averaging.html) vignette highlighted how an analogous fixed-vs-random-effects question can be resolved by combining the candidate models in one ensemble and weighting them by their posterior model probabilities;
the same idea applies here.

Robust Bayesian model-averaged meta-analysis (RoBMA) is the application of that idea to publication-bias adjustment [@maier2020robust; @bartos2021no].
The default `RoBMA()` ensemble averages over the presence vs absence of an effect, the presence vs absence of heterogeneity, and a no-bias spike against several weight-function and PET / PEESE bias models in one fit.
Inclusion Bayes factors quantify evidence for each of the three components, and the model-averaged estimates propagate uncertainty across the bias mechanisms instead of committing to a single one.

In this vignette we fit the default RoBMA ensemble on the same ego-depletion data used in the [*Publication-Bias Adjustment*](v11-metafor-parity-publication-bias.html) vignette, walk through the summary, the marginal posterior weights of the bias components, and the `interpret()` shortcut.
We list the available `model_type` presets and close with a custom ensemble that combines only the three single-model bias adjustments fit one at a time in that vignette.

## Dataset

The example dataset is `dat.lehmann2018` from the `metadat` package: 81 standardized mean differences from the ego-depletion meta-analysis with effect sizes `yi` and sampling variances `vi` already computed.
For the matching single-model bias adjustments and `metafor` package parity calls, see the [*Publication-Bias Adjustment*](v11-metafor-parity-publication-bias.html) vignette.

```{r data, include = TRUE}
library("RoBMA")

data("dat.lehmann2018", package = "metadat")
dat <- dat.lehmann2018
head(dat[, c("Short_Title", "Year", "yi", "vi")])
```

## Default RoBMA Ensemble

`RoBMA()` with default arguments fits the RoBMA-PSMA ensemble [@bartos2021no]: spike-and-slab prior distributions on the effect and heterogeneity, and a 9-component bias mixture (no-bias spike, six weight functions, PET, PEESE).
The default prior distributions on `mu`, `tau`, and the bias parameters match the single-model fits in the [*Publication-Bias Adjustment*](v11-metafor-parity-publication-bias.html) vignette.

```{r fit-RoBMA, eval = FALSE}
fit_RoBMA <- RoBMA(
  yi = yi, vi = vi, measure = "SMD",
  data = dat, seed = 1
)
```
```{r fit-RoBMA-summary}
summary(fit_RoBMA)
```

The *Components* table reports inclusion Bayes factors for the three components.
*Model-averaged estimates* combine all 36 sub-models in the ensemble (effect $\times$ heterogeneity $\times$ bias) weighted by their posterior probabilities; we report the bias parameters only for the bias-adjusted sub-models.

`summary_models()` breaks the mixture prior distributions of the three components down into their individual sub-models and reports their prior weight, posterior weight, and individual inclusion Bayes factor.

```{r summary-models}
summary_models(fit_RoBMA)
```

The posterior model weights show which sub-models the data prefer within each component's mixture.
The bias side of the ensemble concentrates posterior mass on PET (posterior probability $0.77$, individual inclusion BF $\approx 23$) with a smaller share on PEESE; the weight functions and the no-bias spike all receive negligible posterior weight.
The `inclusion BF` column is the Bayes factor for that one sub-model against the rest of its mixture, so it does not need to agree with the component-level inclusion BF reported by `summary()`.

`plot()` overlays the prior and the model-averaged posterior of any model parameter.
On `mu` the spike at zero (right-axis arrow) carries the posterior probability that the effect is null and the slab carries the rest; the grey shapes are the corresponding prior distribution.

```{r plot-RoBMA-mu, fig.width = 6, fig.height = 4, out.width = "75%", fig.align = "center"}
par(mar = c(4, 4, 1, 4))
plot(fit_RoBMA, parameter = "mu", prior = TRUE, xlim = c(-1, 1), lwd = 2)
```

`interpret()` distills the components and the pooled estimates into prose.
With `scope = "all"`, the bias section is included.

```{r interpret-RoBMA}
interpret(fit_RoBMA, scope = "all")
```

## Preset Ensembles

`model_type` selects a preset bias-model ensemble.
Each preset combines the no-bias spike with a different alternative set; the prior weight on "any bias" is $0.5$ in every preset.

| `model_type` | Alternative bias models                                                              |
|:-------------|:-------------------------------------------------------------------------------------|
| `"PSMA"`     | 6 weight functions, PET, PEESE (the default RoBMA-PSMA ensemble [@bartos2021no])     |
| `"PP"`       | PET and PEESE only                                                                   |
| `"6w"`       | 6 weight functions only (no PET / PEESE)                                             |
| `"2w"`       | 2 two-sided weight functions only                                                    |

The PEESE scale is rescaled by $\text{UISD}_{\text{ratio}} = \text{UISD}_{\text{SMD}} / \text{UISD}_{\text{measure}}$ so the prior distribution describes a comparable bias on each effect-size scale (the ratio is $1$ for `measure = "SMD"`).
Bias-adjustment prior distributions are not affected by `rescale_priors`.
Use `print_prior(fit, parameter = "bias")` on a `RoBMA(..., only_priors = TRUE)` fit to inspect the alternative components and prior weights inside any preset:

```{r priors-PP}
fit_priors <- RoBMA(
  yi = yi, vi = vi, measure = "SMD",
  model_type = "PP",
  data = dat, only_priors = TRUE
)
print_prior(fit_priors, parameter = "bias")
```

## Custom Bias Ensembles

Beyond the presets, we can specify the bias model space directly via `prior_bias`.
This is useful when the publication process under study has a specific shape (for example, two-sided selection only or a single weight-function step at a non-default cutoff), or for sensitivity analyses against the preset choice.

The example below builds a 3-model bias ensemble out of the three single-model bias adjustments from the [*Publication-Bias Adjustment*](v11-metafor-parity-publication-bias.html) vignette (default `bPET()`, default `bPEESE()`, and default `bselmodel()`), so the ensemble averages over the same three bias mechanisms that vignette fit one at a time, plus the no-bias spike.

```{r fit-RoBMA-custom, eval = FALSE}
fit_RoBMA_custom <- RoBMA(
  yi = yi, vi = vi, measure = "SMD",
  prior_bias = list(
    prior_PET(
      distribution = "Cauchy",
      parameters   = list(location = 0, scale = 1),
      truncation   = list(lower = 0, upper = Inf)
    ),
    prior_PEESE(
      distribution = "Cauchy",
      parameters   = list(location = 0, scale = 5),
      truncation   = list(lower = 0, upper = Inf)
    ),
    prior_weightfunction(
      "one-sided",
      steps   = c(0.025),
      weights = wf_cumulative(c(1, 1))
    )
  ),
  data = dat, seed = 1
)
```
```{r interpret-RoBMA-custom}
interpret(fit_RoBMA_custom, scope = c("components", "estimates"))
```

The two ensembles agree on all three component inclusion conclusions: strong evidence for heterogeneity, moderate evidence against the effect, and strong evidence in favour of publication bias.
The model-averaged effect on `mu` is essentially zero in both fits with overlapping credible intervals.
The agreement reflects the bias-mixture composition reported by `summary_models()` above: the default ensemble already concentrates posterior weight on PET, so dropping the weight functions that received negligible weight does not move the model-averaged inference.

The prior probability of any bias is $0.75$ in the custom fit and $0.5$ in the default, because the three user-supplied alternative prior distributions all default to `prior_weights = 1` against the no-bias spike's weight of $1$.
The component inclusion Bayes factor is invariant to this asymmetry; pass `prior_weights` explicitly on each prior in `prior_bias` to match a different prior odds split.

## Other Inference Helpers

`RoBMA()` returns an object of class `c("RoBMA", "brma")`, so all `summary()`, `plot()`, `predict()`, `funnel()`, `marginal_means()`, `rstudent()`, `qqnorm()`, and `influence()` methods documented in the [*Bayesian Meta-Analysis*](v02-bayesian-meta-analysis.html) vignette work the same way.
For single-model bias adjustment without averaging see the [*Publication-Bias Adjustment*](v11-metafor-parity-publication-bias.html) vignette; for the model-averaged workflow without publication-bias adjustment see [*Bayesian Model Averaging*](v20-bayesian-model-averaging.html); the prior distributions on `mu`, `tau`, and the moderators are documented in [*Prior Distributions*](v01-prior-distributions.html).

## References

---
title: "Location-Scale Meta-Analysis"
author: "František Bartoš"
date: "28th of April 2026"
output:
  rmarkdown::html_vignette:
    self_contained: yes
bibliography: ../inst/REFERENCES.bib
csl: ../inst/apa.csl
link-citations: true
vignette: >
  %\VignetteIndexEntry{Location-Scale Meta-Analysis}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
  %\VignetteEngine{knitr::rmarkdown_notangle}
---

```{r child = "_vignette-nowrap.md", echo = FALSE, eval = TRUE}
```

```{r setup, include = FALSE}
source("vignette-cache.R", local = knitr::knit_global())

cached_fits <- vignette_cache(
  name     = "v12-metafor-parity-location-scale",
  objects  = c("fit1_brma", "fit2_brma", "fit3_brma"),
  packages = c("metafor", "metadat")
)

knitr::opts_chunk$set(
  collapse    = TRUE,
  comment     = "#>",
  eval        = vignette_cache_eval(cached_fits),
  message     = FALSE,
  warning     = FALSE,
  fig.width   = 7,
  fig.height  = 3.5,
  dev         = "png",
  fig.retina  = 3
)
if (.Platform$OS.type == "windows") {
  knitr::opts_chunk$set(dev.args = list(type = "cairo"))
}
```

```{r load-models, include = FALSE}
library("metafor")
library("RoBMA")

data("dat.bangertdrowns2004", package = "metadat")
dat <- dat.bangertdrowns2004
dat$ni100 <- dat$ni / 100
dat$meta  <- as.factor(dat$meta)
dat$imag  <- as.factor(dat$imag)

vignette_cache_load(cached_fits)
```

```{r precompute-models, include = FALSE, eval = FALSE}
library("metafor")
library("RoBMA")

data("dat.bangertdrowns2004", package = "metadat")
dat <- dat.bangertdrowns2004
dat$ni100 <- dat$ni / 100
dat$meta  <- as.factor(dat$meta)
dat$imag  <- as.factor(dat$imag)

fit1_brma <- brma(
  yi = yi, vi = vi, measure = "SMD",
  data = dat, seed = 1
)

fit2_brma <- brma(
  yi = yi, vi = vi, measure = "SMD",
  mods  = ~ ni100,
  scale = ~ ni100,
  data = dat, seed = 1
)

fit3_brma <- brma(
  yi = yi, vi = vi, measure = "SMD",
  mods  = ~ ni100 + meta,
  scale = ~ ni100 + imag,
  data = dat, seed = 1
)

vignette_cache_save(cached_fits)
```

This vignette illustrates how to fit location-scale meta-analytic models with the `RoBMA` R package [@RoBMA].
We walk through the writing-to-learn dataset (`dat.bangertdrowns2004` in the `metadat` package), showing the `brma()` call with the `scale = ~ ...` argument alongside its `metafor::rma()` counterpart [@metafor], and build from a simple random-effects model up to a model with different predictors in the location and scale parts.

A location-scale model parameterizes the between-study heterogeneity $\tau$ as a regression on study-level covariates, allowing $\tau$ to vary across subgroups or as a function of continuous moderators.
The location side is the usual meta-regression on the effect.

All `brma()` calls keep the default prior distributions; see the [*Prior Distributions*](v01-prior-distributions.html) vignette for the prior definitions and customization options.
For the non-location-scale workflow showing more features (also applicable to location-scale models) see the [*Bayesian Meta-Analysis*](v02-bayesian-meta-analysis.html) vignette.

## Random-Effects Model

The writing-to-learn dataset contains 48 standardized mean differences from studies on the effectiveness of school-based writing interventions on academic achievement.
Effect sizes `yi` and sampling variances `vi` are already provided.

```{r data}
data("dat.bangertdrowns2004", package = "metadat")
dat <- dat.bangertdrowns2004
dat$ni100 <- dat$ni / 100
dat$meta  <- as.factor(dat$meta)
dat$imag  <- as.factor(dat$imag)
```

We start from a standard random-effects model with no scale predictor.

```{r fit1-metafor}
fit1_metafor <- metafor::rma(yi, vi, data = dat)
fit1_metafor
```

```{r fit1-brma, eval = FALSE}
fit1_brma <- brma(
  yi = yi, vi = vi, measure = "SMD",
  data = dat, seed = 1
)
```
```{r fit1-brma-summary}
summary(fit1_brma, include_mcmc_diagnostics = FALSE)
```

The pooled effect and the single heterogeneity parameter `tau` track the REML estimates from the `metafor` package.

## Adding a Scale Predictor

Sample size is a natural candidate for explaining heterogeneity: smaller studies often produce more variable estimates than larger ones.
We add the rescaled total sample size (`ni100 = ni / 100`) on both the location and the scale side.

```{r fit2-metafor}
fit2_metafor <- suppressWarnings(metafor::rma(
  yi, vi,
  mods  = ~ ni100,
  scale = ~ ni100,
  data = dat
))
fit2_metafor
```

```{r fit2-brma, eval = FALSE}
fit2_brma <- brma(
  yi = yi, vi = vi, measure = "SMD",
  mods  = ~ ni100,
  scale = ~ ni100,
  data = dat, seed = 1
)
```
```{r fit2-brma-summary}
summary(fit2_brma, include_mcmc_diagnostics = FALSE)
```

The location coefficient on `ni100` matches the REML estimate from the `metafor` package.
The scale block reports `exp(intercept)`, which is the baseline between-study heterogeneity $\tau$, and a multiplicative log-scale coefficient on `ni100`.
Note that `metafor::rma()` regresses $\log(\tau^2)$ while `brma()` regresses $\log(\tau)$, so each `metafor` package scale coefficient is exactly twice the matching `brma()` coefficient.

`regplot()` with `pi = TRUE` overlays the prediction interval on the bubble plot.
Because $\tau$ varies with `ni100`, the `brma()` prediction band narrows toward larger sample sizes, where the model expects less between-study heterogeneity.
`metafor::regplot()` does not currently draw a prediction interval for location-scale models, so only the regression line and confidence band appear in the left panel.

```{r regplot-pi, fig.width = 7, fig.height = 3.2}
par(mfrow = c(1, 2), mar = c(4, 4, 2, 1), cex.axis = 0.8, cex.lab = 0.8, cex.main = 0.75)
metafor::regplot(fit2_metafor, mod = "ni100", pi = TRUE,
                 main = "metafor", xlim = c(0, 6), ylim = c(-1, 2))
regplot(fit2_brma, mod = "ni100", pi = TRUE,
        main = "RoBMA",   xlim = c(0, 6), ylim = c(-1, 2))
```

## Different Variables in Location and Scale

Variables in the location and scale parts can differ.
The next fit adds metacognitive reflection (`meta`) on the location side and imaginative-component coding (`imag`) on the scale side.

```{r fit3-metafor}
fit3_metafor <- suppressWarnings(metafor::rma(
  yi, vi,
  mods  = ~ ni100 + meta,
  scale = ~ ni100 + imag,
  data = dat
))
fit3_metafor
```

```{r fit3-brma, eval = FALSE}
fit3_brma <- brma(
  yi = yi, vi = vi, measure = "SMD",
  mods  = ~ ni100 + meta,
  scale = ~ ni100 + imag,
  data = dat, seed = 1
)
```
```{r fit3-brma-summary}
summary(fit3_brma, include_mcmc_diagnostics = FALSE)
```

Each scale coefficient is again roughly half of its `metafor` package counterpart, with `brma()` shrinking the larger `metafor` package estimates more strongly because of the weakly informative default prior distribution.

## Scale Predictor Prior Distributions

Scale regression coefficients are not on the effect-size scale: they describe multiplicative changes in $\tau$ on the log scale and are therefore scale-free.
The default prior distribution is `Normal(0, 0.5)`, which corresponds to a weakly informative prior distribution on the multiplicative effect.
A coefficient of $\pm 0.5$ corresponds to roughly $\exp(\pm 0.5) \approx 0.6$ or $1.65$, i.e., a $40\%$ decrease or $65\%$ increase in $\tau$ per unit change in the predictor; values past $\pm 1$ correspond to a halving or doubling, which the prior distribution treats as extreme.

`plot()` with `prior = TRUE` overlays the prior distribution on the posterior so the shift is visible.
`parameter_scale = "ni100"` selects the scale-side coefficient (the analogous selector on the location side is `parameter_mods = "ni100"`).

```{r scale-posterior, fig.width = 5, fig.height = 3.5}
par(mar = c(4, 4, 1, 1))
plot(fit2_brma, parameter_scale = "ni100", prior = TRUE, xlim = c(-2, 2))
```

## Other Inference Helpers

All inference helpers available for any other `brma()` fit work the same way for the location-scale fit specified via `scale = ~ ...`.
This includes meta-regression bubble plots, marginal means, model comparison via Bayes factors, multilevel structures via `cluster`, and the full set of model-fit diagnostics.
See the [*Bayesian Meta-Analysis*](v02-bayesian-meta-analysis.html) vignette for more details.


## References

---
title: "Informed Bayesian Model-Averaged Meta-Analysis in Medicine"
author: "František Bartoš"
date: "`r Sys.Date()`"
output:
  rmarkdown::html_vignette:
    self_contained: yes
bibliography: ../inst/REFERENCES.bib
csl: ../inst/apa.csl
vignette: >
  %\VignetteIndexEntry{Informed Bayesian Model-Averaged Meta-Analysis in Medicine}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
  %\VignetteEngine{knitr::rmarkdown_notangle}
---

```{r setup, include = FALSE}
is_check <- ("CheckExEnv" %in% search()) ||
              any(c("_R_CHECK_TIMINGS_", "_R_CHECK_LICENSE_") %in% names(Sys.getenv())) ||
              !file.exists("../models/MedicineBMA/fit_BMA.RDS")
knitr::opts_chunk$set(
  collapse  = TRUE,
  comment   = "#>",
  eval      = !is_check,
  dev      = "png")
if(.Platform$OS.type == "windows"){
  knitr::opts_chunk$set(dev.args = list(type = "cairo"))
}
```
```{r include = FALSE}
library(RoBMA)
# we pre-load the RoBMA models, the fitting time is around 2-5 minutes
fit_BMA   <- readRDS(file = "../models/MedicineBMA/fit_BMA.RDS")
fit_BMAb  <- readRDS(file = "../models/MedicineBMA/fit_BMAb.RDS")
fit_RoBMA <- readRDS(file = "../models/MedicineBMA/fit_RoBMA.RDS")
```

```{r include = FALSE, eval = FALSE}
# R package version updating
library(RoBMA)

data("Poulsen2006", package = "RoBMA")

fit_BMA <- RoBMA(d = Poulsen2006$d, se = Poulsen2006$se, study_names = Poulsen2006$study,
                 priors_effect        = prior_informed(name = "oral health", parameter = "effect", type = "smd"),
                 priors_heterogeneity = prior_informed(name = "oral health", parameter = "heterogeneity", type = "smd"),
                 priors_bias          = NULL,
                 transformation = "cohens_d", seed = 1, parallel = TRUE)

fit_BMAb <- RoBMA(d = Poulsen2006$d, se = Poulsen2006$se, study_names = Poulsen2006$study,
                  priors_effect        = prior_informed(name = "oral health", parameter = "effect", type = "smd"),
                  priors_heterogeneity = prior_informed(name = "oral health", parameter = "heterogeneity", type = "smd"),
                  priors_bias          = NULL,
                  seed = 1, parallel = TRUE)

fit_RoBMA <- RoBMA(d = Poulsen2006$d, se = Poulsen2006$se, study_names = Poulsen2006$study,
                   priors_effect        = prior_informed(name = "oral health", parameter = "effect", type = "smd"),
                   priors_heterogeneity = prior_informed(name = "oral health", parameter = "heterogeneity", type = "smd"),
                   seed = 1, parallel = TRUE)

saveRDS(fit_BMA, file = "../models/MedicineBMA/fit_BMA.RDS")
saveRDS(fit_BMAb, file = "../models/MedicineBMA/fit_BMAb.RDS")
saveRDS(fit_RoBMA, file = "../models/MedicineBMA/fit_RoBMA.RDS")
```

Bayesian model-averaged meta-analysis allows researchers to seamlessly incorporate available prior information into the analysis [@gronau2017bayesian; @gronau2020primer; @bartos2021bayesian]. This vignette illustrates how to do this with an example from @bartos2021bayesian, who developed informed prior distributions for meta-analyses of continuous outcomes based on the Cochrane database of systematic reviews. Then, we extend the example by incorporating publication bias adjustment with robust Bayesian meta-analysis [@bartos2021no; @maier2020robust].


### Reproducing Informed Bayesian Model-Averaged Meta-Analysis (BMA)

We illustrate how to fit the informed BMA (not adjusting for publication bias) using the `RoBMA` R package. For this purpose, we reproduce the dentine hypersensitivity example from @bartos2021bayesian, who reanalyzed five studies with a tactile outcome assessment that were subjected to a meta-analysis by @poulsen2006potassium. 

We load the dentine hypersensitivity data included in the package.

```{r}
library(RoBMA)

data("Poulsen2006", package = "RoBMA")
Poulsen2006
``` 

To reproduce the analysis from the example, we need to set informed empirical prior distributions for the effect sizes ($\mu$) and heterogeneity ($\tau$) parameters that @bartos2021bayesian obtained from the Cochrane database of systematic reviews. We can either set them manually,
``` r
fit_BMA <- RoBMA(d = Poulsen2006$d, se = Poulsen2006$se, study_names = Poulsen2006$study,
                 priors_effect        = prior(distribution = "t", parameters = list(location = 0, scale = 0.51, df = 5)),
                 priors_heterogeneity = prior(distribution = "invgamma", parameters = list(shape = 1.79, scale = 0.28)),
                 priors_bias          = NULL,
                 transformation = "cohens_d", seed = 1, parallel = TRUE)
```
with `priors_effect` and `priors_heterogeneity` corresponding to the $\delta \sim T(0,0.51,5)$ and $\tau \sim InvGamma(1.79,0.28)$ informed prior distributions for the "oral health" subfield and removing the publication bias adjustment models by setting `priors_bias = NULL`$^1$. Note that the package contains function `NoBMA()` from version 3.1 which skips publication bias adjustment directly.

Alternatively, we can utilize the `prior_informed` function that prepares informed prior distributions for the individual medical subfields automatically.
``` r
fit_BMA <- RoBMA(d = Poulsen2006$d, se = Poulsen2006$se, study_names = Poulsen2006$study,
                 priors_effect        = prior_informed(name = "oral health", parameter = "effect", type = "smd"),
                 priors_heterogeneity = prior_informed(name = "oral health", parameter = "heterogeneity", type = "smd"),
                 priors_bias          = NULL,
                 transformation = "cohens_d", seed = 1, parallel = TRUE)
```
The `name` argument specifies the medical subfield name (use `print(BayesTools::prior_informed_medicine_names)` to check names of all available subfields). The `parameter` argument specifies whether we want prior distribution for the effect size or heterogeneity. Finally, the `type` argument specifies what type of measure we use for the meta-analysis (see `?prior_informed` for more details regarding the informed prior distributions).

We obtain the output with the `summary` function. Adding the `conditional = TRUE` argument allows us to inspect the conditional estimates, i.e., the effect size estimate assuming that the models specifying the presence of the effect are true and the heterogeneity estimates assuming that the models specifying the presence of heterogeneity are true$^2$.

```{r}
summary(fit_BMA, conditional = TRUE)
```

The output from the `summary.RoBMA()` function has 3 parts. The first one under the 'Robust Bayesian Meta-Analysis' heading provides a basic summary of the fitted models by component types (presence of the Effect and Heterogeneity). The table summarizes the prior and posterior probabilities and the inclusion Bayes factors of the individual components. The results  show that the inclusion Bayes factor for the effect corresponds to the one reported in @bartos2021bayesian, $\text{BF}_{10} = 218.53$ and $\text{BF}_{\text{rf}} = 3.52$ (up to an MCMC error). 

The second part under the 'Model-averaged estimates' heading displays the parameter estimates model-averaged across all specified models (i.e., including models specifying the effect size to be zero). We ignore this section and move to the last part.

The third part under the 'Conditional estimates' heading displays the conditional effect size estimate corresponding to the one reported in @bartos2021bayesian, $\delta = 1.082$, 95% CI [0.686,1.412], and a heterogeneity estimate (not reported previously).


### Visualizing the Results

The `RoBMA` package provides extensive options for visualizing the results. Here, we visualize the prior (grey) and posterior (black) distribution for the mean parameter.

```{r fig_mu_BMA, dpi = 300, fig.width = 6, fig.height = 4, out.width = "75%", fig.align = "center"}
plot(fit_BMA, parameter = "mu", prior = TRUE)
```

By default, the function plots the model-averaged estimates across all models; the arrows represent the probability of a spike, and the lines represent the posterior density under models assuming non-zero effect. The secondary y-axis (right) shows the probability of the spike (at value 0) decreasing from 0.50, to 0.005 (also obtainable from the 'Robust Bayesian Meta-Analysis' field in `summary.RoBMA()` function).

To visualize the conditional effect size estimate, we can add the `conditional = TRUE` argument,
 
```{r fig_mu_BMA_cond, dpi = 300, fig.width = 6, fig.height = 4, out.width = "75%", fig.align = "center"}
plot(fit_BMA, parameter = "mu", prior = TRUE, conditional = TRUE)
```
which displays only the model-averaged posterior distribution of the effect size parameter for models assuming the presence of the effect.

We can also visualize the estimates from the individual models used in the ensemble. We do that with the `plot_models()` function, which visualizes the effect size estimates and 95% CI of each specified model included in the ensemble. Model 1 corresponds to the fixed effect model assuming the absence of the effect, $H_0^{\text{f}}$, Model 2 corresponds to the random effect model assuming the absence of the effect, $H_0^{\text{r}}$, Model 3 corresponds to the fixed effect model assuming the presence of the effect, $H_1^{\text{f}}$, and Model 4 corresponds to the random effect model assuming the presence of the effect, $H_1^{\text{r}}$). The size of the square representing the mean estimate reflects the posterior model probability of the model, which is also displayed in the right-hand side panel. The bottom part of the figure shows the model averaged-estimate that is a combination of the individual model posterior distributions weighted by the posterior model probabilities.

```{r fig_models, dpi = 300, fig.height = 4.5, fig.width = 7, out.width = '75%', fig.align = "center"}
plot_models(fit_BMA)
```
We see that the posterior model probability of the first two models decreased to essentially zero (when rounding to two decimals), completely omitting their estimates from the figure. Furthermore, the much larger box of Model 4 (the random effect model assuming the presence of the effect) shows that Model 4 received the largest share of the posterior probability, $P(H_1^{\text{r}}) = 0.77$)

The last type of visualization that we show here is the forest plot. It displays the original studies' effects and the meta-analytic estimate within one figure. It can be requested by using the `forest()` function. Here, we again set the `conditional = TRUE` argument to display the conditional model-averaged effect size estimate at the bottom.

```{r fig_forest, dpi = 300, fig.height = 4.5, fig.width = 7, out.width = '75%', fig.align = "center"}
forest(fit_BMA, conditional = TRUE)
```

For more options provided by the plotting function, see its documentation by using `?plot.RoBMA()`, `?plot_models()`, and `?forest()`.


### Adjusting for Publication Bias with Robust Bayesian Meta-Analysis

Finally, we illustrate how to adjust our informed BMA for publication bias with robust Bayesian meta-analysis [@bartos2021no, @maier2020robust]. In short, we specify additional models assuming the presence of the publication bias and correcting for it by either specifying a selection model operating on $p$-values [@vevea1995general] or by specifying a publication bias adjustment method correcting for the relationship between effect sizes and standard errors -- PET-PEESE [@stanley2017limitations; @stanley2014meta]. See @bartos2020adjusting for a tutorial.

To obtain a proper before and after publication bias adjustment comparison, we fit the informed BMA model but using the default effect size transformation (Fisher's $z$).

``` r
fit_BMAb <- RoBMA(d = Poulsen2006$d, se = Poulsen2006$se, study_names = Poulsen2006$study,
                  priors_effect        = prior_informed(name = "oral health", parameter = "effect", type = "smd"),
                  priors_heterogeneity = prior_informed(name = "oral health", parameter = "heterogeneity", type = "smd"),
                  priors_bias          = NULL,
                  seed = 1, parallel = TRUE)
```
```{r}
summary(fit_BMAb, conditional = TRUE)
```

We obtain noticeably stronger evidence for the presence of the effect. This is a result of placing more weights on the fixed-effect models, especially the fixed-effect model assuming the presence of the effect $H_1^f$. In our case, the increase in the posterior model probability of $H_1^f$ occurred because this model predicted the data slightly better after removing the correlation between effect sizes and standard errors (a consequence of using Fisher's $z$ transformation). Nevertheless, the conditional effect size estimate stayed almost the same.

Now, we fit the publication bias-adjusted model by simply removing the `priors_bias = NULL` argument, which allows us to obtain the default 36 models ensemble called RoBMA-PSMA [@bartos2021no].

``` r
fit_RoBMA <- RoBMA(d = Poulsen2006$d, se = Poulsen2006$se, study_names = Poulsen2006$study,
                   priors_effect        = prior_informed(name = "oral health", parameter = "effect", type = "smd"),
                   priors_heterogeneity = prior_informed(name = "oral health", parameter = "heterogeneity", type = "smd"),
                   seed = 1, parallel = TRUE)
```
```{r}
summary(fit_RoBMA, conditional = TRUE)
```

We notice the additional values in the 'Components summary' table in the 'Bias' row. The model is now extended with 32 publication bias adjustment models that account for 50% of the prior model probability. When comparing the RoBMA to the second BMA fit, we notice a large decrease in the inclusion Bayes factor for the presence of the effect $\text{BF}_{10} = 6.02$ vs. $\text{BF}_{10} = 347.93$, which still, however, presents moderate evidence for the presence of the effect. We can further quantify the evidence in favor of the publication bias with the inclusion Bayes factor for publication bias $\text{BF}_{pb} = 2.30$, which can be interpreted as weak evidence in favor of publication bias.

We can also compare the publication bias unadjusted and publication bias-adjusted conditional effect size estimates. Including models assuming publication bias into our model-averaged estimate (assuming the presence of the effect) slightly decreases the estimated effect to $\delta = 0.838$, 95% CI [-0.035, 1.297] with a much wider confidence interval, as visualized in the prior and posterior conditional effect size estimate plot.

```{r fig_mu_RoBMA_cond, dpi = 300, fig.width = 6, fig.height = 4, out.width = "75%", fig.align = "center"}
plot(fit_RoBMA, parameter = "mu", prior = TRUE, conditional = TRUE)
```

                 
### Footnotes

$^1$ The additional setting `transformation = "cohens_d"` allows us to get more comparable results with the `metaBMA` R package since RoBMA otherwise internally transforms the effect sizes to Fisher's $z$ for the fitting purposes. The `seed = 1` and `parallel = TRUE` options grant us exact reproducibility of the results and parallelization of the fitting process.

$^2$ The model-averaged estimates that `RoBMA` returns by default model-averaged across all specified models -- a different behavior from the `metaBMA` package that by default returns what we call "conditional" estimates in `RoBMA`.

### References

---
title: "Hierarchical Bayesian Model-Averaged Meta-Analysis"
author: "František Bartoš"
date: "`r Sys.Date()`"
output:
  rmarkdown::html_vignette:
    self_contained: yes
bibliography: ../inst/REFERENCES.bib
csl: ../inst/apa.csl
vignette: >
  %\VignetteIndexEntry{Hierarchical Bayesian Model-Averaged Meta-Analysis}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
  %\VignetteEngine{knitr::rmarkdown_notangle}
---

```{r setup, include = FALSE}
is_check <- ("CheckExEnv" %in% search()) ||
              any(c("_R_CHECK_TIMINGS_", "_R_CHECK_LICENSE_") %in% names(Sys.getenv())) ||
              !file.exists("../models/HierarchicalBMA/fit.RDS")
knitr::opts_chunk$set(
  collapse  = TRUE,
  comment   = "#>",
  eval      = !is_check,
  dev      = "png")
if(.Platform$OS.type == "windows"){
  knitr::opts_chunk$set(dev.args = list(type = "cairo"))
}
```
```{r include = FALSE}
library(RoBMA)
# we pre-load the RoBMA models, the fitting time is around 2-5 minutes
fit.0             <- readRDS(file = "../models/HierarchicalBMA/fit.0.RDS")
fit               <- readRDS(file = "../models/HierarchicalBMA/fit.RDS")
fit_BMA           <- readRDS(file = "../models/HierarchicalBMA/fit_BMA.RDS")
hierarchical_test <- readRDS(file = "../models/HierarchicalBMA/hierarchical_test.RDS")
```

```{r include = FALSE, eval = FALSE}
# R package version updating
library(RoBMA)

data("dat.konstantopoulos2011", package = "metadat")
dat <- dat.konstantopoulos2011

fit.0 <- RoBMA(d = dat$yi, v = dat$vi,
             priors_effect_null        = NULL,
             priors_heterogeneity_null = NULL,
             priors_bias               = NULL,
             parallel = TRUE, seed = 1)

fit <- RoBMA(d = dat$yi, v = dat$vi, study_ids = dat$district,
             priors_effect_null        = NULL,
             priors_heterogeneity_null = NULL,
             priors_bias               = NULL,
             parallel = TRUE, seed = 1)

fit_BMA <- RoBMA(d = dat$yi, v = dat$vi, study_ids = dat$district,
                 priors_bias = NULL,
                 parallel = TRUE, seed = 1)

hierarchical_test <- RoBMA(d = dat$yi, v = dat$vi, study_ids = dat$district,
                           priors_heterogeneity_null = NULL,
                           priors_hierarchical_null = prior(distribution = "spike", parameters = list("location" = 0)),
                           priors_bias = NULL,
                           parallel = TRUE, seed = 1)
 
saveRDS(fit.0,             file = "../models/HierarchicalBMA/fit.0.RDS",             compress = "xz")
saveRDS(fit,               file = "../models/HierarchicalBMA/fit.RDS",               compress = "xz")
saveRDS(fit_BMA,           file = "../models/HierarchicalBMA/fit_BMA.RDS",           compress = "xz")
saveRDS(hierarchical_test, file = "../models/HierarchicalBMA/hierarchical_test.RDS", compress = "xz")
```

Hierarchical (or multilevel/3-level) meta-analysis adjusts for the dependency of effect sizes due to clustering in the data. For example, effect size estimates from multiple experiments reported in the same manuscript might be expected to be more similar than effect sizes from a different paper  [@konstantopoulos2011fixed]. This vignette illustrates how to deal with such dependencies among effect size estimates (in cases with simple nested structure) using the Bayesian model-averaged meta-analysis (BMA) [@gronau2017bayesian; @gronau2020primer; @bartos2021bayesian]. (See other vignettes for more details on BMA: [Reproducing BMA](ReproducingBMA.html) or [Informed BMA in medicine](MedicineBMA.html).)

First, we introduce the example data set. Second, we illustrate the frequentist hierarchical meta-analysis with the `metafor` R package and discuss the results. Third, we outline the hierarchical meta-analysis parameterization. Fourth, we estimate the Bayesian model-averaged hierarchical meta-analysis. Finally, we conclude by discussing further extensions and publication bias adjustment.

Note that since version 3.5 of the `RoBMA` package, the hierarchical meta-analysis and meta-regression can use the spike-and-slab model-averaging algorithm described in [Fast Robust Bayesian Meta-Analysis via Spike and Slab Algorithm](FastRoBMA.html). The spike-and-slab model-averaging algorithm is a more efficient alternative to the bridge algorithm, which is the current default in the `RoBMA` package. For non-selection models, the likelihood used in the spike-and-slab algorithm is equivalent to the bridge algorithm. However, the likelihood in the selection models is modified since marginalization of the random-effects is not performed. As such, the selection models use likelihood that closely approximates the exact selection model likelihood. In our testing and a simulation study, found that the approximate selection likelihood performs reasonably well and is better than the non-hierarchical exact likelihood model when within-study dependencies are present. 

### Example Data Set

We use the `dat.konstantopoulos2011` data set from the `metadat` R package [@metadat] that is used for the same functionality in the metafor [@metafor] R package. We roughly follow the example in the data set's help file, `?dat.konstantopoulos2011`. The data set consists of 56 studies estimating the effects of modified school calendars on students' achievement. The 56 studies were run in individual schools, which can be grouped into 11 districts. We might expect more similar effect size estimates from schools in the same district -- in other words, the effect size estimates from the same district might not be completely independent. Consequently, we might want to adjust for this dependency (clustering) between the effect size estimates to draw a more appropriate inference.

First, we load the data set, assign it to the `dat` object, and inspect the first few rows.
```{r}
data("dat.konstantopoulos2011", package = "metadat")
dat <- dat.konstantopoulos2011

head(dat)
``` 

In the following analyses, we use the following variables:

 - `yi`, standardized mean differences,
 - `vi`, sampling variances of the standardized mean differences,
 - `district`, district id which distinguishes among the districts,
 - and `school`, that distinguishes among different schools within the same district.
 
### Frequentist Hierarchical Meta-Analysis with `metafor`

We follow the data set's help file and fit a simple random effects meta-analysis using the `rma()` function from `metafor` package. This model ignores the dependency between effect size estimates. We use this simple model as our starting point and as a comparison with the later models.
``` {r}
fit_metafor.0 <- metafor::rma(yi = yi, vi = vi, data = dat)
fit_metafor.0
```
The model summary returns a small but statistically significant effect size estimate $\mu = 0.128$ ($\text{se} = 0.044$) and a considerable heterogeneity estimate $\tau = 0.297$.

We extend the model to account for the hierarchical structure of the data, i.e., schools within districts, by using the `rma.mv()` function from the `metafor` package and extending it with the `random = ~ school | district` argument.
``` {r}
fit_metafor <- metafor::rma.mv(yi, vi, random = ~ school | district, data = dat)
fit_metafor
```
We find that accounting for the hierarchical structure of the data results in (1) a slightly larger effect size estimate ($\mu = 0.187$) and (2) larger standard error of the effect size estimate ($\text{se} = 0.085$). The larger standard error is a natural consequence of accounting for the dependency between the effect sizes. Because the effect sizes are dependent, they contribute less additional information than independent effect sizes would. Specifying the hierarchical model then accounts for the dependency by estimating similarity between the estimates from the same cluster (school) and discounting the information borrowed from each estimate. The estimate of the similarity among estimates from the same cluster is summarized in the `\rho = 0.666` estimate.

### Specifications of Hierarchical Meta-Analysis

We specify a simple hierarchical meta-analytic model (see @konstantopoulos2011fixed for an example). Using distributional notation, we can describe the data generating process as a multi-stage sampling procedure. In a nutshell, we assume the existence of an overall mean effect $\mu$. Next, we assume that the effect sizes in each district $k = 1, \dots, K$, $\gamma_k$, systematically differ from the mean effect, with the variance of the district-level effects summarized with heterogeneity $\tau_{b}$ (as between). Furthermore, we assume that the true effects $\theta_{k,j}$ of each study $j = 1, \dots J_k$ systematically differ from the district-level effect, with the variance of the study effects from the district-level effect summarized with heterogeneity $\tau_{w}$ (as within). Finally, the observed effect sizes $y_{k,j}$ that differ from the true effects $y_{k,j}$ due to random errors $\text{se}_{k,j}$.

Mathematically, we can describe such a model as:
$$
\begin{aligned}
  \gamma_k     &\sim \text{N}(\mu,          \tau_b^2),\\
  \theta_{k,j} &\sim \text{N}(\gamma_k,     \tau_w^2),\\
   y_{k,j}     &\sim \text{N}(\theta_{k,j}, \text{se}_{k,j}).\\
\end{aligned}
$$
Where N() denotes a normal distribution with mean and variance.

Conveniently, and with a bit of algebra, we do not need to estimate the district-level and true study effects. Instead, we marginalize them out, and we sample the observed effect sizes from each district $y_{k,.}$ directly from a multivariate normal distributions, MN(), with a common mean $\mu$ and covariance matrix S:
$$
\begin{aligned}
   y_{k,.}  &\sim \text{MN}(\mu, \text{S}),\\
   \text{S} &= \begin{bmatrix} 
      \tau_b^2 + \tau_w^2 + \text{se}_1^2 & \tau_w^2 & \dots & \tau_w^2  \\
      \tau_w^2 & \tau_b^2 + \tau_w^2 + \text{se}_2^2 & \dots & \tau_w^2  \\
      \dots  & \dots & \dots & \dots \\
      \tau_w^2 & \tau_w^2 & \dots & \tau_b^2 + \tau_w^2 + \text{se}_{J_k}^2 & \\
  \end{bmatrix}.
\end{aligned}
$$
The random effects marginalization is helpful as it allows us to sample fewer parameters from the posterior distribution (which significantly simplifies marginal likelihood estimation via bridge sampling). Furthermore, the marginalization allows us to properly specify selection model publication bias adjustment models -- the marginalization propagates the selection process up through all the sampling steps at once (we cannot proceed with the sequential sampling as the selection procedure on the observed effect sizes modifies the sampling distributions of all the preceding levels).

We can further re-parameterize the model by performing the following substitution,
$$
\begin{aligned}
   \tau^2 &= \tau_b^2 + \tau_w^2,\\
   \rho   &= \frac{\tau_w^2}{\tau_b^2 + \tau_w^2},
\end{aligned}
$$
and specifying the covariance matrix using the inter-study correlation $\rho$, total heterogeneity $\tau$, and the standard errors $\text{se}_{.}$:
$$
\begin{aligned}
   \text{S} &= \begin{bmatrix} 
      \tau^2 + \text{se}_1^2 & \rho\tau^2 & \dots & \rho\tau^2  \\
      \rho\tau^2 & \tau^2 + \text{se}_2^2 & \dots & \rho\tau^2  \\
      \dots  & \dots & \dots & \dots \\
      \rho\tau^2 & \rho\tau^2 & \dots & \tau^2 + \text{se}_{J_k}^2 & \\
  \end{bmatrix}.
\end{aligned}
$$
This specification corresponds to the compound symmetry covariance matrix of random effects, the default settings in the  `metafor::rma.mv()` function. More importantly, it allows us to easily specify prior distributions on the correlation coefficient $\rho$ and the total heterogeneity $\tau$.

### Hierarchical Bayesian Model-Averaged Meta-Analysis with `RoBMA`

Before we estimate the complete Hierarchical Bayesian Model-Averaged Meta-Analysis (hBMA) with the `RoBMA` package, we briefly reproduce the simpler models we estimated with the `metafor` package in the previous section.

#### Bayesian Random Effects Meta-Analysis
First, we estimate a simple Bayesian random effects meta-analysis (corresponding to `fit_metafor.0`). We use `the RoBMA()` function and specify the effect sizes and sampling variances via the `d = dat$yi` and `v = dat$vi` arguments. We set the `priors_effect_null`, `priors_heterogeneity_null`, and `priors_bias` arguments to null to omit models assuming the absence of the effect, heterogeneity, and the publication bias adjustment components.
``` r
fit.0 <- RoBMA(d = dat$yi, v = dat$vi,
               priors_effect_null        = NULL,
               priors_heterogeneity_null = NULL,
               priors_bias               = NULL,
               parallel = TRUE, seed = 1)
```
We generate a complete summary for the only estimated model by adding the `type = "individual"` argument to the `summary()` function.
``` {r}
summary(fit.0, type = "individual")
```
We verify that the effect size, $\mu = 0.126$ ($\text{95% CI } [0.041, 0.211]$), and heterogeneity, $\tau = 0.292$ ($\text{95% CI } [0.233, 0.364]$), estimates closely correspond to the frequentist results (as we would expect from parameter estimates under weakly informative priors).

#### Hierarchical Bayesian Random Effects Meta-Analysis
Second, we account for the clustered effect size estimates within districts by extending the previous function call with the `study_ids = dat$district` argument. This allows us to estimate the hierarchical Bayesian random effects meta-analysis (corresponding to `fit_metafor`). We use the default prior distribution for the correlation parameter `\rho \sim \text{Beta}(1, 1)`, set via the `priors_hierarchical` argument, which restricts the correlation to be positive and uniformly distributed on the interval $(0, 1)$.
``` r
fit <- RoBMA(d = dat$yi, v = dat$vi, study_ids = dat$district,
             priors_effect_null        = NULL,
             priors_heterogeneity_null = NULL,
             priors_bias               = NULL,
             parallel = TRUE, seed = 1)
```
Again, we generate the complete summary for the only estimated model,
``` {r}
summary(fit, type = "individual")
```
and verify that our estimates, again, correspond to the frequentist counterparts, with the estimated effect size, $\mu = 0.181$ ($\text{95% CI } [0.017, 0.346]$), heterogeneity, $\tau = 0.308$ ($\text{95% CI } [0.223, 0.442]$), and correlation, $\rho = 0.627$ ($\text{95% CI } [0.320, 0.864]$).

We can further visualize the prior and posterior distribution of the $\rho$ parameter using the `plot()` function.
```{r fig_rho, dpi = 300, fig.width = 4, fig.height = 3, out.width = "50%", fig.align = "center"}
par(mar = c(2, 4, 0, 0))
plot(fit, parameter = "rho", prior = TRUE)
```

#### Hierarchical Bayesian Model-Averaged Meta-Analysis
Third, we extend the previous model into a model ensemble that also includes models assuming the absence of the effect and/or heterogeneity (we do not incorporate models assuming presence of publication bias due to computational complexity explained in the summary). Including those additional models allows us to evaluate evidence in favor of the effect and heterogeneity. Furthermore, specifying all those additional models allows us to incorporate the uncertainty about the specified models and weight the posterior distribution according to how well the models predicted the data. We estimate the remaining models by removing the `priors_effect_null` and `priors_heterogeneity_null` arguments from the previous function calls, which include the previously omitted models of no effect and/or no heterogeneity.
``` r
fit_BMA <- RoBMA(d = dat$yi, v = dat$vi, study_ids = dat$district,
                 priors_bias = NULL,
                 parallel = TRUE, seed = 1)
```
Now we generate a summary for the complete model-averaged ensemble by not specifying any additional arguments in the `summary()` function.
``` {r}
summary(fit_BMA)
```
We find the ensemble contains four models, the combination of models assuming the presence/absence of the effect/heterogeneity, each with equal prior model probabilities. Importantly, the models assuming heterogeneity are also specified with the hierarchical structure and account for the clustering. A comparison of the specified models reveals weak evidence against the effect, $\text{BF}_{10} = 0.917$, and extreme evidence for the presence of heterogeneity, $\text{BF}_{\text{rf}} = 9.3\times10^{92}$. Moreover, we find that the `Hierarchical` component summary has the same values as the `Heterogeneity` component summary because the default settings specify that all models assuming the presence of heterogeneity also include the hierarchical structure.

We also obtain the model-averaged posterior estimates that combine the posterior estimates from all models according to the posterior model probabilities, the effect size, $\mu = 0.087$ ($\text{95% CI } [0.000, 0.314]$), heterogeneity, $\tau = 0.326$ ($\text{95% CI } [0.231, 0.472]$), and correlation, $\rho = 0.659$ ($\text{95% CI } [0.354, 0.879]$).

#### Testing the Presence of Clustering
In the previous analyses, we assumed that the effect sizes are indeed clustered within the districts, and we only adjusted for the clustering. However, the effect sizes within the same cluster may not be more similar than effect sizes from different clusters. Now, we specify a model ensemble that allows us to test this assumption by specifying two sets of random effect meta-analytic models. The first set of models assumes that there is indeed clustering and that the correlation of random effects is uniformly distributed on the $(0, 1)$ interval (as in the previous analyses). The second set of models assumes that there is no clustering, i.e., the correlation of random effects $\rho = 0$, which simplifies the structured covariance matrix to a diagonal matrix. Again, we model average across models assuming the presence and absence of the effect to account for the model uncertainty.

To specify this 'special' model ensemble with the `RoBMA()` function, we need to modify the previous model call in the following ways. We removed the fixed effect models by specifying the `priors_heterogeneity_null = NULL` argument.$^1$ Furthermore, we specify the prior distribution for models assuming the absence of the hierarchical structure by adding the `priors_hierarchical_null = prior(distribution = "spike", parameters = list("location" = 0))` argument.
```r
hierarchical_test <- RoBMA(d = dat$yi, v = dat$vi, study_ids = dat$district,
                           priors_heterogeneity_null = NULL,
                           priors_hierarchical_null = prior(distribution = "spike", parameters = list("location" = 0)),
                           priors_bias = NULL,
                           parallel = TRUE, seed = 1)

```
``` {r}
summary(hierarchical_test)
```
We summarize the resulting model ensemble and find out that the `Hierarchical` component is no longer equivalent to the `Heterogeneity` component -- the new model specification allowed us to compare random effect models assuming the presence of the hierarchical structure to random effect models assuming the absence of the hierarchical structure. The resulting inclusion Bayes factor of the hierarchical structure shows extreme evidence in favor of clustering of the effect sizes, $\text{BF}_{\rho\bar{\rho}} = 4624$, i.e., there is extreme evidence that the intervention results in more similar effects within the districts.

### Summary

We illustrated how to estimate a hierarchical Bayesian model-averaged meta-analysis using the `RoBMA` package. The hBMA model allows us to test for the presence vs absence of the effect and heterogeneity while simultaneously adjusting for clustered effect size estimates. While the current implementation allows us to draw a fully Bayesian inference, incorporate prior information, and acknowledge model uncertainty, it has a few limitations in contrast to the `metafor` package. E.g., the `RoBMA` package only allows a simple nested random effects (i.e., estimates within studies, schools within districts etc). The simple nesting allows us to break the full covariance matrix into per cluster block matrices which speeds up the already demanding computation. Furthermore, the computational complexity significantly increases when considering selection models as we need to compute an exponentially increasing number of multivariate normal probabilities with the increasing cluster size (existence of clusters with more than four studies makes the current implementation impractical due to the computational demands). However, these current limitations are not the end of the road, as we are exploring other approaches (e.g., only specifying PET-PEESE style publication bias adjustment and other dependency adjustments) in a future vignette.

### Footnotes

$^1$ We could also model-average across the hierarchical structure assuming fixed effect models, i.e., $\tau \sim f(.)$ and $\rho = 1$. However specifying such a model ensemble is a beyond the scope of this vignette, see [Custom ensembles](CustomEnsembles.html) vignette for some hints.

### References

---
title: "Prior Distributions"
author: "František Bartoš"
date: "28th of April 2026"
output:
  rmarkdown::html_vignette:
    self_contained: yes
bibliography: ../inst/REFERENCES.bib
csl: ../inst/apa.csl
link-citations: true
vignette: >
  %\VignetteIndexEntry{Prior Distributions}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
  %\VignetteEngine{knitr::rmarkdown_notangle}
---

```{r child = "_vignette-nowrap.md", echo = FALSE, eval = TRUE}
```

```{r setup, include = FALSE}
knitr::opts_chunk$set(
  collapse   = TRUE,
  comment    = "#>",
  message    = FALSE,
  warning    = FALSE,
  fig.width  = 6,
  fig.height = 4
)

library(RoBMA)

has_metafor <- requireNamespace("metafor", quietly = TRUE)
has_bcg     <- has_metafor && requireNamespace("metadat", quietly = TRUE)
```

Prior distributions describe which parameter values are plausible before seeing the data.
They are an essential part of any Bayesian analysis, and they matter especially in meta-analysis: many datasets contain only a small or moderate number of studies, and reasonable prior distributions stabilize estimates that the data alone would identify weakly.

The role of prior distributions differs substantially between *estimation* and *hypothesis testing* (or Bayesian model averaging).
For estimation, the goal is to recover a parameter value, and weakly informative prior distributions regularize the posterior without dominating it; the data quickly pull the posterior toward the likelihood as the number of studies grows.
For hypothesis testing and model averaging the prior distribution under the alternative is part of the hypothesis: the Bayes factor compares how well competing prior predictions match the data.
A very wide prior distribution on the alternative spreads probability across implausible values, so ordinary data look surprisingly consistent with the null, and the Bayes factor leans toward the null even when the effect is real.
Bayesian model averaging amplifies this: the alternative model receives little posterior probability and the model-averaged estimate is shrunk toward zero.

The default prior distributions in the `RoBMA` R package are therefore deliberately not "flat".
They are weakly informative: wide enough to regularize estimates, but informative enough to keep hypothesis tests and Bayesian model averaging well-calibrated.
This vignette walks through the practical paths to specifying prior distributions in the `RoBMA` R package, following the discussion of prior distributions in @bartos2025bayesian.
Also see @mulder2024bayesian for further discussion and resources.

## Specifying Prior Distributions

Prior distributions in the `RoBMA` R package come in two parts.
Prior distributions for the *meta-analytic parameters* (the average effect, between-study heterogeneity, and any meta-regression coefficients) can be specified using package defaults, empirical (informed) prior distributions, or fully custom prior distributions.
Prior distributions for the *publication-bias adjustment models*, used by `RoBMA()` and the related single-model functions, are treated separately and select among preset ensembles or a custom ensemble.

| Approach | Description | Main arguments |
|:--|:--|:--|
| Default prior distributions | Standardized prior distributions automatically scaled to the effect-size measure | `measure`, `rescale_priors`, `prior_unit_information_sd` |
| Empirical/informed prior distributions | Empirical prior distributions derived from previous meta-analyses | `prior_informed_field`, `prior_informed_subfield`, `prior_informed()` |
| Custom prior distributions | Fully user-specified prior distributions via `prior()` and helpers | `prior_effect`, `prior_heterogeneity`, `prior_mods`, `prior_scale` |
| Publication-bias prior distributions | Bias-adjustment models and ensembles | `model_type`, `prior_bias`, `prior_bias_null` |

Throughout this vignette we pass `only_priors = TRUE` to ask the package to assemble the prior distributions and stop before fitting the model.
This makes prior specifications visible without running MCMC.
Drop `only_priors = TRUE` to fit the corresponding model.

## Default Prior Distributions

### Basic example

The simplest way to use the defaults is to tell the `RoBMA` R package what effect-size measure is being analyzed; the package then specifies prior distributions on the matching scale.
We start with `brma()` because it fits one meta-analytic model and makes the prior distributions easiest to inspect.

The BCG vaccine dataset from the `metadat` package contains two-by-two tables.
We compute log risk ratios and their standard errors, then use `brma()` to show the default prior distributions for a simple random-effects meta-analysis.

```{r bcg-data, eval = has_bcg}
data("dat.bcg", package = "metadat")

bcg <- metafor::escalc(
  ai = tpos, bi = tneg, ci = cpos, di = cneg,
  measure = "RR", data = dat.bcg
)
```

```{r bcg-default-priors, eval = has_bcg}
bcg_priors <- brma(
  yi = yi, vi = vi, measure = "RR",
  data = bcg, only_priors = TRUE
)

plot_prior(bcg_priors,  parameter = "mu")
plot_prior(bcg_priors,  parameter = "tau")
print_prior(bcg_priors)
```

The first plot shows the prior distribution for the average effect, $\mu$.
The second plot shows the prior distribution for the between-study heterogeneity, $\tau$.
For `measure = "RR"` the package places a `Normal(0, 1)` prior distribution on the log risk ratio and a positive `Normal(0, 0.5)` prior distribution on heterogeneity.
The next subsection explains where these scales come from.

### Effect-size measures and their scales

Different effect-size measures live on different scales: while a standardized mean difference of 1 is large, a Fisher's $z$ of 1 is enormous.
As such, a single fixed prior such as `Normal(0, 1)` would not be a sensible default across different effect sizes.

The `RoBMA` R package deals with this problem by relying on the *unit-information standard deviation* (UISD) concept [@rover2021weakly].
Standardized effect sizes have known UISD values, and for non-standardized inputs (`measure = "GEN"`) the UISD can be estimated from the per-study standard errors and sample sizes (or specified manually):

| Effect-size measure | `measure` | UISD used by default |
|:--|:--|:--|
| Standardized mean difference (Cohen's $d$, Hedges' $g$) | `"SMD"` | `sqrt(2)` |
| Fisher's $z$-transformed correlation | `"ZCOR"` | `1` |
| Log risk ratio | `"RR"` | `sqrt(4)` |
| Log odds ratio | `"OR"` | `sqrt(4)` |
| Log hazard ratio | `"HR"` | `sqrt(4)` |
| Log incidence rate ratio | `"IRR"` | `sqrt(4)` |
| Generic / non-standardized | `"GEN"` | estimated from `sei` and `ni`, or set manually |

The default prior distributions then use fractions of the UISD to specify the scale of each prior distribution.

| Parameter | Meaning | Default prior distribution |
|:--|:--|:--|
| `mu`    | average effect                                 | `Normal(0, 0.5 * UISD)` |
| `tau`   | heterogeneity                                  | positive `Normal(0, 0.25 * UISD)` |
| `mods`  | meta-regression coefficients                   | `Normal(0, 0.25 * UISD)` |
| `scale` | heterogeneity-regression coefficients          | `Normal(0, 0.5)` |
| `rho`   | heterogeneity allocation in multilevel models  | `Beta(1, 1)` |

The `scale` regression coefficients are not on the effect-size scale: they describe changes in log heterogeneity, a multiplicative effect, and are therefore scale-free.
The notation for the meta-regression and multilevel parameters follows the `RoBMA` R package meta-regression and multilevel papers [@bartos2023robust; @bartos2025robust].

The default prior distributions can be made wider or narrower with `rescale_priors`.
Values larger than one make the prior distributions wider; values smaller than one make them narrower.

```{r bcg-rescale, eval = has_bcg}
bcg_wider_priors <- brma(
  yi = yi, vi = vi, measure = "RR",
  rescale_priors = 2,
  data = bcg, only_priors = TRUE
)

print_prior(bcg_wider_priors)
```

The effect prior distribution changed from `Normal(0, 1)` to `Normal(0, 2)`, and the heterogeneity prior distribution from positive `Normal(0, 0.5)` to positive `Normal(0, 1)`.

### Standardized effect-size inputs

Several common effect-size measures live on familiar but inconvenient scales.
Correlations are bounded in $(-1, 1)$ with non-linear behaviour near the boundaries, and their sampling variance depends on the effect size itself.
Odds ratios and risk ratios are positive and asymmetric around the no-effect value of 1.
For each of these, the standard practice is to transform the inputs to a related scale on which the sampling distribution is approximately normal: Fisher's $z$ for correlations, log odds ratio for odds ratios, log risk ratio for risk ratios.
`metafor::escalc()` performs all these transformations, and the `RoBMA` R package is set up to operate on the transformed scale.
Use `measure = "ZCOR"` for Fisher's $z$, `measure = "OR"` for log odds ratios, `measure = "RR"` for log risk ratios, and so on.
The default prior distributions then use the UISD that matches the transformed scale.

The `Kroupova2021` dataset of correlations between student employment and educational outcomes makes the workflow concrete.
We use `escalc()` to convert correlations and sample sizes into Fisher's $z$ effect sizes and their sampling variances:

```{r Kroupova2021-data, eval = has_metafor}
data("Kroupova2021", package = "RoBMA")

Kroupova2021 <- metafor::escalc( 
  ri = r, ni = sample_size, measure = "ZCOR",
  data = Kroupova2021
)
```

`brma()` recognises `measure = "ZCOR"` and uses the default Fisher's-$z$ prior distributions:

```{r Kroupova2021-priors, eval = has_metafor}
Kroupova2021_priors <- brma(
  yi = yi, vi = vi, measure = "ZCOR",
  data = Kroupova2021, only_priors = TRUE
)

print_prior(Kroupova2021_priors)
```

The same pattern applies to other standardized measures: compute the appropriate transformation in `escalc()` and pass the matching `measure` to `brma()`.

The model and its prior distributions then operate on the transformed scale, but posterior summaries can be reported back on the familiar original scale via the `output_measure` and `transform` arguments of the relevant summary functions (e.g., `pooled_effect()`, `marginal_means()`, and `plot()` functions).
For correlations fit with `measure = "ZCOR"`, set `output_measure = "COR"` to obtain a pooled correlation.
For log-scale measures (`"OR"`, `"RR"`, `"HR"`, `"IRR"`), set `transform = "EXP"` to obtain the corresponding ratio.

### Non-standardized inputs

Some studies report effects on a scale specific to the analysis: a regression coefficient on raw units, an unstandardized mean difference, or a measure outside the standardized catalogue.
Use `measure = "GEN"` for these.
Because the package cannot know what counts as a "moderate" effect on an arbitrary scale, the default prior scale must be supplied: either directly via `prior_unit_information_sd`, or implicitly by passing the per-study sample sizes in `ni`, from which the package estimates the UISD via `estimate_unit_information_sd()`.

The `Havrankova2025` dataset reports regression coefficients estimating the beauty premium across studies on a non-standardized scale.

```{r Havrankova-data}
data("Havrankova2025", package = "RoBMA")
head(Havrankova2025)
```

The simplest specification passes the sample sizes in `ni` so that the package estimates the UISD:

```{r Havrankova-priors}
Havrankova_priors <- brma(
  yi = y, sei = se, ni = N, measure = "GEN",
  data = Havrankova2025, only_priors = TRUE
)

print_prior(Havrankova_priors)
```

The UISD can also be set directly when sample sizes are unavailable, when an alternative scale is more appropriate, or when performing subgroup analyses (it is sensible to compute UISD based on the full dataset so the prior distributions match across models):

```{r Havrankova-manual-uisd-calc}
Havrankova_UISD <- estimate_unit_information_sd(
  sei = Havrankova2025$se,
  ni  = Havrankova2025$N
)
Havrankova_UISD
```

```{r Havrankova-manual-uisd}
Havrankova_manual_priors <- brma(
  yi = y, sei = se, measure = "GEN",
  prior_unit_information_sd = Havrankova_UISD,
  data = Havrankova2025, only_priors = TRUE
)

print_prior(Havrankova_manual_priors)
```

The manually specified value takes precedence over the automatic estimate.

### Meta-regression

Meta-regression introduces moderators to explain part of the between-study variation.
The default prior distributions for moderator coefficients reuse the same UISD machinery as the prior distributions on the average effect and heterogeneity, but two specification choices change how those prior distributions are interpreted:
the contrast scheme used for categorical moderators, and the standardization applied to continuous moderators.

We illustrate both with `Havrankova2025`, fitting two moderators at once: the categorical `facing_customer` (No / Some / Direct) and the continuous `N` (per-study sample size).
Each moderator's prior can be inspected separately by passing its name to `parameter_mods` in `print_prior()` or `plot_prior()`.

**Categorical moderators.**
The `set_contrast_factor_predictors` argument controls how categorical predictors are coded, applying a single contrast scheme jointly to every categorical moderator in the model (per-moderator overrides are described in *Per-moderator overrides* below).
The choice affects both how the intercept is interpreted and what the inclusion Bayes factor for the moderator tests.

With *treatment contrasts* (`"treatment"`, the default in `brma()` and matching base R), one level is chosen as a baseline.
The intercept is the average effect for the baseline level, and each remaining coefficient is the difference between its level and the baseline.
In parameter estimation, the regression coefficients are regularized towards the baseline.
In hypothesis testing, removing the moderator coefficient tests whether each level's effect differs from the baseline effect.
The choice of reference level matters for both the prior distribution on the intercept and the resulting inclusion Bayes factor.
Treatment contrasts are most natural when one level has a clear baseline interpretation (a control condition, an absence of treatment, a default policy).

With *mean-difference contrasts* (`"meandif"`, the default in `BMA()` and `RoBMA()`), the levels are treated as exchangeable.
The intercept is the adjusted average effect (averaged with equal weight across levels), and each coefficient describes how a given level deviates from the grand mean.
In parameter estimation, the regression coefficients are regularized towards the average effect.
In hypothesis testing, removing the moderator tests whether any level differs from the average effect, irrespective of which level was chosen as a label.
This is the recommended default for model-averaged inference because it gives a well-defined "is there moderation by this factor?" interpretation that does not depend on relabelling the levels.

The two contrast schemes can be compared on the categorical moderator while the continuous moderator is included in the same model:

```{r Havrankova-treatment}
Havrankova_treatment <- brma(
  yi = y, sei = se, ni = N, measure = "GEN",
  mods = ~ facing_customer + N,
  set_contrast_factor_predictors = "treatment",
  data = Havrankova2025, only_priors = TRUE
)

print_prior(Havrankova_treatment, parameter_mods = "facing_customer")
```

```{r Havrankova-meandif}
Havrankova_meandif <- brma(
  yi = y, sei = se, ni = N, measure = "GEN",
  mods = ~ facing_customer + N,
  set_contrast_factor_predictors = "meandif",
  data = Havrankova2025, only_priors = TRUE
)

print_prior(Havrankova_meandif, parameter_mods = "facing_customer")
```

The treatment-contrast version places a `Normal(0, 12.9)` prior distribution on each non-reference level's deviation from the reference level.
The mean-difference version places a multivariate `mNormal(0, 12.9)` prior distribution on the deviations of all levels from the grand mean.
The continuous moderator `N` is unaffected by the contrast choice and receives the same default prior distribution in both fits.

**Continuous moderators.**
Continuous predictors are standardized internally so that the default prior scale applies regardless of the predictor's units.
The default prior distribution on a meta-regression coefficient therefore describes a plausible change per one *standard deviation* of the predictor, not per one original unit.
By default, `plot_prior()` displays the prior distribution on this standardized scale; pass `standardized_coefficients = FALSE` to obtain the prior distribution per one original unit of the moderator.

We inspect the prior distribution on the continuous moderator `N` from the same `Havrankova_meandif` fit:

```{r Havrankova-continuous}
plot_prior(Havrankova_meandif, parameter_mods = "N")
plot_prior(Havrankova_meandif, parameter_mods = "N", 
  standardized_coefficients = FALSE)
```

The first plot shows the prior distribution on the standardized scale used internally (a plausible change per one standard deviation of `N`).
The second shows the same prior distribution on the original scale of `N` (a plausible change per one unit of `N`); because `N` ranges over orders of magnitude, the per-unit prior distribution is correspondingly tight.

### Per-moderator overrides

The `set_contrast_factor_predictors` argument is a joint setting and `rescale_priors` scales every default prior distribution by the same factor.
Finer control is available through `prior_mods`, which accepts a named list keyed by moderator name.
Each entry can be a regular `prior()` (for continuous moderators) or a `prior_factor()` with its own `contrast` setting (for categorical moderators), so different moderators can use different contrast schemes and prior scales within the same model.

For example, override only the prior distribution on `facing_customer`, switching its contrast and tightening its scale, while leaving `N` on the default per-standard-deviation prior distribution:

```{r Havrankova-per-moderator}
Havrankova_per_mod <- brma(
  yi = y, sei = se, ni = N, measure = "GEN",
  mods = ~ facing_customer + N,
  prior_mods = list(
    facing_customer = prior_factor(
      distribution = "mnormal",
      parameters = list(mean = 0, sd = 5),
      contrast   = "meandif"
    )
  ),
  data = Havrankova2025, only_priors = TRUE
)

print_prior(Havrankova_per_mod)
```

The `facing_customer` prior distribution is now the user-specified `mNormal(0, 5)` mean-difference contrast, while `N` retains the default per-standard-deviation prior distribution.
The named list extends naturally: with multiple moderators, each entry can specify its own prior family, scale, and (for categorical moderators) `contrast` value.

## Informed Empirical Prior Distributions

Default prior distributions are designed to work broadly.
*Informed empirical prior distributions* use previous meta-analyses to define what effects and heterogeneity values have looked like in a specific research area.

For medical meta-analyses, the `RoBMA` R package can use empirical prior distributions based on the Cochrane Database of Systematic Reviews.
Continuous outcomes are described in @bartos2021bayesian, and binary or time-to-event outcomes in @bartos2023empirical.
The shortcut is `prior_informed_field = "medicine"`.
If no subfield is given, the package uses the full Cochrane database.

```{r bcg-informed, eval = has_bcg}
bcg_informed_priors <- brma(
  yi = yi, vi = vi, measure = "RR",
  prior_informed_field    = "medicine",
  prior_informed_subfield = "Cochrane",
  data = bcg, only_priors = TRUE
)

print_prior(bcg_informed_priors)
```

The package also re-exports `prior_informed()` from the `BayesTools` package.
This can be used for other informed prior distributions from psychology, such as the informed prior distribution for the effect size from @gronau2017bayesian or the heterogeneity prior distribution from @erp2017estimates.
Pass them directly to the relevant arguments:

```{r psychology-informed, eval = has_metafor}
psychology_priors <- brma(
  yi = yi, vi = vi, measure = "ZCOR",
  prior_effect        = prior_informed("Oosterwijk"),
  prior_heterogeneity = prior_informed("van Erp", parameter = "heterogeneity"),
  data = Kroupova2021, only_priors = TRUE
)

print_prior(psychology_priors)
```


## Custom Prior Distributions

Custom prior distributions are useful when there is a specific substantive reason for the scale of the effect, heterogeneity, moderator coefficients, or when specifying highly informed hypothesis tests.
They are created with `prior()` and related helper functions.

```{r custom-effect-priors, eval = has_bcg}
bcg_custom_priors <- brma(
  yi = yi, vi = vi, measure = "RR",
  prior_effect        = prior(
    distribution = "normal",
    parameters = list(mean = 0, sd = 0.5)
  ),
  prior_heterogeneity = prior(
    distribution = "normal",
    parameters = list(mean = 0, sd = 0.25),
    truncation = list(lower = 0, upper = Inf)
  ),
  data = bcg, only_priors = TRUE
)

print_prior(bcg_custom_priors)
```


## Prior Distributions and Bayesian Model Averaging

The examples above use `brma()` because the prior distributions are easiest to understand in a single model.
`BMA()` and `RoBMA()` use these prior distributions inside Bayesian model averaging: they average across models that include or exclude the effect, heterogeneity, and any moderators.
Each component is then represented by a *spike-and-slab* prior distribution: a point prior distribution at zero (the spike) for the model that excludes the parameter, and a continuous prior distribution (the slab) for the model that includes it.

```{r bma-priors, eval = has_bcg}
bcg_bma_priors <- BMA(
  yi = yi, vi = vi, measure = "RR",
  data = bcg, only_priors = TRUE
)

print_prior(bcg_bma_priors)
```

By default, `BMA()` and `RoBMA()` include the null hypothesis (a spike at zero) for every component.
Removing it (for example by passing `prior_effect_null = NULL`) drops the no-effect model and yields estimation under an assumed effect rather than a test of effect versus no effect.

```{r bma-no-null, eval = has_bcg}
bcg_estimation_priors <- BMA(
  yi = yi, vi = vi, measure = "RR",
  prior_effect_null = NULL,
  data = bcg, only_priors = TRUE
)

print_prior(bcg_estimation_priors, parameter = "mu")
```

## Publication-Bias Prior Distributions

Publication-bias prior distributions describe the bias-adjustment component (`prior_bias`) and are independent of the prior distributions on the effect, heterogeneity, and moderators.
Because they typically attach to specialised single-model fits or to model-averaged ensembles, the argument-level details are documented in `?publication_bias_prior_specification`, with worked examples in the corresponding vignettes.

For *single-model* bias adjustment with `bPET()`, `bPEESE()`, or `bselmodel()`, see the [*Publication-Bias Adjustment*](v11-metafor-parity-publication-bias.html) vignette: it lists the default `prior_bias` for each function, the rationale for the one-sided defaults, and the flexible alternatives that drop the truncation or the monotonicity constraint.
For *model-averaged* bias adjustment with `RoBMA()`, see the [*Robust Bayesian Meta-Analysis*](v21-robust-bayesian-meta-analysis.html) vignette: it describes the `model_type` presets (`"PSMA"`, `"PP"`, `"6w"`, `"2w"`), the weight functions and PET/PEESE components they include, and how to build a custom ensemble via `prior_bias` and `prior_bias_null`.

## General Considerations and Reporting

Specifying prior distributions is a substantive modelling decision rather than a technical detail, and a few principles deserve to be stated explicitly even at the cost of repeating points made above.

**Wide prior distributions are not "uninformative" in testing or model averaging.**
A very wide prior distribution on the alternative can look like a conservative choice for estimation, but it strongly biases hypothesis tests in favour of the null and shrinks model-averaged estimates toward zero, because the alternative model receives little posterior probability.
Bayes factors and model-averaged inference depend on the prior distribution under the alternative just as much as on the data.

**Report everything.**
Fully report the prior distributions.
Run a *prior sensitivity analysis* by re-fitting under at least one alternative reasonable prior distribution; if the conclusions are robust, the analysis stands on firmer ground.

**Decide prior distributions before looking at the data.**
Preregistering the analysis plan, including the prior distributions, avoids the temptation to tune prior distributions after seeing partial results and keeps the inference honestly Bayesian.

## References

---
title: "Guide to RoBMA Vignettes"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{ Guide to RoBMA Vignettes}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

The `RoBMA` package provides a comprehensive set of vignettes to help users navigate different aspects of Robust Bayesian Meta-Analysis. This guide outlines the available vignettes and their specific focus to help you find the relevant information for your analysis.

## Introductory Vignettes

### [Tutorial: Adjusting for Publication Bias in JASP and R](Tutorial.html)
This is the main introduction to the RoBMA framework. It covers the basics of adjusting for publication bias using selection models, PET-PEESE, and Robust Bayesian Meta-Analysis. It is the recommended starting point for new users.

### [Reproducing Bayesian Model-Averaged Meta-Analysis](ReproducingBMA.html)
This vignette demonstrates how to perform a classic Bayesian model-averaged meta-analysis. It focuses on reproducing standard BMA results and understanding the core components of the method.

## Advanced Modeling Features

### [Robust Bayesian Model-Averaged Meta-Regression](MetaRegression.html)
Learn how to incorporate moderators into your meta-analysis using `RoBMA.reg()`. This vignette explains how to fit meta-regression models to account for heterogeneity explained by study-level covariates.

### [Multilevel Robust Bayesian Meta-Analysis](MultilevelRoBMA.html)
This vignette demonstrates how to perform multilevel meta-analysis to account for dependent effect sizes (e.g., multiple estimates from the same study). It uses the spike-and-slab algorithm (`algorithm = "ss"`) to efficiently estimate models with within-study and between-study heterogeneity while adjusting for publication bias.

### [Multilevel Robust Bayesian Model-Averaged Meta-Regression](MultilevelRoBMARegression.html)
This vignette demonstrates how to perform multilevel meta-regression. In addition, it illustrates how to rescale default prior distributions to work with non-standardized effect sizes.

### [Z-Curve Publication Bias Diagnostics](ZCurveDiagnostics.html)
This vignette details the use of meta-analytic z-curves for diagnosing publication bias. It explains how to interpret z-curve plots and statistics provided by the package.

## Specialized Applications

### [Informed Bayesian Model-Averaged Meta-Analysis in Medicine](MedicineBMA.html)
This vignette focuses on applying RoBMA in medical contexts. It discusses the use of informed priors tailored for medical research questions and continuous outcomes.

### [Informed Bayesian Model-Averaged Meta-Analysis with Binary Outcomes](MedicineBiBMA.html)
Similar to the Medicine BMA vignette, but specifically for binary outcomes. It covers the `BiBMA` models (Binomial-Normal) and appropriate prior settings for medical meta-analysis with binary data.

## Customization and Performance

### [Fitting Custom Meta-Analytic Ensembles](CustomEnsembles.html)
For advanced users who need to go beyond the default model ensembles. This vignette demonstrates how to customize the ensemble of models, including specifying custom priors and model combinations.

### [Fast Robust Bayesian Meta-Analysis via Spike and Slab Algorithm](FastRoBMA.html)
For computationally intensive problems or quick approximations, the "spike-and-slab" algorithm (`algorithm = "ss"`) can be used. This vignette explains how to use this faster alternative to the default bridge sampling approach.

### [Hierarchical Bayesian Model-Averaged Meta-Analysis](HierarchicalBMA.html)
This vignette introduces multilevel models. It shows how to handle dependencies in the data (e.g., multiple effect sizes from the same study) using the `study_ids` argument to specify a hierarchical structure. Note that this vignette relies on multivariate parameterization that is relevant only for the bridge sampling algorithm. However, it is still helpful for describing the parameterization.

---
title: "Multilevel Robust Bayesian Model-Averaged Meta-Regression"
author: "František Bartoš"
date: "17th of December 2025 (updated: 29th of April 2026)"
output:
  rmarkdown::html_vignette:
    self_contained: yes
bibliography: ../inst/REFERENCES.bib
csl: ../inst/apa.csl
link-citations: true
vignette: >
  %\VignetteIndexEntry{Multilevel Robust Bayesian Model-Averaged Meta-Regression}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
  %\VignetteEngine{knitr::rmarkdown_notangle}
---

```{r child = "_vignette-nowrap.md", echo = FALSE, eval = TRUE}
```

```{r setup, include = FALSE}
source("vignette-cache.R", local = knitr::knit_global())

cached_fits <- vignette_cache(
  name     = "v33-robma-multilevel-metaregression",
  objects  = "fit_reg"
)

knitr::opts_chunk$set(
  collapse    = TRUE,
  comment     = "#>",
  eval        = vignette_cache_eval(cached_fits),
  message     = FALSE,
  warning     = FALSE,
  fig.width   = 7,
  fig.height  = 3.5,
  dev         = "png",
  fig.retina  = 3
)
if (.Platform$OS.type == "windows") {
  knitr::opts_chunk$set(dev.args = list(type = "cairo"))
}
```

```{r load-models, include = FALSE}
library("RoBMA")

data("Havrankova2025", package = "RoBMA")

unit_information_sd_361 <- estimate_unit_information_sd(
  sei = Havrankova2025$se,
  ni  = Havrankova2025$N
)
prior_scale_361 <- unit_information_sd_361 * 0.5

vignette_cache_load(cached_fits)
```

```{r precompute-models, include = FALSE, eval = FALSE}
library("RoBMA")

data("Havrankova2025", package = "RoBMA")

unit_information_sd_361 <- estimate_unit_information_sd(
  sei = Havrankova2025$se,
  ni  = Havrankova2025$N
)
prior_scale_361 <- unit_information_sd_361 * 0.5

prior_effect_361        <- prior("normal",   list(mean = 0, sd = prior_scale_361))
prior_heterogeneity_361 <- prior("invgamma", list(shape = 1, scale = 0.15 * prior_scale_361))
prior_mods_361          <- list(
  facing_customer = prior_factor("mnormal", list(mean = 0, sd = 0.25 * prior_scale_361), contrast = "meandif")
)
prior_bias_361          <- list(
  prior_weightfunction("two-sided", c(0.05),             wf_cumulative(c(1, 1)),       prior_weights = 1/12),
  prior_weightfunction("two-sided", c(0.05, 0.10),       wf_cumulative(c(1, 1, 1)),    prior_weights = 1/12),
  prior_weightfunction("one-sided", c(0.05),             wf_cumulative(c(1, 1)),       prior_weights = 1/12),
  prior_weightfunction("one-sided", c(0.025, 0.05),      wf_cumulative(c(1, 1, 1)),    prior_weights = 1/12),
  prior_weightfunction("one-sided", c(0.05, 0.5),        wf_cumulative(c(1, 1, 1)),    prior_weights = 1/12),
  prior_weightfunction("one-sided", c(0.025, 0.05, 0.5), wf_cumulative(c(1, 1, 1, 1)), prior_weights = 1/12),
  prior_PET("Cauchy", list(0, 1), list(0, Inf), 1/4),
  prior_PEESE("Cauchy", list(0, 5 / prior_scale_361), list(0, Inf), 1/4)
)

fit_reg <- RoBMA(
  yi = y, sei = se, ni = N, measure = "GEN",
  mods = ~ facing_customer,
  cluster = study_id,
  prior_unit_information_sd = unit_information_sd_361,
  prior_effect              = prior_effect_361,
  prior_heterogeneity       = prior_heterogeneity_361,
  prior_mods                = prior_mods_361,
  prior_bias                = prior_bias_361,
  sample = 20000, burnin = 10000, adapt = 10000,
  thin = 5, parallel = TRUE, seed = 1,
  data = Havrankova2025
)

vignette_cache_save(cached_fits)
```

**This vignette accompanies the manuscript [Robust Bayesian Multilevel Meta-Analysis: Adjusting for Publication Bias in the Presence of Dependent Effect Sizes](https://doi.org/10.31234/osf.io/9tgp2_v1) preprinted at *PsyArXiv* [@bartos2025robust].**

This vignette reproduces the second example from the manuscript, re-analyzing data from @havrankova2025beauty on the relationship between beauty and professional success (1,159 effect sizes from 67 studies).
We investigate whether the effect depends on the type of customer contact ("no", "some", or "direct").
For the first example describing a publication-bias-adjusted multilevel meta-analysis, see the [*Multilevel Robust Bayesian Meta-Analysis*](v32-robma-multilevel.html) vignette.

## Rescaling Prior Distributions for Non-Standardized Effect Sizes

The effect sizes in this dataset are not standardized: they are percent increases in earnings associated with a one-standard-deviation increase in beauty.
Therefore, the prior distributions need to be defined on this outcome scale.
Overly narrow prior distributions would make it difficult to find evidence for the effect because both models would predict effects very close to zero.
Overly wide prior distributions would bias the results towards the null hypothesis because alternative models would predict unrealistically large effects.
The 4.0 version of the `RoBMA` R package handles this automatically by estimating the unit-information standard deviation (UISD) and rescaling the prior distributions appropriately (see the [*Prior Distributions*](v01-prior-distributions.html) vignette).

To reproduce the RoBMA 3.6.1 version results that used different prior distribution formulation and scaling settings we need to specify the prior distributions manually.[^scaling]

[^scaling]: This reproduces the RoBMA 3.6.1 vignette results. It does not match the current default UISD scaling conventions exactly; in particular, the current default moderator and PEESE scales differ and are therefore specified below.

```{r prior-specification}
prior_effect_361        <- prior("normal",   list(mean = 0, sd = 25.8))
prior_heterogeneity_361 <- prior("invgamma", list(shape = 1, scale = 0.15 * 25.8))
prior_mods_361          <- list(
  facing_customer = prior_factor("mnormal", list(mean = 0, sd = 0.25 * 25.8), contrast = "meandif")
)
prior_bias_361          <- list(
  prior_weightfunction("two-sided", c(0.05),             wf_cumulative(c(1, 1)),       prior_weights = 1/12),
  prior_weightfunction("two-sided", c(0.05, 0.10),       wf_cumulative(c(1, 1, 1)),    prior_weights = 1/12),
  prior_weightfunction("one-sided", c(0.05),             wf_cumulative(c(1, 1)),       prior_weights = 1/12),
  prior_weightfunction("one-sided", c(0.025, 0.05),      wf_cumulative(c(1, 1, 1)),    prior_weights = 1/12),
  prior_weightfunction("one-sided", c(0.05, 0.5),        wf_cumulative(c(1, 1, 1)),    prior_weights = 1/12),
  prior_weightfunction("one-sided", c(0.025, 0.05, 0.5), wf_cumulative(c(1, 1, 1, 1)), prior_weights = 1/12),
  prior_PET("Cauchy", list(0, 1), list(0, Inf), 1/4),
  prior_PEESE("Cauchy", list(0, 5 / 25.8), list(0, Inf), 1/4)
)
```

We fit the robust Bayesian multilevel meta-regression using the `RoBMA()` function.
We specify the effect sizes via `yi`, standard errors via `sei`, and the effect-size scale via the `measure = "GEN"` argument.
The moderator `facing_customer` is specified via the formula syntax using the `mods` argument.
The multilevel structure, i.e., effect sizes nested within studies `study_id`, is specified using the `cluster` argument.
Finally, we specify the RoBMA 3.6.1 equivalent prior distributions in the `prior_*` arguments (omitting them would result in a model fit using the current default prior distributions based on the UISD).

```{r fit-reg, eval = FALSE}
fit_reg <- RoBMA(
  yi = y, sei = se, measure = "GEN",
  mods = ~ facing_customer,
  cluster = study_id,
  prior_effect              = prior_effect_361,
  prior_heterogeneity       = prior_heterogeneity_361,
  prior_mods                = prior_mods_361,
  prior_bias                = prior_bias_361,
  sample = 20000, burnin = 10000, adapt = 10000,
  thin = 5, parallel = TRUE, seed = 1,
  data = Havrankova2025
)
```

The meta-regression can take more than an hour to run with parallel processing enabled.

## Interpreting the Results

The `summary()` function reports inclusion Bayes factors and model parameters.

```{r summary}
summary(fit_reg, include_mcmc_diagnostics = FALSE)
```

The effect-size estimates are obtained with dedicated helpers: `pooled_effect()` for the average effect and `marginal_means()` for the moderator levels.

The multilevel RoBMA meta-regression finds extreme evidence for the presence of the average effect, moderation by the degree of consumer contact, between-study heterogeneity, and publication bias (all BFs > 10,000).

```{r pooled-effect}
pooled_effect(fit_reg)
```

The model-averaged effect estimate is accompanied by considerable overall heterogeneity and a cluster-level allocation close to 1, indicating a high degree of similarity among estimates from the same study.
MCMC diagnostics are included in `summary()` by default, and chain-level checks are available via `plot_diagnostic_trace()`, `plot_diagnostic_density()`, and `plot_diagnostic_autocorrelation()`.

The average effect is moderated by the degree of consumer contact. To examine the effect at each level of the moderator, we use the estimated marginal means.

```{r marginal-means}
mm_reg <- marginal_means(fit_reg)
summary(mm_reg)
```

The estimated marginal means reveal moderate evidence for the absence of an effect for jobs with no consumer contact and extreme evidence for an effect for jobs with some or direct consumer contact.

## Visual Fit Assessment

We visualize model fit using the meta-analytic zplot [@bartos2025zcurve]; see the [*Zplot Publication-Bias Diagnostics*](v36-zplot.html) vignette for more details.
The zplot shows the distribution of the observed $z$-statistics, computed as the effect size divided by the standard error, with dotted red horizontal lines highlighting the typical steps on which publication bias operates.
These are $z = \pm 1.65$ and $z = \pm 1.96$, corresponding to marginally significant and statistically significant $p$-values, and $z = 0$, corresponding to selection on the direction of the effect.

```{r zplot, fig.width = 6, fig.height = 4}
par(mar = c(4, 4, 0, 0))
zplot(fit_reg, by.hist = 0.25, plot_extrapolation = FALSE, from = -4, to = 8)
```

We can notice clear discontinuities corresponding to selection on the direction of the effect and marginal significance.
The posterior predictive density from the multilevel RoBMA meta-regression indicates that the model approximates the observed data reasonably well, capturing the discontinuities and the proportion of statistically significant results.
This supports the model fit even after adjusting for publication bias.

## References

---
title: "Tutorial: Adjusting for Publication Bias in JASP and R - Selection Models, PET-PEESE, and Robust Bayesian Meta-Analysis"
author: "František Bartoš, Maximilian Maier, Daniel S. Quintana & Eric-Jan Wagenmakers"
date: "2022"
output:
  rmarkdown::html_vignette:
    self_contained: yes
bibliography: ../inst/REFERENCES.bib
csl: ../inst/apa.csl
vignette: >
  %\VignetteIndexEntry{Tutorial: Adjusting for Publication Bias in JASP and R - Selection Models, PET-PEESE, and Robust Bayesian Meta-Analysis}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
  %\VignetteEngine{knitr::rmarkdown_notangle}
---

```{r setup, include = FALSE}
is_check <- ("CheckExEnv" %in% search()) ||
              any(c("_R_CHECK_TIMINGS_", "_R_CHECK_LICENSE_") %in% names(Sys.getenv())) ||
              !file.exists("../models/Tutorial/fit_RoBMA_Lui2015.RDS")
knitr::opts_chunk$set(
  collapse  = TRUE,
  comment   = "#>",
  eval      = !is_check,
  dev      = "png")
if(.Platform$OS.type == "windows"){
  knitr::opts_chunk$set(dev.args = list(type = "cairo"))
}
```
```{r include=FALSE}
knitr::opts_chunk$set(echo = TRUE)
data("Lui2015", package = "RoBMA")
df <- Lui2015
# preload the fitted model
fit_RoBMA  <- readRDS(file = "../models/Tutorial/fit_RoBMA_Lui2015.RDS")
fit_RoBMA2 <- readRDS(file = "../models/Tutorial/fit_RoBMA_perinull_Lui2015.RDS")
```
```{r include = FALSE, eval = FALSE}
# R package version updating
library(RoBMA)
data("Lui2015", package = "RoBMA")
df <- Lui2015
fit_RoBMA <- RoBMA(r = df$r, n = df$n, seed = 1, model = "PSMA", parallel = TRUE, save = "min")

fit_RoBMA2 <- RoBMA(r = df$r, n = df$n, seed = 2, parallel = TRUE, save = "min",
                    priors_effect      = prior("normal", parameters = list(mean = 0.60, sd = 0.20), truncation = list(0, Inf)),
                    priors_effect_null = prior("normal", parameters = list(mean = 0,    sd = 0.10)))

saveRDS(fit_RoBMA,  file = "../models/Tutorial/fit_RoBMA_Lui2015.RDS")
saveRDS(fit_RoBMA2, file = "../models/Tutorial/fit_RoBMA_perinull_Lui2015.RDS")
```

**This R markdown file accompanies the tutorial [Adjusting for publication bias in JASP and R: Selection models, PET-PEESE, and robust Bayesian meta-analysis](https://doi.org/10.1177/25152459221109259) published in *Advances in Methods and Practices in Psychological Science* [@bartos2020adjusting].**


The following R-markdown file illustrates how to:

- Load a CSV file into R,
- Transform effect sizes,
- Perform a random effect meta-analysis,
- Adjust for publication bias with:
  - PET-PEESE [@stanley2014meta; @stanley2017limitations], 
  - Selection models [@iyengar1988selection; @vevea1995general],
  - Robust Bayesian meta-analysis (RoBMA) [@maier2020robust; @bartos2021no].

See the full paper for additional details regarding the data set, methods, and interpretation.


### Set-up

Before we start, we need to install `JAGS` (which is needed for installation of the `RoBMA` package) and the R packages that we use in the analysis. Specifically the `RoBMA`, `weightr`, and `metafor` R packages.

JAGS can be downloaded from the [JAGS website](https://sourceforge.net/projects/mcmc-jags/). Subsequently, we install the R packages with the `install.packages()` function.

```{r
install.packages(c("RoBMA", "weightr", "metafor"))
```
If you happen to use the new M1 Mac machines with Apple silicon, see [this blogpost](https://www.dsquintana.blog/jags-apple-silicon-m1-mac/) outlining how to install JAGS on M1. In short, you will have to install Intel version of R (Intel/x86-64) from [CRAN](https://cran.r-project.org/bin/macosx/), not the Arm64 (Apple silicon) version. Note that there might have been some changes in the installation process since the blogpost was written and there might be a JAGS version compatible with Apple silicon available now.

Once all of the packages are installed, we can load them into the workspace with the `library()` function.

```{r, message = FALSE}
library("metafor")
library("weightr")
library("RoBMA")
```


### Lui (2015)
@lui2015intergenerational studied how the acculturation mismatch (AM) that is the result of the contrast between the collectivist cultures of Asian and Latin immigrant groups and the individualist culture in the United States correlates with intergenerational cultural conflict (ICC). @lui2015intergenerational meta-analyzed 18 independent studies correlating AM with ICC. A standard reanalysis indicates a significant effect of AM on increased ICC, r = 0.250, p < .001.

#### Data manipulation
First, we load the Lui2015.csv file into R with the `read.csv()` function and inspect the first six data entries with the `head()` function (the data set is also included in the package and can be accessed via the `data("Lui2015", package = "RoBMA")` call).

```r
df <- read.csv(file = "Lui2015.csv")
```
```{r}
head(df)
```

We see that the data set contains three columns. The first column called `r` contains the effect sizes coded as correlation coefficients, the second column called `n` contains the sample sizes, and the third column called `study` contains names of the individual studies.

We can access the individual variables using the data set name and the dollar (`$`) sign followed by the name of the column. For example, we can print all of the effect sizes with the `df$r` command.

```{r}
df$r
```

The printed output shows that the data set contains mostly positive effect sizes with the largest correlation coefficient r = 0.54.

#### Effect size transformations
Before we start analyzing the data, we transform the effect sizes from correlation coefficients $\rho$ to Fisher's *z*. Correlation coefficients are not well suited for meta-analysis because (1) they are bounded to a range (-1, 1) with non-linear increases near the boundaries and (2) the standard error of the correlation coefficients is related to the effect size. Fisher's *z* transformation mitigates both issues. It unwinds the (-1, 1) range to ($-\infty$, $\infty$), makes the sampling distribution approximately normal, and breaks the dependency between standard errors and effect sizes. 

To apply the transformation, we use the `combine_data()` function from the `RoBMA` package. We pass the correlation coefficients into the `r` argument, the sample sizes to the `n` argument, and set the `transformation` argument to `"fishers_z"` (the `study_names` argument is optional). The function `combine_data()` then saves the transformed effect size estimates into a data frame called `dfz`, where the `y` column corresponds to Fisher's *z* transformation of the correlation coefficient and `se` column corresponds to the standard error of Fisher's *z*.

```{r}
dfz <- combine_data(r = df$r, n = df$n, study_names = df$study, transformation = "fishers_z")
head(dfz)
```

We can also transform the effect sizes according to Cohen's *d* transformation (which we utilize later to fit the selection models).

```{r}
dfd <- combine_data(r = df$r, n = df$n, study_names = df$study, transformation = "cohens_d")
head(dfd)
```


#### Re-analysis with random effect meta-analysis
We now estimate a random effect meta-analysis with the `rma()` function imported from the `metafor` package [@metafor] and verify that we arrive at the same results as reported in the @lui2015intergenerational paper. The `yi` argument is used to pass the column name containing effect sizes, the `sei` argument is used to pass the column name containing standard errors, and the `data` argument is used to pass the data frame containing both variables.

```{r}
fit_rma <- rma(yi = y, sei = se, data = dfz)
fit_rma
```

Indeed, we find that the effect size estimate from the random effect meta-analysis corresponds to the one reported in @lui2015intergenerational. It is important to remember that we used Fisher's *z* to estimate the models; therefore, the estimated results are on the Fisher's *z* scale. To transform the effect size estimate to the correlation coefficients, we can use the `z2r()` function from the `RoBMA` package,

```{r}
z2r(fit_rma$b)
```

Transforming the effect size estimate results in the correlation coefficient $\rho$ = 0.25.


### PET-PEESE
The first publication bias adjustment that we perform is PET-PEESE. PET-PEESE adjusts for the relationship between effect sizes and standard errors. To our knowledge, PET-PEESE is not currently implemented in any R-package. However, since PET and PEESE are weighted regressions of effect sizes on standard errors (PET) or standard errors squared (PEESE), we can estimate both PET and PEESE models with the `lm()` function. Inside the `lm()` function call, we specify that `y` is the response variable (left hand side of the `~` sign) and `se` is the predictor (the right-hand side). Furthermore, we specify the `weights` argument that allows us to weight the meta-regression by inverse variance and set the `data = dfz` argument, which specifies that all of the variables come from the transformed, `dfz`, data set. 

```{r}
fit_PET <- lm(y ~ se, weights = 1/se^2, data = dfz)
summary(fit_PET)
```

The `summary()` function allows us to explore details of the fitted model. The `(Intercept)` coefficient refers to the meta-analytic effect size (corrected for the correlation with standard errors). Again, it is important to keep in mind that the effect size estimate is on the Fisher's *z* scale. We obtain the estimate on correlation scale with the `z2r()` function (we pass the estimated effect size using the `summary(fit_PET)$coefficients["(Intercept)", "Estimate"]` command, which extracts the estimate from the fitted model, it is equivalent to simply pasting the value directly `z2r(-0.0008722083)`).

```{r}
z2r(summary(fit_PET)$coefficients["(Intercept)", "Estimate"])
```
Since the Fisher's *z* transformation is almost linear around zero, we obtain an almost identical estimate.

More importantly, since the test for the effect size with PET was not significant at $\alpha = .10$, we interpret the PET model. However, if the test for effect size were significant, we would fit and interpret the PEESE model. The PEESE model can be fitted in an analogous way, by replacing the predictor of standard errors with standard errors squared (we need to wrap the `se^2` predictor in `I()` that tells R to square the predictor prior to fitting the model).

```{r}
fit_PEESE <- lm(y ~ I(se^2), weights = 1/se^2, data = dfz)
summary(fit_PEESE)
```


### Selection models
The second publication bias adjustment that we will perform is selection models. Selection models adjust for the different publication probabilities in different *p*-value intervals. Selection models are implemented in `weightr` package (`weightfunct()` function; @weightr) and newly also in the `metafor` package (`selmodel()` function; @metafor). First, we use the `weightr` implementation and fit the "4PSM" selection model that specifies three distinct *p*-value intervals: (1) covering the range of significant *p*-values for effect sizes in the expected direction (0.00-0.025), (2) covering the range of "marginally" significant *p*-values for effect sizes in the expected direction (0.025-0.05), and (3) covering the range of non-significant *p*-values (0.05-1). We use Cohen's *d* transformation of the correlation coefficients since it is better at maintaining the distribution of test statistics. To fit the model, we need to pass the effect sizes (`dfd$y`) into the `effect` argument and variances (`dfd$se^2`) into the `v` argument (note that we need to pass the vector of values directly since the `weightfunct()` function does not allow us to pass the data frame directly as did the previous functions). We further set `steps = c(0.025, 0.05)` to specify the appropriate cut-points (note that the steps correspond to one-sided *p*-values), and we set `table = TRUE` to obtain the frequency of *p* values in each of the specified intervals. 

```{r}
fit_4PSM <- weightfunct(effect = dfd$y, v = dfd$se^2, steps = c(0.025, 0.05), table = TRUE)
fit_4PSM
```

Note the warning message informing us about the fact that our data do not contain a sufficient number of *p*-values in one of the *p*-value intervals. The model output obtained by printing the fitted model object `fit_4PSM` shows that there is only one *p*-value in the (0.025, 0.05) interval. We can deal with this issue by joining the "marginally" significant and non-significant *p*-value interval, resulting in the "3PSM" model.

```{r}
fit_3PSM <- weightfunct(effect = dfd$y, v = dfd$se^2, steps = c(0.025), table = TRUE)
fit_3PSM
```

The new model does not suffer from the estimation problem due to the limited number of *p*-values in the intervals, so we can now interpret the results with more confidence. First, we check the test for heterogeneity that clearly rejects the null hypothesis `Q(df = 17) = 75.4999, $p$ = 5.188348e-09` (if we did not find evidence for heterogeneity, we could have proceeded by fitting the fixed effects version of the model by specifying the `fe = TRUE` argument). We follow by checking the test for publication bias which is a likelihood ratio test comparing the unadjusted and adjusted estimate  `X^2(df = 1) = 3.107176, $p$ = 0.077948`. The result of the test is slightly ambiguous -- we would reject the null hypothesis of no publication bias with $\alpha = 0.10$ but not with $\alpha = 0.05$.

If we decide to interpret the estimated effect size, we have to again transform it back to the correlation scale. However, this time we need to use the `d2r()` function since we supplied the effect sizes as Cohen's *d* (note that the effect size estimate corresponds to the second value in the `fit_3PSM$adj_est` object for the random effect model, alternatively, we could simply use `d2r(0.3219641)`).

```{r}
d2r(fit_3PSM$adj_est[2])
```

Alternatively, we could have conducted the analysis analogously but with the `metafor` package. First, we would fit a random effect meta-analysis with the Cohen's *d* transformed effect sizes.

```{r}
fit_rma_d <- rma(yi = y, sei = se, data = dfd)
```
Subsequently, we would have used the `selmodel` function, passing the estimated random effect meta-analysis object and specifying the `type = "stepfun"` argument to obtain a step weight function and setting the appropriate steps with the `steps = c(0.025)` argument. 

```{r}
fit_sel_d <- selmodel(fit_rma_d, type = "stepfun", steps = c(0.025))
fit_sel_d
```

The output verifies the results obtained in the previous analysis.


### Robust Bayesian meta-analysis
The third and final publication bias adjustment that we will perform is robust Bayesian meta-analysis (RoBMA). RoBMA uses Bayesian model-averaging to combine inference from both PET-PEESE and selection models. We use the `RoBMA` R package (and the `RoBMA()` function; @RoBMA) to fit the default 36 model ensemble (called RoBMA-PSMA) based on an orthogonal combination of models assuming the presence and absence of the effect size, heterogeneity, and publication bias. The models assuming the presence of publication bias are further split into six weight function models and models utilizing the PET and PEESE publication bias adjustment. To fit the model, we can directly pass the original correlation coefficients into the `r` argument and sample sizes into the `n` argument -- the `RoBMA()` function will internally transform them to the Fisher's *z* scale and, by default, return the estimates on a Cohen's *d* scale which is used to specify the prior distributions (both of these settings can be changed with the `prior_scale` and `transformation` arguments, and the output can be conveniently transformed later). We further set the `model` argument to `"PSMA"` to fit the 36 model ensemble and use the `seed` argument to make the analysis reproducible (it uses MCMC sampling in contrast to the previous methods). We turn on parallel estimation by setting the `parallel = TRUE` argument (the parallel processing might in some cases fail, try rerunning the model one more time or turning the parallel processing off in that case).
```r
fit_RoBMA <- RoBMA(r = df$r, n = df$n, seed = 1, model = "PSMA", parallel = TRUE)
```
This step can take some time depending on your CPU. For example, this will take approximately 1 minute on a fast CPU (e.g., AMD Ryzen 3900x 12c/24t) and up to ten minutes or longer on slower CPUs (e.g., 2.7 GHz Intel Core i5). 

We use the `summary()` function to explore details of the fitted model.

```{r}
summary(fit_RoBMA)
```

The printed output consists of two parts. The first table called `Components summary` contains information about the fitted models. It tells us that we estimated the ensemble with 18/36 models assuming the presence of an effect, 18/36 models assuming the presence of heterogeneity, and 32/36 models assuming the presence of the publication bias. The second column summarizes the prior model probabilities of models assuming either presence of the individual components -- here, we see that the presence and absence of the components are balanced a priori. The third column contains information about the posterior probability of models assuming the presence of the components -- we can observe that the posterior model probabilities of models assuming the presence of an effect slightly increased to 0.552. The last column contains information about the evidence in favor of the presence of any of those components. Evidence for the presence of an effect is undecided; the models assuming the presence of an effect are only 1.232 times more likely given the data than the models assuming the absence of an effect. However, we find overwhelming evidence in favor of heterogeneity, with the models assuming the presence of heterogeneity being 19,168 times more likely given the data than models assuming the absence of heterogeneity, and moderate evidence in favor of publication bias. 

As the name indicates, the second table called `Model-averaged estimates` contains information about the model-averaged estimates. The first row labeled `mu` corresponds to the model-averaged effect size estimate (on Cohen's *d* scale) and the second row labeled `tau` corresponds to the model-averaged heterogeneity estimates. Below are the estimated model-averaged weights for the different *p*-value intervals and the PET and PEESE regression coefficients. We convert the estimates to the correlation coefficients by adding the `output_scale = "r"` argument to the summary function.

```{r}
summary(fit_RoBMA, output_scale = "r")
```

Now, we have obtained the model-averaged effect size estimate on the correlation scale. If we were interested in the estimates model-averaging only across the models assuming the presence of an effect (for the effect size estimate), heterogeneity (for the heterogeneity estimate), and publication bias (for the publication bias weights and PET and PEESE regression coefficients), we could have added the `conditional = TRUE` argument to the summary function. A quick textual summary of the model can also be generated with the `interpret()` function.

```{r}
interpret(fit_RoBMA, output_scale = "r")
```

We can also obtain summary information about the individual models by specifying the `type = "models"` option. The resulting table shows the prior and posterior model probabilities and inclusion Bayes factors for the individual models (we also set the `short_name = TRUE` argument reducing the width of the output by abbreviating names of the prior distributions).

```{r}
summary(fit_RoBMA, type = "models", short_name = TRUE)
```

To obtain a summary of the individual model diagnostics, we set the `type = "diagnostics"` argument. The resulting table provides information about the maximum MCMC error, relative MCMC error, minimum ESS, and maximum R-hat when aggregating over the parameters of each model. As we can see, we obtain acceptable ESS and R-hat diagnostic values.

```{r}
summary(fit_RoBMA, type = "diagnostics")
```

Finally, we can also plot the model-averaged posterior distribution with the `plot()` function. We set the `prior = TRUE` argument to include the prior distribution as a grey line (and arrow for the point density at zero) and `output_scale = "r"` to transform the posterior distribution to the correlation scale (the default figure output would be on Cohen's *d* scale). (The `par(mar = c(4, 4, 1, 4))` call increases the left margin of the figure, so the secondary y-axis text is not cut off.)

```{r, dpi = 300, fig.width = 6, fig.height = 4, out.width = "50%", fig.align = "center"}
par(mar = c(4, 4, 1, 4))
plot(fit_RoBMA, prior = TRUE, output_scale = "r", )
```

#### Specifying Different Priors
The `RoBMA` package allows us to fit ensembles of highly customized meta-analytic models. Here we reproduce the ensemble for the perinull directional hypothesis test from the Appendix (see the R package vignettes for more examples and details). Instead of using the fully pre-specified model with the `model = "PSMA"` argument, we explicitly specify the prior distribution for models assuming presence of the effect with the `priors_effect = prior("normal", parameters = list(mean = 0.60, sd = 0.20), truncation = list(0, Inf))` argument, which assigns Normal(0.60, 0.20) distribution bounded to the positive numbers to the $\mu$ parameter (note that the prior distribution is specified on the Cohen's *d* scale, corresponding to 95% prior probability mass contained approximately in the $\rho$ = (0.10, 0.45) interval). Similarly, we also replace the default prior distribution for the models assuming absence of the effect with a perinull hypothesis with the `priors_effect_null = prior("normal", parameters = list(mean = 0, sd = 0.10))` argument that sets 95% prior probability mass to values in the $\rho$ = (-0.10, 0.10) interval.

```r
fit_RoBMA2 <- RoBMA(r = df$r, n = df$n, seed = 2, parallel = TRUE,
                    priors_effect      = prior("normal", parameters = list(mean = 0.60, sd = 0.20), truncation = list(0, Inf)),
                    priors_effect_null = prior("normal", parameters = list(mean = 0,    sd = 0.10)))
```

As previously, we can use the `summary()` function to inspect the model fit and verify that the specified models correspond to the settings.

```{r}
summary(fit_RoBMA2, type = "models")
```


### References

---
title: "Bayesian Meta-Analysis"
author: "František Bartoš"
date: "28th of April 2026"
output:
  rmarkdown::html_vignette:
    self_contained: yes
bibliography: ../inst/REFERENCES.bib
csl: ../inst/apa.csl
link-citations: true
vignette: >
  %\VignetteIndexEntry{Bayesian Meta-Analysis}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
  %\VignetteEngine{knitr::rmarkdown_notangle}
---

```{r child = "_vignette-nowrap.md", echo = FALSE, eval = TRUE}
```

```{r setup, include = FALSE}
source("vignette-cache.R", local = knitr::knit_global())

cached_fits <- vignette_cache(
  name     = "v02-bayesian-meta-analysis",
  objects  = c("fit1_brma", "fit2_brma", "fit3_brma"),
  packages = c("metafor", "metadat")
)

knitr::opts_chunk$set(
  collapse    = TRUE,
  comment     = "#>",
  eval        = vignette_cache_eval(cached_fits),
  message     = FALSE,
  warning     = FALSE,
  fig.width   = 7,
  fig.height  = 3.5,
  dev         = "png",
  fig.retina  = 3
)
if (.Platform$OS.type == "windows") {
  knitr::opts_chunk$set(dev.args = list(type = "cairo"))
}
```

```{r load-models, include = FALSE}
library("metafor")
library("RoBMA")

data("dat.bcg", package = "metadat")

dat <- metafor::escalc(
  ai = tpos, bi = tneg, ci = cpos, di = cneg,
  measure = "RR", data = dat.bcg
)

vignette_cache_load(cached_fits)
```

```{r precompute-models, include = FALSE, eval = FALSE}
library("metafor")
library("RoBMA")

data("dat.bcg", package = "metadat")

dat <- metafor::escalc(
  ai = tpos, bi = tneg, ci = cpos, di = cneg,
  measure = "RR", data = dat.bcg
)

fit1_brma <- brma(
  yi = yi, vi = vi, measure = "RR",
  data = dat, seed = 1
)
fit1_brma <- add_loo(fit1_brma)
fit1_brma <- add_marglik(fit1_brma)

fit2_brma <- brma(
  yi = yi, vi = vi, measure = "RR",
  mods = ~ ablat,
  data = dat, seed = 1
)
fit2_brma <- add_loo(fit2_brma)
fit2_brma <- add_marglik(fit2_brma)

fit3_brma <- brma(
  yi = yi, vi = vi, measure = "RR",
  mods = ~ ablat + alloc,
  data = dat, seed = 1
)
fit3_brma <- add_loo(fit3_brma)
fit3_brma <- add_marglik(fit3_brma)

vignette_cache_save(cached_fits)
```

This vignette is a starting point for using the `RoBMA` R package [@RoBMA].
We walk through the BCG vaccine dataset from the `metafor` package [@metafor], showing each `brma()` call alongside its `metafor::rma()` counterpart and building from a simple random-effects model up to meta-regression with a continuous and a categorical moderator.
The interface is intentionally close to the `metafor` package: the same effect-size columns, moderator formulas, and standard plotting and diagnostic functions apply.

All `brma()` calls keep the default prior distributions; see the [*Prior Distributions*](v01-prior-distributions.html) vignette for the prior definitions and customization options.

## Random-Effects Model

The BCG vaccine dataset from the `metadat` package contains 13 randomized trials on tuberculosis prevention.
`escalc()` computes log risk ratios `yi` and their sampling variances `vi`.

```{r escalc}
data("dat.bcg", package = "metadat")

dat <- metafor::escalc(
  ai = tpos, bi = tneg, ci = cpos, di = cneg,
  measure = "RR", data = dat.bcg
)
```

`metafor::rma()` fits the random-effects model.

```{r fit1-metafor}
fit1_metafor <- metafor::rma(yi, vi, data = dat, method = "REML")
fit1_metafor
```

The matching Bayesian fit reuses the same effect-size columns.
`measure = "RR"` tells `brma()` to use the default prior distributions for log risk ratios.
The default prior distributions are weakly informative and provide slight regularization toward zero.

```{r fit1-brma, eval = FALSE}
fit1_brma <- brma(
  yi = yi, vi = vi, measure = "RR",
  data = dat, seed = 1
)
```

`summary()` reports posterior means, credible intervals, and (suppressed here) MCMC convergence diagnostics.
Convergence diagnostics are included by default; chain-level visual diagnostics are available via `plot_diagnostic_trace()`, `plot_diagnostic_density()`, and `plot_diagnostic_autocorrelation()`.

```{r fit1-brma-summary}
summary(fit1_brma, include_mcmc_diagnostics = FALSE)
```

The posterior mean for the log risk ratio is close to the REML estimate from the `metafor` package; the slight pull toward zero reflects the default `Normal(0, 1)` prior distribution.
`pooled_effect()` returns the same summary, backtransformed to the risk-ratio scale via `transform = "EXP"`:

```{r fit1-brma-pooled}
pooled_effect(fit1_brma, transform = "EXP")
```

`plot()` visualizes the posterior distribution, and `prior = TRUE` overlays the prior distribution so the shift from prior to posterior is visible.

```{r fit1-brma-plot, fig.width = 5, fig.height = 3.5}
par(mar = c(4, 4, 1, 1))
plot(fit1_brma, parameter = "mu", prior = TRUE, xlim = c(-3, 3))
```

## Heterogeneity

`metafor::confint()` reports profile-likelihood confidence intervals for `tau^2`, `tau`, `I^2`, and `H^2`.
`summary_heterogeneity()` is the Bayesian counterpart, returning posterior summaries of the same quantities.

```{r heterogeneity-confint}
confint(fit1_metafor)
```

```{r heterogeneity-bayes}
summary_heterogeneity(fit1_brma)
```

## Funnel Plot

`funnel()` plots observed effect sizes against the fitted sampling distribution.
We turn off the `sampling_heterogeneity` argument, since the `RoBMA` R package shows the full sampling distribution under the model by default.

```{r funnel-plot, fig.width = 7, fig.height = 3.2}
par(mfrow = c(1, 2), mar = c(4, 4, 2, 1), cex.axis = 0.8, cex.lab = 0.8, cex.main = 0.75)
metafor::funnel(fit1_metafor, main = "metafor", xlim = c(-2, 1), ylim = c(0, 0.8))
funnel(fit1_brma, main = "RoBMA", xlim = c(-2, 1), ylim = c(0, 0.8), sampling_heterogeneity = FALSE)
```

## Meta-Regression with a Continuous Moderator

A continuous moderator can be added with the standard formula syntax via the `mods` argument.
`ablat` is the absolute latitude of the trial, frequently used to explain BCG effect-size variation.

```{r fit2-metafor}
fit2_metafor <- metafor::rma(yi, vi, mods = ~ ablat, data = dat)
fit2_metafor
```

```{r fit2-brma, eval = FALSE}
fit2_brma <- brma(
  yi = yi, vi = vi, measure = "RR",
  mods = ~ ablat,
  data = dat, seed = 1
)
```
```{r fit2-brma-summary}
summary(fit2_brma, include_mcmc_diagnostics = FALSE)
```

`regplot()` produces the familiar bubble plot with the fitted regression line.

```{r regplot-continuous, fig.width = 7, fig.height = 3.2}
par(mfrow = c(1, 2), mar = c(4, 4, 2, 1), cex.axis = 0.8, cex.lab = 0.8, cex.main = 0.75)
metafor::regplot(fit2_metafor, main = "metafor", xlim = c(10, 60), ylim = c(-2, 0.5))
regplot(fit2_brma, main = "RoBMA", xlim = c(10, 60), ylim = c(-2, 0.5))
```

`predict()` returns posterior summaries for user-specified moderator values.

```{r fit2-predict}
new_data <- data.frame(ablat = c(15, 35, 55))
predict(fit2_brma, newdata = new_data, type = "terms")
```

## Comparing Models with Bayes Factors

Marginal likelihoods enable model comparison via Bayes factors [@kass1995bayes].
After attaching marginal likelihoods to both fits with `add_marglik()`, `bf()` returns the Bayes factor for one model against another.

```{r add-marglik, eval = FALSE}
fit1_brma <- add_marglik(fit1_brma)
fit2_brma <- add_marglik(fit2_brma)
```
```{r bayes-factor}
bf(fit2_brma, fit1_brma)
```

The Bayes factor quantifies how much the data shift the relative predictive support between the two models.
Here we obtain strong evidence in favor of the model including the `ablat` predictor.

## Adding a Categorical Moderator

`alloc` is a three-level factor describing the trial allocation method (random, alternate, systematic).
The same formula syntax extends the meta-regression.

```{r fit3-metafor}
fit3_metafor <- metafor::rma(yi, vi, mods = ~ ablat + alloc, data = dat)
```

```{r fit3-brma, eval = FALSE}
fit3_brma <- brma(
  yi = yi, vi = vi, measure = "RR",
  mods = ~ ablat + alloc,
  data = dat, seed = 1
)
```
```{r fit3-brma-summary}
summary(fit3_brma, include_mcmc_diagnostics = FALSE)
```

The `RoBMA` R package also allows direct visualization of categorical moderators in the bubble plot
(the `metafor` package only supports a single dummy coefficient at the moment).

```{r regplot-categorical, fig.width = 7, fig.height = 3.2}
par(mfrow = c(1, 2), mar = c(4, 4, 2, 1), cex.axis = 0.8, cex.lab = 0.8, cex.main = 0.75)
metafor::regplot(fit3_metafor, mod = "allocrandom", main = "metafor: random dummy", ylim = c(-2, 0.5))
regplot(fit3_brma, mod = "alloc", main = "RoBMA: alloc factor", ylim = c(-2, 0.5))
```

`marginal_means()` computes estimated marginal means for each moderator term, averaging over the other moderators in the fit.
Continuous moderators (here `ablat`) are evaluated at $-1$, $0$, and $+1$ standard deviations of the predictor; categorical moderators (here `alloc`) at each level.

```{r marginal-means}
emm <- marginal_means(fit3_brma)
summary(emm)
```

The `plot()` function visualizes the posterior distribution of the estimated marginal means.

```{r marginal-means-plot, fig.width = 4.5, fig.height = 3.25}
par(mar = c(4, 4, 1, 1), cex.axis = 0.8, cex.lab = 0.8, cex.main = 0.75)
plot(emm, parameter = "alloc", xlim = c(-2, 0.5), lwd = 2)
```

## Model Diagnostics

The `RoBMA` R package implements a wide set of model fit diagnostics. Here, we highlight only a few.

`metafor::rstudent()` returns frequentist studentized residuals computed by deletion;
`RoBMA::rstudent()` returns LOO-PIT residuals, which propagate posterior uncertainty and leverage through leave-one-out prediction.
The two definitions are comparable but not identical; the `RoBMA` R package's LOO-PIT residuals first require LOO computation via `add_loo()`.

```{r add-loo, eval = FALSE}
fit3_brma <- add_loo(fit3_brma)
```
```{r residuals}
head(rstudent(fit3_metafor)$z)
head(rstudent(fit3_brma)$z)
```

`qqnorm()` checks the residuals against a standard normal reference.

```{r qqplot, fig.width = 7, fig.height = 3.2}
par(mfrow = c(1, 2), mar = c(4, 4, 2, 1), cex.axis = 0.8, cex.lab = 0.8, cex.main = 0.75)
qqnorm(fit3_metafor, main = "metafor: studentized", xlim = c(-1.5, 1.5), ylim = c(-3, 3))
qqnorm(fit3_brma, main = "RoBMA: LOO-PIT", xlim = c(-1.5, 1.5), ylim = c(-3, 3))
```

`influence()` collects DFFITS, Cook's distance, COVRATIO, leverage, and DFBETAS where the diagnostic is available.

```{r influence}
inf_brma <- influence(fit3_brma)
head(inf_brma$inf)
```

## References

---
title: "Bayesian Model Averaging"
author: "František Bartoš"
date: "4th of May 2026"
output:
  rmarkdown::html_vignette:
    self_contained: yes
bibliography: ../inst/REFERENCES.bib
csl: ../inst/apa.csl
link-citations: true
vignette: >
  %\VignetteIndexEntry{Bayesian Model Averaging}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
  %\VignetteEngine{knitr::rmarkdown_notangle}
---

```{r child = "_vignette-nowrap.md", echo = FALSE, eval = TRUE}
```

```{r setup, include = FALSE}
source("vignette-cache.R", local = knitr::knit_global())

cached_fits <- vignette_cache(
  name     = "v20-bayesian-model-averaging",
  objects  = c("fit_RE", "fit_FE", "fit_BMA_he", "fit_BMA", "fit_BMA_mod"),
  packages = c("metafor", "metadat")
)

knitr::opts_chunk$set(
  collapse    = TRUE,
  comment     = "#>",
  eval        = vignette_cache_eval(cached_fits),
  message     = FALSE,
  warning     = FALSE,
  fig.width   = 7,
  fig.height  = 3.5,
  dev         = "png",
  fig.retina  = 3
)
if (.Platform$OS.type == "windows") {
  knitr::opts_chunk$set(dev.args = list(type = "cairo"))
}
```

```{r load-models, include = FALSE}
library("RoBMA")

data("dat.baskerville2012", package = "metadat")
dat <- dat.baskerville2012

vignette_cache_load(cached_fits)
```

```{r precompute-models, include = FALSE, eval = FALSE}
library("RoBMA")

data("dat.baskerville2012", package = "metadat")
dat <- dat.baskerville2012
dat$tailor <- as.factor(dat$tailor)

fit_RE <- brma(
  yi = smd, sei = se, measure = "SMD",
  data = dat, seed = 1
)
fit_RE <- add_marglik(fit_RE)

fit_FE <- brma(
  yi = smd, sei = se, measure = "SMD",
  prior_heterogeneity = prior("spike", list(location = 0)),
  data = dat, seed = 1
)
fit_FE <- add_marglik(fit_FE)

fit_BMA_he <- BMA(
  yi = smd, sei = se, measure = "SMD",
  prior_effect_null = NULL,
  data = dat, seed = 1
)

fit_BMA <- BMA(
  yi = smd, sei = se, measure = "SMD",
  data = dat, seed = 1
)

fit_BMA_mod <- BMA(
  yi = smd, sei = se, measure = "SMD",
  mods = ~ tailor,
  data = dat, seed = 1
)

vignette_cache_save(cached_fits)
```

In meta-analysis, analysts often have to balance different assumptions about the data-generating process and decide what kind of model to fit, which covariates to include, and what publication-bias adjustment method to use.
Bayesian model averaging (BMA) tries to alleviate this problem by combining competing meta-analytic models based on their prior predictive performance.
Models that predicted the data better receive higher posterior model probability and more weight in the overall inference.
The *model-averaged estimates* allow us to propagate uncertainty about the selected model and *inclusion Bayes factors* quantify evidence across sets of models [@hoeting1999bayesian].

In this vignette we illustrate Bayesian model-averaged meta-analysis using `BMA()` on the `dat.baskerville2012` dataset to highlight the basics of Bayesian model-averaging using the `RoBMA` R package [@RoBMA].
In a later vignette [*Robust Bayesian Meta-Analysis*](v21-robust-bayesian-meta-analysis.html), we illustrate robust Bayesian meta-analysis that extends this concept to model-averaging across publication-bias adjustment models.
All models use the default prior distributions; see the [*Prior Distributions*](v01-prior-distributions.html) vignette for the prior definitions and customization options.

We start by estimating simple fixed-effect and random-effects Bayesian meta-analytic models, turn the model-comparison into a model-averaged fit, and finally also average over the presence vs absence of the effect.
Finally, we extend the model into a Bayesian model-averaged meta-regression.
Note that we do not demonstrate all of the diagnostics and summary features available to all models fitted via the `RoBMA` R package (see the [*Bayesian Meta-Analysis*](v02-bayesian-meta-analysis.html) vignette for more details).

## Dataset

The example dataset is distributed with the `metadat` package as `dat.baskerville2012` and contains 23 randomized and controlled trials of practice-facilitation interventions in primary care.
The effect size `smd` is the standardized mean difference for adherence to evidence-based primary-care guidelines and `se` is its standard error.
The binary `tailor` indicator records whether the intervention was tailored to the participating practice (1) or applied as a uniform protocol (0); we treat it as a factor and use it as a moderator later in the vignette.

```{r set-up, include = TRUE}
library("RoBMA")

data("dat.baskerville2012", package = "metadat")
dat <- dat.baskerville2012
dat$tailor <- as.factor(dat$tailor)
head(dat[, c("year", "smd", "se", "tailor")])
```

## Comparing Fixed and Random Effects

`metafor::rma.uni()` fits the conventional random-effects model on the standardized mean differences.

```{r metafor-RE}
res <- metafor::rma.uni(
  yi = smd, sei = se,
  data = dat, method = "REML"
)
res
```

The Bayesian counterpart uses `brma()` with the default prior distributions: a `Normal(0, 0.71)` unit-information prior distribution on the effect and a `Normal(0, 0.35)[0, Inf]` half-normal prior distribution on the heterogeneity.

```{r fit-RE, eval = FALSE}
fit_RE <- brma(
  yi = smd, sei = se, measure = "SMD",
  data = dat, seed = 1
)
fit_RE <- add_marglik(fit_RE)
```
```{r fit-RE-summary}
summary(fit_RE, include_mcmc_diagnostics = FALSE)
```

Setting the heterogeneity prior distribution to a spike at zero turns the random-effects model into a fixed-effect model.

```{r fit-FE, eval = FALSE}
fit_FE <- brma(
  yi = smd, sei = se, measure = "SMD",
  prior_heterogeneity = prior("spike", list(location = 0)),
  data = dat, seed = 1
)
fit_FE <- add_marglik(fit_FE)
```
```{r fit-FE-summary}
summary(fit_FE, include_mcmc_diagnostics = FALSE)
```

With marginal likelihoods attached to the models via `add_marglik()`, `bf()` returns the Bayes factor between the two fits.

```{r bf-FE-RE}
bf(fit_RE, fit_FE)
```

The Bayes factor is close to one and neither model is decisively preferred.
Rather than picking one of the two models, we can model-average over them and weight the inference by the posterior model probabilities.

## From Two Fits to Model Averaging

`BMA()` fits a mixture of the two models in one call.
Each top-level prior-distribution argument has a paired `_null` version; when we supply both, the model averages over that component.
Setting `prior_effect_null = NULL` drops the null component on the effect (spike at zero) and leaves a 2-model average over presence vs absence of heterogeneity, which is the BMA analogue of the comparison above.

```{r fit-BMA-he, eval = FALSE}
fit_BMA_he <- BMA(
  yi = smd, sei = se, measure = "SMD",
  prior_effect_null = NULL,
  data = dat, seed = 1
)
```
```{r fit-BMA-he-summary}
summary(fit_BMA_he)
```

The *Components* table reports the posterior probability and inclusion Bayes factor for heterogeneity.
This BF closely matches `bf(fit_RE, fit_FE)` above up to MCMC noise.
The two routes compute the same quantity differently: `bf()` bridge-samples marginal likelihoods from two separate fits, while `BMA()` runs a single MCMC over the joint model space with a model indicator as a latent variable and recovers the posterior model probabilities from the indicator frequencies [@lodewyckx2011tutorial].
Adding a third or fourth component to this product-space sampler costs one more indicator instead of a set of full model fits and bridge-sampling runs, which is what makes the larger ensembles in `RoBMA()` tractable.

## Adding Effect Averaging

The default `BMA()` call also averages over the presence vs absence of the effect, giving a 4-model ensemble (effect $\times$ heterogeneity, each present or absent).

```{r fit-BMA, eval = FALSE}
fit_BMA <- BMA(
  yi = smd, sei = se, measure = "SMD",
  data = dat, seed = 1
)
```
```{r fit-BMA-summary}
summary(fit_BMA)
```

Both inclusion Bayes factors come from the same MCMC.
The effect BF is overwhelming on this dataset; the heterogeneity BF is the same as before.
The model-averaged estimates are spike-and-slab mixtures: a point mass at the null value with weight equal to the posterior probability that the parameter is zero, and a slab carrying the remaining weight.

`plot()` draws both components together.
The arrow at zero is the spike's posterior probability (right axis), the curve is the slab's density, and the grey shapes show the corresponding visualization for the prior distribution.

```{r plot-BMA-mu, fig.width = 6, fig.height = 4, out.width = "75%", fig.align = "center"}
par(mar = c(4, 4, 1, 4))
plot(fit_BMA, parameter = "mu", prior = TRUE, xlim = c(-1, 1.5), lwd = 2)
```
```{r plot-BMA-tau, fig.width = 6, fig.height = 4, out.width = "75%", fig.align = "center"}
par(mar = c(4, 4, 1, 4))
plot(fit_BMA, parameter = "tau", prior = TRUE, xlim = c(0, 1), lwd = 2)
```

The *Model-averaged estimates* summarize the marginal posterior (the full mixture).
For estimates conditional on the parameter being non-zero (slab only) use the `conditional = TRUE` argument.
The marginal and conditional summaries diverge for $\tau$, which keeps substantial posterior weight on the spike, and coincide for $\mu$, where the spike's posterior weight is essentially zero.

```{r pooled-conditional}
summary(fit_BMA, conditional = TRUE)
```

## Meta-Regression with Model Averaging

Adding a moderator extends the averaging to its inclusion.
With `mods = ~ tailor`, `BMA()` averages over 8 sub-models (effect $\times$ heterogeneity $\times$ `tailor` coefficient, each present or absent).

```{r metafor-mod}
res_mod <- metafor::rma.uni(
  yi = smd, sei = se, mods = ~ tailor,
  data = dat, method = "REML"
)
res_mod
```

```{r fit-BMA-mod, eval = FALSE}
fit_BMA_mod <- BMA(
  yi = smd, sei = se, measure = "SMD",
  mods = ~ tailor,
  data = dat, seed = 1
)
```
```{r fit-BMA-mod-summary}
summary(fit_BMA_mod)
```

The *Components* table now also lists `tailor` and its inclusion BF.
Because we average the moderation across models in which the slope is zero, `regplot()` shrinks the difference towards zero when the inclusion BF is modest.

```{r regplot-BMA-mod, fig.width = 5, fig.height = 3.5, fig.align = "center"}
par(mar = c(4, 4, 1, 1))
regplot(fit_BMA_mod, mod = "tailor")
```

Alternatively, we can directly compute the estimated marginal means via the `marginal_means()` function and visualize the prior and posterior distributions at each moderator level.

```{r regplot-BMA-emm, fig.width = 5, fig.height = 3.5, fig.align = "center"}
par(mar = c(4, 4, 1, 1))
emm <- marginal_means(fit_BMA_mod)
emm

plot(emm, parameter = "tailor", prior = TRUE)
```


## Customizing Prior Components

In Bayesian model-averaged meta-analysis, every component has a paired `_null` argument (i.e. `prior_effect` and `prior_effect_null`).
Either side can be `NULL` (remove the component), a single prior distribution, or a list of prior distributions (a sub-mixture).
Common customizations are tightening or widening the slab, replacing the point-zero null with a region of practical equivalence, or forcing a moderator into every model.
The examples below use `only_priors = TRUE` to print the prior structure without fitting the models.

Tighten the effect slab with `rescale_priors = 0.5`:

```{r priors-rescale}
fit_priors <- BMA(
  yi = smd, sei = se, measure = "SMD",
  rescale_priors = 0.5,
  data = dat, only_priors = TRUE
)
print_prior(fit_priors, parameter = "mu")
```

Replace the point-zero null on the effect with a small spread around zero (a perinull) to test against a region of practical equivalence:

```{r priors-perinull}
fit_priors <- BMA(
  yi = smd, sei = se, measure = "SMD",
  prior_effect_null = prior("normal", list(mean = 0, sd = 0.05)),
  data = dat, only_priors = TRUE
)
print_prior(fit_priors, parameter = "mu")
```

Force `tailor` into every model by dropping its null:

```{r priors-force-mod}
fit_priors <- BMA(
  yi = smd, sei = se, measure = "SMD",
  mods = ~ tailor,
  prior_mods_null = list(tailor = NULL),
  data = dat, only_priors = TRUE
)
print_prior(fit_priors, parameter_mods = "tailor")
```

A tighter half-normal slab on $\tau$:

```{r priors-tau}
fit_priors <- BMA(
  yi = smd, sei = se, measure = "SMD",
  prior_heterogeneity = prior(
    distribution = "normal",
    parameters   = list(mean = 0, sd = 0.25),
    truncation   = list(lower = 0)
  ),
  data = dat, only_priors = TRUE
)
print_prior(fit_priors, parameter = "tau")
```

Empirical informed prior distributions based on the Cochrane database are available via `prior_informed_field` and `prior_informed_subfield`; see [*Informed Bayesian Model-Averaged Meta-Analysis in Medicine*](v34-bma-norm-medicine.html).

## Other `brma` Features

A `BMA()` fit is also a `brma` object: `summary()`, `plot()`, `print_prior()`, `predict()`, `marginal_means()`, `funnel()`, `rstudent()`, `qqnorm()`, and `influence()` work the same way as on a single-model fit; see [*Bayesian Meta-Analysis*](v02-bayesian-meta-analysis.html).
Prior-distribution families and informed prior distributions are covered in [*Prior Distributions*](v01-prior-distributions.html).
For binary or count outcomes, `BMA.glmm()` does the same product-space averaging on a binomial-normal or Poisson-normal likelihood.
For model averaging that also adjusts for publication bias, see [*Robust Bayesian Meta-Analysis*](v21-robust-bayesian-meta-analysis.html).

## References

---
title: "Robust Bayesian Model-Averaged Meta-Regression"
author: "František Bartoš"
date: "11th of December 2024 (updated: 29th of April 2026)"
output:
  rmarkdown::html_vignette:
    self_contained: yes
bibliography: ../inst/REFERENCES.bib
csl: ../inst/apa.csl
link-citations: true
vignette: >
  %\VignetteIndexEntry{Robust Bayesian Model-Averaged Meta-Regression}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
  %\VignetteEngine{knitr::rmarkdown_notangle}
---

```{r child = "_vignette-nowrap.md", echo = FALSE, eval = TRUE}
```

```{r setup, include = FALSE}
source("vignette-cache.R", local = knitr::knit_global())

cached_fits <- vignette_cache(
  name     = "v31-robma-metaregression",
  objects  = c("fit_BMA", "fit_RoBMA"),
  packages = c("metafor", "emmeans")
)

knitr::opts_chunk$set(
  collapse    = TRUE,
  comment     = "#>",
  eval        = vignette_cache_eval(cached_fits),
  message     = FALSE,
  warning     = FALSE,
  fig.width   = 7,
  fig.height  = 3.5,
  dev         = "png",
  fig.retina  = 3
)
if (.Platform$OS.type == "windows") {
  knitr::opts_chunk$set(dev.args = list(type = "cairo"))
}
```

```{r load-models, include = FALSE}
library("RoBMA")

data("Andrews2021", package = "RoBMA")

Andrews2021_z <- metafor::escalc(
  ri = ri, ni = ni,
  measure = "ZCOR", data = Andrews2021
)

vignette_cache_load(cached_fits)
```

```{r precompute-models, include = FALSE, eval = FALSE}
library("RoBMA")

data("Andrews2021", package = "RoBMA")

Andrews2021_z <- metafor::escalc(
  ri = ri, ni = ni,
  measure = "ZCOR", data = Andrews2021
)

prior_effect_361 <- prior("normal", list(mean = 0, sd = 0.5))
prior_heterogeneity_361 <- prior("invgamma", list(shape = 1, scale = 0.075))

fit_BMA <- BMA(
  yi = yi, vi = vi, measure = "ZCOR",
  mods = ~ measure + age,
  prior_effect        = prior_effect_361,
  prior_heterogeneity = prior_heterogeneity_361,
  data = Andrews2021_z, parallel = TRUE, seed = 1
)

fit_RoBMA <- RoBMA(
  yi = yi, vi = vi, measure = "ZCOR",
  mods = ~ measure + age,
  prior_effect        = prior_effect_361,
  prior_heterogeneity = prior_heterogeneity_361,
  data = Andrews2021_z, parallel = TRUE, seed = 1
)

vignette_cache_save(cached_fits)
```

**This vignette accompanies the manuscript [Robust Bayesian Meta-Regression: Model-Averaged Moderation Analysis in the Presence of Publication Bias](https://doi.org/10.1037/met0000737) published at *Psychological Methods* [@bartos2023robust].**

Robust Bayesian model-averaged meta-regression (RoBMA-reg) extends the robust Bayesian model-averaged meta-analysis (RoBMA) by including covariates in the meta-analytic model. 
RoBMA-reg allows for estimating and testing the moderating effects of study-level covariates on the meta-analytic effect in a unified framework (e.g., accounting for uncertainty in the presence vs. absence of the effect, heterogeneity, and publication bias). 
This vignette illustrates how to fit a robust Bayesian model-averaged meta-regression using the `RoBMA` R package. 
We reproduce the example from @bartos2023robust, who re-analyzed a meta-analysis of the effect of household chaos on child executive functions with the mean age and assessment type covariates based on @andrews2021examining's meta-analysis. 

First, we fit a frequentist meta-regression using the `metafor` package.
Second, we explain the Bayesian meta-regression model specification, the default prior distributions for continuous and categorical moderators, and the standardized effect-size input specification.
Third, we estimate Bayesian model-averaged meta-regression (without publication bias adjustment).
Finally, we estimate the complete robust Bayesian model-averaged meta-regression.

## Data

We start by loading the `Andrews2021` dataset included in the `RoBMA` R package, which contains 39 source rows from the meta-analysis of the effect of household chaos on child executive functions. 
The dataset includes raw correlation coefficients (`ri`), sample sizes (`ni`), study labels (`slab`), and study-level moderators such as the type of executive function assessment (`measure`) and the mean age of the children (`age`). 
The analyses below use the 36 rows with complete effect-size inputs.

```{r}
library(RoBMA)
data("Andrews2021", package = "RoBMA")
head(Andrews2021)[,c("slab", "ri", "ni", "measure", "age")]
```

We use `metafor::escalc()` to compute Fisher's $z$ effect sizes directly from correlation coefficients and sample sizes which we will use for all the subsequent analyses.

```{r escalc}
Andrews2021_z <- metafor::escalc(
  ri = ri, ni = ni,
  measure = "ZCOR", data = Andrews2021
)
```

## Frequentist Meta-Regression

We start by fitting a frequentist meta-regression using the `metafor` package [@metafor].
While @andrews2021examining estimated univariate meta-regressions for each moderator, we directly proceed by analyzing both moderators simultaneously.
The model is estimated on Fisher's $z$ scale and the pooled estimates are transformed to correlation coefficients only for reporting.

```{r}
fit_rma <- metafor::rma(yi, vi, mods = ~ measure + age, data = Andrews2021_z)
fit_rma
```

The results reveal a statistically significant moderation effect of the executive function assessment type on the effect of household chaos on child executive functions (p = 0.013).
To explore the moderation effect further, we estimate the estimated marginal means for the executive function assessment type using the `emmeans` R package [@emmeans] and back-transform the estimates to correlations.

```{r}
emm_rma <- emmeans::emmeans(metafor::emmprep(fit_rma), specs = "measure")
emm_rma
metafor::transf.ztor(as.data.frame(emm_rma)[, c("emmean", "asymp.LCL", "asymp.UCL")])
```

Studies using the informant-completed questionnaires show a stronger effect of household chaos on child executive functions, r = 0.227, 95\% CI [0.160, 0.293], than the direct assessment, r = 0.111, 95\% CI [0.050, 0.172]; both types of studies show statistically significant effects.

The mean age of the children does not significantly moderate the effect (p = 0.600) with the estimated regression coefficient of b = 0.003, 95\% CI [-0.009, 0.016] on Fisher's $z$ scale.
As usual, frequentist inference only allows us to fail to reject the null hypothesis.
Here, we try to overcome this limitation with Bayesian model-averaged meta-regression.


## Bayesian Meta-Regression Specification

Before we proceed with the Bayesian model-averaged meta-regression, we provide a quick overview of the regression model specification. 
In contrast to frequentist meta-regression, we need to specify prior distributions on the regression coefficients, which encode the tested hypotheses about the presence vs. absence of the moderation (specifying different prior distributions corresponds to different hypotheses and results in different conclusions). 
Importantly, the treatment of continuous and categorical covariates differs in the Bayesian model-averaged meta-regression.

### Continuous vs. Categorical Moderators and Prior Distributions

The package automatically standardizes continuous moderators, which makes the moderator prior distributions scale-invariant and keeps the intercept prior distribution centered on the grand mean effect.
This setting can be overridden with `standardize_continuous_predictors = FALSE`.

Categorical moderators use mean-difference contrasts by default for model-averaged models (`set_contrast_factor_predictors = "meandif"`), modeling deviations of each category from the grand mean effect.
The `"meandif"` contrasts achieve label invariance and the prior distribution for the intercept corresponds to the grand mean effect.
Alternatively, `"treatment"` contrasts place prior distributions on differences from the reference level, with the intercept corresponding to the effect in that level.

### Prior Distributions

Note that the effect-size and heterogeneity default prior distributions in the current (4.0) version of the package differ from those used by `RoBMA` 3.6.1.
The current default prior distributions for Bayesian meta-analyses are calibrated to the unit-information standard deviation of the effect-size measure, specified via the `measure` argument (see the [*Prior Distributions*](v01-prior-distributions.html) vignette).

To reproduce the prior specification used in @bartos2023robust, we specify the prior distributions explicitly below.
The old effect-size and heterogeneity prior distributions, expressed on the fitted Fisher's $z$ scale, are specified explicitly.[^prior-scale]

```{r prior-specification}
prior_effect_361 <- prior("normal", list(mean = 0, sd = 0.5))
prior_heterogeneity_361 <- prior("invgamma", list(shape = 1, scale = 0.075))
```

[^prior-scale]: `RoBMA` 3.6.1 used Cohen's $d$ as the prior scale for correlation inputs while fitting Fisher's $z$ internally. The prior distributions above express the old effect-size and heterogeneity scales on Fisher's $z$ using the local approximation $z \approx d / 2$.

## Bayesian Model-Averaged Meta-Regression

We fit the Bayesian model-averaged meta-regression using `BMA()`, which omits publication-bias adjustment.
We specify the Fisher's $z$ effect sizes as `yi` and their sampling variances as `vi`, and set the `measure` argument to `"ZCOR"`, which tells `BMA()` and `RoBMA()` how to interpret the input scale.
The moderators are passed via the `mods` formula, here `~ measure + age`.
We also set `parallel = TRUE` to speed up computation and `seed = 1` for reproducibility.

```{r fit-bma, eval = FALSE}
fit_BMA <- BMA(
  yi = yi, vi = vi, measure = "ZCOR",
  mods = ~ measure + age,
  prior_effect        = prior_effect_361,
  prior_heterogeneity = prior_heterogeneity_361,
  data = Andrews2021_z, parallel = TRUE, seed = 1
)
```

`BMA()` specifies the combination of all models assuming presence vs. absence of the effect, heterogeneity, moderation by `measure`, and moderation by `age`, which corresponds to 2\*2\*2\*2=16 models.
The new version of the package uses product-space parameterization, which greatly improves estimation time by fitting all of the models simultaneously within a single MCMC.

Once the ensemble is estimated, `summary()` reports model-averaging results on the fitted Fisher's $z$ scale.
Dedicated effect-summary helpers handle transformations to the correlation scale.

```{r}
summary(fit_BMA, include_mcmc_diagnostics = FALSE)
pooled_effect(fit_BMA, output_measure = "COR")
```

The summary function produces output with multiple sections.
The first section contains the `Components summary` with the hypothesis test results for the overall effect size and heterogeneity. 
We find overwhelming evidence for both with inclusion Bayes factors (`Inclusion BF`) above 10,000 (the numerical precision of Bayes factors above 100 is usually limited by the number of posterior samples).

The second section contains the `Meta-regression components summary` with the hypothesis test results for the moderators. 
We find weak evidence for the moderation by the executive function assessment type, $\text{BF}_{\text{measure}} = 2.67$. 
Furthermore, we find moderate evidence for the null hypothesis of no moderation by mean age of the children, $\text{BF}_{\text{age}} = 0.122$ (i.e., BF for the null is 1/0.122 = 8.20). 
These findings extend the frequentist meta-regression by disentangling the absence of evidence from the evidence of absence.

The third section contains the `Model-averaged estimates` with the model-averaged estimates for mean effect and between-study heterogeneity.
`pooled_effect()` reports the pooled mean effect on the correlation scale via `output_measure = "COR"`.

The fourth section contains the `Model-averaged meta-regression estimates` with the model-averaged regression coefficient estimates. 
The main difference from the usual frequentist meta-regression output is that the categorical predictors are summarized as a difference from the grand mean for each factor level. 
Here, the `intercept` regression coefficient estimate corresponds to the grand mean effect and the `measure[dif: direct]` regression coefficient estimate of -0.044, 95\% CI [-0.105, 0.000] corresponds to the difference between the direct assessment and the grand mean. 
As such, the results suggest that the effect size in studies using direct assessment is lower in comparison to the grand mean of the studies. The `age` regression coefficient of 0.000, 95\% CI [-0.000, 0.009] corresponds to the increase in the mean effect for a one-unit increase in mean age of children.

Similarly to the frequentist meta-regression, we can use `marginal_means()` and `summary()` to obtain the marginal estimates for each factor level on the correlation scale.
```{r}
BMA_marginal <- marginal_means(fit_BMA, output_measure = "COR")
summary(BMA_marginal)
```
The estimated marginal means are similar to the frequentist results. 
Studies using the informant-completed questionnaires again show a stronger effect of household chaos on child executive functions, r = 0.210, 95\% CI [0.128, 0.289], than the direct assessment, r = 0.124, 95\% CI [0.054, 0.196]. 

The last column summarizes results from a test against a null hypothesis of marginal means equals 0. 
Here, we find extreme evidence that the effect size for studies using the informant-completed questionnaires differs from zero, $\text{BF}_{10} > 10{,}000$, and strong evidence that the effect size for studies using direct assessment differs from zero, $\text{BF}_{10} = 23.6$.
The test is performed using the change from prior to posterior distribution at 0 (i.e., the Savage-Dickey density ratio) assuming the presence of the overall effect or the presence of difference according to the tested factor. 
Because the tests use prior and posterior samples, calculating the Bayes factor can be problematic when the posterior distribution is far from the tested value. 
In such cases, warning messages are printed and $\text{BF}_{10} = \infty$ returned (like here); while the actual Bayes factor is less than infinity, it is still too large to be computed precisely given the posterior samples.

The full model-averaged posterior marginal means distribution can be visualized by the `plot()` method.
```{r fig_BMA, fig.width = 6, fig.height = 4, out.width = "75%", fig.align = "center"}
par(mar = c(4, 4, 1, 4))
plot(BMA_marginal, parameter = "measure", lwd = 2)
```

## Publication-Bias-Adjusted Model-Averaged Meta-Regression

Finally, we adjust the Bayesian model-averaged meta-regression model by fitting the robust Bayesian model-averaged meta-regression.
In contrast to the previous publication-bias-unadjusted ensemble, `RoBMA()` extends the model ensemble with a publication-bias component specified via 6 weight functions and PET-PEESE [@bartos2021no].
The estimation time further increases as the ensemble now contains 144 models.

```{r fit-robma, eval = FALSE}
fit_RoBMA <- RoBMA(
  yi = yi, vi = vi, measure = "ZCOR",
  mods = ~ measure + age,
  prior_effect        = prior_effect_361,
  prior_heterogeneity = prior_heterogeneity_361,
  data = Andrews2021_z, parallel = TRUE, seed = 1
)
```
```{r}
summary(fit_RoBMA, include_mcmc_diagnostics = FALSE)
pooled_effect(fit_RoBMA, output_measure = "COR")
```
All previously described functions for manipulating the fitted model work identically with the publication bias adjusted model. 
As such, we just briefly mention the main differences found after adjusting for publication bias.

RoBMA-reg reveals moderate evidence of publication bias $\text{BF}_{\text{pb}} = 4.26$. 
Furthermore, accounting for publication bias turns the previously found evidence for the overall effect into a weak evidence for the effect $\text{BF}_{10} = 1.25$ and notably reduces the mean effect estimate to r = 0.061, 95\% CI [-0.020, 0.189].

```{r}
RoBMA_marginal <- marginal_means(fit_RoBMA, output_measure = "COR")
summary(RoBMA_marginal)
```

The estimated marginal means now suggest that studies using the informant-completed questionnaires show a much smaller effect of household chaos on child executive functions, r = 0.128, 95\% CI [0.000, 0.116] with moderate evidence in favor of the effect, $\text{BF}_{10} = 6.33$, 
while studies using direct assessment even provide weak evidence against the effect of household chaos on child executive functions, $\text{BF}_{10} = 0.626$, with most likely effect sizes around zero, r = 0.013, 95\% CI [-0.115, 0.175].

A visual summary of the estimated marginal means highlights the considerably wider model-averaged posterior distributions of the marginal means, a consequence of accounting and adjusting for publication bias.

```{r fig_RoBMA, fig.width = 6, fig.height = 4, out.width = "75%", fig.align = "center"}
par(mar = c(4, 4, 1, 4))
plot(RoBMA_marginal, parameter = "measure", lwd = 2)
```

The Bayesian model-averaged meta-regression models are compatible with the remaining custom specification, visualization, and summary functions included in the `RoBMA` R package.
For prior customization, see the [*Prior Distributions*](v01-prior-distributions.html) vignette; for shared meta-regression plotting and diagnostics, see the [*Bayesian Meta-Analysis*](v02-bayesian-meta-analysis.html) vignette.

## References

```{r vignette-nowrap-options, include = FALSE, eval = TRUE}
options(width = 10000)
```

<style type="text/css">
pre,
pre code,
pre:not([class]),
pre.sourceCode,
pre.sourceCode code,
pre > code.sourceCode {
  white-space: pre !important;
  word-break: normal !important;
  overflow-wrap: normal !important;
  word-wrap: normal !important;
}

pre,
div.sourceCode {
  overflow-x: auto !important;
}
</style>

---
title: "Publication-Bias Adjustment"
author: "František Bartoš"
date: "28th of April 2026"
output:
  rmarkdown::html_vignette:
    self_contained: yes
bibliography: ../inst/REFERENCES.bib
csl: ../inst/apa.csl
link-citations: true
vignette: >
  %\VignetteIndexEntry{Publication-Bias Adjustment}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
  %\VignetteEngine{knitr::rmarkdown_notangle}
---

```{r child = "_vignette-nowrap.md", echo = FALSE, eval = TRUE}
```

```{r setup, include = FALSE}
source("vignette-cache.R", local = knitr::knit_global())

cached_fits <- vignette_cache(
  name     = "v11-metafor-parity-publication-bias",
  objects  = c(
    "fit_PET_brma", "fit_PEESE_brma", "fit_selmodel_brma",
    "fit_PET_brma_flex", "fit_PEESE_brma_flex", "fit_selmodel_brma_flex"
  ),
  packages = c("metafor", "metadat")
)

knitr::opts_chunk$set(
  collapse    = TRUE,
  comment     = "#>",
  eval        = vignette_cache_eval(cached_fits),
  message     = FALSE,
  warning     = FALSE,
  fig.width   = 7,
  fig.height  = 3.5,
  dev         = "png",
  fig.retina  = 3
)
if (.Platform$OS.type == "windows") {
  knitr::opts_chunk$set(dev.args = list(type = "cairo"))
}
```

```{r load-models, include = FALSE}
library("metafor")
library("RoBMA")

data("dat.lehmann2018", package = "metadat")
dat <- dat.lehmann2018

vignette_cache_load(cached_fits)
```

```{r precompute-models, include = FALSE, eval = FALSE}
library("metafor")
library("RoBMA")

data("dat.lehmann2018", package = "metadat")
dat <- dat.lehmann2018

fit_PET_brma <- bPET(
  yi = yi, vi = vi, measure = "SMD",
  data = dat, seed = 1
)

fit_PEESE_brma <- bPEESE(
  yi = yi, vi = vi, measure = "SMD",
  data = dat, seed = 1
)

fit_selmodel_brma <- bselmodel(
  yi = yi, vi = vi, measure = "SMD",
  data = dat, seed = 1
)

fit_PET_brma_flex <- bPET(
  yi = yi, vi = vi, measure = "SMD",
  prior_bias = prior_PET(
    distribution = "cauchy",
    parameters = list(location = 0, scale = 1),
    truncation = list(lower = -Inf, upper = Inf)
  ),
  data = dat, seed = 1
)

fit_PEESE_brma_flex <- bPEESE(
  yi = yi, vi = vi, measure = "SMD",
  prior_bias = prior_PEESE(
    distribution = "cauchy",
    parameters = list(location = 0, scale = 5),
    truncation = list(lower = -Inf, upper = Inf)
  ),
  data = dat, seed = 1
)

fit_selmodel_brma_flex <- bselmodel(
  yi = yi, vi = vi, measure = "SMD",
  prior_bias = prior_weightfunction(
    "one-sided",
    steps   = c(0.025),
    weights = wf_independent(
      prior("normal", parameters = list(mean = 0, sd = 0.5)),
      scale = "log_omega"
    )
  ),
  data = dat, seed = 1
)

vignette_cache_save(cached_fits)
```

This vignette illustrates how to use the `RoBMA` R package [@RoBMA] for single-model publication-bias adjustment.
We walk through the ego-depletion dataset from the `metadat` package, fitting three single-model bias adjustments alongside their `metafor` package counterparts [@metafor]: PET and PEESE [@stanley2014meta], and a weight-function selection model [@vevea1995general].
All of the models can also be combined with Bayesian model-averaging into a Robust Bayesian meta-analysis as described in the [*Robust Bayesian Meta-Analysis*](v21-robust-bayesian-meta-analysis.html) vignette.

The vignette is organized in three parts.
First, we describe the default prior distributions on the bias parameters.
Second, we fit each model with its default prior distribution, check parity against the `metafor` package, and contrast the implied funnels.
Third, we relax each default in turn, placing the restricted and relaxed posteriors side by side.

For the non-bias-adjusted workflow showing more features (also applicable to the bias-adjustment models) see the [*Bayesian Meta-Analysis*](v02-bayesian-meta-analysis.html) vignette. The [*Prior Distributions*](v01-prior-distributions.html) vignette covers setting prior distributions for the meta-analytic parameters.

## Default Bias-Adjustment Prior Distributions

The three model types use a separate prior distribution on the bias-adjustment parameter, independent of the prior distributions on `mu`, `tau`, and the meta-regression coefficients.

| Model         | Bias parameter | Default prior distribution |
|:--------|:---------------|:------------------------------------------------------------------|
| `bPET()`      | `PET`   | Cauchy($0$, $1$), truncated to $[0, \infty)$ |
| `bPEESE()`    | `PEESE` | Cauchy($0$, $5 \cdot \text{UISD}_{\text{ratio}}$), truncated to $[0, \infty)$ |
| `bselmodel()` | `omega` | one-sided weight function with `steps = c(0.025)` and `alpha = c(1, 1)` |

The PET prior distribution is invariant to the effect-size measure.
The PEESE scale is rescaled by $\text{UISD}_{\text{ratio}} = \text{UISD}_{\text{SMD}} / \text{UISD}_{\text{measure}}$ so the prior distribution describes a comparable bias on each scale; the ratio is $1$ for `measure = "SMD"`.
The weight-function default specifies a Dirichlet prior distribution on the relative publication probabilities of one-sided p-values below and above $0.025$, equivalent to a three-parameter selection model (3PSM).
Unlike the prior distributions for meta-analytic parameters, the bias-adjustment prior distributions are not affected by `rescale_priors`; they can be replaced via `prior_bias`, and the weight-function `steps` can be changed directly via the `steps` argument.
The bias-adjustment models rely on the `effect_direction` argument.
Selection models penalize one-sided p-values evaluated against the expected sign of the effect, and the truncated PET and PEESE prior distributions constrain the small-study coefficient to pull the estimate toward zero from the dominant direction.
The `effect_direction` argument tells the model which side that is.
The default `effect_direction = "detect"` infers the direction from the median of `yi` (positive median yields `"positive"`, negative median yields `"negative"`).
Set `effect_direction = "positive"` or `effect_direction = "negative"` explicitly when the dominant direction is known a priori.

These prior distributions correspond to the default settings tested in the larger RoBMA-PSMA ensemble in @bartos2021no.
Note that all three defaults are *restricted to assume the presence of publication bias*: the truncated PET / PEESE coefficients are constrained to pull the estimate toward zero, and the cumulative-Dirichlet weight function forces every non-significant interval's `omega` to be no greater than the most-significant interval's reference value of $1$.
The "no publication bias" case is therefore on the parameter-space boundary.

In a model-averaged `RoBMA()` ensemble this is exactly the desired behaviour: a separate no-bias model carries the "no selection" hypothesis, and the bias-adjustment models are deliberately one-sided so the contrast against the no-bias spike is sharp and the inclusion Bayes factor is diagnostic.
For the single-model fits in this vignette there is no companion no-bias model in the ensemble, so the same restriction can shrink estimates toward the implied bias even when the data are consistent with no selection.
The *Flexible Bias Prior Distributions* section below relaxes the restriction model by model via the `prior_bias` argument when the single-model fit is the analysis of interest.

## Default Fits

The `dat.lehmann2018` dataset from the `metadat` package collects 81 standardized mean differences from the ego-depletion meta-analysis.
Effect sizes `yi` and sampling variances `vi` are already provided.

```{r data}
data("dat.lehmann2018", package = "metadat")
dat <- dat.lehmann2018
```

We fit each of the three bias-adjustment models with its default prior distribution and contrast against the matching `metafor` package call.
The funnel-plot view at the end of the section then visualizes how each adjustment encodes a different bias mechanism.

### PET

The PET model regresses effect sizes on the standard error.
With `metafor::rma()` it is fit by passing `mods = ~ sqrt(vi)`; the slope on `sqrt(vi)` is the PET coefficient.
The `RoBMA` R package implements PET as a meta-regression (with a random-effects heterogeneity component) rather than the UWLS variant, for consistency with the rest of the framework.

```{r fit-PET-metafor}
fit_PET_metafor <- metafor::rma(yi, vi, mods = ~ sqrt(vi), data = dat)
fit_PET_metafor
```

`bPET()` fits the same model with the default Cauchy prior distribution on the (positive) PET coefficient.
`measure = "SMD"` selects the default prior distributions for standardized mean differences.

```{r fit-PET-brma, eval = FALSE}
fit_PET_brma <- bPET(
  yi = yi, vi = vi, measure = "SMD",
  data = dat, seed = 1
)
```
```{r fit-PET-brma-summary}
summary(fit_PET_brma, include_mcmc_diagnostics = FALSE)
```

The posterior mean of `mu` is the bias-corrected average effect; the `PET` coefficient quantifies how strongly small studies pull the meta-analytic mean.
The left panel shows the prior and posterior of the `PET` coefficient with `prior = TRUE` overlaying the prior distribution; the right panel shows the implied PET regression of effect on standard error from `plot_pet_peese()`, with the prior band overlaid.

```{r PET-posterior, fig.width = 7, fig.height = 3.2}
par(mfrow = c(1, 2), mar = c(4, 4, 2, 1), cex.axis = 0.8, cex.lab = 0.8, cex.main = 0.75)
plot(fit_PET_brma,  parameter = "PET", prior = TRUE, main = "PET coefficient", xlim = c(0, 4))
plot_pet_peese(fit_PET_brma, prior = TRUE, main = "PET regression")
```

The half-Cauchy prior distribution on `PET` is wide; the posterior concentrates near $1$, broadly consistent with the `metafor` package estimate but pulled toward zero by the prior distribution.
The implied regression line intersects the distribution of effect sizes and standard errors.

### PEESE

The PEESE model regresses effect sizes on the sampling variance instead of the standard error.
With `metafor::rma()` it is fit by passing `mods = ~ vi`; the slope on `vi` is the PEESE coefficient.
As with PET, the `RoBMA` R package PEESE is the meta-regression version (retaining the heterogeneity component), not the UWLS variant.

```{r fit-PEESE-metafor}
fit_PEESE_metafor <- metafor::rma(yi, vi, mods = ~ vi, data = dat)
fit_PEESE_metafor
```

`bPEESE()` fits the same model with the default Cauchy prior distribution on the (positive) PEESE coefficient, with the prior scale rescaled to the effect-size measure.

```{r fit-PEESE-brma, eval = FALSE}
fit_PEESE_brma <- bPEESE(
  yi = yi, vi = vi, measure = "SMD",
  data = dat, seed = 1
)
```

```{r fit-PEESE-brma-summary}
summary(fit_PEESE_brma, include_mcmc_diagnostics = FALSE)
```

As for PET, the left panel shows the prior and posterior of the `PEESE` coefficient and the right panel shows the implied PEESE regression of effect on sampling variance.

```{r PEESE-posterior, fig.width = 7, fig.height = 3.2}
par(mfrow = c(1, 2), mar = c(4, 4, 2, 1), cex.axis = 0.8, cex.lab = 0.8, cex.main = 0.75)
plot(fit_PEESE_brma, parameter = "PEESE", prior = TRUE, main = "PEESE coefficient", xlim = c(0, 8))
plot_pet_peese(fit_PEESE_brma, prior = TRUE, main = "PEESE regression")
```

The bias-corrected effect under PEESE is also closer to zero than the random-effects estimate due to the weakly informative prior distributions.

### Weight Function Selection Model

The selection model multiplies the likelihood of each study by a p-value-dependent publication probability.
With the `metafor` package, a one-sided three-parameter model is fit by passing the random-effects fit to `metafor::selmodel()`.

```{r fit-selmodel-metafor}
fit_rma_metafor      <- metafor::rma(yi, vi, data = dat, method = "ML")
fit_selmodel_metafor <- metafor::selmodel(fit_rma_metafor, type = "stepfun", steps = .025, decreasing = TRUE)
fit_selmodel_metafor
```

`bselmodel()` fits the same model in one call.
The default weight function uses a single one-sided cutoff at $0.025$.
Passing `steps = c(0.025, 0.5)` (or any other vector) specifies a finer step function; the prior distribution on the omega vector is uniform across the implied intervals.
The default Dirichlet prior distribution on `omega` in the `RoBMA` R package enforces a monotonically decreasing weight function (non-significant intervals cannot exceed the reference interval), which matches `metafor::selmodel(..., decreasing = TRUE)` above.

```{r fit-selmodel-brma, eval = FALSE}
fit_selmodel_brma <- bselmodel(
  yi = yi, vi = vi, measure = "SMD",
  data = dat, seed = 1
)
```
```{r fit-selmodel-brma-summary}
summary(fit_selmodel_brma, include_mcmc_diagnostics = FALSE)
```

The omega vector is reported with `omega[0, 0.025]` fixed to $1$ as the reference interval; `omega[0.025, 1]` is the publication probability of one-sided p-values above $0.025$ relative to the significant interval.
Values below $1$ indicate that non-significant studies are reported less often than significant ones.

The omega vector has one panel per interval (the reference $[0, 0.025]$ panel shows the spike at $1$).
`plot_weightfunction()` then visualizes the same weights as a step function over the one-sided p-value scale, with the prior band overlaid.

```{r selmodel-omega, fig.width = 7, fig.height = 3.2}
par(mfrow = c(1, 2), mar = c(4, 4, 2, 1), cex.axis = 0.8, cex.lab = 0.8, cex.main = 0.75)
plot(fit_selmodel_brma, parameter = "omega", prior = TRUE, xlim = c(0, 1))
```

```{r selmodel-weightfunction, fig.width = 4.5, fig.height = 3.2}
par(mar = c(4, 4, 2, 1), cex.axis = 0.8, cex.lab = 0.8, cex.main = 0.75)
plot_weightfunction(fit_selmodel_brma, prior = TRUE, main = "Weight-function")
```

The bias-corrected `mu` posterior is shifted toward zero relative to a `brma()` random-effects fit, with the magnitude of the shift driven by the gap between the prior and posterior of `omega[0.025, 1]`.

### Asymmetric Sampling Funnels

`funnel()` extends the standard funnel plot by drawing the *full sampling distribution implied by the fitted model*, including the publication-bias adjustment.
With a no-bias `brma()` fit the funnel is symmetric; with a bias-adjustment model the funnel becomes asymmetric in the direction the model identifies as suppressed.
Comparing the three default fits side by side shows how each adjustment encodes a different mechanism.

```{r funnel-plots, fig.width = 7, fig.height = 3.2}
par(mfrow = c(1, 3), mar = c(4, 4, 2, 1), cex.axis = 0.8, cex.lab = 0.8, cex.main = 0.75)
xlim <- c(-2, 2.5); ylim <- c(0, 0.8)
funnel(fit_PET_brma,      main = "PET",       xlim = xlim, ylim = ylim)
funnel(fit_PEESE_brma,    main = "PEESE",     xlim = xlim, ylim = ylim)
funnel(fit_selmodel_brma, main = "Sel. Mod.", xlim = xlim, ylim = ylim)
```

PET tilts the funnel by adding a linear effect of the standard error, so the centerline shifts away from zero in the direction of the dominant effect as precision drops.
PEESE produces the same kind of shift but curved (quadratic in standard error), so the displacement grows faster for imprecise studies and the funnel walls bow outward.
The selection-model funnel keeps the centerline on the random-effects `mu`, but the implied sampling density on each side of the $p < 0.025$ contour differs: the non-significant region is down-weighted by `omega[0.025, 1]`, so non-significant studies are predicted to appear less often than significant ones at the same precision.

## Flexible Bias Prior Distributions {#flexible-bias-priors}

The default prior distributions above encode a one-sided publication-selection mechanism: PET and PEESE are constrained to be non-negative, and the weight function is constrained to be monotonically non-increasing in p-value.
These restrictions are usually appropriate in practice and help regularize estimates from the bias-adjusted models. Analysts might still be interested in specifying more flexible prior distributions for the bias adjustment.
Each of `bPET()`, `bPEESE()`, and `bselmodel()` accepts a `prior_bias` argument that replaces the default with any user-specified prior distribution of the matching type.
Below we relax each default in turn and place the restricted and relaxed posteriors side by side.
The relaxed prior distributions let the posterior settle wherever the data prefer rather than being pinned against the boundary; conversely, when a relaxed posterior concentrates on the same side as its default counterpart, that agreement is data-driven and not a consequence of the prior constraint.

### PET without truncation

For PET the natural relaxation is to drop the `[0, Inf]` truncation on the bias coefficient.
We keep the default Cauchy family and scale (`Cauchy(0, 1)`) so the comparison isolates the effect of removing the lower bound; the prior distribution now puts equal mass on negative-bias regions, where small studies pull the estimate *up* rather than down.

```{r fit-PET-brma-flex, eval = FALSE}
fit_PET_brma_flex <- bPET(
  yi = yi, vi = vi, measure = "SMD",
  prior_bias = prior_PET(
    distribution = "cauchy",
    parameters = list(location = 0, scale = 1),
    truncation = list(lower = -Inf, upper = Inf)
  ),
  data = dat, seed = 1
)
```

```{r fit-PET-brma-flex-summary}
summary(fit_PET_brma_flex, include_mcmc_diagnostics = FALSE)
```

The two panels below show the prior and posterior of the `PET` coefficient under the default truncated prior distribution (left) and the relaxed full-real-line prior distribution (right).
The dashed line at `PET = 0` marks where the default prior distribution is bounded and where the relaxed prior distribution places mass on both sides.

```{r PET-flex-comparison, fig.width = 7, fig.height = 3.2}
par(mfrow = c(1, 2), mar = c(4, 4, 2, 1), cex.axis = 0.8, cex.lab = 0.8, cex.main = 0.85)
plot(fit_PET_brma,      parameter = "PET", prior = TRUE,
  main = "PET (default)",     xlim = c(-2, 4))
abline(v = 0, lty = 2, col = "grey50")
plot(fit_PET_brma_flex, parameter = "PET", prior = TRUE,
  main = "PET (untruncated)", xlim = c(-2, 4))
abline(v = 0, lty = 2, col = "grey50")
```

### PEESE without truncation

PEESE is relaxed analogously: keep `Cauchy(0, 5)` and remove the lower bound, so the prior distribution is symmetric around zero.

```{r fit-PEESE-brma-flex, eval = FALSE}
fit_PEESE_brma_flex <- bPEESE(
  yi = yi, vi = vi, measure = "SMD",
  prior_bias = prior_PEESE(
    distribution = "cauchy",
    parameters = list(location = 0, scale = 5),
    truncation = list(lower = -Inf, upper = Inf)
  ),
  data = dat, seed = 1
)
```

```{r fit-PEESE-brma-flex-summary}
summary(fit_PEESE_brma_flex, include_mcmc_diagnostics = FALSE)
```

```{r PEESE-flex-comparison, fig.width = 7, fig.height = 3.2}
par(mfrow = c(1, 2), mar = c(4, 4, 2, 1), cex.axis = 0.8, cex.lab = 0.8, cex.main = 0.85)
plot(fit_PEESE_brma, parameter = "PEESE", prior = TRUE, main = "PEESE (default)", xlim = c(-5, 10))
abline(v = 0, lty = 2, col = "grey50")
plot(fit_PEESE_brma_flex, parameter = "PEESE", prior = TRUE, main = "PEESE (untruncated)", xlim = c(-5, 10))
abline(v = 0, lty = 2, col = "grey50")
```

### Selection-model Weight Prior Distributions

The weight on each non-reference p-value bin is built from a *weight prior* helper exported from the `BayesTools` package.
Choosing a different helper changes which weight vectors are *a priori* possible.

| Helper | What the prior distribution allows | Notes |
|:--|:--|:--|
| `wf_cumulative(alpha)` | monotone decreasing weights via a Dirichlet on cumulative bin probabilities; non-significant intervals cannot exceed the most-significant interval | the default; matches `metafor::selmodel(..., decreasing = TRUE)` and the canonical 3PSM-style assumption |
| `wf_fixed(omega)` | the weight vector is fixed to the supplied values; no parameters are sampled | sensitivity analyses with externally specified weights |
| `wf_independent(prior, scale = "omega")` | independent prior distribution placed directly on each non-reference `omega`; the prior distribution must have non-negative support (e.g., `Beta` or a positive-truncated distribution) | non-monotone weights drawn independently from a custom prior distribution |
| `wf_independent(prior, scale = "log_omega")` | independent prior distribution placed on `log(omega)`, with `omega = exp(log_omega)` | non-monotone weights drawn independently from a custom prior distribution |

The two `wf_independent` scales allow different parameterizations.
The `"omega"` scale breaks monotonicity and specifies weights directly on the relative probability scale.
The `"log_omega"` scale allows specifying the prior distributions as multiplicative factors against the baseline weight via log scale.

We refit the selection model with a `Normal(0, 0.5)` prior distribution on `log_omega`, which is centered at "no selection" (`omega = 1`) and places roughly 95% of its prior mass on `omega` between $0.37$ and $2.7$:

```{r fit-selmodel-brma-flex, eval = FALSE}
fit_selmodel_brma_flex <- bselmodel(
  yi = yi, vi = vi, measure = "SMD",
  prior_bias = prior_weightfunction(
    "one-sided",
    steps   = c(0.025),
    weights = wf_independent(
      prior("normal", parameters = list(mean = 0, sd = 0.5)),
      scale = "log_omega"
    )
  ),
  data = dat, seed = 1
)
```

```{r fit-selmodel-brma-flex-summary}
summary(fit_selmodel_brma_flex, include_mcmc_diagnostics = FALSE)
```

The dashed line at `omega = 1` marks the upper bound of the default Dirichlet weight, where the `log_omega` prior distribution places equal mass on both sides.

```{r selmodel-flex-comparison, fig.width = 7, fig.height = 3.2}
par(mfrow = c(1, 2), mar = c(4, 4, 2, 1), cex.axis = 0.8, cex.lab = 0.8, cex.main = 0.85)
plot(fit_selmodel_brma, parameter = "omega", prior = TRUE, main = "omega (default)", xlim = c(0, 3))
abline(v = 1, lty = 2, col = "grey50")
plot(fit_selmodel_brma_flex, parameter = "omega", prior = TRUE, main = "omega (log_omega)", xlim = c(0, 3))
abline(v = 1, lty = 2, col = "grey50")
```

## Other Inference Helpers

`bPET()`, `bPEESE()`, and `bselmodel()` accept the same `mods`, `scale`, and `cluster` arguments as `brma()`, so meta-regression, location-scale, and multilevel bias-adjustment models are directly available. The fits are also `brma` objects, so all summaries, plots, predictions, and diagnostics examined in the [*Bayesian Meta-Analysis*](v02-bayesian-meta-analysis.html) vignette work the same way here.

To average across bias-adjustment models rather than commit to a single one, use `RoBMA()`; see the [*Robust Bayesian Meta-Analysis*](v21-robust-bayesian-meta-analysis.html) vignette for the model-averaged workflow and the available `model_type` presets.

## References

The `RoBMA` R package fits a comprehensive collection of Bayesian meta-analytic models [@RoBMA].
It provides single-model fits, meta-regression, location-scale models, multilevel models, fully Bayesian model-averaged ensembles, and built-in publication-bias adjustment using selection models, PET-PEESE, and Robust Bayesian Meta-Analysis (RoBMA).
Posterior sampling is performed by JAGS through a custom C++ module that implements the weighted distributions used in selection models.
Prior handling, plotting, and Bayesian diagnostics are delegated to the companion `BayesTools` package.

This introduction maps the package at a high level and points to the more detailed vignettes that follow.

## System Requirements

The `RoBMA` R package requires JAGS 4.3.1 or newer [@plummer2003jags].
Install JAGS from <https://mcmc-jags.sourceforge.io/> before installing the R package.

```r
install.packages("RoBMA")
```

**Note on backwards compatibility.**
The 4.0 release is not backwards compatible with earlier versions of the package.
The *Manuscript Companions* section accompanies the methodological papers in which the `RoBMA` R package was originally developed and illustrated; the vignettes there have been updated to the 4.0 version of the package, and the numerical results differ from the published values because the underlying algorithms have changed.
To reproduce the published analyses exactly, install `RoBMA` version 3.6.1 together with `BayesTools` version 0.2.23:

```r
remotes::install_version("BayesTools", version = "0.2.23")
remotes::install_version("RoBMA",      version = "3.6.1")
```

## Main Functions

The package exposes a small set of high-level fitting functions.

| Function              | What it fits                                                               |
|:----------------------|:---------------------------------------------------------------------------|
| [`brma()`](`r reference_url("brma")`), [`brma.norm()`](`r reference_url("brma")`) | Bayesian random-effects meta-analysis                                      |
| [`brma.glmm()`](`r reference_url("brma.glmm")`) | Bayesian GLMM meta-analysis (binomial / log OR or Poisson / log IRR)       |
| [`bselmodel()`](`r reference_url("bselmodel")`) | Bayesian weight-function selection model                                   |
| [`bPET()`](`r reference_url("bPET")`), [`bPEESE()`](`r reference_url("bPEESE")`) | Bayesian PET / PEESE publication-bias adjustment                           |
| [`BMA()`](`r reference_url("BMA")`), [`BMA.norm()`](`r reference_url("BMA")`) | Bayesian model-averaging across presence / absence of effect and heterogeneity (no bias adjustment) |
| [`BMA.glmm()`](`r reference_url("BMA.glmm")`) | Bayesian model-averaging for GLMM meta-analysis                            |
| [`RoBMA()`](`r reference_url("RoBMA")`) | Robust Bayesian model-averaging including publication-bias models          |

`brma()` and `RoBMA()` are the main workhorses.
`brma()` fits a single model in the same spirit as `metafor::rma()`; `RoBMA()` averages across an ensemble that includes publication-bias adjustment.
`BMA()` is `RoBMA()` without the bias-adjustment models.

All fitting functions

- share an effect-size interface (`yi`, `sei` / `vi`, `measure` arguments; apart from glmm),
- can be extended into a meta-regression (`mods` argument),
- can be extended into a location-scale model (`scale` argument),
- can be extended into a multilevel model (`cluster` argument),
- and all produce objects that work with the same set of inference helpers (`summary()`, `plot()`, `predict()`, `loo()`, `funnel()`, `regplot()`, residuals and influence diagnostics, ...).

## Vignette Map

The vignettes are organized into four sections.

### Foundations

| Vignette                  | Topic                                                                 |
|:--------------------------|:----------------------------------------------------------------------|
| [*Introduction to RoBMA*](`r article_url("v00-introduction")`) | Package overview, fitting functions, and vignette map |
| [*Prior Distributions*](`r article_url("v01-prior-distributions")`) | Default, informed, and custom prior distributions; rescaling guidance |
| [*Bayesian Meta-Analysis*](`r article_url("v02-bayesian-meta-analysis")`) | Random-effects meta-analysis with `brma()`, compared with `metafor::rma()` |
| [*Feature Coverage*](`r article_url("v03-feature-coverage")`) | Overview of available functionality across model families             |

The foundations section gives the package overview, prior specification, a baseline `brma()` workflow, and current feature coverage.
The [*Bayesian Meta-Analysis*](`r article_url("v02-bayesian-meta-analysis")`) vignette walks through the BCG-vaccine example from the `metafor` package in `brma()`, covering summaries, meta-regression, marginal means, residuals, influence, LOO, and the standard meta-analytic plots.

### Correspondence with `metafor`

| Vignette                                          | Topic                                                |
|:--------------------------------------------------|:-----------------------------------------------------|
| [*Multilevel Meta-Analysis*](`r article_url("v10-metafor-parity-multilevel")`) | `brma()` with `cluster` and `metafor::rma.mv()` for 3-level data |
| [*Publication-Bias Adjustment*](`r article_url("v11-metafor-parity-publication-bias")`) | `bselmodel()` / `bPET()` / `bPEESE()` and `metafor::selmodel()` |
| [*Location-Scale Models*](`r article_url("v12-metafor-parity-location-scale")`) | `brma()` with `scale` and `metafor::rma.ls()`          |
| [*Generalized Linear Mixed-Effects Meta-Analysis*](`r article_url("v13-metafor-parity-glmm")`) | `brma.glmm()` and `metafor::rma.glmm()` for binomial and Poisson outcomes |

Each vignette starts from a `metafor` package analysis and shows the matching `RoBMA` R package syntax, output, and diagnostics.

### Bayesian Model Averaging

| Vignette                       | Topic                                                                  |
|:-------------------------------|:-----------------------------------------------------------------------|
| [*Bayesian Model Averaging*](`r article_url("v20-bayesian-model-averaging")`) | Accounting for model uncertainty across presence and absence of effect and heterogeneity, with posterior model probabilities and inclusion Bayes factors |
| [*Robust Bayesian Meta-Analysis*](`r article_url("v21-robust-bayesian-meta-analysis")`) | Extending the ensemble to publication-bias models used in `RoBMA()`, with `PSMA`, `PP`, and bespoke ensemble specifications |

The first vignette introduces model averaging on an ensemble of models with and without effect and heterogeneity.
The second extends the ensemble to publication-bias models used in `RoBMA()`.

### Manuscript Companions

| Vignette                                  | Companion paper / dataset                                          |
|:------------------------------------------|:-------------------------------------------------------------------|
| [*Adjusting for Publication Bias Tutorial*](`r article_url("v30-tutorial")`) | @bartos2020adjusting, JASP / R tutorial on `Lui2015`              |
| [*Robust Bayesian Meta-Regression*](`r article_url("v31-robma-metaregression")`) | @bartos2023robust, RoBMA-reg on `Andrews2021`                     |
| [*Multilevel Robust Bayesian Meta-Analysis*](`r article_url("v32-robma-multilevel")`) | @bartos2025robust, Multilevel RoBMA on `Johnides2025`             |
| [*Multilevel Robust Bayesian Meta-Regression*](`r article_url("v33-robma-multilevel-metaregression")`) | Multilevel RoBMA-reg on `Kroupova2021`                            |
| [*Informed Bayesian Meta-Analysis in Medicine*](`r article_url("v34-bma-norm-medicine")`) | @bartos2021bayesian, informed prior distributions for medical meta-analysis (continuous outcomes) |
| [*Informed Bayesian Meta-Analysis with Binary Outcomes*](`r article_url("v35-bma-glmm-medicine")`) | @bartos2023empirical, informed prior distributions for medical meta-analysis (binary and time-to-event outcomes) |
| [*Zplot Publication-Bias Diagnostics*](`r article_url("v36-zplot")`) | @bartos2025zcurve, zplot diagnostics on `Hoppen2025`              |

These vignettes reproduce or update analyses from published papers and serve as references when citing the corresponding methodological work.

## Where to Start

- For readers familiar with the `metafor` package: read [*Bayesian Meta-Analysis*](`r article_url("v02-bayesian-meta-analysis")`), then the *Correspondence with `metafor`* vignettes for the model family they use most.
- For readers new to Bayesian meta-analysis: read [*Prior Distributions*](`r article_url("v01-prior-distributions")`) and then [*Bayesian Meta-Analysis*](`r article_url("v02-bayesian-meta-analysis")`).
- For publication-bias adjustment: read [*Robust Bayesian Meta-Analysis*](`r article_url("v21-robust-bayesian-meta-analysis")`), with [*Adjusting for Publication Bias Tutorial*](`r article_url("v30-tutorial")`) as a worked example.

## References

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/plot.R
\name{plot_pet_peese}
\alias{plot_pet_peese}
\alias{plot_pet_peese.brma}
\title{Plot PET-PEESE Fit of brma Object}
\usage{
plot_pet_peese(x, ...)

\method{plot_pet_peese}{brma}(
  x,
  show_data = TRUE,
  prior = FALSE,
  plot_type = "base",
  dots_prior = NULL,
  ...
)
}
\arguments{
\item{x}{a fitted brma object with a PET or PEESE component.}

\item{...}{list of additional graphical arguments
to be passed to the plotting function. Supported arguments
are \code{lwd}, \code{lty}, \code{col}, \code{col.fill},
\code{xlab}, \code{ylab}, \code{main}, \code{xlim}, \code{ylim},
\code{pch}, \code{bg}, \code{cex}, and \code{size}
to adjust the line thickness, line type, line color, fill color,
x-label, y-label, title, x-axis range, y-axis range, and observed data
point style respectively.}

\item{show_data}{whether the observed effect sizes should be added to
figure. Defaults to \code{TRUE}.}

\item{prior}{whether prior distribution should be added to
figure. Defaults to \code{FALSE}.}

\item{plot_type}{whether to use a base plot \code{"base"}
or ggplot2 \code{"ggplot"} for plotting. Defaults to
\code{"base"}.}

\item{dots_prior}{list of additional graphical arguments
to be passed to the plotting function of the prior
distribution. Supported arguments are \code{lwd},
\code{lty}, \code{col}, and \code{col.fill}, to adjust
the line thickness, line type, line color, and fill color
of the prior distribution respectively.}
}
\value{
\code{plot_pet_peese} returns either \code{NULL} invisibly if
\code{plot_type = "base"} or a ggplot2 object if
\code{plot_type = "ggplot"}.
}
\description{
\code{plot_pet_peese} visualizes posterior
(and prior) PET-PEESE fit of a brma object.
}
\details{
The plot shows observed \code{yi} values against \code{sei}. PET
regression uses \eqn{\mu + PET \cdot se_i}; PEESE regression uses
\eqn{\mu + PEESE \cdot se_i^2}, with the fitted effect direction.
}
\examples{
\dontrun{
if (requireNamespace("metadat", quietly = TRUE)) {
  data(dat.lehmann2018, package = "metadat")

  fit <- bPET(
    yi      = yi,
    vi      = vi,
    data    = dat.lehmann2018,
    measure = "SMD",
    seed    = 1,
    silent  = TRUE
  )

  plot_pet_peese(fit)
}
}


}
\seealso{
\code{\link[=RoBMA]{RoBMA()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/interpret.R
\name{interpret}
\alias{interpret}
\alias{interpret.default}
\alias{interpret.brma}
\alias{print.interpret.brma}
\title{Interpret brma Results}
\usage{
interpret(object, ...)

\method{interpret}{default}(object, ...)

\method{interpret}{brma}(
  object,
  output_measure = NULL,
  transform = NULL,
  conditional = FALSE,
  scope = "core",
  probs = c(0.025, 0.975),
  central = NULL,
  priors = FALSE,
  digits = 3,
  ...
)

\method{print}{interpret.brma}(x, ...)
}
\arguments{
\item{object}{a fitted model object.}

\item{...}{additional arguments passed to methods. The \code{brma} method
reserves \code{...}; unused arguments error, except deprecated
\code{output_scale}, which is accepted with a warning.}

\item{output_measure}{optional effect-size measure used for the pooled
effect only. Supported conversion targets include \code{"SMD"},
\code{"COR"}, \code{"ZCOR"}, and \code{"OR"} when the input scale allows
conversion.}

\item{transform}{optional display transformation. \code{"EXP"} exponentiates
log-scale \code{"OR"}, \code{"RR"}, \code{"HR"}, and \code{"IRR"}
pooled effects; aliases for no transform are accepted.}

\item{conditional}{whether to summarize conditional estimates for RoBMA
product-space objects. Defaults to \code{FALSE}.}

\item{scope}{character vector specifying sections to include. Use
\code{"core"} for the default concise interpretation, \code{"all"} for all
sections, or any of \code{"components"}, \code{"estimates"},
\code{"moderators"}, \code{"scale"}, and \code{"bias"}.}

\item{probs}{two quantiles used for credible intervals. Defaults to
\code{c(.025, .975)}.}

\item{central}{whether estimates are described by posterior mean or median.
Defaults to \code{NULL}, which uses the posterior mean, except for
\code{transform = "EXP"} pooled effects where the posterior mode is used.}

\item{priors}{whether to print prior distributions after the interpretation.
Defaults to \code{FALSE}.}

\item{digits}{number of digits after the decimal point.}

\item{x}{an \code{interpret.brma} object.}
}
\value{
A character vector with class \code{"interpret.brma"}. The
normalized BayesTools interpretation records are stored in the
\code{"records"} attribute; the \code{brma} method also stores
\code{"scope"}, \code{"conditional"}, and optionally \code{"priors"}.
}
\description{
Creates a concise textual interpretation of fitted RoBMA
\code{brma} objects.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/priors.R
\name{prior_weightfunction}
\alias{prior_weightfunction}
\alias{wf_cumulative}
\alias{wf_fixed}
\alias{wf_independent}
\title{Weightfunction Prior}
\usage{
prior_weightfunction(
  side = "one-sided",
  steps = c(0.025, 0.05),
  weights = wf_cumulative(),
  reference = "most_significant",
  prior_weights = 1
)

wf_cumulative(alpha = NULL)

wf_fixed(omega)

wf_independent(prior, scale = "omega")
}
\arguments{
\item{side}{character. Either \code{"one-sided"} or \code{"two-sided"}.}

\item{steps}{numeric vector of p-value cut points. These define
\code{length(steps) + 1} p-value bins and must be ordered values in
\code{(0, 1)}.}

\item{weights}{a weight-prior object created by \code{wf_cumulative()},
\code{wf_fixed()}, or \code{wf_independent()}.}

\item{reference}{character. Reference bin, currently
\code{"most_significant"}.}

\item{prior_weights}{numeric prior model weight.}

\item{alpha}{optional positive cumulative-Dirichlet concentration parameters,
one per p-value bin. If \code{NULL}, \code{prior_weightfunction()} uses
\code{rep(1, length(steps) + 1)}. Cumulative weights encode monotone
decreasing publication weights relative to the most-significant bin.}

\item{omega}{fixed publication weights, one per bin; values must be
non-missing, nonnegative, and match \code{length(steps) + 1} when used in
\code{prior_weightfunction()}.}

\item{prior}{continuous simple prior distribution for each non-reference
weight. Point, discrete, mixture, and other non-simple priors are invalid.}

\item{scale}{latent scale for independent weights; either \code{"omega"},
\code{"log_omega"}, or the \code{"log"} alias. Direct \code{"omega"} priors
need nonnegative support; \code{"log"} is normalized to \code{"log_omega"}.}
}
\value{
\code{prior_weightfunction()} returns an object inheriting from
\code{prior} and \code{prior.weightfunction}; the \code{wf_*()} helpers
return \code{weightfunction_weights} helper objects with subclass markers.
}
\description{
Create weightfunction publication-bias priors and their
weight-prior helper objects.
}
\details{
Fixed weights must have one value per p-value bin
(\code{length(steps) + 1}), and the reference bin must have weight 1.
}
\examples{
prior_weightfunction("one-sided", steps = 0.025)
prior_weightfunction(
  side    = "one-sided",
  steps   = c(0.025, 0.5),
  weights = wf_fixed(c(1, 0.8, 0.6))
)

}
\seealso{
\code{\link{publication_bias_prior_specification}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/datasets.R
\docType{data}
\name{Weingarten2018}
\alias{Weingarten2018}
\title{582 effect sizes examining the ease-of-retrieval effect
from a meta-analysis by \insertCite{weingarten2018does;textual}{RoBMA}}
\format{
A data.frame with 12 columns and 582 observations:
\describe{
\item{\code{r_xy}}{Correlation between manipulation and outcome.}
\item{\code{N}}{Sample size.}
\item{\code{paper_id}}{Paper identifier.}
\item{\code{published}}{Publication-status indicator.}
\item{\code{USA}}{Whether the study was conducted in the USA.}
\item{\code{number_of_few}}{Number of examples in the few-examples condition.}
\item{\code{number_of_many}}{Number of examples in the many-examples condition.}
\item{\code{episodic_memory}}{Whether the trial targeted episodic memory.}
\item{\code{standard_paradigm}}{Whether the standard paradigm was used.}
\item{\code{proximal_dataset}}{Whether the data set was proximal.}
\item{\code{r_xm}}{Correlation between manipulation and mediator.}
\item{\code{r_my}}{Correlation between mediator and outcome.}
}
}
\usage{
Weingarten2018
}
\description{
The data set contains correlation coefficients between the manipulation
and outcome variable (r_xy), sample sizes (N), and various study characteristics
including publication status, country (USA vs. other), number of few and many
examples requested, whether the trial targeted episodic memory, paradigm type (standard vs. other),
dataset type (proximal vs. distal), and mediation variables (r_xm, r_my).
The meta-analysis examined whether subjective ease mediates the ease-of-retrieval effect,
where participants list either few or many examples and then make judgments
\insertCite{weingarten2018does}{RoBMA}.
}
\references{
\insertAllCited{}
}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/plot.R
\name{plot_diagnostic}
\alias{plot_diagnostic}
\alias{plot_diagnostic.brma}
\alias{plot_diagnostic_autocorrelation}
\alias{plot_diagnostic_autocorrelation.brma}
\alias{plot_diagnostic_trace}
\alias{plot_diagnostic_trace.brma}
\alias{plot_diagnostic_density}
\alias{plot_diagnostic_density.brma}
\title{Plot MCMC Diagnostics}
\usage{
plot_diagnostic(x, ...)

\method{plot_diagnostic}{brma}(
  x,
  parameter = NULL,
  parameter_mods = NULL,
  parameter_scale = NULL,
  type,
  plot_type = "base",
  lags = 30,
  ...
)

plot_diagnostic_autocorrelation(x, ...)

\method{plot_diagnostic_autocorrelation}{brma}(
  x,
  parameter = NULL,
  parameter_mods = NULL,
  parameter_scale = NULL,
  type = "autocorrelation",
  plot_type = "base",
  lags = 30,
  ...
)

plot_diagnostic_trace(x, ...)

\method{plot_diagnostic_trace}{brma}(
  x,
  parameter = NULL,
  parameter_mods = NULL,
  parameter_scale = NULL,
  type = "trace",
  plot_type = "base",
  lags = 30,
  ...
)

plot_diagnostic_density(x, ...)

\method{plot_diagnostic_density}{brma}(
  x,
  parameter = NULL,
  parameter_mods = NULL,
  parameter_scale = NULL,
  type = "density",
  plot_type = "base",
  lags = 30,
  ...
)
}
\arguments{
\item{x}{a fitted brma object}

\item{...}{additional graphical arguments passed through RoBMA's diagnostic
setup to \code{BayesTools::JAGS_diagnostics()}.}

\item{parameter}{base parameter to plot. Defaults to \code{NULL}, which uses
\code{"mu"} or the meta-regression intercept. Valid values include
\code{"mu"}, \code{"tau"}, \code{"rho"}, \code{"PET"}, \code{"PEESE"},
and \code{"omega"} or \code{"weightfunction"} when present.}

\item{parameter_mods}{moderator term for location regression.}

\item{parameter_scale}{term for scale regression.}

\item{type}{diagnostic plot type. Convenience wrappers set a type-specific
default but still forward this argument to \code{plot_diagnostic.brma()}.}

\item{plot_type}{whether to use a base plot \code{"base"} or ggplot2
\code{"ggplot"} for plotting. Defaults to \code{"base"}.}

\item{lags}{number of lags for autocorrelation plots. Defaults to 30.}
}
\value{
\code{plot_diagnostic} returns the object returned by
\code{BayesTools::JAGS_diagnostics()}, invisibly for base graphics.
}
\description{
\code{plot_diagnostic} creates visual MCMC diagnostics for a
fitted brma object. Convenience wrappers are available for trace, density,
and autocorrelation plots.
}
\seealso{
\code{\link[=summary.brma]{summary.brma()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/loo.R
\name{loo_compare.loo}
\alias{loo_compare.loo}
\title{Compare loo Objects Using LOO}
\usage{
\method{loo_compare}{loo}(x, ..., unit = "estimate")
}
\arguments{
\item{x}{a RoBMA-targeted \code{loo} or \code{waic} object (the first model
to compare).}

\item{...}{additional RoBMA-targeted \code{loo}/\code{waic} or \code{brma}
objects to compare.}

\item{unit}{output/deletion unit used when extracting LOO from brma objects.}
}
\value{
A matrix of class \code{"compare.loo"} as returned by
\code{\link[loo]{loo_compare}}.
}
\description{
Method for comparing RoBMA-targeted \code{loo} or \code{waic}
objects directly.
}
\seealso{
\code{\link{loo.brma}}, \code{\link[loo]{loo_compare}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/priors.R
\name{prior_informed}
\alias{prior_informed}
\title{Informed Prior}
\usage{
prior_informed(name, parameter = NULL, type = "smd")
}
\arguments{
\item{name}{character. Informed prior name. Supported examples include
\code{"van Erp"}, \code{"Oosterwijk"}, \code{"Cochrane"}, and medicine
subfields from \code{BayesTools::prior_informed_medicine_names}.}

\item{parameter}{character. Parameter subset. Required for medicine priors
and not needed for psychology priors.}

\item{type}{character. Effect-size type. Medicine priors use values such as
\code{"smd"}, \code{"logOR"}, \code{"logRR"}, \code{"RD"}, and
\code{"logHR"}.}
}
\value{
An object of class \code{prior}.
}
\description{
Create empirical informed prior distributions.
}
\details{
Further details can be found in \insertCite{erp2017estimates;textual}{RoBMA},
\insertCite{gronau2017bayesian;textual}{RoBMA},
\insertCite{bartos2021bayesian;textual}{RoBMA}, and
\insertCite{bartos2023empirical;textual}{RoBMA}.
}
\references{
\insertAllCited{}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/plot.R
\name{plot.brma}
\alias{plot.brma}
\title{Plots brma Object}
\usage{
\method{plot}{brma}(
  x,
  parameter,
  parameter_mods,
  parameter_scale,
  prior = FALSE,
  standardized_coefficients = FALSE,
  conditional = FALSE,
  output_measure = NULL,
  transform = NULL,
  plot_type = "base",
  dots_prior = NULL,
  ...
)
}
\arguments{
\item{x}{a fitted \code{brma}, \code{BMA}, or \code{RoBMA} object.}

\item{parameter}{a parameter to be plotted. Defaults to \code{"mu"} for
the effect size, or to the meta-regression intercept when moderators are
present. Additional options are \code{"tau"}, \code{"rho"} for multilevel
models, \code{"PET"}, \code{"PEESE"}, and \code{"omega"} or
\code{"weightfunction"} for selection models. Use \code{plot_pet_peese()}
for PET/PEESE regression plots.}

\item{parameter_mods}{character. Moderator term to plot. Use
\code{"intercept"} for the adjusted effect in meta-regression models.}

\item{parameter_scale}{character. Scale-regression term to plot. Use
\code{"intercept"} for the heterogeneity intercept in location-scale models.}

\item{prior}{whether prior distribution should be added to
figure. Defaults to \code{FALSE}.}

\item{standardized_coefficients}{whether to plot moderator and
scale-regression coefficients on the standardized predictor scale. Defaults
to \code{FALSE}.}

\item{conditional}{whether to plot the conditional posterior distribution
for RoBMA product-space objects. Defaults to \code{FALSE}.}

\item{output_measure}{effect-size measure for location/effect predictions.
Defaults to the fitted measure. Supported conversions are among \code{"SMD"},
\code{"COR"}, \code{"ZCOR"}, and \code{"OR"}; \code{"RR"}, \code{"HR"},
\code{"IRR"}, \code{"RD"}, and \code{"GEN"} can only be returned on their
fitted measure. Use \code{transform = "EXP"} for ratio-scale output from
log-scale measures.}

\item{transform}{optional display transformation. Currently \code{"EXP"}
exponentiates log-scale measures \code{"OR"}, \code{"RR"}, \code{"HR"},
and \code{"IRR"}.}

\item{plot_type}{whether to use a base plot \code{"base"}
or ggplot2 \code{"ggplot"} for plotting. Defaults to
\code{"base"}.}

\item{dots_prior}{list of additional graphical arguments
to be passed to the plotting function of the prior
distribution. Supported arguments are \code{lwd},
\code{lty}, \code{col}, and \code{col.fill}, to adjust
the line thickness, line type, line color, and fill color
of the prior distribution respectively.}

\item{...}{list of additional graphical arguments
to be passed to the plotting function. Supported arguments
are \code{lwd}, \code{lty}, \code{col}, \code{col.fill},
\code{xlab}, \code{ylab}, \code{main}, \code{xlim}, \code{ylim}
to adjust the line thickness, line type, line color, fill color,
x-label, y-label, title, x-axis range, and y-axis range
respectively.}
}
\value{
\code{plot.brma} returns either \code{NULL} if \code{plot_type = "base"}
or a \code{ggplot2} object if \code{plot_type = "ggplot"}.
}
\description{
\code{plot.brma} visualizes posterior
(and prior) distribution a brma object.
}
\examples{
\dontrun{
if (requireNamespace("metadat", quietly = TRUE)) {
  data(dat.lehmann2018, package = "metadat")
  fit <- bPET(yi = yi, vi = vi, data = dat.lehmann2018, measure = "SMD")

  plot(fit, parameter = "mu")
  plot(fit, parameter = "tau", prior = TRUE)
  plot(fit, parameter = "PET")
}
}


}
\seealso{
\code{\link[=RoBMA]{RoBMA()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/brma_samples.R
\name{print.summary.brma_samples}
\alias{print.summary.brma_samples}
\title{Print summary.brma_samples Object}
\usage{
\method{print}{summary.brma_samples}(x, probs = NULL, ...)
}
\arguments{
\item{x}{a \code{summary.brma_samples} object}

\item{probs}{quantiles for credible intervals. If \code{NULL}, uses
the default stored in the object (typically \code{c(.025, .975)})}

\item{...}{additional arguments passed to
\code{BayesTools::ensemble_estimates_table}}
}
\value{
Returns the summary table invisibly.
}
\description{
Prints a summary table of posterior samples using
\code{BayesTools::ensemble_estimates_table}. Returns the summary table
invisibly.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/dffits.R
\name{dffits.brma}
\alias{dffits.brma}
\alias{dffits}
\title{DFFITS for brma Objects}
\usage{
\method{dffits}{brma}(model, ...)
}
\arguments{
\item{model}{a fitted normal-outcome \code{brma} object without a
weightfunction component.}

\item{...}{additional arguments (currently ignored).}
}
\value{
A named numeric vector of DFFITS values, one for each observation.
}
\description{
Computes DFFITS (Difference in FITS, standardized) for a
fitted brma object. DFFITS measures how much the fitted value for observation
\eqn{i} changes if observation \eqn{i} is removed, standardized by the
estimated standard error of the fit.
}
\details{
DFFITS values are computed as a PSIS leave-one-out deletion diagnostic. For
each observation \eqn{i}, the leave-one-out posterior mean fitted value at
that observation is estimated with normalized PSIS weights and compared to
the full-posterior fitted value:
\deqn{DFFITS_i =
  \frac{\hat{\mu}_i - \hat{\mu}_{i(-i)}}{SD_{(-i)}(\mu_i)}}

This targets deletion influence on fitted values directly. It does not use
LOO-PIT residuals, which are predictive outlier diagnostics rather than
fitted-value deletion diagnostics.

Estimate-unit LOO must first be computed with
\code{model <- add_loo(model, unit = "estimate")}. If the leave-one-out
posterior SD of a fitted value is near zero, the corresponding DFFITS value
is returned as \code{NA}.
}
\examples{
\dontrun{
if (requireNamespace("metadat", quietly = TRUE)) {
  data(dat.lehmann2018, package = "metadat")
  fit <- bPET(yi = yi, vi = vi, data = dat.lehmann2018, measure = "SMD")
  fit <- add_loo(fit)

  dffits(fit)
}
}

}
\seealso{
\code{\link{influence.brma}}, \code{\link{cooks.distance.brma}}, \code{\link{hatvalues.brma}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/BMA.glmm.R
\name{BMA.glmm}
\alias{BMA.glmm}
\title{Bayesian Model-Averaged Generalized Meta-Analysis}
\usage{
BMA.glmm(
  ai,
  bi,
  ci,
  di,
  n1i,
  n2i,
  x1i,
  x2i,
  t1i,
  t2i,
  weights,
  mods,
  scale,
  cluster,
  data,
  slab,
  subset,
  measure = "OR",
  prior_effect,
  prior_heterogeneity,
  prior_mods,
  prior_scale,
  prior_heterogeneity_allocation,
  prior_baserate,
  prior_lograte,
  prior_effect_null,
  prior_heterogeneity_null,
  prior_mods_null,
  prior_scale_null,
  prior_heterogeneity_allocation_null,
  standardize_continuous_predictors = TRUE,
  set_contrast_factor_predictors = "treatment",
  prior_unit_information_sd,
  rescale_priors = 1,
  prior_informed_field,
  prior_informed_subfield,
  sample = 5000,
  burnin = 2000,
  adapt = 500,
  chains = 3,
  thin = 1,
  parallel = FALSE,
  autofit = FALSE,
  autofit_control = set_autofit_control(),
  convergence_checks = set_convergence_checks(),
  seed = NULL,
  silent = TRUE,
  ...
)
}
\arguments{
\item{ai}{a vector of the number of events in the treatment or experimental
group for binomial GLMM models.}

\item{bi}{a vector of the number of non-events in the treatment or
experimental group for binomial GLMM models.}

\item{ci}{a vector of the number of events in the control group for binomial
GLMM models.}

\item{di}{a vector of the number of non-events in the control group for
binomial GLMM models.}

\item{n1i}{a vector of the sample size in the treatment or experimental
group. If omitted for binomial GLMMs, it is computed as \code{ai + bi}.}

\item{n2i}{a vector of the sample size in the control group. If omitted for
binomial GLMMs, it is computed as \code{ci + di}.}

\item{x1i}{a vector of the number of events in the treatment/experimental group
(for Poisson data).}

\item{x2i}{a vector of the number of events in the control group (for Poisson data).}

\item{t1i}{a vector of the person-time in the treatment/experimental group.}

\item{t2i}{a vector of the person-time in the control group.}

\item{weights}{an optional vector of positive likelihood weights. For
normal/effect-size models, each weight powers the estimate likelihood. For
constructors with GLMM raw-count input, each weight powers the paired
two-arm likelihood for one study.}

\item{mods}{an optional matrix, data.frame, or formula specifying
location moderators (meta-regressors). Formula input is evaluated in \code{data}.}

\item{scale}{an optional matrix, data.frame, or formula specifying
scale predictors for location-scale models. Formula input is evaluated in
\code{data}.}

\item{cluster}{an optional vector of cluster identifiers for multilevel
meta-analysis.}

\item{data}{an optional data frame containing the variables.}

\item{slab}{an optional vector of study labels.}

\item{subset}{an optional logical or numeric vector specifying a subset of
data to be used.}

\item{measure}{a character string specifying the effect size measure.
Normal/effect-size constructors require an explicit value and support
\code{"SMD"}, \code{"ZCOR"}, \code{"RR"}, \code{"OR"}, \code{"HR"}, \code{"RD"}, \code{"IRR"}, and \code{"GEN"}.
Use \code{"GEN"} only for general effect sizes without a known unit information
standard deviation. GLMM raw-count constructors support only \code{"OR"} and
\code{"IRR"} and default to \code{"OR"}.}

\item{prior_effect}{prior distribution(s) for the alternative effect
component(s).}

\item{prior_heterogeneity}{prior distribution(s) for the alternative
heterogeneity component(s).}

\item{prior_mods}{prior distribution(s) for alternative moderator
components. A single prior applies to all terms; a named list can specify
term-specific components.}

\item{prior_scale}{prior distribution(s) for alternative scale-regression
components. A single prior applies to all terms; a named list can specify
term-specific components.}

\item{prior_heterogeneity_allocation}{prior distribution(s) for the
alternative cluster-level heterogeneity allocation component(s).}

\item{prior_baserate}{prior distribution for the estimate-specific midpoint
base-rate probability in binomial GLMM models. If omitted or \code{NULL}, defaults
to independent \code{Beta(1, 1)} priors.}

\item{prior_lograte}{prior distribution for the estimate-specific midpoint
log-rate in Poisson GLMM models. If omitted or \code{NULL}, a data-based
unit-information normal prior is used independently for each estimate.}

\item{prior_effect_null}{prior distribution(s) for the null effect
component(s).}

\item{prior_heterogeneity_null}{prior distribution(s) for the null
heterogeneity component(s).}

\item{prior_mods_null}{prior distribution(s) for null moderator components.
A single prior applies to all terms; a named list can specify term-specific
components.}

\item{prior_scale_null}{prior distribution(s) for null scale-regression
components. A single prior applies to all terms; a named list can specify
term-specific components.}

\item{prior_heterogeneity_allocation_null}{prior distribution(s) for the
null cluster-level heterogeneity allocation component(s).}

\item{standardize_continuous_predictors}{logical. Whether to standardize continuous predictors.
Defaults to \code{TRUE}.}

\item{set_contrast_factor_predictors}{character. How to set contrast for factor predictors.
Defaults are constructor-specific and shown in each function usage; single-model
constructors use \code{"treatment"}, while model-averaging constructors use \code{"meandif"}.}

\item{prior_unit_information_sd}{numeric. The unit information standard deviation (\eqn{\sigma_{unit}}).
Cannot be used together with \code{prior_informed_field}.}

\item{rescale_priors}{numeric. A scaling factor for supported prior distributions.
Point and none priors are unchanged. For constructors with publication-bias
prior distributions, \code{rescale_priors} does not rescale them except for the
default PEESE prior's UISD adjustment. Defaults to 1.}

\item{prior_informed_field}{character. The field of the informed prior distributions.
Omit to use the standard default prior specification; explicit \code{NULL} is invalid.}

\item{prior_informed_subfield}{character. The subfield of the informed prior distributions.
Omit to use the field-specific default, such as \code{"Cochrane"} for
\code{prior_informed_field = "medicine"}; explicit \code{NULL} is invalid.}

\item{sample}{numeric. Number of MCMC samples to save. Defaults to \code{5000}.}

\item{burnin}{numeric. Number of burn-in iterations. Defaults to \code{2000}.}

\item{adapt}{numeric. Number of adaptation iterations. Defaults to \code{500}.}

\item{chains}{numeric. Number of MCMC chains. Defaults to \code{3}.}

\item{thin}{numeric. Thinning interval. Defaults to \code{1}.}

\item{parallel}{logical. Whether to run MCMC chains in parallel. Defaults to \code{FALSE}.}

\item{autofit}{logical. Whether to automatically extend the MCMC chains if convergence is not met.
Defaults to \code{FALSE}.}

\item{autofit_control}{list of autofit control settings. See \code{\link[=set_autofit_control]{set_autofit_control()}} for details.}

\item{convergence_checks}{list of convergence check settings. See \code{\link[=set_convergence_checks]{set_convergence_checks()}} for details.}

\item{seed}{numeric. Random seed for reproducibility. Defaults to \code{NULL}.}

\item{silent}{logical. Whether to suppress output. Constructors with no
explicit default use \code{RoBMA.get_option("silent")} when \code{silent} is omitted.
Model-averaging wrappers default to \code{TRUE} unless explicitly changed.}

\item{...}{additional advanced arguments. Fitting functions reject unused
arguments; currently recognized internal arguments include \code{only_data},
\code{only_priors}, \code{is_JASP}, and \code{is_JASP_prefix}.}
}
\value{
A fitted object of class
\code{c("BMA.glmm", "RoBMA", "brma.glmm", "brma")}. The object contains checked
\code{data}, checked mixture \code{priors}, the JAGS \code{fit}, cached \code{summary}, and
cached \code{coefficients}.
}
\description{
Function for fitting Bayesian model-averaged meta-analytic models
directly to binary or count data using a generalized linear mixed model (GLMM)
framework. Unlike \code{\link{RoBMA}}, this function does not adjust for
publication bias, as weight function and regression-based bias adjustment
methods are not available for GLMM outcomes.
}
\details{
\code{BMA.glmm} combines the data input style of \code{\link{brma.glmm}} with
the mixture prior specification of \code{\link{RoBMA}} for Bayesian model-averaging.

Model for odds ratios (\code{measure = "OR"}) uses a binomial-normal model
as described in \insertCite{jackson2018comparison;textual}{RoBMA}.

Model for incidence rate ratios (\code{measure = "IRR"}) uses a Poisson-normal
model as described in \insertCite{bagos2009mixed;textual}{RoBMA}.

When \code{weights} are supplied, they are treated as likelihood weights on
the paired two-arm study contribution.
}
\examples{
\dontrun{
if (requireNamespace("metadat", quietly = TRUE)) {
  data(dat.bcg, package = "metadat")

  fit <- BMA.glmm(
    ai      = tpos,
    bi      = tneg,
    ci      = cpos,
    di      = cneg,
    data    = dat.bcg,
    measure = "OR",
    seed    = 1,
    silent  = TRUE
  )

  summary(fit)
}
}

}
\seealso{
\code{\link[=brma.glmm]{brma.glmm()}} \code{\link[=RoBMA]{RoBMA()}} \code{\link[=summary.brma]{summary.brma()}} \code{\link[=plot.brma]{plot.brma()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/input-data.R
\name{data_input}
\alias{data_input}
\title{Input Data Specification}
\arguments{
\item{yi}{a vector of effect sizes, or a formula with the effect size on the
left-hand side and location moderators on the right-hand side (for example
\code{yi ~ x1 + x2}). If a formula is supplied, \code{mods} must not be specified.}

\item{vi}{a vector of sampling variances. Either \code{vi} or \code{sei} must be
supplied for normal models.}

\item{sei}{a vector of standard errors. Either \code{vi} or \code{sei} must be
supplied for normal models.}

\item{weights}{an optional vector of positive likelihood weights. For
normal/effect-size models, each weight powers the estimate likelihood. For
constructors with GLMM raw-count input, each weight powers the paired
two-arm likelihood for one study.}

\item{ni}{an optional vector of sample sizes. Used for \code{measure = "GEN"}
or when estimating \code{"UISD"}).}

\item{mods}{an optional matrix, data.frame, or formula specifying
location moderators (meta-regressors). Formula input is evaluated in \code{data}.}

\item{scale}{an optional matrix, data.frame, or formula specifying
scale predictors for location-scale models. Formula input is evaluated in
\code{data}.}

\item{cluster}{an optional vector of cluster identifiers for multilevel
meta-analysis.}

\item{data}{an optional data frame containing the variables.}

\item{slab}{an optional vector of study labels.}

\item{subset}{an optional logical or numeric vector specifying a subset of
data to be used.}

\item{measure}{a character string specifying the effect size measure.
Normal/effect-size constructors require an explicit value and support
\code{"SMD"}, \code{"ZCOR"}, \code{"RR"}, \code{"OR"}, \code{"HR"}, \code{"RD"}, \code{"IRR"}, and \code{"GEN"}.
Use \code{"GEN"} only for general effect sizes without a known unit information
standard deviation. GLMM raw-count constructors support only \code{"OR"} and
\code{"IRR"} and default to \code{"OR"}.}

\item{effect_direction}{direction used by publication-bias adjustments.
\code{"positive"} assumes statistically significant positive estimates are more
likely to be selected; \code{"negative"} mirrors the selection direction;
\code{"detect"} infers the direction from the fitted data.}

\item{ai}{a vector of the number of events in the treatment or experimental
group for binomial GLMM models.}

\item{bi}{a vector of the number of non-events in the treatment or
experimental group for binomial GLMM models.}

\item{ci}{a vector of the number of events in the control group for binomial
GLMM models.}

\item{di}{a vector of the number of non-events in the control group for
binomial GLMM models.}

\item{n1i}{a vector of the sample size in the treatment or experimental
group. If omitted for binomial GLMMs, it is computed as \code{ai + bi}.}

\item{n2i}{a vector of the sample size in the control group. If omitted for
binomial GLMMs, it is computed as \code{ci + di}.}

\item{x1i}{a vector of the number of events in the treatment/experimental group
(for Poisson data).}

\item{x2i}{a vector of the number of events in the control group (for Poisson data).}

\item{t1i}{a vector of the person-time in the treatment/experimental group.}

\item{t2i}{a vector of the person-time in the control group.}
}
\description{
Shared data-input arguments used by the RoBMA fitting functions.

Normal models use approximate effect-size estimates supplied through
\code{yi} with either \code{vi} or \code{sei}. GLMM models use the raw two-arm count
arguments for binomial (\code{measure = "OR"}) or Poisson (\code{measure = "IRR"})
outcomes.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/wrappers.R
\name{fitted.brma}
\alias{fitted.brma}
\title{Fitted Values for brma Objects}
\usage{
\method{fitted}{brma}(
  object,
  unit = "estimate",
  conditioning_depth = "marginal",
  component = "location",
  bias_adjusted = FALSE,
  output_measure = NULL,
  transform = NULL,
  conditional = FALSE,
  ...
)
}
\arguments{
\item{object}{a fitted brma object.}

\item{unit}{output unit. Only \code{"estimate"} is implemented currently.}

\item{conditioning_depth}{conditioning depth for location fitted values.
\code{"marginal"} uses fixed effects only, \code{"cluster"} conditions on
cluster-level random effects, and \code{"estimate"} conditions on the full
estimate-level fitted value.}

\item{component}{fitted component to return. \code{"location"} returns
location fitted values, \code{"scale"} returns fitted heterogeneity
\eqn{\tau_i}, and \code{"all"} returns both as a named list.}

\item{bias_adjusted}{whether location fitted values should adjust for
publication bias. Defaults to \code{FALSE}.}

\item{output_measure}{effect-size measure for location/effect predictions.
Defaults to the fitted measure. Supported conversions are among \code{"SMD"},
\code{"COR"}, \code{"ZCOR"}, and \code{"OR"}; \code{"RR"}, \code{"HR"},
\code{"IRR"}, \code{"RD"}, and \code{"GEN"} can only be returned on their
fitted measure. Use \code{transform = "EXP"} for ratio-scale output from
log-scale measures.}

\item{transform}{optional display transformation. Currently \code{"EXP"}
exponentiates log-scale measures \code{"OR"}, \code{"RR"}, \code{"HR"},
and \code{"IRR"}.}

\item{conditional}{whether to return fitted values from conditional posterior
predictions for RoBMA product-space objects.}

\item{...}{additional arguments. Currently only \code{quiet} is honored.}
}
\value{
A numeric vector of fitted values, or a named list with
\code{location} and \code{scale} components when \code{component = "all"}.
}
\description{
Extract in-sample fitted values from a fitted \code{brma}
object.
}
\details{
This method is a compact adapter around \code{\link{predict.brma}}. It
summarizes posterior prediction draws by their column means and returns a
base numeric vector, matching the usual \code{\link[stats]{fitted}} contract.
Use \code{predict()} directly when posterior draws or intervals are needed.

The default \code{conditioning_depth = "marginal"} corresponds to
\code{predict(object, type = "terms")} and matches the usual fitted-value
convention for meta-regression. For normal models,
\code{conditioning_depth = "estimate"} corresponds to BLUP means for the
observed estimates.

For \code{component = "all"}, \code{conditioning_depth},
\code{output_measure}, and \code{transform} apply only to the
\code{location} component. The \code{scale} component always returns
fitted \eqn{\tau_i} values.
}
\seealso{
\code{\link[=predict.brma]{predict.brma()}}, \code{\link[=residuals.brma]{residuals.brma()}}, \code{\link[=blup.brma]{blup.brma()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/qqnorm.R
\name{qqnorm.brma}
\alias{qqnorm.brma}
\title{Normal QQ Plot for brma Object}
\usage{
\method{qqnorm}{brma}(
  y,
  type = "rstudent",
  unit = "estimate",
  conditioning_depth = "marginal",
  envelope = TRUE,
  conf_level = 95,
  bonferroni = FALSE,
  reps = 1000,
  smooth = TRUE,
  xlim,
  ylim,
  xlab,
  ylab,
  plot_type = "base",
  ...
)
}
\arguments{
\item{y}{a fitted brma object.}

\item{type}{the type of standardized residuals to use. Options are:
\itemize{
\item \code{"rstudent"} (alias: \code{"LOO-PIT"}; default): Leave-one-out
probability integral transform residuals. Works for all model types
including GLMMs and selection models. This is the recommended type for
Bayesian models as it properly accounts for estimation uncertainty and
leverage. Note: requires that loo has been computed (see
\code{\link{add_loo}}).
\item \code{"rstandard"}: Internally standardized residuals using the hat
matrix. Only available for normal outcome models without selection
(weightfunction) bias adjustment.
}}

\item{unit}{output unit. Only \code{"estimate"} is implemented currently.}

\item{conditioning_depth}{residual conditioning depth. Options are
\code{"marginal"}, \code{"cluster"}, and \code{"estimate"}.}

\item{envelope}{logical indicating whether to add a pseudo confidence
envelope to the plot. The envelope is created by simulating sets of
standard normal quantiles, providing a reference for the expected variability
under a correctly specified model. Defaults to \code{TRUE}.}

\item{conf_level}{numeric value between 0 and 100 specifying the confidence level
for the envelope bounds. Defaults to \code{95}.}

\item{bonferroni}{logical indicating whether to apply Bonferroni correction
to the envelope, so it can be regarded as a simultaneous confidence region
for all \eqn{k} residuals. Defaults to \code{FALSE}.}

\item{reps}{retained for compatibility with earlier simulation-based
envelopes. The current envelope is deterministic and does not use this
argument.}

\item{smooth}{logical indicating whether to smooth the envelope bounds using
Friedman's SuperSmoother (\code{\link[stats]{supsmu}}). Defaults to
\code{TRUE}.}

\item{xlim}{x-axis limits. If not specified, limits are computed from data.}

\item{ylim}{y-axis limits. If not specified, limits are computed from data.}

\item{xlab}{title for the x-axis. If not specified, defaults to
\code{"Theoretical Quantiles"}.}

\item{ylab}{title for the y-axis. If not specified, defaults to
\code{"Sample Quantiles"}.}

\item{plot_type}{whether to use a base plot \code{"base"} or ggplot2
\code{"ggplot"} for plotting. Defaults to \code{"base"}.}

\item{...}{additional graphical arguments to customize the plot appearance:
\describe{
\item{pch}{point symbol (default: 21, filled circle)}
\item{col}{point border color (default: "black")}
\item{bg}{point fill/background color (default: "#A6A6A6")}
\item{cex}{point size multiplier for base graphics (default: 1)}
\item{size}{point size for ggplot2 (default: 2)}
\item{shade}{fill color for the envelope region (default: "grey90").
Set to \code{NA} to suppress shading.}
\item{col.envelope}{color for envelope boundary lines
(default: "grey50")}
\item{col.refline}{color of the diagonal reference line
(default: "black")}
\item{main}{character string for plot title (default: no title)}
\item{as_data}{if \code{TRUE}, returns plot data instead of creating
the plot}
}}
}
\value{
For \code{plot_type = "base"}, returns an invisible list with
components \code{x} (theoretical quantiles) and \code{y} (sorted
residuals). For \code{plot_type = "ggplot"}, returns a ggplot object.

If \code{as_data = TRUE}, returns a list with all computed plot data
including: \code{x}, \code{y}, \code{points}, \code{refline},
\code{envelope}, \code{xlim}, \code{ylim}, \code{xlab}, and \code{ylab}.
}
\description{
\code{qqnorm.brma} creates a normal QQ plot for a fitted brma
object. The plot displays sorted standardized residuals on the vertical axis
against the theoretical quantiles of a standard normal distribution on the
horizontal axis. Under a correctly specified model, the points should fall
approximately along the diagonal reference line.
}
\details{
The normal QQ plot is a diagnostic tool for assessing whether the
standardized residuals follow a standard normal distribution, as expected
under a correctly specified model. Each point represents one estimate,
plotted at coordinates \eqn{(\Phi^{-1}(p_i), z_{(i)})} where
\eqn{p_i} are the plotting positions and \eqn{z_{(i)}} are the sorted
standardized residuals.

The default residual type is \code{"rstudent"} (LOO-PIT), which differs
from \code{metafor::qqnorm} (which defaults to \code{"rstandard"}).
LOO-PIT residuals are preferred because they are available for all model
types (including GLMMs and selection models) and properly account for
estimation uncertainty via leave-one-out cross-validation.

The confidence envelope is computed in closed form from the distribution of
normal order statistics. For the \eqn{i}-th sorted residual,
\eqn{\Phi(Z_{(i)})} follows a beta distribution with shape parameters
\eqn{i} and \eqn{k + 1 - i}. Pointwise quantile bounds are transformed back
with \eqn{\Phi^{-1}}. When \code{smooth = TRUE}, bounds are smoothed using
Friedman's SuperSmoother. When \code{bonferroni = TRUE}, the confidence
level is adjusted for simultaneous inference across all \eqn{k} comparisons.

For GLMM models, LOO-PIT residuals and the QQ plot are computed on the
approximate effect-size scale used by \code{loo}; they are not exact
count-scale PIT diagnostics.
}
\examples{
\dontrun{
if (requireNamespace("metadat", quietly = TRUE)) {
  data(dat.lehmann2018, package = "metadat")
  fit <- brma(
    yi      = yi,
    vi      = vi,
    data    = dat.lehmann2018,
    measure = "SMD",
    seed    = 1,
    silent  = TRUE
  )
  fit <- add_loo(fit)

  qqnorm(fit)
  qqnorm(fit, type = "rstandard")
  qqnorm(fit, envelope = FALSE)
  qqnorm(fit, plot_type = "ggplot")
  qqnorm(fit, pch = 19, col = "blue", shade = "lightblue")
}
}

}
\seealso{
\code{\link[=rstudent.brma]{rstudent.brma()}}, \code{\link[=rstandard.brma]{rstandard.brma()}}, \code{\link[=residuals.brma]{residuals.brma()}},
\code{\link[=funnel.brma]{funnel.brma()}}, \code{\link[=radial.brma]{radial.brma()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/loo.R
\name{logLik.brma}
\alias{logLik.brma}
\title{Extract Log-Likelihood Matrix from brma Object}
\usage{
\method{logLik}{brma}(object, unit = "estimate", ...)
}
\arguments{
\item{object}{a brma model object.}

\item{unit}{output unit. See \code{\link{add_loo}}.}

\item{...}{currently unused.}
}
\value{
An S x K or S x G matrix of log-likelihood values.
}
\description{
Extract the pointwise log-likelihood matrix from a brma model
object. This is an S x K or S x G matrix where S is the number of posterior
samples, K is the number of estimates, and G is the number of clusters.
This method implements the S3
\'logLik\' generic for \code{brma} objects and returns the matrix of
pointwise log-likelihoods (one column per observation, one row per sample).
}
\details{
The log-likelihood is computed for each observation at each posterior sample.
For binomial and Poisson models, each observation consists of a pair of
counts (ai/ci or x1i/x2i) that together define a single effect size estimate.
}
\seealso{
\code{\link{loo.brma}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/plot.R
\name{print_prior}
\alias{print_prior}
\alias{print_prior.prior}
\alias{print_prior.brma}
\title{Print Prior Distributions}
\usage{
print_prior(x, ...)

\method{print_prior}{prior}(x, ...)

\method{print_prior}{brma}(x, parameter, parameter_mods, parameter_scale, ...)
}
\arguments{
\item{x}{a \code{brma}, \code{BMA}, or \code{RoBMA} object, or a prior
distribution object.}

\item{...}{additional arguments passed to the prior printing method. Use
\code{silent = TRUE} for programmatic inspection without console output.}

\item{parameter}{character. Base parameter to print. If omitted, all stored
prior distributions are printed.
Common options are \code{"mu"}, \code{"tau"}, \code{"rho"}, \code{"PET"},
\code{"PEESE"}, \code{"omega"}, and \code{"bias"}, with aliases
\code{"effect"} = \code{"mu"}, \code{"heterogeneity"} = \code{"tau"},
and \code{"weightfunction"} = \code{"omega"}. GLMM outcome priors
\code{"pi"} and \code{"phi"} are available when present. Moderator and
scale terms can also be selected by name when unambiguous. A character vector
requests multiple base parameters.}

\item{parameter_mods}{character. Moderator term to print.
Use \code{"intercept"} for the adjusted effect in meta-regression models.}

\item{parameter_scale}{character. Scale-regression term to print.
Use \code{"intercept"} for the heterogeneity intercept in location-scale models.}
}
\value{
\code{print_prior} invisibly returns the selected prior distribution.
If multiple parameters are requested, a named list of prior distributions is
returned invisibly.
}
\description{
\code{print_prior} prints prior distributions stored in
\code{brma}, \code{BMA}, and \code{RoBMA} objects.
This is a user-facing helper for inspecting priors without extracting them
from the internal \code{$priors} list.
}
\examples{
\dontrun{
if (requireNamespace("metadat", quietly = TRUE)) {
  data(dat.lehmann2018, package = "metadat")
  priors <- BMA(
    yi          = yi,
    vi          = vi,
    mods        = ~ Preregistered,
    data        = dat.lehmann2018,
    measure     = "SMD",
    only_priors = TRUE
  )

  print_prior(priors)
  print_prior(priors, parameter = "mu")
  print_prior(priors, parameter = "tau")
  print_prior(priors, parameter_mods = "Preregistered")
}
}

}
\seealso{
\code{\link[=plot_prior]{plot_prior()}} \code{\link[=BMA]{BMA()}} \code{\link[=RoBMA]{RoBMA()}} \code{\link[=brma]{brma()}} \code{\link[=prior]{prior()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/datasets.R
\docType{data}
\name{Anderson2010}
\alias{Anderson2010}
\title{23 experimental studies from
\insertCite{anderson2010violent;textual}{RoBMA} that meet the best practice criteria}
\format{
A data.frame with 3 columns and 23 observations:
\describe{
\item{\code{r}}{Correlation coefficient.}
\item{\code{n}}{Sample size.}
\item{\code{name}}{Study label.}
}
}
\source{
\url{https://github.com/Joe-Hilgard/Anderson-meta}
}
\usage{
Anderson2010
}
\description{
The data set contains correlation coefficients, sample
sizes, and labels for 23 experimental studies focusing on the effect of
violent video games on aggressive behavior.
}
\references{
\insertAllCited{}
}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/bridgesampling.R
\name{bridge_sampler.brma}
\alias{bridge_sampler.brma}
\title{Bridge Sampling for brma Objects}
\usage{
\method{bridge_sampler}{brma}(samples, ...)
}
\arguments{
\item{samples}{a brma model object.}

\item{...}{additional arguments (currently not used).}
}
\value{
An object of class \code{"bridge"} as returned by
\code{\link[bridgesampling]{bridge_sampler}}.
}
\description{
Extract the marginal likelihood bridge sampling object from
a brma model. The marginal likelihood must first be computed using
\code{\link{add_marglik}}.
}
\details{
This function extracts the bridge sampling object that was previously
computed and stored using \code{\link{add_marglik}}. If the marginal
likelihood has not been computed, an error is thrown.
Product-space model-averaging objects (\code{BMA.norm}, \code{BMA.glmm},
and \code{RoBMA}) do not expose bridge-sampling marginal likelihoods.

The returned object can be used for Bayesian model comparison via
\code{\link{bf}} and \code{\link{post_prob}}.
}
\examples{
\dontrun{
if (requireNamespace("metadat", quietly = TRUE)) {
  data(dat.lehmann2018, package = "metadat")
  fit <- brma(yi = yi, vi = vi, data = dat.lehmann2018, measure = "SMD")

  fit <- add_marglik(fit)

  bridge <- bridge_sampler(fit)
  print(bridge)
}
}

}
\seealso{
\code{\link{add_marglik}}, \code{\link[bridgesampling]{bridge_sampler}},
\code{\link{logml.brma}}, \code{\link{bf.brma}}, \code{\link{post_prob.brma}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/update.R
\name{update.brma}
\alias{update.brma}
\title{Update a brma Fit}
\usage{
\method{update}{brma}(
  object,
  formula. = NULL,
  ...,
  sample_extend = NULL,
  slab = NULL,
  autofit_control = NULL,
  convergence_checks = NULL,
  recompute = c("all", "drop"),
  parallel = NULL,
  cores = NULL,
  silent = NULL,
  seed = NULL,
  evaluate = TRUE
)
}
\arguments{
\item{object}{a fitted \code{brma} object.}

\item{formula.}{unsupported; included for compatibility with
\code{\link[stats]{update}}.}

\item{...}{unsupported additional arguments.}

\item{sample_extend}{integer. Number of additional samples per chain.}

\item{slab}{optional character vector of study labels. Updating labels does
not refit or extend the model.}

\item{autofit_control}{list of autofit control settings. Values are merged
with the existing settings before extending.}

\item{convergence_checks}{list of convergence check settings. Values are
merged with the existing settings and used to re-check the fit.}

\item{recompute}{whether cached \code{loo}, \code{waic}, and
\code{marglik} values already stored in \code{object} should be recomputed
after extension (\code{"all"}) or dropped with a warning (\code{"drop"}).}

\item{parallel}{logical. Whether to extend chains in parallel.}

\item{cores}{integer. Number of cores to use when \code{parallel = TRUE}.}

\item{silent}{logical. Whether to suppress JAGS output during extension.}

\item{seed}{optional seed used before extending.}

\item{evaluate}{unsupported; included for compatibility with
\code{\link[stats]{update}}.}
}
\value{
The updated \code{brma} object.
}
\description{
Extends an existing fitted \code{brma} object by additional MCMC samples,
updates study labels, and optionally recomputes cached fit-dependent
quantities.
}
\details{
Extending a fit adds posterior samples only. It does not rerun adaptation or
burn-in. Prior, data, and model-structure updates are intentionally not
supported by this method.
}
\examples{
\dontrun{
fit <- update(fit, sample_extend = 1000)
fit <- update(fit, slab = paste("Study", seq_len(nobs(fit))))
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/plot.R
\name{plot_prior}
\alias{plot_prior}
\alias{plot_prior.prior}
\alias{plot_prior.brma}
\title{Plot Prior Distributions}
\usage{
plot_prior(x, ...)

\method{plot_prior}{prior}(x, plot_type = "base", ...)

\method{plot_prior}{brma}(
  x,
  parameter = "mu",
  parameter_mods,
  parameter_scale,
  standardized_coefficients = TRUE,
  output_measure = NULL,
  transform = NULL,
  plot_type = "base",
  ...
)
}
\arguments{
\item{x}{a \code{brma}, \code{BMA}, or \code{RoBMA} object, or a prior
distribution object.}

\item{...}{additional arguments passed to the prior plotting method.}

\item{plot_type}{whether to use a base plot \code{"base"} or ggplot2
\code{"ggplot"} for plotting. Defaults to \code{"base"}.}

\item{parameter}{character. Base parameter to plot. Defaults to \code{"mu"}.
Common options are \code{"mu"}, \code{"tau"}, \code{"rho"}, \code{"PET"},
\code{"PEESE"}, \code{"omega"}, and \code{"bias"}, with aliases
\code{"effect"} = \code{"mu"}, \code{"heterogeneity"} = \code{"tau"},
and \code{"weightfunction"} = \code{"omega"}. \code{"bias"} plots only
non-mixed or homogeneous bias priors; for mixed weightfunction and PET/PEESE
mixtures use \code{"omega"}, \code{"PET"}, or \code{"PEESE"}. Moderator and
scale terms can also be selected by name when unambiguous. A character vector
requests multiple base parameters.}

\item{parameter_mods}{character. Moderator term to plot.
Use \code{"intercept"} for the adjusted effect in meta-regression models.}

\item{parameter_scale}{character. Scale-regression term to plot.
Use \code{"intercept"} for the heterogeneity intercept in location-scale models.}

\item{standardized_coefficients}{whether to plot moderator and scale-regression
priors on the standardized predictor scale. Defaults to \code{TRUE}, which
shows the priors as specified. Set to \code{FALSE} to transform them to the
original predictor scale when continuous predictors were standardized.}

\item{output_measure}{effect-size measure for location/effect predictions.
Defaults to the fitted measure. Supported conversions are among \code{"SMD"},
\code{"COR"}, \code{"ZCOR"}, and \code{"OR"}; \code{"RR"}, \code{"HR"},
\code{"IRR"}, \code{"RD"}, and \code{"GEN"} can only be returned on their
fitted measure. Use \code{transform = "EXP"} for ratio-scale output from
log-scale measures.}

\item{transform}{optional display transformation. Currently \code{"EXP"}
exponentiates log-scale measures \code{"OR"}, \code{"RR"}, \code{"HR"},
and \code{"IRR"}.}
}
\value{
\code{plot_prior} returns either \code{NULL} invisibly if
\code{plot_type = "base"} or a \code{ggplot2} object if
\code{plot_type = "ggplot"}. If multiple parameters are requested, a named
list is returned, invisibly for base plots.
}
\description{
\code{plot_prior} visualizes prior distributions stored in
\code{brma}, \code{BMA}, and \code{RoBMA} objects.
This is especially useful for objects created with \code{only_priors = TRUE}.
}
\details{
\code{output_measure} and \code{transform} transform the prior
plotting scale only for effect-size location priors (\code{"mu"} or the
meta-regression intercept).
}
\examples{
\dontrun{
if (requireNamespace("metadat", quietly = TRUE)) {
  data(dat.lehmann2018, package = "metadat")

  priors <- BMA(
    yi           = yi,
    vi           = vi,
    mods         = ~ Preregistered,
    data         = dat.lehmann2018,
    measure      = "SMD",
    only_priors  = TRUE
  )

  plot_prior(priors, parameter = "mu")
  plot_prior(priors, parameter = "tau")
  plot_prior(priors, parameter_mods = "Preregistered")
}
}

}
\seealso{
\code{\link[=BMA]{BMA()}} \code{\link[=RoBMA]{RoBMA()}} \code{\link[=brma]{brma()}} \code{\link[=prior]{prior()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/marginal_means.R
\name{marginal_means.brma}
\alias{marginal_means.brma}
\title{Estimated Marginal Means for brma Objects}
\usage{
\method{marginal_means}{brma}(
  object,
  null_hypothesis = 0,
  normal_approximation = FALSE,
  n_samples = 10000,
  output_measure = NULL,
  transform = NULL,
  bf = NULL,
  ...
)
}
\arguments{
\item{object}{a fitted \code{brma} object with moderators.}

\item{null_hypothesis}{point null hypothesis used for inclusion Bayes
factors. Defaults to \code{0}.}

\item{normal_approximation}{whether prior and posterior density at the null
should be approximated with a normal distribution. Defaults to \code{FALSE}.}

\item{n_samples}{number of samples/grid points used by BayesTools for
marginal prior densities. Defaults to \code{10000}.}

\item{output_measure}{effect-size measure for location/effect predictions.
Defaults to the fitted measure. Supported conversions are among \code{"SMD"},
\code{"COR"}, \code{"ZCOR"}, and \code{"OR"}; \code{"RR"}, \code{"HR"},
\code{"IRR"}, \code{"RD"}, and \code{"GEN"} can only be returned on their
fitted measure. Use \code{transform = "EXP"} for ratio-scale output from
log-scale measures.}

\item{transform}{optional display transformation. Currently \code{"EXP"}
exponentiates log-scale measures \code{"OR"}, \code{"RR"}, \code{"HR"},
and \code{"IRR"}.}

\item{bf}{whether inclusion Bayes factors should be shown by default in
summaries. Defaults to \code{TRUE} for RoBMA/BMA objects and \code{FALSE}
for single-model \code{brma} objects.}

\item{...}{additional arguments (currently ignored).}
}
\value{
A list of class \code{marginal_means.brma} containing the
BayesTools \code{marginal_inference} object and parameter metadata.
}
\description{
Computes estimated marginal means for a fitted \code{brma}
object with moderators using \code{BayesTools::as_marginal_inference()}.
}
\examples{
\dontrun{
if (requireNamespace("metadat", quietly = TRUE) &&
    requireNamespace("metafor", quietly = TRUE)) {
  data(dat.bcg, package = "metadat")
  dat <- metafor::escalc(
    measure = "RR",
    ai      = tpos,
    bi      = tneg,
    ci      = cpos,
    di      = cneg,
    data    = dat.bcg
  )

  fit <- brma(yi = yi, vi = vi, mods = ~ alloc, data = dat, measure = "RR")
  mm  <- marginal_means(fit)
  summary(mm)
  plot(mm, parameter = "alloc")
}
}

}
\seealso{
\code{\link[=summary]{summary()}}, \code{\link[=plot]{plot()}}, \code{\link[=summary.brma]{summary.brma()}}, \code{\link[=regplot]{regplot()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/priors.R
\name{prior_PEESE}
\alias{prior_PEESE}
\title{PEESE Prior}
\usage{
prior_PEESE(
  distribution,
  parameters,
  truncation = list(lower = 0, upper = Inf),
  prior_weights = 1
)
}
\arguments{
\item{distribution}{character. Prior distribution name.}

\item{parameters}{list. Distribution parameters.}

\item{truncation}{list with \code{lower} and \code{upper} truncation bounds.}

\item{prior_weights}{numeric prior model weight.}
}
\value{
An object inheriting from \code{prior} with the \code{prior.PEESE}
marker class.
}
\description{
Create PEESE publication-bias regression priors.
}
\details{
This forwards to \code{BayesTools::prior_PEESE()} and uses the same
distribution and parameter conventions as \code{prior()}.
}
\seealso{
\code{\link{publication_bias_prior_specification}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/datasets.R
\docType{data}
\name{Havrankova2025}
\alias{Havrankova2025}
\title{1159 effect sizes from a meta-analysis of beauty and professional success
by \insertCite{havrankova2025beauty;textual}{RoBMA}}
\format{
A data.frame with 5 columns and 1159 observations:
\describe{
\item{\code{y}}{Effect size: percent increase in earnings.}
\item{\code{se}}{Standard error of y.}
\item{\code{facing_customer}}{Type of customer contact.}
\item{\code{study_id}}{Study identifier.}
\item{\code{N}}{Sample size.}
}
}
\usage{
Havrankova2025
}
\description{
The data set contains effect sizes (percent increase in earnings), standard errors,
study identifiers, sample sizes, and the type of customer contact (no, some, or direct).
The meta-analysis examined the relationship between perceived beauty and professional success
\insertCite{havrankova2025beauty}{RoBMA}.
}
\references{
\insertAllCited{}
}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/loo.R
\name{loo_compare.brma}
\alias{loo_compare.brma}
\alias{loo_compare}
\title{Compare brma Models Using LOO}
\usage{
\method{loo_compare}{brma}(x, ..., unit = "estimate")
}
\arguments{
\item{x}{a brma model object (the first model to compare).}

\item{...}{additional brma model objects or \code{loo} objects to compare.}

\item{unit}{output/deletion unit used when extracting LOO from brma objects.}
}
\value{
A matrix of class \code{"compare.loo"} as returned by
\code{\link[loo]{loo_compare}}.
}
\description{
Compare multiple brma models using LOO-PSIS cross-validation.
This is a convenience wrapper around \code{\link[loo]{loo_compare}}.
}
\details{
This function compares models based on their expected out-of-sample
predictive performance (ELPD).

\strong{Important for model comparison:} When comparing models via
\code{\link[loo]{loo_compare}}, the selection is based on expected
out-of-sample predictive performance. This evaluates how well models predict
\emph{new} observations, not how well they fit the observed data.
RoBMA rejects comparisons with different outcome targets/data, \code{unit},
or implied \code{conditioning_depth}.
}
\examples{
\dontrun{
if (requireNamespace("metadat", quietly = TRUE)) {
  data(dat.lehmann2018, package = "metadat")

  fit_bias <- RoBMA(yi = yi, vi = vi, data = dat.lehmann2018, measure = "SMD")
  fit_nobias <- BMA(yi = yi, vi = vi, data = dat.lehmann2018, measure = "SMD")

  fit_bias <- add_loo(fit_bias)
  fit_nobias <- add_loo(fit_nobias)

  loo_compare(fit_bias, fit_nobias)
  loo_compare(loo(fit_bias), loo(fit_nobias))
}
}

}
\seealso{
\code{\link{loo.brma}}, \code{\link[loo]{loo_compare}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/vif.R
\name{vif}
\alias{vif}
\title{Variance Inflation Factors}
\usage{
vif(object, ...)
}
\arguments{
\item{object}{a fitted model object}

\item{...}{additional arguments passed to methods}
}
\value{
Method-specific return value containing VIF diagnostics.
}
\description{
Computes variance inflation factors (VIF) for a fitted model
with moderators, to assess multicollinearity among predictors.
}
\references{
\insertCite{fox1992generalized}{RoBMA}
}
\seealso{
\code{\link[=regplot]{regplot()}}, \code{\link[=summary.brma]{summary.brma()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/wrappers.R
\name{coef.brma}
\alias{coef.brma}
\title{Extract Model Coefficients for brma Objects}
\usage{
\method{coef}{brma}(object, ...)
}
\arguments{
\item{object}{a fitted brma object}

\item{...}{additional arguments (currently ignored)}
}
\value{
A named numeric vector of posterior mean coefficients.
}
\description{
Extract model coefficients (posterior means) from a
fitted brma object.
}
\seealso{
\code{\link[=summary.brma]{summary.brma()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/summary_heterogeneity.R
\name{print.summary_heterogeneity.brma}
\alias{print.summary_heterogeneity.brma}
\title{Print Summary of Heterogeneity}
\usage{
\method{print}{summary_heterogeneity.brma}(x, ...)
}
\arguments{
\item{x}{a \code{summary_heterogeneity.brma} object}

\item{...}{additional arguments (currently ignored)}
}
\value{
Returns \code{x} invisibly.
}
\description{
Prints the heterogeneity summary table.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/influence.R
\name{influence.brma}
\alias{influence.brma}
\title{Measure Influence for brma Objects}
\usage{
\method{influence}{brma}(model, ...)
}
\arguments{
\item{model}{a fitted brma object.}

\item{...}{additional arguments (currently ignored).}
}
\value{
An object of class \code{"infl.brma"}, which corresponds to the
structure of \code{metafor::influence} objects. It is a list containing:
\item{inf}{A data frame with columns: \code{rstudent}, \code{dffits},
\code{cook.d}, \code{cov.r}, \code{tau.del}, and \code{hat}, where available.
The \code{tau.del} column is the PSIS leave-one-out posterior mean of the
aggregate heterogeneity. For scale models, aggregation is over the remaining
scale-model rows after each deletion.}
\item{dfbs}{A data frame with DFBETAS values for the model coefficients.}
Undefined determinant- or variance-standardized diagnostics are reported as
\code{NaN} and printed with an explanatory note.
}
\description{
Computes DFFITS, Cook's distance, COVRATIO, and other
influence diagnostics for a fitted brma object.
}
\examples{
\dontrun{
if (requireNamespace("metadat", quietly = TRUE)) {
  data(dat.lehmann2018, package = "metadat")
  fit <- brma(
    yi      = yi,
    vi      = vi,
    data    = dat.lehmann2018,
    measure = "SMD",
    seed    = 1,
    silent  = TRUE
  )
  fit <- add_loo(fit)

  inf <- influence(fit)
  print(inf)
}
}

}
\seealso{
\code{\link{dffits.brma}}, \code{\link{cooks.distance.brma}},
\code{\link{covratio.brma}}, \code{\link{dfbetas.brma}}, \code{\link{hatvalues.brma}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/bridgesampling.R
\name{bf.brma}
\alias{bf.brma}
\alias{bayes_factor.brma}
\title{Bayes Factor for brma Objects}
\usage{
\method{bf}{brma}(x1, x2, log = FALSE, ...)

\method{bayes_factor}{brma}(x1, x2, log = FALSE, ...)
}
\arguments{
\item{x1}{a brma model object (numerator).}

\item{x2}{a brma model object (denominator).}

\item{log}{logical; if \code{TRUE}, the log Bayes factor is returned.
Default is \code{FALSE}.}

\item{...}{additional arguments (currently not used).}
}
\value{
A list of class \code{"bf_default"} with components:
\itemize{
\item \code{bf}: (scalar) value of the Bayes factor in favor of
\code{x1} over \code{x2}.
\item \code{log}: Boolean indicating whether \code{bf} corresponds
to the log Bayes factor.
}
}
\description{
Compute the Bayes factor comparing two brma models.
}
\details{
Computes the Bayes factor in favor of the model \code{x1} over the
model \code{x2}. The marginal likelihoods must first be computed using
\code{\link{add_marglik}}. Both models must be fitted to the same
outcome target/data, including outcome type and, when present, weights and
cluster identifiers.
}
\examples{
\dontrun{
if (requireNamespace("metadat", quietly = TRUE)) {
  data(dat.lehmann2018, package = "metadat")
  fit1 <- brma(yi = yi, vi = vi, data = dat.lehmann2018, measure = "SMD")
  fit2 <- brma(
    yi           = yi,
    vi           = vi,
    data         = dat.lehmann2018,
    measure      = "SMD",
    prior_effect = FALSE
  )

  fit1 <- add_marglik(fit1)
  fit2 <- add_marglik(fit2)

  bf(fit1, fit2)
}
}

}
\seealso{
\code{\link{add_marglik}}, \code{\link{bridge_sampler.brma}},
\code{\link{post_prob.brma}}, \code{\link{logml.brma}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/vif.R
\name{print.vif.brma}
\alias{print.vif.brma}
\title{Print VIF Results}
\usage{
\method{print}{vif.brma}(x, digits = 3, ...)
}
\arguments{
\item{x}{a \code{vif.brma} object}

\item{digits}{number of decimal places. Defaults to 3.}

\item{...}{additional arguments (currently ignored)}
}
\value{
Returns \code{x} invisibly.
}
\description{
Prints variance inflation factors and optional posterior
correlation matrix.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/priors.R
\name{prior_none}
\alias{prior_none}
\title{Empty Prior}
\usage{
prior_none(prior_weights = 1)
}
\arguments{
\item{prior_weights}{numeric prior model weight.}
}
\value{
An object inheriting from \code{prior} and \code{prior.none}.
}
\description{
Create an empty prior used to omit a model component.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/bridgesampling.R
\name{logml.brma}
\alias{logml.brma}
\title{Log Marginal Likelihood for brma Objects}
\usage{
\method{logml}{brma}(x, ...)
}
\arguments{
\item{x}{a brma model object.}

\item{...}{additional arguments (currently not used).}
}
\value{
A scalar numeric value representing the log marginal likelihood.
}
\description{
Extract the log marginal likelihood from a brma model.
The marginal likelihood must first be computed using \code{\link{add_marglik}}.
}
\details{
This function extracts the log marginal likelihood from the bridge sampling
object that was previously computed and stored using \code{\link{add_marglik}}.
Product-space model-averaging objects (\code{BMA.norm}, \code{BMA.glmm},
and \code{RoBMA}) do not expose bridge-sampling marginal likelihoods.
}
\examples{
\dontrun{
if (requireNamespace("metadat", quietly = TRUE)) {
  data(dat.lehmann2018, package = "metadat")
  fit <- brma(yi = yi, vi = vi, data = dat.lehmann2018, measure = "SMD")

  fit <- add_marglik(fit)

  logml(fit)
}
}

}
\seealso{
\code{\link{add_marglik}}, \code{\link{bridge_sampler.brma}},
\code{\link{bf.brma}}, \code{\link{post_prob.brma}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/wrappers.R
\name{ranef.brma}
\alias{ranef.brma}
\title{Random Effects for brma Objects}
\usage{
\method{ranef}{brma}(object, bias_adjusted = FALSE, probs = c(0.025, 0.975), ...)
}
\arguments{
\item{object}{a fitted brma object}

\item{bias_adjusted}{whether to adjust for publication bias. Defaults to
\code{FALSE}. See \code{\link{blup.brma}} for details.}

\item{probs}{quantiles of the posterior distribution to be displayed.
Defaults to \code{c(.025, .975)} for 95\% credible intervals.}

\item{...}{additional arguments forwarded to \code{\link{predict.brma}} for
supported options such as \code{conditional}. \code{newdata}, \code{type},
\code{quiet}, \code{output_measure}, and \code{transform} are controlled by
this method.}
}
\value{
For 2-level models, a \code{brma_samples} object. For 3-level
models, a named list of \code{brma_samples} objects (one per variance
component).
}
\description{
Extracts random effect deviations from a fitted brma object.
These are posterior-sample offsets from the fixed-effect predictions,
analogous to random-effect deviations returned by \code{metafor::ranef()}.
}
\details{
Random effects are computed as the difference between the true
effects (BLUPs) and the fixed-effect predictions:
\deqn{u_i = \hat{\theta}_i - \hat{\mu}_i}

For standard (2-level) models, returns a single \code{brma_samples}
object with the estimate-level random effects.

For multilevel (3-level) models, returns a list with two observation-aligned
\code{brma_samples} matrices, one column per estimate row:
\describe{
\item{\code{cluster}}{Cluster-level random effects
(\eqn{\gamma_j \cdot \tau_{between}}), representing between-cluster
deviations from the fixed effects.}
\item{\code{estimate}}{Estimate-level random effects
(\eqn{\theta_i - \mu_i - \gamma_j \cdot \tau_{between}}),
representing within-cluster deviations from the cluster means.}
}
}
\examples{
\dontrun{
if (requireNamespace("metadat", quietly = TRUE)) {
  data(dat.lehmann2018, package = "metadat")
  fit <- brma(
    yi      = yi,
    vi      = vi,
    data    = dat.lehmann2018,
    measure = "SMD",
    seed    = 1,
    silent  = TRUE
  )

  ranef(fit)
}
}

}
\seealso{
\code{\link[=blup.brma]{blup.brma()}}, \code{\link[=predict.brma]{predict.brma()}}, \code{\link[=pooled_effect]{pooled_effect()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/marginal_means.R
\name{marginal_means}
\alias{marginal_means}
\title{Estimated Marginal Means}
\usage{
marginal_means(object, ...)
}
\arguments{
\item{object}{a fitted model object.}

\item{...}{additional arguments passed to methods.}
}
\value{
A method-specific estimated marginal means object.
}
\description{
S3 generic for estimated marginal means. The \code{brma}
method works with fitted moderator models and stores the BayesTools
marginal-inference object for \code{summary()} and \code{plot()}.
}
\seealso{
\code{\link[=summary]{summary()}}, \code{\link[=plot]{plot()}}, \code{\link[=summary.brma]{summary.brma()}}, \code{\link[=regplot]{regplot()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/loo.R
\name{add_loo.brma}
\alias{add_loo.brma}
\alias{add_loo}
\title{Add LOO-PSIS to brma Objects}
\usage{
\method{add_loo}{brma}(object, unit = "estimate", r_eff = NULL, parallel = FALSE, ...)
}
\arguments{
\item{object}{a brma model object.}

\item{unit}{output/deletion unit. \code{"estimate"} computes one contribution
per effect-size estimate. \code{"cluster"} computes one contribution per
cluster and is available only for multilevel models.}

\item{r_eff}{optional vector of relative effective sample sizes. If not
provided, it is computed from the log-likelihood values.}

\item{parallel}{Logical. If \code{TRUE}, \code{loo::relative_eff()} and
\code{loo::loo()} use \code{RoBMA.get_option("max_cores")}. Log-likelihood
construction is unchanged. If \code{FALSE}, those computations use one core.}

\item{...}{additional arguments (currently ignored).}
}
\value{
The brma object with the LOO result stored in
\code{object[["loo"]][[unit]]}.
}
\description{
Compute approximate leave-one-out cross-validation (LOO-CV)
using Pareto smoothed importance sampling (PSIS) for brma model objects
and store the result in the object.
}
\details{
With \code{unit = "estimate"}, LOO-CV is computed with one contribution per
effect-size estimate. For binomial and Poisson models, each pair of counts
(ai/ci or x1i/x2i) that defines a single effect size estimate is treated as
one contribution.

With \code{unit = "cluster"}, LOO-CV is computed with one joint contribution
per cluster. For unweighted normal models without selection this uses the
analytic cluster block covariance. Selection, data-weighted normal, and
GLMM models integrate the held-out cluster effect with Gauss-Hermite
quadrature.

For selection models, the LOO evaluates the weighted likelihood, conditioning
on the posterior omega samples.

The PSIS object is essential for model comparison via
\code{\link[loo]{loo_compare}} and is automatically saved in the loo result.
RoBMA stores target metadata so comparisons can reject mismatched data,
unit, or conditioning-depth targets.

\strong{Important for model comparison:} When comparing models via
\code{\link[loo]{loo_compare}}, the selection is based on expected
out-of-sample predictive performance. This evaluates how well models predict
\emph{new} observations, not how well they fit the observed data.
}
\examples{
\dontrun{
if (requireNamespace("metadat", quietly = TRUE)) {
  data(dat.lehmann2018, package = "metadat")
  fit <- bPET(yi = yi, vi = vi, data = dat.lehmann2018, measure = "SMD")

  fit <- add_loo(fit)
  loo_fit <- loo(fit)
  print(loo_fit)
  plot(loo_fit)
}
}

}
\references{
\insertCite{vehtari2017practical}{RoBMA}
\insertCite{vehtari2024pareto}{RoBMA}
}
\seealso{
\code{\link{loo.brma}}, \code{\link[loo]{loo}},
\code{\link[loo]{loo_compare}}, \code{\link[loo]{pareto_k_ids}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/summary_models.R
\name{summary_models}
\alias{summary_models}
\alias{summary_models.RoBMA}
\alias{print.summary_models.RoBMA}
\title{Summarize Model-Averaged Component Weights}
\usage{
summary_models(object, ...)

\method{summary_models}{RoBMA}(object, type = "marginal", include_mcmc_diagnostics = TRUE, ...)

\method{print}{summary_models.RoBMA}(x, ...)
}
\arguments{
\item{object}{a fitted RoBMA-class product-space object, including
\code{RoBMA}, \code{BMA}/\code{BMA.norm}, and \code{BMA.glmm}.}

\item{...}{additional arguments}

\item{type}{whether to summarize marginal component prior distributions
(\code{"marginal"}) or individual model combinations (\code{"individual"}).}

\item{include_mcmc_diagnostics}{whether to include Bayes factor MCMC
diagnostics in the output. Defaults to \code{TRUE}.}

\item{x}{a \code{summary_models.RoBMA} object.}
}
\value{
A list of class \code{summary_models.RoBMA} with elements
\code{name} and \code{type}. For \code{type = "marginal"}, element
\code{marginal} contains component tables with columns such as
\code{prior_prob}, \code{post_prob}, and \code{inclusion_BF}. For
\code{type = "individual"}, element \code{individual} contains individual
model combinations and posterior probabilities.
}
\description{
Creates marginal or individual model-weight summaries for
RoBMA-class product-space objects.
}
\details{
Only mixture-prior components are summarized; non-mixture
components are omitted.
}
\examples{
\dontrun{
if (requireNamespace("metadat", quietly = TRUE)) {
  data(dat.lehmann2018, package = "metadat")
  fit <- RoBMA(yi = yi, vi = vi, data = dat.lehmann2018, measure = "SMD")

  summary_models(fit)
  summary_models(fit, type = "individual")
}
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/marginal_means.R
\name{print.summary.marginal_means.brma}
\alias{print.summary.marginal_means.brma}
\title{Print Summary of Estimated Marginal Means}
\usage{
\method{print}{summary.marginal_means.brma}(x, ...)
}
\arguments{
\item{x}{a \code{summary.marginal_means.brma} object.}

\item{...}{additional arguments (currently ignored).}
}
\value{
Returns \code{x} invisibly.
}
\description{
Prints a summary table of estimated marginal means.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/residuals.R
\name{rstudent.brma}
\alias{rstudent.brma}
\title{Externally Standardized (Studentized) Residuals for brma Objects}
\usage{
\method{rstudent}{brma}(model, unit = "estimate", conditioning_depth = "marginal", ...)
}
\arguments{
\item{model}{a fitted brma object.}

\item{unit}{output unit. Only \code{"estimate"} is available for LOO-PIT
residuals.}

\item{conditioning_depth}{unused for LOO-PIT residuals. LOO-PIT residuals
always use the estimate-unit LOO target.}

\item{...}{additional arguments (currently ignored)}
}
\value{
A data frame with columns:
\itemize{
\item \code{resid}: Raw residuals
\item \code{se}: Standard errors of the residuals
\item \code{z}: Externally standardized residuals (LOO-PIT)
}
}
\description{
Computes externally standardized residuals (also called
studentized residuals or standardized deleted residuals) from a fitted brma
object using LOO-PIT (Leave-One-Out Probability Integral Transform).
Returns a data frame with raw residuals, standard errors, and standardized
residuals (z-values).
}
\details{
This function returns a data frame with three columns matching the output
of \code{metafor::rstudent}:
\itemize{
\item \code{resid}: LOO predictive residuals (observed - fitted values)
\item \code{se}: LOO predictive standard errors when available
\item \code{z}: Externally standardized residuals (LOO-PIT transformed)
}

LOO-PIT residuals are the Bayesian equivalent of studentized deleted
residuals. They are computed via leave-one-out probability integral
transformation using Pareto smoothed importance sampling. For each
observation, the LOO-weighted CDF is computed and transformed to a
standard normal quantile.

Under a correctly specified model, LOO-PIT residuals should follow a
standard normal distribution. Large absolute values may indicate outliers
or model misspecification.

The \code{z} column is the primary standardized diagnostic. The \code{resid}
and \code{se} columns are raw-scale companions computed from LOO predictive
moments using the normalized PSIS weights. For selection models, these moments
are computed from the fitted selected-normal predictive distribution. For
GLMMs, they are computed on the approximate effect-size scale used by the
LOO-PIT diagnostic; they are not exact PIT diagnostics for the raw count
likelihood.

Unlike \code{\link{rstandard.brma}} (which uses the hat matrix), LOO-PIT
residuals properly account for estimation uncertainty and leverage without
requiring explicit hat matrix computation. This makes \code{rstudent.brma}
suitable for all model types including selection models and GLMMs.
}
\examples{
\dontrun{
if (requireNamespace("metadat", quietly = TRUE)) {
  data(dat.lehmann2018, package = "metadat")
  fit <- brma(yi = yi, vi = vi, data = dat.lehmann2018, measure = "SMD")
  fit <- add_loo(fit)

  # externally standardized residuals
  rstudent(fit)

  # check Pareto k values
  plot(loo(fit))
}
}

}
\seealso{
\code{\link[=rstandard.brma]{rstandard.brma()}}, \code{\link[=residuals.brma]{residuals.brma()}}, \code{\link[=loo.brma]{loo.brma()}}, \code{\link[=blup.brma]{blup.brma()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/input-priors.R
\name{publication_bias_prior_specification}
\alias{publication_bias_prior_specification}
\alias{bias_prior_specification}
\title{Publication-bias prior specification}
\arguments{
\item{prior_bias}{prior distribution for publication-bias adjustment. For
\code{\link{bselmodel}}, this is usually a weightfunction prior created by
\code{\link{prior_weightfunction}}. For \code{\link{bPET}}, use
\code{\link{prior_PET}}. For \code{\link{bPEESE}}, use
\code{\link{prior_PEESE}}. For \code{\link{RoBMA}}, this can be a single
publication-bias prior distribution or a list of publication-bias prior
distributions. In the single-model bias-adjustment constructors, omitted or
\code{NULL} uses the corresponding default prior distribution.}

\item{prior_bias_null}{prior distribution(s) for null publication-bias
component(s) in \code{\link{RoBMA}}, usually \code{\link{prior_none}()}.
A single prior distribution object creates one null component, a list creates
multiple null components, and \code{NULL} or \code{FALSE} omits the null
publication-bias component.}

\item{model_type}{character string specifying predefined publication-bias
model ensembles for \code{\link{RoBMA}}. One of \code{"PSMA"},
\code{"6w"}, \code{"2w"}, or \code{"PP"}.}

\item{steps}{numeric vector of one-sided p-value cut points for the default
\code{\link{bselmodel}} weightfunction prior. If \code{prior_bias} is supplied,
the prior distribution carries its own p-value cut points.}
}
\description{
Publication-bias prior distributions are specified separately from the
meta-analytic prior distributions documented in
\code{\link{prior_specification}}. They define selection-model weights,
PET/PEESE regression coefficients, or the publication-bias model space in
\code{\link{RoBMA}}.
}
\details{
\subsection{Single-model bias-adjustment priors}{

\code{\link{bselmodel}}, \code{\link{bPET}}, and \code{\link{bPEESE}} fit one
publication-bias adjustment at a time. The \code{prior_bias} argument must
match the fitted bias-adjustment type.

\tabular{lll}{
Constructor \tab Prior constructor \tab Default prior distribution \cr
\code{bselmodel()} \tab \code{\link{prior_weightfunction}} \tab one-sided cumulative weightfunction with \code{steps = 0.025} \cr
\code{bPET()} \tab \code{\link{prior_PET}} \tab positive Cauchy centered at 0 \cr
\code{bPEESE()} \tab \code{\link{prior_PEESE}} \tab positive Cauchy centered at 0, with UISD-adjusted scale
}

The default PET prior distribution uses
\code{RoBMA.get_option("default_bias_PET.scale")}. The default PEESE prior
distribution uses \code{RoBMA.get_option("default_bias_PEESE.scale")} after
rescaling to the analyzed effect-size measure. The default weightfunction
prior distribution uses \code{RoBMA.get_option("default_bias_weightfunction.alpha")}
for each cumulative-Dirichlet alpha parameter.
}

\subsection{Model-averaged publication-bias priors}{

\code{\link{RoBMA}} averages over publication-bias components. By default,
\code{prior_bias_null} is \code{\link{prior_none}()} and \code{model_type}
determines the alternative components:

\tabular{ll}{
\code{"PSMA"} \tab six weight functions, PET, and PEESE \cr
\code{"6w"} \tab six weight functions \cr
\code{"2w"} \tab two two-sided weight functions \cr
\code{"PP"} \tab PET and PEESE
}

Custom \code{prior_bias} replaces the preset alternative components. If
\code{prior_bias} is omitted, \code{model_type} still supplies the default
alternative components even when \code{prior_bias_null} is customized or
omitted. Setting \code{prior_bias = NULL} or \code{prior_bias = FALSE}
omits all alternative bias-adjustment models. Setting
\code{prior_bias_null = NULL} or \code{prior_bias_null = FALSE} omits the
no-bias component.

Each publication-bias component has a prior model weight. User-specified
components receive equal weights unless their prior objects set
\code{prior_weights}. The \code{"2w"} and \code{"6w"} presets split the
alternative bias mass equally across their weight functions, \code{"PP"}
splits it equally across PET and PEESE, and \code{"PSMA"} assigns half of the
alternative bias mass to the six weight functions combined and one quarter
each to PET and PEESE.

Publication-bias prior distributions are not rescaled by \code{rescale_priors}.
The exception is the default PEESE prior distribution, whose scale is adjusted
to the effect-size measure via the unit-information standard deviation.
}
}
\seealso{
\code{\link{prior_weightfunction}}, \code{\link{prior_PET}},
\code{\link{prior_PEESE}}, \code{\link{prior_none}},
\code{\link{RoBMA_prior_specification}},
\code{\link{prior_specification}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/wrappers.R
\name{pooled_heterogeneity}
\alias{pooled_heterogeneity}
\title{Pooled Heterogeneity}
\usage{
pooled_heterogeneity(object, ...)
}
\arguments{
\item{object}{a fitted model object}

\item{...}{additional arguments passed to methods}
}
\value{
Method-specific return value, typically a summary table or
posterior samples of the pooled heterogeneity.
}
\description{
Computes the pooled (aggregated) heterogeneity estimate (tau)
from a fitted model.
}
\seealso{
\code{\link[=predict.brma]{predict.brma()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/bridgesampling.R
\name{post_prob.brma}
\alias{post_prob.brma}
\title{Posterior Model Probabilities for brma Objects}
\usage{
\method{post_prob}{brma}(x, ..., prior_prob = NULL, model_names = NULL)
}
\arguments{
\item{x}{a brma model object.}

\item{...}{additional brma model objects.}

\item{prior_prob}{numeric vector with prior model probabilities or weights.
Values must be finite, nonnegative, have the same length as the retained
models, and have a positive total. If omitted, a uniform prior is used.
Supplied values are normalized internally.}

\item{model_names}{character vector with model names. If \code{NULL}
(the default), names will be derived from deparsing the call.}
}
\value{
A named numeric vector with posterior model probabilities
(i.e., which sum to one).
}
\description{
Compute posterior model probabilities from marginal
likelihoods of brma models.
}
\details{
The marginal likelihoods must first be computed using \code{\link{add_marglik}}.
\code{x} and at least one additional \code{brma} model must be supplied.
Non-\code{brma} objects in \code{...} are ignored with a warning. All retained
models must be fitted to the same outcome target/data, including outcome type
and, when present, weights and cluster identifiers.
}
\examples{
\dontrun{
if (requireNamespace("metadat", quietly = TRUE)) {
  data(dat.lehmann2018, package = "metadat")
  fit1 <- brma(yi = yi, vi = vi, data = dat.lehmann2018, measure = "SMD")
  fit2 <- brma(
    yi           = yi,
    vi           = vi,
    data         = dat.lehmann2018,
    measure      = "SMD",
    prior_effect = FALSE
  )

  fit1 <- add_marglik(fit1)
  fit2 <- add_marglik(fit2)

  post_prob(fit1, fit2)
}
}

}
\seealso{
\code{\link{add_marglik}}, \code{\link{bridge_sampler.brma}},
\code{\link{bf.brma}}, \code{\link{logml.brma}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/funnel.R
\name{funnel.brma}
\alias{funnel.brma}
\alias{funnel}
\title{Funnel Plot for brma Object}
\usage{
\method{funnel}{brma}(
  x,
  residual,
  type = "LOO-PIT",
  unit = "estimate",
  conditioning_depth = "marginal",
  sampling_heterogeneity = TRUE,
  sampling_bias = TRUE,
  max_samples = 10000,
  plot_type = "base",
  ...
)
}
\arguments{
\item{x}{a fitted brma object}

\item{residual}{whether to use residual mode. Defaults to not specified,
which means the function automatically determines the mode:
\itemize{
\item not specified: For intercept-only models without scale
regression, displays observed effect sizes against the fitted sampling
distribution funnel; for models with moderators or scale regression,
automatically uses residual mode.
\item \code{FALSE}: explicitly requests outcome mode.
\item \code{TRUE}: explicitly requests residual mode, displaying residuals
on the x-axis and using \code{type} to determine how those residuals are
computed.
}}

\item{type}{the type of residuals to use when in residual mode.
Options are:
\itemize{
\item \code{"LOO-PIT"} (alias: \code{"rstudent"}; default): Leave-one-out
probability integral transform residuals returned by
\code{\link{rstudent.brma}}.
\item \code{"rstandard"}: Internally standardized residuals using
\code{\link{rstandard.brma}}. Only available for normal outcome models.
\item \code{"outcome"}: Raw outcome residuals from
\code{\link{residuals.brma}} with \code{type = "outcome"}.
}
Only used when funnel is in residual mode.}

\item{unit}{output unit for residual mode. Only \code{"estimate"} is
implemented in this pass.}

\item{conditioning_depth}{residual conditioning depth for residual mode.
Options are \code{"marginal"}, \code{"cluster"}, and \code{"estimate"}.
The default LOO-PIT residual path is available only with marginal
conditioning depth.}

\item{sampling_heterogeneity}{whether heterogeneity should be incorporated
into the sampling distribution funnel. Defaults to \code{TRUE}. Only used
in outcome mode and ignored in residual mode.}

\item{sampling_bias}{whether publication bias should be incorporated into the
sampling distribution funnel. Defaults to \code{TRUE}. Only used when
\code{residual = FALSE} or when automatic mode selects outcome mode. Ignored
in residual mode. When \code{TRUE} and the model
includes selection models (weightfunction), uses selected-normal quantiles.
When \code{TRUE} and the model includes PET/PEESE, incorporates the expected
skew from these regression adjustments.}

\item{max_samples}{maximum number of posterior samples used for
model-averaged publication-bias funnel contours. Defaults to \code{10000}.
Use \code{Inf} to use all posterior samples.}

\item{plot_type}{whether to use a base plot \code{"base"} or ggplot2
\code{"ggplot"} for plotting. Defaults to \code{"base"}.}

\item{...}{additional graphical arguments to customize the plot appearance:
\describe{
\item{xlim, ylim}{numeric vectors of length 2 specifying axis limits}
\item{xlab, ylab}{character strings for axis labels}
\item{main}{character string for plot title (default: no title)}
\item{pch}{point symbol (default: 21, filled circle). Use standard R pch values.}
\item{col}{point border color (default: "black")}
\item{bg}{point fill/background color (default: "#A6A6A6")}
\item{cex}{point size multiplier for base graphics (default: 1)}
\item{size}{point size for ggplot2 (default: 2)}
\item{las}{axis-label style for base graphics (default: 1)}
\item{back}{background region color (default: "grey"). Set to \code{NA} to suppress.}
\item{shade}{funnel region fill color (default: "white"). Set to \code{NA} to suppress.}
\item{lty}{line type for funnel edges and center line (default: "dotted")}
\item{col.line}{color for funnel edge lines (default: "black")}
\item{refline}{numeric override for the reference line. By default,
residual mode uses 0, while outcome mode uses the center of the fitted
sampling distribution, which may be curved when PET/PEESE bias adjustment
is incorporated.}
\item{col.refline}{color of vertical reference line (default: "black")}
\item{as_data}{if \code{TRUE}, returns plot data instead of creating a plot}
}}
}
\value{
If \code{as_data = TRUE}, \code{funnel.brma} returns a list with the
data used for plotting, including the plotted points, funnel polygons,
plotting limits, labels, and reference line. Otherwise, it returns
\code{NULL} invisibly if \code{plot_type = "base"} or a ggplot object if
\code{plot_type = "ggplot"}.
}
\description{
\code{funnel.brma} creates a funnel plot for a fitted brma object.
For intercept-only models without scale regression, the default outcome mode
displays observed effect sizes against the fitted sampling distribution. For
models with location or scale moderators, the default residual mode displays
residuals against a standard-error funnel.
}
\details{
The funnel plot has two modes. If \code{residual} is not specified, the mode
is chosen automatically from the fitted model: intercept-only models without
scale regression use outcome mode, whereas models with location or scale
moderators use residual mode.

\strong{Outcome mode} (intercept-only models without scale regression):
Displays observed effect sizes on the x-axis and standard errors on the
y-axis. The reference line follows the center of the fitted sampling
distribution. When \code{sampling_bias = FALSE}, this center is the pooled
effect; when PET/PEESE bias adjustment is incorporated, the center line can
vary with the standard error. The funnel region represents the central 95\\%
region of the sampling distribution, optionally incorporating heterogeneity
and publication bias.

\strong{Residual mode} (models with moderators or scale regression):
Displays residuals on the x-axis and standard errors on the
y-axis. The funnel region represents the central 95\\% region of
\eqn{N(0, \mathrm{SE}^2)}. With \code{type = "LOO-PIT"}, the plotted
residuals and standard errors are the raw-scale LOO predictive companions
returned by \code{\link{rstudent.brma}}; the PIT-normalized \code{z} values
are used by \code{\link{qqnorm.brma}} and influence diagnostics. With
\code{type = "rstandard"}, the plotted values are internally standardized
residual companions from \code{\link{rstandard.brma}}. With
\code{type = "outcome"}, these are raw outcome residuals. Under a correctly
specified model, most points should fall within this region.

The \code{type} argument controls how residuals are computed in residual
mode. See \code{\link{residuals.brma}} for details on each type.
The \code{sampling_heterogeneity} and \code{sampling_bias} arguments are
ignored in residual mode.

For GLMM models, observed effect sizes are computed from the raw frequency
data using formulas equivalent to \code{metafor::escalc}. Residual-mode
GLMM funnels use approximate effect-size-scale residual/PIT companions, not
exact PIT diagnostics for the raw count likelihood.
}
\examples{
\dontrun{
if (requireNamespace("metadat", quietly = TRUE) &&
    requireNamespace("metafor", quietly = TRUE)) {
  data(dat.bcg, package = "metadat")
  dat <- metafor::escalc(
    measure = "RR",
    ai      = tpos,
    bi      = tneg,
    ci      = cpos,
    di      = cneg,
    data    = dat.bcg
  )

  fit <- brma(yi = yi, vi = vi, data = dat, measure = "RR")
  funnel(fit)
  funnel(fit, pch = 19, col = "blue", bg = "lightblue")

  fit_reg <- brma(
    yi      = yi,
    vi      = vi,
    mods    = ~ ablat + year,
    data    = dat,
    measure = "RR"
  )
  fit_reg <- add_loo(fit_reg)
  funnel(fit_reg)
  funnel(fit_reg, type = "outcome")

  funnel_data <- funnel(fit, as_data = TRUE)
  funnel(fit, plot_type = "ggplot")
}
}

}
\seealso{
\code{\link[=residuals.brma]{residuals.brma()}}, \code{\link[=rstandard.brma]{rstandard.brma()}}, \code{\link[=rstudent.brma]{rstudent.brma()}},
\code{\link[=predict.brma]{predict.brma()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/input-data.R
\name{print.RoBMA_data}
\alias{print.RoBMA_data}
\title{Print method for RoBMA_data objects}
\usage{
\method{print}{RoBMA_data}(x, n = 6, ...)
}
\arguments{
\item{x}{an RoBMA_data object}

\item{n}{maximum number of rows to display for each component. Defaults to 6.}

\item{...}{additional arguments (ignored)}
}
\value{
Invisibly returns the input object.
}
\description{
Prints a concise summary of an RoBMA_data object.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/plot.R
\name{plot_weightfunction}
\alias{plot_weightfunction}
\alias{plot_weightfunction.brma}
\title{Plots Weight Function of brma Object}
\usage{
plot_weightfunction(x, ...)

\method{plot_weightfunction}{brma}(
  x,
  rescale_p_values = TRUE,
  prior = FALSE,
  plot_type = "base",
  show_data = TRUE,
  dots_data = NULL,
  dots_prior = NULL,
  ...
)
}
\arguments{
\item{x}{a fitted \code{brma} object with a weightfunction/selection
component, such as \code{bselmodel()} or a \code{RoBMA()} object with
weightfunction priors.}

\item{...}{list of additional graphical arguments
to be passed to the plotting function. Supported arguments
are \code{lwd}, \code{lty}, \code{col}, \code{col.fill},
\code{xlab}, \code{ylab}, \code{main}, \code{xlim}, \code{ylim}
to adjust the line thickness, line type, line color, fill color,
x-label, y-label, title, x-axis range, and y-axis range
respectively.}

\item{rescale_p_values}{whether to rescale p-values to the interval shown
by the weightfunction plot. Defaults to \code{TRUE}.}

\item{prior}{whether prior distribution should be added to
figure. Defaults to \code{FALSE}.}

\item{plot_type}{whether to use a base plot \code{"base"}
or ggplot2 \code{"ggplot"} for plotting. Defaults to
\code{"base"}.}

\item{show_data}{whether observed one-sided p-values should be shown as rug
marks on the weightfunction axis. Defaults to \code{TRUE}.}

\item{dots_data}{list of additional graphical arguments for observed
p-value rug marks. Supported arguments include \code{col}/\code{color},
\code{alpha}, \code{lwd}/\code{linewidth}/\code{size},
\code{side}/\code{rug_side}, and
\code{height}/\code{rug_height}/\code{ticksize}.}

\item{dots_prior}{list of additional graphical arguments
to be passed to the plotting function of the prior
distribution. Supported arguments are \code{lwd},
\code{lty}, \code{col}, and \code{col.fill}, to adjust
the line thickness, line type, line color, and fill color
of the prior distribution respectively.}
}
\value{
\code{plot_weightfunction.brma} returns either \code{NULL} if
\code{plot_type = "base"} or a \code{ggplot2} object if
\code{plot_type = "ggplot"}.
The method errors for fitted objects without a weightfunction component.
}
\description{
\code{plot_weightfunction.brma} visualizes the posterior
(and optionally prior) publication-bias weight function of a brma object.
}
\examples{
\dontrun{
if (requireNamespace("metadat", quietly = TRUE)) {
  data(dat.lehmann2018, package = "metadat")
  fit <- bselmodel(
    yi      = yi,
    vi      = vi,
    data    = dat.lehmann2018,
    measure = "SMD",
    seed    = 1,
    silent  = TRUE
  )

  plot_weightfunction(fit)
  plot_weightfunction(fit, prior = TRUE)
}
}


}
\seealso{
\code{\link[=RoBMA]{RoBMA()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/summary_heterogeneity.R
\name{summary_heterogeneity.brma}
\alias{summary_heterogeneity.brma}
\title{Summary of Heterogeneity for brma Objects}
\usage{
\method{summary_heterogeneity}{brma}(object, probs = c(0.025, 0.975), ...)
}
\arguments{
\item{object}{a fitted brma object}

\item{probs}{quantiles of the posterior distribution to be displayed.
Defaults to \code{c(.025, .975)} for 95\% credible intervals.}

\item{...}{additional arguments (currently ignored)}
}
\value{
A list of class \code{summary_heterogeneity.brma} containing:
\itemize{
\item \code{estimates}: A \code{BayesTools_table} with heterogeneity statistics
}
}
\description{
Computes the absolute heterogeneity (tau, tau^2) and
relative measures of heterogeneity (I^2, H^2) for a fitted brma object.
}
\details{
For standard (2-level) random-effects models, the function reports:
\itemize{
\item \code{tau}: Between-study standard deviation
\item \code{tau2}: Between-study variance
\item \code{I2}: Percentage of total variance due to heterogeneity
\item \code{H2}: Ratio of total to sampling variance
}

For multilevel (3-level) models with nested effects, the function additionally
partitions heterogeneity into estimate-level and cluster-level components:
\itemize{
\item \code{rho}: Proportion of heterogeneity variance allocated to clusters
\item \code{tau [within]}: Estimate-level standard deviation
\item \code{tau [between]}: Cluster-level standard deviation
\item \code{tau2 [within]}: Estimate-level variance
\item \code{tau2 [between]}: Cluster-level variance
\item \code{I2 [within]}: Percentage of variance due to estimate-level heterogeneity
\item \code{I2 [between]}: Percentage of variance due to cluster-level heterogeneity
}

For location-scale models, \code{tau2} aggregates the observation-specific
heterogeneity variances \eqn{\tau_i^2}; the corresponding \code{tau} summary
is the square root of this aggregate variance. The relative \eqn{I^2} and
\eqn{H^2} measures average the observation-specific indices.

The I^2 and H^2 statistics are computed following the metafor package
implementation, using the "typical" sampling variance formula from
\insertCite{higgins2002quantifying;textual}{RoBMA}. For multilevel models,
the partitioned I^2 follows the approach described in the metafor documentation.
}
\examples{
\dontrun{
if (requireNamespace("metadat", quietly = TRUE)) {
  data(dat.lehmann2018, package = "metadat")
  fit <- brma(
    yi      = yi,
    vi      = vi,
    data    = dat.lehmann2018,
    measure = "SMD",
    seed    = 1,
    silent  = TRUE
  )

  summary_heterogeneity(fit)
}
}

}
\references{
\insertAllCited{}
}
\seealso{
\code{\link[=pooled_heterogeneity]{pooled_heterogeneity()}}, \code{\link[=summary.brma]{summary.brma()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/covratio.R
\name{covratio.brma}
\alias{covratio.brma}
\alias{covratio}
\title{COVRATIO for brma Objects}
\usage{
\method{covratio}{brma}(model, type = "mods", ...)
}
\arguments{
\item{model}{a fitted brma object.}

\item{type}{type of parameters to be summarized. Defaults to \code{"mods"}
(for the effect size and meta-regression coefficients). Use \code{"scale"}
for heterogeneity and scale-regression coefficients.}

\item{...}{additional arguments. The internal \code{.weights} argument can
supply precomputed PSIS weights for callers that already extracted them.}
}
\value{
A named numeric vector of COVRATIO values, one for each observation.
}
\description{
Computes COVRATIO for a fitted brma object. COVRATIO measures
the change in the determinant of the covariance matrix of the estimates
when observation \eqn{i} is removed.
}
\details{
COVRATIO is computed using importance sampling weights to approximate the
leave-one-out covariance matrices without refitting the model.
Estimate-unit LOO must first be computed with
\code{model <- add_loo(model, unit = "estimate")}, unless internal weights
are supplied.

\deqn{COVRATIO_i = \frac{\det(Cov(\beta)_{-i})}{\det(Cov(\beta))}}

Values > 1 indicate that the observation improves precision (decreases
variance), while values < 1 indicate that the observation decreases precision
(increases variance).
If any included parameter has zero posterior variance, or if a full or LOO
covariance determinant is zero or non-finite, COVRATIO is undefined. In that
case, values are reported as \code{NaN} with a printed note when available.
}
\examples{
\dontrun{
if (requireNamespace("metadat", quietly = TRUE)) {
  data(dat.lehmann2018, package = "metadat")
  fit <- bPET(yi = yi, vi = vi, data = dat.lehmann2018, measure = "SMD")
  fit <- add_loo(fit)

  covratio(fit)
}
}

}
\seealso{
\code{\link{influence.brma}}, \code{\link{dffits.brma}}, \code{\link{cooks.distance.brma}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/datasets.R
\docType{data}
\name{Poulsen2006}
\alias{Poulsen2006}
\title{5 studies with a tactile outcome assessment from
\insertCite{poulsen2006potassium;textual}{RoBMA} of the effect of potassium-containing toothpaste
on dentine hypersensitivity}
\format{
A data.frame with 3 columns and 5 observations:
\describe{
\item{\code{d}}{Cohen's d effect size.}
\item{\code{se}}{Standard error of d.}
\item{\code{study}}{Study label.}
}
}
\usage{
Poulsen2006
}
\description{
The data set contains Cohen's d effect sizes, standard errors,
and labels for 5 studies assessing the tactile outcome from a meta-analysis of
the effect of potassium-containing toothpaste on dentine hypersensitivity
\insertCite{poulsen2006potassium}{RoBMA} which was used as an example in
\insertCite{bartos2021bayesian;textual}{RoBMA}.
}
\references{
\insertAllCited{}
}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/dfbetas.R
\name{dfbetas.brma}
\alias{dfbetas.brma}
\title{DFBETAS for brma Objects}
\usage{
\method{dfbetas}{brma}(
  model,
  type = "mods",
  standardized_coefficients = FALSE,
  transform_factors = TRUE,
  return_loo_estimates = FALSE,
  ...
)
}
\arguments{
\item{model}{a fitted brma object.}

\item{type}{type of parameters to be summarized. Defaults to \code{"mods"}
(for the effect size and meta-regression coefficients). The other options are
\code{"scale"} (for the heterogeneity and scale-regression coefficients) and
\code{"bias"} (for omega, PET, and PEESE publication-bias parameters).}

\item{standardized_coefficients}{whether to show standardized meta-regression coefficients.
Defaults to \code{FALSE}. When set to \code{TRUE}, standardized meta-regression
coefficients are returned for the intercept and continuous predictors. These coefficients
correspond to the standardized scale on which prior distributions are specified by default
(i.e., \code{standardize_continuous_predictors = TRUE}).}

\item{transform_factors}{whether to transform factors to their original names.
Defaults to \code{TRUE}.}

\item{return_loo_estimates}{whether to return the leave-one-out coefficient
estimates used to compute DFBETAS instead of standardized DFBETAS values.
Defaults to \code{FALSE}.}

\item{...}{additional arguments (currently ignored).}
}
\value{
If \code{return_loo_estimates = FALSE}, a data frame with \eqn{K} rows
(observations) and \eqn{P} columns (parameters), containing DFBETAS values.
If \code{return_loo_estimates = TRUE}, returns the corresponding leave-one-out
coefficient estimates. Row names correspond to study labels (if available)
or indices.
}
\description{
Computes DFBETAS (Difference in BETAS, standardized) for a
fitted brma object. DFBETAS measures the influence of each observation on
the estimated model coefficients. Positive values indicate that deleting the
observation yields a smaller estimate, negative values indicate that deleting
the observation yields a larger estimate.
}
\details{
This function computes DFBETAS values using the Leave-One-Out (LOO)
approximation based on Pareto Smoothed Importance Sampling (PSIS) weights.
Ideally, DFBETAS is defined as:
\deqn{DFBETAS_{ij} = \frac{\hat{\beta}_j - \hat{\beta}_{j(-i)}}{SE(\hat{\beta}_{j(-i)})}}
where \eqn{\hat{\beta}_j} is the estimate of the \eqn{j}-th coefficient using
the full data, \eqn{\hat{\beta}_{j(-i)}} is the estimate when observation \eqn{i}
is omitted, and \eqn{SE(\hat{\beta}_{j(-i)})} is the standard error of the
coefficient when observation \eqn{i} is omitted.

In the Bayesian context using LOO approximation:
\itemize{
\item \eqn{\hat{\beta}_{j(-i)}} is estimated as the importance sampling
weighted mean of the posterior samples, using PSIS weights \eqn{w_{is}}.
\item \eqn{SE(\hat{\beta}_{j(-i)})} is estimated as the importance sampling
weighted standard deviation of the posterior samples.
}

This approximation allows computing influence statistics without refitting
the model \eqn{K} times, making it computationally efficient.
For \code{type = "bias"}, fixed identification parameters (e.g., the reference
\eqn{\omega = 1} interval) can have zero LOO posterior standard deviation.
These parameters are retained in the output, but their DFBETAS values are
reported as \code{NaN} because the standardized diagnostic is undefined.

Note: This function requires that LOO-CV has been computed for the model
using \code{\link{add_loo}}.
}
\examples{
\dontrun{
if (requireNamespace("metadat", quietly = TRUE)) {
  data(dat.lehmann2018, package = "metadat")
  fit <- bPET(yi = yi, vi = vi, data = dat.lehmann2018, measure = "SMD")
  fit <- add_loo(fit)

  inf <- dfbetas(fit)
  plot(inf[, 1], type = "h")
}
}

}
\seealso{
\code{\link{add_loo}}, \code{\link{loo_weights.brma}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/wrappers.R
\name{nobs.brma}
\alias{nobs.brma}
\title{Number of Observations for brma Objects}
\usage{
\method{nobs}{brma}(object, ...)
}
\arguments{
\item{object}{a fitted brma object}

\item{...}{additional arguments (currently ignored)}
}
\value{
An integer giving the number of observations.
}
\description{
Extract the number of observations (studies/estimates) from
a fitted brma object.
}
\seealso{
\code{\link[=brma]{brma()}}, \code{\link[=brma.glmm]{brma.glmm()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/marglik.R
\name{add_marglik.brma}
\alias{add_marglik.brma}
\alias{add_marglik}
\title{Add Marginal Likelihood to brma Objects}
\usage{
\method{add_marglik}{brma}(object, ...)
}
\arguments{
\item{object}{a brma model object.}

\item{...}{additional arguments (currently not used).}
}
\value{
The brma object with the marginal likelihood result stored in
\code{object[["marglik"]]}.
}
\description{
Compute the marginal likelihood of a brma model using
bridge sampling and store the result in the object.
}
\details{
The marginal likelihood is computed using the \code{bridgesampling} package
via the \code{BayesTools::JAGS_bridgesampling} wrapper. The result is stored
in the object and can be extracted using \code{\link{bridge_sampler.brma}}.

Product-space model-averaging objects (\code{BMA.norm}, \code{BMA.glmm},
and \code{RoBMA}) do not expose a bridge-sampling marginal likelihood;
use predictive comparison methods such as \code{\link{loo.brma}} instead.
}
\examples{
\dontrun{
if (requireNamespace("metadat", quietly = TRUE)) {
  data(dat.lehmann2018, package = "metadat")
  fit <- brma(yi = yi, vi = vi, data = dat.lehmann2018, measure = "SMD")

  fit <- add_marglik(fit)

  bridge <- bridge_sampler(fit)
  print(bridge)

  logml(fit)
}
}

}
\seealso{
\code{\link{bridge_sampler.brma}}, \code{\link{logml.brma}},
\code{\link{bf.brma}}, \code{\link{post_prob.brma}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/zplot.R
\name{lines.zplot_brma}
\alias{lines.zplot_brma}
\title{Add Zplot Density Lines}
\usage{
\method{lines}{zplot_brma}(
  x,
  plot_type = "base",
  probs = c(0.025, 0.975),
  max_samples = 10000,
  plot_ci = TRUE,
  extrapolate = FALSE,
  from = -6,
  to = 6,
  by = 0.05,
  length.out = NULL,
  col = "black",
  as_data = FALSE,
  ...
)
}
\arguments{
\item{x}{a zplot_brma object.}

\item{plot_type}{graphics system: \code{"base"} or \code{"ggplot"}.
Defaults to \code{"base"}.}

\item{probs}{quantiles for credible intervals. Defaults to \code{c(.025, .975)}.}

\item{max_samples}{maximum posterior samples for density. Defaults to 10000.
Use \code{Inf} to use all posterior samples.}

\item{plot_ci}{whether to show credible interval bands. Defaults to \code{TRUE}.}

\item{extrapolate}{whether to remove bias adjustments. Defaults to \code{FALSE}.}

\item{from, to}{z-value range for density. Defaults to \code{-6} and \code{6}.}

\item{by}{step size for density points. Defaults to 0.05.}

\item{length.out}{number of density points (alternative to \code{by}).}

\item{col}{line color. Defaults to \code{"black"}.}

\item{as_data}{whether to return data instead of plotting. Defaults to \code{FALSE}.}

\item{...}{additional graphical parameters.}
}
\value{
\code{NULL} invisibly for base graphics, ggplot2 layers for ggplot,
or a data frame with columns \code{x}, \code{y}, \code{y_lCI}, \code{y_uCI}
if \code{as_data = TRUE}.
}
\description{
Adds model-implied density lines to an existing zplot.
}
\details{
When \code{extrapolate = FALSE}, the density includes all bias adjustments
(PET/PEESE regression, selection weights) representing the fitted model.
When \code{extrapolate = TRUE}, bias adjustments are removed to show the
hypothetical distribution without publication bias. Under selection models,
the curve is scaled by inverse selection probability and need not integrate
to one.
}
\seealso{
\code{\link[=plot.zplot_brma]{plot.zplot_brma()}}, \code{\link[=hist.zplot_brma]{hist.zplot_brma()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/bPET.R
\name{bPET}
\alias{bPET}
\title{Bayesian Precision-Effect Test (PET) Model}
\usage{
bPET(
  yi,
  vi,
  sei,
  weights,
  ni,
  mods,
  scale,
  cluster,
  data,
  slab,
  subset,
  measure,
  prior_effect,
  prior_heterogeneity,
  prior_mods,
  prior_scale,
  prior_heterogeneity_allocation,
  prior_bias,
  standardize_continuous_predictors = TRUE,
  set_contrast_factor_predictors = "treatment",
  prior_unit_information_sd,
  rescale_priors = 1,
  prior_informed_field,
  prior_informed_subfield,
  effect_direction = "detect",
  sample = 5000,
  burnin = 2000,
  adapt = 500,
  chains = 3,
  thin = 1,
  parallel = FALSE,
  autofit = FALSE,
  autofit_control = set_autofit_control(),
  convergence_checks = set_convergence_checks(),
  seed = NULL,
  silent,
  ...
)
}
\arguments{
\item{yi}{a vector of effect sizes, or a formula with the effect size on the
left-hand side and location moderators on the right-hand side (for example
\code{yi ~ x1 + x2}). If a formula is supplied, \code{mods} must not be specified.}

\item{vi}{a vector of sampling variances. Either \code{vi} or \code{sei} must be
supplied for normal models.}

\item{sei}{a vector of standard errors. Either \code{vi} or \code{sei} must be
supplied for normal models.}

\item{weights}{an optional vector of positive likelihood weights. For
normal/effect-size models, each weight powers the estimate likelihood. For
constructors with GLMM raw-count input, each weight powers the paired
two-arm likelihood for one study.}

\item{ni}{an optional vector of sample sizes. Used for \code{measure = "GEN"}
or when estimating \code{"UISD"}).}

\item{mods}{an optional matrix, data.frame, or formula specifying
location moderators (meta-regressors). Formula input is evaluated in \code{data}.}

\item{scale}{an optional matrix, data.frame, or formula specifying
scale predictors for location-scale models. Formula input is evaluated in
\code{data}.}

\item{cluster}{an optional vector of cluster identifiers for multilevel
meta-analysis.}

\item{data}{an optional data frame containing the variables.}

\item{slab}{an optional vector of study labels.}

\item{subset}{an optional logical or numeric vector specifying a subset of
data to be used.}

\item{measure}{a character string specifying the effect size measure.
Normal/effect-size constructors require an explicit value and support
\code{"SMD"}, \code{"ZCOR"}, \code{"RR"}, \code{"OR"}, \code{"HR"}, \code{"RD"}, \code{"IRR"}, and \code{"GEN"}.
Use \code{"GEN"} only for general effect sizes without a known unit information
standard deviation. GLMM raw-count constructors support only \code{"OR"} and
\code{"IRR"} and default to \code{"OR"}.}

\item{prior_effect}{prior distribution for the effect size (\eqn{\mu}) parameter
(the intercept). If omitted, a default prior is constructed. In single-model
functions, explicit \code{NULL} or \code{FALSE} sets a spike at zero.}

\item{prior_heterogeneity}{prior distribution for the heterogeneity (\eqn{\tau})
parameter. If omitted, a default prior is constructed. In single-model
functions, explicit \code{NULL} or \code{FALSE} sets a spike at zero.}

\item{prior_mods}{prior distribution for the moderators (\eqn{\beta}) parameters.
A single prior applies to all terms; a named list can specify term-specific
priors. If omitted or \code{NULL}, default priors are used.}

\item{prior_scale}{prior distribution for the scale (\eqn{\delta}) parameters.
A single prior applies to all terms; a named list can specify term-specific
priors. If omitted or \code{NULL}, default priors are used.}

\item{prior_heterogeneity_allocation}{prior distribution for the fraction of
heterogeneity allocated to the cluster-level component in multilevel models
(\eqn{\rho}). If omitted or \code{NULL}, defaults to \code{Beta(1, 1)}.}

\item{prior_bias}{PET regression-coefficient prior created by
\code{prior_PET()}. If omitted or \code{NULL}, the default is a Cauchy prior
centered at 0, truncated to positive values, with scale from
\code{RoBMA.get_option("default_bias_PET.scale")}.}

\item{standardize_continuous_predictors}{logical. Whether to standardize continuous predictors.
Defaults to \code{TRUE}.}

\item{set_contrast_factor_predictors}{character. How to set contrast for factor predictors.
Defaults are constructor-specific and shown in each function usage; single-model
constructors use \code{"treatment"}, while model-averaging constructors use \code{"meandif"}.}

\item{prior_unit_information_sd}{numeric. The unit information standard deviation (\eqn{\sigma_{unit}}).
Cannot be used together with \code{prior_informed_field}.}

\item{rescale_priors}{numeric. A scaling factor for supported prior distributions.
Point and none priors are unchanged. For constructors with publication-bias
prior distributions, \code{rescale_priors} does not rescale them except for the
default PEESE prior's UISD adjustment. Defaults to 1.}

\item{prior_informed_field}{character. The field of the informed prior distributions.
Omit to use the standard default prior specification; explicit \code{NULL} is invalid.}

\item{prior_informed_subfield}{character. The subfield of the informed prior distributions.
Omit to use the field-specific default, such as \code{"Cochrane"} for
\code{prior_informed_field = "medicine"}; explicit \code{NULL} is invalid.}

\item{effect_direction}{direction used by publication-bias adjustments.
\code{"positive"} assumes statistically significant positive estimates are more
likely to be selected; \code{"negative"} mirrors the selection direction;
\code{"detect"} infers the direction from the fitted data.}

\item{sample}{numeric. Number of MCMC samples to save. Defaults to \code{5000}.}

\item{burnin}{numeric. Number of burn-in iterations. Defaults to \code{2000}.}

\item{adapt}{numeric. Number of adaptation iterations. Defaults to \code{500}.}

\item{chains}{numeric. Number of MCMC chains. Defaults to \code{3}.}

\item{thin}{numeric. Thinning interval. Defaults to \code{1}.}

\item{parallel}{logical. Whether to run MCMC chains in parallel. Defaults to \code{FALSE}.}

\item{autofit}{logical. Whether to automatically extend the MCMC chains if convergence is not met.
Defaults to \code{FALSE}.}

\item{autofit_control}{list of autofit control settings. See \code{\link[=set_autofit_control]{set_autofit_control()}} for details.}

\item{convergence_checks}{list of convergence check settings. See \code{\link[=set_convergence_checks]{set_convergence_checks()}} for details.}

\item{seed}{numeric. Random seed for reproducibility. Defaults to \code{NULL}.}

\item{silent}{logical. Whether to suppress output. Constructors with no
explicit default use \code{RoBMA.get_option("silent")} when \code{silent} is omitted.
Model-averaging wrappers default to \code{TRUE} unless explicitly changed.}

\item{...}{additional advanced arguments. Fitting functions reject unused
arguments; currently recognized internal arguments include \code{only_data},
\code{only_priors}, \code{is_JASP}, and \code{is_JASP_prefix}.}
}
\value{
A fitted object of class \code{c("bPET", "brma")} containing a single
PET publication-bias model fit.
}
\description{
Function for fitting random-effects, meta-regression, multilevel,
and location-scale meta-analytic PET models.
}
\examples{
\dontrun{
if (requireNamespace("metadat", quietly = TRUE)) {
  data(dat.lehmann2018, package = "metadat")

  fit <- bPET(
    yi      = yi,
    vi      = vi,
    data    = dat.lehmann2018,
    measure = "SMD",
    seed    = 1,
    silent  = TRUE
  )

  summary(fit)
  funnel(fit)
}
}

}
\seealso{
\link{publication_bias_prior_specification}, \code{\link[=RoBMA]{RoBMA()}}, \code{\link[=bPEESE]{bPEESE()}},
\code{\link[=bselmodel]{bselmodel()}}, \code{\link[=summary.brma]{summary.brma()}}, \code{\link[=funnel.brma]{funnel.brma()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/brma_samples.R
\name{as.matrix.brma_samples}
\alias{as.matrix.brma_samples}
\title{Convert brma_samples to Matrix}
\usage{
\method{as.matrix}{brma_samples}(x, ...)
}
\arguments{
\item{x}{a \code{brma_samples} object}

\item{...}{additional arguments (ignored)}
}
\value{
A plain matrix of posterior samples.
}
\description{
Converts a \code{brma_samples} object to a plain matrix,
removing all brma_samples-specific attributes.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/zplot.R
\name{zplot.brma}
\alias{zplot.brma}
\alias{zplot}
\alias{zplot.zplot_brma}
\title{Plot Zplot Diagnostics Directly}
\usage{
\method{zplot}{brma}(
  object,
  significance_level = stats::qnorm(0.975),
  summary_max_samples = 10000,
  ...
)

\method{zplot}{zplot_brma}(object, ...)
}
\arguments{
\item{object}{a normal-outcome \code{brma} object, or a
\code{zplot_brma} object.}

\item{significance_level}{z-value threshold for significance. Defaults
to \code{qnorm(0.975)} (two-sided alpha = 0.05).}

\item{summary_max_samples}{maximum number of posterior samples used for the
EDR and missing-study summaries stored in the generated zplot object.
This is separate from the plot-density \code{max_samples} argument accepted
by \code{plot.zplot_brma()} through \code{...}. Defaults to 10000. Use
\code{Inf} to use all posterior samples.}

\item{...}{arguments passed to \code{\link[=plot.zplot_brma]{plot.zplot_brma()}}.}
}
\value{
\code{NULL} invisibly for base graphics, or a ggplot2 object.
}
\description{
Convenience wrapper for creating and plotting zplot diagnostics
from a fitted \code{brma} object.
}
\details{
When \code{object} already inherits from \code{zplot_brma},
\code{zplot()} dispatches directly to \code{plot.zplot_brma()} without
recomputing stored summaries.
}
\examples{
\dontrun{
if (requireNamespace("metadat", quietly = TRUE)) {
  data(dat.lehmann2018, package = "metadat")
  fit <- bPET(yi = yi, vi = vi, data = dat.lehmann2018, measure = "SMD")

  zplot(fit)
}
}

}
\seealso{
\code{\link[=as_zplot.brma]{as_zplot.brma()}}, \code{\link[=plot.zplot_brma]{plot.zplot_brma()}}, \code{\link[=summary.zplot_brma]{summary.zplot_brma()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/regplot.R
\name{regplot.brma}
\alias{regplot.brma}
\alias{regplot}
\title{Regression Plot (Bubble Plot) for brma Object}
\usage{
\method{regplot}{brma}(
  x,
  mod = NULL,
  pred = TRUE,
  ci = TRUE,
  pi = FALSE,
  si = FALSE,
  level = 95,
  at = NULL,
  digits = 2,
  transf = NULL,
  atransf = NULL,
  targs = NULL,
  refline = NULL,
  psize = NULL,
  plim = c(0.5, 3),
  by = NULL,
  legend = TRUE,
  xlab = NULL,
  ylab = NULL,
  xlim = NULL,
  ylim = NULL,
  sampling_bias = TRUE,
  sei = NULL,
  max_samples = 10000,
  plot_type = "base",
  as_data = FALSE,
  ...
)
}
\arguments{
\item{x}{a fitted brma object with moderators}

\item{mod}{index or name of the moderator variable to plot on the x-axis.
If not specified and only one moderator is present, that moderator is used.
If multiple moderators are present, this argument is required.}

\item{pred}{logical; whether to show the prediction line. Defaults to \code{TRUE}.}

\item{ci}{logical; whether to show credible interval bands. Defaults to \code{TRUE}.}

\item{pi}{logical; whether to show prediction interval bands. Defaults to \code{FALSE}.}

\item{si}{logical; whether to show sampling interval bands. Defaults to \code{FALSE}.
The sampling interval shows the expected range of observed effect sizes,
incorporating both heterogeneity (tau) and a representative level of
sampling error (median SE across studies by default). When the model includes
publication bias adjustments and \code{sampling_bias = TRUE}, the
sampling interval incorporates the expected distortion from the
selection process.}

\item{level}{numeric; credible/prediction interval level in percent.
Defaults to \code{95}.}

\item{at}{numeric vector; for continuous moderators, values at which to
evaluate the prediction. If not specified, uses a sequence across the
observed range.}

\item{digits}{integer; number of decimal places for labels. Defaults to \code{2}.}

\item{transf}{function; transformation to apply to the y-axis (effect sizes).
Defaults to \code{NULL} (no transformation).}

\item{atransf}{reserved for axis-label transformations. Currently not
implemented and must be \code{NULL}.}

\item{targs}{reserved for additional transformation arguments. Currently not
implemented and must be \code{NULL}.}

\item{refline}{numeric; position of horizontal reference line.
Defaults to \code{NULL} (no reference line).}

\item{psize}{numeric vector or \code{NULL}; point sizes for each study.
If scalar, it is recycled to all studies. If \code{NULL} (default), sizes
are computed based on inverse sampling variance.}

\item{plim}{numeric vector of length 2; range for point sizes.
Defaults to \code{c(0.5, 3)}.}

\item{by}{character; name of a moderator variable to use for separate
lines/colors. Defaults to \code{NULL}. If omitted and the selected
moderator enters exactly one two-way interaction, the other variable in the
interaction is used automatically. Continuous \code{by} moderators are shown
at mean - SD, mean, and mean + SD.}

\item{legend}{logical; whether to show legend when \code{by} is specified.
Defaults to \code{TRUE}.}

\item{xlab}{character; x-axis label. Defaults to the moderator name.}

\item{ylab}{character; y-axis label. Defaults to "Observed Effect Size".}

\item{xlim}{numeric vector of length 2; x-axis limits. Defaults to data range.}

\item{ylim}{numeric vector of length 2; y-axis limits. Defaults to data range.}

\item{sampling_bias}{whether publication bias should be incorporated into
plotted predictions and sampling intervals. Defaults to \code{TRUE}. For
PET/PEESE models, this includes the expected regression bias in predictions.
For selection models, sampling-bias adjustment applies to sampling intervals
when \code{si = TRUE}; the mean prediction is unchanged.}

\item{sei}{single positive numeric value used as the reference standard
error for sampling-bias and sampling-interval calculations. Defaults to the
median observed standard error.}

\item{max_samples}{maximum number of posterior samples used for prediction
summaries and interval bands. Defaults to \code{10000}. Use \code{Inf} to
use all posterior samples.}

\item{plot_type}{character; whether to use base R graphics (\code{"base"})
or ggplot2 (\code{"ggplot"}). Defaults to \code{"base"}.}

\item{as_data}{logical; if \code{TRUE}, returns plot data instead of
creating plot. Defaults to \code{FALSE}.}

\item{...}{additional graphical arguments:
\describe{
\item{main}{character string for plot title}
\item{pch}{point symbol (default: 21)}
\item{col}{point border color (default: "black")}
\item{bg}{point fill color (default: "#A6A6A6")}
\item{lcol}{line color (default: "black")}
\item{lwd}{line width (default: 2)}
\item{shade}{CI/PI/SI band shading (default: TRUE)}
\item{col.ci}{CI band color (default: "gray70")}
\item{col.pi}{PI band color (default: "gray85")}
\item{col.si}{SI band color (default: "gray92")}
\item{alpha.ci}{CI band transparency (default: 0.4)}
\item{alpha.pi}{PI band transparency (default: 0.2)}
\item{alpha.si}{SI band transparency (default: 0.15)}
\item{jitter}{jitter amount for categorical moderators (default: 0.2)}
\item{box.width}{box width for categorical interval summaries (default: 0.5)}
}}
}
\value{
\code{regplot.brma} returns \code{NULL} invisibly if
\code{plot_type = "base"} or a ggplot object if \code{plot_type = "ggplot"}.
If \code{as_data = TRUE}, returns a list with plot data components.
}
\description{
\code{regplot.brma} creates a regression plot (also known as
bubble plot) for a fitted brma object with moderators. The plot displays
observed effect sizes against a moderator variable, with point sizes
proportional to study precision.
}
\details{
The regression plot (bubble plot) is a standard visualization for
meta-regression results. It displays:
\itemize{
\item Observed effect sizes (y-axis) against moderator values (x-axis)
\item Point sizes proportional to study precision (inverse variance)
\item Prediction line showing the estimated regression relationship
\item Confidence bands showing uncertainty in the mean prediction
\item Optional prediction bands showing expected range of true effects
\item Optional sampling interval bands showing expected range of observed outcomes
}

For continuous moderators, predictions are computed across the observed
range of the moderator. For categorical moderators (factors), predictions
are computed at each factor level with optional jittering of points.

The \code{by} argument allows displaying separate regression lines for
different levels of a second moderator, useful for visualizing interactions.
}
\examples{
\dontrun{
if (requireNamespace("metadat", quietly = TRUE) &&
    requireNamespace("metafor", quietly = TRUE)) {
  data(dat.bcg, package = "metadat")
  dat <- metafor::escalc(
    measure = "RR",
    ai      = tpos,
    bi      = tneg,
    ci      = cpos,
    di      = cneg,
    data    = dat.bcg
  )

  fit <- brma(
    yi      = yi,
    vi      = vi,
    mods    = ~ ablat + year,
    data    = dat,
    measure = "RR"
  )
  regplot(fit, mod = "ablat")
  regplot(fit, mod = "year", pi = TRUE, si = TRUE)
  regplot(fit, mod = "ablat", plot_type = "ggplot")

  fit_cat <- brma(yi = yi, vi = vi, mods = ~ alloc, data = dat, measure = "RR")
  regplot(fit_cat)
}
}

}
\seealso{
\code{\link[=funnel.brma]{funnel.brma()}}, \code{\link[=predict.brma]{predict.brma()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/datasets.R
\docType{data}
\name{Johnides2025}
\alias{Johnides2025}
\title{412 effect sizes from a meta-analysis of secondary benefits of family-based treatments
by \insertCite{johnides2025secondary;textual}{RoBMA}}
\format{
A data.frame with 3 columns and 412 observations:
\describe{
\item{\code{study}}{Study label.}
\item{\code{d}}{Cohen's d effect size.}
\item{\code{se}}{Standard error of d.}
}
}
\usage{
Johnides2025
}
\description{
The data set contains Cohen's d effect sizes, standard errors, and study labels
from a meta-analysis investigating the extent to which family-based treatments for children
with mental health, physical health, and developmental disorders provide secondary benefits
to the children's siblings and caregivers \insertCite{johnides2025secondary}{RoBMA}.
}
\references{
\insertAllCited{}
}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/RoBMA-package.R
\docType{package}
\name{RoBMA-package}
\alias{RoBMA-package}
\alias{RoBMA_package}
\alias{RoBMA.package}
\title{RoBMA: Robust Bayesian Meta-Analysis}
\description{
RoBMA provides Bayesian meta-analysis, meta-regression, multilevel
meta-analysis, model averaging, and publication-bias adjustment. The main
user-facing fitters are \code{\link{RoBMA}}, \code{\link{BMA}},
\code{\link{brma}}, \code{\link{brma.glmm}}, \code{\link{bselmodel}},
\code{\link{bPET}}, and \code{\link{bPEESE}}.
}
\section{User guide}{

See \insertCite{bartos2021no;textual}{RoBMA},
\insertCite{maier2020robust;textual}{RoBMA},
\insertCite{bartos2020adjusting;textual}{RoBMA}, and
\insertCite{bartos2025robust;textual}{RoBMA} for the RoBMA methodology.
Use \code{vignette(package = "RoBMA")} to list installed vignettes.
}

\references{
\insertAllCited{}
}
\seealso{
Useful links:
\itemize{
  \item \url{https://fbartos.github.io/RoBMA/}
  \item Report bugs at \url{https://github.com/FBartos/RoBMA/issues}
}

}
\author{
Frantisek Bartos \email{f.bartos96@gmail.com}
}
\keyword{package}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/wrappers.R
\name{true_effects.brma}
\alias{true_effects.brma}
\title{True Effects for brma Objects}
\usage{
\method{true_effects}{brma}(
  object,
  bias_adjusted = FALSE,
  output_measure = NULL,
  transform = NULL,
  probs = c(0.025, 0.975),
  ...
)
}
\arguments{
\item{object}{a fitted brma object}

\item{bias_adjusted}{whether to adjust for publication bias. Defaults to
\code{FALSE}, which returns estimates including publication bias effects
(i.e., what we expect the true effects to be given the biased observations).
Set to \code{TRUE} to obtain bias-corrected estimates.}

\item{output_measure}{effect-size measure for location/effect predictions.
Defaults to the fitted measure. Supported conversions are among \code{"SMD"},
\code{"COR"}, \code{"ZCOR"}, and \code{"OR"}; \code{"RR"}, \code{"HR"},
\code{"IRR"}, \code{"RD"}, and \code{"GEN"} can only be returned on their
fitted measure. Use \code{transform = "EXP"} for ratio-scale output from
log-scale measures.}

\item{transform}{optional display transformation. Currently \code{"EXP"}
exponentiates log-scale measures \code{"OR"}, \code{"RR"}, \code{"HR"},
and \code{"IRR"}.}

\item{probs}{quantiles of the posterior distribution to be displayed.
Defaults to \code{c(.025, .975)} for 95\% credible intervals.}

\item{...}{additional arguments passed to \code{\link{predict.brma}}; wrapper
arguments such as \code{newdata}, \code{type}, \code{quiet},
\code{output_measure}, and \code{transform} are fixed by this method.}
}
\value{
A \code{brma_samples} object containing posterior draws of BLUP or
empirical-Bayes true-effect summaries with one column per estimate. For
existing normal data, these are conditional BLUP means, not simulated
latent-effect draws. When printed, displays a summary table. Use
\code{summary()} to obtain the summary table directly. The samples can be
converted to \pkg{posterior} draws formats using \code{as_draws()}.
}
\description{
Computes the estimated true effects (theta) for a
fitted brma object. This is an alias for \code{\link{blup.brma}}.
}
\details{
This function is identical to \code{\link{blup.brma}}. See that function
for full details on how true effects are computed.
}
\examples{
\dontrun{
if (requireNamespace("metadat", quietly = TRUE)) {
  data(dat.lehmann2018, package = "metadat")
  fit <- brma(
    yi      = yi,
    vi      = vi,
    data    = dat.lehmann2018,
    measure = "SMD",
    seed    = 1,
    silent  = TRUE
  )

  true_effects(fit)
}
}

}
\seealso{
\code{\link[=blup.brma]{blup.brma()}}, \code{\link[=predict.brma]{predict.brma()}}, \code{\link[=pooled_effect]{pooled_effect()}},
\code{\link[=pooled_heterogeneity]{pooled_heterogeneity()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/loo.R
\name{loo.brma}
\alias{loo.brma}
\alias{loo}
\title{LOO-PSIS for brma Objects}
\usage{
\method{loo}{brma}(x, unit = "estimate", ...)
}
\arguments{
\item{x}{a brma model object.}

\item{unit}{output/deletion unit. See \code{\link{add_loo}}.}

\item{...}{additional arguments (currently unused).}
}
\value{
An object of class \code{c("psis_loo", "loo")} as returned by
\code{\link[loo]{loo}}.
}
\description{
Extract the LOO-PSIS object from a brma model object.
The LOO must first be computed using \code{\link{add_loo}}.
}
\details{
This function extracts the LOO object that was previously computed and
stored using \code{object <- add_loo(object, unit = unit)}. If LOO has not
been computed for the requested unit, an error is thrown.

This is the RoBMA S3 generic and \code{brma} method. Use
\code{\link[loo]{loo}} directly for raw log-likelihood arrays or matrices.
}
\examples{
\dontrun{
if (requireNamespace("metadat", quietly = TRUE)) {
  data(dat.lehmann2018, package = "metadat")
  fit <- bPET(yi = yi, vi = vi, data = dat.lehmann2018, measure = "SMD")
  fit <- add_loo(fit)

  loo_fit <- loo(fit)
  print(loo_fit)
}
}

}
\seealso{
\code{\link{add_loo}}, \code{\link[loo]{loo}},
\code{\link[loo]{loo_compare}}, \code{\link[loo]{pareto_k_ids}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/datasets.R
\docType{data}
\name{Hoppen2025}
\alias{Hoppen2025}
\title{37 studies from a meta-analysis of social comparison as a behavior change technique
by \insertCite{hoppen2025meta;textual}{RoBMA}}
\format{
A data.frame with 9 columns and 37 observations:
\describe{
\item{\code{d}}{Cohen's d effect size.}
\item{\code{v}}{Sampling variance of d.}
\item{\code{outcome}}{Outcome type.}
\item{\code{feedback_level}}{Feedback level.}
\item{\code{social_comparison_type}}{Social-comparison type.}
\item{\code{sessions}}{Number of sessions.}
\item{\code{sample_type}}{Sample type.}
\item{\code{sample_size}}{Sample size.}
\item{\code{country}}{Country.}
}
}
\usage{
Hoppen2025
}
\description{
The data set contains Cohen's d effect sizes, variances, and study
characteristics including outcome type, feedback level, social comparison type,
number of sessions, sample type, sample size, and country. The meta-analysis
examined social comparison as a behavior change technique across the behavioral
sciences \insertCite{hoppen2025meta}{RoBMA}.
}
\references{
\insertAllCited{}
}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/zplot.R
\name{summary.zplot_brma}
\alias{summary.zplot_brma}
\title{Summarize Zplot Results}
\usage{
\method{summary}{zplot_brma}(object, probs = c(0.025, 0.975), ...)
}
\arguments{
\item{object}{a zplot_brma object.}

\item{probs}{quantiles of the posterior distribution to display.
Defaults to \code{c(.025, .975)}.}

\item{...}{additional arguments (currently unused).}
}
\value{
An object of class \code{"summary.zplot_brma"} containing the
estimates table.
}
\description{
Creates summary tables for zplot estimates including
EDR, Soric FDR, and estimated missing studies.
}
\details{
The summary includes:
\describe{
\item{EDR}{Expected Discovery Rate - average power of significant studies}
\item{Soric FDR}{Expected proportion of false discoveries among significant
results, computed from EDR following Soric (1989)}
\item{Missing N}{Estimated number of studies suppressed by publication bias,
computed from the selection model weights}
}

The footer reports the Observed Discovery Rate (ODR) with 95\\% CI for
comparison with the model-estimated EDR.
}
\seealso{
\code{\link[=as_zplot.brma]{as_zplot.brma()}}, \code{\link[=plot.zplot_brma]{plot.zplot_brma()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/datasets.R
\docType{data}
\name{Wang2025}
\alias{Wang2025}
\title{70 effect sizes from a meta-analysis of ChatGPT's impact on student learning
by \insertCite{wang2025effect;textual}{RoBMA}}
\format{
A data.frame with 12 columns and 70 observations:
\describe{
\item{\code{Learning_effect}}{Learning-effect outcome category.}
\item{\code{Author_year}}{Author-year study label.}
\item{\code{N_EG}}{Experimental-group sample size.}
\item{\code{N_CG}}{Control-group sample size.}
\item{\code{g}}{Hedges' g effect size.}
\item{\code{Grade_level}}{Grade level.}
\item{\code{Type_of_course}}{Course type.}
\item{\code{Duration}}{Study duration.}
\item{\code{Learning_model}}{Learning model.}
\item{\code{Role_of_ChatGPT}}{Role of ChatGPT in the intervention.}
\item{\code{Area_of_ChatGPT_application}}{Application area.}
\item{\code{se}}{Standard error of g.}
}
}
\usage{
Wang2025
}
\description{
The data set contains Hedges' g effect sizes, standard errors,
sample sizes for experimental and control groups, and various study characteristics
including grade level, type of course, duration, learning model, role of ChatGPT,
and area of ChatGPT application. The meta-analysis examined the effect of ChatGPT
on students' learning performance, learning perception, and higher-order thinking
\insertCite{wang2025effect}{RoBMA}.
}
\references{
\insertAllCited{}
}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/datasets.R
\docType{data}
\name{Lui2015}
\alias{Lui2015}
\title{18 studies of a relationship between acculturation mismatch and
intergenerational cultural conflict collected by
\insertCite{lui2015intergenerational;textual}{RoBMA}}
\format{
A data.frame with 3 columns and 18 observations:
\describe{
\item{\code{r}}{Correlation coefficient.}
\item{\code{n}}{Sample size.}
\item{\code{study}}{Study label.}
}
}
\usage{
Lui2015
}
\description{
The data set contains correlation coefficients r,
sample sizes n, and labels for each study assessing the
relationship between acculturation mismatch (that is the result of the contrast
between the collectivist cultures of Asian and Latin immigrant groups
and the individualist culture in the United States) and intergenerational cultural
conflict \insertCite{lui2015intergenerational}{RoBMA} which was used as an
example in \insertCite{bartos2020adjusting;textual}{RoBMA}.
}
\references{
\insertAllCited{}
}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/summary.R
\name{summary.brma}
\alias{summary.brma}
\alias{print.summary.brma}
\alias{print.brma}
\title{Summarize brma Object}
\usage{
\method{summary}{brma}(
  object,
  probs = c(0.025, 0.5, 0.975),
  include_mcmc_diagnostics = TRUE,
  standardized_coefficients = FALSE,
  conditional = FALSE,
  logBF = FALSE,
  BF01 = FALSE,
  ...
)

\method{print}{summary.brma}(x, ...)

\method{print}{brma}(x, ...)
}
\arguments{
\item{object}{a fitted brma object}

\item{probs}{quantiles of the posterior samples to be displayed.
Defaults to \code{c(.025, .50, .975)}}

\item{include_mcmc_diagnostics}{whether to include MCMC diagnostics in the output.
Defaults to \code{TRUE}.}

\item{standardized_coefficients}{whether to show standardized meta-regression coefficients.
Defaults to \code{FALSE}. When set to \code{TRUE}, standardized meta-regression
coefficients are returned for the intercept and continuous predictors. These coefficients
correspond to the standardized scale on which prior distributions are specified by default
(i.e., \code{standardize_continuous_predictors = TRUE}).}

\item{conditional}{whether to include conditional estimates for RoBMA
product-space objects. Defaults to \code{FALSE}.}

\item{logBF}{whether to show inclusion Bayes factors on the log scale.
Defaults to \code{FALSE}.}

\item{BF01}{whether to show inverse inclusion Bayes factors. Defaults to
\code{FALSE}.}

\item{...}{additional arguments}

\item{x}{a \code{summary.brma} or fitted \code{brma} object.}
}
\value{
A list of class \code{summary.brma} with model name, optional RoBMA
inclusion tables, common estimates, moderator estimates, scale estimates,
publication-bias estimates, and optional conditional estimates. The printed
form displays the non-empty tables.
}
\description{
\code{summary.brma} creates summary tables for a
brma object. For RoBMA objects, inclusion summaries are printed before
parameter estimates.
}
\examples{
\dontrun{
if (requireNamespace("metadat", quietly = TRUE)) {
  data(dat.lehmann2018, package = "metadat")

  fit <- bPET(
    yi      = yi,
    vi      = vi,
    data    = dat.lehmann2018,
    measure = "SMD",
    seed    = 1,
    silent  = TRUE
  )

  summary(fit)
}
}


}
\seealso{
\code{\link[=brma]{brma()}}, \code{\link[=brma.glmm]{brma.glmm()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/zplot.R
\name{plot.zplot_brma}
\alias{plot.zplot_brma}
\title{Plot Zplot Results}
\usage{
\method{plot}{zplot_brma}(
  x,
  plot_type = "base",
  probs = c(0.025, 0.975),
  max_samples = 10000,
  plot_fit = TRUE,
  plot_extrapolation = TRUE,
  plot_ci = TRUE,
  plot_thresholds = TRUE,
  from = -6,
  to = 6,
  by.hist = 0.5,
  length.out.hist = NULL,
  by.lines = 0.05,
  length.out.lines = NULL,
  dots_hist = NULL,
  dots_fit = NULL,
  dots_extrapolation = NULL,
  dots_thresholds = NULL,
  ...
)
}
\arguments{
\item{x}{a zplot_brma object.}

\item{plot_type}{graphics system: \code{"base"} or \code{"ggplot"}.
Defaults to \code{"base"}.}

\item{probs}{quantiles for credible intervals. Defaults to \code{c(.025, .975)}.}

\item{max_samples}{maximum posterior samples for density estimation.
Defaults to 10000. Use \code{Inf} to use all posterior samples.}

\item{plot_fit}{whether to show fitted density (with bias adjustments).
Defaults to \code{TRUE}.}

\item{plot_extrapolation}{whether to show extrapolated density (bias removed).
Defaults to \code{TRUE}.}

\item{plot_ci}{whether to show credible interval bands. Defaults to \code{TRUE}.}

\item{plot_thresholds}{whether to show significance threshold lines.
Defaults to \code{TRUE}.}

\item{from, to}{z-value range for plotting. Defaults to \code{-6} and \code{6}.}

\item{by.hist}{bin width for histogram. Defaults to 0.5.}

\item{length.out.hist}{number of histogram bins (alternative to \code{by.hist}).}

\item{by.lines}{step size for density lines. Defaults to 0.05.}

\item{length.out.lines}{number of density points (alternative to \code{by.lines}).}

\item{dots_hist}{graphical parameters for histogram (list).}

\item{dots_fit}{graphical parameters for fit lines (list).}

\item{dots_extrapolation}{graphical parameters for extrapolation lines (list).}

\item{dots_thresholds}{graphical parameters for threshold lines (list).}

\item{...}{additional graphical parameters passed to components.}
}
\value{
\code{NULL} invisibly for base graphics, or a ggplot2 object.
}
\description{
Plots a zplot visualization showing the histogram of
observed z-statistics overlaid with model-implied densities.
}
\details{
The plot displays two density curves:
\describe{
\item{Fit (black)}{Model-implied density including publication bias adjustments.
This represents the expected distribution of z-statistics given the estimated
selection process.}
\item{Extrapolation (blue)}{Bias-corrected curve representing the hypothetical
distribution without selective reporting, scaled by the expected suppressed
studies under selection models. This curve is not normalized to integrate
to one when selection implies missing studies.}
}
}
\seealso{
\code{\link[=hist.zplot_brma]{hist.zplot_brma()}}, \code{\link[=lines.zplot_brma]{lines.zplot_brma()}}, \code{\link[=summary.zplot_brma]{summary.zplot_brma()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/brma_samples.as_draws.R
\name{as_draws.brma_samples}
\alias{as_draws.brma_samples}
\alias{as_draws_array.brma_samples}
\alias{as_draws_df.brma_samples}
\alias{as_draws_list.brma_samples}
\alias{as_draws_matrix.brma_samples}
\alias{as_draws_rvars.brma_samples}
\title{Convert brma_samples to posterior Draws Formats}
\usage{
\method{as_draws}{brma_samples}(x, ...)

\method{as_draws_array}{brma_samples}(x, ...)

\method{as_draws_df}{brma_samples}(x, ...)

\method{as_draws_list}{brma_samples}(x, ...)

\method{as_draws_matrix}{brma_samples}(x, ...)

\method{as_draws_rvars}{brma_samples}(x, ...)
}
\arguments{
\item{x}{a \code{brma_samples} object}

\item{...}{additional arguments passed to the corresponding
\pkg{posterior} function}
}
\value{
An object of the corresponding \pkg{posterior} draws class.
}
\description{
Provides an interface to the \pkg{posterior} package
for \code{brma_samples} objects. These functions convert the posterior
samples to various draws formats supported by the \pkg{posterior} package.
}
\details{
The conversion reconstructs the MCMC chain structure from the stored
\code{nchains} and \code{niter} attributes. The samples are assumed
to be ordered with chains concatenated (i.e., all iterations from chain 1,
then all from chain 2, etc.). Conditional RoBMA samples are intentionally
stored as one flattened chain because conditioning subsets posterior rows
across chains.
}
\seealso{
\code{\link[posterior]{draws}}, \code{\link{as_draws.brma}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utilities.R
\name{RoBMA_control}
\alias{RoBMA_control}
\alias{set_autofit_control}
\alias{set_convergence_checks}
\title{Control MCMC fitting process}
\usage{
set_autofit_control(
  max_Rhat = 1.05,
  min_ESS = 500,
  max_error = NULL,
  max_SD_error = NULL,
  max_time = list(time = 60, unit = "mins"),
  sample_extend = 1000,
  restarts = 10,
  max_extend = 10
)

set_convergence_checks(
  max_Rhat = 1.05,
  min_ESS = 500,
  max_error = NULL,
  max_SD_error = NULL
)
}
\arguments{
\item{max_Rhat}{maximum value of the R-hat diagnostic.
Defaults to \code{1.05}.}

\item{min_ESS}{minimum estimated sample size.
Defaults to \code{500}.}

\item{max_error}{maximum value of the MCMC error.
Defaults to \code{NULL}. Be aware that PEESE publication bias
adjustment can have estimates on different scale than the rest of
the output, resulting in relatively large max MCMC error.}

\item{max_SD_error}{maximum value of the proportion of MCMC error
of the estimated SD of the parameter.
Defaults to \code{NULL}.}

\item{max_time}{list with the time and unit specifying the maximum
autofitting process per model. Passed to \link[base]{difftime} function
(possible units are \code{"secs"}, \code{"mins"}, \code{"hours"},
\code{"days"}, and \code{"weeks"}).
Defaults to \code{list(time = 60, unit = "mins")}.}

\item{sample_extend}{number of samples to extend the fitting process if
the criteria are not satisfied.
Defaults to \code{1000}.}

\item{restarts}{number of times new initial values should be generated in case a
model fails to initialize. Defaults to \code{10}.}

\item{max_extend}{number of times after which the automatic fitting function is stopped.
Defaults to \code{10}.}
}
\value{
\code{set_autofit_control} returns a list of autofit control settings
including \code{max_Rhat}, \code{min_ESS}, \code{max_error},
\code{max_SD_error}, \code{max_time}, \code{sample_extend},
\code{restarts}, \code{max_extend}, and delegated
\code{check_indicators}.
\code{set_convergence_checks} returns a list with convergence thresholds
\code{max_Rhat}, \code{min_ESS}, \code{max_error}, \code{max_SD_error},
and delegated \code{check_indicators}.
}
\description{
\code{set_autofit_control} and \code{set_convergence_checks} control settings for the
autofit process of the MCMC JAGS sampler and specify termination criteria and convergence checks.
}
\details{
\code{set_autofit_control} controls the automatic model fitting process, determining when to
stop iterating based on convergence criteria. \code{set_convergence_checks} sets thresholds
for determining whether a model has converged adequately.

The autofit control manages computational resources by setting maximum time limits and
determining how many additional samples to draw if convergence criteria are not met.
The convergence checks determine the quality standards that fitted models must meet.
Autofit thresholds \code{max_Rhat}, \code{min_ESS}, \code{max_error},
\code{max_SD_error}, \code{max_time}, \code{restarts}, and
\code{max_extend} can be set to \code{NULL}; \code{sample_extend} must be a
positive integer.
Thresholds can be disabled with \code{NULL}; otherwise \code{max_Rhat} must be
at least 1, \code{min_ESS} and \code{max_error} must be nonnegative, and
\code{max_SD_error} must be between 0 and 1.
}
\examples{
# Set custom autofit control with shorter time limit
autofit_ctrl <- set_autofit_control(
  max_time = list(time = 30, unit = "mins"),
  sample_extend = 2000
)

# Set custom convergence checks with stricter criteria
conv_checks <- set_convergence_checks(
  max_Rhat = 1.01,
  min_ESS = 1000
)

\dontrun{
if (requireNamespace("metadat", quietly = TRUE)) {
  data(dat.lehmann2018, package = "metadat")

  fit <- RoBMA(
    yi                 = yi,
    vi                 = vi,
    data               = dat.lehmann2018,
    measure            = "SMD",
    autofit_control    = autofit_ctrl,
    convergence_checks = conv_checks
  )
}
}

}
\seealso{
\code{\link[=RoBMA]{RoBMA()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/input-object.R
\name{fitting_specification}
\alias{fitting_specification}
\title{Fitting specification}
\arguments{
\item{sample}{numeric. Number of MCMC samples to save. Defaults to \code{5000}.}

\item{burnin}{numeric. Number of burn-in iterations. Defaults to \code{2000}.}

\item{adapt}{numeric. Number of adaptation iterations. Defaults to \code{500}.}

\item{chains}{numeric. Number of MCMC chains. Defaults to \code{3}.}

\item{thin}{numeric. Thinning interval. Defaults to \code{1}.}

\item{parallel}{logical. Whether to run MCMC chains in parallel. Defaults to \code{FALSE}.}

\item{autofit}{logical. Whether to automatically extend the MCMC chains if convergence is not met.
Defaults to \code{FALSE}.}

\item{autofit_control}{list of autofit control settings. See \code{\link[=set_autofit_control]{set_autofit_control()}} for details.}

\item{convergence_checks}{list of convergence check settings. See \code{\link[=set_convergence_checks]{set_convergence_checks()}} for details.}

\item{seed}{numeric. Random seed for reproducibility. Defaults to \code{NULL}.}

\item{silent}{logical. Whether to suppress output. Constructors with no
explicit default use \code{RoBMA.get_option("silent")} when \code{silent} is omitted.
Model-averaging wrappers default to \code{TRUE} unless explicitly changed.}

\item{...}{additional advanced arguments. Fitting functions reject unused
arguments; currently recognized internal arguments include \code{only_data},
\code{only_priors}, \code{is_JASP}, and \code{is_JASP_prefix}.}
}
\description{
The \code{brma} family of functions uses the following arguments to specify the
MCMC sampling and fitting settings.
}
\seealso{
\code{\link{brma}}, \code{\link{set_autofit_control}}, \code{\link{set_convergence_checks}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/summary_heterogeneity.R
\name{summary_heterogeneity}
\alias{summary_heterogeneity}
\title{Summary of Heterogeneity}
\usage{
summary_heterogeneity(object, ...)
}
\arguments{
\item{object}{a fitted model object}

\item{...}{additional arguments passed to methods}
}
\value{
Method-specific return value containing heterogeneity estimates.
}
\description{
Computes the absolute heterogeneity (tau, tau^2) and
relative measures of heterogeneity (I^2, H^2) for a fitted model.
}
\seealso{
\code{\link[=pooled_heterogeneity]{pooled_heterogeneity()}}, \code{\link[=summary.brma]{summary.brma()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/hatvalues.R
\name{hatvalues.brma}
\alias{hatvalues.brma}
\title{Hat Values for brma Objects}
\usage{
\method{hatvalues}{brma}(model, ...)
}
\arguments{
\item{model}{a fitted brma object}

\item{...}{additional arguments (currently ignored)}
}
\value{
A numeric vector of length \code{K}, one posterior mean leverage per
observation.
}
\description{
Computes hat values (leverages) from a fitted brma object.
Returns posterior mean leverages, one for each observation.
}
\details{
Hat values (leverages) measure the influence of each observation on the
fitted values. In a Bayesian meta-analysis, the random effects variance \eqn{\tau^2}
is uncertain, so the hat matrix depends on the posterior samples of \eqn{\tau^2}.

This function computes the diagonal elements of the hat matrix:
\deqn{h_i = (X (X^T W X)^{-1} X^T W)_{ii}}
where \eqn{W} is the weight matrix inverse to the marginal variance matrix.

The hat matrix is computed for each posterior draw and then averaged over
draws, matching the vector output shape used by \code{metafor}.

This method is available only for normal outcome models without
weightfunction selection.
}
\examples{
\dontrun{
if (requireNamespace("metadat", quietly = TRUE) &&
    requireNamespace("metafor", quietly = TRUE)) {
  data(dat.bcg, package = "metadat")
  dat <- metafor::escalc(
    measure = "RR",
    ai      = tpos,
    bi      = tneg,
    ci      = cpos,
    di      = cneg,
    data    = dat.bcg
  )

  fit <- brma(yi = yi, vi = vi, data = dat, measure = "RR")
  hatvalues(fit)
}
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/datasets.R
\docType{data}
\name{ManyLabs16}
\alias{ManyLabs16}
\title{55 effect sizes from Many Labs 2 replication studies of
\insertCite{tversky1981framing;textual}{RoBMA} framing effects}
\format{
A data.frame with 2 columns and 55 observations:
\describe{
\item{\code{y}}{Standardized mean difference.}
\item{\code{se}}{Standard error of y.}
}
}
\usage{
ManyLabs16
}
\description{
The data set contains standardized mean differences (y) and standard errors (se)
from 55 replication studies of the framing effect originally described by
\insertCite{tversky1981framing;textual}{RoBMA}. These studies were part of the
Many Labs 2 project examining variation in replicability across samples and settings
\insertCite{klein2018many;textual}{RoBMA}.
}
\references{
\insertAllCited{}
}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/input-priors.R
\name{prior_specification}
\alias{prior_specification}
\title{Prior specification}
\arguments{
\item{prior_effect}{prior distribution for the effect size (\eqn{\mu}) parameter
(the intercept). If omitted, a default prior is constructed. In single-model
functions, explicit \code{NULL} or \code{FALSE} sets a spike at zero.}

\item{prior_heterogeneity}{prior distribution for the heterogeneity (\eqn{\tau})
parameter. If omitted, a default prior is constructed. In single-model
functions, explicit \code{NULL} or \code{FALSE} sets a spike at zero.}

\item{prior_mods}{prior distribution for the moderators (\eqn{\beta}) parameters.
A single prior applies to all terms; a named list can specify term-specific
priors. If omitted or \code{NULL}, default priors are used.}

\item{prior_scale}{prior distribution for the scale (\eqn{\delta}) parameters.
A single prior applies to all terms; a named list can specify term-specific
priors. If omitted or \code{NULL}, default priors are used.}

\item{prior_heterogeneity_allocation}{prior distribution for the fraction of
heterogeneity allocated to the cluster-level component in multilevel models
(\eqn{\rho}). If omitted or \code{NULL}, defaults to \code{Beta(1, 1)}.}

\item{prior_baserate}{prior distribution for the estimate-specific midpoint
base-rate probability in binomial GLMM models. If omitted or \code{NULL}, defaults
to independent \code{Beta(1, 1)} priors.}

\item{prior_lograte}{prior distribution for the estimate-specific midpoint
log-rate in Poisson GLMM models. If omitted or \code{NULL}, a data-based
unit-information normal prior is used independently for each estimate.}

\item{prior_unit_information_sd}{numeric. The unit information standard deviation (\eqn{\sigma_{unit}}).
Cannot be used together with \code{prior_informed_field}.}

\item{rescale_priors}{numeric. A scaling factor for supported prior distributions.
Point and none priors are unchanged. For constructors with publication-bias
prior distributions, \code{rescale_priors} does not rescale them except for the
default PEESE prior's UISD adjustment. Defaults to 1.}

\item{standardize_continuous_predictors}{logical. Whether to standardize continuous predictors.
Defaults to \code{TRUE}.}

\item{set_contrast_factor_predictors}{character. How to set contrast for factor predictors.
Defaults are constructor-specific and shown in each function usage; single-model
constructors use \code{"treatment"}, while model-averaging constructors use \code{"meandif"}.}

\item{prior_informed_field}{character. The field of the informed prior distributions.
Omit to use the standard default prior specification; explicit \code{NULL} is invalid.}

\item{prior_informed_subfield}{character. The subfield of the informed prior distributions.
Omit to use the field-specific default, such as \code{"Cochrane"} for
\code{prior_informed_field = "medicine"}; explicit \code{NULL} is invalid.}
}
\description{
Prior distributions can be supplied for all model parameters. When omitted,
the fitting functions construct defaults from the effect-size measure, sample
sizes, or informed-prior settings when available.

This typically includes the pooled effect \eqn{\mu} and between-study heterogeneity
\eqn{\tau}. In the case of meta-regression, the pooled effect \eqn{\mu} corresponds to
the intercept, and additional prior distributions for the regression coefficients
are required. In the case of a location-scale model, the between-study heterogeneity
corresponds to the intercept of the scale regression, and additional prior
distributions for the scale regression coefficients are required.
}
\details{
There are several ways to specify the prior distributions: \enumerate{
\item{via a standardized effect size \code{measure} with known unit information standard deviation,}
\item{by estimating unit information standard deviation using sample sizes \code{ni},}
\item{by manually setting \code{prior_unit_information_sd},}
\item{by specifying informed empirical prior distributions via \code{prior_informed_field}
and \code{prior_informed_subfield},}
\item{or via fully custom specification using the \code{prior_effect}, \code{prior_heterogeneity},
\code{prior_mods}, \code{prior_scale}, and \code{prior_heterogeneity_allocation} arguments.}
}
In all cases, the prior behavior can be further modified by the \code{rescale_priors},
\code{standardize_continuous_predictors}, and \code{set_contrast_factor_predictors} arguments.
\subsection{(1) Specifying prior distributions via standardized effect size measures with known}{

unit information standard deviation
This is the easiest way to specify prior distributions. The width of prior
distributions is based on a fraction of the known unit information standard deviation (\code{UISD})
\insertCite{rover2021weakly}{RoBMA}. The default prior distributions for the parameters
are set as follows:\tabular{ll}{
effect size:               \tab Normal(0, \eqn{\frac{1}{2}} \code{UISD})  \cr
heterogeneity:             \tab Normal+(0, \eqn{\frac{1}{4}} \code{UISD}) \cr
effect moderation:         \tab Normal(0, \eqn{\frac{1}{4}} \code{UISD})  \cr
heterogeneity moderation:  \tab Normal(0, \eqn{\frac{1}{2}})
}
The heterogeneity moderation parameters are multiplicative, as such they are independent of \code{UISD}.

The default fraction of the \code{UISD} can be changed via \code{RoBMA.options()} using one of
the following arguments: \code{"default_UISD.effect"}, \code{"default_UISD.heterogeneity"},
\code{"default_UISD.mods"}, \code{"default_UISD.scale"}.

The known \code{UISD} for standardized effect size measures (\code{measure}) are set as follows:
\tabular{ll}{
\code{"SMD"}:   \tab \eqn{\sqrt{2}} \cr
\code{"ZCOR"}:  \tab \eqn{1}        \cr
\code{"RR"}:    \tab \eqn{\sqrt{4}} \cr
\code{"OR"}:    \tab \eqn{\sqrt{4}} \cr
\code{"HR"}:    \tab \eqn{\sqrt{4}} \cr
\code{"IRR"}:   \tab \eqn{\sqrt{4}}
}
See Chapter 2.4 in \insertCite{spiegelhalter2004bayesian;textual}{RoBMA} and
Chapter 1 in \insertCite{grieve2022hybrid;textual}{RoBMA}.
}

\subsection{(2) Estimating unit information standard deviation using sample sizes}{

When effect sizes are on a non-standardized scale (\code{measure = "GEN"}) or use a
standardized effect size without known \code{UISD}, the \code{UISD} can be estimated from
sample sizes (\code{ni}) and standard errors (\code{sei}) following Equation 6 in
\insertCite{rover2021weakly;textual}{RoBMA}. The estimated \code{UISD} is then used to
scale the default prior distributions as described in section (1).

Note that the known \code{UISD} for standardized effect size measures (section (1)) is
used if available, even when \code{ni} is provided.
}

\subsection{(3) Manually setting unit information standard deviation}{

Alternatively, the \code{UISD} can be specified directly via the \code{prior_unit_information_sd}
argument. This is useful when the appropriate scale for prior distributions is known
a priori or when multiple analyses are to be performed on subsets of the same data
(re-estimating \code{UISD} on different subsets of the data can lead to slightly different
prior distributions for different subsets; see \code{\link[=estimate_unit_information_sd]{estimate_unit_information_sd()}}).
The specified \code{prior_unit_information_sd} is then used to scale the default prior
distributions as described in section (1).

Note that the manually specified \code{prior_unit_information_sd} takes precedence over
the estimated \code{UISD} from \code{ni} (section (2)) and the known \code{UISD} from \code{measure}
(section (1)). It cannot be combined with \code{prior_informed_field}.
}

\subsection{(4) Specifying informed empirical prior distributions}{

Informed prior distributions can be specified via the \code{prior_informed_field} and
\code{prior_informed_subfield} arguments. Currently, only \code{prior_informed_field = "medicine"}
with subfields defined in \code{BayesTools::prior_informed_medicine_names} is supported,
which uses empirically derived prior distributions from medical meta-analyses as described in
\insertCite{bartos2021bayesian;textual}{RoBMA} and
\insertCite{bartos2023empirical;textual}{RoBMA}.

When \code{prior_informed_field = "medicine"}, the default \code{prior_informed_subfield} is
\code{"Cochrane"} (i.e., using the whole CDSR database as a reference). The informed
prior distributions are available for the following effect size measures:
\tabular{ll}{
\code{"SMD"}:  \tab standardized mean difference \cr
\code{"OR"}:   \tab log odds ratio               \cr
\code{"RR"}:   \tab log risk ratio               \cr
\code{"RD"}:   \tab risk difference              \cr
\code{"HR"}:   \tab log hazard ratio
}

Note that informed prior distributions are only available for the effect size (\eqn{\mu})
and heterogeneity (\eqn{\tau}) parameters. For effect moderation (\code{prior_mods}), the
informed effect prior is scaled by a factor of \code{RoBMA.get_option("default_informed_priors.mods")}.
For heterogeneity moderation (\code{prior_scale}), a normal prior with standard deviation
specified by \code{RoBMA.get_option("default_informed_priors.scale")} is used.
}

\subsection{(5) Fully custom prior distributions}{

Prior distributions can be fully customized by directly specifying the \code{prior_effect},
\code{prior_heterogeneity}, \code{prior_mods}, \code{prior_scale}, and \code{prior_heterogeneity_allocation}
arguments. These should be prior distribution objects created via \code{BayesTools::prior()}
or related functions (e.g., \code{prior_factor()}).
}

\subsection{Rescaling prior distributions}{

The \code{rescale_priors} argument allows rescaling supported prior distributions by a
multiplicative factor. For example, \code{rescale_priors = 2} doubles the standard
deviations/scales of normal, Cauchy, t, and inverse-gamma prior distributions,
making them more diffuse. Point and none priors are unchanged. For publication-bias
prior distributions, see \code{\link{publication_bias_prior_specification}}.
}
}
\references{
\insertAllCited{}
}
\seealso{
\code{\link{publication_bias_prior_specification}},
\code{\link[BayesTools]{prior}}, \code{\link{RoBMA.options}},
\code{\link{brma}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/wrappers.R
\name{pooled_heterogeneity.brma}
\alias{pooled_heterogeneity.brma}
\title{Pooled Heterogeneity for brma Objects}
\usage{
\method{pooled_heterogeneity}{brma}(object, probs = c(0.025, 0.975), conditional = FALSE, ...)
}
\arguments{
\item{object}{a fitted brma object}

\item{probs}{quantiles of the posterior distribution to be displayed.
Defaults to \code{c(.025, .975)} for 95\% credible intervals.}

\item{conditional}{whether to return the pooled heterogeneity conditional on
the heterogeneity component for RoBMA product-space objects. Defaults to
\code{FALSE}.}

\item{...}{additional arguments passed to \code{\link{predict.brma}}; wrapper
arguments such as \code{newdata}, \code{type}, and \code{quiet} are fixed.}
}
\value{
A \code{brma_samples} object containing posterior samples. When printed,
displays a summary table. Use \code{summary()} to obtain the summary table directly.
The samples can be converted to \pkg{posterior} draws formats using \code{as_draws()}.
}
\description{
Computes the pooled (aggregated) heterogeneity estimate (tau)
from a fitted brma object by averaging across the scale model matrix.
}
\details{
This function is a convenience wrapper around \code{predict.brma(...,
type = "terms.scale", newdata = TRUE)}.

For location-scale models (with scale regression), the pooled heterogeneity
averages tau across the scale model matrix proportionately to the levels
observed in the data.

For models without scale regression, this returns the single tau parameter.

For multilevel (3-level) models, the returned tau is the total heterogeneity:
\code{tau = sqrt(tau_within^2 + tau_between^2)}.
}
\examples{
\dontrun{
if (requireNamespace("metadat", quietly = TRUE)) {
  data(dat.lehmann2018, package = "metadat")
  fit <- brma(
    yi      = yi,
    vi      = vi,
    data    = dat.lehmann2018,
    measure = "SMD",
    seed    = 1,
    silent  = TRUE
  )

  pooled_heterogeneity(fit)
}
}

}
\seealso{
\code{\link[=predict.brma]{predict.brma()}}, \code{\link[=pooled_effect]{pooled_effect()}}, \code{\link[=blup]{blup()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/priors.R
\name{prior_factor}
\alias{prior_factor}
\title{Factor Prior}
\usage{
prior_factor(
  distribution,
  parameters,
  truncation = list(lower = -Inf, upper = Inf),
  prior_weights = 1,
  contrast = "meandif"
)
}
\arguments{
\item{distribution}{character. Prior distribution name.}

\item{parameters}{list. Distribution parameters.}

\item{truncation}{list with \code{lower} and \code{upper} truncation bounds.}

\item{prior_weights}{numeric prior model weight.}

\item{contrast}{character. Contrast coding used for factor levels. Common
RoBMA model options are \code{"treatment"}, \code{"meandif"}, and
\code{"orthonormal"}; the standalone helper also accepts BayesTools contrast
aliases such as \code{"independent"}.}
}
\value{
An object inheriting from \code{prior}, \code{prior.factor}, and a
contrast-specific marker class.
}
\description{
Create priors on factor contrasts.
}
\details{
Mean-difference and orthonormal contrasts require vector or
multivariate priors. Treatment/dummy and independent contrasts use univariate
simple priors per contrast coefficient.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/residuals.R
\name{rstandard.brma}
\alias{rstandard.brma}
\title{Internally Standardized Residuals for brma Objects}
\usage{
\method{rstandard}{brma}(model, unit = "estimate", conditioning_depth = "marginal", ...)
}
\arguments{
\item{model}{a fitted brma object.}

\item{unit}{output unit. Only \code{"estimate"} is implemented currently.}

\item{conditioning_depth}{conditioning depth. Options are:
\itemize{
\item \code{"marginal"} (default): Residuals from fixed effects predictions
(\eqn{observed - X\beta}).
\item \code{"cluster"}: Residuals from cluster-level predictions (observed -
(\eqn{X\beta + gamma})). Only available for multilevel (3-level) models.
\item \code{"estimate"}: Residuals from BLUPs,
i.e., deviations of the observed effect sizes from the best linear
unbiased predictions of the estimate-specific true effects (observed - theta).
}}

\item{...}{additional arguments (currently ignored)}
}
\value{
A data frame with columns:
\itemize{
\item \code{resid}: Raw residuals
\item \code{se}: Standard errors of the residuals
\item \code{z}: Standardized residuals
}
}
\description{
Computes internally standardized residuals from a fitted brma
object using the hat matrix. Returns a data frame with raw residuals,
standard errors, and standardized residuals (z-values). Available for normal
outcome models only.
}
\details{
This function returns a data frame with three columns matching the output
of \code{metafor::rstandard}:
\itemize{
\item \code{resid}: Raw residuals (observed - fitted values)
\item \code{se}: Standard errors of the residuals
\item \code{z}: Standardized residuals (resid / se)
}

Internally standardized residuals divide the observed residuals by their
corresponding standard errors computed using the hat matrix. For a correctly
specified model, these residuals should approximately follow a standard
normal distribution.

This function is only available for normal outcome models without selection
(weightfunction) bias adjustment. For other model types, use
\code{\link{rstudent.brma}} which uses LOO-PIT.
}
\examples{
\dontrun{
if (requireNamespace("metadat", quietly = TRUE)) {
  data(dat.lehmann2018, package = "metadat")
  fit <- brma(yi = yi, vi = vi, data = dat.lehmann2018, measure = "SMD")

  # marginal internally standardized residuals (default)
  rstandard(fit)

  # estimate-level (BLUP-based) residuals
  rstandard(fit, conditioning_depth = "estimate")
}
}

}
\seealso{
\code{\link[=rstudent.brma]{rstudent.brma()}}, \code{\link[=residuals.brma]{residuals.brma()}}, \code{\link[=blup.brma]{blup.brma()}}, \code{\link[=predict.brma]{predict.brma()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/datasets.R
\docType{data}
\name{Bem2011}
\alias{Bem2011}
\title{9 experimental studies from
\insertCite{bem2011feeling;textual}{RoBMA} as described in
\insertCite{bem2011must;textual}{RoBMA}}
\format{
A data.frame with 3 columns and 9 observations:
\describe{
\item{\code{d}}{Cohen's d effect size.}
\item{\code{se}}{Standard error of d.}
\item{\code{study}}{Study label.}
}
}
\usage{
Bem2011
}
\description{
The data set contains Cohen's d effect sizes, standard errors,
and labels for 9 experimental studies of precognition from the infamous
\insertCite{bem2011feeling;textual}{RoBMA} as analyzed in his later meta-analysis
\insertCite{bem2011must}{RoBMA}.
}
\references{
\insertAllCited{}
}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/priors.R
\name{prior}
\alias{prior}
\title{Prior Distribution}
\usage{
prior(
  distribution,
  parameters,
  truncation = list(lower = -Inf, upper = Inf),
  prior_weights = 1
)
}
\arguments{
\item{distribution}{character. Prior distribution name.}

\item{parameters}{list. Distribution parameters.}

\item{truncation}{list with \code{lower} and \code{upper} truncation bounds.}

\item{prior_weights}{numeric prior model weight.}
}
\value{
An object inheriting from \code{prior}.
}
\description{
Create prior distribution objects used by RoBMA and brma
fitting functions.
}
\details{
This is RoBMA's re-export of \code{BayesTools::prior()}; supported
distribution names and parameter lists follow BayesTools.
}
\examples{
prior("normal", list(mean = 0, sd = 0.5))
prior(
  "normal",
  list(mean = 0, sd = 0.5),
  truncation = list(lower = 0, upper = Inf)
)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/marginal_means.R
\name{plot.marginal_means.brma}
\alias{plot.marginal_means.brma}
\title{Plot Estimated Marginal Means}
\usage{
\method{plot}{marginal_means.brma}(
  x,
  parameter,
  type = NULL,
  prior = FALSE,
  plot_type = "base",
  dots_prior = NULL,
  output_measure = NULL,
  transform = NULL,
  ...
)
}
\arguments{
\item{x}{a \code{marginal_means.brma} object.}

\item{parameter}{moderator term to plot. Use the original term name, for
example \code{"measure"}, \code{"intercept"} for the intercept when
available, \code{"mu"} as an intercept alias, or the internal parameter name,
for example \code{"mu_measure"}.}

\item{type}{for RoBMA product-space objects, whether to plot model-averaged
(\code{"averaged"}) or conditional (\code{"conditional"}) marginal means.
Defaults to \code{"averaged"} and is available only for RoBMA marginal
means.}

\item{prior}{whether the marginal prior distribution should be added to the
plot. Defaults to \code{FALSE}.}

\item{plot_type}{whether to use base R graphics (\code{"base"}) or ggplot2
(\code{"ggplot"}). Defaults to \code{"base"}.}

\item{dots_prior}{list of additional graphical arguments passed to the prior
plotting function.}

\item{output_measure}{effect-size measure for location/effect predictions.
Defaults to the fitted measure. Supported conversions are among \code{"SMD"},
\code{"COR"}, \code{"ZCOR"}, and \code{"OR"}; \code{"RR"}, \code{"HR"},
\code{"IRR"}, \code{"RD"}, and \code{"GEN"} can only be returned on their
fitted measure. Use \code{transform = "EXP"} for ratio-scale output from
log-scale measures.}

\item{transform}{optional display transformation. Currently \code{"EXP"}
exponentiates log-scale measures \code{"OR"}, \code{"RR"}, \code{"HR"},
and \code{"IRR"}.}

\item{...}{additional graphical arguments passed to
\code{BayesTools::plot_marginal()}.}
}
\value{
\code{NULL} invisibly if \code{plot_type = "base"} or a ggplot object
if \code{plot_type = "ggplot"}.
}
\description{
Plots estimated marginal means stored in a
\code{marginal_means.brma} object using \code{BayesTools::plot_marginal()}.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/datasets.R
\docType{data}
\name{Andrews2021}
\alias{Andrews2021}
\title{39 study rows on household chaos and child executive functions
from a meta-analysis by
\insertCite{andrews2021examining;textual}{RoBMA}}
\format{
A data.frame with 15 columns and 39 observations:
\describe{
\item{\code{study_id}}{Source row identifier.}
\item{\code{slab}}{Study label.}
\item{\code{ri}}{Raw correlation coefficient.}
\item{\code{vi}}{Sampling variance of the raw correlation coefficient.}
\item{\code{sei}}{Sampling standard error of the raw correlation coefficient.}
\item{\code{ni}}{Sample size.}
\item{\code{year}}{Publication year.}
\item{\code{percent_female}}{Percentage of female children in the sample.}
\item{\code{age}}{Mean age of the children in years.}
\item{\code{assessment_interval_months}}{Time between household-chaos and executive-function assessments in months.}
\item{\code{dissertation}}{Whether the study was a dissertation.}
\item{\code{percent_minority}}{Percentage of minority participants in the sample.}
\item{\code{percent_low_parental_education}}{Percentage of parents with high school, GED, or lower education.}
\item{\code{hc_dimension}}{Household-chaos dimension using the source coding.}
\item{\code{measure}}{Executive-function assessment type, direct or informant.}
}
}
\usage{
Andrews2021
}
\description{
The data set contains raw correlation coefficients and
sampling information using \pkg{metafor}-compatible column names
(\code{ri}, \code{vi}, \code{sei}, \code{ni}, and \code{slab}), together with
study-level moderators from the original data set. Three source rows have
missing effect-size inputs and are omitted automatically by model-fitting
functions that require complete outcomes. The original data set assessed the
effect of household chaos on child executive functions
\insertCite{andrews2021examining}{RoBMA} and was used as an example in
\insertCite{bartos2023robust;textual}{RoBMA}.
}
\references{
\insertAllCited{}
}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/wrappers.R
\name{pooled_effect.brma}
\alias{pooled_effect.brma}
\title{Pooled Effect Size for brma Objects}
\usage{
\method{pooled_effect}{brma}(
  object,
  bias_adjusted = TRUE,
  output_measure = NULL,
  transform = NULL,
  probs = c(0.025, 0.975),
  conditional = FALSE,
  ...
)
}
\arguments{
\item{object}{a fitted brma object}

\item{bias_adjusted}{whether to adjust for publication bias. Defaults to
\code{TRUE}. For PET/PEESE models this removes the regression bias term from
the pooled location effect. Selection-model weighting affects response
predictions, not this \code{type = "terms"} wrapper.}

\item{output_measure}{effect-size measure for location/effect predictions.
Defaults to the fitted measure. Supported conversions are among \code{"SMD"},
\code{"COR"}, \code{"ZCOR"}, and \code{"OR"}; \code{"RR"}, \code{"HR"},
\code{"IRR"}, \code{"RD"}, and \code{"GEN"} can only be returned on their
fitted measure. Use \code{transform = "EXP"} for ratio-scale output from
log-scale measures.}

\item{transform}{optional display transformation. Currently \code{"EXP"}
exponentiates log-scale measures \code{"OR"}, \code{"RR"}, \code{"HR"},
and \code{"IRR"}.}

\item{probs}{quantiles of the posterior distribution to be displayed.
Defaults to \code{c(.025, .975)} for 95\% credible intervals.}

\item{conditional}{whether to return the pooled effect conditional on the
effect component for RoBMA product-space objects. Defaults to \code{FALSE}.}

\item{...}{additional arguments passed to \code{\link{predict.brma}}; wrapper
arguments such as \code{newdata}, \code{type}, and \code{quiet} are fixed.}
}
\value{
A \code{brma_samples} object containing posterior samples. When printed,
displays a summary table. Use \code{summary()} to obtain the summary table directly.
The samples can be converted to \pkg{posterior} draws formats using \code{as_draws()}.
}
\description{
Computes the pooled (aggregated) effect size estimate
from a fitted brma object by averaging across the moderator model matrix.
}
\details{
This function is a convenience wrapper around \code{predict.brma(...,
type = "terms", newdata = TRUE, bias_adjusted = TRUE, quiet = TRUE)}.

For meta-regression models, the pooled effect averages the effect size
estimate across moderator levels proportionately to the levels observed
in the data. This provides an estimate representative of the sample of
studies.

For models without moderators, this returns the single mu parameter.
}
\examples{
\dontrun{
if (requireNamespace("metadat", quietly = TRUE)) {
  data(dat.lehmann2018, package = "metadat")
  fit <- brma(
    yi      = yi,
    vi      = vi,
    data    = dat.lehmann2018,
    measure = "SMD",
    seed    = 1,
    silent  = TRUE
  )

  pooled_effect(fit)
}
}

}
\seealso{
\code{\link[=predict.brma]{predict.brma()}}, \code{\link[=pooled_heterogeneity]{pooled_heterogeneity()}}, \code{\link[=blup]{blup()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/zplot.R
\name{hist.zplot_brma}
\alias{hist.zplot_brma}
\title{Histogram of Z-Statistics}
\usage{
\method{hist}{zplot_brma}(
  x,
  plot_type = "base",
  from = -6,
  to = 6,
  by = 0.5,
  length.out = NULL,
  add = FALSE,
  plot_thresholds = TRUE,
  dots_thresholds = NULL,
  dots_hist = NULL,
  dots_all = NULL,
  ...
)
}
\arguments{
\item{x}{a zplot_brma object.}

\item{plot_type}{graphics system: \code{"base"} or \code{"ggplot"}.
Defaults to \code{"base"}.}

\item{from, to}{z-value range for plotting. Defaults to \code{-6} and \code{6}.}

\item{by}{bin width. Defaults to 0.5.}

\item{length.out}{number of bins (alternative to \code{by}).}

\item{add}{whether to add to existing plot. Defaults to \code{FALSE}.}

\item{plot_thresholds}{whether to show significance threshold lines.
Defaults to \code{TRUE}.}

\item{dots_thresholds}{graphical parameters for threshold lines (list).}

\item{dots_hist}{graphical parameters for histogram (list).}

\item{dots_all}{graphical parameters passed to all components (list).}

\item{...}{additional graphical parameters.}
}
\value{
\code{NULL} invisibly for base graphics, or a ggplot2 object.
}
\description{
Plots a histogram of observed z-values from the meta-analysis.
}
\details{
Z-statistics are computed as effect size divided by standard error
(\code{yi / sei}). Histogram bins are adjusted to align with significance
thresholds when selection model priors are present.
}
\seealso{
\code{\link[=plot.zplot_brma]{plot.zplot_brma()}}, \code{\link[=lines.zplot_brma]{lines.zplot_brma()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/residuals.R
\name{residuals.brma}
\alias{residuals.brma}
\title{Residuals for brma Objects}
\usage{
\method{residuals}{brma}(
  object,
  type = "outcome",
  unit = "estimate",
  conditioning_depth = "marginal",
  bias_adjusted = FALSE,
  ...
)
}
\arguments{
\item{object}{a fitted brma object}

\item{type}{the type of residuals to compute. Options are:
\itemize{
\item \code{"outcome"} (default): Raw residuals (observed - fitted).
Available for all model types.
\item \code{"pearson"}: Pearson (semi-standardized) residuals, dividing
raw residuals by the marginal standard error \eqn{\sqrt{v_i + \tau^2}}.
Only available for normal outcome models without selection (weightfunction)
bias adjustment.
\item \code{"rstandard"}: Internally standardized residuals, dividing
raw residuals by their standard errors computed using the hat matrix.
Only available for normal outcome models without selection (weightfunction)
bias adjustment.
\item \code{"LOO-PIT"} (alias: \code{"rstudent"}): Leave-one-out probability
integral transform (PIT) residuals computed via Pareto smoothed importance
sampling. The LOO-CDF value for each observation is computed and transformed
to standard normal quantiles via \eqn{\Phi^{-1}(u_i)}. Under a correctly
specified model, these residuals should follow a standard normal distribution.
This is the recommended standardized residual for Bayesian models as it properly
accounts for estimation uncertainty and leverage. Available for all model
types. Note: This requires that the loo has been computed previously (see
\code{add_loo()} function).
}}

\item{unit}{output unit. Only \code{"estimate"} is implemented currently..}

\item{conditioning_depth}{conditioning depth for non-LOO residuals. \code{"marginal"}
uses fixed effects only, \code{"cluster"} conditions on cluster-level random
effects, and \code{"estimate"} conditions on the full estimate-level fitted
value. LOO-PIT residuals always use the estimate-unit LOO target.}

\item{bias_adjusted}{whether residuals should be computed from bias-adjusted
fitted values. Defaults to \code{FALSE}, which means residuals are computed
as the difference between observed values and raw (biased) predictions
including PET/PEESE terms. Set to \code{TRUE} to compute residuals from
bias-corrected fitted values. Applies to outcome and Pearson residuals. Note
that bias-adjusted residuals are not residuals in the traditional sense.}

\item{...}{additional arguments.}
}
\value{
A numeric vector of residual means, one per estimate.
}
\description{
Computes residuals (observed minus fitted values) from a
fitted brma object, with options for standardization.
}
\details{
Raw residuals (\code{type = "outcome"}) are computed as:
\deqn{e_i = y_i - \hat{\mu}_i}
where \eqn{y_i} is the observed effect size and \eqn{\hat{\mu}_i} is the
fitted value (prediction from the fixed effects).

Pearson residuals (\code{type = "pearson"}) divide raw residuals by the
marginal standard error:
\deqn{r_i^{Pearson} = \frac{e_i}{\sqrt{v_i + \tau^2}}}
where \eqn{v_i} is the sampling variance and \eqn{\tau^2} is the
relevant heterogeneity component. Only available for normal outcome models
without selection (weightfunction) bias adjustment.

Standardized residuals (\code{type = "rstandard"}) use the hat matrix to
compute residual standard errors that account for the uncertainty in
estimated coefficients:
\deqn{z_i = \frac{e_i}{\sqrt{[(I-H)M(I-H)']_{ii}}}}
where \eqn{H} is the hat matrix and \eqn{M} is the marginal variance-covariance
matrix. For models without moderators, this simplifies to the Pearson formula.
Only available for normal outcome models without selection (weightfunction)
bias adjustment.

LOO-PIT residuals (\code{type = "LOO-PIT"}) are the Bayesian equivalent of
studentized deleted residuals \insertCite{vehtari2017practical}{RoBMA}. They
are computed via leave-one-out probability integral transformation:
\deqn{r_i = \Phi^{-1}(u_i)}
where \eqn{u_i = \sum_s w_{is} F(y_i | \theta^{(s)})} is the LOO-weighted CDF
value, \eqn{w_{is}} are the normalized PSIS weights, and \eqn{F} is the
cumulative distribution function of the estimate-unit predictive
distribution used by LOO. Under a correctly specified model, LOO-PIT
residuals should follow a standard normal distribution. Unlike traditional
standardized residuals, LOO-PIT residuals properly account for estimation
uncertainty and leverage without requiring a hat matrix. This is the
recommended method for standardized residuals in Bayesian meta-analysis.

For meta-regression models, fitted values incorporate moderator effects.
For models without moderators, all fitted values equal the pooled effect.

For GLMM models (binomial or Poisson), observed effect sizes and their
sampling variances are computed from the raw frequency data using the
same formulas as \code{metafor::escalc} with the default zero-cell
adjustment (adding 0.5 to all cells when any cell is zero). GLMM residuals
and LOO-PIT values are therefore approximate effect-size-scale diagnostics,
not exact PIT diagnostics for the raw count likelihood.

The residuals are computed separately for each posterior sample,
naturally propagating uncertainty in model parameters to the residuals.
}
\examples{
\dontrun{
if (requireNamespace("metadat", quietly = TRUE)) {
  data(dat.lehmann2018, package = "metadat")

  fit <- bPET(
    yi      = yi,
    vi      = vi,
    data    = dat.lehmann2018,
    measure = "SMD",
    seed    = 1,
    silent  = TRUE
  )

  # raw residuals (default)
  residuals(fit)

  # Pearson and internally standardized residuals
  residuals(fit, type = "pearson")
  residuals(fit, type = "rstandard")

  # LOO-PIT residuals require stored LOO
  fit <- add_loo(fit)
  residuals(fit, type = "LOO-PIT")

  # residuals from bias-adjusted predictions
  residuals(fit, bias_adjusted = TRUE)

  # check Pareto k diagnostics
  plot(loo(fit))
}
}

}
\references{
\insertAllCited{}
}
\seealso{
\code{\link[=predict.brma]{predict.brma()}}, \code{\link[=blup.brma]{blup.brma()}}, \code{\link[=pooled_effect]{pooled_effect()}}, \code{\link[=rstandard.brma]{rstandard.brma()}}, \code{\link[=rstudent.brma]{rstudent.brma()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/zplot.R
\name{print.summary.zplot_brma}
\alias{print.summary.zplot_brma}
\title{Print Zplot Summary}
\usage{
\method{print}{summary.zplot_brma}(x, ...)
}
\arguments{
\item{x}{a summary.zplot_brma object.}

\item{...}{additional arguments (currently unused).}
}
\value{
Invisibly returns the summary object.
}
\description{
Print Zplot Summary
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/wrappers.R
\name{blup}
\alias{blup}
\title{Best Linear Unbiased Predictions (BLUPs)}
\usage{
blup(object, ...)
}
\arguments{
\item{object}{a fitted model object}

\item{...}{additional arguments passed to methods}
}
\value{
Method-specific return value, typically a summary table or
posterior samples of BLUP or empirical-Bayes true-effect summaries.
}
\description{
Computes the estimated true effects (theta) from a
fitted model. These correspond to Best Linear Unbiased Predictions (BLUPs)
or empirical Bayes estimates.
}
\seealso{
\code{\link[=predict.brma]{predict.brma()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/predict.R
\name{predict.brma}
\alias{predict.brma}
\title{Predict From brma Object}
\usage{
\method{predict}{brma}(
  object,
  newdata = NULL,
  type = "terms",
  as_measure = TRUE,
  output_measure = NULL,
  transform = NULL,
  probs = c(0.025, 0.975),
  bias_adjusted = FALSE,
  quiet = FALSE,
  conditional = FALSE,
  ...
)
}
\arguments{
\item{object}{a fitted brma object}

\item{newdata}{specification for prediction data. Defaults to \code{NULL} which
corresponds to prediction for the observed data. Alternatives are:
\itemize{
\item{\code{TRUE} returns a single aggregated prediction (either the scalar
parameter for models without moderators/scale, or the average across the
model matrix for regression models). Not available for \code{type = "response"}
since observation-level sampling variance is required.}
\item{A data.frame (for meta-regression) or a named list with effect size
measure and variability metrics (for meta-analysis) for new studies. The input
must contain the variables used by the requested prediction type. Outcome
columns are optional for \code{type = "terms"} unless PET/PEESE bias terms
are included via \code{bias_adjusted = FALSE}.}
}}

\item{type}{type of prediction to be performed. Options are:
\itemize{
\item{\code{"terms"} (alias: \code{"marginal"}): Fixed-effect parameters only (mu).
Produces the mean effect size estimate at the given predictor levels,
not accounting for random effects.}
\item{\code{"cluster"}: Fixed effects plus cluster-level random effects (mu + gamma).
Only available for multilevel (3-level) models.}
\item{\code{"estimate"} (aliases: \code{"effect"}, \code{"blup"}): True
effects (mu + gamma + theta). For existing normal data, returns posterior
draws of conditional BLUP means, not simulated latent-effect draws. For
new data, samples estimate-level random effects.}
\item{\code{"response"} (alias: \code{"outcome"}): Predicted observed values (yi).
Incorporates both heterogeneity and sampling variability.}
\item{\code{"terms.scale"}: Scale parameter (tau), incorporating scale
regression if present.}
}}

\item{as_measure}{logical; whether to convert GLMM response predictions from
simulated counts to continuity-corrected effect-size estimators (logOR for
binomial, logIRR for Poisson). Defaults to \code{TRUE}. Only relevant for
GLMM models with \code{type = "response"}. When \code{FALSE}, returns raw
frequency data (counts). Use \code{type = "estimate"} for latent logOR/logIRR
predictions.}

\item{output_measure}{effect-size measure for location/effect predictions.
Defaults to the fitted measure. Supported conversions are among \code{"SMD"},
\code{"COR"}, \code{"ZCOR"}, and \code{"OR"}; \code{"RR"}, \code{"HR"},
\code{"IRR"}, \code{"RD"}, and \code{"GEN"} can only be returned on their
fitted measure. Use \code{transform = "EXP"} for ratio-scale output from
log-scale measures.}

\item{transform}{optional display transformation. Currently \code{"EXP"}
exponentiates log-scale measures \code{"OR"}, \code{"RR"}, \code{"HR"},
and \code{"IRR"}.}

\item{probs}{quantiles of the posterior samples to be displayed when the
returned \code{brma_samples} object is printed. Defaults to \code{c(.025, .975)}.}

\item{bias_adjusted}{whether predictions should adjust for publication bias.
Defaults to \code{FALSE}. When \code{TRUE}:
\itemize{
\item{PET/PEESE terms are NOT added to the mean parameter (mu), returning
the bias-corrected effect estimate.}
\item{For \code{type = "response"} with selection models, samples from
the ordinary normal predictive distribution instead of the selected-normal
distribution, simulating what would be observed without publication bias.}
}
When \code{FALSE}:
\itemize{
\item{PET/PEESE terms ARE added to mu, returning predictions that include
the expected bias (i.e., what we expect to observe given publication bias).}
\item{For \code{type = "response"} with selection models, samples from
the selected-normal distribution reflecting the selective publishing
process.}
}}

\item{quiet}{logical; whether to suppress informational messages about
prediction scale and bias adjustment.}

\item{conditional}{whether to return conditional posterior predictions for
RoBMA product-space objects. For location predictions, samples are conditioned
on the effect component; for \code{type = "terms.scale"}, samples are
conditioned on the heterogeneity component. Conditional samples are flattened
to one chain after subsetting posterior rows.}

\item{...}{additional arguments}
}
\value{
A \code{brma_samples} object containing posterior samples. When printed,
displays a summary table via \code{BayesTools::ensemble_estimates_table}. The
underlying samples matrix can be accessed directly (the object inherits from matrix)
or via \code{summary()} to obtain the summary table. The samples can also be converted
to \pkg{posterior} draws formats using \code{as_draws()} and related functions.
}
\description{
\code{predict.brma} predicts values
}
\details{
\strong{Type hierarchy:}
\itemize{
\item{\code{"terms"}: mu (fixed effects only)}
\item{\code{"cluster"}: mu + gamma (adds cluster-level random effect)}
\item{\code{"estimate"}: mu + gamma + theta (adds estimate-level random effect)}
\item{\code{"response"}: mu + gamma + theta + epsilon (adds sampling error)}
}
For existing normal observations, \code{type = "estimate"} reports the
conditional BLUP mean \eqn{E(\theta_i \mid y_i, \mu_i, \tau_i)} for each
posterior draw. It is therefore an empirical-Bayes summary, not a draw from
the full latent-effect posterior \eqn{\theta_i \mid y_i}. For new data,
estimate-level random effects are sampled from their model distribution.

For RoBMA product-space objects, conditional posterior predictions subset
posterior rows according to model indicators. This removes the original
chain structure, so returned \code{brma_samples} objects are intentionally
stored as one flattened chain.

Note that in contrast to \link[metafor]{predict}, the \code{type = "response"} produces
predictions for the new effect size estimates. To obtain results corresponding to
metafor's predict function, use \code{type = "terms"} for the mean effect size
and \code{type = "estimate"} for true effects (prediction interval).

\strong{Likelihood weights:} If the model was fitted with \code{weights},
the weights affect the posterior fit, log-likelihood, LOO/WAIC, and
existing-data conditional diagnostics such as BLUP shrinkage and leverage.
They do not change the observation-level sampling error used by
\code{type = "response"} for normal models: response predictions simulate
raw future effect-size estimates with the supplied \code{sei}, not
\code{sei / sqrt(weight)}.
}
\examples{
\dontrun{
if (requireNamespace("metadat", quietly = TRUE) &&
    requireNamespace("metafor", quietly = TRUE)) {
  data(dat.bcg, package = "metadat")
  dat <- metafor::escalc(
    measure = "RR",
    ai      = tpos,
    bi      = tneg,
    ci      = cpos,
    di      = cneg,
    data    = dat.bcg
  )

  fit <- brma(
    yi      = yi,
    vi      = vi,
    mods    = ~ ablat + year,
    data    = dat,
    measure = "RR",
    seed    = 1,
    silent  = TRUE
  )

  predict(fit, type = "terms")
  predict(fit, newdata = TRUE, type = "terms")
  predict(fit, type = "estimate")
}
}

}
\seealso{
\code{\link[=pooled_effect]{pooled_effect()}}, \code{\link[=pooled_heterogeneity]{pooled_heterogeneity()}}, \code{\link[=blup]{blup()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/priors.R
\name{contr.BayesTools}
\alias{contr.BayesTools}
\alias{contr.orthonormal}
\alias{contr.meandif}
\alias{contr.independent}
\title{BayesTools Contrast Matrices}
\usage{
contr.orthonormal(n, contrasts = TRUE)

contr.meandif(n, contrasts = TRUE)

contr.independent(n, contrasts = TRUE)
}
\arguments{
\item{n}{a vector of levels for a factor, or the number of levels}

\item{contrasts}{logical indicating whether contrasts should be computed}
}
\description{
BayesTools provides several contrast matrix functions for Bayesian factor analysis.
These functions create different types of contrast matrices that can be used with factor
variables in Bayesian models.
}
\details{
The package includes the following contrast functions:
\describe{
\item{\code{contr.orthonormal}}{Return a matrix of orthonormal contrasts.
Code is based on \code{stanova::contr.bayes} and corresponding to description
by \insertCite{rouder2012default;textual}{BayesTools}. Returns a matrix with n rows and
k columns, with k = n - 1 if \code{contrasts = TRUE} and k = n if \code{contrasts = FALSE}.}
\item{\code{contr.meandif}}{Return a matrix of mean difference contrasts.
This is an adjustment to the \code{contr.orthonormal} that ascertains that the prior
distributions on difference between the grand mean and factor level are identical independent
of the number of factor levels (which does not hold for the orthonormal contrast). Furthermore,
the contrast is re-scaled so the specified prior distribution exactly corresponds to the prior
distribution on difference between each factor level and the grand mean -- this is approximately
twice the scale of \code{contr.orthonormal}. Returns a matrix with n rows and k columns,
with k = n - 1 if \code{contrasts = TRUE} and k = n if \code{contrasts = FALSE}.}
\item{\code{contr.independent}}{Return a matrix of independent contrasts -- a level for each term.
Returns a matrix with n rows and k columns, with k = n if \code{contrasts = TRUE} and k = n
if \code{contrasts = FALSE}.}
}
}
\examples{
# Orthonormal contrasts
contr.orthonormal(c(1, 2))
contr.orthonormal(c(1, 2, 3))

# Mean difference contrasts
contr.meandif(c(1, 2))
contr.meandif(c(1, 2, 3))

# Independent contrasts
contr.independent(c(1, 2))
contr.independent(c(1, 2, 3))

}
\references{
\insertAllCited{}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/loo.R
\name{add_waic.brma}
\alias{add_waic.brma}
\alias{add_waic}
\title{Add WAIC to brma Objects}
\usage{
\method{add_waic}{brma}(object, unit = "estimate", ...)
}
\arguments{
\item{object}{a brma model object.}

\item{unit}{output/deletion unit. See \code{\link{add_loo}}; the same
accepted values and multilevel constraint apply.}

\item{...}{additional arguments passed to \code{\link[loo]{waic}}.}
}
\value{
The brma object with the WAIC result stored in
\code{object[["waic"]][[unit]]}.
}
\description{
Compute the Widely Applicable Information Criterion (WAIC)
for brma model objects and store the result in the object.
}
\details{
WAIC is an alternative to LOO-CV for estimating out-of-sample predictive
accuracy. Like LOO, it evaluates expected predictive performance for new
observations.

In most cases, LOO-PSIS (via \code{\link{add_loo}}) is preferred over WAIC
because it provides better estimates and includes diagnostics (Pareto k
values) that indicate when the approximation may be unreliable.
}
\seealso{
\code{\link{waic.brma}}, \code{\link{add_loo}}, \code{\link[loo]{waic}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/zplot.R
\name{as_zplot.brma}
\alias{as_zplot.brma}
\alias{as_zplot}
\title{Transform brma Object to Zplot}
\usage{
\method{as_zplot}{brma}(
  object,
  significance_level = stats::qnorm(0.975),
  max_samples = 10000,
  ...
)
}
\arguments{
\item{object}{a normal-outcome \code{brma} object.}

\item{significance_level}{z-value threshold for significance. Defaults
to \code{qnorm(0.975)} (two-sided alpha = 0.05).}

\item{max_samples}{maximum number of posterior samples for estimation.
Defaults to 10000. Use \code{Inf} to use all posterior samples.}

\item{...}{additional arguments (currently unused).}
}
\value{
The input object with added class \code{"zplot_brma"} and a new
\code{zplot} list component containing:
\describe{
\item{estimates}{a list with posterior samples for \code{EDR} and
missing-study \code{weights}}
\item{data}{a list with \code{significance_level}, observed
\code{z}-statistics, \code{N_significant}, and \code{N_observed}}
}
}
\description{
Transforms an estimated brma model into a zplot object
that can be summarized and plotted to assess replicability.
}
\details{
Zplot analysis estimates the Expected Discovery Rate (EDR), the posterior
mean probability that an exact replication would be statistically significant
at the supplied threshold. This provides insight into the replicability of
findings in a literature.

The implementation uses extrapolation mode by default, which removes
selection-model weights when evaluating the model-implied z-value density.
PET/PEESE regression terms remain part of the fitted location model.
Zplot diagnostics are available only for normal outcome models. GLMM
objects are rejected because their raw likelihood is on a count scale while
zplot diagnostics require observed effect-size z-statistics with standard
errors.

The resulting object retains all original brma properties while adding
zplot results, enabling both standard meta-analytic summaries and
zplot diagnostics on the same object.
}
\examples{
\dontrun{
if (requireNamespace("metadat", quietly = TRUE)) {
  data(dat.lehmann2018, package = "metadat")
  fit <- bPET(yi = yi, vi = vi, data = dat.lehmann2018, measure = "SMD")

  zfit <- as_zplot(fit)
  summary(zfit)
  plot(zfit)
}
}

}
\seealso{
\code{\link[=summary.zplot_brma]{summary.zplot_brma()}}, \code{\link[=plot.zplot_brma]{plot.zplot_brma()}}, \code{\link[=hist.zplot_brma]{hist.zplot_brma()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/wrappers.R
\name{ranef}
\alias{ranef}
\title{Random Effects}
\usage{
ranef(object, ...)
}
\arguments{
\item{object}{a fitted model object}

\item{...}{additional arguments passed to methods}
}
\value{
Method-specific return value, typically posterior samples of the
random effect deviations.
}
\description{
Extracts the estimated random effects (deviations from the
fixed-effect predictions) from a fitted model.
}
\seealso{
\code{\link[=blup]{blup()}}, \code{\link[=predict.brma]{predict.brma()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/loo.R
\name{waic.brma}
\alias{waic.brma}
\alias{waic}
\title{WAIC for brma Objects}
\usage{
\method{waic}{brma}(x, unit = "estimate", ...)
}
\arguments{
\item{x}{a brma model object.}

\item{unit}{output/deletion unit. See \code{\link{add_loo}}.}

\item{...}{additional arguments (currently unused).}
}
\value{
An object of class \code{"waic"} as returned by
\code{\link[loo]{waic}}.
}
\description{
Extract the WAIC object from a brma model object.
The WAIC must first be computed using \code{\link{add_waic}}.
}
\details{
This function extracts the WAIC object that was previously computed and
stored using \code{object <- add_waic(object, unit = unit)}. If WAIC has not
been computed for the requested unit, an error is thrown.

This is the RoBMA S3 generic and \code{brma} method. The method is also
registered for \code{\link[loo]{waic}}, so \code{loo::waic(fit)} extracts
the cached WAIC object for \code{brma} fits. Use \code{\link[loo]{waic}}
directly for raw log-likelihood arrays or matrices.

In most cases, LOO-PSIS (via \code{\link{loo.brma}}) is preferred over WAIC
because it provides better estimates and includes diagnostics (Pareto k
values) that indicate when the approximation may be unreliable.
}
\seealso{
\code{\link{add_waic}}, \code{\link{loo.brma}}, \code{\link[loo]{waic}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/datasets.R
\docType{data}
\name{Kroupova2021}
\alias{Kroupova2021}
\title{881 estimates from 69 studies of a relationship between employment and
educational outcomes collected by \insertCite{kroupova2021student;textual}{RoBMA}}
\format{
A data.frame with 11 columns and 881 observations:
\describe{
\item{\code{r}}{Partial correlation coefficient.}
\item{\code{se}}{Standard error of r.}
\item{\code{study}}{Study label.}
\item{\code{sample_size}}{Sample size.}
\item{\code{education_outcome}}{Type of educational outcome.}
\item{\code{employment_intensity}}{Employment intensity.}
\item{\code{students_gender}}{Gender composition of the student sample.}
\item{\code{location}}{Study location.}
\item{\code{design}}{Study design.}
\item{\code{endogenity_control}}{Whether endogeneity was controlled.}
\item{\code{motivation_control}}{Whether motivation was controlled.}
}
}
\source{
\url{http://meta-analysis.cz/students}
}
\usage{
Kroupova2021
}
\description{
The data set contains partial correlation coefficients, standard errors,
study labels, sample sizes, type of the educational outcome, intensity of the
employment, gender of the student population, study location, study design, whether
the study controlled for endogeneity, and whether the study controlled for motivation.
The original data set including additional variables and the publication are
available from the project page.
(Note that some standard errors and employment intensities are missing.)
}
\references{
\insertAllCited{}
}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/marginal_means.R
\name{print.marginal_means.brma}
\alias{print.marginal_means.brma}
\title{Print Estimated Marginal Means}
\usage{
\method{print}{marginal_means.brma}(x, ...)
}
\arguments{
\item{x}{a \code{marginal_means.brma} object.}

\item{...}{additional arguments passed to \code{summary()}.}
}
\value{
Returns \code{x} invisibly.
}
\description{
Prints the estimated marginal means summary.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/radial.R
\name{radial.brma}
\alias{radial.brma}
\alias{radial}
\alias{galbraith}
\alias{galbraith.brma}
\title{Radial (Galbraith) Plot for brma Object}
\usage{
\method{radial}{brma}(
  x,
  center = FALSE,
  xlim,
  zlim,
  xlab,
  zlab,
  atz,
  aty,
  steps = 7,
  level = 95,
  digits = 2,
  transf,
  targs,
  plot_type = "base",
  ...
)

\method{galbraith}{brma}(x, ...)
}
\arguments{
\item{x}{a fitted brma object. Must be an intercept-only model
(no moderators or scale regression).}

\item{center}{logical indicating whether to center the plot at the
pooled estimate. When \code{TRUE}, the pooled estimate is shifted to
zero on the z-axis. Defaults to \code{FALSE}.}

\item{xlim}{x-axis limits. If not specified, limits are computed from data.}

\item{zlim}{z-axis limits. If not specified, limits are computed from data.}

\item{xlab}{title for the x-axis. If not specified, a default label is used.}

\item{zlab}{title for the z-axis. If not specified, a default label is used.}

\item{atz}{numeric vector of positions for tick marks on the z-axis (left
vertical axis). If not specified, positions are chosen automatically.}

\item{aty}{numeric vector of effect size values to mark on the y-axis arc
(right side). If not specified, values are chosen automatically.}

\item{steps}{integer specifying the number of tick marks on the y-axis arc
when \code{aty} is not specified. Defaults to \code{7}.}

\item{level}{numeric value between 0 and 100 specifying the confidence level
for the confidence band and pooled-effect CI arc. Defaults to \code{95}.}

\item{digits}{integer specifying the number of decimal places for y-axis
arc labels. Defaults to \code{2}.}

\item{transf}{optional transformation function applied to the y-axis arc
labels (e.g., \code{transf = exp} when effect sizes are on a log scale).}

\item{targs}{optional list of additional arguments passed to the
\code{transf} function.}

\item{plot_type}{whether to use a base plot \code{"base"} or ggplot2
\code{"ggplot"} for plotting. Defaults to \code{"base"}.}

\item{...}{additional graphical arguments to customize the plot appearance:
\describe{
\item{pch}{point symbol (default: 21, filled circle)}
\item{col}{point border color (default: "black")}
\item{bg}{point fill/background color (default: "#A6A6A6")}
\item{cex}{point size multiplier for base graphics (default: 1)}
\item{size}{point size for ggplot2 (default: 2)}
\item{las}{axis-label style for base graphics (default: 1)}
\item{back}{background color for the confidence band (default: "grey90").
Set to \code{NA} to suppress.}
\item{arc.res}{integer specifying the number of line segments used to
draw arcs (default: 100)}
\item{main}{character string for plot title (default: no title)}
\item{as_data}{if \code{TRUE}, returns plot data instead of creating
the plot}
}}
}
\value{
\code{radial.brma} returns \code{NULL} invisibly if
\code{plot_type = "base"} or a ggplot object if \code{plot_type = "ggplot"}.

If \code{as_data = TRUE}, returns a list with the computed plot data
including: \code{points}, \code{refline}, \code{band}, \code{arc},
\code{arc_ticks}, \code{ci_arc}, \code{ci_ticks}, \code{xlim}, and
\code{zlim}.
}
\description{
\code{radial.brma} creates a radial (Galbraith) plot for a
fitted brma object. The plot displays each study as a point with
precision on the x-axis and the standardized effect size on the z-axis.
Without centering, a line from the origin through any point has slope equal
to the observed effect size. With centering, slopes are relative to the
pooled estimate. An arc on the right side maps standardized values back to
the effect size scale.
}
\details{
The radial (Galbraith) plot transforms each study's effect size and
standard error into a point in precision-standardized space:

\itemize{
\item x-axis: precision = \eqn{1/\sqrt{v_i + \hat{\tau}^2}}
\item z-axis: standardized effect = \eqn{y_i/\sqrt{v_i + \hat{\tau}^2}}
}

Under the random-effects model, studies consistent with the pooled effect
should fall within the sloped parallelogram confidence band around the
pooled-effect line. The arc on the right side allows reading individual
effect sizes by projecting from the origin through a point to the arc; when
\code{center = TRUE}, the plotted slope is relative to the pooled effect.

This function requires an intercept-only model; radial plots are not
meaningful for meta-regression or location-scale models where the
pooled effect varies across studies.

\code{galbraith()} is a same-argument alias for \code{radial()}.
}
\examples{
\dontrun{
if (requireNamespace("metadat", quietly = TRUE) &&
    requireNamespace("metafor", quietly = TRUE)) {
  data(dat.bcg, package = "metadat")
  dat <- metafor::escalc(
    measure = "RR",
    ai      = tpos,
    bi      = tneg,
    ci      = cpos,
    di      = cneg,
    data    = dat.bcg
  )

  fit <- brma(yi = yi, vi = vi, data = dat, measure = "RR")
  radial(fit)
  radial(fit, center = TRUE)
  radial(fit, plot_type = "ggplot")
  galbraith(fit)
}
}

}
\references{
Galbraith, R. F. (1988). Graphical display of estimates having differing
standard errors. \emph{Technometrics}, 30(3), 271-281.

Galbraith, R. F. (1994). Some applications of radial plots. \emph{Journal
of the American Statistical Association}, 89(428), 1232-1242.
}
\seealso{
\code{\link[=funnel.brma]{funnel.brma()}}, \code{\link[=pooled_effect.brma]{pooled_effect.brma()}}, \code{\link[=pooled_heterogeneity.brma]{pooled_heterogeneity.brma()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/brma.norm.R
\name{brma}
\alias{brma}
\alias{brma.norm}
\title{Bayesian Meta-Analysis}
\usage{
brma(
  yi,
  vi,
  sei,
  weights,
  ni,
  mods,
  scale,
  cluster,
  data,
  slab,
  subset,
  measure,
  prior_effect,
  prior_heterogeneity,
  prior_mods,
  prior_scale,
  prior_heterogeneity_allocation,
  standardize_continuous_predictors = TRUE,
  set_contrast_factor_predictors = "treatment",
  prior_unit_information_sd,
  rescale_priors = 1,
  prior_informed_field,
  prior_informed_subfield,
  sample = 5000,
  burnin = 2000,
  adapt = 500,
  chains = 3,
  thin = 1,
  parallel = FALSE,
  autofit = FALSE,
  autofit_control = set_autofit_control(),
  convergence_checks = set_convergence_checks(),
  seed = NULL,
  silent,
  ...
)
}
\arguments{
\item{yi}{a vector of effect sizes, or a formula with the effect size on the
left-hand side and location moderators on the right-hand side (for example
\code{yi ~ x1 + x2}). If a formula is supplied, \code{mods} must not be specified.}

\item{vi}{a vector of sampling variances. Either \code{vi} or \code{sei} must be
supplied for normal models.}

\item{sei}{a vector of standard errors. Either \code{vi} or \code{sei} must be
supplied for normal models.}

\item{weights}{an optional vector of positive likelihood weights. For
normal/effect-size models, each weight powers the estimate likelihood. For
constructors with GLMM raw-count input, each weight powers the paired
two-arm likelihood for one study.}

\item{ni}{an optional vector of sample sizes. Used for \code{measure = "GEN"}
or when estimating \code{"UISD"}).}

\item{mods}{an optional matrix, data.frame, or formula specifying
location moderators (meta-regressors). Formula input is evaluated in \code{data}.}

\item{scale}{an optional matrix, data.frame, or formula specifying
scale predictors for location-scale models. Formula input is evaluated in
\code{data}.}

\item{cluster}{an optional vector of cluster identifiers for multilevel
meta-analysis.}

\item{data}{an optional data frame containing the variables.}

\item{slab}{an optional vector of study labels.}

\item{subset}{an optional logical or numeric vector specifying a subset of
data to be used.}

\item{measure}{a character string specifying the effect size measure.
Normal/effect-size constructors require an explicit value and support
\code{"SMD"}, \code{"ZCOR"}, \code{"RR"}, \code{"OR"}, \code{"HR"}, \code{"RD"}, \code{"IRR"}, and \code{"GEN"}.
Use \code{"GEN"} only for general effect sizes without a known unit information
standard deviation. GLMM raw-count constructors support only \code{"OR"} and
\code{"IRR"} and default to \code{"OR"}.}

\item{prior_effect}{prior distribution for the effect size (\eqn{\mu}) parameter
(the intercept). If omitted, a default prior is constructed. In single-model
functions, explicit \code{NULL} or \code{FALSE} sets a spike at zero.}

\item{prior_heterogeneity}{prior distribution for the heterogeneity (\eqn{\tau})
parameter. If omitted, a default prior is constructed. In single-model
functions, explicit \code{NULL} or \code{FALSE} sets a spike at zero.}

\item{prior_mods}{prior distribution for the moderators (\eqn{\beta}) parameters.
A single prior applies to all terms; a named list can specify term-specific
priors. If omitted or \code{NULL}, default priors are used.}

\item{prior_scale}{prior distribution for the scale (\eqn{\delta}) parameters.
A single prior applies to all terms; a named list can specify term-specific
priors. If omitted or \code{NULL}, default priors are used.}

\item{prior_heterogeneity_allocation}{prior distribution for the fraction of
heterogeneity allocated to the cluster-level component in multilevel models
(\eqn{\rho}). If omitted or \code{NULL}, defaults to \code{Beta(1, 1)}.}

\item{standardize_continuous_predictors}{logical. Whether to standardize continuous predictors.
Defaults to \code{TRUE}.}

\item{set_contrast_factor_predictors}{character. How to set contrast for factor predictors.
Defaults are constructor-specific and shown in each function usage; single-model
constructors use \code{"treatment"}, while model-averaging constructors use \code{"meandif"}.}

\item{prior_unit_information_sd}{numeric. The unit information standard deviation (\eqn{\sigma_{unit}}).
Cannot be used together with \code{prior_informed_field}.}

\item{rescale_priors}{numeric. A scaling factor for supported prior distributions.
Point and none priors are unchanged. For constructors with publication-bias
prior distributions, \code{rescale_priors} does not rescale them except for the
default PEESE prior's UISD adjustment. Defaults to 1.}

\item{prior_informed_field}{character. The field of the informed prior distributions.
Omit to use the standard default prior specification; explicit \code{NULL} is invalid.}

\item{prior_informed_subfield}{character. The subfield of the informed prior distributions.
Omit to use the field-specific default, such as \code{"Cochrane"} for
\code{prior_informed_field = "medicine"}; explicit \code{NULL} is invalid.}

\item{sample}{numeric. Number of MCMC samples to save. Defaults to \code{5000}.}

\item{burnin}{numeric. Number of burn-in iterations. Defaults to \code{2000}.}

\item{adapt}{numeric. Number of adaptation iterations. Defaults to \code{500}.}

\item{chains}{numeric. Number of MCMC chains. Defaults to \code{3}.}

\item{thin}{numeric. Thinning interval. Defaults to \code{1}.}

\item{parallel}{logical. Whether to run MCMC chains in parallel. Defaults to \code{FALSE}.}

\item{autofit}{logical. Whether to automatically extend the MCMC chains if convergence is not met.
Defaults to \code{FALSE}.}

\item{autofit_control}{list of autofit control settings. See \code{\link[=set_autofit_control]{set_autofit_control()}} for details.}

\item{convergence_checks}{list of convergence check settings. See \code{\link[=set_convergence_checks]{set_convergence_checks()}} for details.}

\item{seed}{numeric. Random seed for reproducibility. Defaults to \code{NULL}.}

\item{silent}{logical. Whether to suppress output. Constructors with no
explicit default use \code{RoBMA.get_option("silent")} when \code{silent} is omitted.
Model-averaging wrappers default to \code{TRUE} unless explicitly changed.}

\item{...}{additional advanced arguments. Fitting functions reject unused
arguments; currently recognized internal arguments include \code{only_data},
\code{only_priors}, \code{is_JASP}, and \code{is_JASP_prefix}.}
}
\value{
A fitted object of class \code{c("brma.norm", "brma")}. The object
contains checked \code{data}, checked \code{priors}, the JAGS \code{fit}, cached \code{summary},
and cached \code{coefficients}. If the corresponding package options are enabled,
it can also contain cached LOO, WAIC, or marginal likelihood results. The
advanced internal \code{only_data = TRUE} and \code{only_priors = TRUE} paths return
partially constructed objects.
}
\description{
Function for fitting normal-likelihood/effect-size
random-effects, meta-regression, multilevel, and location-scale
meta-analytic models. \code{brma.norm()} is an alias for \code{brma()}; raw-count
GLMM models use \code{brma.glmm()}.
}
\details{
\subsection{Prior distributions}{

Prior distributions must be specified for all model parameters.
This typically includes the pooled effect \eqn{\mu} and between-study heterogeneity
\eqn{\tau}. In the case of meta-regression, the pooled effect \eqn{\mu} corresponds to
the intercept, and additional prior distributions for the regression coefficients
are required. In the case of a location-scale model, the between-study heterogeneity
corresponds to the intercept of the scale regression, and additional prior
distributions for the scale regression coefficients are required.

There are several ways to specify the prior distributions: \enumerate{
\item{via a standardized effect size \code{measure} with known unit information standard deviation,}
\item{by estimating unit information standard deviation using sample sizes \code{ni},}
\item{by manually setting \code{prior_unit_information_sd},}
\item{by specifying informed empirical prior distributions via \code{prior_informed_field}
and \code{prior_informed_subfield},}
\item{or via fully custom specification using the \code{prior_effect}, \code{prior_heterogeneity},
\code{prior_mods}, \code{prior_scale}, and \code{prior_heterogeneity_allocation} arguments.}
}
In all cases, the prior behavior can be further modified by the \code{rescale_priors},
\code{standardize_continuous_predictors}, and \code{set_contrast_factor_predictors} arguments.
\subsection{(1) Specifying prior distributions via standardized effect size measures with known}{

unit information standard deviation
This is the easiest way to specify prior distributions. The width of prior
distributions is based on a fraction of the known unit information standard deviation (\code{UISD})
\insertCite{rover2021weakly}{RoBMA}. The default prior distributions for the parameters
are set as follows:\tabular{ll}{
effect size:               \tab Normal(0, \eqn{\frac{1}{2}} \code{UISD})  \cr
heterogeneity:             \tab Normal+(0, \eqn{\frac{1}{4}} \code{UISD}) \cr
effect moderation:         \tab Normal(0, \eqn{\frac{1}{4}} \code{UISD})  \cr
heterogeneity moderation:  \tab Normal(0, \eqn{\frac{1}{2}})
}
The heterogeneity moderation parameters are multiplicative, as such they are independent of \code{UISD}.

The default fraction of the \code{UISD} can be changed via \code{RoBMA.options()} using one of
the following arguments: \code{"default_UISD.effect"}, \code{"default_UISD.heterogeneity"},
\code{"default_UISD.mods"}, \code{"default_UISD.scale"}.

The known \code{UISD} for standardized effect size measures (\code{measure}) are set as follows:
\tabular{ll}{
\code{"SMD"}:   \tab \eqn{\sqrt{2}} \cr
\code{"ZCOR"}:  \tab \eqn{1}        \cr
\code{"RR"}:    \tab \eqn{\sqrt{4}} \cr
\code{"OR"}:    \tab \eqn{\sqrt{4}} \cr
\code{"HR"}:    \tab \eqn{\sqrt{4}} \cr
\code{"IRR"}:   \tab \eqn{\sqrt{4}}
}
See Chapter 2.4 in \insertCite{spiegelhalter2004bayesian;textual}{RoBMA} and
Chapter 1 in \insertCite{grieve2022hybrid;textual}{RoBMA}.
}

\subsection{(2) Estimating unit information standard deviation using sample sizes}{

When effect sizes are on a non-standardized scale (\code{measure = "GEN"}) or use a
standardized effect size without known \code{UISD}, the \code{UISD} can be estimated from
sample sizes (\code{ni}) and standard errors (\code{sei}) following Equation 6 in
\insertCite{rover2021weakly;textual}{RoBMA}. The estimated \code{UISD} is then used to
scale the default prior distributions as described in section (1).

Note that the known \code{UISD} for standardized effect size measures (section (1)) is
used if available, even when \code{ni} is provided.
}

\subsection{(3) Manually setting unit information standard deviation}{

Alternatively, the \code{UISD} can be specified directly via the \code{prior_unit_information_sd}
argument. This is useful when the appropriate scale for prior distributions is known
a priori or when multiple analyses are to be performed on subsets of the same data
(re-estimating \code{UISD} on different subsets of the data can lead to slightly different
prior distributions for different subsets; see \code{\link[=estimate_unit_information_sd]{estimate_unit_information_sd()}}).
The specified \code{prior_unit_information_sd} is then used to scale the default prior
distributions as described in section (1).

Note that the manually specified \code{prior_unit_information_sd} takes precedence over
the estimated \code{UISD} from \code{ni} (section (2)) and the known \code{UISD} from \code{measure}
(section (1)). It cannot be combined with \code{prior_informed_field}.
}

\subsection{(4) Specifying informed empirical prior distributions}{

Informed prior distributions can be specified via the \code{prior_informed_field} and
\code{prior_informed_subfield} arguments. Currently, only \code{prior_informed_field = "medicine"}
with subfields defined in \code{BayesTools::prior_informed_medicine_names} is supported,
which uses empirically derived prior distributions from medical meta-analyses as described in
\insertCite{bartos2021bayesian;textual}{RoBMA} and
\insertCite{bartos2023empirical;textual}{RoBMA}.

When \code{prior_informed_field = "medicine"}, the default \code{prior_informed_subfield} is
\code{"Cochrane"} (i.e., using the whole CDSR database as a reference). The informed
prior distributions are available for the following effect size measures:
\tabular{ll}{
\code{"SMD"}:  \tab standardized mean difference \cr
\code{"OR"}:   \tab log odds ratio               \cr
\code{"RR"}:   \tab log risk ratio               \cr
\code{"RD"}:   \tab risk difference              \cr
\code{"HR"}:   \tab log hazard ratio
}

Note that informed prior distributions are only available for the effect size (\eqn{\mu})
and heterogeneity (\eqn{\tau}) parameters. For effect moderation (\code{prior_mods}), the
informed effect prior is scaled by a factor of \code{RoBMA.get_option("default_informed_priors.mods")}.
For heterogeneity moderation (\code{prior_scale}), a normal prior with standard deviation
specified by \code{RoBMA.get_option("default_informed_priors.scale")} is used.
}

\subsection{(5) Fully custom prior distributions}{

Prior distributions can be fully customized by directly specifying the \code{prior_effect},
\code{prior_heterogeneity}, \code{prior_mods}, \code{prior_scale}, and \code{prior_heterogeneity_allocation}
arguments. These should be prior distribution objects created via \code{BayesTools::prior()}
or related functions (e.g., \code{prior_factor()}).
}

\subsection{Rescaling prior distributions}{

The \code{rescale_priors} argument allows rescaling supported prior distributions by a
multiplicative factor. For example, \code{rescale_priors = 2} doubles the standard
deviations/scales of normal, Cauchy, t, and inverse-gamma prior distributions,
making them more diffuse. Point and none priors are unchanged.
}

\subsection{Handling of continuous and factor predictors}{

When \code{standardize_continuous_predictors = TRUE}, continuous moderator and
scale predictors are internally standardized before fitting. Reported
summaries are transformed back to the original predictor scale by default;
use \code{standardized_coefficients = TRUE} in \code{summary()} or related methods to
inspect coefficients on the standardized scale.

Factor predictors use the contrast specified by
\code{set_contrast_factor_predictors}. The default \code{"treatment"} follows standard
treatment coding. Model-averaging functions default to \code{"meandif"} so
inclusion priors for factor levels are centered on deviations from the grand
mean.
}

}
}
\examples{
\dontrun{
if (requireNamespace("metadat", quietly = TRUE) &&
    requireNamespace("metafor", quietly = TRUE)) {
  data(dat.bcg, package = "metadat")
  dat <- metafor::escalc(
    measure = "RR",
    ai      = tpos,
    bi      = tneg,
    ci      = cpos,
    di      = cneg,
    data    = dat.bcg
  )

  fit <- brma(
    yi      = yi,
    vi      = vi,
    mods    = ~ ablat + year,
    data    = dat,
    measure = "RR",
    seed    = 1,
    silent  = TRUE
  )

  summary(fit)
  predict(fit, type = "terms")
}
}

}
\references{
\insertAllCited{}
}
\seealso{
\code{\link[=RoBMA]{RoBMA()}}, \code{\link[=BMA]{BMA()}}, \code{\link[=brma.glmm]{brma.glmm()}}, \code{\link[=summary.brma]{summary.brma()}},
\code{\link[=plot.brma]{plot.brma()}}, \code{\link[=predict.brma]{predict.brma()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/bridgesampling.R
\docType{import}
\name{reexports}
\alias{reexports}
\alias{bridge_sampler}
\alias{logml}
\alias{post_prob}
\alias{bf}
\alias{bayes_factor}
\title{Objects exported from other packages}
\keyword{internal}
\description{
These objects are imported from other packages. Follow the links
below to see their documentation.

\describe{
  \item{bridgesampling}{\code{\link[bridgesampling:bf]{bayes_factor}}, \code{\link[bridgesampling]{bf}}, \code{\link[bridgesampling]{bridge_sampler}}, \code{\link[bridgesampling]{logml}}, \code{\link[bridgesampling]{post_prob}}}
}}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/cooks.distance.R
\name{cooks.distance.brma}
\alias{cooks.distance.brma}
\title{Cook's Distance for brma Objects}
\usage{
\method{cooks.distance}{brma}(model, ...)
}
\arguments{
\item{model}{a fitted brma object.}

\item{...}{additional arguments (currently ignored).}
}
\value{
A numeric vector of Cook's distance values, one for each observation.
}
\description{
Computes Cook's distance for a fitted brma object. Cook's
distance measures the aggregate influence of each observation on the
model coefficients.
}
\details{
Cook's distance is computed as a PSIS leave-one-out deletion diagnostic. For
each observation \eqn{i}, normalized PSIS weights estimate the fitted values
under the leave-one-out posterior. The distance is the posterior Mahalanobis
distance between the full-data and leave-one-out fitted-value vectors:
\deqn{D_i = \frac{\Delta_i' V_\mu^+ \Delta_i}{P}}

where \eqn{\Delta_i = \hat{\mu} - \hat{\mu}_{(-i)}}, \eqn{V_\mu^+} is the
generalized inverse of the full-posterior fitted-value covariance, and
\eqn{P} is the rank of the fixed-effect model matrix.
}
\examples{
\dontrun{
if (requireNamespace("metadat", quietly = TRUE)) {
  data(dat.lehmann2018, package = "metadat")
  fit <- bPET(yi = yi, vi = vi, data = dat.lehmann2018, measure = "SMD")
  fit <- add_loo(fit)

  cooks.distance(fit)
}
}

}
\seealso{
\code{\link{influence.brma}}, \code{\link{dffits.brma}}, \code{\link{hatvalues.brma}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/wrappers.R
\name{true_effects}
\alias{true_effects}
\title{True Effects}
\usage{
true_effects(object, ...)
}
\arguments{
\item{object}{a fitted model object}

\item{...}{additional arguments passed to methods}
}
\value{
Method-specific return value, typically a summary table or
posterior samples of BLUP or empirical-Bayes true-effect summaries.
}
\description{
Computes the estimated true effects (theta) from a
fitted model. This is a separate S3 generic whose \code{brma} method
delegates to \code{\link{blup.brma}}.
}
\seealso{
\code{\link[=blup]{blup()}}, \code{\link[=predict.brma]{predict.brma()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/as_draws.R
\name{as_draws.brma}
\alias{as_draws.brma}
\alias{as_draws}
\alias{as_draws_array}
\alias{as_draws_df}
\alias{as_draws_list}
\alias{as_draws_matrix}
\alias{as_draws_rvars}
\alias{as_draws.default}
\alias{as_draws_array.default}
\alias{as_draws_df.default}
\alias{as_draws_list.default}
\alias{as_draws_matrix.default}
\alias{as_draws_rvars.default}
\alias{as_draws_array.brma}
\alias{as_draws_df.brma}
\alias{as_draws_list.brma}
\alias{as_draws_matrix.brma}
\alias{as_draws_rvars.brma}
\title{Convert brma Objects to posterior Draws Formats}
\usage{
as_draws(x, ...)

as_draws_array(x, ...)

as_draws_df(x, ...)

as_draws_list(x, ...)

as_draws_matrix(x, ...)

as_draws_rvars(x, ...)

\method{as_draws}{default}(x, ...)

\method{as_draws_array}{default}(x, ...)

\method{as_draws_df}{default}(x, ...)

\method{as_draws_list}{default}(x, ...)

\method{as_draws_matrix}{default}(x, ...)

\method{as_draws_rvars}{default}(x, ...)

\method{as_draws}{brma}(x, ...)

\method{as_draws_array}{brma}(x, ...)

\method{as_draws_df}{brma}(x, ...)

\method{as_draws_list}{brma}(x, ...)

\method{as_draws_matrix}{brma}(x, ...)

\method{as_draws_rvars}{brma}(x, ...)
}
\arguments{
\item{x}{an object to convert. The \code{brma} methods expect a fitted
\code{brma} object; default methods forward non-\code{brma} objects to the
corresponding \pkg{posterior} conversion function.}

\item{...}{additional arguments passed to the corresponding
\pkg{posterior} function.}
}
\value{
An object of the corresponding \pkg{posterior} draws class.
}
\description{
Provides an interface to the \pkg{posterior} package
for \code{brma} objects. These functions convert the MCMC samples
from a fitted \code{brma} model to various draws formats supported
by the \pkg{posterior} package, enabling the use of posterior's
rich set of diagnostics and summary functions.
}
\details{
These functions are S3 generics. Their default methods forward to the
corresponding \pkg{posterior} generics so attaching \pkg{RoBMA} preserves
the usual \pkg{posterior} behavior for non-\code{brma} objects.

The following conversion functions are available:
\itemize{
\item \code{as_draws}: converts to the default draws format
\item \code{as_draws_array}: converts to a 3-D array (iteration x chain x variable)
\item \code{as_draws_df}: converts to a data frame with columns for iterations, chains, and variables
\item \code{as_draws_list}: converts to a list of lists
\item \code{as_draws_matrix}: converts to a 2-D matrix (draw x variable)
\item \code{as_draws_rvars}: converts to random variable objects
}

These methods require the \pkg{posterior} package to be installed.
For \code{brma} methods, conversion is performed by first extracting the MCMC
samples as a \code{mcmc.list} object and then using the corresponding
\pkg{posterior} conversion function. \code{brma_samples} objects have
separate methods documented at \code{\link{as_draws.brma_samples}}.
}
\seealso{
\code{\link[posterior]{draws}}, \code{\link{brma}},
\code{\link{as_draws.brma_samples}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/loo.R
\name{check_loo.brma}
\alias{check_loo.brma}
\alias{check_loo}
\title{Check LOO Diagnostics for brma Objects}
\usage{
\method{check_loo}{brma}(object, unit = "estimate", ...)
}
\arguments{
\item{object}{a brma model object.}

\item{unit}{output/deletion unit. See \code{\link{add_loo}}.}

\item{...}{currently unused.}
}
\value{
NULL (throws warning if diagnostics are unreliable).
}
\description{
Check Pareto k diagnostics for a brma model object and
warn if any values are high.
}
\details{
LOO must first be computed with
\code{object <- add_loo(object, unit = unit)}. The method warns when any
Pareto \eqn{k} diagnostic is greater than 0.7.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/vif.R
\name{vif.brma}
\alias{vif.brma}
\title{Variance Inflation Factors for brma Objects}
\usage{
\method{vif}{brma}(object, posterior_correlation = TRUE, ...)
}
\arguments{
\item{object}{a fitted brma object with moderators}

\item{posterior_correlation}{logical; whether to also compute and return
the posterior correlation matrix of regression coefficients. Defaults
to \code{TRUE}.}

\item{...}{additional arguments (currently ignored)}
}
\value{
An object of class \code{vif.brma} containing:
\item{vif}{A data frame with columns \code{term}, \code{df}, \code{GVIF},
and \code{GVIF^(1/(2*df))} (= \eqn{GVIF^{1/(2 \cdot df)}}). For single-df
terms, GVIF equals the standard VIF.}
\item{posterior_correlation}{A correlation matrix of posterior regression
coefficient samples when requested and available; otherwise \code{NULL}.}
}
\description{
Computes variance inflation factors (VIF) and generalized
VIF (GVIF) for a fitted brma meta-regression model. Also optionally
returns the posterior correlation matrix of regression coefficients.
}
\details{
VIF is computed from the correlation matrix derived from the
coefficient variance-covariance matrix \eqn{(X'WX)^{-1}}. For standard
meta-regression models, \eqn{W = \mathrm{diag}(w_i/(v_i + \hat\tau^2))}
with \eqn{\hat\tau} equal to the posterior mean heterogeneity. For scale
and multilevel models, the coefficient covariance is averaged across
posterior heterogeneity draws, using observation-specific \eqn{\tau_i}
and block-structured multilevel covariance where applicable.

A VIF of 1 indicates no collinearity; values above 5 or 10 are
commonly considered problematic.

For multi-column terms, such as factor contrasts,
the Generalized VIF (GVIF) of \insertCite{fox1992generalized;textual}{RoBMA}
is reported. GVIF captures the joint inflation for all coefficients
belonging to the same term. To enable comparison across terms with different
degrees of freedom, \eqn{GVIF^{1/(2 \cdot df)}} is also reported; this
value can be compared against the usual VIF thresholds (after squaring).

When \code{posterior_correlation = TRUE}, the function also returns the
posterior correlation matrix of the regression coefficients. This
Bayesian diagnostic complements VIF: while VIF diagnoses the
\emph{potential} for collinearity problems (a data property), the
posterior correlation shows the \emph{realized} identification
given the data and priors. Informative priors can mitigate
collinearity, reducing posterior correlations even when VIF is high.
}
\examples{
\dontrun{
if (requireNamespace("metadat", quietly = TRUE) &&
    requireNamespace("metafor", quietly = TRUE)) {
  data(dat.bcg, package = "metadat")
  dat <- metafor::escalc(
    measure = "RR",
    ai      = tpos,
    bi      = tneg,
    ci      = cpos,
    di      = cneg,
    data    = dat.bcg
  )
  fit <- brma(
    yi      = yi,
    vi      = vi,
    mods    = ~ ablat + year,
    data    = dat,
    measure = "RR",
    seed    = 1,
    silent  = TRUE
  )

  vif(fit)
}
}

}
\references{
\insertAllCited{}
}
\seealso{
\code{\link[=regplot]{regplot()}}, \code{\link[=summary.brma]{summary.brma()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/marginal_means.R
\name{summary.marginal_means.brma}
\alias{summary.marginal_means.brma}
\title{Summarize Estimated Marginal Means}
\usage{
\method{summary}{marginal_means.brma}(
  object,
  type = NULL,
  probs = c(0.025, 0.5, 0.975),
  logBF = FALSE,
  BF01 = FALSE,
  bf = NULL,
  output_measure = NULL,
  transform = NULL,
  ...
)
}
\arguments{
\item{object}{a \code{marginal_means.brma} object.}

\item{type}{for RoBMA product-space objects, whether to summarize
model-averaged (\code{"averaged"}) or conditional (\code{"conditional"})
marginal means. Defaults to \code{"averaged"} and is available only for
RoBMA marginal means.}

\item{probs}{quantiles of the posterior distribution to be displayed.
Defaults to \code{c(.025, .50, .975)}.}

\item{logBF}{whether to show inclusion Bayes factors on the log scale.
Defaults to \code{FALSE}.}

\item{BF01}{whether to show inverse inclusion Bayes factors. Defaults to
\code{FALSE}.}

\item{bf}{whether to show inclusion Bayes factors. Defaults to the setting
stored by \code{marginal_means()}.}

\item{output_measure}{effect-size measure for location/effect predictions.
Defaults to the fitted measure. Supported conversions are among \code{"SMD"},
\code{"COR"}, \code{"ZCOR"}, and \code{"OR"}; \code{"RR"}, \code{"HR"},
\code{"IRR"}, \code{"RD"}, and \code{"GEN"} can only be returned on their
fitted measure. Use \code{transform = "EXP"} for ratio-scale output from
log-scale measures.}

\item{transform}{optional display transformation. Currently \code{"EXP"}
exponentiates log-scale measures \code{"OR"}, \code{"RR"}, \code{"HR"},
and \code{"IRR"}.}

\item{...}{additional arguments (currently ignored).}
}
\value{
A \code{BayesTools_table} of class
\code{summary.marginal_means.brma}.
}
\description{
Summarizes estimated marginal means stored in a
\code{marginal_means.brma} object using
\code{BayesTools::marginal_estimates_table()}.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/input-priors.R
\name{estimate_unit_information_sd}
\alias{estimate_unit_information_sd}
\title{Estimate Unit Information Standard Deviation}
\usage{
estimate_unit_information_sd(sei, ni)
}
\arguments{
\item{sei}{a complete numeric vector of strictly positive standard errors
for each study.}

\item{ni}{a complete numeric vector of strictly positive sample sizes for
each study, with the same length as \code{sei}.}
}
\value{
Returns a single numeric value representing the estimated unit
information standard deviation.
}
\description{
Estimates the unit information standard deviation (UISD) from
sample sizes and standard errors. The UISD is used to scale weakly informative
prior distributions for meta-analytic parameters.
}
\details{
The unit information standard deviation is computed following Equation 6 in
\insertCite{rover2021weakly;textual}{RoBMA}:
\deqn{\text{UISD} = \sqrt{\frac{\sum n_i}{\sum \text{se}_i^{-2}}}}
where \eqn{n_i} is the sample size and \eqn{\text{se}_i} is the standard error
for each study.

This function is useful when you want to compute the UISD once and pass it
to multiple analyses via the \code{prior_unit_information_sd} argument in
\code{\link[=brma]{brma()}} or related functions. This ensures consistent prior scaling
across analyses performed on different subsets of the same data.
}
\examples{
# Example with simulated data
sei <- c(0.2, 0.3, 0.25, 0.15)
ni  <- c(50, 30, 40, 80)
estimate_unit_information_sd(sei = sei, ni = ni)

}
\references{
\insertAllCited{}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/loo.R
\name{loo_weights.brma}
\alias{loo_weights.brma}
\alias{loo_weights}
\title{Extract Normalized PSIS Weights from brma Object}
\usage{
\method{loo_weights}{brma}(object, unit = "estimate", ...)
}
\arguments{
\item{object}{a brma model object.}

\item{unit}{output/deletion unit. See \code{\link{add_loo}}.}

\item{...}{currently unused.}
}
\value{
An \code{S x K} matrix for estimate-unit LOO, or \code{S x G}
matrix for cluster-unit LOO, with posterior samples in rows and LOO targets
in columns. Columns are normalized to sum to one.
}
\description{
Extract the normalized Pareto smoothed importance sampling
(PSIS) weights from a brma model object.
}
\details{
LOO must first be computed with
\code{object <- add_loo(object, unit = unit)}. This method extracts the
stored PSIS object and does not compute LOO.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utilities.R
\name{RoBMA_options}
\alias{RoBMA_options}
\alias{RoBMA.options}
\alias{RoBMA.get_option}
\title{Options for the RoBMA package}
\usage{
RoBMA.options(...)

RoBMA.get_option(name)
}
\arguments{
\item{...}{named option(s) to change. Names must be exact public option names,
nonempty, and unique.}

\item{name}{a single non-missing character string matching one public option
exactly; for available options, see details below.}
}
\value{
\code{RoBMA.options()} invisibly returns a named list with all current
options after applying any changes. \code{RoBMA.get_option()} returns the current
value of the requested option.
}
\description{
Inspect or change package-level defaults used by RoBMA.
}
\details{
The available options are:
\describe{
\item{\code{max_cores}}{number of cores to use for parallel computing (default is one fewer than detected logical cores, with a minimum/fallback of 1)}
\item{\code{check_scaling}}{whether to check scaling of predictors (default \code{TRUE})}
\item{\code{silent}}{whether to suppress output (default \code{FALSE})}
\item{\code{autocompute.loo}}{whether to automatically compute LOO (default \code{FALSE})}
\item{\code{autocompute.waic}}{whether to automatically compute WAIC (default \code{FALSE})}
\item{\code{autocompute.marglik}}{whether to automatically compute marginal likelihood (default \code{FALSE})}
\item{\code{cluster_likelihood.n_gamma}}{number of Gauss-Hermite nodes used for cluster-unit log-likelihoods (default \code{15})}
\item{\code{default_UISD.effect}}{default scaling of the unit information standard deviation for the effect size parameter (default \code{0.5})}
\item{\code{default_UISD.heterogeneity}}{default scaling of the unit information standard deviation for the heterogeneity parameter (default \code{0.25})}
\item{\code{default_UISD.mods}}{default scaling of the unit information standard deviation for the moderators (default \code{0.25})}
\item{\code{default_UISD.scale}}{default scaling of the unit information standard deviation for the scale parameter (default \code{0.5})}
\item{\code{default_informed_priors.mods}}{default scaling of informed priors for moderators (default \code{0.5})}
\item{\code{default_informed_priors.scale}}{default scaling of informed priors for the scale parameter (default \code{0.5})}
\item{\code{default_bias_weightfunction.alpha}}{default alpha for the weightfunction (default \code{1})}
\item{\code{default_bias_PET.scale}}{default scale for the PET (default \code{1})}
\item{\code{default_bias_PEESE.scale}}{default scale for the PEESE (default \code{5})}
}

Boolean options require scalar \code{TRUE} or \code{FALSE} values.
\code{max_cores} must be integer-like and at least 1;
\code{cluster_likelihood.n_gamma} must be integer-like and at least 3.
Scale and alpha defaults must be finite positive numbers.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/input-priors.R
\name{RoBMA_prior_specification}
\alias{RoBMA_prior_specification}
\title{Prior specification for model-averaging}
\arguments{
\item{prior_effect}{prior distribution(s) for the alternative effect
component(s).}

\item{prior_heterogeneity}{prior distribution(s) for the alternative
heterogeneity component(s).}

\item{prior_mods}{prior distribution(s) for alternative moderator
components. A single prior applies to all terms; a named list can specify
term-specific components.}

\item{prior_scale}{prior distribution(s) for alternative scale-regression
components. A single prior applies to all terms; a named list can specify
term-specific components.}

\item{prior_heterogeneity_allocation}{prior distribution(s) for the
alternative cluster-level heterogeneity allocation component(s).}

\item{prior_bias}{prior distribution(s) for alternative publication-bias
component(s), such as weight functions, PET, or PEESE.

Alternative prior arguments can be:
\itemize{
\item A single prior distribution object (creates a mixture with one alternative)
\item A list of prior distributions (creates a mixture with multiple alternatives)
\item \code{NULL} or \code{FALSE} (omits the alternative hypothesis component)
}
See \code{\link{publication_bias_prior_specification}} for details on
specifying publication-bias priors and \code{\link{prior_specification}} for
details on specifying meta-analytic parameter priors.}

\item{prior_effect_null}{prior distribution(s) for the null effect
component(s).}

\item{prior_heterogeneity_null}{prior distribution(s) for the null
heterogeneity component(s).}

\item{prior_mods_null}{prior distribution(s) for null moderator components.
A single prior applies to all terms; a named list can specify term-specific
components.}

\item{prior_scale_null}{prior distribution(s) for null scale-regression
components. A single prior applies to all terms; a named list can specify
term-specific components.}

\item{prior_heterogeneity_allocation_null}{prior distribution(s) for the
null cluster-level heterogeneity allocation component(s).}

\item{prior_bias_null}{prior distribution(s) for null publication-bias
component(s), usually \code{prior_none()}. See
\code{\link{publication_bias_prior_specification}}.

Null prior arguments can be:
\itemize{
\item A single prior distribution object (creates a mixture with one null)
\item A list of prior distributions (creates a mixture with multiple nulls)
\item \code{NULL} or \code{FALSE} (omits the null hypothesis component)
}
Defaults to a point mass (spike) at zero for effect and heterogeneity parameters.}

\item{model_type}{character string specifying predefined publication-bias model
ensembles. One of:
\itemize{
\item \code{"PSMA"} (default): Full RoBMA-PSMA ensemble with 6 weight functions + PET + PEESE
\item \code{"6w"}: Six weight function models
\item \code{"2w"}: Two weight function models
\item \code{"PP"}: PET-PEESE models only
}
Custom \code{prior_bias} replaces the preset alternative bias components. If
\code{prior_bias} is omitted, \code{model_type} determines the default alternatives even
when \code{prior_bias_null} is customized or omitted.}
}
\description{
The \code{\link{RoBMA}} function extends the prior specification described in
\code{\link{prior_specification}} by allowing mixture priors that define competing
hypotheses for Bayesian model-averaging. This enables multimodel inference via the
product space method \insertCite{carlin1995bayesian,lodewyckx2011tutorial}{RoBMA},
where a single MCMC chain explores a joint model space containing all competing models.
}
\details{
\subsection{The product space method for model-averaging}{

RoBMA implements Bayesian model-averaging using the product space method
\insertCite{carlin1995bayesian,lodewyckx2011tutorial}{RoBMA}. Rather than fitting
each model separately and combining results after the fitting is complete,
this approach embeds all competing models within a single joint model space.
A discrete model indicator variable selects which model component is "active"
at each MCMC iteration, and the posterior probability of each model is estimated
from the proportion of iterations spent in that model's subspace.

This is achieved by specifying \strong{mixture priors} that combine:
\itemize{
\item \strong{Null hypothesis component(s)}: Typically point masses (spikes) representing
the absence of an effect, heterogeneity, or publication bias
\item \strong{Alternative hypothesis component(s)}: Continuous distributions representing
the presence of the parameter
}

The mixture structure enables direct computation of:
\itemize{
\item \strong{Inclusion Bayes factors}: Evidence for the presence vs. absence of each
parameter (e.g., effect, heterogeneity, bias)
\item \strong{Model-averaged estimates}: Posterior estimates that account for model
uncertainty by weighting across all models
\item \strong{Conditional estimates}: Posterior estimates given that a particular
hypothesis is true
}
}

\subsection{Default mixture prior structure}{

By default, \code{\link{RoBMA}} creates the following mixture priors:

\tabular{lll}{
\strong{Parameter}      \tab \strong{Null component}         \tab \strong{Alternative component}   \cr
Effect (\eqn{\mu})        \tab Spike(0)                   \tab Normal(0, UISD-scaled)      \cr
Heterogeneity (\eqn{\tau}) \tab Spike(0)                   \tab Normal+(0, UISD-scaled)     \cr
Publication bias   \tab No bias (prior_none)       \tab Weight functions + PET + PEESE \cr
Allocation (\eqn{\rho})   \tab (none by default)          \tab Beta(1, 1)
}

See \code{\link{prior_specification}} for details on how the UISD scaling is determined.
}

\subsection{Specifying custom mixture priors}{
\subsection{Single prior vs. list of priors}{

Each \verb{prior_*} and \verb{prior_*_null} argument accepts either a single prior or a list:
\itemize{
\item \strong{Single prior}: Creates one component in the mixture
\item \strong{List of priors}: Creates multiple components, enabling model-averaging
across different prior specifications (e.g., different effect size scales)
}

When multiple priors are specified in a list, they are all treated as alternative
(or null) hypothesis components for the main inclusion Bayes factor. The inclusion
Bayes factor tests the combined evidence for all alternative components against all
null components. The detailed summary output shows posterior probabilities and
inclusion Bayes factors for each individual component separately, allowing finer-grained
inference about which specific prior specification is most supported by the data.
}

\subsection{Omitting hypothesis components}{

For top-level mixture components, setting a prior argument to \code{NULL} or \code{FALSE}
removes that hypothesis component:
\itemize{
\item \code{prior_effect_null = NULL}: No null hypothesis for effect (assumes effect exists)
\item \code{prior_effect = NULL}: No alternative hypothesis (assumes no effect)
\item \code{prior_bias_null = NULL} or \code{FALSE}: No "no bias" model (assumes some bias mechanism)
\item \code{prior_bias = NULL} or \code{FALSE}: No bias model (assumes no bias mechanism)
}
For moderator and scale regression terms, an omitted or whole-argument \code{NULL}
\code{prior_mods} / \code{prior_scale} means "use defaults". To omit a component for a
specific term, use a named list entry such as \code{prior_mods_null = list(x1 = NULL)}.
}

\subsection{Parameter-specific customization for moderators}{

For moderator priors (\code{prior_mods}, \code{prior_mods_null}), you can specify different
priors for different predictors using named lists. A single prior object applies
to all moderator terms:
\preformatted{
RoBMA(...,
  mods = ~ x1 + x2,
  prior_mods_null = list(x1 = NULL)  # No null for x1, default null for x2
)
}
This allows testing whether specific moderators have effects while assuming others
are always included or excluded.
}

}

\subsection{Mixture priors for different parameter types}{
\subsection{Effect size (\eqn{\mu}) and heterogeneity (\eqn{\tau})}{

Effect and heterogeneity priors combine spike-and-slab components:
\preformatted{
# Default: spike(0) null + normal alternative
# Custom: multiple alternatives with different scales
RoBMA(...,
  prior_effect = list(
    prior("normal", list(mean = 0, sd = 0.5)),
    prior("normal", list(mean = 0, sd = 1.0))
  )
)
}
}

\subsection{Publication bias}{

Bias priors combine different publication bias adjustment mechanisms:
\itemize{
\item No publication bias (null hypothesis)
\item Weight functions: Selection models with different step functions
\item PET/PEESE: Regression-based bias adjustment
}

The \code{model_type} argument provides convenient presets:
\preformatted{
RoBMA(..., model_type = "PSMA")  # Full ensemble (default)
RoBMA(..., model_type = "PP")    # Only PET-PEESE models
}
See \code{\link{publication_bias_prior_specification}} for the bias-prior
constructors and preset model spaces.
}

\subsection{Heterogeneity allocation (\eqn{\rho}) for multilevel models}{

When \code{cluster} is specified, \eqn{\rho} controls the cluster-level variance
allocation. The decomposition is
\eqn{\tau_{between} = \tau\sqrt{\rho}} and
\eqn{\tau_{within} = \tau\sqrt{1 - \rho}}. By default, only an alternative
(Beta(1,1)) is specified, but null hypotheses can be added:
\preformatted{
RoBMA(...,
  cluster = study,
  prior_heterogeneity_allocation_null = prior("spike", list(location = 0.5))
)
}
}

\subsection{Moderator and scale regression terms}{

When \code{mods} is specified, the effect prior becomes the intercept prior, and each
moderator receives a mixture prior. The \strong{intercept} inherits the effect mixture
structure, while \strong{regression coefficients} get separate spike-and-slab mixtures.

Similarly, when \code{scale = ~ ...} is specified for location-scale models, the
heterogeneity prior becomes the scale intercept prior, and each scale predictor
receives a mixture prior following the same pattern as moderator terms.
Note that the scale intercept (i.e., baseline heterogeneity) is always positive and
the multiplicative scale coefficients require non-zero prior distribution on the
intercept.
}

}

\subsection{Model weights and prior odds}{

Each component in a mixture prior has an associated prior weight (prior model
probability). User-specified components receive equal weights unless their
prior objects set \code{prior_weights}. Publication-bias presets use fixed weights:
\code{"2w"} and \code{"6w"} split the alternative bias mass equally across their
weight functions, \code{"PP"} splits it equally across PET and PEESE, and \code{"PSMA"}
assigns half of the alternative bias mass to the six weight functions
combined and one quarter each to PET and PEESE. Custom weights can be
specified via the \code{prior_weights} argument in individual prior objects.

The prior odds between null and alternative hypotheses affect the Bayes factor
interpretation. With equal prior weights on null and alternative, the posterior
odds equal the Bayes factor.
}
}
\examples{
\dontrun{
if (requireNamespace("metadat", quietly = TRUE)) {
  data(dat.lehmann2018, package = "metadat")

  # Default mixture priors (spike-and-slab for effect and heterogeneity)
  fit1 <- RoBMA(
    yi      = yi,
    vi      = vi,
    data    = dat.lehmann2018,
    measure = "SMD",
    seed    = 1,
    silent  = TRUE
  )

  # Custom alternative priors (two effect size scales)
  fit2 <- RoBMA(
    yi           = yi,
    vi           = vi,
    data         = dat.lehmann2018,
    measure      = "SMD",
    prior_effect = list(
      prior("normal", list(mean = 0, sd = 0.35)),
      prior("normal", list(mean = 0, sd = 0.70))
    ),
    seed         = 1,
    silent       = TRUE
  )

  # No null hypothesis for effect (assume effect exists)
  fit3 <- RoBMA(
    yi                = yi,
    vi                = vi,
    data              = dat.lehmann2018,
    measure           = "SMD",
    prior_effect_null = NULL,
    seed              = 1,
    silent            = TRUE
  )

  # Only selection model bias adjustment (no PET-PEESE)
  fit4 <- RoBMA(
    yi         = yi,
    vi         = vi,
    data       = dat.lehmann2018,
    measure    = "SMD",
    model_type = "6w",
    seed       = 1,
    silent     = TRUE
  )

  # Meta-regression with different null specifications per moderator
  fit5 <- RoBMA(
    yi              = yi,
    vi              = vi,
    mods            = ~ Preregistered,
    data            = dat.lehmann2018,
    measure         = "SMD",
    prior_mods_null = list(Preregistered = NULL),
    seed            = 1,
    silent          = TRUE
  )
}
}

}
\references{
\insertAllCited{}
}
\seealso{
\code{\link{prior_specification}} for base prior specification options,
\code{\link{publication_bias_prior_specification}} for publication-bias priors,
\code{\link{RoBMA}} for the main model-averaging function,
\code{\link[BayesTools]{prior}} for creating prior distribution objects
}
